
#include "or.h"
#include "vector.h"
#include "matrix.h"
#include "rng.h"
#include <stdexcept>
#include <cassert>
#include <cmath>
#include <algorithm>


using namespace std;


// static fractal noise function
class PerlinNoise2D
{
public:
	static const unsigned int psize = 256;

	PerlinNoise2D(unsigned int levels, RNG& rng)
	: m_levels(levels)
	, m_amplitudeDecay(0.5)
	{
		using namespace std;
		for (size_t i=0; i<psize; i++) m_value[i] = rng.gauss();
		for (size_t i=0; i<psize; i++) m_permutation[i] = i;
		for (size_t i=1; i<psize; i++) swap(m_permutation[i], m_permutation[rng.discrete(0, psize)]);
	}


	double operator () (double x, double y) const
	{
		x -= psize * std::floor(x / psize);
		y -= psize * std::floor(y / psize);
		double ret = 0.0;
		double a = 1.0;
		for (unsigned int l=0; l<m_levels; l++)
		{
			a *= m_amplitudeDecay;
			ret += a * primitive(x, y);
			x *= 2.0;
			y *= 2.0;
			if (x >= psize) x -= psize;
			if (y >= psize) y -= psize;
		}
		return ret;
	}

	double amplitideDecay() const
	{ return m_amplitudeDecay; }

	void setAmplitideDecay(double decay)
	{ m_amplitudeDecay = decay; }

private:
	double vertex(unsigned int x, unsigned int y) const
	{ return m_value[(x + m_permutation[y]) % psize]; }

	double spline(double t) const
	{ return 3.0 * t * t - 2.0 * t * t * t; }

	double interpolate(double v, double w, double t) const
	{ double s = spline(t); return (1.0 - s) * v + s * w; }

	double primitive(double x, double y) const
	{
		// lower and upper indices
		unsigned int xl = (unsigned int)std::floor(x);
		unsigned int xu = (xl + 1) % psize;
		unsigned int yl = (unsigned int)std::floor(y);
		unsigned int yu = (yl + 1) % psize;
		// remainders
		double xr = x - xl;
		double yr = y - yl;
		// grid values
		double a = vertex(xl, yl);
		double b = vertex(xu, yl);
		double c = vertex(xl, yu);
		double d = vertex(xu, yu);
		// interpolation
		return interpolate(interpolate(a, b, xr), interpolate(c, d, xr), yr);
	}

	unsigned int m_levels;
	double m_amplitudeDecay;

	unsigned int m_permutation[psize];
	double m_value[psize];
};

// This problem formulation does not have a definite source. It seems to
// appear frequently in the OR literature, in different variants. Here
// we place a fixed number of facilities on a 2D square map. The map is
// characterized by a population density and a price of land, both given
// as continuous functions based on Perlin noise.
// The extreme solutions are easily characterized: the lowest investment
// cost is realized by placing all facilities at the spot with lowest
// price of land. This solution has very bad coverage. The best coverage
// is achieved by ignoring the price of land. There should be a
// non-trivial trade-off between the two extremes, and due to the pseudo
// random functions the problem should be multi-modal.
Vector facility_placement(Vector const& x)
{
	// number of facilities
	assert((x.size() & 1) == 0);
	std::size_t n = x.size() / 2;

	// generate two pseudo-random functions:
	// pd : population density (exp(perlin noise) modulated by Gaussian)
	// ic : investment cost (exp(perlin noise))
	RNG rng(42);
	PerlinNoise2D pd_noise(16, rng);
	PerlinNoise2D ic_noise(16, rng);
	auto pd = [&pd_noise] (double x1, double x2) { return std::exp(-10.0 * ((x1-0.5)*(x1-0.5) + (x2-0.5)*(x2-0.5))) * std::exp(pd_noise(x1, x2)); };
	auto ic = [&ic_noise] (double x1, double x2) { return std::exp(ic_noise(x1, x2)); };

	// compute investment cost: sum of position costs
	double investmentcost = 0.0;
	for (std::size_t i=0; i<n; i++) investmentcost += ic(x[2*i], x[2*i+1]);

	// compute expected distance: average distance weighted by population density
	// (really dumb but stable and deterministic numerical integration)
	double average = 0.0;
	for (double x1=0.0005; x1<1.0; x1+=0.001)
	{
		for (double x2=0.0005; x2<1.0; x2+=0.001)
		{
			double bestdist2 = (x1-x[0]) * (x1-x[0]) + (x2-x[1]) * (x2-x[1]);
			for (std::size_t i=1; i<n; i++)
			{
				double dist2 = (x1-x[2*i]) * (x1-x[2*i]) + (x2-x[2*i+1]) * (x2-x[2*i+1]);
				bestdist2 = (std::min)(bestdist2, dist2);
			}
			average += pd(x1, x2) * std::sqrt(bestdist2);
		}
	}

	// return normalized result
	Vector f(2, 0.0);
	if (investmentcost < 5.0) f[0] = 0.02 * investmentcost;
	else if (investmentcost < 7.5) f[0] = 0.1 + 0.8 * (investmentcost - 5.0) / 2.5;
	else f[0] = 0.9 + 0.1 * std::tanh(investmentcost - 7.5);
	double tmp = std::tanh(4e-5 * average);
	if (tmp < 0.9) f[1] = tmp / 9.0;
	else f[1] = 9.0 * (tmp - 0.9) + 0.1;
	return f;
}
