
#include "design.h"
#include "data.h"
#include "matrix.h"
#include <stdexcept>
#include <cmath>
#include <iostream>
#include <cassert>


using namespace std;


// source:
// Reference Point Based Multi-Objective Optimization Using Evolutionary Algorithms
// http://dl.acm.org/citation.cfm?id=1144112
// http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.420.4706&rep=rep1&type=pdf
Vector welded_beam_design(Vector const& x)
{
	// problem encoding
	assert(x.size() == 4);
	double b = 0.125 + 4.875 * x[0];
	double h = 0.125 + (b - 0.125) * x[1];   // encoding of g3
	double l = 0.1 + 9.9 * x[2];
	double t = 0.1 + 9.9 * x[3];

	// auxiliary terms
	double sigma = 504000.0 / (t*t * b);
	double tau1 = 6000.0 / (std::sqrt(2.0) * h * l);
	double tau2 = 6000.0 * (14.0 + 0.5*l) * std::sqrt(0.25 * (l*l + (h+t)*(h+t))) / (1.414 * h*l * (l*l/12.0 + 0.25*(h+t)*(h+t)));
	double tau = std::sqrt(tau1*tau1 + tau2*tau2 + l*tau1*tau2 / std::sqrt(0.25 * (l*l + (h+t)*(h+t))));
	double P_c = 64746.022 * (1.0 - 0.0282346*t) * t * b*b*b;

	// fitness functions
	double f1 = 1.10471 * h*h * l + 0.04811 * t * b * (14.0 + l);
	double f2 = 2.1952 / (t*t*t * b);

	// scaling roughly to range [0, 1]
	f1 /= 40.0;
	f2 /= 0.006;

	// constraints
	double g1 = tau - 13600.0;
	double g2 = sigma - 30000.0;
	double g3 = 6000.0 - P_c;

	// penalty for constraint violations
	double penalty = ((std::max)(1e-6 * g1, 0.0) + (std::max)(1e-6 * g2, 0.0) + (std::max)(1e-7 * g3, 0.0));
	if (penalty > 0.0) penalty += 0.5;
	f1 += penalty;
	f2 += penalty;

	// non-linearly rescaled fitness
	Vector ret(2, 0.0);
	ret(0) = std::tanh(f1);
	ret(1) = std::tanh(f2);
	return ret;
}


// based on: https://arxiv.org/abs/1303.6336
Vector disk_brake_design(Vector const& x)
{
	// problem encoding
	assert(x.size() == 4);
	double r = 55.0 + 25.0 * x[0];
	double R = r + 20.0 + (90.0 - r) * x[1];   // g1 encoded
	double F = 1000.0 + 2000.0 * x[2];
	double s = 2.0 + 9.0 * x[3];               // g2 encoded

	// auxiliary terms
	double r2 = R*R - r*r;
	double r3 = R*R*R - r*r*r;

	// fitness functions
	double f1 = 4.9e-5 * r2 * (s - 1.0);
	double f2 = 9.82e6 * r2 / (F * s * r3);

	// scaling roughly to range [0, 1]
	f1 /= 2.0;
	f2 /= 20.0;

	// non-linearly rescaled fitness
	Vector ret(2, 0.0);
	ret(0) = std::tanh(f1);
	ret(1) = std::tanh(f2);
	return ret;
}

Vector optical_filter(Vector const& x)
{
	assert(x.size() == 11);
	Vector xx = Vector(11, 100.0) + 50.0 * x;

	constexpr int m = 47;

	constexpr double XAir = 1.0;
	constexpr double XSub = 4.0;
	constexpr double lower = 7.7;

	constexpr double stepsize = 100.0;
	constexpr double ConsI = XAir * XSub * 4.0;

	double Y11, Y12, Y21, Y22;

	double xMerit = 0.0;
	double maxMerit = 0.0;
	for (unsigned int i=1; i<=m; i++)
	{
		const double wale = lower * 1000.0 + stepsize * (i - 1.0);
		const double delta = 2.0 * M_PI * xx[0] * 2.2 / wale;

		Y11 = Y22 = std::cos(delta);
		const double AA = std::sin(delta);

		Y12 = AA / 2.2;
		Y21 = AA * 2.2;

		for (unsigned int j=2; j<=11; j++)
		{
			const double delta = 2.0 * M_PI * xx[j-1] * ((j & 1) ? 2.2 : 4.2) / wale;

			const double X11 = std::cos(delta);
			const double X22 = X11;
			const double AA = std::sin(delta);

			const double ref = (j & 1) ? 2.2 : 4.2;
			const double X21 = AA * ref;
			const double X12 = AA / ref;

			const double Z11 =  Y11 * X11 - Y12 * X21;
			const double Z12 =  Y11 * X12 + Y12 * X22;
			const double Z21 =  Y21 * X11 + Y22 * X21;
			const double Z22 = -Y21 * X12 + Y22 * X22;

			Y11 = Z11;
			Y12 = Z12;
			Y21 = Z21;
			Y22 = Z22;
		}

		const double QQ1 = XAir * Y11 + XSub * Y22;
		const double QQ2 = XAir * XSub * Y12 + Y21;
		const double refl = 1.0 - ConsI / (QQ1 * QQ1 + QQ2 * QQ2);

		xMerit += refl * refl;
		maxMerit = (std::max)(maxMerit, refl);
	}
	for (unsigned int i=1; i<11; i++) maxMerit += xx[i];

	Vector f(2, 0.0);
	f[0] = std::tanh(10.0 * (sqrt(xMerit / (double)m) - 0.25));
	f[1] = (maxMerit - 1000.0) / 500.0;
	return f;
}

Vector heat_exchanger(Vector const& x)
{
	assert(x.size() == 16);
	Vector indx(16, 0.0);
	for (std::size_t i=0; i<16; i++) indx[i] = 1000.0 * std::pow(4000.0, x[i]);

	constexpr int n_c = 4;
	constexpr int n_h = 4;
	constexpr int n_pos = 16;

	constexpr double T_start[8] = {470.0, 450.0, 370.0, 310.0, 200.0, 150.0, 185.0, 140.0};
	constexpr double T_ziel[8] = {320.0, 240.0, 150.0, 200.0, 420.0, 400.0, 330.0, 300.0};

	// Heat capacities
	constexpr double CPM[8] = {22400.0, 17500.0, 28500.0, 20100.0, 16800.0, 23200.0, 35100.0, 17250.0};

	double Q_zu[8];

	// Kostendaten
	constexpr double k_kw = 5.0E-5;
	constexpr double k_k = 350.0;
	constexpr double b = 8500.0;
	constexpr double k_hd = 1.303E-6;
	constexpr double k_e= 0.6;
	constexpr double a_kap = 0.1;
	double Q_kw = 0;
	double Q_hd = 0;
	double m_kw;
	double k_inv = 0.0;
	double k_bm;
	double A_kw = 0.0;
	double A_hd = 0.0;
	constexpr double A_ges_max = 250000.0;

	constexpr double T_kws = 100.0;
	constexpr double T_kwmax = 180.0;
	constexpr double T_hd = 456.0;
	constexpr double alpha_k2 = 300.0;
	constexpr double alpha_h2 = 300.0;
	constexpr double alpha_kw = 300.0;
	constexpr double alpha_hd = 600.0;
	constexpr double c_fkw = 1.014;

	double T[20][20];

	double H[20][20];
	double A[20];

	double dCrossT = 0;
	int nCrossT = 0;
	double pCrossT = 10000;

	// The streams have to be cooled (first 4) or heated (second 4) by
	// heat exchange and coolant (fuel).

	for (int i=0; i<n_pos; i++)
		for (int j=0; j<n_c+n_h; j++) 
			H[i][j] = 0.0;
	for (int i=1; i<n_pos; i++)
	{
		int hot  = (i-1) / n_c;
		int cold = (i-1) % n_c;
		H[i][hot] = -indx[i];
		H[i][n_h+cold] = indx[i];
	}

	for (int i=0; i<n_h; i++) T[0][i] = T_start[i];
	for (int i=0; i<n_c; i++) T[n_pos][n_h+i] = T_start[n_h+i];
	for (int i=0; i<n_pos; i++)
	{
		for (int j=0; j<n_h; j++)
			T[i+1][j] = T[i][j] + H[i][j] / CPM[j];
		for (int j=0; j<n_c; j++)
			T[n_pos-i-1][n_h+j] = T[n_pos-i][n_h+j] + H[n_pos-i-1][n_h+j] / CPM[n_h+j];
	}

	for (int i=0; i<n_pos; i++)
	{
		int hot  = (i-1) / n_c;
		int cold = (i-1) % n_c;
		double DT1 = T[i][hot] - T[i][n_h+cold];
		double DT2 = T[i+1][hot] - T[i+1][n_h+cold];

		// Crossing Temperatures ?
		if (i == 0) A[i] = 0;
		else if (DT1 < 0)
		{
			dCrossT -= DT1;
			nCrossT++;
		}
		else if (DT2 < 0)
		{
			dCrossT -= DT2;
			nCrossT++;
		}
		else
		{
			double TLOG = (fabs(DT1 - DT2) > 0.0002 && DT1 != 0.0)
			       ? (DT1 - DT2) / std::log(DT1 / DT2)
			       : (DT1 + DT2) / 2.0;
			double k = 1.0 / (1.0 / alpha_h2 + 1.0 / alpha_k2);
			if (TLOG != 0) A[i] = indx[i] / (k * TLOG);
			else A[i] += 1000000000.0;
		}
	}

	for (int i=0; i<n_h; i++) Q_zu[i] = (T_ziel[i] - T[n_pos][i]) * CPM[i];
	for (int i=0; i<n_c; i++) Q_zu[i+n_h] = (T_ziel[i+n_h] - T[0][i+n_h]) * CPM[i+n_h];

	for (int i=0; i<n_c+n_h; i++)
	{
		if (Q_zu[i] >= 0.0)
		{
			Q_hd += Q_zu[i];

			double DT1 = (i >= n_h) ? T_hd - T[0][i] : T_hd - T[n_pos][i];
			double DT2 = T_hd - T_ziel[i];
			double TLOG = (DT1 + DT2) / 2.0;

			if (TLOG != 0.0)
			{
				double k = (i >= n_h) ? 1.0 / (1.0/alpha_k2 + 1.0/alpha_hd)
				               : 1.0 / (1.0/alpha_h2 + 1.0/alpha_hd);
				if (TLOG != 0.0) A_hd -= (Q_zu[i]) / (k * TLOG);
				else A_hd += 10000000000.0;
			}
		}	   
		else 
		{
			Q_kw -= Q_zu[i];

			double DT1 = (i >= n_h) ? T[0][i] - T_kwmax : T[n_pos][i] - T_kwmax;
			double DT2 = T_ziel[i] - T_kws;

			if (DT1 < 0)
			{
				dCrossT -= DT1;
				nCrossT++;
			}
			else if (DT2 < 0)
			{
				dCrossT -= DT2;
				nCrossT++;
			}
			else
			{
				double TLOG = (fabs(DT1 - DT2) > 0.0000001 && DT1 != 0.0)
				       ? (DT1 - DT2) / std::log(DT1 / DT2)
				       : (DT1 + DT2) / 2.0;
				double k = (i >= n_h) ? 1.0 / (1.0/alpha_k2 + 1.0/alpha_kw)
				               : 1.0 / (1.0/alpha_h2 + 1.0/alpha_kw);
				if (TLOG > 0.0) A_kw -= (Q_zu[i]) / (k * TLOG);
				else A_kw += 1000000000.0;
			}
		}
	}
	m_kw = Q_kw / ((T_kwmax - T_kws) * c_fkw);

	for (int i=0; i<n_pos; i++) k_inv += std::pow(A[i], k_e);
	k_inv *= k_k;

	k_bm = b * (k_kw * m_kw + k_hd * Q_hd);

	if (nCrossT > 0)
	{
		// solution is infeasible, return high penalty
		double penalty = dCrossT * pCrossT + A_ges_max;
		double p = 0.99 + 0.01 * std::tanh(1e-6 * penalty);
		return Vector(2, p);
	}
	else
	{
		// encode results into the unit interval
		Vector f(2, 0.0);
		double a = std::log(a_kap * k_inv);
		f[0] = a;
		if (a < 5.0) f[0] = 0.1 * a / 5.0;
		else if (a < 10.0) f[0] = 0.1 + 0.8 * (a - 5.0) / 5.0;
		else f[0] = 0.9 + 0.09 * std::tanh(a - 10.0);
		if (k_bm < 60000.0) f[1] = 0.1 * k_bm / 60000.0;
		else if (k_bm < 200000.0) f[1] = 0.1 + 0.8 * (k_bm - 60000.0) / 140000.0;
		else f[1] = 0.9 + 0.09 * std::tanh(1e-5 * (k_bm - 200000.0));
		return f;
	}
}

// Maximize EI and PI of a GP model for the Yacht hydro-dynamics data set.
Vector hydro_dynamics(Vector const& x)
{
	assert(x.size() == 6);

	// rescale input vector to the following range:
	// min: [-5, 0.53, 4.34, 2.81, 2.73, 0.125]
	// max: [0, 0.6, 5.14, 5.35, 3.64, 0.45]
	Vector xx(6, 0.0);
	xx(0) = 5.0 * x(0) - 5.0;
	xx(1) = 0.07 * x(1) + 0.53;
	xx(2) = 0.8 * x(2) + 4.34,
	xx(3) = 2.54 * x(3) + 2.81;
	xx(4) = 0.91 * x(4) + 2.73;
	xx(5) = 0.325 * x(5) + 0.125;

	// hard-coded Gaussian process data
	static const Vector z { 0.23919053834928311053, -0.22045703556262313816, -1.5207881523512525312, -0.19949942037532761296, -0.54535080350173648078, 2.2081991246514123262, 1.3201945837712945053, 4.8292255197438818115, 2.3200168606632338708, -8.1942201176836313437, -2.2340299953701694413, -7.0251034308525124317, 6.3523704373099221243, 2.8248479698165835927, 0.1254172014467515861, -0.0089684947837797082681, -0.50127672269333944932, -0.13121492139057880366, 0.47220538823534119466, 0.21887154140513359613, -0.055077433376872930726, -0.73540933827008658774, 2.0672504457911289322, 2.1132347940722397972, -4.3631854551044950696, -7.5092851034216350214, 12.271660278307018999, -3.946452940677129817, -0.3720456352101630948, 0.33214046184839551845, 0.41041936645613219525, -0.10576610421950882746, -0.47329939649917079914, -0.75403148119680696304, 0.059980467692251737688, 3.7115190969955476064, -3.5469559807788328953, 0.063590347274323111293, 0.35139068953285956809, 5.8147325694700882082, -11.929029943731574548, 6.5027019873676170647, 0.6915020194788749075, -0.84081314639976567626, 1.1586988159613964555, -1.0074681516564520933, -1.1643739501379453127, -1.2655041285653032546, 1.717084283017938251, 5.9808904539788763088, -5.3861106047602129721, 3.4863324613686863529, -2.171114440061803208, -2.9848149759357740507, -7.2510631887444763066, 9.8267098901600942185, 0.39515290677061054581, -1.0696871753443633057, 0.46350459635675828052, -0.87786507422530102307, 0.55290285016393270201, 1.4510913820174238253, 1.8896588722116769432, 1.9875677429242046834, -8.772121286251159944, 0.96287738381381871022, -3.8484295933594907524, 5.3787155876228851525, 11.801495959219622023, -10.425918211009076231, -0.40950731314765337343, 0.80711008790790361722, 0.074722666322390010607, -0.23459498349711155329, -0.53401202839605255157, -0.58813880220047920488, -0.9532054076475178217, -0.59363150828450694796, 4.5715166124679083026, 3.8102636454682947686, -2.0082128662283551535, -3.0027335324894037782, -12.551286087344536213, 11.802751603417934945, 0.082316088518684715769, -0.11976759372444603646, -0.51833078628481776828, 0.33592569183948328693, 0.72527532619960388693, 0.73901998945669644669, 1.094892477147426435, -6.7530287729355507054, 3.467908660352139627, 0.42001025533743341356, 1.7275597519363179444, -6.321929827283900849, 13.524858314216814392, -8.6184247164059790691, -0.46086405696873472548, -0.58756788610705890097, 0.84643162149131012661, 2.9377343131493645068, 1.0778355786901752111, -0.93747984658827321347, -4.5545646979341709581, -2.0376532770575837006, -6.0363612590381103828, -2.95069708267076658, 11.65258639943200869, 12.018654785156916986, 13.142419776004825138, -25.107059854918311714, -0.28206349245162254391, 0.45744048162014427916, 0.38327831226890179961, 1.2082226318759006212, -0.45854002758090905978, -2.7904521069765357844, -2.609950612293155725, -0.24233059779850973925, 8.1993016457161154165, 0.20475376655145227645, -3.1038891872657785598, -5.6629206516978438657, 10.062312749918030264, -5.5834422948015767219, 0.76510795463919900605, -0.35834588685051371737, 0.11095547165748934071, -1.0092259974841155667, -0.60545254726569364134, -0.14679246261663411732, 1.0580178192707883067, 1.8647736654978575022, 4.3809479558393391585, -4.4351776863399035022, -2.1478606227206604729, -12.599187775692582747, 15.854097379210239538, -2.2728528515908799079, 0.52119210771883583977, 0.055076760408938760238, 0.36216418546973416781, -0.086095531624775012425, -1.3142417030994548721, -0.39052505711525542509, -0.80935946521236534767, 3.1978331765784275831, -2.7816617293925145482, -1.1436186822986882383, 4.4322374130203705533, -9.0516426096275317548, 16.664388499529650289, -9.5886705662587878862, -0.5620552695068798954, 0.38905827686030397539, 0.068381310662816358059, 0.41070178046134547811, 0.15067456830581249516, 0.53139173467130718098, -0.29788037552477208969, -3.4208023171187447176, -0.47006775248840726444, 1.4200153820273233585, 5.7127746701574011468, 2.9532448157372552799, -9.904162951706588558, 2.6647758509342800437, -0.50527304279186346925, 0.41072561866772355765, -0.30437892346768902208, 0.70671199365370596102, 0.82288102387303696528, -0.51029864491782672697, -2.1349098839668059213, 1.0659807987553471254, -0.27529078857252620161, 0.033277223793895582826, -0.29821085918080614263, 3.6287656253854301269, -0.26073033079696816827, -2.7040243784024244889, -0.18156448904403429956, 0.59379342890633945995, -0.53334378337182142271, -0.33243165879193864454, -0.1471811784701370085, 0.51904981901616165096, 1.8588699047103927064, 0.048514052590033274193, -3.2001433834567372116, -0.58837702971480609904, 0.91081277632526824739, 0.90782027781189000848, 1.4459197575343081255, -1.3467920149854979428, 0.90003811261643862363, -0.56463895569995059542, -0.32634822088470005763, -0.23854855659180845828, 0.25400187150824893312, 0.38589655659888538741, 0.078471115400766289216, -1.3387839907328589195, -0.21899129082807233804, -1.4819744151679268285, 3.453414656891677037, 10.530194982130044679, -20.428561369292989269, 9.2426733545320765018, -0.21704017316092333312, 0.41191040266207146203, -0.38711250847618788873, -0.089050951800387023738, 0.58644182138588885067, -0.88832326986057219553, 2.1210527076675509406, 0.55360004811498064736, -3.9570184571547697416, -1.407616809712097794, 2.6781759007437844211, 1.5651297329720890161, 1.2812262761306487668, -2.315227459590637249, -0.3092622321107458272, 0.58465869568293982184, 0.033833457733297045422, -0.7578004382723328991, -0.32539244273064604762, 1.0789392448505084054, 0.90173432253069707443, 0.41302764398037550286, -4.9048334502648227584, 0.08062723992592424338, 5.3735862990874281309, 5.3751136389819595962, -14.215769321743014331, 6.7481326902298199855, 0.079020758670090063269, 0.050653157513672074364, -0.43759136886675276124, -1.2657320695410831934, 0.48816200939127751823, 2.809077451583989582, 0.2766320390009774921, -3.1675229741029964359, 0.69886582950046527429, 4.4823250367411979767, -2.9934459749592505773, 4.2598468595006657722, -19.938995614594965389, 15.001063064891532406, -0.39241837760925513123, -0.19332833767879942233, 0.35307858267492431281, 0.22166488735001560473, 0.49755202858420222167, 0.54372204361178777532, -1.5692136600692216941, 0.17532671211059480676, -1.7486548044995489093, 3.4329026729970171239, -3.8383115998279722092, 7.0068721237062128182, -6.1640006553707209491, 1.53403441417884423, -0.57809006754085279223, 0.17453472323371488351, 0.22935120377174017525, 0.13357765174764649196, -0.43631439158809837142, 1.0247203293642894728, -0.8488485881208451822, 0.52404295672531442385, 1.3737393210837425261, -1.0774979706812213287, -1.508424829202436257, 4.3676819695405608712, -9.6604177277759131215, 6.2750855566574799127, 0.0091964526672445034983, 0.34411144974319335388, -0.2297651914087656122, -0.3961092465816800634, -0.34527619228362155246, -0.8026990230436608087, 1.1858786744441789374, -0.52515927900866898881, 4.7187904199236516689, 3.3996514770089416579, -11.844735803828104537, -0.5916266053525180979, 2.5432372371180038684, 2.6683134871382359066, -0.25800743960827526857, 0.47081227592442609353, -0.27039730325600441763, 0.84536559222794149804, 0.34169531199745695194, -1.1583202000195291781, -2.363620309038276357, -1.5574729708942305351, 4.1070557867555175235, 5.5354883614499694744, -5.3008008656046641249, 2.3584091860658737971, -7.3342760268634492604, 4.5145719495193361581 };
	static const Vector CF { -0.0060626065713787480327, 0.0043353676481243472868, 0.0029801157642482993884, -0.00071705498143898212249, -0.00084271365057759367589, -0.00050010176653607352204, -0.00012618492525161284501, -0.00012618492474774701574, -0.00050010176641940295994, -0.00084271365047117186009, -0.00071705498163181224935, 0.0029801157637713848736, 0.0043353676475354763209, -0.0060626065704686149527, 0.0022693616874569361164, -0.0023241592336158416361, 0.00025198229251240013976, -1.4871676462037061728e-05, 0.0001236444807839569282, 3.8181749301598177926e-05, 9.1335238844319002749e-05, 9.1335239176777022196e-05, 3.8181749055683777971e-05, 0.00012364448074041016479, -1.4871676051936288943e-05, 0.00025198229207671132834, -0.0023241592335305760741, 0.0022693616875025524049, 0.0029761305676032868625, -0.0039526504253289719559, 0.00063815433270728234069, 0.00064267880116664847373, 0.00021117088190536221354, -0.00012964177145968237836, -3.3290152778831308411e-05, -3.329015285503250654e-05, -0.00012964177154733969033, 0.00021117088218690349688, 0.00064267880083190842555, 0.00063815433300620783008, -0.0039526504255478212002, 0.0029761305676977434231, -0.0031483872934571473581, 0.00022424710862617956708, 0.0019563131719278099852, 0.00024326645869947663425, -0.00015810371002736821333, -0.00030962311410304838022, -0.00014641628223024949484, -0.0001464162822556874799, -0.00030962311341145923316, -0.00015810370988978555973, 0.00024326645888916864635, 0.0019563131707792669189, 0.00022424710866050800995, -0.0031483872929922995798, 0.00070082743334199445323, -0.00092965563586396710027, 0.00080315806984952026193, -0.00038164979750097667924, -0.00014492081522336081045, -8.6768358534272671534e-06, 0.00015395437367515807381, 0.00015395437360750385825, -8.6768356736266483153e-06, -0.00014492081511163767971, -0.00038164979761812212189, 0.00080315806961039903911, -0.00092965563584179733425, 0.00070082743347832376912, 0.0014478451723907834105, -0.0010467056293338395864, 0.00026682160339094808561, -0.00054356806329620456497, -6.9179117550945481985e-05, 0.00010229938994452472678, 0.00025435761762328430724, 0.00025435761748637819579, 0.00010229938995130749557, -6.9179117602215234317e-05, -0.00054356806341659957837, 0.00026682160371999108506, -0.0010467056294635179725, 0.001447845172366069022, 0.0016532399283710314496, -0.001447231156139498244, 6.3800198888407837552e-05, -0.00026517208823195054235, 0.00015807269345154140172, 8.4140652569206914047e-05, 0.00010165761802461911745, 0.00010165761782580939876, 8.4140652570618978956e-05, 0.00015807269340389895618, -0.00026517208782589340998, 6.3800198631197568844e-05, -0.0014472311563749618024, 0.0016532399285146825996, -0.0040348310620526850467, 0.0030475122070235637073, 0.0019932820114953141155, -0.00093978925156597628576, -0.0007077812439282883672, -0.00023668732333842479321, 0.00016516896926670882051, 0.00016516896929492236312, -0.00023668732354026941062, -0.00070778124358693295104, -0.00093978925186757183985, 0.0019932820117160021667, 0.0030475122074381522097, -0.0040348310625175076716, -0.0029864528756856222221, 0.0029235378280477774052, 0.0015953721107653385906, -0.0011224559585437067152, -0.00068885032398235263029, -0.00016739707222239441631, 0.00023070889005567286889, 0.00023070889037254960208, -0.00016739707312804497086, -0.00068885032360449516275, -0.0011224559593410454728, 0.0015953721120165464953, 0.0029235378286227090006, -0.0029864528766431752693, -0.0016849653025179439289, 0.00049773554744028062485, 0.00067395273952029072229, -0.00036847884833996445708, 0.00013094371339461390036, 1.0502273945859827431e-05, -4.1111073745360438014e-05, -4.1111073929553376693e-05, 1.0502274021551016858e-05, 0.00013094371306299201763, -0.00036847884827739818547, 0.00067395273974191205468, 0.00049773554782228668358, -0.0016849653029341604693, -0.0019585174070040414773, 0.00045848770670777264091, 0.00090644057174519253317, -0.00027302450953855389448, 0.00010093826924203325168, -3.5444720231581116135e-05, -8.0805345831255759315e-05, -8.0805345723397592472e-05, -3.5444719827332216155e-05, 0.00010093826871658551081, -0.00027302450886279583386, 0.00090644057155797597147, 0.000458487706395505068, -0.0019585174068940114367, 0.0025060742087450174273, -0.0023127049646348781138, -0.00013406559702023798653, 0.00017889606542378401477, 0.00039902645674377672236, 3.3941867959366711638e-06, -9.0617617954511647804e-05, -9.0617618088529444664e-05, 3.3941864797798482201e-06, 0.00039902645639685457857, 0.00017889606544174707636, -0.00013406559593867606317, -0.0023127049653628439785, 0.0025060742088156562346, 0.0024211692307585766688, -0.0023382762982018601497, -1.4666402879783338453e-05, 0.0001693952582648167865, 0.00034095950299070513015, -5.0171267495062454866e-06, -6.2563526680017161885e-05, -6.2563527090938458874e-05, -5.017126731354099034e-06, 0.00034095950274725880802, 0.00016939525828238433114, -1.466640266959209904e-05, -0.0023382762979884397225, 0.0024211692305366183664, 0.0016364187882353092157, -0.0015727591115837248614, 0.00039020674157309170882, -3.9978648964305446567e-05, 1.9460738933736038137e-05, -7.486865779161855361e-05, 8.6101583585074248361e-05, 8.6101583990846886074e-05, -7.4868658056392867195e-05, 1.9460739199450571846e-05, -3.9978649571212332425e-05, 0.00039020674241685426864, -0.0015727591119504835682, 0.0016364187881835468019, -0.0033437684182346539413, 0.00078984980305708829551, 0.0012145881315901618965, -0.00024452256517325937635, 4.5212570113493943946e-05, -3.2565124781981646951e-05, -2.8975054389952271949e-05, -2.8975054236921038431e-05, -3.2565124686856350422e-05, 4.5212569754954357037e-05, -0.00024452256506815595039, 0.0012145881315855922011, 0.00078984980323559828008, -0.0033437684183546534378, 0.0025183823192644217565, -0.0036639544860399804807, 0.00045635897903629574299, 0.0008960249385220413719, 0.00050196939394004931427, -0.00018780813197231721774, -0.00030127020446982927249, -0.00030127020482009769786, -0.00018780813180301753063, 0.00050196939415081409663, 0.00089602493811515027122, 0.00045635897916701528843, -0.0036639544857782202492, 0.0025183823191147203246, 0.0019208613691885248562, -0.00155872810194730314, -0.00039187774314660469427, -0.00013339795582233571897, 0.00052458701588457140597, 0.00013313264788526735227, -0.00011127072516786684631, -0.00011127072542027605095, 0.00013313264788220036117, 0.00052458701627861037409, -0.00013339795583364177922, -0.00039187774347604342751, -0.0015587281021083310151, 0.0019208613694379206294, 0.0020668483443275469796, -0.0016103928741009439565, -0.00035174543518492508432, -0.0001795579090384880408, 0.00046693615324172252357, 0.00013654007016516978013, -7.1598535883420078019e-05, -7.1598535972716703668e-05, 0.00013654006997459999795, 0.00046693615350096306926, -0.0001795579092733479154, -0.00035174543538651903418, -0.0016103928736170836089, 0.002066848344103311419, 0.0039618540056300718261, -0.0017837303851494200335, -0.0012306322817074583087, -0.00051781387005577980687, 0.00054868977407384819656, 0.00029725381092855079901, 5.0883239949714442751e-05, 5.0883239905798183234e-05, 0.00029725381117662666441, 0.00054868977440300503723, -0.00051781387022598046599, -0.001230632281705931752, -0.001783730385817305919, 0.0039618540061667206764, 0.0037248804575868797034, -0.001793955894947970621, -0.0010187499446243312301, -0.00047297422403950063058, 0.0004816915331591906449, 0.00025796435729202524256, 5.4896882277430392172e-05, 5.4896882310348504852e-05, 0.0002579643568242223628, 0.00048169153334814973633, -0.00047297422317610360309, -0.0010187499456350279647, -0.0017939558948869560595, 0.0037248804578570438029, 0.0029207320240203183889, -0.0020391652860151958035, -0.0004561021829833662028, -0.00029615034263699019368, 0.00028237126205770193832, 0.00017338830667286441768, 0.00014320670847123989633, 0.00014320670831616255647, 0.00017338830735744159561, 0.00028237126127615523596, -0.00029615034235030979204, -0.00045610218329300566964, -0.0020391652852186810396, 0.0029207320235599132374, 0.0025095520991110110282, -0.0032098853936865425623, 0.00038361116767998693078, 0.00035884032972712844034, 0.00033932722562655948106, -1.2691821790297708183e-05, -9.2841852999125951129e-05, -9.2841852601159569058e-05, -1.2691821983490392256e-05, 0.0003393272254034185309, 0.00035884033001143400554, 0.00038361116779761506024, -0.0032098853937441162998, 0.0025095520990578989995 };
	static const double FCF = 0.0028148874920630615956;
	std::vector<double> l1 = { 16.867977764311795852 , -13.576253430816452195, -4.8889148365464558665, -1.0653722251614687888, 0.45501148853589035159, 0.99255217188830413555, 1.0442087813010518804, 0.80539064003565186134, 0.45990746798015075258, 0.14941470237775592023, -0.056211537438256901911, -0.1518070051571251966, -0.16408085071333927685, -0.1253265631575768424, 0.2126836963764726085, 0.18057304985660765917, -0.29901639086890835495, -0.18912136110335109329, -0.051991632625357185604, 0.039388376447544858061, 0.081782206808085802785, 0.060691897980090621367, 0.019989157458014751578, -0.0057035427365226603821, -0.011486444467258813407, -0.0084340624761361365769, -0.0045828220891256325059, -0.0020135102047151363751, 0.15663971706541507389, 0.15317021732836130776, -0.36672200360020629217, -0.093987721150498887424, 0.096365057211218227429, 0.055407223421361562199, 0.010424443073840682084, 0.0022865913236809845034, 0.0048207460512661179722, 0.0034043255957204969497, -0.0005176477499168615588, -0.0034476894478911190335, -0.0035478610458782484927, -0.0011397054975181531659, -1.8658146649820588348, 0.87438709961832650208, 1.0389353072098379549, 0.596804949954148789, -0.0079965577559053310081, -0.34083261656682328766, -0.33453926023424745972, -0.18317467592125755149, -0.037862415873331570215, 0.044421195435831013765, 0.068029108919588507365, 0.053863746650537926153, 0.021062223800870298607, -0.017954572199757122541, -0.38767806876219446099, 0.74432871124034938326, 0.032260794248241651816, -0.22348521741482013647, -0.20144930744112038434, -0.10637003227635673241, 0.0098881820926181116654, 0.067357709840819360303, 0.06332672230424729598, 0.034748224583312552205, 0.010009561078492460948, -0.0044478295174088565769, -0.01145663654614789001, -0.014243496357190108675, -0.68079024646332841897, 0.86923964431831501898, 0.21994393341528858943, -0.20581695435719019049, -0.24801875468811235281, -0.10297593472214827681, 0.037570815939602375533, 0.075781101057938560217, 0.046496016515390564683, 0.0075214409845570668472, -0.011634147369076109149, -0.012251140781985020781, -0.0051694857204316073049, 0.0026314696990080892539, -0.87058446096756414434, 0.84000013931714045956, 0.24894953278760234561, -0.080025248554313740956, -0.10301149462786177158, -0.049349762055813227446, -0.0040179101948304898201, 0.00021690511501361661598, -0.010239735777582146051, -0.013119365219300500522, -0.0057409309080769825237, 0.0030573349136772463679, 0.0071802835881548286376, 0.0061413166005967112157, -4.4867845718083190221, 2.1076300290830074502, 2.5759202967278129748, 1.102369092374423154, -0.15442900660381561617, -0.62135409331895741225, -0.54747608970612260393, -0.31375859867169275663, -0.11069916973765903112, 0.011313554687425671913, 0.063857663803670902669, 0.069948136279567349471, 0.049269629108337889556, 0.015635132444937492602, -3.604571933793747629, 1.9499300181139289645, 2.0520461367921698503, 0.72872884703161122655, -0.26213680409046824504, -0.55455426357533788018, -0.41895274627081385832, -0.1995516252972911464, -0.041240004890629587686, 0.037326360781753761819, 0.060411722275989690467, 0.052014856634169985561, 0.02848431247706631475, -0.00049974936167896711909, -3.3935844854596259879, 1.2632233885378176819, 1.5369934518013963487, 0.92132922437365849522, 0.35802612882544343487, -0.047445519572979676093, -0.27173464982112699451, -0.33152756052362603034, -0.26975363336807167691, -0.15734323073605385734, -0.048873670552437298298, 0.029698285687058775806, 0.074803879781278712069, 0.093443155070854266864, -3.0526377728215683405, 1.16943825707557858, 1.4078261365687423368, 0.83670076224619083227, 0.28892633558759445167, -0.085878047331622672789, -0.26801551111013960371, -0.29834741838498357591, -0.2279794272130359345, -0.12297682305871465713, -0.029641171643401367786, 0.032827290609430417878, 0.064526370944505584415, 0.073403237725143108405, 0.6926662385801017896, 0.056855794828087974357, -0.6973928210131689065, -0.35237284047706812773, 0.042696612557876767458, 0.14455098138520366957, 0.10511949720930999219, 0.054629626780123521912, 0.029109590352040998207, 0.013375574141502626774, -0.0010195385382906580959, -0.011648511664266834528, -0.014773897364190043818, -0.010857931299941942296, 0.67767493027480241174, 0.002566590796462872321, -0.65284687878851133114, -0.29771642706599688033, 0.041642010338120108437, 0.12218323662457901413, 0.09397740183198857733, 0.054632998164588782164, 0.030704104629373469432, 0.01282059184739163353, -0.002367108757700820635, -0.012023138217201245759, -0.013926748526840960693, -0.0094028335823641059216, 0.021809886816080457234, 0.04488581823875473914, -0.21686045867355493999, 0.04426937978143069552, 0.098923526084616630882, 0.031405979790397005325, 0.0028875806267716786629, -0.0003971908026191987612, -0.0048309317207471729957, -0.013411697349524203185, -0.01633079316339837439, -0.010679186577813498671, 0.00031150143234857826909, 0.012756049717213068656, -3.1291375660093181388, 1.097434879397417351, 1.4619154617021092868, 0.90360785279001420989, 0.30004248271502265144, -0.077273419162617901224, -0.23775912490768061347, -0.27517309959613195369, -0.23549572235578150603, -0.15399984810962880077, -0.062092024864474473778, 0.015546389059030572832, 0.067989122265265669598, 0.096545479064203354458, 0.32178076893281792881, -0.16928211365295245394, -0.47160030999510066385, 0.060407977660486757154, 0.28299363826657880638, 0.12721562986157594621, -0.045902443495571279564, -0.089028131186476622272, -0.044399108029295315803, 0.0032357722569425436711, 0.020260033468394497008, 0.014664858244314451716, 0.0029075965962648965372, -0.0069136178692926720732, -0.57582713554250586352, 1.0409125980017568391, 0.041517086027672657489, -0.36906164147369086015, -0.2098626753795043709, -0.04650979201260177498, 0.013119088831791697256, 0.030353678539686318461, 0.043532994222454990396, 0.043644163858040020687, 0.027980552715584507129, 0.0064242760379006492305, -0.011354837505620898885, -0.022577471593103751873, -0.3766076771572278159, 0.8738020031666086096, -0.07836612847274444793, -0.37032573378766542316, -0.16776826470533673374, -0.0019161528128413156998, 0.042630588160259851038, 0.038448417345034811843, 0.035472517479985185562, 0.029801790638529664784, 0.01651966820048669099, 0.00096906682626786162099, -0.010428101402965100908, -0.016193747072193530323, 1.5096649776763977346, 0.47823369964294148993, -1.1206503848838571535, -1.1099952520752587315, -0.41722447564207593151, 0.10422528243309170382, 0.28813942359710736918, 0.26393144909108978702, 0.17834289043347900661, 0.093447167431278094218, 0.026303913527260604915, -0.02071655172691293928, -0.049319887895161500235, -0.063577846329291282101, 1.3097911104087578416, 0.37795345264574037314, -0.99836022300489146719, -0.92913535910327582634, -0.32710542231090916632, 0.10478782097303213083, 0.24958334197957768308, 0.22103695324168429193, 0.14544168606833374202, 0.07372826840157242767, 0.018464048817720058598, -0.019245428714923649982, -0.041118067455046251879, -0.05081813336145637322, 0.26721821364159392687, 0.28715200083332487191, -0.52629640156906309922, -0.37827702805594848323, 0.01896573326878443766, 0.20673346358571803583, 0.18715382091943080778, 0.079183020242705931535, -0.013423427065554469767, -0.055375944347787427191, -0.053208464338350386358, -0.028767447604999735883, 0.00027322204129049667762, 0.025807828008882930521, -0.15655122731470069719, 0.18325389683269888152, -0.27610330879411604244, 0.020950537766697002739, 0.22495088628435738753, 0.12924523687892114365, -0.0015533769353869044099, -0.062619936811786250375, -0.063084522527178665818, -0.038532875784898723559, -0.01261644677619117548, 0.0046744753228353704, 0.012899427221644318986, 0.014842931139313023103, -13.576253430904051456, 29.510511112531276012, -10.789071434173068909, -4.9059961920730090768, -1.6084627552791215255, -0.16721033388491238214, 0.35764969013866249048, 0.53972055541048069571, 0.51714550240047107987, 0.37592663599314557477, 0.1930998612100928713, 0.029815626377583039075, -0.088913205574174938861, -0.16408085290291868352, 0.18057305240384888267, -0.46144923167629636751, 0.52386947916627824018, -0.09673336226056086784, -0.11782308112113390464, -0.057618180242225182364, -0.037574211062654808146, 0.015931200591563440122, 0.040866717588472586953, 0.030314356276209916685, 0.0078412514465487706539, -0.0050919891811024610428, -0.0074172965671542006455, -0.0045828254136551322892, 0.15317022248833900289, -0.55922556363958064729, 0.75244751082040983725, -0.17245059311697058524, -0.26817398518551283093, 0.015793398250742010647, 0.059587106854910962261, 0.022501378801314163292, 0.0017336738183063153153, 0.00061699599363320865517, 0.0018883668427413884388, 0.0015740831897435010687, -0.00058852698015925933479, -0.0035478667919727931157, 0.87438709939634529089, -1.6669203156642713459, 0.092086977525353649887, 0.29104343304267027426, 0.44329931318575066523, 0.27944193752775992046, -0.021180514534830240103, -0.15755289428513036265, -0.14533895084057302238, -0.07375922040209861863, -0.012047350957315856868, 0.021254050243792125402, 0.029357210100030083078, 0.021062217628399250435, 0.74432871121083388211, -1.4238841880449102373, 0.61708914077255982988, 0.24233728117421601822, -0.034649476989061386845, -0.059775185807056811782, -0.090141778266317321155, -0.049020495574196423783, 0.0081707306609264274827, 0.034754840931725240749, 0.028712087380018430638, 0.012335714390681583666, -0.0018591318877483413961, -0.01145663389404719297, 0.86923964446403378936, -1.6413901487102366961, 0.46615066886716610695, 0.41191856057118436674, 0.10660674433810291006, -0.082745329625322841527, -0.13192504192454690415, -0.054553366245997249318, 0.015447971182546472307, 0.036209676468889458179, 0.02168113150200286221, 0.0023811748065488734863, -0.006793941399568492398, -0.005169490414394193982, 0.84000013888963875885, -1.7375422793319839077, 0.56040337694500774735, 0.40938774790292875583, 0.065262230434569939996, -0.043302322353860958626, -0.06782321780038215453, -0.029305389619980425769, -0.0020683004483348960308, 0.0022110977602082223029, -0.0040612210024811136802, -0.0057034975036193398881, -0.00085988761141413756935, 0.0071802842902385288423, 2.1076300302234391992, -3.9378338541302682074, -0.3654654522525401239, 1.2815875035369881374, 1.1798639536308370701, 0.43327693718015186652, -0.14302042609114290483, -0.30936478977157833015, -0.2529578254809561888, -0.13699152779600948238, -0.042801174042479468718, 0.014607637311575783975, 0.041923409321224042179, 0.049269628828303949075, 1.9499300190737729466, -3.5670407714825311807, -0.1446100658079480239, 1.1319478799973414773, 0.93492876472960040335, 0.28985045637932138618, -0.17250326247216929421, -0.26795039458062180904, -0.18927775402989344133, -0.084579614870801622661, -0.013711065987097835472, 0.021755706796415610199, 0.03256086924967521723, 0.028484315524048873375, 1.2632233875315934668, -3.0416613781319661314, 0.20303099398995028135, 0.90548916684695379686, 0.62883647043156432055, 0.31688714171790111074, 0.076368700213666235177, -0.071830368890107743063, -0.15465183515429137273, -0.16214620607481405878, -0.11740146422775829538, -0.0506282058624506065, 0.016414278295975030103, 0.074803882889507816722, 1.1694382575731698815, -2.7496011874350165805, 0.18724560223386277347, 0.78666202552096120559, 0.58283688030628844068, 0.29478056145940678689, 0.055832605677162673585, -0.07917019265825940566, -0.14543995972535644912, -0.14365042015197176939, -0.098790122377359429473, -0.038743790829628019234, 0.017836651779961361364, 0.06452636984769416495, 0.056855793046526506562, -0.35170556892028420615, 0.86877866404125014821, -0.20008885536052636867, -0.41719689678351734585, -0.13243813746722268854, 0.049709153970520447707, 0.082922012210747755856, 0.048020249759332195305, 0.021769512826258097943, 0.010562773462985496423, 0.0038650514715417735, -0.0046935083421020349498, -0.01477389486043504882, 0.0025665879819265140054, -0.22558743584137511129, 0.7883365147693254249, -0.25089159278035638723, -0.36949799174805025848, -0.099061877869950354381, 0.043783361475230340998, 0.068782119095963850652, 0.042473963747066245666, 0.022479327847866367729, 0.011800558561227908819, 0.0037468010404151119198, -0.0050789538956262977182, -0.01392674482541741976, 0.044885819185553357813, -0.32794487508147568411, 0.52283277885929912987, -0.23730060458617618924, -0.10788271711820879295, 0.072852661576574689883, 0.043787428929019066304, 0.0027872701594623201966, -0.0082169650459688534094, -0.0026551874793181466303, -0.001698732986705799421, -0.0037840909278474667635, -0.003621985116291131547, 0.00031150381950412648095, 1.0974348787222827362, -2.6454185971968353641, 0.085274364471320779346, 0.74576389087950778478, 0.67055100764768094113, 0.34322089916777370755, 0.041987315867235711075, -0.095280888541695238358, -0.14139131896519643861, -0.135865404017591479, -0.10179885620459079576, -0.049784823201491905376, 0.0096408604141279066746, 0.067989121323621656834, -0.16928211491923450538, -0.060919139271101915845, 0.73709815672972200939, -0.3758441860093819642, -0.34780925841529369835, 0.051953348444303472375, 0.15987412141539739108, 0.073974541278686162249, -0.022416706942235666011, -0.043750717552939639987, -0.020776592999613593266, 0.0028639194421709378117, 0.0098488283009287580505, 0.0029076000739367731385, 1.040912596054401229, -2.0320068995524662725, 0.80336520828739810263, 0.54597745953599718938, -0.12806113203362903818, -0.19815683486580754069, -0.072849870599572083552, 0.005299185544830702782, 0.012494155427079347426, 0.013445197111061568673, 0.01835139640181398385, 0.017024558980214973825, 0.0056603894424085759648, -0.011354834104851956392, 0.87380200259754003955, -1.7555349640769470909, 0.7993473508213257972, 0.43530767477614873417, -0.16293144724445118854, -0.19298145514418804658, -0.058319357094698702493, 0.021909125014770183121, 0.024051823737637856998, 0.016089726319278491173, 0.014388488030079167948, 0.011340225838023607757, 0.0022359143031375292715, -0.010428101232151485919, 0.47823370158428235577, -0.77123947423118888178, 1.1290453512188147744, 0.07392266594557718451, -0.61684605230492250971, -0.48761141849549527461, -0.13789906312147778844, 0.09815308909160583295, 0.1549043597568312558, 0.12653321930507993587, 0.076083288395808579274, 0.027389046205408790896, -0.014769214949686654964, -0.049319894478674165816, 0.37795345421023135435, -0.64823810092962785845, 1.0267744616845291272, 0.010302252348128451004, -0.54086862423301551583, -0.40338957632938482467, -0.10712697385359225277, 0.089732676902311708789, 0.13252287410092408737, 0.10513172636663344062, 0.061690360813387308536, 0.021295654632783957844, -0.013136683353323663423, -0.041118062970113708476, 0.28715199866461571521, -0.87589767592981238042, 0.83565123232597959824, 0.07597692382300209879, -0.2932464690779419958, -0.1694092885736814269, -0.00024212361542364532578, 0.085406641492067350563, 0.075010640986513318906, 0.030972279090060821777, -0.0074947132089020433027, -0.023359226787989077062, -0.01815489759259945185, 0.00027322441249177215582, 0.18325389486733603683, -0.78109890511142776592, 0.80007948331120593011, -0.085262309025671473295, -0.25520652677412597464, 0.035761622779735768862, 0.10470868275094666333, 0.050578015115967239757, -0.0071250539878311052247, -0.030073091697459405913, -0.028233478466013390423, -0.014648001119038881954, 0.00044143466266846793922, 0.012899427202712943422, -4.8889148377033198045, -10.78907143224234666, 31.834481171287933421, -9.9690212748118298691, -4.999245347136024975, -2.053398745650818924, -0.59221373028501289593, 0.064927002324575974757, 0.39108028029135127879, 0.47724956277013774741, 0.3966063781985359471, 0.22478402440432102005, 0.029815623338876450638, -0.15180700418385817407, -0.29901639267773716391, 0.52386948198554184053, -0.52798607833088906371, 0.57563038799498411624, -0.066143131236650415383, -0.15815321967196144248, -0.084278132267421354551, -0.046112054565587867749, 0.015226718230300818543, 0.040782755700773623553, 0.031248050130110678613, 0.0090623638660135130002, -0.0050919960038392315591, -0.0084340558742287165261, -0.36672201392480213711, 0.75244752113469115518, -0.72731819689199339596, 0.62165733459720551313, -0.1073022237619860042, -0.2510309072387530982, 0.0030214959320808768006, 0.048638988882233523403, 0.019356396877138744989, 0.0018128842477412724418, 0.0023398284809868896, 0.0035439437611284804741, 0.0015740734092570614518, -0.0034476788545507023633, 1.0389353078201535308, 0.092086976500361128473, -2.1441781188059252372, 0.023969275120785435429, 0.41061638210280182459, 0.55330793209600293014, 0.36954521092163006291, 0.030380313977231693939, -0.1448875296918923361, -0.16160502076836227858, -0.099564273861959123879, -0.030513991832846303065, 0.021254043829820774836, 0.053863753575611177105, 0.032260793785241499709, 0.61708914054937469285, -1.4280090958525002431, 0.64888014033146801385, 0.34388381427260444889, -0.020054167776084340369, -0.076708197012844384077, -0.10812554328032047968, -0.062012208349099387505, -0.0016311531039067484264, 0.029671834111326397115, 0.027674358363170385722, 0.012335719102488830121, -0.0044478359137588583186, 0.21994393434084383299, 0.46615066818612110877, -1.5589498788336131341, 0.56222049602214629882, 0.46669287193584935158, 0.089039154393871000615, -0.10350536258338809137, -0.14457847566005471718, -0.05769311823915816706, 0.016187524220552741011, 0.038321201557960392348, 0.023695504309537263055, 0.0023811687049253467408, -0.012251134742572578409, 0.24894953671878816692, 0.56040337368991066658, -1.7052561809791484393, 0.5331289199152359215, 0.42354628989220804991, 0.060959402203220013072, -0.036124630118595665862, -0.056280566080438243493, -0.019915603452612600649, 0.0007815803970879091388, 0.0011498073319435352449, -0.0058669081529053165522, -0.0057034946109127916333, 0.0030573323336077571494, 2.5759202933326950102, -0.3654654494406776033, -4.3663192326534820964, -0.41342802054762345376, 1.3725925231486857392, 1.3189075797148612779, 0.58764006677676472457, -0.03699886737750573279, -0.2653137258830930989, -0.25612753801008258936, -0.16015572383507142851, -0.062352129160213390358, 0.014607636661344288592, 0.069948136762365942243, 2.052046134442847336, -0.14461006363639508643, -3.8548802006924152153, -0.12760702853886884967, 1.2479825449409853722, 1.0521994660942377564, 0.4039425165807113971, -0.10198262635481661342, -0.24620048078537160041, -0.20092131545850513974, -0.10661936191698638543, -0.029346427801279419134, 0.021755713236989260989, 0.052014850445198705653, 1.5369934523243280466, 0.2030309929694665283, -3.2838288069577372674, -0.12174595471959243076, 0.75307272573034678143, 0.63878531585078024424, 0.42283842198217480579, 0.19098417864597125315, 0.016917841927288106846, -0.10310481519883579749, -0.1426986012491645861, -0.11620337145777204224, -0.050628206441140591321, 0.029698286695824849951, 1.4078261368499269679, 0.18724560232452863717, -3.0158415010251751198, -0.083149552097322151911, 0.67481145583159762857, 0.60386893068075631064, 0.39660170587462389991, 0.15863015345673336931, -0.0047402068630289298512, -0.10604155423831608529, -0.13181505638161383764, -0.10072323229027663194, -0.038743792586808580014, 0.032827292332973746769, -0.69739281719200396648, 0.86877866056962493868, -0.35637326665236995771, 0.77439576007372667998, -0.18824519670214651113, -0.41313085288392825678, -0.14834107143984279209, 0.021617729579866063611, 0.064327123309323344658, 0.042482585859134711082, 0.024630523417720248669, 0.014683336398970429568, 0.0038650588989225490195, -0.011648519768483883707, -0.65284687300600408033, 0.78833650892115803011, -0.28385910490581273136, 0.72277202383404914876, -0.22502133720455572141, -0.37066789600175459318, -0.11898852566972296929, 0.016499505324048119315, 0.052281276185349688512, 0.038109083068985374898, 0.02546679659214448263, 0.015739485939128487757, 0.0037468088039715494002, -0.012023146310885435631, -0.21686046339320183551, 0.52283278312699066337, -0.58515737376330567709, 0.480643201283845245, -0.18240719712483233206, -0.11486201912296638283, 0.053613209130637440314, 0.03464735752971076499, 0.0075380772666322100328, 0.00077824150915172939525, 0.0053621982708592199132, 0.0030702054730454985017, -0.0037840890605304005993, -0.010679189068018223388, 1.4619154618463230388, 0.085274362813054047083, -2.9532985009711589797, -0.12778468318253224933, 0.63930886678843068527, 0.65185369936786685674, 0.41500118423665854905, 0.1410130067914127372, -0.008994639006615512411, -0.087010463411845803683, -0.11189984506571584721, -0.096774650786597443686, -0.049784820656774045333, 0.015546386762461152087, -0.47160030761520443932, 0.7370981548433144237, -0.3440778143465193506, 0.49867339420670764305, -0.3712930629468348398, -0.29746667007700638985, 0.085871966893151255307, 0.16541511385486926122, 0.070574099711051288542, -0.026375890448393422144, -0.046793464325654676728, -0.023022183796587377541, 0.0028639249911572924219, 0.014664853118937287366, 0.041517091423281207718, 0.80336520416553891533, -1.8084504768573199662, 0.7283951384807783036, 0.55642298322753935835, -0.075483081628177559064, -0.16819965010654819304, -0.084850454845824554995, -0.018244730625219793274, -0.0049852616212370717136, 0.0048771290335092257889, 0.015497004836426613722, 0.017024568138036804549, 0.0064242664155768607753, -0.078366127163048082682, 0.7993473489513491792, -1.5533876581745040824, 0.73201458346842063207, 0.4349655414217496463, -0.13047207219377432441, -0.17392930942708056952, -0.071470240803754470083, 0.0020035289213047279, 0.011407769702801795822, 0.01152006947225378726, 0.013697187603367962788, 0.011340225861517269634, 0.00096906647151856115628, -1.1206503880647709082, 1.1290453534909672406, -0.31575457458560812762, 1.2435382739460709711, 0.12239377392886496843, -0.62906348484776819507, -0.54007921539772330011, -0.2141488453379347956, 0.034757182190806423794, 0.1183954066705372038, 0.11321970166520134438, 0.07470753495910432973, 0.027389032385292663013, -0.020716537062129461444, -0.9983602254411199084, 1.0267744649032641835, -0.30251994274629973036, 1.1184482045705363973, 0.051163274823422540127, -0.55702211018560177092, -0.44944464159656799529, -0.17141655728078622478, 0.037427710107232441128, 0.10351516143694748651, 0.095649674432773001964, 0.061569068253687214698, 0.021295657871897733654, -0.01924543137205408927, -0.52629639912269365887, 0.83565122944650604975, -0.7161897949662912044, 0.78017584451841837101, 0.021092488583885432574, -0.36075087816895595738, -0.21356736777407839267, -0.017159901476209844101, 0.094057230547987272229, 0.094255674492479790683, 0.048858316401181350885, 0.0025274061246394518079, -0.023359226583145195622, -0.028767447021833151416, -0.27610330640530733604, 0.80007948151322760832, -0.93803668642301407665, 0.57373965548737138054, -0.089943312580889930863, -0.24814520697762515455, 0.042977259433240211361, 0.11883442876919667053, 0.067314977486303800847, 0.003798270120156547764, -0.025737773635903092584, -0.028040767409609323318, -0.014648003049145832954, 0.0046744784239196113598, -1.0653722265345562104, -4.905996192722215099, -9.9690212720248911893, 32.416810472357411754, -9.8912888460478338715, -5.1487655203164495177, -2.230534299172941104, -0.71096063775492246428, 0.0080508297815306481549, 0.37774208448985358011, 0.48331849866505910418, 0.39660637485594102403, 0.19309985848549990339, -0.056211532751349949089, -0.18912136533541451522, -0.096733359502653826212, 0.57563038978224423392, -0.55000571896364802704, 0.5278354565163019263, -0.077778406393755994408, -0.15210208042217054913, -0.076109620696425170361, -0.045495146405080767738, 0.012637893810213650381, 0.039346704469609687649, 0.031248051963242796936, 0.0078412582718831776857, -0.01148645212594034265, -0.093987714761107968298, -0.17245059870715531702, 0.62165733260599342902, -0.71544369251928663367, 0.6409980929313991993, -0.096656671955117887118, -0.2571068259198242556, -0.0038898295826507276081, 0.046210828507428525214, 0.020649142824237187721, 0.003387532084514954453, 0.0023398314624845906705, 0.0018883694426340110264, -0.00051765274568084829621, 0.59680494841988018795, 0.29104343386865333398, 0.023969274768038461437, -1.9956420910875438679, 0.10593718245336392059, 0.43307783071776684292, 0.54755985647366933744, 0.3637765460821936192, 0.024049462285488796415, -0.15339691360893009664, -0.1697022857500659343, -0.09956427283633509373, -0.012047357025909158357, 0.068029115122183181352, -0.22348521751127534141, 0.24233728171213209102, 0.6488801393803123041, -1.3876400218846511869, 0.61569605999236698413, 0.3166605082604420085, -0.029900493774978824379, -0.074436678076453308139, -0.10671329725960153334, -0.06363735208375838337, -0.003616595511862767242, 0.029671831147017380115, 0.028712085315712104067, 0.010009565468387212478, -0.20581695665348967439, 0.41191856272290056085, 0.56222049612859448242, -1.5562579303956447507, 0.48820778226787042708, 0.42668902666255703604, 0.087104942228751439104, -0.087102846284196511206, -0.13303970777501455958, -0.05505738224962046512, 0.015307857241425467526, 0.038321202838118122169, 0.021681130971363100912, -0.011634147505206813081, -0.080025255225366978551, 0.40938775384598136142, 0.53312892112457810345, -1.7346654903543992088, 0.48481339959972236109, 0.4177202433882474808, 0.07575191517068226299, -0.020087466809062946849, -0.050673228855700545858, -0.021292397599701948924, -0.0015867075172597859198, 0.0011498076528570115931, -0.0040612226450208559203, -0.0057409295082125107873, 1.1023690993866663845, 1.2815874983863904113, -0.41342802320792537696, -4.1841988959326590347, -0.36789718194077303259, 1.3739630519081067472, 1.3279126314782849771, 0.61005135792008380324, -0.021608859116734872863, -0.26369714064912186391, -0.2619283404599756504, -0.1601557254924254603, -0.042801176766943817287, 0.063857667545592911029, 0.72872885472134052431, 1.1319478738686055141, -0.12760703072704424055, -3.6895396089697598541, -0.10810042838148392619, 1.2337482696460684295, 1.0517826444189686264, 0.41887923628250955943, -0.091681455760877675343, -0.24632321335735146706, -0.20615198728447073351, -0.10661936236499536257, -0.013711066426594351042, 0.060411723196751268561, 0.92132922203665668093, 0.90548916924758504621, -0.12174595409894284348, -3.3750136379604289694, -0.13434556011036247192, 0.80271127188418278653, 0.6972112991595196263, 0.47201616659106077067, 0.22334862847941894248, 0.034266125839663860719, -0.09699674765311680491, -0.14269859998232253817, -0.11740145652324734948, -0.048873678799582043197, 0.83670076170849749797, 0.78666202525721118377, -0.083149551086314443804, -3.0744219416581111126, -0.084857045768911149031, 0.72081129824283840524, 0.65277068548926164837, 0.43617331801868969299, 0.18352935345019377733, 0.0076809284827663720785, -0.10242296047286697513, -0.13181505723140782282, -0.09879012485622171158, -0.029641168841459133998, -0.35237283977407413138, -0.20008885570681225752, 0.77439576018439271188, -0.42242229679044829282, 0.77956254985846173966, -0.16291477569235479628, -0.41639785187525768517, -0.16299455909161963962, 0.011759563888864234607, 0.063514626847655922259, 0.044991839812584817626, 0.024630521500339851854, 0.010562768463216408096, -0.0010195329895615894318, -0.29771643201517139499, -0.2508915882673524167, 0.72277202521171191041, -0.32894962138046168754, 0.72652517146708261997, -0.20758471958802143797, -0.37684158909209813926, -0.13174860461850088078, 0.0089374318961042333725, 0.05212509975719883254, 0.040369329679994042648, 0.02546679414620316162, 0.011800554227462758256, -0.0023671030389243214072, 0.044269386314298463103, -0.23730060971330024788, 0.48064319940049182733, -0.56420551307600674829, 0.48535823926910459658, -0.18813140504297495847, -0.12555420820706714169, 0.054601445090185322939, 0.041791615487426776732, 0.014460672592505264406, 0.004415272128118265596, 0.0053621970887579931131, -0.0016987319606835509216, -0.016330793540136941816, 0.9036078570208191385, 0.74576388894244172167, -0.12778468584817401221, -2.9985841408224871429, -0.1637788334706303528, 0.65376057849531565846, 0.69708447098792725161, 0.46895210336531989448, 0.18104437472279027999, 0.012416653763261625495, -0.079490042415460870906, -0.11189984426866955713, -0.10179885595757302841, -0.062092025856019264496, 0.060407975522648330036, -0.37584418459300455995, 0.49867339498479451576, -0.37571244117407370799, 0.58280062328204729205, -0.30286905249437690024, -0.29700979308157371417, 0.061730826241845011826, 0.14947891092831466775, 0.068046679713592558847, -0.024705727871862360329, -0.046793465308114561008, -0.020776594272250242107, 0.020260035275604524468, -0.36906164782444272898, 0.54597746309262795705, 0.72839514159440688079, -1.8729602666954736634, 0.714856186621374734, 0.58008313129363642258, -0.070910814322808629617, -0.17840049546082972776, -0.095882441326432560347, -0.023489372886606081636, -0.0065652471688466343402, 0.0048771256710309256535, 0.018351388916310707594, 0.027980562735505615235, -0.3703257344999931755, 0.43530767602662279181, 0.73201458283608900057, -1.6287346302255163444, 0.70976536853084670753, 0.45646698724030698724, -0.12404035985628955596, -0.18128985088153551364, -0.080230246172344393818, -0.0016609673892264663342, 0.010857785649922424911, 0.011520070318907674883, 0.014388493720852209984, 0.016519662296923944694, -1.1099952514073319065, 0.073922666660087135715, 1.243538272157474367, -0.42075620802383700836, 1.1715483573990004285, 0.11784851761315005836, -0.63549469975725025872, -0.56086304975320866451, -0.23913894399055640738, 0.017770048244730305559, 0.1117219649888550187, 0.11321970541177805991, 0.07608329475704107836, 0.026303904892778076641, -0.9291353607622560018, 0.010302253077355032901, 1.1184482055963644864, -0.39584043113625067001, 1.0561306098526208608, 0.04934635442810337419, -0.56202383450577253488, -0.46680883012279911215, -0.1918987074328732112, 0.024108005171336696626, 0.098675465152679123215, 0.095649675358310592976, 0.061690362640218925072, 0.018464046367396576287, -0.3782770290628273413, 0.075976924896174760637, 0.78017584514225524206, -0.84861286208344421933, 0.69963731691184083417, 0.016767281822341675857, -0.34244430650388468784, -0.19463569543992179089, -0.0061625331212217752258, 0.099962583562180740149, 0.09778573565808529322, 0.048858317785540533729, -0.0074947114721095149312, -0.053208467403573718457, 0.02095054056681217719, -0.085262311820804917439, 0.57373965572915486266, -0.98571423072943398225, 0.59124224124575552874, -0.054497268642192023269, -0.23278320209877262359, 0.045817087279528365129, 0.11815083755108625463, 0.067686176938849831886, 0.0045392147581315408014, -0.025737771276441771817, -0.028233475261257457573, -0.012616451522128818361, 0.45501148829639770366, -1.608462754959743668, -4.9992453474402500646, -9.8912888452943441564, 32.509275725240982524, -9.8523938251403269106, -5.1426064623973015699, -2.2355516177818928192, -0.71479912563758152277, 0.0075969301593153681917, 0.37774208213204107931, 0.47724956063192708333, 0.37592663309655943715, 0.14941470781430668535, -0.051991628968610287265, -0.11782308377580152925, -0.066143132269855686722, 0.52783545569687739807, -0.57172941860575587913, 0.53325368853168020333, -0.065469572112155940968, -0.14716869573262003623, -0.076387759534467497557, -0.046834604016325519227, 0.012637894137907284761, 0.040782753536762778035, 0.030314350809386089725, -0.0057035362459340859659, 0.096365054814657291615, -0.26817398339027115339, -0.10730222207779409327, 0.64099809198948309863, -0.71660575586425345751, 0.63767749913394977668, -0.097192199519644917327, -0.2577701871991213145, -0.0033746855311493888924, 0.047188952248498269859, 0.020649142121331585942, 0.0018128839645027733807, 0.00061699804502252759367, 0.0034043243386891386057, -0.0079965537045476563216, 0.44329931076523032552, 0.41061638064644301327, 0.10593718062779487743, -1.9739137374166728645, 0.08449091805759119389, 0.40330643693596951449, 0.5242087674288054, 0.35123715073826788924, 0.019027566175642291607, -0.15339691116601494869, -0.1616050189556651473, -0.0737592119564390164, 0.044421185490626062531, -0.20144930388611045902, -0.034649479013445130815, 0.34388381292566272407, 0.61569605857076770139, -1.4348666243212240268, 0.59207845769574474204, 0.31532317862045350365, -0.02577950434895868656, -0.071576833993772209785, -0.1062099648140556879, -0.063637352002453059785, -0.00163115231151233879, 0.034754842346011671772, 0.034748222757641888947, -0.24801875370489392947, 0.10660674305235995352, 0.46669287288052591078, 0.48820778173668760713, -1.6149097837033532521, 0.46873730675812197433, 0.43631693312298042153, 0.10119170296537211007, -0.079561794936979610138, -0.13172721983948282309, -0.055057381422693646433, 0.016187523102935294506, 0.036209676992515572869, 0.007521441023820497028, -0.10301149141963698852, 0.065262226965013048208, 0.42354629066034904517, 0.48481339871918099904, -1.7574528402247753256, 0.48455388225219991893, 0.42736406553859518365, 0.081115278422061301566, -0.019931598950804974707, -0.052308910687375906323, -0.021292397227751180128, 0.00078157969569257897645, 0.0022110951602386423945, -0.013119362584041930134, -0.15442901304727782441, 1.1798639585270696006, 1.372592525010517317, -0.36789718102007101797, -4.1835208876147671475, -0.40304308747776623756, 1.3422659951871220851, 1.3116594132341146128, 0.60567723916028148867, -0.022375739105595100764, -0.26369714101203811651, -0.25612753544486227542, -0.13699152349953777619, 0.011313548697498300097, -0.26213681306793945591, 0.93492877240426253138, 1.2479825469718635045, -0.10810042803122822674, -3.7122843834231535887, -0.15024404507159136446, 1.2041067756707439784, 1.0392850296077456029, 0.41714816715856994911, -0.091314469391768793294, -0.24632321270232943045, -0.2009213157126644167, -0.084579618085304272834, 0.03732636367904716701, 0.35802613241177982939, 0.6288364675517250646, 0.75307272391898694153, -0.13434555915356474753, -3.3228625583233282192, -0.094203016774406955758, 0.82524068299477570942, 0.70696434314836686941, 0.47611616273619722373, 0.22532539542306156966, 0.03426612591263698826, -0.10310481448719874265, -0.16214620547826572028, -0.15734323204499386928, 0.28892633586515992228, 0.58283688044686832175, 0.67481145521323127046, -0.084857045680749476624, -3.0266209908382553095, -0.053401007657162338771, 0.73501754823885123713, 0.65681298318225023181, 0.43723686116840748683, 0.18418278654040357023, 0.0076809283584481288787, -0.10604155188150517219, -0.1436504145673659083, -0.12297683015849827926, 0.042696607352962004378, -0.41719689344593580138, -0.18824519470191117265, 0.77956255090822934584, -0.4137236745050577591, 0.77948114735856355306, -0.16468336572447142552, -0.41796446799460945654, -0.16225135527735082275, 0.013278360825910601531, 0.06351462565589890974, 0.042482584729901969045, 0.021769510655134920868, 0.01337557755191094884, 0.041642013869540889237, -0.36949799473315492193, -0.22502133826245510062, 0.72652517174043707016, -0.32539309343511768624, 0.7247222881723315302, -0.20852596219931437971, -0.37718785903502510815, -0.13049951242293203468, 0.010404890928040125622, 0.052125099597824367526, 0.038109085249886591462, 0.02247933173774410362, 0.012820586650352041452, 0.098923522468708974253, -0.10788271453356498841, -0.18240719578059993999, 0.48535823938177297121, -0.57430505855226765544, 0.47947979094623555119, -0.18359114330607090304, -0.11886332333600663214, 0.060118127528338997922, 0.044717173148817589146, 0.014460672220368739685, 0.00077824087035260338611, -0.0026551876143703365477, -0.013411696602728347127, 0.30004248329807692297, 0.6705510072558661383, 0.63930886593476377033, -0.16377883278341232032, -2.9781545632962695969, -0.13352321421246324262, 0.68191360065505857158, 0.7155360557664035559, 0.47874989077107071855, 0.18515923088177713152, 0.012416653775400465776, -0.08701046488864232209, -0.13586541249777148166, -0.15399983905600928002, 0.28299364058113191112, -0.34780926066104861372, -0.37129306277123058644, 0.58280062277212230093, -0.31225468632486491627, 0.59375219537941259507, -0.3176749214621666928, -0.31361664742254757865, 0.054344710366809753643, 0.14883201429203365973, 0.068046679487807362752, -0.026375888792220143764, -0.043750717268609816413, 0.0032357707373015928351, -0.20986266956863022792, -0.12806113571949617747, 0.55642298097649600663, 0.71485618519716787045, -1.8784277358971255278, 0.69772634232217589645, 0.56675565257565807542, -0.07673031521951520495, -0.17913872688553872381, -0.095656097632435113653, -0.023489371984168679053, -0.0049852578594553140917, 0.013445202553940875656, 0.043644155445622973, -0.16776826477805412141, -0.16293144658053362872, 0.43496554058812819665, 0.70976536863415851109, -1.633887569490922953, 0.69871802460018272996, 0.44875827438022097837, -0.12681445254012446711, -0.18091097271240563771, -0.079665474655794335512, -0.0016609676786556032413, 0.011407767823256623543, 0.016089717964027937247, 0.029801800121577570746, -0.41722447677860852178, -0.61684605180684792813, 0.12239377475279625274, 1.1715483572991967076, -0.46510669437134627424, 1.1529625012661641748, 0.11476496119778632765, -0.6371129641748027872, -0.56457110918816832967, -0.24303363071983008492, 0.017770048320417487775, 0.11839540575845798609, 0.12653321954828453122, 0.093447167913049550281, -0.32710542216956800532, -0.54086862466055896093, 0.051163274954256883487, 1.0561306102134151441, -0.43131471480284311326, 1.0423263669369600137, 0.047931421114602350275, -0.56268918525014144372, -0.46927733825544620805, -0.19470249815739154586, 0.024108004735840307675, 0.10351516088594138465, 0.10513172736400321072, 0.073728268244231731643, 0.018965735487016251171, -0.29324647070189691078, 0.021092487211096015709, 0.69963731792500938322, -0.87096412239140807987, 0.71667305499229061461, 0.04671049382555027274, -0.32233853658086725202, -0.18697816011967469119, -0.0049009376367656343199, 0.09996258354288876935, 0.094255674052298052401, 0.030972282027147805122, -0.055375946745247081604, 0.22495087878303340889, -0.2552065218071879471, -0.089943309638219309798, 0.59124224240340650383, -0.96024384599766865467, 0.61031710860699062149, -0.045590492566272289299, -0.2340173364773981135, 0.042232350231957337117, 0.11642769903511900054, 0.067686177754498602632, 0.0037982678123485713595, -0.030073095198219934598, -0.038532871384626944666, 0.99255218063262995454, -0.16721034053653377471, -2.0533987485827438491, -5.1487655214746750332, -9.8523938235558734533, 32.60784706164374569, -9.7720672474756060666, -5.0992842868541146117, -2.2215248859481775412, -0.71479912733602324337, 0.0080508303491958581322, 0.3910802857296980517, 0.51714551491008786677, 0.45990745212555911303, 0.039388379254616204261, -0.057618183015610745235, -0.15815322021881034198, -0.07777840541946383357, 0.53325368740000300694, -0.56157971978604237506, 0.53801538554319028318, -0.064530946249571896622, -0.14745138573335264676, -0.076387759764803289664, -0.045495146736091472695, 0.015226718132134322778, 0.040866718883036202981, 0.0199891566163329093, 0.055407226844519368025, 0.015793395764648766183, -0.25103090877098682698, -0.096656671673811742074, 0.63767749881917779042, -0.71821081253987839599, 0.63894526978379218018, -0.094811232952439691846, -0.25639708194085908399, -0.0033746852599737774125, 0.04621082856382719567, 0.019356395399304452548, 0.0017336716188176364527, 0.0048207491164550435661, -0.34083262588201046484, 0.27944194621433254833, 0.55330793544501899461, 0.43307782954204615544, 0.084490916372857716166, -2.0062761013896475148, 0.053663268495242749667, 0.38251762790683169202, 0.51521370581323489013, 0.35123715355836343521, 0.024049464026996418553, -0.14488753027015915942, -0.14533895320009720953, -0.037862415477497336769, -0.10637003307853773459, -0.059775185529189388267, -0.02005416702554755487, 0.31666050884036534008, 0.59207845632102318323, -1.442799094690135675, 0.59114628982107353394, 0.31663535230062483938, -0.024705232413667607694, -0.071576833180307095073, -0.10671329752370968535, -0.062012209704488681361, 0.0081707281232672249183, 0.063326725623633700613, -0.1029759329597608325, -0.082745330985450282757, 0.089039153274950738615, 0.42668902677406533863, 0.46873730747875086422, -1.6129419825615984863, 0.47571004591693966113, 0.44111554114239764912, 0.10291548848119462523, -0.079561794381856121294, -0.13303970739324877393, -0.057693117867345176941, 0.015447973995836597472, 0.046496013167894811868, -0.049349762166686100828, -0.043302322255818717633, 0.060959401270893540103, 0.41772024442898808561, 0.48455388270730725964, -1.7545813783128085817, 0.48394155902882335374, 0.42499360330091839888, 0.079297629120553728144, -0.019931598962328343771, -0.050673228651762110275, -0.019915603183354495254, -0.0020682986274824757995, -0.010239737722237221909, -0.62135409661863438835, 0.43327693989060639401, 1.3189075813470487741, 1.3739630505235758129, -0.4030430867557476815, -4.22457911285019172, -0.44014500866934630485, 1.3189467824310998623, 1.3030816326940002359, 0.60567723867916301028, -0.021608858408996303724, -0.2653137271018932708, -0.25295782550491674456, -0.11069916888486110162, -0.55455426412556407811, 0.28985045594128389101, 1.0521994666741636415, 1.2337482704887334783, -0.15024404503192395621, -3.7522887595409950556, -0.18168716744778051764, 1.1867892689204466716, 1.0340050830095124379, 0.41714816719219977026, -0.091681455640926626249, -0.24620047969862784987, -0.18927775283352157865, -0.041240006927110488266, -0.047445517690177581305, 0.31688713944445867643, 0.63878531573104690988, 0.8027112720796578671, -0.094203016244202103513, -3.3045055899242603559, -0.094079860499478640268, 0.81927488332872477894, 0.70303014595390123098, 0.47611616259545813667, 0.22334862799468452321, 0.016917838327879342747, -0.1546518439858077798, -0.2697536219430493043, -0.085878052299965812466, 0.29478056545162656388, 0.60386893177015765755, 0.72081129926382037088, -0.053401008074680865467, -3.0163206807823050859, -0.058419882626732859421, 0.72665330312581555283, 0.65209362797219527774, 0.43723686113390397567, 0.18352935333625156056, -0.0047402078895036163508, -0.14543996209728668467, -0.22797942397955534011, 0.14455098409324518283, -0.13243813891977215813, -0.41313085496232354599, -0.16291477553281646928, 0.77948114793488887386, -0.41770357311508293785, 0.78093658227154538753, -0.1594881571770514761, -0.41444534183704800911, -0.16225135690234429586, 0.011759563499538151082, 0.064327125115119995757, 0.048020251347166399702, 0.02910958828769480769, 0.12218323939096976949, -0.099061879896652021982, -0.37066789762944873932, -0.20758471931002650535, 0.7247222886288118282, -0.32771208376027022302, 0.7275226582941783926, -0.20324759745668455335, -0.37397666387507028096, -0.13049951210296689696, 0.0089374308606942776817, 0.052281273808711795081, 0.042473956254106208508, 0.030704113862118551787, 0.031405980759310697659, 0.072852660202759988994, -0.11486201930693198947, -0.18813140386277599214, 0.47947979031774340397, -0.56909624537172076586, 0.48802101363780542309, -0.17729844523028193137, -0.11615462193813679004, 0.060118127022419802497, 0.041791614320010987693, 0.0075380784722984166615, -0.0082169639784712613512, -0.004830932772481237418, -0.077273424767818982284, 0.343220904026016016, 0.65185370111797946713, 0.65376057789205721527, -0.13352321421278315339, -2.9527322316713600081, -0.12399146215472622856, 0.6817230172451256287, 0.71394709054128213666, 0.4787498904719620918, 0.18104437457558336666, -0.0089946372875500738009, -0.14139131343479466318, -0.23549572858380504736, 0.12721562535334082766, 0.051953352285678711109, -0.29746666935830973211, -0.30286905203394870467, 0.59375219562425907149, -0.32491437364045783198, 0.58267055552539570584, -0.3217033390400896864, -0.31430336116257440171, 0.054344710705303680975, 0.14947891056283288402, 0.070574098767289664136, -0.022416710087888837982, -0.044399104185594860661, -0.046509793575356832285, -0.19815683365728056775, -0.075483080568008883993, 0.58008313176240400288, 0.69772634026260460516, -1.9021239263810540976, 0.68508732658197346588, 0.56514568026194966777, -0.075316952374080509136, -0.17913872579579795397, -0.095882442062358338997, -0.018244732858169940976, 0.012494148768508019434, 0.043533002296179290624, -0.0019161509417834918321, -0.19298145706010463862, -0.13047207147716533004, 0.45646698610669600438, 0.698718024561075568, -1.6506381819219881901, 0.6906698134363421504, 0.44915540878151016813, -0.12488223233998096973, -0.18091097290909802231, -0.080230245769182029214, 0.0020035282159337544639, 0.024051823434768582199, 0.035472518169526262577, 0.104225285231592385, -0.48761142027353404638, -0.62906348540367551525, 0.11784851642916363557, 1.1529625015166764523, -0.47870760393301609037, 1.1498720954204109113, 0.11765711532088241054, -0.63523303419774501677, -0.56457110896651985232, -0.23913894286870610029, 0.034757181385315340238, 0.15490436159923276493, 0.17834288860213826089, 0.10478782589921363522, -0.40338958044102879441, -0.55702211071814811127, 0.049346353033346043726, 1.0423263672890361597, -0.44127776827988368646, 1.0409734999667767941, 0.051343774436432711161, -0.56060272182597692403, -0.46927733871128018972, -0.19189870657922328268, 0.037427710318916451726, 0.13252287419125310941, 0.14544168535081405946, 0.20673346111305818118, -0.16940928714929393006, -0.36075087760519819469, 0.016767282410758518885, 0.71667305589992524872, -0.84533079606753225299, 0.73715852250386415534, 0.057660696128236106628, -0.31911684610802765194, -0.1869781599700690855, -0.0061625339352470191387, 0.094057231266590018981, 0.075010641701992314401, -0.013423427854958109445, 0.12924523967624867993, 0.03576162122260648224, -0.24814520898354036627, -0.054497268929009093363, 0.61031711003430677565, -0.95213476696307353109, 0.61126890360253249845, -0.047462382780107476021, -0.23592106300970003274, 0.042232349201991023491, 0.11815083720502278508, 0.067314981200734702593, -0.0071250482149293140749, -0.063084530245824249617, 1.0442087760393496954, 0.35764969440753363594, -0.59221372971767771709, -2.2305342976749757966, -5.1426064625892111692, -9.7720672477507157794, 32.683246339903945454, -9.7354048320632777802, -5.0992842860049165665, -2.2355516188851987103, -0.71096063762621719562, 0.064927000624394728967, 0.53972055009746988663, 0.80539064652542158296, 0.081782202458265090828, -0.037574207675677991081, -0.084278130901863715363, -0.15210208045443418534, -0.065469571479072161124, 0.53801538486554434382, -0.56113708294856934344, 0.53774270616375730025, -0.064530946734287455979, -0.14716869639127180824, -0.076109619624908253499, -0.046112054901285400021, 0.015931200402927551962, 0.060691898132422462953, 0.010424440164163816316, 0.059587108490936631366, 0.0030214983017569784114, -0.2571068270294248892, -0.097192198804740601026, 0.63894526913845350879, -0.71554287583453968224, 0.6409085337457750553, -0.094811233954568363558, -0.2577701876052099772, -0.0038898278148556189593, 0.048638987970920956327, 0.022501379876682405662, 0.0022865903032118152678, -0.33453925264104655835, -0.02118052103191384733, 0.36954520976430588197, 0.54755985735054124586, 0.40330643416851424954, 0.053663268755399902432, -2.0287601887725421612, 0.0423843910053053885, 0.38251762935790073472, 0.52420877014270206296, 0.36377654408295445609, 0.030380311595513458112, -0.15755289375937492347, -0.18317467465044368069, 0.0098881759505059455689, -0.090141773839470529239, -0.076708194544743452115, -0.029900493298011796006, 0.31532317829667372289, 0.5911462890905968548, -1.443729575345195304, 0.59086486740373722615, 0.31663535212139864994, -0.025779503877271115614, -0.07443667871386325241, -0.10812554317249910285, -0.049020497361391635049, 0.067357711624474736478, 0.037570812004669808248, -0.13192503859334250849, -0.10350536168796661829, 0.087104942908367108401, 0.43631693247190256324, 0.47571004568175123683, -1.6110836203896490293, 0.47547836756794442659, 0.441115541860480187, 0.10119170272611716088, -0.087102847037360722759, -0.14457847663244147163, -0.054553369939000541378, 0.075781105650224167558, -0.0040179086530124303103, -0.067823218846811597627, -0.036124631009626179123, 0.075751915539378641973, 0.42736406496077355976, 0.48394155966687030235, -1.7593638577762471531, 0.48055624380984918176, 0.42499360379662082243, 0.081115277599698901168, -0.020087467234681005007, -0.056280565006021777974, -0.029305388270117343347, 0.00021690349704589525332, -0.54747608132194580666, -0.14302043263190572508, 0.58764006381860967032, 1.3279126312223243911, 1.3422659956120506219, -0.44014500809822199034, -4.2534936188528744694, -0.45492408449522153679, 1.3189467822626648186, 1.3116594140663089263, 0.61005135676320021432, -0.03699886563530058875, -0.30936478828982688638, -0.31375860105328701177, -0.41895274417705707926, -0.17250326395445905736, 0.40394251601550368314, 1.0517826433361732175, 1.2041067765088493413, -0.18168716722369601668, -3.7750961133958651494, -0.19251880619477093615, 1.1867892693843618002, 1.0392850301587222006, 0.41887923495716938849, -0.1019826266518978064, -0.26795039660995983999, -0.19955162264060682942, -0.27173464954785397918, 0.076368700592096092072, 0.42283842172408281357, 0.69721129860351593877, 0.82524068277956352802, -0.094079860083328636922, -3.3128953571257375721, -0.10085333238484928131, 0.81927488165650563801, 0.70696434303447086567, 0.47201616740813734374, 0.19098418194370037071, -0.071830362069162781435, -0.3315275695704386627, -0.26801550386497674516, 0.055832599821408153418, 0.3966017040709495789, 0.6527706841413235539, 0.73501754927467022327, -0.058419881988410263252, -3.0269525745932654992, -0.06580842808132872368, 0.72665330320030885325, 0.65681298320100567345, 0.43617331790861391116, 0.15863015375115907535, -0.079170190316966662647, -0.29834742074301812131, 0.10511949812289858586, 0.049709152929924609043, -0.14834107146370681396, -0.41639785239198906508, -0.1646833650798768256, 0.78093658260805376425, -0.41221018753886107255, 0.78554686219847325024, -0.1594881579452569198, -0.41796446789108609998, -0.16299455862815129259, 0.021617730268532073246, 0.082922013955058229695, 0.054629624404790619319, 0.093977394690999679905, 0.043783367122727311072, -0.1189885240247547521, -0.37684158808878781555, -0.2085259624450223892, 0.72752265800004745078, -0.32199009582981591437, 0.73179125119881205386, -0.20324759891836952086, -0.3771878592410097264, -0.13174860467916005291, 0.016499507095200555834, 0.068782120725063441991, 0.05463299569040913628, 0.0028875815571327229439, 0.043787428599696420761, 0.053613208881609822343, -0.12555421009826364664, -0.1835911415623403542, 0.48802101400026393074, -0.56230861951235100449, 0.4911498681302045588, -0.17729844594069932295, -0.1188633235718695863, 0.054601445798930822728, 0.034647358141188216996, 0.0027872696517324034929, -0.00039719103921219092073, -0.23775913001405424296, 0.041987319736301195738, 0.41500118620369441658, 0.69708447108261439862, 0.68191360005265466704, -0.12399146210329450857, -2.9550941256061782347, -0.12837396412611321028, 0.68172301832876491812, 0.71553605543965492775, 0.46895210377364909826, 0.1410130038826434562, -0.095280894673270277728, -0.27517309182395432998, -0.045902435683287931556, 0.15987411497162332119, 0.085871964815953757966, -0.29700979353391893101, -0.31767492147072634578, 0.58267055645152088239, -0.32714061198489946536, 0.58367588897284949923, -0.32170333946138834902, -0.31361664790833015948, 0.061730826712979358639, 0.16541511560956473348, 0.073974547244788718681, -0.089028138185821392714, 0.013119084004389113679, -0.072849866472018146024, -0.16819964958565220003, -0.070910813224574265057, 0.56675565301543406882, 0.68508732499977709196, -1.9068245357353394986, 0.6849721365955354635, 0.56514568027875899947, -0.076730314201293919885, -0.17840049623484546815, -0.084850454840655537003, 0.0052991854927387834623, 0.03035367856465657696, 0.042630586470491296791, -0.058319355790201338197, -0.17392930905356854265, -0.12404035870488398785, 0.44875827299395609771, 0.69066981337047261835, -1.6528158253816913614, 0.69148122338266271303, 0.44915540901445405453, -0.12681445253440945531, -0.18128985036456612967, -0.071470239543463112697, 0.021909128026587464805, 0.038448413132831225913, 0.28813942172694134713, -0.13789906217908393105, -0.54007921438415051885, -0.63549469899442956322, 0.11476496063199369269, 1.1498720949829470683, -0.47807935771641313361, 1.1519466416462442737, 0.11765711647747824353, -0.63711296490953028826, -0.56086304934139874501, -0.21414884657319058214, 0.098153087033571395636, 0.26393145178984595001, 0.24958334034511686639, -0.10712697231039701451, -0.44944464157024988094, -0.56202383363183605614, 0.047931420975820392127, 1.0409734988039964776, -0.43960348376339924981, 1.0433737815728187925, 0.051343774235667952177, -0.56268918443590643097, -0.46680883041222015484, -0.171416557228820432, 0.089732676268787670004, 0.22103695365129494288, 0.18715382073960182274, -0.00024212318153025107594, -0.21356736801242034995, -0.34244430712507295533, 0.046710494493900169, 0.73715852293686090491, -0.83517158995246898456, 0.74069608701165123854, 0.057660696267011403437, -0.32233853759187719445, -0.19463569514491965706, -0.017159902425901783773, 0.085406637859544637381, 0.079183024670469032147, -0.0015533736599415506625, 0.10470868060426437307, 0.042977257908674584841, -0.23278320232815921664, -0.045590492662729166506, 0.6112689045474726246, -0.95354492322148920458, 0.61008305629195025421, -0.047462382556258148059, -0.23401733626671894495, 0.045817087349301566235, 0.11883442711478216236, 0.050578012952670906022, -0.062619933631609211999, 0.80539064581358910111, 0.53972055089637560599, 0.064927000647307470116, -0.71096063768168316077, -2.2355516187148198881, -5.0992842864214864562, -9.7354048310145930856, 32.683246338779994744, -9.7720672479474899319, -5.1426064619456060001, -2.2305342976345476913, -0.59221372937857164054, 0.35764969278961161736, 1.0442087771045902578, 0.060691898183014569845, 0.015931200256548854605, -0.046112054500247495747, -0.076109620351230880542, -0.14716869577605148911, -0.064530947174733108151, 0.53774270707166649252, -0.56113708435803588692, 0.53801538616262511994, -0.06546957236586709028, -0.15210208001660088795, -0.084278130942226928135, -0.037574207772501005964, 0.081782202496124029034, 0.0022865905564131562382, 0.022501379022867012275, 0.048638989335692807414, -0.0038898292807810828757, -0.25777018671149287687, -0.094811233981127632231, 0.64090853404306147301, -0.71554287740134758788, 0.63894527104718168431, -0.097192200036962161924, -0.25710682598544232214, 0.0030214970643553645119, 0.059587109331497606357, 0.010424439945471089194, -0.18317467297433628892, -0.15755289708159722628, 0.030380311318889990924, 0.36377654754647520408, 0.52420877086901951358, 0.38251762656090676495, 0.042384388711561353513, -2.0287601859581760877, 0.053663270933888333747, 0.40330643330165938387, 0.54755985438651344488, 0.36954520967464749015, -0.021180517733415869103, -0.33453925418300156647, 0.067357711592423541447, -0.049020497199575636948, -0.10812554366819733531, -0.074436678277600601406, -0.025779503697108459986, 0.31663535159354061488, 0.59086486815774519332, -1.4437295764343258675, 0.59114629040699107065, 0.31532317733383768976, -0.029900493233861520637, -0.076708194530488729712, -0.090141773436401090303, 0.0098881757263807861652, 0.075781105844346163813, -0.054553370324180297524, -0.14457847654123098802, -0.087102846995555330278, 0.1011917030075509677, 0.44111554149802911784, 0.47547836811165350435, -1.6110836213636972047, 0.47571004643764464026, 0.43631693263493764867, 0.08710494236740141949, -0.10350536191708117895, -0.13192503783674724982, 0.037570811668278268625, 0.0002169034867467934584, -0.029305388159041229762, -0.056280565292427198187, -0.02008746700913565178, 0.081115277817290568074, 0.42499360329248730972, 0.48055624400265262164, -1.7593638575233407906, 0.48394155912682473852, 0.42736406578825741898, 0.075751914760043670416, -0.036124630746031304029, -0.067823218732379356166, -0.004017908738869531253, -0.31375860107919500974, -0.3093647883703620205, -0.036998865699753441527, 0.61005135730738702105, 1.3116594136676107407, 1.3189467825939129586, -0.45492408559086627839, -4.2534936177547661984, -0.44014500850847471103, 1.3422659958034466321, 1.3279126311639037894, 0.58764006362146969931, -0.1430204322899343572, -0.54747608150428539453, -0.19955162294808126422, -0.26795039612772603155, -0.10198262634207620747, 0.41887923420990563761, 1.0392850299763514155, 1.1867892699544655422, -0.19251880627153869496, -3.7750961131041527175, -0.18168716723986580441, 1.2041067752379563949, 1.0517826443252684676, 0.40394251595282831824, -0.17250326352250713868, -0.41895274459824499713, -0.33152756950773115685, -0.0718303612864448493, 0.19098418154055371576, 0.47201616676535157469, 0.70696434208510194885, 0.81927488275313342658, -0.10085333222062467518, -3.312895356743245312, -0.094079861440753045909, 0.82524068335949041231, 0.69721129997763042763, 0.42283842128155851858, 0.076368699431770631358, -0.27173464903250149627, -0.29834742080432741229, -0.079170189947462332736, 0.15863015328315724362, 0.43617331857911989523, 0.65681298170878787968, 0.72665330453584064951, -0.065808428623099407551, -3.02695257363926018, -0.058419883556361511256, 0.73501755006501579004, 0.65277068445833197341, 0.39660170400462918527, 0.055832599264275431517, -0.26801550354088377048, 0.054629623945548329289, 0.082922014836861540621, 0.021617730168690119097, -0.16299455968375112236, -0.41796446712735385587, -0.15948815766360535662, 0.78554686214816904499, -0.41221018767641176561, 0.78093658211852978734, -0.16468336484272175224, -0.4163978524897332667, -0.14834107085835218642, 0.049709152488886729215, 0.10511949817832695608, 0.054632995148215741932, 0.068782121853970229486, 0.016499506676132456168, -0.13174860503394697497, -0.37718785988644842888, -0.20324759756605370598, 0.73179125167489833537, -0.32199009729113026168, 0.72752265766637991273, -0.20852596115865951187, -0.37684158884100832365, -0.11898852334139978226, 0.043783366508581618437, 0.093977394882642562024, -0.00039719148877056349115, 0.0027872706521276111383, 0.034647358121783509988, 0.054601444863277358299, -0.11886332359115632518, -0.1772984460196411205, 0.49114986970806639999, -0.56230862070637555039, 0.48802101415359205872, -0.18359114293653813754, -0.12555420752496809778, 0.053613208021832191308, 0.043787427669307509015, 0.0028875821145512353336, -0.27517309132202705246, -0.095280895353281547244, 0.14101300439098543582, 0.46895210277635823726, 0.71553605626631966441, 0.68172301803332002734, -0.12837396392082539576, -2.9550941254516027712, -0.12399146256236380503, 0.68191360052050320473, 0.69708447025123032592, 0.41500118681777187213, 0.041987320339656158952, -0.23775913063659454005, -0.089028138383408408907, 0.073974547656821931785, 0.1654151154066758056, 0.06173082683815936389, -0.3136166483183486231, -0.32170333897750452712, 0.58367588838100659121, -0.32714061116524989492, 0.58267055577914828035, -0.3176749212429970104, -0.29700979352112849563, 0.08587196501388053016, 0.15987411461109354849, -0.045902435527611362698, 0.030353678775177032195, 0.0052991849234114664011, -0.084850455308257299869, -0.1784004943564350687, -0.076730315551986075229, 0.56514568101719941939, 0.68497213576625626263, -1.9068245354457855623, 0.68508732554833051687, 0.5667556531252239127, -0.070910815065041113425, -0.16819964825022820643, -0.072849866236692428778, 0.013119083716045379373, 0.038448413458114547026, 0.021909127641328268732, -0.071470239754208367278, -0.18128985060556890296, -0.12681445088708315216, 0.44915540754878535701, 0.69148122288600766261, -1.6528158239077754832, 0.69066981295562368537, 0.44875827275310098408, -0.12404035909813548866, -0.1739293088896460282, -0.058319355062895213149, 0.042630585978444872353, 0.26393145186018818205, 0.098153086474243186887, -0.21414884649386445314, -0.56086304876252446316, -0.637112964195286402, 0.11765711553725806404, 1.151946641368740476, -0.47807935719617061787, 1.1498720943831779451, 0.11476496168431561673, -0.63549469947555303762, -0.54007921533185920815, -0.13789906133631044116, 0.28813942167954620377, 0.22103695368678530353, 0.08973267600454921944, -0.17141655720780146765, -0.4668088296893753153, -0.56268918499652109677, 0.051343774568372801959, 1.0433737808394747404, -0.43960348378950098169, 1.0409734999127082666, 0.047931420445652878071, -0.56202383383688259144, -0.44944464127210270732, -0.10712697283838742979, 0.24958334067539544865, 0.079183024169802412873, 0.085406638913419569148, -0.017159902219017877933, -0.19463569601264199549, -0.3223385389169560189, 0.057660697791877039975, 0.74069608813999432417, -0.83517159113701988638, 0.73715852197611486663, 0.046710494855611822684, -0.34244430558142552323, -0.21356736814841414951, -0.00024212494160044120087, 0.18715382165864674624, -0.062619933737616637059, 0.050578013080490556719, 0.1188344275713194842, 0.045817086451269029013, -0.23401733559520845884, -0.047462383161287646782, 0.61008305676810026252, -0.95354492321108519359, 0.61126890414898371606, -0.045590492544655768226, -0.23278320184889422784, 0.042977257955530756561, 0.10470867983089381104, -0.0015533732639845898669, 0.45990745253320192143, 0.51714551410316467184, 0.39108028618702156409, 0.0080508306088596121725, -0.71479912712587578838, -2.2215248872029813576, -5.0992842860009659489, -9.7720672485161177434, 32.607847063838121926, -9.8523938250718519072, -5.1487655210567675468, -2.053398748961670961, -0.16721034048765753832, 0.99255218080209439702, 0.019989156754133170085, 0.040866718730968767948, 0.015226717980776937819, -0.045495146254542741249, -0.076387760678757896105, -0.14745138456319348275, -0.064530947233418067821, 0.53801538646722202319, -0.56157972082890084131, 0.53325368842239495581, -0.077778406121667226225, -0.1581532204595700597, -0.057618182105623812206, 0.039388378794855516407, 0.0048207492471619723534, 0.0017336711823303062319, 0.019356395744957476113, 0.046210828945793036571, -0.0033746862528004173236, -0.25639708073592154536, -0.094811234159548452016, 0.6389452708813599946, -0.7182108139459685292, 0.63767750046821236598, -0.096656672428185957946, -0.25103090897615176758, 0.015793395930208262851, 0.055407226872440991372, -0.037862416252654441939, -0.14533895240061062148, -0.14488752988876002492, 0.024049464152248165028, 0.35123715211280159831, 0.51521370709751368455, 0.38251762761570046179, 0.053663268839135441768, -2.0062761029251157119, 0.084490917858250191852, 0.4330778296890879786, 0.55330793526086963219, 0.27944194502109676037, -0.34083262503317057757, 0.063326725589950311202, 0.0081707285432783965939, -0.062012210594720000578, -0.10671329656755551563, -0.071576834507318679024, -0.024705230806185436015, 0.31663535138475290553, 0.59114629048859579363, -1.4427990958758913909, 0.59207845748164533717, 0.31666050792865535479, -0.020054166159914163514, -0.059775186108113650962, -0.10637003291024280105, 0.04649601305231786802, 0.015447974085088569604, -0.057693117921641973167, -0.13303970705120021778, -0.079561794671513835775, 0.10291548841220669874, 0.44111554127371377376, 0.47571004603875943761, -1.6129419831685571918, 0.46873730821605896768, 0.42668902650820511591, 0.089039153427649661565, -0.082745331408708128196, -0.10297593271049609864, -0.010239737769736270007, -0.0020682982793413648079, -0.019915603809418824105, -0.050673228349500269718, -0.019931598803649547974, 0.079297628950170243045, 0.42499360320762430421, 0.48394155943084687932, -1.7545813786980817284, 0.48455388304979712943, 0.4177202435900134736, 0.060959402131765358379, -0.043302322375590160397, -0.049349762302234832789, -0.11069916889163229634, -0.252957825788203583, -0.26531372646675560611, -0.021608859169914017978, 0.60567723964807429482, 1.3030816317955480343, 1.3189467829032270885, -0.44014500884985374718, -4.2245791126521066161, -0.40304308722950260568, 1.3739630511136007307, 1.3189075810219244023, 0.43327693990921084533, -0.62135409657606333056, -0.041240006877185354794, -0.18927775283283149177, -0.24620047993681690168, -0.091681455644947479344, 0.41714816785512964792, 1.0340050821071793408, 1.1867892695564012939, -0.18168716751341859594, -3.7522887595817673301, -0.15024404530568738481, 1.2337482704591467009, 1.0521994669728356175, 0.28985045587987329219, -0.55455426415709896393, -0.26975362181065859479, -0.1546518439866000072, 0.016917838247558235343, 0.22334862778118907967, 0.47611616235199860636, 0.70303014718384948534, 0.81927488209919285822, -0.094079860257585190553, -3.3045055898284760865, -0.09420301612734612029, 0.80271127197815383969, 0.63878531585047659824, 0.31688713970793191965, -0.047445518027443733811, -0.22797942408310112317, -0.14543996233368666759, -0.0047402069683754946561, 0.18352935240377604709, 0.43723686197247552121, 0.65209362670334003464, 0.72665330421557872143, -0.058419883270289803268, -3.0163206798533681585, -0.053401008885312906127, 0.72081129969191359841, 0.60386893095291471756, 0.29478056643271738357, -0.085878052685106162634, 0.02910958861165227618, 0.048020251219308487944, 0.064327124242471014903, 0.011759564040765006082, -0.16225135623318409639, -0.41444534238011832006, -0.15948815745476907502, 0.78093658264051191153, -0.41770357266791335871, 0.77948114707209170682, -0.16291477568220955052, -0.41313085422524298274, -0.13243813899606002304, 0.14455098391862528096, 0.030704114159085804658, 0.04247395592952472182, 0.052281273390588439698, 0.0089374315556383452513, -0.13049951241474691543, -0.37397666369759618998, -0.20324759755428165042, 0.72752265831919737948, -0.32771208392916684193, 0.72472228893038226349, -0.20758471989224119847, -0.37066789698007135412, -0.099061880069208602118, 0.12218323932429084322, -0.004830932767186850961, -0.0082169637899141782095, 0.0075380781383769766835, 0.041791614636351302747, 0.060118126493209238237, -0.11615462095870640602, -0.17729844676644304635, 0.48802101520891094655, -0.56909624645356582384, 0.47947979118688710543, -0.18813140450248103108, -0.11486201919845853703, 0.072852660387211734538, 0.031405980661123968078, -0.23549572825720332947, -0.14139131375844740668, -0.0089946376066130304117, 0.18104437450327356962, 0.47874989115763205705, 0.71394708977319631948, 0.68172301869262386109, -0.12399146335816120523, -2.9527322313434405388, -0.13352321515786541584, 0.65376057892367578361, 0.65185370088663030508, 0.34322090461078974544, -0.077273425324200520348, -0.044399104153168389431, -0.022416709881032843843, 0.070574098165710447139, 0.14947891104544433483, 0.054344710596549397619, -0.3143033610214502338, -0.32170333954337160254, 0.58267055613612506182, -0.32491437406105139196, 0.59375219613378937833, -0.3028690527597449611, -0.29746666881784167291, 0.051953352100098659783, 0.12721562537789446456, 0.0435330020021250097, 0.012494149298401282383, -0.018244733206983658469, -0.095882441745379881626, -0.1791387262942406855, -0.075316952423647276493, 0.56514568156167410695, 0.68508732508464442734, -1.9021239256645405824, 0.69772633992251376078, 0.58008313188365190349, -0.07548308022648921034, -0.1981568342721450604, -0.046509793259409709654, 0.035472518241454219223, 0.024051823614421843217, 0.0020035278489401592283, -0.08023024618229324112, -0.18091097221005952855, -0.12488223222297956116, 0.44915540861622782476, 0.69066981326453669165, -1.6506381819952715695, 0.69871802454712217401, 0.45646698634364407665, -0.1304720716157255489, -0.1929814569261462387, -0.0019161510357831835431, 0.17834288843278905956, 0.15490436107683952516, 0.034757183377293138227, -0.23913894488354006662, -0.56457110769296692254, -0.63523303537905917437, 0.11765711635017982173, 1.1498720944822458101, -0.47870760290707653484, 1.1529625007187043195, 0.11784851758120978849, -0.62906348691984370625, -0.48761141977775829259, 0.10422528540786223095, 0.14544168503983778695, 0.1325228746536626645, 0.03742771045657061757, -0.19189870687406776484, -0.46927733884840910905, -0.56060272153362067371, 0.051343774339937496565, 1.0409734994600416957, -0.4412777674283691054, 1.0423263665487556562, 0.049346353752146299154, -0.55702211083286379267, -0.40338958117685908, 0.10478782637928657495, -0.013423427861535714226, 0.075010641811387876454, 0.094057231146584180381, -0.0061625336838502927858, -0.18697816032200550995, -0.31911684615382157615, 0.057660696268131167441, 0.73715852239930823586, -0.84533079581280490267, 0.71667305587770113728, 0.01676728232899035026, -0.36075087763835678167, -0.16940928713476272049, 0.20673346111186760576, -0.063084530353775772338, -0.0071250480927175091628, 0.067314981304022955544, 0.11815083697232720006, 0.042232349676176332565, -0.23592106356202868778, -0.047462382937142118255, 0.61126890407857326082, -0.95213476677478769972, 0.61031710954618179077, -0.054497268834779226399, -0.24814520891613123244, 0.035761621344593090555, 0.12924523956346478659, 0.14941470765349085248, 0.37592663291012945459, 0.47724956057509965168, 0.3777420831524104794, 0.0075969312324909027412, -0.71479912765622399817, -2.2355516188755553131, -5.1426064614676345599, -9.8523938242875992444, 32.509275725625691678, -9.8912888466068356053, -4.9992453469102322572, -1.6084627544074625582, 0.45501148776930999684, -0.0057035357533043746364, 0.03031434977122975799, 0.040782754052677554213, 0.012637894086877116823, -0.046834603518206521788, -0.07638776004772172834, -0.14716869571043086906, -0.065469572086377908482, 0.53325368852406351827, -0.57172941861186243884, 0.5278354560728217848, -0.06614313257593289197, -0.11782308396354053415, -0.051991628783665415348, 0.0034043242693505867784, 0.00061699795885221788219, 0.0018128843744855828012, 0.020649141985151206469, 0.047188951994678353385, -0.003374685495027839998, -0.25777018721981898031, -0.097192199407518456433, 0.63767749991102951057, -0.71660575755716293678, 0.64099809334987456477, -0.10730222274573533725, -0.2681739830215789655, 0.096365054674736119744, 0.04442118320964405509, -0.073759208534002515534, -0.16160501728490880846, -0.15339691506294877987, 0.019027564217334744234, 0.35123715386871406929, 0.52420877072819227482, 0.40330643323300169323, 0.084490915930512541787, -1.9739137355148319575, 0.10593718341119698345, 0.41061637969265935011, 0.44329930731317385151, -0.007996551555849159415, 0.034748222873211567774, 0.034754841903453494167, -0.0016311517878615773656, -0.063637352386000725635, -0.10620996448429911863, -0.071576834206207307076, -0.025779503883277991166, 0.31532317804200121758, 0.59207845714831708328, -1.4348666231132733984, 0.61569605849080422111, 0.34388381240337051459, -0.034649479086264894923, -0.20144930369053465347, 0.0075214412868182803015, 0.036209676297406866163, 0.016187524174299135898, -0.055057382418827328918, -0.13172721982750776326, -0.079561794460328502665, 0.10119170295417853622, 0.43631693253757836137, 0.4687373083133128393, -1.6149097853193246088, 0.48820778253961194526, 0.46669287199093623109, 0.10660674392835296775, -0.24801875396392630835, -0.013119362302637804438, 0.0022110944389117531163, 0.00078158025664110835663, -0.021292397329253150151, -0.052308910779975521799, -0.019931598510508383854, 0.081115277952961806207, 0.42736406512617797215, 0.48455388333459026295, -1.7574528412449328396, 0.48481339975123871611, 0.42354628959813489519, 0.065262227593524874125, -0.10301149156445880972, 0.011313548365927789466, -0.1369915229243856003, -0.25612753572044527628, -0.26369714064547505883, -0.022375739983658850946, 0.60567723966484310338, 1.3116594134386063697, 1.3422659953367168661, -0.40304308758953549807, -4.183520888455031006, -0.36789718014521988199, 1.3725925245322971868, 1.1798639589081316714, -0.15442901319377935754, 0.037326364079103308002, -0.084579618148952775969, -0.20092131650488292971, -0.24632321247643371165, -0.091314469431691455692, 0.41714816749847233313, 1.0392850297510654034, 1.2041067757013439454, -0.15024404518886011473, -3.7122843836340631007, -0.10810042798409363884, 1.2479825468678320544, 0.93492877251473893718, -0.26213681302534924678, -0.15734323162944552355, -0.16214620628788750656, -0.10310481422798026263, 0.034266125874811703689, 0.2253253966942674047, 0.47611616127885492133, 0.70696434312560407776, 0.82524068314915099975, -0.094203015789127736435, -3.3228625591445415388, -0.13434555988507385083, 0.75307272457424734302, 0.62883646830309658515, 0.35802613180287562455, -0.12297682939021513382, -0.14365041556868130002, -0.10604155261483284856, 0.0076809298486684371432, 0.18418278614778926805, 0.43723686152473756339, 0.65681298164328927314, 0.73501755018644032713, -0.053401009143130506729, -3.0266209896362337162, -0.084857047123912149322, 0.67481145644655682325, 0.58283688015339074617, 0.28892633576316578692, 0.013375578212678177967, 0.021769509689622452281, 0.042482584268547632367, 0.063514626902351906113, 0.013278360872796475187, -0.1622513563997300956, -0.41796446758255684228, -0.16468336537615185788, 0.77948114771594423367, -0.41372367496686784927, 0.7795625503856944416, -0.18824519452920668305, -0.41719689276002708267, 0.042696606962659006124, 0.012820586801251889625, 0.022479331814413371776, 0.03810908446595331206, 0.052125100355002645935, 0.010404890777383865585, -0.13049951227028289136, -0.37718785967036788653, -0.20852596182041188699, 0.72472228902237023629, -0.32539309440256997119, 0.72652517142910000292, -0.2250213381429945203, -0.36949799378266917627, 0.041642013282160175724, -0.01341169579732723327, -0.002655189092631875334, 0.00077824078668603806526, 0.014460673447281535003, 0.044717173898663378773, 0.060118126004542998819, -0.11886332405353566832, -0.18359114227591497026, 0.47947979153505609906, -0.57430505871025727682, 0.48535823794519994401, -0.18240719481818118419, -0.10788271396143789271, 0.098923521934472360972, -0.1539998387431054383, -0.13586541246733804811, -0.087010465200252601337, 0.012416652840651999645, 0.18515923127830113959, 0.47874989170071574618, 0.71553605593577718214, 0.68191360042857240842, -0.13352321489737783988, -2.9781545630829073801, -0.16377883292117739744, 0.63930886632247851953, 0.67055100709748505228, 0.30004248332188254711, 0.0032357703547654529178, -0.043750716514834010751, -0.026375889294007946045, 0.068046679760753736588, 0.14883201417065539118, 0.054344710594487595501, -0.31361664803640237897, -0.31767492120526485966, 0.5937521959118891024, -0.31225468739975786869, 0.5828006237209035767, -0.37129306312933424428, -0.34780926043383519897, 0.28299364040692098765, 0.043644154906956458828, 0.013445203995521060741, -0.0049852588651280865806, -0.023489372164768308054, -0.095656097934244191383, -0.1791387261144138654, -0.076730315229813605971, 0.56675565355360801867, 0.69772633918479687321, -1.8784277330337024736, 0.71485618468058687025, 0.55642298078245189252, -0.12806113639678048943, -0.20986266906137956978, 0.029801799891261318565, 0.016089718119517898431, 0.011407768481739688493, -0.0016609682886214310882, -0.079665475364558271942, -0.18091097204516032315, -0.12681445156533266427, 0.44875827302164394972, 0.69871802505761748669, -1.6338875697687944566, 0.70976536913067733714, 0.43496554104685603015, -0.16293144798281697372, -0.16776826405800782105, 0.093447167394252197981, 0.12653322003071626467, 0.11839540622712421691, 0.017770048426913050144, -0.24303363226449375967, -0.56457110777243879607, -0.63711296443134357581, 0.11476496092978077934, 1.1529625008600590252, -0.46510669364375656354, 1.1715483573913623161, 0.1223937749649720852, -0.61684605295224947952, -0.41722447615151131428, 0.073728267461683510975, 0.1051317286168998999, 0.10351516062738039259, 0.024108004510286454303, -0.19470249776637987837, -0.46927733940706678473, -0.56268918397227474237, 0.047931420474546591881, 1.0423263665731616889, -0.43131471393648213031, 1.0561306103557197567, 0.051163274059640058422, -0.540868624375761331, -0.32710542207514753388, -0.055375946121044683501, 0.030972281115485365138, 0.09425567373857363096, 0.099962583836886817479, -0.0049009358569326458635, -0.18697816133242575742, -0.32233853829879205444, 0.046710495133478385776, 0.71667305611976450663, -0.87096412289795044348, 0.69963731625526082158, 0.021092488127716128854, -0.29324646960708716792, 0.018965734678368192934, -0.038532871316270499162, -0.030073095211003375571, 0.0037982675711479511188, 0.067686178021694617946, 0.11642769919884952967, 0.042232349635947026567, -0.23401733583272107952, -0.045590493286091764158, 0.61031710974576092177, -0.96024384716056121469, 0.59124224274020309444, -0.089943309589205572285, -0.25520652151917117001, 0.22495087856054066244, -0.056211533081548269164, 0.19309985927708231412, 0.39660637458120118204, 0.4833184985996872296, 0.37774208329686204122, 0.0080508307532439742776, -0.71096063817004950813, -2.2305342975890196655, -5.1487655207180624828, -9.8912888463225652202, 32.416810471126531468, -9.9690212721275148766, -4.9059961910943732377, -1.0653722272291432649, -0.011486452271352125959, 0.0078412583599019678587, 0.031248052319362087242, 0.039346703982975059222, 0.012637894300567066239, -0.045495147071410745043, -0.076109620251482740239, -0.15210208071680311748, -0.077778405658679217183, 0.52783545567761847028, -0.55000571847464752029, 0.57563038916322251115, -0.096733358726847712417, -0.1891213656684749278, -0.00051765265226220177983, 0.0018883693110463657325, 0.0023398313393734938327, 0.0033875325898786840494, 0.020649142292100179652, 0.046210828213980606471, -0.0038898283537768412778, -0.25710682690339742296, -0.096656671739138610988, 0.64099809278255026612, -0.71544369214976888394, 0.62165733234844144928, -0.17245059865625014806, -0.093987714753390586009, 0.068029115136234941108, -0.012047355949152675769, -0.099564273585627763308, -0.16970228717224991599, -0.15339691386461354328, 0.024049464197873095944, 0.36377654689110167086, 0.54755985538462503825, 0.43307782965271979236, 0.10593718212336766804, -1.9956420893362099012, 0.023969274971619675824, 0.29104343247947561668, 0.59680494897265645715, 0.010009565494970056043, 0.028712084651309550259, 0.029671833025256857258, -0.0036165969654349400253, -0.06363735249857835341, -0.10671329697299659911, -0.07443667714686005854, -0.029900494220898150149, 0.31666050850973109787, 0.6156960586167582461, -1.3876400207351744509, 0.64888013963775648119, 0.24233728112939190158, -0.22348521735277956934, -0.011634147647548345739, 0.021681131649596136618, 0.038321201906645806223, 0.015307857762032460805, -0.055057382587591532919, -0.13303970726663397195, -0.087102846732012323416, 0.087104942529689838304, 0.42668902616558113117, 0.48820778277912724175, -1.5562579306500008425, 0.56222049654528205576, 0.41191856222218525829, -0.20581695647516146108, -0.0057409297379599341302, -0.0040612219520079961485, 0.0011498069465101404089, -0.0015867073957645894209, -0.021292397385886438488, -0.05067322878906233008, -0.020087467054606133326, 0.075751915118582022202, 0.41772024359958898065, 0.48481339956782637524, -1.7346654904709917222, 0.53312892127849831336, 0.40938775374769070803, -0.080025255206049444867, 0.063857668013255319828, -0.042801177585417328475, -0.16015572549242348965, -0.26192834017994015561, -0.26369714039203595579, -0.021608858599146699153, 0.61005135694819745407, 1.3279126317526028789, 1.3739630508692053379, -0.36789718035869650725, -4.1841988963784144673, -0.41342802237222242567, 1.2815874965880786895, 1.1023691002526678773, 0.060411723299346110905, -0.01371106667450207317, -0.10661936252454960739, -0.2061519870997970405, -0.24632321257822553662, -0.091681455864875457951, 0.41887923521990189091, 1.0517826440951594247, 1.2337482705575000264, -0.10810042777416394022, -3.6895396098822130782, -0.1276070301996285461, 1.131947873285846784, 0.72872885500075379461, -0.048873678774265898184, -0.11740145682526179072, -0.1426986005570597682, -0.096996746018271060019, 0.034266125413466776151, 0.22334862840243285831, 0.47201616572375526548, 0.6972112999062100025, 0.8027112713367776653, -0.1343455591813133565, -3.3750136389230864786, -0.12174595367570548166, 0.90548916941165558114, 0.92132922185827592632, -0.029641168995902626715, -0.098790124779043142489, -0.13181505693865136486, -0.10242296102762643573, 0.0076809296910253788138, 0.18352935249585156097, 0.43617331838877182459, 0.6527706841104319313, 0.72081129994639048242, -0.0848570467295486075, -3.0744219411503537209, -0.083149551463088847147, 0.78666202580043631531, 0.83670076138885762784, -0.0010195332900998117787, 0.010562768541619421345, 0.024630522148195617416, 0.044991839609509878817, 0.063514626551576316982, 0.01175956386888736202, -0.16299455921433908534, -0.41639785173131149687, -0.16291477568994250369, 0.77956255035773547313, -0.42242229710659845443, 0.77439575968053886434, -0.20008885526168238012, -0.35237283983422651445, -0.0023671030055274610432, 0.011800553903166770622, 0.025466794629950426088, 0.040369329327605056601, 0.052125100484144676338, 0.0089374311459479646724, -0.13174860510544500469, -0.37684158824113539454, -0.20758471944851761282, 0.72652517116826387245, -0.32894962129566801545, 0.72277202519546313031, -0.25089158871607258883, -0.29771643165510180973, -0.01633079345511323846, -0.0016987324889562900598, 0.0053621975644254065568, 0.0044152721838276376073, 0.014460673014150660087, 0.041791614527872972262, 0.054601445851337152004, -0.12555420917633650091, -0.18813140364241232039, 0.48535823838911928974, -0.56420551310836597469, 0.48064319909689051658, -0.23730060872182012766, 0.044269385786649406711, -0.062092026704427524808, -0.1017988547194202259, -0.11189984456820192826, -0.079490041774653921736, 0.012416653176513487572, 0.18104437425796249261, 0.46895210346097404575, 0.69708447093122705152, 0.65376057858481018137, -0.16377883301300183438, -2.9985841409106370747, -0.12778468558928221532, 0.74576388790834924336, 0.90360785762183759573, 0.020260035450114649569, -0.020776594717292408421, -0.046793465039201093125, -0.024705727823196473036, 0.068046679616275709823, 0.1494789109127496185, 0.061730826574675587426, -0.29700979345291966816, -0.3028690524446421839, 0.58280062347136152301, -0.37571244122706104474, 0.49867339449343467406, -0.37584418395652985367, 0.060407975294294515012, 0.027980562555676460196, 0.018351389182536904932, 0.0048771264877314262839, -0.006565248702293692909, -0.023489372410314790507, -0.095882441436007173619, -0.1784004947736586022, -0.070910814780354450293, 0.58008313168526659531, 0.71485618534500938726, -1.8729602650528720531, 0.72839514064823340789, 0.54597746333392949136, -0.36906164783712019917, 0.016519662070168215917, 0.014388494145762171367, 0.011520070197454851391, 0.010857786046997360344, -0.0016609687378069185049, -0.080230245439914776862, -0.18128985052319351867, -0.12404035917493400065, 0.45646698541673874638, 0.70976536969402170829, -1.6287346304287966259, 0.73201458266083174742, 0.43530767629286193365, -0.37032573463013240822, 0.026303905272027347517, 0.076083294743449408903, 0.11321970481595072799, 0.11172196476561448153, 0.017770048592707518575, -0.2391389439883086776, -0.56086304945162257507, -0.63549469941777836368, 0.11784851698328878411, 1.1715483571971034849, -0.42075620798628254882, 1.2435382727701966843, 0.073922666306900280953, -1.109995251401520111, 0.01846404698340758932, 0.061690361605340741924, 0.095649675692142985262, 0.09867546529268533817, 0.02410800442334340779, -0.19189870628026531318, -0.46680883083124707911, -0.5620238333535148012, 0.049346352951613631221, 1.0561306105339165473, -0.39584043147888142711, 1.1184482060175642282, 0.010302252904602753406, -0.92913536077719593997, -0.053208466841467259856, -0.0074947128745466830638, 0.048858318076388337903, 0.097785736965827832479, 0.09996258332327956353, -0.0061625337884157410998, -0.19463569570211353144, -0.34244430636686684633, 0.016767282557248439773, 0.6996373166293149426, -0.84861286251777545697, 0.78017584495604952455, 0.07597692581542893342, -0.37827702954568298566, -0.012616451387763145259, -0.028233475209630102404, -0.025737772086409219097, 0.0045392149621054965827, 0.067686178485712622521, 0.11815083666479067603, 0.045817086167062960611, -0.23278320128337764983, -0.054497268562975771655, 0.59124224175719386487, -0.98571423127475532588, 0.57373965530171566751, -0.08526231120894675608, 0.02095054039467662832, -0.15180700204421854194, 0.029815619970295269281, 0.22478402706459030269, 0.39660637523205860555, 0.47724956092347897707, 0.39108028587527227105, 0.064927000849133068017, -0.59221372964873308931, -2.0533987492489513826, -4.9992453467634447861, -9.9690212721754498659, 31.834481171231882257, -10.789071431765163922, -4.8889148389604724088, -0.0084340568974568819183, -0.0050919937412442738134, 0.0090623617493410454915, 0.031248052226939455195, 0.04078275386755669113, 0.015226718560909881239, -0.046112054768125576576, -0.084278130774917692225, -0.15815322076282811725, -0.06614313186066685113, 0.57563038884840866949, -0.52798607772052064124, 0.52386948059489468754, -0.29901639204939295391, -0.0034476795851514788552, 0.0015740763204874061356, 0.0035439393153935483741, 0.0023398313381772705907, 0.0018128842165158248669, 0.019356395991411829566, 0.048638988579261042966, 0.003021497755375513166, -0.25103090927965693302, -0.10730222209710342518, 0.62165733218894592138, -0.72731819392807250946, 0.75244751940787502065, -0.36672201357622385975, 0.05386375423662188161, 0.021254043612542308783, -0.030513994118601291128, -0.09956427319535414433, -0.16160501720619507848, -0.14488753059775139054, 0.030380311321473105202, 0.36954521030934511794, 0.55330793596484928454, 0.41061637938396772229, 0.023969274850065121207, -2.1441781186489503597, 0.092086979122728257474, 1.0389353059339754282, -0.0044478357535976337581, 0.012335718437198260958, 0.02767436003487240212, 0.029671832730844671655, -0.0016311523669291000307, -0.062012209865724809976, -0.10812554377562669239, -0.076708194786265004872, -0.020054166892903201036, 0.34388381241989257608, 0.64888014006378369469, -1.4280090958620483832, 0.61708914155937555357, 0.032260793288478410712, -0.012251135273498837963, 0.0023811703563393843766, 0.023695502839866884603, 0.038321201721096063986, 0.016187524232523585466, -0.057693117766163440341, -0.14457847649924346345, -0.10350536141475512319, 0.089039153651777919185, 0.4666928716391892662, 0.56222049672429363909, -1.558949879114189585, 0.46615066818003236815, 0.21994393433565309626, 0.0030573319741023013627, -0.0057034939951518147036, -0.0058669081085844969684, 0.0011498068708055597838, 0.00078158046436237660261, -0.019915603664982777649, -0.056280565429189258198, -0.036124630252108727579, 0.060959401749738240106, 0.42354628974722097201, 0.53312892119397281565, -1.7052561823989209611, 0.56040337419114916173, 0.24894953674618902628, 0.06994813765548894835, 0.014607635674916219157, -0.062352128861635616741, -0.16015572532554528395, -0.25612753561348083897, -0.2653137273076148217, -0.036998865973342544289, 0.58764006410130109614, 1.3189075812891810635, 1.3725925248100709908, -0.41342802304973236271, -4.3663192316844599006, -0.3654654494870902548, 2.5759202933000007185, 0.052014850230480906557, 0.021755713433613221158, -0.029346426591831717423, -0.1066193627553469192, -0.20092131669874113697, -0.2462004795695355297, -0.10198262678900413269, 0.40394251581969792264, 1.0521994665642258049, 1.2479825470024050738, -0.12760703011512611327, -3.8548802004564231005, -0.14461006489611130665, 2.0520461354170183021, 0.029698284543443888989, -0.050628205461342529192, -0.11620336993761594402, -0.14269860065517012138, -0.10310481441326660446, 0.016917837923165406772, 0.19098418259387822671, 0.42283842052594933225, 0.63878531643468161594, 0.75307272384486989569, -0.12174595319782581526, -3.2838288077057047332, 0.20303099161582568999, 1.5369934539881509039, 0.032827291642933006943, -0.038743792117232681016, -0.1007232319855393271, -0.13181505682491634346, -0.10604155307629339888, -0.0047402070038255750273, 0.15863015352428908034, 0.39660170418323215102, 0.60386893104766037332, 0.67481145600634828696, -0.08314955104479739012, -3.015841500820136023, 0.1872456008513037895, 1.4078261374760860924, -0.011648519402655132735, 0.0038650581274287183937, 0.014683338143076528695, 0.024630521983083991655, 0.042482584735524124564, 0.0643271244203546394, 0.021617729928435264891, -0.14834107129854762208, -0.41313085460329207477, -0.18824519400290398718, 0.77439575929848902813, -0.35637326752109982353, 0.86877866032594230905, -0.69739281664294650032, -0.012023146006739223943, 0.0037468081787871343122, 0.015739487503252738015, 0.025466794558692981698, 0.038109084630511595726, 0.052281273635001436484, 0.016499506590041984611, -0.11898852361948654421, -0.37066789742552214149, -0.22502133772339136253, 0.72277202494053816029, -0.28385910597342656203, 0.78833650901308571779, -0.65284687271276287124, -0.010679189730143066189, -0.0037840884608851595855, 0.0030702064676756818384, 0.0053621974967133691908, 0.00077824013223665839366, 0.00753807864197307987, 0.034647358238327333191, 0.053613208274977640355, -0.11486202018351651666, -0.18240719504605079626, 0.48064319991330944903, -0.58515737346221996518, 0.52283278301022761969, -0.21686046324143290476, 0.015546385040323797491, -0.049784819083709387633, -0.096774649870625104042, -0.1118998446684857373, -0.08701046512568759439, -0.0089946376276643525555, 0.14101300375113398622, 0.4150011866769811042, 0.6518537009846642194, 0.63930886653892959881, -0.12778468603916950941, -2.9532985000021518829, 0.085274361648262436697, 1.4619154633312489988, 0.014664853432427746652, 0.0028639242926634207384, -0.023022183251226610579, -0.046793465073404441612, -0.026375889021295256809, 0.070574098173868296535, 0.16541511569478339894, 0.08587196480142911259, -0.29746666910902680447, -0.37129306288666796831, 0.49867339443490577011, -0.3440778149913184536, 0.73709815533937605991, -0.47160030777309225325, 0.0064242674985780894026, 0.017024566048648766964, 0.015497007339937253836, 0.0048771259961166261998, -0.004985258490957739097, -0.018244732994103229906, -0.084850455233599880733, -0.16819964905281231671, -0.075483079756282406292, 0.55642297996868361754, 0.72839514111761494153, -1.8084504788007618181, 0.80336520579435244382, 0.041517090607202496166, 0.00096906596962996049017, 0.011340226412262467345, 0.013697187913173168583, 0.011520069954735725343, 0.011407768852767483786, 0.0020035273063198754534, -0.071470239099420990181, -0.17392930935768785572, -0.1304720713276619759, 0.43496554066940551486, 0.73201458305359745626, -1.5533876589665667201, 0.79934735001697221346, -0.078366127278452560323, -0.020716535909921101472, 0.027389033030626643711, 0.074707530227035440218, 0.11321970504945963465, 0.11839540635290421444, 0.034757182459689617282, -0.21414884620767504764, -0.54007921451157780979, -0.62906348617080898311, 0.12239377486830683994, 1.2435382727199622011, -0.31575457247217453283, 1.129045354150816749, -1.120650389671309588, -0.019245429407247546111, 0.021295654264581000137, 0.06156906916809763014, 0.095649675595747496448, 0.10351516093213913861, 0.037427710386829529665, -0.17141655713065237499, -0.44944464207763135111, -0.55702210999388224177, 0.051163273606536441296, 1.1184482058275186933, -0.30251994266026349667, 1.0267744653681707412, -0.99836022611401431526, -0.028767447477479932139, -0.023359226260170617984, 0.002527405464466179856, 0.048858318017549771939, 0.094255673272472931257, 0.094057231753159817322, -0.017159902068594540953, -0.21356736841938386839, -0.36075087765672297913, 0.021092487883091686623, 0.78017584537640716214, -0.71618979512493874395, 0.8356512291908999579, -0.5262963989695771394, 0.0046744783970617464694, -0.014648002721360721401, -0.028040768960357152079, -0.02573777152175714053, 0.003798267624617949513, 0.067314981329141015953, 0.11883442712574339428, 0.042977258600784239151, -0.2481452091053825415, -0.089943308766544133315, 0.57373965431678297655, -0.93803668686825747436, 0.80007948159954356271, -0.27610330630204577007, -0.16408085483167317298, -0.088913199025270839293, 0.029815622233996544321, 0.19309985804389884212, 0.37592663159913936033, 0.51714551472014158584, 0.53972055100960114604, 0.35764969312830052761, -0.16721034103456580699, -1.6084627546615803961, -4.9059961906781248686, -10.789071431936912759, 29.510511114309359471, -13.576253433742749621, -0.0045828276356288831159, -0.0074172921773623477076, -0.0050919937485798668839, 0.0078412581163820865154, 0.030314350164347834277, 0.040866718608133643931, 0.015931200261879732705, -0.037574207523992729507, -0.057618182358183805958, -0.1178230841810937457, -0.096733358843814259975, 0.52386948067482685953, -0.46144923524922920599, 0.18057305381330013283, -0.0035478681890107940239, -0.00058852160070094097343, 0.0015740762887851218917, 0.0018883692395032069655, 0.00061699814591244239725, 0.0017336710538797337876, 0.022501379408977950636, 0.059587109117167059991, 0.01579339576836190015, -0.26817398302662570631, -0.17245059894043140991, 0.75244751970599388358, -0.55922556963695890797, 0.15317022401343866056, 0.021062220023279957376, 0.029357212121289968237, 0.021254044013820583281, -0.012047355686084679563, -0.073759209459584765267, -0.14533895319126297063, -0.15755289622574927821, -0.021180518971441329029, 0.27944194535396649215, 0.44329930874237272187, 0.29104343194586068666, 0.092086979046379524583, -1.6669203107563477229, 0.87438709511077261638, -0.0114566328623306804, -0.0018591343615492520423, 0.012335717773168761169, 0.028712084869118591895, 0.034754843021160018723, 0.0081707279980334727304, -0.049020498411372272207, -0.090141772281737031491, -0.059775185633266093854, -0.034649479293132418034, 0.24233728078710214948, 0.6170891419312289905, -1.423884189530526978, 0.74432871208969497712, -0.0051694925991022577447, -0.0067939356031067200253, 0.0023811702791888722693, 0.021681131761857153073, 0.036209676540679101853, 0.015447973775909345714, -0.054553370500771108942, -0.13192503837330984973, -0.082745331452815137907, 0.10660674412728037008, 0.41191856195991077838, 0.46615066845274927276, -1.6413901488906583737, 0.86923964438858869475, 0.0071802844092631965683, -0.00085988932980881282141, -0.0057034938632182802937, -0.0040612220545468511596, 0.0022110944255404763742, -0.0020682983337701951743, -0.029305387930719954681, -0.067823219339994175736, -0.04330232215628537773, 0.065262227663530403277, 0.40938775367421342821, 0.56040337431905340448, -1.7375422814902130586, 0.84000014057283889368, 0.049269630158661023578, 0.04192340967833906723, 0.014607635199596421746, -0.042801177300029941142, -0.13699152338750486546, -0.25295782515904180476, -0.30936478757457180899, -0.14302043317687695989, 0.43327694013921769223, 1.1798639587930381811, 1.2815874973528338376, -0.36546544993892349229, -3.9378338551674345602, 2.1076300309411690748, 0.02848431428967554413, 0.032560867316035649321, 0.021755712657295873946, -0.01371106627955076887, -0.084579617598520243082, -0.18927775281888822856, -0.26795039638224726897, -0.17250326357909567809, 0.28985045607778692212, 0.93492877215731240437, 1.1319478732788610387, -0.14461006478867363678, -3.5670407706552773597, 1.9499300188118282584, 0.074803883897363962485, 0.016414273381208033947, -0.05062820598175797604, -0.11740145687107576833, -0.16214620564699105953, -0.1546518434815222498, -0.07183036171269907566, 0.076368700057538652093, 0.31688713936872681121, 0.62883646834978101925, 0.90548916903676057721, 0.20303099173782163112, -3.041661376534828154, 1.2632233879672782884, 0.064526368783080823022, 0.017836654297663068203, -0.038743792639086255258, -0.098790124886672200932, -0.14365041498087227634, -0.14543996221056748608, -0.079170189836538118477, 0.055832599213996476906, 0.29478056634406574199, 0.58283687985885523908, 0.78666202573378096741, 0.18724560102249912519, -2.74960118742490911, 1.1694382579201296757, -0.014773896677002008176, -0.0046935080568331882106, 0.0038650584123547408397, 0.010562768558149977782, 0.021769509285647775654, 0.048020251514125755765, 0.082922014890841819201, 0.049709152265028103135, -0.13243813891356351342, -0.41719689325426878579, -0.200088854608734934, 0.86877866013558235903, -0.35170556893691473643, 0.05685579444172162572, -0.013926744308178844872, -0.0050789577729102239334, 0.0037468083346371395224, 0.011800553847836804319, 0.022479331613527910733, 0.04247395611311221425, 0.068782121717133035133, 0.04378336658829076744, -0.099061879911388026065, -0.36949799414511674822, -0.25089158837757058462, 0.78833650868054627736, -0.22558743267282974077, 0.0025665878596435910686, 0.00031150425118409261378, -0.0036219877064378649262, -0.0037840888709935347964, -0.0016987327328628566616, -0.0026551879118263304116, -0.0082169638578228447467, 0.0027872697312827394295, 0.043787428570592298871, 0.072852660378808331565, -0.10788271396012220904, -0.23730060959289883682, 0.52283278378227138194, -0.32794487842118813514, 0.044885820483047768337, 0.067989119665279484916, 0.0096408610301089211342, -0.049784819478600063347, -0.10179885423250407961, -0.13586541273693050758, -0.14139131354715112066, -0.095280895115830252506, 0.041987320026906554959, 0.343220904537466176, 0.67055100688029911637, 0.74576388842519325539, 0.085274361268958001614, -2.6454185981947726525, 1.0974348819478365602, 0.0029076018815180557087, 0.0098488229963336355183, 0.0028639245494263576054, -0.020776594769747115665, -0.043750716773860741371, -0.022416709903405454551, 0.073974547734904055885, 0.15987411456469491355, 0.051953352174026308508, -0.34780926049973215353, -0.3758441838213909536, 0.73709815537680711817, -0.060919139361382240727, -0.16928211456718306183, -0.011354832128000893338, 0.0056603848436464392416, 0.017024565542201840901, 0.018351389779094520738, 0.013445203852884068246, 0.012494148734080787536, 0.0052991845783243798906, -0.07284986555943064046, -0.19815683391603688968, -0.12806113633781732175, 0.54597746315355288882, 0.80336520563127544747, -2.0320068974060583322, 1.0409125948902229197, -0.010428101582865945129, 0.0022359136837194217538, 0.011340226197103986036, 0.014388494487096564409, 0.016089718085253897073, 0.024051823658950387264, 0.021909126894199933838, -0.058319354860072183944, -0.1929814565925287162, -0.16293144783730245662, 0.43530767589403507323, 0.79934735016850899658, -1.7555349652100940006, 0.8738020036921028133, -0.049319894670016940774, -0.014769208141279297972, 0.027389033263773041732, 0.076083294764065903526, 0.12653321946844145462, 0.15490436125277218249, 0.098153086547560386421, -0.13789906178373753654, -0.48761141990499690113, -0.61684605255118041267, 0.073922666100552161805, 1.1290453541670508741, -0.7712394744495780774, 0.47823369983141744699, -0.041118059647299254333, -0.013136688400504874588, 0.021295654422286963009, 0.061690361789518435853, 0.10513172817754391775, 0.13252287434212611794, 0.089732676519732393339, -0.10712697224442337129, -0.40338958121979595628, -0.54086862384287748284, 0.010302253058550332049, 1.0267744648917878081, -0.64823810085073196863, 0.37795345237871846855, 0.00027322435287575739431, -0.018154899172455871148, -0.023359226430088461668, -0.0074947130281814251448, 0.030972282026447195413, 0.075010641570544142609, 0.085406638497919032393, -0.00024212403141645155421, -0.16940928751866082824, -0.29324646997773629042, 0.075976925790743596312, 0.83565122949007508701, -0.87589767364856685461, 0.28715199800402957386, 0.012899427644300558854, 0.00044143358006918943228, -0.01464800185321753212, -0.02823347576198992348, -0.030073095933710984606, -0.0071250479201419528869, 0.0505780143059472026, 0.10470867885461394853, 0.035761621022800407776, -0.25520652138967803157, -0.085262310508420258159, 0.80007948105713466713, -0.7810989025048580281, 0.18325389389443566546, -0.12532656272983042078, -0.16408085457737608959, -0.15180700341228287042, -0.056211532412120061852, 0.14941470834468920104, 0.45990745238723407251, 0.80539064571153273775, 1.0442087770179917516, 0.99255218089009422666, 0.45501148798934643036, -1.065372227403374783, -4.8889148388065883921, -13.576253433831960038, 16.867977768458825238, -0.0020135074324507205162, -0.0045828276273721180525, -0.0084340569298706948714, -0.011486452106520193134, -0.0057035359292295986777, 0.019989156732255247001, 0.060691898203442229409, 0.081782202409112311758, 0.039388378965726819625, -0.051991628769167581858, -0.18912136554170624891, -0.29901639204382302051, 0.1805730536763339722, 0.21268369478005624829, -0.001139703694122335938, -0.0035478681013130716759, -0.0034476797252162632787, -0.00051765250245084856989, 0.0034043241398166276986, 0.0048207493574179008725, 0.0022865904200690960642, 0.010424439936647298538, 0.055407227039118794876, 0.096365054526867069407, -0.093987714635267546881, -0.36672201358523465187, 0.15317022395069557206, 0.15663971545221397341, -0.01795457396172756015, 0.0210622201376926034, 0.053863753979129194716, 0.068029114894893535914, 0.044421183657796882438, -0.037862415665792306663, -0.18317467330584805141, -0.33453925374441462193, -0.34083262542478676149, -0.0079965523445208533393, 0.59680494933338279484, 1.0389353060854418231, 0.8743870950872629777, -1.8658146607577417164, -0.01424349759455913883, -0.011456632972959859829, -0.0044478354100580246716, 0.010009565442047126116, 0.034748222438168260839, 0.06332672569933367468, 0.067357712154045443209, 0.0098881751936236855771, -0.1063700330022582935, -0.20144930359069593284, -0.22348521719768393856, 0.032260792966947927918, 0.74432871226790153774, -0.38767806935937731971, 0.0026314721454706773221, -0.0051694925864240529045, -0.012251135212636066543, -0.011634147737523081734, 0.0075214411858510275138, 0.046496013227990018457, 0.075781105840124221573, 0.037570811898987824251, -0.10297593280268786675, -0.24801875398989736698, -0.20581695631991456241, 0.21994393421074343031, 0.86923964436275180656, -0.68079024638579321849, 0.0061413168064748122413, 0.0071802844546690752781, 0.0030573319617377072051, -0.0057409297857165166729, -0.013119362279515081951, -0.010239737747235341803, 0.00021690341219400346442, -0.0040179085619152085304, -0.049349762351162132501, -0.10301149163078553161, -0.080025255059531480439, 0.24894953656281135213, 0.84000014065150574449, -0.87058446243133702502, 0.01563513021159520372, 0.04926963011224691491, 0.069948137927478545861, 0.063857667894993669355, 0.01131354846309530135, -0.11069916913577619588, -0.31375860128576943398, -0.54747608119971702401, -0.62135409674252239931, -0.1544290133554257205, 1.1023691000281112817, 2.5759202935366447562, 2.1076300309914972608, -4.4867845729027919788, -0.0004997482534924846528, 0.028484314179947553547, 0.052014850678618974877, 0.060411723251195793838, 0.037326363620490708717, -0.041240006986697733127, -0.19955162275227056323, -0.41895274441872837334, -0.55455426419272135785, -0.26213681283252288079, 0.72872885502972994942, 2.0520461352431169644, 1.9499300188397654665, -3.6045719345878617368, 0.093443155817117845574, 0.07480388383347857939, 0.029698284831679258589, -0.048873678774800630165, -0.15734323187913781816, -0.2697536220316995581, -0.33152756957658224835, -0.27173464921576545583, -0.047445517674162232535, 0.35802613193912236067, 0.92132922182005827505, 1.5369934539155132303, 1.263223387955324295, -3.3935844868971263288, 0.073403239073982676222, 0.064526368659105381997, 0.032827291953954187975, -0.029641169040059586054, -0.12297682956112229724, -0.22797942414088293028, -0.29834742080440707079, -0.2680155035735113378, -0.085878052641917418386, 0.28892633596758454662, 0.83670076133919846217, 1.407826137438280556, 1.1694382578901807435, -3.0526377733381324653, -0.010857930122841065457, -0.014773896591882849927, -0.011648519513605064268, -0.0010195332662829255984, 0.013375578290873997578, 0.029109588428674747218, 0.054629623926537231038, 0.10511949833156679357, 0.14455098396559024065, 0.042696607137057307479, -0.35237284003226643225, -0.69739281668777142187, 0.056855794488715555302, 0.69266623710248564905, -0.0094028344279996223937, -0.013926744321450582748, -0.012023146011895004259, -0.0023671030062495375217, 0.012820586912015992376, 0.030704114023397774402, 0.054632995162660527577, 0.093977394778797407326, 0.12218323945854610291, 0.041642013304576203092, -0.29771643166991035301, -0.65284687260803131448, 0.0025665877804919470621, 0.6776749301679433346, 0.012756049533694507198, 0.00031150414177653738693, -0.010679189568681524136, -0.016330793199109884528, -0.013411696330302704319, -0.0048309328562686016176, -0.00039719101601221593924, 0.0028875816247325857174, 0.03140598086685288115, 0.098923521993393465079, 0.04426938612685151625, -0.21686046377998013934, 0.044885820710984777882, 0.021809886002658204068, 0.096545482289634373951, 0.067989119721600835278, 0.015546385257754282566, -0.062092026895260714336, -0.15399983864895483432, -0.23549572848214533893, -0.27517309148870522373, -0.23775913036314821891, -0.077273425350704458281, 0.30004248354162937895, 0.90360785743993476959, 1.4619154634568682916, 1.0974348817993744287, -3.1291375704203039376, -0.00691361972115957929, 0.0029076019059738043328, 0.014664853348224361068, 0.020260035385327175572, 0.0032357705375663154725, -0.04439910418307531359, -0.089028138412598933904, -0.045902435523333673384, 0.127215625470034982, 0.28299364038107127639, 0.060407975216806748808, -0.47160030788523626955, -0.16928211441058704789, 0.32178076894780799355, -0.022577473953307936744, -0.011354832339763499052, 0.0064242679091165164409, 0.027980562209968092013, 0.043644155013641119933, 0.043533002205148406405, 0.03035367896993182088, 0.01311908345708773585, -0.046509793485393233214, -0.20986266910356818349, -0.3690616476605173002, 0.041517090668558909172, 1.0409125948242563542, -0.57582713402347474041, -0.016193746149866340006, -0.010428101594179254794, 0.00096906611644964244477, 0.016519661925469275587, 0.029801799849294305367, 0.035472518262058258531, 0.038448413678117014658, 0.042630585927406393354, -0.0019161511873433218975, -0.16776826395160313043, -0.3703257346259253846, -0.078366127313591465997, 0.87380200367877070011, -0.37660767824220348077, -0.063577847155227740217, -0.049319894648085449429, -0.020716536115915284988, 0.026303905249523595183, 0.093447167659854210586, 0.17834288856949492841, 0.26393145189035466247, 0.28813942169876649579, 0.1042252852363018123, -0.41722447635555043322, -1.1099952514511670643, -1.1206503893331132282, 0.478233699699860626, 1.5096649803777049126, -0.050818138077760935389, -0.041118059642595197745, -0.019245429500954900487, 0.018464046874985411834, 0.073728267628768814901, 0.14544168528096734772, 0.22103695344805704415, 0.24958334053745157011, 0.10478782621226766236, -0.32710542231807610003, -0.92913536084630266032, -0.99836022586028494441, 0.37795345241416994364, 1.3097911124616916645, 0.025807827909793395332, 0.00027322431217551311272, -0.028767447385906391721, -0.05320846680170662063, -0.055375946417224326812, -0.013423427888093352259, 0.079183024315557362649, 0.18715382138370145748, 0.2067334613025380019, 0.018965734991393444486, -0.37827702975097177163, -0.52629639912826320369, 0.28715199814020081481, 0.26721821467217904411, 0.014842930713995318437, 0.012899427775236662827, 0.0046744779610954478172, -0.012616451167974364334, -0.038532871074522127841, -0.063084530335535585199, -0.062619934285514405881, -0.0015533728383021788884, 0.12924523969007017943, 0.22495087838487703991, 0.020950540175549099697, -0.27610330602174043557, 0.18325389383873355587, -0.15655122629045853389, 0.21268369488920191301, 0.18057305335667900303, -0.29901639205561869606, -0.1891213650973611049, -0.051991628777297176078, 0.039388378671249134333, 0.081782202393318237377, 0.06069189816947408167, 0.019989156942623802909, -0.0057035359012350518929, -0.011486452303583912643, -0.0084340568611060671989, -0.0045828275058337874376, -0.0020135075418931769158, 4.1734184928835107442, -7.5214006972329476852, 2.3652509044634317803, 1.3950822258680064536, 0.040622199066780655119, -0.26632498885846644932, -0.14701071350371136548, -0.028866757922750983095, 0.013465698381658307292, 0.016226667005320199316, 0.0079094756128653825877, -4.7660196481320011674e-05, -0.0029977572774080312075, -0.00084784280016691831137, -0.4119638257067586018, 0.19610897155072823383, 0.61216027082775570367, -0.054850161051397755552, -0.44319265014231801203, -0.09305260267968233634, 0.1335783481298053188, 0.081648002284034001952, -0.0011569208376189198378, -0.022274900619898629212, -0.010821605648926225021, 0.00044788952630658851461, 0.0033076569012372336771, 0.00036186574608634088004, -0.98530954744256415978, 1.3311548115709841422, 0.065995205068065818099, -0.38186019307377988552, -0.10678842186686353377, 0.030922975685806191465, 0.027057589791600809154, 0.00018500441531576355698, -0.0066319588029764191608, -0.0018364802534435288942, 0.0030589071316546511889, 0.0043365716945122749493, 0.0025126918089693204075, -0.0012623151733168937877, 0.46783129637050996363, -0.49195740461590309334, -0.26607832585983159923, 0.15941423453460054627, 0.16653397066819886074, 0.054013649346276886021, -0.0358149513921188975, -0.045521868798203975193, -0.013641731428250879893, 0.0059965266188850433263, 0.0075100620417298171588, 0.0030223875823058446791, -0.00035083945258054837543, -0.0016258417844869112439, -1.1626053692057618516, 1.6842684178666487149, 0.13639916330012630263, -0.63959560186002595294, -0.23412285612396285805, 0.082491166301937840499, 0.10264389329540160589, 0.037854134666380441498, -0.0031938573638279495713, -0.013376318381554237746, -0.0080687394354065772317, -0.0010138434380097026757, 0.0020675038255173424674, 0.0010418841921978247954, 1.0630247804740151629, -1.3042134499832609418, -0.49364030177611195604, 0.56563473613724279954, 0.37872116402662397983, -0.013933815251888869574, -0.13463130134386383507, -0.077885654058357317586, -0.00010767921721594371509, 0.025361986210584074042, 0.013611970216997717684, -0.00099389324673036013452, -0.0047803783769917451707, 1.6521405466427338329e-07, -1.3921413095186374598, 1.553213263375198494, 0.46972246587936311579, -0.3597741213040103192, -0.31740804629128371861, -0.084489851284398370668, 0.041415049610297682836, 0.046864104115278366891, 0.013825249963370858305, -0.0048561495098726169262, -0.0037552241821848428814, 0.00197738137926056377, 0.0033783912382688574549, 0.00014459135431977090946, 0.27291746503143260449, -0.48070430093715788011, -0.11503346552473124598, 0.34265445245739106772, 0.13316670273647251976, -0.064076463272538533578, -0.074367079691593312996, -0.033859072644299943733, -0.0046485435694684270475, 0.0070013925419244787615, 0.0087644498255915071622, 0.0060565027856307870543, 0.0019579275725402117865, -0.0021404012758181665925, -0.025749663507063591528, 0.032651150800982409184, -0.10576602027601933464, 0.082398139760951574773, 0.048596277822587796713, -0.013304003614943419112, -0.010062681705548326763, -0.0024660528168857654557, -0.0049736887517986411167, -0.0069006010948665565813, -0.0038049943322133713067, 5.1512528169638752834e-05, 0.0022039059336066446106, 0.0029337621180405715898, 0.0010532714808167082179, -0.0047904415446136097353, -0.098962278231388089988, 0.093763283419695087062, 0.042602703000851602222, -0.017822067555190737598, -0.0089977013720718007628, -0.00079019623110538651789, -0.0041466322632077862176, -0.0064321785267238860626, -0.0035152536250329046406, 0.00015176556587880387934, 0.0020831866923116237195, 0.0025363760262273279435, -0.60146356229582720321, 1.0249814845119085138, -0.12917147210371671129, -0.36185240887422714939, -0.047312267180161148017, 0.063675374157341216308, 0.043737741040301023054, 0.01132186559078598477, -0.0016796958694212068327, -0.0033846853074675064404, -0.0022993488073181737874, -0.0013164795018685419067, -0.0004973640516334101435, 0.00021975783714437750424, -0.52754921893665540633, 0.93037250528518311565, -0.15165817877284185067, -0.33056589765115179658, -0.027078486108017319617, 0.067082382787448005534, 0.038876569980159529638, 0.0057139570650832872609, -0.0038514486536726671474, -0.0027526769150325687205, -0.0012011500041380271647, -0.00085423102310025955317, -0.00054093283763181258892, 7.9626842962941379173e-05, -0.81145104603154494249, 1.2580254515586555275, -0.027644008726966123879, -0.44026442717718999864, -0.13126733120002170851, 0.058165260498084563057, 0.079327196101068267464, 0.030381072011546383926, -0.007160961173565644905, -0.014523441699434614174, -0.0069626464233904193646, 0.00051416119290200270643, 0.002812058715438149354, 0.0004551265429290838856, 0.047599506554463592856, 0.058615639129592896073, -0.21066235667589403024, 0.033836230033619053847, 0.065889630975424995363, 0.008906528860530938535, 0.0075038858662433236685, 0.008278753748561807832, -0.0022456665496897012038, -0.010004512779597827393, -0.0093029929473647694771, -0.0042651473478594031236, 0.0011400441681132572527, 0.0053934683303582028197, -0.53387460957115440152, 0.95458796723452876698, -0.20586785278797831156, -0.36098211729547480253, 0.047767706280313057543, 0.11438558163168524584, 0.012263796001770115723, -0.030349312860626368599, -0.013867055978623357357, 0.0038298354381399754391, 0.0062297489317394748667, 0.0020528630263968138114, -0.00075430639319433705897, -0.00094720794343241545513, 0.13615443477348826429, -0.086494130986563624863, -0.18109805288404787205, 0.075283061827222566564, 0.081580212649171965356, 0.0077800159745023420368, -0.017870449873264738355, -0.012269213023165010876, -0.0034838369729626914549, 2.0014679246460511785e-05, 0.0018549155644116501786, 0.0022466276679175959698, 0.00093869421181535942292, -0.0012574763500298028786, 0.14966708863760724846, -0.11751239137593293138, -0.16764883210542333436, 0.089337702242235944583, 0.075230609282287808082, 0.0015300528460328007355, -0.017879099513795654319, -0.010539190503967777596, -0.0022363966056471603315, 0.000529963268911541629, 0.001670088293297219105, 0.0017399463647506312018, 0.00061062339741715736773, -0.0010808801153923846584, 0.170080011159788963, -0.14982827594438696828, -0.19408847204389462204, 0.1633903632750437207, 0.093948565216800644651, -0.049389474051820286826, -0.048110867988779032478, -0.0063142984488209208394, 0.01294298615119774537, 0.011953693971425155204, 0.0053054664203931126837, -0.00017291755092259337772, -0.0026215088674531390918, -0.0024951842825611117785, 0.17776031115710075325, -0.17875875526526191495, -0.17801866320402368049, 0.17658502592523422914, 0.08782122539895191593, -0.052987561922575254769, -0.046164753723678120245, -0.0050412120447628095549, 0.012227291619546923687, 0.010733556006342587913, 0.004577850157531758099, -0.00023515602577194849566, -0.0022853005498080650529, -0.0021201045406026340981, -0.24136096190446476584, 0.68222115844700803589, -0.36261323595887967608, -0.29016350641270094446, 0.11237539894809583929, 0.12480395703895286186, 0.013291670554367276263, -0.025012970328458825064, -0.010883189610628786828, 0.00079649131107123612681, 0.00078519560235569790634, -0.0016063179020118470604, -0.0013579212341254035239, 0.0011208262982260570276, 0.1166205099862567679, -0.22042401679695497463, 0.019948424972028089286, 0.11667617625904724843, 0.0082913344986694972188, -0.019029115366911843882, -0.026274214170406571395, -0.011727564321814088616, 0.008726620057515203574, 0.010133735790953090836, 0.001721565948821449565, -0.0033583049745964585475, -0.0026238442483980588087, 0.0013888925972878221442, 0.18057305304368684307, -0.46144923370670398732, 0.52386948034441382926, -0.096733359740561672413, -0.11782308393193477447, -0.057618182246293940529, -0.037574207106950842372, 0.015931199973879907328, 0.040866718846685433364, 0.030314349812077708013, 0.0078412580723074658262, -0.0050919936745380074347, -0.0074172924342620879037, -0.004582827319031401378, -7.5214006971639175703, 18.146760929319519562, -12.505951542056225279, 0.014355783622894689386, 1.5043461080566919996, 0.53970643610495017128, -0.040572057232535853333, -0.1190386920936117171, -0.054252708851728462247, -0.0095423783309478427128, 0.0053148295851008706125, 0.0079612900555882989279, 0.0042194720176754072979, -0.0029977573436453914223, 0.19610897169118976513, 0.11923624079078622151, -0.96136506915608110901, 0.3496077161489162255, 0.72803610891060199783, -0.21123525019416533599, -0.31161386068826324447, -0.0072781370143912447676, 0.08380018222242616166, 0.035719555989323886758, -0.0054429865534699377674, -0.012194495038323427405, -0.0048258615051265090629, 0.0033076568005486954409, 1.3311548114632325568, -2.8190742433012565549, 1.1455204529997240481, 0.77172239971669176484, -0.22997737574987431319, -0.19435650183051156326, -0.032563692980241282227, 0.021349850884761288616, 0.011060054399692476751, -0.0010022059685638527553, -0.0030255137427813833494, -0.00081085771817335043892, 0.0011228890488881447963, 0.0025126918069302253862, -0.4919574045844691268, 0.82481288385654683015, 0.052131819069066817618, -0.47132723849363539204, -0.0740613663339356465, 0.07059041242131565419, 0.097539917736450862384, 0.034980369141551635737, -0.024050997326215701327, -0.023009892753437344304, -0.0047337072028507433397, 0.0039637825735474264618, 0.0034715399810040913769, -0.00035083939444484438805, 1.6842684178452749233, -3.4254308124428063387, 1.1345933569480326497, 1.2170285906953628441, -0.24074777185548196234, -0.38044106383798770299, -0.079950697010682192367, 0.049769810304826474223, 0.044171220410338350726, 0.016248123644249688052, -0.002121739062469454179, -0.0070389923411927016877, -0.0040406927346109920707, 0.0020675037302687066659, -1.3042134499771433909, 2.3030029264671387779, -0.062252586106830445911, -1.3328426175026346634, -0.075044536116210452947, 0.37853436191590683624, 0.18787993133464062079, -0.0086498723302689071307, -0.073999531717442196266, -0.039984603184126228659, 0.0031801971362808924455, 0.015186943259097680434, 0.0068066959601971359708, -0.0047803783818807709868, 1.5532132634445261488, -3.0015644831277725757, 0.72070711117565644521, 1.001055593756046358, 0.095499749071322179161, -0.2127658745996718892, -0.16499680937199701081, -0.039766270369018148767, 0.023755662568840665849, 0.024974751903847312862, 0.0054104817104654095647, -0.0040382563231831890282, -0.0022982986562510975687, 0.0033783912022502019488, -0.4807043009125568922, 0.82983124092835824914, 0.034930384698711361424, -0.6686188209838940244, 0.089752800339003402375, 0.22614666700370045227, 0.036114441558074583671, -0.036330016392274913639, -0.031821559136048054028, -0.012816918113374119181, -0.0014137456557444900637, 0.0033475506026676682923, 0.0038522139659131122459, 0.0019579275440630952883, 0.032651150879405343996, -0.20709296831955378848, 0.35067500075106755597, -0.23216765553971610569, -0.011045960162981266137, 0.08149139528771745522, 0.0043720925692777405647, -0.014847760337680817114, -0.0038056396234623494391, 0.0010486589057701017728, -0.0025361850106090438536, -0.0032227930808237867377, -0.00062790004248291836266, 0.0022039058799747375354, -0.0047904414541627068913, -0.11722086449016200227, 0.3026616295089646913, -0.25566485065461963533, 0.0096417780222608460983, 0.084919626671598219536, -0.0015346423892911714715, -0.015999116665142806143, -0.0025820070208927015099, 0.0017948717672424938964, -0.0021890510478911446267, -0.0030785290976863746021, -0.00067535675189505717849, 0.0020831865852675027564, 1.024981484513547203, -2.2598296807916637974, 1.1202773206190266286, 0.53004554242714774226, -0.2967196000673573697, -0.16072538749041093142, -0.0020078254414939789366, 0.033171300189875976838, 0.017289356736980718288, 0.0034763452220916823271, -0.0012536072486399835006, -0.0014715873480348840106, -0.0010348372831131226121, -0.0004973640715546253201, 0.93037250527598924776, -2.0789118736705956003, 1.0779993214901941734, 0.46063761991364404746, -0.29374783263930587252, -0.14548810645607984648, 0.0069587395448395547928, 0.036083068194798798012, 0.01505355445162469101, 0.00039688937848175915874, -0.0022496839606080875837, -0.00094532991358516028305, -0.00041934183182552696575, -0.00054093278886531643082, 1.2580254515848228181, -2.6720581069993945, 1.1362568319605010903, 0.72147598738837415144, -0.23005164867832464393, -0.2164340006008102224, -0.062924728116323921934, 0.031109752069315411982, 0.039663218632366673, 0.013881742449366711242, -0.0043377311020950607562, -0.007651824452817154211, -0.0036146880234410832915, 0.0028120586930693103681, 0.058615639277030062715, -0.3149737228769802222, 0.51195116226920933666, -0.25454555996515632055, -0.087289902645693251415, 0.078169310418606877122, 0.013371394020407613087, -0.0072688434538590949302, 0.0028805778430203380497, 0.0048688088362267038442, -0.0011609488657173903242, -0.0042248862947859375933, -0.0028691851666235719358, 0.0011400439931427521541, 0.95458796724326344663, -2.1931340500252742842, 1.1759558790663107164, 0.54994853210447380842, -0.42294156178534214607, -0.18017279533699837035, 0.078139646076356122917, 0.064404319508141336104, -0.0041231636330459524656, -0.018835396130030622464, -0.0058387443471813920207, 0.00345759686443020001, 0.0033176087122647980251, -0.00075430632564644660085, -0.086494130978067032545, 0.022945326511820893223, 0.32060460986916161641, -0.28407349801096920539, -0.066485943509556819087, 0.069918840073834059856, 0.036718992843990300168, -0.00067741483961416409931, -0.0094538403511733540979, -0.0031318474865132824546, -0.0011161575682464952141, -0.00053121196641575923331, 0.00052609755254717500798, 0.00093869435555470797164, -0.11751239140856500898, 0.09953299187320334418, 0.27277671679478826539, -0.30382299624880121369, -0.04311292386817425909, 0.075320983428661045234, 0.030834207840912099263, -0.0029334609770135314233, -0.0091297913117537685818, -0.0026639571600481984359, -0.00064175269220823928651, -0.00020599800298144277204, 0.00049384554476454112053, 0.00061062334106709852555, -0.14982827602426174174, 0.1360225057412491334, 0.37709333231323133306, -0.46250093124178570658, -0.053361711216988960693, 0.16507460733076542714, 0.032898962004824477967, -0.033273997806130016508, -0.021357266253608443024, -0.00079955848315001166147, 0.0060562388075056233783, 0.0053180051224129094509, 0.0016931584537415791947, -0.0026215088864942732126, -0.17875875528337664688, 0.20729887703791757825, 0.32631353373995447553, -0.47629652849346493237, -0.027573470731882262358, 0.1681744109231844142, 0.025953710891309211828, -0.034577181660159349297, -0.019928797852038167282, -0.00025657737328508472066, 0.005708527035313709988, 0.0047650404537793802917, 0.0014217682470474726601, -0.0022853002692677206256, 0.6822211584400931228, -1.7176710791471314632, 1.1951867178440105377, 0.27128296334474061302, -0.45836353168438498384, -0.11173079491650728112, 0.10398504888777404531, 0.062652755435874840639, -0.005768089081260783374, -0.015428830778032193508, -0.0037895751984882240088, 0.0018842765793932729924, 0.00090442203170581656965, -0.001357921184678919941, -0.22042401687746168704, 0.46562571859065920998, -0.17480700321507766648, -0.20137948814383629936, 0.1096092889874292764, 0.023152279952882885011, 0.01561488155577430044, 0.0002300753691053393328, -0.021259378919495194177, -0.0078919826205432232202, 0.0056778199150482990334, 0.0059023440403966013351, 0.00094551073753688843343, -0.0026238442625249867347, -0.2990163909975099843, 0.52386947853072773906, -0.52798607663078367391, 0.57563038915009878682, -0.066143132585570502124, -0.15815322001403539498, -0.084278131843835091264, -0.046112053922777042347, 0.015226717936950830146, 0.040782754185042811856, 0.031248052937309277188, 0.0090623612653269736139, -0.0050919935559651631832, -0.0084340571932528026616, 2.3652509044146228234, -12.505951541967910146, 20.796052860825117392, -12.090432980787937822, -0.2709288407471254212, 1.3437254752568665861, 0.5292266709852383455, -0.01752209087519460487, -0.11079547819731873359, -0.055994783195909801632, -0.010069820537919488626, 0.0066965520840258129301, 0.0079612901359610386415, -4.7660149374708173954e-05, 0.61216027053852306583, -0.96136506884413575058, 0.35270233818267693326, -0.34231327433399511007, 0.14452010050545360809, 0.56446407808323828537, -0.17219017929586014071, -0.27637478246681668193, -0.0079167356567543326062, 0.077124945258777125034, 0.033504237639818329475, -0.0060902627790218556367, -0.012194494979914118765, 0.0004478896552695642839, 0.065995205137935927797, 1.1455204529480540465, -2.5040381472032393617, 0.89866069707780082965, 0.76329273690943688635, -0.1838564241709612912, -0.1720925753759453003, -0.031428673917776241142, 0.018714660920176352799, 0.0090954338120175556165, -0.0028441986238673350011, -0.0048045463396578025686, -0.00081085766373944925898, 0.0043365717017063206556, -0.26607832594747321542, 0.05213181919747694737, 0.53563720192167252332, 0.047244374066303843063, -0.48690008516737920408, -0.050216437042389719392, 0.085485185387974588211, 0.08927541426173167538, 0.031980024624026374969, -0.022584224632913405678, -0.022350097352990778293, -0.0050282033277531754806, 0.0039637824951434964121, 0.0030223874954597288768, 0.1363991633447148577, 1.1345933568955055559, -2.7216728207969462794, 0.88404041243067243983, 1.1158128097952777491, -0.19571852483985599513, -0.35834620680214457344, -0.084066817729819928728, 0.047191890853742715639, 0.044020120323388291228, 0.016117348732561515012, -0.0028584081754770896723, -0.0070389922455815915128, -0.0010138434043231983519, -0.49364030176943091144, -0.062252585998814627399, 1.3716735665120995957, -0.02105148144496991891, -1.1918687299804875401, -0.053124762685403392748, 0.37253534933520682237, 0.17065682863819320314, -0.015097197560643409869, -0.070470101610868113307, -0.035866525406319425817, 0.0052188260827873231978, 0.015186943257492950196, -0.00099389326373943158617, 0.46972246561463126824, 0.72070711147014754339, -2.604655731281872999, 0.69250445512880565158, 0.98588822596089986661, 0.097180254565134877032, -0.20185747038452356295, -0.15532796078861593747, -0.036248755712434466147, 0.020437809034882088027, 0.020925519042657457491, 0.0032157447296775752872, -0.0040382564600671922925, 0.001977381533713869384, -0.11503346556889611219, 0.034930384696776457798, 0.2452801958515833658, 0.14351804400088497005, -0.57111478693557149455, 0.068336012785229507949, 0.22142201716093259933, 0.043868197115895131433, -0.031458623086382586642, -0.032857651648730910876, -0.015652762972980828471, -0.0033710369742139170007, 0.0033475507593114565107, 0.006056502697004791623, -0.10576602046584254735, 0.35067500100293624854, -0.47987187949699428025, 0.37500555241302385712, -0.17493399404880216608, -0.031338915231885908219, 0.070353311338850879619, 0.0089262496719523728994, -0.0083526319232592131231, -0.0021085932755941358061, 0.0011681996708022982304, -0.0023934765514193193439, -0.003222792986971224731, 5.1512626824048155828e-05, -0.098962278403973924501, 0.30266162969216503509, -0.40152578887799056373, 0.34811265185592971116, -0.19991067686804964154, -0.015602957090844955454, 0.074245144009414607011, 0.0036798493498053207879, -0.0098364122026977293678, -0.0011862024038245119403, 0.001746304716478380703, -0.0021768160302500674623, -0.0030785291325853619479, 0.00015176570531649086204, -0.12917147209579538103, 1.1202773206820098029, -2.0347952412164675984, 0.88345791357631087148, 0.53832273884834047628, -0.26081316244415986283, -0.15376528369069061841, -0.0079208303052709129466, 0.029498360577288933182, 0.016960589573510872347, 0.0044478897819821604578, -0.00070380320957507498322, -0.0014715876477860857351, -0.0013164794700409869922, -0.1516581787680631177, 1.0779993216182066629, -1.8911425770863117268, 0.848478457225956606, 0.46812753765395914707, -0.25884041094976018016, -0.13660361514022861806, 0.0014378043625638469924, 0.031865314817589245056, 0.014634218864659287798, 0.0015197109526596144606, -0.001632220706258509773, -0.00094533000305014984466, -0.00085423108751558635837, -0.02764400874007202219, 1.1362568318088779318, -2.3377339816765014824, 0.9473366441168189489, 0.70670362669830388835, -0.22523977319532179076, -0.20849442859600691991, -0.058087004124218744039, 0.031788082229291747538, 0.03918459634044602069, 0.013125971837671805351, -0.0056474567571233955793, -0.0076518246325528553597, 0.00051416125917924247919, -0.21066235694000193246, 0.51195116246516392255, -0.59100268635761532554, 0.50337912104289928017, -0.18058149942153542411, -0.11184205063214902598, 0.057891716445252476564, 0.013605511508472542537, -0.0015251107184752164814, 0.006774125541700900914, 0.0075080465108333669169, 0.00024502592585110027391, -0.0042248863671061663919, -0.0042651471791750762003, -0.20586785285825678904, 1.1759558793203261917, -1.9111953359178792677, 0.80605085932924147585, 0.52446933300213616214, -0.3333933600812022946, -0.15424850250774085891, 0.065266873744031075888, 0.05576120936421480323, -0.0043752312001782804607, -0.017133288106927258704, -0.0052745268829934785493, 0.0034575968798714684996, 0.0020528629620712858343, -0.1810980530213887052, 0.32060461038292381586, -0.23687695823021476249, 0.31334451467934842928, -0.2402541639663207107, -0.065845894379011338238, 0.063798712126687390223, 0.035677114820292593622, 0.0029750593964068972747, -0.0090118361441867949474, -0.0045885384492451424671, -0.0021157077666051349318, -0.00053121170198036193454, 0.0022466273743586028834, -0.1676488320027924861, 0.27277671663856106754, -0.17286664178559238425, 0.28613037020861598059, -0.26025200164390527302, -0.046902454566413356862, 0.069374759631404714355, 0.030029050158040294233, 0.00021096780544465308649, -0.0087889499371271266603, -0.0038519190675782408358, -0.0013990228732109833212, -0.00020599833470614527502, 0.0017399464935470721622, -0.19408847190207625943, 0.37709333205645734211, -0.30969749976723176532, 0.39090810339614029401, -0.33749737550703445832, -0.05794494823092535829, 0.14706724638512491676, 0.029892483705442618458, -0.033118862196852720536, -0.023037017158998217281, -0.0016995599198420413035, 0.0062674988072488318352, 0.0053180049157428464432, -0.00017291736890245320646, -0.178018663290264334, 0.32631353406098068115, -0.24590426601418682861, 0.36148513451461733448, -0.35492512230883860314, -0.038432841427234118448, 0.15048825805047935655, 0.024232175792193277886, -0.033979158399762149356, -0.021389176840470123381, -0.001088774241163640525, 0.0058515240594956934028, 0.0047650408820829734044, -0.00023515636615402225949, -0.36261323599962674846, 1.1951867179782560413, -1.6042061159522214009, 0.84668777973832964889, 0.28235347036899599571, -0.38239886540361517353, -0.10179911288295305505, 0.086332641490578274079, 0.056634750378197866449, -0.00287052987080646093, -0.012186128967879112858, -0.0024773220632009528355, 0.0018842768409391180869, -0.0016063179923332623379, 0.019948425115247300082, -0.17480700343790248064, 0.31418327804299728268, -0.071479739801077565864, -0.22983233460602076059, 0.12389219119772602873, 0.034581546742757852952, 0.0027495090579354969189, -0.0047142915384832794617, -0.019134688247528234412, -0.0055089292327061764598, 0.0070500335106489283449, 0.0059023438986813411486, -0.0033583048436985479812, -0.18912136626431699793, -0.096733357271837044911, 0.57563038745920203088, -0.55000571758555849566, 0.52783545601027426208, -0.077778406438301292414, -0.15210207954154705834, -0.0761096210688432806, -0.04549514690235466341, 0.01263789411318621235, 0.039346703493186661882, 0.031248053043694577063, 0.0078412579375092463646, -0.011486452009965881713, 1.3950822257863726428, 0.014355783617980915137, -12.090432980748607505, 21.48279085006946687, -12.037863483809614351, -0.39308617647560911701, 1.2784456832881070376, 0.51832030313165911029, -0.011442852155950646378, -0.10385801322725378426, -0.052701646781513435125, -0.010069820112425504829, 0.0053148292539947203025, 0.0079094756120647417841, -0.054850160816802148722, 0.34960771575220056517, -0.34231327403520134256, 0.17972645881098467391, -0.53086088350970361827, 0.13370098957041226528, 0.61237265013054875684, -0.14231983769837114329, -0.27617629635550700051, -0.016367480118553486695, 0.072899955814516512187, 0.033504237958540697417, -0.0054429867962343427001, -0.010821605673429096975, -0.38186019302322571356, 0.77172239955724175609, 0.89866069703511064493, -2.6016375771811675044, 0.86743200787694652032, 0.7902249861807462139, -0.16234019073403652134, -0.16678845929646513024, -0.033769849195253866248, 0.01575492032906543674, 0.007362137020308919845, -0.0028441986521699274756, -0.0030255136429925997762, 0.0030589071709463530255, 0.1594142345599605104, -0.47132723851035274176, 0.047244373907302614213, 0.52523719447674954708, 0.11502089357614686538, -0.45852201522317126425, -0.06724899383355333049, 0.069994581215252663076, 0.085091197113861372081, 0.03387902445691066966, -0.020918511692412333075, -0.022350097492077454497, -0.0047337070077392668851, 0.0075100620541849738565, -0.63959560190136488522, 1.2170285907584961205, 0.88404041233409580425, -2.9237539469096094358, 0.81841657033861880954, 1.1415828922599335904, -0.1594822665396646022, -0.34249870339146420983, -0.085212729822119137602, 0.04226206452179846057, 0.041279309364427523021, 0.01611734912856016208, -0.002121739519889974477, -0.0080687392612158199273, 0.56563473613633619141, -1.3328426177179719669, -0.021051481261820468444, 1.5431412653427842852, 0.10994509832103849245, -1.1782270142245276912, -0.10870775081447565935, 0.33852139849013834771, 0.17053754167323448288, -0.004420319525230236718, -0.064692876363092802205, -0.035866525301751911448, 0.0031801970247133382053, 0.013611970266710486308, -0.35977412097375621736, 1.0010555935259848326, 0.69250445494928070111, -2.689944773243280185, 0.59386008591630101705, 0.9708149244039261383, 0.12627886009965466618, -0.17663822668332293531, -0.15049487227502258979, -0.041103744408257633647, 0.01642524799644987446, 0.020925519174760535968, 0.0054104817889029273303, -0.00375522427400801391, 0.3426544524971131267, -0.66861882073824796535, 0.14351804373477700105, 0.35073267119573603434, 0.17062520055676688724, -0.57663280248733750355, 0.052437266837620670479, 0.21570626302705470101, 0.043656282939132970422, -0.030901952307234106104, -0.033077852345592177741, -0.015652762758387964159, -0.0014137460965915403348, 0.0087644500372396159232, 0.082398139868391062635, -0.23216765584168688741, 0.37500555280493197419, -0.44611290860404451131, 0.36516446165028609849, -0.18177987110767515611, -0.032135534188533862454, 0.074610974001268218214, 0.010661443221390370928, -0.00830425306891572311, -0.00221211357569367148, 0.001168199727083851433, -0.002536185148346945667, -0.0038049944511224825967, 0.093763283505077760216, -0.25566485074864309102, 0.34811265191224854965, -0.36641082801813179559, 0.33536675702740720295, -0.20740892735500876531, -0.016310288726515135421, 0.078551050375220726751, 0.0053655619641314288015, -0.0099582886483620663282, -0.0014452510231487067371, 0.0017463046633874861752, -0.0021890510721217877263, -0.0035152536382052359219, -0.36185240879374624967, 0.53004554232087608412, 0.88345791362323422558, -2.1137339445658422932, 0.86174153674572784301, 0.56084561283835310252, -0.24595874529223388238, -0.15164012284354203008, -0.010023879746948549316, 0.028423444355768691383, 0.016882698156131598771, 0.0044478892028747161677, -0.001253606780470588539, -0.0022993488684256669849, -0.33056589764879873439, 0.46063761967970706657, 0.84847845755109974064, -1.9665237249134273068, 0.83264729531808845397, 0.49358610214142606409, -0.24543079181025048707, -0.13659120604811553212, -0.001335795301102548141, 0.031085734886890342699, 0.014873716744229263653, 0.0015197105988289246695, -0.0022496836546964170227, -0.0012011500147840290399, -0.44026442721937841807, 0.72147598778056831748, 0.94733664371775994084, -2.4477029083135928289, 0.87720113125923082631, 0.7250034913091171207, -0.19341987797446411057, -0.19547486573330047355, -0.059605188183239968258, 0.026797129606377475403, 0.036200576802843741075, 0.013125971501902762573, -0.0043377308494621010654, -0.0069626464908812751506, 0.033836230189140288627, -0.25454555985529275786, 0.5033791207498164999, -0.55319777835123284948, 0.4883193895823472741, -0.18908786997514370753, -0.11348606249130675783, 0.061921910898882992491, 0.016949017763042955459, 3.8460947519320092675e-05, 0.0072297119608671842061, 0.0075080467501939598951, -0.0011609490687916688544, -0.0093029930206254898839, -0.36098211716953054751, 0.5499485315794803153, 0.80605085989311664463, -2.0192448519976533561, 0.84430966860371003957, 0.57430934389213084135, -0.32636818257789790287, -0.16553174183372240669, 0.058928575397392865542, 0.056734861223056876067, -0.0025631322082067154074, -0.017133288192917560389, -0.0058387442500484124097, 0.0062297489205357147249, 0.075283062012731272183, -0.28407349896501915554, 0.31334451568313165248, -0.2127106260535633897, 0.33130236025570464209, -0.23759386050586320249, -0.071926838454508895682, 0.062334903994251768256, 0.034703570444745791834, 0.0020990009544364475816, -0.0095108449376127609382, -0.0045885384302297793982, -0.0011161576300325487011, 0.0018549157516886990819, 0.089337702195786086734, -0.30382299615404201321, 0.28613037007074199058, -0.14595311149763506853, 0.30135042775668008375, -0.25892662750994988041, -0.053537188373279802345, 0.067815925589197384449, 0.02935001671517812466, -0.0003245432132718993265, -0.0090984411059126611004, -0.003851919632839388187, -0.00064175224922975162775, 0.0016700881197235875088, 0.16339036321382180428, -0.46250093107568446271, 0.39090810313994767133, -0.20618107863123621981, 0.41042188984584870859, -0.34935250985303317739, -0.074561528625546491522, 0.14135398890254136095, 0.030608586543098899313, -0.031062995865265959072, -0.022097797548494151759, -0.0016995601486475592386, 0.0060562391596718119502, 0.005305466223583631645, 0.17658502610302778679, -0.47629652921108478481, 0.36148513522010133903, -0.1428396888913111562, 0.37633190441442793128, -0.36786121351811068569, -0.05427363487948745624, 0.1456394726919196525, 0.025301274413990014478, -0.032016710504079101685, -0.020614283137070942009, -0.0010887739281546034327, 0.0057085268638496248703, 0.0045778502018869607695, -0.2901635063611018861, 0.27128296306135257776, 0.84668778017704082739, -1.675165999451929677, 0.89217242164847099239, 0.32402802378159423213, -0.38372962401349236528, -0.11334068140421700577, 0.082463445027004839138, 0.058817173621165771968, -0.00079558885680397787785, -0.012186128720240685733, -0.0037895755157855109778, 0.00078519567590443553606, 0.11667617622593115523, -0.20137948788331874739, -0.071479740030009383811, 0.29877558180008589872, -0.038395964256375321455, -0.22579515906258176838, 0.1045628720704195741, 0.027669220023274140874, 0.0055800886668384147571, -0.00029421423107591296048, -0.017102014025446855144, -0.0055089295583213102742, 0.0056778202679090708851, 0.001721565738993765908, -0.051991628560767576273, -0.11782308498203077074, -0.066143131129378252009, 0.52783545540582144096, -0.57172941851301517513, 0.53325368849922460956, -0.065469572628830408068, -0.14716869501241197637, -0.076387760205615898301, -0.046834603728327726946, 0.012637894648633260161, 0.040782753257649129175, 0.030314350390297117865, -0.0057035359901578958075, 0.040622199139072008578, 1.5043461079828315263, -0.27092884057459004632, -12.037863484095423061, 21.592776015355283903, -12.024686134896001022, -0.41985846182837860407, 1.2622654448034866803, 0.51860513471593372437, -0.0072359098081589490539, -0.10385801295875389527, -0.055994783736668267415, -0.0095423780451439638856, 0.016226667042975592009, -0.44319265029553650548, 0.72803610934367235519, 0.14452009985151934024, -0.53086088294862954218, 0.18130114434281618951, -0.49790133538028008564, 0.14817215034080546965, 0.61580575445092966547, -0.14327293454439810061, -0.27869925670374962046, -0.016367480175842667039, 0.077124945435811637084, 0.035719555805265694459, -0.022274900558846996712, -0.10678842166234495115, -0.22997737589277519876, 0.7632927370279923851, 0.8674320074547637871, -2.6115179819256963967, 0.86444827305450777644, 0.79362888789871477435, -0.16001711942474167172, -0.16733669366651096611, -0.035273009073685704373, 0.015754920577613251015, 0.0090954339404836215266, -0.0010022059639254879762, -0.0018364804696078579523, 0.1665339708055728063, -0.074061366622555968586, -0.48690008460275352453, 0.11502089293656307667, 0.5627210548685843694, 0.1106810587537756102, -0.4687690752552542417, -0.073003772804718009004, 0.068603802496492630181, 0.085504959561295945658, 0.03387902489953355939, -0.02258422487741149412, -0.023009892793776406256, 0.0059965266419635685283, -0.23412285609607988435, -0.24074777193809887565, 1.1158128100451254472, 0.81841657001373391278, -2.9442450408589668776, 0.81909859692718745006, 1.155324792516864818, -0.15107894066812019296, -0.34242217312634215531, -0.087783136270877470952, 0.042262064876907562416, 0.044020119780944062804, 0.016248124034334922999, -0.013376318494459287639, 0.37872116401455557799, -0.075044536002899814298, -1.1918687299535897228, 0.10994509807198055151, 1.5852907134269731415, 0.089517159685067251851, -1.2039007817396740752, -0.11721031960923382342, 0.34191941705940526175, 0.17506422586338732383, -0.0044203193644292140471, -0.070470102057341804658, -0.039984602804593136205, 0.025361986121298907976, -0.31740804628020113931, 0.09549974868983110432, 0.98588822656735430883, 0.59386008558425085191, -2.7383477950092998121, 0.58984488238278542926, 0.98494658125674294791, 0.13485238179795649982, -0.17576006701846536018, -0.15241507267926246438, -0.041103743828015619488, 0.02043780850743895372, 0.024974752245813743728, -0.0048561496978102754174, 0.13316670278634437641, 0.089752799837268651406, -0.57111478635466117471, 0.17062520039696169039, 0.35906844608017551845, 0.16975329256252238985, -0.58246992054934643157, 0.046393344612121696635, 0.21347578308463410757, 0.043551366652176966288, -0.030901952094382435787, -0.03285765201559145976, -0.012816917837139524955, 0.0070013924918904679728, 0.048596277845560233166, -0.011045960017607259901, -0.17493399444194837611, 0.36516446191100893071, -0.45427878638566265845, 0.36806887124432646763, -0.17703407404216472965, -0.031912369702765619461, 0.073961346090392218744, 0.010529263279137895601, -0.008304253495160464002, -0.0021085928625772668368, 0.0010486588068856867936, -0.0069006010900617316406, 0.042602703002261085863, 0.0096417780864930177631, -0.19991067693998931798, 0.33536675698419643465, -0.3738874335010701766, 0.33888514836208227798, -0.20287049478591326346, -0.016232235276416366598, 0.077826521727382622085, 0.0051286962232299174863, -0.009958288732107146643, -0.0011862020931180609212, 0.0017948716062538760604, -0.006432178575142073057, -0.047312267391871956623, -0.2967195998590363426, 0.53832273872938596337, 0.86174153700442968074, -2.1234979858152844834, 0.8595160153678246262, 0.56509878565755577995, -0.24289042309312053325, -0.15092373815019136107, -0.010326175648245023569, 0.028423443839537862465, 0.016960590037082008685, 0.0034763449400286123778, -0.0033846851737681388911, -0.027078486099120956804, -0.29374783248200375674, 0.46812753725599243193, 0.83264729574096185694, -1.9710559096090554121, 0.83104146291082503506, 0.49617589065997036446, -0.2434107665080502414, -0.13598377211710796497, -0.0014583230860569676261, 0.031085734723790958006, 0.014634219388306779486, 0.00039688894015961533263, -0.0027526768556288120672, -0.13126733130306120212, -0.23005164871202452015, 0.70670362701139533002, 0.8772011309769944809, -2.4655616778907232245, 0.88609922291410336825, 0.737663222369094429, -0.18851103726199122668, -0.1970258772521482904, -0.062548646011684583268, 0.026797129335229853686, 0.039184596543224224641, 0.013881742324230395003, -0.014523441622636551473, 0.065889631066603573939, -0.087289902886733633514, -0.18058149916924687295, 0.48831938943803154807, -0.56608277171723675458, 0.49110625415164471663, -0.18247456706032663676, -0.11062976428681091379, 0.062876657997724294535, 0.017144437541137935027, 3.8460948750756920184e-05, 0.0067741252070611570474, 0.004868809041431832775, -0.01000451266629545817, 0.047767706205379741047, -0.42294156140128619237, 0.52446933240860271752, 0.84430966901425086313, -2.00268201010435698, 0.83569991046630076603, 0.56815404081101605893, -0.32759641322410498754, -0.16481254287215610366, 0.059633645918536842823, 0.05673486124367198985, -0.0043752311123065766238, -0.018835396212037159441, 0.0038298354333571658741, 0.08158021269135794229, -0.066485943011197590202, -0.24025416505885324736, 0.33130236108875510403, -0.21021727122605868288, 0.33099388607692803443, -0.2365966316291952809, -0.074061061968212021922, 0.060127529449858857502, 0.033919884237350143852, 0.002099001232220197577, -0.0090118357770394980077, -0.0031318480092875334263, 2.0014727287178454146e-05, 0.075230609067606979412, -0.043112923298054370291, -0.26025200202130621685, 0.30135042761947328227, -0.14349239676877331706, 0.30090321362242361536, -0.25839538020438584942, -0.055539345523054825526, 0.066039391355199386635, 0.02882145531128365093, -0.0003245443968263777082, -0.0087889492712197106461, -0.0026639572437527443347, 0.00052996335457112923352, 0.093948565086421326198, -0.053361710920903387967, -0.33749737558289766293, 0.41042188958997183335, -0.21627575960662445698, 0.40286198377188603414, -0.35344657713456673065, -0.076963664642244133951, 0.14200820275085279665, 0.031945738664846214105, -0.03106299641666106226, -0.023037016730044910379, -0.00079955854738434609282, 0.011953693915983127499, 0.087821225391471191557, -0.027573470589810959719, -0.35492512279445764811, 0.37633190502276320233, -0.15267819503453122776, 0.36989164930947787679, -0.3714859296718959758, -0.056462920546730251126, 0.14627372868774543568, 0.026495269364176454729, -0.032016710092598051063, -0.021389176976284222931, -0.00025657750372177394807, 0.01073355606873148925, 0.11237539893738283936, -0.45836353149636105409, 0.28235346991692206187, 0.89217242209535640995, -1.6618749539576060936, 0.88284349980880838515, 0.32041863649766932909, -0.38318793324138589185, -0.11232697674458447246, 0.083143679052632160631, 0.058817173675419914813, -0.0028705297288911424576, -0.015428830971309363818, 0.00079649137679975792747, 0.0082913343897453618825, 0.10960928887026978595, -0.22983233445194656275, -0.038395964084816283646, 0.31046454624859737281, -0.053904337330199857126, -0.23346696163061164686, 0.10615225372402561921, 0.031326470237490397497, 0.0076371364090891151119, -0.00029421486024226690481, -0.019134687749059482026, -0.0078919827988533520352, 0.010133735872606975878, 0.039388378777483287752, -0.057618181332219034485, -0.15815322137483914244, -0.077778405920451543221, 0.53325368816506257819, -0.56157972045607351674, 0.53801538607405752224, -0.064530947149179382349, -0.14745138478530628312, -0.076387759655614950738, -0.045495147323878021295, 0.015226718822950914209, 0.040866718365264877166, 0.019989156707592107415, -0.26632498888215566657, 0.53970643640264770458, 1.3437254747117988174, -0.39308617598958361938, -12.024686135202644621, 21.624459708806572422, -12.017429250004859398, -0.42173393719981400185, 1.2620267924306471929, 0.51860513499842775254, -0.011442852711930264625, -0.11079547765487991751, -0.054252709117362123148, 0.013465698427663277958, -0.093052602555932090178, -0.21123525044275234808, 0.56446407859362535131, 0.13370098872918800836, -0.49790133473490660876, 0.19245014223782691554, -0.5030738757153517815, 0.14310105853649049412, 0.6142934820767009807, -0.14327293466491755614, -0.27617629597354459436, -0.0079167364452683384318, 0.083800182873154036023, -0.0011569210007102007694, 0.0309229753722797783, -0.19435650152031158644, -0.18385642427982012981, 0.7902249866061645811, 0.8644482726740124745, -2.6207032936731224915, 0.85910285480450598872, 0.79213523305737720825, -0.16062029632574767457, -0.16733669346230167774, -0.033769849344198556162, 0.018714661236835579416, 0.011060053892208611057, -0.0066319585827408706127, 0.054013649239991606832, 0.070590412840541488793, -0.050216437743749488265, -0.45852201452454949271, 0.11068105829354525638, 0.55750190128938603706, 0.11013359077389764384, -0.46815385926015923301, -0.072926739813253621536, 0.068603802941224159562, 0.085091196499911286866, 0.031980025041680466835, -0.024050997516620972738, -0.013641731351728260055, 0.082491166311902508479, -0.3804410636844304805, -0.19571852531670846931, 1.1415828929099134381, 0.81909859644938420864, -2.9480424397562337013, 0.81625387915879854539, 1.1544839461753837462, -0.15170259349890463652, -0.34242217276925079883, -0.085212730453228538452, 0.047191891564762097899, 0.044171220086086239454, -0.003193857371830107969, -0.013933815283472240909, 0.37853436193787681763, -0.053124762775394566983, -1.1782270139392991837, 0.089517159419132921738, 1.574544267107854667, 0.093782537603816065741, -1.197701582678690313, -0.11453928539471457115, 0.34191941706766726394, 0.17053754151241531933, -0.015097197096679051617, -0.073999532152580801969, -0.00010767908970984790482, -0.084489851615427977172, -0.21276587401969362934, 0.09718025390883139325, 0.97081492513464384686, 0.58984488195445083925, -2.738647902068485962, 0.58597433395407372281, 0.98171649685836848143, 0.13361479790662855738, -0.17576006644431849124, -0.15049487315046078395, -0.036248754906541613696, 0.023755661961128868298, 0.013825250231153480771, -0.06407646344106697478, 0.22614666746756034721, 0.068336011930639276035, -0.57663280158955354526, 0.16975329218882379001, 0.35990733275976732886, 0.1681863702081448364, -0.58484348590157408943, 0.045071216564228071266, 0.21347578349525059282, 0.043656282147296360252, -0.031458622350014467384, -0.031821559350089317908, -0.004648543612531739512, -0.013304003621560483647, 0.0814913951924652602, -0.031338915232667657884, -0.18177987105412726776, 0.368068871325300917, -0.44868356708984202807, 0.36920053127604346432, -0.17777565145527729484, -0.032363345797308928908, 0.073961345892298802074, 0.010661443332365568709, -0.0083526322991951187924, -0.0038056395541705501442, -0.0049736884889077462996, -0.017822067594084979592, 0.084919626449892676723, -0.015602957020790611184, -0.20740892705786304062, 0.33888514827629417914, -0.36815956458984527844, 0.33995681674263067418, -0.20373758070030908995, -0.016795113034440811717, 0.077826521687500663238, 0.0053655619189122236851, -0.0098364123207355363271, -0.0025820069306815244398, -0.0041466321956895625667, 0.063675374357158739946, -0.16072538797771274144, -0.26081316175601654272, 0.5608456122005888167, 0.85951601563590496191, -2.1291613465505534464, 0.85776524940092147009, 0.56620082668542870419, -0.24214214380268894722, -0.15092373870854630669, -0.010023879059998276608, 0.02949836014208281551, 0.017289356971611390817, -0.001679695969594107623, 0.067082382854168176589, -0.14548810659661376432, -0.25884041049162648251, 0.49358610129884805673, 0.83104146352432561873, -1.9774478719843131547, 0.82898581178429331384, 0.49736960040457256493, -0.24255273872830704351, -0.13598377228941888051, -0.0013357949112565447886, 0.031865313927107334624, 0.015053555096137065036, -0.0038514487104711215421, 0.058165260749522017247, -0.2164340011041885059, -0.22523977308087295146, 0.72500349169275668526, 0.88609922256771200821, -2.4653096755964201137, 0.88360561694512418374, 0.73567856497048578213, -0.19007170318877017468, -0.19702587767272494723, -0.059605188030774657049, 0.03178808245638871377, 0.039663218508361117176, -0.0071609611903681955752, 0.0089065286526911653653, 0.07816931059118828462, -0.11184205095063209923, -0.1890878694237245472, 0.49110625392328710603, -0.55896749100798781562, 0.49534672348646230366, -0.18101994760844389298, -0.11053638761410079783, 0.062876658290865847234, 0.016949017331725518831, -0.0015251101449421735918, 0.0028805774140864366314, -0.0022456664376551771847, 0.11438558156308770153, -0.18017279543006056719, -0.33339335974596928658, 0.57430934345775430661, 0.83569991091518303072, -2.018348211294252792, 0.8312661900531866932, 0.57041465962967197179, -0.32592319439004086012, -0.16481254288268967767, 0.058928575508969599506, 0.055761208988469902759, -0.0041231633452255842323, -0.013867056004812359712, 0.0077800158694904348808, 0.069918840127818029928, -0.065845893731898974544, -0.23759386196143444936, 0.33099388733274370811, -0.20756771030437096148, 0.33099231946581975095, -0.23812296006387459468, -0.075079475522969216228, 0.060127529625050565865, 0.034703570849183601466, 0.0029750580098852749331, -0.0094538391792838970956, -0.0034838372318946510958, 0.0015300533333296513928, 0.075320982012689691176, -0.046902452990729988358, -0.25892662800481291985, 0.30090321321078999928, -0.14099222646965564598, 0.30106992568132284749, -0.25959113726262694755, -0.056333418797672278611, 0.066039389796189770787, 0.029350018303826601318, 0.00021096733400957287663, -0.0091297915324669940557, -0.0022363964625536311183, -0.049389473693655816522, 0.16507460652894401232, -0.057944947346197414606, -0.34935251008349671453, 0.40286198329695221965, -0.21722521513205100829, 0.40338626529164800516, -0.35257599844675235889, -0.076291640650056780082, 0.14200820194816068942, 0.030608587389405320089, -0.033118862300379817154, -0.021357266511389982194, 0.012942986212863367632, -0.052987562081492141364, 0.16817441161607474265, -0.038432841727467338677, -0.3678612143409873414, 0.36989165018849090583, -0.15288057958408907133, 0.37061858053562435344, -0.37074343763705192023, -0.055956674262038028078, 0.14627372934506152613, 0.025301274258237166359, -0.033979159011405489499, -0.019928797200962789643, 0.012227291371844528634, 0.12480395702513806533, -0.11173079490682856496, -0.3823988654357297623, 0.32402802352894422011, 0.8828435003318625407, -1.6721180101219690339, 0.88283395066864212009, 0.32379018449135965607, -0.38168001614465113525, -0.11232697649058748235, 0.082463445014713906978, 0.056634749863676850379, -0.005768088431810292592, -0.010883189873067647402, -0.019029115243278195424, 0.023152279642468333309, 0.12389219158254984043, -0.22579515931931842809, -0.053904337273238477068, 0.30522407604906226508, -0.051029889512856765887, -0.22987463680641381791, 0.10784543811746272191, 0.031326469705449770664, 0.0055800892862610000297, -0.0047142915758119604067, -0.021259379113090354912, 0.0087266200527278646459, 0.08178220283120229972, -0.03757420898260448866, -0.084278130083847130338, -0.15210208067449229596, -0.065469571403922413588, 0.53801538605215626365, -0.56113708367498127405, 0.53774270599596218911, -0.06453094685684013232, -0.14716869601432161496, -0.07610962024466984277, -0.046112054854752913824, 0.015931200464434559783, 0.060691898350436097875, -0.1470107134604568544, -0.040572057689250470847, 0.52922667172335180474, 1.278445682730863231, -0.41985846144697497317, -12.017429250127293017, 21.636752895095582971, -12.012704409955773599, -0.42173393711539230999, 1.2622654443866205831, 0.51832030378385252067, -0.017522091463482651619, -0.11903869178190677747, -0.02886675799192803954, 0.1335783480197244022, -0.31161386052355827259, -0.17219017962311458647, 0.61237265088914261124, 0.14817214937079067893, -0.50307387499885192206, 0.18471296160537731268, -0.50686440930552845163, 0.14310105834111086498, 0.61580575493639755358, -0.14231983851832083476, -0.27637478145071819879, -0.0072781377421414824441, 0.081648002493103749577, 0.027057589711154104301, -0.032563692772652252927, -0.17209257511751932523, -0.16234019118526069048, 0.79362888795245145612, 0.85910285466496749684, -2.6258231806249665574, 0.85678892204655621168, 0.79213523316596079482, -0.16001711933773377083, -0.16678845899863445346, -0.03142867458687286647, 0.021349851137918190308, 0.0001850044629864783447, -0.035814951405284969899, 0.097539917319990301903, 0.085485185949361622293, -0.067248994014487264992, -0.46876907486209362652, 0.11013359028362285086, 0.56017149730756621295, 0.11159584945894672481, -0.46815385884917626313, -0.073003773265870833264, 0.069994581511999928725, 0.089275413685789270501, 0.034980369710736966549, -0.045521868954072002011, 0.10264389325431379218, -0.079950697293122804932, -0.35834620601185029898, -0.15948226744781024422, 1.1553247932700849709, 0.81625387877759847388, -2.9541509066136111272, 0.81314790854158514044, 1.154483946483926271, -0.15107894137060065565, -0.34249870254719688711, -0.084066818515754759833, 0.049769810632702998765, 0.037854134693195977401, -0.13463130129786707312, 0.18787993134851008192, 0.37253534924997278033, -0.10870775083184565368, -1.2039007815559350512, 0.093782537334982138399, 1.5860789361253755381, 0.099490558145224200493, -1.1977015826540833299, -0.11721031959527726529, 0.33852139845175488464, 0.17065682865780834598, -0.0086498722736351642965, -0.077885654116564381755, 0.041415049807172275853, -0.16499680943824124313, -0.20185747011880464385, 0.12627885935962335218, 0.98494658186993611082, 0.58597433342925386857, -2.7469941536091422307, 0.58159194736583763596, 0.9817164975021036577, 0.13485238106370764366, -0.17663822607193396808, -0.15532796110008847967, -0.039766270163384472014, 0.046864103951929553249, -0.074367079546867886908, 0.036114441136027339296, 0.22142201815701084699, 0.052437265616020273251, -0.58246991988178786759, 0.16818636991852742302, 0.35961657862977358668, 0.16779650818275651525, -0.58484348559155019398, 0.046393343946308923664, 0.21570626410010382279, 0.043868196076078072587, -0.036330015885656929964, -0.033859072732967107622, -0.010062681641610745839, 0.0043720928751170180793, 0.070353310967481930138, -0.032135534074625361767, -0.17703407419639199949, 0.36920053116891321521, -0.4491065759888478448, 0.36890230923127109719, -0.17777565135234954519, -0.031912369561211156865, 0.074610973870119584594, 0.008926250120291332607, -0.014847760374657358734, -0.0024660531557356727902, -0.0089977012946056995207, -0.0015346420418314151921, 0.07424514367409133786, -0.01631028881665307767, -0.20287049475651891539, 0.33995681652793230398, -0.3686410829759325658, 0.33955868575521014341, -0.20373758052785606587, -0.016232235254046500222, 0.078551050294966590082, 0.0036798495722144344833, -0.015999116719758038618, -0.00079019636477652371951, 0.043737740927310364558, -0.0020078248853463440474, -0.15376528470543393556, -0.24595874448936988488, 0.56509878532176027743, 0.85776524955574096065, -2.1306582713846791144, 0.85752822973885711733, 0.56620082615326394748, -0.24289042250713746562, -0.15164012331520165811, -0.0079208298237399953762, 0.033171299898099124159, 0.011321865640897967042, 0.038876569842616399975, 0.00695873968804486525, -0.1366036155446948841, -0.24543079102565076854, 0.4961758899836247183, 0.82898581226211187634, -1.9785299481770746155, 0.82904712137033187425, 0.49736960025845938782, -0.24341076624280402751, -0.13659120663512006044, 0.0014378051677776638399, 0.036083067714511467272, 0.005713957156870961944, 0.079327196071123012855, -0.062924727760816742683, -0.20849442921061495104, -0.19341987771730245749, 0.73766322258802341505, 0.88360561667376180672, -2.470913548309754848, 0.88041288178278998, 0.73567856442053913835, -0.1885110366812079441, -0.19547486589189916328, -0.058087004410656707665, 0.03110975248809060012, 0.03038107182198836187, 0.007503885885723341094, 0.013371393849911896584, 0.057891717092661221233, -0.11348606320176463091, -0.18247456663064010041, 0.4953467230745136618, -0.55716405430689741607, 0.49573453105106396466, -0.18101994710509999864, -0.11062976497269698395, 0.061921911634061826968, 0.01360551078145375882, -0.0072688429804977153226, 0.0082787536040333474519, 0.012263796192018313128, 0.078139645981581129508, -0.15424850261538239349, -0.32636818240901804833, 0.56815404028119720348, 0.83126619056402628605, -2.0181452587815038591, 0.83249435721856523962, 0.57041465951209313623, -0.32759641300141828513, -0.1655317421239432818, 0.065266874333564914301, 0.064404319093636100413, -0.030349312842223148678, -0.017870449957204184027, 0.036718992686209574183, 0.063798712097105206298, -0.071926837298486576144, -0.23659663330787078506, 0.33099232056214700215, -0.20833868560647647583, 0.33049461339838281937, -0.23812295970458816075, -0.074061061736376082587, 0.062334902782860468484, 0.035677116255489520202, -0.00067741576334505622381, -0.012269212736505788475, -0.017879099915222164408, 0.030834209181110273923, 0.069374757622530508483, -0.053537186967031351326, -0.25839538017388097302, 0.30106992503703938757, -0.14138070077397396629, 0.3007907602384104484, -0.25959113870209482577, -0.055539343739687253854, 0.067815924611498559771, 0.030029050487673028536, -0.0029334610478555065045, -0.010539190540270205673, -0.048110868395398860564, 0.032898962887496836782, 0.14706724522854208725, -0.074561527906531546717, -0.35344657676152468895, 0.40338626475168581909, -0.21520114853187868875, 0.40465687072124634227, -0.35257599928538929657, -0.076963663855500560884, 0.14135398863922077339, 0.029892483433800368098, -0.033273997730990691191, -0.0063142982826889801698, -0.046164753650992193523, 0.025953710301400288035, 0.15048825874576807204, -0.05427363476997192071, -0.37148593060194323723, 0.37061858159246985345, -0.15095071908556900731, 0.3717085065701372093, -0.37074343701442852295, -0.056462921113598046885, 0.14563947240247715476, 0.024232176547318729637, -0.034577182257058462767, -0.0050412118044407703762, 0.013291670627593786189, 0.10398504868767367848, -0.10179911265088850691, -0.38372962400190874233, 0.32041863603171710784, 0.88283395121381269544, -1.6697889929420122179, 0.88451632537635627163, 0.32379018479680260478, -0.38318793348873997351, -0.11334068173968502025, 0.086332642038534246343, 0.062652755111487420936, -0.02501297021011243335, -0.02627421428956246488, 0.015614882238276263371, 0.03458154590612894097, 0.10456287231848816754, -0.23346696168027830653, -0.05102988938899062632, 0.30951030928238115747, -0.049067508862688284654, -0.22987463736630081579, 0.10615225424184111802, 0.027669219955140302025, 0.0027495088891457486815, 0.00023007543539924288638, -0.011727564295318002041, 0.060691897822223235537, 0.015931201340549410589, -0.046112054819628808811, -0.076109620475750483504, -0.14716869636219573336, -0.064530946954641510427, 0.53774270615938701834, -0.56113708302714204734, 0.53801538573352924999, -0.065469571957826544484, -0.15210208014919698871, -0.084278131112533474778, -0.037574207487271880923, 0.081782202205949713902, -0.02886675796536209912, -0.11903869186542019587, -0.017522091332021216131, 0.51832030362659586853, 1.2622654445442262894, -0.42173393722989516164, -12.012704409881656886, 21.636752895022585363, -12.017429250063296209, -0.41985846152223754713, 1.2784456829098278519, 0.52922667143532919898, -0.040572057458819796305, -0.14701071353147157628, 0.081648002374574993323, -0.00727813734417381928, -0.27637478199196618966, -0.14231983823900712371, 0.61580575524227454398, 0.14310105756923416354, -0.50686440864174386878, 0.18471296147848029201, -0.503073875239738455, 0.14817214950408547081, 0.61237265109432625909, -0.17219017996807214899, -0.31161386032442950356, 0.13357834797963089035, 0.00018500466667572155132, 0.021349850603633414164, -0.031428674170586028136, -0.16678845904498793007, -0.16001711935562201727, 0.79213523316668432717, 0.85678892189532585011, -2.6258231804440397283, 0.85910285471107272759, 0.79362888777140150331, -0.16234019104835062386, -0.17209257516842604852, -0.032563692818163139653, 0.027057589753634508911, -0.045521868831391615329, 0.034980369332855539011, 0.089275414138969499955, 0.069994581207832473613, -0.073003773123953619173, -0.46815385873999931787, 0.11159584918223446059, 0.56017149737682458976, 0.11013359036705080984, -0.4687690748818278963, -0.067248993887261493407, 0.085485185827858328755, 0.097539917208429541273, -0.035814951298392273815, 0.037854134738908890023, 0.049769810516212965867, -0.084066818407958307513, -0.3424987026425798109, -0.15107894130254345688, 1.1544839466111918025, 0.81314790830220096129, -2.9541509065071118734, 0.816253878720053061, 1.1553247933092944955, -0.1594822672189273316, -0.35834620637920977515, -0.079950697143711682569, 0.10264389325862538493, -0.077885654041828303318, -0.0086498724519494304269, 0.17065682885481139874, 0.3385213982389180809, -0.11721031951533505966, -1.1977015823800276628, 0.099490557772017942928, 1.5860789362292906368, 0.0937825374431361658, -1.2039007817387898935, -0.10870775055972245471, 0.37253534892576412041, 0.18787993159459812076, -0.13463130138320350548, 0.046864104102832149223, -0.039766270422893858216, -0.15532796101258050658, -0.1766382262172121187, 0.13485238128913626388, 0.98171649765016522071, 0.58159194692679461181, -2.7469941531817267943, 0.58597433311294089275, 0.98494658183894356895, 0.12627885977201197543, -0.20185747048296812101, -0.16499680929288801368, 0.041415049760055569483, -0.033859072633283061882, -0.036330016184188088091, 0.043868196464280563573, 0.21570626374318896401, 0.046393344088181659446, -0.58484348534060226488, 0.16779650771263218689, 0.35961657898454163496, 0.1681863698355480774, -0.58246992009139542112, 0.052437266068721538614, 0.22142201778386691458, 0.036114441220863033322, -0.074367079531552610083, -0.0024660530129242720643, -0.014847760654527494481, 0.0089262502216448449954, 0.074610973882946601821, -0.031912369528343927094, -0.17777565130838307583, 0.36890230911647492462, -0.44910657583824875605, 0.36920053092752602719, -0.17703407406005042257, -0.03213553402869147313, 0.070353311111966715385, 0.0043720926193241346769, -0.010062681568420288672, -0.00079019630083555802459, -0.015999116849098524856, 0.0036798495926837326148, 0.078551050360714275156, -0.016232235340301166826, -0.20373758036282413331, 0.33955868557965895871, -0.36864108287916130813, 0.33995681638915542511, -0.20287049457276534747, -0.016310288791648711038, 0.074245143662173343491, -0.0015346422306743504049, -0.0089977011899020232361, 0.011321865626135066071, 0.03317129989179700561, -0.0079208296653176454849, -0.15164012354084616119, -0.24289042253758832968, 0.56620082640860314172, 0.85752822970355913057, -2.1306582717398305782, 0.85776524997669800765, 0.56509878530583557144, -0.24595874502487449775, -0.15376528411959963782, -0.0020078250864084463446, 0.043737740927750901054, 0.0057139571098696381041, 0.036083067937313619, 0.001437804781250278871, -0.13659120637877966709, -0.24341076613470727374, 0.49736959990695017719, 0.82904712167074368079, -1.9785299481851756909, 0.82898581199647913831, 0.49617589024727115632, -0.24543079119844851688, -0.13660361539146770316, 0.0069587395944240935247, 0.03887656986453603275, 0.030381071840104759141, 0.031109752220429436054, -0.058087003782490979897, -0.1954748662897042577, -0.18851103699227342303, 0.73567856499891026711, 0.88041288165656894638, -2.4709135489172169287, 0.88360561756327127281, 0.73766322217607749323, -0.1934198779746832686, -0.20849442863064740483, -0.062924728223682613315, 0.07932719621875275462, 0.0082787537256648717487, -0.0072688431839503693521, 0.013605510814105476955, 0.061921911541889820829, -0.11062976474320955489, -0.18101994707011132002, 0.49573453086088642427, -0.5571640542793551143, 0.49534672300598719952, -0.18247456643660775466, -0.11348606321088974824, 0.05789171700748042132, 0.013371393842167286076, 0.0075038858959430758747, -0.030349313035390516197, 0.06440431971198358474, 0.065266873520872556447, -0.16553174156225919367, -0.32759641304072212309, 0.57041465901610577394, 0.8324943577842131015, -2.0181452589398021225, 0.83126619047379313177, 0.56815404035356020795, -0.32636818253410621038, -0.15424850250550495301, 0.078139646019343173, 0.012263796139796912313, -0.012269212868591145016, -0.00067741510636906048307, 0.035677114944547748399, 0.062334903856075755957, -0.07406106145934854279, -0.23812296097883867141, 0.33049461422242804032, -0.20833868521894610693, 0.33099231938742418313, -0.2365966323640347202, -0.071926837359681458639, 0.063798711792280707322, 0.036718992683445646208, -0.017870449856418571533, -0.010539190566424853565, -0.0029334612313233663326, 0.03002905116772809635, 0.06781592414952744563, -0.055539344585935249143, -0.25959113697824692801, 0.30079075943355915168, -0.14138070186813003759, 0.30106992684119671111, -0.2583953806851796986, -0.053537188055418502564, 0.069374758848187709215, 0.030834208718222122503, -0.01787909986958720332, -0.0063142982637503517265, -0.03327399801224053405, 0.029892484160216926947, 0.14135398814563099323, -0.07696366440532345643, -0.35257599812517526061, 0.40465687024089552759, -0.21520114928186145908, 0.40338626588668236028, -0.35344657701079545964, -0.074561528796845494282, 0.14706724630065015624, 0.032898962348193179495, -0.048110868268801391945, -0.0050412118697334773812, -0.034577181761231069734, 0.024232175430735955479, 0.14563947327854881442, -0.056462920602817698312, -0.37074343857122876722, 0.37170850756603512899, -0.15095071854009931012, 0.37061858026388788678, -0.3714859300086167937, -0.054273634367371176301, 0.15048825824642944604, 0.025953710426802251154, -0.046164753624757151607, -0.025012970296006058896, 0.062652755469412638178, 0.086332641417720512589, -0.11334068126779954511, -0.38318793333518047506, 0.32379018411411458755, 0.88451632587371176175, -1.669788992727046617, 0.88283395057247404658, 0.32041863630552003261, -0.38372962367098956138, -0.1017991129425881297, 0.10398504861844523606, 0.013291670719344681265, -0.011727564176068806454, 0.00023007506397072511532, 0.0027495093812741943154, 0.027669219652656066449, 0.10615225415048194779, -0.22987463714467812337, -0.049067508760633128662, 0.30951030886487390648, -0.051029888873897318036, -0.23346696196638969112, 0.10456287189534037296, 0.034581546754820537137, 0.015614881757193027623, -0.026274214216381739512, 0.01998915720401513163, 0.040866717628661959616, 0.015226718217044110312, -0.045495146129282182379, -0.076387759703776980658, -0.14745138524945225478, -0.064530946531475133487, 0.53801538500154377775, -0.56157971989235844212, 0.53325368843311204969, -0.077778406711520975403, -0.15815321982685445756, -0.057618182660557827612, 0.03938837912466496699, 0.013465698330791074991, -0.054252708768069944634, -0.11079547817426545209, -0.011442852284429171072, 0.51860513475330038968, 1.2620267925331489778, -0.42173393718208213032, -12.017429250064502355, 21.624459708777212796, -12.024686135009320154, -0.3930861763267993747, 1.34372547506838802, 0.53970643619109626155, -0.26632498883050459426, -0.0011569208849314604418, 0.083800182531706665645, -0.0079167360501602012518, -0.27617629618565175953, -0.14327293476204183698, 0.61429348257952831602, 0.14310105768029135476, -0.5030738747706269276, 0.19245014158533579884, -0.49790133469016967238, 0.13370098926331816513, 0.56446407787383445775, -0.21123524996358761241, -0.093052602693419861346, -0.0066319587743874121311, 0.011060054397981602253, 0.01871466067094377031, -0.033769848908125128795, -0.16733669368180800752, -0.16062029637183322661, 0.79213523330770929665, 0.8591028544531098543, -2.620703293568900083, 0.86444827306207849826, 0.79022498602716717553, -0.18385642381553821134, -0.19435650183469155294, 0.030922975492358617533, -0.013641731438842123053, -0.024050997117476640569, 0.031980024326947666113, 0.085091197140352292649, 0.068603802634151250794, -0.072926739961023237524, -0.46815385865565606327, 0.11013359022883088822, 0.55750190138069333212, 0.11068105860826994713, -0.45852201524594593529, -0.050216436915079785963, 0.070590412404678032354, 0.054013649319282097672, -0.0031938575061010326429, 0.04417122056905744415, 0.04719189085785844423, -0.085212729894813876719, -0.3424221729415347637, -0.15170259382120795943, 1.1544839468476124633, 0.81625387850557507008, -2.9480424393947459727, 0.8190985965752201059, 1.1415828923482236323, -0.19571852480288876031, -0.38044106385320147767, 0.082491166316228325961, -0.00010767936669629626304, -0.073999531359698916821, -0.015097198044025563618, 0.17053754224268280471, 0.34191941668986336467, -0.11453928552957459708, -1.1977015821047602984, 0.093782536910587438461, 1.5745442674960987706, 0.089517159690227179891, -1.1782270146741244954, -0.053124762122581513368, 0.37853436160197850846, -0.013933815191952471715, 0.013825249941734740061, 0.02375566263963814484, -0.036248755633803357579, -0.15049487241994483067, -0.17576006677669908651, 0.13361479731505626622, 0.98171649790430992599, 0.5859743330340886347, -2.7386479014971243373, 0.58984488211497365651, 0.9708149243837940201, 0.097180254614022770765, -0.21276587448327724528, -0.084489851429274315242, -0.0046485438127801181549, -0.031821558625338169923, -0.031458623421989413693, 0.043656283020448521937, 0.21347578325562746593, 0.045071215930977737152, -0.58484348475631309228, 0.16818636931019109726, 0.35990733310260120126, 0.16975329250408768833, -0.57663280256590321304, 0.068336012977870563834, 0.22614666690391188597, -0.06407646330420259384, -0.0049736886317327212367, -0.0038056392415555190603, -0.0083526323123122164782, 0.010661443146500203943, 0.073961345869556466281, -0.032363345868517509607, -0.17777565122096425365, 0.36920053108699307787, -0.44868356695119265831, 0.36806887134677140905, -0.18177987110158710937, -0.031338915319399168746, 0.081491395192760454624, -0.013304003563202427413, -0.0041466322645643659967, -0.0025820068085183264794, -0.0098364122859662273674, 0.0053655617954379581508, 0.077826521771689083362, -0.016795113198239119967, -0.20373758041014070064, 0.33995681640467650952, -0.36815956425652335282, 0.33888514810294018131, -0.20740892718933193195, -0.015602956870743216536, 0.084919626456325184027, -0.017822067611980203089, -0.0016796958680858353492, 0.017289356623417354714, 0.029498360646820730169, -0.010023879477864388113, -0.15092373860322366719, -0.24214214352010984821, 0.56620082619656564304, 0.85776524984136337526, -2.1291613467141421445, 0.8595160154169906308, 0.56084561264491683552, -0.26081316214951622001, -0.16072538778420919492, 0.063675374315005486836, -0.0038514486004952944859, 0.01505355473098588838, 0.031865314440710446053, -0.0013357953301199106556, -0.13598377211830023348, -0.24255273864927637262, 0.49736960013107395229, 0.82898581223035694432, -1.9774478723736246266, 0.83104146348621377172, 0.49358610177970924093, -0.25884041104536875677, -0.14548810627804389939, 0.067082382772948798144, -0.0071609610887356138845, 0.039663218231459559937, 0.031788082826652457047, -0.059605188321254158612, -0.1970258777749035739, -0.19007170261559638846, 0.73567856428067313068, 0.88360561742014098741, -2.4653096757119392635, 0.88609922248478012463, 0.72500349158375254621, -0.22523977284478860494, -0.21643400115489902502, 0.058165260710893479379, -0.0022456664756896589988, 0.0028805776475855707219, -0.0015251103969016542149, 0.016949017406846313832, 0.062876658250125894156, -0.11053638781541787195, -0.18101994707410445917, 0.49534672311535632439, -0.55896749090260144932, 0.49110625391482404245, -0.18908786957312936905, -0.111842050753288208, 0.078169310424112820046, 0.0089065287431436494098, -0.013867055853042927621, -0.0041231639242615136798, 0.055761209925563715029, 0.058928574671570074783, -0.16481254260326025363, -0.32592319385177370705, 0.57041465845274563762, 0.83126619129197620861, -2.0183482120263565029, 0.83569991073188720776, 0.57430934454205873063, -0.33339336104064248856, -0.18017279466502683638, 0.11438558137153206906, -0.0034838370747937758183, -0.0094538396974296994013, 0.0029750588185639042019, 0.034703569924519873935, 0.060127530371210394233, -0.075079475731727698951, -0.23812296053143722219, 0.33099232048725518718, -0.20756771106044913378, 0.33099388688214381737, -0.23759386056114961683, -0.065845895179911739792, 0.069918841036578263659, 0.0077800155996456676899, -0.0022363963070144536605, -0.0091297921214654360889, 0.00021096841316542783162, 0.029350017224851081277, 0.066039390030957142996, -0.056333418078986383093, -0.25959113811343220712, 0.30106992596680581808, -0.1409922261000413346, 0.30090321256906066294, -0.25892662748084699365, -0.046902453373864998187, 0.075320982288245061365, 0.0015300532337817567257, 0.012942986305983329026, -0.021357267040479011727, -0.033118861395491319688, 0.030608586664317607995, 0.14200820215258352919, -0.076291640237856034323, -0.35257599908508469033, 0.40338626545014588665, -0.21722521478763828773, 0.40286198285713725742, -0.34935250964216280067, -0.057944947825233086847, 0.16507460693542125307, -0.049389473853628969302, 0.012227291398438587186, -0.019928797439021712912, -0.03397915856135417384, 0.025301273834959461806, 0.14627372974155206098, -0.055956674503046810454, -0.37074343799064352067, 0.37061858139501202558, -0.15288058035035093085, 0.36989165028381010281, -0.36786121355929052301, -0.038432842731803869285, 0.16817441213821124335, -0.0529875621905875685, -0.010883189794541281437, -0.0057680886162047679752, 0.056634750071167083707, 0.082463444803935817551, -0.11232697627060432566, -0.38168001629766512561, 0.3237901844044269728, 0.8828339510542427826, -1.6721180105478483657, 0.88284350042856163387, 0.32402802385554124731, -0.38239886595899136301, -0.1117307944983548379, 0.12480395688613166172, 0.0087266200490028859871, -0.021259379307760159494, -0.0047142910406216169752, 0.0055800888436034019532, 0.031326469699291495385, 0.1078454385452864428, -0.22987463750887299208, -0.051029888991133948162, 0.30522407589637728931, -0.05390433725620596972, -0.22579515916611458959, 0.12389219132917755894, 0.023152279718641255446, -0.019029115217880414324, -0.0057035358840122618584, 0.03031434965360364775, 0.040782754705464205336, 0.012637893383804376171, -0.046834603470808186532, -0.076387759685941872267, -0.14716869604896887225, -0.065469571128182676989, 0.53325368771922454503, -0.5717294187609796019, 0.52783545656575547866, -0.066143133107003357352, -0.117823083776589288, -0.051991628715780231762, 0.016226667157295576044, -0.0095423784734120793011, -0.055994783082612846481, -0.10385801347277449913, -0.0072359096588315305679, 0.51860513489014903321, 1.2622654444365069004, -0.41985846144031607752, -12.024686135012133903, 21.592776015093878783, -12.037863483711769064, -0.27092884086361496143, 1.5043461081417628389, 0.040622199092062903636, -0.022274900631451516331, 0.035719556028124585523, 0.077124945127833466341, -0.016367479849671861397, -0.27869925699864833701, -0.14327293455510917153, 0.61580575511011348411, 0.14817214913663595555, -0.49790133436409361734, 0.18130114434964578196, -0.5308608840533015627, 0.14452010125269357244, 0.72803610847830091402, -0.44319265006740327095, -0.0018364803660589621489, -0.001002206291934904913, 0.0090954345278326693969, 0.01575491995906883741, -0.035273008870165732664, -0.16733669339773765694, -0.16001711989732220198, 0.79362888844252588161, 0.86444827267944934768, -2.6115179821944689564, 0.86743200827684285859, 0.76329273630623462044, -0.22997737553668790023, -0.10678842176202793524, 0.005996526742078628, -0.023009893162758067453, -0.022584224161207366866, 0.03387902406317513937, 0.085504959909066563517, 0.068603802916872069773, -0.073003773623628837042, -0.46876907444276622039, 0.11068105837668539004, 0.56272105446139197849, 0.11502089389258646213, -0.48690008565219877745, -0.074061365865411371567, 0.16653397055260268922, -0.013376318286820407591, 0.016248123338093417578, 0.044020120822633332369, 0.042262063949801198515, -0.087783135924503902237, -0.34242217267386371171, -0.15107894164331223119, 1.1553247935134076663, 0.81909859640547222348, -2.9442450411695353374, 0.81841657083693475361, 1.1158128094275472364, -0.24074777171980032975, -0.23412285613361549785, 0.025361986420791904351, -0.039984603769123411299, -0.070470100756865453051, -0.0044203203749630460076, 0.17506422635298485302, 0.34191941717770862885, -0.11721032038092803196, -1.2039007806639980824, 0.089517159005760668644, 1.5852907131918096972, 0.10994509898112450608, -1.1918687307812265708, -0.075044535632281403337, 0.37872116394109756055, -0.0048561494670402187077, 0.02497475144474273559, 0.020437809840243816722, -0.041103745010929865211, -0.15241507239128163076, -0.1757600662499887445, 0.13485238045350075864, 0.98494658235178811356, 0.58984488184146421919, -2.7383477951721868493, 0.59386008655513677912, 0.98588822547185472889, 0.095499749269377515182, -0.31740804641489617222, 0.0070013927506684088889, -0.012816918722511182888, -0.032857650592911145926, -0.030901953385943842223, 0.043551366995970938745, 0.21347578397146710349, 0.046393343163909056548, -0.58246991944434445276, 0.16975329207975478618, 0.35906844573094498507, 0.17062520157501873119, -0.57111478771689805356, 0.089752800748414310328, 0.13316670250033058043, -0.0069006010403038749515, 0.0010486587049568147668, -0.0021085929126116401827, -0.0083042533780870991106, 0.010529263451220280401, 0.073961345958622831676, -0.031912369879490337654, -0.17703407390722741832, 0.36806887126533821553, -0.45427878660281828527, 0.36516446202859215164, -0.17493399434556738958, -0.011045959904670917331, 0.048596277700297903224, -0.0064321784983368737038, 0.0017948714775437540729, -0.0011862020775106505918, -0.0099582886653574020225, 0.0051286962499799589044, 0.077826521774447710023, -0.016232235635632412396, -0.20287049429043974014, 0.33888514794679686082, -0.37388743345877906155, 0.33536675725824199867, -0.19991067711204674184, 0.009641778185090111794, 0.042602702934818596769, -0.003384685387538683126, 0.0034763456681599709704, 0.016960588935343117595, 0.028423444754700634418, -0.010326175791539287177, -0.15092373881919829426, -0.24289042229184476529, 0.56509878523489276425, 0.85951601537536992392, -2.1234979854428570611, 0.86174153654160701166, 0.53832273904938443554, -0.29671960009120701418, -0.04731226729114235835, -0.0027526770308575914704, 0.00039688945044578314561, 0.014634218764910604527, 0.031085735230635024723, -0.001458323457928795662, -0.1359837719834644798, -0.24341076621303819882, 0.49617588986365362969, 0.83104146374765619498, -1.9710559096580919647, 0.8326472949882981478, 0.46812753808609736117, -0.29374783297014739336, -0.027078485955501354482, -0.014523441712786614236, 0.013881742865959273772, 0.039184595408266686967, 0.026797130221870177347, -0.062548645743121786267, -0.19702587840603383307, -0.18851103617732123396, 0.73766322217245794413, 0.88609922226215587227, -2.4655616773514625883, 0.87720113111557695973, 0.70670362649584417269, -0.23005164829761473588, -0.13126733143399194059, -0.010004512710598070097, 0.0048688089380211647256, 0.0067741256240341236189, 3.8460608704980227268e-05, 0.017144437622558014039, 0.062876658225277298753, -0.1106297648978086523, -0.18247456659283603497, 0.49110625404724594922, -0.56608277187972855238, 0.48831938977925931633, -0.18058149945610332821, -0.087289902629433008263, 0.065889630924775982601, 0.0038298352825830027542, -0.018835395631754892759, -0.0043752321475829213104, 0.056734862361559459942, 0.059633645323022718698, -0.1648125432962102288, -0.32759641179205101214, 0.56815403898927951953, 0.83569991160850742862, -2.0026820096646922309, 0.84430966720465860931, 0.52446933440485510314, -0.42294156259521853425, 0.047767706528692457923, 2.0014574792764385208e-05, -0.0031318475957659739836, -0.0090118361950374970687, 0.0020990018035988719861, 0.033919883010632194409, 0.060127530584287047777, -0.074061061465744357224, -0.23659663384369233885, 0.3309938881299192559, -0.21021727115348068371, 0.33130235890190512515, -0.24025416258463691555, -0.066485944313959377538, 0.081580212970584556342, 0.00052996311219475537944, -0.00266395607280452704, -0.008788951531226850411, -0.00032454259101493809075, 0.028821455845916089944, 0.066039388800080561781, -0.055539343354519224438, -0.25839538011651802529, 0.30090321194894381618, -0.14349239556269355322, 0.30135042746606843789, -0.26025200197704440042, -0.043112923569819168457, 0.075230609205864856781, 0.011953693698383675154, -0.00079955749199788666084, -0.023037018642676671681, -0.031062994962294886159, 0.03194573888861065164, 0.14200820107535597203, -0.07696366314965437383, -0.35344657703571158436, 0.40286198267322653743, -0.21627575885741043282, 0.41042188952836394789, -0.33749737556937109417, -0.053361711262367592101, 0.093948565307896914067, 0.01073355599961541354, -0.00025657744502070260056, -0.021389176796748819026, -0.03201671009789168465, 0.026495268385610337702, 0.14627373007785865466, -0.056462920612789943564, -0.37148593145494734902, 0.36989165150865788156, -0.15267819556703210448, 0.37633190354790846399, -0.35492512103308498927, -0.027573471454706177663, 0.087821225576289629888, 0.00079649135096922056565, -0.015428830992043788412, -0.0028705296041517296482, 0.058817173658880110565, 0.083143678653160413283, -0.11232697606102309829, -0.38318793342780654854, 0.32041863573394141484, 0.8828435008934578665, -1.6618749543007418357, 0.89217242134685614818, 0.28235347094004781265, -0.45836353197938561888, 0.11237539900739969279, 0.010133735665712314752, -0.0078919820628259047118, -0.019134688927699594618, -0.00029421388333485695796, 0.0076371363890618522777, 0.031326469357235549318, 0.10615225462663492595, -0.23346696202191291603, -0.053904337405417841744, 0.31046454648228821727, -0.038395964042085090062, -0.22983233471842690054, 0.109609288988605863, 0.0082913343982925057779, -0.011486453809752874763, 0.0078412624941458335115, 0.031248050024317797213, 0.039346703964282893351, 0.012637893753397785868, -0.04549514716990131491, -0.076109619901473121728, -0.15210208123794718982, -0.077778405513773546387, 0.5278354564307540242, -0.5500057189904739019, 0.5756303900129832174, -0.096733359273783819932, -0.18912136579184063434, 0.0079094755084069788553, 0.0053148296022846128889, -0.010069820622050317666, -0.052701646330849991062, -0.10385801333912700328, -0.01144285251610402572, 0.51832030375570214975, 1.2784456827130032952, -0.39308617627664382832, -12.037863483578954416, 21.482790849701196123, -12.090432980389955731, 0.014355783333522486103, 1.3950822258892934258, -0.010821605568820240675, -0.0054429871781461851288, 0.033504238605638662907, 0.072899955148233358537, -0.016367479697363924301, -0.27617629632591139677, -0.14231983828899658073, 0.61237265102801552441, 0.13370098899585255992, -0.53086088374232154319, 0.17972645971843290158, -0.34231327505455344529, 0.34960771639128435107, -0.054850160997728852896, 0.0030589070780655440787, -0.0030255132969262887518, -0.0028441992684133227641, 0.0073621374756057461272, 0.015754920243721864254, -0.033769849421862896399, -0.16678845855200341575, -0.16234019160280963034, 0.79022498665389062644, 0.86743200789089502933, -2.6016375777740803343, 0.89866069776816570602, 0.77172239921449881095, -0.38186019296712825355, 0.0075100619865847255407, -0.0047337067266072958838, -0.022350098105307136809, -0.020918510904732255662, 0.033879024201053369036, 0.085091196505989161181, 0.069994582122097040444, -0.067248994674759546286, -0.45852201468361958314, 0.115020893762502463, 0.52523719369151378178, 0.047244374806439566605, -0.47132723914915775065, 0.15941423477100311046, -0.0080687393751044562962, -0.0021217390464053063372, 0.016117348249915650765, 0.041279310266202712265, 0.042262064175996802429, -0.085212730272500258888, -0.34249870246979369215, -0.1594822674271152152, 1.1415828926933542231, 0.81841657063295492058, -2.9237539477293696955, 0.88404041307501657343, 1.2170285904232474028, -0.63959560183571595449, 0.013611970158319110571, 0.0031801975520482834471, -0.035866526353813529826, -0.064692875352125714383, -0.0044203198904238802727, 0.17053754127636255866, 0.3385213994782765301, -0.10870775186488153274, -1.1782270136944426042, 0.10994509852701324359, 1.5431412645571396247, -0.021051480386980753884, -1.3328426182257151478, 0.56563473626457538845, -0.0037552241953597905594, 0.0054104820609565422093, 0.020925517829060921726, 0.01642524923569014389, -0.041103744486893092902, -0.15049487288743165769, -0.17663822546366247934, 0.12627885897427820749, 0.97081492477376674088, 0.59386008628587882985, -2.6899447744056654841, 0.69250445616858369302, 1.0010555929967048794, -0.35977412089599658529, 0.0087644499640744371466, -0.0014137456210351955974, -0.015652764116901071723, -0.033077850812820122628, -0.030901952779313193731, 0.043656282222859894848, 0.21570626438880899367, 0.052437265545653678755, -0.57663280187029730683, 0.17062520108136666752, 0.35073266980350076283, 0.14351804511052926316, -0.66861882162643204719, 0.34265445279793954603, -0.0038049943164080112477, -0.0025361855504332374463, 0.0011681997940756324791, -0.002212113244826249478, -0.0083042533829614708502, 0.010661443413626910115, 0.074610974061827803339, -0.03213553421605964272, -0.18177987111919985397, 0.36516446181322514963, -0.44611290885787224791, 0.37500555273333019413, -0.23216765582815296343, 0.082398139979740159244, -0.0035152535555855015173, -0.0021890512430436962973, 0.0017463045284177181743, -0.0014452508563984299561, -0.0099582885900658044132, 0.0053655619127967388304, 0.078551050643561756703, -0.016310289104280107503, -0.20740892706618935248, 0.33536675703812546256, -0.36641082830465815379, 0.34811265196972857039, -0.25566485066619709654, 0.093763283503422292786, -0.0022993486246135059863, -0.0012536075880510166337, 0.0044478903975715064489, 0.016882697270224896607, 0.028423444502739703132, -0.010023879319298444751, -0.15164012350682359864, -0.24595874472275158773, 0.56084561259649534648, 0.86174153648931361982, -2.1137339439212206038, 0.88345791297464459912, 0.53004554274125426083, -0.36185240893549991403, -0.0012011498256460240858, -0.0022496842391189447556, 0.0015197113991073131489, 0.014873716168204289406, 0.031085735128956678425, -0.001335795285824830465, -0.13659120651786391121, -0.24543079096565267871, 0.49358610153901272266, 0.83264729513510127301, -1.9665237240793091988, 0.84847845664955889866, 0.46063762024743171164, -0.33056589782277279177, -0.0069626464382452381285, -0.0043377312229496301638, 0.013125972276423069837, 0.036200576143145250796, 0.026797129743496270887, -0.059605187807408151202, -0.19547486642925890332, -0.1934198776230435235, 0.72500349160834698381, 0.87720113092143481293, -2.4477029081454939607, 0.94733664355865443429, 0.72147598780949617758, -0.44026442718122660303, -0.0093029929223922870618, -0.0011609491770209899759, 0.0075080463146102440972, 0.0072297125783807999685, 3.8460660937992847269e-05, 0.016949017753721699886, 0.061921911294690032412, -0.11348606279472647773, -0.18908786986702530442, 0.48831938967259480577, -0.55319777881059861713, 0.50337912110106386443, -0.2545455600060939072, 0.033836230275185855132, 0.0062297491575648332798, -0.0058387451533674483012, -0.017133286682645878368, -0.002563133633284179428, 0.056734861833618449589, 0.058928575906913902016, -0.16553174322225902571, -0.32636818102397724406, 0.57430934311620207566, 0.84430966798601958523, -2.0192448502939521759, 0.80605085814418764301, 0.54994853255770348444, -0.36098211741340568981, 0.001854916035665828437, -0.0011161585126256464828, -0.0045885372492013264123, -0.0095108459854988565418, 0.0020990018008401906818, 0.034703570286797338906, 0.062334902814825392248, -0.071926836526656628967, -0.23759386170137056227, 0.33130235979331063012, -0.21271062430490972339, 0.31334451388799566818, -0.28407349799663811352, 0.075283061779108137457, 0.0016700883915984081254, -0.00064175334106549215309, -0.0038519179010462782295, -0.0090984422226054670335, -0.00032454355978417913364, 0.029350018215306257485, 0.067815923861060292688, -0.053537187455396886837, -0.2589266273849928357, 0.30135042714535981156, -0.14595311060426924277, 0.28613036899263749602, -0.30382299543678237264, 0.089337702012831893272, 0.0053054664926488184426, 0.0060562383032731282109, -0.0016995587811326098686, -0.022097798608809594845, -0.031062995842362276649, 0.030608587507779509013, 0.14135398768268461756, -0.074561528066791352143, -0.34935250984594712342, 0.41042188955676123241, -0.20618107809181260293, 0.39090810252756263976, -0.46250093047411688918, 0.16339036292589090293, 0.0045778503817573571247, 0.0057085264732318237446, -0.00108877356654965389, -0.020614283379262375584, -0.032016710107184910561, 0.025301274109499248544, 0.14563947217520592514, -0.054273633635945346676, -0.36786121471647270642, 0.37633190454304460415, -0.14283968786093725889, 0.36148513412137550649, -0.4762965286138238219, 0.17658502592116614993, 0.00078519579733961537317, -0.0037895758323075678063, -0.01218612846095096236, -0.00079558892363200861726, 0.058817173738132909333, 0.082463444917227096687, -0.11334068167272988437, -0.38372962350070982662, 0.32402802339468211867, 0.89217242171013133589, -1.6751659989881688695, 0.84668777947723050659, 0.27128296343093910581, -0.29016350641714439007, 0.0017215659677962910114, 0.0056778196140670955314, -0.0055089287448488359969, -0.017102014694837429365, -0.00029421411361701781118, 0.0055800892352345756245, 0.027669219407119322734, 0.10456287242767842161, -0.22579515924576371022, -0.038395964354142629593, 0.29877558207628868159, -0.071479740348504627878, -0.20137948757803131317, 0.11667617609673702761, -0.0084340550805995020966, -0.005091998488011837895, 0.0090623640972603957539, 0.031248052705986226196, 0.040782754377414853764, 0.015226718685803060371, -0.046112055284597254501, -0.084278130309049001379, -0.15815322086023519965, -0.066143132783021668653, 0.57563038952255718428, -0.52798607845591938315, 0.52386948096832097743, -0.29901639184706729768, -4.7660163983192118233e-05, 0.0079612901063429754311, 0.0066965521715538541686, -0.010069820578152680404, -0.055994783279666504594, -0.11079547790552049524, -0.017522091451701457543, 0.52922667165553416435, 1.343725474959503785, -0.27092884097719838188, -12.090432980296862198, 20.796052860293364972, -12.505951541591310061, 2.3652509042995246702, 0.00044788943131462559703, -0.012194494120043824006, -0.0060902641655708545168, 0.033504238751342202418, 0.077124945045728379722, -0.007916736243792020164, -0.27637478169407991846, -0.17219017962986646331, 0.56446407788524899374, 0.1445201009284987903, -0.34231327478768691641, 0.35270233862558697524, -0.96136506916480390927, 0.61216027064505695865, 0.0043365717388687517742, -0.00081085770916402948039, -0.0048045459623597816176, -0.0028441993300761972234, 0.0090954341891903781048, 0.018714661162168366032, -0.031428674572503971507, -0.17209257490602178353, -0.18385642434701612857, 0.76329273688217080807, 0.89866069754858313168, -2.5040381477173356828, 1.1455204531817080316, 0.065995205057695002782, 0.0030223875705880759966, 0.0039637820672831737198, -0.0050282024579102555656, -0.022350098130123327578, -0.022584224396179821553, 0.031980024919658726901, 0.089275413422956750864, 0.085485186334040411116, -0.050216437456137373585, -0.48690008530872969095, 0.047244374616850250137, 0.53563720134377290627, 0.052131819365397624733, -0.26607832591199742645, -0.0010138435064826888544, -0.0070389921164993240685, -0.002858407923797139788, 0.016117348203235480619, 0.044020120468229971544, 0.047191891329792262644, -0.084066818508679752586, -0.35834620614411477613, -0.19571852513496668102, 1.1158128095820096792, 0.88404041317557435775, -2.7216728216149830288, 1.1345933573068816003, 0.13639916326480766573, -0.00099389330958158197439, 0.01518694315575226661, 0.0052188267217548800145, -0.035866526170625225023, -0.070470101572567014903, -0.015097196846403730316, 0.1706568277221834351, 0.37253534999726323207, -0.053124762891624231764, -1.1918687302881094681, -0.021051480666064183861, 1.3716735655950973438, -0.062252585401696029777, -0.49364030194053015732, 0.0019773814141413647086, -0.0040382564025606771177, 0.0032157455410956603536, 0.020925518017777068602, 0.020437809214272132513, -0.036248755363267437524, -0.15532796146213628807, -0.20185746975619545118, 0.097180254281193798049, 0.98588822574880285998, 0.69250445594965248652, -2.6046557319938523634, 0.72070711172097101649, 0.46972246555052471484, 0.0060565026066929390086, 0.0033475505361439386251, -0.0033710357766535514572, -0.015652764192687219913, -0.032857651315563021766, -0.031458622701738762906, 0.043868196002035682313, 0.22142201834066360666, 0.068336012183415506538, -0.57111478720941266651, 0.14351804518777944164, 0.24528019471991019396, 0.034930385190782294036, -0.11503346568714012399, 5.1512435652590368544e-05, -0.0032227926440015399606, -0.0023934765918469432661, 0.0011681995488324499466, -0.0021085929809350201225, -0.0083526323671416942418, 0.008926249846786915576, 0.070353311314444125224, -0.031338915298690303568, -0.17493399418122113076, 0.37500555284182335303, -0.47987187965593181005, 0.35067500098478110448, -0.10576602050620861895, 0.00015176551930055302017, -0.0030785288119269250788, -0.0021768158543646037167, 0.0017463043323103404417, -0.0011862023112983602693, -0.0098364121861212686776, 0.0036798492915411096238, 0.07424514387022528028, -0.015602956904169993507, -0.19991067693941966255, 0.34811265201463614805, -0.40152578884891509947, 0.30266162950970554313, -0.09896227835834030262, -0.001316479606086070394, -0.001471587288105336358, -0.00070380361451708286016, 0.0044478901005305286967, 0.01696058927031919572, 0.029498360652048614267, -0.0079208298459751136367, -0.15376528440417419619, -0.2608131619903207854, 0.53832273899901272873, 0.88345791288196218183, -2.0347952404629285894, 1.1202773202062470403, -0.12917147193926850623, -0.00085423115748591742832, -0.00094532972791485324598, -0.0016322211900949038996, 0.0015197112618402068302, 0.014634219099491689586, 0.031865314145574724636, 0.0014378050101214944704, -0.13660361537426232714, -0.25884041104135896427, 0.46812753791886924093, 0.84847845681841360665, -1.8911425766805836091, 1.0779993213175218525, -0.15165817864617006649, 0.0005141612737747373223, -0.0076518248105261702646, -0.0056474564898552851583, 0.013125971927792733432, 0.039184595751955111909, 0.031788082854696524115, -0.058087004118872874459, -0.20849442929842618155, -0.22523977244682999443, 0.7067036265947357343, 0.94733664335064093898, -2.3377339805697920916, 1.1362568310982235076, -0.027644008549352089288, -0.0042651473683819364491, -0.004224886168953142021, 0.00024502638166657157888, 0.0075080460489925886225, 0.0067741255322925787025, -0.0015251105983476792236, 0.013605511017988833333, 0.057891716759834677131, -0.11184205051560050515, -0.18058149941523882198, 0.50337912125038952738, -0.59100268662601007641, 0.51195116237406534943, -0.21066235684279893681, 0.0020528627468318683992, 0.0034575977639608032263, -0.0052745284154712543043, -0.017133286790975980202, -0.0043752314586931444901, 0.055761208494090303001, 0.06526687520561620659, -0.15424850369256609683, -0.33339335975216999319, 0.52446933341060208988, 0.8060508585508965318, -1.9111953351936536993, 1.1759558789658468569, -0.2058678527878391451, 0.0022466270710936354915, -0.00053121058224093412742, -0.002115709458771540194, -0.0045885372184102770146, -0.0090118359055396980717, 0.002975057683232124224, 0.035677116517523407846, 0.063798711675922104014, -0.065845894957735531805, -0.24025416326694781777, 0.31334451451576733616, -0.23687695819951162801, 0.32060461007727408811, -0.18109805282952198824, 0.0017399462276405271155, -0.00020599782260954956938, -0.0013990230976262832494, -0.0038519187998850043346, -0.0087889508679915052947, 0.00021096863529706383344, 0.030029050669831285764, 0.069374757624977800852, -0.0469024524371243548, -0.26025200203316473102, 0.28613036863482332883, -0.17286663980762254078, 0.27277671553772658841, -0.16764883174182659831, -0.00017291751282572028229, 0.0053180054301028385419, 0.0062674981066532329393, -0.0016995594583466438511, -0.023037017800139208273, -0.033118861482269633389, 0.029892484011093622964, 0.14706724513022098599, -0.057944946919144937891, -0.33749737581069072512, 0.39090810217302829344, -0.30969749817977187956, 0.37709333124665356252, -0.19408847172441928697, -0.00023515637941537767932, 0.0047650413222037795238, 0.0058515228367001107315, -0.0010887733907079511463, -0.021389176297681555816, -0.033979159827926722071, 0.024232177061798504392, 0.1504882581050840662, -0.0384328424832406923, -0.35492512175383983397, 0.36148513457755082623, -0.24590426598762354926, 0.32631353387995870596, -0.17801866319388298665, -0.0016063181184528402268, 0.0018842772855800697489, -0.0024773228659203017261, -0.012186128420803462452, -0.0028705295240957417677, 0.056634749571568832049, 0.086332642017349914276, -0.10179911275311950836, -0.38239886592384031427, 0.28235347059154153326, 0.84668777986889831588, -1.6042061160901057715, 1.1951867180204409635, -0.36261323599895084469, -0.003358304857756211001, 0.0059023439747296355465, 0.0070500333864332610789, -0.005508929087359233645, -0.019134688593021385861, -0.0047142910868692329551, 0.0027495091187480979675, 0.0345815461776676214, 0.12389219159524002267, -0.22983233445900161374, -0.071479740636056374181, 0.31418327913481536307, -0.17480700402034868302, 0.019948425211226698123, -0.004582827969059582833, -0.0074172909468447545053, -0.0050919945873390522542, 0.0078412583422098423164, 0.0303143496355711764, 0.04086671836592706275, 0.015931200810692665304, -0.03757420785403272484, -0.057618182107509914591, -0.11782308399116303566, -0.096733358968297197822, 0.52386948083026951828, -0.46144923527669606811, 0.18057305364039036299, -0.0029977573220535242392, 0.0042194719924432003491, 0.0079612900597470451919, 0.005314829570561010845, -0.0095423782955421764668, -0.054252708980552469265, -0.11903869177635841015, -0.040572057609247300092, 0.53970643629209680281, 1.5043461081334017493, 0.014355783310954375528, -12.505951541579092279, 18.146760928945148805, -7.5214006970511952943, 0.0033076569508734658676, -0.0048258621040769243693, -0.012194494049593369508, -0.0054429873206218275239, 0.035719555993294023477, 0.083800182870658310175, -0.0072781376557392819457, -0.31161386059813622751, -0.21123524985915306806, 0.7280361086465467757, 0.34960771613023139404, -0.96136506902424312404, 0.11923624073305835736, 0.19610897169080901414, 0.0025126918569940764545, 0.0011228889814743773026, -0.0008108580211610766017, -0.0030255131561078506784, -0.0010022061661105518063, 0.011060054172657811378, 0.021349851129950043332, -0.032563693105013800932, -0.19435650166125623128, -0.2299773758106120336, 0.77172239938677367199, 1.1455204532559630781, -2.8190742432542728046, 1.331154811419163142, -0.00035083945528977347229, 0.003471540344918848213, 0.0039637819291322359816, -0.0047337067675740970157, -0.023009892917634618137, -0.024050997355242419856, 0.03498036961712736731, 0.09753991719135202132, 0.070590412594509058253, -0.074061366275516918312, -0.4713272387958871712, 0.052131819323845571268, 0.82481288399707608505, -0.49195740474223009775, 0.0020675038090736024346, -0.0040406928848893287204, -0.0070389923560138841635, -0.0021217388908074661692, 0.016248123625308318746, 0.044171220161937925497, 0.049769810669019177995, -0.079950697293191347326, -0.38044106369194857775, -0.24074777175907086102, 1.217028590214639161, 1.1345933575246356373, -3.4254308127359105463, 1.6842684179007392231, -0.0047803783502566024252, 0.0068066959612040327937, 0.015186943015920269934, 0.0031801974735228633398, -0.039984603107550600876, -0.073999532256397437813, -0.0086498717734588888989, 0.18787993103104225212, 0.37853436196541007108, -0.075044535919825960946, -1.3328426179258174855, -0.062252585570171180274, 2.303002926039860121, -1.3042134498255146813, 0.0033783911949208040096, -0.0022982985989296609089, -0.004038256664627096898, 0.0054104822008430282845, 0.024974751765050769159, 0.023755662382481695927, -0.039766270109715826098, -0.16499680967775906537, -0.21276587427819121379, 0.095499749128209659998, 1.0010555933066886958, 0.7207071114118648314, -3.0015644831913608215, 1.5532132635173083735, 0.0019579275541851243712, 0.0038522142054312731932, 0.0033475499489538582121, -0.0014137451522822383574, -0.012816918271390795314, -0.031821559235875145066, -0.036330015783334827784, 0.036114440854560579675, 0.22614666738372796173, 0.089752800493938064696, -0.6686188218176899456, 0.034930385406499682432, 0.82983124078525460909, -0.4807043009242721876, 0.0022039058986410226329, -0.00062790006540925128679, -0.0032227930242939200126, -0.0025361850932652356405, 0.0010486588354862869804, -0.0038056395057799114259, -0.014847760289894790242, 0.004372092532254101202, 0.081491395314021095042, -0.011045960099794777109, -0.23216765588128132625, 0.3506750009747733321, -0.20709296836980817247, 0.032651150936569603433, 0.0020831866391214930341, -0.0006753568169693945902, -0.0030785292364431168094, -0.0021890508206425709353, 0.0017948717399006953663, -0.0025820071009337074511, -0.015999116642726053661, -0.001534642225968778969, 0.084919626412268661708, 0.0096417781172262809197, -0.25566485066710276097, 0.30266162945922014949, -0.11722086440914418481, -0.0047904414608909802117, -0.00049736407706681433263, -0.0010348372597651673156, -0.0014715873820482542559, -0.0012536072926049410432, 0.0034763454462863004925, 0.017289356551873500539, 0.033171299975490030343, -0.0020078249495922642631, -0.16072538791872281738, -0.29671959995534397114, 0.530045542815706483, 1.1202773200461737524, -2.2598296804844637542, 1.0249814844496751842, -0.00054093279410756935164, -0.00041934191028357377428, -0.00094532966068291231134, -0.0022496841338478113653, 0.00039688913709194894577, 0.015053555013021451686, 0.036083067750376000837, 0.0069587395127713638082, -0.14548810611799872783, -0.29374783287190736614, 0.46063762000899777238, 1.0779993214176213367, -2.0789118736130900444, 0.93037250524964709708, 0.0028120586344037496775, -0.0036146876612754680856, -0.0076518250222700921401, -0.0043377309467156994691, 0.013881742892428749689, 0.039663217991646279592, 0.03110975234889643351, -0.062924727551626258304, -0.21643400161543696614, -0.23005164833632657517, 0.72147598804112411663, 1.1362568309749885298, -2.6720581063143509226, 1.2580254513691209173, 0.0011400440226452571938, -0.0028691851295000623742, -0.0042248865457506640242, -0.0011609486784604232303, 0.0048688088875736574429, 0.0028805777002689257962, -0.0072688431689501096372, 0.01337139388678398183, 0.078169310332992389401, -0.087289902649277439917, -0.25454556006527834233, 0.51195116233906545755, -0.31497372281307794983, 0.058615639274580508267, -0.00075430618124165097101, 0.003317608137534073208, 0.0034575978295550134113, -0.0058387451278085417475, -0.018835396083416545282, -0.004123162940267788637, 0.064404318583807323995, 0.078139646597000400607, -0.18017279533742888709, -0.4229415619412427163, 0.54994853221833261969, 1.175955879010564864, -2.19313405002209727, 0.95458796724848338222, 0.00093869457017447437489, 0.00052609673067177979353, -0.0005312106345516117269, -0.0011161585412434670526, -0.0031318479456021309885, -0.0094538386352269743185, -0.00067741609444367956866, 0.036718992571436806505, 0.069918841217190386139, -0.066485944135720745352, -0.28407349851756547032, 0.32060461048839483666, 0.022945326577873362273, -0.086494131157646800268, 0.0006106234660924618517, 0.00049384554362992484806, -0.00020599850657519963668, -0.00064175253350108162759, -0.0026639560844139969359, -0.0091297927873596196324, -0.002933460730664316958, 0.030834209506126957628, 0.075320981211488641227, -0.043112923208622908056, -0.30382299510139604282, 0.27277671527569508081, 0.099532992708962209072, -0.11751239160475132617, -0.0026215088431013160164, 0.0016931583660013090672, 0.0053180050487061543418, 0.0060562388761841194282, -0.00079955780917104505787, -0.021357267333990972991, -0.033273997713211038441, 0.032898963101627459737, 0.16507460600367140913, -0.053361710821482742639, -0.46250093023490879141, 0.37709333105992676183, 0.13602250628873840954, -0.14982827613088975371, -0.0022853002772740523343, 0.0014217678559748348453, 0.0047650415752373137535, 0.0057085262465500687806, -0.00025657804408326951445, -0.019928796261713312837, -0.034577182743857931557, 0.0259537103437979845, 0.16817441245358488722, -0.027573471361042205252, -0.47629652897022778379, 0.3263135341543493273, 0.20729887706341842968, -0.1787587553731691814, -0.0013579211607406690369, 0.00090442179262433940613, 0.0018842772603453038661, -0.0037895757647128137935, -0.0154288311294891161, -0.0057680882081462166167, 0.062652755028880152688, 0.10398504850701544011, -0.11173079425510937246, -0.45836353193291973218, 0.27128296303743515416, 1.195186718284348526, -1.7176710793368799024, 0.68222115845100284037, -0.002623844367932534876, 0.00094551097657833891273, 0.0059023437729470908802, 0.0056778201118064010322, -0.0078919823541215780582, -0.021259379537021321632, 0.00023007546160044219118, 0.015614881984657304204, 0.023152279590941786391, 0.10960928899981385587, -0.20137948745055508915, -0.17480700421619865814, 0.46562571903309096788, -0.22042401689456736502, -0.0020135076206274052542, -0.0045828273711678363042, -0.0084340568904180401866, -0.011486452359222737787, -0.00570353573565758349, 0.019989156832429570698, 0.06069189804527973936, 0.081782202473708973112, 0.039388378872058316116, -0.051991628734536832201, -0.1891213655915915115, -0.29901639206461250176, 0.18057305376579452227, 0.21268369476763432369, -0.00084784276982338626239, -0.0029977573420837079141, -4.7660120970144167628e-05, 0.0079094755055934152371, 0.016226667094402978664, 0.013465698408645736944, -0.028866758005671764448, -0.14701071347665481404, -0.2663249888703330126, 0.040622199120883814571, 1.3950822258792572317, 2.3652509042959404262, -7.5214006970502964577, 4.1734184928212210153, 0.00036186573553359063738, 0.0033076569549321402711, 0.00044788940282639429181, -0.010821605523418853165, -0.02227490059921199278, -0.0011569210250537453107, 0.08164800248155618112, 0.13357834806387938742, -0.09305260274040022539, -0.44319265010912406444, -0.054850160908380601876, 0.6121602705808540934, 0.19610897170755084429, -0.41196382574202455862, -0.0012623152073533046302, 0.0025126917574008300442, 0.0043365719057888866064, 0.0030589070403495358108, -0.0018364803616378838451, -0.0066319587573758871299, 0.00018500441715415543603, 0.027057589913391941888, 0.030922975495735458007, -0.10678842175945929549, -0.38186019299931922566, 0.065995204998512052574, 1.331154811466658483, -0.98530954735740106099, -0.0016258417813565248858, -0.00035083951491480364591, 0.0030223876702574438077, 0.0075100620100810892496, 0.0059965266521723814899, -0.013641731404215222748, -0.045521868903678076868, -0.035814951351262704071, 0.054013649308325195619, 0.16653397071748837766, 0.15941423461654474769, -0.26607832592938784888, -0.49195740469461735067, 0.46783129643612886284, 0.0010418842130124599671, 0.0020675037558719302003, -0.0010138433640343169628, -0.0080687394607734577029, -0.013376318374777653245, -0.0031938573643315523731, 0.037854134656777976842, 0.10264389330190615279, 0.082491166276012634073, -0.2341228561257449603, -0.63959560174253760056, 0.13639916314012121012, 1.6842684179466076433, -1.1626053692184288302, 1.6521786947950407306e-07, -0.0047803783810358253456, -0.00099389322746753984601, 0.013611970163446264481, 0.025361986209577601359, -0.00010767908452426120356, -0.077885654228851311509, -0.13463130124635011575, -0.013933815272755216419, 0.37872116402027522497, 0.56563473614873160944, -0.49364030185045842947, -1.3042134498533999309, 1.0630247804092323172, 0.00014459135348520715456, 0.0033783911893935873044, 0.0019773814760815624711, -0.0037552242317607782522, -0.0048561495587948052752, 0.013825250036590181271, 0.046864104028127170909, 0.041415049806401683929, -0.084489851464231630063, -0.31740804637757635831, -0.35977412108304851479, 0.46972246579798848654, 1.5532132634121231796, -1.3921413095640826629, -0.0021404012389311471119, 0.0019579274676172985468, 0.0060565028821852658764, 0.0087644497694405222588, 0.0070013926063270002717, -0.0046485435948813618248, -0.033859072764062345118, -0.074367079473828687219, -0.064076463395813243462, 0.13316670257340104677, 0.34265445283661349896, -0.11503346576531284229, -0.4807043009018844848, 0.27291746503257202638, 0.0029337621754126261796, 0.0022039059210449069362, 5.1512483324506316753e-05, -0.0038049943808374513295, -0.0069006011362340293966, -0.0049736885506925874317, -0.0024660530116080809845, -0.010062681669046133134, -0.013304003588524191196, 0.048596277846766025577, 0.082398139841229192659, -0.10576602034347523662, 0.03265115083227494025, -0.025749663558392733853, 0.0025363760549372660162, 0.0020831866358489224515, 0.00015176562101622281548, -0.0035152536363314908392, -0.0064321786368700369432, -0.0041466321679209851986, -0.00079019626978085646257, -0.0089977013109546807174, -0.017822067533862385319, 0.042602702962287491983, 0.093763283395299268363, -0.098962278206744414, -0.0047904415462724364572, 0.0010532714733344466628, 0.00021975786375312756542, -0.00049736408184669774445, -0.0013164795521870829072, -0.0022993487445250658345, -0.0033846853224827402884, -0.0016796958235092823968, 0.01132186560402661038, 0.043737740905582724038, 0.063675374346998048214, -0.047312267353660335256, -0.36185240893928227734, -0.12917147186555238969, 1.0249814844141527104, -0.60146356230548181365, 7.9626837195064995413e-05, -0.00054093277216866390671, -0.0008542311879994103109, -0.0012011498734830386003, -0.0027526769190984331191, -0.0038514486848848818179, 0.0057139571569064162224, 0.038876569899330651325, 0.067082382687259328735, -0.027078485983908644891, -0.33056589770555205865, -0.15165817871022008179, 0.93037250526073889123, -0.5275492189453508951, 0.00045512656906123333559, 0.0028120586092342586201, 0.00051416136254678172519, -0.0069626464978526040467, -0.014523441796892102434, -0.0071609609925229007407, 0.03038107188257118535, 0.079327195947362746931, 0.058165260884084295234, -0.13126733140303212255, -0.44026442728478476507, -0.027644008485991286572, 1.2580254513536817118, -0.81145104595816908155, 0.0053934684120263143292, 0.001140044046903632884, -0.00426514733120926786, -0.0093029930016832376033, -0.010004512674564093946, -0.0022456665713948067986, 0.0082787537873095940616, 0.0075038858089763729384, 0.0089065287962656492643, 0.065889630979719435167, 0.033836230145039405959, -0.21066235666677005089, 0.058615639168287339555, 0.047599506461778108568, -0.00094720800870100221435, -0.00075430616919899037397, 0.0020528627117322333952, 0.0062297491603828438617, 0.0038298354090007965221, -0.01386705614116409864, -0.030349312695059013634, 0.012263796007262758522, 0.11438558153546075913, 0.047767706336883195006, -0.36098211730883561499, -0.2058678527927313151, 0.95458796723990635424, -0.53387460957078380908, -0.0012574764431572814805, 0.00093869451751994552584, 0.0022466271446760547291, 0.0018549160209379643265, 2.0014714278954290512e-05, -0.0034838374213583219764, -0.01226921265713669211, -0.017870449790054573902, 0.007780015518852520344, 0.081580212952118036362, 0.075283061957084979054, -0.18109805306288975779, -0.086494131057400822082, 0.13615443486494863135, -0.0010808801182235700314, 0.00061062332121566856389, 0.0017399465939473002241, 0.0016700881319117854423, 0.00052996302953956225118, -0.0022363960459333468399, -0.01053919073284632392, -0.017879100052638607216, 0.0015300536121386171078, 0.075230609041000012782, 0.089337701912018313566, -0.16764883163733637628, -0.11751239162617532763, 0.14966708868331854432, -0.0024951843076755643967, -0.0026215089717417042886, -0.00017291721082993513235, 0.0053054662201540972108, 0.011953693751039475013, 0.012942986505126676672, -0.0063142984685813181508, -0.048110868371087960049, -0.049389473566384220482, 0.093948565064589845131, 0.16339036298057690866, -0.19408847174775700761, -0.14982827609097476573, 0.17008001123299712476, -0.0021201046111225568987, -0.0022853002948805063894, -0.0002351564054378731171, 0.0045778503810968915189, 0.010733556236358840486, 0.01222729106318733161, -0.0050412117060081654274, -0.046164753513413668562, -0.052987562400378988847, 0.087821225614555908923, 0.17658502608968859593, -0.1780186634095620446, -0.17875875527301079981, 0.17776031122019489428, 0.0011208263044430869567, -0.0013579211532698080844, -0.0016063180987832062496, 0.00078519574202328920078, 0.0007964913979917775293, -0.010883189912570076507, -0.025012970142541592067, 0.013291670710234806951, 0.1248039567660066268, 0.1123753990632679195, -0.29016350630516407616, -0.36261323611714779691, 0.68222115848373565683, -0.24136096188791550388, 0.00138889264250745075, -0.0026238443673170107513, -0.003358304780650485117, 0.0017215657718926233744, 0.010133735760230128467, 0.008726620168441352049, -0.011727564303842556367, -0.026274214246835313202, -0.0190291152100682745, 0.0082913343457711938256, 0.11667617609021785352, 0.019948425333881456001, -0.22042401696735791639, 0.11662050999430902121, 0.15663971579156782399, 0.15317022310914105154, -0.36672201316201902266, -0.093987714704801050836, 0.096365054775942760035, 0.055407226919363380957, 0.010424440085090500713, 0.0022865904286320857912, 0.0048207492484521642651, 0.003404324169270572624, -0.00051765276922574685742, -0.0034476795220157461022, -0.003547868165618851561, -0.001139703630628873714, -0.41196382572064554894, 0.19610897163796611742, 0.61216027064340960973, -0.054850160886275985128, -0.44319265021092252299, -0.093052602651995192096, 0.13357834809066868043, 0.081648002346541362351, -0.0011569208507682825129, -0.022274900715206172319, -0.010821605537146056566, 0.0004478894687086445222, 0.0033076569298462017363, 0.00036186573486013401272, 3.2818432734618729896, -6.2647946700329155689, 2.5174384518018348977, 0.8797369237428876998, -0.12403098868599485294, -0.18801759390591799015, -0.07768707799992659313, -0.013344020409160660054, 0.0079657821325304720844, 0.010508585572695797877, 0.0051531508436467275691, -0.00061011677088392089058, -0.0022666909379603591458, -0.00011622875602963050055, 0.32815969934557359666, -0.28090700466650725398, -0.20156510761452195446, 0.064819094815062275328, 0.073793398726161835155, 0.0388745044353178415, 0.0023162006465260575466, -0.0098986283842645468589, -0.0022316115305920781846, 0.0011868172115439507095, -0.0013981899805398565438, -0.003297213028195996546, -0.0019382990362857776493, 0.0015466051179465326905, -0.62063735768149874694, 1.0587334930513667697, -0.18116792166312120638, -0.34307307241492090988, -0.00062485272365169289028, 0.056698828557935135319, 0.022718400121075232878, 0.0042398703687896435385, -3.2574835890488630205e-06, -0.0010259776009279510285, -0.0013697549317697938539, -0.00085067173001864850555, -6.9219651269865770994e-05, 0.00011120508133334982138, 0.67717832122387910943, -0.80828905266992445711, -0.27830742971770666694, 0.31097622383373069166, 0.16971045960797973007, 0.01614895065623568951, -0.032856633831233744703, -0.040012814573645347216, -0.016466056384260466944, 0.0050826605266032628613, 0.0085594233966762404975, 0.0028524334249325706062, -0.0013007751680569801158, -0.0012797122053532648984, -1.8144508650372503933, 2.630592359413987058, 0.22563641087268260232, -1.0465229938437534685, -0.3534546625077055193, 0.16647979882086591519, 0.16900639624485597201, 0.040959844264026865068, -0.020707299096471506322, -0.021100713853026316358, -0.0059578316730487740696, 0.002417089133449026872, 0.0030779781411148813761, -0.00010010178663234661853, 0.21667253594600577316, -0.3872606710050354728, -0.12514139893966913508, 0.30518686805702222031, 0.14171532286439980641, -0.057757179354107436586, -0.082009723969828474122, -0.038922302990361423713, -0.0014463646085425177056, 0.012778064277895649636, 0.011135020812280406813, 0.0045333342496584223824, -0.00030142108305664512503, -0.0019586054207497045898, -0.87992108716183403683, 1.1337512730305436826, 0.082336165395794677635, -0.28560064529529788357, -0.080255402633939126877, 0.002914466574468226566, 0.00030202203285775963389, -0.0032115146654487698209, -2.61749368006439284e-05, 0.0028296843291696258636, 0.0044960569231702190995, 0.0040373541529901838323, 0.0016445769680858116253, -0.0016843231552624772941, -0.10924291372016679824, 0.23207203651506011499, -0.17742192198300130168, 0.00061415212091269280603, 0.073339222329291908764, 0.00067260843731335425225, -0.016234805399926183905, -0.0029956453037284168882, -3.2344115158056553194e-05, -0.0033804515560058863259, -0.0031824401684265512733, -0.00063226363034878509204, 0.0013668235103532688432, 0.0021701946566018183449, -0.11760043795822680557, 0.22937784130416280504, -0.15469659800986693132, 0.00021682030077243240589, 0.060945875512224988169, -0.0012348070941759355882, -0.01425827404733680942, -0.0031449122960811874868, -0.00039049603848757298823, -0.0027014293263446661174, -0.0023031665881774691235, -0.00024568788511475782126, 0.0012458904587462774419, 0.0017446023238046276874, 0.17605209839937935312, -0.26573774407844918954, -0.06192727178635813734, 0.16898525479102133318, 0.051867435262248247352, -0.036126328666930267175, -0.037419477021317884557, -0.010116564332173182822, 0.007696895407086278354, 0.0077980003190384894474, 0.0024683492412107882127, -0.00061197133561946158355, -0.0010676609686260788747, -0.00041803738947157289456, 0.15761914069149482454, -0.24323184326134508759, -0.042220428305067940478, 0.15158875928080672857, 0.033048870831401490344, -0.033385335688715093994, -0.028103875587388933016, -0.006570171741112334346, 0.0066869505973120935485, 0.0061436151710146458607, 0.0016449600625146677032, -0.0006970219513630964087, -0.00085895345452470311335, -0.00027760837795615106579, 0.23461991001245402488, -0.31220889290234610236, -0.097321762042062623377, 0.128107405805777258, 0.085822850797908620257, 0.027224450629416481473, -0.030847131813994788097, -0.039174876916946863026, -0.011020752751697435179, 0.0073194522778712916775, 0.0083024774471397323056, 0.0031438637102491613466, -0.00050443090276708021404, -0.0012697102009455848905, 0.10317763297895798935, 0.044689496910588231982, -0.22447093923846070385, -0.027153704643287784881, 0.067754813451493423049, 0.044773229644956252793, 0.012183894036456978729, -0.00087907134387548054033, -0.0033679296551854475371, -0.0056507213535753643419, -0.005839191159300547232, -0.0034988154193379009463, -0.00016799673709217067446, 0.0031294228954392207948, 0.0084609295705947145522, -0.19452951024786910295, 0.17979928070155906661, 0.15454207520000817766, -0.10693515517635909795, -0.075107850216935362342, 0.0056877962849252094654, 0.016474389896879524514, 0.0087077031314566778059, 0.0024740258691537883595, -0.0012176728472905922265, -0.0018055861069054797777, -0.00053434089037944249457, 0.0007606266201517602717, -0.12287864976685398954, 0.3182520852383043386, -0.21546043024751160289, -0.032215610536836913891, 0.064355717853014443586, -0.011320722590708112709, -0.012906291756215035327, 0.008814896467949313863, 0.0075871413751454220425, -0.00036380383933582230698, -0.0022790090773709137384, -0.001021690673569095693, -9.5681429459032633976e-05, -7.9773295543991755647e-05, -0.11843490054936992029, 0.30445719450120417582, -0.20164456031312022466, -0.032389606336032432388, 0.056086596490141950311, -0.0096925755069887879062, -0.0092850483963399562531, 0.0088022920666337642337, 0.0065056042927944170845, -0.0006857016462005456961, -0.0022132061140723436991, -0.0010129406083292118081, -0.00011928408992974005149, 1.1285640977724925403e-06, -0.028110257223234874635, 0.13822339516034887197, -0.12992027309545375879, -0.0037105089704911574378, 0.028850307637223151364, -0.0074399716550489292458, -0.0068745464626265495242, 0.0064889151241188493605, 0.006699172227567524486, 0.00014125588306293312231, -0.0015082265001920495528, 5.2879690155684859563e-05, 0.00041478135310086725572, -0.0012664816865437069412, -0.045047841912051252156, 0.14256800726275167812, -0.1058869071973307513, -0.005388034188532439539, 0.016837403163707600345, -0.0082789209685817828632, -0.0039565509420196112722, 0.0066522315974736978897, 0.005901422820372761828, -1.5097500352301821856e-05, -0.0014095051167449022898, 1.3534847777298133977e-05, 0.00037022939044553985436, -0.0010692747459795624593, 0.22733329974831270603, -0.25778694861439671593, -0.1553961507375108686, 0.15045226335866990208, 0.087529709875859632029, -0.013573466207342297155, -0.028447715177925221586, -0.010750454035581752388, 0.0015339875773181736295, 0.002239562281529423101, -2.3398684490091958523e-06, -0.00051528140983886881743, 0.00019207502902982980268, 0.00063120110706189110766, -0.8308288337707331328, 1.3476462016613424133, -0.15585865826974526493, -0.44973037566962259515, -0.0317225626040565159, 0.07297776755454889186, 0.041370768322021279906, 0.0081727129686083692384, -0.0060330194633015209876, -0.0062030016075645018347, -0.0021620872156821123976, 0.00062424090530453725436, 0.001282707931903761308, 0.00027689919263228007437, 0.15317022273592306725, -0.5592255665653405039, 0.75244751832155398752, -0.17245059907323417869, -0.2681739833192481326, 0.015793395385054961877, 0.05958710922622777878, 0.022501379094252677016, 0.0017336717735798588044, 0.00061699812711763293849, 0.0018883694744925656663, 0.0015740761319777332033, -0.00058852160159615185368, -0.0035478683493624756096, 0.19610897168602894891, 0.11923624082057887597, -0.9613650691984485519, 0.34960771623741054848, 0.72803610877095448206, -0.21123525011973665055, -0.31161386060156059941, -0.0072781372456168025656, 0.083800182388653041143, 0.035719556090798881831, -0.0054429869628233075451, -0.012194494517331620417, -0.0048258617861752072659, 0.003307656841305229873, -6.2647946699699028628, 15.632090378758586979, -11.80063399213148756, 1.077003275443723318, 1.2808111205387073817, 0.22395025343505936544, -0.08084308612397503313, -0.069029382933298591429, -0.026322979483323735023, -0.0066343506838075589835, 0.0026889109964138178596, 0.0058434878817894457884, 0.0028577343408968847345, -0.0022666910426512297018, -0.28090700460485457102, 0.46284697798337209873, 0.082324482761386971741, -0.31362679237192780946, -0.02826753354954211761, 0.024236232855814349857, 0.040797126202699131314, 0.025702571265909263803, -0.0046248303469270295521, -0.0071809661912848009313, -0.00041235533397814574897, 0.0015949957327659163974, -0.00020569572912266126949, -0.0019382991236760859638, 1.0587334931643201941, -2.3945592471544849822, 1.289173043007698638, 0.4879442482403788528, -0.3765645594141481256, -0.11332917948708159939, 0.024749900843171350473, 0.022440242293679777547, 0.0062679354284184054386, 0.0011925280097518963299, 0.00048625248834465063926, -0.00036239502766572875966, -0.0010312553108364173882, -6.9219668296800917107e-05, -0.80828905262887529304, 1.4786690016471015774, -0.14716770841346038523, -0.79684509728278740681, 0.065050263000415678438, 0.16477985876948653132, 0.053721125494055527172, 0.022775545373356308881, -0.01267603066129341198, -0.023161590855754452672, -0.0083458286090569294702, 0.0042880338077364386212, 0.0049953851406938174634, -0.0013007752695871461178, 2.6305923594583235925, -5.3349056151508653656, 1.6874786197420201805, 2.027523641266682386, -0.41152260603635837377, -0.64517561803156497913, -0.11744434888313519394, 0.10706862085612070534, 0.077670198379386254373, 0.013162754868345782899, -0.012643682270274739926, -0.010085085592752722483, -0.0020613995249230770111, 0.0030779781571255090955, -0.3872606709270695613, 0.62453528749287745558, 0.12915438240693619121, -0.59283635354173558341, 0.031776421976070397257, 0.21138595387147515736, 0.052536661921519241247, -0.031451143942802642361, -0.037434250264209795178, -0.017326610250272982128, -0.0003206324812370012602, 0.0068655040106502468814, 0.0055035824847428646903, -0.00030142096502493978451, 1.1337512730985563891, -2.4034967445933492769, 0.98215376173041668562, 0.62918623939427698044, -0.19400888571431948448, -0.13183642343848539502, -0.016174750022994996423, 0.0011389717943458778273, -0.0025217313209359324896, -0.00058987407838474151169, 0.00019376560439161250797, 0.0010116671107821201877, 0.0017462457856192547655, 0.0016445769839977717429, 0.2320720364437889871, -0.67603455124762035222, 0.64051528361557330538, -0.12554059841541567732, -0.15267844715360456842, 0.070126439983457716787, 0.0337417135195415413, -0.014131359191547310586, -0.0095331028374030565148, 0.0015293671200753162359, 0.00048694806565005002621, -0.0017557661538432638333, -0.00096336119140073644095, 0.001366823695067197024, 0.22937784129814400846, -0.65029089751287660892, 0.59403824005736083347, -0.11097763175635949606, -0.1297429383254757318, 0.06057364752833439836, 0.026208753875271672412, -0.012163878378766484686, -0.0080710759304957798571, 0.00093275406646747892163, 0.0002002914392018983851, -0.0013758469876953898989, -0.00061810598002824787729, 0.0012458905127183546833, -0.26573774411429440523, 0.48411692924815741268, -0.013497199273268277453, -0.35783977299432045971, 0.063672998284980866779, 0.10152090908868906249, 0.021529090552554700638, -0.018922774966627954607, -0.021028437597115492985, -0.0036658525638872310054, 0.0048125516486565607377, 0.0037551205449345322063, 0.00062051788699811927675, -0.0010676609524221630614, -0.24323184331380945378, 0.45852147036647233902, -0.044021353921982364632, -0.32125431741553334142, 0.083335155544199054911, 0.082888864104807696842, 0.0095903132559842490384, -0.015466063053822046158, -0.015983459420472856394, -0.0023623368935026122124, 0.0041698298624669779405, 0.0029522934411318761845, 0.00026102201718359137805, -0.0008589534150845297639, -0.31220889289937725497, 0.55902569524904344167, -0.061961737160742290376, -0.28072009400838748849, 0.023491193355774892676, 0.025177487392264531751, 0.054173554849293231728, 0.027700806907369000809, -0.020470709819427012482, -0.021113180535390893566, -0.004154490958467430689, 0.0047720931990416744564, 0.0042975599081018145062, -0.00050443086415126665075, 0.044689496926132700472, -0.26266341633807932521, 0.4369930386730643268, -0.15342599700245918015, -0.12631551885653710676, 0.010592603003133168119, 0.037715485703239320925, 0.018312514202766624893, 0.00052922943149298455612, -0.00053804410456929517223, -0.0010922338391098308438, -0.0022826132420496426587, -0.0018858926963446695502, -0.00016799668407054154078, -0.19452951029690990215, 0.53599427132313848254, -0.39029357447116302948, -0.21279193076725094746, 0.3257591518851579715, 0.031398285181813791189, -0.091347312176586037413, -0.020990433851414521532, 0.0056101078597996694233, 0.0055560044205838245157, 0.0041187292122128113525, 0.0014556056611013229972, -0.00089182198723920249554, -0.00053434082515507795672, 0.31825208527682175008, -0.86607847382027003391, 0.77945063855821028476, -0.12157092310148179259, -0.19671150073028828564, 0.082860754366260447323, 0.025582802595067021673, -0.02402531833880828388, -0.0048179813694063794438, 0.0080873220926614679954, 0.0033562109232867583415, -0.0010939285016067566529, -0.0014368332579499010129, -9.5681463939776455674e-05, 0.30445719458509962196, -0.82361966085076288024, 0.73497862830651905508, -0.11439686022597388149, -0.17758779650626155266, 0.073858530404359704069, 0.019962524915580773244, -0.020756790710929992616, -0.0028654869161535549173, 0.0074644304237487853898, 0.0027445253805002890919, -0.0011056322655009929914, -0.0013165161869338661699, -0.00011928409773223580222, 0.13822339527315652408, -0.39899101737328906969, 0.41802710927556085974, -0.12035451132855876089, -0.080430820836355054748, 0.040570641504834407953, 0.014091159404465833024, -0.012912997343515900464, -0.0049030686507810913519, 0.0070459371327869944543, 0.0039760799622625599831, -0.0012812801404202848674, -0.0018368033137023767089, 0.00041478122637927266649, 0.1425680072931872211, -0.39014674850053593369, 0.37701616086313483045, -0.1004082071083928912, -0.058129272060771827735, 0.030694624484936837427, 0.0065307516784157909964, -0.010878156777161158281, -0.0032541952550977459695, 0.0064492447208938053979, 0.0034520516780906398795, -0.0011773486425090979451, -0.0016308157621618957472, 0.00037022943312484417246, -0.25778694869467855266, 0.397831393381045062, 0.14527803860053456475, -0.39848312264692825613, -0.0074691220563660827222, 0.11248775873937771896, 0.034500188232129011534, -0.011671491328091121095, -0.015668453379953744403, -0.0037875985937667385611, 0.0019332418407226232854, 0.0012780239930615458488, -0.00031573780304239946679, 0.00019207507221846611793, 1.3476462015510446424, -2.9898894969045066716, 1.4854743173422786384, 0.70588113261524776387, -0.42803978557933902716, -0.16931280588860903036, 0.0088971546576114574761, 0.030346715514370972699, 0.017542571747717988079, 0.0030113315882022585052, -0.0031624963135626605291, -0.0029028205980807463525, -0.00099547154727659876473, 0.0012827080103315550812, -0.36672201226008249719, 0.75244751594549752483, -0.72731819306473111109, 0.62165733379091792088, -0.10730222245044196416, -0.25103090828178697969, 0.0030214969034341068559, 0.048638989312896209471, 0.019356394996089088673, 0.0018128841476871863318, 0.0023398311744771181989, 0.0035439394315656637044, 0.0015740761966173009902, -0.0034476792124435440279, 0.61216027061500777329, -0.96136506916681241375, 0.35270233871138140236, -0.34231327491469559776, 0.14452010096334574896, 0.5644640779756074922, -0.17219017949213080709, -0.27637478217850597328, -0.0079167358022348716801, 0.077124945097099814606, 0.033504238154638751879, -0.0060902634564462604025, -0.012194494596590887969, 0.00044788959777930986489, 2.5174384517065657718, -11.800633992133993999, 18.971631541038586732, -11.649297495855712725, 0.69308573662969796825, 1.1761570297759211812, 0.24037336078338489287, -0.062921241138032851792, -0.06781986549513452367, -0.028304301498639963114, -0.0059579658773788422843, 0.0039253649854806768604, 0.0058434876974332609564, -0.00061011657624396992399, -0.20156510764371599054, 0.0823244828758327718, 0.27039614690938162944, 0.12823723784199012043, -0.32898840507416637902, -0.029214607397642346209, 0.028788775449664967543, 0.031761680220599747271, 0.022252370613573997649, -0.0022692751316807577289, -0.0041681833067030273424, 0.0012000205712037477056, 0.0015949955434689433355, -0.0032972129567417555844, -0.1811679218516168699, 1.2891730432694379349, -2.2314512671246560238, 0.97810671193285947389, 0.54519186209741710147, -0.32745315160748628136, -0.11476106918275086333, 0.016905352210725588247, 0.020259745901926959344, 0.0070485876301848212483, 0.0017901054880166550332, 0.00025849813079858432931, -0.00036239509081093456647, -0.00085067176151063590879, -0.27830742977273448258, -0.14716770830275796578, 0.94707662702733719229, 0.0035640805613220108601, -0.77987523001041691995, 0.026687488596517608408, 0.18153857348682533868, 0.05942438491338884049, 0.017210239482741036821, -0.014383866338235976046, -0.0214618174366920135, -0.0070676232921810782511, 0.0042880337168237385101, 0.0028524335312256252593, 0.22563641081951349432, 1.6874786199265392472, -4.0904684689115571317, 1.2445274568468951415, 1.8233075734886370345, -0.33747739941718613021, -0.58987974688622946129, -0.11533434991447820417, 0.097700572248973618916, 0.073290361853752747368, 0.013043987800515329314, -0.012813958270654393287, -0.010085085572386843361, 0.0024170890950979706004, -0.12514139917871508478, 0.12915438295230802224, 0.089180657160416523799, 0.17806403582280533326, -0.49139549125458659162, 0.030300291080853403863, 0.21015038197397914455, 0.055723734370697351748, -0.031012966348336695743, -0.039409502964958952054, -0.018668845115935348011, -0.00064140118167350294951, 0.0068655040284952536037, 0.0045333341135595540961, 0.08233616525885047166, 0.98215376204641280378, -2.1833496051486949163, 0.7577134913211488465, 0.6477081227941645647, -0.14099058730225697, -0.11917592873441600065, -0.01763814416883511052, 0.00049352097814811784603, -0.0046179874776754951254, -0.003061719402851622359, -0.001339155629710725914, 0.0010116671032970505489, 0.004037354097084974551, -0.17742192190767905324, 0.64051528387082212745, -0.89307359038092726244, 0.54257578381782489707, -0.036367755625735784242, -0.14481737544530701167, 0.051528697188439810672, 0.031760927788221066104, -0.0081463227654349074092, -0.0072486223428529610036, 0.0015658204757604875321, 0.00039014033862762764922, -0.0017557659782493823843, -0.0006322638706644948068, -0.15469659804638630218, 0.59403824039357733255, -0.85331260532958741472, 0.51316225987447061652, -0.032453941611083121643, -0.12436768592126168365, 0.045993091258135998523, 0.025079905722180686339, -0.0072360820848106841008, -0.0063427003662894360092, 0.00087026516656629575308, 6.4407105656741868438e-05, -0.0013758470449942166065, -0.00024568796638332212777, -0.061927271746965543986, -0.013497199313808456306, 0.18198279644505213626, 0.069357149753434488293, -0.31844848994978036671, 0.062913302323495456925, 0.097184804712258363035, 0.015834372277261664352, -0.01853519449391180865, -0.019951917211436003691, -0.0037275785623653761094, 0.0045980560006332136386, 0.0037551205772408269984, -0.00061197139225547527927, -0.042220428230344511644, -0.044021354005520964281, 0.18856734600838062632, 0.056068575146968342715, -0.29558026555357591603, 0.076631269821804570697, 0.081013760550042612785, 0.0051551765644741558148, -0.015324074392532414587, -0.014995930941477941811, -0.0023523783711977374992, 0.0039387071459145691146, 0.002952293449140693752, -0.00069702200064924904176, -0.097321762095857411445, -0.061961737134557250128, 0.37562052038428322964, -0.021206616019619796421, -0.32508138564408456395, 0.043227823557785725883, 0.04989149430239887828, 0.048853700391812256154, 0.022839535688803210001, -0.020560143140844130971, -0.020549482283801743382, -0.0039525864905332044538, 0.0047720933043079562588, 0.0031438636416702531694, -0.22447093930982542354, 0.43699303913857290294, -0.40379145840845093174, 0.39178744842835333539, -0.13362477658682775372, -0.12332338321459671016, 0.003086457709897768957, 0.030459142229327242773, 0.020070061060380908757, 0.0040445690094571768239, 0.001844204285769922378, 0.00019950230563109308101, -0.0022826133036442800353, -0.0034988155934021240509, 0.17979928077778131734, -0.39029357454173380093, 0.33226137849735498664, -0.09561134545264081408, -0.24223204938233305894, 0.26704996805357145151, 0.045377704478284278111, -0.08606365377008962092, -0.023993094318949425431, 0.005153281080890562485, 0.0061292690047412964546, 0.0041401725298761517594, 0.0014556056305722578272, -0.0018055861418123117599, -0.21546043026605984916, 0.7794506386204612669, -1.1008470915867918372, 0.65992984876934179805, 0.0054687921703158500625, -0.19282993205571921647, 0.054931863768424307737, 0.024146199498359055546, -0.018611476533384577886, -0.0044452891947320436958, 0.007178730201494151579, 0.003103985707866024988, -0.0010939286783374260314, -0.0010216906588099263289, -0.20164456044681880464, 0.73497862854223217077, -1.0459053398394260181, 0.63137777016659535345, 0.0012043098879222214042, -0.17703923611400573801, 0.049240162274120334585, 0.019209675758921105487, -0.016111504117292625715, -0.0026464407533302938824, 0.0067638078224080263809, 0.0026067900808894903068, -0.0011056320142710002359, -0.0010129407127058938018, -0.12992027335870853788, 0.41802710941946347178, -0.56910188405693162839, 0.39015395921102330457, -0.062156043481927790439, -0.076676645333099549551, 0.025304204983562519904, 0.0098139160637922863428, -0.0081834357643084318368, -0.0037892754839180938994, 0.0052292337384854921736, 0.0026271778945558751928, -0.0012812800523625622953, 5.2879903767161456993e-05, -0.10588690722602756011, 0.37701616069205107218, -0.54370539530824502084, 0.36451895317242311023, -0.053055356860051132439, -0.057330774046434565538, 0.0189463126929673778, 0.003070811859903522633, -0.0069756040633901185075, -0.0023642922868148093588, 0.004878216983184304259, 0.0022949236316167258634, -0.0011773486475707335176, 1.3534904523212010013e-05, -0.15539615065126974325, 0.14527803846967565771, 0.067936138965437861681, 0.18691696219299486215, -0.34262004509213866976, -0.014622500904334825855, 0.10291662418260581757, 0.031236504508602512542, -0.0092630448431810577525, -0.013457544214015650791, -0.0034159341399504963149, 0.0014404386210262502214, 0.001278023997468529958, -0.00051528139065171022805, -0.1558586581551185124, 1.4854743172552744568, -2.6986772153985163492, 1.1159525458375092466, 0.7528043836001995226, -0.37815562947339098621, -0.16751550713316298857, 0.0061183689158782472717, 0.029039569831386759236, 0.017306178095077909274, 0.0030733342946570535795, -0.0036083677145967662202, -0.0029028204368046536116, 0.00062424085116959202989, -0.093987715049960285363, -0.17245059677094595751, 0.62165733236721876231, -0.71544369476962033083, 0.64099809336520130465, -0.096656672216549652243, -0.2571068258694157449, -0.0038898289284546418501, 0.046210828779904303265, 0.020649142164316264259, 0.0033875328011014975947, 0.0023398315758981018918, 0.0018883693242030313766, -0.00051765308138905529768, -0.054850160967085101471, 0.34960771627990727684, -0.34231327478883116777, 0.17972645953187957324, -0.53086088391871977255, 0.13370098936291185954, 0.61237265063134727061, -0.14231983804346123867, -0.27617629631055556905, -0.016367479840125456064, 0.072899955419204368057, 0.033504238320675103213, -0.0054429870387070651561, -0.010821605604576805051, 0.87973692377542622722, 1.0770032755772931399, -11.649297496042438027, 19.497904092017527944, -11.629650285176301949, 0.6011946480325675024, 1.1373293013602245871, 0.23458789550023323445, -0.058895296004295473458, -0.062861284711523648938, -0.026005261622356209972, -0.0059579659789224049052, 0.0026889111561211645994, 0.0051531507200336902283, 0.064819094618014452824, -0.31362679226378897823, 0.12823723799490543973, 0.24973263789311495042, 0.15353019356220112823, -0.31550228856038536884, -0.040170837275855665605, 0.02103836918961752081, 0.031225114603778317268, 0.024991252442836529857, -0.00040821911259911642126, -0.0041681837643089128184, -0.00041235524716980533934, -0.001398189818366378772, -0.34307307236172335241, 0.48794424804681091334, 0.97810671216388145321, -2.2688368149504243299, 0.96981754096117012143, 0.56388481273642931946, -0.31770403985250356138, -0.11419972929395919015, 0.016164062822511166689, 0.019803596321329629271, 0.0065329912278276440485, 0.0017901052624354420358, 0.00048625246448466780212, -0.0013697547964828815387, 0.31097622381031664318, -0.79684509730820651807, 0.0035640805974659259779, 0.98197881048250434954, 0.033123418591258828314, -0.75350122519171403024, 0.014465238652631313629, 0.16392488689682588321, 0.053408049379718117167, 0.019812512944686673511, -0.011270509198012170632, -0.021461817455734031462, -0.0083458286968665206657, 0.0085594234530651930326, -1.0465229939132374426, 2.0275236412193624602, 1.2445274569064657122, -4.4447338625091266096, 1.1423849471292348845, 1.8876043812350224993, -0.26915516192554450603, -0.57218761645766591517, -0.12430953419828583151, 0.089233288769130142404, 0.07086505259366798759, 0.013043987889313320844, -0.01264368230107730702, -0.0059578316555536151022, 0.3051868681929726379, -0.59283635390786304509, 0.17806403569511389251, 0.19325556013903871788, 0.21862544883788953554, -0.49143824125893581511, 0.011021054689552401903, 0.19918524639921886266, 0.054551315633189885468, -0.028223855492573935583, -0.037631914117324243019, -0.018668845263730485251, -0.00032063229686991504648, 0.011135020851061354016, -0.28560064527754253083, 0.62918623932248252117, 0.75771349114254338275, -2.2385670681308962315, 0.74562617733081448801, 0.66188775336971117369, -0.12885482820533772363, -0.11448097100479984189, -0.018662716665168169877, -0.0014766826424729721703, -0.005816246956171926101, -0.0030617193329907036597, 0.00019376551335855692971, 0.0044960570059675714072, 0.00061415193303189269565, -0.12554059866195693695, 0.54257578397757533217, -0.84596185571646353996, 0.54841878846872382791, -0.043320859023501634422, -0.14887560471657859673, 0.054101950976848035835, 0.033936208883947195647, -0.0076642731605201398076, -0.0074863184042746958036, 0.0015658206185414573052, 0.00048694808118327588195, -0.003182440122131451575, 0.00021682021946824948304, -0.11097763197099164145, 0.51316225989984731726, -0.81437869332831824831, 0.51610736332796758319, -0.037140938811909572637, -0.12737598781825337779, 0.048140398797284877919, 0.026766786006056308977, -0.0068886487354147254644, -0.0065565842716360685227, 0.00087026516291617344798, 0.00020029148775887496703, -0.0023031664625210324787, 0.16898525493293120592, -0.35783977320027809954, 0.069357149893859534884, 0.22577662379388666047, 0.094311074753250678793, -0.32518734816520322717, 0.046040635766082631253, 0.093359853251310118138, 0.018233832588046051254, -0.016349250408461409728, -0.019533212078842211723, -0.003727578524587872328, 0.0048125516029623892009, 0.0024683492611191335843, 0.15158875932744900195, -0.32125431755784172871, 0.05606857531641035175, 0.21906453558560537442, 0.072688227399907226478, -0.30105589407594041118, 0.062657015968666032735, 0.078346981369086285407, 0.0074058328275562659154, -0.01356744177532779802, -0.014782940242950400739, -0.0023523785450354495197, 0.0041698300327327421019, 0.0016449600240780297162, 0.12810740590555025342, -0.28072009409987180861, -0.021206616025295478567, 0.34073560328332802571, 0.02759224577628432637, -0.29778052026590334789, 0.030989241354015758384, 0.037208086589798058941, 0.045087044540968700124, 0.024803016272256870473, -0.018546343992736410294, -0.020549482164827055641, -0.0041544910904101686744, 0.0083024775063685243154, -0.027153704646287888519, -0.15342599750098404043, 0.391787448759084167, -0.41503688249804554511, 0.40188669858181796934, -0.12730694785451307505, -0.12776802955992616617, 0.0022006961108127676315, 0.031364317010335103497, 0.021384295912637203002, 0.0049319517411570354404, 0.0018442040970590373519, -0.0010922338032433407274, -0.0058391909699665022621, 0.15454207520842352941, -0.2127919308970303014, -0.095611345225275284898, 0.32195994110184994463, -0.13869737426561096205, -0.2493992042782779972, 0.26461496244907650732, 0.049064852089167353133, -0.082401002728559907817, -0.023051928026150726725, 0.0046805805357197100697, 0.0061292686212443400415, 0.0041187296400393008655, -0.001217672999951218988, -0.032215610611446142408, -0.12157092309021103049, 0.65992984887353123291, -1.0270706382435061954, 0.65614857325627484741, -0.0085260722117316382895, -0.19553435934973489863, 0.058327481881573266242, 0.02583975942842497639, -0.018891648589514375284, -0.0051067121018240554942, 0.0071787296874714539088, 0.0033562111235767616742, -0.0022790089697736293575, -0.032389606271167513396, -0.11439686052286468509, 0.6313777705652072747, -0.98114418224018784542, 0.62467397472039265516, -0.010536443955545484255, -0.17867984538925862914, 0.052349228164140691744, 0.020598455452214222555, -0.016376759024106835733, -0.0031974570490057247288, 0.0067638082906979401718, 0.0027445247910197305419, -0.0022132059376187766635, -0.0037105086265705757541, -0.1203545116448631741, 0.39015395942688213538, -0.54089768886153155592, 0.39376531218692117919, -0.071156875026073568291, -0.081731909829809892831, 0.029229195511136472907, 0.011676794633297129036, -0.0095359864152594897568, -0.005397083998417795056, 0.0052292342237483819406, 0.0039760795925767259545, -0.0015082265975828340089, -0.0053880341327718159899, -0.10040820692661107938, 0.36451895302619513606, -0.52422528977532034666, 0.36436684206056219049, -0.060189336255057401814, -0.061204641046027130624, 0.022733408774115018702, 0.0047338311639569538927, -0.0082124806015620201161, -0.0037864686829097906479, 0.0048782167575363690262, 0.0034520519503448547118, -0.0014095053340960374266, 0.15045226337651237958, -0.39848312259366563959, 0.1869169620122722586, 0.11789530462982697467, 0.20727375978763099007, -0.3470197960119791869, -0.027688914983831029032, 0.099706149033305893625, 0.03326120805205681763, -0.0076437697488583320263, -0.013535915165910612432, -0.0034159341801367538484, 0.0019332420304228360719, -2.3399955491903074351e-06, -0.44973037553824285473, 0.70588113238447269815, 1.1159525459356374189, -2.7627379568435364021, 1.0890984062259847498, 0.77660611998602790962, -0.35682818333861426874, -0.16350410228428463588, 0.0037473042808854237738, 0.026849075623476895136, 0.016280677358107596003, 0.0030733346429266616105, -0.0031624963814022057085, -0.0021620873261076140395, 0.096365054583738743443, -0.26817398407727732756, -0.10730222295146812916, 0.64099809588562361995, -0.71660575793098457709, 0.63767749949320173819, -0.097192199823270472581, -0.25777018715300920038, -0.0033746857582468009484, 0.047188952038533273081, 0.020649142313388508391, 0.0018128838757206547881, 0.00061699817623750994114, 0.0034043243781404965848, -0.44319265014133230052, 0.72803610877201418994, 0.1445201006906333685, -0.53086088364193506628, 0.18130114452823045679, -0.49790133485547705128, 0.14817214951673854384, 0.61580575508667445561, -0.14327293490081621785, -0.27869925667168826688, -0.01636747995197449651, 0.077124945136779365917, 0.035719556018106522755, -0.022274900622964225599, -0.12403098870522748232, 1.2808111204921035498, 0.69308573663758277217, -11.629650285019467404, 19.603183073261828184, -11.618402627376754666, 0.57830655385917384415, 1.1266557115615418727, 0.23632093139187335118, -0.055806334655461212191, -0.062861284990830337827, -0.028304301061687162783, -0.0066343509723328607242, 0.010508585647738237154, 0.073793398857442349392, -0.02826753360572552623, -0.32898840506231419312, 0.15353019339578816149, 0.2680053154837159779, 0.15007652583694530701, -0.31971439083390279512, -0.04059208368435210379, 0.022509691551021168004, 0.032342647122879708554, 0.02499125215443746828, -0.0022692747843508385627, -0.0071809662573713885225, 0.0011868171257177334826, -0.00062485276888331992787, -0.37656455915409448032, 0.54519186185291912405, 0.9698175408267094566, -2.284889635442514777, 0.96379478979927446503, 0.56761781282889889333, -0.31418236918975922523, -0.11361244395753400394, 0.015442607781951795537, 0.019803596376327805423, 0.0070485876424779573754, 0.0011925280808199887131, -0.001025977674039962699, 0.16971045964135800244, 0.065050262914598283581, -0.77987522982716561692, 0.033123418397207074482, 1.0281112309540170191, 0.040958735012349291316, -0.76740036201365335877, 0.0049905058181765570013, 0.16338812660645804509, 0.055448160525761368278, 0.019812512826657201898, -0.014383866095217542544, -0.023161590985113614943, 0.0050826605483385196849, -0.35345466234406602846, -0.41152260639817250798, 1.8233075739816488881, 1.1423849467001623292, -4.4532237403880907678, 1.1535892394425362184, 1.9025738315223059693, -0.26224504536145276656, -0.57300464553641494714, -0.12701422134780904205, 0.089233289108745769691, 0.07329036151251032416, 0.013162754963582245712, -0.021100713832958688243, 0.14171532301489669914, 0.031776421631765479348, -0.49139549031089629905, 0.2186254477998934298, 0.20355001508110204256, 0.21258457232074923215, -0.50195427264637204612, 0.0044308547480050031186, 0.19909052435256524571, 0.056410061523545190765, -0.028223855302766314573, -0.039409503165732016894, -0.017326610122055097901, 0.012778064165653213805, -0.080255402544936182974, -0.19400888596362853011, 0.64770812350834205962, 0.74562617652561569148, -2.2553507114899904451, 0.73740850838857052718, 0.66442669688212518952, -0.12702641474042805636, -0.11474865480224633041, -0.019363402735109308539, -0.0014766825331670179085, -0.0046179877342805708551, -0.00058987392888151931811, 0.0028296843080424876303, 0.073339222626756433998, -0.15267844731996541019, -0.036367755265886390903, 0.54841878767545138018, -0.86243302161915147419, 0.54550757188316301161, -0.038406747473825379735, -0.14745021403123367842, 0.053701722550739725004, 0.0336424331226589865, -0.0076642728044076791072, -0.00724862262722828464, 0.0015293672131261799082, -0.0033804516390849954563, 0.060945875731673558595, -0.12974293851414886447, -0.032453941190746585232, 0.51610736266197521083, -0.8279464343353176714, 0.51375981535554926527, -0.033143027402666107306, -0.12633990293424932072, 0.047778449119701561776, 0.026523394912282293595, -0.0068886482714995266735, -0.0063427006619273555363, 0.00093275422487872246269, -0.0027014295026349852656, 0.051867435060574264283, 0.063672998542022976598, -0.31844849014819182065, 0.094311074970074140755, 0.22665696779929719873, 0.084996693841927511981, -0.32895251829775606911, 0.045842447492303206591, 0.094309680339592735776, 0.018832980269176593374, -0.016349250437419499127, -0.019951917426071722611, -0.0036658523702386399193, 0.007798000309339798812, 0.033048870762714101235, 0.083335155750506428873, -0.29558026581864965676, 0.072688227539534577826, 0.22108712593495369325, 0.065242810654850727503, -0.30407008102548993556, 0.062735011908029839067, 0.079215372800541972165, 0.0077990499783945664916, -0.013567441836721365334, -0.014995930702994306194, -0.002362337022979288121, 0.0061436151494120841074, 0.085822850680015647096, 0.023491193681979872371, -0.32508138604688607831, 0.027592246007469553581, 0.37573758677355673097, 0.022237937789300495778, -0.30867092780335264601, 0.025983802625029590871, 0.037089331932330396602, 0.046318740196850456681, 0.024803016283065582037, -0.020560143120618726287, -0.021113180529784243006, 0.0073194522534383351514, 0.067754813558049728495, -0.12631551869485230233, -0.13362477653770554742, 0.40188669806830457087, -0.41392152084811045798, 0.40027727877678748181, -0.12429088874271530807, -0.12595794866758863728, 0.0027048133001388572662, 0.031618259815755563047, 0.021384295913499468123, 0.0040445687610194264172, -0.00053804391998026534702, -0.0056507213866494121826, -0.10693515523434590775, 0.32575915222509221891, -0.242232050032165247, -0.1386973737051165334, 0.32707818352163037234, -0.13918996693686180866, -0.25150137717852050212, 0.26670008476581030799, 0.05142876762665479079, -0.082083814282513128946, -0.023051928521469946393, 0.0051532818809126379814, 0.005556003812176250195, 0.0024740260390996617057, 0.064355717854905805653, -0.19671150064957906856, 0.0054687921795497224098, 0.65614857305000440224, -1.053931682464436026, 0.65389908552324504942, -0.0024852375937052736954, -0.19433802463649779702, 0.057675723076324701688, 0.02543451393640043523, -0.018891648720467720673, -0.0044452888046691389767, 0.0080873220068359650409, -0.00036380396376208879762, 0.056086596407833776146, -0.1775877962358108908, 0.0012043095295997613761, 0.62467397491314491376, -1.0049786366263844251, 0.62330532182109754658, -0.0050526838286010893053, -0.17764636934113314792, 0.051826781174613964498, 0.020272788308160937659, -0.016376758894577861381, -0.002646441182046047752, 0.0074644307140857759503, -0.00068570162655362477868, 0.028850307339345768881, -0.08043082041500924928, -0.062156044055664341763, 0.3937653128437522132, -0.55573392301075597022, 0.38963137188249030718, -0.06578976693897846395, -0.080136368009354433384, 0.027646235669840341087, 0.010249894394648543725, -0.0095359864735290872373, -0.0037892754599447584299, 0.007045937112845139863, 0.00014125594454782460061, 0.016837403011078520831, -0.058129272007759288932, -0.053055356879879542187, 0.36436684223287568729, -0.53672244454330531038, 0.36082693067015947364, -0.055446559117665557281, -0.059690371093863964225, 0.021436267351448528434, 0.0034997133975762551202, -0.0082124810086779062579, -0.0023642919039646098991, 0.006449244539697161771, -1.5097415404573852164e-05, 0.087529709893826995737, -0.0074691221101393220558, -0.34262004493088488166, 0.20727375957396215744, 0.1189501019128267334, 0.20317503159040326, -0.34837072892590392748, -0.027932286333618862317, 0.099738395142482949574, 0.033025780378337789289, -0.00764376973797414698, -0.013457543963290380384, -0.0037875988293019414478, 0.0022395623188905125836, -0.031722562777623294905, -0.42803978530947922598, 0.75280438333729704503, 1.0890984065314512907, -2.7816557649904805061, 1.088789829574271284, 0.7834145878435424537, -0.35350516519463787146, -0.16387002312491702471, 0.0024815628985166825887, 0.02684907562172096987, 0.017306177770005985034, 0.0030113317471001581176, -0.0062030015706169587195, 0.055407226835385403607, 0.015793396398307631145, -0.25103090828577417915, -0.096656674437066311678, 0.6376775003968135902, -0.71821081275307818714, 0.63894527050486638053, -0.094811233583618093057, -0.25639708126474852756, -0.0033746854777534957166, 0.046210827944567850478, 0.019356395965681196664, 0.001733671009768593483, 0.0048207495381879204854, -0.093052602614547577642, -0.21123525019320671392, 0.56446407821533239879, 0.13370098896882814943, -0.4979013346580510313, 0.19245014176535060746, -0.50307387497400402054, 0.14310105781064785813, 0.61429348264952643444, -0.14327293497491591645, -0.27617629600551740765, -0.0079167362440480601449, 0.083800182727113800563, -0.0011569209585384234145, -0.18801759386755434456, 0.22395025341741128799, 1.1761570299248103044, 0.60119464763480690017, -11.618402627127064619, 19.62513313730161002, -11.614211940226240927, 0.57742174439796944974, 1.1268112936783691858, 0.23632093130402059833, -0.058895295661810190269, -0.067819866042127391337, -0.026322979137334435312, 0.0079657820753634735189, 0.038874504389960679029, 0.024236232735448285519, -0.029214607470452728932, -0.31550228808995622876, 0.1500765256458245489, 0.26654973864810582729, 0.1526215830966364817, -0.31689829300261035394, -0.039444169630402939331, 0.022509691555635234078, 0.031225114536629332979, 0.0222523708291901684, -0.0046248302966042723061, -0.0022316116710078873729, 0.05669882872275455038, -0.11332917980927294244, -0.32745315161419769057, 0.56388481319686845072, 0.96379478930217299393, -2.2909158072546875751, 0.96315083544628066203, 0.56871492056721362207, -0.3140905655945481767, -0.11361244390753771905, 0.016164062562241766907, 0.020259746215587289692, 0.0062679353295906651763, -3.2574762820910948822e-06, 0.01614895066722177655, 0.16477985888015161953, 0.026687488277163124578, -0.75350122495652294141, 0.040958735022772967127, 1.0227992184067580261, 0.038251279104657677499, -0.76624313348859263151, 0.006443258573741404531, 0.16338812679463293742, 0.053408049421668497836, 0.017210239152618558856, -0.012676030257067217796, -0.01646605657439488557, 0.16647979863566159109, -0.64517561743837603672, -0.33747740031463718813, 1.8876043819950008018, 1.1535892391783433286, -4.4663349975662587354, 1.1427559126923532151, 1.9008247386863408313, -0.2616160337097629407, -0.5730046450160191096, -0.12430953481132321625, 0.097700572650875464054, 0.077670198259809378194, -0.020707299105274207945, -0.057757179623292895232, 0.21138595416775110558, 0.030300290630554053972, -0.49143824039197531173, 0.21258457199354929834, 0.198763337810270585, 0.21022584405735053603, -0.50201376315690848795, 0.0051240158835988720148, 0.19909052454541562027, 0.054551315410427285435, -0.031012965988078046442, -0.037434250544735252209, -0.0014463645431498936642, 0.0029144665065365710395, -0.13183642338332129951, -0.14099058783016640839, 0.66188775433285607086, 0.73740850808151647833, -2.2631352509034710074, 0.73285675455183818539, 0.66303741753101952394, -0.12724750646033519685, -0.1147486546671090818, -0.018662716925420604064, 0.00049352134433786477418, -0.0025217313519585371137, -2.6175053955944139621e-05, 0.00067260836266643387986, 0.070126440067686549029, -0.14481737562906377148, -0.043320858527288515216, 0.54550757159704499344, -0.85852200167320646962, 0.54753694411254261176, -0.038510275167475085945, -0.14795544465208543272, 0.053701722739629920156, 0.033936208744105848811, -0.0081463228576294718919, -0.009533102850118201424, -3.2344037533647536711e-05, -0.0012348072329483928723, 0.060573647560894332276, -0.12436768591588946709, -0.037140938473035754741, 0.51375981518599411757, -0.82501273776007810756, 0.51517295424810671811, -0.033302629995120042183, -0.12676340851436490587, 0.047778449427481814449, 0.026766785645000405053, -0.0072360819319624396229, -0.008071075940635804094, -0.00039049606136778655403, -0.036126328596863759024, 0.10152090902601385025, 0.062913302343238081127, -0.32518734822307604482, 0.08499669385670595434, 0.22641563897857841514, 0.087139410485453830391, -0.32803247421136905171, 0.045843446454483578745, 0.094309680170725190829, 0.018233832761889230689, -0.018535194438480378371, -0.021028437807345239791, 0.0076968955104344278367, -0.033385335633706277048, 0.082888864064324496717, 0.076631269832314816148, -0.30105589408257754647, 0.065242810554394223277, 0.2212201471858748214, 0.067228398450106474638, -0.30331311372284969208, 0.062635701395855045148, 0.079215372751316889821, 0.0074058328603685075545, -0.015324074542095340867, -0.01598345952414776186, 0.0066869507933044114098, 0.027224450766413835906, 0.025177486942170597406, 0.043227824257401722963, -0.29778052084570694191, 0.022237937977050297594, 0.36802630290380600186, 0.020054887587818488737, -0.30783687035410339794, 0.026934261610330120784, 0.037089331705003308892, 0.045087044923760449811, 0.022839535312304991982, -0.020470709595783280477, -0.011020752816118577211, 0.044773229662090456571, 0.010592602708856046351, -0.12332338296508503495, -0.12730694763428879179, 0.40027727866853318739, -0.4110035631646817289, 0.40354098476286470731, -0.12264311583542886197, -0.12553088975013421136, 0.0027048130673853142646, 0.031364317395988570991, 0.020070061092761354304, 0.00052922941138333851139, -0.0033679298057542024838, -0.075107850185532065712, 0.031398284744644365529, 0.26704996906562677994, -0.24939920537839546655, -0.13918996625028545733, 0.32731405849450678769, -0.13791721947197441156, -0.25072657763115036023, 0.26672729247138166064, 0.051428767166065696148, -0.082401001763932779642, -0.023993095456642816315, 0.0056101084554508063446, 0.0087077030578531673666, -0.011320722449373097837, 0.082860754106298936672, -0.19282993214057178588, -0.0085260718432907025427, 0.6538990853283814797, -1.0495735021178003166, 0.65542461729380319113, -0.002869722711994930793, -0.1948695368898208935, 0.057675723014963167778, 0.025839759201526379256, -0.018611476269238202974, -0.004817981543905807372, 0.007587141488267595997, -0.0096925754350511723195, 0.073858530421913523578, -0.17703923632174875524, -0.010536444002921446783, 0.62330532215575074062, -1.0011469235063805705, 0.62456131970326411729, -0.0053334943186414979915, -0.17805866006413761937, 0.051826781394434591477, 0.020598455355881707673, -0.016111504330392174139, -0.0028654865681509420795, 0.0065056041487839728898, -0.0074399715856508742348, 0.040570641456731920682, -0.076676645171165902504, -0.071156875490914997529, 0.38963137220805238448, -0.5509135197713252996, 0.39160016504893530032, -0.067210157373491702382, -0.081644946088474099066, 0.027646235721324931983, 0.011676794665905466941, -0.0081834361374977775822, -0.0049030682747121662191, 0.0066991721364810319708, -0.0082789207839683090495, 0.030694624249670145882, -0.057330773965155491789, -0.060189336246873656466, 0.36082693050760689113, -0.53264636790220976081, 0.36244820911011277342, -0.05668545082366924226, -0.060975518170549775077, 0.021436266876672077641, 0.0047338313814382030481, -0.0069756038453688750875, -0.0032541956141128995129, 0.0059014229895787079574, -0.013573466232008894045, 0.11248775873779133516, -0.014622500852365400564, -0.3470197959054518444, 0.20317503137150952353, 0.12109353023512438308, 0.20591364298482819906, -0.34785343661412271299, -0.028667182466677335051, 0.099738395128567386427, 0.03326120805569722505, -0.0092630449038692934394, -0.015668453374732715933, 0.0015339875916121738578, 0.072977767579869859405, -0.1693128060807556623, -0.37815562898555860061, 0.77660611953666569551, 1.0887898297263178815, -2.7875967326033088689, 1.0860321550245122246, 0.78338706253961032377, -0.35363157821782631451, -0.16387002353931356957, 0.0037473047657532827445, 0.029039569801856544157, 0.017542571542405266394, -0.0060330194204885941728, 0.010424440050725684165, 0.059587109121434132797, 0.0030214966605921454723, -0.2571068247811444385, -0.0971921998590098285, 0.63894526951332919484, -0.71554287681551298483, 0.64090853343009468457, -0.094811233181807175185, -0.25777018729936707064, -0.003889828236409220881, 0.048638988949380811133, 0.022501379290639039049, 0.0022865900912143687605, 0.13357834800578374801, -0.31161386047545902578, -0.172190179719652392, 0.61237265103640636799, 0.14817214930152988739, -0.50307387502999278972, 0.18471296149658320007, -0.5068644090175478123, 0.14310105805702388881, 0.61580575501939172067, -0.14231983832311817517, -0.27637478171972534868, -0.0072781375794159254566, 0.081648002441251699146, -0.077687078037020740329, -0.080843086053593013474, 0.24037336056448571631, 1.1373293017109851188, 0.57830655371187666969, -11.614211940276128132, 19.634407177871981531, -11.610924826310208502, 0.57742174453599692807, 1.1266557115380755327, 0.23458789534954277434, -0.062921240901434929982, -0.06902938305542211106, -0.013344020403485720369, 0.0023162009355638768331, 0.040797126019960390508, 0.02878877539748671277, -0.040170837740977065333, -0.31971439031099002648, 0.15262158290856672749, 0.2704832376204108213, 0.15455597621052480362, -0.31689829282278186851, -0.040592084125097772918, 0.021038369491603849115, 0.031761680003616891976, 0.025702571498177076426, -0.0098986284709301711887, 0.022718400068913555123, 0.024749900793374659397, -0.11476106902138699661, -0.31770404001466656663, 0.56761781337449979112, 0.9631508347316510843, -2.2922319762752736416, 0.96239580485236975171, 0.56871492079885221926, -0.31418236906272517528, -0.11419972946206469189, 0.016905351902902227695, 0.022440242674760652736, 0.0042398702523000972309, -0.032856633834997643617, 0.053721125363481801729, 0.1815385736964199026, 0.014465238675472410698, -0.76740036221460095067, 0.03825127922330728153, 1.0268816099980235812, 0.041383576430300918203, -0.76624313322171722085, 0.0049905055934345351337, 0.16392488685674769244, 0.059424385078038870212, 0.022775545159281221363, -0.040012814447403684259, 0.16900639644501760928, -0.1174443495134873533, -0.58987974596544945349, -0.26915516270402523036, 1.9025738318214555633, 1.1427559126972346437, -4.4763992160633154782, 1.1395963688853685269, 1.9008247392285704258, -0.26224504629932426925, -0.57218761567814779845, -0.11533435031918996827, 0.1070686210246131892, 0.040959844219834243284, -0.082009723783750779202, 0.052536661883837043607, 0.21015038197261504127, 0.011021054091607241504, -0.50195427229502642685, 0.2102258442099078084, 0.20061251695596929445, 0.21182683000498364612, -0.50201376298886901672, 0.0044308541956537715253, 0.19918524684026195359, 0.055723734070056894774, -0.031451143818518864648, -0.038922302957169432891, 0.00030202209204688713721, -0.016174750178193208344, -0.1191759281259604647, -0.12885482911959256214, 0.66442669716183833106, 0.73285675466947086676, -2.2675110316865780824, 0.73116302238663333224, 0.66303741762522716563, -0.127026415052747782, -0.11448097041812900498, -0.017638144925186307471, 0.0011389720527652915164, -0.0032115145774433721468, -0.016234805404534712742, 0.033741713285871076378, 0.051528697603587629572, -0.14887560515552353269, -0.038406747302932066346, 0.5475369443645395906, -0.8586467138820690348, 0.54714941782943637971, -0.038510275420925166312, -0.14745021434382882264, 0.054101951353237379139, 0.0317609275604870539, -0.014131359080935249409, -0.0029956452596728903003, -0.014258273956581227551, 0.026208753664457649923, 0.045993091447895498303, -0.12737598808827957364, -0.033143027306189135062, 0.51517295448747413289, -0.82525188339054511566, 0.51481559797160192371, -0.03330263012153955593, -0.12633990332712494165, 0.048140399226279204092, 0.025079905411211630278, -0.012163878321431020185, -0.0031449121612952025961, -0.037419477107384413461, 0.021529090706553417195, 0.097184804756504747369, 0.046040635590544769229, -0.32895251823724130924, 0.087139410502276387627, 0.22841873975534779384, 0.087614192409699642727, -0.32803247424072617955, 0.045842447664506097427, 0.093359853031805925538, 0.015834372566781336478, -0.018922775121951528055, -0.010116564349615768392, -0.028103875747376626248, 0.0095903133913719466191, 0.081013760650746863279, 0.062657015818674749452, -0.30407008077341191044, 0.067228398234713077075, 0.22289644890818652301, 0.067504624727386400296, -0.30331311385588005658, 0.062735012123823624686, 0.078346981357738695873, 0.0051551764555345441429, -0.015466062921313203718, -0.0065701718311977404946, -0.030847131847377494418, 0.054173554929181917672, 0.049891493994308963589, 0.030989241962457569207, -0.30867092819948155347, 0.020054887576889553896, 0.37008216472673710618, 0.021774314539292898008, -0.30783687044866853144, 0.025983803180501599539, 0.037208085784766496695, 0.048853700963800567036, 0.027700806641642962413, -0.03917487683679113375, 0.01218389383598123682, 0.037715485853951409567, 0.0030864576868251665347, -0.12776802956967805991, -0.12429088875084111365, 0.40354098502258239911, -0.40904358022885967117, 0.40438361924276455817, -0.12264311596828032791, -0.12595794854392575179, 0.0022006957206660957178, 0.030459142142521686974, 0.018312514286398698382, -0.00087907111682483435235, 0.0056877962367470873833, -0.091347311581971971095, 0.04537770324985646031, 0.26461496370120241028, -0.2515013781843360996, -0.13791721890574545895, 0.32821065063938487105, -0.13825457938968507832, -0.25072657793637087575, 0.2667000854220161199, 0.049064851022584454743, -0.086063652494673986704, -0.020990434538396225694, 0.01647438997303124833, -0.012906291853360441629, 0.02558280278298061261, 0.054931863660215574097, -0.19553435933910978672, -0.0024852375338307239327, 0.65542461723334766166, -1.0501662858704998715, 0.65492642197958994554, -0.00286972307330497561, -0.19433802407964093084, 0.058327481872733587254, 0.024146198826416689215, -0.024025317699802192783, 0.008814896250153201121, -0.0092850482692037415028, 0.019962524453486452103, 0.049240162797663591943, -0.17867984518950622008, -0.0050526847180408463941, 0.62456132033747358001, -1.0016548247415424377, 0.62416292015396768189, -0.0053334940261639002951, -0.17764636959024796004, 0.052349227891308323934, 0.019209676573344037359, -0.020756791314314907565, 0.0088022921700288414559, -0.0068745463657246064573, 0.014091159277330015526, 0.025304204979745552329, -0.081731909506940370047, -0.06578976745314278618, 0.39160016531686742125, -0.5523529709276926436, 0.390146639038045695, -0.067210157329565881068, -0.080136367928647228576, 0.029229195202799901354, 0.0098139168835665001323, -0.012912998015090637258, 0.0064889152318302239023, -0.0039565511105555360716, 0.0065307523277249859492, 0.018946311996662676402, -0.061204640902347749731, -0.055446558853476879403, 0.36244820883074707618, -0.53398249323676538758, 0.36115559889549125883, -0.056685451499047141699, -0.059690370178491852782, 0.022733408603173077545, 0.0030708115026419858726, -0.0108781563405131107, 0.00665223134218970516, -0.028447715270548974004, 0.034500188312068823249, 0.10291662418716925054, -0.027688915005977643435, -0.34837072879150909888, 0.20591364288992472509, 0.12282993677697451318, 0.20584221872434313449, -0.34785343686066039215, -0.027932286105195783255, 0.099706149170785934865, 0.031236504052522831576, -0.011671491080680471836, -0.010750454015747491418, 0.041370768242440680851, 0.0088971549703493347377, -0.16751550740494702829, -0.35682818349415768022, 0.78341458803209440553, 1.0860321550918143885, -2.7909331658518685693, 1.0846428882205558786, 0.78338706220193488061, -0.35350516490573402617, -0.16350410252232583375, 0.0061183692173532918529, 0.030346715232454048866, 0.0081727131037146200782, 0.0022865905877414234755, 0.022501378873906852096, 0.048638989905713359008, -0.0038898305417878527995, -0.25777018662007422645, -0.094811232698934808094, 0.64090853369520117688, -0.7155428765872293706, 0.63894526970435727709, -0.097192199584092239806, -0.25710682657606948043, 0.0030214972255841470068, 0.05958710935211767451, 0.010424440223531018443, 0.081648002401492406066, -0.0072781374438920512157, -0.27637478185357511329, -0.14231983830974748173, 0.61580575504787771202, 0.14310105810835513362, -0.50686440924288289622, 0.1847129618981778465, -0.50307387549776427793, 0.14817214976124873371, 0.61237265074942015541, -0.17219017969830241466, -0.31161386043362215847, 0.13357834800681903098, -0.013344020387115759427, -0.069029383069430461095, -0.062921240943346973262, 0.23458789548690969773, 1.1266557113492101649, 0.57742174464795148481, -11.610924826442314384, 19.634407178182801346, -11.614211940622926278, 0.57830655405440078987, 1.1373293012754492892, 0.24037336096607458957, -0.080843086279264800287, -0.077687077971963072742, -0.0098986285093123401596, 0.02570257138386080123, 0.031761680115605935015, 0.021038369684082066507, -0.040592084245479713445, -0.3168982925882111723, 0.15455597578564764238, 0.27048323767481941005, 0.15262158295411099029, -0.31971439034271803559, -0.040170837622092406272, 0.028788775366311698811, 0.040797126168228677046, 0.0023162007849727864542, 0.0042398702463346694264, 0.022440242498947375605, 0.016905352140470803618, -0.11419972919592460969, -0.3141823696302245561, 0.56871492129935496962, 0.96239580428382620614, -2.2922319760349649798, 0.96315083501032883184, 0.56761781306667191682, -0.31770403978308708881, -0.11476106909387093213, 0.024749900579186755872, 0.022718400226852387275, -0.040012814487998468738, 0.022775545202186841393, 0.059424385070183528834, 0.16392488682008962697, 0.0049905058542139323421, -0.7662431335648629549, 0.041383576450975609329, 1.026881610250100163, 0.038251278926214660192, -0.7674003620456238961, 0.014465238820855673418, 0.18153857339772569435, 0.053721125514731168926, -0.032856633853834533998, 0.040959844147778937118, 0.10706862125273336472, -0.1153343505706297123, -0.57218761576860222018, -0.26224504582898278127, 1.9008247388879291329, 1.139596368868901699, -4.476399215767798978, 1.1427559121168979761, 1.9025738322670127101, -0.26915516261949296029, -0.58987974632686723808, -0.11744434921573666808, 0.16900639633462760658, -0.038922302987066503821, -0.031451144054726667965, 0.055723734378352825225, 0.19918524681216176475, 0.0044308544100253597145, -0.50201376306008649308, 0.21182682984238435098, 0.20061251693897474979, 0.21022584395225923926, -0.50195427233077083429, 0.011021054539206893291, 0.2101503820036467185, 0.052536661917245687514, -0.08200972398908537575, -0.0032115146863447132167, 0.001138971858200525742, -0.017638144477107979557, -0.11448097039424297294, -0.12702641506018771955, 0.66303741754333478386, 0.7311630222019691594, -2.2675110316991111681, 0.73285675466161259717, 0.66442669702306655921, -0.12885482854686422516, -0.11917592834381408873, -0.01617475025688111176, 0.00030202207315800991272, -0.0029956454082122095997, -0.014131358971564186855, 0.03176092755493516856, 0.054101951454996556612, -0.14745021426467042058, -0.03851027545906112598, 0.54714941776659453687, -0.85864671417972227019, 0.54753694456051693606, -0.038406747348852202095, -0.14887560487024167899, 0.051528697461476161146, 0.033741713372835394069, -0.016234805529204469898, -0.003144912339840586369, -0.012163878155322004754, 0.02507990549556649712, 0.048140399076747344975, -0.12633990309534348651, -0.033302630188817745893, 0.51481559788226061158, -0.82525188360755175054, 0.51517295464960488616, -0.033143027246540120168, -0.12737598797196725275, 0.045993091278789179077, 0.026208753844598219052, -0.014258274092480661618, -0.010116564292965219557, -0.018922774991641317516, 0.015834372109665154765, 0.093359853398292988458, 0.04584244750782397182, -0.32803247426259773967, 0.087614192570771104451, 0.22841873979097715441, 0.087139410381862084098, -0.3289525183397596364, 0.046040635783102482059, 0.097184804661737803078, 0.021529090674718313198, -0.037419477062303606318, -0.0065701717082141111378, -0.015466063108642935306, 0.0051551764703796198325, 0.078346981576893334753, 0.062735011706115242913, -0.30331311359690610319, 0.067504624783126215837, 0.2228964488603543681, 0.067228398398575445194, -0.30407008112443806302, 0.062657015883389455269, 0.0810137609489467031, 0.009590313095603840432, -0.028103875644059906486, -0.039174877004077256204, 0.027700807058654089432, 0.048853700491643163584, 0.037208086253803819565, 0.025983802823812497673, -0.30783687030956052855, 0.021774314393617367341, 0.3700821648426511068, 0.020054887541425987785, -0.3086709280436764069, 0.030989242015382546108, 0.049891493640338971383, 0.054173555083322888493, -0.030847131819668457964, -0.00087907107152999746074, 0.018312514019828318029, 0.030459142260136566616, 0.0022006959880239221339, -0.12595794856155675956, -0.12264311612686117081, 0.40438361924626720079, -0.4090435802753257799, 0.40354098499050833349, -0.12429088890956030422, -0.12776802911747345282, 0.0030864576339436691765, 0.037715485724136223189, 0.012183893807093998732, 0.016474389954104027428, -0.020990434383520602951, -0.086063652734070794637, 0.049064851074414043974, 0.2667000855692099881, -0.25072657818718185885, -0.13825457909959829039, 0.32821065056067216803, -0.13791721911472196171, -0.25150137784853909828, 0.26461496333323492047, 0.045377703509638787904, -0.091347311649756277041, 0.0056877962303635686697, 0.0088148964026773940494, -0.024025318311335384924, 0.024146199485953423469, 0.058327482007632519556, -0.19433802457991428136, -0.0028697230361702263414, 0.65492642222954466913, -1.0501662862928051734, 0.65542461797095907805, -0.0024852381914981392513, -0.19553435930147497479, 0.05493186427696809121, 0.025582802086160524868, -0.0129062915902292321, 0.0088022918997113803324, -0.020756790434565752096, 0.019209675475231918762, 0.052349228211702659996, -0.17764636871071914914, -0.0053334952500701101816, 0.62416292079512059043, -1.0016548247666290372, 0.62456131981148488297, -0.0050526840455383369849, -0.17867984506258044464, 0.049240161917544475467, 0.019962525110367956155, -0.0092850484236178480041, 0.0064889150583604011946, -0.012912997207409444497, 0.0098139158853615994504, 0.029229195350023461086, -0.080136367446502446144, -0.067210157859607377873, 0.39014663945944488832, -0.55235297098499769319, 0.39160016512403184041, -0.065789767234387841466, -0.081731909473117286824, 0.025304204343698979279, 0.014091159912589643413, -0.0068745465209343933205, 0.0066522316558240762382, -0.010878157130519587603, 0.0030708122658260421887, 0.022733408553427581628, -0.059690370961684606488, -0.056685450773365911314, 0.36115559886134263046, -0.53398249366882788447, 0.36244820972343905208, -0.055446559753832745654, -0.061204640995134881876, 0.018946312914734156352, 0.0065307515151332446585, -0.0039565508241075544177, -0.010750453978038733427, -0.011671491234037666834, 0.031236504338204240661, 0.099706149027030316345, -0.027932286319387551743, -0.34785343663942769332, 0.20584221869337174238, 0.1228299366587115582, 0.20591364323363908762, -0.34837072898298532264, -0.027688915189521535326, 0.10291662449595452566, 0.034500188007491466857, -0.028447715125608761394, 0.0081727131282851341043, 0.030346715448454807812, 0.00611836863403516628, -0.16350410203981635693, -0.35350516531252024199, 0.78338706266882995877, 1.0846428881246834575, -2.7909331659343274978, 1.0860321550630274157, 0.78341458787906081973, -0.35682818331852694804, -0.16751550745340851312, 0.0088971551293493215257, 0.041370768116247180757, 0.004820749031962414527, 0.0017336717925475742235, 0.019356394751133914472, 0.046210829968620345531, -0.003374685991804985026, -0.25639708189001542138, -0.094811233583397894198, 0.6389452700601500057, -0.71821081298891598088, 0.63767750020550106882, -0.096656672422074874085, -0.25103090857768090016, 0.015793395550380787823, 0.055407226864000534705, -0.0011569209410645419148, 0.083800182687696692097, -0.0079167361929162027162, -0.27617629617475536463, -0.14327293455647960307, 0.61429348204346878592, 0.14310105849351983931, -0.50307387562250716062, 0.19245014221910572427, -0.49790133494569033257, 0.13370098914968492343, 0.56446407816350629982, -0.21123525016482813665, -0.093052602645378665325, 0.007965782154397941725, -0.026322979491449727957, -0.067819865369623671958, -0.058895296409415535954, 0.23632093189737310435, 1.126811293282000026, 0.57742174470515539308, -11.614211940544727497, 19.625133137617968515, -11.618402627541918548, 0.60119464821375934171, 1.1761570294059837671, 0.22395025369815824146, -0.18801759394595052299, -0.0022316116571354286356, -0.0046248301979398675349, 0.022252370718651732356, 0.031225114529140157033, 0.022509691472035960741, -0.039444169842036712337, -0.31689829243547473769, 0.1526215827449879936, 0.26654973897878520006, 0.15007652550631528365, -0.31550228837222232814, -0.029214607223383776829, 0.024236232496620432192, 0.038874504562968509314, -3.2574087131212308362e-06, 0.0062679354113202032442, 0.020259745846915831746, 0.016164062748731045788, -0.1136124438624661892, -0.31409056581693300947, 0.56871492133346923659, 0.96315083455371974797, -2.290915806739958871, 0.963794789007260011, 0.56388481311735172419, -0.32745315129414309663, -0.11332917984118585919, 0.0566988286309477868, -0.016466056587632935115, -0.012676030078368628884, 0.017210238869944067402, 0.053408049485600989081, 0.1633881268862076297, 0.0064432584065929141698, -0.76624313316024939002, 0.038251278855565395975, 1.0227992183847038898, 0.040958735307596874642, -0.75350122544391928692, 0.026687488662158857788, 0.16477985878259501784, 0.016148950654830521861, -0.020707299169840039793, 0.077670198468789239388, 0.097700572376079569681, -0.12430953465809109482, -0.57300464486636337735, -0.26161603420595758163, 1.900824739239194372, 1.1427559123620290027, -4.4663349974931847441, 1.1535892394864768473, 1.8876043814658829412, -0.33747739999226344931, -0.64517561756348140722, 0.16647979868821266569, -0.0014463646214601642587, -0.037434250183281309043, -0.031012966463185399207, 0.054551315751057538606, 0.19909052422452092057, 0.0051240157903690002755, -0.5020137627536027658, 0.21022584388071643979, 0.19876333815059768484, 0.21258457217251791604, -0.49143824131065599881, 0.030300291042600593222, 0.21138595392435582943, -0.057757179346179035595, -2.6174981570294473507e-05, -0.0025217310639814513595, 0.00049352058564000893161, -0.018662716446664662923, -0.11474865497017901206, -0.12724750639079535075, 0.66303741797124393642, 0.73285675427061403031, -2.2631352506724962126, 0.73740850816292491388, 0.66188775338789529457, -0.14099058709329620465, -0.13183642355685792902, 0.002914466576424743112, -3.2343922232079764037e-05, -0.0095331029324250277091, -0.0081463229171375491772, 0.033936208771561698905, 0.053701722516913845407, -0.14795544438634239404, -0.038510275267536010768, 0.54753694441386713265, -0.85852200185250793218, 0.54550757173741915018, -0.043320858954149235598, -0.14481737547652048792, 0.070126439991650788008, 0.00067260851250020687285, -0.0003904959410464498926, -0.0080710760023772580574, -0.0072360821737307481211, 0.026766785929551568934, 0.047778449142560894258, -0.12676340835665733664, -0.033302629957085896906, 0.51517295451686551111, -0.82501273803465979828, 0.51375981532359160653, -0.037140938877644948113, -0.12436768566637924904, 0.060573647475331976464, -0.0012348071319577386649, 0.0076968954562173794512, -0.021028437870076670951, -0.018535194111210778062, 0.018233832416939252385, 0.094309680401672812966, 0.045843446477533376926, -0.32803247448016076282, 0.087139410576810905074, 0.22641563902387135121, 0.084996693853300303578, -0.32518734801460402739, 0.062913301915853411805, 0.10152090939887328802, -0.036126328737059641705, 0.0066869507756001506094, -0.015983459683869053608, -0.015324074155222173685, 0.0074058326205716047339, 0.079215372746552381589, 0.062635701599231666248, -0.30331311398840354876, 0.067228398407499764811, 0.22122014737878453738, 0.065242810451382943215, -0.30105589390201048516, 0.07663126961155532213, 0.0828888642112267926, -0.033385335692404247143, -0.011020752731027908325, -0.020470709696756395762, 0.022839535386161492758, 0.045087044556925984229, 0.037089332187075482739, 0.026934261456622112657, -0.30783687026857464719, 0.0200548874341242786, 0.36802630308300970929, 0.022237937481868877421, -0.29778052028243962024, 0.043227823949106791712, 0.025177487181000095251, 0.027224450636715441521, -0.0033679298269255063432, 0.0005292295874779774905, 0.020070060929717109532, 0.031364317356582363527, 0.0027048128703924590767, -0.12553088944403634208, -0.12264311583222378688, 0.40354098473859778551, -0.41100356301717416718, 0.40027727871705048868, -0.12730694820628482566, -0.12332338268699859218, 0.010592602669400850296, 0.044773229762822899791, 0.0087077031245654488989, 0.0056101080572740916544, -0.023993094717871855243, -0.082401002401010017828, 0.051428767486022985977, 0.2667272925241540027, -0.25072657809597798106, -0.13791721892132682847, 0.32731405820884346136, -0.139189966363723161, -0.24939920487901479085, 0.26704996850298096112, 0.031398285036191173059, -0.075107850244265667028, 0.0075871416198967480196, -0.0048179818726031380105, -0.018611475857346848617, 0.025839758890528946589, 0.057675722771336909722, -0.19486953604117149275, -0.0028697236353452336942, 0.65542461808738627926, -1.0495735025919363803, 0.65389908535980234561, -0.0085260719222891891145, -0.19282993190921238202, 0.082860754086705540322, -0.0113207225184142838, 0.0065056041329423506051, -0.002865486465710286295, -0.016111504282977050229, 0.020598454896318604823, 0.051826781688372319801, -0.17805865965621447966, -0.0053334951196487914715, 0.62456132059374080967, -1.0011469243646287097, 0.62330532266575711109, -0.010536443870876590831, -0.17703923703404886303, 0.073858531156372281279, -0.0096925757070236220891, 0.0066991721271424534492, -0.0049030684714709326169, -0.0081834357503260696937, 0.011676794522032811832, 0.027646235679131086344, -0.081644945868631330077, -0.067210157922349300641, 0.39160016554825377577, -0.55091352027435613881, 0.38963137257303898098, -0.071156875179956377564, -0.07667664571023249942, 0.040570641811195268245, -0.0074399717424509480279, 0.0059014229827351102203, -0.0032541958235587937348, -0.0069756034172581884295, 0.0047338312738070421198, 0.02143626663806833324, -0.060975517782603030903, -0.056685451391341255822, 0.36244820942303757061, -0.53264636789366615055, 0.36082693043107860742, -0.06018933611300840425, -0.057330773905858431472, 0.030694624128194458734, -0.0082789207887022688015, 0.0015339877213491233399, -0.015668453723781009673, -0.0092630445948486458269, 0.033261208002721635857, 0.09973839501640857641, -0.028667182283832422229, -0.34785343675512681028, 0.20591364308325738053, 0.12109353011053529403, 0.20317503131774392044, -0.34701979581392988772, -0.014622500729756628471, 0.11248775867456213551, -0.013573466270523725993, -0.0060330194954827595644, 0.017542571448354485342, 0.029039570241341912393, 0.0037473044695437102494, -0.163870023399393272, -0.35363157839381470504, 0.78338706245614875279, 1.0860321549968616761, -2.7875967321493484441, 1.0887898293357460844, 0.77660611991822314781, -0.37815562943960573383, -0.16931280597051326375, 0.072977767651531882942, 0.0034043241927689143513, 0.00061699802166701920268, 0.0018128845659808789838, 0.020649142140167026094, 0.047188951697676458086, -0.0033746856094109870493, -0.25777018763175801874, -0.097192198807855137432, 0.63767749976083998309, -0.71660575759203248847, 0.64099809359416626275, -0.10730222291671819013, -0.26817398281350462996, 0.096365054465689783947, -0.022274900527100290754, 0.035719555728814231244, 0.077124945487870286143, -0.016367480103477778652, -0.2786992569087570204, -0.14327293438451166319, 0.61580575460476927319, 0.14817214983312435028, -0.4979013348951152329, 0.18130114428633339885, -0.5308608833849514097, 0.1445201005243358372, 0.72803610886238101507, -0.44319265015977948874, 0.010508585537382316571, -0.0066343505405090704569, -0.028304301815817081017, -0.062861284236567130379, -0.055806335130738883576, 0.23632093156265734923, 1.1266557116813258332, 0.57830655350731463571, -11.618402627015949946, 19.603183073164782257, -11.629650285225608286, 0.69308573690633046294, 1.2808111203302798842, -0.12403098865451941957, 0.0011868171085926276787, -0.0071809663324438098922, -0.002269274570874055437, 0.024991251856084015226, 0.032342647414122814575, 0.022509691650116920442, -0.040592084136197088828, -0.31971439042836768962, 0.15007652536999180337, 0.26800531565932844691, 0.15353019384664698399, -0.32898840548900420977, -0.028267533422045972025, 0.073793398768561530021, -0.0010259775681216358554, 0.0011925279210987967116, 0.0070485878762542543072, 0.019803595725944304895, 0.015442608392739738329, -0.11361244390906030666, -0.31418236958772394196, 0.56761781359950902459, 0.96379478876727564529, -2.2848896350247995812, 0.96981754106775142077, 0.54519186145689468859, -0.37656455879554628918, -0.00062485291958803718809, 0.005082660696358678315, -0.023161591416377143304, -0.014383865613465544875, 0.019812512631462071006, 0.055448160228547919981, 0.1633881272573966259, 0.0049905053562971602929, -0.76740036193354443927, 0.040958735209364258156, 1.0281112304010675551, 0.033123419073117735256, -0.77987523021965510228, 0.065050263034153871233, 0.16971045961915146028, -0.02110071364287211329, 0.013162754341790922422, 0.073290362293548988459, 0.089233288840239718964, -0.12701422185943181398, -0.57300464460379563203, -0.26224504621187455511, 1.9025738318758087519, 1.1535892397409215349, -4.4532237411654431725, 1.14238494750264441, 1.8233075736106190146, -0.41152260635565218738, -0.3534546623441140456, 0.012778064281195360835, -0.017326610535634064114, -0.039409502496897980128, -0.028223855796652576322, 0.056410061667146120523, 0.19909052461054843608, 0.0044308539866119277945, -0.50195427206748555182, 0.2125845719955131996, 0.20355001492764321358, 0.21862544891824622906, -0.49139549115889835651, 0.031776421870668064562, 0.14171532288201568739, 0.0028296844833393319449, -0.00058987436781950683753, -0.0046179871072517609046, -0.0014766833025890706722, -0.019363402217414758116, -0.11474865458792989081, -0.1270264154553680247, 0.66442669756054051344, 0.73740850778582367475, -2.2553507116330822058, 0.74562617766383354478, 0.64770812263350663329, -0.19400888557701664872, -0.080255402743036177715, -0.0033804515943658942044, 0.0015293672386585330496, -0.0072486226502423006801, -0.0076642730057501513158, 0.033642433332549848135, 0.053701722561841573611, -0.14745021417235637529, -0.038406747488928402734, 0.54550757177813380405, -0.86243302160399337719, 0.54841878804001154002, -0.03636775526121341362, -0.15267844732721663759, 0.073339222436708734199, -0.0027014294578514897772, 0.0009327541673030512355, -0.0063427004854575905202, -0.0068886485579828298995, 0.026523395000740610655, 0.047778449253999349944, -0.12633990313307177922, -0.033143027453268789073, 0.51375981546447857529, -0.82794643448378046902, 0.51610736312421801042, -0.032453941383097123508, -0.12974293850054685606, 0.060945875633504037472, 0.0077980002864516811956, -0.0036658523392195246915, -0.019951917553188865811, -0.016349250095376934716, 0.018832980075420825, 0.094309680063056677835, 0.045842447864025033522, -0.32895251848876416734, 0.084996694019263865916, 0.22665696783997626995, 0.094311074480742884685, -0.31844848945643405846, 0.063672997939342321216, 0.051867435311787696284, 0.0061436149905086008538, -0.0023623365292321269089, -0.014995931318007886188, -0.013567441527854735145, 0.0077990501831742336733, 0.079215372282747970045, 0.062735012300285289411, -0.30407008106548155668, 0.065242810439412907142, 0.22108712634379082407, 0.072688227133813551606, -0.29558026564658729241, 0.083335155641201377752, 0.033048870845964029019, 0.0073194524056060450076, -0.02111318080302967623, -0.020560142974675794936, 0.024803016285413169439, 0.046318740062035353156, 0.037089331939644011593, 0.025983802965896477671, -0.30867092804067552958, 0.022237937679824060916, 0.37573758708077931168, 0.027592245385016908599, -0.32508138535993430507, 0.023491193425002519779, 0.085822850672521849846, -0.0056507212788991737634, -0.00053804409817735611085, 0.0040445689087055163372, 0.021384295689818484665, 0.031618260156621939838, 0.0027048130081946033382, -0.12595794866114323751, -0.12429088858414642782, 0.40027727840661192493, -0.41392152076333715849, 0.40188669871287607815, -0.13362477690202062552, -0.1263155186588017509, 0.06775481345396058841, 0.0024740258557215152706, 0.0055560044349028244945, 0.0051532808943485271941, -0.023051927503848150536, -0.082083814890136969078, 0.051428767573597267138, 0.26670008530141781078, -0.25150137797833027609, -0.1391899662174519714, 0.32707818342399047618, -0.1386973742613253624, -0.24223204930272751456, 0.32575915173695574323, -0.10693515508652415291, -0.00036380421939613842488, 0.0080873230454537627099, -0.0044452903133275836756, -0.018891648016534492993, 0.025434514314108067762, 0.057675722259828428318, -0.19433802351059070723, -0.00248523847178809109, 0.65389908538771124302, -1.0539316818599684389, 0.65614857291397266081, 0.0054687917892007132173, -0.19671150027699857432, 0.064355717752554775224, -0.00068570131971580963997, 0.0074644298997232315374, -0.0026464406477708490167, -0.01637675838694583122, 0.020272787190590394607, 0.051826781776369137966, -0.17764636878290918087, -0.0050526850377840930306, 0.62330532314771824876, -1.0049786376644145314, 0.62467397451018447097, 0.0012043113832994089326, -0.17758779778221947532, 0.056086596848210518163, 0.00014125599898334137731, 0.0070459367664238879356, -0.0037892750316741042466, -0.0095359863839925181062, 0.010249893948023535914, 0.027646235821138932798, -0.080136367646764850026, -0.065789767540714763627, 0.38963137270553405145, -0.55573392350000394568, 0.39376531219253152472, -0.062156042828243704412, -0.080430821461761656255, 0.028850307790361937738, -1.5097792442067417898e-05, 0.006449245654892890936, -0.0023642931770282479137, -0.0082124806609772463273, 0.0034997141720621845022, 0.021436266300397716078, -0.059690370355603823704, -0.055446559223760218116, 0.36082693004804849979, -0.53672244364581311871, 0.36436684195131957553, -0.053055357619934240265, -0.058129271308778068805, 0.016837402891582881753, 0.0022395621354460966759, -0.0037875981641236986658, -0.013457544731459417445, -0.007643769601220630483, 0.033025780730164773502, 0.099738394788159309701, -0.027932286037277563595, -0.34837072906195604149, 0.20317503146232643307, 0.11895010212327583599, 0.20727375959001845263, -0.34262004526589262543, -0.0074691217871896819352, 0.087529709800454338908, -0.0062030016683015014659, 0.0030113318077242632459, 0.017306177929830274254, 0.026849075505102681927, 0.0024815631719639344097, -0.16387002376665021908, -0.35350516483822375102, 0.78341458784995332554, 1.0887898294207645211, -2.7816557644524420034, 1.0890984059116586291, 0.75280438377753366996, -0.42803978567123535104, -0.031722562621642982905, -0.0005176526328993466021, 0.0018883691277688482209, 0.0023398314669395335955, 0.0033875320376057205772, 0.020649142762821457581, 0.046210828541761249377, -0.0038898280018526995398, -0.25710682731394590705, -0.096656671817362704857, 0.6409980930140832811, -0.71544369306078037241, 0.621657332937211482, -0.17245059908533980075, -0.093987714299519062311, -0.010821605707663009555, -0.0054429866746821898085, 0.033504237780710384442, 0.072899955882948383468, -0.016367480073451515926, -0.27617629625157086304, -0.14231983811128545714, 0.61237265073331226262, 0.13370098911628958138, -0.53086088340756798498, 0.17972645902421455211, -0.34231327445559206479, 0.3496077160867635536, -0.054850160909213303839, 0.0051531507952522993474, 0.0026889108928049437752, -0.0059579656028043333216, -0.026005261830289422287, -0.062861284853064192113, -0.058895295521328328014, 0.23458789470126575893, 1.1373293023003125946, 0.60119464725316618114, -11.629650284763014767, 19.497904092026011824, -11.649297496274295227, 1.0770032757585714656, 0.87973692372061396139, -0.001398189888800951226, -0.00041235504102516025698, -0.0041681840890588827894, -0.00040821864961098638198, 0.024991252004150903837, 0.031225114613155295629, 0.021038369584442061927, -0.040170837558962758929, -0.3155022884866222066, 0.15353019366869291584, 0.24973263746728352475, 0.12823723843246859211, -0.31362679248854202774, 0.064819094697297921503, -0.0013697550194528599569, 0.00048625281794363231893, 0.0017901049691847062753, 0.0065329918981835918834, 0.0198035957184957076, 0.016164062698489623654, -0.11419972907077198532, -0.31770404016551934134, 0.56388481321932326651, 0.96981754114487939145, -2.2688368155524440972, 0.97810671254893466475, 0.48794424775958977758, -0.34307307221646360373, 0.0085594232971534095489, -0.0083458282911543311272, -0.021461817950788428244, -0.011270508759548117372, 0.019812512870647485536, 0.053408048984252172453, 0.16392488727788931402, 0.014465238448865878548, -0.75350122509082051447, 0.033123418884463781686, 0.98197880987074914749, 0.0035640810579413103633, -0.79684509749691390113, 0.31097622385453643767, -0.0059578318158966542956, -0.012643681735228626609, 0.013043987076003460412, 0.070865053063497052133, 0.089233288918331668538, -0.12430953468891939306, -0.57218761588170241605, -0.26915516234312153454, 1.8876043812680034506, 1.1423849475244622909, -4.4447338631949060428, 1.2445274573988505118, 2.0275236411257302471, -1.0465229939332902909, 0.011135020733460428491, -0.00032063191262902474747, -0.018668845997161460926, -0.037631913542718338495, -0.0282238554960804848, 0.054551315290553625825, 0.19918524725186215196, 0.011021053947743981657, -0.49143824094027327432, 0.21862544876190401211, 0.19325555929299309543, 0.17806403652204455534, -0.59283635392231537331, 0.30518686811680112436, 0.0044960567760523708139, 0.00019376600146811098009, -0.0030617199717756457024, -0.0058162462306182481989, -0.0014766829131451911835, -0.018662717111340591802, -0.11448097014037417196, -0.12885482907693329713, 0.66188775400946553429, 0.74562617744915515861, -2.2385670691386403419, 0.75771349191332715289, 0.62918623908010473667, -0.2856006451731929463, -0.0031824402706627490794, 0.0004869481028781705691, 0.0015658207417554245462, -0.0074863182925289165071, -0.0076642730902964622403, 0.03393620853259161213, 0.054101951355437438718, -0.14887560494626148055, -0.04332085872556810241, 0.54841878817038325344, -0.8459618558548096523, 0.54257578388392901925, -0.12554059847239243397, 0.0006141520090028730703, -0.0023031665283995425911, 0.00020029146465568026247, 0.00087026514526568820793, -0.0065565841003227768452, -0.0068886485630663754631, 0.026766785629993780637, 0.048140399124647396956, -0.12737598800338995053, -0.03714093874936851003, 0.51610736331662876442, -0.81437869363537707112, 0.51316226008412235693, -0.11097763191510777447, 0.00021682023692764357519, 0.0024683493750549314466, 0.0048125514380570810422, -0.0037275783828887173142, -0.019533212381479189701, -0.016349250267487638877, 0.018233832943586734954, 0.093359852785172850465, 0.046040636086432551455, -0.32518734841062241214, 0.094311074730786967346, 0.22577662417970129072, 0.069357149362540046478, -0.35783977270993422293, 0.16898525471541853138, 0.0016449602039945328529, 0.0041698296075172959149, -0.0023523780774921520392, -0.014782940610620144348, -0.013567441760366323253, 0.0074058332138132558498, 0.078346980933062651542, 0.062657016255406899896, -0.30105589418928280221, 0.07268822725930858264, 0.2190645358831850853, 0.056068575039465329279, -0.32125431729820802529, 0.15158875919335268456, 0.008302477344035162285, -0.0041544909195898533164, -0.020549482116681883764, -0.018546343899502777675, 0.024803016081428954503, 0.045087044752154512017, 0.037208085972118630502, 0.030989241977546669315, -0.29778052045576519635, 0.027592245797470674185, 0.34073560350662351892, -0.021206616389465998695, -0.28072009405602998955, 0.12810740599478551149, -0.0058391911332215020586, -0.0010922335894057524659, 0.0018442039015084705101, 0.0049319520052072468239, 0.021384295824373695388, 0.031364317032601174429, 0.0022006962589382244103, -0.12776802986869284506, -0.12730694756920385435, 0.40188669856092540389, -0.41503688308120245676, 0.39178744903359125074, -0.15342599729268954767, -0.027153704695666171232, -0.0012176728182585921153, 0.0041187290947396829055, 0.0061292694422216487463, 0.0046805796809005510359, -0.023051927657229881463, -0.082401002307666726399, 0.049064851250350052791, 0.26461496335046669204, -0.24939920493918482647, -0.13869737430913053888, 0.32195994180352044411, -0.095611346043532483008, -0.21279193037447674541, 0.15454207505211417506, -0.002279008848320782446, 0.0033562104722467717317, 0.0071787308631618453048, -0.0051067129632048355661, -0.018891648338421104109, 0.025839759565916425171, 0.058327481171899078771, -0.1955343586283501689, -0.0085260721994876269292, 0.65614857280616933899, -1.0270706376446829822, 0.65992984829501077915, -0.12157092278806215513, -0.032215610682461634506, -0.0022132060799058628567, 0.0027445249453577204128, 0.0067638086188039885788, -0.0031974577307086527768, -0.016376758423927051389, 0.020598455177456825166, 0.052349227807431175652, -0.17867984476728002163, -0.010536444352128065266, 0.62467397490983256336, -0.98114418175899975605, 0.63137776940775314216, -0.11439685969628954343, -0.03238960646428917789, -0.0015082264737979994128, 0.0039760795222510891475, 0.0052292342346450110258, -0.0053970842741596861819, -0.0095359861446742757995, 0.011676794549544676147, 0.02922919536438544838, -0.081731909506236460894, -0.071156875364108695958, 0.39376531221286992235, -0.54089768819387840093, 0.39015395856499562877, -0.12035451098132683712, -0.0037105089431136808499, -0.0014095050340412651653, 0.0034520512017834547799, 0.0048782176381429055018, -0.0037864692387927882666, -0.0082124806931919303726, 0.0047338315646751200072, 0.022733408349476603993, -0.061204640814756308687, -0.060189336007288637975, 0.36436684151512349628, -0.52422528931510758987, 0.36451895297496539339, -0.10040820698936472888, -0.0053880341767965165417, -2.3399099015755009235e-06, 0.0019332416485889332512, -0.0034159336953652045567, -0.013535915243305583072, -0.007643769935764857873, 0.033261208197238295581, 0.09970614877483752192, -0.027688914761576321338, -0.34701979597495513641, 0.20727375972301476104, 0.1178953046154123524, 0.18691696203129376519, -0.39848312262426643926, 0.15045226339622666534, -0.0021620871453929008804, -0.0031624965260078708243, 0.0030733344243603177674, 0.016280677455085951905, 0.026849075458044345532, 0.0037473049472281751496, -0.1635041027487873222, -0.356828183023175316, 0.77660611958869041338, 1.0890984061010977602, -2.7627379565347323087, 1.1159525457941470439, 0.70588113264967133897, -0.44973037571375873434, -0.0034476788537005794398, 0.0015740752304013077081, 0.0035439399629549695261, 0.0023398310316679890933, 0.0018128841186214512699, 0.019356396035187253823, 0.048638988260371864014, 0.0030214981554014717045, -0.25103090943169814597, -0.10730222231497772711, 0.62165733319704730153, -0.72731819449243562126, 0.75244751973150447721, -0.36672201399238235631, 0.00044788955232225632474, -0.012194494645986221279, -0.0060902631695707153819, 0.033504237820640506151, 0.077124945433180075449, -0.0079167362251555545294, -0.27637478185470831793, -0.1721901795791268841, 0.56446407809501231156, 0.14452010056204481203, -0.34231327443611136996, 0.35270233841073733716, -0.96136506906050445131, 0.61216027060417965711, -0.00061011663957833816435, 0.0058434879203171806103, 0.0039253647665269102845, -0.0059579660305947848217, -0.028304300874046253028, -0.067819866439312911921, -0.062921240106638243184, 0.2403733599910831964, 1.1761570302492487805, 0.69308573643396576003, -11.649297496137302588, 18.971631541675119337, -11.800633992598921651, 2.5174384518326955451, -0.0032972128680268724885, 0.0015949954791191402886, 0.0012000207513653254193, -0.00416818391784495694, -0.0022692746253040398281, 0.022252370770495955715, 0.031761679850775534351, 0.028788775366732528849, -0.029214607003860935952, -0.32898840548671681727, 0.12823723828212454934, 0.27039614662039268556, 0.082324483017970226384, -0.20156510772019356481, -0.00085067165845683408859, -0.00036239551809727813501, 0.00025849879906149271806, 0.0017901050516719258041, 0.0070485876889722822586, 0.020259746066990391017, 0.0169053518759700247, -0.11476106907525537859, -0.32745315113410483621, 0.54519186133230823632, 0.9781067126273542689, -2.2314512674309465723, 1.2891730431200878471, -0.18116792170040366128, 0.0028524334851635999716, 0.0042880337346820332142, -0.0070676230318704119995, -0.021461817965696464855, -0.014383865836662027005, 0.017210239322642241488, 0.059424384601180248178, 0.18153857405000883851, 0.02668748806978141902, -0.77987522983628121409, 0.0035640809249844643096, 0.94707662656994440198, -0.14716770816758314999, -0.27830742975671945993, 0.0024170891188597452663, -0.010085085744184778614, -0.012813957807342284123, 0.013043987197250998464, 0.073290362310874310192, 0.097700572026478277321, -0.11533434999515755626, -0.58987974649025731733, -0.33747739992461128766, 1.8233075736577708525, 1.2445274573127544926, -4.0904684694964155156, 1.6874786201127320862, 0.22563641082843688412, 0.0045333342476872274737, 0.0068655037798052405765, -0.00064140071534253128172, -0.01866884570052790987, -0.039409502793648146235, -0.031012966085073555289, 0.055723733838589473155, 0.21015038235793281896, 0.030300290792235718557, -0.49139549092130524466, 0.17806403618128244237, 0.089180656796652671425, 0.12915438266466378048, -0.12514139898639781157, 0.0040373540832028577408, 0.0010116667856291657133, -0.0013391547780122420395, -0.0030617198945581338039, -0.0046179876264978185138, 0.00049352134431900757106, -0.017638144991689540947, -0.11917592802500784344, -0.14099058758122853807, 0.64770812295093260769, 0.75771349183221803347, -2.1833496056951902098, 0.98215376188765979304, 0.082336165438860367538, -0.00063226380555711438232, -0.0017557660449051043419, 0.00039014025515348902361, 0.0015658207495775032514, -0.0072486226187984496849, -0.0081463227016486065152, 0.031760927632925838038, 0.051528697437262759029, -0.14481737567869601424, -0.036367755290577445659, 0.54257578375217996314, -0.89307359028438926263, 0.6405152835385936605, -0.17742192178561810856, -0.00024568807523369761824, -0.0013758467639664187857, 6.4406991228113403408e-05, 0.00087026512633656821771, -0.0063427004716386601452, -0.0072360819558798467402, 0.025079905508379466555, 0.045993091378391366686, -0.12436768592211711049, -0.032453941424507803948, 0.51316225999394349255, -0.85331260555374488508, 0.5940382402637665038, -0.15469659795107915112, -0.00061197150494818862172, 0.0037551208212385890975, 0.0045980558133985405603, -0.0037275784139727953341, -0.019951917369737868363, -0.01853519460972085503, 0.015834372736249064523, 0.097184804353107626107, 0.062913302374217924817, -0.31844848986318741169, 0.069357149584535135456, 0.18198279665395372673, -0.013497199477511450175, -0.061927271677804915373, -0.00069702200255357176992, 0.002952293434771884658, 0.003938707084684530714, -0.0023523781278534435651, -0.014995931112434737406, -0.015324074614044197462, 0.0051551769795825600623, 0.081013760247176241913, 0.07663127018004313451, -0.29558026594347014404, 0.056068575139582424449, 0.18856734638020106187, -0.044021354346855272321, -0.042220428105801455221, 0.0031438635954879746819, 0.0047720934431625975855, -0.0039525867333719552979, -0.02054948211934179772, -0.020560142978498775063, 0.022839535312830731156, 0.048853700829664593874, 0.049891493731381381105, 0.043227824040055595756, -0.3250813857221851455, -0.021206616056981195118, 0.37562052023973357961, -0.061961737004467727641, -0.097321762117431931394, -0.00349881560343490031, -0.0022826133074392409086, 0.00019950254691419692932, 0.0018442042143752466164, 0.0040445685652368834617, 0.020070061409398029517, 0.030459141818397508089, 0.0030864581276562829369, -0.12332338317091748325, -0.13362477671878628116, 0.39178744873610088506, -0.40379145847567815553, 0.43699303863370475565, -0.22447093902745415828, -0.0018055861690382322797, 0.0014556058277904277906, 0.0041401720828939036698, 0.0061292693742969828544, 0.0051532812106484604878, -0.023993095049634899363, -0.086063652686995839591, 0.045377703608653473244, 0.26704996832496852299, -0.24223204906703629802, -0.095611346209031319865, 0.33226137927456467214, -0.39029357492593358403, 0.17979928084912377706, -0.0010216906129840021104, -0.00109392890130666459, 0.0031039858982774694122, 0.0071787306224919810588, -0.0044452899521808583033, -0.018611476267043607774, 0.024146199720724712684, 0.054931863324635982204, -0.19282993151034319435, 0.0054687920358438158602, 0.65992984800043996696, -1.1008470901515232665, 0.77945063742217179925, -0.21546042986362623495, -0.0010129410561426647943, -0.0011056308728424528529, 0.0026067885480338731305, 0.0067638087098154404259, -0.0026464405365852232063, -0.016111504837807037738, 0.01920967616225304625, 0.049240162493627059881, -0.1770392369535657151, 0.001204310851652709612, 0.63137776989394811533, -1.0459053402390654508, 0.73497862888721954278, -0.20164456051463544584, 5.2879637302015095176e-05, -0.0012812790855887506818, 0.0026271765063154658461, 0.0052292343939868363503, -0.0037892752043342040058, -0.008183435968429284002, 0.0098139161013713338555, 0.025304205028078313733, -0.076676645748244010026, -0.062156043039108960224, 0.3901539588672283676, -0.56910188393352201253, 0.41802710956286898147, -0.12992027343042264476, 1.353509909267107869e-05, -0.0011773491278038433067, 0.0022949237777631480845, 0.004878217531251309233, -0.0023642931695749356408, -0.0069756033559419203943, 0.0030708119233563840877, 0.018946312197822796214, -0.057330773648161109657, -0.053055357204146794126, 0.36451895266003758733, -0.54370539394500172481, 0.37701615976466690272, -0.10588690700106230536, -0.00051528135567244421657, 0.0012780238154556076459, 0.0014404387682285673972, -0.003415934070415783258, -0.013457544307939411921, -0.0092630447770330470725, 0.031236504521028045367, 0.10291662397139830776, -0.014622500583678010519, -0.34262004530393341817, 0.18691696204530822145, 0.06793613943247744924, 0.14527803806095165862, -0.15539615051541660895, 0.00062424082526909695657, -0.0029028202494966117049, -0.0036083680788462830537, 0.0030733345869688637533, 0.017306177880345625236, 0.029039569978120162441, 0.0061183692145166607493, -0.16751550781181145933, -0.37815562892106713289, 0.75280438327908016927, 1.1159525459984476203, -2.6986772154633449361, 1.4854743173607107831, -0.15585865823115283013, -0.0035478690756933490939, -0.00058852013816436757072, 0.0015740753088280959923, 0.0018883701130841862217, 0.00061699770492806870533, 0.0017336710579517564546, 0.022501379614293715026, 0.059587108736684088472, 0.015793396132955956485, -0.26817398326201780101, -0.172450599126394849, 0.75244751973372769882, -0.55922556964022673842, 0.15317022419512876796, 0.0033076568897087992201, -0.0048258618067957672129, -0.012194494663574790769, -0.005442986722759525324, 0.035719555800533452083, 0.083800182725332961198, -0.0072781374620139449583, -0.31161386058146828271, -0.21123525009587038048, 0.72803610886180358808, 0.34960771603759643877, -0.96136506901714247064, 0.11923624075368695341, 0.19610897168515764588, -0.0022666910304491730624, 0.0028577343050658677616, 0.0058434878522174030571, 0.0026889112438736964029, -0.0066343511735252758826, -0.026322978847201392721, -0.069029383537936542248, -0.080843085749977269816, 0.22395025327316758768, 1.2808111205580918757, 1.0770032757019358804, -11.800633992610841005, 15.632090379110506362, -6.2647946700687482391, -0.0019382991302179628131, -0.0002056957462972963831, 0.0015949955855854643268, -0.00041235497958733450998, -0.0071809663029131810819, -0.0046248305941518708453, 0.025702571464528606998, 0.040797126298187089133, 0.024236232632714956731, -0.028267533309729881569, -0.31362679266827353519, 0.08232448289270291597, 0.46284697805308988583, -0.28090700464636503275, -6.9219613617298265062e-05, -0.0010312551216786205403, -0.00036239553095893654215, 0.00048625256868107033905, 0.0011925282895731660008, 0.0062679352732805376086, 0.022440242534714199135, 0.024749900911834737949, -0.11332918011775502953, -0.376564558873391797, 0.48794424783328571582, 1.2891730432012502572, -2.3945592468551186727, 1.058733492911729801, -0.0013007751650011402465, 0.0049953850245888480747, 0.0042880336521393894872, -0.0083458282646521660814, -0.023161591287571847014, -0.01267603029994423397, 0.022775545483340094849, 0.053721125048665198387, 0.16477985921265339564, 0.06505026267710993515, -0.79684509732871844356, -0.1471677081691682154, 1.4786690016128933856, -0.80828905268453432598, 0.0030779781839324221217, -0.002061399585349610146, -0.010085085676136296834, -0.012643681855074752687, 0.013162754268697439766, 0.077670198826515574053, 0.10706862076732170941, -0.11744434916424631443, -0.6451756175442526775, -0.41152260636554932605, 2.02752364114753858, 1.6874786200894698052, -5.3349056153534455404, 2.6305923594991242886, -0.00030142103503552489588, 0.0055035825645194137723, 0.006865503865734115703, -0.00032063218871061392387, -0.017326610380985783821, -0.037434250406872926487, -0.031451143709483646294, 0.052536661710383021628, 0.21138595411573646299, 0.031776421841095893461, -0.59283635378885712885, 0.12915438250355978922, 0.6245352877523722146, -0.38726067104920475215, 0.0016445771124051920733, 0.0017462458644291853523, 0.001011666544453247718, 0.00019376582862677051047, -0.00058987391558318422874, -0.0025217314022480592263, 0.0011389722382136847033, -0.016174750426983703494, -0.13183642342101475897, -0.19400888577316513506, 0.62918623911462989717, 0.98215376203024218338, -2.4034967443319401603, 1.1337512728751713009, 0.0013668236861154479597, -0.00096336118599701947053, -0.001755766130767954375, 0.00048694787294438512427, 0.0015293673399688670865, -0.0095331028384415348459, -0.014131359109118045783, 0.033741713369308888593, 0.070126440060941278021, -0.15267844739297703471, -0.12554059825928617911, 0.6405152834996590272, -0.67603455098493725473, 0.23207203632808195981, 0.0012458906549933372772, -0.00061810625665522939826, -0.0013758468743665262443, 0.00020029138095731290666, 0.00093275412837136827127, -0.0080710758598084193782, -0.012163878293725146132, 0.026208753741841537416, 0.060573647563690775908, -0.12974293853357668493, -0.11097763174079437742, 0.59403824022858897624, -0.65029089745316537297, 0.22937784122348009452, -0.0010676609233370656666, 0.00062051770694128882862, 0.0037551208345527713195, 0.0048125515283793992433, -0.0036658525379937288192, -0.021028437616245652247, -0.018922775186579657292, 0.021529090756378238547, 0.10152090912149992763, 0.063672998293117941127, -0.35783977295291691245, -0.013497199399730800198, 0.48411692923263988098, -0.26573774406372269175, -0.00085895349848940164006, 0.00026102217295502396898, 0.0029522934423311460346, 0.0041698296696122015212, -0.0023623366633121292181, -0.015983459497609008026, -0.015466063259902606761, 0.0095903135565766628456, 0.082888863757360251028, 0.083335155937556124051, -0.32125431739695781141, -0.044021354352258214304, 0.45852147066434495448, -0.24323184337502981589, -0.00050443071589008499876, 0.0042975596317446056877, 0.0047720934590494397745, -0.0041544912310415241935, -0.021113180557262568177, -0.02047070951605871672, 0.027700806799445840817, 0.0541735550768918106, 0.025177486955532575597, 0.023491193443471586333, -0.28072009409753423403, -0.061961736781555872, 0.55902569507108634816, -0.31220889292175718621, -0.00016799664662821574388, -0.001885892758168005931, -0.0022826133235720261205, -0.0010922338621220861396, -0.0005380437694816104675, 0.00052922917308339091982, 0.01831251439594021313, 0.037715485503681388246, 0.010592602977039036055, -0.12631551871450033553, -0.15342599717535085335, 0.43699303859313654019, -0.26266341596353531829, 0.044689496748569118578, -0.00053434085399621598488, -0.00089182205424194796982, 0.0014556059036482430456, 0.0041187291336575769865, 0.0055560041700671754239, 0.0056101083183708981561, -0.020990434445580880007, -0.091347311817527948596, 0.031398285289931909814, 0.32575915152060769264, -0.21279193020898024002, -0.3902935750293211048, 0.53599427154545764562, -0.1945295103037056883, -9.5681494634857210168e-05, -0.0014368328641373281831, -0.0010939290827920032829, 0.0033562106854650689165, 0.0080873229346189196703, -0.0048179817600286908505, -0.024025318378963603921, 0.025582802907396257064, 0.082860753808197309955, -0.19671150055727706918, -0.12157092245243927842, 0.77945063727996977132, -0.86607847263719650588, 0.31825208483876488152, -0.00011928366462100152014, -0.0013165174769251620276, -0.0011056308011899097961, 0.0027445248304262233677, 0.0074644298540694721186, -0.0028654860847717473354, -0.020756790949273708907, 0.019962524519162042069, 0.073858531353699394262, -0.17758779755146500423, -0.11439686017421329955, 0.73497862918658174447, -0.82361966149898546696, 0.30445719471320553673, 0.00041478144611579846222, -0.0018368041787913723731, -0.0012812789188483913148, 0.0039760795268121073773, 0.0070459367024044870439, -0.0049030683412654083031, -0.012912997401433945824, 0.014091159286075193718, 0.040570641970445131541, -0.08043082122483334595, -0.12035451131212031306, 0.41802710958958544385, -0.39899101776824952292, 0.13822339541497666304, 0.00037022916419830796762, -0.0016308150618097540154, -0.0011773491632582500782, 0.0034520514325119397089, 0.0064492456365036789079, -0.0032541961211007960561, -0.010878156704638408322, 0.0065307521448696517763, 0.030694623878558841001, -0.058129271495465567599, -0.10040820670936617676, 0.37701615952520722841, -0.39014674753942618635, 0.14256800705495387138, 0.00019207503068081560149, -0.00031573753764288449344, 0.0012780236353601870814, 0.0019332418595074517959, -0.0037875984110668611989, -0.0156684534449905849, -0.011671491307757899877, 0.034500188329666933718, 0.11248775842779579204, -0.0074691218646811023338, -0.39848312241282368706, 0.14527803804673034005, 0.39783139387988153413, -0.25778694887427183735, 0.0012827079286713185657, -0.00099547167156696430776, -0.002902820113267541146, -0.0031624965710894512201, 0.0030113317573190827824, 0.017542571408400906702, 0.030346715350087333923, 0.0088971551622451568869, -0.1693128062767145503, -0.42803978505119399944, 0.70588113228097870699, 1.4854743173512447996, -2.9898894970798575166, 1.3476462017210726341, -0.0011397033602526267151, -0.0035478687212740227415, -0.0034476791954171958146, -0.00051765299043022704618, 0.0034043244219441504313, 0.0048207492710140385089, 0.0022865903889196827203, 0.010424440099785282562, 0.055407226833182478265, 0.096365054730027266405, -0.093987714643509495671, -0.36672201366987511317, 0.15317022402334376507, 0.15663971539008658684, 0.00036186574987681445741, 0.003307656880871777394, 0.00044788956499903093958, -0.010821605687538007196, -0.022274900558625833347, -0.0011569209508417947069, 0.081648002401781549775, 0.13357834805299031999, -0.093052602658654740009, -0.44319265016462933149, -0.054850160894278035539, 0.61216027058698474494, 0.19610897168993154938, -0.41196382573347528622, -0.00011622873617506356253, -0.0022666910199228532292, -0.00061011663105835158179, 0.0051531506999845282449, 0.0105085857043429199, 0.0079657819978827225371, -0.01334402027693843383, -0.077687078109185792041, -0.18801759384098670758, -0.12403098870078702043, 0.87973692371608880336, 2.5174384518538808209, -6.2647946700788139651, 3.2818432734785694116, 0.0015466051015551192903, -0.001938299073014919079, -0.0032972129094929231337, -0.0013981899988320409343, 0.0011868170920583530781, -0.0022316114504685653254, -0.0098986284142313883672, 0.0023162006977654123197, 0.03887450439931131041, 0.073793398723150716023, 0.064819094876208613321, -0.20156510764712862183, -0.28090700474680585508, 0.32815969940663219928, 0.00011120502968736837104, -6.9219641551437058932e-05, -0.00085067165546930871393, -0.0013697548641764044179, -0.001025977752139875087, -3.2574012491754799378e-06, 0.0042398703199529718527, 0.022718400013849730096, 0.056698828762548315885, -0.00062485281996144993472, -0.34307307231353334487, -0.18116792173314155701, 1.058733492943668697, -0.62063735758875659965, -0.0012797122243504478484, -0.001300775188168555932, 0.0028524335266228856289, 0.0085594233011890251406, 0.0050826606559600129992, -0.016466056554926524619, -0.040012814542760823111, -0.032856633737385100569, 0.016148950540694220757, 0.16971045974428333736, 0.31097622378235584284, -0.27830742975971689557, -0.80828905267288997383, 0.67717832124905530389, -0.00010010179221552124753, 0.003077978180569155859, 0.0024170891009168945021, -0.0059578317836847692265, -0.021100713608768528751, -0.020707299285481945567, 0.040959844289265287531, 0.16900639634906319819, 0.16647979865857437387, -0.35345466235978056924, -1.0465229939023628081, 0.22563641082114538339, 2.6305923594977476121, -1.8144508650721589138, -0.001958605426006902045, -0.0003014210785845742994, 0.0045333342693960540504, 0.011135020791876540716, 0.012778064253202218309, -0.0014463645595130904861, -0.038922303069163687039, -0.082009723822713584407, -0.057757179469405318939, 0.1417153228014612909, 0.30518686822840634942, -0.12514139898823625763, -0.38726067105989547823, 0.2166725359629173564, -0.0016843232008226024111, 0.00164457697805508226, 0.0040373542616473469605, 0.0044960568821137449069, 0.0028296843131706307661, -2.6174974607657741421e-05, -0.0032115147433270279168, 0.00030202216239117314319, 0.0029144665799674326911, -0.080255402724155877614, -0.28560064512902150202, 0.082336165300778710319, 1.1337512729504788389, -0.87992108709971239566, 0.0021701946731937228445, 0.0013668235441834043677, -0.00063226364248728157171, -0.0031824402034283227197, -0.0033804516285153496033, -3.2344002663534446307e-05, -0.002995645410924906854, -0.016234805356023684214, 0.00067260843753630806911, 0.073339222434548198559, 0.00061415203601493872313, -0.1774219219202351483, 0.23207203643647941749, -0.10924291370532578094, 0.0017446022860664136579, 0.0012458905342654033339, -0.00024568791763220629906, -0.002303166537056294122, -0.0027014294020166408801, -0.00039049605639971464747, -0.003144912327576392802, -0.014258273925964523934, -0.0012348071640346219152, 0.060945875590800148436, 0.00021682026845988705273, -0.15469659806182739503, 0.22937784131593652592, -0.11760043794816581186, -0.00041803738191608524628, -0.0010676609025032305755, -0.00061197152443540124653, 0.0024683493293427647558, 0.0077980003721676441689, 0.0076968954078972849978, -0.010116564313871395653, -0.037419477064827434687, -0.036126328657870132588, 0.051867435189456953526, 0.16898525479580747688, -0.061927271699323875598, -0.26573774406670458426, 0.17605209835747559488, -0.00027760835003420289207, -0.00085895348253345706665, -0.00069702201984221505166, 0.0016449601820851271396, 0.0061436150334272963611, 0.0066869507544666580839, -0.0065701717216303419378, -0.028103875771196037969, -0.033385335543123555169, 0.033048870739102641225, 0.15158875921564568534, -0.042220428083716163292, -0.24323184338975001317, 0.1576191407040614112, -0.001269710266077168177, -0.0005044307635113846975, 0.0031438635897344511562, 0.0083024775528796578017, 0.007319452270322019706, -0.01102075284558248075, -0.039174876928460555803, -0.030847131836061462973, 0.027224450777828864551, 0.085822850728134711384, 0.12810740591627867158, -0.097321762246389623141, -0.31220889282045177771, 0.23461991002161661224, 0.003129422923494727931, -0.00016799675126937473578, -0.003498815450008834621, -0.0058391911571893228422, -0.0056507213550117691564, -0.0033679296433300016586, -0.00087907129322843170251, 0.012183894023006219126, 0.044773229576709712429, 0.067754813409077949138, -0.02715370453777301088, -0.22447093918836699689, 0.04468949683690001673, 0.10317763297290671876, 0.00076062661873223550921, -0.00053434083148869377169, -0.0018055862141259622702, -0.0012176728310175004115, 0.0024740259423517456036, 0.0087077030567368190334, 0.016474389960045108389, 0.0056877963074204018062, -0.075107850344415597044, -0.10693515502388731275, 0.15454207500276601062, 0.17979928090169863819, -0.19452951032566451195, 0.0084609295697206775982, -7.9773311988296969504e-05, -9.5681527091537565662e-05, -0.001021690529216652283, -0.0022790088940853902612, -0.00036380423970521532772, 0.0075871416000384385783, 0.00881489647526991979, -0.012906291910652182425, -0.011320722417820490088, 0.064355717868966502704, -0.032215610821337323233, -0.21546042977996734957, 0.31825208480860589511, -0.12287864960008738902, 1.1284218246306080234e-06, -0.00011928365042785553981, -0.0010129410986859932464, -0.0022132060222681869746, -0.00068570133141052688319, 0.0065056040254948335755, 0.0088022920791595050949, -0.0092850482761068029586, -0.0096925758073829453715, 0.056086596849424553979, -0.032389606333116507941, -0.20164456064271185642, 0.30445719474673665905, -0.11843490060647630935, -0.0012664817646266231591, 0.00041478155987888438213, 5.287946831274194519e-05, -0.0015082264665259939324, 0.00014125604727141608799, 0.0066991720870765549337, 0.0064889151741055212802, -0.0068745464378825666199, -0.0074399717675758050839, 0.028850307737547171288, -0.003710508925958604879, -0.12992027329929442536, 0.13822339531222599396, -0.028110257241515383686, -0.0010692747023444185671, 0.00037022922877554730265, 1.3535047125953501347e-05, -0.001409505099595836702, -1.5097801290783773899e-05, 0.0059014231108530067205, 0.006652231559738539142, -0.0039565511103348688371, -0.0082789206869658040627, 0.016837402959836721833, -0.0053880343411665844455, -0.10588690678486434305, 0.14256800694521887252, -0.045047841812546333962, 0.00063120111221415470606, 0.00019207497307910251574, -0.00051528126567605290944, -2.3399298270191320627e-06, 0.0022395621989259634384, 0.0015339875860336466343, -0.010750453992993583285, -0.02844771519318805425, -0.013573466141848976318, 0.087529709843011949122, 0.1504522632552810768, -0.15539615052056582334, -0.25778694882593033988, 0.22733329982937244829, 0.00027689923012640035036, 0.0012827079651301345666, 0.00062424077085865704428, -0.0021620871945997985515, -0.0062030016044210371795, -0.0060330194160131394437, 0.0081727130487140205173, 0.041370768211109090751, 0.072977767694397330245, -0.031722562892332717421, -0.44973037551335509621, -0.15585865822703373618, 1.3476462016904964258, -0.83082883382738470512, -1.8658146605737357948, 0.87438709456528240782, 1.0389353073236116121, 0.59680494805188133434, -0.007996553012026633156, -0.34083262400917024859, -0.33453925369788617417, -0.18317467306750867539, -0.037862418151021010049, 0.044421185366102206027, 0.068029116192392741214, 0.053863753638439482163, 0.021062218341580032227, -0.01795457298922832276, -0.98530954773601542307, 1.3311548123207868066, 0.065995204578846930477, -0.38186019334035342654, -0.1067884215508196899, 0.030922976045451702753, 0.02705758922698679389, 0.00018500424074729694126, -0.0066319583462758346926, -0.0018364798993821824635, 0.0030589064457074554432, 0.0043365721103281636151, 0.0025126914009291094129, -0.0012623149418256424895, 0.32815969921983267898, -0.2809070037854385915, -0.2015651091603571643, 0.064819095782941801853, 0.073793398233685844612, 0.038874505161228714578, 0.0023162003012776608349, -0.0098986287406842860676, -0.0022316108382452147507, 0.0011868165837354198531, -0.0013981901671957944287, -0.0032972121343370344379, -0.0019382995846001190628, 0.0015466051845423383834, 14.222791113888577641, -13.184100384850490073, -4.1570422459971956286, -0.21802985841634464625, 1.398670349094998544, 1.5391360213926492673, 1.042945571130587723, 0.44148435308075500982, -0.00093366127404451669785, -0.21659516901354305962, -0.24906198935610429634, -0.17460021409656004132, -0.055525130371209624469, 0.071659139319917769528, 0.033064999505953082681, 0.051659764006665782887, 0.19497327172670117501, -0.076544209155356549013, -0.30679343553004206901, -0.11169518113252642488, 0.08301837768582094601, 0.11850697376171745456, 0.071983223618235558927, 0.014923338436958455064, -0.019520756337028412342, -0.02720239176511664303, -0.01631722141081783789, 0.0039167558392268399942, 0.41959444468201140266, -0.21686875114892020799, -0.26743975218517723746, -0.032813898126939418376, -0.013933675618667814428, 0.033196576567088291465, 0.072241513675432036945, 0.05453506347184034625, 0.015306800551396580051, -0.013194293507923752704, -0.020332333969859429923, -0.013577969882677657157, -0.0031459463545382344331, 0.0059150342777415142528, -0.85563921762596628628, 1.0769099461928426376, 0.20771299784829327928, -0.2767070656048299937, -0.21229396163215960569, -0.053207341000553520782, 0.040471605818659946163, 0.047168381915204511678, 0.020166785135420445263, 0.00017144631187741621097, -0.0049105011759845709415, -0.0033615189288361633363, -0.0012316199633038218963, 3.022864326770607912e-05, -1.1328517721620678937, 1.0724631420387786829, 0.45525705862415821734, -0.025798856121812475256, -0.20712294082797419459, -0.21485267471628582969, -0.094581389846617736072, 0.017021474523577946603, 0.057731651679390788567, 0.044552018070345707212, 0.016980929892405922238, -0.0025469229454365641813, -0.0093903204410265422863, -0.0068973289506034114044, -1.1689437940428366591, 0.53105896384202511129, 0.66504104185605306121, 0.41597454128441530008, -0.022019910328590901194, -0.24137877725126224981, -0.22147883290823547342, -0.1146270451109210331, -0.016324455657079368215, 0.037307657227071036599, 0.049671953020736403506, 0.035949576368661861903, 0.010589474043219594332, -0.017386328600108115483, -0.47651165608361023551, 0.16305934474801031531, 0.066215861142087190094, 0.2624889917604489864, 0.19521853569862807065, -0.0013559685128444586705, -0.10945716525832446808, -0.11494797077712161359, -0.068706073344452806007, -0.017947279929982164204, 0.015972619387176582861, 0.027616545167201092537, 0.020778498451153114279, 0.0025929780230800857088, -0.48628934638651716282, 0.14502533641905934747, 0.11682376153496495941, 0.29138164297300644945, 0.16863644671330960123, -0.036629032095212184228, -0.12410765348744663694, -0.11067375530107867454, -0.056648013606987734914, -0.0071302825276391007536, 0.021480224012434045999, 0.027887645654675542045, 0.017757350025870984045, -0.0017332256938957552213, -3.3063607094889424509, 2.5594840131675851147, 1.6205429516891609598, 0.11895803679774033523, -0.41717335110664183206, -0.478719198288631409, -0.31076603648949913117, -0.10802494098400974853, 0.027875737296041244728, 0.080384182164056600106, 0.074692190300131064418, 0.042657733512342539683, 0.0064297337515994553922, -0.024915956514353163748, -2.8829330996760806904, 2.3341435028139452257, 1.3527528725879864879, 0.024773253455115655519, -0.37782026212722652492, -0.3913505467188291842, -0.24321491992826008155, -0.0823885453790799277, 0.023589916031764846666, 0.065215877351740364487, 0.06083863978378940135, 0.034818373166691958531, 0.0051735138594334309156, -0.020469593335438188847, -3.5186879589415687164, 2.6156886268146468666, 1.7132388422813393802, 0.16126255582734988492, -0.37051188613223173807, -0.42874913187591057184, -0.30816652642152092545, -0.15338469982865537977, -0.024568472865193702326, 0.050866110036127919203, 0.071871355103542811094, 0.055542019154048639928, 0.021588725618757683561, -0.017368443468114336259, 0.044180053547014271065, 0.31490399665015905706, -0.1826164367645326525, -0.19025820863279011008, -0.11493372045166011519, -0.015347097415120490582, 0.06526208555783871712, 0.080329153272960196475, 0.052059577423206410129, 0.012917632572188848458, -0.014173870388755966526, -0.022393765201412521582, -0.014913507918653919754, 0.0018210787641795779697, 0.050335736911922057679, -0.3181172928660034871, -0.010357191384593825503, 0.30820144562389378828, 0.1427818171278190662, -0.033056695349814112228, -0.093456398271724791527, -0.076970351608985965175, -0.028459789885935924542, 0.0089376264581302496126, 0.021767941768006868891, 0.017009609425055943754, 0.0056533796994493344321, -0.0060117383193596229257, 1.4045124123599366239, -0.38446584086793816892, -1.1566562591726787801, -0.46455779920669715288, 0.13661733398180972432, 0.28409144923218415668, 0.2196104280333783243, 0.11528806445046670126, 0.031704612282682960445, -0.020895362976797657306, -0.042367855125293944762, -0.038911264647746961198, -0.019074301070065236702, 0.0089660943200513161522, 1.2744046492542762117, -0.39737234645514746267, -1.0448288698723688928, -0.37476728045843887838, 0.14425377791458030008, 0.25361276822473088899, 0.18949602430281753396, 0.096157548258539324304, 0.023156962972906511983, -0.02158053502788361, -0.038453719490238014167, -0.033679551687350081646, -0.015337906034896587465, 0.0093407908155476813278, 1.17676551956662645, -0.24570956001035645921, -1.0027524593653445173, -0.47187371686601553566, 0.075382858043356454347, 0.27956228181564668489, 0.24395630726956171741, 0.11791036918857224214, 0.011439161947059947599, -0.041605144302961111213, -0.049455118992756189056, -0.033337999283929102556, -0.0097427081934507789307, 0.013324014093860723357, 1.0403707327149669393, -0.25774475032459387425, -0.88470544912284554417, -0.37458095545895542244, 0.084673641829074677845, 0.24264122514197766778, 0.2055378688143754129, 0.095566651454810386257, 0.005648454783972137877, -0.037483363908392297592, -0.042524021062455494291, -0.027921863239778351579, -0.0074610436072909645799, 0.012267418413474855468, -3.2608013253505236939, 2.4284155181737943074, 1.5248153694661013002, 0.15693704608612049634, -0.29580859313175283942, -0.37170738138364073366, -0.27726521664710040982, -0.14695275512375399751, -0.040744975518705289774, 0.028418776244152808841, 0.059498184501652277845, 0.05600797173416118685, 0.027906588796969777899, -0.01303283475671838422, -0.16423900852798031802, -0.21376762039358099265, 0.21413570079017693604, 0.25103398107768243319, 0.043052453654037071973, -0.0051131523735202973058, -0.048919456135808170316, -0.081262449584010021386, -0.057987808305297636191, -0.012591611945727736524, 0.018584743280228896434, 0.025497881571750789093, 0.014819876195495759394, -0.0044550666767782303931, 0.8743870958237524027, -1.6669203114699535639, 0.092086977359387672815, 0.29104343365965090751, 0.44329930960965407971, 0.27944194482284545611, -0.021180519292508581675, -0.15755289650969109272, -0.14533895146584349667, -0.073759211496189192037, -0.012047356521196936688, 0.02125404451860146704, 0.029357214433988303159, 0.021062218597212408605, 1.3311548117096281274, -2.8190742438342142329, 1.145520453584168763, 0.77172239917307006163, -0.22997737546840141265, -0.19435650202129411968, -0.032563692657482677661, 0.021349851100551188454, 0.011060053801419923725, -0.0010022061456282562592, -0.0030255132742410539848, -0.00081085774000278019579, 0.0011228891841452582484, 0.0025126916158548389703, -0.28090700464188794738, 0.46284697756561288262, 0.082324483857418070953, -0.31362679320753455992, -0.028267532978048164755, 0.024236232140564129695, 0.04079712640352398112, 0.025702571609154269139, -0.0046248309065118151789, -0.0071809655133495600005, -0.00041235562478429200247, 0.0015949956641824213945, -0.0002056957490303805956, -0.0019382990698867961013, -13.184100385087363705, 27.382713710828937792, -10.32236376494598673, -4.1638497225828201564, -1.477907704368542019, 0.10090969642411291851, 0.68013681483535093353, 0.68608586776027513654, 0.44812440082154753052, 0.17850777311578505668, -0.01336015482206858937, -0.10057411833198021123, -0.10402752586610301833, -0.055525129432739857294, 0.051659763821410159546, 0.20910447488796074511, -0.41931420488854931339, -0.0056261502741440641215, 0.36682118330261948014, -0.071419780797922216564, -0.17439445069139017375, -0.043640063808485930752, 0.036194948378505606423, 0.049908598238646573186, 0.030390412498920896245, 0.0061979886444054853312, -0.0098541282208994308844, -0.016317221684110001095, -0.21686875121122106691, 0.36541314198775981525, 0.14725705049049844786, -0.32880281470872185778, 0.0045313354914762486378, 0.0060319862289264667907, -0.027338204407075041708, 0.010710780407699852651, 0.030945627595364243873, 0.02556156198349017819, 0.0069421393624500005171, -0.006840843451629255792, -0.0093558584327084019783, -0.0031459457096365277436, 1.0769099462356241936, -2.1321247120063815395, 0.67686826381359299631, 0.59734436082158404169, 0.011975249584232097039, -0.13668136339448019867, -0.11077223026091935487, -0.025860234498175083823, 0.021249118349612836076, 0.023056358274158264554, 0.0080458840717069629589, -0.0011734398973571814797, -0.0029111091322088331231, -0.001231619941355190399, 1.0724631417931091981, -2.0517563783903658425, 0.55711711388371265574, 0.44534728493083669543, 0.13384390005508750576, 0.032884284656202300567, -0.088637003499961106168, -0.10379952016233555401, -0.047262593330766468347, 0.0077533340270368299868, 0.026661776115131593373, 0.020350691534678447964, 0.0054317023865398279814, -0.0093903205125693747085, 0.53105896358367177257, -0.99881852778752511224, 0.063764350352252127663, 0.095032347902482577906, 0.32935218627197176966, 0.19499280348289296794, -0.02620123043647261396, -0.1064993598710378131, -0.093867482496392026792, -0.045648489388118981824, -0.0049117725200709637379, 0.016472973188340862133, 0.019700726553953797765, 0.010589473812865766306, 0.16305934491411944176, -0.59480943637286110004, 0.47306159196913449838, -0.13574291901505439184, -0.027731294499943139054, 0.13353046900158921373, 0.078770266237751643268, -0.0060886742151961825709, -0.049345818794502262217, -0.047134222616622159507, -0.027762396562105270265, -0.0064385821225443305715, 0.010134087243456463279, 0.020778497958925799816, 0.14502533639656314812, -0.50946687039035154498, 0.39253145942510736521, -0.16264996955721283678, 0.018173434812420140305, 0.15164315399754624281, 0.066199879503262948344, -0.020372475994982923442, -0.054740344831111856228, -0.045057113262344322679, -0.022558474292409019107, -0.0021360588551551061565, 0.011186997098950005047, 0.017757350095687778752, 2.5594840133726250997, -4.7652241672499595637, 0.61735852926035716948, 1.547254734305459678, 0.49836138642663335707, 0.020654069034611092365, -0.19120315788355859987, -0.21160369833565942366, -0.13402059930404150445, -0.044399295908697240798, 0.014812757530274162435, 0.03689292281655156408, 0.029880423143883720616, 0.0064297334808794508187, 2.3341435031226676067, -4.387742434166549188, 0.67783342532515433732, 1.4089671414102884484, 0.38951250176191071217, -0.022113128478093227458, -0.17472116823726696611, -0.17110727847673318225, -0.10376247491227606756, -0.033995540978915245178, 0.012062849721998832736, 0.029887190104957016279, 0.024361050697646203372, 0.0051735133382267272512, 2.6156886267547592162, -4.9280639497289673301, 0.58916745545859217081, 1.6644269870323424865, 0.54117008199621774445, -0.0082032753089872779279, -0.18871543446318897463, -0.18603571895791037871, -0.1299483769896574159, -0.063723444183195143697, -0.0075805443058415784494, 0.024931883026603762765, 0.031882552217862895894, 0.021588725209954574141, 0.31490399653915457368, -0.66596119934331632351, 0.484706834156719224, -0.016734210347925149859, -0.058304815526713141272, -0.050776275055360128863, -0.05357600409101274247, -0.006695788837987365523, 0.025152583979763630528, 0.03255843310969021992, 0.021459983891618357454, 0.0055297661901756323816, -0.0072870614222139426866, -0.014913507745574689986, -0.31811729260633814231, 0.51478720460265581238, 0.019714108353324957118, -0.43489881806690550148, 0.076815556289110284438, 0.15021192707424760315, 0.05582919671698205466, -0.0064392416008705556305, -0.039995600463047821516, -0.034347566032724617668, -0.01242643598817361715, 0.004904351061915941494, 0.010208167776011252414, 0.0056533793085267727493, -0.3844658408232323743, 0.36362846938327247903, 0.9679298546612503662, -0.49964852628067246698, -0.60224929615951272588, -0.14603840987390320771, 0.093963708731651579553, 0.13090795833973281614, 0.090019947149445680989, 0.045373691089787476782, 0.010673078985505578958, -0.010208938264036479179, -0.018948965202030425281, -0.01907430133824594698, -0.39737234657882419819, 0.40917805091250686322, 0.86958411261470514031, -0.51244384124660702362, -0.52990155023958529501, -0.10690377448127924565, 0.09032854043212845685, 0.11529838087664255608, 0.077178212056623632376, 0.037795483619030623146, 0.0075485435106498015301, -0.010175142843144668534, -0.016729882180158951732, -0.015337905838764737121, -0.2457095603160257824, 0.14403212281871080602, 0.86088414099788390654, -0.34849253498472615442, -0.4809150328825395615, -0.18210751290043369899, 0.037588835448250978999, 0.12689287234520851833, 0.10568104191534462288, 0.049745670511122754009, 0.0026208032188234052506, -0.02006063951513653576, -0.021487035715525737434, -0.0097427073987065779259, -0.25774475049183326325, 0.19097662622968789026, 0.76247837546510999207, -0.36569289100274071647, -0.41049777488883232301, -0.13864641820784731618, 0.038034879068274934921, 0.11020925710076912019, 0.089000140738992261635, 0.040007431665213415151, 0.00041041964709071648842, -0.017745444417820627897, -0.018048117721056293472, -0.0074610441897159989963, 2.4284155186781997138, -4.6559682229122811847, 0.68129426850135399452, 1.5101127175311823692, 0.44929435279033302919, 0.015121471218578347306, -0.16129180513185151224, -0.17388837701262058189, -0.11544292856213242415, -0.055786848756309730113, -0.013652015386484715742, 0.013472863245473041052, 0.026420133729467342593, 0.027906588437458131752, -0.21376762002537025298, 0.44777332465983948895, -0.3311045790217979623, -0.1447772541201329255, 0.25593935602827438958, 0.023614545129479660307, -0.011146132986757869426, 0.018793298446609874031, -0.013894509677024503499, -0.034987983632045638327, -0.025821149639568857781, -0.0054918717253540161663, 0.0092596676329361349983, 0.014819876060348662755, 1.0389353057484340681, 0.092086980190743156216, -2.1441781199177953887, 0.023969275537718603974, 0.41061638019939300515, 0.55330793423239954532, 0.36954520974932392186, 0.030380311656916094559, -0.14488752838648544241, -0.16160501780365008062, -0.099564274855655263141, -0.030513993443138886369, 0.021254044706229806988, 0.053863753351029233785, 0.065995205260024586424, 1.1455204525832634044, -2.5040381472074342284, 0.89866069775617274384, 0.76329273647422180016, -0.18385642447529326748, -0.17209257486312154461, -0.031428674020996673344, 0.01871466068649501463, 0.0090954340727822814011, -0.0028441988670490450132, -0.0048045462401904610872, -0.00081085765176444298593, 0.004336571703701425258, -0.20156510751277142979, 0.082324482242104868313, 0.27039614783205867532, 0.12823723734022471343, -0.32898840489918818042, -0.029214607665412357235, 0.028788775632786915704, 0.031761680366083587523, 0.022252370105148241292, -0.002269274418057555965, -0.0041681834374332421903, 0.0012000199436648364623, 0.0015949959470797627978, -0.0032972129585199204393, -4.1570422458357114692, -10.322363765387754242, 29.803901155145876345, -10.204518271503058457, -4.6583543332617400878, -2.0526871415851450919, -0.29012656322542790255, 0.51782938442371662102, 0.68901390953001384609, 0.53124280006820479016, 0.27276307417181805404, 0.046121318699976619759, -0.10057411755411073406, -0.17460021488882213059, 0.19497327217765969554, -0.41931420556278548029, 0.20199184775138417924, -0.0060380624370028335157, -0.017866755718851556195, 0.29368790784685055995, -0.09729773881544441505, -0.19941880253532529021, -0.058913097804005720715, 0.035581080577891173455, 0.058869791314450678477, 0.039606869875377301304, 0.0061979881362982259391, -0.027202390927966061535, -0.26743975214387283312, 0.14725705024221566464, 0.17973530897240022619, 0.29809274562008492016, -0.31517241248039457036, -0.053360167635049228496, -0.024406114495820724342, -0.038522077661759218659, 0.013110548612746174627, 0.037137342241766561901, 0.030948955809419285284, 0.010052547243156609649, -0.0068408427886177177596, -0.013577970605281619934, 0.20771299750856797584, 0.6768682638607727009, -1.9299638475987281883, 0.65973785874894763648, 0.61545006548734393359, 0.0080702236828662215345, -0.13835289961859187358, -0.11046173358703006984, -0.026808510271597273122, 0.018739474600536876192, 0.022137574952598722267, 0.0083724764314092349482, -0.0011734401235402383556, -0.0033615186737629330037, 0.45525705897917245757, 0.55711711372604211245, -2.1331707842252241036, 0.54769246882297906165, 0.57988626612902427482, 0.19201471320101054507, 0.036834337023251319254, -0.10225283542980266593, -0.11660451175177620475, -0.055994136678930626905, 0.004987871587649418334, 0.027877243844847385962, 0.020350691701484600188, -0.0025469229535134314812, 0.66504104226002058731, 0.06376434990286473925, -1.3962823306504659193, 0.076565073800240371349, 0.18108089629188306446, 0.3949565606781092475, 0.25695586708559953193, 0.0068176078834034231635, -0.10188034097960363411, -0.10822116751793268985, -0.064512166610548898893, -0.01751443731631120021, 0.016472972769703535939, 0.035949576919594460056, 0.066215860955979588542, 0.4730615920619136161, -0.93108754583808861494, 0.32515442645303954672, -0.11574262805964417866, 0.0011428211767622181716, 0.17052365761550031253, 0.11187739815452123204, 0.014891754548590221963, -0.046367845622532641303, -0.055848172595829088938, -0.03736918489476769023, -0.0064385820485695147153, 0.027616545146608446892, 0.11682376137268557503, 0.39253145948613876737, -0.88117307011690537699, 0.27985091169265297051, -0.12579914866977087717, 0.050318690319698697322, 0.18962054300061181866, 0.09721921345282286131, -0.0033141122115560028849, -0.054942070193623321339, -0.055282182066496536599, -0.03242693372217656661, -0.0021360584084715895925, 0.027887645368698966292, 1.6205429513130082952, 0.61735852968330084956, -4.6103613287877198701, 0.46266285481830105031, 1.6815638267331212141, 0.66256331058191053618, 0.12237234565649611484, -0.15648328669864822182, -0.22168178739971594604, -0.16174051344094944604, -0.069441810927659372576, 0.0011764849909768650138, 0.036892922176532082223, 0.042657734384082081347, 1.3527528720402206552, 0.67783342582766148432, -4.1997526354610359434, 0.53050990770596340163, 1.5155154492732876292, 0.52478068071645667114, 0.064371121612449039295, -0.14539794973907160802, -0.180343322003087142, -0.12723532948256249808, -0.054535180709486118233, 0.0010346490391992485634, 0.02988718923756542023, 0.034818374196342545546, 1.7132388429912082106, 0.58916745479497711901, -4.7109493669293707185, 0.4047632548420467824, 1.7225278666080283685, 0.68778280001865543802, 0.11830374640164180444, -0.12886686499624300239, -0.18148884754929553442, -0.15240330385312511607, -0.090858929795751691705, -0.02581685441713282847, 0.024931882450320089206, 0.055542019696015534902, -0.18261643675384162111, 0.48470683405959413825, -0.72442979734092749489, 0.56624407932289577694, 0.040536853693350985428, -0.078660708474081381114, -0.080773581566329841208, -0.081250744241082695218, -0.020581530348431596733, 0.024913299470688512333, 0.04073069370027683489, 0.029714441614997991192, 0.005529766308467187752, -0.022393765015819619424, -0.010357191678724963352, 0.019714108594898919341, 0.092111383752321071561, 0.050975857149974740379, -0.41302196078811820312, 0.088349259014919953592, 0.18256298160442119283, 0.072281540872262711739, -0.0039616633973282759998, -0.044749644971007959682, -0.040642765081963669327, -0.01680600233901782467, 0.0049043503667485915942, 0.017009610082146173204, -1.1566562596387148787, 0.96792985464560199471, 0.35699287027077469814, 0.8784944846852406064, -0.49457328681076540011, -0.64941616503420451778, -0.21535492009338441655, 0.044732220757424424884, 0.11748125011629492742, 0.10003194375621887735, 0.06396814640457987311, 0.024780680429463325387, -0.010208938707264017184, -0.038911263838308965402, -1.0448288698735752611, 0.86958411255510437154, 0.34547880416180898333, 0.79544488541915081559, -0.5028444354310603881, -0.57595902832170764452, -0.16890823816579284, 0.048636882101228573472, 0.10559340656828367089, 0.087229944103882711692, 0.054397124263298485269, 0.019740071310571913332, -0.010175142311545284229, -0.033679552029374670852, -1.0027524589516303433, 0.86088414091305398568, 0.21361196708896398877, 0.83960355178920220709, -0.38545319756455864724, -0.56451992728880928674, -0.24623071055238376914, 0.0049205447029435425699, 0.12477002310444434974, 0.11940963292170804488, 0.066123875287755901065, 0.012912730259524596993, -0.020060638033751972131, -0.033338001157890373438, -0.88470544891299918167, 0.76247837527905426569, 0.19472829123997417722, 0.75111789480927326768, -0.39418013164566140238, -0.4863218594161005659, -0.19258692401006832773, 0.011699289072852350757, 0.10936922408581933874, 0.1012773402090302588, 0.054235468414351339916, 0.009196813866538210841, -0.017745444788082612164, -0.027921862979251650672, 1.5248153690081103218, 0.68129426900319012272, -4.4871791418684985331, 0.46731798584581979439, 1.5765391704582729648, 0.57273917902451931816, 0.11907357258520953613, -0.09705284473014230695, -0.15591728809468693995, -0.13139807866604960584, -0.083346779648084234893, -0.034227004099612476362, 0.01347286299619063904, 0.056007971770410579226, 0.21413570030562051327, -0.33110457838150264553, 0.27142036335166863648, -0.19557754318836567675, -0.25452967280489652291, 0.23480841470388233438, 0.082579883607607276752, 0.024730485476563045938, 0.029035897553866905213, -0.017595153475810320493, -0.04444690335387518898, -0.034302604531911765062, -0.0054918718996186562109, 0.02549788142555618059, 0.5968049479301807958, 0.29104343303157270695, 0.0239692763278260626, -1.9956420897482072263, 0.10593718133320999264, 0.43307782984124287839, 0.54755985559809694507, 0.36377654681927529268, 0.024049464104450999496, -0.15339691332489591158, -0.16970228684317464896, -0.099564274828563475994, -0.012047356666546273421, 0.068029116325126384734, -0.38186019293613893133, 0.77172239922179952654, 0.89866069749691590562, -2.601637577295505821, 0.86743200780784501802, 0.79022498604554569646, -0.16234019085582504505, -0.1667884590058995864, -0.033769848973097330025, 0.015754919642279596737, 0.0073621378914644269517, -0.0028441994766423596419, -0.0030255131718224291884, 0.0030589070558133986399, 0.064819094824151310408, -0.31362679256301578823, 0.1282372381575307696, 0.24973263767940243052, 0.15353019369654063997, -0.31550228864232277015, -0.040170837112451417139, 0.021038369239991406429, 0.031225114716112636198, 0.024991251711163408461, -0.00040821835498070760308, -0.0041681842056571693708, -0.00041235490469692152003, -0.0013981899774822405345, -0.21802985772846444901, -4.1638497232542563964, -10.204518271600901969, 30.051624852729148074, -10.211568061970879739, -4.7368565347721052206, -2.118735845383723948, -0.3187215536419114037, 0.52476705967862480318, 0.71763359552988537082, 0.55928661092485243422, 0.27276307435929475487, -0.013360153151520146736, -0.24906199092253797356, -0.076544209294392429044, -0.0056261497677787810842, -0.0060380629532178807864, 0.19061724628347401223, -0.093293902511472917571, -0.07222377328892785886, 0.27710642098620336116, -0.087501611773066326205, -0.18694102658622707702, -0.050342494667888246229, 0.040280965484328627724, 0.058869791917598321418, 0.0303904132061702352, -0.019520757304926257547, -0.032813898038953071001, -0.32880281451236925694, 0.29809274518777911789, 0.17105708016826207118, 0.25847037339531525602, -0.33500792853356536893, -0.057374125369402929897, -0.015828142785725939079, -0.032146071168296631859, 0.014816968827424443003, 0.037199096103500181587, 0.030948955567264034572, 0.0069421398786136583395, -0.020332334222947359009, -0.27670706507948855668, 0.59734436067692020522, 0.65973785837500698026, -1.9756009170518153972, 0.58861718183585798325, 0.60141524878210372851, 0.023491712783825908356, -0.12322308927601371975, -0.10663583234874786376, -0.028410325205751232186, 0.017273739952235144662, 0.022137575107217508791, 0.008045884346641175508, -0.0049105015268757420954, -0.025798856593824193162, 0.44534728462697642515, 0.54769246994439768628, -2.0420836854419515127, 0.54320764126310117081, 0.54828758463723736849, 0.16865010249794645802, 0.034727414913852568124, -0.095334831563018862521, -0.11037011681705830024, -0.05355933260527676737, 0.0049878720398087855736, 0.026661775708341748509, 0.016980929744118682112, 0.41597454057405658379, 0.095032348359219226408, 0.076565074037310457133, -1.2939121108310975838, 0.12855794944470017027, 0.19628919801689997948, 0.3885352538088513441, 0.25050614674202864585, 0.0002944653148685902333, -0.10856739687682619155, -0.11372797148363283926, -0.064512166662719708232, -0.0049117722711063335761, 0.049671952781266785759, 0.26248899210667947512, -0.13574291892840209539, 0.3251544259705599349, -0.92218636006640331004, 0.36960361530020424681, -0.076959388492612357258, 0.014412642510996748202, 0.17256192559074304271, 0.10741041355675698399, 0.0081671663693776014314, -0.051821383928802325358, -0.055848172894559795776, -0.027762397930215836073, 0.015972620589936665364, 0.29138164325245247133, -0.16264996970020223488, 0.27985091153571495237, -0.85042853599024204048, 0.32756023842191117623, -0.09181568836249709431, 0.058246495044588730838, 0.18883526153618399035, 0.092069573752717098758, -0.0099131909653579902697, -0.06026688310475250554, -0.055282182044570041257, -0.022558474671126309274, 0.02148022399578697339, 0.11895803651772263176, 1.5472547343647200524, 0.46266285512064053709, -4.506312904349484505, 0.47459033530514005417, 1.6809586748130398437, 0.66183707627080046176, 0.12131285748363986043, -0.15892020963431224834, -0.22568480203351837243, -0.16641031711143350336, -0.069441810175908086356, 0.01481275763669747704, 0.074692189790277641048, 0.024773253801836699756, 1.4089671412417492657, 0.53050990758410665471, -4.1332126420026051861, 0.52644570924598887984, 1.5176012665206617402, 0.52926508062776955743, 0.065335877621174834329, -0.1476997021143043598, -0.18407470843744044853, -0.13115916637023433688, -0.054535179910855303975, 0.012062849758358544852, 0.060838639245978924641, 0.16126255454869678485, 1.6644269882657485304, 0.40476325515157424162, -4.6755267454221574042, 0.40076318368715224683, 1.7475409825962719967, 0.71223200620940685379, 0.12697656835444173784, -0.13279695912252406642, -0.19075266502050827744, -0.16098655934743355544, -0.090858929682036682074, -0.0075805442344975323271, 0.071871354999315975598, -0.19025820871084064834, -0.016734210370509840921, 0.56624407931278286643, -0.71916432138696739784, 0.5270796785740602397, 0.014796044133759171657, -0.093794594469113798385, -0.081311228885368944108, -0.075159180880448636652, -0.013065527691698979734, 0.030085133912550261065, 0.04073069385637633627, 0.02145998378847642496, -0.01417387041789121456, 0.30820144557812029218, -0.43489881791396545196, 0.05097585722362406313, 0.12709138921834764457, 0.10108644692819659938, -0.38367378057315937312, 0.082755923455525598476, 0.17060745453846154485, 0.064495783756920055385, -0.0064370881972802516946, -0.045331954588784320603, -0.040642764884632330558, -0.012426435771805211319, 0.021767941399254000373, -0.46455779814658959559, -0.49964852645170260148, 0.87849448311150057922, 0.27058000211801219725, 0.85816180694727262068, -0.49161112597631417831, -0.65792318294253926414, -0.22264367619654401587, 0.04519911413118156418, 0.12507140581987660766, 0.10772791177265565732, 0.063968146918807422541, 0.010673079518431728349, -0.042367856175647672368, -0.37476728010039966899, -0.51244384148200372575, 0.79544488486088316126, 0.27478050780295992794, 0.77915000724444316571, -0.49967348549092960308, -0.58365050133420393053, -0.17478702950035618358, 0.049563623604005666023, 0.11232905213326493943, 0.093824219491764809797, 0.054397123774954259101, 0.0075485430568258456846, -0.038453718707186453074, -0.47187371672325667538, -0.3484925356946357855, 0.83960355224603178126, 0.089173740331195627817, 0.77330488841441891168, -0.3851334279135999461, -0.55375379191205187368, -0.23815408864305956382, 0.0082908886868955644633, 0.12750344830132007146, 0.12244755946965510396, 0.066123874265439558373, 0.0026208025962009447174, -0.049455117660058216245, -0.37458095513913813868, -0.36569289125209469837, 0.75111789451520338812, 0.090622638846409642177, 0.69562015350907469013, -0.39159867101557394031, -0.47710779454347856943, -0.18606228780077593843, 0.014444671685919938109, 0.1118159121442999504, 0.10397611722649122146, 0.054235468373569052647, 0.00041041903970190016238, -0.042524020457809084961, 0.15693704582076284382, 1.5101127180776372505, 0.46731798555865822209, -4.4486677349730125997, 0.45409027890740938371, 1.5934829081280232899, 0.60500493117781128927, 0.13845822039828717398, -0.098436174842200380164, -0.1689172295873077323, -0.14358043528009784007, -0.083346779995947917197, -0.013652015697547835449, 0.05949818492889821786, 0.25103398070665366992, -0.1447772543264841949, -0.19557754213249073372, 0.19292449727351182465, -0.19017872078854292095, -0.21001571147301828502, 0.25680974668558170793, 0.083882457487227252146, 0.014936546237320555766, 0.019687716441431919734, -0.022725955853948539426, -0.044446903280542522652, -0.025821150262901229611, 0.018584743929819038527, -0.0079965526444680853219, 0.44329930907064762913, 0.41061638037199488238, 0.10593718160410574292, -1.9739137359108478442, 0.084490916387038136892, 0.4033064346318072646, 0.52420876960376550091, 0.35123715286870732522, 0.019027564941228813922, -0.15339691340903235939, -0.16160501734383275751, -0.073759210994261845129, 0.0444211844630820335, -0.10678842185465294012, -0.22997737546465135683, 0.76329273632375915781, 0.86743200849544188635, -2.6115179825696053229, 0.86444827298831905527, 0.79362888812342480271, -0.160017119753910253, -0.16733669322631122833, -0.035273009317120400385, 0.015754920652729244301, 0.0090954340131821406062, -0.0010022062316254690822, -0.0018364803045027545351, 0.073793398771314744344, -0.02826753337477070499, -0.32898840564309050905, 0.1535301941887055599, 0.26800531499253965606, 0.1500765260780471122, -0.31971439074767044053, -0.040592084064401623422, 0.022509691880991548885, 0.032342646605990392483, 0.024991252773992226471, -0.0022692751630324203144, -0.0071809660070065962564, 0.0011868170035680013132, 1.398670348514759576, -1.4779077034899095189, -4.6583543340260380461, -10.211568061435650989, 30.225487740275756465, -10.053773099420503456, -4.6413883720401170052, -2.076190412999582513, -0.3030153527699752658, 0.53210614526018562565, 0.71763359538525850478, 0.53124279963243015157, 0.17850777219840666121, -0.21659516779281753895, -0.30679343600097264133, 0.36682118345454262087, -0.017866755177810132726, -0.093293902741077752716, 0.14142627909239890993, -0.1156746727086927079, -0.068865492388450535999, 0.29275751490741352967, -0.072878673634362664546, -0.17990841354533412821, -0.050342494305643366304, 0.035581080739318662887, 0.04990859712751141708, 0.014923339145104066658, -0.013933675589909370884, 0.0045313353772947458434, -0.31517241226602471516, 0.25847037313314702445, 0.15377389825817991409, 0.26116331801215580954, -0.32255980411042145528, -0.048606514870164499453, -0.012965335370340991955, -0.032327118163602334977, 0.014816968934743820346, 0.037137342389521536823, 0.025561561633974674845, -0.013194293363318721629, -0.21229396188825130909, 0.0119752498546803407, 0.61545006524295187145, 0.58861718210803648521, -2.014934609948864086, 0.58264784536108493729, 0.6106268182963637825, 0.030682954592690343115, -0.1205465952873782387, -0.10660551104178968285, -0.028410325197805938457, 0.018739474592452579071, 0.023056358526533645564, 0.0001714460739086178171, -0.207122940184692933, 0.13384390009765942398, 0.57988626498832251954, 0.54320764163217538645, -2.0938581767843169601, 0.49943133007653428557, 0.53424680187175188628, 0.17280704697371698786, 0.044127958101607897556, -0.089532663955372043807, -0.11037011729775718372, -0.055994137328670645637, 0.0077533345858804246509, 0.044552018331186396538, -0.022019909511540987357, 0.32935218578573061121, 0.1810808961219958213, 0.12855794908759637929, -1.2776697830017182955, 0.11289897873423897223, 0.17359216890877773865, 0.37070399348833815711, 0.24156954577445857724, -0.0028719074184887765505, -0.10856739698609188527, -0.1082211668696503315, -0.045648489627733969398, 0.0373076570770112248, 0.19521853501514419005, -0.027731294148714086267, -0.11574262734489615401, 0.36960361518155254723, -0.88484004222593060085, 0.38551433068168972085, -0.077046648742971263202, 0.0035171520508584241949, 0.160894957785467696, 0.10079392414506055387, 0.008167165794124367495, -0.04636784538948650336, -0.047134222341500539577, -0.017947279965777031274, 0.16863644638149857813, 0.018173435125311783456, -0.12579914855060750356, 0.32756023846768927976, -0.816261390980081325, 0.33796804749810011037, -0.096162252314756280214, 0.046028706283591847015, 0.1777059087976422258, 0.086126846467962314691, -0.0099131921259024285975, -0.054942069942719454212, -0.045057113498482639913, -0.0071302818726194788779, -0.4171733503516109054, 0.49836138590926387648, 1.6815638264251466794, 0.47459033516962606525, -4.5451388069666327496, 0.41871842704220157971, 1.643628475428291269, 0.64773443106575556705, 0.12100261078631438516, -0.15712424577326133956, -0.22568480146762723293, -0.16174051390498908609, -0.044399295469844522743, 0.080384181794174752733, -0.37782026245476513227, 0.38951250193760383889, 1.5155154492651050635, 0.52644570951234825884, -4.1685627895562493705, 0.48107698999885828428, 1.4880978875534791595, 0.51841777685218126681, 0.065417374233219666158, -0.14618485179611645508, -0.18407470816567142458, -0.12723533038399131856, -0.033995540985191995131, 0.065215877874656991153, -0.37051188549074010048, 0.54117008135370137101, 1.7225278666274279615, 0.40076318346796829806, -4.6876426660255585688, 0.36867830737356349635, 1.7187187365148097751, 0.69572256139263910502, 0.11987436447584202959, -0.13613191584379649712, -0.19075266456489303746, -0.15240330393640116879, -0.063723444664301895357, 0.050866110293479094295, -0.114933720579918075, -0.058304815464549922255, 0.040536853925255980635, 0.52707967839935998544, -0.75660958841377645712, 0.50768263115054712564, 0.015986808305871312574, -0.083490373779399773557, -0.070386198700167490472, -0.069002615942312911512, -0.013065527721377631631, 0.024913300083361573922, 0.032558433672002978776, 0.012917631641466467349, 0.14278181702774561668, 0.076815556324566436341, -0.41302196094926080239, 0.10108644718093136305, 0.16718045358164940195, 0.10524976943960838438, -0.39689920766436598276, 0.069385831887629645376, 0.16458930588067186007, 0.063485984927559302449, -0.0064370883119590866991, -0.044749645369982701815, -0.034347565918918201988, 0.0089376266880368850087, 0.13661733308581627822, -0.60224929583023989466, -0.49457328555486468336, 0.8581618062777088829, 0.25977352055513958096, 0.86439428301633691731, -0.4754997564870934168, -0.6447062270046460819, -0.21371301867534703289, 0.05016980696261710837, 0.12507140482650180857, 0.10003194424558256581, 0.04537369031152421045, -0.020895362066038772331, 0.14425377784109666424, -0.52990155043800635237, -0.50284443448590820935, 0.77915000651639720886, 0.26725718401183751194, 0.78636622494426589025, -0.48455580683458659985, -0.57202059511217229204, -0.16734632247164277219, 0.053588008459495170555, 0.112329052182786715, 0.087229945603905095775, 0.037795483738014425124, -0.021580536159671610125, 0.075382858082509413, -0.48091503270131247261, -0.38545319765751484553, 0.77330488824715748475, 0.07601954173307645668, 0.79465889824078794756, -0.35697878737063842847, -0.53764439694971954964, -0.23330859736382256187, 0.008560948350201075252, 0.12750344819975314992, 0.11940963318938115079, 0.04974567077450705449, -0.041605144670242066862, 0.084673641736684027159, -0.41049777483810317991, -0.3941801312672018609, 0.69562015304980839847, 0.082804429999398021511, 0.71553021312248377139, -0.36756084690466495069, -0.46386363924222434596, -0.18219370403031026617, 0.014622773903458785583, 0.11181591212813983549, 0.10127734005230354386, 0.040007432036612874948, -0.03748336415347000572, -0.2958085933539425505, 0.44929435272060386231, 1.5765391709417184707, 0.45409027873129037545, -4.4673417563500263938, 0.43574397292685601579, 1.5795184306174041566, 0.59153326978682219117, 0.12663702840073276512, -0.10622344615035764215, -0.16891722996103336607, -0.13139807870052613858, -0.055786848960875115577, 0.028418776588098201841, 0.043052454589173290433, 0.25593935598541978083, -0.25452967445967944471, -0.19017872020473750028, 0.25063551368367298355, -0.16528356408240993392, -0.2139057268372950571, 0.24292906429702643134, 0.072203611217195304284, 0.0089943521021088202599, 0.019687717170989646465, -0.017595154718753079626, -0.034987982262330422079, -0.012591612687091675513, -0.3408326238518182838, 0.27944194310807807247, 0.55330793581864856634, 0.4330778297461083115, 0.084490917053748157417, -2.0062761016694650174, 0.053663269505231370182, 0.38251762720822929387, 0.51521370555226075894, 0.35123715257405480061, 0.024049464623338366598, -0.14488752977291505242, -0.14533895195774512721, -0.037862416791619848966, 0.030922975423455210248, -0.1943565014945062841, -0.18385642437934471283, 0.79022498653004891178, 0.86444827252233658044, -2.620703292998368017, 0.85910285418610043973, 0.79213523340815072959, -0.16062029667261817911, -0.16733669321619809578, -0.033769849365835553912, 0.018714661052340993141, 0.011060054218261020087, -0.00663195875984691572, 0.038874504356705578967, 0.024236232977354821988, -0.029214607786274989637, -0.315502287777508994, 0.15007652513041264131, 0.26654973917953805884, 0.15262158249355911588, -0.31689829251321344206, -0.039444169665619997767, 0.022509691687950510869, 0.031225114209178280594, 0.022252371076382754245, -0.0046248305721153107686, -0.0022316115165386006189, 1.5391360216507548042, 0.10090969691318130153, -2.0526871417345073922, -4.7368565357327900855, -10.053773099459792917, 30.415176706071772372, -9.9246667420353453792, -4.5807891308553232435, -2.0592161026578312821, -0.30301535343095525876, 0.52476706032590769713, 0.68901390943266627165, 0.44812439961666294952, -0.00093366031198836439871, -0.11169518115902857258, -0.071419780535983745651, 0.29368790753260998194, -0.072223773123116008499, -0.11567467282493687497, 0.14380896970938492063, -0.10378526643678646912, -0.056268006527956819507, 0.30045259465982609992, -0.072878673437968000126, -0.18694102676768936666, -0.058913097976537175116, 0.036194949110855648144, 0.071983223087528247497, 0.03319657620763039052, 0.006031986700276926322, -0.053360167686271275933, -0.33500792848092558751, 0.26116331785452290282, 0.16807130454961585753, 0.27189235027554670143, -0.3185106283726590326, -0.048207015287184296071, -0.01296533531771411861, -0.032146071437888910916, 0.013110548329724112243, 0.030945627485784082955, 0.015306800979893312939, -0.053207340902710294206, -0.13668136339274480906, 0.0080702237497043641939, 0.60141524855004635519, 0.58264784519487566872, -2.015644696561561755, 0.5818830335454353575, 0.610970316778589706, 0.031366468512543210778, -0.12054659533712096253, -0.10663583243101458498, -0.026808510576301465556, 0.021249118317778706994, 0.02016678542481752795, -0.21485267443492606088, 0.032884284482486002277, 0.19201471329807887067, 0.54828758413960299034, 0.49943133038879689023, -2.1156544638715573114, 0.49561459648477568241, 0.53909005657500130848, 0.17772350009535714532, 0.044127957795971277377, -0.095334831349652440768, -0.11660451118811179794, -0.047262594064420133944, 0.057731651889240849507, -0.2413787771299563123, 0.19499280348924022399, 0.39495656066079942725, 0.19628919757965512805, 0.11289897909516728725, -1.3015099687751867652, 0.090366957000361791685, 0.15918003293984078761, 0.3649512158690778052, 0.24156954622155291101, 0.00029446501235822972187, -0.10188034133018349092, -0.093867482519625289328, -0.016324455365521062938, -0.0013559683533633029584, 0.13353046860546421981, 0.0011428210425399182304, -0.076959388151041932535, 0.38551433086494657498, -0.88413330411748258797, 0.37547524184881786447, -0.089592483113014931284, -0.0045461645074331702432, 0.16089495813057713702, 0.10741041354192934487, 0.014891755372863496598, -0.049345817975734757033, -0.068706074755767954088, -0.036629031296125040551, 0.15164315321520760449, 0.05031869038736266897, -0.091815689049070059458, 0.33796804819821624211, -0.81945911068840759928, 0.32637043042799029902, -0.10861348924116207004, 0.038506458007111582509, 0.17770590833810495646, 0.092069574591745279024, -0.0033141125632213941365, -0.054740344453931812085, -0.056648013983153028272, -0.47871919859320988255, 0.020654069378546618257, 0.6625633103539064761, 1.680958675286033932, 0.4187184266489690776, -4.6014169151100077215, 0.38314920992260603771, 1.6306650530127642096, 0.64648596382899981005, 0.12100261112277711784, -0.15892021012952245451, -0.22168178727522214189, -0.13402059919130984711, 0.027875737274964503681, -0.39135054694259574504, -0.022113128425038288544, 0.52478068080751161251, 1.5176012669111953457, 0.4810769896753940289, -4.2158883803210640551, 0.45103759911960350504, 1.4775281599351732797, 0.51765247319788987301, 0.06541737426278737555, -0.14769970212966265777, -0.18034332122377064178, -0.10376247432016373617, 0.023589914902131407792, -0.42874913187518370883, -0.008203275201945260317, 0.68778279952833942801, 1.7475409832364170448, 0.36867830710117760473, -4.7359740464507131819, 0.33000855421616021035, 1.6987009785287381369, 0.68917762998850817802, 0.11987436476958533704, -0.13279695974110769519, -0.18148884779574131487, -0.12994837595526731455, -0.02456847334581474529, -0.01534709710368784702, -0.050776274977371034514, -0.078660708925845673756, 0.014796044121386187076, 0.5076826311424924576, -0.75849213574683738148, 0.51648514537172673933, 0.027241905054525540619, -0.076459892643445034888, -0.070386198842220762395, -0.075159180666225733347, -0.020581530946604281729, 0.025152583521638127717, 0.052059578209868062615, -0.033056694793694378631, 0.15021192692644788624, 0.08834925849148431154, -0.38367378052279410561, 0.10524976915133849031, 0.15865997829559938537, 0.095053338333873152455, -0.40310103902704602641, 0.067220541905313832554, 0.16458930575828900622, 0.064495783729149450125, -0.0039616629133208567867, -0.039995600428701136297, -0.028459790235313368184, 0.28409144927619972565, -0.14603840989633262715, -0.64941616485221942412, -0.49161112642295867747, 0.86439428340914747917, 0.27667416838997133111, 0.88523521482626532642, -0.45889140357807356452, -0.63678140779486314127, -0.21371301870226105946, 0.045199115119344268621, 0.11748124997638650924, 0.090019947116015894273, 0.031704611843659684378, 0.25361276765948298495, -0.10690377360606494184, -0.57595902886869154713, -0.49967348539726619272, 0.78636622545850287658, 0.28381756212709940979, 0.80532829371590108902, -0.47024782512584129002, -0.56550723644118305877, -0.16734632129913307175, 0.04956362328722345445, 0.10559340609868553273, 0.077178212061899148511, 0.023156963123440024721, 0.27956228133267463365, -0.18210751258978644107, -0.56451992696322228493, -0.38513342824806962028, 0.79465889878956530445, 0.1035605253651252361, 0.81616652892552110643, -0.34555059397901816132, -0.53438992832527354437, -0.23330859722663457867, 0.0082908887633026742692, 0.12477002304870449256, 0.10568104146447113156, 0.011439162307688398312, 0.24264122465250109739, -0.13864641802989019492, -0.48632185916665870984, -0.39159867060411346351, 0.71553021278906459024, 0.10693307170892271674, 0.73392650879395482377, -0.35794305568916967575, -0.46121053729785760344, -0.1821937040477359937, 0.014444671602578326719, 0.10936922445746061205, 0.08900014017645754838, 0.0056484550611236791773, -0.37170738074491760372, 0.015121470565854498377, 0.57273917901807125386, 1.5934829081023904607, 0.43574397271933262843, -4.4992202053680934171, 0.4022666412169321748, 1.5559461057058103783, 0.58069071117337134336, 0.12663702804519033629, -0.098436174830469388985, -0.15591728821992981158, -0.1154429281774956989, -0.040744975667333836744, -0.0051131525843872002376, 0.023614545069290438645, 0.23480841501556945494, -0.21001571115447900673, -0.16528356466186705509, 0.24941580234884333511, -0.17897586784173780128, -0.22666040269226073178, 0.23623997752870395939, 0.072203610890912858777, 0.014936545627004260001, 0.029035897873554573911, -0.013894510317340919708, -0.057987807520925548077, -0.33453925328930672611, -0.021180519494335567354, 0.3695452089197590051, 0.54755985597726686276, 0.40330643469387417177, 0.053663269984802156043, -2.0287601873030052246, 0.042384390487830991834, 0.38251762676829076382, 0.52420876956221729159, 0.36377654610137349911, 0.030380312673983594224, -0.15755289598302088527, -0.18317467373194257219, 0.027057589805584262438, -0.032563692822831405427, -0.17209257506532754656, -0.1623401918605488492, 0.79362888876434933838, 0.85910285442396738809, -2.6258231803160314577, 0.85678892166767295357, 0.79213523325320012169, -0.16001711932414044992, -0.16678845936856420695, -0.031428673997250133887, 0.021349850878628760104, 0.00018500448022414531343, 0.0023162007454764527332, 0.040797126145042619805, 0.028788775643989322761, -0.040170838038436455786, -0.31971438987558969957, 0.15262158244146054042, 0.27048323786091221788, 0.15455597585401040228, -0.3168982926622010976, -0.04059208370193301274, 0.021038368997309313257, 0.031761680278018532275, 0.025702571336564512816, -0.0098986284128804152144, 1.0429455700510747995, 0.68013681485553734163, -0.29012656155965099725, -2.1187358453214475418, -4.6413883726554132636, -9.9246667419860123971, 30.508592097035382551, -9.8847689122008457474, -4.5807891317399525022, -2.0761904116650962138, -0.31872155385162304064, 0.51782938389504717946, 0.68608586785699499089, 0.44148435314182293832, 0.083018378444727480847, -0.17439445138867967766, -0.09729773891591927959, 0.27710642106165128684, -0.068865492747651579908, -0.10378526604715321291, 0.15515993754604118227, -0.097290590926518427661, -0.056268006587159614951, 0.29275751476778505422, -0.087501611229037074002, -0.19941880297961522839, -0.043640063474177884228, 0.1185069735513543665, 0.072241514174886861666, -0.027338204833386128834, -0.024406114598654823217, -0.057374125314051685476, -0.32255980435030440212, 0.27189235051680610278, 0.17270529370994713858, 0.27273799541373899835, -0.31851062863753504262, -0.04860651475905393426, -0.015828142480581858936, -0.038522077315369426809, 0.010710780478746319971, 0.054535062976368425813, 0.040471605528457556267, -0.11077223011941633724, -0.13835289937105677005, 0.023491712752595369368, 0.61062681840278709711, 0.58188303344259384442, -2.0187077852298260794, 0.58055458339037246951, 0.61097031651326794144, 0.030682954751717721187, -0.12322308914251688117, -0.11046173346633580969, -0.025860234899433391653, 0.047168382115652260356, -0.094581390866681949725, -0.08863700267890405049, 0.036834337260947480619, 0.16865010301860858255, 0.53424680127328172397, 0.49561459648570649339, -2.1137911271203537744, 0.49825216208420941566, 0.53909005640341900456, 0.17280704716171582658, 0.034727414872449957617, -0.10225283561658644882, -0.10379952026153722866, 0.017021474723788315719, -0.22147883316706232226, -0.026201230082885065831, 0.25695586653419832945, 0.38853525497519081355, 0.17359216798734602993, 0.090366957211679782569, -1.3169023093507146172, 0.083151791520184675788, 0.15918003320886575658, 0.37070399291451533452, 0.25050614729642406431, 0.0068176077202831029608, -0.10649935944874971971, -0.11462704554368123577, -0.1094571651328947598, 0.078770266381062437056, 0.17052365747326214684, 0.014412641886538124503, -0.077046647867552109368, 0.37547524118193487119, -0.89616236975058150005, 0.36763625819599804467, -0.089592481846267085777, 0.0035171511696038548767, 0.17256192573555897973, 0.11187739742764811479, -0.0060886745443190642968, -0.11494796993811617902, -0.12410765432817803811, 0.066199880595687238261, 0.18962054227510299897, 0.058246496479274711966, -0.096162253510997885986, 0.32637043068499860254, -0.83138963657075048985, 0.3189670525205481888, -0.10861348938014400711, 0.046028706961895954342, 0.18883526053480784168, 0.097219214250973573499, -0.020372476525555016647, -0.1106737551179636514, -0.31076603651501588654, -0.191203157954828451, 0.12237234594480597694, 0.66183707575744443741, 1.6436284760687123185, 0.38314920960048609588, -4.621820838626431005, 0.37607988453894192205, 1.6306650531087496514, 0.6477344306546966024, 0.1213128578187185197, -0.15648328671559005865, -0.21160369873765541571, -0.10802494065253936961, -0.24321491906120795701, -0.1747211687548843595, 0.06437112147628536174, 0.52926507964230318226, 1.4880978883478823693, 0.45103759904885643017, -4.2332762593622677016, 0.44512925455681850995, 1.4775281600892655742, 0.51841777647298348253, 0.065335877948306350649, -0.14539795000905969546, -0.17110727855428981004, -0.082388545163060086796, -0.30816652636855584868, -0.18871543468825330292, 0.11830374639155211147, 0.71223200648375406541, 1.7187187364675802215, 0.33000855431341641344, -4.7636211705808104711, 0.31755541576812523763, 1.6987009789343985311, 0.69572256076103633493, 0.12697656865819448302, -0.12886686484441942091, -0.18603571919732822537, -0.15338469972735074842, 0.065262086330060808947, -0.053576004515552062835, -0.080773581722870718691, -0.093794594824483379369, 0.015986808313885894628, 0.51648514522873467669, -0.74819065723397537582, 0.52309250830305287217, 0.027241904441860920194, -0.083490372867339263108, -0.081311229223075681682, -0.081250743397919628075, -0.0066957886494521503054, 0.080329152379820023544, -0.093456399170428333667, 0.055829196995092471301, 0.18256298240837903757, 0.08275592339882166526, -0.39689920712779763745, 0.095053337661058556751, 0.15405796591143627872, 0.093647150105668250375, -0.403101038731339123, 0.069385831646824214225, 0.17060745469903346172, 0.072281540159220170061, -0.0064392417889888899085, -0.076970350934601639015, 0.21961042840879180238, 0.093963708664913797586, -0.21535492066416841994, -0.65792318220107137705, -0.47549975738844196815, 0.8852352150552117438, 0.2949085086757856633, 0.89546025582631616491, -0.4588914034627385452, -0.644706226999305132, -0.22264367612395366547, 0.044732220052152736611, 0.13090795912962041592, 0.11528806420558242174, 0.18949602466705545156, 0.090328539702543161893, -0.16890823753060515977, -0.5836505014940562841, -0.48455580730335612305, 0.80532829409508477347, 0.29981491637830964825, 0.81401133452858720485, -0.47024782395909980881, -0.57202059624511358837, -0.1747870289121163101, 0.048636881177116564334, 0.11529838109594719997, 0.096157548648736712593, 0.24395630729486020294, 0.037588835552386831718, -0.2462307107464219158, -0.55375379159172422039, -0.35697878824353262983, 0.81616652975120151936, 0.11585303448130136739, 0.8211359533000814892, -0.34555059405631433078, -0.53764439682142195487, -0.23815408899587270364, 0.0049205451199345329183, 0.12689287229413404035, 0.1179103690945149524, 0.20553786881673405396, 0.038034879181144412574, -0.19258692402083799067, -0.47710779517382151393, -0.36756084610046202421, 0.73392650836369821565, 0.11751193969872744671, 0.73817759947049077507, -0.35794305566241574335, -0.46386363961774229603, -0.1860622873017410972, 0.011699288440843752523, 0.11020925762009313453, 0.095566651343868075363, -0.27726521643995616362, -0.16129180513489602133, 0.11907357180096775351, 0.60500493166908431153, 1.5795184310942551598, 0.40226664079435514587, -4.5297420524241900353, 0.38607547667553526738, 1.5559461055919734385, 0.59153326964792363185, 0.13845822030576973583, -0.097052843929553808677, -0.17388837668944859272, -0.14695275587776149284, -0.048919456584979321978, -0.011146132713899820771, 0.082579884257548225035, 0.25680974569783171502, -0.21390572591156953308, -0.1789758682664820133, 0.23726027472663907547, -0.1855716753833184518, -0.22666040334897133435, 0.24292906487490656398, 0.083882457256228906695, 0.024730486391006188673, 0.018793297852707665369, -0.08126244970584306504, -0.18317467353064925684, -0.15755289562855567809, 0.030380311975052327195, 0.36377654608803122738, 0.52420876894391899015, 0.38251762801344357801, 0.042384390013443781764, -2.0287601872912008893, 0.053663269993280901526, 0.40330643458918347122, 0.54755985628107994856, 0.36954520853192246443, -0.021180519230027630706, -0.33453925337438239485, 0.00018500446156499239601, 0.021349850850578919875, -0.031428673806251486433, -0.1667884596669885755, -0.16001711899204956735, 0.79213523289171128283, 0.85678892195844524871, -2.6258231805200042963, 0.85910285465025548746, 0.79362888851971769011, -0.16234019172174982137, -0.1720925750167912327, -0.032563692950550858496, 0.027057589860824994565, -0.0098986283607730880807, 0.025702571211162476839, 0.031761680464627649323, 0.021038368713913468272, -0.040592083363274182595, -0.31689829303173439046, 0.15455597626781242271, 0.27048323746724395322, 0.15262158281626450607, -0.31971439020410219456, -0.040170837939097057967, 0.028788775804420976834, 0.040797125959086549818, 0.002316200806790162485, 0.44148435319398443566, 0.68608586783315883562, 0.51782938364166231082, -0.31872155345802882787, -2.0761904118338048164, -4.580789131802331049, -9.8847689119060433427, 30.50859209655926918, -9.9246667415386937705, -4.641388373199424322, -2.1187358448350153139, -0.29012656172110751207, 0.68013681500694234128, 1.0429455699122744949, 0.11850697367428934814, -0.043640063844568185991, -0.19941880228519104357, -0.087501612194071820561, 0.29275751566613866794, -0.056268007217089147098, -0.097290590602308019141, 0.15515993753222356855, -0.10378526627881894118, -0.068865492270210937331, 0.27710642049607836501, -0.0972977385770846942, -0.17439445144577997437, 0.083018378422732672095, 0.054535062916194899929, 0.010710780633838935805, -0.03852207737937860671, -0.01582814262971155711, -0.048606514491294522395, -0.3185106289512110167, 0.27273799561485623233, 0.17270529370841636307, 0.27189235052048521535, -0.32255980439218867595, -0.057374125450086910227, -0.024406114355934233062, -0.027338204932112999263, 0.072241514169027215564, 0.047168382040965815827, -0.025860234808316513722, -0.11046173343880934248, -0.12322308921606141274, 0.030682954904835173932, 0.61097031622451913879, 0.58055458368089107868, -2.0187077854773112229, 0.5818830336759985844, 0.61062681818884156915, 0.023491712897723213138, -0.13835289936372602293, -0.1107722302157316957, 0.040471605574284523488, 0.017021474593329564279, -0.10379952011704748904, -0.10225283569489733437, 0.034727415048492915606, 0.17280704706392893688, 0.53909005649444352581, 0.49825216201504657354, -2.1137911272714076105, 0.49561459640048383113, 0.53424680174298933011, 0.16865010265737492845, 0.036834337462718427791, -0.088637002901725284176, -0.094581390753805297256, -0.11462704552543118153, -0.10649935978315129559, 0.0068176084314338965958, 0.25050614654035940898, 0.37070399373484780314, 0.15918003253601220592, 0.083151791727267859722, -1.31690230913383588, 0.090366956370439593949, 0.17359216942273145157, 0.38853525356765122867, 0.25695586751376436885, -0.026201230623188020985, -0.22147883300276302587, -0.11494797012987564844, -0.0060886744137279152911, 0.11187739704255778017, 0.17256192692810765066, 0.0035171500441381065016, -0.089592481079887681394, 0.36763625743851513494, -0.89616236893063949687, 0.37547524060648562738, -0.077046647805744106208, 0.014412641723640734481, 0.17052365780011333185, 0.078770266216543427529, -0.10945716507066377554, -0.11067375512543613236, -0.020372476683559501143, 0.097219214527078115506, 0.18883526028737829261, 0.046028707453027672969, -0.10861349014720488027, 0.31896705327957941556, -0.83138963718438674277, 0.32637043128042797502, -0.096162254225786780726, 0.058246497095848337455, 0.18962054209291329121, 0.066199880478366265879, -0.1241076542616323386, -0.10802494073940523134, -0.2116036985209796828, -0.15648328682334594619, 0.12131285783002358769, 0.64773443052762968986, 1.6306650531652751024, 0.37607988465870217931, -4.6218208388822175081, 0.38314920998148549458, 1.6436284756315728828, 0.6618370763488038433, 0.12237234536364448301, -0.19120315773929424941, -0.31076603651013562368, -0.082388545161742918199, -0.17110727855320537194, -0.14539794991736071372, 0.065335877753414484115, 0.51841777660443544296, 1.4775281600984127017, 0.44512925444057166446, -4.2332762594223014574, 0.45103759943224969176, 1.4880978879498343304, 0.52926508005386674327, 0.064371120933559733412, -0.17472116839426579671, -0.24321491913838971799, -0.15338469968171891677, -0.18603571940218557867, -0.12886686436804928313, 0.12697656787085725982, 0.69572256167645907254, 1.6987009781454842638, 0.317555416210283048, -4.7636211703868598377, 0.33000855362201209076, 1.7187187372112957551, 0.71223200587584445387, 0.1183037467669465076, -0.18871543484215427822, -0.30816652632595858963, 0.080329152041267781748, -0.006695787979491942335, -0.081250744223248308473, -0.081311228041876029704, -0.08349037400333278458, 0.027241905200639807139, 0.52309250788393579779, -0.74819065745780799492, 0.51648514612351814623, 0.015986807347791114065, -0.093794594033954542533, -0.080773582469801027961, -0.053576004079842830252, 0.065262086251086495747, -0.076970350769998155394, -0.0064392422574685720393, 0.072281540650185077679, 0.17060745442758837309, 0.06938583177204998842, -0.40310103877352532198, 0.093647150168613621513, 0.15405796580231181925, 0.095053337694948850101, -0.39689920716353388475, 0.082755923548026813497, 0.18256298229950443379, 0.055829196962824269301, -0.093456399127944414507, 0.11528806440993619176, 0.13090795858993231615, 0.044732220906785288839, -0.22264367711357666013, -0.64470622644576580651, -0.45889140335704392548, 0.89546025549617447314, 0.29490850863686363104, 0.88523521540302452237, -0.47549975747463985032, -0.65792318240019886932, -0.21535492046920623466, 0.093963708619705446634, 0.21961042837666325256, 0.096157548566581860228, 0.11529838130036948463, 0.048636880903059932779, -0.17478702861101680366, -0.57202059632516100152, -0.47024782412069349169, 0.81401133425086202866, 0.2998149175605295258, 0.80532829253429527583, -0.48455580611703413796, -0.58365050214109492899, -0.16890823730314055706, 0.090328539705378643743, 0.18949602464614315189, 0.11791036919091216206, 0.12689287241042201959, 0.0049205447153035791752, -0.2381540888305106185, -0.53764439663299712358, -0.34555059450414976618, 0.82113595359104374349, 0.11585303466807042594, 0.81616652936471090207, -0.35697878800292659474, -0.55375379158661675039, -0.24623071077337771451, 0.03758883558199010988, 0.2439563072418265699, 0.095566651440856659194, 0.11020925751885055077, 0.011699288353630009352, -0.18606228700561017186, -0.46386364032467614082, -0.35794305453756425361, 0.7381775984985576855, 0.11751194013937921179, 0.73392650819427540654, -0.36756084589270582308, -0.47710779550114584824, -0.1925869237662543032, 0.038034879186279346719, 0.20553786875595025929, -0.14695275574282989917, -0.17388837705037743508, -0.097052843661165871803, 0.13845822027982912505, 0.59153326962421770574, 1.555946105657880052, 0.38607547656931723301, -4.5297420523793139324, 0.40226664087821811799, 1.5795184310316108256, 0.60500493151904732869, 0.11907357197120839454, -0.1612918051763528593, -0.27726521643694063135, -0.081262449853874416106, 0.01879329828875572414, 0.024730485598822217475, 0.083882458304071760846, 0.24292906402663230758, -0.22666040291181896449, -0.18557167544943933835, 0.23726027440780153577, -0.17897586769027420961, -0.21390572669024704444, 0.25680974669580342029, 0.082579883322208030849, -0.01114613221473059225, -0.048919456693151251725, -0.037862417859244643092, -0.14533894988850934449, -0.14488752982466246433, 0.024049463240773942541, 0.35123715323084025997, 0.51521370488335271887, 0.38251762753587842392, 0.053663270067402409069, -2.0062761020929396061, 0.084490916988880768179, 0.43307782961280705258, 0.55330793520421128751, 0.27944194451869791562, -0.34083262444563983129, -0.0066319588035263802792, 0.011060054306846062167, 0.018714660940259888328, -0.033769849102178013667, -0.16733669364819561687, -0.16062029615691580942, 0.79213523275252528677, 0.85910285494571347886, -2.6207032936986549565, 0.86444827306241711629, 0.79022498614404190764, -0.1838564240678197681, -0.19435650173500126914, 0.030922975515674383518, -0.0022316116110137318961, -0.0046248302255049080628, 0.022252370497757693185, 0.031225114865353346266, 0.022509691091158322285, -0.039444169175395601967, -0.31689829300718080907, 0.15262158310201409672, 0.26654973848932855818, 0.15007652588480646649, -0.31550228863748813746, -0.029214607014066241303, 0.024236232575283542939, 0.038874504445128341135, -0.00093366035281699393845, 0.4481243998077982793, 0.68901390938166728883, 0.52476705973978154951, -0.30301535285112329854, -2.0592161025044593003, -4.580789131162609884, -9.9246667418813618866, 30.415176705987139627, -10.053773099618959819, -4.7368565356599328098, -2.0526871415304732693, 0.1009096967999551786, 1.5391360216359344371, 0.07198322313375048187, 0.03619494890416825178, -0.0589130976170970308, -0.18694102701710618719, -0.072878673336007823713, 0.30045259470617108333, -0.056268007020793452122, -0.10378526575626591255, 0.14380896927330674706, -0.11567467247267279595, -0.072223773405881985799, 0.29368790766470215248, -0.071419780671226909541, -0.11169518107220653691, 0.015306801107263378822, 0.030945627157952904085, 0.013110548624748032467, -0.032146071570520864946, -0.012965335334410185014, -0.048207014964547774405, -0.31851062897932352902, 0.27189235097452180145, 0.16807130387057039878, 0.26116331823129712575, -0.33500792848206589758, -0.053360167685212289701, 0.0060319865837905207312, 0.033196576267202390631, 0.020166785587178510247, 0.021249117917713684217, -0.026808510241788392064, -0.10663583252419812919, -0.12054659548239572697, 0.03136646884172320654, 0.61097031642922150496, 0.58188303390790019343, -2.0156446968959862431, 0.58264784525693835704, 0.60141524877353713663, 0.0080702236002153059091, -0.1366813634487963891, -0.053207340849043785469, 0.057731652120327096867, -0.047262594240472147189, -0.11660451169773569657, -0.09533483086639869819, 0.044127957604868879582, 0.17772350048631743724, 0.53909005650061392334, 0.49561459594295420583, -2.1156544629748950115, 0.49943132929100231276, 0.54828758525531484658, 0.19201471249771556726, 0.032884284903221261409, -0.21485267459083345898, -0.016324455234252376046, -0.093867482842486429928, -0.10188034127861540779, 0.00029446533842996357178, 0.24156954587305357052, 0.36495121626165200057, 0.15918003269944736022, 0.090366956666417780353, -1.3015099681236417251, 0.11289897838355696102, 0.19628919853489423808, 0.39495655993354156221, 0.19499280363988844011, -0.2413787771103312052, -0.068706074577186720953, -0.049345818750722095225, 0.014891756490753305098, 0.10741041208274634822, 0.16089496017486271118, -0.0045461668442714764371, -0.089592480635520388432, 0.37547523954189304662, -0.88413330212370166983, 0.38551432922994433028, -0.076959386902304319222, 0.0011428201085092876133, 0.13353046914715532845, -0.0013559685257364534339, -0.056648014005409697003, -0.0547403442184980632, -0.0033141133079945400697, 0.092069575522255781008, 0.17770590716594836245, 0.038506459816229104964, -0.10861349107510287482, 0.3263704318668252391, -0.81945911214456357285, 0.33796804977621375254, -0.091815690218394863109, 0.050318691081868455184, 0.15164315284876830892, -0.036629031230575676004, 0.027875737353000723528, -0.13402059935438098703, -0.22168178727475160161, -0.15892020965596651005, 0.12100261047359249511, 0.64648596410294933978, 1.6306650529423201146, 0.38314921008612257136, -4.6014169154287429819, 0.41871842712250029184, 1.6809586748940743561, 0.66256331059668005512, 0.020654069174103971418, -0.47871919849730965035, 0.023589914976045501388, -0.10376247448058062428, -0.18034332120544155353, -0.1476997017690788172, 0.065417373883538965096, 0.51765247321037588524, 1.4775281598808314154, 0.45103759933994069797, -4.2158883803713598226, 0.48107698964413170284, 1.5176012668895504376, 0.52478068088637130906, -0.022113128582965810298, -0.39135054684810943737, -0.024568473433148187851, -0.12994837579495707724, -0.18148884799968656445, -0.13279695942990044077, 0.11987436453978406181, 0.68917763003895071705, 1.6987009784585986871, 0.33000855439100656996, -4.7359740467917150752, 0.36867830764556142054, 1.7475409827330363743, 0.68778279976520406702, -0.0082032752329795032009, -0.42874913188932939345, 0.052059578392598865326, 0.025152582898994517435, -0.020581530229779265717, -0.075159181219154214126, -0.070386198488042686638, -0.076459892688673369854, 0.027241904693347855637, 0.51648514604550166407, -0.75849213649441771246, 0.50768263212997655209, 0.014796043282694217036, -0.078660708294886139136, -0.050776275695504434482, -0.015347096774081428472, -0.028459790325052573834, -0.039995600200323777607, -0.003961663176979417049, 0.064495784023326854784, 0.16458930552859366014, 0.067220541745817125046, -0.40310103845408984391, 0.095053337665007078816, 0.158659978825478859, 0.10524976891837388437, -0.38367378063236751151, 0.088349258703666236348, 0.15021192683350545471, -0.033056694784866898218, 0.031704611896921863767, 0.090019947086194540553, 0.11748124982440130704, 0.045199115419490648593, -0.21371301892431660008, -0.63678140761821966365, -0.45889140391269234165, 0.88523521512778080567, 0.27667416811346962069, 0.86439428386879402222, -0.49161112692917952049, -0.64941616429372450714, -0.14603841044393270354, 0.28409144949956793846, 0.023156963214552499924, 0.077178211634065282953, 0.10559340683312549991, 0.049563622514557467702, -0.16734632011694042242, -0.56550723836856242599, -0.47024782311390483924, 0.80532829220337986964, 0.28381756324571738581, 0.78636622459650062922, -0.49967348479913525905, -0.57595902915626651097, -0.10690377361903705133, 0.25361276772864282947, 0.011439162289828951752, 0.10568104189264140857, 0.12477002210990562059, 0.0082908897421382771187, -0.23330859794095282211, -0.53438992774254456819, -0.3455505945631399678, 0.81616652949577817022, 0.1035605248290006053, 0.7946588990356190374, -0.3851334284130266683, -0.56451992673106565768, -0.18210751262435681541, 0.27956228129622678891, 0.0056484552078041342857, 0.089000139891549201465, 0.10936922475223748163, 0.014444671173576811851, -0.18219370333101203574, -0.46121053826606023263, -0.35794305431610201618, 0.7339265070679165115, 0.1069330733688934193, 0.71553021133538541321, -0.39159866967315970498, -0.48632185961795226792, -0.13864641769920627024, 0.24264122452145617248, -0.040744975672996279481, -0.11544292824191977609, -0.1559172881122078691, -0.098436174925386366041, 0.12663702812791832741, 0.58069071125950166845, 1.5559461055990222444, 0.4022666410945000548, -4.4992202052736613993, 0.43574397283340926545, 1.5934829079861574375, 0.57273917899414150678, 0.015121470652250936992, -0.37170738078393034121, -0.057987807506581903749, -0.013894510139564338058, 0.029035897250480276893, 0.01493654640568669624, 0.072203610140529769779, 0.23623997825906595249, -0.22666040304815340822, -0.17897586773725751574, 0.24941580213142863776, -0.16528356452268633259, -0.21001571115241379761, 0.23480841501672969351, 0.023614545113627379014, -0.005113152631889671812, 0.044421184155780210623, -0.073759209931414229655, -0.16160501773963992234, -0.15339691419891429702, 0.019027565089861468717, 0.35123715304980512864, 0.52420876997191467872, 0.40330643403896920818, 0.08449091694399296637, -1.9739137359803780036, 0.10593718115690049786, 0.4106163805055828564, 0.44329930954841834101, -0.0079965529620946492612, -0.0018364803549019406616, -0.00100220601940092407, 0.0090954336403642026265, 0.015754921048020783386, -0.035273009577684207028, -0.16733669322179547945, -0.16001711954385414161, 0.79362888786346885706, 0.86444827321305250933, -2.6115179827381926891, 0.86743200854133051259, 0.7632927363660175768, -0.22997737547238458755, -0.10678842186961644278, 0.0011868169694324920069, -0.0071809658749827790306, -0.0022692753530397348841, 0.024991252898907755781, 0.032342646584739787041, 0.022509691903317270018, -0.040592084242338510869, -0.31971439046159105812, 0.15007652592975725958, 0.26800531484702544427, 0.15353019456741301418, -0.3289884060320545367, -0.02826753316383022463, 0.07379339872254880861, -0.21659516771102810262, 0.17850777211890886398, 0.53124279942766039486, 0.71763359584474206709, 0.53210614491737306953, -0.30301535272069191063, -2.0761904129809094499, -4.641388371931440382, -10.053773099631772681, 30.225487740428324202, -10.21156806155019936, -4.6583543337395738604, -1.4779077037589385402, 1.3986703485798950286, 0.014923339040747329864, 0.04990859760690228436, 0.035581079766781105456, -0.050342493136466916204, -0.1799084144767847393, -0.072878673117629466649, 0.29275751486733742057, -0.068865492847689502298, -0.11567467198899950009, 0.14142627829118062377, -0.093293901992132111101, -0.017866755560111923645, 0.36682118347416403648, -0.30679343596241359649, -0.013194293500813799452, 0.025561561965840397248, 0.037137341990200710418, 0.014816969435277153877, -0.032327118609357441414, -0.012965335306752721889, -0.048606514563253100381, -0.32255980464254241369, 0.2611633184555620657, 0.15377389812076866504, 0.25847037325853788881, -0.31517241245603222488, 0.0045313354242074149053, -0.013933675567225304454, 0.00017144606684944328312, 0.023056358603694475373, 0.018739474341150898218, -0.028410324872489178816, -0.10660551123911816973, -0.12054659525050813451, 0.030682954737360924186, 0.61062681810085983614, 0.58264784532291069574, -2.0149346096966045394, 0.5886171818344524409, 0.61545006548088709852, 0.011975249691706057326, -0.21229396183670454223, 0.044552018430161065277, 0.0077533341869321960316, -0.055994136752227655074, -0.11037011816687085364, -0.089532662731520343269, 0.04412795718559191227, 0.17280704744745911672, 0.53424680194562723656, 0.4994313291043074865, -2.0938581753296054977, 0.54320764020374434633, 0.57988626615205052239, 0.13384389943352992258, -0.20712294000382688863, 0.037307656951364120967, -0.045648489404083024412, -0.10822116711691574797, -0.10856739679044423441, -0.0028719073699975429526, 0.24156954564285901266, 0.3707039935674817932, 0.17359216883008959931, 0.11289897868510576084, -1.2776697826103915467, 0.12855794838963202631, 0.18108089671482505145, 0.32935218553879813674, -0.022019909469446107692, -0.017947280056481003385, -0.047134221860043366836, -0.046367846726036499039, 0.0081671678996864949251, 0.1007939217694216183, 0.1608949603293347852, 0.0035171493158746847119, -0.077046646090103865223, 0.38551432835608478999, -0.88484004004637450969, 0.36960361345190417826, -0.11574262642726597383, -0.027731294651903715121, 0.19521853522871809861, -0.0071302816913719566788, -0.045057114301294463876, -0.054942068780821237683, -0.0099131935836734297318, 0.086126848762096308576, 0.17770590603018088238, 0.046028709228123465547, -0.096162255286810660593, 0.33796805022745007863, -0.81626139339146774887, 0.32756024022983842858, -0.12579914963185090393, 0.018173435673258700462, 0.16863644624817519801, 0.080384181801945175927, -0.044399295695913006121, -0.16174051322705143563, -0.22568480231128806857, -0.15712424525026608135, 0.12100261072823667585, 0.64773443093139670967, 1.6436284754159669053, 0.41871842705951139996, -4.5451388067696107953, 0.47459033497244951239, 1.6815638263867982438, 0.49836138606263774564, -0.41717335041806874463, 0.065215877875529112973, -0.033995541086031047517, -0.1272353300318910807, -0.18407470865910810498, -0.14618485139675041484, 0.065417374013393425614, 0.51841777682893253054, 1.4880978878409176769, 0.48107698961000799898, -4.1685627892497061353, 0.52644570926314637749, 1.5155154494706359891, 0.38951250185989072516, -0.3778202624532400189, 0.050866110356961882766, -0.063723444785060034956, -0.15240330402123519749, -0.19075266412183689013, -0.13613191656319501011, 0.11987436524751715616, 0.69572256098053375517, 1.7187187363243758842, 0.36867830799738199321, -4.6876426666611532568, 0.40076318395527948946, 1.7225278662501590787, 0.54117008149463774469, -0.37051188547999774903, 0.012917631515320486535, 0.032558433819719310998, 0.024913299868926645775, -0.013065527352232042535, -0.069002615920707333075, -0.070386199140660743456, -0.083490373215715507627, 0.015986807472331915347, 0.50768263219157994115, -0.75660958941919032039, 0.52707967938545385866, 0.040536853130980446402, -0.058304815191440068078, -0.11493372057183709467, 0.008937626801889064429, -0.034347566144049372638, -0.044749645188001084539, -0.0064370885691591231414, 0.063485985215524329273, 0.16458930593168510925, 0.06938583139695486679, -0.39689920693299540755, 0.10524976889468425079, 0.16718045367521328082, 0.10108644727390526941, -0.41302196101603605438, 0.076815556390572581336, 0.14278181699459063791, -0.020895361976468254661, 0.045373690020868898132, 0.10003194459096759195, 0.1250714044473832387, 0.050169807789315831736, -0.21371301998477409967, -0.64470622572733193856, -0.47549975739069022529, 0.864394283870799085, 0.259773519041953449, 0.85816180826718169161, -0.49457328708982972731, -0.60224929510457547721, 0.13661733290805552765, -0.021580536196934001214, 0.037795483963739433253, 0.087229945072175602006, 0.11232905336562358511, 0.053588006234304527697, -0.16734631963735319138, -0.57202059751836797918, -0.48455580562875860684, 0.78636622456985738605, 0.26725718437257561799, 0.77915000622644803308, -0.50284443441637927119, -0.5299015504841503299, 0.144253777874625011, -0.041605144444422544059, 0.049745670034303975104, 0.11940963453245995174, 0.12750344659295365091, 0.008560949599953140976, -0.23330859824785288303, -0.5376443960066601413, -0.3569787884189502547, 0.79465889923819021945, 0.076019540799486251537, 0.77330488879103020938, -0.38545319767699953717, -0.48091503271781799178, 0.075382858030660512827, -0.037483364112024075698, 0.040007432128317602094, 0.10127733977248677688, 0.11181591254978656691, 0.014622773293768362435, -0.18219370329018358401, -0.46386364064431040433, -0.36756084453203224927, 0.71553021050891207899, 0.082804431976587877795, 0.69562015198663873505, -0.39418013071181134954, -0.41049777514786672805, 0.084673641818027167716, 0.028418776540962448207, -0.055786848769142846627, -0.1313980790987399061, -0.16891722955540119933, -0.10622344617844262094, 0.12663702809814858519, 0.59153326984816134715, 1.5795184309170029469, 0.43574397286705579546, -4.4673417567406215056, 0.45409027923411232619, 1.5765391705959803659, 0.44929435281719803941, -0.29580859333971343261, -0.012591612467873183195, -0.034987983090114943563, -0.017595153338069095816, 0.019687715727996945442, 0.0089943532300010321789, 0.072203610519585334493, 0.2429290644772960639, -0.21390572645407521235, -0.16528356475876579412, 0.25063551448665294386, -0.19017872119980411272, -0.2545296735585874659, 0.2559393555374542828, 0.043052454681321843111, 0.068029115711218446361, -0.012047356144272887424, -0.09956427554180734274, -0.16970228490154451451, -0.1533969139166987683, 0.024049463872502500894, 0.36377654609713377942, 0.54755985605233148394, 0.43307782938534777895, 0.10593718181643824172, -1.995642089745731651, 0.023969276080005451346, 0.29104343326497372768, 0.59680494787423243874, 0.0030589072049066394314, -0.0030255136265154793269, -0.0028441988230969786344, 0.0073621371883008024026, 0.015754920221807858921, -0.033769849315211375451, -0.16678845866736008441, -0.16234019131009755355, 0.79022498644273431179, 0.86743200753285176763, -2.6016375770713309201, 0.89866069722013974808, 0.77172239948382470587, -0.38186019303323692764, -0.0013981898944902797916, -0.00041235529713571684746, -0.0041681835215807540654, -0.00040821900179918143908, 0.024991252220844670934, 0.031225114235946264257, 0.021038369827568820236, -0.040170837858089399053, -0.31550228799136975644, 0.15353019332689005205, 0.24973263790888555747, 0.12823723795895577982, -0.31362679244748664642, 0.064819094797156001642, -0.24906199094650857684, -0.013360153459981819607, 0.27276307508886143971, 0.55928661056135386964, 0.71763359553219430165, 0.52476705935801071234, -0.31872155338469237984, -2.1187358453052933527, -4.7368565348762672329, -10.211568061762852366, 30.051624852796745557, -10.204518272110163934, -4.1638497229924071874, -0.21802985770626090445, -0.01952075725919236171, 0.030390413047556120302, 0.058869792227871940715, 0.040280964913866894317, -0.05034249404574316078, -0.18694102692558672851, -0.087501611471874726766, 0.27710642074200614315, -0.072223773397153079312, -0.093293902354625590778, 0.19061724631195345325, -0.0060380630696022213194, -0.0056261496450084489893, -0.076544209344223138047, -0.020332334058039313457, 0.0069421394570217745401, 0.030948956001384194531, 0.037199095665504257335, 0.014816969319874457797, -0.032146071492919058021, -0.015828142645788162435, -0.057374125367904475759, -0.3350079285361000081, 0.25847037322788302127, 0.17105708027333896282, 0.29809274523380319133, -0.3288028145301381544, -0.032813898050387674266, -0.0049105014348475480024, 0.0080458841413205469534, 0.022137575350254713735, 0.017273739653856198112, -0.028410324943322903118, -0.10663583241406145163, -0.12322308941315485153, 0.023491712901977324091, 0.60141524885532360312, 0.58861718174456656438, -1.9756009170247705864, 0.65973785825811848138, 0.59734436084148756496, -0.27670706514560505784, 0.016980929864406822671, 0.026661775674921062151, 0.0049878718461269008072, -0.053559332134157408567, -0.11037011772312806113, -0.095334830880711540768, 0.034727414374866143187, 0.16865010301441910046, 0.54828758492506168842, 0.5432076405117288731, -2.0420836845384267022, 0.54769246899025658681, 0.44534728504099829749, -0.025798856612630004237, 0.049671953067453047947, -0.0049117725550660235392, -0.064512166655879499277, -0.11372797165030759658, -0.10856739697544531831, 0.00029446557891437373511, 0.25050614665126957892, 0.38853525421571033505, 0.19628919792750379458, 0.12855794865947187122, -1.2939121097935415428, 0.076565073249876214323, 0.095032348748469427924, 0.41597454048672954929, 0.01597262093716710421, -0.027762398426010290176, -0.05584817173858652728, -0.051821386337867383909, 0.0081671686351047212415, 0.10741041154865037244, 0.17256192787096097385, 0.014412640201619114633, -0.076959386241402347184, 0.36960361332880098395, -0.92218635848029117952, 0.32515442523113607942, -0.13574291887687484071, 0.26248899210182419228, 0.021480223912816162612, -0.022558474036479005814, -0.055282183336204972213, -0.060266881141403230715, -0.0099131939175061412062, 0.092069576960347981887, 0.18883525817420343529, 0.058246498826987930242, -0.091815691601636134633, 0.32756024093032493072, -0.85042853803009688107, 0.27985091285237273517, -0.16264997027854841138, 0.29138164338141381293, 0.074692189765718466909, 0.014812757841899988273, -0.069441810767127090664, -0.16641031659027333278, -0.22568480201715837574, -0.15892020973175968734, 0.12131285745616618432, 0.66183707632545274446, 1.6809586748439606652, 0.47459033499371910958, -4.5063129042709721972, 0.46266285549850255654, 1.5472547340947215844, 0.11895803655485501227, 0.060838639178420819698, 0.01206285000280788236, -0.054535180373320925329, -0.1311591660200134013, -0.18407470851867457906, -0.14769970197985646254, 0.065335877815708209848, 0.52926507997206340139, 1.5176012668597009814, 0.52644570925320555155, -4.1332126420679742296, 0.5305099077410877495, 1.4089671411408661861, 0.024773253807654872088, 0.071871355038604339827, -0.0075805441732614310679, -0.090858929751411479647, -0.16098655962341312797, -0.1907526645096274398, -0.13279695966166296661, 0.12697656882304544923, 0.71223200587934654138, 1.7475409828852888072, 0.40076318335956701011, -4.6755267451221627084, 0.4047632550025369591, 1.6644269883075697436, 0.16126255453082175007, -0.014173870087781439975, 0.021459983356829248624, 0.040730694053414402689, 0.030085133416683753488, -0.013065527413100366805, -0.075159180494571903308, -0.08131122932022626093, -0.093794593763622202687, 0.014796042959543540679, 0.52707967952818191026, -0.71916432212794001444, 0.56624407998383130547, -0.016734210433740241003, -0.19025820890078509739, 0.021767941272583045398, -0.012426435460066956318, -0.040642765204279687752, -0.04533195422294725857, -0.006437088572890646912, 0.064495783848403237526, 0.17060745467611207471, 0.082755923233846270803, -0.3836737804484706138, 0.10108644713678723026, 0.12709138888531937295, 0.050975857470876420607, -0.43489881809862962481, 0.30820144565210322307, -0.042367856346142318547, 0.010673080001519712895, 0.06396814600179934196, 0.10772791313287934101, 0.12507140411478318009, 0.045199116020205570843, -0.22264367767502629691, -0.65792318180743869505, -0.49161112747842306003, 0.85816180894832949644, 0.27057999995416004424, 0.87849448477077562991, -0.49964852722470054491, -0.46455779796551349703, -0.038453718614207647863, 0.007548542914912722071, 0.054397124024887402727, 0.093824218503022291382, 0.11232905388775904498, 0.049563621846864053255, -0.17478702788114824918, -0.58365050274246421669, -0.49967348458231702901, 0.77915000653706656397, 0.27478050835155548581, 0.79544488475400199068, -0.51244384156130273755, -0.3747672800833868334, -0.049455117996804194569, 0.0026208030822304044516, 0.066123873677699757101, 0.12244756076633735165, 0.12750344689633216766, 0.0082908897855713398206, -0.23815408974511823037, -0.55375379093808652087, -0.38513342880400280865, 0.77330488943059938478, 0.08917373954388294055, 0.83960355244032003341, -0.34849253576314032088, -0.47187371660850263577, -0.042524020676827034571, 0.0004104192444442170129, 0.054235468742288635291, 0.10397611646968590859, 0.11181591318162371895, 0.014444670329481226756, -0.18606228613511949233, -0.4771077966827564909, -0.39159866883415600514, 0.69562015198600912758, 0.090622639713764580049, 0.75111789405257634034, -0.36569289118477710243, -0.37458095507035649163, 0.059498184884077279233, -0.013652015585391025568, -0.083346779859661546119, -0.14358043558359032121, -0.16891722962710603584, -0.098436174721175884184, 0.13845822046610306577, 0.60500493120250209422, 1.593482907977284313, 0.45409027912695254514, -4.4486677353692218873, 0.46731798600595314541, 1.5101127178628461767, 0.15693704584316808814, 0.018584743843736119956, -0.025821150034657031763, -0.044446903504188478573, -0.022725955628061990377, 0.019687716317985548553, 0.014936545971379930342, 0.083882457888257030976, 0.25680974621932056712, -0.21001571071279681591, -0.19017872153923867651, 0.19292449780226372802, -0.19557754249913872346, -0.14477725415729622571, 0.25103398067533433391, 0.053863755924773992767, 0.021254039055695232024, -0.03051399166585797712, -0.099564273071916412072, -0.16160501759477152728, -0.14488752930650389272, 0.030380312259002693887, 0.36954520871560375683, 0.55330793547935352983, 0.41061637962515667999, 0.023969275939358526684, -2.1441781195370013258, 0.092086978087119839009, 1.0389353070161106984, 0.004336571818405833556, -0.0008108578250922583025, -0.0048045461571112283578, -0.0028441991323542329881, 0.0090954344106679036602, 0.018714660716090455311, -0.031428674119533685227, -0.17209257502854119481, -0.18385642449320285796, 0.76329273707954914396, 0.89866069704889750547, -2.5040381469808639103, 1.1455204526732083448, 0.065995205199934930951, -0.0032972128068261092812, 0.0015949953928878417113, 0.0012000207500494916926, -0.0041681840095454149309, -0.0022692744502893920887, 0.022252370698027650053, 0.0317616797883579563, 0.028788775783694667276, -0.029214607647833630566, -0.32898840492197323204, 0.128237237696630757, 0.27039614719244747798, 0.082324482657934391905, -0.20156510760659068771, -0.17460021511459855592, -0.10057411733230942064, 0.046121318568118886694, 0.27276307501040103487, 0.53124279917397998485, 0.6890139095943335068, 0.51782938413649859388, -0.29012656248608825749, -2.0526871418904373279, -4.6583543330311494302, -10.204518271955505426, 29.803901155299726611, -10.32236376544024381, -4.1570422457323523702, -0.02720239110689709125, 0.0061979885596670980172, 0.039606869466794349111, 0.058869791312886526891, 0.035581080820535661291, -0.05891309785837936519, -0.19941880231299938253, -0.097297738981700132577, 0.29368790721292054036, -0.017866754611421040849, -0.0060380634948046588151, 0.20199184851748774872, -0.41931420583389988899, 0.19497327218190599329, -0.01357797063026307087, -0.0068408428611964255603, 0.010052547451073527707, 0.030948955719808466519, 0.037137342298325201617, 0.013110548255309911747, -0.038522077143225454121, -0.024406114785095193304, -0.053360167678456665863, -0.31517241212092850233, 0.29809274505412464151, 0.17973530937425963305, 0.14725705014586876196, -0.26743975214979692767, -0.0033615188716693456157, -0.0011734396305143001381, 0.0083724759957905792174, 0.022137575215469308676, 0.01873947451442510917, -0.026808510455351354518, -0.1104617334109166682, -0.13835289965907812682, 0.0080702236979889475227, 0.61545006569965632348, 0.65973785817484853489, -1.929963847142112332, 0.67686826383808174068, 0.20771299743354579292, -0.0025469236325987077742, 0.020350692882097353581, 0.027877243532824690087, 0.004987872092866146162, -0.055994137807122437711, -0.11660451142834762917, -0.10225283556154607378, 0.036834337539710160225, 0.19201471315534546225, 0.57988626538264487209, 0.54769246933928639187, -2.1331707842852543067, 0.55711711401557018064, 0.45525705876038463638, 0.035949576606572872239, 0.016472973475887317529, -0.017514437447600519904, -0.064512167136312650872, -0.10822116706932151386, -0.10188034143020319178, 0.0068176082758691462782, 0.25695586719590957081, 0.39495656027492037632, 0.18108089651889902605, 0.076565073265856667395, -1.3962823300775695223, 0.063764350062201283209, 0.66504104200956437598, 0.027616544891616332663, -0.0064385808687336225364, -0.037369185987884030731, -0.055848171644265989788, -0.046367847677495271919, 0.014891757077512168816, 0.11187739576194864755, 0.17052365952509296321, 0.0011428194242132649894, -0.11574262639417992604, 0.32515442531654437097, -0.93108754507764657227, 0.47306159192252217327, 0.066215860812251114975, 0.027887645370537176431, -0.0021360585039934196216, -0.032426932693238388894, -0.055282183624169700664, -0.054942068627211390619, -0.0033141145328964144445, 0.097219215892272337065, 0.18962054093168606572, 0.050318692277569271321, -0.1257991504529039628, 0.2798509131939885819, -0.88117307130678679261, 0.39253146023251872521, 0.1168237611744660509, 0.042657734370792690926, 0.036892922455488766365, 0.0011764842292660664114, -0.069441810448978541492, -0.16174051330461397535, -0.22168178745292529963, -0.15648328668505437355, 0.12237234552783432528, 0.66256331083705588547, 1.6815638260577352447, 0.46266285556252351174, -4.6103613292539282753, 0.61735853006134566723, 1.6205429511187614544, 0.034818374093621901055, 0.029887189618759298776, 0.0010346484423202446587, -0.054535180314002708324, -0.1272353299962170059, -0.18034332104952588427, -0.14539795004207445306, 0.064371121088404578692, 0.52478068100007690955, 1.515515449233870271, 0.53050990777055850867, -4.1997526353638265917, 0.67783342575674865316, 1.352752872007822571, 0.055542019732520861441, 0.02493188242727326781, -0.025816854238813403272, -0.090858930332767096072, -0.15240330314429462177, -0.18148884812545984513, -0.12886686497676058671, 0.11830374710071492328, 0.68778279911029105431, 1.7225278670837884665, 0.40476325478446856243, -4.7109493669024598006, 0.58916745476069898313, 1.7132388429867786428, -0.022393764982554128368, 0.0055297667570599409737, 0.029714440993559434329, 0.040730693891962925501, 0.024913299589516352261, -0.020581530846781107696, -0.081250743322727317364, -0.080773582490882442353, -0.078660708069999571435, 0.040536852899367106784, 0.56624408030395645408, -0.72442979770100712766, 0.48470683466820901897, -0.1826164372403450975, 0.017009610056614620638, 0.0049043504367635342225, -0.016806002275953958042, -0.040642765376490724849, -0.044749644863066762035, -0.0039616629023083455713, 0.072281539968845162636, 0.18256298235912071792, 0.088349258715097064854, -0.41302196082214220896, 0.050975857324471116061, 0.09211138356410381034, 0.019714108697729895714, -0.010357191701053402796, -0.038911264011517475436, -0.01020893860286471895, 0.024780680892804548415, 0.063968145428026743082, 0.1000319446354601266, 0.11748124958214689606, 0.044732221609195044199, -0.21535492136627379378, -0.6494161639308941858, -0.49457328731044336578, 0.87849448459470014239, 0.35699287032017978927, 0.96792985480793081177, -1.1566562597121134992, -0.033679552110923723085, -0.010175142017944255324, 0.019740070978345503755, 0.054397124009524761012, 0.087229944640756063023, 0.1055934068730911557, 0.048636880928813741587, -0.16890823657681663938, -0.5759590299552895809, -0.50284443431831027382, 0.79544488477541730465, 0.34547880448545897902, 0.86958411267582302884, -1.044828870045316771, -0.033338001336431152821, -0.020060638323042844638, 0.012912731365795297039, 0.066123873928542645229, 0.1194096342646348563, 0.12477002179158008799, 0.0049205459073694626793, -0.2462307114413970488, -0.56451992640357739894, -0.3854531984007662504, 0.8396035527603236126, 0.21361196601181620114, 0.86088414110120659029, -1.0027524587249028176, -0.02792186339925113378, -0.017745444056777342351, 0.0091968133087716834317, 0.054235469081402039315, 0.10127733889906259668, 0.10936922587761588321, 0.011699287437323506456, -0.19258692296591806148, -0.48632185988400583598, -0.39418013088357262141, 0.7511178939192336701, 0.19472829160084245692, 0.76247837496727355333, -0.88470544869375178365, 0.056007971759416387558, 0.013472863077342083979, -0.034227003843122098814, -0.083346780155159130854, -0.13139807870959926972, -0.15591728787467423745, -0.097052844520629677483, 0.11907357226772401215, 0.57273917896740145217, 1.5765391708261060622, 0.46731798569611177063, -4.4871791418004463026, 0.68129426872622778077, 1.5248153691740513604, 0.025497881214978323355, -0.0054918712783452354811, -0.034302605159525045853, -0.044446902803571379514, -0.017595153976289617387, 0.029035897654145399055, 0.024730485321769846846, 0.082579883851981644716, 0.23480841517484069003, -0.25452967390016711935, -0.19557754208959707287, 0.27142036262603386199, -0.3311045781364405638, 0.21413570029350473267, 0.021062218548244207583, 0.029357214442478886396, 0.021254044783547025588, -0.012047357012854715413, -0.073759210683196657632, -0.14533895258114640225, -0.15755289588088369768, -0.021180519295634879506, 0.27944194482852108274, 0.44329930980121790007, 0.29104343331938392625, 0.092086977854602472537, -1.6669203123991984672, 0.87438709635170563939, 0.0025126917673079997469, 0.0011228889201840464192, -0.00081085755082630807389, -0.0030255135076355083555, -0.0010022060119633073583, 0.01106005397223321432, 0.02134985096471189378, -0.032563692707919679459, -0.1943565020749260519, -0.22997737531090900465, 0.77172239928493047145, 1.1455204531266507395, -2.8190742434163573726, 1.3311548115737230624, -0.0019382990147631723368, -0.00020569594993286426381, 0.0015949958281313242423, -0.0004123554678286651216, -0.0071809657909675918622, -0.0046248308531623393003, 0.025702571777758760618, 0.040797126157519007783, 0.024236232296906780326, -0.028267533060626921088, -0.31362679285512218286, 0.0823244832184023001, 0.4628469780176859838, -0.28090700475462115904, -0.055525129314812980696, -0.1040275265672703231, -0.10057411766773290129, -0.013360154407002600846, 0.17850777252530844463, 0.44812440067798370302, 0.68608586791178649555, 0.68013681518014179606, 0.10090969585282767718, -1.4779077037498395963, -4.1638497228117845594, -10.322363765017597004, 27.382713710508227223, -13.184100384773451253, -0.016317221767410058919, -0.0098541279990075267281, 0.0061979880807195089631, 0.030390413469979199457, 0.04990859726520782691, 0.03619494907868735234, -0.043640064005973262795, -0.17439445098869041373, -0.07141978040620376722, 0.36682118282324727154, -0.0056261495509461545775, -0.41931420564082255709, 0.2091044753517140864, 0.051659763693682032859, -0.0031459456106381900042, -0.0093558586848769449695, -0.006840843331559079768, 0.0069421395491973871689, 0.025561561743314979017, 0.030945627719653430454, 0.010710780369091794928, -0.027338204554862105544, 0.0060319866065350743675, 0.0045313351791911203256, -0.32880281450895421091, 0.14725705013059819937, 0.36541314234039951048, -0.216868751320010128, -0.0012316199178423806442, -0.0029111091027995915879, -0.0011734399906160257031, 0.0080458838925404393277, 0.023056358720212027924, 0.021249118197143087444, -0.025860234750942109883, -0.11077222998097548123, -0.13668136351725251876, 0.011975249645582688313, 0.59734436070070939806, 0.67686826378666620219, -2.1321247117414556804, 1.0769099460776416777, -0.0093903205131380968612, 0.0054317019554368725642, 0.020350692361238636313, 0.026661775517697730575, 0.0077533346423463937042, -0.047262593987040107946, -0.1037995200457896422, -0.088637003173285613955, 0.032884284413962024118, 0.13384389994073611629, 0.44534728503632187158, 0.55711711392206686444, -2.0517563782484082857, 1.0724631416569450071, 0.010589473840851055739, 0.019700726700719851936, 0.016472972653309634478, -0.0049117719653919934483, -0.045648489605476856579, -0.093867482519786896167, -0.10649935984718213178, -0.02620123042673598579, 0.19499280387417028404, 0.32935218524827286624, 0.09503234895559274642, 0.063764349863731487056, -0.99881852771232637611, 0.53105896358880333441, 0.020778497984691127187, 0.010134087292552966184, -0.0064385815972406411356, -0.027762397869998758221, -0.047134221754530164494, -0.049345818927976750623, -0.0060886739009748253038, 0.078770265265831446166, 0.13353046985708486027, -0.027731294946112622357, -0.13574291854215869702, 0.47306159166952704842, -0.59480943609388703397, 0.16305934469014213284, 0.017757350071218678395, 0.011186997283478699455, -0.0021360592201553167679, -0.02255847356000926715, -0.045057114586741588147, -0.05474034331089857458, -0.020372477461687854561, 0.066199881013374678451, 0.15164315258604246783, 0.018173435933675707865, -0.16264997043702120183, 0.39253146009749279033, -0.50946687077470431682, 0.14502533651708307549, 0.0064297335786438775812, 0.029880423083078484769, 0.036892922595006198694, 0.014812757535856464416, -0.044399295552583949165, -0.13402059930945345312, -0.21160369865746200824, -0.19120315768177353899, 0.020654069078443294122, 0.49836138645402733305, 1.5472547336075943569, 0.61735853019008002374, -4.7652241675898396878, 2.5594840133564944473, 0.0051735132715512730309, 0.024361050962046115059, 0.029887189696986820459, 0.012062849973068999129, -0.033995541140076725173, -0.10376247465320254104, -0.17110727848564427633, -0.17472116834306936073, -0.022113128752078835965, 0.38951250223543970597, 1.4089671408943869046, 0.67783342576006710978, -4.3877424342523472234, 2.3341435030636774606, 0.021588725027425304664, 0.031882552772083690151, 0.024931882447507665301, -0.007580543804870876895, -0.063723444873298895619, -0.1299483762124240549, -0.1860357195143887421, -0.18871543460568451778, -0.008203274513226818665, 0.54117008105195152368, 1.6644269878939086382, 0.58916745495331723781, -4.9280639495994869037, 2.615688626737385114, -0.014913507592698119208, -0.0072870615709983467545, 0.0055297662004828865637, 0.021459983360709013189, 0.032558433899096753117, 0.025152583612861923595, -0.0066957883330741804762, -0.053576004291873605134, -0.050776275763774997296, -0.058304815108541054036, -0.016734210496649574995, 0.48470683458309787861, -0.66596119934369901738, 0.31490399633885002295, 0.0056533791535097925943, 0.010208168283554133401, 0.0049043504123218711194, -0.012426435566420049317, -0.034347566110101618309, -0.039995600575532364029, -0.0064392416360667175876, 0.055829196987590923307, 0.15021192674206779172, 0.076815556592234615629, -0.43489881836887211985, 0.019714108706068662891, 0.5147872043256330743, -0.31811729252142073587, -0.019074301591392908706, -0.018948964436003148948, -0.01020893964243865977, 0.010673080651713167527, 0.045373689823779470021, 0.090019947772341249448, 0.13090795817000555612, 0.093963708944319712657, -0.14603841037885753606, -0.60224929566016671689, -0.49964852644487139921, 0.96792985455737079459, 0.36362846934550430156, -0.3844658407134892153, -0.015337905747701724088, -0.016729882345553025169, -0.010175142727349791374, 0.0075485432354946320477, 0.037795483863618271891, 0.077178212279409103003, 0.11529838101869298306, 0.090328539083917344654, -0.10690377277413594392, -0.52990155121725157628, -0.51244384094382455785, 0.86958411242626287851, 0.4091780511727949321, -0.39737234671295734723, -0.0097427073015953370022, -0.021487036381897028914, -0.02006063847529911473, 0.0026208028948039793771, 0.049745669980618981154, 0.10568104258814997598, 0.12689287184491360172, 0.037588835652392509068, -0.18210751282983378418, -0.48091503257619661049, -0.34849253551308789767, 0.86088414099094023868, 0.14403212292072403722, -0.24570956025106627241, -0.0074610443406979143344, -0.018048117536921984977, -0.017745444229758648713, 0.00041041910550318472239, 0.040007432628885564174, 0.089000139352614582311, 0.11020925813598457521, 0.038034878617792623745, -0.13864641786038031657, -0.41049777512301854898, -0.36569289065335874866, 0.76247837482063141046, 0.19097662662120162835, -0.25774475054458007017, 0.027906588339856572972, 0.0264201341605564341, 0.013472862875009889577, -0.013652015263431528311, -0.055786848931146056085, -0.11544292883054693699, -0.17388837643890525664, -0.16129180520765046269, 0.015121471034214425608, 0.44929435267687878319, 1.5101127177955504521, 0.68129426846391372141, -4.6559682229960364097, 2.4284155186968048312, 0.014819876165911911869, 0.0092596672707904482957, -0.0054918708297061122336, -0.02582115102707984583, -0.034987982362804107006, -0.013894510499827679786, 0.018793298658325713335, -0.011146132564555349953, 0.02361454459019513008, 0.25593935652295624061, -0.14477725502184526918, -0.33110457781149355005, 0.44777332386796836472, -0.21376761982718006649, -0.017954574110933419623, 0.021062221167727403909, 0.053863752762304366317, 0.068029114910183610676, 0.04442118461155123571, -0.03786241632692551734, -0.18317467384631583305, -0.33453925308611781597, -0.34083262475519338075, -0.0079965529756836472436, 0.59680494807312156613, 1.0389353068883362408, 0.87438709639257838901, -1.8658146617174922177, -0.0012623152265749616654, 0.0025126919438606628611, 0.0043365716287947679922, 0.0030589071280662689933, -0.0018364804848602354383, -0.0066319584137774752544, 0.00018500429554941592012, 0.027057589631568606386, 0.030922975906407353802, -0.10678842205312351343, -0.38186019293603257196, 0.065995204865266304806, 1.3311548118025544607, -0.98530954753454125239, 0.0015466050022570796644, -0.0019382988739748349682, -0.0032972131074878942465, -0.0013981896678449462143, 0.00118681668602713228, -0.0022316111615224704101, -0.0098986286536085349441, 0.0023162006575922961839, 0.038874504784860243467, 0.073793398460951550866, 0.06481909509716972384, -0.20156510802222890599, -0.28090700454567785194, 0.32815969940029110496, 0.07165913934475809377, -0.055525129683812242398, -0.17460021492678470811, -0.24906199025195785435, -0.21659516780019363869, -0.00093366097594976165449, 0.44148435297427374158, 1.0429455701635599318, 1.5391360222793784018, 1.3986703483059934605, -0.21802985792300336176, -4.1570422459396025872, -13.184100384541954654, 14.222791113509460459, 0.0039167560443872035184, -0.016317221965140380618, -0.02720239084427693993, -0.019520757232588857588, 0.014923339032379101879, 0.071983223056763884218, 0.11850697370485752113, 0.083018378287239333324, -0.11169518113290551053, -0.30679343582465007367, -0.076544209254870113068, 0.19497327204150666224, 0.051659763720567221379, 0.033064999623344026869, 0.0059150341284471569941, -0.0031459458594427841191, -0.013577970333330391037, -0.02033233396902359158, -0.01319429347229254465, 0.015306800844851071661, 0.054535063031304009673, 0.072241514071812840125, 0.033196576244332455519, -0.01393367554555050708, -0.032813897966173309617, -0.26743975216206478107, -0.21686875134335054094, 0.41959444476161633641, 3.022868647802455741e-05, -0.0012316201766072196275, -0.0033615186498056818902, -0.0049105011271966102909, 0.00017144589418522637454, 0.020166785410509140142, 0.047168382072879641576, 0.040471605557841169865, -0.05320734088389827976, -0.21229396190190952254, -0.27670706500206149192, 0.2077129975104041737, 1.0769099459735431701, -0.85563921742958870542, -0.0068973285708662755711, -0.0093903207516764612695, -0.0025469234173412722, 0.016980929915885897147, 0.044552018530474198077, 0.057731651883490803234, 0.017021474642741497157, -0.094581390748202875951, -0.21485267441406830091, -0.20712294006784665035, -0.025798856798154058689, 0.45525705889674555893, 1.0724631416350474122, -1.1328517719154707066, -0.017386328538302499302, 0.010589473562273632457, 0.035949577008186739402, 0.049671952854323742077, 0.037307657174036057413, -0.016324455427418622933, -0.11462704539535102655, -0.22147883314929098231, -0.24137877731062251652, -0.02201990920731189405, 0.41597454030610675479, 0.66504104219785498131, 0.53105896350171144515, -1.1689437938361575409, 0.0025929779818593388409, 0.020778497808967628935, 0.027616545170412565291, 0.015972620638549941963, -0.017947279889884669435, -0.068706074589092031024, -0.11494796994005265778, -0.10945716507789736427, -0.0013559687816999677068, 0.19521853547335588086, 0.26248899176954776369, 0.0662158610830609895, 0.16305934462970986254, -0.47651165582236992835, -0.0017332256986496378387, 0.017757349864643695581, 0.027887645633824599545, 0.021480223774281118065, -0.0071302814941260738574, -0.056648014309180025339, -0.11067375473484478143, -0.12410765456889102931, -0.036629031245783053028, 0.16863644630804699998, 0.29138164332751426144, 0.11682376135193234257, 0.1450253363980370247, -0.48628934638976994975, -0.024915956579823348577, 0.0064297335738475042133, 0.042657734361640407827, 0.074692189866254435349, 0.08038418166596707215, 0.027875737368517981146, -0.10802494062936336683, -0.31076603659568191595, -0.47871919852920324923, -0.41717335056809878857, 0.1189580368956816292, 1.6205429510075410882, 2.5594840133258300874, -3.3063607093501183876, -0.020469593172887098204, 0.0051735132769018020785, 0.034818374080801919868, 0.06083863919534433784, 0.065215877844943703767, 0.023589915056597111909, -0.082388545187292397398, -0.24321491917699808449, -0.39135054680178837927, -0.37782026259950651781, 0.024773253957452598395, 1.3527528719995924877, 2.3341435030341282086, -2.8829330996147524147, -0.017368443361267311975, 0.021588725111148291541, 0.055542019659872890469, 0.071871355003550796425, 0.05086611011806821403, -0.024568473093190389828, -0.15338469947894445289, -0.30816652659889287591, -0.42874913218057908448, -0.37051188538081891721, 0.16126255486194626676, 1.7132388429065454893, 2.6156886266495869009, -3.5186879589199904217, 0.0018210784442722826541, -0.014913507810480954377, -0.022393764874315106456, -0.014173869915263948363, 0.012917631577834537762, 0.052059578078320178918, 0.080329152269709061529, 0.065262086202811209201, -0.015347096675372288652, -0.11493372057234031713, -0.19025820906810331223, -0.18261643701854665678, 0.31490399624940396128, 0.044180054125544404586, -0.0060117381244141950697, 0.0056533790843439806276, 0.017009610086475477481, 0.021767941466449645255, 0.008937626597764584957, -0.028459790196520732414, -0.076970350927218655901, -0.093456399081487659397, -0.033056694800951823277, 0.14278181677887225032, 0.30820144593256632026, -0.0103571917072679473, -0.31811729260260718233, 0.050335736821503107574, 0.0089660947022704492321, -0.019074301929094185371, -0.038911263398283227333, -0.042367856538791645593, -0.020895361822412566316, 0.031704611478810346137, 0.11528806445246224388, 0.21961042843917893963, 0.28409144940184061223, 0.13661733325442074194, -0.46455779839906485851, -1.1566562596094727144, -0.38446584067425193521, 1.4045124122417551593, 0.0093407907343848704806, -0.015337906025483100114, -0.033679551607598647023, -0.03845371881148793719, -0.021580536039208030041, 0.023156962835434419246, 0.096157548659263000768, 0.18949602490120914799, 0.25361276742564942133, 0.1442537782555338155, -0.37476728048423285644, -1.0448288698780443529, -0.39737234671020721377, 1.2744046494679623915, 0.013324014236757131605, -0.0097427074562676015818, -0.03333800103805516224, -0.049455118004096271234, -0.041605144355092327224, 0.01143916194755092812, 0.11791036910276896099, 0.24395630746294005853, 0.27956228135165384074, 0.075382858098041682915, -0.4718737168198479659, -1.0027524587455813876, -0.24570956017113168568, 1.1767655193064110453, 0.012267418940745938166, -0.0074610444417616255536, -0.027921863125777636461, -0.042524020724582410613, -0.037483364242622670137, 0.0056484552714927234104, 0.095566651159319282449, 0.20553786904926396595, 0.24264122460159764927, 0.08467364185484869965, -0.37458095533037116809, -0.88470544869491696272, -0.25774475045885364421, 1.0403707325771378578, -0.013032834634299141074, 0.027906588265631540602, 0.056007971883877918917, 0.05949818483703261246, 0.028418776450530303795, -0.040744975413086970129, -0.14695275583361230853, -0.27726521648327528968, -0.37170738106584272353, -0.29580859330669584395, 0.15693704604330271923, 1.5248153693313604151, 2.4284155185027365143, -3.2608013254966423666, -0.0044550666451388118774, 0.014819876196964579251, 0.025497880986308243956, 0.018584744304099041351, -0.012591612831491473187, -0.057987807384990361359, -0.081262449904295167014, -0.048919456611021129577, -0.0051131523997239997081, 0.043052454198381745476, 0.25103398109692071083, 0.21413570013509222734, -0.213767619817241461, -0.1642390087016198652, -0.3876780694289247986, 0.74432871219515883698, 0.032260793227889668355, -0.22348521730383891226, -0.20144930377073563821, -0.10637003279163254055, 0.0098881754560438440971, 0.067357711916249549122, 0.063326725434643668655, 0.034748222625719769885, 0.010009565472813092279, -0.0044478355006400006194, -0.011456632991571508509, -0.01424349750086366731, 0.46783129636462056355, -0.49195740459417564017, -0.26607832588904145599, 0.15941423446448096946, 0.1665339708331919355, 0.054013649365180181616, -0.035814951566567443386, -0.045521868754599334461, -0.013641731505021606821, 0.0059965269293854785854, 0.0075100617142252763109, 0.0030223877852782977602, -0.00035083958215698627953, -0.0016258417330581476524, -0.62063735764139771334, 1.058733493109889956, -0.18116792194142361527, -0.34307307216882693135, -0.00062485287914053333491, 0.056698828793864147757, 0.02271839996667835565, 0.0042398703253175035882, -3.2574077052157755384e-06, -0.0010259776504604671381, -0.001369755011644488386, -0.00085067150889145024765, -6.9219753456492471503e-05, 0.00011120506717439026698, 0.033064999719403902201, 0.051659763520518944824, 0.194973272283949145, -0.076544209387498243768, -0.30679343604021852565, -0.1116951810868475059, 0.083018378811506921, 0.1185069733760401034, 0.071983222880777700259, 0.014923339117431308476, -0.019520757149863794677, -0.027202390894181895098, -0.016317221785617629787, 0.0039167558860347525748, 7.3361120421384953261, -9.966772811105366614, 0.14927643171883323014, 1.5199230573155646518, 1.0960350273568588708, 0.4097583927180746266, -0.031683463386221119573, -0.18669630136059275904, -0.16103448017988289398, -0.073365196269601681833, -0.0029335907499376959531, 0.025634173639562940483, 0.020979387308508106597, -0.0019182390964253598527, -1.1876655058300467083, 1.4493484758382986222, 0.27868790420223960025, -0.32702736699468593073, -0.22329119841504829758, -0.1016707565187678769, -0.0048565607991041693375, 0.046282454514658663558, 0.043182905355475947884, 0.018120277800480054375, -0.00030126322029331182315, -0.0067459419842282684204, -0.0055627701043034103942, -0.00080923449950858698364, 0.73459587553746374944, -0.65568196047421511974, -0.52195596691710566262, 0.19324503790420091698, 0.25782736591284155514, 0.083950984945085835309, -0.0087552572983547165969, -0.030632183163663152803, -0.025780203384739880379, -0.014732635195036026393, -0.0038923998830838802722, 0.0027363865204074297628, 0.0041988494808839526076, 0.0019284775472832975125, 0.43847184523097293418, -0.093255878547400167244, -0.54720362866590699369, -0.14755073801247350906, 0.16485915996750893919, 0.19663612003543382012, 0.10319098866885913213, 0.0043413538754884656301, -0.047244143742535778252, -0.050831048717759910738, -0.027378040361510834477, -0.0021851659157725118396, 0.011194984797358762188, 0.012066888507822108589, -1.5198218630121000938, 1.3124694170640121005, 0.74774406512294422367, 0.0026860411282694494385, -0.3333268188498964002, -0.3066417572291999516, -0.10287556075206988337, 0.042850776563749547188, 0.075823462957966714093, 0.046494484622207878088, 0.010065181136814226048, -0.0093741710964799496164, -0.0105052411063567986, -2.319279872357195288e-05, -0.96831225982001789543, 0.65208530124000796402, 0.35609093513681572274, 0.18542834692504633898, 0.011326057427575711434, -0.12817832924193020561, -0.12592647772367165926, -0.063673713896663763911, -0.012674256514347946243, 0.010821622394413383389, 0.016579046712279374431, 0.013785444377579449474, 0.0074694807331396112157, 4.4996923765866793355e-05, -1.0654578598205604045, 0.74449369962451650284, 0.43514684644789236589, 0.17406641209201034748, -0.039069267159384676869, -0.16360476712351229578, -0.1329412558904378705, -0.053283444306017557768, 0.00079481008748590173049, 0.019554725510253941761, 0.019395761255901252523, 0.012709028745480691752, 0.0048929333269168507128, -0.0023678498134487189694, -0.089962021403143466181, -0.075435469450954820148, 0.24631821757762809022, 0.11493684094700314102, -0.1665249047131528437, -0.10487744269002677311, 0.013844438990369824921, 0.045095420000344728362, 0.027125598790860346726, 0.0021111312080855627014, -0.010006753008930989804, -0.0091043463069387080971, -0.0022986774404195882242, 0.005243024847314012131, 0.012287513469365742361, -0.12757299824881945804, 0.15351200186912020151, 0.089461036727597240126, -0.12521756109954634839, -0.061759940762349627152, 0.024896903683092410503, 0.033870673843631807221, 0.014320193720591494618, -0.0026359999163793963912, -0.0083874538645884819804, -0.0060320106286449824001, -0.00089391132902555652927, 0.003963566697089230921, 0.60905283902575846344, -0.66202308182258085001, -0.34201089184686589428, 0.11602591913192032702, 0.22407433146188973638, 0.21070093640334136564, 0.037459094850177621083, -0.10514425732591091422, -0.10401201154325018239, -0.033710661697944059589, 0.019133053694066853162, 0.030835860813144797682, 0.01593933551252384867, -0.0089414527786114486596, -0.21881262216396532572, 0.34282141008603500554, -0.13245192328518798353, -0.087672111167753141725, 0.03001327312507916234, 0.063277367627666974981, 0.046649996875297083943, 0.0046824456642655717248, -0.024286396040819276804, -0.029474564524491056461, -0.019269385323555727335, -0.005252047457240869438, 0.006266965733787207421, 0.013722208638168992972, -0.34266966218520977794, 0.62729268140535843212, -0.20100474590419975529, -0.17852748230844564525, 0.060955432747808908911, 0.049112507616338237082, -0.0021882462458048114289, -0.013545960572300355662, -0.0073048210136134354362, -0.0013581102806791188822, 0.0017782550156678410096, 0.0023665131850208235252, 0.0013326746133532924726, -0.00044417408391368235149, 0.90828074882802956047, -0.25821309812027182229, -1.0203464359669527362, -0.22937506193529658138, 0.37381501779138326391, 0.31182914957299584513, 0.079736809666066818525, -0.04513325511268854312, -0.057324303810745290411, -0.028472234330157636384, -0.0040540034762495099144, 0.0047265063803276184673, 0.0034879275273581584853, -0.0013536373430601521001, 0.71477822041999028446, -0.13422367190329045905, -0.87560778705499187069, -0.21061706383549630428, 0.31455317226181572776, 0.25881713099832065472, 0.064948093025163716741, -0.03453380688988767766, -0.043560142034926023491, -0.022425367058362469241, -0.004924138221543241864, 0.0017700395053464241558, 0.0020092210635980524425, 0.00011854475736141882754, 0.083247365480542317995, 0.24173670883110259799, -0.23062190656158007562, -0.13954972534474091583, -0.034427215562697240092, 0.0046452380614929959096, 0.034050173162224015022, 0.039741556934968447212, 0.028343677072934257005, 0.011355775259266152194, -0.0019688057732237290894, -0.0081168374838686586925, -0.0078261346389579319188, -0.003695565050033395877, 0.021785834732986380374, 0.28784127787724533842, -0.19999909532141077473, -0.14890422378639339707, -0.044353930407756571308, 0.0072853618209372909831, 0.039237175895056410069, 0.04008266356553122356, 0.024669572058364647171, 0.0073111054392098430085, -0.004209242643512243684, -0.0083311101494930334815, -0.0066547794334561229807, -0.001810471283325922531, -0.4568631115036111412, -0.061834285838260276413, 0.46013592875313308328, 0.39178145989914037184, -0.050182906888380646171, -0.19944094433312251069, -0.14066429441040234627, -0.046526029094123885765, 0.013853104610649431827, 0.030555171478151631898, 0.024135405125499057727, 0.011857913598201152983, 0.00097751638535809016149, -0.0069853187464097464196, -4.3747797787160012106, 4.6459224056864849572, 1.8839922568247351453, -1.0603454203552087343, -1.1329583074056872594, -0.4250257608360131778, 0.044178759333394782005, 0.17183749001161521819, 0.12982165309599566494, 0.054627543761347784057, 0.00063552226459714239188, -0.021125919075099201111, -0.016709753998774008743, 0.0028885343166027725378, 0.74432871252344967594, -1.4238841899113938716, 0.61708914197536490764, 0.24233728075974889093, -0.034649479282161381455, -0.059775185938446079303, -0.090141772736487074669, -0.049020497970288755274, 0.0081707286354232144115, 0.03475484263101604604, 0.028712085045155075896, 0.012335717726653979559, -0.0018591341972936886815, -0.011456633179271511619, -0.49195740460852865894, 0.82481288381431117074, 0.052131819331890996705, -0.47132723872031156986, -0.0740613662761164665, 0.070590412261250831683, 0.097539917853686319837, 0.03498036910734150795, -0.024050997055567117922, -0.023009893350725955236, -0.0047337064132322442811, 0.0039637818848521095411, 0.0034715404072493045474, -0.00035083954843511967183, 1.058733493025323158, -2.3945592472078303103, 1.2891730436970885165, 0.48794424734108260422, -0.37656455857383358543, -0.11332918024541399626, 0.02474990111350851632, 0.022440242431343137425, 0.0062679352149370608574, 0.0011925282749264199452, 0.00048625257732102410884, -0.00036239551836338384887, -0.0010312550682819872881, -6.9219650569560912403e-05, 0.051659763305170643133, 0.20910447616811600469, -0.41931420609498348817, -0.0056261499072951370176, 0.36682118376767741674, -0.071419780664340015708, -0.1743944521707705142, -0.043640063112155909331, 0.0361949494583912873, 0.049908596968780492842, 0.030390413178584922693, 0.0061979884609951870975, -0.009854128469879984073, -0.016317221477228169985, -9.9667728111257947177, 21.838228998508906642, -11.497485299261187564, -1.9244733901199975179, 0.29186058113623825072, 0.70677275199610045675, 0.48755226877786483897, 0.19436376195040935833, 0.0012166129245994553205, -0.077425677730745504079, -0.072460542839830305306, -0.034422792637089597023, 5.092345068548449706e-05, 0.020979387390287745213, 1.4493484758082164632, -2.9022793336527219488, 0.97849744767307256232, 0.7956132889701517108, -0.088269519859832445974, -0.11839823504915884811, -0.087924516065559649158, -0.057371305771133883356, -0.0064942637166558311521, 0.022497177256489427066, 0.019740718736345554823, 0.0075336407404821446529, -0.0013453759414065913531, -0.0055627700461837294513, -0.65568196043366278047, 0.98471797377334435719, 0.35007916698798374755, -0.76740578776860646215, -0.18884616095726108886, 0.16555537696090510957, 0.11038453672654718885, 0.029537349460109520866, -0.0047184574399787770557, -0.011305076662559681439, -0.010903212117031239534, -0.0068358215812303092795, -0.0011187519797328040767, 0.0041988494662026434878, -0.093255878591787327858, -0.1717107799269868007, 0.6869285822091882876, -0.22558876314176315447, -0.29698791311733457876, -0.055196256372524316047, 0.067264956759229971306, 0.081409546052118864079, 0.044395933754189480236, 0.0051512271154655411842, -0.020292432548349295662, -0.023631363787136813043, -0.010051919382121399932, 0.011194984845613858288, 1.3124694170230877255, -2.3871738382600722339, 0.37488205833132226541, 0.6278989420341216654, 0.32485218213160782508, 0.04665088658435459501, -0.16648071038372597164, -0.15258014764378541317, -0.051550569528880159365, 0.018377630086225820588, 0.034717293038975978459, 0.022541786132074134036, 0.0043002235057977193991, -0.010505241215006023633, 0.65208530116555529776, -1.4019793751924196634, 0.48072806254878724941, 0.18038160641710077048, 0.090675627222645896475, 0.12143949657377411944, -0.00066205914394817585999, -0.068251650432976221361, -0.059757127866736828303, -0.02672288181611901306, -0.00397984144913176522, 0.0063510133382545831682, 0.0089485627169067303438, 0.0074694807128112992653, 0.74449369947938370817, -1.5262222503043358923, 0.4406646617680499145, 0.24468601369602821505, 0.13395614835344965554, 0.11638428853348758985, -0.023260924276818376044, -0.083861757643117124106, -0.062606044574734606711, -0.021869913606223010533, 0.0020844692935034049922, 0.0099550888699083978778, 0.0092418875007454527604, 0.0048929334036351135317, -0.075435469306071978313, 0.35340190792268377784, -0.41051453159489131872, -0.12117471965174686499, 0.35776971221902331699, 0.02211566732300948232, -0.11335182278914206211, -0.0463891745891325466, 0.0054635106700185315232, 0.020564963843949235489, 0.012884873029909236358, 0.0012266023569376838676, -0.0042255887998575488579, -0.0022986775848650017537, -0.12757299810328104339, 0.40403999732351070229, -0.3419444880324677416, -0.14295758601024957657, 0.29251740266633968757, 0.0017215723390112485335, -0.091481192914374134806, -0.02249802949711312422, 0.012086374376571007438, 0.01517665144102762001, 0.0065325803712552284533, -0.0011012995759801854495, -0.0034168811004278896136, -0.00089391147123453267801, -0.66202308182345326326, 1.1044676879258019664, -0.086754668481824767645, -0.4110449100166262415, -0.040393867756126453084, -0.072626258931448542389, 0.079116392104943283892, 0.14007788164151283028, 0.03264204767847605354, -0.049790014231780733989, -0.050332618978799432974, -0.015703773521357704268, 0.0098560128215276054248, 0.015939335519270486641, 0.34282141002172783484, -0.8896076662982858041, 0.58736763159588567529, 0.060675352313431386853, -0.092021253506619563045, -0.031052806966917850112, -0.0037295656236163389138, 0.022918089533066239583, 0.018170782165071886621, 0.00063544281263960009434, -0.011192344375042783192, -0.011598927808095670436, -0.0044731621388698116284, 0.0062669657848047934101, 0.62729268144840277799, -1.5164486147255293869, 0.97282769495796261872, 0.19875651624672094342, -0.29772607002991846592, -0.038330567378950505808, 0.049275455843369023179, 0.017437374311732492527, -0.0043626253228183893909, -0.0054094137922104101879, -0.002934585329013961446, -0.00056775663266151203592, 0.00087959122774957136052, 0.0013326745496077599656, -0.25821309823534638328, -0.10547956818695310155, 1.232663929685462012, -0.43473221101417175438, -0.70628018841760031865, -0.058415542949074701629, 0.21194907747512334306, 0.15092905099555684045, 0.034307725890399501378, -0.022147643586625652412, -0.027414269996304523047, -0.013215632857419681856, -0.001028684651014963921, 0.0034879275021243274016, -0.13422367188492245749, -0.28053812618811491131, 1.1557506185422736777, -0.34966925876343035728, -0.62266305441381830654, -0.04649831034296342952, 0.18209867535807372496, 0.12319925108026869021, 0.025835441412772400249, -0.017434092649029037103, -0.020691875711724143361, -0.010099177171427317634, -0.0013117338162331253996, 0.0020092210505986895716, 0.24173670882774622726, -0.58188066716818676039, 0.54978520277450637632, -0.074679106153486995989, -0.12829910204758682157, -0.02496799568467998956, -0.014025641057135732059, 0.0040061483273604956562, 0.015224027305472075805, 0.017400340708074151647, 0.012164843043177523893, 0.0042778983237360540537, -0.0030367984715934347173, -0.0078261345411296517199, 0.28784127782642726645, -0.66853430646474398813, 0.54303709698933144701, -0.024909228342364725384, -0.11388186843923876757, -0.036146662494734214943, -0.021425107658137869382, 0.004794304090751826293, 0.017546634507588365626, 0.017388246367172829898, 0.010436567445787303665, 0.0026843374055135842647, -0.0033912237838723685579, -0.0066547794896078425395, -0.061834285756023038561, 0.27779830837935798327, -0.39701939543882525108, -0.21215561650173675456, 0.3842580823205805074, 0.17927038331216035849, -0.039905178483214294449, -0.089713308175305567227, -0.064482476212211231603, -0.018784013475486745909, 0.0080254491745386825896, 0.014307706113491965721, 0.0097747915573124944527, 0.00097751634053334923724, 4.6459224057994479296, -8.5483991495810833783, 1.287616897527497084, 3.3470839730223511665, 0.46966015858282184059, -0.63263828188098125693, -0.52895682328657955829, -0.18283309839618297477, 0.01969964229581540327, 0.070199463930393182776, 0.055523425423177001636, 0.026222425492014117543, 0.00019676579464997836491, -0.016709754033534231221, 0.032260792876806818064, 0.6170891418553942076, -1.4280090962745122241, 0.64888014014775419191, 0.34388381338156354117, -0.02005416700120864551, -0.076708195087841504756, -0.1081255433770212937, -0.062012210065943366988, -0.0016311527800871982591, 0.029671832112248726354, 0.027674360858855138329, 0.012335717705394839278, -0.0044478351148324575884, -0.26607832591387009513, 0.05213181934379208099, 0.535637201164786525, 0.047244375060901254437, -0.48690008576032001164, -0.050216436841527503643, 0.085485185488589798819, 0.089275414193336011248, 0.031980024559294786823, -0.02258422454143955288, -0.022350097760918646861, -0.0050282028445345931453, 0.0039637822854681759033, 0.0030223875471191508427, -0.18116792171716702975, 1.2891730431866728068, -2.2314512676056343921, 0.97810671302245211223, 0.54519186091861815413, -0.32745315087952214794, -0.11476106943856177744, 0.01690535216578181732, 0.020259746184798310686, 0.0070485872735768385217, 0.0017901057475820157747, 0.00025849799478421610398, -0.00036239501054488685184, -0.00085067179757199792162, 0.19497327259581045955, -0.41931420683038328479, 0.2019918487461892509, -0.0060380622190993795545, -0.017866756232869457355, 0.29368790792144588986, -0.097297738225693389258, -0.19941880338980180754, -0.058913097994766329413, 0.035581081646680293162, 0.058869791148417977, 0.039606868594833262442, 0.0061979895383672050854, -0.027202391425798482655, 0.14927643171378773834, -11.497485299342658394, 23.825545183803232163, -11.637273447664187032, -2.2878276109300696994, 0.099809820813981361964, 0.68399404445336475789, 0.53393790747286917764, 0.2348906560267157162, 0.01717003089703898408, -0.080107437579393828142, -0.081722753991664273676, -0.034422792534565080369, 0.025634173589705166324, 0.27868790418897843031, 0.9784974476929019227, -2.6601856615585561805, 0.8251195791746014363, 0.88456503254804241809, -0.037671354051911992711, -0.1288312548717631123, -0.10192261408744712758, -0.063060142196895319078, -0.0095831908357914396407, 0.022237213044257988365, 0.021385500702074660556, 0.0075336409170088380893, -0.0067459420788033496938, -0.521955967035803714, 0.35007916724922488738, 0.48478693387875765319, 0.32256341791000486818, -0.70328040606333974694, -0.21345774099555475667, 0.14744943411838495884, 0.11321881194176769414, 0.039311703204622475138, 0.00029094894055464021448, -0.011027146655773307007, -0.012198614578404820552, -0.0068358217678639900117, 0.0027363865964063916608, -0.54720362855006443592, 0.6869285820862782721, -0.23868418548810294011, 0.58301313824781364481, -0.26813258042727350361, -0.34189926763585082314, -0.076266328852279921535, 0.072668612252048264688, 0.10071872361602399182, 0.058133490302112511561, 0.0090441512805905545369, -0.021749158604360715324, -0.023631363904613813393, -0.0021851659015896252036, 0.74774406522126235508, 0.37488205827930776159, -2.4357354247549167781, 0.4664057449773723607, 0.78634671796952293477, 0.36512671269067814572, 0.043070683937330350721, -0.18001001740535410578, -0.16561109097428960979, -0.058827922770772339556, 0.018970559163321781132, 0.03895461812492533038, 0.022541785727279179535, -0.0093741709275503067728, 0.35609093527516805233, 0.4807280627582088961, -1.6402374136902475588, 0.38472659088814958883, 0.26950494536948910174, 0.13772799368765814565, 0.14415757347517699216, 0.013991170040679002523, -0.061259602621719491666, -0.061766346809149788588, -0.03227559733506293882, -0.0080894415195884410152, 0.0063510131793524590704, 0.013785444209144957425, 0.43514684662589614428, 0.44066466183570612847, -1.744835815858118977, 0.37082254553289878407, 0.34594306158485138569, 0.18551741675496785655, 0.138479372530344369, -0.011807489095991979128, -0.080218918336485228737, -0.066736408451154333177, -0.028117787108642856853, -0.0018644100946144806855, 0.0099550887553793276225, 0.012709028601665355376, 0.24631821733806394392, -0.41051453113879254886, 0.19800080059276581324, -0.09161748706876551851, -0.13164168893443051189, 0.30248620394098091335, 0.015323479528946339079, -0.11711918667358582613, -0.046983980404326469715, 0.0077316552164289134574, 0.023558401477218573333, 0.015080567074634740679, 0.0012266018281524957052, -0.0091043459902182943128, 0.15351200161162750768, -0.34194448760600587578, 0.27027142614566440626, -0.061977316941810765838, -0.17565341639898121584, 0.23769892213002805081, 0.0014254932977138014272, -0.091452361879989660376, -0.022479756149674526289, 0.013733127647768067839, 0.017354141962669926424, 0.0079994150431009561353, -0.0011013001661411218456, -0.006032010281965885512, -0.34201089187582200957, -0.086754668335185414207, 0.97951878406500125784, -0.053461884901669989856, -0.62344498746489163121, -0.084607022510804721915, 0.005667110666873555018, 0.12447589800340941313, 0.14772705766100424629, 0.02667457838153613689, -0.058919026307615730842, -0.058700074297510292098, -0.015703773479296634363, 0.030835860743870562029, -0.13245192302494718239, 0.58736763153410032068, -0.93686221410906345142, 0.53926314639492789471, 0.055332696082835426676, -0.12089959508908419694, -0.037767692849504271024, 0.0017220217792586569629, 0.032520570480156357762, 0.026952398063824885288, 0.0065987104844037920703, -0.0088098924554466020304, -0.011598928074963392801, -0.0052520475258867240637, -0.20100474592633124638, 0.97282769498422871912, -1.512918081808612758, 0.73374026833087757637, 0.25169401012529940553, -0.25625856414921471815, -0.039707300537848544408, 0.045446710484431260269, 0.018635867202209881188, -0.0039420221262307280144, -0.0060786558924215399607, -0.0036972626789486466407, -0.00056775690044198691333, 0.0023665133321018098043, -1.0203464357977716226, 1.2326639292992644936, -0.22803440254397927545, 0.8603521863751892873, -0.45677846961948775917, -0.68490422318780797006, -0.059463389064183622135, 0.2089438735917042711, 0.15442396909141659078, 0.038206378359075116036, -0.020348292551772607234, -0.027789025803148629384, -0.013215633274660567795, 0.0047265066601563507018, -0.8756077870854629408, 1.1557506185899042439, -0.39669655333265346764, 0.8214641748288635803, -0.35252048953175679769, -0.59841169827634443923, -0.051245636669631049187, 0.17565721465376712973, 0.12541360865089515308, 0.02996700494054298225, -0.014781622395728531011, -0.02017606969272585371, -0.010099177001634217915, 0.0017700394828895301726, -0.2306219066709925547, 0.54978520269090291883, -0.6837648282816942702, 0.54978296671835291676, -0.018504049427940507527, -0.13360519337729759304, -0.040342518941611470173, -0.028754431771913471078, -0.004574120905189074833, 0.012915756510409708013, 0.019144351964844083525, 0.014440985707456981818, 0.0042778991987925587409, -0.0081168377419659244743, -0.19999909520257755369, 0.54303709650897435957, -0.74081997785423958636, 0.54509540325498606617, 0.021539468625957625125, -0.12297867715923697929, -0.050580247630996319486, -0.034230283954857131101, -0.0023253677346321832697, 0.016192132906224607214, 0.019685697736006355085, 0.012914042231831211055, 0.0026843373052272976112, -0.0083311099611019390904, 0.46013592854009022215, -0.39701939497578481175, -0.11955975709558494535, -0.20394883862730092683, -0.15421983406113479176, 0.40387463032119080841, 0.20504575846690334395, -0.031748577731795311896, -0.090778298569546098107, -0.071210196106864007137, -0.025960416232216513427, 0.0042180633763738246619, 0.014307705868879409816, 0.011857913656176919531, 1.8839922565668520971, 1.2876168981059450402, -7.0601710981925052124, 1.2341086539247361209, 3.3466272302054171917, 0.55553026084130385787, -0.5916346164908606875, -0.54557588701959103261, -0.21632628870792930398, 0.00306900469326562347, 0.073049794597219927517, 0.064117782099743406143, 0.026222425598760819299, -0.021125919156711692182, -0.22348521743322954936, 0.24233728144257085657, 0.64888014001737359671, -1.3876400205750309969, 0.61569605755567025085, 0.31666050810761708911, -0.029900493335461336003, -0.074436677895613423406, -0.10671329695010828287, -0.063637351788341484093, -0.0036165964472225742114, 0.029671832203236478409, 0.028712085274551601977, 0.010009565005530546289, 0.15941423464289192213, -0.47132723887329658341, 0.047244374775360482854, 0.52523719350447595211, 0.11502089392758670094, -0.45852201512075813072, -0.067248993972300469268, 0.069994581446849238326, 0.085091196804993218805, 0.033879024474895616525, -0.02091851130494111477, -0.022350097819755845863, -0.0047337069310677305778, 0.0075100620468539240268, -0.34307307236480205637, 0.48794424806944353135, 0.97810671225205980672, -2.2688368153750952949, 0.96981754125843189218, 0.56388481284850999753, -0.31770403975267169638, -0.11419972938552949704, 0.016164062699669138473, 0.019803596135753380458, 0.0065329913013886476911, 0.0017901055322114267903, 0.0004862523689102887705, -0.0013697548379274027477, -0.076544209497272683307, -0.0056261490589659375952, -0.0060380635654812536001, 0.19061724588637901645, -0.093293901372620086088, -0.072223773861376239225, 0.27710642012455821703, -0.087501610661535222935, -0.18694102663486439342, -0.050342495328164524326, 0.040280965862310821879, 0.058869792613292802608, 0.030390412161867240426, -0.019520756939513092543, 1.5199230575231481577, -1.9244733902916848489, -11.637273447843698548, 24.4019989733938516, -11.420238084300189385, -2.2605899947981678011, 0.068519163331840202136, 0.64324606906593906341, 0.50565287929100666275, 0.22189753246926341435, 0.01260025068903970237, -0.080107437322642138788, -0.072460542810944744696, -0.0029335908635015578265, -0.32702736686368433228, 0.79561328879327120056, 0.82511957912056310693, -2.6523506509326870706, 0.7843511207119598927, 0.85848081460511127538, -0.038612571516763949386, -0.11855129249693023563, -0.094376486932959496867, -0.060274930245174898613, -0.0091261023509002902471, 0.022237212961007325057, 0.019740718926989933579, -0.0003012633317181356179, 0.19324503806761333946, -0.76740578815724114214, 0.32256341802983745648, 0.53442490254850472109, 0.35900133262630701259, -0.69033508321392811347, -0.22017130755684530263, 0.14209319099207146087, 0.11036991127869530405, 0.037835469135503715932, -0.00048609575174149378864, -0.011027146915578564176, -0.010903211871492214657, -0.0038923999559922176257, -0.14755073822152858787, -0.2255887628118063426, 0.58301313819070055278, -0.33253973583798707159, 0.55764580810602404348, -0.24757760659774832868, -0.3185572666628778693, -0.061798348524194489373, 0.073580625637657312565, 0.095431968760926744189, 0.054097629277076210441, 0.0090441513129131990978, -0.020292432429324692988, -0.027378040354074085344, 0.0026860409843824329876, 0.62789894219847564028, 0.46640574514742644263, -2.3122367743409681928, 0.42424147652207122583, 0.72619430392228212501, 0.33859546482517349775, 0.049027363041023508394, -0.16480803129972654286, -0.15470130025449346789, -0.054544006337476275137, 0.018970558902942417229, 0.034717293296589402107, 0.010065181108985032238, 0.18542834691138815328, 0.18038160586813276054, 0.38472659144220289207, -1.5711286891910121177, 0.40774885686485140557, 0.27078568731070834508, 0.13242329868244290259, 0.14485381785166365209, 0.016416791007140572878, -0.060478122279483396972, -0.062733067969096237282, -0.032275597734473117362, -0.0039798415777644114213, 0.016579047094615821656, 0.17406641210973095024, 0.24468601313727503244, 0.37082254618814797231, -1.6631050551207247068, 0.39011314656555912128, 0.33973764877955436869, 0.17639137474386493154, 0.13883241523665285078, -0.0085646401253901549566, -0.078868208915404053316, -0.067549957401918755973, -0.028117787652599160464, 0.0020844697879626606501, 0.019395761247295303237, 0.11493684097920768561, -0.12117471990483362487, -0.091617486607754039518, 0.20424137879406351059, -0.14288027774554967753, -0.16532574930548249448, 0.2936101547019624447, 0.026401189846688054991, -0.10710801309541123638, -0.043374529868909349195, 0.0080181280460404919658, 0.023558401251171635221, 0.01288487331700234606, -0.010006753123789914259, 0.08946103685582749987, -0.14295758628860424921, -0.061977316800145947073, 0.25057269013482519338, -0.10795885278936730167, -0.19488272670417003396, 0.2352705733633873586, 0.010485765586895914869, -0.085344577962458215148, -0.020889562972799112506, 0.013634987625420352114, 0.017354141585092023453, 0.0065325809602307677998, -0.0083874541153218899747, 0.11602591928306778268, -0.41104491054090819269, -0.053461884262285028213, 0.78064726718211574497, -0.037326700546426894167, -0.51808039532092142743, -0.037636895705071893303, -0.0047262884584225780399, 0.097735339772218488008, 0.13187174220752775899, 0.022002592772935845172, -0.058919026379912732561, -0.050332619045065980778, 0.019133053768278149992, -0.087672111464281096693, 0.06067535245883326489, 0.53926314653574447533, -0.98673175069254570602, 0.50025322077951994704, 0.061190835022510457752, -0.10674354713179515231, -0.026867008460011568283, 0.0064654108300066480408, 0.034147144470879860212, 0.02754690296744050082, 0.0065987102941165512449, -0.011192344295932556142, -0.019269385158522014945, -0.17852748245650604253, 0.19875651657955745177, 0.73374026799903369067, -1.5269039091113738582, 0.74718057022659944977, 0.26884105276754410863, -0.25180273654988155618, -0.041290528460267969524, 0.043211346392579327857, 0.017724525382310736543, -0.0042531972409899946805, -0.0060786560487036851719, -0.0029345851802886004263, 0.0017782549740638603324, -0.2293750619666960755, -0.43473221057227634523, 0.86035218564670545849, -0.32987413449192826143, 0.91953208374895922805, -0.38190835692111768651, -0.66773023007651743921, -0.075390824005619833104, 0.19206942684853731129, 0.1485123825893553684, 0.038423988588080965068, -0.020348292617376174951, -0.027414269155821360813, -0.0040540039974518203916, -0.21061706375038360961, -0.34966925895429379345, 0.82146417501575397235, -0.47525549438232689292, 0.87347782603665691248, -0.2923724925541957842, -0.58734165344168154377, -0.064983280779789792669, 0.16295744385469634885, 0.121889182366572979, 0.030810264940754457375, -0.014781621650942454449, -0.020691876444036697708, -0.0049241380327152686411, -0.13954972513384492072, -0.074679106383670895708, 0.54978296698864748127, -0.67655851892445106888, 0.53445639520594645511, -0.023664466644489283442, -0.14039088726148155062, -0.043244561388255602086, -0.028564610642604328083, -0.0030420351278428493791, 0.014013711475518888031, 0.019144352862292192408, 0.012164841999001004338, -0.0019688055535147771027, -0.14890422394273591933, -0.024909227579394021407, 0.54509540248689136721, -0.74328953421972687732, 0.52274003120232748199, 0.016201992707076923184, -0.1267978298231957901, -0.051281521237456245688, -0.033317644634304921514, -0.00074923495283161709035, 0.017277779186922172239, 0.019685697824091331898, 0.010436567546191953537, -0.0042092428405876144634, 0.39178146004018943316, -0.21215561702843260816, -0.20394883812920142074, -0.023862088976284240999, -0.17048021777726984904, -0.16849885881235149299, 0.38078241041117083521, 0.20002085382784451362, -0.029525081492415703593, -0.089130604153531675649, -0.071596373408841645913, -0.02596041702720318714, 0.0080254498195122848936, 0.024135404964679276624, -1.0603454203040068027, 3.3470839725594569991, 1.2341086545924022655, -7.2085752286846869197, 0.99197162653454240555, 3.2734383325075944171, 0.58696145299813862195, -0.55107648096342742861, -0.52202340551267967239, -0.20585910669897311553, 0.007044302893184870365, 0.073049794674690179952, 0.055523424809977177385, 0.00063552265989029075796, -0.20144930354735160472, -0.034649479213870043592, 0.34388381278276181074, 0.6156960574141481235, -1.4348666230697759705, 0.59207845795609725936, 0.31532317822900524096, -0.025779504169936417329, -0.071576834162566244091, -0.10620996445352220983, -0.063637352493999904679, -0.0016311522029297128265, 0.034754842334785186753, 0.034748222814869916819, 0.16653397071108511085, -0.074061366310719162209, -0.48690008513191940231, 0.11502089344099068746, 0.56272105480379341902, 0.11068105834788789787, -0.46876907471550127715, -0.073003773345017217156, 0.068603802838139021913, 0.085504959816356251867, 0.033879024219730442813, -0.022584224331227129023, -0.023009893061211626059, 0.0059965267252120209307, -0.0006248527410269697803, -0.37656455920182879682, 0.54519186191855184553, 0.96981754066207348508, -2.2848896349753826662, 0.96379478908750304544, 0.56761781331263727601, -0.31418236939285354437, -0.11361244400245633368, 0.015442608247963155246, 0.019803595856231637901, 0.0070485879240118134992, 0.0011925279118036951308, -0.0010259776037011018781, -0.30679343596098168634, 0.36682118322820578271, -0.017866754874163358335, -0.093293902641681289434, 0.14142627844976374263, -0.11567467220165000841, -0.068865492007879614, 0.29275751432182228751, -0.072878673751153782989, -0.17990841336078691737, -0.050342493985369593523, 0.035581080193662351085, 0.049908597565602841328, 0.014923338987306224157, 1.0960350271026997326, 0.29186058153835570117, -2.2878276108716821824, -11.420238084495911934, 24.68312122902475636, -11.304514084375087535, -2.2686629144540750858, 0.02230895506396395056, 0.60752721474102366894, 0.49109244952334502043, 0.22189753276067342158, 0.0171700307154509825, -0.077425677784978150231, -0.073365196223458828406, -0.22329119845967929092, -0.088269519820016503764, 0.88456503263437924467, 0.78435112065452228247, -2.7141840041467069078, 0.75577507573473357017, 0.86167366708609638426, -0.028380609350445208661, -0.11135157072286354374, -0.091572419066686908118, -0.060274930313564817341, -0.0095831905918525479848, 0.0224971769288168158, 0.018120277948707904508, 0.25782736577080200835, -0.18884616070246440511, -0.70328040610207154248, 0.35900133262027184022, 0.55670710692492009208, 0.37326420405621690835, -0.68628867600749665634, -0.2250299747443483267, 0.13653903139767559205, 0.10781071826865336327, 0.037835468865193878918, 0.00029094922455877471465, -0.01130507668276584457, -0.014732635238741145778, 0.16485916004573314408, -0.29698791333343410592, -0.26813258054001731789, 0.55764580848776890143, -0.31957921254865073735, 0.58889949272971209293, -0.22185360374729246047, -0.31331665948224696194, -0.068885580022879089324, 0.066777085948892367373, 0.095431969099186234051, 0.058133490156285050343, 0.0051512270716408308305, -0.0508310487590936666, -0.33332681871193142742, 0.32485218181005454108, 0.78634671773942144135, 0.42424147713283616445, -2.3926027386920498685, 0.37061271037339682533, 0.71370977888616904394, 0.34981149485882884198, 0.065023774885674914814, -0.1558614605322933, -0.15470130027333642286, -0.058827922753341130302, 0.01837763007295058898, 0.046494484643612235542, 0.011326057354050209774, 0.090675627369141906242, 0.26950494532779051271, 0.40774885693930068564, -1.5795375780048215297, 0.39320717115164105504, 0.26271157800097461044, 0.1285740488115596214, 0.14412160920058492142, 0.016616257520757998084, -0.060478122176503376628, -0.061766346995781887252, -0.026722881686621291486, 0.010821622343866250937, -0.03906926740204380577, 0.13395614910314004864, 0.34594306064117852362, 0.39011314727068563402, -1.6800344786294998389, 0.36904247954189872694, 0.32942051960880969208, 0.17327711568330519953, 0.13952653149748567296, -0.0075850850191177959791, -0.078868209470550482143, -0.066736407800592256745, -0.021869914295146744565, 0.019554725777162627887, -0.16652490461142460676, 0.3577697121211135256, -0.13164168931983677435, -0.14288027726239657667, 0.18065868285420227934, -0.15282164431219091849, -0.16128124124659920513, 0.30243563387126332076, 0.032534334708149974247, -0.10538046988079736999, -0.043374530034260685984, 0.0077316552791667795749, 0.020564964064834979213, 0.0021111310788298094654, -0.12521756115083990113, 0.29251740272274501242, -0.17565341635141018539, -0.10795885275372635315, 0.24081769852815163024, -0.1099709273584989766, -0.19067942374592697363, 0.24094465915468085693, 0.014002802610652316864, -0.084622325258023364802, -0.020889563018443049053, 0.013733127888585265619, 0.015176651031499104744, -0.0026359997504407694027, 0.22407433123253880924, -0.040393867162471841503, -0.62344498815273741155, -0.037326700142216166778, 0.89996323019488411177, 0.023559938855744862879, -0.52143511406387599294, -0.063555226757651378522, -0.024964957175560603675, 0.089292394695686713724, 0.13187174209321128604, 0.026674578322225375726, -0.049790014050398012646, -0.033710661760364801776, 0.030013273106678978186, -0.092021253474430297903, 0.055332696081486158757, 0.50025322089926882452, -0.99178114422665553107, 0.51153021227823536332, 0.074113579565152840067, -0.10025173017018408628, -0.025319326468636233624, 0.0060863463095634940531, 0.034147144576506041624, 0.026952397780433741403, 0.00063544338562999957976, -0.029474564814404864699, 0.060955432885580369817, -0.29772607054155214579, 0.25169401096721849553, 0.74718056963973988882, -1.5321678032348924337, 0.74446138923079607785, 0.27084461789618474725, -0.2521245698812137892, -0.04266382870069777844, 0.04245353033760902034, 0.017724525410522683211, -0.0039420221427790957253, -0.0054094137310391320755, -0.0013581103275403202165, 0.37381501785694537521, -0.70628018873519027299, -0.45677846905161217927, 0.91953208348863602062, -0.26991495301820572772, 0.95206937954004711155, -0.37630298065829892451, -0.6784989495099584822, -0.085998158659442450191, 0.18798364432148384484, 0.14851238275003036948, 0.038206378709967389606, -0.022147644599391995424, -0.028472233858297890663, 0.31455317222583428771, -0.62266305423459056279, -0.3525204897957048833, 0.87347782603973100901, -0.43039365440864008061, 0.89654547239099025013, -0.28776948474439550951, -0.59475961124880805109, -0.072377558176507594823, 0.16026879934279833484, 0.12188918289077918522, 0.02996700378746305557, -0.017434091503940186973, -0.022425367446991004844, -0.034427215660772593486, -0.12829910146584116992, -0.018504050440617314521, 0.53445639565802682824, -0.69269138381260386694, 0.5259704931765906144, -0.023704786611537721763, -0.13738927566885156262, -0.040056154232958392225, -0.027038202772358334303, -0.0030420350629137855175, 0.012915756038802737524, 0.017400341168560901073, 0.011355775121475696066, -0.044353930065246710557, -0.113881869312847675, 0.021539469479017699094, 0.52274003068957697327, -0.76018846509029791569, 0.51601986497707541535, 0.017852953841324736839, -0.12294535117525047518, -0.047919324536495611078, -0.031852744995454110311, -0.00074923480309929506304, 0.016192133000104358409, 0.017388246086950023134, 0.0073111055921249246473, -0.050182906948352153098, 0.38425808255122589907, -0.15421983431372476847, -0.17048021758771764178, -0.024431484138376157594, -0.19296008783307033729, -0.18507277541975566959, 0.37468755013431936929, 0.20069222602705336089, -0.027872830501882175924, -0.089130604765019702529, -0.071210195116383032121, -0.018784014406599922054, 0.030555171850134271033, -1.1329583073154076978, 0.46966015857008497347, 3.3466272297540320402, 0.99197162689575046901, -7.3753434487159958977, 0.91579648416972814218, 3.2731371055482698651, 0.62050617006041208779, -0.52077485945519430555, -0.50913716481440496331, -0.20585910680330760103, 0.0030690043543234392549, 0.070199464801415034643, 0.054627543330639370678, -0.10637003339631039001, -0.05977518577554098711, -0.020054167669427871989, 0.31666051057256372481, 0.59207845705948747295, -1.4427990955284331065, 0.59114628932759616387, 0.31663535213982030303, -0.024705231659896419866, -0.071576834639049652376, -0.10671329651710410868, -0.062012209860386323379, 0.0081707280030619347522, 0.063326725822120483178, 0.054013649242046421484, 0.070590412621230447243, -0.050216437257856474941, -0.45852201500545802793, 0.11068105857778109957, 0.55750190148188849548, 0.11013359006769722626, -0.46815385849671309337, -0.072926740263042966195, 0.068603803157109052857, 0.085091196386643905281, 0.031980025024763061858, -0.024050997491606974921, -0.013641731335398409389, 0.056698828653204913497, -0.11332917978812685755, -0.32745315146485770574, 0.56388481312134486334, 0.96379478925393280431, -2.2909158070038539989, 0.96315083476175988864, 0.56871492127583078702, -0.31409056602391977053, -0.11361244348885896072, 0.016164062235853690619, 0.020259746361855367203, 0.006267935187400036981, -3.2573956180280667788e-06, -0.11169518084932472668, -0.071419781090420977887, 0.29368790797521610036, -0.072223773351026310241, -0.11567467300199561764, 0.14380896950276678137, -0.10378526546363926741, -0.056268007034069492112, 0.30045259437293347915, -0.072878672963648263861, -0.186941027108464497, -0.05891309796466620563, 0.036194949254645593373, 0.071983223036160920438, 0.40975839267476138472, 0.70677275195631050764, 0.099809821003892754132, -2.2605899949831833595, -11.304514084246523709, 24.753980707070361689, -11.29788011645830359, -2.2897095273470582377, 0.0049868960407505498467, 0.60752721507664397826, 0.50565287902670452702, 0.23489065593309507718, 0.0012166129358165830247, -0.16103448007808290887, -0.10167075664825869774, -0.11839823480420147428, -0.037671354007150249787, 0.8584808143256985602, 0.75577507585627501374, -2.7237428972394823745, 0.75632358082211714745, 0.86574811281001473606, -0.025252603684564611797, -0.11135157070030490822, -0.094376486886747587568, -0.063060142143004885296, -0.0064942639749359645437, 0.043182905490160856943, 0.083950985063092342653, 0.16555537675000245312, -0.21345774067363218474, -0.69033508372595353553, 0.37326420435435209155, 0.56786702556462897995, 0.37621348403544507821, -0.68882434887364174525, -0.22789735270648528442, 0.13653903126385202871, 0.1103699115332509445, 0.039311702991868709145, -0.0047184574530418719324, -0.02578020333395076788, 0.19663612042951189585, -0.055196256849312841375, -0.34189926725366753812, -0.24757760723586635931, 0.58889949289484544437, -0.2937532643024300949, 0.59781161473335853174, -0.22537944262002332363, -0.31906829172075068524, -0.068885579978516103816, 0.073580625697119927953, 0.1007187233797572079, 0.044395934053952243459, -0.047244143910045360846, -0.3066417571636567696, 0.046650886598967503116, 0.36512671299007015691, 0.72619430318251676493, 0.37061271065026263338, -2.4145202049929443966, 0.37044653556527828986, 0.72278524482830408537, 0.35749698829851822568, 0.065023774931513983466, -0.16480803122310583309, -0.16561109114218058869, -0.051550569437531543193, 0.075823462899428956119, -0.12817832922584473931, 0.12143949707095567925, 0.13772799313627351325, 0.27078568740363040401, 0.39320717084036688194, -1.5908861205999205168, 0.38545776756382366113, 0.25859181921079621436, 0.12719525855157140515, 0.14412160914226396224, 0.016416790494483263912, -0.061259602050873815382, -0.059757127590307407283, -0.012674256869439811876, -0.1636047665526639272, 0.11638428750892172714, 0.18551741791522752267, 0.33973764741595791072, 0.36904248017080398814, -1.6944126779185344489, 0.36089695325534887349, 0.32607497142472580132, 0.17263954291962452237, 0.13952653035253312352, -0.0085646385717833804774, -0.080218919468009436402, -0.062606043824477902726, 0.00079480974018521373912, -0.10487744278701004175, 0.022115667448318339361, 0.30248620400498155147, -0.16532574951965375853, -0.15282164409018617057, 0.18597106256965065252, -0.14609915168922862905, -0.15742187848306643527, 0.30386372587464727246, 0.032534334668858432005, -0.10710801303097583492, -0.046983980402935651699, 0.0054635105840479076741, 0.02712559884307780278, -0.061759940877007007964, 0.0017215725381896931044, 0.23769892207982351584, -0.19488272689275970562, -0.10997092706065930701, 0.24644600805098773599, -0.10552272543721873121, -0.18859347291385869516, 0.24158950436368314962, 0.014002802529420689537, -0.085344578134224788335, -0.022479756199463549887, 0.012086374923655123295, 0.014320193453913979478, 0.21070093646233933882, -0.072626258970303725349, -0.084607022647963534001, -0.51808039522627169493, 0.023559939008359585821, 0.90155474771923360677, 0.0032906124781300252063, -0.53803429254407209026, -0.071356131797015262541, -0.024964957297455587365, 0.097735339634505591366, 0.14772705819983314446, 0.032642047066988623405, -0.10401201131248639931, 0.063277367906879364301, -0.031052807320660652524, -0.12089959469776208334, 0.061190834479933529921, 0.51153021236272777639, -0.98173801718078212986, 0.51701959487777882085, 0.076055847689058084238, -0.10029203476827794339, -0.0253193263875996534, 0.0064654112216481537251, 0.032520570221188302429, 0.018170781953223705263, -0.024286395997358444648, 0.049112507702505539542, -0.038330567257558352534, -0.25625856494833637589, 0.26884105352541287637, 0.74446138909030368236, -1.5355616248996035722, 0.74272135127952831901, 0.27025356031788416722, -0.25253563097128650661, -0.042663828363707599622, 0.043211345865415320966, 0.018635867482939456191, -0.0043626252582733568777, -0.0073048210690563955549, 0.311829149359704072, -0.058415542740105901032, -0.68490422280056917348, -0.38190835761204489085, 0.95206937992380968883, -0.26318484132621361438, 0.94984538408790986086, -0.37964046658439976056, -0.68101442407103707222, -0.085998158689165438151, 0.1920694273702912569, 0.15442396782323100646, 0.034307727151131006682, -0.057324304203673029368, 0.2588171308865986342, -0.046498310282659029158, -0.59841169826175499846, -0.2923724919708683978, 0.8965454714738778641, -0.42546526284442448551, 0.8960800374431480142, -0.28907591434640145245, -0.59613403324254510274, -0.072377558156169030656, 0.16295744345495194461, 0.12541360937570100731, 0.025835440714997291722, -0.043560141743049042573, 0.0046452380444985509958, -0.024967996146837385246, -0.13360519264797118821, -0.023664466682187999119, 0.52597049286754038455, -0.69495068486450006784, 0.52868934206761764028, -0.019726084837133717881, -0.13522221147995269752, -0.04005615477075176456, -0.028564610376640903189, -0.0045741207217869558205, 0.015224026998699381752, 0.028343677260339841112, 0.0072853613068661719765, -0.036146661461471767118, -0.12297867808784557209, 0.01620199342428012973, 0.51601986471179595029, -0.7612295820502287258, 0.51909042971643393205, 0.0218281370240829449, -0.12082085947221896105, -0.047919323865665885975, -0.033317645298993281422, -0.0023253676771568566883, 0.017546634812153620586, 0.024669571954109192208, -0.19944094429116893119, 0.17927038332071473792, 0.40387463013760505071, -0.16849885893154817329, -0.19296008734335642898, -0.037808149498795989851, -0.20044378604833881785, -0.18821389654935280489, 0.3744064924667845351, 0.20069222579808335127, -0.029525081281563139018, -0.090778298814301869313, -0.064482475859603954893, 0.013853104427314279212, -0.42502576070205055903, -0.6326382819810658642, 0.55553026075708833442, 3.2734383326034452999, 0.91579648408280078797, -7.4191247343854476171, 0.90927272137934167517, 3.2887632513019071645, 0.63482114599353522966, -0.52077485952942137448, -0.52202340557306681212, -0.2163262881704991758, 0.01969964161131917274, 0.12982165331384953877, 0.009888176298416369997, -0.090141773246900411265, -0.076708194390493061876, -0.02990049507857236652, 0.3153231783325872728, 0.59114628970806437636, -1.4437295760127084598, 0.59086486790164649552, 0.3166353519634431124, -0.025779503318796564287, -0.074436678161112032615, -0.10812554361979548334, -0.049020497339100313483, 0.06735771130739130752, -0.03581495141265240989, 0.097539917550072616215, 0.085485185645678640753, -0.067248993934809112027, -0.4687690750028000175, 0.11013359040430099889, 0.5601714975980115474, 0.11159584897989019425, -0.46815385859436531302, -0.073003773523713746951, 0.069994581903728730898, 0.089275413613740847718, 0.034980369569854445622, -0.045521868918630359335, 0.022718400024638540913, 0.024749901008592079671, -0.11476106945032725493, -0.31770403952625625443, 0.56761781284615187015, 0.96315083517291921744, -2.2922319761639147195, 0.96239580426347925979, 0.56871492152057712133, -0.31418236991416076398, -0.1141997287734863864, 0.016905351652322056499, 0.022440242689505111862, 0.0042398702634799615568, 0.08301837799003095042, -0.17439445019578134199, -0.097297739791055895542, 0.27710642054727951944, -0.068865491366057796085, -0.10378526645362037839, 0.15515993618204657056, -0.097290589634010951614, -0.05626800623197984047, 0.29275751375114456021, -0.087501611265718981514, -0.19941880173565471046, -0.043640064601357261842, 0.11850697387970886287, -0.031683463261324984472, 0.48755226868592344003, 0.68399404418465603595, 0.06851916356540996289, -2.2686629145183632161, -11.29788011631242739, 24.757165721057614149, -11.301850293496729094, -2.2897095273405718707, 0.022308954740466603595, 0.64324606921689020389, 0.53393790775801208692, 0.19436376183452169775, -0.18669630143217996743, -0.0048565606740379038339, -0.087924516292860341315, -0.12883125496872843696, -0.03861257121904806966, 0.86167366695856018044, 0.75632358091975337988, -2.7238730535230692986, 0.75694727436651476626, 0.86574811278761742983, -0.028380609426856825278, -0.11855129236816099036, -0.10192261439024702385, -0.057371305327729348045, 0.046282454324917017552, -0.0087552574257199044377, 0.11038453701582502664, 0.14744943369234039809, -0.22017130702572673684, -0.68628867630154721802, 0.37621348406376092788, 0.56822000704532771298, 0.37534276720781623027, -0.68882434907731915352, -0.2250299744930477075, 0.14209319085878807654, 0.11321881200755994079, 0.029537349404154599614, -0.03063218314252898547, 0.10319098822009235839, 0.06726495726667604258, -0.076266328995102911081, -0.31855726627754815672, -0.22185360389798877701, 0.59781161455481368616, -0.29649727523701580045, 0.59322377706542772646, -0.2253794425908598742, -0.31331665942075898057, -0.061798348685963920557, 0.072668612405461185944, 0.08140954572344272222, 0.0043413541401325847491, -0.10287556084576157667, -0.16648071032527136937, 0.043070683678095203484, 0.33859546535526718891, 0.71370977869969298713, 0.37044653557753592921, -2.4095235187830614798, 0.37480648347098810547, 0.72278524477522876346, 0.34981149464790567416, 0.049027362979220556749, -0.18001001720321554989, -0.1525801476455550254, 0.042850776570028885348, -0.12592647765072167987, -0.00066206014472644687807, 0.14415757481254948269, 0.13242329770890870488, 0.26271157898919722529, 0.3854577673462095122, -1.5962617271863066559, 0.38291479722247356854, 0.25859181931655150777, 0.12857404836669206194, 0.1448538188134218252, 0.013991168672037899295, -0.06825164985520044969, -0.063673713767425033083, -0.13294125645409970993, -0.023260923363584928847, 0.13847937148823416531, 0.17639137609051219746, 0.32942051882804895824, 0.36089695354988871046, -1.6994023308637620229, 0.35876626486757617895, 0.32607497073407976096, 0.17327711693587516395, 0.13883241361714049322, -0.01180748799501346244, -0.083861758204734099253, -0.053283444002489382951, 0.013844439148798475328, -0.1133518231561683326, 0.015323479931277873486, 0.293610154550051905, -0.16128124140724087177, -0.1460991515961873044, 0.18942591561622607044, -0.14545374552465720197, -0.15742187862527221021, 0.30243563408764922995, 0.026401189871816142102, -0.11711918687953615237, -0.046389174439323159482, 0.045095419930029349842, 0.024896903993975830205, -0.091481193566800039774, 0.0014254935826488715402, 0.23527057348065216758, -0.19067942395999673555, -0.1055227252311032593, 0.24837708998664043136, -0.10534701173474718405, -0.18859347287806979554, 0.24094465909476273091, 0.0104857658516660842, -0.09145236214730674551, -0.022498029441871437994, 0.033870673839998380328, 0.037459095065470765507, 0.079116391430390550776, 0.0056671115462487197181, -0.037636896369092301406, -0.52143511395661346075, 0.0032906127867784857287, 0.88720753781412253769, -0.0026171645661366127987, -0.5380342931010684282, -0.063555226470466033573, -0.0047262881384354631098, 0.12447589710789974671, 0.14007788255679953782, -0.10514425765366497423, 0.046649996712773068863, -0.0037295658071618978738, -0.037767692973908598142, -0.10674354661603352445, 0.074113579726840128625, 0.51701959486567605762, -0.98051013031586486957, 0.51678472298778521665, 0.076055847464749304354, -0.10025173021793210015, -0.02686700897826930709, 0.0017220221098589970354, 0.022918089629980025607, 0.0046824457716046433836, -0.0021882464124339633116, 0.049275456035764900187, -0.039707300145094374944, -0.25180273705655009664, 0.27084461798630976581, 0.74272135122906723925, -1.5372521959160565519, 0.74209755348951977894, 0.27025356061408534192, -0.25212457053587411515, -0.041290527479557456725, 0.04544670986455077305, 0.017437374278494031932, -0.013545960478061277316, 0.079736809660354401874, 0.21194907794735623474, -0.059463390366448917734, -0.66773022869416698288, -0.37630298143595286575, 0.94984538436302634601, -0.26430155581438541379, 0.95004651600143685819, -0.37964046599886619449, -0.67849894994930759484, -0.075390824520147392551, 0.20894387502819009028, 0.15092904984642677158, -0.045133254821768373222, 0.064948093004345841539, 0.18209867547849745151, -0.051245636486469153814, -0.58734165457540155053, -0.28776948306894833962, 0.89608003634826338679, -0.42505720977625360835, 0.89692158505975616567, -0.28907591443926311525, -0.59475961109801644966, -0.064983280633335957499, 0.17565721450528543057, 0.12319925102264324412, -0.034533806874880286497, 0.034050173127666519268, -0.014025640422864081233, -0.040342519591677741819, -0.14039088734475124731, -0.023704786278985565812, 0.52868934156462921159, -0.69143394191378326763, 0.53081458798048375414, -0.019726085121775677189, -0.13738927464066125195, -0.043244562348484706649, -0.02875443096130468959, 0.0040061479055799337906, 0.039741556899401356273, 0.039237176193313891281, -0.021425108065791873685, -0.050580247178546533515, -0.12679783024144197778, 0.017852953817421114702, 0.51909042968901686343, -0.75790810193746704915, 0.52104855887191603436, 0.021828138072702861494, -0.12294535242165892386, -0.051281520189897039641, -0.034230284282562001286, 0.0047943039780151269769, 0.040082663550787253626, -0.14066429417865863871, -0.039905179165808116193, 0.20504575907026414971, 0.38078241047841232536, -0.18507277608337033281, -0.20044378539629592595, -0.040675847746948622097, -0.20141376575625913325, -0.18821389675272445774, 0.37468755001106812674, 0.20002085418724663279, -0.031748578329041036994, -0.08971330753304776473, -0.046526029310130105621, 0.044178759317979571231, -0.52895682358695794711, -0.59163461577124354385, 0.58696145249096220997, 3.2731371057104379219, 0.90927272131293068735, -7.4194665366085592595, 0.91324398678110596528, 3.2887632509471571574, 0.62050617050059864166, -0.55107648120247343382, -0.54557588732979100943, -0.18283309795108765305, 0.17183748989993602452, 0.067357711673092651727, -0.04902049802301872844, -0.10812554324225727431, -0.074436678278953755106, -0.025779503401790068717, 0.31663535220440747198, 0.59086486791777825811, -1.4437295762729347448, 0.59114628956443693308, 0.31532317777971952122, -0.029900493351169152995, -0.07670819490999557233, -0.090141773602600283599, 0.0098881762723564189249, -0.04552186884533129807, 0.034980369398802627079, 0.089275413805174452797, 0.069994581676382741642, -0.073003773293254095589, -0.46815385877682863525, 0.11159584923882603924, 0.56017149725340276145, 0.11013359065181459184, -0.46876907514051935344, -0.067248993841688434192, 0.085485185626397452996, 0.097539917548827112515, -0.035814951423921999318, 0.0042398703181007269328, 0.022440242451882627672, 0.016905352129240724252, -0.11419972941112532805, -0.31418236921047226762, 0.56871492086408281796, 0.9623958046765910268, -2.2922319762546137234, 0.96315083508212939734, 0.56761781299066782491, -0.31770403970036509289, -0.11476106930351340218, 0.024749900958585806243, 0.02271840002185922594, 0.1185069739661729904, -0.043640064844045922676, -0.19941880180691728963, -0.087501610776422225624, 0.29275751369705116378, -0.056268006588360099107, -0.097290589689486006719, 0.15515993638114616848, -0.10378526614895687563, -0.068865491685661597843, 0.27710642050039996365, -0.097297739788777232173, -0.1743944501511904277, 0.083018378010027274461, -0.18669630135202600063, 0.19436376187730783349, 0.53393790747856573198, 0.64324606920717541936, 0.022308955003271437045, -2.289709527493363872, -11.301850293219063204, 24.757165720738242953, -11.297880116216729718, -2.26866291450737112, 0.068519163542823655066, 0.68399404406140684731, 0.48755226897301845446, -0.031683463411194226489, 0.046282454377703050041, -0.05737130548700675875, -0.10192261416355992421, -0.11855129255722231252, -0.028380609334470049965, 0.86574811267656970326, 0.75694727450447274464, -2.7238730535152062551, 0.75632358082545969591, 0.86167366712328885558, -0.038612571624409085247, -0.12883125454986149583, -0.087924516420346321421, -0.0048565606890610306737, -0.030632183124039105704, 0.029537349266292197564, 0.11321881224751446926, 0.14209319077146287391, -0.22502997460938817254, -0.68882434897529376538, 0.37534276723294690559, 0.56822000687825902077, 0.3762134843014698915, -0.68628867632699674939, -0.22017130729824957447, 0.14744943400717622217, 0.11038453680512595734, -0.0087552573456367698695, 0.0043413539079967681036, 0.081409546052305561958, 0.072668612364691562155, -0.061798348606699603225, -0.31331665950376025309, -0.2253794427327653338, 0.59322377697858541534, -0.29649727499750699877, 0.59781161464213627976, -0.22185360376052751241, -0.31855726676535761621, -0.076266328922709514249, 0.067264957281119891475, 0.10319098833887195565, 0.042850776477470077508, -0.15258014737735986488, -0.1800100173652215696, 0.049027362770180007989, 0.34981149495200802857, 0.7227852446823708199, 0.37480648353818030172, -2.4095235188855821384, 0.37044653550752809634, 0.71370977904409527159, 0.33859546491372966015, 0.043070683801425591053, -0.16648071025414795665, -0.10287556084904074527, -0.063673713848679244776, -0.068251649772213554557, 0.013991168658481024922, 0.14485381902866778514, 0.12857404806882161141, 0.25859181939689962482, 0.3829147971278569762, -1.596261726695533012, 0.38545776690628041905, 0.26271157899896824262, 0.13242329762578750052, 0.14415757498626460831, -0.00066206027912048617812, -0.12592647755349764549, -0.05328344402001980451, -0.083861758291438159851, -0.01180748782042886566, 0.13883241368211898803, 0.17327711669455475585, 0.32607497084267483833, 0.35876626484685542051, -1.699402330730441335, 0.36089695355796003184, 0.32942051865362831364, 0.17639137605204627257, 0.13847937161198078937, -0.023260923523761706938, -0.13294125632353884225, 0.045095419861354499313, -0.04638917422234744431, -0.11711918711334866339, 0.026401190004700859221, 0.30243563393964895125, -0.15742187841447502494, -0.14545374565202839889, 0.18942591550084914021, -0.14609915139478596435, -0.16128124154649603472, 0.29361015482574365221, 0.015323479602618005069, -0.11335182301725797716, 0.013844439133045104498, 0.033870673782953199249, -0.022498029273299594011, -0.091452362244633586541, 0.010485765750201870433, 0.24094465917745072625, -0.18859347284126665167, -0.10534701176384626831, 0.24837708997118734811, -0.10552272532696641005, -0.19067942373073043516, 0.23527057335598086651, 0.001425493570690073656, -0.091481193541564267968, 0.024896903983885460693, -0.10514425761866932896, 0.1400778825807786343, 0.1244758969582013125, -0.0047262879132433186641, -0.063555226884955115518, -0.53803429275320124869, -0.002617164384335148291, 0.88720753744815905595, 0.0032906130931716196832, -0.52143511457638114237, -0.037636895732719388707, 0.005667111271035673789, 0.079116391596042459633, 0.037459094968576543694, 0.0046824456842107370358, 0.022918089874499360942, 0.0017220218340527134078, -0.026867008845834934594, -0.10025173029919454137, 0.076055847667917356181, 0.51678472285878163106, -0.98051013036285428193, 0.51701959504893146846, 0.074113579739329818463, -0.10674354682797652927, -0.037767693208237231028, -0.003729565492615517612, 0.046649996694364342931, -0.013545960454738954076, 0.017437374088690425072, 0.045446710317805207036, -0.041290528024590572587, -0.2521245701031885611, 0.27025356041372489146, 0.74209755346939587639, -1.537252195836747104, 0.74272135105208647587, 0.27084461832175155127, -0.25180273737684688573, -0.039707299886245356335, 0.049275455823209905748, -0.0021882463299326164116, -0.045133254739975814385, 0.15092904973759879717, 0.20894387520027701832, -0.075390825225974802959, -0.67849894887834727442, -0.37964046668637901893, 0.9500465161199612707, -0.26430155541349553694, 0.9498453834258302475, -0.37630298020198565956, -0.66773022973072770903, -0.05946338992211507718, 0.21194907797375672764, 0.079736809586715140519, -0.034533806769361602429, 0.1231992509122986218, 0.175657214213343571, -0.064983279793649678657, -0.59475961236384722319, -0.28907591316529918091, 0.89692158441987868134, -0.42505721000426127665, 0.89608003723843587807, -0.28776948410759928265, -0.58734165402087346042, -0.051245636547844565512, 0.18209867549115260621, 0.064948092963737380501, 0.039741556882599338163, 0.0040061479896162637687, -0.028754431308189519478, -0.043244561540757682494, -0.13738927561606728567, -0.019726084445470401518, 0.53081458757364052836, -0.6914339419340417292, 0.52868934215239649088, -0.023704787091423699941, -0.1403908866166716507, -0.040342519927931259904, -0.01402564035895723267, 0.034050173090150487321, 0.040082663518200084551, 0.0047943040087709857741, -0.034230284137282851464, -0.051281520694218110656, -0.12294535155639890744, 0.021828137307192956029, 0.52104855895450863379, -0.75790810140693420927, 0.51909042871174881917, 0.017852954987228032702, -0.12679783101042621518, -0.050580246861281083492, -0.021425108122284548806, 0.039237176153117933608, -0.046526029402398229318, -0.08971330731698617944, -0.031748578292741150286, 0.20002085367846331176, 0.37468755058579811257, -0.18821389712647010328, -0.20141376547572942268, -0.040675847732622893993, -0.20044378592849365628, -0.18507277546537903579, 0.3807824101461763644, 0.20504575919644874027, -0.039905179147504021842, -0.14066429422463549992, 0.17183748970565601577, -0.18283309757821103458, -0.5455758877225050929, -0.55107648055684554933, 0.6205061698693569161, 3.2887632511697284521, 0.9132439867278581147, -7.4194665368078904777, 0.90927272187459540209, 3.2731371050404773904, 0.58696145327882187637, -0.59163461645898751673, -0.52895682337947858009, 0.044178759348114951533, 0.063326725353180041145, 0.0081707288966297933475, -0.062012209980904571538, -0.1067132970810218684, -0.071576834620412185806, -0.024705231405576236897, 0.31663535198314718411, 0.591146289505466771, -1.4427990948751612255, 0.59207845747874554565, 0.31666050845855719942, -0.020054167267488074572, -0.059775185189289509191, -0.10637003336637725071, -0.013641731381389908193, -0.024050997389400221443, 0.031980024855843371123, 0.085091196725564088865, 0.068603802705771432802, -0.07292673980599166661, -0.46815385904908934922, 0.11013359066989915347, 0.55750190105450769806, 0.11068105881556843106, -0.45852201514810358995, -0.050216437141406965183, 0.070590412477990721607, 0.054013649319863542286, -3.2574417506288723851e-06, 0.0062679353643279788963, 0.020259746049610092428, 0.016164062640427887679, -0.11361244397147525487, -0.31409056553089409469, 0.56871492083178065702, 0.96315083519410227275, -2.2909158074372091285, 0.96379478969715370873, 0.56388481266524859148, -0.32745315109201889037, -0.11332917999732458614, 0.056698828714450082156, 0.071983223016462996591, 0.036194949316094668368, -0.058913097937604283483, -0.18694102727829650656, -0.072878673064499244205, 0.30045259487662373532, -0.056268007282567374105, -0.10378526555249159569, 0.14380896932193221183, -0.11567467278182783086, -0.07222377331068197126, 0.29368790781317566374, -0.071419780751447517475, -0.11169518107347770064, -0.16103448016502422413, 0.0012166129238517600121, 0.23489065611747572682, 0.50565287919403034689, 0.60752721470144022131, 0.0049868962799017365478, -2.2897095278866950174, -11.297880115851993921, 24.753980706882149576, -11.304514084152021525, -2.2605899949427463724, 0.099809820785482478156, 0.70677275194407307435, 0.40975839277534648053, 0.0431829055379437457, -0.0064942641449294491104, -0.06306014194889938973, -0.094376486975906226373, -0.11135157064395671822, -0.025252603696410583917, 0.86574811272175622445, 0.75632358088989781741, -2.7237428972206494393, 0.75577507572177371475, 0.85848081455987113042, -0.037671354110487685618, -0.11839823489090949005, -0.10167075658304741692, -0.025780203240735106995, -0.0047184576642285884843, 0.0393117031403612191, 0.11036991151382784249, 0.13653903121303548329, -0.22789735256528159679, -0.68882434910589107613, 0.37621348429402590163, 0.56786702541883560258, 0.37326420420001227285, -0.69033508345960425956, -0.21345774072711251601, 0.16555537671471173833, 0.083950985057560642555, -0.047244143633056262521, 0.044395933569927152618, 0.10071872359890803039, 0.073580625468178284532, -0.06888557964200017758, -0.31906829171071116047, -0.22537944269452087065, 0.59781161470102073263, -0.29375326454386280206, 0.58889949271177388646, -0.24757760660244443329, -0.34189926738371617665, -0.055196256722449744958, 0.19663612019517467955, 0.075823462982351388928, -0.051550569766957078655, -0.16561109075066859919, -0.16480803142679173456, 0.065023775101362327189, 0.35749698826886089353, 0.72278524462698268138, 0.37044653567978236275, -2.4145202051020731027, 0.37061271049637140029, 0.72619430377120963715, 0.36512671271821145469, 0.046650886577438446801, -0.30664175719589925606, -0.012674256839689724066, -0.059757127779705251913, -0.061259601667731602981, 0.01641679005482767767, 0.14412160969423121193, 0.12719525798067099487, 0.25859181967164207583, 0.38545776700606437792, -1.5908861202452979633, 0.39320717059479948219, 0.27078568777406747392, 0.13772799310561048558, 0.12143949702600549678, -0.12817832930066241914, 0.00079480979217131145741, -0.062606043772556019267, -0.080218919623344348002, -0.0085646387288902357804, 0.13952653064998021826, 0.1726395430005958076, 0.32607497130203610958, 0.36089695302947122357, -1.6944126780060304593, 0.36904248045239906872, 0.33973764753054069843, 0.18551741785863085088, 0.11638428753450912045, -0.16360476665416467945, 0.027125598891015290093, 0.0054635105637193285763, -0.046983980580325879206, -0.10710801270306674615, 0.032534334367500826468, 0.30386372592074795129, -0.15742187830323844944, -0.14609915173164189639, 0.18597106256380230294, -0.15282164425590408219, -0.16532574940277594555, 0.3024862040733767854, 0.022115667295700545958, -0.10487744270655319223, 0.014320193493970026499, 0.012086374901403086277, -0.022479756329992023334, -0.085344577944971605765, 0.014002802529292380715, 0.24158950407850476871, -0.18859347264197462435, -0.10552272537716814171, 0.2464460079169707718, -0.10997092705854440153, -0.19488272687977628572, 0.23769892205158457688, 0.0017215725321144393061, -0.061759940845221704409, -0.10401201144135618915, 0.032642047295129907047, 0.14772705794655890954, 0.097735340138054030423, -0.024964957837128355644, -0.07135613147009767232, -0.53803429292929638716, 0.0032906126704627214168, 0.9015547478996892572, 0.023559938956896120388, -0.51808039529110982979, -0.084607022607272278081, -0.072626259142709329031, 0.21070093658560071215, -0.024286395876353967377, 0.018170781725715099558, 0.032520570398862150285, 0.0064654111167398199689, -0.025319326245071008263, -0.10029203496957593555, 0.076055847757663760933, 0.51701959494227645031, -0.98173801737949617507, 0.51153021231565531934, 0.061190834868940696678, -0.12089959469360379851, -0.031052807551317299084, 0.063277367951180399364, -0.0073048211031220330186, -0.0043626251093194269581, 0.018635867263375495428, 0.043211346012257463578, -0.0426638284351099073, -0.25253563101476261821, 0.27025356051771814991, 0.74272135113756798663, -1.535561624952489268, 0.74446138933406436067, 0.26884105307746869462, -0.2562585645384500288, -0.038330567389438026382, 0.049112507695820546705, -0.057324304212249724277, 0.034307727084157392528, 0.15442396802981728232, 0.1920694271088688454, -0.085998158254209672302, -0.68101442475129603249, -0.37964046605522355771, 0.94984538394174844456, -0.26318484147224163738, 0.95206938026354481774, -0.38190835784820575949, -0.68490422270676820649, -0.058415542910682349831, 0.31182914947226020308, -0.043560141822408700307, 0.025835440734635822108, 0.1254136094397692025, 0.1629574437052300484, -0.072377558396497537752, -0.59613403358371608309, -0.2890759137866582007, 0.89608003695607196804, -0.42546526220740493729, 0.89654547079265622944, -0.29237249153721184269, -0.59841169833239438169, -0.04649831048125047106, 0.25881713102070719179, 0.028343677299342933534, 0.015224027046275219119, -0.0045741210358583384368, -0.028564610037772462936, -0.040056155013494888772, -0.13522221133062903298, -0.019726084632568550992, 0.52868934169092607345, -0.69495068442422425203, 0.52597049240497495415, -0.023664466477277899814, -0.13360519276982529835, -0.024967996045564929547, 0.0046452380374913341404, 0.024669571964836687489, 0.017546634735029060376, -0.0023253674530476532711, -0.033317645790472305567, -0.047919323178073097713, -0.12082086017215351581, 0.021828137665997570521, 0.51909042942642380503, -0.76122958206898361233, 0.51601986480287775905, 0.016201993273733630851, -0.12297867811558862994, -0.036146661345514273511, 0.0072853612918282904462, 0.013853104425800387769, -0.064482475875532269116, -0.090778298774578422559, -0.029525081388242644526, 0.20069222600816821744, 0.37440649235211120871, -0.18821389663778234635, -0.20044378597699902267, -0.037808149563190104891, -0.19296008707599346388, -0.1684988592005168262, 0.40387463019824021426, 0.17927038329669947592, -0.19944094425483188671, 0.12982165339678503058, 0.01969964174636705187, -0.21632628875076356789, -0.52202340513607659034, -0.52077485981306281815, 0.63482114643646891938, 3.2887632511575102257, 0.90927272136144621228, -7.4191247345538879898, 0.9157964840033351317, 3.2734383327979403866, 0.55553026074043265758, -0.63263828185313464303, -0.42502576083165244381, 0.034748222659408599389, 0.03475484276966160635, -0.0016311528233110402994, -0.063637351660757970895, -0.10620996486468770037, -0.071576834558365790007, -0.025779503584167369479, 0.31532317811218468773, 0.59207845750706655785, -1.4348666234726690227, 0.61569605824978690389, 0.3438838128688586071, -0.034649479094492563469, -0.20144930389013582239, 0.0059965266580779092928, -0.02300989288364926888, -0.022584224548094829904, 0.033879024407764704585, 0.085504959788650455876, 0.068603802636597738251, -0.073003773073925998238, -0.46876907495845243634, 0.11068105857937111935, 0.56272105463175914331, 0.11502089351202951506, -0.48690008517184629744, -0.074061366249610932622, 0.16653397067845576718, -0.0010259776196143907939, 0.001192527981114720366, 0.0070485877538439224363, 0.019803596148080582218, 0.015442607936702733296, -0.11361244382122923091, -0.31418236942191440919, 0.56761781322806181826, 0.96379478933795115214, -2.2848896354959817856, 0.96981754136874942507, 0.54519186131092789438, -0.37656455888044509939, -0.00062485282287961325855, 0.014923338920506902874, 0.049908597867444247365, 0.035581079940450781318, -0.050342494133428110359, -0.17990841325856038457, -0.072878673767109353188, 0.29275751475372047139, -0.068865492366525282719, -0.11567467241881843054, 0.14142627859192810624, -0.093293902444214454883, -0.017866754633632405541, 0.36682118254556250747, -0.3067934356314459543, -0.073365196322504905302, -0.077425677632146749918, 0.017170030765815361834, 0.22189753264896106, 0.49109244955034281332, 0.60752721457108904701, 0.022308955330696721847, -2.2686629147048456012, -11.304514084113295169, 24.683121228889547183, -11.420238084655359501, -2.2878276103159693733, 0.29186058088009408928, 1.0960350273746071181, 0.018120277756955720933, 0.022497177437013717027, -0.0095831911142344582827, -0.060274929980401299734, -0.091572419234840662927, -0.11135157071076257074, -0.028380609291482693235, 0.86167366698734881858, 0.75577507576940128331, -2.7141840040405682544, 0.78435112070878720836, 0.88456503238857187021, -0.088269519715214697642, -0.22329119844518832694, -0.014732635336073553536, -0.011305076298057496059, 0.00029094868651027020129, 0.037835469132212244669, 0.10781071832395368315, 0.13653903119183116721, -0.22502997440278107999, -0.68628867633328083375, 0.37326420412686356354, 0.55670710707963622088, 0.35900133253951321821, -0.70328040624061483044, -0.18884616054941699703, 0.25782736573003678382, -0.050831048702401765427, 0.0051512271951279931664, 0.058133489757322305558, 0.095431969243860673147, 0.066777085878275965758, -0.068885579875090835467, -0.31331665907504635538, -0.22185360441472440352, 0.58889949294424581705, -0.31957921218585016998, 0.55764580799110796683, -0.26813258031386033586, -0.29698791350573228165, 0.16485916016978047205, 0.046494484588059888097, 0.018377630247817407921, -0.058827923096841282236, -0.15470129984168115578, -0.15586146080324542185, 0.065023774884288551568, 0.34981149496405772314, 0.71370977864673734725, 0.3706127107095151807, -2.3926027384076244964, 0.42424147637198067473, 0.78634671814685663804, 0.32485218162151768828, -0.33332681859124024815, 0.010821622470876802319, -0.026722881848246705927, -0.061766347293920401129, -0.060478121572610663015, 0.016616257055732285702, 0.14412160948316132814, 0.12857404855238460661, 0.26271157835040259698, 0.39320717098387303334, -1.5795375778174842729, 0.40774885666188553923, 0.26950494530728369424, 0.090675627428746102821, 0.011326057390240848802, 0.019554725805299356944, -0.021869914528397347425, -0.066736407592290269974, -0.078868209501312430687, -0.0075850846654100696764, 0.13952653090549202441, 0.17327711601637885153, 0.32942051937801836159, 0.36904247992085226215, -1.6800344787903975785, 0.39011314718466749696, 0.34594306060665458435, 0.13395614913646966548, -0.039069267374008641303, 0.0021111311468773818166, 0.020564963812190639736, 0.0077316558188387490141, -0.043374530779392392321, -0.10538046929680364161, 0.032534334571398149105, 0.30243563367809039866, -0.16128124104370472613, -0.15282164443855744684, 0.1806586831288956585, -0.14288027773602685055, -0.13164168909021048415, 0.35776971228467135955, -0.16652490474597461678, -0.0026359996929173831748, 0.015176650819417039293, 0.013733128257965835153, -0.020889563325824343082, -0.084622325236889214817, 0.014002802825771095557, 0.24094465886114205255, -0.19067942359463582735, -0.1099709271372077829, 0.24081769808650574372, -0.1079588524734221705, -0.17565341647181115237, 0.29251740284594324182, -0.12521756121436458664, -0.033710661735860139587, -0.049790014340053381781, 0.026674578937912713644, 0.1318717413894243129, 0.089292395511881006431, -0.024964957845231595945, -0.063555226742997614209, -0.52143511367858952354, 0.023559938394471090228, 0.89996323091609187195, -0.037326700858345190914, -0.62344498787379898364, -0.040393867183465818893, 0.2240743312360309325, -0.029474564894212032518, 0.00063544357627566983721, 0.026952397669381442102, 0.034147144536201490261, 0.0060863462261471993481, -0.025319326330273787123, -0.10025173015226709894, 0.074113579420857292335, 0.51153021260389019975, -0.99178114425580687907, 0.500253220496220119, 0.055332696328582689493, -0.092021253565532951102, 0.030013273166321546848, -0.001358110201803489547, -0.0054094140126618875733, -0.0039420219905316293552, 0.017724525499108273463, 0.042453530088453414471, -0.04266382829446761088, -0.25212457030324542107, 0.27084461813789456608, 0.74446138934668526499, -1.5321678038096877561, 0.74718057043704944054, 0.2516940102427687731, -0.29772607005922407941, 0.060955432727426532569, -0.028472233897834549332, -0.022147644651410124811, 0.038206378827256325115, 0.14851238318687273732, 0.18798364327551059239, -0.085998157653046830573, -0.67849895014184347097, -0.37630298072308948631, 0.9520693805155946432, -0.26991495455322073171, 0.91953208494651450522, -0.45677846984040493261, -0.70628018862639108111, 0.37381501791364063481, -0.022425367523069176384, -0.017434091375729130619, 0.029967004187813783833, 0.1218891816510861148, 0.16026880066189502783, -0.072377558803444375091, -0.59475961148092004738, -0.28776948376142930108, 0.89654547085409996843, -0.43039365278067492904, 0.87347782515393057601, -0.35252048986663975239, -0.62266305393268717161, 0.3145531721352699539, 0.011355775155034011914, 0.017400340962987238813, 0.012915756713113433346, -0.0030420362472777795482, -0.027038201516586608114, -0.040056155006242231964, -0.13738927546710230643, -0.023704786276226550573, 0.52597049210580448797, -0.6926913826385987516, 0.53445639504407960185, -0.018504050168030906215, -0.12829910152971168991, -0.034427215693539472974, 0.0073111057469878127854, 0.017388245851988594348, 0.016192132830890715312, -0.00074923391416381775187, -0.031852746439440228488, -0.047919323129202939793, -0.12294535195270792083, 0.017852953907850341309, 0.51601986545725153643, -0.76018846611316859274, 0.52274003164916482689, 0.021539469266737734654, -0.11388186944171449499, -0.044353930029780740119, 0.030555171931272110536, -0.018784014662077113539, -0.071210195152255545414, -0.089130604045476777286, -0.027872831295898684717, 0.20069222652308366106, 0.37468754969472012339, -0.18507277532509036622, -0.19296008724376104149, -0.024431485016901677287, -0.17048021694314208219, -0.15421983464493377314, 0.38425808271482980727, -0.050182907004840217324, 0.054627543550547573437, 0.070199464220213447785, 0.0030690050794569298347, -0.20585910767561083068, -0.50913716394979924296, -0.52077485986994909162, 0.62050617010230113557, 3.2731371057231641863, 0.9157964837111641776, -7.3753434480075830137, 0.99197162606576849697, 3.3466272302856476806, 0.46966015854806286711, -1.1329583074041116308, 0.010009564946291087464, 0.028712085065487984864, 0.029671832133700039302, -0.0036165958686790421195, -0.063637351783817144857, -0.10671329676743021575, -0.074436678892544960728, -0.029900492776353967961, 0.31666050820145069666, 0.6156960577808705537, -1.3876400209184445167, 0.64888014007965033603, 0.24233728112429486767, -0.2234852171520405606, 0.0075100620699195853869, -0.0047337070752883235111, -0.022350097543569150821, -0.020918511605901711492, 0.033879024782321694176, 0.085091196492340592794, 0.069994581677909700757, -0.067248994076021223476, -0.45852201495841810042, 0.11502089355078275179, 0.52523719387585865448, 0.047244374627555152379, -0.47132723886632660326, 0.15941423464998619175, -0.0013697548563761082017, 0.0004862524342835838724, 0.0017901055117194885659, 0.0065329910473290106476, 0.019803596715051777905, 0.016164062009128638692, -0.11419972877498739405, -0.31770404020042880555, 0.56388481309681137699, 0.96981754127456143433, -2.2688368157146512338, 0.97810671273356042388, 0.48794424776265338251, -0.34307307228845573821, -0.01952075704330022074, 0.030390412345536237104, 0.058869792975501343546, 0.040280965020855555303, -0.050342494947984893494, -0.18694102654576824629, -0.087501610691267536768, 0.27710641990925943823, -0.072223773648112554735, -0.093293901493149547499, 0.19061724619071893083, -0.0060380638940022123007, -0.0056261488904689443274, -0.07654420955848219299, -0.0029335906409904762178, -0.072460543203436533766, -0.080107436977355020447, 0.012600250198609403568, 0.22189753279828017263, 0.50565287930545765871, 0.64324606916291260372, 0.068519163368996827135, -2.2605899950813919119, -11.420238084352087427, 24.40199897373158322, -11.637273448259488617, -1.9244733897966437297, 1.5199230572788209326, -0.00030126324568391013841, 0.019740718712470850527, 0.022237213270322141873, -0.0091261027137383016195, -0.060274930113029652479, -0.094376486714019811042, -0.11855129278743574417, -0.038612571236517033435, 0.85848081442936408081, 0.78435112059155909225, -2.6523506507566327883, 0.82511957900903143415, 0.79561328897916516745, -0.32702736697720313774, -0.0038923999131851303238, -0.01090321210707051805, -0.011027146454717388566, -0.00048609610788661385692, 0.037835469155532285013, 0.11036991147243564726, 0.14209319067308509577, -0.22017130725001968217, -0.69033508328140757992, 0.35900133241106452209, 0.53442490290280864063, 0.32256341779703240347, -0.76740578813677318148, 0.19324503809526460962, -0.027378040341776752054, -0.020292432449852109561, 0.0090441516046045489913, 0.054097628848502957166, 0.095431968830914384871, 0.07358062582399130247, -0.061798348912633332963, -0.31855726624191088581, -0.24757760677561907459, 0.55764580809666663974, -0.33253973560825195177, 0.58301313791236109019, -0.22558876271663383384, -0.14755073821997499728, 0.010065181260859409895, 0.034717292997510328101, 0.018970559281796955248, -0.054544006704983639955, -0.15470130026159459313, -0.16480803099620131014, 0.049027362736077842453, 0.33859546523109385063, 0.72619430357171488311, 0.42424147642325360508, -2.3122367739985070223, 0.466405744775218567, 0.62789894261730050395, 0.0026860407866780996143, 0.016579047097355387175, -0.0039798415533524242327, -0.032275597749283957416, -0.062733068023651278322, -0.060478122122699339014, 0.016416790712052833912, 0.14485381813041009469, 0.13242329858388901531, 0.2707856872951735494, 0.40774885681429201556, -1.5711286891118010356, 0.3847265914088532357, 0.1803816059255195503, 0.18542834687707784869, 0.019395761241193999747, 0.0020844699630778893915, -0.028117787935264548088, -0.067549957070753535726, -0.078868209601506436557, -0.0085646392944748565845, 0.13883241463704715235, 0.17639137517828196167, 0.33973764845448028993, 0.39011314675044350686, -1.6631050552991926139, 0.37082254638765738131, 0.24468601309587778569, 0.17406641207515194392, -0.010006753167944457128, 0.012884873374079472072, 0.023558401057812650148, 0.0080181285475477408892, -0.043374530218590619246, -0.10710801325317352573, 0.026401190099439357517, 0.29361015462095568784, -0.16532574926588772257, -0.14288027774972730799, 0.20424137883394360982, -0.091617486680790935494, -0.12117471997844096732, 0.11493684106528105626, -0.0083874541320922407078, 0.0065325809202109660911, 0.017354141630529361373, 0.013634987714754919635, -0.020889562912621860435, -0.085344578351673791583, 0.01048576591728889694, 0.23527057331803313822, -0.1948827268106674282, -0.10795885254884982074, 0.25057268986850156356, -0.06197731668582137643, -0.14295758634221011896, 0.089461036893066406273, 0.019133053924132184531, -0.050332619211603639442, -0.058919026527407283855, 0.022002592854803590394, 0.13187174230731241176, 0.097735339736420068224, -0.0047262881835537177283, -0.037636896117789406246, -0.51808039501152602924, -0.037326700894843974077, 0.78064726731572608998, -0.053461884217553115595, -0.41104491041987856192, 0.1160259191732501971, -0.019269385142056366655, -0.011192344359924596975, 0.0065987104874692826401, 0.027546902896817427459, 0.034147144203259718886, 0.0064654110884465772202, -0.026867008698342968037, -0.10674354683803542743, 0.061190834886582175234, 0.50025322074870437472, -0.9867317505522383847, 0.53926314639855676969, 0.060675352496347499665, -0.087672111456943535579, 0.0017782549022743864443, -0.0029345850159417908275, -0.0060786562333684478199, -0.0042531971080417025483, 0.017724525532327454341, 0.043211345882478581493, -0.041290527836062775047, -0.25180273715690038028, 0.26884105313476447385, 0.74718057037075558124, -1.5269039095373355686, 0.73374026843830486566, 0.19875651619927187097, -0.17852748229817175241, -0.0040540040262882076627, -0.027414268857936276158, -0.020348293044354176518, 0.038423988638011830066, 0.14851238272326794387, 0.19206942675195307069, -0.075390823864171924162, -0.66773022985167429511, -0.38190835764441621869, 0.91953208451816781377, -0.32987413522626890527, 0.86035218627050791262, -0.43473221077451901229, -0.22937506199575674604, -0.0049241379743531115201, -0.020691876449991764009, -0.014781622035126527892, 0.030810265756387650304, 0.12188918151215150099, 0.16295744422941185747, -0.06498328051218360546, -0.58734165398068005626, -0.29237249195151882031, 0.8734778253825823402, -0.47525549408389156092, 0.82146417523822357065, -0.34966925923291475309, -0.21061706367339122492, -0.0019688057097074660937, 0.012164842468374107223, 0.019144352075202115265, 0.014013712436162294867, -0.0030420359271700527642, -0.028564610141212642969, -0.043244561660717287244, -0.14039088721431844387, -0.023664466532203169008, 0.53445639509778686271, -0.67655851888240536862, 0.54978296702719542388, -0.07467910647212405717, -0.13954972509665872216, -0.0042092429400309230858, 0.010436567699620586183, 0.019685697772405495731, 0.017277779039864452071, -0.00074923453457801172939, -0.033317645084112841591, -0.051281520907046525271, -0.1267978300658614832, 0.016201992655177126224, 0.52274003158873139085, -0.74328953460775060158, 0.54509540269047462857, -0.024909227677192009864, -0.1489042239088684827, 0.024135404978419108812, 0.0080254498194105190756, -0.025960417114377527803, -0.071596373532853682664, -0.089130603674414360382, -0.029525081870990690924, 0.2000208540658534595, 0.38078241021056224991, -0.16849885887916671301, -0.17048021743012178875, -0.023862089380449796189, -0.20394883800811036689, -0.21215561686499087579, 0.39178145993746288367, 0.0006355224532535927711, 0.055523425088944394767, 0.073049794572644086155, 0.007044303182668419476, -0.20585910697541937031, -0.52202340559578375156, -0.55107648083261251415, 0.58696145264610188885, 3.27343833303938192, 0.99197162634921798308, -7.2085752285327613365, 1.2341086543224970562, 3.3470839725342149684, -1.0603454201865634143, -0.0044478348318863917057, 0.012335717167990825627, 0.027674361187059310502, 0.029671831863144849939, -0.0016311528213697860674, -0.062012209876385893104, -0.10812554304644272918, -0.076708195429736364401, -0.020054167076750083198, 0.34388381352813190972, 0.64888014007429950514, -1.428009095910722337, 0.61708914128505798491, 0.032260793116416650039, 0.0030223875520458054131, 0.0039637823315748889688, -0.0050282030013021651593, -0.022350097522078792717, -0.022584224884621938167, 0.031980024985978475205, 0.089275413814497911469, 0.08548518587113058953, -0.050216437378414510717, -0.48690008514710342302, 0.047244374536455094327, 0.53563720151831484273, 0.052131819201193439761, -0.26607832589751906349, -0.00085067175832428032402, -0.00036239517485268729017, 0.00025849829928751769234, 0.0017901054667717279916, 0.0070485872803086480759, 0.0202597465480293705, 0.016905351525550407765, -0.11476106864891673365, -0.32745315173803263287, 0.5451918617679548662, 0.97810671236076607293, -2.2314512672540067761, 1.2891730430581029854, -0.1811679216872314202, -0.027202391328937421816, 0.0061979894329851408993, 0.039606868227656084891, 0.058869791886370842482, 0.035581081302289450929, -0.058913098015549045239, -0.19941880361274103017, -0.097297737827961083834, 0.29368790784702097918, -0.017866756203031863154, -0.0060380626312506267134, 0.20199184895700245113, -0.41931420663830171192, 0.1949732724808217188, 0.0256341735331428143, -0.034422792226201828769, -0.081722754585071541444, -0.080107437125085667606, 0.01717003076356608815, 0.23489065615146442756, 0.53393790732818113831, 0.68399404451105083602, 0.099809820770549229074, -2.2878276108310720005, -11.637273447889317168, 23.825545183909827784, -11.497485299157634842, 0.14927643157301112575, -0.0067459420758259102183, 0.0075336409508406591146, 0.021385500619137080142, 0.022237213093349611492, -0.0095831907909642907034, -0.063060142365632135264, -0.101922613834685441, -0.12883125508166334838, -0.037671353944577143369, 0.88456503263009067517, 0.82511957879072894695, -2.6601856610573344497, 0.97849744739919686243, 0.27868790425411860134, 0.0027363865990053031982, -0.0068358218178014339486, -0.012198614523136910512, -0.011027146556563074964, 0.00029094877848287865841, 0.039311703192020153175, 0.1132188121148906268, 0.14744943393747733662, -0.21345774091346095269, -0.70328040596696450582, 0.32256341758165651967, 0.48478693431715058626, 0.35007916695765967097, -0.52195596695704560286, -0.0021851660742953256389, -0.023631363708337810997, -0.021749158599990613822, 0.0090441514085581667381, 0.058133490299152767999, 0.10071872323277490657, 0.07266861255065742986, -0.076266329121584985895, -0.34189926724871144703, -0.26813258052960259326, 0.58301313801803311687, -0.23868418524459286689, 0.68692858188025218702, -0.54720362843893988192, -0.0093741710534457553528, 0.02254178598400984751, 0.03895461794114160603, 0.018970559200813481848, -0.058827922600002864606, -0.16561109131928875016, -0.18001001694353196969, 0.043070683453060254475, 0.36512671287792564101, 0.78634671810391365554, 0.46640574466757428462, -2.4357354243653479564, 0.37488205803940005678, 0.74774406527175196757, 0.013785444202626317631, 0.0063510132001928787382, -0.0080894415233747490879, -0.032275597273694611733, -0.061766346869572705258, -0.06125960259569247085, 0.013991169974198806175, 0.14415757348355534573, 0.13772799373714048055, 0.26950494517515494186, 0.38472659118325341909, -1.6402374138712245699, 0.48072806288110309092, 0.35609093520724838244, 0.012709028576676571765, 0.009955088840428626823, -0.0018644100669096908343, -0.028117787387890471468, -0.066736407966360891675, -0.080218918990505011513, -0.011807488410967036108, 0.1384793719775609111, 0.18551741719531897323, 0.34594306115927619949, 0.3708225459619482467, -1.7448358161978803071, 0.4406646619712271673, 0.43514684661914460051, -0.0091043460801687042805, 0.0012266019640756625243, 0.015080567110519922658, 0.023558401436727965939, 0.0077316549621794204705, -0.046983979989391029541, -0.1171191869560399279, 0.015323479540707198726, 0.30248620401635128996, -0.13164168901570225123, -0.091617486729256875089, 0.19800080018622473466, -0.41051453100148799225, 0.24631821734081696396, -0.0060320103462850223341, -0.001101300000871506795, 0.0079994149445774612173, 0.017354141850013805881, 0.01373312780968311865, -0.02247975608944299819, -0.091452362180098195621, 0.0014254936107394042613, 0.23769892184420093884, -0.17565341612736318222, -0.061977316953688021961, 0.27027142589712643828, -0.3419444873955354014, 0.1535120015494547685, 0.0308358606995395787, -0.015703773246770337535, -0.058700074788466727327, -0.058919025646193071144, 0.026674577576209484303, 0.14772705843245390045, 0.12447589761433662525, 0.0056671107338019544725, -0.084607022391161634811, -0.62344498790524427445, -0.053461884420325761491, 0.9795187839346797265, -0.08675466837786592611, -0.34201089186696292943, -0.0052520475286488869579, -0.011598928034397189626, -0.0088098924529788728843, 0.0065987105302384588615, 0.026952398004602279302, 0.032520570359051613474, 0.0017220219684567247657, -0.037767693151723617795, -0.12089959481263977403, 0.055332696048161662816, 0.53926314643549588812, -0.93686221429439442421, 0.58736763173353534473, -0.13245192310909917333, 0.0023665133863001405236, -0.00056775717607844946103, -0.0036972621679729317755, -0.0060786563761770131403, -0.0039420218851164935334, 0.018635867267275264825, 0.045446710149446421789, -0.039707300041038624816, -0.25625856466944257894, 0.25169401046672695843, 0.73374026827602412126, -1.5129180818781584605, 0.97282769500022137077, -0.20100474591299988258, 0.0047265067911093968517, -0.013215633672076287919, -0.027789025097028952765, -0.020348293724311818792, 0.038206379754283890937, 0.15442396810714742461, 0.20894387402068706483, -0.059463389113031416, -0.68490422341466894451, -0.45677846921621911847, 0.86035218604077057236, -0.22803440264325061659, 1.2326639296724510864, -1.0203464359724243593, 0.0017700395579793195937, -0.010099177183768310445, -0.020176069722389739641, -0.01478162183400916381, 0.029967004065023190168, 0.12541360934960443796, 0.17565721444847626209, -0.051245636993135258719, -0.59841169747120204203, -0.35252049068436625001, 0.82146417580608910303, -0.39669655380685842072, 1.1557506187913062501, -0.87560778716289211498, -0.0081168377098093918293, 0.0042778990961733834308, 0.014440985713102533552, 0.019144352214604660328, 0.012915756054681298143, -0.0045741205510665700756, -0.02875443174659904691, -0.040342519412323751826, -0.13360519229581541145, -0.018504050777399035088, 0.54978296771396084619, -0.68376482880853539648, 0.54978520282798337693, -0.23062190664897308512, -0.0083311099894347873107, 0.002684337308897625108, 0.012914042465443309471, 0.019685696977200399516, 0.01619213404256966693, -0.0023253685862452654065, -0.034230283824796849113, -0.050580247106302336779, -0.12297867802323413089, 0.021539469510215011189, 0.54509540271083012364, -0.74081997782098152339, 0.5430370966651387743, -0.19999909526044254426, 0.011857913642535765353, 0.014307705894235633881, 0.0042180635793808247611, -0.025960416795488931718, -0.071210195570773945906, -0.090778298852449645917, -0.03174857761486268859, 0.20504575861336854503, 0.40387462990466077573, -0.15421983371844552591, -0.20394883873177782618, -0.11955975699229391684, -0.39701939512487377515, 0.46013592859597507445, -0.021125919255426842819, 0.026222425867090448648, 0.064117781703610571276, 0.073049795224725905252, 0.0030690039293550289666, -0.21632628809845064199, -0.54557588754321884039, -0.59163461608673495462, 0.5555302607878828125, 3.346627229902964018, 1.2341086545107886607, -7.0601710986822485694, 1.2876168981689317672, 1.8839922566376161583, -0.011456632599320056703, -0.0018591352204074325805, 0.01233571830626230964, 0.028712084668664563519, 0.034754843108377862781, 0.0081707281953511245315, -0.049020497648551271397, -0.09014177295124198297, -0.059775185845485412361, -0.034649479312820662125, 0.24233728109456842414, 0.61708914131479231102, -1.4238841893932607796, 0.74432871237689846033, -0.00035083946670383560481, 0.0034715402555487188604, 0.003963782065558411026, -0.0047337067434826060608, -0.023009892995680038957, -0.024050997251102865238, 0.034980369358949839342, 0.097539917349386537349, 0.070590412769077653943, -0.074061366521713911659, -0.47132723848896013186, 0.052131818899824523827, 0.82481288416000364538, -0.49195740470146759327, -6.921963279029051978e-05, -0.0010312551371634522731, -0.00036239553721290382977, 0.00048625287738975699102, 0.0011925278769931034386, 0.0062679353697599237208, 0.022440242692808823333, 0.024749900398379584554, -0.11332917927474814301, -0.37656455961468543459, 0.48794424848071654699, 1.2891730426518497321, -2.3945592466276282018, 1.0587334928872960127, -0.016317221426183733946, -0.0098541289295615806798, 0.0061979888163067268364, 0.030390413614559318167, 0.049908596718909153944, 0.036194949136099247022, -0.043640063152163122151, -0.17439445190874194869, -0.071419780512558164731, 0.36682118368536814623, -0.0056261500882642499766, -0.41931420611317454794, 0.20910447623930580874, 0.051659763324218878344, 0.020979387255615589841, 5.0923465589652620249e-05, -0.034422792432476798097, -0.072460542708206829299, -0.077425677924331068325, 0.0012166128042262775322, 0.19436376205681424323, 0.48755226853545607923, 0.70677275227138336433, 0.291860581236711214, -1.9244733901922623787, -11.49748529922321616, 21.838228998147375393, -9.9667728108728521619, -0.005562770023506632272, -0.0013453761019303425456, 0.0075336408717956342487, 0.019740718917160345047, 0.022497177089686147544, -0.0064942637895155294614, -0.057371305736919231411, -0.08792451612094363822, -0.11839823488763909221, -0.088269519912788141358, 0.79561328912032558414, 0.97849744730216126065, -2.9022793334478396154, 1.4493484758022476822, 0.004198849439248771194, -0.0011187517762885660581, -0.0068358219497312751323, -0.010903212014132818847, -0.011305076395516065052, -0.004718457632681218733, 0.029537349440899560854, 0.11038453678664450486, 0.16555537691601227013, -0.18884616087699199194, -0.76740578785873958534, 0.35007916682227574467, 0.98471797414875938959, -0.65568196061586403456, 0.011194984833651775921, -0.010051919422300849977, -0.023631363858516711562, -0.020292432236703822346, 0.0051512268798754841728, 0.044395933867642227433, 0.081409545775258163047, 0.06726495736725014718, -0.055196256841360820389, -0.29698791332283414057, -0.22558876282887010967, 0.68692858206697160472, -0.17171077958735961322, -0.093255878831707758603, -0.010505241291330355904, 0.0043002236864728860671, 0.022541786032343216067, 0.034717293075467628438, 0.018377629922275023205, -0.051550569361862273732, -0.15258014784133205888, -0.1664807100626922165, 0.046650886561338970882, 0.3248521816507388138, 0.62789894268814050449, 0.37488205797329160474, -2.387173838312695473, 1.312469417115845749, 0.0074694805069815215642, 0.0089485631610049885115, 0.0063510134489103958061, -0.0039798420840328613268, -0.026722881782430839281, -0.059757127339967838764, -0.068251650712590039571, -0.00066205922227925359524, 0.12143949651615193464, 0.090675627633731270549, 0.1803816059980742903, 0.4807280629105126768, -1.4019793755065030894, 0.65208530126686592432, 0.0048929333592416550044, 0.0092418874604380842064, 0.0099550891001305048811, 0.0020844692880284563727, -0.02186991396788985284, -0.062606044189688916934, -0.083861758082927492786, -0.023260923593131175441, 0.11638428784632784407, 0.13395614880253725598, 0.2446860133434259621, 0.44066466203065546292, -1.5262222503941842433, 0.74449369949106036781, -0.0022986774559831330347, -0.004225588819866801793, 0.0012266020161143175934, 0.012884872981764687247, 0.020564964420468789807, 0.0054635103299675909921, -0.046389174413858681734, -0.11335182286016130815, 0.022115667234671725072, 0.35776971214087399664, -0.12117471983242480982, -0.41051453105859625481, 0.35340190773495333554, -0.075435469368563434767, -0.00089391144565190692496, -0.003416880970649294625, -0.0011012999956742936924, 0.0065325807556542827073, 0.015176651040623916761, 0.012086374869766382603, -0.022498029343763922827, -0.091481193609478414497, 0.0017215727407690866856, 0.29251740241603124693, -0.14295758596085222925, -0.34194448753103917626, 0.40403999682671437599, -0.12757299797981430656, 0.015939335358232203238, 0.0098560129802968271528, -0.015703773059034430853, -0.050332619752479768049, -0.049790013612113145913, 0.032642047030197074431, 0.14007788199021950493, 0.079116392176100544531, -0.072626259259939707214, -0.040393867245086791107, -0.41104491012431548791, -0.086754668698718187536, 1.1044676878744541515, -0.66202308170422363709, 0.0062669658312582690249, -0.0044731621816925378962, -0.011598928069189266832, -0.01119234436236820826, 0.00063544339514565325276, 0.018170781877903333795, 0.022918089656116764213, -0.0037295655930626022898, -0.031052807497467957432, -0.092021253360453358905, 0.060675352396005702293, 0.58736763185578144686, -0.88960766637135058055, 0.3428214099251085667, 0.001332674421398110905, 0.0008795917731117628563, -0.00056775739805540884411, -0.0029345848332371594555, -0.0054094141264507756076, -0.0043626249938009250759, 0.017437374133340496907, 0.049275455898297314261, -0.038330567288754772981, -0.29772607032687581352, 0.19875651647506009656, 0.97282769488970743943, -1.516448614610149237, 0.62729268136043403459, 0.0034879272999651799066, -0.0010286842040630706733, -0.013215633727901694716, -0.027414268396648407145, -0.02214764528932476606, 0.034307727032933721389, 0.15092905049618807189, 0.21194907724558140205, -0.058415542348704704856, -0.70628018880077270136, -0.43473221083674895571, 1.2326639296767387677, -0.1054795684440946596, -0.25821309805069542209, 0.002009220952577286945, -0.0013117336437814225251, -0.010099176974432024292, -0.020691876823936670299, -0.017434090961069459436, 0.025835440384692848204, 0.12319925100886548741, 0.18209867593212403225, -0.04649831131949189178, -0.6226630529148524662, -0.34966926005508647268, 1.1557506190732540485, -0.28053812628473917545, -0.13422367187425246482, -0.0078261344466950596294, -0.0030367989004188118349, 0.0042778991649333377584, 0.012164842046695814246, 0.017400341666376158839, 0.015224026544260062244, 0.0040061484781225625493, -0.014025640547936094743, -0.024967996592459157368, -0.12829910079045855897, -0.074679107312019515907, 0.54978520324358781668, -0.58188066718124764609, 0.2417367088156587851, -0.0066547795440590334914, -0.00339122362030967869, 0.0026843370959975096021, 0.010436568279537719595, 0.017388245016187447306, 0.017546635472174373915, 0.0047943038758950021205, -0.021425108160559622889, -0.03614666129887378343, -0.11388186956598321531, -0.024909227756654231389, 0.54303709678118083914, -0.66853430651309408983, 0.28784127789986263535, 0.00097751621377860408238, 0.0097747918927178767401, 0.014307705930153664584, 0.0080254495142627977483, -0.018784014426462096459, -0.064482475546566012103, -0.089713308182504516997, -0.039905178603521426761, 0.17927038353020938177, 0.38425808195221150543, -0.21215561617022843377, -0.39701939542602954214, 0.27779830805411326322, -0.061834285571374172918, -0.016709753851235006894, 0.00019676553869149546769, 0.026222425839856507851, 0.055523424494740243607, 0.070199464992762458659, 0.019699641563469188349, -0.18283309775668907249, -0.52895682346776939919, -0.63263828204591177062, 0.46966015854044101951, 3.3470839726713537132, 1.2876168982574367483, -8.5483991497843057061, 4.6459224056975685357, -0.014243498068926423866, -0.011456631890721808917, -0.0044478360174173021263, 0.010009565592042884985, 0.034748222396574941773, 0.063326725574314401968, 0.067357711769175598882, 0.0098881755935735955731, -0.10637003276557699394, -0.20144930384707626647, -0.22348521745173061115, 0.032260793456169771509, 0.74432871217478746573, -0.38767806947580452093, -0.0016258417803329508651, -0.00035083950145577770573, 0.0030223877091526368636, 0.0075100618531121132379, 0.005996526800841844515, -0.013641731472148289905, -0.045521868840262103006, -0.035814951361764366111, 0.05401364920668801084, 0.16653397085499066499, 0.15941423441096849722, -0.26607832569262118705, -0.49195740477572397209, 0.46783129641981530122, 0.00011120505292645724695, -6.9219689721660687524e-05, -0.00085067155840327316613, -0.0013697551021860480958, -0.0010259774689807005937, -3.2575300585761519421e-06, 0.0042398702862944991424, 0.022718400215186652924, 0.056698828429137371199, -0.00062485249196482398146, -0.34307307264187697493, -0.18116792144557480881, 1.0587334928172782433, -0.62063735757185178876, 0.0039167558450348121529, -0.016317221523956652179, -0.027202391017789460254, -0.019520757531428602205, 0.014923339310946441214, 0.071983223065129553486, 0.11850697349637075162, 0.083018378572529999415, -0.11169518122855057185, -0.30679343597137997968, -0.076544209172368135596, 0.19497327226295449454, 0.051659763389913460063, 0.03306499975682337944, -0.0019182390244534062822, 0.02097938726256881603, 0.025634173629933393324, -0.0029335908971465274493, -0.073365196150016284693, -0.16103448012662233135, -0.18669630140711374056, -0.031683463263268873966, 0.40975839258249641128, 1.0960350272796386406, 1.5199230574081263878, 0.14927643168449825084, -9.9667728109761828392, 7.3361120420463326042, -0.00080923450351637359924, -0.0055627700395784818524, -0.0067459420444206689635, -0.00030126331162744972129, 0.018120277888384741377, 0.043182905413566570474, 0.046282454462780266147, -0.0048565607444588524202, -0.10167075664351678255, -0.22329119838053815283, -0.3270273669908979608, 0.27868790430642320644, 1.4493484757586507783, -1.1876655058261140763, 0.0019284775654377543999, 0.0041988493755418391032, 0.0027363866974338770834, -0.0038923999399578866447, -0.014732635313841757271, -0.02578020327600712841, -0.030632183186456590079, -0.0087552573002100831517, 0.083950984947070678155, 0.25782736587236720904, 0.19324503799937675641, -0.52195596690066570211, -0.65568196063754569103, 0.73459587562986561338, 0.012066888571140417519, 0.011194984737672213082, -0.0021851659004381547916, -0.027378040511942303842, -0.050831048606931182599, -0.04724414369900598043, 0.0043413539828780923352, 0.10319098836431361832, 0.19663612020472656661, 0.16485916009320014664, -0.14755073811527702476, -0.54720362864203686559, -0.093255878700566147144, 0.43847184534371524967, -2.3192705044568420064e-05, -0.010505241305153279613, -0.0093741710401603652036, 0.010065181171732238974, 0.046494484706339392344, 0.075823462899644339386, 0.042850776575357976683, -0.10287556082468030105, -0.30664175725481235268, -0.33332681860282725728, 0.0026860408180261622149, 0.74774406524569458909, 1.3124694171350130834, -1.5198218630691844311, 4.4997077942745292113e-05, 0.0074694804134366327153, 0.013785444319949411579, 0.016579047067723801795, 0.010821622393259695133, -0.012674256785508510731, -0.063673713768927914236, -0.1259264776611868919, -0.12817832919448915452, 0.01132605723070376387, 0.18542834706964261837, 0.35609093497428112585, 0.65208530139927267655, -0.96831225986447755361, -0.0023678497632475315783, 0.0048929332785726037044, 0.01270902862274237051, 0.019395761328319400391, 0.019554725644339961826, 0.0007948099787983588238, -0.053283444190908822724, -0.13294125617539720441, -0.16360476680662319748, -0.039069267327523644895, 0.17406641219597404624, 0.43514684637337008866, 0.74449369962786315913, -1.0654578598123765065, 0.0052430247952316502702, -0.0022986774621404216894, -0.0091043461110450890628, -0.010006752993446520195, 0.0021111309584263917606, 0.027125598894075956397, 0.045095419941012182541, 0.013844439034973949135, -0.10487744264131142191, -0.16652490466970013006, 0.11493684097153317181, 0.24631821738587697523, -0.075435469385213962945, -0.089962021369365430257, 0.003963566702070999391, -0.00089391145707744853587, -0.0060320103449441843033, -0.0083874540638456205471, -0.0026359997459373301146, 0.014320193467847986205, 0.033870673792998726159, 0.024896903993990783521, -0.061759940889418343812, -0.12521756104888503991, 0.089461036723913547886, 0.15351200162013325934, -0.12757299799189938949, 0.012287513402720445441, -0.0089414526789839045978, 0.015939335368023419687, 0.030835860665875167164, 0.019133053992273250238, -0.033710661869737132412, -0.10401201136011822834, -0.10514425745835763448, 0.037459094828843846059, 0.2107009365539588841, 0.22407433128868847749, 0.11602591906680193024, -0.34201089170001430784, -0.66202308177849111814, 0.60905283895708206554, 0.013722208625835676313, 0.0062669657191206313077, -0.0052520473245006799382, -0.019269385323897492146, -0.029474564805709355109, -0.024286395883692975251, 0.0046824455793495652067, 0.046649996915265112829, 0.063277367854857713425, 0.030013273054787494021, -0.087672111224297286092, -0.1324519233801208784, 0.3428214100636969297, -0.21881262208219390275, -0.00044417401691572081087, 0.0013326743461966013889, 0.0023665135289132090135, 0.0017782548407931353645, -0.0013581101761856570389, -0.0073048211786278993349, -0.013545960431269392713, -0.0021882463482540666713, 0.049112507667675338319, 0.060955432832940040222, -0.17852748242831667502, -0.20100474585582298048, 0.62729268135744087331, -0.34266966214894056808, -0.0013536372493505753264, 0.0034879273282943839066, 0.0047265067356720867273, -0.0040540040779339378899, -0.028472233732485283803, -0.057324304202556686239, -0.045133254942010279931, 0.079736809783606490964, 0.31182914933088251574, 0.37381501786040838287, -0.22937506189964865233, -1.020346435959500031, -0.25821309808487169501, 0.90828074878155418137, 0.00011854479816653148007, 0.0020092210007538044078, 0.0017700394240826686654, -0.0049241377767788403405, -0.022425367749758624847, -0.043560141630168608406, -0.034533806823280312204, 0.06494809279133220703, 0.25881713133319411702, 0.31455317173486285753, -0.2106170633583588081, -0.87560778725470633788, -0.13422367188495165635, 0.71477822043061434165, -0.0036955651023573006346, -0.0078261344076438459777, -0.008116837884532481992, -0.0019688053331041489857, 0.011355774847521798304, 0.028343677405598840385, 0.039741556865083460526, 0.034050172966529006213, 0.0046452383011501424681, -0.034427215932623600469, -0.13954972491749412233, -0.23062190670146501259, 0.24173670879176470394, 0.083247365489761498969, -0.0018104712437800670084, -0.0066547795230370672057, -0.008331110015228503432, -0.0042092429544686119774, 0.007311105918944172527, 0.024669571735615451746, 0.040082663639061204275, 0.039237176074089559485, 0.0072853613446988577454, -0.044353929999341928925, -0.14890422396851801801, -0.19999909522024550412, 0.28784127787780977581, 0.0217858346995033203, -0.0069853186585501119213, 0.00097751616452165573587, 0.011857913670827346236, 0.024135405075012633658, 0.030555171871447361215, 0.01385310429650156612, -0.046526029123935351561, -0.14066429435026017725, -0.19944094436358975053, -0.050182906766194405879, 0.3917814597420176681, 0.46013592870388247968, -0.061834285601808237909, -0.45686311162688170162, 0.00288853426015111929, -0.016709753963648692893, -0.021125919149397272279, 0.00063552263319387441304, 0.054627543339019916935, 0.12982165336334333672, 0.17183748977952445491, 0.044178759364740277649, -0.42502576076311732178, -1.132958307292704303, -1.0603454203224011998, 1.8839922565516411535, 4.6459224057908796723, -4.3747797786839761613, -0.6807902459107011417, 0.86923964365077333039, 0.21994393445128310183, -0.20581695663846710787, -0.24801875400407247252, -0.10297593229308234231, 0.037570812248459865834, 0.075781105702747930675, 0.046496012394026446646, 0.0075214412215581314541, -0.011634147476375401864, -0.012251134864544170958, -0.0051694925679649167752, 0.0026314719286505191403, -1.1626053692895514935, 1.6842684181172993263, 0.13639916307407887186, -0.63959560186342434562, -0.23412285613875907786, 0.082491166613646652039, 0.10264389290386859155, 0.037854134917903042856, -0.0031938576794353318138, -0.01337631785873638797, -0.0080687399791883520489, -0.001013842983092065695, 0.0020675034844992957669, 0.0010418843215666540297, 0.67717832116547160837, -0.8082890523443623243, -0.27830743028423238217, 0.31097622422782578155, 0.16971045942388621186, 0.016148950945148191072, -0.032856634191623282359, -0.040012814252282480343, -0.016466056718384958907, 0.0050826608535828763855, 0.0085594229575002584581, 0.0028524340200891439359, -0.0013007755800713814449, -0.0012797121024348684371, 0.41959444423761427245, -0.21686875086647883215, -0.26743975166707467661, -0.032813897955345582014, -0.01393367574175301965, 0.033196575433683622736, 0.072241514018355143523, 0.054535063432263088112, 0.015306801443495920631, -0.013194293571584267064, -0.020332333974911756536, -0.013577970603951067946, -0.0031459463341500138074, 0.0059150345759868116352, -1.1876655060102123684, 1.4493484760501209596, 0.27868790417582745, -0.32702736684941718925, -0.22329119841248837308, -0.10167075662900208755, -0.0048565611865609642842, 0.046282454673369638476, 0.043182905712065837445, 0.018120277754322185182, -0.00030126319216331609455, -0.0067459423070184362792, -0.0055627699508830515118, -0.00080923448491777589392, 6.2849462788848553529, -9.3873905091617739771, 0.92480887610424789003, 1.7316789486667345521, 0.88860522355235460612, 0.106916490985446061, -0.21205026728364775868, -0.19500339595862445363, -0.085022835677189781411, -0.0061085233751399922383, 0.02250844367410696073, 0.019796273153369504244, 0.0067854771878959954656, -0.0062381209738121574243, -0.18956176838140642582, -0.19831583803035590652, 0.48633999248611747834, 0.28167584886824176271, -0.27433788429790978425, -0.21303201314048450521, 0.006037707469711812644, 0.062061292608585283403, 0.032984926778665497882, 0.0052948014536137889136, -0.0041370764688115528943, -0.004096754943321885209, -0.0019028896281774270154, -8.4092961017062717867e-07, -1.6623713526182735123, 1.4771416246833011243, 0.81939499023563078239, -0.08327107610247352909, -0.3612769810650143798, -0.27885992890366534391, -0.10346333625282605473, 0.025620295865463226653, 0.070287587326036002344, 0.052302629783417951825, 0.017336256232065166971, -0.0057770143723035209324, -0.010827263991910644064, -0.0029122176828041347781, 0.45679661534833065906, -0.23306912742434499064, -0.53005467749980073044, -0.0045457859853557467528, 0.22346383690078930329, 0.16669496305002554082, 0.052104830381851492682, -0.037786614825453319899, -0.063743116410873865152, -0.040855695000548794549, -0.0062703253770271093898, 0.014067348609015251776, 0.014505278775339284017, 0.0005650996005977066642, -0.67544051273993621631, 0.65348668501676321352, 0.23685815646648883392, -0.018268463674323141138, -0.12746066055182334043, -0.11218724025339535388, -0.028893346350212529422, 0.021408200048977251567, 0.024900913316494119321, 0.0095457407164234314711, -0.0010295698777515913894, -0.0029038220409879134977, -0.00081942486006745174007, 0.0015451776431751077998, -0.67395669309268491087, 0.67577073758273820125, 0.23301342060535676359, -0.041258670988965295834, -0.13938065567040308634, -0.10849464475281253573, -0.020025423736701953442, 0.027321848009663134149, 0.026285027538324352847, 0.0083903809801576453853, -0.0026267274889384458619, -0.0039188627699261952345, -0.0010678538894349745743, 0.0019121975024599616871, 0.12379833020180587566, -0.1871946315829940144, -0.047493369986119624016, 0.1463280636282170144, 0.011946371049170245032, -0.045933633732564933816, -0.014451392450613621474, 0.0097304918173375474361, 0.0095432914647483411269, -0.00028242530040221443645, -0.0046928160950069309695, -0.0027926705705479054967, 0.00044417357869004658356, 0.0022131207368968090418, 0.13277498052482483715, -0.17968941246029798697, -0.063090228957297142531, 0.13171884352101143056, 0.016447952868721686659, -0.035227477704318187202, -0.010292846231582195721, 0.0074170554748715375806, 0.0063911255626882719494, -0.0011262166365984046124, -0.0038454665918340289084, -0.0019024426586857481544, 0.00058890451187915734532, 0.0016201069345486950046, 0.30328436514135131974, -0.51934267662592614734, -0.18301959207926093987, 0.38764695176988678282, 0.22069948705858821736, -0.064242470156116601721, -0.12681366932181858953, -0.063449022771581348179, 0.00017076470432520022029, 0.022484420418844894285, 0.017916070107535916889, 0.0069144354279788615772, -0.00087105435084823101387, -0.004260047145789456341, 0.35055924583997838395, 0.14326237050061471634, -0.62336454996809986806, -0.21895717836046807436, 0.17579693707784846413, 0.17804039336293503437, 0.073047580785648089097, -4.14240306550636944e-05, -0.023963212504887712595, -0.021620845490988267051, -0.013162252932000720676, -0.0060727166299631594668, 3.1990488923888920746e-05, 0.0059310348084665748952, -3.2476285610418256233, 3.8685583993134313729, 1.1377177376607856552, -1.1586192526755751953, -0.86222495749067240567, -0.15456571397807139512, 0.16110388505510375712, 0.16124367633413164147, 0.066390779517401868337, 0.00053146673191253347972, -0.019031049943257639934, -0.014645934931321843639, -0.0041516229190097890533, 0.0054079265399982340279, 0.027313516489758393052, 0.24421377938777213967, -0.23335535823324193849, -0.16457317323821812893, 0.014589478540158518083, 0.084621771454074970475, 0.054825783699613384048, 0.0084364484311980000864, -0.010184776389707784691, -0.011070264529726455208, -0.0066778962047831210216, -0.0023908872969978186597, 0.00042517585817335341869, 0.0017976540501845601587, 0.043998727900918639666, 0.23184292035272077714, -0.25448880705086990295, -0.16654112868703416228, 0.031366572159571871892, 0.097496339429880951588, 0.054462343599281384798, 0.0013088331238304247786, -0.015708497592033947993, -0.012248569346309929468, -0.005227670631142826603, -0.00069528431822135920888, 0.0011528929821959777412, 0.0012956379824725005809, 0.33589295094152182841, -0.12564601522008389534, -0.50002143223695538765, 0.052132663462966681578, 0.2614137573474248355, 0.091888959540794054592, -0.044519292660223752645, -0.062191445914019045926, -0.02424511251175667792, 0.0079594744765003055953, 0.015974751529008537204, 0.0086635300445259966207, -0.00069812526988187169428, -0.0064062717207648139928, 0.26089896528170147816, -0.057890748044902282099, -0.44908750145758169481, 0.036245514065079892374, 0.22864039773009300904, 0.080118222413355658085, -0.037441676216692150458, -0.052716585875513352621, -0.021258458739381928121, 0.005729247127941902544, 0.012995280949306218382, 0.0074506540359495490242, -0.0001804804798897695307, -0.0051838815563046694232, 0.54273096903734474594, -0.44208149100394833475, -0.44052976050463532598, 0.046216864658756511008, 0.24151645414976025128, 0.18706392474997626807, 0.013105163473159658558, -0.087934875144683746773, -0.069887985783215644808, -0.014799138465108061316, 0.016948688189704529811, 0.018731995830413749465, 0.006858503560983475987, -0.0065563649948215518143, -0.32383196833958555194, 0.68614319973004600506, -0.25295042980460691329, -0.25592719453373496963, 0.060742975234014860708, 0.089213421319249461283, 0.024389950706700899707, -0.0080133748744281012993, -0.013484571530637992631, -0.0103929990042664986, -0.0038354328729573346213, 0.0017020589343928077021, 0.0033067044979114210146, 0.0012275758129701313821, 0.86923964349050986122, -1.6413901473531029573, 0.46615066788204179149, 0.41191856245309838025, 0.10660674404900673251, -0.082745332186173234379, -0.13192503932365345909, -0.054553369799326636991, 0.015447974882838992616, 0.036209676520386861742, 0.021681131557812000771, 0.0023811697144129279416, -0.0067939358106631382583, -0.0051694922166974693542, 1.6842684180016858075, -3.4254308129502537561, 1.1345933575811710803, 1.217028590322055237, -0.24074777153130463847, -0.38044106442979158667, -0.079950696497470358226, 0.049769810019083822461, 0.044171220953617737481, 0.016248122665620062799, -0.002121738097814028761, -0.0070389929584493355144, -0.0040406923862787002405, 0.0020675036001079945963, -0.80828905254670480041, 1.4786690011392367161, -0.14716770745156509959, -0.79684509802985536986, 0.065050263336721758645, 0.16477985837758257759, 0.053721126062632303111, 0.022775544611037465415, -0.012676029686829379944, -0.023161591775650799269, -0.008345827738378297847, 0.0042880329842622591507, 0.0049953855760734560548, -0.0013007753467831257611, -0.21686875028917079744, 0.36541314121267542969, 0.14725704963126706737, -0.32880281475656092383, 0.0045313354721575847817, 0.0060319879680078127351, -0.027338204074831954499, 0.010710779477770731097, 0.030945626534613392133, 0.025561562070159755955, 0.006942139454844287183, -0.0068408430748272946315, -0.0093558573297145494752, -0.0031459466619964560757, 1.4493484761427202212, -2.9022793340339312351, 0.9784974477179200214, 0.79561328860612567926, -0.088269519673142227134, -0.11839823491536775879, -0.087924515423100424449, -0.057371306082523110259, -0.0064942642958019302854, 0.022497177257961107483, 0.019740718864691014084, 0.0075336411460295146444, -0.0013453761116118464081, -0.0055627701143584271093, -9.3873905090987008748, 21.017143400862597247, -11.864312280297738766, -1.5435132043965564819, 0.640218279070838725, 0.82398363844467292783, 0.42687101017378276113, 0.04763750703548323695, -0.093878166367831958339, -0.084173876956694981821, -0.036813795162808735095, -0.001146245510374244549, 0.01151163786478029763, 0.0067854773019826149574, -0.19831583798195828727, 0.76125984100725707648, -0.79794498832702975832, -0.27495511941895345309, 0.66377363045169346378, 0.10124274593796385446, -0.22101387399459243577, -0.092517142850409600507, 0.016573858047545814848, 0.030216715768938928049, 0.013211560034165596364, 0.001050749594228817755, -0.0027658647673958940486, -0.0019028896674846479051, 1.4771416246610349354, -2.6743467309265795961, 0.34878808270944278158, 0.86438564987191079947, 0.30984085141945072461, -0.051699772914322629502, -0.15641328628324407957, -0.12592231127123762469, -0.055102382359511509202, 0.0083048542718489031661, 0.032444812919811209795, 0.025638498241158248769, 0.0073108515411678424642, -0.010827263991073863766, -0.23306912748499949983, 0.080294560300014569676, 0.66552555468583929343, -0.40924129051152574599, -0.2649215612457637703, 0.022060505165506040515, 0.079567360886323434954, 0.075455611186409490831, 0.031337597349239840194, -0.011887949155961608297, -0.028392779843531265604, -0.020294811754653408614, -0.0028265816408758557415, 0.014505278598584017344, 0.65348668490016725841, -1.2899208957134269404, 0.41008310578429413296, 0.22271970331021681555, 0.10445790209289913808, 0.019608833777797218118, -0.071990052412819277583, -0.060422023441545111289, -0.015579157505223793084, 0.011737306439532971952, 0.012494770076240932499, 0.0042218305156943051679, -0.00099108759092421741532, -0.00081942479905384221078, 0.67577073745670901417, -1.3272552262313752625, 0.41493300864388321658, 0.24729557569395183014, 0.10245890485156666783, 0.0067094128794124175547, -0.078206484577615753562, -0.058726540256462018974, -0.01138135162738685216, 0.014409509107780590909, 0.013037500103746844926, 0.0036072020733910613484, -0.0017486088679521407351, -0.0010678537690113486421, -0.18719463162327315553, 0.33711736356050481245, 0.031441417582967728928, -0.34296342869575863199, 0.10233941148137587485, 0.098153553657096645479, -0.013900917626992385576, -0.027783723831877011473, -0.0085048306183860399726, 0.0068380790161631389151, 0.0057162628347359988046, -0.00041758400451850051505, -0.0024801813912687098451, 0.00044417359902605983886, -0.17968941245651362526, 0.3134854497413120944, 0.050765458141763933342, -0.3287834811082400166, 0.087333971414643302156, 0.08687975698238495359, -0.011290271393554875481, -0.021085886151538908023, -0.0057704777033545694501, 0.0055514525556474916823, 0.0039255187446585832964, -0.00085682075903795922121, -0.0020502477897970539794, 0.00058890444806623711959, -0.51934267688165069998, 0.82880838918712584906, 0.17466885606126564268, -0.73329875775614883349, -0.030646656147935141118, 0.27297748317566356002, 0.10788066649831765109, -0.037409194757356080407, -0.062619735582139263164, -0.027980489635610852894, 0.0019071603860190849446, 0.011546051063103298512, 0.0078572827814054058138, -0.00087105421494068852387, 0.14326237068897471061, -0.70360214504796281254, 1.0460525121627728407, -0.16438093031265760269, -0.44374662465759318053, -0.052722295618528057515, 0.096048105155389917886, 0.081383444829401671572, 0.03012117820438814722, -0.003010308328828251688, -0.012363197478638621174, -0.00887383303628139046, -0.0037123559386602659159, 3.1990395305426342775e-05, 3.8685583992570826695, -7.3209529595533382462, 1.4933912228168122471, 2.855903453869754749, 0.10171691236721633333, -0.6777204906112967997, -0.40256089844669756816, -0.065130096419201380376, 0.077758453673139923956, 0.072252451810705650703, 0.027084362354974741288, -0.0014003244031348773598, -0.009184660234803170703, -0.0041516229653079167114, 0.24421377962413592932, -0.64534577865006703679, 0.53194887670881907393, 0.042947704167776108719, -0.14838390033431647641, -0.092348585524065815444, 0.010298361187843494899, 0.047589906865446923723, 0.022893859293541887756, 0.0022292077147230467821, -0.0049409722222241289277, -0.0050407261984067525465, -0.0027145043204083356761, 0.00042517551314956252267, 0.23184292053001034906, -0.63924099769852482655, 0.55287553972155889515, 0.040749019515690994286, -0.16856239282446042194, -0.096994647843741066406, 0.019432077053856341986, 0.055176282514869050277, 0.023283548563988055308, -0.0012093795196700734759, -0.0076224791788498183487, -0.0055178898808499994913, -0.0019310751757442498611, 0.0011528928330345777128, -0.12564601518742707364, -0.14834563638464601243, 0.83267124426796745862, -0.42740107367525642657, -0.36544442746208127826, 0.094461818479115597369, 0.14526559198875221846, 0.046847429611552880713, -0.021254571457004149682, -0.030201865159719620824, -0.011929850180758099984, 0.0044802538773757154961, 0.0073758626443698850975, -0.00069812513740417870764, -0.057890748063168323556, -0.25303644513268852556, 0.80681625833339365439, -0.36804996612136792189, -0.33238585344035498359, 0.084206764445762471993, 0.12564727663087066345, 0.039334262501449980542, -0.017933950417001586625, -0.025263726383746507032, -0.010324697266088495723, 0.0032985117504432392987, 0.0059903610919756992287, -0.00018048064798883146323, -0.44208149118321954374, 0.58789978528662079071, 0.29327026337993128413, -0.37561815214636218307, -0.20520564220247358711, -0.042492174679570386686, 0.11968304874721728825, 0.118764581963912777, 0.011163703006923130986, -0.044263501741717482807, -0.034301595633200288038, -0.005894630956541795086, 0.0093459128865210523707, 0.00685850372805311536, 0.68614319965934023049, -1.6567878138938862964, 1.0426610570239551734, 0.2834159890557185757, -0.34884254837184164311, -0.089102931262777435029, 0.05461987035224549758, 0.035559649949661098489, 0.0058826027554024088242, -0.0019822353909791684719, -0.0050638905048599738498, -0.0051513549973207393637, -0.0017546128544249231865, 0.003306704613948326793, 0.21994393423654440278, 0.46615066790513554063, -1.5589498788260041096, 0.56222049693870490561, 0.46669287137811721333, 0.089039153839786860933, -0.10350536095829186711, -0.14457847720636798905, -0.057693117411989347509, 0.016187524240444530615, 0.038321201425543301167, 0.023695502675634364659, 0.0023811704878576650177, -0.012251135116147601756, 0.1363991632685054578, 1.1345933571776383175, -2.721672821344438109, 0.88404041306166081249, 1.1158128093648076451, -0.19571852476230713891, -0.3583462064859996854, -0.084066818093295692105, 0.047191890817022762172, 0.044020120687451341768, 0.016117348401752787379, -0.0028584080287834658443, -0.0070389923141272086118, -0.001013843368302005379, -0.27830742976125866228, -0.14716770827377356756, 0.94707662687818294778, 0.0035640806971102105549, -0.77987522992147129131, 0.026687488363497189475, 0.18153857352589516938, 0.059424385356389726465, 0.01721023868788338837, -0.014383865640253087545, -0.021461817791589000298, -0.0070676232823300676189, 0.0042880338753587335376, 0.0028524334493826510213, -0.26743975224104776789, 0.14725705096177058029, 0.17973530809745896786, 0.29809274504895211244, -0.31517241124099315552, -0.053360167611641536045, -0.024406115391911432921, -0.038522076824827763963, 0.013110548457801678157, 0.037137341388106315243, 0.03094895598437277448, 0.010052548590003432954, -0.0068408438602130545308, -0.013577970429144967796, 0.27868790419961286808, 0.97849744766214341585, -2.6601856615746539703, 0.82511957925883161469, 0.88456503245405104785, -0.037671353929955991857, -0.1288312550272650292, -0.10192261397534398137, -0.063060142209842712235, -0.0095831907880732022931, 0.022237212797863857572, 0.021385501002187275871, 0.0075336407670733677999, -0.0067459420505470869128, 0.92480887599485872563, -11.864312280061762195, 22.951724945698096292, -11.946334245386188044, -1.7860665214560800251, 0.55319049544847387345, 0.81036210881626347113, 0.44817699959831797107, 0.06879881487125552586, -0.087612622928557343238, -0.086785344425500829391, -0.040883125865686223421, -0.0011462452491274859479, 0.0197962729377059235, 0.48633999228697077921, -0.79794498801456559267, 0.41636368129533191107, -0.23389381099224423877, -0.25894585162336219986, 0.53862596190718592393, 0.099274043645466381047, -0.20356274558470652258, -0.092085528280775541532, 0.012032162936060566624, 0.028530840683909042266, 0.013538996117337692704, 0.0010507494491190628137, -0.0040967548456822214936, 0.81939499035800622551, 0.34878808267281335986, -2.5326198200577554331, 0.39433682129900604618, 0.95077289079893534396, 0.38413969377061185639, -0.03030741165341211546, -0.17331049968824521756, -0.14440997548966108543, -0.064480012707596984223, 0.0079040443338545957053, 0.036081294951419366768, 0.025638497978614303852, -0.005777014291347856885, -0.53005467733072031411, 0.66552555444482064395, -0.17044407593199717432, 0.56531431744338267453, -0.42820821066083980533, -0.3108922565329808041, 0.023664407287783201933, 0.1037354389753724071, 0.094702196869789020961, 0.036793056311538958891, -0.016512078842844345877, -0.035524625364652964721, -0.020294812262010818171, 0.01406734896118773219, 0.23685815663241191475, 0.41008310561115413062, -1.3825536317844688838, 0.44795882383668111926, 0.29713284745576051815, 0.10906121908914316598, 0.012592326582821790001, -0.075510161245427184773, -0.059654269199339124419, -0.015490031470284892218, 0.011158769771163954548, 0.012296856934849890314, 0.0042218307866436297765, -0.0029038222825542867392, 0.233013420713633429, 0.41493300850935743718, -1.3950153315146536226, 0.45716949799146178179, 0.3178610867836045295, 0.1038842337879748029, -0.0019805201308647490954, -0.08235142337524382905, -0.05805514988606585991, -0.011034045284461743361, 0.01419669252910277929, 0.013070496050631309065, 0.0036072022655507032796, -0.0039188628923245896057, -0.047493369980765150207, 0.031441417415331344987, 0.023080826260018935669, 0.16307185238521815229, -0.29408659810138326751, 0.074939157216426605101, 0.084329582448565545971, -0.016578535482380026694, -0.023465813313393625583, -0.0054702373747892805295, 0.0072446167720220545211, 0.0053619012663603552934, -0.00041758408161937090805, -0.0027926703816081198048, -0.063090229040358714041, 0.050765458112095637744, 0.017883505837102596503, 0.17219850728515487259, -0.28802460124401674246, 0.060077392918133278032, 0.075804185724992992346, -0.012578560093673670298, -0.017085047374435509415, -0.0033751808371892720942, 0.0056844729314210282883, 0.0035029261388692181645, -0.0008568210874832771478, -0.0019024423470930310314, -0.18301959164183137507, 0.17466885524040393074, 0.18225737743568845572, 0.15891053146113781613, -0.59978632816475929701, 0.00081376009263313701589, 0.27200819256103725419, 0.10481303792657230278, -0.037789637324029574927, -0.064433628983000632018, -0.030443570206975498899, 0.00058011542555174854148, 0.01154605158179329745, 0.0069144352420635237649, -0.62336455004980162453, 1.0460525124724631141, -0.78114357086451102763, 0.82519720387373174386, -0.14894672682698478772, -0.44470132613267815413, -0.065952839305359806099, 0.089276396652412062793, 0.082473198962241395238, 0.03530475574813393691, 0.0030618062317544101833, -0.0089014060577483917341, -0.0088738335092359670042, -0.0060727164313672827439, 1.1377177377431708649, 1.4933912225765009207, -5.8523911650665256801, 1.3256351399322590101, 2.7682148051348112183, 0.13825129448575404489, -0.65944681918855641012, -0.40643320262607035565, -0.076497663613568495089, 0.071098754564573429793, 0.073432152607888531715, 0.030206929786729137116, -0.0014003243729492963286, -0.014645934865292996319, -0.23335535858654485319, 0.53194887726827999597, -0.6210679193284298627, 0.484828882831664143, 0.007260974754478586278, -0.15503707487875292848, -0.087062578326261283457, 0.0063955917720041281088, 0.046782346472478246457, 0.025171529613261654779, 0.0043291658989624004494, -0.0042251564568465266314, -0.0050407274406824775434, -0.0023908865647043701236, -0.25448880730615153212, 0.55287554029264951438, -0.60892098798455007724, 0.48804485607299485928, -0.0028071647876814591338, -0.17293485498309921211, -0.087818306119055444814, 0.017490462427610545953, 0.054549119945537953569, 0.025153108590239250247, 0.000453460880296112194, -0.0072572277777739700039, -0.0055178899515169370707, -0.00069528428418008810484, -0.50002143240106733391, 0.83267124463320740801, -0.49615529101670174938, 0.6027603969346254642, -0.35802655624254997591, -0.32274322623147227995, 0.10553027285401392343, 0.14492839728707526681, 0.042972048632562855064, -0.02517830469708906907, -0.031974381642631680145, -0.012501289461509844728, 0.0044802544855401058715, 0.0086635296142371174405, -0.44908750142405068351, 0.80681625824455316387, -0.57031179376227891797, 0.59809781835850039933, -0.29951696674706684842, -0.29620467581582787808, 0.091832306048841044754, 0.12553235222111455704, 0.036857151496503023913, -0.020905823588328723006, -0.026819077261518844968, -0.011007348206934278873, 0.003298511153144001451, 0.007450654323697400612, -0.44052976038273161663, 0.29327026302382092071, 0.43576240669774746195, 0.2008217264734369345, -0.49969095447734052096, -0.20169125279884309032, 0.013324122977517751515, 0.13970172188991306084, 0.11779056415710738304, 0.0063461515709916285072, -0.0481333942783091287, -0.037242621443643951495, -0.005894630541024125335, 0.018731995626820752837, -0.2529504297414651437, 1.0426610568316228012, -1.5572930692189843604, 0.77402123212335183222, 0.30718266266059984781, -0.30742293331162934278, -0.092466371607137914923, 0.049176322563918101194, 0.039954814290522167042, 0.0091787585601459994317, -0.0023370966910779807438, -0.0065398483191364299716, -0.0051513546112536478735, 0.0017020587162655134803, -0.20581695568579194688, 0.41191856140666316399, 0.56222049661866524417, -1.5562579314645417217, 0.48820778384697605112, 0.42668902589791546376, 0.087104942553373018721, -0.087102846609246248333, -0.13303970753538327254, -0.055057383127070411155, 0.015307858344581828924, 0.038321202250542175938, 0.021681131799225237394, -0.011634148088077255964, -0.63959560188238295808, 1.2170285907181142004, 0.88404041246462727877, -2.9237539471083389131, 0.81841657024926972586, 1.1415828929221571997, -0.15948226762238446041, -0.34249870236853291461, -0.085212730370565037408, 0.042262064626271772516, 0.041279309450013561278, 0.016117348863213670773, -0.0021217391501656694859, -0.0080687394312836061699, 0.31097622374795030886, -0.79684509709892459295, 0.0035640802812640739496, 0.98197881087017158119, 0.033123418044092058088, -0.75350122455115853271, 0.014465238088090739618, 0.16392488740726199326, 0.053408048917447437098, 0.019812513318269751178, -0.011270509490389001711, -0.021461817338151892992, -0.0083458286556759384706, 0.0085594234172886047435, -0.032813898598671564777, -0.3288028149547724821, 0.29809274710263977681, 0.17105708109738601563, 0.25847037118951793344, -0.33500792901252163247, -0.057374125254806722374, -0.015828142462936786034, -0.032146071048892631283, 0.014816970705523024171, 0.037199095359063426292, 0.030948954026371544579, 0.0069421397929267832094, -0.020332333442896012576, -0.32702736713616409325, 0.79561328908035444663, 0.82511957917931699757, -2.6523506505542995271, 0.78435112014839458716, 0.8584808146173249499, -0.038612571690478769693, -0.11855129231091061925, -0.094376486817293728238, -0.060274930002316963462, -0.0091261024081119735546, 0.02223721256572006727, 0.019740718953789301859, -0.00030126317744376206837, 1.7316789486869177406, -1.5435132046441926157, -11.946334244992620199, 23.682253547083146117, -11.667729042424376473, -1.8089573994148731018, 0.46971603105643849885, 0.75599497807601756527, 0.43055653791864695323, 0.07060205407803886446, -0.084115443639443482349, -0.086785344687636573058, -0.036813795112465519388, 0.022508443732098870615, 0.28167584920554472605, -0.27495512004669109274, -0.23389381067967385253, 0.47106387140010824721, -0.34703770394190175397, -0.31004412871438807864, 0.53975949088881025606, 0.11763236698535833868, -0.19458844493477028825, -0.092196527261258098829, 0.010365765082500885338, 0.028530840663960822107, 0.013211560021974389986, -0.0041370764669099082539, -0.08327107638084027541, 0.86438564998957856389, 0.39433682173405409932, -2.4801307398981964347, 0.37301953501715423567, 0.91314440702960841811, 0.35981557483160764299, -0.029728670360059227668, -0.16209218978998718441, -0.13417536589529691082, -0.05979136801960473202, 0.0079040437838140357313, 0.0324448135135460905, 0.017336256036303576006, -0.0045457862973662865533, -0.409241289993571844, 0.56531431724719616305, -0.2319677215104732626, 0.55465499961942510421, -0.38714385566345010936, -0.28078975421384766431, 0.030910545242584731157, 0.095201259878821900551, 0.083191947646211183987, 0.029780631345699349471, -0.01651207893124298498, -0.028392779561880823053, -0.0062703255203649102198, -0.018268463414475626105, 0.22271970294042181537, 0.44795882361203154254, -1.3424440993609090267, 0.41738146222166155885, 0.27128350171612874986, 0.10238492800754236955, 0.021652011230378354067, -0.067233903069050540702, -0.057304785525156116954, -0.016223483075543505827, 0.011158769944351321773, 0.012494769968773966617, -0.0010295697782372679584, -0.041258670902660561641, 0.24729557564766208033, 0.45716949771102233235, -1.3605839886905766178, 0.42105599882758931729, 0.29066091220028866804, 0.098390435545569143727, 0.0081857814117712989627, -0.073714333836852427573, -0.055756097242037012951, -0.011835031321101266433, 0.014196693012747457416, 0.01303750023124720589, -0.002626727722051279152, 0.1463280636468278495, -0.34296342836745091764, 0.16307185175539923305, 0.068721431651755496284, 0.15645543364510017259, -0.31194416121905282901, 0.063492957812690972719, 0.08854528271871543732, -0.011063062190051842562, -0.022182333066576916114, -0.0063451877077072751238, 0.0072446175093692924754, 0.005716262825431528416, -0.0046928163933083509932, 0.131718843690658588, -0.32878348096768222897, 0.17219850661437627326, 0.053935726753297677516, 0.16531743219425132363, -0.30150505620870510226, 0.05116445863681759465, 0.079403519168092295821, -0.0084642086979310471317, -0.01640097507146718106, -0.0042215409315271854376, 0.0056844732757332103526, 0.0039255190678763570225, -0.0038454669736212843643, 0.38764695162037299259, -0.73329875705237768102, 0.15891053054571885328, 0.32306330720127962586, 0.24447631410244241579, -0.59631372805677507198, -0.032021598406103177203, 0.25345853210109686726, 0.10330035698949745215, -0.033678461573011568564, -0.062316977044536869812, -0.03044356915172904271, 0.0019071598670652592858, 0.017916070062381033584, -0.21895717864955899223, -0.1643809301888924379, 0.82519720413711405982, -0.84096552932956214477, 0.83038902226775601623, -0.11618964814989751755, -0.43417772014138222847, -0.070899755448399909286, 0.084245278314485527726, 0.082930802081794374736, 0.037562500436548761085, 0.0030618058139628440834, -0.012363197265139314399, -0.013162252873419101953, -1.1586192526783221091, 2.8559034541965639953, 1.3256351393560297236, -6.1003411465061905261, 1.0893169811914547029, 2.7281082526400837196, 0.2001960720546331618, -0.60930100917688778672, -0.39157054710262989561, -0.078536066284557351502, 0.067887604712130927842, 0.073432153384526868689, 0.027084361814475679114, -0.019031049854491877832, -0.1645731732333907682, 0.042947703996900746248, 0.48482888315996619522, -0.70137832102511665777, 0.47036541199950226977, 0.028940180257287922982, -0.14571078010245747114, -0.086087289678823578076, 0.0041891661412650927493, 0.045479295976195918683, 0.02501231973440517023, 0.0043291648518079522273, -0.0049409709961922088431, -0.0066778966679281797844, -0.16654112852030558689, 0.040749018680827385486, 0.48804485714654155526, -0.69753131316221017588, 0.47741157911423531823, 0.024977780582221276173, -0.16137535576241299196, -0.088001289863696124938, 0.013701716042598106535, 0.052581756371936225203, 0.024919211617567677691, 0.00045346086565571088953, -0.0076224798946501314684, -0.00522767009159833939, 0.052132663642867047016, -0.42740107441206098837, 0.6027603979663821443, -0.47189580424104698997, 0.67531246981063619561, -0.31270591550261206404, -0.3290565541549585471, 0.083912894674725097, 0.1327570508060639809, 0.042440628593807996904, -0.02281468851357672939, -0.031974381872201841637, -0.011929850750805243154, 0.015974751990324255468, 0.036245514047356458542, -0.36804996609535656216, 0.59809781848096343904, -0.54713044420761713926, 0.65828671800791505042, -0.26106050431601024453, -0.3010812880390637547, 0.074115215948241025545, 0.11543160938269675375, 0.036250786802712903212, -0.019149901222227233566, -0.026819078399262934881, -0.010324696401095193043, 0.012995280767227126995, 0.046216864988226739885, -0.37561815222028077654, 0.20082172571926543148, 0.30001465007587013734, 0.25100690696373079147, -0.41125318906970853572, -0.17964374587625839275, -0.0055133149792498097747, 0.11816952931118807069, 0.11020226091429705106, 0.0067551106896910064967, -0.048133393819784417444, -0.03430159545705575852, 0.016948687897318712864, -0.25592719462617863346, 0.28341598948801571822, 0.77402123158453695062, -1.60753430748326398, 0.78240965504189563529, 0.32784391154954950487, -0.29787261849315810336, -0.089824344702282371555, 0.047689177386516574952, 0.037478076448734805315, 0.0075115145034840317614, -0.0023370961918326572763, -0.0050638910008509131419, -0.0038354327282432596966, -0.24801875404914680567, 0.1066067442645892982, 0.46669287139718679258, 0.48820778308777085241, -1.6149097863160366551, 0.46873730940469299222, 0.4363169328451335649, 0.10119170259966463188, -0.079561795180910735437, -0.13172721920847840771, -0.055057382888352844608, 0.016187524401521904277, 0.036209676661587482382, 0.0075214410088596424242, -0.23412285608304753137, -0.24074777193056587365, 1.115812809925024629, 0.81841657014623681032, -2.9442450407003200041, 0.8190985962784230745, 1.1553247935238490918, -0.15107894157692522419, -0.34242217269113722766, -0.087783136175703241344, 0.042262064416967726199, 0.044020120335340175155, 0.016248123637043671019, -0.013376318362069712642, 0.16971045969524561903, 0.06505026280593693977, -0.77987522979388235189, 0.033123418349711733488, 1.0281112312256706076, 0.040958734652786071839, -0.76740036159443592023, 0.0049905050823668070659, 0.16338812749424236337, 0.055448159855166982701, 0.019812513141904207481, -0.014383865994303052238, -0.023161591262368885769, 0.0050826606662990542543, -0.013933676226270381646, 0.0045313362715900347075, -0.31517241224938502553, 0.25847037281977452983, 0.15377389902853136339, 0.26116331703278561349, -0.32255980480606849525, -0.048606514153201094142, -0.012965334411788509775, -0.032327119018559842545, 0.014816969326453854547, 0.037137342432844520912, 0.025561560700526814172, -0.013194292743686575689, -0.2232911985281265943, -0.088269519642214383515, 0.88456503256832286208, 0.7843511202983235453, -2.7141840034683326621, 0.75577507531842746946, 0.86167366700043834804, -0.028380609306950136533, -0.11135157039406690105, -0.091572419422384476739, -0.06027493020746561847, -0.0095831905155679721064, 0.022497176846930280053, 0.018120277967266274627, 0.88860522353681548058, 0.64021827920018903324, -1.7860665217339433131, -11.667729042214213919, 23.912071282038311892, -11.619940944231908375, -1.8470456139963682407, 0.42688503477102801575, 0.73685316755748786477, 0.42823918626389118058, 0.070602053845744167093, -0.087612622539613910222, -0.084173877199185911935, -0.0061085233210011677385, -0.27433788454232838072, 0.66377363104257081083, -0.25894585220113403023, -0.34703770358378210892, 0.44102840496785955526, -0.34585690162306903517, -0.30592818539842941972, 0.54469420628027298559, 0.1217419857569245617, -0.19349331806602307338, -0.092196527571845265925, 0.012032163055230268509, 0.030216715998030420354, 0.0052948012965240926278, -0.36127698091734977703, 0.30984085145522455296, 0.95077289022342392766, 0.3730195355351805242, -2.54198967648240437, 0.31575158817416965062, 0.89479592102805272091, 0.36854803667596425232, -0.014256468986087069209, -0.15299392156317023761, -0.13417536638532018278, -0.064480012023221955353, 0.0083048535767184118034, 0.052302630075054347414, 0.22346383717095388843, -0.26492156163054142448, -0.42820821061938990626, 0.55465499964223585749, -0.19350113225841975018, 0.59674399346338591332, -0.3700561079439702028, -0.28796947244116960496, 0.016200074533526134801, 0.085472066425882328722, 0.083191947643473165841, 0.036793056498229795293, -0.01188794927427447301, -0.040855695012027300195, -0.12746066070551009819, 0.10445790241354742378, 0.29713284749271406948, 0.41738146181045082672, -1.3804121292250017827, 0.39927791065348239297, 0.27349402414558859897, 0.10938927670773508638, 0.026066060594671838085, -0.066244510565578576533, -0.057304785162804440513, -0.015490031515246922844, 0.011737306254224463831, 0.0095457407946458374542, -0.13938065559945064797, 0.10245890468777053839, 0.31786108724548978177, 0.42105599819439332165, -1.4006395705159027099, 0.40347807888549747091, 0.29407411163250962716, 0.10622912667254419405, 0.012845502322700805276, -0.072779766451746011335, -0.055756096457280968082, -0.011034045969511560581, 0.014409509075759330265, 0.0083903811532069910717, 0.011946371414745386372, 0.10233941075195321369, -0.29408659754457111557, 0.15645543328381011916, 0.057272809597679608107, 0.15131417740164884722, -0.30856344502878424452, 0.066875610015816808196, 0.089506942936491654117, -0.011543835590014813602, -0.022182332459630683258, -0.0054702381260803975213, 0.0068380796882576942719, -0.00028242558616930276034, 0.016447953003413933598, 0.087333970912794128827, -0.28802460052776646116, 0.16531743172861365632, 0.046261355289285321923, 0.16251625531838242988, -0.29834995771331856984, 0.053462442690916690824, 0.079733360245318046777, -0.0090481776663847192366, -0.016400974730209433305, -0.0033751813155099706816, 0.005551452697827735823, -0.0011262166036851836357, 0.22069948744624212633, -0.030646656915924348069, -0.59978632754851501918, 0.24447631366673686082, 0.33887761763825929862, 0.22902769598627212289, -0.61297338667702849779, -0.04074355095693202311, 0.25299920958631211265, 0.10537864122509230103, -0.033678460941680511687, -0.064433629672285372414, -0.027980488974625641885, 0.022484420112427554866, 0.1757969369581543484, -0.4437466245250604735, -0.14894672669441713642, 0.83038902219651455905, -0.82519098589339834326, 0.84566511928150001065, -0.10875115212044920987, -0.43425902924204828892, -0.071950716505325848504, 0.084226669824098168715, 0.082930801999619122022, 0.035304755836393510171, -0.0030103080406156850282, -0.021620845725964763645, -0.86222495742156590737, 0.10171691200211568917, 2.7682148057916817763, 1.0893169806887581519, -6.2278495127425115285, 1.0634782502778685753, 2.7542390775392751046, 0.22964183966630363765, -0.59504103112553652366, -0.38970027619560188015, -0.078536065523323442261, 0.071098753228763966749, 0.072252452817565981924, 0.00053146643908795997147, 0.014589478170260503087, -0.1483838997689405359, 0.0072609746807190730239, 0.47036541188726571638, -0.69075377162632822792, 0.47711143834155500532, 0.03408442733827094645, -0.14422300123199580102, -0.086973921007822543872, 0.0031049012225991580449, 0.04547929585185272322, 0.025171530047726532048, 0.0022292065814246025436, -0.011070263872968624796, 0.031366571669103199271, -0.16856239167106581989, -0.0028071658472913043128, 0.47741157967729197509, -0.68096757031124854898, 0.48677669248166438587, 0.029517370792186321121, -0.16139886548408247102, -0.089857152345712842934, 0.012329123181166783824, 0.052581756542432994683, 0.025153108507132881239, -0.0012093792583025122014, -0.012248569540943695169, 0.26141375716929809991, -0.36544442686484235328, -0.35802655712109909825, 0.67531247039092268114, -0.43337441683155608629, 0.67992341259285271615, -0.32503279969236853475, -0.34207270898597763598, 0.079309406022693276772, 0.1334655721012195928, 0.042440628397399655924, -0.025178304350483279861, -0.030201865390495560848, 0.0079594745455193790379, 0.22864039769668090263, -0.3323858533060468079, -0.29951696700902685766, 0.65828671812861483392, -0.51621071429648268669, 0.66246984129565966715, -0.27058703915994841704, -0.31184952044514158143, 0.070007069896877577309, 0.11575730956486147971, 0.036250786050740259336, -0.020905822484569111519, -0.025263727181208041339, 0.0057292473761908283003, 0.24151645417380462888, -0.20520564224889206728, -0.49969095412403852219, 0.25100690629589644809, 0.39279460010740063769, 0.28506011360338573413, -0.42094400983595248178, -0.19992298171742278745, -0.017356953096890892496, 0.11540525960367660485, 0.11020226122977733185, 0.0063461515862781748415, -0.044263501876388791634, -0.014799138437344913632, 0.060742975828122609816, -0.34884254952229987179, 0.30718266326523013809, 0.78240965500220271966, -1.6123298773261296457, 0.78332775104188845372, 0.33566412798478112434, -0.2958523737705056611, -0.092617165123018319051, 0.045148805521712770394, 0.037478076400243448552, 0.0091787580725998744319, -0.0019822343655075094944, -0.010392999512205763901, -0.10297593300755801538, -0.082745331422457865655, 0.089039154228988934281, 0.42668902613692788472, 0.46873730846003541783, -1.6129419837487011247, 0.47571004562697943863, 0.44111554204235692289, 0.10291548838023462098, -0.079561794673285182733, -0.13303970689202532229, -0.057693118137311136251, 0.01544797354742696166, 0.046496013543482025976, 0.082491166245212951824, -0.38044106368751662295, -0.19571852497618688616, 1.1415828924262485522, 0.81909859670842799506, -2.948042439712768914, 0.81625387894112899634, 1.1544839463425213832, -0.15170259349340303734, -0.34242217298778671042, -0.085212730005820178092, 0.047191891173757528899, 0.044171220168255989624, -0.003193857336572104421, 0.016148950657104525763, 0.16477985890631810517, 0.026687488386293440473, -0.75350122527144025586, 0.040958735268212517755, 1.0227992183460201669, 0.038251279144761708706, -0.76624313359479090391, 0.0064432587450477451715, 0.16338812663512453094, 0.053408049474961548364, 0.01721023923377852996, -0.012676030392250590972, -0.016466056513425027935, 0.033196577249758357775, 0.0060319856399990009685, -0.053360168219044018523, -0.33500792887441493972, 0.26116331850798607528, 0.1680713051090302057, 0.27189235147298479456, -0.31851062941016811125, -0.04820701616367590453, -0.012965335816653506679, -0.032146071073697268794, 0.013110549026081745258, 0.030945628382982766358, 0.015306799967659642792, -0.10167075638229738921, -0.11839823521268035, -0.037671353997067169583, 0.85848081445276769319, 0.75577507571140478682, -2.7237428970487638225, 0.75632358117764209027, 0.86574811240699700576, -0.025252603830957259801, -0.1113515706814380557, -0.094376486727940855914, -0.063060142367827295984, -0.0064942634748720480531, 0.043182905190006663776, 0.10691649092656478615, 0.82398363868982071878, 0.5531904951039855467, -1.8089573992482885778, -11.61994094424021462, 23.950061318311146152, -11.616535350602156029, -1.8574041152686175948, 0.41975908311919951643, 0.73685316759304020362, 0.43055653782061253931, 0.068798815161161916421, -0.093878166783340821766, -0.085022835484457812139, -0.21303201311168670773, 0.1012427456863318892, 0.53862596253307859495, -0.31004412942587089574, -0.34585690123123680184, 0.44838681487162601602, -0.34447369698876922905, -0.30630760991627653533, 0.54503239219528010207, 0.12174198534648444925, -0.19458844426846963471, -0.092085528789128248683, 0.016573858146152364856, 0.032984926823402704876, -0.2788599286506619479, -0.051699773437841288481, 0.38413969407399262135, 0.91314440691842668762, 0.3157515883184338068, -2.5720368623562959165, 0.31165315168232377019, 0.90345939427879684569, 0.37667574260317054913, -0.014256469367437272361, -0.16209218926023558072, -0.14440997614967682194, -0.055102381740437522684, 0.070287587083440364322, 0.16669496315771947792, 0.022060504867966040932, -0.31089225617257870704, -0.38714385583994692652, 0.59674399342147188552, -0.17426058938139324384, 0.59443845058973621143, -0.38144860017813009279, -0.29714742745843220062, 0.016200074554826717627, 0.095201260213313806946, 0.094702196157762236206, 0.031337597730424417153, -0.063743116444968578316, -0.11218724049805130216, 0.019608833842346792137, 0.10906121976532959228, 0.27128350141893209369, 0.39927791032177101238, -1.3828684168136935462, 0.40141053906772611315, 0.27516016024896444714, 0.10994026938368396673, 0.026066060735498931128, -0.067233903038626433024, -0.059654269696369700837, -0.015579156836524592905, 0.024900913013370145482, -0.10849464472855697983, 0.0067094129416551216011, 0.10388423336048127621, 0.2906609130040255784, 0.40347807823987219544, -1.4022359712782594965, 0.40621006614854993089, 0.29601565443453453641, 0.10683203591461687409, 0.012845502872815564577, -0.073714334381212512359, -0.058055149608155665053, -0.011381351840155070135, 0.026285027633177592693, -0.045933634249811096384, 0.098153554454516969696, 0.074939157137787912322, -0.31194416131790725366, 0.15131417729296606223, 0.063502471015403252896, 0.15615000547366403305, -0.3080381109487863589, 0.0659154406867931697, 0.089506943402073158023, -0.01106306234506592226, -0.023465813310272181508, -0.0085048312099714040019, 0.009543291891187109105, -0.03522747823976286885, 0.086879757769406595891, 0.060077392934032733396, -0.30150505636241448082, 0.16251625513898548814, 0.052070223978840916856, 0.1662913337474329345, -0.29828185535688117591, 0.052487202996633029739, 0.079733360155544857895, -0.0084642087346874089748, -0.017085047116874058082, -0.0057704778774301376462, 0.006391125578717624274, -0.064242470802775161753, 0.27297748433981194882, 0.00081375980857353359416, -0.59631372835465712434, 0.22902769603130382348, 0.32971937177743698699, 0.22641015187688517951, -0.61300311519816608197, -0.040282841280486078595, 0.25299920996641939164, 0.10330035720503777386, -0.03778963751349973671, -0.062619736362388378481, 0.00017076527925550334707, 0.17804039367644924896, -0.052722296043855532144, -0.44470132630340580837, -0.11618964785761570768, 0.84566511936840926822, -0.81961886984412979373, 0.84895344705348296799, -0.10586641581082462116, -0.43311608463031819882, -0.071950716122625588134, 0.084245278046766958191, 0.082473198927334401276, 0.03012117814289343859, -0.023963212430951858628, -0.15456571403405366949, -0.6777204903682002568, 0.13825129407588188091, 2.7281082531539584402, 1.0634782497737360618, -6.2399121260631513763, 1.0612636698712605288, 2.7596288851476913351, 0.23426257151373020005, -0.59504103081183790103, -0.39157054785613376602, -0.076497662623565637618, 0.077758453015343578785, 0.066390779692496307285, 0.084621772156880520344, -0.092348586744091215595, -0.15503707480378753347, 0.028940181049481639858, 0.47711143784012821634, -0.68844201602434584863, 0.47835687918213470882, 0.034639300900419785212, -0.1444644409335652524, -0.086973921100112913662, 0.0041891653181795659305, 0.046782347080681054019, 0.022893860354246873406, -0.010184777246394919117, 0.097496339836339498963, -0.096994648562940904979, -0.17293485482417914056, 0.024977780653760585372, 0.48677669272294993297, -0.67921289420647423807, 0.48677291193471983188, 0.029333285538676513726, -0.16189648089426642019, -0.089857151720629524494, 0.013701715303439174556, 0.054549119639735160758, 0.023283549870500363371, -0.015708498284332985534, 0.091888959627826699572, 0.094461818578186779183, -0.32274322691520740403, -0.31270591486551763127, 0.67992341249854637564, -0.44358592156976228704, 0.67233684673205318383, -0.32657676802165469532, -0.34146272097368957077, 0.079309406571758001014, 0.13275704990594944244, 0.042972049142362737928, -0.021254571216837478409, -0.02424511274548863185, 0.080118222488464396758, 0.084206764202357273219, -0.29620467565421837453, -0.26106050403885172617, 0.66246984086548332815, -0.52427466791536969648, 0.65626483553676262517, -0.27205494826509735162, -0.31159961733475671775, 0.070007069580543451748, 0.11543160945708805243, 0.036857151591083797681, -0.017933950321763897362, -0.021258458832739333261, 0.18706392444112052154, -0.042492174447014537109, -0.2016912522694104537, -0.41125318932816912199, 0.28506011325357050179, 0.38648889203283426363, 0.27047964126737189527, -0.42889088650854034768, -0.20213229642167440669, -0.01735695307075534391, 0.11816952978129016683, 0.11779056377645522047, 0.011163702277604166391, -0.069887985186642201385, 0.089213420699612966391, -0.089102930127741203359, -0.3074229335716747169, 0.32784391102830101605, 0.78332775143571742316, -1.6109714304349536107, 0.78369488847662160946, 0.33417318826754149885, -0.29770424768274528216, -0.092617165163726172805, 0.047689177888219365375, 0.039954814323373756546, 0.0058826014705527836929, -0.013484570718498583089, 0.037570811674445231398, -0.13192503795541701828, -0.10350536179836336559, 0.087104942966282475347, 0.43631693234139096171, 0.47571004557167079208, -1.611083621090673601, 0.47547836792148673313, 0.44111554218773824099, 0.10119170312287369129, -0.087102847204170927298, -0.14457847648124974582, -0.054553370655356821228, 0.075781105990483974, 0.1026438932910516133, -0.079950697255140992792, -0.35834620628413654808, -0.15948226712451668563, 1.155324793130008354, 0.81625387889474720904, -2.9541509068727136444, 0.81314790870509101595, 1.1544839464422669284, -0.15107894117643808074, -0.34249870280177985293, -0.084066818406252630247, 0.049769810628493803273, 0.037854134683443417531, -0.032856633851394215218, 0.053721125425609944637, 0.18153857352491864496, 0.01446523891472152254, -0.7674003623205684077, 0.038251279198835134998, 1.0268816098735114029, 0.041383576909846224445, -0.76624313393289444907, 0.0049905062135986149457, 0.16392488656483994447, 0.05942438495209440702, 0.022775545466559063423, -0.040012814585072414841, 0.072241514690218666672, -0.027338204930754159139, -0.024406115643107174212, -0.057374125378837494138, -0.32255980404667056183, 0.2718923515766464849, 0.17270529368586196028, 0.27273799526879977106, -0.31851062953940334488, -0.048606514891011656787, -0.015828142613770045111, -0.03852207644357709404, 0.010710780938105722918, 0.054535062309177130546, -0.0048565606577894802204, -0.087924516245535364045, -0.12883125505339132477, -0.038612571239641485521, 0.8616736668852211789, 0.75632358113837072633, -2.7238730536298088047, 0.75694727448406129433, 0.8657481126008516048, -0.028380609396895774277, -0.11855129240675606195, -0.10192261415597232699, -0.057371305531929317478, 0.046282454376603790469, -0.21205026683835337398, 0.4268710091524727801, 0.81036210967683097728, 0.46971603059040462069, -1.8470456137723707535, -11.616535350630829981, 23.962076953201552953, -11.612430778757181216, -1.857404115298785019, 0.42688503461947729933, 0.75599497838499862823, 0.44817699884394368581, 0.047637508135995126901, -0.19500339647885539529, 0.0060377074387503873801, -0.22101387387757767677, 0.099274043366183881143, 0.53975949134491030268, -0.3059281858417795541, -0.34447369678309286423, 0.44752226823363239205, -0.34548155500170213417, -0.30630761020822938612, 0.54469420692241810045, 0.1176323663016797666, -0.20356274499135082467, -0.092517143283861208447, 0.062061292760298349125, -0.10346333660901228801, -0.15641328562735629104, -0.030307411871376338519, 0.35981557456525364858, 0.89479592114959749516, 0.31165315186216968613, -2.5680630233842918209, 0.31622025873391784723, 0.90345939390672191927, 0.36854803666805363571, -0.029728670322654977831, -0.17331049924740227208, -0.1259223117315440843, 0.025620296017087460655, 0.052104830167322750689, 0.079567361348413551125, 0.023664406775261975086, -0.28078975373993203535, -0.37005610819029083336, 0.59443845065634948188, -0.18418474117025726566, 0.58673537039373080404, -0.38144860035403865828, -0.28796947231141389878, 0.030910544813541079856, 0.10373543959995956687, 0.075455610862001309957, -0.037786614767075288324, -0.02889334694802561665, -0.071990051297855503498, 0.012592325825742323175, 0.10238492845134331, 0.27349402384784354814, 0.40141053922083391736, -1.3830677983291232991, 0.40054674391708733339, 0.27516016014323452232, 0.10938927729163698543, 0.021652010807982385171, -0.075510160635660336648, -0.06042202508119552673, 0.021408201071826237633, -0.020025424356321007424, -0.078206483884020835951, -0.0019805196026556788624, 0.09839043463347091345, 0.29407411217824802607, 0.40621006587795721599, -1.4022784886922055048, 0.4053451007634109593, 0.29601565472643492916, 0.10622912631600733846, 0.008185781947310351303, -0.082351423663268530362, -0.058726540750605273067, 0.027321848430786645212, -0.014451392572520834964, -0.013900917520155639528, 0.084329582622750878795, 0.063492957862293433746, -0.30856344543564534488, 0.15615000561521555356, 0.064746994618136252675, 0.15539469765509569821, -0.30803811080072068807, 0.066875609953670880903, 0.088545282893229312893, -0.016578535468894980281, -0.027783724280173578008, 0.0097304920911427863345, -0.010292846014175897895, -0.011290271667563817753, 0.075804185641253254979, 0.051164458861707229342, -0.29834995783679024717, 0.16629133372396606716, 0.052840629387653451077, 0.16555629468094476553, -0.298281855417670827, 0.053462442907254340008, 0.079403519313752585207, -0.012578560462788901733, -0.02108588618941014195, 0.007417055617840596915, -0.1268136694250938945, 0.10788066618664698959, 0.27200819319443497735, -0.032021597920971811568, -0.6129733876139528137, 0.22641015191935101591, 0.33235428588996235311, 0.22823278483092462876, -0.61300311474915170074, -0.04074355057145317982, 0.25345853167018433938, 0.10481303777162857827, -0.037409194891644570458, -0.063449022544137187452, 0.073047581094768165855, 0.096048104664596220537, -0.065952839213322733691, -0.43417772038210783458, -0.10875115178125252524, 0.84895344720953846984, -0.81625225418839908453, 0.8512464805924946365, -0.10586641592272424961, -0.43425902990400033765, -0.070899755095396491122, 0.089276396658881471136, 0.081383445270318169329, -4.1424335804381970166e-05, 0.16110388477189199752, -0.40256089819242257821, -0.65944681876896316108, 0.20019607136669878567, 2.7542390782713432884, 1.0612636692559949125, -6.2478076770297308329, 1.058022437791375614, 2.7596288854487953657, 0.22964183929314022081, -0.60930100851628055114, -0.40643320320071096852, -0.065130096560666012051, 0.16124367657865326664, 0.054825783937912861421, 0.01029836115400642807, -0.087062578691645770013, -0.14571078065764608844, 0.034084428608022479534, 0.47835687860355519119, -0.68838644162034856588, 0.47822939862772112507, 0.03463930085082174426, -0.14422300208356325646, -0.086087288586156870962, 0.0063955910614436166373, 0.047589907458907551585, 0.0084364481404959994204, 0.054462343882134375828, 0.019432076689568894556, -0.087818306020889455588, -0.16137535610346243065, 0.029517371000881018062, 0.48677291233801417425, -0.67981290357359547283, 0.48648708145161012606, 0.029333286016155223075, -0.16139886626459412633, -0.088001289871311699775, 0.017490463519003886239, 0.055176281444842059654, 0.0013088334657769657397, -0.04451929223577100192, 0.14526559108097036876, 0.10553027397212019767, -0.32905655548863405047, -0.32503279889341574194, 0.6723368468485314553, -0.44341377588362962392, 0.67451765206710723177, -0.32657676748772057618, -0.34207271021204482819, 0.083912895790832692233, 0.14492839675919494402, 0.046847430140098761386, -0.062191446260369376287, -0.037441676127161704901, 0.12564727678978829251, 0.091832305631903338172, -0.30108128820120755309, -0.27058703855534116389, 0.65626483522576350982, -0.52434877241600641451, 0.6578384670019712388, -0.27205494868437141154, -0.31184952047872038783, 0.074115216419376767076, 0.12553235163221021509, 0.03933426319328067744, -0.052716586229470653591, 0.013105162974805276491, 0.11968304954975433896, 0.013324122591001906143, -0.1796437452144177549, -0.42094401059810399035, 0.27047964126956508535, 0.38055724803524715316, 0.27012518312608901905, -0.42889088693825699883, -0.19992298059758423223, -0.0055133155595619987022, 0.13970172175345746401, 0.11876458184920435368, -0.087934874960845482716, 0.024389950556362127027, 0.054619870388192798272, -0.092466371385954163054, -0.29787261815796955977, 0.33566412717523491649, 0.78369488883660187728, -1.6137954532560514487, 0.78163698125141845008, 0.33417318822392849675, -0.29585237348341603125, -0.089824345033837058039, 0.049176322711265076093, 0.03555964998747478123, -0.0080133749172326598081, 0.075781105945448914984, -0.054553370593448828607, -0.14457847643647359082, -0.087102847452400461448, 0.10119170350141915016, 0.44111554178187978303, 0.47547836833252810074, -1.6110836212375942988, 0.47571004540269973271, 0.43631693254188924636, 0.087104942788800571174, -0.10350536152879907803, -0.13192503830908922091, 0.037570811854348622194, 0.037854134668248745621, 0.049769810667372002544, -0.084066818535286844116, -0.34249870247580371796, -0.15107894160261761685, 1.1544839468052205955, 0.8131479083818683451, -2.9541509065590703109, 0.81625387862452836174, 1.1553247933787602619, -0.15948226733765100138, -0.35834620619762236338, -0.079950697233402520658, 0.1026438932691985384, -0.040012814527246462482, 0.022775545309566578439, 0.059424385021015012354, 0.16392488685837711127, 0.00499050553579632062, -0.76624313310504721386, 0.041383576210654769711, 1.0268816102967386339, 0.038251278989669998398, -0.76740036218492302389, 0.014465238786680871541, 0.18153857367093972908, 0.053721125292588448041, -0.032856633799680109997, 0.054535062382690333016, 0.010710780887779306272, -0.038522076635745153284, -0.015828142252012473296, -0.048606515211193974946, -0.31851062943052627086, 0.27273799521061614648, 0.17270529391406722053, 0.27189235137011735777, -0.32255980389904109762, -0.057374125534315145725, -0.024406115522927564893, -0.027338205057446180596, 0.072241514761621106189, 0.04628245427522242883, -0.057371305443511086408, -0.10192261406430375148, -0.1185512924234680826, -0.028380609413627407717, 0.86574811255201722382, 0.75694727427663388575, -2.723873053358686569, 0.75632358096129936875, 0.86167366736756489942, -0.038612571837669085184, -0.1288312547265858532, -0.08792451641953449859, -0.0048565605783452659924, -0.19500339649865133795, 0.047637508200179437545, 0.44817699867536875402, 0.75599497862422149197, 0.42688503449304987525, -1.8574041152496756357, -11.612430778830768574, 23.962076953151218106, -11.616535350472563692, -1.8470456138661328627, 0.46971603070230871779, 0.81036210952684795128, 0.42687100919206955041, -0.21205026681799152288, 0.062061292708123633188, -0.092517143037309138021, -0.20356274545925553965, 0.11763236679028030163, 0.544694206591844865, -0.30630761005168660738, -0.34548155506998817765, 0.44752226827409269383, -0.34447369673532896028, -0.30592818595464277287, 0.53975949139465506654, 0.099274043291348090801, -0.22101387371336125853, 0.0060377073509320158817, 0.025620296126945350801, -0.12592231183895127855, -0.1733104993064733812, -0.029728670344623696342, 0.3685480366905139471, 0.90345939402928343576, 0.31622025878657522568, -2.5680630236013937129, 0.31165315203036053626, 0.89479592103835314809, 0.35981557453016926873, -0.030307411777041676909, -0.15641328557211281458, -0.10346333668207230438, -0.037786614831711591578, 0.075455611009061673844, 0.10373543952682362512, 0.030910544687603475761, -0.28796947210275847251, -0.38144860047015299864, 0.58673537051952506793, -0.1841847414457671861, 0.59443845079675505794, -0.3700561079787118568, -0.28078975398850858358, 0.023664406864194745955, 0.079567361307512324275, 0.052104830189061847323, 0.021408201069094419389, -0.06042202515012619346, -0.075510160513548224559, 0.021652010832801809465, 0.10938927731836110835, 0.27516015985921882248, 0.4005467442017953128, -1.3830677984459127661, 0.40141053926483599712, 0.27349402357570090638, 0.10238492877036645123, 0.012592325853501026497, -0.07199005138557136585, -0.028893346964939448096, 0.027321848490682081045, -0.058726540829633348817, -0.082351423632735412661, 0.0081857820071615013807, 0.10622912625715871804, 0.29601565445064742521, 0.40534510111473942562, -1.4022784887223611605, 0.40621006600547610033, 0.29407411172785141718, 0.098390435028055500677, -0.0019805197344449267799, -0.078206483896216330565, -0.02002542434003141969, 0.0097304920369291128535, -0.027783724068976143468, -0.016578535712876936448, 0.088545282842235492615, 0.06687561016664331226, -0.30803811079993126398, 0.15539469754596613171, 0.064746994780415401127, 0.15615000530240844001, -0.30856344521093648225, 0.063492957837970792623, 0.084329582520471457752, -0.013900917343221408373, -0.014451392664876789446, 0.0074170556654524011711, -0.021085886316877494595, -0.012578560302868897411, 0.079403519111673870334, 0.053462442984651095679, -0.29828185515531036032, 0.16555629443103400678, 0.05284062937038900426, 0.16629133370539478376, -0.29834995768984706732, 0.051164458680475900787, 0.075804185815766894629, -0.011290271734911578802, -0.010292846019674763147, -0.063449022694091541941, -0.037409194614740140516, 0.10481303760597668329, 0.25345853179034033564, -0.040743550539982728576, -0.61300311497734472077, 0.22823278473214625439, 0.33235428640591979654, 0.22641015129773472969, -0.6129733870630164061, -0.032021598067858147996, 0.27200819296252015667, 0.10788066634483281514, -0.12681366943621746879, -4.1424340265618499017e-05, 0.081383445360908329702, 0.089276396563563550912, -0.070899755256567886796, -0.43425902964946411222, -0.10586641590462346485, 0.85124648064444097262, -0.81625225459798900207, 0.84895344746842915651, -0.10875115180043970486, -0.43417772037837604193, -0.065952839219615352895, 0.096048104799932199072, 0.073047580977521578416, 0.16124367659825858401, -0.065130096610413829028, -0.40643320311952851887, -0.60930100852602131489, 0.22964183904846918738, 2.7596288858092115071, 1.0580224375942368642, -6.2478076769883532648, 1.0612636693442947244, 2.7542390780602441502, 0.20019607151860777106, -0.65944681880675304342, -0.40256089815779655394, 0.16110388474452258523, 0.0084364482750118527377, 0.0475899069278859152, 0.0063955919362339982634, -0.086087289361745852889, -0.14422300179169456058, 0.034639301046902971903, 0.47822939849320506012, -0.68838644220841660459, 0.47835687951058070144, 0.034084428359327227642, -0.14571078093398243114, -0.087062578418968555871, 0.010298360901542609122, 0.0548257840686982717, 0.0013088332182147489197, 0.055176282080592428669, 0.017490462812911108242, -0.088001289471991067659, -0.16139886593853330909, 0.029333285087640239425, 0.48648708205862839504, -0.67981290343804390286, 0.4867729119924336656, 0.029517370871326928383, -0.16137535527791194156, -0.087818307153048308367, 0.019432077479605430748, 0.054462343650077113533, -0.062191446246692212718, 0.046847430156077396468, 0.14492839665952486627, 0.083912895831719264139, -0.34207270988806659506, -0.32657676831029319731, 0.67451765270260488627, -0.44341377568708434387, 0.67233684653404435227, -0.32503279909794979829, -0.32905655499135810782, 0.10553027342113921461, 0.1452655914152832517, -0.044519292301706182846, -0.052716586047780157609, 0.039334262563898728104, 0.1255323524997714324, 0.074115215791955763036, -0.31184952039655278178, -0.27205494826741177805, 0.65783846676779200724, -0.52434877291097181651, 0.65626483601769103338, -0.27058703889626106731, -0.30108128848670667743, 0.091832306214480213802, 0.12564727632159561344, -0.037441675968938922803, -0.087934874931739626103, 0.11876458161116390522, 0.13970172221763402232, -0.0055133158172070784522, -0.19992298072925870911, -0.42889088666420616125, 0.27012518292646492402, 0.38055724797409451465, 0.27047964156789927603, -0.42094401084057192497, -0.1796437452811455171, 0.013324122958278518045, 0.11968304919392322438, 0.013105163095333861867, -0.0080133748768258551698, 0.035559650102531772942, 0.04917632236294901954, -0.089824344879680981246, -0.29585237349012305508, 0.33417318835073783756, 0.78163698114829616159, -1.6137954529120071001, 0.78369488820941402007, 0.33566412746576862469, -0.29787261812596055321, -0.092466371453575571526, 0.054619870542829171423, 0.024389950451618819249, 0.046496013752552446674, 0.015447973011650515882, -0.057693117692106769823, -0.13303970699401707067, -0.079561794714929551242, 0.10291548871584758895, 0.44111554131690788871, 0.47571004587931065011, -1.6129419834488387675, 0.46873730834539945045, 0.4266890263568067776, 0.089039153786720573458, -0.082745331089198376207, -0.10297593314611799831, -0.0031938573209500517244, 0.044171220140717032399, 0.047191891248505923873, -0.085212730211325735108, -0.342422172728959473, -0.15170259369657634885, 1.1544839465278438073, 0.81625387875774801927, -2.9480424394929958254, 0.81909859634872705936, 1.141582892834928753, -0.19571852522034197119, -0.38044106361906859792, 0.082491166237267155026, -0.016466056571734048741, -0.012676030213688798884, 0.017210239012115931961, 0.053408049589137793212, 0.16338812674433367267, 0.0064432583922099497323, -0.76624313311606973009, 0.038251278657545580153, 1.022799218803352117, 0.040958734929844635431, -0.7535012251523799387, 0.026687488413105042023, 0.16477985888214533028, 0.016148950655294976725, 0.015306799963262139597, 0.030945628374074492933, 0.013110549158352297408, -0.032146071399040909711, -0.012965335546475206807, -0.048207016130895265216, -0.3185106295245216379, 0.27189235142066364714, 0.16807130526401312109, 0.26116331836939926658, -0.33500792873580825804, -0.05336016845531781766, 0.0060319858603989431889, 0.033196577181481563834, 0.043182905387297992883, -0.0064942638431199854049, -0.063060142029180588086, -0.094376487156691682423, -0.11135157044903223889, -0.025252603802650302328, 0.86574811261234951854, 0.75632358106930219765, -2.7237428971856258997, 0.75577507551458644741, 0.85848081476650972554, -0.037671354082810408481, -0.11839823512787049642, -0.10167075645725762167, -0.085022835468596999475, -0.093878166894152126343, 0.068798815441935515236, 0.43055653753621925262, 0.7368531676732784641, 0.41975908311267229323, -1.8574041152420668332, -11.616535350467332322, 23.950061318140260624, -11.619940944366813795, -1.8089573989850915581, 0.55319049490324223584, 0.82398363890172465762, 0.10691649081191557935, 0.03298492684988719309, 0.016573857992579644344, -0.092085528466467336806, -0.19458844459462779075, 0.12174198552650059013, 0.54503239213968857069, -0.30630760996965716858, -0.344473696855126299, 0.44838681483151798801, -0.34585690147371878078, -0.31004412892404586977, 0.53862596210132906993, 0.10124274581104912574, -0.21303201309731220614, 0.070287587027236836379, -0.05510238180405458569, -0.14440997577262207341, -0.16209218959779669866, -0.014256469294148610763, 0.37667574270324477581, 0.90345939401226693644, 0.31165315190067344187, -2.5720368623826890264, 0.31575158831500566015, 0.91314440705581778612, 0.38413969385995466732, -0.051699773518479541501, -0.27885992850535862297, -0.063743116321594489593, 0.031337597409543636873, 0.094702196530391122287, 0.095201259938935911542, 0.016200074590797499535, -0.29714742724141068875, -0.38144860042396716571, 0.59443845079912471796, -0.17426058944219818825, 0.59674399315653670062, -0.38714385558262875975, -0.3108922562670659584, 0.022060504987672702182, 0.16669496308521386529, 0.024900912900493538116, -0.015579156672076620305, -0.059654269671068745062, -0.067233903173570280787, 0.026066060787123149917, 0.10994026942795434576, 0.27516016011604915725, 0.40141053924200115199, -1.382868416855040028, 0.39927791037808701979, 0.27128350121191968647, 0.1090612197916221715, 0.019608833955200202781, -0.11218724052403912517, 0.026285027535601562104, -0.011381351673180623713, -0.05805514968100698725, -0.073714334412624593629, 0.012845502858690589082, 0.10683203617762106197, 0.29601565409344865465, 0.40621006637599821332, -1.4022359715294023808, 0.40347807854241679593, 0.29066091270735666674, 0.10388423344575237195, 0.0067094130856779499866, -0.10849464481319569253, 0.0095432919258137283852, -0.008504831311002632524, -0.023465813282725726807, -0.01106306216658609759, 0.089506943368959090979, 0.065915440319069568131, -0.30803811063092534361, 0.15615000540963866471, 0.063502471013204914163, 0.15131417740460506582, -0.31194416147016235108, 0.074939157251794286108, 0.098153554322725808379, -0.045933634181235721838, 0.0063911255495047135378, -0.0057704778071090015792, -0.01708504720813361008, -0.0084642085795539750981, 0.079733360080644688495, 0.052487202789620615584, -0.29828185511304661226, 0.16629133366743367728, 0.052070223991940618735, 0.16251625517022544321, -0.30150505642433883535, 0.06007739311660587056, 0.086879757482827738024, -0.035227478104396796332, 0.00017076527901352438416, -0.062619736249430082986, -0.037789637710229542766, 0.10330035716827090653, 0.25299921017779064547, -0.040282841499933663698, -0.61300311481939573355, 0.22641015159500066378, 0.32971937180244920151, 0.22902769593456054342, -0.59631372829983875228, 0.00081375983546615280207, 0.27297748448560527068, -0.064242470927778433376, -0.023963212365369111062, 0.030121177916752672754, 0.082473199122423845875, 0.084245278186300495693, -0.07195071639500937033, -0.43311608455977490539, -0.10586641590374083755, 0.84895344736041156164, -0.81961887004454081485, 0.84566511937354527095, -0.11618964779438038792, -0.44470132635517667374, -0.052722296177336813727, 0.17804039381085159932, 0.066390779633686503036, 0.077758453250154568881, -0.07649766306590696785, -0.39157054742960328841, -0.59504103092639504258, 0.23426257137114539497, 2.75962888536675921, 1.0612636695083872418, -6.2399121256484288978, 1.0634782496420180919, 2.7281082530183273782, 0.13825129425698584607, -0.67772049052805827252, -0.1545657139613356712, -0.010184777208642139049, 0.022893860448090618681, 0.046782346739131869251, 0.0041891656767995590072, -0.086973921534562015356, -0.14446444046918871518, 0.034639300845414937313, 0.47835687907899365756, -0.68844201621988143636, 0.47711143818656287507, 0.028940180507036325497, -0.15503707408387362721, -0.092348587164812950823, 0.08462177222948254196, -0.015708498267403569332, 0.023283550064683852104, 0.054549119067676662964, 0.013701715875609629669, -0.08985715194489393709, -0.16189648107686624434, 0.029333285800697557477, 0.48677291215482204612, -0.67921289470791534892, 0.4867766929853434843, 0.024977780530318353241, -0.17293485472203340425, -0.096994648552667997454, 0.09749633980032922087, -0.024245112876732529544, -0.021254570802119267248, 0.042972048599551018744, 0.13275705028785214634, 0.079309406455218098286, -0.34146272099905244923, -0.32657676802310409148, 0.67233684687211492381, -0.44358592199022300884, 0.67992341298230318802, -0.31270591507542822018, -0.32274322683916162324, 0.094461818535207950354, 0.091888959620067781064, -0.021258458864956895212, -0.017933950256824343628, 0.036857151571699206527, 0.11543160947665359606, 0.070007069449974604347, -0.31159961715647183089, -0.27205494831911408715, 0.65626483536700586097, -0.52427466782548526325, 0.66246984118741214509, -0.26106050458396173886, -0.29620467511501247238, 0.084206763831050121527, 0.080118222595959409649, -0.06988798521461715707, 0.011163702349419065904, 0.11779056366925483301, 0.11816952988438324257, -0.017356953187704072389, -0.20213229623656531198, -0.4288908865606609333, 0.27047964107981065096, 0.38648889217177417876, 0.28506011334111996947, -0.41125318958546441817, -0.20169125205759966257, -0.042492174437380160479, 0.18706392438041291593, -0.013484570882860086663, 0.0058826017080565165193, 0.039954814266827551328, 0.047689177877456821741, -0.092617164898347717483, -0.29770424814663176694, 0.33417318840711590644, 0.78369488834546063849, -1.6109714301766528965, 0.78332775152343869784, 0.3278439109243779237, -0.30742293366901030094, -0.08910293019108396495, 0.089213420802190412928, 0.0075214405822400633508, 0.036209677794314977517, 0.016187523256997789906, -0.05505738211546189198, -0.13172721966189160359, -0.079561795237170163109, 0.10119170302586601162, 0.43631693289000406111, 0.46873730873292285271, -1.6149097857526717448, 0.48820778238150164263, 0.4666928723494688791, 0.10660674368458333827, -0.24801875389909569525, -0.013376318391607832856, 0.016248123640718314942, 0.044020120486911909485, 0.042262064097133879292, -0.087783135644356025118, -0.34242217349277753735, -0.1510789406842492355, 1.1553247927652747773, 0.81909859692232400707, -2.9442450412487644051, 0.81841657043123228199, 1.1158128098842956533, -0.2407477719318733278, -0.23412285609166608169, 0.0050826607312375047046, -0.023161591521742002503, -0.014383865405615835209, 0.019812512145742578601, 0.055448161108057794633, 0.16338812629252524866, 0.0049905060184219102865, -0.76740036218989671202, 0.040958735059033697656, 1.0281112307232420644, 0.033123418984159179457, -0.77987523035353201184, 0.065050263124336538123, 0.16971045960836431132, -0.013194292768754471365, 0.025561560812708491314, 0.037137342386162564001, 0.014816969169195601252, -0.032327118788032162267, -0.012965334802144554138, -0.048606513531535389117, -0.32255980549217744269, 0.26116331761938466061, 0.15377389863537324621, 0.25847037306568410031, -0.31517241232209186608, 0.0045313362232675524072, -0.013933676203710082531, 0.018120277930650813963, 0.022497177085267709706, -0.0095831908578387515335, -0.060274930095269477204, -0.091572419339204569333, -0.11135157066061178444, -0.028380608783916502769, 0.86167366644169618439, 0.75577507593825254961, -2.7141840041722695709, 0.78435112068259960072, 0.88456503248259310546, -0.088269519532029480646, -0.22329119860501656092, -0.0061085233754325099845, -0.084173876922151308344, -0.087612623001641951048, 0.070602054122011301263, 0.42823918624222395701, 0.73685316753988183702, 0.42688503493114315912, -1.8470456143343216837, -11.619940944069277577, 23.912071282243264392, -11.667729042679203744, -1.7860665212509581057, 0.64021827894913707624, 0.88860522358239157903, 0.0052948012738071020128, 0.030216716028862642546, 0.01203216307227596843, -0.092196527625670307571, -0.19349331797681582068, 0.12174198553392123756, 0.5446942067037514601, -0.30592818589321213452, -0.34585690140702807227, 0.4410284052519309328, -0.34703770414362344887, -0.25894585169748757236, 0.66377363074483863858, -0.27433788445480067342, 0.052302630059140937935, 0.008304853787506871407, -0.064480012553279522636, -0.1341753658695738205, -0.15299392170788700396, -0.014256469101432654722, 0.36854803678772735109, 0.89479592089446169378, 0.31575158809132730653, -2.5419896758662949843, 0.3730195348805448452, 0.95077289062072167791, 0.30984085115985160463, -0.36127698079629899652, -0.040855695069794202567, -0.01188794900175553157, 0.036793055956058244582, 0.083191948211315314143, 0.085472066139555852304, 0.016200074574399848937, -0.28796947245411463889, -0.37005610798460092381, 0.59674399340615580378, -0.1935011318616423337, 0.55465499916527971624, -0.4282082102627451925, -0.26492156186720799971, 0.22346383724669927062, 0.0095457408828036258153, 0.011737306173794669062, -0.015490031862957524475, -0.057304784455404329802, -0.066244511243250675525, 0.026066061078274613966, 0.10938927635515265091, 0.27349402462465965957, 0.39927791006570240162, -1.3804121287612689528, 0.4173814615774182335, 0.29713284758009361752, 0.10445790236333704626, -0.12746066068735015242, 0.008390381170419885376, 0.014409509006465114067, -0.011034046018233146375, -0.055756096221372772326, -0.072779766676160789762, 0.012845502460124034322, 0.1062291266448133209, 0.29407411155020229954, 0.40347807884691527791, -1.4006395702361327249, 0.42105599794398679281, 0.31786108743851515523, 0.10245890446980443489, -0.13938065550580250318, -0.00028242559914775278076, 0.0068380794970985640802, -0.0054702375480503119196, -0.022182332988343531066, -0.011543835549528213419, 0.089506943297558497208, 0.066875609704418637058, -0.30856344518914924357, 0.15131417778637687443, 0.057272809379028719046, 0.15645543350885132705, -0.29408659783471863269, 0.10233941086726795755, 0.011946371423966385336, -0.0011262165659382704017, 0.0055514525369176111885, -0.0033751810652061131797, -0.016400974969502778861, -0.0090481773247143276045, 0.079733359863404351664, 0.053462442761430799265, -0.29834995755891113856, 0.16251625531667363012, 0.046261355090007560964, 0.16531743210308214032, -0.28802460108300070862, 0.087333971353796654991, 0.016447952872020409065, 0.022484420290945272836, -0.027980489435876935989, -0.064433629113534479638, -0.033678461502229931068, 0.10537864153090024943, 0.25299920964955613378, -0.040743550845314537023, -0.61297338723053840237, 0.22902769679170875694, 0.33887761685673611911, 0.24447631397775515527, -0.59978632740393156286, -0.030646657061618363949, 0.22069948747042952841, -0.021620845887824164527, -0.0030103077248167576553, 0.035304755610395026355, 0.082930802218443122609, 0.084226669472088885882, -0.071950716150671958915, -0.43425902957610018618, -0.10875115196816741159, 0.84566511937531074761, -0.82519098577687122198, 0.83038902191245400619, -0.14894672657253385517, -0.44374662456329028171, 0.17579693697720838452, 0.00053146655782508166882, 0.072252452342624229331, 0.071098754022181021028, -0.078536066268231466414, -0.38970027568753018432, -0.59504103140757147461, 0.22964183961818909752, 2.7542390779388723487, 1.0634782497804089463, -6.227849512519438413, 1.0893169809591620734, 2.7682148053021196077, 0.10171691226533206842, -0.86222495746037963738, -0.011070263949171972248, 0.0022292068746804553046, 0.025171529649512004723, 0.0454792960668709903, 0.0031049012941846976243, -0.086973921090895800479, -0.14422300183006298013, 0.034084428790043835855, 0.47711143709704656457, -0.69075377152578187889, 0.47036541262680209297, 0.0072609737633810272212, -0.14838389909361984764, 0.014589477939879717458, -0.012248569314177356823, -0.001209380023667180979, 0.02515310966164324008, 0.052581755779868247225, 0.012329122667440890401, -0.08985715070895920753, -0.16139886698950792199, 0.029517371312868974365, 0.48677669256418742982, -0.68096757016757292202, 0.47741157899134212261, -0.0028071649464051409012, -0.16856239227165492611, 0.031366571838759876867, 0.0079594747283873203136, -0.030201865925354322295, -0.025178303724534764618, 0.042440628130327680279, 0.13346557172868550589, 0.079309407053120159481, -0.34207271002528644166, -0.32503279942205864916, 0.67992341287183788445, -0.43337441695746897752, 0.67531247008774675233, -0.35802655665151367215, -0.3654444271405027922, 0.26141375723196452752, 0.0057292473300917844473, -0.025263726875883546485, -0.020905823127010621232, 0.03625078666901505875, 0.11575730940351262876, 0.070007069715353031913, -0.3118495206520148777, -0.27058703809154449615, 0.66246984011727216846, -0.51621071416646413521, 0.65828671900375734793, -0.2995169681410337903, -0.33238585249167285074, 0.22864039743439407149, -0.014799138407561390762, -0.044263501880806660482, 0.0063461513732044165514, 0.11020226148581241388, 0.11540525969064664713, -0.017356953412971321687, -0.19992298157429719274, -0.42094400956188604557, 0.28506011319162011253, 0.39279460010502859069, 0.25100690675260384221, -0.49969095474204772156, -0.20520564177267619499, 0.2415164540155733397, -0.010392999407677992046, -0.0019822348416623971484, 0.0091787587397346905721, 0.037478076075497662734, 0.045148805575457674755, -0.092617165183632652048, -0.29585237357194371643, 0.33566412736646861203, 0.78332775185178982191, -1.6123298778211025972, 0.78240965534583561958, 0.3071826629311357193, -0.34884254940515618815, 0.060742975842133971331, -0.011634147119866945427, 0.021681129773600135713, 0.038321203721168574374, 0.015307857451139049096, -0.055057382978403464346, -0.13303970685401947383, -0.08710284708984876445, 0.087104942645003874402, 0.42668902637066169481, 0.48820778306609924346, -1.5562579307044370758, 0.56222049543574614816, 0.41191856268098464566, -0.20581695619713960554, -0.0080687394829518158113, -0.0021217389985671380147, 0.016117348565227271329, 0.041279309956137634485, 0.042262063957925970203, -0.085212729561505870524, -0.34249870335439086544, -0.15948226663328066155, 1.1415828921473356594, 0.81841657076641682877, -2.9237539473912383947, 0.88404041263733801337, 1.2170285905758346789, -0.63959560182493457869, 0.0085594232875918340592, -0.0083458281584331307779, -0.021461818176990613699, -0.011270508578555013188, 0.019812512478625809997, 0.053408049656301845309, 0.16392488674106642232, 0.01446523873350239256, -0.75350122518939177674, 0.033123418788665578982, 0.98197880994422848122, 0.0035640811571828827885, -0.79684509761028010644, 0.31097622388413320715, -0.020332333552614722388, 0.0069421399809249974278, 0.030948954028843459613, 0.037199095135471621498, 0.014816970822539077268, -0.032146070623367370422, -0.015828143101217138217, -0.057374124796859007902, -0.33500792956723285387, 0.25847037154314467244, 0.17105708108751688235, 0.29809274712956657094, -0.3288028149267202549, -0.032813898661854079553, -0.00030126325738122865369, 0.019740718869991989021, 0.022237212912115222885, -0.0091261024697067167494, -0.060274930111514649078, -0.094376486682231044978, -0.1185512929791842085, -0.038612570914462719662, 0.85848081412272625457, 0.78435112069474111074, -2.6523506507917073982, 0.82511957902360910655, 0.79561328909012274391, -0.32702736706021046098, 0.022508443669508250429, -0.036813795075729217887, -0.086785344511791692734, -0.084115443800364522731, 0.070602053991256075172, 0.43055653813844807498, 0.75599497767561152362, 0.46971603156583030536, -1.8089573996775438758, -11.667729042455194488, 23.682253547323735887, -11.946334245258659834, -1.5435132046166004649, 1.7316789487479276044, -0.0041370764584677307413, 0.013211559961253283024, 0.028530840779443972927, 0.010365765034237361952, -0.092196527374394696142, -0.19458844463648955481, 0.117632366519456244, 0.53975949142091628197, -0.31004412907456035953, -0.34703770402627109881, 0.47106387179282349642, -0.23389381099597239544, -0.2749551199061900375, 0.28167584916679228035, 0.017336256050361622399, 0.032444813401762487293, 0.0079040441360106687002, -0.059791368435038884377, -0.1341753657090306584, -0.16209218971521646702, -0.029728670430097400845, 0.35981557482099568723, 0.91314440707919031226, 0.37301953460110909289, -2.4801307393801423906, 0.39433682147604037782, 0.86438565022626134837, -0.083271076528662377192, -0.0062703256005084044913, -0.028392779721109560886, -0.016512078252106380388, 0.029780630825725563221, 0.083191947721373601943, 0.095201259981491481721, 0.030910544957842411157, -0.28078975396823951938, -0.38714385578090709794, 0.55465499963515185744, -0.2319677211283385776, 0.5653143166779901474, -0.40924128978105367427, -0.0045457862783437729542, -0.0010295698841510385509, 0.012494769927468795528, 0.01115877030852549176, -0.016223483488182994378, -0.05730478478318514618};
	std::vector<double> l2 = { -0.067233903975391098862 , 0.021652011859752764295, 0.10238492748247375597, 0.27128350219744445937, 0.41738146175200485644, -1.3424440989903199117, 0.44795882340127396448, 0.22271970303633334409, -0.018268463424043129145, -0.0026267277255316761123, 0.013037500286180173786, 0.014196693021784157623, -0.011835031552410661682, -0.055756096900484428125, -0.073714333936345882714, 0.0081857813553835000187, 0.098390435399811057571, 0.29066091241191999162, 0.42105599860707226822, -1.3605839884168116072, 0.45716949756140423666, 0.24729557567149934583, -0.041258670907620406676, -0.0046928164142939895673, 0.0057162630848140031162, 0.0072446170199654382274, -0.0063451873770994527552, -0.022182333147267262879, -0.011063062404227979457, 0.088545283070384339652, 0.063492957765881791188, -0.31194416136664920858, 0.15645543379531212658, 0.068721431396413137915, 0.16307185205512467197, -0.34296342856292638679, 0.14632806370640269455, -0.0038454669755386187112, 0.0039255191538106071897, 0.0056844731127104275795, -0.004221540763383711467, -0.016400975279827780667, -0.0084642086661896719785, 0.079403519291626145638, 0.051164458694523448634, -0.30150505628436974348, 0.1653174322342376712, 0.053935726516068052683, 0.17219850691039628732, -0.32878348112157240735, 0.13171884372787545647, 0.017916069985807761555, 0.0019071599898089371377, -0.030443569392403134477, -0.0623169764710008664, -0.033678462076754636656, 0.10330035714114760292, 0.25345853184488309484, -0.032021598029270716135, -0.59631372843702645792, 0.2444763147264739922, 0.32306330669443539261, 0.15891053064695778718, -0.73329875710255487675, 0.38764695168414020632, -0.013162252921849984522, -0.012363197290280448365, 0.0030618061083350824741, 0.03756250014318059105, 0.082930802291756283706, 0.084245278124035635758, -0.070899755442531284255, -0.43417772005718863193, -0.11618964802966183503, 0.83038902194008679114, -0.84096552906694577167, 0.82519720401816487598, -0.16438093022948560562, -0.21895717857576049714, -0.019031049901844138833, 0.02708436211162495999, 0.073432152785319587673, 0.067887605229486419622, -0.078536066488660502549, -0.39157054713630029541, -0.60930100884832116837, 0.20019607159715413491, 2.7281082528449052127, 1.0893169812950189712, -6.1003411468833181885, 1.3256351398204084813, 2.8559034539961509758, -1.1586192526756569077, -0.0066778966913387579479, -0.0049409710710259860389, 0.004329165119352159502, 0.025012319528340338365, 0.045479296220678991358, 0.004189165460732230048, -0.086087288672950665336, -0.14571078126390415997, 0.028940181223038201341, 0.47036541173690099571, -0.70137832121209331415, 0.48482888343339441972, 0.0429477037041283774, -0.16457317310155256118, -0.0052276703756337916662, -0.007622479280372356121, 0.00045346027597402980072, 0.024919212031256934153, 0.052581756649425985539, 0.013701714806026133245, -0.088001288483914433258, -0.16137535671695957262, 0.02497778120172725741, 0.47741157871776196409, -0.69753131279107349894, 0.48804485683821874353, 0.040749018704661257861, -0.16654112845145016553, 0.01597475189709017182, -0.011929850203192647259, -0.031974382732116536643, -0.022814688115927479056, 0.042440628688772991106, 0.13275705036328500075, 0.083912895457484676376, -0.32905655473419620494, -0.31270591537237940605, 0.6753124697966712553, -0.47189580422491328449, 0.60276039806477521577, -0.42740107446404507208, 0.052132663612773862127, 0.012995280849616335297, -0.01032469657833373286, -0.026819078280296899941, -0.019149901346307098005, 0.036250786908301732558, 0.11543160928385283459, 0.07411521640006840661, -0.30108128879348716733, -0.26106050372237876722, 0.65828671785324599508, -0.54713044457175818991, 0.59809781910656190096, -0.36804996646829857276, 0.036245514112966968623, 0.01694868789698587494, -0.034301595445265509188, -0.048133393845485726581, 0.0067551106352699884575, 0.11020226101249705475, 0.11816952934675697962, -0.0055133150214108830184, -0.17964374610173752811, -0.4112531887657380758, 0.2510069069974331657, 0.30001464980185521547, 0.20082172597781125511, -0.3756181524382211645, 0.046216865086289304898, -0.0038354326992472323292, -0.0050638907297188354151, -0.002337096846078323166, 0.00751151490709130263, 0.03747807614858172931, 0.047689177887198959394, -0.089824344931205599152, -0.2978726182244677001, 0.32784391097481768718, 0.78240965537259521412, -1.6075343077910295708, 0.77402123203663120155, 0.28341598935298173423, -0.25592719468056657206, -0.012251136241141390854, 0.0023811728628126346829, 0.023695501539717011424, 0.038321201316812714133, 0.016187524581899273568, -0.057693118302662514674, -0.14457847628375228299, -0.10350536146243941027, 0.089039153819595553974, 0.46669287141360699112, 0.56222049706146914883, -1.5589498785234201517, 0.46615066745681038585, 0.21994393437935358393, -0.0010138433987520046719, -0.007038992192479818534, -0.0028584082836468093815, 0.016117348845769798038, 0.044020120106646097591, 0.047191891334236547861, -0.084066818458594746999, -0.35834620616690060491, -0.19571852518433363599, 1.1158128099108048925, 0.88404041253500664777, -2.721672820990574948, 1.1345933570201647278, 0.1363991633050259944, 0.0028524335311304977944, 0.0042880336637832715263, -0.0070676231749676327043, -0.021461817462205972501, -0.014383866505329311741, 0.017210239919818792753, 0.05942438414839162214, 0.18153857436792175672, 0.026687487847383617456, -0.77987522958611843471, 0.0035640805425810498201, 0.94707662691010363609, -0.14716770828218009304, -0.27830742975662403627, -0.013577970283353538519, -0.0068408440107794940555, 0.010052548336036136109, 0.030948956511621536491, 0.03713734098328609623, 0.013110548563786262113, -0.038522076863812079739, -0.024406115211799861431, -0.053360167772599102765, -0.31517241100759130212, 0.29809274463977758485, 0.17973530848335370069, 0.14725705071614939023, -0.26743975215326781791, -0.0067459420591889605878, 0.0075336408531172177963, 0.021385500761668563285, 0.022237213097256260802, -0.0095831909319771656913, -0.063060142177206734981, -0.10192261407286826169, -0.12883125493104727322, -0.03767135411167100989, 0.88456503297875177516, 0.82511957856986695159, -2.6601856609749678917, 0.97849744720533415876, 0.27868790438136731291, 0.019796272977889370348, -0.0011462452730137263728, -0.040883126032689912077, -0.086785344195150632141, -0.087612622946951435043, 0.068798814789234288836, 0.44817699970173935364, 0.81036210853874890425, 0.55319049570332923249, -1.7860665214763473685, -11.946334245414126585, 22.951724945661773347, -11.86431228002356697, 0.92480887598129624116, -0.0040967548588514068539, 0.0010507495992052649607, 0.013538995711342443132, 0.02853084117119699184, 0.012032162636757602267, -0.092085528283956455398, -0.20356274535553173233, 0.099274043357937089005, 0.53862596221383007844, -0.2589458518723549707, -0.23389381089739316733, 0.41636368127727496624, -0.79794498801044644321, 0.48633999229103963557, -0.0057770141581805030048, 0.025638497823769951234, 0.036081294755072358671, 0.0079040444391144035197, -0.064480012477052692232, -0.14440997579032530118, -0.17331049923580860161, -0.030307412096576898108, 0.38413969415030796428, 0.95077289056979152893, 0.39433682123328611668, -2.5326198200580560815, 0.34878808282314177625, 0.81939499029191320556, 0.014067348992978163588, -0.020294812261472911646, -0.035524625591363150068, -0.016512078743724213636, 0.036793056647523303404, 0.094702196343560973713, 0.10373543960167466715, 0.023664406644988478234, -0.31089225596654845862, -0.42820821093029654003, 0.56531431717460423059, -0.17044407557158616773, 0.66552555442559968579, -0.53005467740438738655, -0.0029038220103943263722, 0.0042218303933515659299, 0.012296856994888961251, 0.01115876964816661307, -0.015490031369796260913, -0.059654269113245603218, -0.07551016104973012899, 0.012592326244740560323, 0.10906121938259859261, 0.29713284713564042105, 0.44795882405788162162, -1.3825536317768523098, 0.41008310552842430763, 0.23685815664980333639, -0.0039188628415368909064, 0.0036072020809726399425, 0.013070496220049691166, 0.014196692435646158376, -0.01103404507884254114, -0.058055150448219086656, -0.082351422682939726116, -0.0019805205570768893292, 0.10388423416287506651, 0.31786108638875365529, 0.45716949833790054836, -1.3950153318298958993, 0.41493300863021276292, 0.23301342072515601744, -0.0027926703929060823703, -0.00041758413792159746088, 0.0053619013841272435486, 0.0072446167554270218814, -0.0054702374592634651213, -0.023465813135153577862, -0.016578535777195743672, 0.084329582836106531385, 0.074939156859687522383, -0.29408659800289499442, 0.16307185250876782279, 0.0230808259984636166, 0.031441417733276383339, -0.047493370123753916867, -0.0019024423046924959085, -0.00085682125512744316664, 0.0035029263673334730574, 0.0056844727861091451024, -0.0033751809258816493214, -0.017085046900443649764, -0.012578560561035220799, 0.075804185804430490592, 0.060077392848242358592, -0.28802460097384147142, 0.17219850702492189853, 0.017883506000267418434, 0.050765458050330468376, -0.06309022903709822494, 0.0069144351451902088396, 0.011546051847228995177, 0.00058011510460260172495, -0.030443569996840488812, -0.064433628986120497495, -0.037789637297585207398, 0.10481303745507052072, 0.27200819344615467799, 0.00081375920133540629968, -0.59978632761905548154, 0.15891053138305991088, 0.18225737732273766811, 0.17466885526404649087, -0.18301959162446571061, -0.0060727162490721272106, -0.0088738337687347835364, -0.0089014061243248791128, 0.0030618062665748926182, 0.035304755878154653326, 0.082473198810991077412, 0.089276397092523676613, -0.06595283978681308612, -0.44470132603408896177, -0.1489467266095603204, 0.82519720368832349955, -0.78114357087927965839, 1.0460525125697046622, -0.62336455009544222694, -0.014645934821276638052, -0.0014003245756039073028, 0.030206930079758984031, 0.073432152616432266878, 0.071098754029984292946, -0.076497662817097911159, -0.40643320337334415404, -0.65944681863811682732, 0.13825129434827213015, 2.7682148048828691955, 1.3256351401670056767, -5.8523911650979005827, 1.4933912225455596712, 1.1377177377519609447, -0.0023908863909661418924, -0.0050407280992505734338, -0.004225155469803162922, 0.0043291651022621027017, 0.025171529927893933343, 0.04678234675149570837, 0.0063955914196447763431, -0.087062578771652521237, -0.15503707396444857403, 0.0072609743094569556207, 0.48482888281571084876, -0.62106791918662873631, 0.53194887705664761768, -0.23335535846961308337, -0.00069528426953527312174, -0.0055178898674688187792, -0.0072572280113211585367, 0.00045346093518139774653, 0.025153109031442851679, 0.054549119278478677697, 0.017490462640715003645, -0.087818305659654524731, -0.17293485546611925274, -0.0028071650667577047536, 0.48804485712297090982, -0.60892098920121306982, 0.55287554110112713257, -0.25448880755211955584, 0.0086635295417685342229, 0.0044802544918635537133, -0.012501289204082468437, -0.031974381916841244922, -0.025178304218931600589, 0.042972047889504566553, 0.14492839754958439236, 0.10553027318047263072, -0.32274322648800773949, -0.35802655656858051181, 0.6027603978607742885, -0.49615529206478015123, 0.83267124522589242019, -0.50002143252659581041, 0.0074506544790194613806, 0.0032985106241373530686, -0.011007347306636917905, -0.026819078093222255971, -0.020905823321914290824, 0.036857151830171887286, 0.12553235191052644626, 0.091832305730087743223, -0.29620467524815830007, -0.29951696692197837812, 0.59809781808982120932, -0.57031179334403114822, 0.8068162579584311489, -0.44908750134079217231, 0.018731995678563866103, -0.0058946307939745723817, -0.037242620964516477688, -0.048133394634131160261, 0.0063461515376951925863, 0.11779056437419288539, 0.13970172178017661868, 0.013324122868495196642, -0.2016912525598030248, -0.4996909546627757992, 0.20082172647735141435, 0.43576240690721779192, 0.29327026278961237926, -0.44052976030216850489, 0.001702058728808816682, -0.0051513546251295942724, -0.0065398483489689648515, -0.0023370966029773914199, 0.0091787583935409033931, 0.039954814493573791168, 0.049176322250182383233, -0.09246637091548710885, -0.30742293417363725228, 0.30718266319996140323, 0.77402123189227134414, -1.5572930691740451969, 1.0426610569605685441, -0.25295042983461574204, -0.0051694917763198339547, -0.0067939371849181666094, 0.0023811702892822354805, 0.02168113241987886311, 0.036209676357448913886, 0.015447974699510352153, -0.054553370391224334512, -0.13192503895664101088, -0.082745332163495069477, 0.10660674457590166297, 0.41191856184314740164, 0.46615066835152585423, -1.6413901482928192621, 0.86923964408940579407, 0.0020675038421043719604, -0.0040406929925833785788, -0.0070389920990494552616, -0.0021217393802339563902, 0.016248124115389065564, 0.044171219835766932504, 0.049769811100627392442, -0.079950697806098106057, -0.38044106329229687713, -0.2407477722474634374, 1.2170285908974209921, 1.1345933569135404628, -3.4254308124291328319, 1.6842684178354656588, -0.0013007752495010288865, 0.0049953850899762822571, 0.0042880339123834371531, -0.0083458288372013876566, -0.023161590496088527158, -0.012676031265001498974, 0.022775546316799580304, 0.053721124486371743534, 0.1647798596613503086, 0.065050262260047550522, -0.79684509688800886362, -0.1471677085913301597, 1.4786690018435459937, -0.80828905273245754692, -0.0031459466999968913395, -0.0093558577352222116463, -0.0068408424121862206752, 0.0069421395464950774387, 0.025561561826798570585, 0.03094562601601435689, 0.010710779907760537011, -0.027338203997009955654, 0.0060319881537190620652, 0.0045313351978015035254, -0.3288028145466672103, 0.14725704916463369654, 0.36541314144226488825, -0.21686875023635029947, -0.0055627700575125007057, -0.0013453761010011645871, 0.0075336409654152324336, 0.019740718795967778648, 0.022497177331615318685, -0.0064942641291392301517, -0.057371305528011909414, -0.087924516337965399893, -0.1183982344747389337, -0.088269520367602730126, 0.79561328949059550997, 0.97849744710964581262, -2.9022793334563932177, 1.4493484758409265201, 0.0067854775086093792671, 0.011511637311408765005, -0.0011462449529941382866, -0.036813795484737302233, -0.08417387674200745018, -0.093878166697899292914, 0.047637507487651761795, 0.42687100997941557212, 0.82398363841287358689, 0.64021827900559913349, -1.543513204379323378, -11.864312280262062416, 21.017143400930770269, -9.3873905091533380585, -0.0019028896322338657941, -0.0027658648779929797916, 0.0010507498371118222623, 0.0132115595689354208, 0.03021671632078469158, 0.016573857800213899483, -0.092517142959846018768, -0.22101387388233353914, 0.10124274580942214941, 0.66377363090812468993, -0.27495512006607342181, -0.79794498796112967032, 0.76125984101101873414, -0.1983158380424927536, -0.010827264255356015485, 0.0073108518240678922046, 0.025638498411345404804, 0.032444813134284711253, 0.0083048535659388869612, -0.055102381951633826185, -0.12592231192400019091, -0.15641328523678035833, -0.051699773584639903412, 0.30984085150978718515, 0.86438565006126455437, 0.34878808265998806348, -2.6743467310593693753, 1.4771416247333801763, 0.014505278732178741169, -0.0028265814937872167298, -0.020294812271048668501, -0.028392779618744784509, -0.011887949432361526875, 0.031337597806556052871, 0.07545561087982398385, 0.079567361562490737659, 0.022060504452141190684, -0.26492156140587130864, -0.40924129003562897955, 0.66552555459577900088, 0.080294560254463451243, -0.23306912748448954664, -0.00081942500970759901864, -0.00099108684790870523428, 0.0042218301757871749477, 0.012494769945462745608, 0.011737305784499751579, -0.015579156324863018684, -0.06042202462114755479, -0.071990051680451236371, 0.019608833301057455867, 0.10445790295263109038, 0.22271970264402282669, 0.4100831060593090327, -1.2899208958121817226, 0.65348668487141736705, -0.0010678537752219819951, -0.0017486087402317801887, 0.0036072019336508727438, 0.013037500444928117396, 0.014409508285921664508, -0.011381350745134796026, -0.058726541181615111997, -0.078206483456177727787, 0.006709411964342019627, 0.10245890545864924648, 0.24729557506934501943, 0.41493300910816677263, -1.3272552262739665263, 0.67577073739305504319, 0.00044417367034909293038, -0.002480181403991947673, -0.00041758432150984200245, 0.005716263011651801125, 0.0068380793273759548495, -0.0085048308091419839994, -0.027783723786709395298, -0.013900917943724494866, 0.098153554129560505181, 0.10233941102192725681, -0.34296342859095241273, 0.031441417935335697065, 0.33711736328245461847, -0.18719463158347179887, 0.00058890429986079701179, -0.0020502473150364702685, -0.00085682138882407082869, 0.0039255190905958768643, 0.0055514525029235822426, -0.005770477660599888578, -0.021085885869700921047, -0.011290272018277751342, 0.086879757357283926567, 0.087333971156993900342, -0.32878348080172842272, 0.050765458072287175662, 0.31348544971195407838, -0.17968941247328690247, -0.0008710541163406404757, 0.0078572826593715245497, 0.01154605134737083684, 0.0019071598067201241469, -0.02798048929881782193, -0.062619735793328773488, -0.037409193926265731089, 0.10788066522511974288, 0.27297748436636287694, -0.03064665724263981525, -0.73329875698796831429, 0.17466885580579571968, 0.82880838928957889511, -0.51934267695372704488, 3.1990468867824201592e-05, -0.0037123558507480692963, -0.0088738333679533650672, -0.012363197472223398654, -0.0030103079543227439195, 0.030121178037420162643, 0.081383444807754959371, 0.096048105333356101343, -0.052722296164137372187, -0.44374662456956531775, -0.16438093006056220258, 1.046052512307167115, -0.70360214490249806119, 0.14326237039847095067, -0.0041516231450834627237, -0.0091846597843334206596, -0.0014003247112519528333, 0.027084362184086170633, 0.072252452412897363865, 0.077758452967674002165, -0.065130096042298940562, -0.40256089861132560159, -0.67772049048923777015, 0.10171691247485817322, 2.8559034537620182626, 1.4933912226975327719, -7.320952959443855157, 3.8685583992442453827, 0.00042517525000053152818, -0.0027145033174862834117, -0.0050407279469242265851, -0.0049409706135470641067, 0.0022292066936192068451, 0.022893859924249172766, 0.047589906553345932838, 0.010298362093771509795, -0.092348587125128431263, -0.1483838992923754152, 0.042947703743378577712, 0.53194887679488767063, -0.64534577839678752831, 0.24421377946095274014, 0.0011528930827390617992, -0.0019310756643793666049, -0.0055178894743837612197, -0.007622479634317780306, -0.0012093796894739874612, 0.023283550004984950033, 0.055176281313487299252, 0.019432077025755344762, -0.096994647748155998768, -0.16856239201803829841, 0.040749017930903838614, 0.55287554128417826238, -0.63924099843293591405, 0.23184292062924294342, -0.00069812506707607328068, 0.0073758619624047696423, 0.0044802549424410960041, -0.011929850567894449004, -0.030201865608411766762, -0.021254570681012933953, 0.046847429124125192224, 0.14526559189575166675, 0.094461818762189925236, -0.36544442704213742079, -0.42740107475840916074, 0.83267124510201895315, -0.14834563692492996889, -0.12564601491435234415, -0.00018048111181390841511, 0.0059903621340384842486, 0.0032985107267882087662, -0.010324696416669464083, -0.025263726732414976128, -0.017933950812008766368, 0.039334262611954336208, 0.12564727724270416398, 0.084206763926552574695, -0.33238585322353142448, -0.36804996579667065282, 0.80681625751619112652, -0.25303644477022180226, -0.057890748015747603428, 0.0068585036305811065971, 0.0093459133249635568952, -0.0058946314059314620579, -0.034301595428971001778, -0.044263501891785315967, 0.011163702793098976424, 0.11876458227133629419, 0.11968304904710030945, -0.042492175288485485585, -0.20520564188974196873, -0.37561815209128945892, 0.29327026322029953187, 0.58789978537622877841, -0.44208149120650069852, 0.0033067044977625002083, -0.001754612845772337594, -0.0051513544980804643753, -0.0050638910810459790077, -0.0019822347247495154289, 0.0058826017111100695894, 0.035559650613372781325, 0.054619869685367125611, -0.089102929873997274712, -0.3488425495575045443, 0.28341598969755332771, 1.0426610567355782955, -1.6567878140829528366, 0.6861431998600145965, 0.002631471951199366912, -0.0051694923039148103425, -0.012251135015360797506, -0.011634147902857317619, 0.0075214411895309918332, 0.046496012793755404913, 0.075781105778898849579, 0.037570812183937152273, -0.10297593235157161129, -0.24801875425605562908, -0.20581695639149713561, 0.21994393418587601663, 0.86923964427231126262, -0.6807902462954101841, 0.0010418841708911103731, 0.0020675038415123403283, -0.0010138434424614753936, -0.0080687393336910627484, -0.01337631854204209686, -0.0031938572205117026805, 0.037854134446564526029, 0.10264389355238354162, 0.082491166101062357585, -0.2341228559422734723, -0.63959560199907705691, 0.13639916336282406628, 1.6842684178409799145, -1.1626053691967477288, -0.0012797121886047207236, -0.0013007752145687122985, 0.0028524334293966685343, 0.0085594234875250897743, 0.0050826604000210520024, -0.016466056194716718192, -0.040012814876523472574, -0.032856633515401197088, 0.016148950346610423112, 0.16971045993554895182, 0.31097622361734839069, -0.2783074296129985914, -0.808289052770161498, 0.67717832127694987943, 0.005915034567663796615, -0.0031459460168651773825, -0.01357797098309173299, -0.020332334217067427051, -0.01319429324987997712, 0.015306801709240843226, 0.05453506323855165927, 0.072241513817509248052, 0.033196575458590525665, -0.013933675704578514856, -0.032813897938230460194, -0.26743975149456145068, -0.21686875092483101635, 0.41959444416733848726, -0.00080923450300201225927, -0.0055627699944665424947, -0.0067459421182079750898, -0.00030126327359412361255, 0.0181202777856897361, 0.043182905564235742057, 0.046282454417645814415, -0.0048565606892275433107, -0.10167075679854262193, -0.22329119822210585777, -0.3270273671512403113, 0.27868790438564772227, 1.4493484758256025557, -1.1876655058835148271, -0.0062381211283332660064, 0.0067854775808569578466, 0.019796272814803603235, 0.022508443819154306392, -0.0061085235036557111385, -0.085022835445305131041, -0.19500339626188026299, -0.21205026709773303062, 0.10691649093594783604, 0.88860522359907123668, 1.7316789486492252248, 0.92480887612027729006, -9.387390509218189294, 6.2849462789139689534, -8.4095137926015409536e-07, -0.0019028895998577113023, -0.0040967549655361295419, -0.0041370763368958644521, 0.0052948012041344888479, 0.032984926913904935586, 0.062061292656267981471, 0.0060377074362315437395, -0.21303201312246519694, -0.27433788452134283409, 0.28167584925424621334, 0.4863399922563752531, -0.19831583803926902676, -0.18956176833920365676, -0.0029122175537008095254, -0.010827264115290195431, -0.0057770144463535804194, 0.017336256080184183487, 0.05230263016017477895, 0.070287587159430356398, 0.025620296148346901471, -0.10346333680809258193, -0.27885992859667491084, -0.36127698097296978563, -0.08327107628399359196, 0.81939499030704632254, 1.4771416246819624174, -1.66237135262042246, 0.00056509950752371630907, 0.014505278686585021397, 0.014067348980700101341, -0.0062703255091900340912, -0.040855694954800708674, -0.06374311653564881941, -0.037786614813355108577, 0.052104830123123981045, 0.166694963344525382, 0.22346383711342232559, -0.0045457862685163294078, -0.53005467749073875705, -0.23306912741521229604, 0.4567966153717544775, 0.0015451776836265943612, -0.00081942520257030767791, -0.0029038218749520977591, -0.0010295696833697660835, 0.0095457410368026476832, 0.024900912571081200431, 0.021408200714267428683, -0.028893346652631587967, -0.11218724010157107851, -0.12746066096529837242, -0.018268463354259495413, 0.2368581563007313151, 0.65348668511134888615, -0.67544051272513583317, 0.0019121974821458332695, -0.0010678539002900444554, -0.0039188627466897000412, -0.0026267276648331214228, 0.0083903814330410064065, 0.026285027152516838439, 0.027321848396678123649, -0.020025424327678720965, -0.10849464429746534522, -0.1393806559180067739, -0.041258670724333416924, 0.23301342043024475537, 0.67577073757161498779, -0.67395669306005545618, 0.0022131206949784908329, 0.00044417356743535138259, -0.0027926703188751226224, -0.004692816252027789091, -0.00028242548346023201977, 0.0095432915514377398419, 0.0097304918603782035447, -0.014451392371012157165, -0.045933633917854647943, 0.011946371330026064536, 0.14632806354052299569, -0.047493370178620125666, -0.18719463148914181061, 0.12379833022565771283, 0.0016201070221303255376, 0.0005889042278578451254, -0.0019024422740777525292, -0.0038454668054666682649, -0.0011262165320341799291, 0.006391125334440196995, 0.0074170554430074793137, -0.010292845827242070103, -0.035227477931185177673, 0.016447952934383482104, 0.1317188434525360663, -0.063090229015172999416, -0.17968941240805744175, 0.13277498053817851642, -0.0042600471726237888401, -0.00087105437446244598099, 0.0069144353837519894126, 0.017916070345797067281, 0.022484420261458869933, 0.00017076481108242852198, -0.063449023116540617662, -0.12681366883778044663, -0.064242470596995393528, 0.22069948753502222072, 0.38764695132908982833, -0.18301959187545802088, -0.51934267668867528656, 0.3032843651710346311, 0.0059310346871620780659, 3.1990536459406523266e-05, -0.0060727164067265063455, -0.013162252902147964992, -0.021620845850807916821, -0.023963212293506043482, -4.1424196748021170113e-05, 0.073047580843227433167, 0.17804039368283608447, 0.17579693698627130161, -0.21895717849919008047, -0.62336455005242930039, 0.14326237036275257797, 0.35055924605546190387, 0.005407926655519713563, -0.0041516231935325512148, -0.014645934757141324672, -0.019031049889730498881, 0.00053146654976733893622, 0.066390779691525569906, 0.16124367634130543081, 0.16110388498341796093, -0.15456571400467453148, -0.86222495749984051638, -1.1586192526308192186, 1.1377177376826701494, 3.8685583992809986498, -3.2476285610361883549, 0.0017976541761022793355, 0.00042517541631418719235, -0.0023908864885163736737, -0.0066778970042291850573, -0.011070263975662695058, -0.010184776821960679627, 0.0084364486351032677569, 0.054825783396553057214, 0.084621772088031732562, 0.014589478020763494565, -0.1645731729178046554, -0.23335535836688686695, 0.24421377932307947178, 0.027313516539405505862, 0.0012956378392177264353, 0.0011528932179371638618, -0.00069528445664525145766, -0.0052276703583502759018, -0.012248569453839978569, -0.015708498183078338506, 0.0013088337080743007133, 0.054462343582500377659, 0.097496339436234299991, 0.031366571782853507055, -0.16654112804429990757, -0.25448880764027803636, 0.23184292056393546178, 0.043998727912000865081, -0.0064062717256376479, -0.00069812486319934083552, 0.0086635292903931085395, 0.015974751867567346664, 0.0079594745999432784528, -0.024245112723806631905, -0.062191445772500818234, -0.04451929262600790721, 0.091888959399072350775, 0.26141375719633797026, 0.052132663847351240483, -0.50002143241529484197, -0.12564601506516051566, 0.33589295079884012862, -0.0051838813007026576449, -0.00018048098768144951883, 0.0074506543945223376729, 0.012995280689197362053, 0.0057292472525342398593, -0.021258458566549624097, -0.052716585901863510788, -0.037441676428354005168, 0.080118222491655607942, 0.22864039766964230904, 0.036245513972809444148, -0.44908750111330519861, -0.057890748151021832746, 0.26089896521208583158, -0.0065563649510164924081, 0.0068585033539978190742, 0.018731995962889002455, 0.016948688196824653746, -0.014799138366171388226, -0.06988798571289950301, -0.087934875328437536735, 0.013105163347178979544, 0.18706392504415172984, 0.24151645399612808873, 0.046216864642691944665, -0.44052976049432723826, -0.44208149094926979528, 0.54273096900835093859, 0.001227575878668628372, 0.0033067044719556812869, 0.0017020586769159982976, -0.0038354325516715770741, -0.010392999361261433949, -0.013484570958094398499, -0.0080133751826446993766, 0.024389950867813727647, 0.089213420753790351192, 0.060742975828284355433, -0.25592719485781356692, -0.25295042969261005705, 0.68614319983448712748, -0.3238319684338896165, -0.87058446184143978286, 0.8400001397792690172, 0.2489495365767999957, -0.080025255016027044275, -0.10301149160706365993, -0.049349761776863960983, -0.0040179088210960649946, 0.00021690363976271327174, -0.010239738510333943472, -0.013119362140272925951, -0.0057409294857598330494, 0.0030573322345481858398, 0.0071802841474409963632, 0.0061413168271420501276, 1.0630247803458772182, -1.3042134497021933282, -0.49364030189268603976, 0.56563473595594160326, 0.37872116420375123713, -0.013933815171714972539, -0.1346313015344403341, -0.077885654001423013093, -0.00010767935815594839079, 0.025361986707933541207, 0.013611969672261036574, -0.0009938928582020694312, -0.004780378707018948653, 1.6536588486896612691e-07, -1.8144508651037560831, 2.630592359710045347, 0.22563641042326906816, -1.0465229936417315137, -0.35345466246628720608, 0.16647979891067285485, 0.16900639606843764762, 0.040959844347923581032, -0.020707299200874072775, -0.02110071360351168171, -0.0059578321359084560754, 0.0024170897336488169141, 0.0030779777183766667166, -0.00010010166733721063373, -0.85563921736065506884, 1.0769099459380850892, 0.20771299752566552144, -0.27670706509103498716, -0.21229396179640044884, -0.053207341211722186669, 0.04047160588332848058, 0.047168382076463691299, 0.020166785472512886029, 0.0001714456419397692138, -0.0049105011237851117767, -0.0033615184041565723295, -0.0012316202092337452407, 3.022859030082826408e-05, 0.73459587550780369725, -0.65568196046503957053, -0.52195596700535495938, 0.19324503818695418222, 0.25782736578198078847, 0.083950985072860653813, -0.0087552577630104749112, -0.030632183043625197533, -0.025780202999310223438, -0.014732635429779140229, -0.0038923997769888443038, 0.0027363863517225919635, 0.0041988496126869542802, 0.0019284774988390883648, -0.1895617683673115339, -0.19831583796351351379, 0.48633999228497476475, 0.28167584903134940122, -0.27433788431767242022, -0.21303201315660855175, 0.0060377073320124650599, 0.062061292869190168298, 0.032984926673342733483, 0.0052948013444027892949, -0.0041370764844554618056, -0.0040967545843390669111, -0.0019028900115155796859, -8.4080410342948256929e-07, 4.9010375125650593731, -8.3336245867022871181, 2.0149360886979019725, 1.6634191398654769589, 0.28699719024566477943, -0.23841297322140092563, -0.20503368337264582522, -0.062950482893214199098, 0.00847429671820652472, 0.017358479733385067528, 0.0076961930550225330153, 0.00023151354454785459365, -0.0016621071242646939282, -4.333587350847017066e-05, 0.63982607471914088038, -0.23357985181918911044, -0.65808401160019835974, -0.12297271626062482308, 0.19947025939489143598, 0.20466734005186781875, 0.092520195657685919133, -0.014081854526526830368, -0.052105177589843704178, -0.03691189454245816931, -0.0086573909709937724166, 0.0078361172784543915715, 0.0086757701672625191225, -0.0010071343237703100654, -1.3265216324846422769, 1.4412914329781980616, 0.50642393905270410848, -0.26338169056298865733, -0.3227133654185010414, -0.17060555408888822271, -0.022409950515638928786, 0.054933113950735655562, 0.061253962157357420837, 0.031312648757691234624, 0.0027346118614888751015, -0.010133423820653169461, -0.0092633393035265083892, -0.00061573387496628305578, -0.11253507766883547037, 0.40874153125453327462, -0.2114308407352176733, -0.14726019859968200731, 0.011849850542755193533, 0.024571731321243434065, 0.01095751367710389064, 0.0099400462971346963126, 0.011482217894364602634, 0.0073172245204856821085, 0.0019137379690090917071, -0.0015723916682472902084, -0.0031367953222997857661, -0.0036904495804194648208, -0.14024832532658512707, 0.39073266038211668105, -0.16603432374812437944, -0.12144959943759252308, -0.0007704157369655289557, 0.011099489054520184084, 0.0081120534704002666276, 0.010854950960306822272, 0.011747010977046924707, 0.0067682388845848708361, 0.0013912052637867424083, -0.0016190939869099735, -0.0027337986960830969359, -0.0030365133556791955305, -0.39060786083578613592, 0.65941889025953437375, -0.10884068409619918061, -0.20849048888313370664, -0.020807057604799075734, 0.040652880518655967745, 0.030597747036207711968, 0.0073995618476503585703, -0.0059109775938553290459, -0.0075924381636016035832, -0.0033318657800985363761, 0.00039241384373937257058, 0.0016019246908895985824, 0.00094876287927124730408, -0.34680519208160437428, 0.59792125222550729369, -0.11422698920342455597, -0.18757801462114681579, -0.013114264649308040103, 0.040664123246589529981, 0.028048170049202798826, 0.0045111781692199082877, -0.0069060710635250538783, -0.0068516125934618837823, -0.0024087816369736001973, 0.00067107513125372233575, 0.0014055790467171374366, 0.00074445512058783928824, -0.47441998203323937355, 0.64605932007964195218, 0.048054955783675364078, -0.1541911343976585802, -0.098972042752437233704, -0.029878372719354787806, 0.020416523927583290349, 0.032309904924718274577, 0.015276114707174626509, -0.0035946004525120745648, -0.0077606307498466319678, -0.0028525720861050647519, 0.0012276067988492267956, 0.0011045067901039875517, 0.19544430749644131629, 0.13998114343521883551, -0.3398808273745181352, -0.15581763943378268422, 0.049879394282606202837, 0.088321118326128392195, 0.044651978756038233909, 0.0076872200597192588689, -0.0021846164480171536594, -0.0029270315667977975758, -0.0024631260896280419113, -0.0022091068189933511789, -0.0021070955942328515285, -0.0019738788741380281118, 0.099279866308225703242, -0.26231751037038220087, 0.18495082383216665534, 0.10269956148032904908, -0.1587365879955721093, -0.030788403951697292799, 0.064937312955956949523, 0.025158845464818068577, -0.012820338925878006775, -0.013449662674435041917, -0.0028316542662490005663, 0.0023876094669398012331, 0.0021786628759050377831, 6.3786684136541949868e-07, -0.40001551084132974667, 0.59963863769952230331, -0.072901037657649075951, -0.12817232717518300955, -0.013540743090235988003, -0.0041494857477234704454, -0.0005553508636849813088, 0.0066347169732852099427, 0.0075891008362562418707, 0.0028346733215834801246, -0.0017491834936217837138, -0.003094689140541044374, -0.0012948049553657941645, 0.0020031980382468417626, -0.37929277928058047209, 0.55942155142460570083, -0.060348649136909654489, -0.11224133315014192536, -0.018257906689186569754, -0.0088583584282294625112, -0.00084938142893296864327, 0.0066729683153860779726, 0.0074823993715611540611, 0.0032862882225936263132, -0.0010267723209733698592, -0.0026759684453452753372, -0.0013722581418757016231, 0.0015425410814587155617, 0.15489351665376563383, -0.028566669710795487835, -0.36784904492123898789, 0.029681630168536018988, 0.22837286403309436933, 0.10366121326073360343, -0.038516484078828433502, -0.066545648086859213177, -0.031039565840105108607, -0.0011771800651598320672, 0.0080688085412648664907, 0.0063310209121039814559, 0.0027052256852278301104, 0.00035725022541145993408, 0.12018584061743639579, -0.01616169631562882833, -0.31117516165670533468, 0.038074350264420110668, 0.19173868323289625093, 0.078780299289058122736, -0.037558975793776185059, -0.05558758920890680133, -0.023091130958766411391, 0.0010158815039570336016, 0.0069675278229423516854, 0.0046763425652902824645, 0.0017707115977520265955, 0.00037452228564473737235, -0.7802744772152920838, 1.082154979972553388, 0.075882333403322704823, -0.33117366556658445198, -0.13795609483735593592, 0.022969776360906059287, 0.047318021963678624786, 0.010682933879770773952, -0.006983099405727537061, -0.00049300945887217548225, 0.0059248017308690670291, 0.0036857355113944093375, -0.00091748244616531200724, -0.0029647735822497306088, 0.36453948804856334265, -0.46788134149841270615, -0.24648849918487636401, 0.3269646983015746522, 0.18992805604305371125, -0.075231206820246351863, -0.10237990110205089089, -0.026868130235555155727, 0.019351978527817063208, 0.021208228201457519912, 0.0081475393579252628623, -0.0012630807014670259337, -0.003821586022706075815, -0.0019693875188736384603, 0.8400001400976786492, -1.7375422809768839016, 0.56040337479451585168, 0.40938775334473376244, 0.065262227743822495696, -0.043302322835649650901, -0.067823219168894402498, -0.029305387892156514046, -0.00206829783787256926, 0.0022110944102579094955, -0.0040612221529148610777, -0.0057034945437705484791, -0.00085988871345838975655, 0.0071802842951014054368, -1.3042134498136637166, 2.3030029259786428675, -0.062252585530349513476, -1.332842617820310549, -0.075044536057015623354, 0.37853436184343330906, 0.18787993153297968818, -0.0086498724661640338796, -0.073999531486451775031, -0.039984603911039730273, 0.0031801980779429411296, 0.015186942549561740015, 0.006806696393654082522, -0.004780378534871797952, 2.6305923594540403521, -5.334905615318009886, 1.6874786202041411887, 2.0275236410278845156, -0.41152260625032671681, -0.64517561795876054997, -0.11744434859112273806, 0.10706862039371452255, 0.077670198961325556675, 0.01316275416952982294, -0.012643681465131034747, -0.010085086362979667207, -0.0020613990487715226282, 0.0030779780228091403044, 1.0769099459162323473, -2.1321247114338777173, 0.67686826365788621551, 0.59734436036761218158, 0.01197525003993816714, -0.13668136325217511429, -0.11077223044481876446, -0.025860234578704160285, 0.021249118138309618098, 0.023056358649832513208, 0.0080458840199116882924, -0.0011734402458145954447, -0.0029111088481546020315, -0.0012316199630187318025, -0.65568196043306581355, 0.98471797391351600925, 0.35007916703928892987, -0.76740578847159268694, -0.18884616028627118633, 0.16555537656002844438, 0.11038453736563429153, 0.02953734928875503124, -0.0047184580719417910932, -0.01130507627154697832, -0.010903211965543710416, -0.0068358217719490053238, -0.0011187519216361136408, 0.0041988494637412608276, -0.19831583798197133239, 0.76125984086382136873, -0.79794498796245461048, -0.27495511972607933338, 0.66377363043571224743, 0.10124274620123785429, -0.22101387412555872913, -0.092517143024180431565, 0.016573858158827732556, 0.030216715900391457467, 0.013211560040844285216, 0.0010507492290285334619, -0.0027658643713214616823, -0.0019028898056159680892, -8.3336245867101599316, 19.582898354177693534, -12.578220825160398633, -0.69701076814843165863, 1.3726216796682595422, 0.74473945289637910605, 0.09108178482921115704, -0.12334011390117677331, -0.088500034552915524122, -0.021653349734732103837, 0.00617388981126588271, 0.0085827124929261731229, 0.0034128141571982046655, -0.0016621071306339358677, -0.2335798519849112409, 0.082425170758225474454, 0.73502820192308215841, -0.38626442774833008587, -0.33520400113742715442, -0.042475450545466322816, 0.076988083916961283504, 0.10010349502780782616, 0.049699470649146870738, -0.0036685626651244518208, -0.024634986473906390902, -0.018708841232054254644, -0.004371522536481359418, 0.0086757702325118632447, 1.4412914329199548735, -2.7209278363850240012, 0.63023469084626759784, 0.84281419837207438572, 0.11601117559591839212, -0.1265738430356314892, -0.13723874156728146922, -0.080666552332575178208, -0.017008813020293746948, 0.023839453232634217478, 0.028123556410664621308, 0.014511769344183939046, 0.00013816654356572777968, -0.0092633395474964932287, 0.40874153127855261669, -0.98862685370670233542, 0.7343631807519408472, 0.041468086042225918431, -0.19832583503410822212, -0.034773803463912476841, 0.020375911246696624957, 0.01164460221872783835, 0.002190601915066423596, 0.0051631695702863290159, 0.0056164990771236126751, 0.0025727918357195790969, -0.00090326147559855569317, -0.0031367955004729798467, 0.39073266034609244191, -0.92803213096962144935, 0.6613798525244870552, 0.034600403754574272919, -0.15371898416457344116, -0.025454557492091384185, 0.0093121861233841571326, 0.0051149269447510557068, 0.001337551490553038052, 0.0056289619588219506896, 0.0057469470563041129274, 0.0023985737932115361488, -0.00094813241077387506051, -0.0027337986435874267986, 0.65941889033398781716, -1.4900386055960388454, 0.80106536526848692858, 0.27342517313035580306, -0.17952676910100459629, -0.085284895900696572335, -0.0053747032398836838696, 0.017940275005899179034, 0.013642983110702366564, 0.0039877369341324445454, -0.0031793825329361775217, -0.0038000830697956858306, -0.0012121446313322083686, 0.001601924600009336716, 0.59792125234707760306, -1.3643459782019065063, 0.75708675672673908519, 0.23221260384778477648, -0.16834303361965491375, -0.077168044139911587398, -0.0030080328727418548067, 0.019407166883256273676, 0.012770236067113525588, 0.0022242556474011553232, -0.0037667929947079910324, -0.0033103141694393344897, -0.00068896031548427289198, 0.00140557888857356433, 0.64605932006018407243, -1.3416812125485018115, 0.56313698234598463799, 0.27071845324501464525, -0.036259832173614295336, -0.054101618717058964914, -0.048481548344133332251, -0.023544750293636576566, 0.0066473923022845491856, 0.019083562529449472206, 0.00846754017807694985, -0.0032685338930852985383, -0.004994034926973120285, 0.0012276066679043407849, 0.1399811434435763724, -0.47106321341742057873, 0.59938224574939769074, -0.078482797921340094471, -0.21277189550436695353, -0.068929734334149325714, 0.037253409666593399752, 0.04840707470809183649, 0.017422342827322235409, 0.002841798396157229397, -0.0010947710738711947769, -0.0019555519718484509241, -0.0020732804488072559441, -0.0021070955373412828156, -0.26231751031104605376, 0.72535092165224690497, -0.56401922391135184132, -0.17623661707258436726, 0.41891135593491507505, -0.049324416913565143072, -0.14437885445257933603, 0.012609860156496755304, 0.043698467079541405966, 0.0089162261150558443784, -0.0091765214518320301074, -0.0065142835233015011429, -0.0003711856121127866232, 0.0021786627256742694195, 0.5996386378121695282, -1.360144748447400298, 0.76335655888179909656, 0.17044899354271939029, -0.14599594669230989141, -0.021700265326766561558, -0.00052536357311868731902, -0.0059437708403129634396, -0.0034940462189647376184, 0.0021182529043715214029, 0.0035707097006046264326, 0.0020044504769682557586, -0.00035218613455929430886, -0.0012948051325200231679, 0.55942155146629979345, -1.2642987729593540003, 0.70751397317977615664, 0.14754692440514163843, -0.12155818545142123288, -0.017559334418342387868, -0.0059807387493755236399, -0.0068677346623738314696, -0.0027200237838617791035, 0.0020612287708840293723, 0.0033402168877082606506, 0.0021555317167583848761, -7.6128377535796353903e-05, -0.0013722581802006039026, -0.028566669841223496507, -0.29175530257757920083, 0.69148147725657382878, -0.24167188280358822827, -0.30121414628212994646, 0.037852553379971905223, 0.14010845953960743282, 0.053181897404058708911, -0.026473947688073869861, -0.032916249038339444322, -0.013050070752274687408, 0.0013310739276670843171, 0.0052798949739439847428, 0.0027052260326916733983, -0.016161696444877783618, -0.26810945575214151226, 0.61276082059969150251, -0.23120559280415853287, -0.24681552319858121591, 0.04211285118935711308, 0.11788344296497553498, 0.040667327413455883756, -0.025632014288849029604, -0.02758278368483289486, -0.0093421436528561586921, 0.0021232331644506427515, 0.0044282174366474229971, 0.0017707113832527356535, 1.0821549800838912159, -2.2437560513853380684, 0.85811308040274236753, 0.62879333242482104982, -0.11518826206660846834, -0.18872236830617583614, -0.069651664021672665683, 0.026562400813056157023, 0.029326299579557971176, 0.0013895110657361525903, -0.00712512297586726049, 1.6096413290892824701e-06, 0.0037271094573896935455, -0.00091748257153672274286, -0.4678813414839124718, 0.72620661634994676525, 0.2732951262917188151, -0.7419780376801745625, -0.069313734854336936375, 0.28165542085403788564, 0.090743545958437810128, -0.053482986418229329362, -0.052491301975204442132, -0.0097780966356322222138, 0.010851757986022243091, 0.010199975785568344763, 0.002814277857320069548, -0.0038215859970956991484, 0.24894953622337345767, 0.56040337533357620359, -1.7052561831726886776, 0.53312892154340618678, 0.42354628926838783221, 0.060959401929472183113, -0.036124629864747348496, -0.056280566038865241507, -0.019915603007102992278, 0.00078158016938251672831, 0.0011498067109407810339, -0.0058669077893977513577, -0.005703494406131446974, 0.0030573321859408684499, -0.49364030181900497851, -0.06225258565591229587, 1.3716735657110425972, -0.021051480553402140972, -1.1918687304014030648, -0.053124762788162888105, 0.37253534952680633641, 0.17065682859269643012, -0.015097197648327266573, -0.070470101249076891325, -0.035866526057840027175, 0.0052188266518072047584, 0.015186943055900953628, -0.00099389326086721414832, 0.22563641095000261383, 1.6874786196245350478, -4.0904684687941141874, 1.2445274568364568246, 1.8233075737663972937, -0.33747739967440781861, -0.58987974704749945865, -0.11533434932555669539, 0.097700571515158290969, 0.073290362488885138492, 0.013043987372456581145, -0.012813958162011583261, -0.010085085522324186216, 0.0024170890789917088767, 0.20771299764128112186, 0.67686826343535555495, -1.9299638471785256488, 0.65973785890519820452, 0.61545006477498098718, 0.0080702242091962046039, -0.1383528993676239327, -0.1104617340618640059, -0.02680851063233622919, 0.018739475437021910209, 0.022137574750368926857, 0.0083724760477092187438, -0.0011734397797242620155, -0.0033615187815970348946, -0.52195596694176504826, 0.35007916695377561123, 0.48478693402789674316, 0.32256341828298656926, -0.70328040667797619978, -0.21345774061190050941, 0.14744943405755880916, 0.11321881187133768509, 0.039311703359391193813, 0.00029094895634816989299, -0.011027147177416225104, -0.012198613953978986596, -0.0068358220592216667119, 0.002736386656066065997, 0.48633999222755402991, -0.79794498785819734099, 0.41636368113827398885, -0.23389381096294931117, -0.2589458514874931061, 0.53862596152831865304, 0.099274044221668522603, -0.20356274594044368453, -0.092085528255424986543, 0.012032162976777945745, 0.028530840656232975966, 0.013538996096855064957, 0.0010507494951381708258, -0.0040967548564276720299, 2.0149360885274854027, -12.578220824988365578, 21.810810209818420446, -12.209489802874488262, -0.89964423743554766677, 1.2291263813227550905, 0.70789551553216800706, 0.10411338269002232138, -0.10776940478355706565, -0.084131999833229581531, -0.02190702568237878553, 0.0059963320128113648672, 0.0085827124178531055687, 0.00023151362133023515037, -0.65808401125199045278, 0.73502820145709901745, -0.080428105859034385028, 0.63655052491878139076, -0.42532047072959100475, -0.38371589289736252049, -0.043705612484177681909, 0.092032875398408237055, 0.11168078039304557558, 0.053345528759741522862, -0.0057510059815151712281, -0.028799631323889487294, -0.018708841093581323606, 0.007836117208322756697, 0.50642393915944416971, 0.63023469076637705921, -2.4756767044735403793, 0.61543156023301293089, 0.90699320660328408561, 0.16616707055861293374, -0.12568114973606089868, -0.15675443119188578467, -0.093356922929865693295, -0.02179813320761428419, 0.025008994232510538269, 0.031588197017226017638, 0.014511768638746728663, -0.010133423325956091021, -0.21143084078666654624, 0.73436318082192075796, -1.0518977795781969675, 0.62875090147507095573, 0.087689631687657698667, -0.17533566582194656802, -0.039949074968866966906, 0.0096231567864433972814, 0.0077662653700889296149, 0.0010247978063000475965, 0.0046302797187177347979, 0.0054103358297857891829, 0.0025727917304099426517, -0.001572391646377337528, -0.166034323864819644, 0.66137985267890764263, -1.0051562777477660582, 0.58454280314919704331, 0.08025867441932076074, -0.13680599205405369334, -0.030283360265438208359, 0.00016658723510099785609, 0.0018883821351242306794, 0.00035083621823948072695, 0.0051309532953594735871, 0.0055386174903410221454, 0.002398573824794559238, -0.0016190940333850050055, -0.10884068426932400064, 0.80106536543546247309, -1.4207371295929027699, 0.66982485223183474421, 0.28688493912572560207, -0.16804094101260561023, -0.084444726368124306548, -0.0051240196946903847006, 0.019689099478372322494, 0.013999286603823060671, 0.003867985552125339805, -0.0034379766866577082633, -0.0038000834869513409019, 0.00039241423084318034645, -0.11422698946216287841, 0.75708675699751182275, -1.3163946907940144992, 0.63775011177022333797, 0.2432569354118506999, -0.15878874771749831862, -0.074390344312456002651, -0.0022106335177776610668, 0.020488399360557144058, 0.012780083722871767027, 0.0021873051231948393777, -0.0038748392723663119437, -0.0033103146142863240295, 0.00067107552839359041908, 0.048054955991339312615, 0.56313698211330920351, -1.305987623359718075, 0.53624981377957048334, 0.30903192338301765929, -0.033724076642646351465, -0.068791601198521373783, -0.054020811078305093489, -0.018264074197690677998, 0.009330560004649398817, 0.017805796756486060028, 0.0067971463569872778501, -0.0032685340980209177941, -0.0028525719172881273261, -0.33988082731131902259, 0.59938224564936160021, -0.51385399948857135577, 0.52823124081743377811, -0.089687156354170086137, -0.20640081895202261331, -0.068795125897659209491, 0.027411295770678110323, 0.045218331460416393008, 0.018034674276097808349, 0.0038198686404320686232, -0.0006080409883506603325, -0.001955552132596303453, -0.0022091067243424143511, 0.18495082367886483965, -0.56401922370470602974, 0.61314057302219926449, -0.15763140998670255311, -0.28992652615670999605, 0.30846379722702843562, -0.019520071062406114948, -0.12068087034858618334, 0.011316824635348264952, 0.039402010166735515484, 0.0081991042143665926467, -0.0090262610124604396311, -0.0065142836735982527496, 0.0023876096637027621923, -0.072901037975701535321, 0.76335655926138312743, -1.4045244049139902209, 0.64949217908503542596, 0.22832780952711473343, -0.12632885509686062053, -0.025977947510192250791, -0.0064923397281907288189, -0.0078590573984604686353, -0.0028852309268503845488, 0.0038793789893264421162, 0.0048214831403449506103, 0.0020044500791621035901, -0.0030946887040331646189, -0.060348649109693335013, 0.70751397307187446906, -1.3209103755237936451, 0.61512110951560128314, 0.20224447740076001434, -0.10656437095773979573, -0.020339409982050443704, -0.010672042453584698954, -0.0090872754916207533482, -0.0026626832206299785866, 0.0035636732819231469946, 0.0045331968995999611174, 0.002155531715041346906, -0.0026759684386320715596, -0.36784904464991047268, 0.69148147692278993937, -0.50179970416289032542, 0.46977816045029108238, -0.23191070350201448136, -0.2467486341338255007, 0.055871656359081658949, 0.13527586220611781398, 0.05308687472293174836, -0.024292740252100156889, -0.032232015205065238372, -0.013442974764357153855, 0.0013310747271284757189, 0.0063310201593153158511, -0.31117516146692197898, 0.61276082041219404051, -0.48175748739159862666, 0.43472016755747849848, -0.21463234559806643187, -0.1995815783441484137, 0.05662716511965132371, 0.11196889501512632958, 0.040056699295629868973, -0.023594380325623748318, -0.026796772930057100731, -0.0095291569622823740077, 0.0021232329765435830476, 0.0046763427230933742387, 0.075882333163144799482, 0.85811308079768566781, -1.9779307958757945674, 0.74120797395734305812, 0.59689565411953338092, -0.10757605802174746712, -0.16049574834416366342, -0.059110552541839680996, 0.018944156445753639051, 0.021314342184754989096, 0.00029608759464131914878, -0.0058160912902543151701, 1.6095633843928919591e-06, 0.0036857355386287239674, -0.24648849920004947678, 0.27329512617760054471, 0.080646457626439016497, 0.26443859062911173607, -0.5628469388049839317, -0.050578742126905107401, 0.25423069887818028345, 0.079184928271641905062, -0.052882509539827385336, -0.051797704467252202176, -0.0099799478317028709334, 0.011185506182117591414, 0.010199975808295558971, -0.0012630808093961598797, -0.080025255029393088657, 0.4093877533906459254, 0.53312892160411318176, -1.7346654909297285485, 0.48481340025821401118, 0.41772024343153568715, 0.07575191461627955869, -0.020087466699257410657, -0.050673228923396151924, -0.021292397506588300188, -0.001586706998393246559, 0.0011498066357496729197, -0.0040612217436059357828, -0.0057409298328334910574, 0.56563473619354465161, -1.3328426180137054047, -0.021051480659384794181, 1.5431412647277968908, 0.10994509853258849219, -1.1782270139227959405, -0.10870775133374496746, 0.33852139889344246004, 0.17053754156304842815, -0.0044203198024497739743, -0.064692875751705528065, -0.03586652599122017171, 0.0031801974475570041027, 0.013611970163694251876, -1.0465229939613844845, 2.0275236413575594696, 1.2445274568749171706, -4.4447338625828072267, 1.1423849471282012669, 1.8876043812083245221, -0.26915516187397481307, -0.57218761649215310605, -0.1243095342257009156, 0.089233288888004705375, 0.070865052508976983381, 0.013043987975323046227, -0.012643682390675708635, -0.0059578316319923250305, -0.27670706530817479329, 0.59734436096176113384, 0.65973785846668886457, -1.9756009172707444943, 0.58861718219059078194, 0.60141524833413007034, 0.023491712506595392035, -0.1232230886335991954, -0.10663583218926120377, -0.028410325890141200494, 0.017273740444647314729, 0.022137575171864418144, 0.0080458837455345162681, -0.0049105011591160504722, 0.19324503797654624115, -0.76740578810591864034, 0.32256341816936512368, 0.53442490263685915597, 0.35900133234295372731, -0.69033508307511981528, -0.22017130781356153424, 0.14209319116148710904, 0.11036991151206818063, 0.037835468765549239156, -0.00048609529164096298027, -0.011027147299356653345, -0.010903211823995534457, -0.00389239989903427102, 0.28167584923905986116, -0.27495512017727585796, -0.23389381058053354545, 0.4710638715051421177, -0.34703770397230765399, -0.31004412882199822166, 0.53975949071273365831, 0.11763236732578211119, -0.19458844515214140936, -0.092196527132937383664, 0.010365765156436194602, 0.028530840418821406396, 0.013211560184466324824, -0.0041370765024611039218, 1.6634191402742528609, -0.69701076866404143573, -12.209489802804759151, 22.601689838963604728, -12.078570739396349509, -1.0288892980578350222, 1.136940578571383087, 0.68767638142371534382, 0.11242817846888229771, -0.098656751497054578914, -0.081236407803498528191, -0.021907025678155483267, 0.0061738898615347394955, 0.0076961929951828772153, -0.12297271665538052343, -0.38626442727925186515, 0.63655052489763930268, -0.17358457253033426904, 0.61994973168152822662, -0.37967858009021077104, -0.35381999051125095912, -0.04055927160106676671, 0.081209496956950302793, 0.10138274256695664621, 0.048460669793962907193, -0.0057510057879729056807, -0.024634986323036869166, -0.0086573912067940128612, -0.26338169092341179311, 0.84281419863690665295, 0.61543156044529989668, -2.4718897377729280862, 0.57104649455845069017, 0.8723773573889940236, 0.15513219961351590315, -0.11743125246219712765, -0.14614622003550015172, -0.08686474782316525034, -0.019394561106305250459, 0.02500899377215523578, 0.028123556923844057753, 0.002734611674191901598, -0.14726019835320974805, 0.041468085661374332052, 0.62875090156625990101, -1.0617934396158330568, 0.63493585990892420501, 0.095604578008121332555, -0.17864847373910769379, -0.042753131811619847491, 0.0083456210363252285311, 0.0073923706072641188689, 0.00071904692658053266478, 0.004630279506488981997, 0.0056164989464264625241, 0.0019137382440886949002, -0.121449599162898908, 0.034600403298079555892, 0.58454280332372432483, -1.010604700553440205, 0.58550412885376756922, 0.084990453360008502703, -0.14008370413213952488, -0.031808841623884469285, -0.00023906685675434066751, 0.0016984714147530825812, -1.6869042966946492385e-05, 0.0051309531849169616805, 0.0057469474242265927363, 0.001391205011043268689, -0.20849048891528262906, 0.273425173232532015, 0.6698248520776068915, -1.4693504888984443291, 0.64793604323276798862, 0.29870021943607694137, -0.15536792062812668225, -0.079300788439361849513, -0.0062869294920343387484, 0.017729100325116920606, 0.013135321257787510238, 0.0038679854943475863033, -0.0031793820661805685321, -0.0033318661401997040107, -0.18757801436884996571, 0.23221260363009421646, 0.63775011163039130313, -1.3637768167996486834, 0.61814555528937442652, 0.25690145323201390992, -0.14677426398150425491, -0.070608700639107402353, -0.0039387310831066309549, 0.018721436954662305091, 0.012221145464448109411, 0.0021873051366905955396, -0.0037667926299545786625, -0.0024087819673511002942, -0.15419113481183968628, 0.27071845380569481421, 0.53624981358396506259, -1.3161613831690879284, 0.502104176058222329, 0.29094004492950253926, -0.030781865260810104684, -0.054990236324861399564, -0.047558648414396019477, -0.019973181942252438842, 0.0061730054706449202437, 0.01780579656551190687, 0.008467540344178785261, -0.0077606307867233038228, -0.15581763971006246106, -0.078482797640160639752, 0.5282312408701211881, -0.57314861628679913963, 0.53457336324423188056, -0.067907902065145922532, -0.20444401560464031675, -0.072510724192442796432, 0.023555956740333890836, 0.043783637954982190799, 0.018052183062095147892, 0.0038198683526579884501, -0.0010947708752770861709, -0.0024631261200824040337, 0.10269956170780879245, -0.17623661745535826317, -0.15763140981008200159, 0.52276818403246072542, -0.2498286066526695659, -0.28111273799406238938, 0.33216686838828041184, -0.0098630544872136197143, -0.12415446085641354546, 0.0069750012488402303099, 0.038352506099112053384, 0.0081991047055801017512, -0.0091765217877766147581, -0.002831654227347982309, -0.12817232678372544874, 0.17044899299619864497, 0.64949217931478875343, -1.4032919534371297576, 0.64089224293183166825, 0.23005781090939866895, -0.12703251056763378846, -0.02551204580776999381, -0.0044614554475361370769, -0.0056217439943931695953, -0.0019302123687563921186, 0.0038793795424370623552, 0.0035707096786467638211, -0.0017491837993240159053, -0.11224133341912201356, 0.14754692505706484829, 0.61512110898821481531, -1.3191914110801543636, 0.60419742671050324923, 0.20406065214570903721, -0.10685646619437091931, -0.020011695198129821061, -0.0090268080572395246325, -0.0070639575782236643103, -0.0017162277655923922903, 0.0035636733062257902258, 0.0033402164325114190591, -0.0010267719323798463974, 0.029681629975709295649, -0.24167188240398271293, 0.46977816018167711798, -0.52755787320224600823, 0.54531463673633340239, -0.18621692600170219878, -0.2542103978309522283, 0.039672233893260896997, 0.12837588883236641935, 0.054229357604010298977, -0.022208216719602956457, -0.032232015480169039223, -0.013050071040235316708, 0.0080688089439375134532, 0.038074350447915981943, -0.23120559321988187751, 0.434720167939037061, -0.49728270533377832985, 0.50051320857533698749, -0.17767530572351589169, -0.20885116559464644848, 0.042469761417104519297, 0.10693378189255781097, 0.041660230512920122525, -0.021658939761229210191, -0.026796772808955912626, -0.0093421438383081943646, 0.0069675279519640980971, -0.33117366524624597401, 0.62879333172306983712, 0.7412079744070296794, -2.0829704940902620258, 0.68526216143810969594, 0.62581567439480156878, -0.078118520324802015464, -0.15794146560609995111, -0.067041241859996353325, 0.015952156980821127763, 0.022465056068285863683, 0.00029608758477451169091, -0.0071251230140468120242, 0.0059248018194827329133, 0.32696469816164913658, -0.74197803723280830734, 0.26443859007741282729, 0.25542596552468638249, 0.33093706210868889528, -0.58296349768347122122, -0.087984229709477432269, 0.24131798014923461215, 0.082933889485725512114, -0.047198725041038412653, -0.049711557986374062257, -0.0099799481088233936432, 0.010851758147880688141, 0.0081475394017485724268, -0.10301149133844600403, 0.065262227173827083715, 0.4235462893192313838, 0.48481340056432381536, -1.7574528417946739811, 0.48455388345911593051, 0.42736406549355354656, 0.08111527763461423568, -0.019931598492109434823, -0.052308910598174010453, -0.02129239795492365861, 0.00078158095655651990078, 0.0022110942593528775356, -0.013119362375853090702, 0.37872116395334565198, -0.075044535762275210722, -1.1918687304174293562, 0.10994509854390382975, 1.5852907133649882798, 0.089517159203186091876, -1.2039007810380111252, -0.11721032019789848211, 0.34191941737416192693, 0.17506422596828202787, -0.0044203199314077571552, -0.070470101244780800065, -0.039984603415124823911, 0.025361986309842269455, -0.3534546623478878602, -0.41152260641351834369, 1.8233075739009421135, 1.1423849469183886551, -4.4532237405802668206, 1.1535892395610254368, 1.9025738315891747021, -0.26224504572486950771, -0.57300464492223490698, -0.12701422203442408465, 0.089233289575763555246, 0.073290361328176401035, 0.013162755004867547279, -0.021100713833502655892, -0.21229396185202067948, 0.01197524971296765918, 0.6154500655094294892, 0.58861718185853029173, -2.0149346096487068536, 0.58264784493491272865, 0.6106268185254556391, 0.030682954480920267282, -0.12054659518571179955, -0.10660551084891647933, -0.02841032550378826127, 0.01873947474026315882, 0.023056358480381747289, 0.0001714460815981460847, 0.25782736573820907999, -0.18884616045270130602, -0.70328040646314216033, 0.35900133248994176016, 0.5567071075855967166, 0.37326420342374500017, -0.68628867564890838615, -0.22502997481867259566, 0.13653903122629526545, 0.10781071857430725436, 0.037835468495753574181, 0.0002909495179483351365, -0.011305076759834411376, -0.014732635259277426787, -0.27433788452922702739, 0.6637736310904089887, -0.25894585230905459161, -0.3470377035432957169, 0.44102840478551980086, -0.34585690118470219234, -0.30592818568046159511, 0.54469420616397512536, 0.12174198614627745885, -0.19349331844418277138, -0.092196527541279882367, 0.012032163376940468655, 0.030216715680552352213, 0.0052948014000527487619, 0.28699718996947060079, 1.3726216801670343415, -0.8996442377214584063, -12.078570739336260687, 22.720172764946461541, -12.070922463209262432, -1.057594439684330867, 1.1189416240270972924, 0.68561287232448553386, 0.11519743815515647023, -0.098656751655834332149, -0.084131999828330070179, -0.021653349509333428419, 0.017358479576788335419, 0.19947025968482567926, -0.33520400126178090394, -0.42532047111545934204, 0.61994973196463520893, -0.13361296533972477629, 0.66123992538178133493, -0.36399865551329135815, -0.36043023050260752926, -0.053016796336424913683, 0.07344029865345018393, 0.10138274245155845521, 0.053345528870214528605, -0.0036685631071707225023, -0.036911894178274083644, -0.32271336497817365574, 0.11601117529506416171, 0.90699320619107259844, 0.57104649486112801782, -2.5366450069250503496, 0.5293648712619797303, 0.86851882753274078919, 0.16744798077964428984, -0.10533550946051255248, -0.1404588789757662548, -0.086864748151796788989, -0.021798132663130756481, 0.023839452967787687354, 0.031312648754093723447, 0.011849850269744707787, -0.19832583453039984156, 0.087689631448890914545, 0.63493585990616008274, -1.0696500819693044448, 0.62744503425994369028, 0.095880950048952071074, -0.17768857447921457271, -0.042633729284205459931, 0.0080574035939938188494, 0.0073923706007183054983, 0.0010247979001629695887, 0.005163169331335768289, 0.0073172246876664808271, -0.00077041596616034312486, -0.15371898357920685263, 0.080258673906001068432, 0.58550412906928595014, -1.0199439211429293106, 0.57805740983917131004, 0.08550910221243872289, -0.1387141162804291783, -0.031391356936458662819, -0.00045239578568493808361, 0.0016984711330502236248, 0.000350836526927064251, 0.0056289613664094999973, 0.0067682393196899321916, -0.02080705724842695642, -0.17952676966479172438, 0.28688493945946252905, 0.64793604305693708323, -1.4781397279067907125, 0.65072834680122271234, 0.30521858108824828637, -0.15358552125468999394, -0.080252915950286440872, -0.0073471969314532507497, 0.017729100402137223935, 0.013999286698815312582, 0.0039877369833443568684, -0.0075924383002539215168, -0.013114264767056457023, -0.16834303366708439076, 0.24325693572093956552, 0.61814555517650116023, -1.3690141346860080773, 0.62145408119204059716, 0.26213232131326907881, -0.14579356461637393849, -0.07154118826736036163, -0.0047472843389691063085, 0.018721437063290637115, 0.012780083575472335825, 0.0022242554250840483251, -0.0068516123318007448456, -0.098972042403222285079, -0.036259832826990558152, 0.30903192388920697198, 0.50210417577028576552, -1.3507921347816755908, 0.49403200778793077141, 0.30251048096427779477, -0.02261671984804621352, -0.054619061064667927263, -0.049854728499809708442, -0.019973182048704754554, 0.0093305602962176648757, 0.01908356242770712502, -0.0035946004933167636196, 0.049879394226238438215, -0.21277189531419801738, -0.089687156555697755689, 0.53457336344498940228, -0.5620111788115210727, 0.53706536319979325533, -0.065441104053155979559, -0.20467798419561716861, -0.073992916278027093258, 0.022652609589258045769, 0.04378363770060749588, 0.018034674670850518036, 0.0028417984860017726605, -0.0029270318241348485498, -0.15873658824877723283, 0.41891135639446974714, -0.28992652642145760877, -0.24982860662282024844, 0.54705449929671412512, -0.22496859223662490912, -0.27772182010351292858, 0.32916271242324718349, -0.011612602367117988433, -0.12493608309291402703, 0.0069750017341204239435, 0.039402009303786109107, 0.0089162267055159279833, -0.013449662797788124874, -0.013540743512471996313, -0.14599594609235178044, 0.22832780929562401795, 0.64089224302312364223, -1.4196049670439883084, 0.63222703142534120158, 0.23077765717652007349, -0.12373969850144471427, -0.02255180078073201061, -0.0032245912659758604359, -0.0056217433392301797632, -0.0028852318567229315663, 0.0021182529788976956972, 0.0028346736382184557421, -0.018257906454972137184, -0.12155818618323596669, 0.20224447835240505489, 0.60419742601923098757, -1.3338630668655402367, 0.59648147780162563691, 0.2042852619005992576, -0.10413151537947584746, -0.017217727346788998938, -0.0077575968953396160985, -0.0070639572061931589497, -0.0026626836353905846549, 0.0020612294647122794342, 0.0032862877234092744728, 0.22837286421699309891, -0.30121414673821572672, -0.23191070308470121408, 0.5453146365152666819, -0.48883863766512630766, 0.54464472602503566367, -0.19473195701208964592, -0.2615195491731129307, 0.036783136414730896457, 0.12858871524943252251, 0.054229357288744062249, -0.024292739823212432942, -0.032916249246135217965, -0.0011771800599924776756, 0.1917386830909451334, -0.24681552275150508979, -0.21463234618311921587, 0.50051320891354311726, -0.46625560510876140619, 0.49716495342453037587, -0.18556630047615552126, -0.21447791376388145035, 0.040661909172152552183, 0.1073933105638127522, 0.041660230805556995592, -0.023594380557205182902, -0.027582783570599109235, 0.0010158814542826943069, -0.13795609511600331887, -0.1151882615060730447, 0.59689565372742570126, 0.68526216153842967049, -2.0826820474219696067, 0.69374633186732947987, 0.62557534409906367934, -0.08133816671377398777, -0.15776987009143778051, -0.065561614891298539143, 0.015952157185450489829, 0.021314341969515992331, 0.0013895111216695392019, -0.00049300941805234380945, 0.18992805656676897685, -0.069313735820853886183, -0.56284693795987106313, 0.33093706159373187381, 0.24470524701681867219, 0.31075412880844172081, -0.5929848277459032202, -0.090399606873520013672, 0.24352315222159587327, 0.085410205427707594494, -0.047198725169448646488, -0.051797704264186311685, -0.0097780964859873996387, 0.021208228006019066297, -0.04934976203530741512, -0.043302322761647735194, 0.060959402577164145676, 0.4177202425536864494, 0.48455388394747439529, -1.7545813787237820591, 0.48394155906120767119, 0.42499360373482125697, 0.079297628531577693534, -0.019931598803230986955, -0.050673228106896790746, -0.019915603987587713469, -0.0020682983184061830284, -0.010239737689925542535, -0.013933815239047910375, 0.37853436180704291836, -0.053124762527479017782, -1.1782270141886959092, 0.089517159350159539177, 1.5745442675889118611, 0.09378253693455096407, -1.1977015820738226015, -0.11453928576132627049, 0.34191941711901252532, 0.1705375417542630001, -0.015097197530018667544, -0.073999531765255324722, -0.00010767923005219439997, 0.16647979865112857989, -0.64517561739997320025, -0.33747740038230478188, 1.8876043819682164493, 1.153589239128938182, -4.4663349973518791103, 1.1427559124879536068, 1.9008247388947681067, -0.26161603399189348185, -0.57300464470595013911, -0.12430953511493951502, 0.097700572882367275884, 0.077670198152255051638, -0.020707299081121052775, -0.053207340616714081383, -0.13668136391132687524, 0.008070223846121321945, 0.60141524884806540907, 0.58264784487319398742, -2.0156446963991361265, 0.58188303378297212554, 0.61097031659956368976, 0.031366468256844318163, -0.12054659512746139505, -0.10663583247121218023, -0.026808510708244498444, 0.02124911870505914388, 0.020166785194378658574, 0.083950985166094560519, 0.16555537656313515371, -0.21345774069481485147, -0.69033508348874794702, 0.37326420402177012958, 0.56786702594886839801, 0.37621348382476460515, -0.6888243487013638866, -0.22789735293590282494, 0.13653903124994307139, 0.11036991174130697524, 0.039311702666955695984, -0.0047184570816308597438, -0.025780203490453464099, -0.21303201304863894627, 0.10124274561345802947, 0.53862596250326499892, -0.31004412945322962214, -0.34585690107826505546, 0.44838681473891633944, -0.34447369693320312178, -0.30630760982801302772, 0.54503239203028608362, 0.1217419854018210873, -0.19458844430493033029, -0.092085528677023922861, 0.016573858077624785412, 0.032984926829426525208, -0.23841297315558546144, 0.7447394526373386503, 1.2291263816928104102, -1.0288892981594099929, -12.070922463393035429, 22.758597911458213758, -12.057472665106690357, -1.059595142982437288, 1.1163601706070089836, 0.68561287185750063866, 0.1124281787877859512, -0.10776940468668143314, -0.088500034927339113477, 0.0084742969294697087501, 0.20466734015611029451, -0.042475450833867521572, -0.38371589252942694959, -0.37967858030643436562, 0.66123992549893029214, -0.11661353129688510621, 0.65973795726156481134, -0.37210464172024326768, -0.36697745704486128826, -0.053016796520524399339, 0.081209497451623577358, 0.1116807797567455629, 0.049699471041743319144, -0.05210517770696204698, -0.17060555406614039131, -0.12657384323559306427, 0.16616707107529327114, 0.87237735684272221803, 0.5293648715268511884, -2.5530342466702902193, 0.53030634896254968957, 0.8760035259957122955, 0.17317923510171057133, -0.10533550964324214783, -0.14614621949307474158, -0.093356923571107655602, -0.017008812676454241469, 0.061253962102793949041, 0.024571731140576785757, -0.034773803310652121046, -0.1753356657278888342, 0.095604578121557176607, 0.62744503402455176211, -1.0729398592709071814, 0.62767018528525564403, 0.096785068284875888645, -0.17751541929282990684, -0.042633729293813184269, 0.0083456210008415222801, 0.0077662654517383512137, 0.0021906024933610623341, 0.011482217313797933242, 0.011099489395980502437, -0.025454558211598865558, -0.13680599134279300411, 0.084990452884926764865, 0.57805741002376953563, -1.0227128314954025967, 0.57847196940986356228, 0.086373555079420100133, -0.13856967161641398079, -0.031391357400904207342, -0.00023906604546665086164, 0.0018883815472109828637, 0.0013375518171831904477, 0.011747010811309610209, 0.040652880217704405463, -0.085284895359548085203, -0.16804094130214400349, 0.2987002194698898383, 0.65072834683390889943, -1.4777217361394170414, 0.64959347293930858047, 0.30445569870507277477, -0.15403060925274619786, -0.080252915691432599732, -0.0062869297655987182311, 0.019689099668860562259, 0.01364298281902093235, -0.0059109774127962351128, 0.040664122914681508081, -0.077168043501587749167, -0.15878874805131987236, 0.25690145315899748368, 0.62145408136164537183, -1.3692911976421349163, 0.61990607569729250237, 0.26146320643710802001, -0.14602677201766670056, -0.071541188072642591944, -0.0039387314104760142927, 0.020488399674153244534, 0.0127702364548537943, -0.0069060715241213762455, -0.029878372868339400648, -0.054101618219908186036, -0.033724077265526397618, 0.29094004522931221635, 0.49403200780951900262, -1.3448151850193437173, 0.49812092498126558926, 0.30168193260506187547, -0.024645228625611009105, -0.054619060992052138881, -0.047558648488392391007, -0.01826407393338242216, 0.0066473919156123583915, 0.015276114874122205534, 0.088321118624493918925, -0.0689297347600706517, -0.20640081869964646555, -0.067907902478549192482, 0.5370653635980154883, -0.56134135347564118046, 0.53799964387485299522, -0.064836105180718223151, -0.20474050694154335983, -0.073992916301164307624, 0.023555956986439367112, 0.045218330971567584353, 0.017422342959077942448, -0.0021846163233000506949, -0.030788403656183217783, -0.04932441735658457671, 0.30846379736248924486, -0.28111273782872209237, -0.22496859252813905528, 0.55247158470175072864, -0.2294972646098676905, -0.28077696424879744486, 0.32874492342003619472, -0.011612602255525718287, -0.12415446129653731622, 0.011316825217402981876, 0.043698466902651271482, -0.012820339011433836557, -0.0041494856055000984771, -0.021700265414913155748, -0.12632885533475471385, 0.23005781111216311308, 0.63222703138655012012, -1.4237700900807899185, 0.63309124637614200282, 0.23343882338101437646, -0.1223226223298103793, -0.022551800264260974532, -0.004461456182181498982, -0.0078590570849146022814, -0.0034940454892346322832, 0.0075891002101265278221, -0.0088583583660665869697, -0.017559334511422587666, -0.10656437137573521179, 0.20406065332486322816, 0.59648147665612216706, -1.3381446133372800755, 0.59686097062636056521, 0.20670728853992992313, -0.10270601765121480098, -0.017217726998542209893, -0.0090268091092344793225, -0.009087274258750423217, -0.0027200243329955852388, 0.0074823994190169537677, 0.1036612131949019161, 0.03785255337164114553, -0.24674863413558675851, -0.18621692568791142763, 0.54464472566482047267, -0.49645471680549874138, 0.54132925082395288996, -0.19508359674639294545, -0.26143770504178781566, 0.036783136248748710129, 0.12837588895682777723, 0.053086874800709782374, -0.026473948012628222004, -0.031039565623696221269, 0.078780298929940359431, 0.042112851750051215338, -0.19958157857616179442, -0.17767530576867132663, 0.49716495361926577035, -0.47329204806169627995, 0.49493201192145663825, -0.18533283531349833972, -0.2141985284333680073, 0.040661909794325219736, 0.10693378139498836699, 0.040056699192622308159, -0.02563201403697754413, -0.023091131011698465963, 0.022969776386669280704, -0.18872236806838255108, -0.10757605847595116955, 0.62581567462187337458, 0.69374633182029810108, -2.093476673005188804, 0.68355683058727212398, 0.62484735209964725744, -0.078788171955231800503, -0.15776986982933946524, -0.067041242228522701652, 0.018944156860228829331, 0.029326299136864908412, -0.0069830992065546149822, -0.075231207304223363108, 0.28165542184615172117, -0.050578742903548197518, -0.58296349730312368997, 0.31075412876982572152, 0.24167345712797472235, 0.31425419850200048133, -0.59030141605347397071, -0.089338724165650820752, 0.24352315256953616096, 0.082933889277255479211, -0.05288250934690466587, -0.052491302852721202776, 0.019351979170644868083, -0.004017909034937728989, -0.067823218480258656848, -0.036124630875927793427, 0.075751915943451236801, 0.4273640643264150496, 0.48394155954011597309, -1.7593638573496899191, 0.48055624361220855567, 0.42499360339899744377, 0.08111527841999371935, -0.020087467513377983697, -0.056280565492016543239, -0.029305387868795194223, 0.00021690345033782546837, -0.13463130134171610863, 0.18787993141464626179, 0.3725353492615734452, -0.10870775099389316798, -1.2039007812879263248, 0.093782537044154554162, 1.5860789365259044814, 0.099490557724335959566, -1.1977015824713879155, -0.11721031959898606245, 0.33852139847201850875, 0.17065682863655526336, -0.0086498723265256427517, -0.077885654070516036063, 0.1690063964090581794, -0.11744434944925503939, -0.58987974595593273275, -0.26915516286644741717, 1.9025738321986782609, 1.1427559122358568189, -4.4763992158090557538, 1.1395963689484591708, 1.9008247388279551071, -0.2622450457812580682, -0.57218761593716127578, -0.11533435029539712557, 0.10706862105066485003, 0.040959844202333610397, 0.040471605431662525942, -0.11077222957962006011, -0.1383529000348560356, 0.02349171256345415279, 0.61062681890480774438, 0.58188303370683991389, -2.0187077857680852944, 0.58055458357273681536, 0.61097031687540626077, 0.030682953967497914038, -0.12322308893698502574, -0.11046173293030543683, -0.025860235204126068992, 0.047168382097794371577, -0.0087552574241923364351, 0.11038453696380111368, 0.14744943384536043918, -0.22017130705634968546, -0.68628867653528580206, 0.37621348433395079836, 0.56822000694341501248, 0.37534276693849005779, -0.68882434853689200605, -0.22502997492539919544, 0.14209319098760789246, 0.11321881230951248709, 0.029537348953798645695, -0.030632182967472495333, 0.006037707363908682498, -0.22101387373743044962, 0.099274043268276240815, 0.53975949148289215085, -0.30592818592479331663, -0.34447369695135271472, 0.44752226840278930275, -0.34548155496183785562, -0.30630761044006465932, 0.54469420731147544235, 0.11763236609573174407, -0.20356274524820860972, -0.092517142886892417275, 0.062061292604428212882, -0.20503368350314696289, 0.091081785076616780805, 0.70789551541495521203, 1.1369405784004738003, -1.0575944393349012707, -12.057472665316847582, 22.773955153017553243, -12.053161157605124743, -1.059595143647701132, 1.1189416247851151631, 0.68767638118801022173, 0.1041133824776681871, -0.12334011372940013873, -0.062950482908140772964, 0.092520195285052203582, 0.076988084363086489725, -0.043705612406359145572, -0.35381999078492559763, -0.36399865553680621488, 0.65973795750111008385, -0.12335408142124562603, 0.65481601470785355534, -0.37210464204735438365, -0.36043023038157095961, -0.040559271644048953964, 0.092032875758784674081, 0.10010349448805579142, -0.014081854269637339785, -0.0224099505623324112, -0.13723874137706898457, -0.12568115023673057107, 0.15513220036290376047, 0.8685188269894815738, 0.53030634908179208331, -2.5501120578458502663, 0.53299998324644015657, 0.87600352598179354047, 0.16744798107196987025, -0.11743125315445358814, -0.15675443057110127754, -0.080666552459840099032, 0.05493311386502430782, 0.010957513803264646093, 0.020375910972743618843, -0.039949074687695818042, -0.17864847418192139616, 0.095880950444547433098, 0.62767018526117346333, -1.0727437560057959676, 0.6277466132602331772, 0.09678506836027406246, -0.17768857449600250487, -0.042753131664279103719, 0.0096231565326012555306, 0.011644601906094360216, 0.0099400466941589830233, 0.0081120530810543864386, 0.0093121868767266583605, -0.030283360930932246707, -0.14008370362487679306, 0.085509101817108440491, 0.57847196964998459645, -1.0225704608561032583, 0.57843251889112246023, 0.086373554853526060548, -0.13871411561606306062, -0.031808842434979299474, 0.00016658780292491681685, 0.0051149265945180317008, 0.010854951099415099253, 0.03059774699162792419, -0.0053747033060992438144, -0.084444726165430070153, -0.15536792060353502021, 0.30521858094287590557, 0.64959347291295388427, -1.4800846314551676652, 0.64854070876347502672, 0.30445569878578476741, -0.15358552147286572143, -0.079300787963042199125, -0.0051240201983410891759, 0.017940275049073032415, 0.0073995619856580133536, 0.02804817056082947388, -0.0030080337452128990024, -0.074390343835656291827, -0.14677426398005055663, 0.26213232103043521137, 0.61990607586218704483, -1.3713750952198155986, 0.61916206112311156229, 0.26146320670751654891, -0.14579356496101159024, -0.070608700209115596302, -0.0022106338394356217503, 0.01940716674382909554, 0.004511178369243260336, 0.020416523930592338221, -0.048481548610019180312, -0.068791600737983377156, -0.030781865248674825791, 0.30251048055273777049, 0.49812092517881079168, -1.3474974637299805824, 0.49501974799911724334, 0.30168193261805703598, -0.022616719742020990197, -0.054990235967104304782, -0.054020812156814028426, -0.023544749409305204546, 0.032309904693315440183, 0.044651979059421559315, 0.037253409432787883138, -0.068795126318979216151, -0.2044440150375245735, -0.065441104415602663336, 0.53799964403829281245, -0.56074399507140937349, 0.5384161515586917135, -0.064836105228951682045, -0.20467798422788710533, -0.072510724487503439017, 0.027411296438645513934, 0.048407074625980914262, 0.0076872198040868625415, 0.064937312528914020726, -0.14437885396628868939, -0.019520070947949476337, 0.33216686797962363675, -0.27772181956545205495, -0.22949726502559697483, 0.54930238322615010382, -0.23034917938954652117, -0.28077696432557291972, 0.32916271252007689441, -0.0098630543865783504992, -0.12068087042346455073, 0.012609859665885797195, 0.025158845893570324498, -0.00055535040672780402495, -0.00052536445156871444846, -0.025977946834887431782, -0.12703251099659335077, 0.23077765750987458415, 0.63309124612116041142, -1.4224075693150370903, 0.63392378148966788132, 0.23343882410276203809, -0.12373969940540439916, -0.02551204554790596768, -0.0064923391334317870696, -0.0059437715909167196729, 0.0066347172521551136745, -0.00084938170707503357886, -0.0059807379342632933233, -0.020339410451096121002, -0.1068564672194780435, 0.20428526392863533001, 0.59686096920564946089, -1.3369450398626967491, 0.5977146596507463272, 0.2067072887778237944, -0.10413151608992657349, -0.02001169400621478911, -0.010672043883132418737, -0.0068677338666355688609, 0.0066729682229711585056, -0.038516484277291075655, 0.14010846028018233977, 0.055871655667692921021, -0.25421039800270978093, -0.19473195638701806809, 0.54132925042145696448, -0.49560601434817674882, 0.5426230733260662964, -0.19508359686159904012, -0.26151954920185244147, 0.039672234093956794398, 0.13527586182870784959, 0.053181898012462612557, -0.066545648427795520141, -0.037558975450812444319, 0.11788344199556204106, 0.056627166190003408452, -0.20885116608797113513, -0.18556630062464421482, 0.49493201227112504359, -0.47202627602836472187, 0.49622085311288566256, -0.18533283461277824666, -0.21447791479264666226, 0.04246976190723165262, 0.1119688954237587214, 0.040667326821331915843, -0.055587588999571994897, 0.047318022225743855413, -0.069651664994207690418, -0.16049574696194096601, -0.078118521154924416838, 0.62557534417130422622, 0.68355683070241368693, -2.0973794054366048378, 0.68456828682595827207, 0.62484735218335596407, -0.081338166885720181165, -0.15794146510481610335, -0.059110553212537962631, 0.026562401459676119458, 0.010682933559341642807, -0.10237990112098668827, 0.090743545702097208605, 0.25423069942210674288, -0.087984230000002092154, -0.5929848279283908008, 0.31425419883106503471, 0.24681478739706894099, 0.31656346862805340159, -0.59030141585536166637, -0.090399607450966326661, 0.24131798089068748348, 0.079184927610098737594, -0.053482985753266674978, -0.026868130577951811894, 0.0002169035116639832434, -0.029305388052974636753, -0.056280565403827004234, -0.020087467330023384832, 0.081115278590038281115, 0.42499360257308771427, 0.480556244057019466, -1.7593638573128038693, 0.48394155945084171844, 0.42736406496715068082, 0.075751915128911648245, -0.036124630066330820632, -0.067823219440592774632, -0.0040179086039462998114, -0.077885654068290163798, -0.0086498723319216892086, 0.17065682859725733733, 0.33852139863051999846, -0.11721031981548712952, -1.1977015823537351391, 0.099490557805598220376, 1.5860789361693696797, 0.093782537571003327104, -1.2039007816890840985, -0.10870775079322174406, 0.37253534915557084961, 0.18787993145427536734, -0.13463130134382653158, 0.040959844150631169457, 0.10706862117479067076, -0.11533435042226350409, -0.57218761578268551027, -0.26224504603109716205, 1.9008247391004395865, 1.139596368821858885, -4.476399215978722701, 1.1427559127063751099, 1.9025738315530389411, -0.26915516217293139478, -0.58987974653749875031, -0.11744434913602211368, 0.16900639633201466894, 0.047168382054235742651, -0.025860235011985258785, -0.1104617331250082185, -0.12322308896959244251, 0.030682954078364008121, 0.61097031666108003733, 0.5805545840063138785, -2.0187077858999789015, 0.58188303337983282582, 0.61062681906508631258, 0.02349171268033982754, -0.1383529000187413982, -0.11077222976660457132, 0.040471605536885911503, -0.030632183176698461746, 0.029537349325294993269, 0.11321881224367189023, 0.14209319069666209723, -0.225029974302497493, -0.68882434935894920613, 0.37534276733467558662, 0.56822000698368246852, 0.37621348406129023756, -0.68628867581529362507, -0.22017130766963088906, 0.14744943385642922395, 0.11038453705134923521, -0.008755257399919415956, 0.062061292671518941688, -0.092517143025383441479, -0.20356274530015330826, 0.11763236652851172259, 0.54469420670938140105, -0.30630760999772155273, -0.34548155502802940786, 0.44752226803752359618, -0.34447369640996472473, -0.30592818634259544286, 0.53975949167730563616, 0.099274043408820414336, -0.22101387407299441512, 0.0060377075242128317931, -0.062950482807588539758, -0.12334011391771389221, 0.10411338244488305138, 0.68767638163345656821, 1.1189416240538527791, -1.0595951430391403747, -12.053161157651626212, 22.773955152512833422, -12.057472664558678943, -1.0575944399929908535, 1.1369405785713526669, 0.70789551566075192834, 0.091081784858607670041, -0.20503368345187186761, -0.014081854309026428998, 0.10010349482373816221, 0.092032875197710958814, -0.040559271432700709503, -0.36043023046851552227, -0.37210464175875362924, 0.65481601482107310019, -0.1233540817550720925, 0.6597379575414169528, -0.36399865555117155713, -0.35381999095129673671, -0.043705612199207120872, 0.076988084569947698443, 0.092520195085070117136, 0.054933113771212585541, -0.080666552156935522611, -0.15675443096699859247, -0.11743125296478812425, 0.16744798113621950364, 0.8760035258989812279, 0.5329999833951389876, -2.5501120581118983388, 0.53030634911969509737, 0.86851882729161566932, 0.15513219998041061221, -0.12568115000444898821, -0.13723874142923819797, -0.02240995056605857641, 0.0099400467386176645512, 0.011644601810387570501, 0.0096231564801754054111, -0.042753131364747211374, -0.1776885747745321209, 0.09678506842750403405, 0.62774661322994007584, -1.0727437557443475491, 0.62767018504280325253, 0.095880950270232245813, -0.17864847404274017406, -0.039949074628293175515, 0.020375910995791571279, 0.010957513762535104826, 0.010854951115495123717, 0.0051149264273593082059, 0.00016658798869453959185, -0.031808842327739728084, -0.13871411575469017041, 0.086373554787568099234, 0.57843251897079228652, -1.0225704607532968282, 0.57847196948323154064, 0.085509101743262150297, -0.14008370366325845202, -0.030283360622176121874, 0.0093121867787718196757, 0.0081120530303559826302, 0.0073995620384575263792, 0.017940274894094117297, -0.0051240198910398083998, -0.079300788406958727661, -0.15358552122066163048, 0.30445569890329499119, 0.64854070858824386381, -1.4800846314235736045, 0.64959347290757840643, 0.30521858104035060055, -0.1553679207547196961, -0.084444726150800827891, -0.0053747031603141436767, 0.030597746901695716676, 0.0045111782873375853853, 0.019407167013704297587, -0.0022106340360420633807, -0.070608700482302125256, -0.14579356444328273046, 0.2614632064443873638, 0.61916206107467308684, -1.3713750950236580639, 0.61990607559227506318, 0.26213232127911534342, -0.14677426398316983924, -0.074390344204953107177, -0.0030080333001208810727, 0.028048170387099537193, 0.032309904843615580572, -0.023544750068861468317, -0.054020811265600460205, -0.054990236392280629485, -0.02261671949860570921, 0.30168193227153650149, 0.49501974800514331188, -1.3474974635743015572, 0.49812092516140282772, 0.30251048069118458006, -0.030781865525613610196, -0.068791600335533817367, -0.048481549077804025893, 0.020416524136536259859, 0.0076872196822418172035, 0.048407074978277835064, 0.02741129617942458796, -0.072510724637528611858, -0.20467798401800402219, -0.064836105145556599538, 0.53841615153281230377, -0.56074399519934803315, 0.53799964402325861634, -0.065441104289206131206, -0.20444401526071651065, -0.068795126256899666317, 0.037253409667505114899, 0.044651978918316792555, 0.025158845865680429243, 0.012609859701120294806, -0.12068087040947443611, -0.0098630543161551120779, 0.32916271216916637155, -0.28077696381343003607, -0.23034917975423077618, 0.54930238326140767846, -0.22949726464700143103, -0.27772182024715813542, 0.33216686863568761767, -0.019520071453833189729, -0.14437885360387245348, 0.064937312394369950397, 0.0066347168051992845561, -0.0059437704368585806899, -0.0064923401407523577758, -0.025512045329132065907, -0.12373969914413932014, 0.23343882391604811599, 0.63392378124912163173, -1.4224075688209347756, 0.63309124560933049963, 0.23077765786626430744, -0.12703251054909636686, -0.025977948161523385651, -0.000525363342497838187, -0.00055535073038545722073, 0.0066729686966344554586, -0.0068677353153493774993, -0.010672042293168535962, -0.020011694407150875524, -0.10413151698334437156, 0.20670729009517266328, 0.59771465880286034356, -1.3369450399993338952, 0.59686097002317073379, 0.20428526330567398084, -0.10685646732768258604, -0.020339409491377965361, -0.0059807391188862286599, -0.00084938121747275472078, -0.066545648005092092325, 0.053181896853416184279, 0.1352758629772549126, 0.039672233731953887581, -0.26151954956534811947, -0.19508359626830754086, 0.54262307295811040841, -0.49560601435530959868, 0.54132925068722681505, -0.19473195668670725822, -0.25421039803161682435, 0.055871656349250953255, 0.14010845945046171424, -0.038516483971674259079, -0.055587589329761023793, 0.040667327887445953172, 0.11196889432686386867, 0.042469761967921917933, -0.21447791404063934473, -0.1853328354059987082, 0.49622085336840837666, -0.47202627541932634658, 0.49493201119505053631, -0.18556629986712031011, -0.20885116580912063688, 0.056627164915244165844, 0.11788344315344846669, -0.037558975821081787283, 0.01068293341591158678, 0.02656240183985976791, -0.059110553689043755787, -0.15794146471822004263, -0.081338166898864347476, 0.62484735189638795916, 0.6845682868707099189, -2.0973794052003711386, 0.68355683057651228651, 0.6255753440517821673, -0.078118520778117678249, -0.16049574751699424713, -0.069651664563336587732, 0.047318022088816087678, -0.02686813036955681272, -0.053482986177650251869, 0.0791849280166307401, 0.24131798030257328724, -0.09039960677644744147, -0.59030141627553256445, 0.31656346888931519157, 0.24681478755429792082, 0.31425419828348560847, -0.59298482768103610852, -0.087984230125753029594, 0.25423069969991879402, 0.090743545654707311043, -0.10237990120072598677, -0.010239737717825190405, -0.0020682983794695236103, -0.019915603251678713065, -0.050673229336489739039, -0.019931598595464068702, 0.079297629548438802738, 0.42499360288343041026, 0.48394155950901668328, -1.7545813788246740206, 0.48455388297067386594, 0.41772024400574814251, 0.06095940150138911412, -0.043302321914965975769, -0.049349762417619805199, -0.00010767920752698169362, -0.073999531797759449114, -0.015097197551705424101, 0.170537541877636728, 0.3419194167963853781, -0.11453928520492180632, -1.1977015827054879882, 0.093782537525349665697, 1.5745442671201490548, 0.089517159547088220872, -1.17822701417213338, -0.05312476254389755792, 0.37853436178243754506, -0.013933815227044973536, -0.02070729918072012829, 0.077670198581464316989, 0.097700572061378845867, -0.12430953416520056087, -0.57300464544281504153, -0.26161603364089985879, 1.900824738812751713, 1.1427559125724604527, -4.4663349975654433877, 1.1535892395111122521, 1.8876043813936997928, -0.33747739983284963383, -0.64517561766880737739, 0.1664797987013390268, 0.020166785258868891173, 0.0212491184932072745, -0.026808510431486961417, -0.1066358326743790802, -0.12054659503343342908, 0.031366468464182443998, 0.6109703160224096985, 0.5818830342356923202, -2.0156446965074779065, 0.58264784495941535081, 0.60141524882335695157, 0.0080702236063114694137, -0.13668136363677399481, -0.05320734070803349175, -0.025780203294982175638, -0.0047184574635100980447, 0.03931170303273069655, 0.11036991128083362135, 0.13653903144905157885, -0.22789735255663501334, -0.688824349198939756, 0.37621348451365532739, 0.56786702508853981186, 0.37326420425919337731, -0.69033508323935277584, -0.21345774091863645161, 0.16555537684262322529, 0.083950984995168537184, 0.032984926921596866012, 0.016573857809661501905, -0.092085528353057527484, -0.19458844449745352212, 0.12174198533760996477, 0.54503239235277600727, -0.3063076102461733119, -0.34447369665541166528, 0.44838681480309677818, -0.3458569014467603453, -0.31004412902831962606, 0.53862596211110747024, 0.10124274593348209228, -0.21303201317098380274, 0.008474296992466399947, -0.088500035218766953671, -0.10776940422681575982, 0.1124281785345461887, 0.68561287167823081745, 1.1163601712181239201, -1.0595951437732504807, -12.057472664674381946, 22.758597911571328609, -12.070922463800235036, -1.0288892976130825652, 1.2291263812288955037, 0.74473945279481890136, -0.23841297315267803714, -0.052105177665127115627, 0.049699470598013432976, 0.11168078052915264564, 0.081209497073765013475, -0.053016796398928360545, -0.36697745711173723748, -0.37210464219259714591, 0.65973795775616594916, -0.11661353118224862535, 0.66123992511302598363, -0.37967857959297063308, -0.38371589333260464061, -0.042475450619649780803, 0.20466734023225952033, 0.061253962286175273055, -0.017008813250940210682, -0.093356922822203466938, -0.14614622010782624129, -0.10533550930353408059, 0.17317923520305328333, 0.87600352547555271343, 0.53030634956387223689, -2.5530342469771167835, 0.52936487126777620471, 0.87237735767348956184, 0.16616707017991266815, -0.12657384276304592396, -0.17060555417475026241, 0.011482217258240482241, 0.002190602549714444798, 0.0077662656002881625544, 0.0083456206217253098983, -0.042633728847925149441, -0.17751541973317738754, 0.096785068653495801017, 0.62767018495819315582, -1.0729398588882956833, 0.62744503389525163595, 0.095604578049707594767, -0.17533566586205306947, -0.034773803189629706734, 0.024571731154551666038, 0.011747010783913840529, 0.0013375518547283221649, 0.0018883815827979628411, -0.00023906622755505424482, -0.031391357146205063622, -0.13856967167366676663, 0.08637355500160938393, 0.57847196923185240092, -1.02271283117731282, 0.57805740989014575693, 0.08499045305944821771, -0.13680599164406290202, -0.025454558092627654203, 0.011099489413400061288, -0.0059109774222995698639, 0.013642982870428721834, 0.019689099435390811782, -0.0062869293263942997371, -0.080252915899246338172, -0.15403060954180841646, 0.3044556990912962724, 0.64959347276958989692, -1.4777217361575378796, 0.65072834701631832033, 0.29870021912961042254, -0.16804094076906800348, -0.085284895884204500738, 0.040652880419054522598, -0.006906071580285738469, 0.012770236629957226054, 0.020488399376208540903, -0.0039387310454437563997, -0.071541188212747450215, -0.14602677232205452107, 0.26146320682839530036, 0.61990607563660526935, -1.3692911978983346444, 0.62145408174201310914, 0.25690145282815102235, -0.15878874779448351462, -0.077168043700854449529, 0.040664122992152412772, 0.015276114907440210139, 0.0066473919901221195997, -0.018264074151096123394, -0.047558648424865512805, -0.054619061094386835153, -0.024645228344676355092, 0.30168193261008913186, 0.49812092491415560502, -1.3448151852994549849, 0.49403200822299869754, 0.29094004460084366137, -0.033724076550456845436, -0.054101618501954204521, -0.029878372876973636335, -0.0021846162298297874815, 0.017422342680607338233, 0.04521833123433273427, 0.023555956947793995487, -0.073992916254496998718, -0.2047405071859026704, -0.064836105143601205358, 0.53799964420498835871, -0.56134135374944338359, 0.53706536368663515546, -0.067907902327501087902, -0.20640081898815140771, -0.0689297347452503123, 0.088321118725269404104, -0.012820339133828484793, 0.043698467383013805954, 0.011316824350751219369, -0.12415446038089338054, -0.011612602772266958362, 0.32874492337228722327, -0.2807769637709258137, -0.22949726526210323385, 0.55247158512911065387, -0.22496859243907202375, -0.28111273830653699113, 0.30846379792503014761, -0.04932441774936405926, -0.030788403532511665617, 0.0075891001657510540404, -0.0034940451142422522535, -0.0078590578095245246931, -0.0044614557603399095445, -0.022551800018601179942, -0.12232262308474491086, 0.23343882444417113353, 0.63309124553479401154, -1.4237700898642247083, 0.63222703149756442897, 0.23005781093792543324, -0.12632885525997139542, -0.021700265249065570083, -0.0041494857405714700993, 0.0074823995462607647416, -0.0027200244710894566462, -0.0090872744876091936933, -0.0090268086670996852683, -0.017217727219951742962, -0.10270601796001209427, 0.20670728948206118503, 0.59686096948481415225, -1.3381446124178391077, 0.59648147618130764602, 0.2040606528771093664, -0.10656437017214512308, -0.017559335427997983109, -0.0088583581213393658427, -0.031039565711022589628, -0.026473947798064063897, 0.053086874549454751704, 0.12837588918209952293, 0.036783136230049209847, -0.26143770530203008784, -0.19508359627619087373, 0.54132925015335742192, -0.49645471630036874799, 0.54464472569708888283, -0.18621692626289812478, -0.2467486332667765403, 0.037852552761995117425, 0.10366121335145984939, -0.023091131195941888565, -0.025632013541316372301, 0.04005669865578583011, 0.10693378167392267319, 0.040661909996736515793, -0.21419852910595621465, -0.18533283462381339168, 0.4949320115405003162, -0.47329204801564522809, 0.49716495386703080106, -0.17767530606140891081, -0.19958157857457264894, 0.042112851969844010414, 0.078780298814397450857, -0.0069830993042636442178, 0.02932629945288896603, 0.018944156510177750929, -0.067041242118855592924, -0.15776986981527671428, -0.078788171973337192688, 0.62484735225250465085, 0.6835568304362203973, -2.093476673066368754, 0.69374633201010105221, 0.62581567446822206069, -0.10757605854098709064, -0.18872236784670701204, 0.022969776278580319312, 0.019351979006371353154, -0.052491302608981514621, -0.052882509480253310497, 0.082933889544477626399, 0.24352315241504543519, -0.089338724337344327386, -0.59030141609998854069, 0.31425419865686599064, 0.24167345697520367653, 0.3107541294534199583, -0.58296349812172110205, -0.050578742409240026734, 0.28165542142096428035, -0.075231207080594664216, -0.013119361904290572501, 0.0022110936651046114894, 0.00078158004251216407158, -0.021292396569637026676, -0.052308910895949391195, -0.019931598798217733254, 0.08111527873800909727, 0.42736406457930825598, 0.48455388304314100933, -1.7574528406473359787, 0.48481339908858933629, 0.42354629016689288035, 0.065262227243780682184, -0.10301149144323942797, 0.025361986228584605663, -0.039984603214574178376, -0.070470101363315981735, -0.0044203201317593848046, 0.17506422651456596262, 0.34191941660257169078, -0.11721031940953947903, -1.2039007815826270331, 0.089517159430621953686, 1.5852907133966389619, 0.10994509827486594056, -1.1918687300983275001, -0.075044535916592769453, 0.37872116397906491203, -0.021100713630707861046, 0.013162754275977057647, 0.073290362565401750983, 0.089233288144935210551, -0.12701422080749805166, -0.57300464561247710371, -0.26224504559869787901, 1.9025738318208700317, 1.1535892392821587293, -4.4532237405586894141, 1.1423849471681659651, 1.823307573625615241, -0.41152260627773679103, -0.35345466237509898244, 0.00017144609754877570933, 0.023056358603098115606, 0.018739474330412096009, -0.028410324977002871655, -0.10660551105138231509, -0.12054659551842991649, 0.03068295496467982747, 0.61062681805188012696, 0.58264784550235526606, -2.0149346099779497088, 0.5886171817055274591, 0.61545006589312167833, 0.011975249455043862276, -0.2122939617924265443, -0.014732635269659203822, -0.011305076492173261188, 0.00029094885490527685329, 0.037835469255072889938, 0.10781071790503699015, 0.13653903148609863871, -0.2250299742490760313, -0.68628867679782246114, 0.3732642047249093431, 0.55670710656705457708, 0.35900133256501554069, -0.70328040580718609309, -0.18884616091522027359, 0.25782736582268928061, 0.0052948012211574482788, 0.030216716252638470275, 0.012032162736504216013, -0.092196527475190082068, -0.19349331783005910279, 0.12174198529194174556, 0.54469420683967384456, -0.30592818580469899326, -0.34585690169638377833, 0.44102840552034111576, -0.34703770427952979061, -0.25894585164590777637, 0.66377363075257989067, -0.27433788447253465437, 0.017358479272327968779, -0.021653348662346381948, -0.08413200065545516082, -0.098656751504621581494, 0.11519743885725046428, 0.68561287099407741685, 1.1189416252408785901, -1.0575944401112757909, -12.070922463659963242, 22.720172765915396695, -12.078570740262414063, -0.89964423711926766458, 1.3726216798640895611, 0.28699719005201296262, -0.036911894450664006762, -0.0036685624621880272116, 0.053345528343376448499, 0.10138274291658783133, 0.073440298405151482886, -0.05301679679852847743, -0.36043022991160955648, -0.36399865601023839723, 0.66123992555818533745, -0.13361296453556231745, 0.6199497308421751951, -0.42532047025857183664, -0.33520400192619964619, 0.19947025994080519462, 0.031312648503726864258, 0.023839453709780397683, -0.021798133585991291, -0.086864747296198335791, -0.14045887950129493094, -0.1053355096444719835, 0.16744798148574355223, 0.86851882662884904995, 0.52936487187255232278, -2.5366450066051062784, 0.57104649372974058963, 0.90699320741342803753, 0.11601117459098979345, -0.32271336481254558892, 0.0073172246397331663914, 0.0051631694377219658834, 0.0010247977705563474961, 0.0073923710123539989117, 0.0080574030238808221072, -0.042633728973971676823, -0.17768857469362353552, 0.095880950278842011492, 0.62744503399309148328, -1.069650081787587359, 0.63493585996332835286, 0.087689631599434159082, -0.1983258348648193059, 0.011849850384683677587, 0.0067682392453482353306, 0.0056289616028494257952, 0.00035083630578126943217, 0.001698471373158314153, -0.00045239607854477949592, -0.031391356957728058785, -0.13871411616071982476, 0.085509102347356868989, 0.57805740959839324589, -1.0199439209345066981, 0.58550412895316672479, 0.080258674114662073573, -0.15371898386985663465, -0.0007704158571501312433, -0.0075924384289690893812, 0.0039877370852594963904, 0.013999286977976484192, 0.017729100012090593524, -0.0073471969170743472999, -0.080252915505957367337, -0.15358552181899959299, 0.30521858132549906095, 0.65072834687990199765, -1.4781397281369001995, 0.64793604362948264974, 0.28688493857830593692, -0.17952676903293016131, -0.020807057415730554478, -0.0068516122217307994999, 0.0022242549512848375226, 0.012780084263893300156, 0.018721436671219490733, -0.0047472844390098242057, -0.071541187747122009233, -0.14579356517441507957, 0.26213232134878211577, 0.62145408169398208553, -1.3690141353366078736, 0.61814555575432328371, 0.24325693537672987699, -0.16834303362703414986, -0.013114264721892440399, -0.0035946005618902464492, 0.019083562779603865933, 0.0093305598366766641372, -0.019973181960866764201, -0.04985472843579112634, -0.054619061037302039929, -0.022616719716682255614, 0.30251048080338571733, 0.49403200800825503025, -1.3507921354949079529, 0.50210417685864161541, 0.30903192280506985856, -0.036259832040810069387, -0.098972042675105190157, -0.0029270320567225025475, 0.0028417988573463549146, 0.018034674554320496292, 0.04378363797214705061, 0.022652609154448225676, -0.073992916054550633698, -0.20467798430366071494, -0.065441104477307221465, 0.53706536388424153117, -0.56201117896539065644, 0.53457336330674343206, -0.089687156318619815076, -0.21277189556891157318, 0.049879394298015224118, -0.013449662593177794459, 0.0089162259154377566939, 0.039402010687688629464, 0.0069750002874674216918, -0.12493608214793253119, -0.011612602583371690879, 0.32916271190086804355, -0.27772181910956150741, -0.22496859316777317872, 0.54705449974169506611, -0.24982860637019174699, -0.28992652709410449141, 0.41891135684730856914, -0.1587365883465007288, 0.0028346741518548166164, 0.0021182514188726664014, -0.0028852299535431398525, -0.0056217443946731844584, -0.0032245915831009795693, -0.022551799443368193698, -0.12373969996436449958, 0.23077765807193054148, 0.63222703142243508179, -1.4196049676419080221, 0.64089224324187077375, 0.22832780985983364141, -0.14599594674719124088, -0.013540743291169225065, 0.0032862872666277174495, 0.002061230664352913227, -0.0026626845056211554189, -0.0070639576256377508562, -0.0077575958520578712496, -0.017217727783065472219, -0.10413151615589813503, 0.20428526372308358772, 0.59648147585723521047, -1.3338630660125130323, 0.60419742694334321875, 0.20224447596926825588, -0.12155818397905046591, -0.018257907219653306263, -0.0011771802321678922262, -0.032916248651168926453, -0.024292740502580581485, 0.054229357302264712126, 0.12858871568422822196, 0.036783136353272447017, -0.26151954978860325607, -0.19473195584845751083, 0.54464472507975769666, -0.48883863762701640354, 0.54531463744503172464, -0.23191070475410943685, -0.30121414534497698012, 0.2283728637914637094, 0.0010158820475683633805, -0.027582785287146936726, -0.023594378828440339818, 0.041660230359975472525, 0.10739330968040323999, 0.040661910741861276297, -0.21447791500823445987, -0.18556630018039699248, 0.49716495406481336916, -0.46625560618441563854, 0.50051320930408926824, -0.21463234537921985212, -0.24681552380694313231, 0.19173868349363720753, -0.00049300911123922191732, 0.0013895102739615709277, 0.021314342845071089505, 0.015952156759668822961, -0.065561614960782915729, -0.15776986969538167593, -0.081338167013558809737, 0.62557534417709836916, 0.6937463319656665961, -2.0826820477419860644, 0.68526216165173636785, 0.59689565415942325011, -0.11518826201052494163, -0.13795609494768665271, 0.021208227918224218517, -0.0097780964160314367412, -0.051797704229143176624, -0.047198724964303447049, 0.085410204975161316265, 0.24352315268373792012, -0.090399607236079967953, -0.59298482791837447969, 0.3107541295461730968, 0.24470524612783575713, 0.3309370627279179522, -0.56284693911597405425, -0.069313735273289808769, 0.18992805649495447806, -0.0057409306169226327621, -0.0040612203501286629759, 0.0011498068958168141361, -0.0015867081945249051234, -0.021292397200969460924, -0.050673228551655014051, -0.020087468109170544817, 0.075751915855500465136, 0.41772024380009181721, 0.48481339937615941516, -1.7346654901312223984, 0.53312892067727846168, 0.40938775409890537649, -0.080025255268447115231, 0.013611970195103696366, 0.0031801972897376917966, -0.035866525749936740974, -0.064692875822624132387, -0.004420319921006620606, 0.17053754173102714242, 0.33852139861863389525, -0.10870775101931888507, -1.178227014047493304, 0.10994509835590493552, 1.543141265179932109, -0.021051481132890539211, -1.3328426177851364631, 0.56563473615520509785, -0.0059578317912308206838, -0.012643681823144644824, 0.013043987102045052906, 0.070865053287174073482, 0.08923328851366467418, -0.12430953433332909974, -0.57218761600673195833, -0.26915516252491961113, 1.887604381674660825, 1.1423849471293974212, -4.4447338629759425288, 1.2445274573548064101, 2.0275236410575279145, -1.0465229938814701871, -0.0049105010624127479885, 0.0080458834540194576479, 0.022137575589345107052, 0.017273739869193928398, -0.028410325401511414556, -0.10663583204576111696, -0.12322308911644921925, 0.02349171277994132126, 0.60141524816319713587, 0.58861718219236869309, -1.975600916955892572, 0.65973785798696127358, 0.59734436143771885153, -0.27670706552314910942, -0.0038923998563134901141, -0.010903212247905512758, -0.011027146479990684175, -0.00048609590383258914886, 0.037835469190644469006, 0.11036991132323192844, 0.14209319052884322732, -0.2201713069361312114, -0.69033508352444272749, 0.35900133257516964047, 0.53442490298066824828, 0.3225634173393202575, -0.76740578771393219704, 0.19324503798221381867, -0.0041370763845159710947, 0.013211559732854920302, 0.028530841088507567088, 0.010365764780622213972, -0.092196527283901805183, -0.19458844465489982811, 0.11763236667009190073, 0.53975949120112443058, -0.31004412890006666226, -0.34703770424376623183, 0.47106387212591349645, -0.23389381126988909032, -0.27495511986414339356, 0.28167584920559607387, 0.0076961932681731803213, 0.0061738890635948213623, -0.021907024752270225915, -0.081236408312142252552, -0.098656751651568300177, 0.11242817916131474243, 0.68767638061972247154, 1.1369405792441236169, -1.0288892983480475429, -12.078570739752152008, 22.601689839710967789, -12.209489803504869343, -0.69701076824116547925, 1.6634191401514397679, -0.0086573908273594873952, -0.024634987053205652241, -0.0057510052944769324279, 0.048460669153222356298, 0.10138274292266513665, 0.081209497605252783936, -0.040559272158962419996, -0.35381999010391734783, -0.37967858052577813099, 0.61994973117967833165, -0.17358457163073745533, 0.6365505243314035777, -0.38626442680441208655, -0.12297271688427713343, 0.0027346119373659757111, 0.028123556041166265784, 0.025008994903252975556, -0.019394561897532369005, -0.086864747611795786741, -0.1461462196246329226, -0.11743125329652294442, 0.15513220068930863582, 0.87237735661142745691, 0.57104649431553056171, -2.4718897366763799006, 0.61543155936312277099, 0.84281419912137089945, -0.26338169098696267989, 0.0019137383239843182205, 0.0056164987477744362718, 0.0046302795342206531232, 0.0007190468500937684769, 0.0073923708630659057395, 0.0083456210683441663795, -0.042753131914685618686, -0.17864847376637849052, 0.095604577935870696526, 0.63493586004140267764, -1.0617934396236154981, 0.62875090139292277858, 0.041468085813015316543, -0.14726019837952805669, 0.0013912050983718973925, 0.0057469471895387307103, 0.0051309533743372137335, -1.6869281173567370405e-05, 0.0016984715900386646373, -0.00023906663722027265973, -0.031808841804275483911, -0.14008370416435309003, 0.084990453340907448632, 0.58550412887941016837, -1.0106047005636964453, 0.58454280322581086971, 0.034600403499207219027, -0.12144959925325758687, -0.0033318661035334601848, -0.0031793819496252579901, 0.0038679851084644886923, 0.013135321618755978895, 0.017729100208923626036, -0.0062869297909754701462, -0.079300787824862994935, -0.15536792101879390304, 0.29870021950376562936, 0.64793604328464404762, -1.4693504892498105985, 0.66982485269686986573, 0.27342517280529876622, -0.20849048881027004709, -0.0024087820145328809121, -0.003766792374999941833, 0.0021873047335241378557, 0.012221145766726577292, 0.018721436917017972995, -0.0039387314984551270031, -0.070608700015733619204, -0.14677426425190839843, 0.25690145317862478347, 0.61814555543476690236, -1.3637768171521169602, 0.63775011208051290978, 0.23221260340665084154, -0.18757801434085605297, -0.0077606308382698998075, 0.0084675402665675865432, 0.01780579676708676784, 0.006173005549183694686, -0.019973182021961095184, -0.047558648610134569845, -0.054990236250578389265, -0.03078186534655147702, 0.29094004508328080716, 0.50210417638848303667, -1.3161613837786889647, 0.53624981409610206207, 0.27071845330043398103, -0.15419113455626801135, -0.0024631258552153222074, -0.0010947713883451553208, 0.0038198686125296078993, 0.018052182856651995818, 0.043783638141997989479, 0.02355595683623723513, -0.072510724429362835575, -0.20444401509516466064, -0.06790790257055276391, 0.53457336328584392771, -0.5731486161033835236, 0.52823124073399552181, -0.078482797561700817779, -0.15581763973726203698, -0.0028316543709338230089, -0.0091765212074469415526, 0.0081991036648359581079, 0.038352507120315681988, 0.0069750007893280225116, -0.1241544611541815224, -0.0098630535387566903999, 0.33216686726757682635, -0.28111273726967378472, -0.2498286067157038659, 0.52276818336897890038, -0.15763140884753212845, -0.17623661802841666946, 0.10269956183235696978, -0.0017491840562857642258, 0.0035707104925690263085, 0.0038793783724475582042, -0.0019302113791140988741, -0.0056217440800322175318, -0.0044614563538553355868, -0.025512044757814083401, -0.12703251134613652296, 0.23005781126078284071, 0.64089224329757288334, -1.4032919541218635828, 0.64949217983312190761, 0.17044899267522903297, -0.12817232666908037753, -0.0010267718444333907239, 0.0033402162087951942603, 0.0035636732284807202953, -0.0017162270325510628304, -0.0070639582328094153649, -0.0090268082031740963511, -0.02001169449916606613, -0.1068564672109218322, 0.20406065304241149949, 0.6041974267376244434, -1.319191412017812759, 0.61512111031331107025, 0.14754692394692925372, -0.11224133301910144134, 0.0080688089300954505406, -0.013050070928783691002, -0.03223201571581224556, -0.022208216271885294901, 0.054229357133920470802, 0.12837588883195619194, 0.039672234483254777138, -0.25421039866093175785, -0.18621692553188004537, 0.54531463691421289131, -0.52755787397151565621, 0.46977816120552273738, -0.2416718830391220707, 0.029681630113123233516, 0.0069675277058464362562, -0.009342142984344849882, -0.026796773896651481656, -0.021658939199842557749, 0.041660230750801267863, 0.10693378113896179205, 0.042469762236956351553, -0.20885116610375925594, -0.17767530591296165632, 0.50051320929902098911, -0.49728270593826673363, 0.43472016812111391459, -0.23120559312778929395, 0.038074350369200635202, 0.0059248016483758898093, -0.0071251226043554400508, 0.00029608715780005855555, 0.022465056393711171911, 0.015952156909933200291, -0.06704124209684511615, -0.15794146530600136624, -0.078118520576974867509, 0.62581567441889673908, 0.68526216177198973067, -2.0829704943502127534, 0.74120797441946950634, 0.62879333166327389115, -0.33117366517442220486, 0.0081475395009074705333, 0.010851758160322749411, -0.0099799483948758548751, -0.049711557905945835678, -0.047198724990330988283, 0.082933889409080793764, 0.24131798070408963519, -0.087984230097018015226, -0.58296349794489787932, 0.33093706237840964102, 0.2554259650537867854, 0.26443859082702270147, -0.74197803752991586812, 0.32696469812349782114, 0.0030573325188802320286, -0.0057034951426642808286, -0.0058669079109178639067, 0.0011498078139578508591, 0.00078157978148105989626, -0.019915603842639538479, -0.056280564629752484473, -0.03612463093714320117, 0.060959401980211193328, 0.42354628965227764104, 0.53312892103611164707, -1.7052561819484053363, 0.5604033740305468525, 0.2489495366739621629, -0.00099389325417653814031, 0.015186943150629628185, 0.0052188262967319820618, -0.035866525512289568201, -0.070470101869292547936, -0.01509719700787496427, 0.17065682819466398978, 0.37253534952404787628, -0.053124762739223958807, -1.1918687301231405407, -0.021051481070786394467, 1.3716735661666734636, -0.062252585877550850602, -0.49364030177432866031, 0.0024170891084420802718, -0.010085085656371480314, -0.012813958016868062795, 0.013043987564435933699, 0.07329036175935985109, 0.097700572618274500458, -0.11533435047777507165, -0.58987974618334393373, -0.33747740002347359489, 1.8233075735928065964, 1.244527457411057858, -4.0904684695143807005, 1.6874786201085276716, 0.22563641081584204229, -0.0033615188879318738321, -0.0011734395008139069175, 0.008372475710298346932, 0.022137575155641835839, 0.018739475123139975876, -0.026808510825759811896, -0.11046173367833636569, -0.13835289928928440339, 0.0080702238480126822773, 0.6154500651884118323, 0.65973785830907571981, -1.9299638465737176674, 0.67686826303817992923, 0.20771299778508398037, 0.002736386416829720853, -0.0068358214133350334724, -0.012198614691408518965, -0.011027146767132317562, 0.00029094910555736198682, 0.039311702871432825657, 0.11321881224865899818, 0.14744943401763321855, -0.21345774116510912721, -0.70328040560968607675, 0.32256341735965066153, 0.48478693453088422816, 0.3500791666729753393, -0.52195596683488032497, -0.0040967549123334294839, 0.001050749726186054776, 0.013538995573174348111, 0.028530841398004963549, 0.01203216232600626788, -0.092085527995925231526, -0.20356274561666179435, 0.099274043409235984692, 0.53862596234392945416, -0.25894585180944856839, -0.23389381122308502969, 0.41636368164694065319, -0.79794498822159587448, 0.48633999233536878704, 0.0002315136074833627983, 0.0085827126224725970582, 0.0059963314311018162811, -0.021907024872122541487, -0.084132000582851737547, -0.10776940434950958714, 0.10411338286633593786, 0.70789551473194678444, 1.2291263822629743263, -0.89964423791409064979, -12.209489803144728981, 21.810810210539351317, -12.578220825526335247, 2.0149360886742324617, 0.0078361170259113373943, -0.018708840546822470136, -0.028799632015371157828, -0.0057510053742389714154, 0.053345528489817058138, 0.11168078005677825604, 0.092032875976657393946, -0.043705612949440295067, -0.38371589246226361913, -0.42532047079934193201, 0.63655052446282489598, -0.080428105507078267999, 0.73502820144908709299, -0.65808401128979610029, -0.01013342359404777529, 0.014511769425507838049, 0.031588196099786254589, 0.025008994712462956356, -0.02179813306785370608, -0.093356923533167240903, -0.15675443037564776305, -0.1256811506292331293, 0.16616707118514181807, 0.90699320678927286554, 0.61543155943981964207, -2.4756767038486282573, 0.63023469053995473388, 0.50642393920116723915, -0.0015723915830392271639, 0.0025727917725284540212, 0.0054103358384171287496, 0.0046302794155419976785, 0.0010247980758508535345, 0.00776626512824717119, 0.0096231570866327203334, -0.039949075176039745838, -0.17533566556568688788, 0.087689631349931657089, 0.62875090159459023909, -1.0518977795661006436, 0.73436318097840624919, -0.21143084091875105623, -0.001619093991191928282, 0.0023985738125467683489, 0.005538617445305107044, 0.005130953352626692332, 0.00035083618983162393086, 0.0018883819255384425018, 0.00016658756884988944365, -0.030283360504833566063, -0.13680599173446819639, 0.0802586741137156362, 0.58454280324023266679, -1.005156277737335957, 0.66137985264076804004, -0.16603432383541205652, 0.00039241435151034004172, -0.0038000838884792698634, -0.0034379761596865822393, 0.00386798510180032468, 0.013999286883654721986, 0.019689099582957649004, -0.0051240201023894206903, -0.084444725983750607012, -0.16804094128618071724, 0.28688493922718638629, 0.66982485233947053338, -1.4207371297465329896, 0.80106536551936735613, -0.10884068429321165639, 0.00067107546714307501727, -0.0033103144874343033884, -0.0038748392909133465116, 0.002187304880198127216, 0.012780083992708194535, 0.02048839950234957949, -0.0022106339939920852979, -0.074390343779132089841, -0.15878874834367204261, 0.24325693590988869963, 0.63775011182921104158, -1.3163946913582382781, 0.75708675755184173806, -0.11422698965836418361, -0.002852571787127614316, -0.0032685345058501460817, 0.0067971468354951674121, 0.017805796521836617791, 0.0093305599841246502496, -0.018264073924233830554, -0.05402081163784664547, -0.068791600584224274484, -0.033724077127620190331, 0.30903192370205817552, 0.53624981359194912045, -1.3059876230432925226, 0.56313698171768489242, 0.048054956151202324333, -0.0022091067585110367097, -0.0019555518829586513233, -0.00060804124946150605972, 0.0038198685498729203353, 0.018034674571241266633, 0.045218331084199071823, 0.027411296295979801357, -0.068795126369313258641, -0.20640081882189689755, -0.0896871562339446593, 0.52823124066944915356, -0.51385399949850840695, 0.59938224584382504556, -0.33988082742942943293, 0.0023876097496866462676, -0.0065142839895634182396, -0.0090262605301165872718, 0.008199103952277186913, 0.039402009835366978707, 0.011316825531986608386, -0.12068087147803108206, -0.019520070160689684258, 0.30846379703133647299, -0.28992652681053865393, -0.1576314088614014508, 0.61314057202406369029, -0.56401922318289543501, 0.18495082355148889652, -0.0030946889782631657392, 0.0020044505980970957898, 0.0048214830626416320045, 0.0038793784093378944761, -0.0028852302277328832451, -0.007859057658128122334, -0.0064923400435368408892, -0.025977946854339822869, -0.12632885571410953918, 0.22832780955679246593, 0.6494921802122988197, -1.4045244068296098838, 0.7633565606383968527, -0.072901038346917690358, -0.0026759679454524689345, 0.0021555301707776208971, 0.0045331987316083247053, 0.0035636723842805287724, -0.0026626834694422531896, -0.0090872745857043541601, -0.010672043121164948715, -0.020339410290160460371, -0.10656436989528571269, 0.20224447621255822738, 0.61512111013609793719, -1.3209103750519868292, 0.70751397217429223829, -0.060348648740793708889, 0.0063310204927960267787, 0.0013310735186422319176, -0.013442973188747540589, -0.032232016150418735245, -0.02429274008695740672, 0.053086875038197839372, 0.13527586183592610336, 0.055871656399425956296, -0.24674863379734807145, -0.2319107039707637985, 0.46977816071088762673, -0.50179970395263784333, 0.69148147644353818997, -0.36784904441695370769, 0.0046763423515928598598, 0.0021232338379847575174, -0.0095291573754028788268, -0.026796773571180692447, -0.023594379266310909482, 0.040056698599839235531, 0.11196889500768707204, 0.056627165790185111927, -0.19958157910836762938, -0.21463234539525105626, 0.43472016826607384754, -0.48175748891602937496, 0.61276082178854951454, -0.31117516193426020266, 0.0036857353804898647549, 1.610015265899291111e-06, -0.0058160916093590022888, 0.00029608742144980018836, 0.021314342519455343333, 0.018944156277910177699, -0.059110552647476943822, -0.16049574799989149709, -0.10757605825447984216, 0.59689565393663779336, 0.74120797455933729037, -1.9779307967242376609, 0.8581130814957874664, 0.075882332918880857076, -0.0012630806699738790846, 0.010199975445248074535, 0.01118550672592766998, -0.0099799485112047284924, -0.051797703943128658999, -0.05288250965721935648, 0.079184927926495507489, 0.25423069967548073134, -0.050578742949905372561, -0.5628469383023603223, 0.26443859022443338969, 0.080646457932704862515, 0.27329512615744810944, -0.24648849925992269427, 0.0071802845442775469625, -0.00085988931979054470528, -0.005703494058951014134, -0.0040612225282648335353, 0.0022110950093918915055, -0.0020682981012382431474, -0.029305388385987164029, -0.067823218925894185327, -0.043302322443198787427, 0.065262227665539407351, 0.40938775389687037665, 0.56040337392159900176, -1.7375422813402394695, 0.84000014063793060348, -0.0047803783566972880836, 0.006806695969090060562, 0.015186943192214634457, 0.0031801970423920851265, -0.039984602794935854286, -0.073999532226690381553, -0.0086498719530482860718, 0.18787993114248680016, 0.37853436201027162999, -0.075044536083931731585, -1.332842617695008558, -0.062252585888671961578, 2.3030029263435412012, -1.3042134499431168315, 0.0030779781600788385168, -0.0020613995627719963223, -0.010085085565712808772, -0.01264368224671739982, 0.013162754918354304529, 0.077670198150428790274, 0.1070686212660112574, -0.1174443493840587327, -0.64517561755532937262, -0.41152260624699643632, 2.027523641000183563, 1.6874786201349420978, -5.3349056153006753078, 2.6305923594709210711, -0.0012316200412533452515, -0.0029111091655067565694, -0.0011734396531975872444, 0.0080458842381454908493, 0.023056357958534289682, 0.021249118378808735247, -0.025860234656280650656, -0.11077223023956128678, -0.13668136331417890528, 0.011975249927003019795, 0.59734436080286001935, 0.67686826313691128565, -2.132124711487698665, 1.0769099461339421975, 0.0041988496428049396428, -0.0011187521909719788833, -0.0068358217662345850674, -0.01090321189580494704, -0.011305076722367369754, -0.0047184572087883309283, 0.029537349359452055864, 0.11038453659101736537, 0.16555537713171050718, -0.18884616118143254804, -0.76740578768620204553, 0.35007916683996742613, 0.98471797418591411333, -0.65568196066484285556, -0.0019028896432820561514, -0.0027658648562474638134, 0.0010507498572898004742, 0.013211559402480853656, 0.030216716610071681776, 0.016573857503789448525, -0.092517142773175117809, -0.2210138737823717503, 0.101242745613020102, 0.66377363087563590049, -0.27495511990758142451, -0.7979449881747328055, 0.76125984124888657067, -0.1983158381409624027, -0.0016621072151076245454, 0.0034128143219856431345, 0.0085827125223273351834, 0.0061738892870900324156, -0.021653348736198441832, -0.088500035380423627829, -0.12334011391125164481, 0.091081785462785308516, 0.74473945214657588654, 1.3726216803782853582, -0.69701076852366328218, -12.578220825458293675, 19.582898354742400926, -8.3336245869437082234, 0.0086757700611075991987, -0.0043715226607763816094, -0.018708840632170087998, -0.024634986510793245584, -0.0036685629281615725512, 0.049699470702760234486, 0.1001034944422323214, 0.076988084805294315638, -0.042475450985415878769, -0.3352040013528427842, -0.38626442714183889526, 0.73502820152393100184, 0.082425170854342963422, -0.23357985199910041851, -0.0092633393182349858808, 0.00013816624715735776317, 0.014511769413279029806, 0.028123556275694877593, 0.023839453199321233162, -0.017008812657593894158, -0.080666552656597348414, -0.13723874101696759142, -0.12657384338275812219, 0.11601117486715581406, 0.8428141992209929878, 0.63023469057902470336, -2.7209278360953916831, 1.4412914327031995931, -0.0031367955847044974409, -0.00090326113867208242692, 0.0025727917800889210306, 0.0056164989000726172827, 0.0051631690973704159886, 0.00219060282715388948, 0.011644601464814321665, 0.020375911489254161513, -0.034773803582692663916, -0.19832583461428063276, 0.041468085697484349805, 0.73436318100097153216, -0.98862685385544757377, 0.40874153127785040063, -0.002733798681473250617, -0.00094813237413060829761, 0.002398573821448759906, 0.005746947256688408022, 0.0056289615824765600743, 0.0013375517356401188905, 0.0051149265047964144124, 0.0093121868802985494074, -0.02545455815010394876, -0.15371898386472235276, 0.034600403595923819033, 0.66137985256234177456, -0.92803213080215629649, 0.3907326602478261024, 0.0016019245342026645386, -0.0012121442240938464126, -0.0038000839551327473056, -0.0031793818605377957565, 0.003987736978665341013, 0.013642982943913167793, 0.017940274991652422482, -0.0053747033656143458624, -0.08528489552100182447, -0.1795267694833663219, 0.27342517313234404597, 0.80106536543690032293, -1.4900386055251626516, 0.65941889022682353971, 0.0014055789214018322118, -0.00068896029453417641664, -0.0033103144782811429539, -0.0037667924975072646955, 0.0022242550704528823123, 0.012770236693450558174, 0.01940716691472152422, -0.0030080337632647767034, -0.077168043293479565703, -0.16834303394868077586, 0.23221260358518150402, 0.75708675754056087293, -1.3643459789014551475, 0.59792125254488759367, 0.0012276065581690762899, -0.0049940343062547130557, -0.0032685346713317983344, 0.0084675403476150400139, 0.019083562678084448061, 0.0066473920463478910342, -0.023544749716004797352, -0.048481548903317184251, -0.054101618294249510799, -0.036259832868630333114, 0.27071845409087830214, 0.56313698149928281822, -1.3416812117003120708, 0.64605931967302887564, -0.0021070954571735397119, -0.0020732805494618582187, -0.0019555521343361742873, -0.0010947709532234213212, 0.0028417985410083674257, 0.017422342828394093289, 0.048407074799021503064, 0.037253409633054665462, -0.068929734721034516265, -0.21277189560242668054, -0.078482797606015425318, 0.59938224588783628466, -0.4710632133444698777, 0.13998114325579286144, 0.0021786625680794728319, -0.0003711851531325261952, -0.0065142839589042191017, -0.0091765214428206065556, 0.0089162265465659879676, 0.043698466525705699504, 0.012609860448049955289, -0.14437885444715664596, -0.049324417247570133405, 0.41891135673049861854, -0.17623661802122878028, -0.56401922320954989143, 0.72535092129865874622, -0.26231751022175536781, -0.0012948048911884558077, -0.00035218663075941821286, 0.0020044504187784221647, 0.0035707105598906719625, 0.0021182515972456160587, -0.0034940450085035662173, -0.0059437708965260002691, -0.0005253642746147688114, -0.021700264998415350753, -0.14599594641346869128, 0.17044899220329889911, 0.76335656092029924302, -1.360144749783547935, 0.59963863814952444997, -0.0013722586667920565803, -7.6126659086699587347e-05, 0.0021555297503434577185, 0.0033402171144479462714, 0.002061229813911972901, -0.0027200244849555060139, -0.0068677344784709671785, -0.0059807381795132308691, -0.017559335840003287982, -0.12155818416701388851, 0.14754692407438380064, 0.70751397225084233789, -1.2642987715074989197, 0.55942155081685851492, 0.0027052256462718419355, 0.0052798957281478039322, 0.0013310734158788607544, -0.013050070520615634356, -0.032916248930340320789, -0.026473948231584578522, 0.053181897729166975608, 0.14010845972740210152, 0.037852553070902794641, -0.30121414590372014297, -0.24167188268992212397, 0.69148147632139222996, -0.29175530191298798544, -0.028566669920243974168, 0.0017707115480254276719, 0.0044282167340992367038, 0.0021232339673128417835, -0.0093421433380092566701, -0.027582784939804715768, -0.025632013336588641006, 0.040667327087966928523, 0.11788344240642771588, 0.042112852456840502158, -0.24681552368716094814, -0.23120559343548818809, 0.61276082185427482862, -0.26810945717657941323, -0.016161695812728225224, -0.00091748236357684480064, 0.003727109021317672171, 1.6101252790626770033e-06, -0.0071251231124787751253, 0.0013895106482468286627, 0.029326299799962853021, 0.026562401024472964695, -0.069651664195190490791, -0.18872236823862245658, -0.11518826187847247544, 0.62879333171744733466, 0.85811308145342546361, -2.2437560521822965676, 1.0821549803290526626, -0.0038215862510918787061, 0.0028142783004495813652, 0.010199975538047530155, 0.010851758362461933274, -0.0097780968122128351927, -0.052491302467985022362, -0.053482986163637294408, 0.090743545610594972262, 0.28165542204911364621, -0.069313735938467443964, -0.74197803696348396674, 0.27329512577135256501, 0.72620661630735194869, -0.46788134130193792526, 0.0061413166365634382082, 0.0071802846750717080546, 0.003057331948353595244, -0.0057409295937124972184, -0.013119362511659748288, -0.010239737862167037907, 0.00021690354389166502966, -0.0040179086751429300001, -0.049349762222134690992, -0.10301149160485191525, -0.08002525524731582407, 0.24894953682196474531, 0.84000014053069793629, -0.8705844624384116992, 1.6520513936042559906e-07, -0.0047803783675692225127, -0.00099389326221380126965, 0.013611970281965760782, 0.025361986102777126645, -0.0001076790949195417465, -0.077885654184987510029, -0.13463130124142433375, -0.013933815319996373586, 0.3787211640858992312, 0.56563473607297587442, -0.49364030175682432899, -1.3042134499368713829, 1.0630247804404555634, -0.0001001017772776828193, 0.0030779781481069597333, 0.0024170890950567860966, -0.0059578316689640935591, -0.02110071381713030797, -0.020707299069387598289, 0.040959844130970493592, 0.16900639641237841237, 0.16647979865978557168, -0.3534546623806408272, -1.0465229938567253143, 0.22563641079625118557, 2.6305923594779341279, -1.8144508650574606712, 3.0228669218028136618e-05, -0.0012316199804219489077, -0.0033615188141743711225, -0.0049105014570409123362, 0.00017144628066613303279, 0.020166785423721085568, 0.047168382022491288363, 0.040471605577382475938, -0.053207340940281282193, -0.21229396191707844421, -0.27670706515986231988, 0.20771299770907442062, 1.0769099460792381784, -0.85563921755819882797, 0.0019284774795890560241, 0.0041988495475820042033, 0.0027363866295449278468, -0.0038923999895313526366, -0.014732635200555137603, -0.025780203421690565946, -0.030632183171049681691, -0.008755257234786331666, 0.083950984925291516348, 0.25782736593267324698, 0.1932450379423287512, -0.52195596693226375962, -0.65568196060174088746, 0.73459587562715511488, -8.4093242586331255872e-07, -0.0019028896369805733783, -0.004096754963746262676, -0.0041370762721030136799, 0.0052948011031809611923, 0.032984927020795161423, 0.062061292597234259116, 0.0060377073775660069013, -0.21303201304934810123, -0.27433788449963353306, 0.28167584921956539956, 0.48633999230122404445, -0.19831583812341227469, -0.18956176829629181602, -4.3335848564486260199e-05, -0.0016621072186640917321, 0.00023151365195768708674, 0.0076961931688106846594, 0.017358479306138763221, 0.0084742970826452990174, -0.06295048285913601438, -0.20503368361197185643, -0.23841297299221020367, 0.28699718992460071521, 1.6634191402196261134, 2.0149360886617109223, -8.3336245869457474811, 4.9010375126996192918, -0.0010071341501669860748, 0.0086757701029503368068, 0.0078361170667204417539, -0.0086573911512128291085, -0.036911894217781383853, -0.052105177596599258627, -0.014081854255392134076, 0.092520195080380757124, 0.20466734028751257246, 0.19947025971210699535, -0.12297271669305975816, -0.65808401136472960413, -0.23357985197498948304, 0.63982607479457531685, -0.00061573396769304487158, -0.0092633393080887166393, -0.010133423603972570651, 0.0027346118475783331124, 0.031312648689340964736, 0.061253962112640406024, 0.054933113923132700918, -0.022409950713363411406, -0.17060555397368321051, -0.32271336484679552514, -0.26338169107431380622, 0.50642393914165351188, 1.4412914327841592765, -1.3265216323225448303, -0.0036904495818771286715, -0.0031367955822365007137, -0.0015723916134579892911, 0.0019137383087806671654, 0.0073172247403985833658, 0.011482217187007089793, 0.0099400468310711989628, 0.010957513589573570842, 0.024571731311526488067, 0.011849850309639501852, -0.14726019838636533743, -0.21143084090219718685, 0.40874153127740681102, -0.11253507759194927851, -0.003036513373042424007, -0.0027337986703786032808, -0.0016190940194891394275, 0.0013912051092473786806, 0.0067682392252593160872, 0.011747010806525180412, 0.010854951174338023021, 0.0081120529569606806414, 0.011099489458119820434, -0.00077041583685605037439, -0.12144959934285377845, -0.16603432377378468665, 0.39073266023777775136, -0.14024832525108854586, 0.00094876287435246610142, 0.0016019245199082559998, 0.00039241438922933279294, -0.0033318661643490627387, -0.007592438332947220106, -0.0059109774724239979096, 0.0073995619700649682693, 0.030597747024806224575, 0.040652880314537204187, -0.020807057310128333066, -0.20849048891259217009, -0.10884068426338398827, 0.65941889023399802294, -0.39060786075037579046, 0.0007444551389909453672, 0.0014055789126338543316, 0.00067107547046018657666, -0.0024087819850073391784, -0.0068516122241608209617, -0.006906071622441374222, 0.0045111782851832609234, 0.028048170595235840524, 0.040664122861980442347, -0.013114264655063661777, -0.18757801438373453151, -0.11422698966619894423, 0.59792125256089057039, -0.34680519214742394629, 0.0011045068304433218551, 0.0012276064874419600689, -0.0028525717051014133803, -0.0077606307887740149862, -0.0035946005458227523258, 0.015276114803162633982, 0.032309904738219978604, 0.020416524077141728899, -0.029878372855277404718, -0.098972042375386842794, -0.15419113494292502331, 0.048054956268708454159, 0.64605931968890950579, -0.47441998186228750978, -0.0019738790020762471585, -0.0021070954980345582737, -0.0022091066108698275555, -0.0024631260971765354256, -0.0029270318975077045179, -0.0021846162537949567403, 0.0076872197415739076595, 0.044651978961970074933, 0.088321118671008252976, 0.049879394289245218563, -0.15581763968681180965, -0.33988082745198766599, 0.13998114325780694478, 0.19544430773630355747, 6.3797469295214016848e-07, 0.0021786625694403859502, 0.0023876097367585510413, -0.0028316542950446430861, -0.013449662781898720498, -0.012820338881673234988, 0.025158845647671831958, 0.064937312650839823513, -0.030788403668570860777, -0.15873658832354378689, 0.10269956182171820214, 0.18495082357757947067, -0.26231751023444893622, 0.099279866273519035147, 0.0020031980239833442614, -0.0012948049331026169687, -0.0030946888626272798899, -0.0017491841091715723717, 0.002834674087672402025, 0.007589100105025505319, 0.0066347170293519601436, -0.00055535048585026813937, -0.0041494857855761150753, -0.013540743401873660268, -0.12817232649614598849, -0.072901038495271464068, 0.59963863820154217343, -0.40001551097238224886, 0.0015425412121336525787, -0.0013722587227587477142, -0.0026759677463048881299, -0.0010267721664498680092, 0.0032862875486940733988, 0.0074823995382824852463, 0.0066729684246009113302, -0.00084938158111578251994, -0.0088583579607499282615, -0.018257907159215880366, -0.11224133303992842037, -0.06034864881180993651, 0.55942155085717726326, -0.37929277899536217866, 0.00035725038890568682577, 0.0027052256646975096112, 0.0063310205306354380705, 0.0080688087781441639179, -0.0011771801542584416338, -0.031039565554622498833, -0.066545648273826862873, -0.038516484110175081002, 0.1036612132546432391, 0.22837286394573586001, 0.029681630068063642042, -0.36784904439936083609, -0.028566669937791232953, 0.15489351657714456323, 0.00037452239787306328871, 0.0017707115609133072375, 0.0046763423041708187394, 0.006967527816647505097, 0.001015881940885921637, -0.023091131268238144353, -0.055587589061033837456, -0.037558975583462286818, 0.078780298649644184272, 0.19173868341602634535, 0.038074350538811668754, -0.31117516196348365964, -0.016161695843473176948, 0.12018584033981476178, -0.0029647736398375353593, -0.00091748239683335129028, 0.0036857353274456130382, 0.0059248019174140877791, -0.00049300927132806058716, -0.0069830994640445957083, 0.010682933672615526136, 0.047318021987152382446, 0.022969776444425119011, -0.137956094968441495, -0.33117366527141212096, 0.075882332978154859449, 1.0821549803172825222, -0.78027447732799481983, -0.0019693874057785413709, -0.003821586192524211497, -0.0012630807478666275425, 0.0081475394061878911312, 0.021208228086725799261, 0.019351978994434606424, -0.026868130397325301617, -0.10237990114990376533, -0.075231207339054723127, 0.18992805668701817634, 0.32696469796607435798, -0.24648849906141057087, -0.46788134139555920266, 0.36453948794332785566, -4.4867845714718681549, 2.1076300296989454175, 2.5759202929155988748, 1.1023690993639105873, -0.15442901322863841762, -0.62135409399767893568, -0.54747608306350625806, -0.31375859939162792367, -0.11069917320065685018, 0.011313549801533281589, 0.063857669617462461953, 0.069948138098291104403, 0.049269629071787002239, 0.015635130372620559963, -1.3921413099557680138, 1.5532132643338525391, 0.46972246563489888915, -0.35977412241065731546, -0.31740804526963506405, -0.084489850781320027795, 0.041415048527638859133, 0.046864104210124254979, 0.013825249702483327977, -0.0048561475657645904527, -0.0037552264463160335735, 0.0019773830674339014368, 0.0033783897401922489447, 0.00014459204203508946697, 0.21667253554880683275, -0.38726066935731223717, -0.12514140150112920691, 0.30518686968729669129, 0.14171532233612066376, -0.05775717856729453703, -0.082009724756492194397, -0.038922303048456050045, -0.0014463639791052361125, 0.012778064180464183963, 0.011135019144278619774, 0.0045333371235810205474, -0.0003014230712120001543, -0.0019586049063314101698, -1.1328517711166996573, 1.0724631408243188169, 0.45525705892116635765, -0.025798857295143574203, -0.20712293900450315376, -0.21485267651299216896, -0.09458138871001187109, 0.01702147522672721805, 0.057731651427815770805, 0.044552015868720734604, 0.016980931423149030896, -0.0025469214951648480941, -0.0093903218424076732168, -0.006897328909031560408, 0.43847184523622023677, -0.093255878781769704911, -0.54720362849266257488, -0.14755073812201510708, 0.16485916087515264938, 0.19663612004028063751, 0.10319098696804293591, 0.0043413542151492087381, -0.047244143214656142415, -0.050831048448029929332, -0.027378039533758002921, -0.0021851678298164078744, 0.01119498608037771889, 0.012066888120346740826, -1.6623713524514769357, 1.4771416244375596971, 0.81939499061225617993, -0.083271076907523680632, -0.36127698068676344922, -0.27885992804790593613, -0.10346333769207857745, 0.025620297201056595193, 0.0702875862218596964, 0.052302630255567215956, 0.01733625599160263514, -0.0057770130431398234241, -0.010827265749670028291, -0.0029122169981116528131, 0.6398260744483447171, -0.23357985170398518071, -0.65808401036008679075, -0.12297271845185209216, 0.19947026067984735542, 0.20466734011201254129, 0.092520195607349489664, -0.014081854863887556528, -0.052105178006310440275, -0.036911893666996667407, -0.0086573907319628020524, 0.0078361164861576182439, 0.0086757701617978365966, -0.0010071340737473116669, 11.397207747406545408, -12.828629401912643004, -2.050751879158218216, 1.3305305816184955781, 1.6024975471696230045, 0.93944287057822695175, 0.31044067344443893042, -0.036518576125356715101, -0.14730071186303633768, -0.1341693244328853174, -0.080633262263664115532, -0.030912856705953169895, 0.0012636412359323372784, 0.017410836952313681752, 0.69126548954627897725, 0.47083468069342776507, -0.51914610187261722718, -0.75386833659585428258, -0.49120643166847016703, 0.013490706512946652046, 0.34750211498731076176, 0.34805132899921381595, 0.16905589540295931639, -0.005413862237105151673, -0.090654903351270132994, -0.090455979602904273973, -0.040632373069174235192, 0.028438284556954570442, -0.44378342223963568314, 0.93449391401540093938, 0.081379427825660280438, -0.33854343856100765908, -0.28804525713455908509, -0.11843426957092716145, 0.0053687361813506443917, 0.059810584252141342498, 0.075289776800520610833, 0.064131851134179601592, 0.037059307383221995469, 0.0067343796667549103707, -0.017612445400900837711, -0.033568871906465985733, -0.4837367761361712204, 0.93538594162813082367, 0.099541227530431708459, -0.32355286516599646163, -0.28187363429759693112, -0.11643728876552346396, 0.0084679326444604980512, 0.061997429867362466305, 0.073215189895879190241, 0.058463821991276115897, 0.030957834944724114956, 0.0030604440835129598404, -0.017305424123270098796, -0.028820102842578972091, 0.18842937968502049273, 0.20646098590800368577, -0.56448374117974908781, -0.19595221625427691747, 0.2231560130072247905, 0.21491509228265431863, 0.055648286589731275897, -0.044605656384962645566, -0.056388354088921573137, -0.030844969260909994857, -0.0067762763420234923412, 0.0048813512675615830516, 0.0075709862073975887214, 0.0059160768039554856648, 0.037050940319115786836, 0.28733360451462397478, -0.44450710811238813847, -0.16154694129202845643, 0.18228094449456497261, 0.16458347607579229543, 0.034789410726868384705, -0.040141063967589367401, -0.045228554621539261704, -0.023649504941789446788, -0.0045501193497166402338, 0.0045583828898786218908, 0.006572604333498755641, 0.0050400517178276768929, 0.34340333007539436405, -0.082522493494577839335, -0.35932857347137214798, 0.0058250860543134271724, 0.053802354208100824917, -0.010708897935761919548, 0.014098182193458812239, 0.044977330889250009827, 0.036370070719591196529, 0.0072897313297527336143, -0.013506100659372735281, -0.017107643300991653978, -0.0085614388365636874545, 0.0050401110327120733981, -1.3189472242938571345, 0.74787509197846557818, 0.90010463704506971094, 0.42600666298977252211, -0.22280643206410469248, -0.43618952184234194291, -0.29031166583102341328, -0.085144581648646952554, 0.05035448326212320358, 0.10053678602077328885, 0.090765112571008610409, 0.050035927064170787615, -0.00065606848330967840535, -0.050131187725596541849, 0.85067142349617919095, -0.3425506802328050493, -0.78880786204418751328, -0.18260162467282639587, 0.22915954832894519511, 0.24813272398124136542, 0.11022910755781351044, -0.0026623688298266527316, -0.039617613893146615134, -0.033392055053396776931, -0.017177890950303861123, -0.004709957475001636841, 0.0023053505329146384804, 0.0053318467687351791401, -1.3670486405096871074, 1.5657927847839376856, 0.42873713953855013825, -0.4024304152419233005, -0.30418107663441534827, -0.053526664899878340087, 0.048702709158662586819, 0.046584469314685794661, 0.020151256489194371846, -0.0015387278777475943761, -0.010825258077000396439, -0.0092583899652118658435, -0.0014371878766815152865, 0.0085247054060803677256, -1.3599544510240542117, 1.5474697281132148152, 0.42485982675329886371, -0.38860276806480514855, -0.29156026021032721562, -0.052759546148185447467, 0.041928166168033742522, 0.039247928886369978785, 0.016319231641093345092, -0.0010652416147069751831, -0.0078481979742100957398, -0.0063673802081834632652, -0.00043111023007807005095, 0.0069153438868634672237, 0.34786009034152531871, -0.3426231352694926735, -0.46572681359554590763, 0.10537285105808501251, 0.29445230409097611357, 0.19788884130173922671, 0.043528347002208082894, -0.063224567838376169449, -0.084981485720225638514, -0.057065145823025478777, -0.021824853180016035026, 0.0022517541362051286813, 0.014767180265522383137, 0.019753603148506524573, 0.3389391791274104504, -0.2973062531687481469, -0.44221558563609902892, 0.083244979105027164246, 0.25436585679999662579, 0.17247339508271658515, 0.04399874562493726049, -0.048272176369863251111, -0.0707346457165406034, -0.049882146426429442576, -0.020614630271722250709, 0.00055719021618262816464, 0.012314561459487553818, 0.01756435123658581926, -2.0612194560993084558, 1.9766154132203930249, 0.91280978778457011469, -0.22371153741800228754, -0.42460837645985555211, -0.28783083183208219014, -0.10064907839751638385, 0.02502751516005350188, 0.066123663604964785789, 0.052617894150534381958, 0.025299762820486303844, 0.0044154463760870564168, -0.006740002544089079492, -0.010871961524747239883, -0.1565694570233314431, 0.30802196183906244142, -0.13778326966187667835, -0.038567133473464500215, 0.065782836150788615015, 0.0036130836268969445182, -0.049925357796288830081, -0.034310540222164391533, 0.0035362833895466062178, 0.022375177465117454534, 0.020353058735792198264, 0.0086510658160931525745, -0.0032279172819054007849, -0.011469107313514130542, 2.1076300303730350905, -3.937833854787315957, -0.36546544935644315055, 1.2815874976879522151, 1.179863958536146118, 0.43327693862973470917, -0.14302043153101429152, -0.30936478922856958107, -0.2529578234179594709, -0.13699152321783705788, -0.042801178794615035716, 0.014607634852886734744, 0.041923411311151403957, 0.049269629499150341401, 1.5532132636118276547, -3.0015644835375767663, 0.72070711142781507252, 1.0010555940568235389, 0.095499748628406253004, -0.21276587489372092432, -0.16499680857532331157, -0.039766270616095771018, 0.023755662909767050195, 0.024974750377222472431, 0.0054104834619220221192, -0.0040382575418926574401, -0.0022982977957416821091, 0.003378390872054401993, -0.38726067083226423104, 0.6245352869077644975, 0.12915438379743954256, -0.59283635450529048594, 0.031776421950767547164, 0.21138595366110191076, 0.052536662469444661561, -0.03145114430274532219, -0.037434250266206885172, -0.017326610443762824332, -0.00032063136814877333473, 0.0068655025490510796643, 0.0055035835038010402082, -0.00030142132388149547526, 1.0724631411057141683, -2.051756377277231369, 0.55711711336641023262, 0.44534728486754732169, 0.13384390004095131355, 0.032884285553990526274, -0.088637004229715626868, -0.10379952014019211415, -0.047262593835555033461, 0.0077533350264361693138, 0.026661775386072100041, 0.020350691654694552657, 0.0054317021191434382288, -0.0093903201403831414867, -0.093255878691004268077, -0.17171077928408659541, 0.68692858146136082098, -0.22558876299341495963, -0.29698791311365130285, -0.055196256844748839421, 0.067264957578496639701, 0.081409545914114353482, 0.04439593409078144981, 0.0051512260006347865629, -0.020292432220201928744, -0.023631362767871016278, -0.010051920387912680055, 0.011194985125464778503, 1.4771416244158188658, -2.6743467305017993851, 0.3487880823846578604, 0.8643856502877774739, 0.3098408515703685584, -0.051699774160345948748, -0.15641328499387341378, -0.12592231216504973235, -0.05510238143632898572, 0.008304853479849858694, 0.032444813309006177715, 0.025638497902628364089, 0.0073108520772327198478, -0.01082726428735301602, -0.23357985172585418732, 0.08242517052143361056, 0.73502820117095168939, -0.38626442622083956291, -0.33520400198254296509, -0.042475450862711414124, 0.076988084251121982615, 0.10010349525868904308, 0.049699470726300584711, -0.0036685630785028722906, -0.024634986884085330328, -0.018708840400434842421, -0.0043715226298641366234, 0.0086757700391839817633, -12.828629402500387968, 27.004855313814161377, -11.559888162826233327, -4.0202401395388482896, -0.46261637052449761098, 0.72667547138230537129, 0.74605166352194773882, 0.42863855624361169028, 0.14224516680908666499, -0.013840488644916257299, -0.068122800096896141131, -0.063720934841110979385, -0.034477139497084002828, 0.0012636412341625452831, 0.47083468041123616166, -0.74392062829920924116, 0.50642437733286638224, 0.11312546795763959995, -0.050096067090819908219, -0.27029053070138597814, -0.2533924695477814959, -0.015189396212031587574, 0.14307421215736360054, 0.14867158139927549865, 0.072239659727733801509, -0.0028313833367283819734, -0.040124801549394781275, -0.040632373693143639115, 0.93449391418187865987, -1.6949417039517080674, 0.61922013539025899664, 0.4223645515165432851, -0.031712143862707820041, -0.16701094741855035397, -0.11710911518379177321, -0.026528648822514315497, 0.014656294967343108659, 0.030078722113752721229, 0.032394662654888203845, 0.023673501621265004291, 0.0053964544170107126009, -0.017612446380494608394, 0.93538594140326503101, -1.7103974735544673624, 0.60869358929405992242, 0.4274645787449388723, -0.018474960171666502534, -0.15625079217338924686, -0.11655098311396681898, -0.029429820920961621833, 0.013843985010745456232, 0.03001253333131751258, 0.031267160363522103483, 0.021852641927760808188, 0.0041576717500588465659, -0.01730542379524166105, 0.20646098623336595068, -0.87331039936095733456, 1.0615781118191056631, -0.057195080066910070071, -0.45047572714965239404, -0.084161771368009535843, 0.12227767971694053495, 0.10609806506408353544, 0.022197567532305402527, -0.023200771807663153612, -0.026334398173910044533, -0.01249235977717984454, 0.00023202969333232218883, 0.0075709856922121501407, 0.28733360522786094426, -0.97671207197149345625, 0.99436998970357104444, -0.013830049864136767859, -0.38159250072494471651, -0.061109641847145200477, 0.10136126552046825777, 0.079903584808015404306, 0.012305311876073411065, -0.020389090076986842587, -0.02080571062409748706, -0.0093433503370008042649, 0.00069312541541582375817, 0.0065726033648390571815, -0.082522493659982168457, -0.021250025293980435992, 0.55081687775879073499, -0.43133757181065662811, -0.1433897333963981835, 0.10478506626104679145, 0.023397888899274042751, -0.019625729463603344693, -0.00064164380329897292236, 0.019895821283347894792, 0.017203291461433317666, 0.0038159943274522776027, -0.0062715339017448057043, -0.0085614396434400864833, 0.74787509161488519016, -1.1455023148179879211, -0.082867218178919302418, 0.12010580397968614874, 0.48867849616039849003, 0.22479212457366015698, -0.12168236744809028671, -0.18393021361251263301, -0.11043980027183916692, -0.024727944200591007312, 0.026018002127474730567, 0.039714178250006039461, 0.026816159278195118854, -0.00065606795861572475559, -0.34255067988009013646, 0.26850394701607949166, 0.7483951017795651417, -0.41439325018403183609, -0.42675834678532381661, -0.072061996796582303282, 0.11625972406586293495, 0.12230166898360914085, 0.047781614010744370702, -0.0039370248518329816617, -0.019006348141104922184, -0.014991047555188581272, -0.0062701535690416372593, 0.0023053495012423305677, 1.5657927854814137625, -3.0522948954679054445, 0.75636298244651689604, 1.0480065443726971353, 0.051416498429929846647, -0.25007282842400080369, -0.13822780246764596046, -0.018328498581844793192, 0.01642985168828474743, 0.01711092609262716488, 0.008535165647420992624, 0.00085572927756401555993, -0.0026073560500742363338, -0.0014371889868521996181, 1.5474697281473488442, -3.0230652609664439012, 0.7586090463100775283, 1.0319487000348535055, 0.045164741298978280937, -0.24340531338857063215, -0.13258658020793381849, -0.017462175706065691944, 0.014642217817742057134, 0.014510808767100974795, 0.0068286620792750891973, 0.0008227676057298450446, -0.001566372895896243813, -0.00043111029684977280121, -0.34262313569585428619, 0.22378721416977898073, 0.55891574497872043015, -0.43967767846521810249, -0.20896633435382458432, 0.054989945623785410622, 0.11270258377992638243, 0.08996681013948470973, 0.016729445527939507121, -0.033103822642462740755, -0.039716740397735393608, -0.022329927538177717083, -0.0017055273969731098507, 0.014767182054680815656, -0.29730625341964583841, 0.17939399060535596298, 0.54265255712484439243, -0.41878582775556172368, -0.19207933469298540952, 0.051653723135215226514, 0.095011361663951049428, 0.077344859177580049803, 0.017362712398813408277, -0.025527345287442861077, -0.033067848296414258102, -0.019604691132477404525, -0.0023129537367706943379, 0.012314559876253955775, 1.9766154139517562172, -3.6693244400899520663, 0.65539497762037335349, 1.2079012762856684571, 0.25034978106528893438, -0.12359304076419322382, -0.18429450504141448031, -0.13080747029639461232, -0.048081236632740019521, 0.013045316730257373178, 0.031172893935434379392, 0.023263174011855361489, 0.0078220187108323895042, -0.0067400033030596108799, 0.30802196234260031948, -0.78122836637353176847, 0.61688972849912804275, -0.0038466871107449072779, -0.18571190914736110389, 0.0048546401243074171344, 0.060547828908236109302, 0.0089946191566039907644, -0.024579867254863236453, -0.014944950199992354523, 0.0015955561315245776147, 0.0093301670871855230116, 0.0067045442623058296161, -0.0032279171299595295352, 2.5759202924052937433, -0.36546544837559380436, -4.3663192316404257909, -0.41342802296982905652, 1.3725925253977848683, 1.3189075792883977467, 0.58764006452828554394, -0.036998866320680973141, -0.26531372452592349998, -0.25612753699429818832, -0.16015572580429143779, -0.062352128770464067264, 0.014607635132966381719, 0.069948138139275889413, 0.46972246599507483378, 0.72070711085143424057, -2.6046557315583465098, 0.69250445641340263592, 0.98588822500415784766, 0.097180254511846211884, -0.20185747021336208906, -0.15532796062908060808, -0.036248755541671645597, 0.020437808361432746596, 0.020925519591418886151, 0.0032157443316471120975, -0.0040382557171061983503, 0.0019773810445212420053, -0.12514139876937452334, 0.12915438161279896301, 0.089180658176762073186, 0.17806403549956345556, -0.49139549059937648323, 0.030300290450327151126, 0.21015038234765928693, 0.055723734502877535213, -0.031012966902770954664, -0.039409502563115574036, -0.018668844748118057497, -0.00064140263745505977575, 0.0068655051111473231612, 0.004533333972314501116, 0.45525705870596405722, 0.55711711336354485802, -2.1331707831875745818, 0.54769246995309373016, 0.57988626325657799399, 0.19201471477839532342, 0.036834337488088765566, -0.10225283693018730191, -0.11660451196095152071, -0.055994134264040623661, 0.0049878708231859544484, 0.027877241517471305909, 0.020350694084230638858, -0.0025469236490181196113, -0.54720362856618942615, 0.68692858195970252222, -0.23868418519156631152, 0.58301313824022482635, -0.26813258134986178716, -0.34189926702354911381, -0.076266328033989125301, 0.072668612279415928379, 0.10071872272805418302, 0.058133490484560873535, 0.0090441507182076937704, -0.021749157802588944821, -0.023631364307847950446, -0.0021851657080292738629, 0.81939499034390217425, 0.34878808247122133901, -2.5326198199562592883, 0.39433682188665358792, 0.95077288988811425874, 0.38413969423687305271, -0.030307411384223010425, -0.17331050002624301398, -0.14440997556024320891, -0.064480012292200669655, 0.0079040442138126852656, 0.03608129372414318331, 0.025638499570272924727, -0.0057770148449533785528, -0.658084011233722177, 0.73502820144478353548, -0.080428106050600689514, 0.63655052544471601994, -0.42532047130550681979, -0.38371589264867134172, -0.043705612578207049534, 0.092032875593587248897, 0.11168078060586122957, 0.053345528023248579841, -0.0057510054640918161981, -0.028799631578458516346, -0.018708840792840991663, 0.0078361170528371566074, -2.0507518777735165294, -11.559888164160097901, 28.516300289065824813, -11.542093594332017759, -4.4341302439508920585, -0.81020590603299513788, 0.56807841159685379395, 0.72377703118791036907, 0.4675623995630847074, 0.18542363298234240676, 0.012231609274372950633, -0.059603940861710280608, -0.063720934306457838181, -0.030912856928444427951, -0.51914610132685690136, 0.50642437706401466357, -0.45647528116399926912, 0.92728010370489177827, 0.20197015614079422141, -0.20226468065739888758, -0.42079619296953185703, -0.34460017363442313476, -0.043134253111204223785, 0.1599482321021922504, 0.18503346540114878893, 0.10205049335944917088, -0.0028313848085037784073, -0.090455978084905019898, 0.081379427644048957236, 0.61922013568766176572, -1.5606332434680012522, 0.66318487951310622908, 0.47542375909421319413, 0.0095011600503514814786, -0.14617008192404698508, -0.13357917153679321753, -0.056122254879737973332, -0.0097707724632026676792, 0.016778065005767076046, 0.02780580888194632283, 0.02367350087339320508, 0.0067343798576339035378, 0.099541227439113255659, 0.60869358927896033418, -1.5797983270401774192, 0.65395641197709308834, 0.48196363523523477745, 0.017870835708135947295, -0.14030887642190445486, -0.1338224408688797884, -0.05718225557620171956, -0.0078181929116517279166, 0.019429154918934092577, 0.028469684010143953695, 0.021852642325187479622, 0.0030604436538540997007, -0.56448374200921569432, 1.0615781123240377593, -0.89031546703818842392, 0.77870256369616053238, -0.10121231972032318502, -0.43554393918174938438, -0.070173738223945097969, 0.1278993166812694704, 0.1122757193383469021, 0.027824652769027190952, -0.020230448186860938015, -0.026328697493225656362, -0.01249236095846072668, 0.0048813531695785254347, -0.44450710957151362557, 0.9943699908741160387, -1.0137732968896147057, 0.73995945480066072086, -0.034447203941444584641, -0.36182677212438629777, -0.049333097839855073741, 0.10557737828235630451, 0.085354348400676210984, 0.017011662865542939715, -0.018217019081458271645, -0.021057346526246439877, -0.0093433518606707288701, 0.0045583849058619075809, -0.3593285724977101081, 0.55081687749931096398, -0.38108109988420513226, 0.5714865801929316369, -0.30189539729694581105, -0.15877775672832236675, 0.060867684795584020563, -0.00068867101696386365684, -0.023598432432947312465, 0.0041714080406006527613, 0.027002418043271542852, 0.022609152430362874225, 0.0038159931027388796325, -0.01710764224557794963, 0.90010463762747294769, -0.082867218594680436272, -1.6414802202759142258, 0.069372442988444776857, 0.30749971068268738827, 0.57894754054457520098, 0.29629949119071247177, -0.092778684151553492132, -0.19793307628639170148, -0.14671495581712076794, -0.060501517024582743276, 0.0047996338150409906254, 0.039714178303379941981, 0.050035927395251519889, -0.78880786270108671498, 0.74839510215711868568, 0.16608249876959077507, 0.60962069383164996239, -0.46777417105326762892, -0.45491774564979869933, -0.080678497784562167028, 0.11133920091979672418, 0.12596101889338612301, 0.05545054336809825124, 0.0018892383812637149419, -0.017108529572135988683, -0.014991048913016382857, -0.0047099560536218157067, 0.42873713789343648184, 0.75636298355289910411, -2.592165547974721207, 0.69591187310141477962, 1.0048029271505394266, 0.07441990445976245927, -0.23361692189538094122, -0.1450418930682240426, -0.024051592849895893356, 0.017564536540234175488, 0.021834304105955500142, 0.012395575212200665716, 0.00085572699107399480267, -0.0092583871349905047521, 0.42485982664776922313, 0.75860904634690273785, -2.5764393039180126799, 0.68861306621461049193, 0.98919549542101836703, 0.070371920394059650494, -0.22396539400104539808, -0.13656537856292982891, -0.022393603593212671465, 0.014787499792373301477, 0.017742353643035474464, 0.0096214878771398761048, 0.00082276699747363749027, -0.0063673797832676966302, -0.46572681253547115743, 0.55891574416808542658, -0.062206791202730080115, 0.45124251463279846375, -0.49656594567820505626, -0.25675048613911494177, 0.064458427943830054474, 0.13472416370235462679, 0.1091150490309053811, 0.030702199274926171568, -0.025171128871417786099, -0.037661214720699602576, -0.02232992446528691069, 0.0022517501067488105912, -0.44221558504427760372, 0.54265255641722442981, -0.098509716846321440653, 0.45856945617844757646, -0.45865627646906814707, -0.23874898329350224824, 0.055587284333112108037, 0.11265935694794239341, 0.093933896025020613418, 0.029883861805027549241, -0.018157995378580243234, -0.030957162227697128387, -0.019604692154451944242, 0.00055719113189087697557, 0.91280978689078162702, 0.65539497879231367605, -3.3631917171750109752, 0.61146738479358664353, 1.2956481810470572569, 0.34005065653399690895, -0.094836762576302538474, -0.19199072735388911237, -0.14529249969560220634, -0.061369832681151641196, 0.0051176253946721289578, 0.028903731913118985153, 0.02326317342154850254, 0.004415446393770422473, -0.13778327017568833823, 0.61688972857666646288, -0.89922268685083550643, 0.46607202363822269886, 0.038649050136745577433, -0.12551659808433576759, 0.022071713162666272562, 0.054564566892974362966, 0.0017726954635251461305, -0.031980601817199569292, -0.020359150884758508981, -0.00072300708854353850676, 0.0093301681458580372486, 0.0086510645299164370958, 1.1023690995875712328, 1.2815874974982273127, -0.41342802282660157109, -4.1841988955766264979, -0.36789718117118491714, 1.3739630509848810291, 1.3279126320665450844, 0.6100513564085180418, -0.021608857747882619094, -0.26369714071051408855, -0.26192834071086390502, -0.16015572522868803951, -0.042801177759030098469, 0.063857668242040727113, -0.35977412100602684797, 1.0010555932366749232, 0.69250445558470408169, -2.6899447735440857876, 0.59386008583687777129, 0.97081492438952710078, 0.12627886025470921871, -0.17663822646684632089, -0.15049487261010488504, -0.041103744801572465739, 0.016425248938123197112, 0.02092551825511406588, 0.0054104822826377824141, -0.0037552243507834141834, 0.30518686826967023062, -0.59283635411718937913, 0.17806403681418031537, 0.19325555890286699734, 0.21862544888949339561, -0.49143824117832424214, 0.01102105426331893151, 0.19918524726794867297, 0.054551315196995284196, -0.028223855499356086812, -0.037631913344942792554, -0.018668846284790041601, -0.00032063174848951444141, 0.011135020677431214237, -0.025798857407802359282, 0.44534728634807529968, 0.54769246832564888461, -2.0420836846454504254, 0.54320764175620761538, 0.54828758440902336524, 0.1686501011871955058, 0.034727415565002542341, -0.095334830178547519397, -0.11037011803861566939, -0.053559332694543403119, 0.0049878725102478621567, 0.026661774925015403748, 0.016980930295021293835, -0.14755073813433189356, -0.22558876306583974802, 0.58301313851557112766, -0.33253973556242444332, 0.55764580744563241321, -0.24757760687779395425, -0.31855726567363951141, -0.061798349032137656067, 0.073580625252151476978, 0.095431969751115341971, 0.054097628227446500904, 0.009044151812120732048, -0.020292432231292401035, -0.0273780405771717264, -0.083271076340850930242, 0.86438565012217150052, 0.39433682138421849439, -2.4801307397470453431, 0.37301953525845316095, 0.91314440668006402468, 0.35981557506764827448, -0.029728670880466309678, -0.16209218902881708524, -0.13417536632041943134, -0.05979136790867093515, 0.0079040437129460662524, 0.032444813607343581108, 0.017336255987720275429, -0.12297271664118374079, -0.38626442709028085964, 0.63655052414859969456, -0.1735845714342288959, 0.6199497310357577895, -0.37967857987437264589, -0.35381999083668369543, -0.040559271668634828967, 0.081209497481246020745, 0.1013827425375889435, 0.04846066920635058034, -0.0057510051424870421363, -0.02463498671432979778, -0.0086573910920791589257, 1.3305305796973461163, -4.0202401384539863116, -11.542093593461158818, 29.0259770666500998, -11.299906243965716968, -4.4128513461781890825, -0.86955259096131465313, 0.51555171046022219006, 0.69758459210346945856, 0.46259957934380629974, 0.18851462446553959862, 0.012231610585835252236, -0.06812279994022225893, -0.080633263252405315558, -0.75386833857444035534, 0.1131254686914689167, 0.92728010521471782024, -0.47387939378103849375, 0.71718628941805584454, 0.064973688009259608633, -0.23394351327388870909, -0.39176669968897515073, -0.30785782355079965855, -0.017857592837778636652, 0.17361895353663606145, 0.18503346550985230734, 0.072239660153662732522, -0.090654903824412644275, -0.33854343768881223209, 0.42236455078075657577, 0.66318487935842596048, -1.5755953818330319383, 0.64090718129513202239, 0.48012809206972145715, 0.0057016450742232854809, -0.1573764003220163854, -0.1457354565177535477, -0.063717554815237650856, -0.013089065979068018053, 0.016778063899428652395, 0.032394661701224021499, 0.037059309178942484442, -0.32355286474784117345, 0.42746457817363547882, 0.65395641236987767098, -1.5917287539905176086, 0.62710502147204660872, 0.48139756796091781599, 0.012998692384133550548, -0.1495605236624150125, -0.14323659758276280352, -0.062522443147719722334, -0.0098459327526589646151, 0.019429154804367479437, 0.03126716122050470642, 0.030957834028984942965, -0.19595221634034520886, -0.057195080200846147578, 0.77870256369360379978, -1.0046531799274136532, 0.79985066685228112249, -0.040428613887463868415, -0.41351210119072295823, -0.075661448203549502844, 0.11840037867240069125, 0.10905901463405713592, 0.028197623709131935249, -0.020230447160423711134, -0.026334397256512064034, -0.0067762778052804064147, -0.16154693992980528883, -0.013830050664163250421, 0.73995945392118767092, -1.0999750940602330829, 0.75960271730943540458, 0.014698157566210486372, -0.3451465751695177997, -0.053660436894022958687, 0.098524107627225701589, 0.083095186553475211544, 0.017257527635363232449, -0.018217017851097885156, -0.020805710114400115945, -0.0045501207756022503392, 0.0058250845595126240703, -0.43133757167881031647, 0.57148658187658230023, -0.29504703157776718259, 0.57178834366305919534, -0.3302211528885140579, -0.18388675039410107281, 0.05822721182913105642, 0.0075965820330371426269, -0.016029448541361693281, 0.0075555520297441508201, 0.027002417383217198338, 0.017203291847234989526, -0.013506100683408050686, 0.42600666193247271885, 0.12010580433347287088, 0.069372443769374653177, -1.4371350930182285044, 0.12484835856547041311, 0.30382634010194575769, 0.55785175638172712631, 0.28053056111789687055, -0.10737639209601093637, -0.21299970092471115501, -0.15831179113640569955, -0.060501516997274483456, 0.026018002119876509903, 0.090765112718661097868, -0.18260162400238172387, -0.414393250444120953, 0.60962069328656542933, 0.056980685128555759456, 0.6176099601235657266, -0.42108153586556962278, -0.43982037957031921982, -0.085732907679339803719, 0.10358814070691280129, 0.12320237510912024348, 0.055821826153546187654, 0.0018892389165544307746, -0.01900634805558291951, -0.017177891409666606881, -0.40243041361844517079, 1.0480065439666375049, 0.69591187198499560029, -2.7080368637038865209, 0.61735531019991796153, 0.99846396845057616698, 0.090610855836196241286, -0.21772901149262793852, -0.13659157888357034705, -0.020126466669763594386, 0.01922996045824497402, 0.021834306073961924632, 0.0085351666234937256794, -0.010825260709505585971, -0.38860276818636640961, 1.0319487005036358518, 0.68861306592218674005, -2.6906941728993967189, 0.61420442164690614018, 0.98636832009562225299, 0.088202184836140715141, -0.20864000802672089119, -0.12964141412466140268, -0.01975928675574529364, 0.015787982436518618878, 0.017742352642428477899, 0.0068286619310336631183, -0.0078481968741409523227, 0.10537285038975316431, -0.43967767802538404442, 0.45124251494719663791, -0.14647326284028042154, 0.45059610883562029038, -0.44830494030493028834, -0.23309954200591859186, 0.066979532173292471997, 0.13171035143582832116, 0.10923267587267296297, 0.032483545369449209894, -0.025171131107590560827, -0.039716740976842447586, -0.02182485088788285385, 0.083244979345776307089, -0.41878582767249061813, 0.45856945587963338085, -0.16755593388253767206, 0.45371932180908419729, -0.41942889827645174483, -0.2189925879116602081, 0.058452209498753363026, 0.11070687075567309243, 0.094275451481443672819, 0.031445675675280830574, -0.018157994770203769036, -0.033067849770410155663, -0.020614629154753903034, -0.22371153698850929081, 1.2079012754960394282, 0.61146738524930388525, -3.3338959347224430552, 0.57281647503416355249, 1.2631945904826547, 0.33147507967687411812, -0.086833117673819834592, -0.18407869309858185258, -0.14266359257358202095, -0.062108129252463374625, 0.0051176243288681344187, 0.031172893817202795147, 0.02529976368281866908, -0.038567134133697611453, -0.0038466865448314587969, 0.46607202335232456036, -0.89547865657664293959, 0.51835394174705595738, 0.064366965337571760419, -0.13379913181427938174, 0.0084226518495168663236, 0.046250057407054748215, -0.00010916964117672054747, -0.031713314215944116148, -0.020359151610325490478, 0.0015955551937118781555, 0.020353060219419558119, -0.15442901289505578388, 1.1798639582803907011, 1.3725925249560848584, -0.36789718101430146646, -4.1835208875580232046, -0.40304308719893833235, 1.3422659946998378722, 1.3116594138684862791, 0.60567723909034898444, -0.022375740083529653546, -0.26369714018675843858, -0.25612753550262329361, -0.1369915233126800802, 0.011313548432902460067, -0.31740804629840996265, 0.095499748834779685236, 0.98588822628743577603, 0.59386008561706971065, -2.7383477945055516578, 0.58984488157999137581, 0.98494658205749152646, 0.1348523811667542982, -0.17576006684790010381, -0.15241507220183614368, -0.041103744681961901231, 0.020437809309369340399, 0.024974751769573772658, -0.0048561495678597008108, 0.14171532282247456536, 0.031776421741399531185, -0.49139549072780541739, 0.21862544853069329487, 0.20355001527469793188, 0.21258457161195343232, -0.50195427182359542151, 0.0044308540061939669097, 0.19909052457018347471, 0.05641006143793078631, -0.028223855576626818292, -0.039409502571208523958, -0.017326610453532308165, 0.012778064238700259647, -0.20712293975750187269, 0.13384389907283497645, 0.57988626593878356275, 0.54320764103463770045, -2.0938581760101691032, 0.49943132944002516238, 0.53424680134412683152, 0.17280704755443448972, 0.044127958345783577232, -0.089532663655858255036, -0.11037011801147095236, -0.055994137280773237375, 0.0077533349794717812062, 0.044552018111017416269, 0.16485916040358117307, -0.296987914133854225, -0.26813257987415828332, 0.55764580811084418777, -0.31957921256724503012, 0.5888994930793960414, -0.22185360387704228247, -0.31331665976146744201, -0.068885579779381825305, 0.066777086240738442457, 0.095431968920378418275, 0.058133489775483625228, 0.0051512273128427503791, -0.050831048743730455897, -0.36127698077913789065, 0.30984085120564131, 0.95077289042280888065, 0.37301953518088099004, -2.5419896762776068577, 0.31575158840982264774, 0.89479592089966741852, 0.36854803680258413356, -0.014256469436889384825, -0.15299392114388471464, -0.13417536636592755639, -0.064480012390025154279, 0.0083048537749795415391, 0.052302630086541568311, 0.19947025971932411115, -0.33520400134184696883, -0.42532047085406438081, 0.61994973111705886648, -0.13361296430469041252, 0.66123992503600037551, -0.36399865568934181015, -0.36043023022990045012, -0.053016796657939575799, 0.073440298768074022129, 0.10138274275957082082, 0.053345528407717349162, -0.0036685628397997334382, -0.036911894241168280439, 1.6024975484838126505, -0.46261637047456466432, -4.4341302450997153173, -11.299906244914796005, 29.336315028103015834, -11.145405860385052321, -4.3864761112189976089, -0.89908620735832822657, 0.48400126841810486855, 0.68396666606583378378, 0.46259957998101736143, 0.18542363179097676174, -0.013840490239319377247, -0.13416932228950037564, -0.49120642912110090705, -0.05009606851741353617, 0.20197015464484857628, 0.71718628932449812741, -0.63352124956958322244, 0.67494961887870774664, 0.10139102369105212897, -0.17943409971479781784, -0.3529249897569518013, -0.29155906492700023325, -0.017857592997239164767, 0.15994823227712612335, 0.14867158201498922865, -0.0054138629237994494517, -0.28804525810312892053, -0.031712143220684267664, 0.47542375947577064643, 0.64090718139384461605, -1.5958099670777172729, 0.61646553924862934615, 0.46365978587263734223, -0.0017272462655415765642, -0.15889588087042666453, -0.14595626541622019889, -0.063717554753998581485, -0.0097707729904881607252, 0.030078722061830234086, 0.064131851582005086621, -0.28187363453007024949, -0.018474959413321345031, 0.48196363396694041903, 0.6271050225129005673, -1.6167313495144328694, 0.60182073897165566656, 0.46673486386326185071, 0.008079471079743605344, -0.14896136099168069689, -0.14226987736492677228, -0.06252244430405515252, -0.0078181929124950255705, 0.030012531746033341284, 0.058463824067111112848, 0.22315601419813152484, -0.45047572709253730361, -0.10121232094234844379, 0.79985066629636714097, -0.96144572020019891667, 0.82491614114698474935, -0.033990279652995239457, -0.41871477500399850724, -0.081834725088555174133, 0.11568063169782355493, 0.10905901502419679261, 0.027824653888668626073, -0.023200772058222348532, -0.030844970031099209345, 0.18228094333921671111, -0.38159249976151982109, -0.034447203703088798143, 0.75960271742777463189, -1.0681000915742306923, 0.7771653467049559616, 0.019542448592386806733, -0.34903321916067231445, -0.058384024317000779436, 0.096381566401902540875, 0.083095186160693640387, 0.01701166214170092561, -0.020389090889652010385, -0.023649503448509377745, 0.053802355316272268415, -0.14338973412742464508, -0.3018953978049917497, 0.57178834378904286329, -0.32727657015295030796, 0.55473686271197131248, -0.32771119099765638749, -0.17507182442747723505, 0.065946316920605194123, 0.011279231872986532695, -0.016029449500313614624, 0.0041714090718994508572, 0.0198958215771817791, 0.0072897307718540983387, -0.22280643225053659018, 0.48867849636872640051, 0.30749971075539378473, 0.12484835847796921238, -1.4411179999167547106, 0.087423145596843596095, 0.26394031038935039657, 0.53023439878889588872, 0.26670474206968936004, -0.11326333163452624586, -0.2129997001936735912, -0.14671495468882156388, -0.024727943085578990567, 0.10053678379131106291, 0.22915954749251130074, -0.42675834600133133456, -0.46777417089528916527, 0.61760996006239587963, 0.097428997382416490658, 0.64610922044821872401, -0.40992642199286333238, -0.44154207627902980482, -0.090889216985248003078, 0.10058431412580148834, 0.1232023751087912844, 0.055450543065369485574, -0.0039370252186997772847, -0.033392054445922463024, -0.30418107809046185386, 0.051416498959078044129, 1.0048029274510017483, 0.61735531161509160647, -2.749428126509783521, 0.59972186251298476023, 1.0049388399377987557, 0.10411811426896121213, -0.20829690717133109157, -0.13329887822250949259, -0.020126467749319303102, 0.017564534717496103899, 0.017110925933803693144, -0.0015387258441081851611, -0.29156025972785648559, 0.045164740079405622797, 0.98919549599297207365, 0.61420442219968918174, -2.7281545477948765388, 0.59873407179438553616, 0.99233090175764770979, 0.099846062627214349217, -0.20078458853158806519, -0.12706861721759757367, -0.019759286014425232747, 0.014787501012780278342, 0.014510809885624397983, -0.0010652439908375977394, 0.2944523045156294816, -0.20896633467371814041, -0.49656594579768814635, 0.45059610882603368109, -0.095825903496405531734, 0.48886632222028214567, -0.43605967367461506701, -0.23613919053012555338, 0.06216914393003298428, 0.12976230546488057294, 0.10923267616338855346, 0.030702199858742732635, -0.033103823442442122105, -0.05706514566325242982, 0.25436585668312822195, -0.19207933437623403439, -0.4586562765043381007, 0.45371932131902570351, -0.12684293609041213946, 0.48671651826796458185, -0.40764253560686142253, -0.22071239094978034601, 0.054680268454546883394, 0.10905850484749823881, 0.094275452400144316156, 0.029883860600192907464, -0.025527344738113109252, -0.04988214657553598913, -0.42460837720214017477, 0.25034978190705836631, 1.2956481805777688709, 0.57281647557641468538, -3.403320151016202999, 0.52106686392015866272, 1.2501792060841248055, 0.33808153003060337216, -0.077774975438437712882, -0.1799371311800761486, -0.14266359332140102212, -0.061369832356877422608, 0.013045316234221273829, 0.052617894685611642236, 0.065782837596194299046, -0.18571190868153822362, 0.03864904867246123088, 0.51835394026183756466, -0.88165329043942963771, 0.50750229109300482033, 0.052630627686559745826, -0.14234223813007029769, 0.0054319968161730368406, 0.04656507246486888546, -0.0001091691699465571503, -0.031980600495442469688, -0.01494495062996403778, 0.022375176739023845685, -0.62135409623402226842, 0.43327693944136369542, 1.3189075808916015387, 1.3739630515850600467, -0.40304308760636331499, -4.2245791125828082713, -0.44014500844358800391, 1.3189467827453527082, 1.3030816315258273441, 0.60567723930969952217, -0.021608858399238508458, -0.26531372699538546822, -0.25295782543253098051, -0.11069916902854962104, -0.084489851553929795624, -0.21276587398104537807, 0.09718025396381145542, 0.97081492477032405031, 0.58984488202931195655, -2.7386479014283544586, 0.58597433286032751809, 0.98171649766420365779, 0.13361479757997402062, -0.17576006626469128924, -0.15049487320669766022, -0.036248754857575997412, 0.023755661909215648048, 0.013825250223703235489, -0.057757179482945258941, 0.21138595422899603138, 0.030300290560169362841, -0.49143824072687769533, 0.21258457180099243899, 0.19876333841454826845, 0.21022584371607017095, -0.50201376287393584263, 0.0051240158621321130888, 0.19909052443573113655, 0.05455131526306811196, -0.031012965816702240535, -0.037434250606670348149, -0.0014463645172391665413, -0.21485267380987640995, 0.032884283478948304891, 0.19201471328689778684, 0.54828758511087261063, 0.49943132912277854452, -2.115654463333041857, 0.49561459744078545375, 0.53909005573142354528, 0.1777234997462000532, 0.04412795816862480347, -0.095334831291693913902, -0.11660451153734448237, -0.04726259324103364301, 0.057731651361383765075, 0.19663611996627330947, -0.055196256067221997899, -0.34189926799687614389, -0.24757760693303990096, 0.58889949364028371104, -0.2937532645382701646, 0.59781161376460290047, -0.22537944189565689568, -0.31906829173674161604, -0.068885580362555653977, 0.073580625818116390802, 0.10071872390818351062, 0.044395933345921977098, -0.047244143599278698442, -0.27885992864847214401, -0.051699773352768839396, 0.3841396938758882551, 0.91314440710596511686, 0.31575158811122755464, -2.5720368621032245748, 0.31165315157239603661, 0.90345939423211008013, 0.37667574254367497399, -0.01425646908166054716, -0.16209218978019895863, -0.14440997545450121042, -0.055102382270534898401, 0.070287587247335747698, 0.20466734020628860047, -0.042475450795361184564, -0.3837158926537866388, -0.37967858032718598826, 0.66123992535682063476, -0.11661353115317751872, 0.65973795774971144557, -0.37210464228458967062, -0.36697745701984119115, -0.05301679637320257793, 0.081209497313041584254, 0.11168078003001555287, 0.049699470772157416931, -0.0521051776116896459, 0.93944287167406514971, 0.72667547025920586901, -0.81020590619654786479, -4.4128513456085078914, -11.145405861117756885, 29.463366048162836108, -11.095472961886986951, -4.3875376393047380574, -0.91174001809607752023, 0.48400126865863973835, 0.69758459189035559866, 0.46756239945216088172, 0.14224516757803598588, -0.14730071247481987817, 0.013490706511546561933, -0.27029053027152583066, -0.20226468093194582831, 0.064973687847731428913, 0.6749496186851114965, -0.60463693631529380479, 0.72487435764921248627, 0.14181993599426823138, -0.15938904621443725218, -0.35292498991014370491, -0.30785782369157166372, -0.043134253465759594959, 0.14307421197213984754, 0.16905589597055734052, -0.11843427030792337551, -0.16701094634450047716, 0.00950115946363577768, 0.48012809273490564266, 0.61646553863189879063, -1.6215229274105276147, 0.60204359139300933634, 0.46060970921628185959, -0.00078818683554793267376, -0.15889588085456204936, -0.14573545650606906099, -0.056122252257861368618, 0.014656295947598492468, 0.075289773825147135455, -0.11643728747674478297, -0.15625079318412846141, 0.01787083623578642616, 0.48139756624079788061, 0.60182073966133653453, -1.6405100341602250058, 0.58988396353577543518, 0.46566857610510115073, 0.010112969660060477928, -0.14896136210921839815, -0.14323659597202959004, -0.057182256088921523673, 0.013843985841013588256, 0.073215188921994203697, 0.21491509148813439656, -0.084161771101430457587, -0.43554393847128963779, -0.040428613845152505668, 0.82491614111314415236, -0.9575308266586993744, 0.82276816995700596813, -0.036183566102816408505, -0.42041012713222791763, -0.081834724441143771867, 0.11840037885309515064, 0.1122757183497687522, 0.022197567356135535638, -0.056388353503687517476, 0.16458347543993384621, -0.061109641802337522076, -0.36182677122248141366, 0.014698157406867792818, 0.77716534677722937108, -1.0661228648161740207, 0.77522554853188796464, 0.01778547097859840137, -0.35044699580568000608, -0.058384024725613534179, 0.098524108565335957732, 0.085354349735191417659, 0.01230531294794986151, -0.045228557040033728898, -0.010708898479844557866, 0.10478506751329660918, -0.15877775760536327887, -0.33022115317861205908, 0.55473686346219608367, -0.32298468287768222762, 0.5664674190223346617, -0.31879022383277294539, -0.17133273310353780428, 0.065946316771741134155, 0.0075965827220566186501, -0.023598432637109780646, -0.00064164467062532092555, 0.036370071316341713585, -0.43618952045009057317, 0.22479212375635590493, 0.57894753986959968106, 0.30382634008062331388, 0.087423145343807490137, -1.4836105645541637355, 0.051840917904999374299, 0.24218195066171421814, 0.52152573041896232997, 0.26670474272228600343, -0.10737639280817026788, -0.1979330774617377442, -0.11043980066093539438, 0.050354484693916116977, 0.24813272527090460273, -0.072061997563297092162, -0.45491774629719677892, -0.42108153595973729644, 0.64610922044172458545, 0.1103783370700055777, 0.65092998069437502107, -0.40950225914449023268, -0.44302672933180720083, -0.090889217070203864868, 0.10358814085947003203, 0.12596101939955287041, 0.047781614717171808082, -0.039617615029336937515, -0.053526664263068277072, -0.25007282954957404941, 0.074419905562669141919, 0.99846396735853770732, 0.59972186303125263329, -2.7578142866288106738, 0.60111501295269842426, 1.0108851839798196348, 0.10824806297787861742, -0.20829690864983552157, -0.13659157703522545324, -0.024051592251020062485, 0.01642985276727156746, 0.020151254393120182629, -0.052759547030844865245, -0.24340531213433411195, 0.070371920055568970565, 0.9863683195936597814, 0.59873407263902922448, -2.7368308353092491991, 0.59847583415981975286, 0.99674319302209435367, 0.10320307674402236231, -0.20078458764478651633, -0.12964141475959234384, -0.022393603895432240281, 0.01464221876790986962, 0.016319231085077042648, 0.19788884057403682837, 0.054989945718259124197, -0.25675048529562383148, -0.44830494055239861195, 0.48886632253059703057, -0.079937935276057442424, 0.49161000833835172275, -0.43758361997255651454, -0.23799777932882054676, 0.062169144371137262384, 0.13171035100978070731, 0.10911504833895717403, 0.016729446066571125595, -0.084981485545338811272, 0.17247339407774914255, 0.051653723199437576175, -0.23874898275457501207, -0.41942889676906447427, 0.48671651707814489463, -0.11185663849224913857, 0.49018320992303432426, -0.40844000883161757365, -0.2222211300855479188, 0.054680268388628827658, 0.11070687012705464347, 0.093933896743570638455, 0.017362711973610500671, -0.070734645451936181693, -0.28783083105864049056, -0.1235930412394871114, 0.34005065670227019314, 1.2631945896401703955, 0.52106686401343194071, -3.4328641760913454206, 0.51137318058211267413, 1.2516134908188867314, 0.34153564439344469283, -0.077774976326418818195, -0.18407869225520581691, -0.1452924998260255729, -0.04808123705303415929, 0.066123663903966400923, 0.0036130831750669917257, 0.0048546396571439841966, -0.12551659716901422681, 0.064366966054205124181, 0.50750229028511717466, -0.89558891467248546991, 0.49906795787301211353, 0.050231233764940881603, -0.14209518129652423668, 0.0054319977735493244625, 0.046250056631165985255, 0.0017726937269519947929, -0.024579867738323060711, 0.0035362853144340865297, -0.5474760814697108291, -0.14302043280446605222, 0.58764006474821284254, 1.3279126309973636744, 1.342265994573186072, -0.44014500774734899302, -4.2534936177175710625, -0.45492408530503081199, 1.3189467817334414868, 1.3116594145352395984, 0.61005135711859526104, -0.036998866417477016544, -0.30936478749516144182, -0.31375860137973621278, 0.041415049667265370348, -0.16499680924499143808, -0.20185747034185230753, 0.1262788595676229142, 0.98494658172592941536, 0.58597433356191386267, -2.7469941532607071721, 0.58159194678669756673, 0.98171649773111435699, 0.13485238104491159561, -0.17663822607839946288, -0.15532796110997246219, -0.03976627031420161551, 0.046864104109190236891, -0.082009723962536848485, 0.052536662279617202254, 0.21015038143383649172, 0.011021054731540788629, -0.50195427257779590136, 0.21022584435439170547, 0.20061251662068319002, 0.21182682995164536788, -0.50201376289915333739, 0.0044308544014231765126, 0.19918524667219983382, 0.055723734473286122559, -0.031451144187834650934, -0.038922302914554549047, -0.094581391231669323383, -0.088637001600956125125, 0.036834336508866125093, 0.16865010180611297863, 0.53424680312928318759, 0.49561459620764841683, -2.1137911276586676124, 0.49825216279818707044, 0.5390900562400935403, 0.1728070456869004734, 0.034727415967851976708, -0.10225283462019671421, -0.10379952177091769649, 0.017021475275737156518, 0.10319098821574780589, 0.06726495787181722974, -0.076266329524796616646, -0.3185572668014079456, -0.22185360321464117628, 0.59781161423090700691, -0.29649727549669258053, 0.59322377734765463497, -0.22537944220510230942, -0.31331666015450382323, -0.061798347960828038661, 0.072668612294951015484, 0.081409545580856751412, 0.0043413540893209026303, -0.10346333677680050189, -0.15641328540716201156, -0.030307411657579809428, 0.35981557415904491526, 0.89479592156424880844, 0.31165315145101513128, -2.568063023601289796, 0.31622025899975736429, 0.90345939445375011978, 0.36854803588263418401, -0.029728669860050644519, -0.17331049915420562724, -0.12592231215004895395, 0.025620296205353724345, 0.09252019521431721627, 0.076988084251041311035, -0.043705612128734777633, -0.35381999052868173816, -0.36399865612108872615, 0.65973795788139566465, -0.12335408175638366834, 0.65481601492845464829, -0.37210464215152722067, -0.36043023012234309821, -0.040559271809337042736, 0.092032875491728838302, 0.10010349485464796471, -0.014081854392317858307, 0.31044067150099341479, 0.74605166408294465441, 0.5680784128562584856, -0.86955258997574402802, -4.3864761113619756827, -11.095472962634207903, 29.495516909680915774, -11.087607581305881865, -4.387537638882643698, -0.89908620745317602285, 0.51555171001697752686, 0.7237770307550898119, 0.42863855638615733223, -0.036518575698889221837, 0.34750211436126376974, -0.25339246917692542782, -0.42079619247986527197, -0.23394351347173159006, 0.10139102374338718193, 0.72487435755159923545, -0.56713754755904077776, 0.74353885709926492709, 0.14181993554672647773, -0.17943410006777860133, -0.39176669856644547041, -0.34460017343480042751, -0.015189396377637588709, 0.3480513285165837134, 0.0053687364184762968386, -0.11710911645917146395, -0.14617008006188608737, 0.0057016435541768634154, 0.46365978710068878055, 0.60204359082033431783, -1.6275740331083718981, 0.60130306674649003362, 0.46060971024330532586, -0.0017272467523570811387, -0.15737639996304442214, -0.13357917452098255762, -0.026528648754434808188, 0.059810586324646961009, 0.008467930701330565052, -0.11655098190950322068, -0.14030887616665810302, 0.012998694108768152344, 0.46673486266937630473, 0.58988396359320660611, -1.6445791415653006062, 0.5902029821883510996, 0.465668575679321739, 0.0080794723481245611652, -0.14956052513200274179, -0.13382244025528472475, -0.029429821619739983396, 0.061997430505695262648, 0.055648286743883355887, 0.12227767889782725774, -0.070173737440230324336, -0.41351210065893212464, -0.033990280585407400826, 0.82276817019491510496, -0.95932407705158573297, 0.82225081220970708262, -0.036183565982370442804, -0.41871477483284935683, -0.075661448444652529277, 0.12789931570453966225, 0.10609806557852449749, -0.044605656028492889076, 0.034789412825987617306, 0.10136126414210445967, -0.049333098433941242889, -0.3451465756799902973, 0.019542448405650367921, 0.77522554888256223204, -1.0677916475043238709, 0.77462245168816323826, 0.01778547118713633482, -0.34903321877327048162, -0.053660436928985325022, 0.10557737693060439965, 0.07990358535177960464, -0.040141063590879731693, 0.014098181932404396308, 0.023397888600343089843, 0.060867685773225703449, -0.18388675034971588285, -0.32771119206752957753, 0.56646741978590642308, -0.31323073374129251079, 0.57081757714299397044, -0.31879022399614975747, -0.17507182449781233391, 0.058227212494064893578, -0.00068867197854117340444, -0.019625729144321226588, 0.044977331024025089812, -0.29031166444605754906, -0.12168236803475386554, 0.29629948998808830485, 0.55785175651050644685, 0.26394031084257962849, 0.051840917395842824045, -1.5072958569053285238, 0.040684016682417306621, 0.24218195033710765429, 0.53023439895390567234, 0.2805305611929613252, -0.092778682516352489751, -0.18393021303037160319, -0.085144583562195563786, 0.1102291056779043793, 0.11625972450052984852, -0.080678495950060435948, -0.43982037939763884982, -0.40992642215739905076, 0.65092998048943362388, 0.11255554716928795056, 0.65148297725691428628, -0.40950225901552628205, -0.4415420760690191293, -0.085732907737457328401, 0.1113391990546536009, 0.12230166876606496373, -0.0026623672119349480088, 0.048702709822867971956, -0.13822780119274458155, -0.2336169245034764308, 0.09061085577160424287, 1.0049388406209340818, 0.60111501291316615792, -2.7569602267461741363, 0.60223146232595581306, 1.0108851833297136658, 0.10411811493527148298, -0.217729012299517799, -0.14504189172394787999, -0.018328499712584218895, 0.046584469686475955819, 0.041928166546455127028, -0.13258658041272092243, -0.22396539418786298725, 0.088202184783809228774, 0.99233090154037562058, 0.59847583468521148298, -2.7374123966494949833, 0.59888340806557305296, 0.99674319395795651921, 0.099846061476988531846, -0.20864000786978026403, -0.13656537818273975504, -0.017462177862302662168, 0.039247930774301574564, 0.04352834751457242829, 0.11270258402398942998, 0.064458426449152117499, -0.23309954095236307126, -0.43605967436269216231, 0.49161000880423422199, -0.079530718425429372531, 0.49143049606791561201, -0.43758361958892538768, -0.23613919077024619897, 0.066979531172196765665, 0.13472416596189651372, 0.089966809423520296973, -0.063224568334980441042, 0.043998745841450886962, 0.095011361871080210473, 0.055587284162932820708, -0.2189925892068427804, -0.40764253395402483271, 0.49018320912050028504, -0.11094990537026071242, 0.49016646932061658237, -0.40844000889972786883, -0.22071239087400953971, 0.058452209469047528012, 0.1126593563313899593, 0.077344860121761671579, -0.048272176750068347684, -0.10064907848027443427, -0.18429450489030990701, -0.094836762555434842015, 0.33147507980440654718, 1.2501792057124798685, 0.5113731805533494601, -3.4376314700268997271, 0.51046125192954805438, 1.2516134900530258012, 0.33808153130384938079, -0.086833118059338798567, -0.19199072705515146664, -0.13080746927959951975, 0.025027513834788538138, -0.049925358150063971208, 0.060547827971307956185, 0.022071714348462945221, -0.13379913077226965323, 0.052630626828215752722, 0.49906795774338252869, -0.8981784950922385713, 0.4994535905608908366, 0.050231233696683530443, -0.14234223788341143768, 0.0084226513527806898263, 0.054564567006408500049, 0.0089946209605654050406, -0.034310541782238579422, -0.31375860156651452781, -0.30936478683937818079, -0.03699886690274638984, 0.61005135727870063445, 1.311659414240180288, 1.3189467817222191304, -0.45492408540227430302, -4.2534936171812081085, -0.44014500828059721638, 1.3422659947070441078, 1.3279126313602613862, 0.5876400647088045881, -0.14302043293382954436, -0.54747608154794680235, 0.046864104019834186554, -0.039766270140721086956, -0.15532796122157571661, -0.17663822603697795777, 0.13485238113469283361, 0.98171649752266298972, 0.58159194683180404084, -2.746994153124428184, 0.58597433342746119145, 0.9849465818326261779, 0.12627885953830542132, -0.20185747036798803999, -0.16499680926629142763, 0.041415049694564234906, -0.038922302972986218272, -0.031451143927145992774, 0.055723734027412319403, 0.19918524703634851725, 0.0044308542976055336621, -0.50201376299376709866, 0.21182683006777047741, 0.20061251656117690123, 0.21022584448130127166, -0.5019542727296125717, 0.011021054580981575963, 0.21015038187521245527, 0.052536661956144294072, -0.082009723883831042723, 0.017021474927599304638, -0.10379952094135168339, -0.10225283458636311162, 0.034727414850808074887, 0.17280704583380074313, 0.5390900566666877447, 0.49825216369310976194, -2.1137911285078088142, 0.4956145958431206755, 0.53424680291826642797, 0.16865010273777195082, 0.036834336640324601142, -0.088637002503270029274, -0.094581390830675349246, 0.0043413540856659443498, 0.0814095453589985113, 0.072668612948656133299, -0.061798348483277830423, -0.31331666013813075367, -0.22537944211293955976, 0.59322377722361530061, -0.29649727543549209141, 0.59781161451169029242, -0.22185360362670131229, -0.31855726653145349792, -0.076266329585212566666, 0.067264957807003145085, 0.10319098825080535997, 0.025620296155266425453, -0.12592231199215755955, -0.17331049923568722648, -0.029728670032424634601, 0.36854803605335045846, 0.90345939461235380552, 0.31622025855292645868, -2.56806302309858836, 0.31165315103395013763, 0.89479592170962041209, 0.35981557435302535808, -0.030307411892949370075, -0.15641328531716416905, -0.10346333679306199405, -0.014081854459930279178, 0.1001034949677227226, 0.0920328755667800813, -0.040559272111399223559, -0.36043022986122930096, -0.37210464228236361794, 0.6548160150042795502, -0.12335408162022523659, 0.65973795747865959793, -0.36399865581813539706, -0.3538199904304166199, -0.0437056124652861408, 0.076988084478136986633, 0.092520195164391749731, -0.036518575642014945715, 0.42863855641473730396, 0.72377703061319420552, 0.5155517099202725495, -0.89908620741840883372, -4.3875376386540398954, -11.087607581164432347, 29.495516909430051555, -11.095472962794996619, -4.3864761111487347023, -0.8695525900773395378, 0.56807841285948945664, 0.74605166417672763579, 0.31044067145493264848, 0.34805132852657555409, -0.015189396382655730861, -0.34460017352331434592, -0.39176669842407346733, -0.1794341001878724251, 0.14181993587216251695, 0.74353885681592501999, -0.5671375478420010996, 0.72487435766502861245, 0.10139102446914287359, -0.23394351410400401892, -0.42079619251337163632, -0.25339246905646289898, 0.34750211437157130234, 0.059810586427763254291, -0.026528649278952369517, -0.13357917445626871222, -0.15737639932039104718, -0.0017272467552060799534, 0.46060970992728367257, 0.601303066736102787, -1.6275740330261285749, 0.60204359106403870783, 0.46365978683969677476, 0.005701643174577164519, -0.14617007986824753774, -0.11710911628845160903, 0.0053687364127012063633, 0.06199743068150678077, -0.029429821989384076508, -0.13382244025190323522, -0.14956052496338037572, 0.0080794726553334861308, 0.46566857527038918452, 0.59020298232627643742, -1.6445791414974542111, 0.58988396366533923931, 0.46673486239970396472, 0.012998694090347359675, -0.14030887602807906589, -0.11655098190449356133, 0.0084679306920606815801, -0.044605656047650668605, 0.10609806553100747939, 0.1278993159448935657, -0.075661448650971324659, -0.41871477476992624567, -0.036183566044638973414, 0.82225081225136453789, -0.95932407726777568779, 0.82276817030416227183, -0.033990280138386001763, -0.41351210093031881554, -0.070173737781532907754, 0.12227767926244549612, 0.055648286641540124775, -0.040141063491301955113, 0.07990358505588432414, 0.10557737726778693155, -0.053660437103303010309, -0.34903321863805630842, 0.017785471140322622191, 0.7746224513898440911, -1.0677916474601554242, 0.77522554950463873347, 0.019542447855341302243, -0.34514657546465510318, -0.049333098680279151294, 0.10136126438587376997, 0.034789412740461211904, 0.044977331151542100651, -0.019625729510184639071, -0.0006886718250679594493, 0.058227212596207819639, -0.17507182408073793112, -0.3187902247042078141, 0.57081757731826510138, -0.31323073355328567846, 0.56646741957369450837, -0.32771119150925431818, -0.18388675114237679198, 0.060867685963830957252, 0.023397888972043413142, 0.014098181725005768522, -0.085144583891565581157, -0.18393021261298797553, -0.092778682578262272607, 0.28053056132261089406, 0.53023439861325782907, 0.24218195078862544967, 0.040684016146191319396, -1.5072958567107872518, 0.051840917878400512298, 0.26394031040645060671, 0.55785175652288909731, 0.29629948987997217769, -0.12168236810411860171, -0.29031166425065341175, -0.002662367147350496073, 0.12230166862170510633, 0.11133919910482786841, -0.085732907661992388548, -0.44154207604133616227, -0.40950225932609951585, 0.65148297781911268522, 0.11255554662027433821, 0.65092998081044273206, -0.40992642227244435738, -0.43982037940668322618, -0.080678495924627127711, 0.1162597245169755128, 0.11022910566316086733, 0.046584469779944175094, -0.018328499993711711707, -0.14504189118525259716, -0.21772901258581006667, 0.10411811436749571214, 1.0108851841928472304, 0.60223146188846221616, -2.7569602269310884424, 0.60111501360752006562, 1.0049388397277241314, 0.090610856470082684622, -0.23361692452839810663, -0.1382278015474246724, 0.048702709966623854154, 0.039247930341052575387, -0.017462176952800827212, -0.13656537863516252096, -0.20864000805790336446, 0.099846061955562642742, 0.99674319318284188007, 0.59888340861077316024, -2.7374123962559870904, 0.59847583332451881333, 0.99233090313523242276, 0.088202184020814977661, -0.22396539438052384674, -0.13258658010527826776, 0.041928166481346403061, -0.063224568331096617224, 0.089966809846313167132, 0.1347241653808529116, 0.066979531072564990613, -0.23613919022640647571, -0.43758362022786162493, 0.49143049640535974021, -0.07953071793313033766, 0.49161000777486263402, -0.43605967365442332984, -0.23309954124695464461, 0.064458426833033488634, 0.11270258386890250057, 0.043528347419428715936, -0.048272176416258513743, 0.077344859467911655782, 0.11265935661422746272, 0.058452209910775154711, -0.22071239199115916252, -0.40844000765431986988, 0.49016646880372682915, -0.11094990564763322627, 0.49018320969631612449, -0.40764253474059403137, -0.21899258860840148389, 0.055587284288752016048, 0.095011361568256888499, 0.043998745893241472671, 0.025027514266984084851, -0.13080747026626513008, -0.19199072682221296482, -0.086833117371991230815, 0.33808153110394073559, 1.2516134900781570316, 0.51046125130523900548, -3.4376314697487373451, 0.51137318135703468958, 1.2501792047890820658, 0.33147507992234770402, -0.094836762429523838169, -0.18429450475825234768, -0.10064907858206370572, -0.034310541626295362749, 0.008994620691939344409, 0.054564566990191354368, 0.0084226514324600154637, -0.14234223770185422397, 0.05023123351476400239, 0.49945359079926032742, -0.89817849541191407958, 0.49906795785501717466, 0.052630626900893789921, -0.13379913081002547948, 0.022071714170360339641, 0.060547828234726694396, -0.049925358250688549322, -0.11069916840650918799, -0.25295782650592191221, -0.26531372666110508707, -0.021608858839637035132, 0.6056772394886313915, 1.3030816325838407987, 1.3189467822072320491, -0.4401450085618485164, -4.2245791126459808495, -0.40304308757811113662, 1.3739630513000700152, 1.3189075811037505037, 0.4332769397993799787, -0.62135409652115713985, 0.013825250105491501684, 0.023755662184718993879, -0.03624875520184128036, -0.15049487278737994078, -0.17576006662831453187, 0.13361479779581231231, 0.98171649729128640516, 0.58597433350767380578, -2.738647901941571039, 0.5898448822361801458, 0.97081492451687456668, 0.097180254435275878477, -0.21276587438018834675, -0.084489851426921183664, -0.0014463646110357900043, -0.037434250241781534541, -0.03101296634875119751, 0.054551315659033025596, 0.19909052414304073686, 0.0051240162426922390357, -0.50201376342082160154, 0.21022584448046960359, 0.19876333761179723614, 0.21258457241990563746, -0.49143824129821717106, 0.030300291149562130072, 0.21138595386519978825, -0.057757179393854184146, 0.057731651465918340516, -0.047262593095576871383, -0.11660451193620212373, -0.095334831313723014667, 0.044127958450406463975, 0.17772349986794094834, 0.53909005544116184172, 0.49561459729352358439, -2.1156544631088340935, 0.49943132923498628761, 0.54828758473097505277, 0.19201471349335455518, 0.032884283836019392866, -0.21485267411950154104, -0.047244143568867552618, 0.044395933324500598538, 0.10071872383981252119, 0.073580625990063222575, -0.068885580462202833374, -0.31906829178904749789, -0.22537944207855706291, 0.59781161414222905304, -0.29375326462421541551, 0.58889949362110105557, -0.24757760719785543158, -0.34189926763467509696, -0.055196256234458111745, 0.19663611998685859295, 0.070287587097791079382, -0.055102381813701743762, -0.14440997593172685831, -0.16209218959832991103, -0.014256469015908745587, 0.37667574228473055253, 0.90345939454029167148, 0.31165315145484173698, -2.5720368621552465171, 0.31575158815712200955, 0.91314440702806998207, 0.38413969409787357412, -0.051699773561506977282, -0.27885992858725350274, -0.052105177695019169737, 0.049699470923850332105, 0.11168077994458139313, 0.081209497435456898251, -0.053016796473296878833, -0.36697745721829816423, -0.37210464197189641045, 0.65973795763615428189, -0.11661353117512696098, 0.66123992530457809114, -0.37967858014016953039, -0.38371589276400486224, -0.042475450876857258342, 0.20466734028000907464, -0.14730071258332502748, 0.14224516785275775899, 0.46756239925994602658, 0.69758459201818068252, 0.48400126868446341488, -0.91174001845109531317, -4.387537638849669186, -11.095472962471937706, 29.463366048610694747, -11.145405860993641056, -4.4128513459245848338, -0.81020590603060427259, 0.72667547012947497631, 0.93944287174534957252, 0.16905589588919245347, 0.1430742119388650202, -0.043134253315669099371, -0.30785782338576112016, -0.35292499018301420888, -0.15938904649303958783, 0.14181993621402097516, 0.72487435743469574856, -0.60463693596519574314, 0.67494961855222079805, 0.064973687853500119993, -0.20226468083936943732, -0.27029053048907786261, 0.013490706619197484831, 0.07528977368326758568, 0.014656295991371158516, -0.056122252051873378065, -0.14573545670342419323, -0.15889588034807003436, -0.00078818769335281535504, 0.46060971007028761459, 0.60204359073917557144, -1.6215229269759665609, 0.61646553823925764881, 0.48012809308394993613, 0.0095011591883806256753, -0.16701094628311802248, -0.11843427024194687347, 0.0732151887076757657, 0.013843986322208692852, -0.057182256213772125752, -0.14323659634074809022, -0.1489613617793027478, 0.010112969647670860818, 0.46566857614101209162, 0.58988396303445156388, -1.6405100337325857485, 0.60182073968840066325, 0.48139756628639962521, 0.017870836061708465781, -0.15625079314223011551, -0.11643728746771606075, -0.056388353567891576212, 0.022197567300990771882, 0.112275718660762025, 0.11840037882495169086, -0.081834724824168397772, -0.42041012701569946408, -0.036183565987797670915, 0.82276817010907876693, -0.95753082696829727727, 0.82491614125140544367, -0.040428613601855532822, -0.43554393877542624258, -0.084161771153884998298, 0.21491509160798055733, -0.04522855698205362901, 0.012305312685266691156, 0.085354350105892387579, 0.098524108390143444458, -0.058384024639614638674, -0.35044699599737849649, 0.017785470831560117044, 0.77522554907570528915, -1.0661228650342196023, 0.77716534657506008976, 0.014698157732312627094, -0.3618267714879803143, -0.061109641805410550019, 0.16458347552166530114, 0.036370071259124364671, -0.0006416445864348725844, -0.023598432646109942373, 0.0075965826043384856656, 0.065946316811631613963, -0.17133273287029257559, -0.31879022404043266814, 0.56646741911458620145, -0.32298468311872968739, 0.5547368638811311925, -0.33022115356890818783, -0.15877775744321862006, 0.10478506750755323118, -0.010708898484342194504, 0.050354484557443082715, -0.11043980054901657295, -0.19793307730197273742, -0.10737639253725461685, 0.26670474219059464582, 0.52152573041134431264, 0.24218195059782243739, 0.051840918150826142963, -1.483610564579151081, 0.087423145524494899616, 0.30382633988747437259, 0.57894753998514436599, 0.22479212333595258166, -0.43618952014695711039, -0.039617614927745514741, 0.047781614351163501087, 0.12596101991055763669, 0.10358814048117259699, -0.09088921698106895708, -0.4430267291287010023, -0.40950225938495704581, 0.6509299808088215844, 0.11037833699196233916, 0.64610922041469820432, -0.4210815356698376366, -0.45491774671982021516, -0.072061997296177432437, 0.24813272520669338306, 0.020151254584330875741, 0.016429852280718075519, -0.024051591609663969251, -0.13659157759425949452, -0.20829690825219310257, 0.10824806272713347055, 1.0108851839003629713, 0.60111501291329305641, -2.7578142859130330145, 0.59972186200909249898, 0.99846396814475080195, 0.074419905173997963344, -0.25007282938524011406, -0.053526664331771098282, 0.016319231042924958741, 0.014642218707793367566, -0.022393603613625967519, -0.12964141499063375207, -0.20078458735107501676, 0.10320307602847507422, 0.99674319366530372832, 0.59847583414279603709, -2.7368308355009132171, 0.59873407241484288832, 0.98636832031717669445, 0.07037191919913238225, -0.24340531162138093202, -0.052759547148063128119, -0.084981485585011146133, 0.016729446498070406346, 0.10911504764977703053, 0.13171035114769202856, 0.062169144565674488989, -0.23799777929352733397, -0.43758362006549017531, 0.49161000846042146595, -0.079937935489900149855, 0.48886632262058721254, -0.44830494051020419732, -0.25675048543959694269, 0.054989946065038299017, 0.19788884034902426423, -0.070734645173719482014, 0.017362711385368975736, 0.093933897010097827662, 0.1107068701581282455, 0.054680268364212969434, -0.22222113004099613986, -0.40844000865830343283, 0.49018320965804679235, -0.11185663821952410157, 0.48671651660343828727, -0.4194288965946150749, -0.23874898250436038527, 0.051653723115304181479, 0.17247339402019809529, 0.066123663875646332411, -0.048081237033017115712, -0.14529249989155901268, -0.18407869224325584812, -0.077774976279080976216, 0.34153564457126828202, 1.2516134906867919518, 0.5113731805031027644, -3.4328641760885085787, 0.52106686408434832458, 1.2631945895752312303, 0.34005065676343915193, -0.12359304128271890455, -0.28783083104131407248, 0.0035362852723952327948, -0.024579867661718830807, 0.0017726936255444321073, 0.046250056837026168322, 0.0054319972915475404907, -0.14209518052600741966, 0.050231233141949008236, 0.49906795827177674596, -0.8955889153272823533, 0.50750229115514422507, 0.064366965498623460173, -0.12551659704055348477, 0.0048546396517973756241, 0.0036130831884414937524, 0.011313547742351888481, -0.13699152199997841062, -0.25612753576891539309, -0.26369714052155945883, -0.022375740647500381303, 0.60567723995121647285, 1.3116594135592436476, 1.3422659952263500394, -0.40304308811138106483, -4.1835208875227944958, -0.36789718036355895103, 1.3725925240489982304, 1.1798639592920716623, -0.15442901329681280043, -0.0048561495705649751833, 0.024974751683368327099, 0.02043780954362323099, -0.041103744878176286248, -0.15241507214964941652, -0.17576006696445450883, 0.13485238154263584631, 0.98494658157024284062, 0.58984488198792250913, -2.7383477949536452201, 0.59386008614480467749, 0.98588822589741731406, 0.09549974898588585448, -0.3174080463202961222, 0.012778064211799318103, -0.017326610392924522014, -0.039409502500070615894, -0.028223856001492349937, 0.056410062199723616516, 0.19909052367776078341, 0.0044308548902036101746, -0.50195427257579272595, 0.21258457203217615006, 0.20355001522757665278, 0.21862544837003891618, -0.49139549054353187652, 0.031776421643626068247, 0.14171532284264537482, 0.044552018739663225289, 0.0077533339377604103965, -0.055994137368221695572, -0.1103701174423972059, -0.089532662798088830103, 0.044127957267427228571, 0.17280704681619771246, 0.53424680234372312615, 0.49943133005941198199, -2.0938581768220108081, 0.54320764079432648508, 0.57988626581715951858, 0.13384390012983291007, -0.20712294036759365534, -0.050831048717041006635, 0.0051512272945795321843, 0.058133489870492202622, 0.09543196860674203752, 0.06677708667025528122, -0.068885580151142739025, -0.31331665962573529427, -0.22185360379271756859, 0.5888994930347685175, -0.31957921262834704246, 0.55764580833669641358, -0.26813258043262372388, -0.29698791355933673586, 0.1648591602014847779, 0.052302630135131367162, 0.0083048536566507104306, -0.064480012396722879986, -0.13417536625686307672, -0.1529939211357471962, -0.014256469308139835608, 0.36854803652239320755, 0.89479592095791293804, 0.31575158852591800329, -2.5419896765147789175, 0.37301953538945953337, 0.95077289053556723886, 0.30984085100565261772, -0.36127698072795344508, -0.036911894093314544241, -0.0036685632571531409142, 0.053345528877593972883, 0.10138274254528094331, 0.073440298604544498873, -0.053016796277967917495, -0.36043023053899586605, -0.36399865567248423925, 0.66123992542024712105, -0.1336129648784556978, 0.61994973143811626937, -0.42532047069335704448, -0.33520400169136899127, 0.19947025986713357026, -0.13416932212243448674, -0.013840490531778069927, 0.18542363208290005283, 0.46259957952655594848, 0.6839666663972904237, 0.48400126849620034308, -0.89908620784523474168, -4.3864761106818974667, -11.145405860414063781, 29.336315028025484963, -11.299906245086104306, -4.4341302449120556517, -0.4626163706138075038, 1.6024975485391295127, -0.0054138629263914004111, 0.14867158180416245594, 0.15994823278911893261, -0.017857593544580517564, -0.29155906464093783725, -0.35292498956452067471, -0.17943410031455289677, 0.10139102441887257178, 0.67494961825629140595, -0.63352124932710762728, 0.71718628958808461338, 0.20197015423416234126, -0.050096068297056553509, -0.49120642917807788574, 0.064131851596392425163, 0.030078722040171899238, -0.0097707730536618145334, -0.063717554494970740642, -0.145956265959940934, -0.15889588043895364056, -0.0017272465310113338521, 0.46365978646738958968, 0.61646553843617990509, -1.595809966533479507, 0.64090718152398040708, 0.47542375935835590184, -0.031712143703299638942, -0.28804525777622597138, 0.058463824105462455782, 0.030012531586889702045, -0.0078181927768891795738, -0.062522444299524276845, -0.14226987752406092258, -0.14896136090671110996, 0.008079471329531795476, 0.46673486353863247178, 0.60182073914970279915, -1.6167313496634121428, 0.62710502288385117087, 0.4819636337580527341, -0.018474959710575503818, -0.28187363430025663691, -0.03084497019550455979, -0.023200771758241148929, 0.027824653871769140789, 0.10905901481819266219, 0.11568063159922976901, -0.081834724677736517262, -0.41871477512461763304, -0.033990279934207723311, 0.82491614127874435258, -0.96144572008148065922, 0.79985066611963173688, -0.10121232066650462378, -0.45047572736393098891, 0.22315601429707535019, -0.023649503615509236132, -0.020389090604196909856, 0.017011662168875493761, 0.083095185812959881932, 0.096381566855637548774, -0.058384024934089312486, -0.34903321850142737803, 0.01954244834628864752, 0.77716534637331102786, -1.0681000911389422203, 0.75960271727184003332, -0.034447203681543955844, -0.38159249981433518428, 0.18228094337508413125, 0.0072897306429247224827, 0.019895821785769946671, 0.0041714090808937388105, -0.016029449475706978451, 0.011279231273677506678, 0.065946317500826842628, -0.17507182409379531962, -0.32771119174078600089, 0.55473686327125559803, -0.32727657089637046672, 0.57178834467414241249, -0.30189539795719877446, -0.14338973470254331999, 0.053802355657625082486, 0.1005367838935306013, -0.024727943374339911992, -0.14671495450077359779, -0.21299970012752536541, -0.11326333177303370137, 0.2667047423494742775, 0.53023439841194675015, 0.26394031036103660082, 0.0874231457096925485, -1.4411179994695837436, 0.12484835820272996432, 0.30749971042137524968, 0.48867849670246210625, -0.22280643234548516696, -0.033392054451390040803, -0.0039370251932049557531, 0.055450543085361583828, 0.12320237501138130154, 0.10058431428072636682, -0.090889217380790685019, -0.44154207570914028036, -0.40992642238280463385, 0.64610922059394781947, 0.097428997545431883509, 0.61760995959201969718, -0.46777417039503216278, -0.42675834634408493784, 0.22915954761607038037, -0.0015387258730293058677, 0.017110926138629023452, 0.017564534337182649559, -0.020126467741837072989, -0.13329887735013096206, -0.20829690830910430055, 0.10411811471233961557, 1.0049388402670900167, 0.59972186170882746392, -2.7494281254232473266, 0.61735531066804794076, 1.0048029275608882926, 0.051416499517096800331, -0.30418107840419356114, -0.0010652437101750017109, 0.014510809379490988494, 0.014787500840245174083, -0.019759284742865632079, -0.12706861903425195615, -0.20078458675135946176, 0.099846061657272958123, 0.99233090112021271612, 0.5987340734482979876, -2.7281545489808300964, 0.6142044221492869438, 0.98919549686932073396, 0.045164739361996358857, -0.29156025953042979948, -0.057065145507962204696, -0.033103823575289494097, 0.030702199324839994643, 0.10923267734613878455, 0.1297623043583514546, 0.062169144675318817328, -0.23613919076557365884, -0.43605967420721736261, 0.48886632327563495082, -0.095825904403698869416, 0.45059610884076151116, -0.49656594515695978709, -0.20896633508724030137, 0.29445230459025600878, -0.049882146694868539916, -0.025527344276922232097, 0.029883860133880099652, 0.094275452157669553666, 0.10905850596077429349, 0.054680267178826028052, -0.22071239051736218917, -0.40764253484964529939, 0.48671651683698891055, -0.12684293466331533762, 0.45371932018192684399, -0.45865627608243281221, -0.19207933408189392654, 0.25436585645099668573, 0.052617894419969953035, 0.013045316725795917667, -0.061369832491863035573, -0.14266359356063793018, -0.17993713097570704362, -0.077774975848153973801, 0.33808153085900177759, 1.2501792057521046164, 0.52106686314096839485, -3.4033201501148937496, 0.57281647553077641444, 1.2956481803031174582, 0.25034978182285216874, -0.42460837706480181186, 0.02237517672807392069, -0.014944950613017076585, -0.031980600620659944966, -0.00010916887380199298378, 0.046565072184667391508, 0.0054319968886008783426, -0.14234223797594836025, 0.052630627453943874816, 0.50750229127074142355, -0.8816532906301214334, 0.51835394043485349957, 0.038649048807664718175, -0.18571190902558648572, 0.065782837753215334109, 0.063857669310089584802, -0.042801180214380341038, -0.16015572474155601035, -0.26192833938485726719, -0.26369714015759088133, -0.021608859308127632143, 0.61005135748146377406, 1.3279126310599373983, 1.373963051683954939, -0.36789718108533736585, -4.1841988960068343673, -0.41342802230875774772, 1.2815874964876636799, 1.1023691002391606819, -0.0037552242656817441148, 0.0054104821876786741291, 0.02092551830507270727, 0.016425248625095614091, -0.04110374437574036699, -0.15049487278143997004, -0.17663822620577276923, 0.12627885970598418108, 0.970814924732384954, 0.59386008598536665914, -2.6899447738679933551, 0.69250445573739083382, 1.0010555932423419456, -0.35977412102477096534, 0.011135020837560316923, -0.00032063236171854124414, -0.018668845373536857579, -0.037631914012956596549, -0.028223855316752404637, 0.05455131540400287582, 0.19918524701879536387, 0.011021054145863512855, -0.49143824076229364328, 0.21862544837741018644, 0.19325555969133445533, 0.17806403587623398366, -0.59283635356550090645, 0.30518686815034401505, 0.016980930270830960072, 0.026661774604380870934, 0.0049878725895568763737, -0.053559331633278237095, -0.11037011862053318356, -0.095334830845190163262, 0.034727415611396077189, 0.16865010160819962626, 0.54828758460201831859, 0.54320764214687167204, -2.0420836855391564235, 0.54769246843185948048, 0.44534728646101690019, -0.025798857334729228402, -0.027378040748510610453, -0.020292431678701372544, 0.0090441508198790855622, 0.054097629105176528452, 0.095431969416695877628, 0.07358062553770812475, -0.061798349150857807832, -0.3185572659537382334, -0.24757760681726848073, 0.55764580769338267796, -0.33253973581572349172, 0.58301313887818351223, -0.22558876347613035507, -0.14755073796316026113, 0.017336255994376732065, 0.032444813371075659214, 0.0079040441278002259579, -0.059791367999863065386, -0.13417536641950497667, -0.16209218923560186898, -0.029728670481431768158, 0.35981557478189118982, 0.91314440669904117787, 0.3730195356064335277, -2.480130740121571975, 0.39433682135695596882, 0.86438565032715641845, -0.083271076414070138716, -0.0086573911910000703185, -0.024634986412741623207, -0.0057510054896994901655, 0.048460669353596212694, 0.10138274269946315198, 0.081209497088626986239, -0.040559271261203344583, -0.35381999106779521425, -0.37967858000172605148, 0.61994973158128385826, -0.17358457209364366869, 0.63655052443989634803, -0.38626442698728036262, -0.12297271674334349367, -0.080633263558065229604, -0.068122799381342608038, 0.012231610298223252048, 0.18851462494803145065, 0.46259957862852363331, 0.69758459215757739891, 0.5155517107588498682, -0.86955259092419079359, -4.4128513462225349429, -11.299906244284647627, 29.025977066924429693, -11.542093593676536756, -4.0202401380086607574, 1.3305305794320547719, -0.090654903970312866734, 0.072239660718494178027, 0.18503346490505309663, 0.17361895375207250791, -0.017857592936543473239, -0.30785782382292642412, -0.39176669891453524963, -0.23394351362029178198, 0.064973688035939225482, 0.71718628898476510791, -0.47387939374726917308, 0.92728010569065044688, 0.11312546867672171325, -0.75386833874879233264, 0.037059309156806802266, 0.032394662355561154587, 0.016778063621379527004, -0.013089066707686697333, -0.063717554890353758634, -0.14573545571244200802, -0.1573764006611988453, 0.0057016447389717228589, 0.48012809235162934307, 0.64090718144903924269, -1.5755953819535022387, 0.66318487948124482667, 0.42236455090173141702, -0.33854343791886404302, 0.030957833835310036991, 0.0312671618351204883, 0.019429154487103458826, -0.009845932499061835208, -0.062522444079318173493, -0.14323659704734542641, -0.14956052386002249843, 0.012998693085435775174, 0.48139756717959347743, 0.6271050220587619517, -1.5917287544641010033, 0.65395641269309578991, 0.42746457808916871191, -0.32355286477682593249, -0.0067762777091035005655, -0.02633439734197590143, -0.020230447397741168486, 0.028197623804502892908, 0.10905901489082212019, 0.11840037869242261981, -0.075661448356444532815, -0.41351210109489944244, -0.040428613843730733246, 0.79985066626337952833, -1.0046531793731654503, 0.77870256361715217697, -0.057195080138013568161, -0.19595221642683627317, -0.0045501207666335044544, -0.020805710050514479581, -0.018217018188974046233, 0.017257528112079048577, 0.083095185895677964316, 0.098524108521164541208, -0.053660437092802999726, -0.34514657560607281406, 0.014698157580875037195, 0.75960271738984042056, -1.0999750939457320076, 0.7399594540247723673, -0.013830050793951520829, -0.16154693992861574037, -0.013506100726906980838, 0.017203292093883774255, 0.027002417312537309313, 0.0075555516141473004835, -0.016029448238832108564, 0.007596582116937322833, 0.058227211602696629655, -0.18388675005552940211, -0.33022115329525419947, 0.57178834377518417131, -0.29504703133435994378, 0.57148658167881616343, -0.43133757160895852545, 0.0058250845267293575175, 0.090765112953913734684, 0.026018001981385305288, -0.060501517149639590099, -0.15831179126717784267, -0.21299970045319502332, -0.1073763927324880435, 0.28053056171155854592, 0.55785175626330929521, 0.30382633976488832017, 0.12484835841701642789, -1.4371350925386625619, 0.069372443907394942819, 0.12010580424837277724, 0.42600666176217955039, -0.017177891382544566157, -0.01900634802359582598, 0.0018892388171242965097, 0.05582182613917729963, 0.12320237511640548533, 0.10358814100671057123, -0.085732908276265396008, -0.43982037897892178879, -0.42108153608580373151, 0.61760995987371880567, 0.056980685507752586172, 0.60962069309733701861, -0.41439325040072227946, -0.18260162401128457454, -0.010825260805237456319, 0.0085351667073403497676, 0.021834305710212786183, 0.019229961395962447135, -0.020126467783614626628, -0.13659157773106878064, -0.21772901237424802101, 0.090610856071072137796, 0.99846396856417318766, 0.61735531022510925503, -2.7080368640162966187, 0.6959118724978958781, 1.0480065433969170119, -0.40243041334539048082, -0.0078481967833256404549, 0.0068286618030223913967, 0.017742352783519540343, 0.015787981471718601223, -0.019759284994363024435, -0.12964141500473028157, -0.20864000844336624829, 0.0882021855570599006, 0.98636831949121828167, 0.61420442216246273759, -2.690694173272599965, 0.6886130659204974247, 1.0319487008302052899, -0.38860276837653834336, -0.021824851076207754663, -0.039716741350661825816, -0.025171129811667376092, 0.032483544239639568219, 0.10923267652838157538, 0.13171035112936574363, 0.066979531973101247289, -0.23309954148900022264, -0.44830494074669025029, 0.45059610929788240785, -0.14647326290965501094, 0.45124251418437022476, -0.43967767752711545137, 0.10537285042787834788, -0.020614629479223711694, -0.033067849322491578434, -0.0181579947637353252, 0.0314456755152164244, 0.094275451268605287014, 0.11070687143509343187, 0.058452208867336277742, -0.21899258817092559504, -0.41942889716315162607, 0.45371932083543337066, -0.16755593315988309699, 0.45856945519036351744, -0.4187858275413890996, 0.083244979493541246263, 0.025299763461874943959, 0.031172894355220047424, 0.0051176244708781935128, -0.062108129955146730572, -0.14266359277050300047, -0.1840786925655946149, -0.08683311769263823976, 0.3314750797267986826, 1.2631945903443677626, 0.5728164747545989588, -3.3338959341794076607, 0.61146738520048027343, 1.2079012751108271218, -0.2237115367984059966, 0.020353060176577141005, 0.0015955551842270353398, -0.020359151096519662733, -0.031713314879605422614, -0.00010916944438424240095, 0.046250057126172097743, 0.0084226522154901249662, -0.13379913181932620581, 0.064366965501033560071, 0.51835394117757915833, -0.89547865609809418341, 0.46607202315063084308, -0.0038466864249490466399, -0.03856713419738098253, 0.069948137716522168494, 0.014607635501884208562, -0.062352128648550657342, -0.16015572559501051186, -0.25612753574945595902, -0.26531372707810002565, -0.036998865507551902121, 0.58764006380212852054, 1.3189075811699941809, 1.3725925250093131691, -0.41342802325063732205, -4.3663192321403583307, -0.36546544879100761882, 2.5759202930836861967, 0.0019773814617809167341, -0.0040382565108572597568, 0.0032157450070701290926, 0.020925518801528548984, 0.020437808840223507106, -0.036248755039075347684, -0.1553279612701871093, -0.2018574700763701435, 0.097180254153829831454, 0.9858882260138610576, 0.69250445554108130963, -2.6046557315938625443, 0.72070711136006193609, 0.46972246576121351458, 0.0045333342147933679625, 0.0068655041825469945854, -0.00064140150836767390562, -0.018668844872323953066, -0.039409503558185517147, -0.031012965579825427231, 0.055723733605621425524, 0.21015038247392123227, 0.030300290517001598339, -0.49139549049189440444, 0.17806403594056868789, 0.089180657094160661891, 0.12915438239044976787, -0.12514139895491599441, -0.0025469238520851129534, 0.020350693438474492386, 0.027877242847778854057, 0.0049878714595832998602, -0.055994136335218741563, -0.11660451140903189959, -0.10225283654884188833, 0.036834338499647296949, 0.19201471296388705756, 0.57988626442595059274, 0.54769246984676145384, -2.1331707833089730286, 0.55711711252443896214, 0.45525705941724303738, -0.0021851658024490032764, -0.023631364007503367908, -0.021749158536182169182, 0.0090441514994284797485, 0.05813348980158476037, 0.10071872374656831717, 0.072668612258374551405, -0.076266329022867507259, -0.34189926720400748428, -0.26813258055853939066, 0.58301313822148215227, -0.23868418541284178169, 0.68692858202990969563, -0.54720362859118520937, -0.0057770144266528662691, 0.02563849840677462355, 0.036081294704418530317, 0.0079040439648290113095, -0.064480011970015418954, -0.14440997623189300025, -0.1733104994501154239, -0.030307411605998278714, 0.38413969417008586582, 0.95077289014287991176, 0.39433682162386057835, -2.5326198200265181981, 0.3487880827075049428, 0.81939499026276441107, 0.0078361171432828821692, -0.018708840808004344874, -0.028799631877790079354, -0.0057510054224698013356, 0.053345528599884048382, 0.11168078024981736851, 0.092032875159018701083, -0.043705611847138933768, -0.38371589320629045927, -0.42532047067510236893, 0.63655052455698746261, -0.080428105576716660274, 0.73502820167366333948, -0.65808401145810657695, -0.030912857306776793387, -0.063720934141309082399, -0.059603940511407924041, 0.012231609536388491841, 0.18542363248111906615, 0.46756239964961698874, 0.72377703063385079307, 0.56807841270241121734, -0.81020590630696920353, -4.4341302451137751817, -11.542093592820549475, 28.516300288090764781, -11.559888163603362798, -2.0507518779879227999, -0.090455977940674295312, -0.0028313847610757902212, 0.10205049286056792901, 0.18503346537258691384, 0.1599482325252661874, -0.043134253369505895503, -0.34460017328745656817, -0.42079619276079177626, -0.20226468064883781905, 0.20197015465192125205, 0.9272801052069170602, -0.45647528165339140704, 0.50642437729802114443, -0.51914610148563800074, 0.0067343802391588115583, 0.023673501067578334861, 0.027805808865209041131, 0.016778063805135079445, -0.0097707726392503718715, -0.056122253288229659307, -0.13357917276073491553, -0.14617008175168749728, 0.0095011603283291085065, 0.47542375927490793108, 0.66318487933885628127, -1.5606332432266047938, 0.61922013579899859437, 0.081379427288666789098, 0.0030604440370512616364, 0.021852641634300899037, 0.028469684402696034814, 0.019429154961224825993, -0.0078181932031672202177, -0.057182255669138502829, -0.13382244092447181938, -0.14030887547167045226, 0.017870835249330363492, 0.48196363482971776859, 0.65395641211541022386, -1.5797983272185993631, 0.6086935897149322594, 0.099541227275665486451, 0.0048813531848715422876, -0.01249236049672633117, -0.026328698933207866223, -0.02023044733972931622, 0.027824653570907090239, 0.11227571880180639563, 0.12789931629636383525, -0.070173737985103790127, -0.4355439388725785288, -0.10121232033726218491, 0.77870256339299226411, -0.89031546598463895403, 1.0615781120655165548, -0.56448374220223895836, 0.0045583847271728561781, -0.0093433511507772810251, -0.021057347824374978568, -0.018217017868878419146, 0.017011661620652176324, 0.085354350318594340852, 0.1055773775227522443, -0.049333099077277162148, -0.36182677129027185359, -0.034447203630266758079, 0.73995945395150142243, -1.0137732958163365726, 0.99436999044517515589, -0.44450710963952971833, -0.017107642273913061665, 0.0038159935580054770571, 0.022609151742499829923, 0.027002418235165762117, 0.0041714084893387307074, -0.023598432990327256747, -0.00068867131229456766787, 0.060867685674059056333, -0.15877775724069881358, -0.3018953972494466953, 0.57148658049529554859, -0.38108110024999197751, 0.55081687768777964909, -0.3593285725562218591, 0.050035928058581806421, 0.039714177738024103448, 0.0047996332109564877194, -0.060501517396488083134, -0.14671495439252435489, -0.19793307702020326855, -0.092778682832194744878, 0.296299490016092959, 0.57894753998947245943, 0.30749971023775640733, 0.069372444253991860652, -1.6414802204740879255, -0.082867217696766595769, 0.90010463668617468613, -0.0047099564176293576973, -0.014991047895357274833, -0.017108530625165471739, 0.0018892391140993515172, 0.055450542711598015122, 0.12596101952862739926, 0.11133920004167302542, -0.080678496785601211405, -0.45491774631984271959, -0.46777417044910080124, 0.60962069303564603295, 0.16608249954527953829, 0.74839510167231215743, -0.78880786256361168363, -0.0092583874120957618953, 0.00085572756082180536502, 0.012395574008795484655, 0.021834305974519914451, 0.017564534565067888017, -0.024051591325911787778, -0.14504189251894883395, -0.23361692386381127662, 0.074419905428966218186, 1.0048029272427934089, 0.69591187256893849433, -2.5921655472357163497, 0.7563629828791064158, 0.42873713820680098285, -0.0063673798192383493019, 0.00082276737438720902451, 0.0096214876463085359515, 0.017742352173571895868, 0.014787502315270505127, -0.022393604966046906646, -0.13656537808920687915, -0.22396539418841088231, 0.070371919533656679402, 0.98919549663114680893, 0.68861306574652636492, -2.576439303979133566, 0.7586090467251560554, 0.42485982636560865044, 0.002251749839460793351, -0.022329925269222450357, -0.037661212281432314419, -0.025171130695231563185, 0.030702200361841854026, 0.10911504761888288217, 0.13472416480455826071, 0.064458427542267979216, -0.25675048600017957856, -0.49656594538900028502, 0.45124251466667003596, -0.062206792141453519374, 0.55891574427836765437, -0.46572681208158805966, 0.00055719048307103146919, -0.019604691119133408478, -0.030957162354029323093, -0.018157995125477284504, 0.02988386068318229366, 0.093933896677248593376, 0.11265935699627180333, 0.055587283357937257444, -0.23874898150578072187, -0.45865627725382257518, 0.45856945619584332796, -0.098509716911535705197, 0.54265255583915072979, -0.44221558453250159326, 0.00441544655881925905, 0.023263173402047903487, 0.028903732010983906181, 0.0051176253720554400542, -0.061369833378993876583, -0.14529249954912276666, -0.19199072687216037214, -0.094836762694707116284, 0.3400506567224583776, 1.2956481807870399159, 0.61146738475724404793, -3.3631917167831315574, 0.65539497827645398775, 0.91280978710609650406, 0.0086510644504614927358, 0.0093301683229506660044, -0.00072300683767749702492, -0.020359151480539058876, -0.031980601068848438229, 0.0017726942242728388043, 0.054564567201882283043, 0.022071714019042273924, -0.12551659777741056012, 0.038649048939898580923, 0.46607202374204093065, -0.89922268599387766841, 0.61688972789348861525, -0.13778326998706258433, 0.04926962810911886953, 0.041923414174732748227, 0.014607633661531723573, -0.042801178606292497597, -0.13699152274729003986, -0.2529578254002479154, -0.30936478863814675933, -0.14302043178405834944, 0.43327693991662669104, 1.1798639586477168706, 1.2815874971077600986, -0.36546544895409011078, -3.937833856090663609, 2.1076300311951419175, 0.003378391304182118849, -0.0022982987409214021904, -0.004038256465526416511, 0.0054104820387591671949, 0.024974751772382307313, 0.023755662285635158193, -0.039766270106406605145, -0.16499680970441960026, -0.21276587395798149394, 0.095499748722426613945, 1.0010555936992238113, 0.72070711102743167942, -3.0015644827853269483, 1.5532132633001931588, -0.00030142105185648981848, 0.0055035824224360504683, 0.0068655041697299528866, -0.00032063246512169826623, -0.017326610081117750639, -0.037434250550519958256, -0.031451143656178348618, 0.052536661512091097082, 0.2113859545045618793, 0.031776421214897820899, -0.59283635324984107484, 0.12915438215493044161, 0.62453528797542123829, -0.38726067110207673672, -0.0093903206635368857702, 0.0054317020459858107034, 0.020350692748739320631, 0.026661775591402942648, 0.0077533336417653277856, -0.047262593522356635689, -0.10379951952417909578, -0.088637003805674630108, 0.03288428426294709439, 0.13384390062328424542, 0.44534728535254514137, 0.55711711306445665937, -2.0517563782269867545, 1.072463141893012617, 0.011194985057358407265, -0.010051920105463376526, -0.023631363247647797082, -0.02029243220866117775, 0.0051512264631580397672, 0.044395933769467051111, 0.081409546868270954434, 0.067264956395230876551, -0.055196256913229178875, -0.29698791271476232545, -0.22558876291029353323, 0.68692858126726485946, -0.17171077893459393837, -0.093255878923055215979, -0.010827263898976502779, 0.0073108511902281146655, 0.025638498366051663824, 0.032444813606341993406, 0.0083048531249667635612, -0.055102381587301720034, -0.12592231140339929518, -0.15641328594586995093, -0.051699773475928516298, 0.30984085137511491181, 0.86438565005980894096, 0.34878808271732886226, -2.6743467308096455781, 1.4771416245681610047, 0.0086757701511562222707, -0.0043715226651725577409, -0.018708840881552786517, -0.024634986387801153379, -0.0036685630747553140373, 0.049699470877321044737, 0.10010349480968004643, 0.076988084232154668785, -0.042475450417170323991, -0.33520400192796023786, -0.38626442698311230783, 0.73502820163169613199, 0.082425170776244144344, -0.23357985196051928023, 0.0012636412444703567452, -0.034477140037955683605, -0.063720934546619228955, -0.068122799083019339172, -0.013840489689342670709, 0.14224516705761497404, 0.4286385556721311052, 0.74605166501112341049, 0.72667547040140578751, -0.46261637104342839821, -4.0202401388448310016, -11.559888162798776179, 27.004855313614772427, -12.828629402439077012, -0.040632373254680845265, -0.040124801754269565079, -0.0028313842981634463475, 0.072239660129830615687, 0.14867158190716514565, 0.14307421198095734982, -0.015189396190779958645, -0.25339246919071284347, -0.27029053076127906863, -0.050096068189562609207, 0.1131254689513542383, 0.50642437707776633005, -0.74392062825301863427, 0.47083468040337700389, -0.017612446262265483987, 0.0053964546586505041104, 0.023673501478167567613, 0.03239466193094117652, 0.03007872202456625399, 0.014656296531531505523, -0.026528650302762961377, -0.11710911489736947066, -0.16701094736160324028, -0.031712143320654320733, 0.42236455098991276813, 0.61922013573050360691, -1.6949417039395384688, 0.93449391397402992165, -0.017305423496226681607, 0.004157671054521101979, 0.021852642131481990473, 0.031267161078112733208, 0.030012532265774057028, 0.013843985818457409109, -0.029429821671732032951, -0.11655098198929178532, -0.15625079294088695114, -0.018474960074453500997, 0.42746457845912999574, 0.60869358956085906076, -1.7103974733250488782, 0.93538594121943352544, 0.0075709860447519703938, 0.00023202939129367760544, -0.012492360551973776678, -0.026334397412941159405, -0.02320077133125964669, 0.022197567081463899508, 0.10609806483909224595, 0.1222776797141991445, -0.084161770654179068885, -0.45047572813159686422, -0.057195079817710152426, 1.0615781120284881744, -0.87331039910831798068, 0.20646098595747702786, 0.0065726033415336110668, 0.00069312583594211018591, -0.0093433511669566999408, -0.020805710272418256185, -0.020389090084809546877, 0.012305312370797762828, 0.079903584732507942912, 0.1013612648449013498, -0.061109641854927329585, -0.38159249998786132219, -0.013830050703439381379, 0.99436999053318964048, -0.97671207232415058197, 0.28733360520686496153, -0.008561439507636013635, -0.0062715339304182018271, 0.0038159942027059371751, 0.017203291527501250041, 0.019895821240942808406, -0.00064164397915763052814, -0.019625729096811896784, 0.023397888252013686328, 0.10478506734523967969, -0.14338973480417008255, -0.43133757034803454911, 0.55081687672269707967, -0.021250024714097384693, -0.082522493886177786249, -0.00065606756409191356594, 0.026816158867705650048, 0.03971417770337096198, 0.026018001913246390638, -0.024727942954849247564, -0.11043980082833546041, -0.18393021264452105723, -0.12168236798989606751, 0.22479212334006065666, 0.48867849668815288577, 0.12010580398629629173, -0.082867217517524890957, -1.145502314620190365, 0.74787509109089300718, 0.002305349263878321911, -0.0062701530214194934471, -0.01499104785955373588, -0.019006348192308692574, -0.0039370249225038310426, 0.047781614609188720333, 0.12230166766118530375, 0.1162597254467901825, -0.072061997473657699032, -0.42675834661614331367, -0.41439325019792772054, 0.7483951016095664599, 0.26850394731996168884, -0.34255068003521349507, -0.0014371893429409915346, -0.0026073551702945497316, 0.00085572822531665860418, 0.0085351662255441634874, 0.017110925881204389759, 0.016429852000355554126, -0.018328498417864477754, -0.1382278021880454788, -0.25007283032526433919, 0.051416500661804441508, 1.0480065429452791737, 0.75636298299605786433, -3.0522948955590352149, 1.5657927855222599778, -0.00043110997067225150219, -0.0015663733302221697317, 0.00082276747487625666311, 0.0068286623538144214374, 0.014510807866232927968, 0.014642220107977468174, -0.017462177225720694473, -0.13258658091019251413, -0.24340531182016506007, 0.045164739809925573411, 1.0319487008906469416, 0.75860904649620308771, -3.0230652613682051921, 1.5474697282236322682, 0.014767182139724312206, -0.0017055285249805906753, -0.02232992558510005221, -0.039716740641292144764, -0.033103824071898796388, 0.016729446426910883627, 0.089966810089245952131, 0.11270258324972318165, 0.054989946705635527058, -0.20896633517824547699, -0.43967767797994888879, 0.55891574463602211598, 0.22378721375292420781, -0.34262313522164361546, 0.012314559188223599129, -0.0023129525313189923386, -0.019604691114939624269, -0.03306784917135893731, -0.025527344307167850607, 0.017362711036164112793, 0.077344859560806361598, 0.095011362414435238111, 0.05165372220440071771, -0.19207933312502598056, -0.4187858287420707093, 0.54265255660748035194, 0.17939399094627125897, -0.29730625329800214329, -0.0067400032598048421895, 0.0078220189363956242612, 0.023263173994392014748, 0.031172893542695791214, 0.013045316967588609439, -0.048081237287986439144, -0.13080746975615858974, -0.18429450459412402563, -0.12359304157128059298, 0.2503497820286127995, 1.2079012752963216304, 0.65539497812054292591, -3.6693244400728226573, 1.9766154138501972337, -0.0032279171480715470634, 0.006704544136906077402, 0.0093301676338599596333, 0.0015955554947075609607, -0.0149449493738818915, -0.024579868604532876392, 0.0089946197148521624121, 0.060547829138976883012, 0.0048546408086527865697, -0.18571191025165009747, -0.0038466866359350556856, 0.61688972843662170931, -0.78122836649561988587, 0.30802196244270552139, 0.015635131403248920395, 0.049269627455038107755, 0.06994813886401943237, 0.063857668625667718398, 0.011313548211228173501, -0.1106991689490563896, -0.31375860095422358764, -0.54747608189107077603, -0.62135409665356988729, -0.15442901313469278857, 1.1023691001582238691, 2.5759202930157267808, 2.1076300312754367994, -4.4867845729001105681, 0.00014459128405685064446, 0.0033783912790307370037, 0.001977381484391855463, -0.0037552243059789606333, -0.0048561494670420098096, 0.013825249982013539454, 0.046864104046599713849, 0.041415049852828234034, -0.084489851624038714295, -0.31740804617974166524, -0.35977412125262525633, 0.4697224659384747758, 1.5532132632796404881, -1.3921413094966708091, -0.0019586054093747784152, -0.00030142109475015483808, 0.0045333343041964891534, 0.011135020724318437507, 0.012778064294437509532, -0.0014463646404730497088, -0.038922303011386487348, -0.082009723738224446499, -0.057757179665789540324, 0.14171532310759948903, 0.30518686794563931741, -0.12514139883762392946, -0.38726067112078804699, 0.21667253597480939931, -0.0068973285113865029014, -0.0093903207483493882773, -0.0025469234889848209324, 0.016980929965134197029, 0.044552018710717795902, 0.057731651569368533372, 0.017021474503590261484, -0.094581390469223647344, -0.21485267421016715494, -0.20712294034861963854, -0.025798857069866915764, 0.45525705917733516603, 1.0724631419256955844, -1.1328517721774225002, 0.01206688838259611582, 0.011194985180593104018, -0.0021851661822580329894, -0.027378040598096915065, -0.050831048300338889379, -0.047244143768551724594, 0.0043413534592176934376, 0.10319098889611726111, 0.19663612023820087327, 0.16485915978446089025, -0.14755073814468999682, -0.54720362816038170539, -0.093255879098692123774, 0.43847184543391348743, -0.0029122176782545180183, -0.010827263909791533941, -0.0057770144127442107918, 0.017336255910900114086, 0.052302630341585235052, 0.070287587018006053574, 0.02562029590734216794, -0.10346333653798420182, -0.27885992860452646358, -0.36127698086427750823, -0.083271076339649738318, 0.81939499026164852591, 1.477141624579846102, -1.6623713525320642503, -0.0010071342152515351531, 0.0086757701478699777303, 0.0078361171620180657127, -0.0086573911861064292705, -0.036911894164269223872, -0.052105177724145024243, -0.014081854314525200575, 0.092520195196360302536, 0.20466734011279733019, 0.19947026000407652724, -0.12297271681102948937, -0.65808401137809824366, -0.23357985200476535925, 0.63982607481571784902, 0.017410837439484860151, 0.0012636412245603401844, -0.030912857123437709889, -0.080633263573946706293, -0.13416932266716810607, -0.14730071219981633046, -0.036518575291854371567, 0.31044067106951023227, 0.93944287169090068268, 1.6024975488038246674, 1.3305305797048982974, -2.0507518783885414493, -12.828629402358361133, 11.397207747648325338, 0.028438284148740763935, -0.040632373154688726513, -0.090455978177040360921, -0.090654903773968911751, -0.0054138629683179980215, 0.16905589593371522783, 0.34805132841581076786, 0.34750211444462336674, 0.013490706743268309625, -0.49120642933260638685, -0.7538683387919272727, -0.5191461013556968318, 0.47083468035838160759, 0.69126548981820823414, -0.033568872336037440529, -0.017612446091098608175, 0.0067343799699616678153, 0.037059309357008435315, 0.064131851498160946656, 0.075289773594132886059, 0.059810586740766800984, 0.005368735931052127236, -0.11843426988163012481, -0.28804525799331515135, -0.33854343789976193424, 0.081379427354907926673, 0.93449391395144287831, -0.4437834217501135381, -0.028820103284617100414, -0.017305423340987145103, 0.0030604437669739157322, 0.030957834151410911633, 0.058463823784873997225, 0.073215188891765717316, 0.06199743058425395692, 0.0084679306316070134669, -0.11643728739973739639, -0.28187363430774492468, -0.32355286487310580545, 0.099541227346618438299, 0.93538594122287122001, -0.4837367759209451612, 0.0059160765232089407295, 0.0075709860310763064939, 0.004881353185931162561, -0.00677627767840227168, -0.030844970353496571625, -0.056388353465327140102, -0.044605655749624426032, 0.055648286468075423838, 0.21491509134093303435, 0.22315601465906947931, -0.19595221658376332785, -0.5644837421581552217, 0.20646098594981912577, 0.1884293800765415583, 0.0050400519300993927227, 0.0065726033256970602461, 0.0045583847378316321233, -0.0045501207116629173444, -0.023649503795803928502, -0.045228556855787513391, -0.040141063343722221957, 0.034789412615467502077, 0.16458347547113269527, 0.18228094347587514568, -0.1615469399918607607, -0.44450710968024897962, 0.2873336052469404045, 0.037050940369548097941, 0.0050401109987294719431, -0.008561439389201780728, -0.017107642566484101976, -0.013506100664214057069, 0.0072897310510245090856, 0.036370071021769569131, 0.044977330834922002456, 0.014098182157308219303, -0.010708898459032327455, 0.053802355584490091367, 0.0058250840847515741525, -0.35932857204457985256, -0.082522494052099312767, 0.34340333023338476259, -0.050131188736902240755, -0.00065606750802268715573, 0.050035928023364609207, 0.090765113029939503697, 0.10053678362234276522, 0.05035448466159690023, -0.085144583737634740861, -0.2903116644038495342, -0.43618952008887584837, -0.22280643234785646456, 0.42600666179444351966, 0.90010463667167694979, 0.74787509111865535516, -1.3189472230172423206, 0.0053318472306539211117, 0.0023053492968296944443, -0.0047099564520873913459, -0.017177891334119142763, -0.03339205454763125025, -0.039617614993810446045, -0.0026623668043517451434, 0.11022910532184468368, 0.24813272521486812172, 0.22915954777974614776, -0.18260162411473368493, -0.78880786250415313443, -0.34255068007579020373, 0.85067142349794711009, 0.008524705992057351675, -0.0014371891723644375678, -0.0092583878265778529054, -0.010825260545630764272, -0.0015387258122884339001, 0.020151254690868349029, 0.046584469197263618512, 0.048702710207345741023, -0.053526663892669856804, -0.30418107891726720027, -0.40243041316023903109, 0.4287371381536304038, 1.5657927855407069995, -1.3670486408368140996, 0.006915343492452199689, -0.00043110990987524312398, -0.0063673799392355012738, -0.007848196925942419952, -0.0010652432189447581627, 0.01631923059879178059, 0.039247930400842782894, 0.041928166829521766601, -0.05275954708300635565, -0.29156025977682037409, -0.38860276831291157285, 0.42485982647490466713, 1.5474697281655098724, -1.3599544508143599497, 0.019753603367640284322, 0.014767182055253865944, 0.0022517500139856887545, -0.021824851281108108292, -0.057065145397579938868, -0.08498148553086809498, -0.06322456844940015086, 0.043528347649581064505, 0.19788884013625571145, 0.29445230467731720037, 0.10537285063236814797, -0.46572681238055690933, -0.34262313511435932289, 0.34786008952309815934, 0.017564352565721110672, 0.012314559175067575983, 0.00055719045018123468035, -0.020614629458335809176, -0.049882146723114681452, -0.070734644978764360523, -0.048272176542671227373, 0.043998745623876830968, 0.17247339432127159475, 0.25436585615571399854, 0.083244979978579936031, -0.4422155849469799338, -0.29730625321458342691, 0.33893917864620559666, -0.010871961499653940916, -0.0067400031627753327623, 0.0044154462574813460943, 0.025299763713988201314, 0.052617894467945708348, 0.06612366395349458803, 0.02502751400838940099, -0.10064907865116763663, -0.28783083089304389857, -0.42460837727426165022, -0.22371153676216321005, 0.91280978715630489706, 1.9766154138357208137, -2.0612194563334500508, -0.011469107223515835267, -0.0032279172520284517799, 0.0086510648038727998993, 0.020353060055672847484, 0.02237517621711156493, 0.0035362856303498978357, -0.034310541184836652884, -0.049925358689563041625, 0.0036130827173446069915, 0.06578283839011331291, -0.038567134219370760584, -0.13778327026491929419, 0.30802196255179431494, -0.15656945728669036111, -3.6045719351464740043, 1.9499300210022116886, 2.0520461339751734364, 0.72872885428485001391, -0.2621368132353056879, -0.55455426253517448032, -0.41895274541992472139, -0.19955162142960428118, -0.041240010041384569939, 0.03732636510777310318, 0.060411724186973977946, 0.052014850598002579374, 0.028484313835249890773, -0.00049974830499160071383, 0.27291746472244582389, -0.48070430026498972564, -0.11503346564642596972, 0.34265445155623996953, 0.13316670360919602434, -0.064076463032193678449, -0.07436708032919858713, -0.033859072726721520241, -0.0046485438379902056447, 0.0070013942313681973642, 0.0087644480484703728845, 0.0060565040204139915583, 0.00195792638932209195, -0.0021404006987302197189, -0.87992108744332830916, 1.1337512742166580004, 0.08233616351361751351, -0.28560064399950219061, -0.0802554031773340254, 0.0029144672344932590242, 0.00030202148545632094903, -0.0032115147671016980833, -2.6174470206446651902e-05, 0.0028296842773747402089, 0.0044960556066566239111, 0.0040373564901657714443, 0.0016445753123049185879, -0.0016843227219646812135, -1.168943792883298638, 0.53105896277146114048, 0.66504104157356636051, 0.41597453976888149052, -0.022019907815567851123, -0.24137877865133217092, -0.22147883166951842782, -0.11462704482358085367, -0.016324455948224240315, 0.037307654284239960552, 0.04967195447582919976, 0.035949579062876532709, 0.010589472569836787946, -0.017386328992072954541, -1.5198218629288611226, 1.3124694167682631196, 0.74774406533555393128, 0.002686040886831622676, -0.33332681801315156012, -0.3066417572342899911, -0.10287556192073854522, 0.04285077664513525092, 0.075823463323307402351, 0.046494484753417046619, 0.010065181909345808156, -0.0093741724781812047623, -0.010505240130312934813, -2.3193171060232642443e-05, 0.45679661555971268205, -0.23306912784657146664, -0.53005467703107100874, -0.0045457866429779699227, 0.22346383723984478142, 0.16669496364532584987, 0.052104829450200403074, -0.037786614013941871582, -0.063743117150652423009, -0.04085569486957160712, -0.0062703255048494653881, 0.014067349938419378488, 0.014505277318708357029, 0.00056510005160669777701, -1.3265216325438187184, 1.4412914328680257459, 0.50642393998404500177, -0.26338169241055742198, -0.32271336407989542616, -0.17060555410330696668, -0.022409950414946130937, 0.054933113506797082159, 0.061253961784891154674, 0.031312649289160249255, 0.0027346120093738662468, -0.010133424081245743023, -0.0092633392234304830909, -0.00061573390081497581666, 0.69126548973382329066, 0.47083468086815777331, -0.51914610266675265127, -0.75386833721848367507, -0.49120643013696657642, 0.013490706026964463526, 0.34750211582957063472, 0.3480513281082345256, 0.16905589624214242828, -0.0054138648025472812381, -0.090654902515821686459, -0.090455977437858242629, -0.040632373771726730538, 0.028438284014994895388, 11.938922675921885386, -13.108469677748516347, -2.5027115382297440149, 1.176093015100442285, 1.7332236205151194675, 1.1375721885283882617, 0.43423433640934944178, -0.013500244043608883651, -0.18716535634818673439, -0.18761991942624342555, -0.11719495345683207665, -0.042973871691041276577, 0.0085180878508298897567, 0.036168874294003477976, -1.2221536397877168501, 1.3580258339964199621, 0.513003735889808965, -0.20568144455991069219, -0.33136334414190926623, -0.22348878071548813895, -0.077809278288742428353, 0.02452748885944667015, 0.070689307452878186178, 0.068640148880471790749, 0.041058709971652104787, 0.009610746250058596965, -0.013670112548103303951, -0.026767156139120097669, -0.88815827620230203276, 1.0752202783725479662, 0.34002284106486380111, -0.19069971045409955801, -0.27608070130373102424, -0.17599786673999329167, -0.047489414295379846209, 0.034695294029538713276, 0.066102950396944512579, 0.057901108080286214919, 0.030837629076671931039, 0.0035389784293561521097, -0.014276239491363581363, -0.021872986490255857256, 0.76867369787855621688, -0.27791692483749835807, -0.76751408973702683891, -0.19901921857801638249, 0.22818126146046771585, 0.26685220964422862178, 0.12350233806843104178, -0.0052891122848352546681, -0.051657606617384811687, -0.044867446557479040425, -0.02289557930236028882, -0.0047591876094115297124, 0.0057915442052313996155, 0.010381882062277631232, 0.65429739439135981005, -0.32622412326651389902, -0.6133083580644425048, -0.076198817717638075986, 0.20384681498892487372, 0.19150694791129621586, 0.079105085259636304107, -0.010969267675673638127, -0.040193008585835521274, -0.033224149291632004699, -0.016620698110026560501, -0.0030603038374252622758, 0.0048825261352723279412, 0.0082266368913096778182, 0.20666232430880762672, 0.3979379720126483555, -0.43286486041527366675, -0.39450513169730577756, -0.042604191015920153673, 0.13404168266032598367, 0.12771122036465437644, 0.065428654860367932433, 0.017381558112080888051, -0.0092311071244918080358, -0.019286682590281930411, -0.018111769032405387186, -0.0093271204865390654887, 0.0038759205915197170352, -3.3426912947119253872, 2.4291396009083232599, 1.8525738344429174553, 0.24996892644401066952, -0.50038418049806199228, -0.58009699995333063871, -0.36676370778318745369, -0.12306654994455125396, 0.038117475515738248604, 0.10268312434349501083, 0.10065465153995299674, 0.062707031521140335628, 0.010153860434839338811, -0.044491692123297982286, 0.28444581616060193863, 0.15176672731454826648, -0.49036490542867317011, -0.18713155703400596863, 0.11074316542328513369, 0.11108599143193885472, 0.042165068028976154801, 0.011147923485644552904, 0.0077666007799121871552, 0.0026667700757052044229, -0.0070970541626812636352, -0.01254243020484432751, -0.0084667671720273698227, 0.0034291092093977201469, 0.1442377900198045515, 0.027955633533051138517, -0.40233640647733620588, -0.024323542129500573539, 0.21389256174038262515, 0.15041035452652973259, 0.010197529727802006477, -0.067055564820379845448, -0.059620296760995131202, -0.021305612403330304733, 0.0057065299350179858953, 0.013092414143374614399, 0.0089092135330087664918, 0.00043558453076159811979, 0.2263187835666529979, -0.12183341367028174829, -0.39446515387078084913, 0.050083593228200662395, 0.22929286480775865065, 0.13415618630822323532, -0.0017280603434378593274, -0.070957617500298775237, -0.060505767950853106552, -0.02127691561260242753, 0.0066612306166204792585, 0.014691186809689510059, 0.010047606339414009644, -3.0492191312435584869e-05, -1.4104485211212181017, 1.6504644807115480543, 0.39829920319996542366, -0.45013200046118739817, -0.30515757836456786078, -0.036479503085134527829, 0.070183951414144574166, 0.063896305666851740956, 0.023753452499880430004, -0.01078507828875684138, -0.024066402408980523686, -0.018357864813689624139, -0.0023921052699698371231, 0.016660246679067600845, -1.0265501458657224632, 1.3076355650863635827, 0.18773138649684359414, -0.40483801098646510219, -0.20258385462879280325, 0.024401217210269967456, 0.079004952449253554314, 0.045373967033169551599, 0.004977867778877737015, -0.018084734933566867787, -0.021976489816937380684, -0.013452897109990744062, 0.00024229469260720107836, 0.01452732333125432429, 0.0620578111692890802, -0.1844319222926207591, -0.017557087152074529052, 0.22851399942020514389, 0.032818131957983706637, -0.087970687764341012893, -0.057897279598867000538, -0.011043909589914486374, 0.010962842460471236214, 0.013794978773699428584, 0.010215919863085927768, 0.0054050335885343541234, -0.00012357800030477942383, -0.0064129311561595602961, 0.35480830971935706231, 0.027858903853840471382, -0.51787977102536186802, -0.13024687838963910913, 0.1435215533416607514, 0.13162665288134092756, 0.056636449272164407498, -0.0034680584620427703547, -0.024658184707097716842, -0.018329405806662077416, -0.0055212487786957800337, 0.0020325777611170225273, 0.0029596756075299358001, -0.00055058228339580149013, 1.949930020094843508, -3.5670407734896558694, -0.14461006364364142884, 1.1319478736753769699, 0.93492877267677432229, 0.28985045528458985853, -0.17250326342071420349, -0.26795039671794101954, -0.18927775136283511714, -0.08457961854945018898, -0.013711066757097196619, 0.021755713033174577109, 0.032560867498090367844, 0.028484314130190806214, -0.48070430073997694187, 0.82983124049775358611, 0.0349303849439949049, -0.66861882058579813481, 0.089752799737087524123, 0.22614666708116998306, 0.036114441479869488494, -0.036330015889523080419, -0.031821558879302380696, -0.012816919483796228307, -0.0014137443009451374984, 0.0033475495947537354212, 0.0038522147301344746538, 0.0019579272598750081628, 1.133751273164843365, -2.4034967451204902744, 0.98215376298531520405, 0.62918623842206700125, -0.19400888538181923493, -0.13183642389890762314, -0.016174749753750009007, 0.0011389718034685316315, -0.002521731376354756967, -0.00058987403490856778719, 0.00019376643186875581884, 0.001011665520985398619, 0.0017462468607675331149, 0.0016445767151745131675, 0.53105896279993791698, -0.99881852758212730237, 0.063764351392848708788, 0.095032349149490727425, 0.32935218365123036799, 0.19499280446976507974, -0.026201230450811446165, -0.10649936071356995437, -0.093867482483339440003, -0.045648486836316018445, -0.0049117733089723402545, 0.016472970524900192729, 0.019700727909340638722, 0.010589474143380989135, 1.3124694168607944356, -2.387173837692114553, 0.37488205774891486532, 0.62789894242066779384, 0.32485218175173813382, 0.046650886263011517863, -0.16648070982436052878, -0.15258014752555351667, -0.051550569278109026339, 0.018377629414888213555, 0.034717293109844671317, 0.022541786673767529037, 0.0043002227754966801387, -0.010505240860168678363, -0.23306912769567453281, 0.080294560832283076723, 0.66552555409654545748, -0.40924128991318509119, -0.26492156145050804739, 0.022060504545308383395, 0.079567361565795274858, 0.075455610713120291333, 0.031337597904526712234, -0.011887949242650276693, -0.028392779812893498975, -0.020294812478325669619, -0.0028265809827417694503, 0.014505278454032514632, 1.4412914329249770784, -2.7209278362928288608, 0.63023469030507717203, 0.84281419969980597706, 0.11601117447500899904, -0.12657384320028880476, -0.13723874122892987248, -0.080666552317522371984, -0.017008812491371941289, 0.023839452706761344453, 0.028123556333095031667, 0.014511769586810803825, 0.00013816624459846382983, -0.0092633393617386448893, 0.47083468042668108477, -0.74392062885266663219, 0.5064243783631240392, 0.11312546825914146509, -0.050096068408642405179, -0.27029053016785431485, -0.25339246959566724682, -0.015189396376247546114, 0.14307421173590298413, 0.14867158337566288573, 0.072239658935223632907, -0.0028313851886727463281, -0.040124800232283118484, -0.040632373692226775308, -13.108469677777264906, 27.328387899471401568, -11.308241125363053925, -4.1629882910087419035, -0.69323294630336795841, 0.65387308451396319864, 0.79817551928932684024, 0.50986436978167970491, 0.19434232160998055061, -0.0046653344478082238098, -0.086581661047451449753, -0.086937883429335768848, -0.046175420218046804577, 0.008518087278849936772, 1.3580258343077202809, -2.4856700602707109127, 0.58089433046867167132, 0.71722940385327882407, 0.13167272538291430184, -0.10713647088646907624, -0.13791422949917803153, -0.083091442532295750611, -0.028646119259708849314, 0.012651066611823892638, 0.031169894641329940616, 0.028195040966678235062, 0.010516059286537492862, -0.013670113384015221761, 1.0752202783692959009, -1.979960309241997729, 0.54048197693188648749, 0.51320735269442896431, 0.089737455610842481457, -0.089321747583264068937, -0.1167977375099901044, -0.063691502439994482354, -0.016324778799210754726, 0.016011669494258886498, 0.028631007660655162506, 0.023752435165997072553, 0.0071599747379322694826, -0.014276239294079194619, -0.27791692449307681123, 0.12221167296830054427, 0.77538390474318741319, -0.34866030022314470349, -0.41739558427039935573, -0.092929658395138420302, 0.1102442348684673501, 0.12936234203113908259, 0.053425263943052192506, -0.0052698030981484286078, -0.024503094851722740516, -0.019480547980445898915, -0.0070093170065417776879, 0.005791543784621598287, -0.32622412264601319265, 0.29236144587865536826, 0.59278062802539921439, -0.41665234389239602653, -0.29683378503161494955, -0.022784956494923212095, 0.093034225843259293898, 0.09179522754346500224, 0.033346013173208045866, -0.0068918262711633914222, -0.018636649073580351804, -0.014196942392075631231, -0.0048853866954793523023, 0.0048825253540801015831, 0.39793797204120717748, -0.97135575077197156002, 0.78461344448412950747, 0.14924269660073877231, -0.28890932563828802815, -0.1889408930905839834, -0.0017281728983027100366, 0.070184467880750320279, 0.053356691419018990874, 0.025306203665630751243, 0.0053941772445801703523, -0.004521045758088730257, -0.0083981340915004962233, -0.0093271212127151153365, 2.4291396008473102874, -4.2966299559382576945, 0.18790940945116177407, 1.4806072590524881338, 0.71175641297157732357, 0.048770183530968541641, -0.23772089527767695238, -0.2508149655745012252, -0.15031837754202415081, -0.041751139303332004959, 0.02218916109414277929, 0.042725076762731573066, 0.034403886279942015591, 0.010153860864333426675, 0.15176672756684941867, -0.6022715682653897451, 0.85139390668688774699, -0.13844249924395010676, -0.35508657367582086017, -0.042066721706804414971, 0.081277607804531165625, 0.055498135321174646428, 0.010769469283156132486, -0.0010136940684682040764, 0.0020608810725026454648, 0.0029206612454649961627, -0.0016763184255549630052, -0.0084667680187184811424, 0.027955634110652854973, -0.43317481513226524514, 0.75731622547582555516, -0.19534630548346298973, -0.28163420319639348444, -0.013786349238803825498, 0.10094862794901041103, 0.082121610638236475221, 0.0068081267647110005337, -0.032753562364199004242, -0.027997159999753586102, -0.0084205753496383144163, 0.004897780836439590349, 0.0089092126810111152424, -0.12183341356278847012, -0.121756186360693755, 0.643144592329458753, -0.3177600905943276377, -0.23662084777265385283, 0.028518725813390832097, 0.10103334634878752274, 0.072589813179577647451, 0.002357021118572270868, -0.033348365564082840806, -0.02791406818601165768, -0.0086238318041408149051, 0.0050088819158114784857, 0.010047606317054659161, 1.6504644803005235065, -3.260464418668199027, 0.86149144929559207107, 1.1046011816209135858, 0.013643354521602304649, -0.25909490571901844236, -0.13590171808390599795, -0.018718103075324201562, 0.021090223073746652838, 0.023569308305495070832, 0.010071318512312636001, -0.0023938691459556979066, -0.0064890480745943097604, -0.0023921038260562204128, 1.307635564818587337, -2.6667946961201409728, 0.85401356141989470672, 0.85150561737638152948, -0.068319146818906217877, -0.23425550385510335305, -0.089676096821989056962, 0.012473496909577642955, 0.027770335130329216328, 0.015933872275439672772, 0.0014380090636560895895, -0.0061601426349507165323, -0.0058591417340929721758, 0.00024229354255593689698, -0.18443192157236959483, 0.31816072112356452717, 0.092442606129580801033, -0.43501737188773031972, 0.11471805863594666808, 0.15628613622133888383, -0.0077659436270093810256, -0.043946834597921681509, -0.023472773905200899208, -0.0014114337744203695434, 0.0061299730668209675283, 0.0051185911512015907707, 0.0020770972071916383832, -0.00012357871296440745739, 0.027858904305786729577, -0.38763713311923736482, 0.81541408094262368689, -0.24606865187716212007, -0.3189015330553646721, -0.017194237078125063078, 0.062396990128375323659, 0.061657932673780446731, 0.029069394001121469373, -0.00075993863294335824848, -0.011881709620059606805, -0.009154212476126900172, -0.0028403351178200508462, 0.0029596757533073437402, 2.0520461361108943699, -0.14461006755518074351, -3.8548801986424683363, -0.12760702964122866421, 1.2479825468931957655, 1.0521994659347986278, 0.40394251701391103992, -0.10198262843024233359, -0.24620047800473185418, -0.20092131669407997663, -0.10661936341070860079, -0.029346427341271134198, 0.021755713744656447056, 0.052014850567121649616, -0.1150334655200069417, 0.034930384729260390009, 0.24528019528957889506, 0.14351804510014420924, -0.57111478775928381513, 0.068336012589060454725, 0.22142201805760960975, 0.043868196542994716824, -0.031458622975067122851, -0.032857652113152766016, -0.015652762369696704425, -0.0033710372154744307378, 0.0033475511046127644568, 0.0060565024116973169288, 0.082336165475162606531, 0.98215376141128218901, -2.1833496049381384552, 0.75771349159011514018, 0.64770812267319477495, -0.14099058715985710033, -0.1191759286031742332, -0.017638144062546622093, 0.00049352028589348423838, -0.0046179871759810080245, -0.0030617190219729899063, -0.0013391564252323931179, 0.0010116677319620061455, 0.0040373539470906919316, 0.6650410410298819297, 0.063764352212482236504, -1.3962823315979764338, 0.076565073574177286431, 0.18108089683571185202, 0.39495656082888175309, 0.25695586500572942246, 0.0068176090995044043644, -0.10188034040321612694, -0.10822116708879817026, -0.064512167269320810914, -0.017514437337352469631, 0.016472972023141286418, 0.035949577617893961612, 0.74774406504033463783, 0.37488205857328521509, -2.435735424808061822, 0.46640574483344049428, 0.78634671768049280161, 0.36512671311428179788, 0.043070683995139406874, -0.18001001690796880617, -0.16561109165216872374, -0.058827922714067636123, 0.018970558604133654773, 0.038954619026910816681, 0.022541785193651700525, -0.0093741707198720033989, -0.53005467757194724143, 0.66552555471714258672, -0.17044407587801591042, 0.5653143173356465212, -0.42820821052873209167, -0.31089225683793325627, 0.023664407577160942453, 0.10373543895933448611, 0.094702196966014881574, 0.036793055905346684631, -0.01651207836913363633, -0.03552462625668528079, -0.020294811217542219756, 0.014067348560472463931, 0.50642393916648908991, 0.63023469075573601561, -2.475676704653116289, 0.61543156077519189928, 0.90699320610617051308, 0.1661670708856973766, -0.12568115051289585615, -0.1567544301559215536, -0.093356923391527790734, -0.021798133183884786657, 0.025008994188596082864, 0.031588196758417605436, 0.014511769233496818118, -0.010133423630655374187, -0.51914610152282314459, 0.50642437730673417473, -0.45647528094115008557, 0.92728010408704963918, 0.20197015516726180007, -0.20226468052063587, -0.42079619323577938106, -0.34460017313846247333, -0.043134253694464587403, 0.15994823372202485179, 0.18503346481547797153, 0.10205049238895957942, -0.0028313846538712923381, -0.09045597774839789218, -2.5027115380515216891, -11.30824112509011492, 28.93618078841338459, -11.359234478763026388, -4.6463801822071459924, -1.0964654865562852848, 0.46682909571180153385, 0.77658630747668877525, 0.563949748503403403, 0.25330723878154970397, 0.03154504465181386802, -0.074163063815994481609, -0.08693788474825156487, -0.042973870473404889025, 0.51300373586052538943, 0.58089433078621321993, -2.3529705198480961315, 0.58813189731436388286, 0.79874425475573718192, 0.20655810368054022819, -0.072008021659738477571, -0.14762914044939978719, -0.10890021797021258443, -0.051560783893702347158, 3.0631524962410061619e-05, 0.027234082630789795132, 0.02819504040580429749, 0.0096107462949056258333, 0.34002284106200564295, 0.54048197693101429628, -1.910307901776127304, 0.57171895608675549205, 0.58901537599842601622, 0.1419578697164790293, -0.066645675526008640865, -0.12967122432606628668, -0.087734202160012930705, -0.035225432415970238387, 0.0073868206039654077388, 0.027113398256075296744, 0.023752435460314091226, 0.0035389780653420511297, -0.76751409041443108272, 0.77538390508187793326, 0.060568947484718513308, 0.632890469654152521, -0.42482521384649118978, -0.45450346160528870554, -0.10019877604202850896, 0.10999135348973709814, 0.13747390037074264102, 0.06389947613935885451, 0.0020077634987078962422, -0.022178423470135814888, -0.019480548726602743559, -0.0047591862312742006305, -0.613308359275573034, 0.59278062903171802045, 0.11722800145784363635, 0.51089514691297521765, -0.45609012003498766719, -0.32972978900112032141, -0.029533981811495114977, 0.093325582580108798481, 0.09900804443459744475, 0.041890021103187062568, -0.0014246213108947414216, -0.017130179460703776062, -0.014196943551042373252, -0.0030603022930086416428, -0.43286485975867472487, 0.7846134444054155832, -0.79113732520160318096, 0.7155104903538094252, 0.11440248671077556109, -0.30007321112494655946, -0.20687313423571349058, -0.025148186418173023443, 0.060557170345731783923, 0.055995376265648370839, 0.033798301017618152808, 0.012448687989424217851, -0.0045210467004089503751, -0.018111768247786052405, 1.8525738348582250214, 0.18790940941789854857, -4.225874923214337997, 0.17652668156416378986, 1.6094173966250080454, 0.8831211691480841397, 0.16976267588726556834, -0.19176935469679939894, -0.25886747819673555648, -0.18779120479973301916, -0.082191626394359812435, -0.0027508106356075199758, 0.042725076698055051982, 0.06270703174556223336, -0.49036490595980347651, 0.85139390699902084414, -0.68343280439460829534, 0.6852095614212645458, -0.10541724257001180987, -0.33036269974731302668, -0.053088080343164062436, 0.058693909023452728313, 0.044808591709043908702, 0.012763854105243841244, 0.0070728641978309962635, 0.0088489606349855670642, 0.0029206600910163067406, -0.012542429054335802255, -0.40233640805475862079, 0.75731622690351574345, -0.59562443259501351989, 0.57486661459877896885, -0.23366746324277457014, -0.27612782118550904809, 0.016514036874399166954, 0.11789126608223236825, 0.087528514922379258012, 0.0076028173739157399458, -0.034207145508565607595, -0.030683113047164020659, -0.0084205775903717951236, 0.013092416583619549339, -0.39446515425736655702, 0.6431445929092463043, -0.37034924358565451952, 0.49979873535904878468, -0.34434981571174899662, -0.23969739681211729709, 0.056698040948669953254, 0.12051737902850770401, 0.08009399670510580771, 0.0033993089810814419716, -0.035836008616174128139, -0.031638380838245064308, -0.0086238323501843412722, 0.014691187190806939342, 0.39829920384746275719, 0.86149144879331129854, -2.7592424797251471347, 0.77599768914407685205, 1.0697785857200234894, 0.032749263353918534047, -0.25834272678292502512, -0.14980872038999581974, -0.022260701690945780551, 0.027605159179160190797, 0.033666156778511785452, 0.017523752235611110561, -0.0023938665884244486257, -0.018357867883129291853, 0.18773138679457471167, 0.85401356100431613516, -2.2605457492491063398, 0.75073313243378547721, 0.80161768836058244769, -0.059449298924529152266, -0.23061408946551836818, -0.09802883024771416598, 0.011742348624255816852, 0.034364945602167670802, 0.024436790552906716839, 0.0070898477974237206212, -0.0061601435088424680298, -0.013452896341149414919, -0.017557087891853152828, 0.092442607106067400946, -0.12825534268067945165, 0.21595738263794439504, -0.3465214096296068691, 0.098056884595395868431, 0.15292140473580986826, -0.0041189509195136229242, -0.04245094261333467317, -0.02654817251994552596, -0.006201722883914726793, 0.0027369949783351971151, 0.0051185906376179732963, 0.0054050336279652336918, -0.51787977140672936116, 0.81541408095881662277, -0.5413138267860873043, 0.68674959030967319684, -0.22590326558872522722, -0.33883314074649906189, -0.024827560037480780886, 0.065979331663154353049, 0.065890746153226473214, 0.031137706224346091421, -0.00066627505836096528446, -0.013200557136325768298, -0.0091542112925801916168, 0.0020325765870121896388, 0.72872885299004097615, 1.1319478775417255445, -0.12760703209811530279, -3.6895396099080075558, -0.10810042759460924966, 1.2337482694893449064, 1.0517826439001198846, 0.41887923545076066745, -0.091681454423492389494, -0.24632321337717724696, -0.20615198754030264472, -0.10661936185388727616, -0.013711067135012848517, 0.060411723435011388528, 0.34265445276158640331, -0.66861882150893536814, 0.14351804456938740961, 0.35073267062309843034, 0.17062520097250521989, -0.57663280257615701085, 0.052437266581436617341, 0.21570626343624027532, 0.0436562828407343409, -0.030901952847215190257, -0.033077851643169037255, -0.015652763548324063331, -0.0014137452719026708404, 0.0087644497042211853832, -0.28560064506917298699, 0.62918623881137480947, 0.75771349239957763722, -2.2385670696558186421, 0.74562617806334907655, 0.6618877528879082428, -0.12885482799054656655, -0.11448097098371549929, -0.018662716175931737261, -0.0014766834723270118226, -0.0058162459643732724798, -0.0030617203190965026731, 0.00019376617267408106243, 0.0044960567728103148846, 0.41597454136715594775, 0.095032346842812767673, 0.076565074198664873073, -1.2939121095922527793, 0.12855794790602562139, 0.19628919895537019769, 0.38853525491174295547, 0.25050614551657157669, 0.00029446466229565357309, -0.10856739575462820424, -0.11372797196123764218, -0.06451216752708362201, -0.0049117706406574300829, 0.049671952066340320509, 0.0026860412708794548405, 0.62789894150108860327, 0.46640574568886661311, -2.3122367739094307204, 0.4242414756186459468, 0.72619430365284620699, 0.33859546632971737612, 0.049027361922857094767, -0.16480803155988252295, -0.15470129923482642265, -0.054544007184793207033, 0.018970559151252182339, 0.034717293602760469773, 0.010065180870685625988, -0.0045457860771238268668, -0.40924129041348966673, 0.56531431727587710956, -0.23196772083082431437, 0.55465499868862944144, -0.38714385517165278427, -0.28078975399359851206, 0.030910544595013134117, 0.095201260063041845405, 0.08319194826177145341, 0.029780630839394677672, -0.016512079056230814095, -0.028392779266485684148, -0.0062703256243535291509, -0.26338169088929980211, 0.84281419881785002435, 0.61543155973156837835, -2.471889737002676668, 0.57104649433084653243, 0.87237735713694020134, 0.15513220023948634529, -0.11743125324959746703, -0.14614621949834949444, -0.086864748056507665175, -0.019394561280990579161, 0.025008994439854582981, 0.028123556310290155624, 0.0027346118616686822256, -0.75386833861033830662, 0.11312546882761985445, 0.92728010487820877739, -0.47387939398344552933, 0.71718629031618275071, 0.06497368796268042368, -0.23394351478719771498, -0.39176669818339698859, -0.30785782337984873847, -0.017857593842234769504, 0.17361895414109640434, 0.1850334657055954235, 0.072239659206727757645, -0.090654903254062280116, 1.1760930136234035714, -4.1629882906526640696, -11.359234477384815065, 29.354870434379481736, -11.152756106708480388, -4.6165914077837388874, -1.1351460172995522235, 0.42773852494378944433, 0.75455023461848602651, 0.55912744097641553509, 0.25656354879259746227, 0.031545044566343383163, -0.086581660913430905091, -0.11719495358006946917, -0.20568144434588653269, 0.71722940368152621193, 0.5881318973365382563, -2.3270977840748834353, 0.57567494606575764227, 0.79262907992782383193, 0.19633794621406586156, -0.080689924045313998757, -0.15347834711339133906, -0.11148189537265794369, -0.052797587172676276046, 3.0630573890293166794e-05, 0.031169893348135233047, 0.041058711800779086609, -0.19069971059771631539, 0.51320735308778786532, 0.57171895567976771524, -1.8878033610647999563, 0.55542924056819786838, 0.58123341354440649642, 0.13071351386689333829, -0.075030441262524810719, -0.13414425616369962668, -0.088582709160317546493, -0.035114500048025007461, 0.0073868212944456987379, 0.028631007614195739391, 0.030837628671344437248, -0.19901921896427365866, -0.34866029955898020232, 0.63289046915807811722, -0.073529277958890512656, 0.62871318500281436759, -0.3761122857642989481, -0.43225163225462787109, -0.1004025517844266413, 0.10349571005670944734, 0.13465818543925403561, 0.064227614034184046798, 0.0020077648347251162743, -0.024503094806658777538, -0.022895580134963611912, -0.076198816817912420896, -0.41665234415646335009, 0.51089514593826879096, 0.035685276482694594513, 0.51435064564336563375, -0.42023552981257167316, -0.31727618568837145929, -0.030925829557759302452, 0.089025417646544396932, 0.097436771555395679245, 0.042171385545967859931, -0.0014246200989439905104, -0.018636649025819879144, -0.016620699027678997178, -0.39450513330575426307, 0.14924269753252666915, 0.71551049141735878489, -0.900493498506512835, 0.68061176443143367454, 0.12504135387641790444, -0.29367920440835570961, -0.206316386955103237, -0.025862384073036533833, 0.063203109523457964847, 0.059669319289841621368, 0.033798301137031167007, 0.0053941765404105597101, -0.019286682117007427489, 0.24996892551624919077, 1.480607259224166139, 0.17652668231665447074, -4.0977734193617294522, 0.20122896701297954269, 1.6059776058318329106, 0.8822161470895025781, 0.17134022279131572164, -0.1995785048618122226, -0.27414880159670407256, -0.20138489158936143109, -0.082191626643388429607, 0.022189161399707675593, 0.10065465160284260693, -0.18713155646052936909, -0.13844249942510283846, 0.68520956080935868648, -0.71341643992598113844, 0.71114059128218087835, -0.083910041305754745622, -0.34030865308272734149, -0.066729582661977482339, 0.056076031805384715656, 0.051034160340144345269, 0.018771201522675348694, 0.0070728646137254991277, 0.0020608810716630471087, -0.0070970544483813938555, -0.024323540469580469392, -0.19534630666758365236, 0.57486661417775586358, -0.6789000646975020814, 0.60006683209053668371, -0.17060684318644200874, -0.25535793351394031614, 0.0078647883252575753743, 0.10557424632560274969, 0.082459394187029022572, 0.0068876718253834297301, -0.034207144518018900436, -0.027997158174553230664, 0.0057065273616226271092, 0.05008359352292562211, -0.317760091258084576, 0.49979873613348685391, -0.43206726262255096271, 0.52695243426316684676, -0.28550014624889985537, -0.22111404796097441916, 0.048681780270701489888, 0.10882061997606007342, 0.074457174588041158403, 0.0018439137740025313139, -0.035836009652817399362, -0.027914067894140800336, 0.006661231216706273682, -0.45013200063843394894, 1.1046011813462361939, 0.77599768990604967023, -2.8754893033712476402, 0.68743359067635234627, 1.051302580674047249, 0.046156741580616791887, -0.2381802717271971348, -0.13580967587382197781, -0.014297569453438928322, 0.031420214671151698582, 0.033666155245011292196, 0.010071318191608662207, -0.024066400919599523212, -0.40483801043244266316, 0.85150561684347125802, 0.75073313252620488267, -2.3876330528829852184, 0.67756132274439229146, 0.80023473447168369521, -0.041311951780015593316, -0.21417131928035634814, -0.089163324683271760751, 0.016500405431954003055, 0.036946469021727802728, 0.024436790316163716635, 0.0014380086892670090881, -0.021976489309366385033, 0.2285139991603729881, -0.43501737135451040706, 0.21595738234064973393, -0.043096747287117799508, 0.22431746790190479346, -0.35765892603680099482, 0.084387867724312740081, 0.15082038803049341302, -0.003339257715799534218, -0.043620456063735615904, -0.028701602080316763327, -0.0062017233118526722893, 0.0061299721209542778685, 0.010215920990438821098, -0.13024687912667240464, -0.24606865095577365454, 0.68674958972817279346, -0.58048124452388483174, 0.67976688771778426279, -0.1972358310581752705, -0.32426567940732575757, -0.025991992717819639019, 0.059574928203202863475, 0.061300065291091707098, 0.029258858870851535416, -0.00066627508671270919711, -0.011881711033293806609, -0.0055212473394082731484, -0.26213681307744451932, 0.93492877238848037802, 1.2479825471992769259, -0.10810042811944933772, -3.7122843841078969618, -0.15024404447762751413, 1.2041067756353036611, 1.0392850299597657937, 0.41714816695528872392, -0.091314469489066157903, -0.24632321234231391305, -0.20092131664158741078, -0.084579617744200100149, 0.037326363882140820016, 0.13316670264137009383, 0.08975280023561862941, -0.57111478677723670483, 0.17062520059217736734, 0.35906844620786093625, 0.16975329201520289457, -0.58246991959873062328, 0.046393343667727883484, 0.21347578341855530382, 0.043551367102601108716, -0.030901952980323950382, -0.032857651085088854237, -0.01281691849350270751, 0.0070013927074134901449, -0.080255402735466177533, -0.19400888585087508464, 0.64770812320133075435, 0.74562617719116752735, -2.2553507113817614638, 0.73740850769814458854, 0.66442669751388294674, -0.12702641509034096057, -0.11474865512091289577, -0.019363402038086837548, -0.0014766833670561429776, -0.004617986962952257135, -0.00058987433756962394382, 0.0028296844133841258205, -0.022019909210758938783, 0.32935218550388828396, 0.18108089601845625549, 0.12855794855859922765, -1.2776697819402071943, 0.11289897839957126202, 0.17359216849759542822, 0.37070399392339764555, 0.24156954598020288771, -0.0028719081330398250895, -0.10856739692378121476, -0.10822116640092468987, -0.045648489414362773686, 0.03730765669606058571, -0.33332681842334838107, 0.32485218123699471926, 0.78634671825580926274, 0.42424147667021389063, -2.392602738733439427, 0.37061271083805530369, 0.71370977891175668706, 0.34981149453651638748, 0.065023774975720927127, -0.15586146054186686416, -0.1547013000822320683, -0.058827923176729038368, 0.01837763047206756109, 0.046494484500608092292, 0.22346383725256027675, -0.26492156179065773358, -0.42820821039801065799, 0.55465499909521631672, -0.19350113156212359056, 0.59674399314367065905, -0.37005610792992982283, -0.28796947243843623587, 0.016200074595701306063, 0.085472066322459516741, 0.083191947839624441619, 0.036793056287727464448, -0.01188794927459915253, -0.040855694942747017295, -0.32271336485658885795, 0.11601117499624903628, 0.90699320662085536338, 0.57104649413054264784, -2.5366450062353540496, 0.52936487135936083348, 0.8685188268985329918, 0.16744798116478706862, -0.10533550951133417717, -0.14045887911043367091, -0.086864747732204039554, -0.021798133168622207262, 0.023839453232494797752, 0.031312648699860494428 };
	std::vector<double> l3 = { -0.49120642921255469604 , -0.050096068048959288554, 0.20197015390721639871, 0.71718628959265495748, -0.63352124919745156362, 0.67494961846391055005, 0.10139102393832312432, -0.17943409986566016823, -0.35292498974184016713, -0.29155906510279272492, -0.017857593092094222603, 0.15994823305000319125, 0.14867158157787321859, -0.0054138629738383218029, 1.7332236225339807678, -0.69323294767498122582, -4.6463801831814173227, -11.152756106720627116, 29.672367786824505487, -10.969144259792820151, -4.5660403718585467203, -1.1556156070981882245, 0.39516783735868316052, 0.73834974636211714305, 0.55912744069957698301, 0.25330723896526990968, -0.00466533365461772008, -0.18761992019555567834, -0.33136334484647378407, 0.13167272584914532785, 0.79874425461227616108, 0.57567494695360821311, -2.3609374833195704291, 0.53572585616160095512, 0.77015995767358669255, 0.19077009176783679689, -0.077674711183481057186, -0.15059596025084473481, -0.11148189613844375345, -0.051560783784617293668, 0.012651066534728382107, 0.068640149164176961216, -0.27608070124230865705, 0.089737455123775788923, 0.58901537682344262681, 0.55542924012946859325, -1.9186282392050180157, 0.52117910632163721996, 0.56277200492591972214, 0.12730468942801465015, -0.070879094971564471361, -0.13064567361561935099, -0.088582708677328383939, -0.035225434469736215115, 0.016011669360228866776, 0.057901109477429574257, 0.22818126254059628577, -0.41739558476041244139, -0.42482521419276586538, 0.62871318419221045026, -0.032182244684826090408, 0.66301985392710627298, -0.36030357592868611505, -0.43238740138211195774, -0.10567087097915132954, 0.10011788655128162073, 0.13465818647511243467, 0.063899476111312303317, -0.0052698028619255592175, -0.044867447205606147964, 0.20384681422473005519, -0.29683378471271298915, -0.45609011919179437422, 0.51435064524525675456, 0.069174167830460667195, 0.53943289607846311995, -0.40934105503844686602, -0.31794510030945649648, -0.035007274436523259931, 0.086465543547753834219, 0.097436771734825206037, 0.041890020145635413484, -0.0068918268147098527346, -0.03322414815073718164, -0.042604189786882150359, -0.28890932720450340776, 0.11440248705302927257, 0.68061176453236016393, -0.91021690376000130573, 0.68132410415230293399, 0.13397151098234405531, -0.2858236698155374178, -0.20104461156353414797, -0.023087276572745139835, 0.063203109505306581672, 0.055995375311092616955, 0.025306205203588186259, -0.009231107851934344799, -0.50038418073556079069, 0.71175641323261396032, 1.6094173966382461227, 0.20122896707927095949, -4.1303144814699495058, 0.14460880930410804912, 1.5653296638669778051, 0.85978801698497309136, 0.16027729362457318607, -0.20530271420976128893, -0.27414880116693779444, -0.1877912036145715502, -0.041751138446608451749, 0.10268312250767334948, 0.11074316475488980849, -0.35508657307951657645, -0.10541724228764930149, 0.71114059113059002737, -0.70675322081875857627, 0.70382539573346236672, -0.087996907827122120005, -0.33871130637626506443, -0.061131407610141220987, 0.060861986069276327593, 0.051034160314558499327, 0.012763853983412424761, -0.0010136945468491903762, 0.0026667706268321900512, 0.21389256081128835141, -0.2816342031623465525, -0.23366746272011887364, 0.60006683318686915296, -0.62401187801564805824, 0.62649565917703575746, -0.17144593842345529078, -0.26659269005147706588, -0.00065121748414712771202, 0.10231359753953411795, 0.082459393034118783072, 0.0076028165173139920785, -0.032753563007559757803, -0.021305610668680145109, 0.22929286507668508732, -0.23662084786412102, -0.34434981704696920879, 0.52695243590275708812, -0.37659942434438153924, 0.55442451826256533565, -0.28613719064456671903, -0.23295426234077498684, 0.039168736386143460093, 0.1047921799973360879, 0.074457174298618228336, 0.0033993104209542567007, -0.033348364686051699957, -0.021276917473746971565, -0.30515757810617649692, 0.013643354869589160994, 1.0697785846799399145, 0.68743359094254508079, -2.9321922277574166138, 0.66601943867677471101, 1.0631082774541935621, 0.066130765105817904126, -0.22444076265911205859, -0.13052657660149960539, -0.014297569285266750749, 0.027605159495784677842, 0.023569307735487862732, -0.010785078111503967543, -0.20258385482404714373, -0.068319146121508994574, 0.80161768755302464928, 0.67756132305665872639, -2.4218031606593219962, 0.66921124422128441811, 0.81273706617313612455, -0.026426138210810315443, -0.20527518574273301288, -0.086160968546014016223, 0.016500405008587765154, 0.03436494656538409409, 0.015933871390844570742, -0.018084734684121445136, 0.032818131857518945127, 0.11471805779913342438, -0.34652140824506744154, 0.22431746755537537719, -0.046994105369482674917, 0.21803364034343594713, -0.36210368204724024954, 0.080982995599868878855, 0.14876801503307102159, -0.004653374894881361773, -0.043620456164421214651, -0.026548173377201108453, -0.0014114332485015731837, 0.013794978938005478253, 0.14352155446737521749, -0.31890153285029781971, -0.22590326633720228533, 0.67976688606334456644, -0.56258474406362146691, 0.70040475014661174313, -0.1889504977350256465, -0.32741456339871716219, -0.031852823163236322623, 0.05595271781198805916, 0.061300066408392708817, 0.031137706644691412472, -0.00075993879072173823631, -0.018329406360072374171, -0.5545542639982633526, 0.28985045594178893147, 1.0521994663365685785, 1.2337482710732183833, -0.15024404571974855904, -3.7522887592960341152, -0.18168716715300492481, 1.1867892692965464896, 1.0340050815304140386, 0.41714816798786269469, -0.091681455400362543706, -0.24620047997413918583, -0.1892777524466605632, -0.041240007188129493287, -0.064076463443943867948, 0.22614666749306380766, 0.068336012199007367673, -0.57663280210814549243, 0.16975329223960097802, 0.35990733340285580022, 0.16818636915888707239, -0.584843484906563571, 0.045071215800031511067, 0.21347578399419422923, 0.043656282166449532356, -0.03145862247653711391, -0.031821559523935699643, -0.0046485434735372170231, 0.0029144665067996895595, -0.13183642312809640273, -0.14099058799279379217, 0.66188775430295798685, 0.73740850740067553915, -2.2631352498328993761, 0.73285675358244217747, 0.66303741844901775693, -0.12724750696600498068, -0.11474865438203611656, -0.018662717206149599669, 0.00049352159646825306469, -0.0025217316714386769395, -2.6174880540165637322e-05, -0.24137877751613576183, 0.19499280430585499979, 0.39495655997816814242, 0.19628919795326371656, 0.11289897931813575038, -1.3015099689999978239, 0.090366955860050154903, 0.15918003404755248775, 0.36495121651127665485, 0.24156954508340588994, 0.00029446539565077325951, -0.10188034054444192189, -0.093867483923877295671, -0.01632445472442145204, -0.30664175751525618585, 0.046650887303465331246, 0.36512671221965831148, 0.72619430331524803535, 0.37061271165353942836, -2.4145202052344667543, 0.37044653424707890688, 0.72278524584965730959, 0.35749698838651550181, 0.065023774376353399296, -0.16480803112994785176, -0.16561109047577973752, -0.051550570219868245558, 0.075823463206482297205, 0.16669496319463505962, 0.022060504979061010961, -0.31089225653383245618, -0.38714385569430076472, 0.59674399357362495344, -0.1742605894374339992, 0.59443845065722311638, -0.38144860036961009131, -0.29714742726804793671, 0.016200074372350475177, 0.095201260018581326805, 0.094702196894759518853, 0.031337597096050433254, -0.063743116264009025418, -0.17060555408039013159, -0.12657384312163638684, 0.1661670708391617679, 0.8723773571353624634, 0.52936487118286956743, -2.5530342465509594518, 0.53030634924774000538, 0.87600352573801831824, 0.17317923495340614704, -0.10533550908341711427, -0.14614622015634742835, -0.093356923130001823186, -0.017008812906350166683, 0.061253962184867311036, 0.013490706732016286007, -0.27029053100471756022, -0.20226468022442214978, 0.064973688010317637298, 0.67494961789432328381, -0.604636936124795632, 0.72487435851325043856, 0.14181993533375616967, -0.15938904646387239139, -0.35292498958207718651, -0.30785782358883273524, -0.043134253966595502428, 0.14307421262566483167, 0.1690558956703494542, 1.1375721886104104286, 0.65387308472373395496, -1.0964654866995378057, -4.6165914080022432131, -10.96914425994225617, 29.824746762399616529, -10.906430210557413574, -4.5646449521113581227, -1.1697072544133810812, 0.39516783710524894246, 0.75455023475765092833, 0.56394974792120733387, 0.1943423215353652922, -0.18716535586102053812, -0.22348878077886505378, -0.10713647156353652001, 0.20655810534449711713, 0.79262907842120489121, 0.53572585682919049432, -2.3950260476459037662, 0.51927360834421321112, 0.76850536972310035111, 0.19360585511520903013, -0.0776747123150733354, -0.15347834519775160889, -0.10890021809347334714, -0.028646116845710046378, 0.070689304639675265562, -0.17599786567148500982, -0.089321747951326344039, 0.14195786830552167834, 0.58123341436828024342, 0.52117910580264248388, -1.94634812233060428, 0.50935662577383644667, 0.5635064389952917363, 0.13096655053765182686, -0.070879094758265370757, -0.1341442569723371725, -0.087734200712786300502, -0.016324778803150696849, 0.06610294968010900496, 0.2668522093012360985, -0.092929658713006485504, -0.45450346090641996488, -0.37611228523625650499, 0.66301985327562540373, -0.015271564312857215454, 0.66887138313487704, -0.3600774681724949744, -0.43418555700946909193, -0.10567087008572190188, 0.10349570941878061259, 0.13747389996502207277, 0.053425264232511801565, -0.051657606496899252657, 0.19150694750898003549, -0.022784956549748096311, -0.32972978853010287414, -0.42023552938989616745, 0.53943289569216468227, 0.082440216246645256493, 0.54437273137661057465, -0.40923061189818660255, -0.31956545975748279487, -0.035007274307375046973, 0.089025417750413879392, 0.099008045417092044316, 0.033346014690943788505, -0.040193010779668400523, 0.1340416826258752081, -0.18894089284267548412, -0.30007321178594892341, 0.12504135428091076032, 0.68132410432538959011, -0.90667745991546422335, 0.68845902433014893873, 0.14199411759972133118, -0.28113992018598521483, -0.2010446113962822412, -0.025862384535936980973, 0.060557171057520127466, 0.053356690921041817488, 0.017381558254892747611, -0.58009699890580102011, 0.048770182866000613697, 0.88312116882934244, 1.6059776054022321112, 0.1446088094876894492, -4.1883576557208188618, 0.099702686126840578407, 1.5399155808746387564, 0.85048610724033202324, 0.16027729409844557185, -0.19957850502867241493, -0.25886747926360054173, -0.15031837847544474163, 0.038117477041746568311, 0.11108599255522934313, -0.042066722752366302041, -0.33036269973692128366, -0.083910041778964494785, 0.70382539605126315241, -0.71257322778806875263, 0.70795434748212349163, -0.079362497534497658536, -0.33302742180415639828, -0.061131407868408850492, 0.056076032312979424976, 0.044808591422299939855, 0.010769470664872984531, 0.0077665995412660336281, 0.15041035427445736961, -0.013786348130925619657, -0.27612782212124470371, -0.17060684375351564679, 0.62649566013844670831, -0.62498900685963099111, 0.61850964658766094395, -0.17714677263020583986, -0.26947561613161158389, -0.00065121853554778175166, 0.10557424764345713086, 0.087528515782103771814, 0.0068081270380694770133, -0.059620298095510407266, 0.13415618560830244732, 0.028518726161695942545, -0.23969739541093509905, -0.28550014782329441454, 0.55442451892936384095, -0.37568848635896201493, 0.54669599582589889497, -0.29297526692134667758, -0.23685477767923274195, 0.039168735278001007294, 0.10882062122702294837, 0.080093996368617248471, 0.0023570209862121827005, -0.060505767957103800958, -0.036479503781467512158, -0.25909490542026775151, 0.032749263873479136921, 1.0513025809524405574, 0.66601943820619258219, -2.9360744842104407049, 0.67359321196627031103, 1.0734880599516329891, 0.072367061973643773132, -0.22444076289662176493, -0.13580967620297179632, -0.022260701164884228592, 0.021090222437970672364, 0.023753452958282454677, 0.02440121623235143658, -0.23425550352326390824, -0.059449297732514422288, 0.80023473375737119717, 0.6692112447583877799, -2.4220468197495650386, 0.67500245482065879266, 0.81992299048023475905, -0.022341962927077511913, -0.20527518625140220299, -0.089163323652653225793, 0.011742347146655538609, 0.027770336058063695844, 0.004977867684046968777, -0.087970686971428846412, 0.1562861358789311983, 0.098056883634178401699, -0.3576589254706755705, 0.21803364007088019139, -0.047909991451239225535, 0.21634403780083658053, -0.36505101777530807317, 0.078792151503389293299, 0.14876801517624679838, -0.0033392580491052826705, -0.042450942325598137184, -0.023472773723301764565, 0.010962842288350090753, 0.13162665278457894025, -0.017194238020524919147, -0.33883313981898810496, -0.19723582989301441382, 0.70040474876856739783, -0.55480151862606930102, 0.70041005122631871949, -0.19174072019363011532, -0.33016409668141555311, -0.031852821556697039074, 0.059574926709170299144, 0.065890745034629133459, 0.029069393903446982819, -0.024658183436620420559, -0.41895274389223846612, -0.17250326417954919833, 0.40394251601064340429, 1.0517826429995495996, 1.2041067772872613428, -0.1816871679876328749, -3.7750961137281926483, -0.19251880551861411117, 1.1867892696619273263, 1.0392850295060467314, 0.41887923522892422934, -0.10198262635742695592, -0.26795039665613790136, -0.19955162286675193206, -0.074367079570078875594, 0.03611444125819337353, 0.22142201771315886405, 0.05243726602731452191, -0.5824699198161800151, 0.16818636985419174734, 0.35961657845508310238, 0.16779650818766639331, -0.58484348519988704762, 0.046393343660735254785, 0.2157062638654582698, 0.043868196350164145869, -0.036330015949997171432, -0.033859072714828263029, 0.00030202207737815020896, -0.016174750070871848212, -0.11917592859288141105, -0.12885482861913907793, 0.66442669728216341429, 0.73285675442670261681, -2.2675110314873765383, 0.7311630216122275705, 0.6630374186536375225, -0.12702641588070665013, -0.11448097001336454692, -0.017638144846975269131, 0.0011389720249697370298, -0.0032115146656337698397, -0.22147883293566289487, -0.026201231560884139654, 0.25695586876941473031, 0.38853525458427429218, 0.17359216690766721136, 0.090366956850450305683, -1.3169023072149808051, 0.083151789725145600074, 0.15918003270347413913, 0.37070399510212126426, 0.25050614650234537262, 0.0068176064807210726842, -0.10649935846395537642, -0.11462704567488567198, -0.10287556096377271919, -0.16648070984957230589, 0.043070683559375128047, 0.33859546483925190286, 0.71370977919744105211, 0.37044653520460329998, -2.4095235189775059403, 0.37480648370478242271, 0.72278524507549035683, 0.34981149435671105019, 0.049027363336015712669, -0.18001001730707807957, -0.15258014788307228127, 0.042850776660171188692, 0.052104830058413299787, 0.079567361601913119684, 0.023664406755298891816, -0.28078975394298216761, -0.37005610826622165055, 0.59443845110545623367, -0.18418474189870825963, 0.58673537114027618511, -0.3814486007099613385, -0.2879694724042173859, 0.030910544940983802897, 0.10373543974368421017, 0.075455610741626516491, -0.037786614783484738511, -0.022409950679274742019, -0.13723874134935198921, -0.1256811498608907951, 0.15513219995266819273, 0.86851882709425154427, 0.5303063490228362431, -2.5501120575389006895, 0.53299998303589180004, 0.87600352584640617248, 0.1674479810218919007, -0.11743125271331464332, -0.15675443108980471885, -0.080666552109375649904, 0.054933113757691560963, 0.34750211439837674909, -0.2533924695139896377, -0.42079619211593705108, -0.23394351330149570911, 0.10139102323638514436, 0.72487435786576104313, -0.56713754700726926039, 0.74353885619318504929, 0.14181993560599781556, -0.17943409954029621023, -0.39176669883654541282, -0.34460017378796187204, -0.015189395871061522891, 0.34805132836080032721, 0.43423433573199304858, 0.7981755199751839891, 0.46682909560025526163, -1.1351460170265996741, -4.5660403718530773176, -10.906430210718387031, 29.860263947723598932, -10.898860025210975522, -4.5646449524927419361, -1.1556156071423804299, 0.4277385253150770561, 0.7765863081572607074, 0.5098643693833656565, -0.01350024432147929436, -0.07780927879716711959, -0.13791422814538423225, -0.072008023660678266431, 0.1963379477939002915, 0.77015995733231379372, 0.51927360832059976659, -2.4008637430623909204, 0.51926943013266269045, 0.76850537024682241682, 0.1907700925972528605, -0.080689926615306578617, -0.14762913925837645168, -0.083091444738379427459, 0.024527491083667713606, -0.047489415772261941995, -0.11679773710260760167, -0.066645673647645675697, 0.1307135125930830033, 0.56277200639894209289, 0.50935662471300879872, -1.9489134011533790769, 0.51082085483779660073, 0.56350643976873215912, 0.12730468788688953374, -0.075030439934746395725, -0.12967122539907388612, -0.063691502225911911239, 0.034695294354929122338, 0.12350233808215689557, 0.11024423478353456707, -0.10019877622886474988, -0.43225163180319442491, -0.36030357578858818179, 0.66887138261079104318, -0.013534200580860253696, 0.66898922900019286431, -0.36007746775496252134, -0.43238740168634737238, -0.10040255198137788828, 0.10999135331310802566, 0.12936234186689810066, -0.0052891118504690163779, 0.079105086916321162849, 0.093034224936957538032, -0.029533982811322783552, -0.31727618595459894379, -0.40934105468777565173, 0.54437273137900299425, 0.084181736780436500212, 0.54443198780696211436, -0.40923061119409315234, -0.31794510068749748388, -0.030925829498688228031, 0.093325582113368485415, 0.091795227441702778592, -0.010969267262728519055, 0.1277112196060864735, -0.0017281712283011910203, -0.20687313522318118153, -0.29367920503499783669, 0.13397151182201560982, 0.68845902439571660114, -0.89975186956187946929, 0.69322307884469769323, 0.14199411747667664563, -0.28582367074830694609, -0.20631638638349236725, -0.025148185403097246537, 0.070184466529109743038, 0.065428655383026543269, -0.36676370666855817593, -0.23772089613003433151, 0.16976267543500092594, 0.88221614727542729817, 1.5653296637773568278, 0.099702685988401595907, -4.2211579323998842384, 0.084224033809003145956, 1.5399155804530317848, 0.85978801712397157075, 0.1713402228740045774, -0.19176935403139810932, -0.25081496432415395503, -0.12306655164759587595, 0.04216506634403249798, 0.081277608940729106912, -0.053088079978593334662, -0.34030865196296999908, -0.087996908656487002331, 0.70795434749753138881, -0.70362025093820101951, 0.71448266630221546336, -0.079362497342021123825, -0.33871130580962677259, -0.066729583545056606053, 0.05869390856306055948, 0.055498134327315096215, 0.011147925021018089548, 0.01019753079542268237, 0.10094862676431458348, 0.016514037763654509927, -0.25535793476864199203, -0.17144593854848372283, 0.61850964726543444794, -0.62983363590521690423, 0.61675702246551877828, -0.17714677254928981598, -0.26659269043553984746, 0.0078647895667097872385, 0.11789126481925202294, 0.082121611022811916158, -0.067055564776093298796, -0.0017280602785550080655, 0.10103334676019296023, 0.056698040346203550099, -0.2211140478198332926, -0.28613719082018629347, 0.54669599593995732434, -0.38121802036466345687, 0.54409131164569191075, -0.29297526776588850295, -0.23295426059271642782, 0.048681778881781385959, 0.12051737913125172563, 0.072589812002061798735, -0.070957616188238550881, 0.070183952083154241275, -0.13590171891395633197, -0.25834272642878974846, 0.046156741224461721174, 1.0631082778039175896, 0.67359321150990780769, -2.9310290816792461399, 0.67672157995748771775, 1.0734880597491553988, 0.066130765839728003486, -0.23818027218884296747, -0.14980872040337167572, -0.018718102184385292108, 0.063896304869827710848, 0.079004952902189520825, -0.089676097022168957396, -0.23061409050837958223, -0.041311950128151767037, 0.81273706461573769211, 0.67500245563398075799, -2.418963061221948152, 0.67690421473933837238, 0.81992299040896576834, -0.026426137240329244937, -0.21417132069535699257, -0.098028829448985360662, 0.012473496902071549985, 0.045373966897099229922, -0.057897279738288641437, -0.0077659428172510876209, 0.15292140378304203141, 0.084387867463065313611, -0.36210368114735541623, 0.21634403738304441855, -0.049914129158856825053, 0.21439503889513297774, -0.36505101717997046062, 0.080982994436425315143, 0.15082038839650216411, -0.0041189494878544740536, -0.043946834852140799488, -0.01104391035617533455, 0.056636449054546608717, 0.062396989555355476809, -0.024827559561954017087, -0.32426567863567190564, -0.18895049760668122807, 0.70041005051371063583, -0.55695045613015226049, 0.69868295494770904508, -0.19174071985125049067, -0.32741456380349798216, -0.025991993181953949865, 0.06597933203453955342, 0.061657934159044608391, -0.0034680597688402977491, -0.19955162352236266443, -0.2679503954905893548, -0.10198262684351053209, 0.41887923589141612402, 1.0392850284440953157, 1.1867892699195941031, -0.19251880575275159835, -3.7750961131107034774, -0.18168716798927420086, 1.2041067765299855452, 1.051782643566703257, 0.40394251568042177647, -0.17250326343479729929, -0.4189527443825249442, -0.033859072705115865476, -0.036330016054254843372, 0.043868196536778571171, 0.21570626366498463211, 0.046393344031167335628, -0.58484348570447286519, 0.16779650857793521057, 0.35961657810926783707, 0.16818637028802954636, -0.58246992027264743452, 0.052437266419155255381, 0.22142201745618392272, 0.036114441361321546198, -0.074367079587637135996, -0.0032115146998988406551, 0.0011389721045758313157, -0.017638144821572616927, -0.11448097033268801737, -0.1270264151945108555, 0.66303741768106017052, 0.73116302249667386448, -2.2675110319044722296, 0.73285675452865883805, 0.66442669715277047349, -0.12885482854368987549, -0.11917592845732355444, -0.016174750238470442937, 0.00030202212908321307474, -0.11462704580410376654, -0.10649935801808514679, 0.0068176064108795807195, 0.25050614577061569577, 0.3707039955211012261, 0.15918003252278858928, 0.083151790670324396415, -1.316902307975832187, 0.090366956746134291523, 0.17359216654991643791, 0.38853525550283701806, 0.25695586873832121366, -0.026201232194487760613, -0.22147883266448031847, 0.042850776618621876091, -0.15258014781337930765, -0.18001001716051912016, 0.049027363011629868095, 0.34981149455280219351, 0.7227852449715783667, 0.3748064838713900393, -2.4095235193650466066, 0.37044653573630226262, 0.71370977874028729282, 0.33859546533286311032, 0.043070683079715087194, -0.16648070959613309183, -0.10287556102705856798, -0.037786614771783667843, 0.075455610726348931738, 0.10373543986796217309, 0.030910544640537867406, -0.28796947224279151323, -0.38144860049919976364, 0.58673537070159720308, -0.1841847413791076471, 0.59443845063481781654, -0.37005610807062278012, -0.28078975385174004264, 0.023664406685032259026, 0.079567361534423730607, 0.052104830107347691048, 0.05493311382270073262, -0.080666552311255065866, -0.15675443078885192838, -0.117431253037896699, 0.1674479812221784103, 0.87600352590690477861, 0.53299998288741878838, -2.5501120575619848907, 0.53030634913194474311, 0.86851882701047722346, 0.15513220012194592345, -0.1256811500052031072, -0.13723874137886696301, -0.02240995062834241075, 0.3480513284368341731, -0.015189395830863396814, -0.34460017387785224496, -0.39176669911531103452, -0.17943409940613141984, 0.14181993583816546178, 0.74353885617225612403, -0.56713754709240238316, 0.72487435769699459787, 0.10139102369304785201, -0.23394351371407309803, -0.42079619195121870101, -0.25339246965415518353, 0.34750211449216006354, -0.013500244241766145084, 0.50986436938764601035, 0.77658630784945759107, 0.42773852560498010211, -1.1556156072797578727, -4.5646449523196315212, -10.8988600253934802, 29.860263947834226883, -10.90643021077255348, -4.566040371538393039, -1.1351460174924838942, 0.46682909569335173661, 0.79817552006699421518, 0.43423433572370390143, 0.024527491300835535526, -0.083091445283852843895, -0.14762913935667432153, -0.080689925984961136107, 0.19077009260311017513, 0.76850537007732666606, 0.51926943002690584272, -2.4008637427851318158, 0.51927360804263500427, 0.77015995740930121016, 0.196337947603057561, -0.072008023438913063008, -0.13791422828357224883, -0.077809278703825535128, 0.034695294561524529786, -0.063691502503356339782, -0.12967122555637586312, -0.075030439680678240322, 0.12730468782172207898, 0.56350643994384475022, 0.51082085465511839484, -1.9489134009572179895, 0.50935662442522589011, 0.5627720066065329263, 0.13071351236605133206, -0.066645673240789946035, -0.11679773749863155141, -0.047489415625897250794, -0.0052891119832393310857, 0.12936234205784311047, 0.10999135334992646285, -0.10040255198897767019, -0.43238740192035074461, -0.36007746751384112072, 0.66898922894180579135, -0.013534200960567470753, 0.66887138306797822018, -0.3603035756437669157, -0.43225163203945843327, -0.10019877652445487171, 0.11024423517034714903, 0.1235023379685669942, -0.010969267275184279437, 0.091795227320244893177, 0.093325582468075637954, -0.030925829891601031019, -0.31794510021320770976, -0.40923061164287444225, 0.54443198779448775948, 0.084181736831074188698, 0.54437273177405032243, -0.40934105499846995535, -0.31727618591416884009, -0.029533982984756058238, 0.093034225172153234285, 0.07910508683834756305, 0.065428655519124634399, 0.070184466270838741231, -0.025148185474342641538, -0.2063163860031557717, -0.28582367090071975069, 0.14199411732402394959, 0.69322307917309500258, -0.89975187006417622904, 0.68845902502308153892, 0.13397151126080569794, -0.2936792046609941198, -0.20687313558876382902, -0.0017281708980890994089, 0.1277112194934881817, -0.12306655174381778139, -0.25081496428318489356, -0.19176935394920047173, 0.17134022279496627372, 0.85978801711287722309, 1.5399155809258169292, 0.084224033194841993466, -4.2211579323998540403, 0.099702686102259768797, 1.5653296642482950052, 0.88221614688647542746, 0.16976267538322084572, -0.23772089622331007974, -0.36676370651869749651, 0.011147925035969476898, 0.055498134269058578549, 0.05869390862387122848, -0.066729583598413425016, -0.33871130560532719533, -0.079362497790431668632, 0.71448266682128880323, -0.70362025127156035964, 0.70795434761221254227, -0.087996908600920201171, -0.34030865223019274302, -0.053088079660749978861, 0.081277608791417529921, 0.042165066366719551605, -0.067055564755858471093, 0.082121610997016827871, 0.11789126507165273439, 0.0078647891258456850694, -0.26659269022393622395, -0.17714677258706290619, 0.61675702229566120582, -0.62983363531455760498, 0.61850964638144700913, -0.17144593774219640814, -0.25535793517910565509, 0.016514037952161560174, 0.10094862663584146245, 0.010197530824236952784, -0.07095761629150434957, 0.072589812308914897221, 0.12051737855971791924, 0.048681779819194706538, -0.23295426182098874501, -0.29297526656406264189, 0.5440913110323263302, -0.38121802051100328601, 0.54669599609762209447, -0.28613719066479753605, -0.22111404775158297054, 0.0566980401245013102, 0.10103334673349968764, -0.0017280601941498528913, 0.063896304896021577369, -0.018718101961810854061, -0.14980872090631441873, -0.23818027155519133631, 0.066130764937358629818, 1.0734880606584689122, 0.67672157931406307085, -2.9310290813227726225, 0.67359321142105077485, 1.0631082776167029014, 0.046156741457783936888, -0.25834272625767212839, -0.13590171919343646545, 0.070183952138462860071, 0.045373966957765868302, 0.01247349669608805757, -0.09802882906991641776, -0.21417132109653386207, -0.02642613691961374911, 0.81992299006446489607, 0.67690421484947216335, -2.4189630606588927719, 0.67500245469837316659, 0.81273706531097822392, -0.04131195063164470882, -0.23061409014548175933, -0.08967609703883763772, 0.079004952819113308249, -0.011043910120058880037, -0.043946835470522645217, -0.004118949125685313073, 0.15082038838028108407, 0.08098299483991173997, -0.3650510176336559387, 0.21439503872209489388, -0.049914128688963610436, 0.21634403725610695801, -0.36210368125039327225, 0.084387867274525504846, 0.15292140396304829131, -0.0077659426720641322808, -0.057897279856210133453, -0.0034680596686334267363, 0.061657933898711228538, 0.065979332167457424574, -0.025991993154244441566, -0.32741456367521909465, -0.19174072013821524818, 0.69868295536668578727, -0.55695045670128229265, 0.70041005114909415941, -0.18895049814417932965, -0.32426567835736774814, -0.024827559713347494702, 0.062396989680253256427, 0.056636449014904777233, -0.041240004873005581265, -0.18927775692907400362, -0.24620047866841421325, -0.091681454845905382967, 0.41714816795967696317, 1.0340050823886064446, 1.186789269113247558, -0.18168716736701995274, -3.7522887593390672478, -0.1502440458236920795, 1.2337482708247511365, 1.0521994666780056793, 0.28985045590849983776, -0.5545542640711397242, -0.0046485436490499090767, -0.031821559045871035398, -0.031458623127021030641, 0.043656282961037830292, 0.21347578300729791456, 0.045071216847913486703, -0.58484348597486013333, 0.16818637028650701426, 0.35990733245072970004, 0.16975329285032458859, -0.57663280255787074946, 0.068336012675166923902, 0.22614666710321579268, -0.06407646330832492132, -2.6174997753345988298e-05, -0.0025217312066874160263, 0.00049352083872154870609, -0.018662716508796407117, -0.1147486550096380592, -0.12724750609070892282, 0.66303741718367559255, 0.73285675508639169795, -2.2631352512116902354, 0.73740850854829409489, 0.66188775317805259579, -0.14099058704305439393, -0.13183642357439806503, 0.0029144665848412350077, -0.01632445488967134567, -0.093867483393291539784, -0.10188034093515377365, 0.00029446544180824520132, 0.24156954494555182844, 0.36495121697329885313, 0.15918003352760212721, 0.090366956015572041938, -1.3015099691846041541, 0.11289897970536209859, 0.19628919752083032191, 0.39495656017321773001, 0.19499280471475549659, -0.2413787778666281747, 0.075823463096371307435, -0.051550570030582262837, -0.16561109063840859235, -0.16480803076275932217, 0.065023773903486140058, 0.35749698862295292701, 0.72278524542745725689, 0.37044653499557012033, -2.4145202056231358512, 0.37061271166302794944, 0.72619430318764843779, 0.36512671245520272878, 0.046650887180772530882, -0.30664175749777095037, -0.063743116524774823262, 0.031337597731729165129, 0.094702196363665516743, 0.095201260245866309306, 0.016200074328949261998, -0.29714742743609640074, -0.3814486001380271718, 0.59443845052862342992, -0.17426058938855717995, 0.59674399360661867231, -0.38714385593639699668, -0.31089225619794857991, 0.022060504860204006861, 0.16669496317296925159, 0.061253962032516123815, -0.017008812584730911149, -0.093356923296334853868, -0.14614622004762742313, -0.10533550936885113791, 0.17317923510554486533, 0.87600352590170904588, 0.53030634911411800303, -2.5530342467123192662, 0.52936487132948106815, 0.87237735727742249381, 0.16616707071372041615, -0.12657384320873951711, -0.1706055540020347816, 0.16905589548092075924, 0.14307421298174036184, -0.043134253897884104745, -0.30785782376938497507, -0.35292498942242467308, -0.15938904693470024032, 0.14181993582469265558, 0.72487435788667575753, -0.60463693563984888257, 0.67494961790190710627, 0.064973687873859806152, -0.20226468014105847271, -0.27029053107567146963, 0.013490706770154681227, -0.18716535620225213665, 0.19434232189285738368, 0.56394974841455691728, 0.75455023438746160469, 0.39516783698416624304, -1.1697072545506361774, -4.5646449520383729492, -10.906430210764384015, 29.824746762645148124, -10.9691442597377673, -4.6165914081411205672, -1.0964654867301415475, 0.65387308458657311672, 1.13757218872775967, 0.070689304178932571565, -0.028646116217747309013, -0.10890021784068890098, -0.15347834565740650636, -0.07767471184533515649, 0.19360585425860021846, 0.7685053704130366814, 0.51927360772256592192, -2.395026047072339459, 0.53572585649824300269, 0.7926290786991316839, 0.20655810506231037094, -0.10713647151057914542, -0.22348878069776864641, 0.066102949432549892861, -0.01632477832749683333, -0.087734200564215436557, -0.13414425764643808647, -0.070879093999341502563, 0.1309665497588070604, 0.56350643963720692398, 0.50935662490440003491, -1.946348121399219977, 0.52117910521604204899, 0.58123341481918555651, 0.14195786777503066611, -0.089321747550800786875, -0.17599786577750242689, -0.051657606412563727982, 0.053425263972634737253, 0.13747390019609162515, 0.10349570948041321439, -0.1056708704088907963, -0.43418555677249820457, -0.36007746821843400475, 0.66887138328039719148, -0.015271564586349721873, 0.66301985339054281265, -0.37611228499550386495, -0.45450346109936828887, -0.092929658888279947915, 0.26685220945517362789, -0.040193010681897650693, 0.033346014461179483701, 0.099008045548291734317, 0.089025417760830644065, -0.03500727426958628874, -0.319565459891541781, -0.40923061204403621272, 0.54437273200449487476, 0.082440215694257440449, 0.53943289582842579577, -0.42023552924958645916, -0.32972978863714264008, -0.022784956674390382386, 0.19150694761976094771, 0.017381558172235356263, 0.053356691169376711803, 0.060557170601221239903, -0.025862383928040522107, -0.20104461206873183721, -0.28113991944528765599, 0.14199411686574736624, 0.68845902489417898362, -0.90667746023924700172, 0.6813241046898290687, 0.12504135381208120181, -0.30007321155617427966, -0.1889408928873853033, 0.13404168265246502179, 0.038117476641044728547, -0.15031837790193033899, -0.25886747910683793883, -0.1995785048867115552, 0.16027729364303949788, 0.85048610700077942326, 1.5399155809563724873, 0.099702686139126292519, -4.1883576553614885185, 0.1446088090360052314, 1.6059776058454831027, 0.88312116875364599089, 0.04877018246409180191, -0.5800969986302426662, 0.0077665997004378803381, 0.010769470277471283534, 0.044808591816586501588, 0.056076031972133122405, -0.061131407695471845498, -0.3330274216487718042, -0.07936249779191287268, 0.70795434773044962906, -0.71257322814956869372, 0.70382539635330432226, -0.083910041682180455669, -0.33036270007959095407, -0.042066722590527183545, 0.11108599255556031449, -0.059620298073762463542, 0.0068081267638113399793, 0.08752851672052899823, 0.1055742461980775232, -0.00065121705165209486932, -0.2694756174573378904, -0.17714677187493838018, 0.61850964636975835909, -0.62498900644610833055, 0.6264956593007549035, -0.17060684293553191049, -0.27612782261765694836, -0.013786347848103862324, 0.15041035415640569006, -0.060505768129645261244, 0.0023570211586723087416, 0.080093996535645473056, 0.10882062112942245125, 0.03916873522560752191, -0.23685477804782162359, -0.29297526621304559313, 0.54669599509170230256, -0.37568848550830369337, 0.55442451788836832893, -0.28550014676366031185, -0.23969739610633797322, 0.02851872637554853096, 0.13415618559482542227, 0.023753453029837227201, 0.021090222690451181087, -0.022260701932050513951, -0.13580967609153185571, -0.22444076243470489906, 0.07236706170089096668, 1.0734880605557335365, 0.67359321108400882494, -2.9360744832977156982, 0.66601943734031443523, 1.0513025813379313078, 0.032749263756898190636, -0.25909490508232496664, -0.036479504028148930506, 0.0049778678236985368644, 0.027770335795197415019, 0.011742347269907825458, -0.089163324099155755187, -0.20527518558087579081, -0.022341963305276046903, 0.81992299088899411874, 0.67500245462003816144, -2.4220468199083979854, 0.66921124484071670135, 0.80023473365975006377, -0.0594492976092355907, -0.23425550338660011684, 0.024401216079810484488, 0.010962842158738234782, -0.023472773447015322923, -0.042450942524268626965, -0.003339258225655400078, 0.14876801562806060697, 0.078792151249206507369, -0.36505101775674708708, 0.21634403781791256005, -0.047909991478221780614, 0.21803364027536814573, -0.35765892582802538691, 0.0980568837822282241, 0.15628613592254464448, -0.087970686991166391344, -0.024658183473849636796, 0.029069394205058711467, 0.065890744398529405013, 0.059574927399245802828, -0.031852822407415930239, -0.33016409548818514974, -0.19174072111338172353, 0.70041005166025915862, -0.55480151927285736857, 0.70040474961427656719, -0.19723583028186114641, -0.33883313989305324787, -0.01719423791701281401, 0.13162665277270271802, 0.037326362483768735223, -0.08457961481493458944, -0.20092131725364342776, -0.2463232136844144271, -0.091314469604993786511, 0.41714816768025292371, 1.0392850298870943693, 1.2041067758349963679, -0.15024404540646851824, -3.7122843837462808914, -0.10810042738254412031, 1.2479825468708323211, 0.93492877207548252816, -0.26213681290734847096, 0.0070013926192332145634, -0.012816918336447611151, -0.032857651157075125348, -0.030901952912607919682, 0.043551367066878239609, 0.21347578319733698238, 0.046393344078029967459, -0.58246991997181685186, 0.16975329228070304421, 0.35906844599244053251, 0.17062520074902748357, -0.57111478684923222549, 0.089752800273961458322, 0.13316670262261282032, 0.0028296843897177028325, -0.00058987419876414370467, -0.0046179871929026812885, -0.0014766832434405425829, -0.019363401985072373201, -0.11474865530000381841, -0.12702641470926673284, 0.66442669691027844703, 0.7374085082309921324, -2.2553507116180493419, 0.74562617727350521957, 0.64770812313887971001, -0.1940088858352859158, -0.080255402726860866625, 0.037307657170379031342, -0.045648490154979688704, -0.1082211665598338951, -0.10856739647133638316, -0.0028719075936490447848, 0.24156954545819406266, 0.37070399328837810859, 0.17359216900212681711, 0.1128989790580101904, -1.2776697825431471145, 0.12855794843710124931, 0.18108089590865247831, 0.32935218600947430145, -0.022019909449545228136, 0.046494484640211553028, 0.018377630108816551957, -0.058827922753452319138, -0.15470130049304886555, -0.15586146012198692778, 0.065023774774577589031, 0.34981149431714497755, 0.71370977929928502981, 0.37061271061386075187, -2.3926027385476156262, 0.42424147641450610235, 0.78634671829533742127, 0.32485218137984406273, -0.33332681848469797314, -0.040855694929262061088, -0.011887949311505165498, 0.036793056251269384838, 0.083191947947261049578, 0.085472066342712482712, 0.016200074534791483738, -0.28796947247457715946, -0.3700561080219101906, 0.59674399342161887905, -0.19350113189983067508, 0.55465499942734741712, -0.42820821057409269583, -0.26492156177223519231, 0.2234638372581551069, 0.031312648798395549543, 0.02383945297639156885, -0.021798132994912008259, -0.086864747727953744993, -0.14045887906444962101, -0.10533550968137855852, 0.16744798125204338191, 0.86851882698640681113, 0.52936487142411003948, -2.5366450065324666063, 0.57104649424796438684, 0.90699320673579475471, 0.11601117491420627481, -0.32271336484641233167, -0.0054138627021979328011, 0.14867158122056833847, 0.15994823286127640705, -0.017857593014109815477, -0.29155906489578298224, -0.35292498951889134107, -0.17943410014838573341, 0.10139102403388208784, 0.67494961844692413777, -0.63352124919668739711, 0.71718628947611162694, 0.20197015391674943974, -0.050096067899092953934, -0.49120642928822227935, -0.18761992009607011966, -0.004665333991709080981, 0.25330723926706255611, 0.55912744056240215595, 0.73834974641152073538, 0.39516783751402556613, -1.1556156072194319062, -4.5660403719482482998, -10.969144259767194427, 29.672367786810781354, -11.152756106496305222, -4.6463801833816704701, -0.69323294760866549513, 1.7332236225047126243, 0.068640149398418504045, 0.012651066027167000294, -0.051560783780288270417, -0.11148189572933926705, -0.15059596039425521274, -0.077674711363620599358, 0.19077009203824879524, 0.77015995766668821076, 0.53572585616701284827, -2.3609374834509107011, 0.57567494710008737524, 0.79874425433148477094, 0.1316727260303214575, -0.33136334485354534962, 0.057901109477899843914, 0.016011669184055308396, -0.035225434412192682676, -0.088582708428133594136, -0.1306456739313702764, -0.070879094584957294534, 0.12730468913002226938, 0.56277200510658098587, 0.52117910612621609623, -1.9186282390086588645, 0.55542924004172955499, 0.58901537689355498717, 0.089737454932345112057, -0.27608070112701754706, -0.044867447358436736427, -0.005269802623667099381, 0.063899476369152496957, 0.13465818585423619536, 0.10011788677416320892, -0.10567087089701017455, -0.43238740134280057026, -0.36030357592563394542, 0.6630198536660092401, -0.032182244401547675794, 0.62871318409503895719, -0.42482521416145258009, -0.41739558479503036104, 0.22818126254893505989, -0.033224148178488517591, -0.0068918267780911082396, 0.041890020293525655204, 0.097436771492354440349, 0.086465543592863750022, -0.035007274505734514713, -0.31794510002285408801, -0.40934105519631469594, 0.53943289599693111658, 0.069174167831877436674, 0.51435064536641439403, -0.45609011925637993334, -0.29683378469869275973, 0.2038468142160103358, -0.0092311079211255444438, 0.025306205118494663753, 0.055995375838663194457, 0.063203108834868362842, -0.023087276162859375617, -0.20104461177785129533, -0.2858236696217925088, 0.13397151100015886049, 0.68132410382636121593, -0.91021690348258887937, 0.68061176446590954114, 0.11440248720861052878, -0.28890932749590325646, -0.04260418964834423311, 0.10268312259401984887, -0.041751138573346474603, -0.18779120376364411005, -0.27414880101469185636, -0.20530271423680068743, 0.16027729381868893732, 0.85978801664995485599, 1.5653296641187901539, 0.14460880935259989855, -4.1303144811638201617, 0.20122896638760001675, 1.6094173966969691492, 0.71175641358927688174, -0.5003841808655874468, 0.0026667707187229788661, -0.0010136947606522554721, 0.012763854113227508791, 0.05103416039818082367, 0.060861985935778434298, -0.061131407721619283879, -0.33871130596005016855, -0.087996908319919853181, 0.70382539619208051018, -0.70675322125372697091, 0.71114059145008789997, -0.10541724240298827597, -0.35508657310737362645, 0.11074316478599777192, -0.021305610541271228359, -0.032753563177568258136, 0.0076028162748224131784, 0.082459393665979041632, 0.10231359701523561057, -0.00065121701154625261099, -0.26659269047468464597, -0.17144593867114824026, 0.6264956599684823324, -0.62401187847908456963, 0.60006683308164809887, -0.23366746246373479301, -0.2816342032766401271, 0.21389256082465862852, -0.02127691739081711414, -0.033348364844898059633, 0.0033993105632164032082, 0.074457174079972043845, 0.10479218060725359329, 0.039168735509897946445, -0.2329542620713875023, -0.28613719011823396388, 0.55442451763160638389, -0.37659942371089205615, 0.52695243516760259528, -0.34434981668631770013, -0.23662084777415393066, 0.22929286498387471682, -0.010785078125717762615, 0.023569307857394111716, 0.027605159233818066244, -0.014297568921530022512, -0.13052657661512187537, -0.22444076305866658694, 0.066130765196565008401, 1.0631082779734757349, 0.66601943775344163434, -2.9321922269018845242, 0.68743359028219430673, 1.0697785852304551035, 0.01364335464808220412, -0.30515757810871135813, -0.018084734514016865409, 0.015933871208060043645, 0.034364946334386137217, 0.016500405605838597278, -0.086160969098446643644, -0.2052751853465668852, -0.026426138351058080234, 0.81273706574912174538, 0.66921124489129402146, -2.4218031611706569706, 0.67756132330328189273, 0.80161768762269169919, -0.068319146254007256935, -0.20258385479321597278, 0.013794978879303344821, -0.0014114332126329543447, -0.026548173493819569868, -0.043620455641005494163, -0.004653375606754881974, 0.14876801539570697064, 0.080982995850199163135, -0.36210368254374036834, 0.21803364046322640246, -0.046994105340900239887, 0.22431746787430117962, -0.34652140852392787629, 0.11471805772779729293, 0.03281813194927268168, -0.01832940632659086494, -0.00075993893443531378151, 0.031137706843653266409, 0.061300066298563875289, 0.055952717792194892055, -0.031852823135232501139, -0.32741456338649493896, -0.18895049756879270242, 0.70040474976737721047, -0.56258474379512291019, 0.67976688596569612155, -0.22590326614611991318, -0.31890153307376611913, 0.14352155454078038832, 0.060411724292287617699, -0.013711068688812794863, -0.10661936172201347262, -0.20615198696700465231, -0.24632321284717731413, -0.091681455313456380884, 0.4188792357209643602, 1.051782643426865782, 1.2337482702453796968, -0.10810042733778632562, -3.6895396099304811344, -0.12760703083005237035, 1.1319478738321391287, 0.72872885497998651783, 0.0087644498213387440899, -0.0014137455419033274386, -0.015652763151859466068, -0.033077852298587756552, -0.03090195210259657696, 0.043656282194076176384, 0.21570626419300531484, 0.052437265704600936889, -0.57663280185434495628, 0.17062520055798352514, 0.35073267082642850534, 0.14351804443719184956, -0.66861882142270412377, 0.34265445273060624043, 0.0044960568979230553649, 0.00019376566363019602892, -0.0030617194120575682605, -0.0058162469819523282197, -0.0014766825858876873184, -0.018662716874026416547, -0.11448097032197426803, -0.12885482878406825935, 0.6618877536797139749, 0.74562617737263769957, -2.2385670688451928534, 0.75771349157447298595, 0.62918623924825145544, -0.28560064515508765126, 0.049671952022979470875, -0.0049117708695918726666, -0.064512167298113334835, -0.11372797162394486514, -0.10856739577598030488, 0.00029446410837281541961, 0.25050614560479983473, 0.38853525519651405107, 0.19628919910397610304, 0.12855794788665431128, -1.2939121101359287724, 0.076565074197120053245, 0.095032347356213811573, 0.41597454117913723604, 0.010065180739005809632, 0.034717294079716291832, 0.018970558417638547222, -0.054544006764310926094, -0.15470129914160590801, -0.16480803180305086952, 0.049027362399747656529, 0.33859546561859965585, 0.72619430410709306933, 0.42424147550149904351, -2.3122367738919100688, 0.46640574565319248279, 0.62789894157404457786, 0.0026860412284503373294, -0.0062703255335947189614, -0.028392779583755470979, -0.016512078729445430619, 0.029780630694700088029, 0.083191948456690673064, 0.095201259749388519871, 0.030910544869443234289, -0.28078975405964629042, -0.38714385540605888991, 0.554654999097014767, -0.23196772113197369758, 0.56531431719357660981, -0.40924129010992704725, -0.004545786218063076721, 0.0027346118097190206392, 0.028123556419211593466, 0.025008994384435025149, -0.019394561273948895952, -0.086864748010652872989, -0.14614621968566418597, -0.11743125302637126561, 0.1551322001398782735, 0.87237735713407538185, 0.57104649446529764933, -2.4718897373313586385, 0.61543155989836118014, 0.84281419898629950094, -0.26338169101807834549, -0.090654903623216764874, 0.072239659636395295306, 0.18503346587548069602, 0.1736189542023426069, -0.017857594192835307689, -0.30785782347277884652, -0.39176669817601544876, -0.23394351426585685561, 0.064973687569987809143, 0.71718629010426415693, -0.47387939406611612192, 0.92728010494440327172, 0.1131254695678773764, -0.75386833910358108657, -0.11719495366520114865, -0.086581660766902823756, 0.031545044338022631969, 0.25656354921974861005, 0.55912744069438546912, 0.75455023452184832866, 0.42773852517228305725, -1.1351460171416536404, -4.6165914081226633314, -11.152756106898545241, 29.354870434325725626, -11.35923447718855428, -4.1629882900699044512, 1.176093013159797529, 0.041058711720092740116, 0.031169893897047429993, 3.0630221239454080706e-05, -0.052797587233762842407, -0.1114818961768050537, -0.15347834571803142811, -0.080689924808762900077, 0.19633794657202324752, 0.79262907915575075979, 0.57567494673586083742, -2.3270977843315798772, 0.58813189755253225677, 0.71722940370199572691, -0.20568144444559299799, 0.030837628549729940997, 0.028631007919010617591, 0.0073868212001916497278, -0.035114499990578043609, -0.088582709548781335407, -0.1341442558790806594, -0.07503044133819652961, 0.13071351407495085395, 0.58123341322945043519, 0.55542924074635047216, -1.88780336127024162, 0.57171895573668185531, 0.51320735334867217947, -0.19069971073619493285, -0.022895580058837201182, -0.024503094819613727151, 0.0020077644869552742704, 0.06422761428172087117, 0.13465818569302548391, 0.10349570981138354886, -0.10040255177078365179, -0.4322516322512458542, -0.37611228562695891986, 0.62871318465254466545, -0.073529277687344171688, 0.6328904691739086763, -0.34866029959807387506, -0.19901921899187019482, -0.016620699069597729552, -0.018636648917142121379, -0.0014246202901382820291, 0.042171385813671441456, 0.097436771187914425485, 0.089025417965919270702, -0.030925829371455104688, -0.31727618618519176641, -0.42023552972548866569, 0.51435064581532285999, 0.035685276502983503955, 0.51089514588724205257, -0.41665234415121810141, -0.076198816840698638253, -0.019286682133786831744, 0.0053941767405218674428, 0.033798300984931445301, 0.059669319007197657578, 0.063203109943753069122, -0.025862384378847181471, -0.20631638657928691449, -0.29367920487046672573, 0.12504135413722219927, 0.6806117642141166213, -0.90049349806999201551, 0.71551049120877752152, 0.14924269740416939989, -0.39450513322276647976, 0.10065465184820894529, 0.02218916102448842062, -0.082191626574741161182, -0.20138489146700885857, -0.27414880170475713905, -0.19957850477959759861, 0.17134022296787332351, 0.88221614709355944406, 1.6059776054293550818, 0.20122896674901033509, -4.097773418893474684, 0.176526682248444311, 1.480607259537318976, 0.24996892526576830473, -0.0070970545301563322388, 0.0020608812815665275564, 0.007072864541220635036, 0.0187712012967456085, 0.051034160640275813559, 0.056076031647762217269, -0.066729582710851984007, -0.34030865297989748575, -0.083910041339555832551, 0.71114059132676199493, -0.71341643997750070483, 0.68520956088917628346, -0.13844249955093310134, -0.18713155639928238938, 0.0057065272287021365227, -0.027997157857649841717, -0.03420714522866016849, 0.0068876729490579095827, 0.082459393154429835859, 0.10557424716785136776, 0.0078647875948892401188, -0.25535793295442499318, -0.17060684357749053564, 0.60006683250307102373, -0.67890006519266143936, 0.57486661427786234224, -0.19534630638657979329, -0.024323540618290145909, 0.0066612313278490709045, -0.027914068008533288873, -0.035836009786249685416, 0.0018439137679396103153, 0.074457174661922087555, 0.10882062062469925745, 0.048681779262759086735, -0.2211140473477369861, -0.28550014656120303869, 0.52695243454563067687, -0.43206726303763992592, 0.49979873653114492704, -0.31776009135498534119, 0.050083593480303896517, -0.024066401043077962491, 0.010071317999568235244, 0.03366615603002413859, 0.031420214021813289151, -0.01429756936162106945, -0.13580967559688131296, -0.23818027226169774524, 0.046156741941975776289, 1.0513025808220402002, 0.68743359068876597195, -2.8754893033221176069, 0.7759976893075809512, 1.1046011817204273164, -0.45013200065130315464, -0.021976489583017531915, 0.0014380092392021689984, 0.024436790022017478019, 0.03694646903421747175, 0.016500405322132545571, -0.089163324398521537351, -0.21417131987613705069, -0.041311951171889016443, 0.80023473437596615021, 0.67756132282425929336, -2.3876330529500404687, 0.75073313223431059615, 0.85150561703127347624, -0.40483801043574813017, 0.010215920902924507593, 0.0061299724069430617385, -0.0062017232071894228679, -0.028701602647391757928, -0.043620456048005844252, -0.0033392573789547026764, 0.15082038800098296338, 0.084387867733565102846, -0.35765892628293727284, 0.22431746806691615914, -0.04309674701470843361, 0.21595738210510753707, -0.43501737150952912714, 0.22851399929439961123, -0.0055212473738716193877, -0.011881711092371667268, -0.00066627463561464100959, 0.029258858319210812732, 0.061300065635898404182, 0.05957492771024314876, -0.0259919921896604178, -0.3242656797115964773, -0.19723583066954625265, 0.67976688723800704484, -0.58048124424926739184, 0.6867495896177912007, -0.24606865088018034515, -0.13024687915532778826, 0.052014848389592402422, 0.021755717635639350988, -0.029346428743090895164, -0.10661936302779431718, -0.20092131604638993614, -0.24620048034538735693, -0.10198262749043418562, 0.40394251638644113056, 1.0521994672512684499, 1.2479825465184590794, -0.1276070302774759968, -3.8548801994180359465, -0.1446100655349145947, 2.0520461353074197497, 0.0060565028087593278036, 0.0033475503176145942653, -0.0033710366166034180617, -0.015652762891244977744, -0.032857651790431020855, -0.031458622550017530106, 0.043868195896999979955, 0.22142201833966490554, 0.068336012118159150663, -0.57111478681080074526, 0.14351804433688156171, 0.24528019537370090997, 0.034930384994031567214, -0.11503346564812279296, 0.0040373541797032869161, 0.0010116668197689012397, -0.0013391552249063308315, -0.0030617193795285093869, -0.0046179877878550924103, 0.00049352117277008070596, -0.017638144781978157272, -0.11917592820604461235, -0.14099058748358145321, 0.64770812302325719845, 0.75771349172134028205, -2.1833496056374444017, 0.98215376200003867702, 0.082336165313250275766, 0.035949577625813022852, 0.016472971238855649984, -0.01751443640282921918, -0.064512166428479997271, -0.10822116879335924589, -0.10188034014460710386, 0.0068176095208527113037, 0.25695586578805168498, 0.39495655947508756523, 0.18108089785358066903, 0.076565073323015570006, -1.3962823316006298668, 0.063764351346408898458, 0.665041041716483039, -0.0093741707554718771422, 0.022541785442875380413, 0.0389546182581884462, 0.018970559478155676197, -0.058827923483521955828, -0.16561109061325479663, -0.18001001712593633908, 0.043070683443970400606, 0.36512671266205282139, 0.78634671846677184348, 0.46640574480035784655, -2.4357354248435236777, 0.37488205842957844638, 0.7477440650985699433, 0.014067349004103422289, -0.020294812328348215552, -0.035524625397799547422, -0.016512078538565039265, 0.036793055918602615706, 0.094702196850073055989, 0.1037354392290310845, 0.02366440723770256091, -0.3108922566646661334, -0.42820821040668238799, 0.56531431715278723793, -0.17044407591619051279, 0.66552555477048491728, -0.53005467754586510498, -0.010133423447570386045, 0.014511768959391862371, 0.031588196673023025141, 0.025008994309993370919, -0.021798133003708995703, -0.093356923294227331378, -0.15675443066544361703, -0.12568115017783854093, 0.16616707069623476434, 0.9069932066581981589, 0.61543156001293553015, -2.4756767041954370612, 0.63023469067877380123, 0.50642393913538108485, -0.090455977783190852359, -0.0028313849067092338938, 0.10205049259833520736, 0.18503346517679583605, 0.15994823317952533426, -0.043134253524163175542, -0.34460017339042192663, -0.42079619253501532317, -0.20226468058067745281, 0.20197015423481992635, 0.92728010529765247938, -0.45647528151747002179, 0.50642437706183451862, -0.5191461013085105769, -0.042973870091961192952, -0.0869378849265953213, -0.074163064562685995784, 0.031545044661804570729, 0.2533072394736432531, 0.56394974819519994025, 0.77658630776775061655, 0.46682909602310335151, -1.0964654867104630664, -4.6463801833364222205, -11.359234477575830269, 28.93618078809298666, -11.308241125187615594, -2.5027115380015927393, 0.0096107468296295628374, 0.028195040068094076591, 0.027234082763451294973, 3.0630306334218648789e-05, -0.051560783369146240274, -0.10890021712549838306, -0.14762914123635414998, -0.072008021683435952109, 0.20655810425835985367, 0.79874425453300479383, 0.58813189739830029623, -2.3529705196856243177, 0.58089433073278295971, 0.51300373562130441307, 0.0035389785009223135269, 0.023752434781812391956, 0.027113398359754261463, 0.0073868209456290634396, -0.035225433105678424783, -0.087734201839737915862, -0.12967122449853685229, -0.066645674580336988768, 0.14195786919242406321, 0.58901537594942676712, 0.57171895601171907053, -1.9103079017473079126, 0.54048197685352605824, 0.34002284113922476294, -0.0047591862285867756951, -0.019480548453077794341, -0.022178424420954561541, 0.0020077642955567539737, 0.063899476310506977961, 0.13747390035956894017, 0.10999135304891147213, -0.10019877609114261063, -0.45450346093445648243, -0.42482521465571077002, 0.63289046955007000061, 0.060568948175043754723, 0.77538390505939247532, -0.7675140906319997125, -0.0030603024461125444625, -0.01419694300684439317, -0.017130180397611268622, -0.0014246205119528513455, 0.04189002048012600149, 0.099008045615888440327, 0.093325581913624819363, -0.029533982514767367911, -0.32972978862448354409, -0.45609011950501043842, 0.51089514607036323746, 0.11722800228327748262, 0.59278062876009396387, -0.61330835933493021983, -0.018111768217442307527, -0.0045210464436330928539, 0.01244868747340804882, 0.03379830146190797302, 0.055995375779535123961, 0.060557170809823958335, -0.025148186836428963742, -0.20687313389489281334, -0.30007321109889095734, 0.11440248616993500819, 0.71551049108013475841, -0.79113732562029903672, 0.78461344459828952225, -0.43286485985579686853, 0.062707032427542316877, 0.042725075961753282427, -0.0027508111054120416503, -0.082191626872028758499, -0.18779120320794787102, -0.25886747904012907817, -0.19176935356101895702, 0.16976267483885337795, 0.88312116887273583998, 1.6094173965904454704, 0.1765266823257137796, -4.2258749234216290702, 0.18790940981091552864, 1.8525738343513253881, -0.012542429355322867388, 0.0029206608118142844688, 0.0088489600081504693474, 0.0070728646855990895403, 0.012763853556811716877, 0.044808592249427937892, 0.058693908174035268333, -0.053088079253972429716, -0.33036270051694588412, -0.10541724201069777034, 0.68520956067871174877, -0.6834328036770576098, 0.851393906690188218, -0.4903649059301015134, 0.013092416434064058084, -0.0084205772720815173088, -0.030683113718757024091, -0.034207144755253868218, 0.0076028169984165110129, 0.087528514980265231804, 0.11789126690279444487, 0.016514035937574340906, -0.2761278216080232939, -0.23366746207758620768, 0.57486661363470259456, -0.59562443186556845998, 0.75731622624738592631, -0.40233640772245754524, 0.014691187474558038348, -0.0086238327000552530283, -0.031638380875439742201, -0.035836009161621260044, 0.0033993099848466205859, 0.080093996836926015237, 0.12051737886671867739, 0.056698039812103584456, -0.2396973956585286003, -0.34434981642071027519, 0.49979873595512103179, -0.37034924379354600132, 0.64314459298739201643, -0.39446515435715323594, -0.018357867962306324838, -0.0023938675223334491063, 0.017523754355259095472, 0.03366615581443706956, 0.027605159078791081939, -0.022260701948663169647, -0.14980871990439548203, -0.25834272753847070669, 0.032749264250988313618, 1.0697785851880079466, 0.77599768974394178755, -2.7592424805147666156, 0.86149144861157811448, 0.39829920436479049339, -0.013452896962335142775, -0.0061601425995652138445, 0.0070898480407722598573, 0.024436789997373704886, 0.034364945994694404141, 0.011742347674431108384, -0.098028829560326963266, -0.23061408972270922013, -0.059449298658623195291, 0.80161768848789416442, 0.75073313225868221199, -2.2605457494354719294, 0.85401356090868318915, 0.18773138702576289139, 0.0054050336691444872753, 0.0051185907722332358097, 0.0027369952274507174279, -0.0062017235245497267029, -0.02654817235400343256, -0.042450943054601501803, -0.0041189503408122288275, 0.15292140507323659371, 0.0980568839992415342, -0.34652140944614195917, 0.21595738253176052779, -0.12825534237276386973, 0.092442606843916891313, -0.017557087832221214552, 0.0020325765436577237881, -0.0091542112671681181296, -0.013200556868325654877, -0.0006662753061427638358, 0.03113770647943469444, 0.065890745303142914113, 0.065979331970711652122, -0.024827559565096860711, -0.33883314022435689905, -0.22590326672290192422, 0.686749590597533488, -0.54131382637642522226, 0.81541408055452391235, -0.51787977127513862285, 0.028484314705197515127, 0.03256086581583070988, 0.02175571396960855261, -0.013711066277065503405, -0.084579617819870389628, -0.18927775302849586536, -0.26795039598966075012, -0.17250326373810156988, 0.2898504561359690479, 0.93492877238040306143, 1.1319478728803049616, -0.14461006522548983111, -3.5670407700918098648, 1.9499300187676902318, 0.0019579275473102109723, 0.0038522141730051380655, 0.0033475502161468902347, -0.0014137453084797962834, -0.012816918573210555471, -0.031821558981803929611, -0.036330015801321544067, 0.036114440762089361492, 0.22614666758133022961, 0.089752800132969670632, -0.66861882109885639736, 0.034930384627008495557, 0.82983124119293916188, -0.48070430101979133575, 0.0016445769127784080031, 0.0017462461058239213667, 0.0010116664605142671688, 0.00019376620735731550985, -0.0005898745469761679152, -0.0025217308105517830495, 0.001138971752829252837, -0.016174750322773427669, -0.13183642314893734282, -0.19400888617165384931, 0.62918623957107377986, 0.98215376169859180955, -2.4034967443470778292, 1.1337512729772876163, 0.010589473760107057446, 0.019700727723936283659, 0.016472971521194843675, -0.0049117729529904851499, -0.045648488145548206385, -0.093867482366645174641, -0.10649936034545436081, -0.02620122981116553007, 0.1949928032870499317, 0.32935218437540431102, 0.095032349425297077405, 0.063764351135120939107, -0.9988185283851138685, 0.53105896342680836142, -0.010505240953311380248, 0.0043002229702965794403, 0.022541786401271874524, 0.03471729315753630668, 0.018377629613580456769, -0.051550569359630725452, -0.15258014689868845681, -0.1664807108124015389, 0.046650886356554607803, 0.3248521820565347662, 0.62789894250788080843, 0.37488205753751246974, -2.3871738375421922562, 1.3124694168002377648, 0.014505278637084338331, -0.0028265815144360431285, -0.020294812143350278444, -0.028392779500560787198, -0.011887949551705277035, 0.031337597553925296279, 0.075455611499356034511, 0.079567360870467826461, 0.02206050495344874865, -0.26492156160289437317, -0.40924129004324188985, 0.66552555456741202544, 0.080294560250461707862, -0.23306912743807159383, -0.0092633394060654891677, 0.00013816664122216369606, 0.014511768814317831458, 0.028123556661817220026, 0.023839453109149817689, -0.017008812756002422012, -0.080666552555195156216, -0.13723874110204109522, -0.12657384306325256129, 0.11601117471546119886, 0.84281419898298060023, 0.63023469073750770786, -2.7209278361796518375, 1.4412914327811068294, -0.040632373820459387337, -0.04012480029397772624, -0.0028313850592508287007, 0.072239659598997460477, 0.14867158255074969242, 0.1430742119801420853, -0.015189397028106376916, -0.25339246847032703691, -0.27029053063068869678, -0.050096068871448566628, 0.11312546932228127738, 0.50642437775683901702, -0.74392062914843937182, 0.47083468068053185851, 0.0085180873251126196927, -0.046175419825001222385, -0.086937884100371079499, -0.086581661171378082598, -0.0046653339621098091627, 0.19434232129517739485, 0.50986437002915852013, 0.79817551938033337589, 0.65387308463661841795, -0.69323294711988769912, -4.1629882900324055583, -11.308241125531026228, 27.328387898805555523, -13.108469677377579288, -0.01367011349470584336, 0.010516060082480394727, 0.028195040830963920275, 0.031169893702156378779, 0.012651066416740595899, -0.028646118082414261108, -0.083091443659444014225, -0.13791422933928171313, -0.10713647066669511332, 0.13167272614928524077, 0.717229403309637914, 0.58089433084553943054, -2.4856700607640962453, 1.3580258343509257202, -0.014276239178605801619, 0.0071599745204705914775, 0.023752435342028956067, 0.028631007923775160517, 0.016011668911457055398, -0.01632477864425347161, -0.063691502717952319657, -0.11679773665653195303, -0.089321748231459677236, 0.089737455741491223282, 0.51320735275167195244, 0.54048197703331690711, -1.9799603093158582023, 1.0752202783111703965, 0.0057915440380951624735, -0.0070093170900220793523, -0.019480548848108392851, -0.024503094352406874085, -0.005269802355147270613, 0.053425263495138791692, 0.12936234182608194465, 0.11024423509986179537, -0.092929658262078607134, -0.41739558500743451086, -0.34866030002388409548, 0.77538390528825884562, 0.12221167282729067383, -0.2779169246133841864, 0.0048825252954557809074, -0.0048853862244905860851, -0.014196943327103651519, -0.018636648468026224129, -0.0068918267529324457599, 0.033346014181416895605, 0.091795227331476089705, 0.093034224986377922484, -0.022784956082848029091, -0.29683378496455992135, -0.41665234438405174267, 0.59278062897523986496, 0.292361445382867291, -0.32622412262770938973, -0.0093271212374821544505, -0.0083981338416980726852, -0.0045210459462247205506, 0.005394177127599963581, 0.025306203988347285294, 0.053356690899802466932, 0.070184468016895387721, -0.0017281727039402431604, -0.18894089293963847731, -0.28890932609922465035, 0.14924269723501651375, 0.78461344383707209982, -0.97135575036589039133, 0.39793797191066060392, 0.01015386107630424431, 0.03440388626851900028, 0.04272507630915237109, 0.02218916077708200929, -0.041751138188970672582, -0.15031837845515957919, -0.25081496448055173865, -0.23772089582193220436, 0.048770182535757329589, 0.7117564132802909338, 1.4806072595597046249, 0.18790941009847200727, -4.2966299562202179274, 2.4291396004544334453, -0.0084667682006116917054, -0.0016763178411656504564, 0.0029206606378921083286, 0.0020608813527106021289, -0.0010136944447911785949, 0.01076947022338701869, 0.05549813385718869313, 0.081277609104732181922, -0.042066722200993542091, -0.35508657364163559444, -0.13844249933548555265, 0.85139390667782266497, -0.60227156805586479216, 0.1517667274421384549, 0.008909212372789757084, 0.0048977814879989090002, -0.0084205762380701737468, -0.027997158971690715157, -0.032753563594081415433, 0.0068081280379150132392, 0.082121610628672445098, 0.1009486271764337556, -0.013786349369253905056, -0.28163420255934640268, -0.19534630598441252158, 0.75731622600693304648, -0.43317481568507371081, 0.02795563437869888479, 0.010047606316951779651, 0.0050088820747640682915, -0.0086238319140082009273, -0.027914068874293058303, -0.033348364639190733461, 0.0023570211268325678283, 0.072589812764225686292, 0.10103334659238552562, 0.028518725393757833608, -0.23662084735946622982, -0.31776009083701345359, 0.6431445926041637895, -0.12175618648474403893, -0.12183341358948907018, -0.0023921038844167863587, -0.0064890488467700593173, -0.0023938675960070727317, 0.010071317794277194133, 0.023569308319889223369, 0.021090222811504553274, -0.018718102643624660864, -0.13590171820081761722, -0.25909490565854570443, 0.013643354454548051874, 1.1046011815313478976, 0.86149144904628893737, -3.2604644186491613667, 1.6504644805679549169, 0.000242293079121136813, -0.005859140996439984872, -0.0061601426798473931865, 0.0014380089203961995276, 0.015933871880760376583, 0.027770335612437725692, 0.012473496677600035693, -0.089676097243030508666, -0.23425550301948569465, -0.068319146758571008826, 0.85150561701376603629, 0.85401356120561067176, -2.6667946963640818403, 1.307635565233102648, -0.00012357872673842979893, 0.002077097490240942864, 0.0051185909810520712968, 0.0061299732213792902658, -0.0014114343644172407805, -0.023472773849006919183, -0.043946834194187919809, -0.0077659437570104464937, 0.15628613632081167495, 0.11471805870039836206, -0.43501737235931153824, 0.092442606864699100599, 0.31816072053210603388, -0.18443192139019221054, 0.0029596756909765677829, -0.0028403349793068030996, -0.0091542122345940699774, -0.011881710309594456162, -0.0007599377417533193825, 0.029069392959018277439, 0.061657932910372185986, 0.062396990664168650031, -0.017194236829494738417, -0.31890153397135517777, -0.24606865150298937173, 0.81541408127997405675, -0.3876371335650770078, 0.027858904459963702849, -0.00049974770915505459593, 0.028484313309453876406, 0.052014850685774563499, 0.060411723378382256588, 0.037326363723376679293, -0.041240006643105737993, -0.19955162286263633531, -0.41895274446993252537, -0.55455426440058930293, -0.26213681292743712392, 0.72872885527118502758, 2.052046135365470203, 1.9499300186331962603, -3.6045719345410356382, -0.0021404012758645461593, 0.0019579275303500216429, 0.0060565028858359419919, 0.0087644496203721049266, 0.0070013928653162001495, -0.0046485437952427889272, -0.033859072697888015213, -0.074367079418373699395, -0.06407646351681629826, 0.13316670272090805338, 0.34265445255913812295, -0.11503346542583495205, -0.48070430109482098491, 0.27291746507636760466, -0.0016843231162103578999, 0.0016445768410759209564, 0.0040373544147571417684, 0.0044960565378739970888, 0.0028296847458428760202, -2.6175322527879470928e-05, -0.003211514498343180906, 0.00030202215079016869791, 0.0029144663824336621005, -0.080255402470915657154, -0.28560064540775270681, 0.082336165511609743506, 1.1337512729143559564, -0.87992108712638339441, -0.017386328764955792658, 0.01058947360674595585, 0.035949577444358636613, 0.049671953188893808173, 0.037307656589319337925, -0.016324455732039045569, -0.11462704534765262576, -0.22147883319551397996, -0.24137877679210029136, -0.022019909023733900966, 0.41597454004513390036, 0.66504104177579448365, 0.53105896368175187217, -1.1689437937310085402, -2.3192977963810018784e-05, -0.010505240759283445903, -0.0093741712771188093895, 0.010065181035470242496, 0.04649448500640858406, 0.075823462772944758004, 0.042850776137208321237, -0.10287556043602003042, -0.30664175710745772419, -0.33332681886865200838, 0.0026860408612827874932, 0.74774406558604988859, 1.312469416646159015, -1.5198218628683672904, 0.0005650995412190749868, 0.014505278734647897992, 0.014067348876604841151, -0.0062703256308116045448, -0.040855694734348607655, -0.06374311651188679928, -0.03778661507053414359, 0.052104830364957908928, 0.16669496323346352784, 0.22346383713301856688, -0.0045457862811000152373, -0.5300546774122678606, -0.23306912747898150817, 0.45679661537728794007, -0.00061573397285035803318, -0.0092633393933826274624, -0.010133423371334162114, 0.0027346116593567346489, 0.031312648761497044458, 0.061253962103851040955, 0.054933113922345067071, -0.022409950746430228052, -0.17060555406329605543, -0.32271336471975980986, -0.26338169106208142445, 0.50642393912274363821, 1.4412914327880901322, -1.3265216323412079014, 0.028438284364475110988, -0.040632373809889023308, -0.090455977728656655756, -0.090654903447718510057, -0.00541386345329857889, 0.16905589592454564607, 0.34805132886431083827, 0.34750211403060665694, 0.0134907066791177567, -0.49120642886466869825, -0.75386833900973504452, -0.5191461017491373342, 0.47083468084836271883, 0.69126548965789891277, 0.036168874048046667502, 0.0085180875908377143985, -0.042973870553486712143, -0.11719495356297204847, -0.18761992016882442225, -0.1871653557909183363, -0.013500244592139038977, 0.43423433602922695274, 1.137572188698065645, 1.7332236222013555071, 1.1760930132892641886, -2.5027115378487647668, -13.108469677469608783, 11.938922675813715912, -0.026767156436771001593, -0.013670113232425572941, 0.0096107464533362173958, 0.04105871172107875694, 0.068640149090274465982, 0.070689305100088431488, 0.02452749073580285516, -0.077809278469768330955, -0.22348878107114561464, -0.3313633449567980338, -0.20568144402966090367, 0.51300373545489097449, 1.358025834351233474, -1.2221536395815535414, -0.021872986814088002383, -0.014276238995713830954, 0.0035389782207639820973, 0.030837628556673313956, 0.057901109391485280908, 0.066102949820090045607, 0.034695294441219819148, -0.047489415857788756792, -0.17599786564552619139, -0.27608070143021518073, -0.19069971036248203888, 0.34002284096232021637, 1.0752202783528062024, -0.88815827615876541401, 0.010381881859235475227, 0.0057915440021033114809, -0.0047591860257324303946, -0.022895580310525551898, -0.04486744750209208793, -0.051657606162817469586, -0.0052891117622377980395, 0.1235023379081392203, 0.26685220904415690724, 0.2281812628079673555, -0.19901921885413839663, -0.7675140907076080099, -0.27791692463529710233, 0.7686736981354987952, 0.0082266370854644464694, 0.0048825252759912875222, -0.0030603022836370031949, -0.016620699299432974139, -0.033224148240555098632, -0.040193010534282723467, -0.010969267120677701766, 0.079105086843243424544, 0.19150694724770503385, 0.20384681444533225836, -0.076198816753577050531, -0.61330835944229211698, -0.32622412260983979504, 0.65429739441827061697, 0.0038759206255075742686, -0.0093271210280201077847, -0.018111768444254699312, -0.019286682576167755304, -0.0092311073101981680478, 0.017381558350118800349, 0.065428654891891660506, 0.12771122011131963436, 0.13404168248887871062, -0.042604190170404884763, -0.39450513303086803818, -0.43286485941374597264, 0.39793797163707522735, 0.20666232440652049229, -0.044491692915578827383, 0.010153861144982397752, 0.0627070322719467943, 0.10065465201627166181, 0.10268312227752149357, 0.038117476924490419077, -0.12306655176015891817, -0.36676370659845825006, -0.58009699864222896704, -0.50038418071257018127, 0.24996892528303571446, 1.8525738340950721472, 2.4291396006249628137, -3.3426912938312751677, 0.0034291095818068514575, -0.0084667682103454558473, -0.012542429271630167048, -0.0070970546044669853558, 0.0026667706664187640873, 0.0077665996462761323349, 0.011147925329434189498, 0.042165066120524605231, 0.11108599239440725137, 0.11074316502598904877, -0.18713155646580134639, -0.49036490593075077182, 0.15176672743280258948, 0.28444581619386544169, 0.00043558496747268967125, 0.0089092126042064276326, 0.013092415787742617272, 0.0057065278188365743672, -0.021305610465650069718, -0.059620298433411898775, -0.067055564935692826833, 0.010197530924959679818, 0.15041035485156534213, 0.21389256032611084213, -0.024323540743593139585, -0.40233640761250422102, 0.02795563439265890196, 0.14423778962149544358, -3.0492453717265410569e-05, 0.010047606513871569153, 0.014691186981668307582, 0.0066612317137451759058, -0.021276917424215172764, -0.060505768135955199927, -0.070957616593019232076, -0.0017280600509001538961, 0.13415618605857185242, 0.22929286468049542291, 0.050083593330538515864, -0.39446515417401262216, -0.12183341362944762054, 0.22631878372362690399, 0.016660246875705692071, -0.002392103878428339641, -0.018357868038039178854, -0.024066400799171872721, -0.010785078310425634876, 0.023753452911092497873, 0.063896305094327809915, 0.07018395183825072714, -0.036479503745668273584, -0.30515757802570314672, -0.45013200069035785855, 0.39829920413954356118, 1.6504644806936619172, -1.4104485217267048736, 0.014527324315787121592, 0.00024229307011822977767, -0.013452897000154586532, -0.021976489342912186586, -0.018084734790089096457, 0.0049778678403435260708, 0.045373967040086941871, 0.079004952838247322067, 0.024401216113708636135, -0.2025838546841880472, -0.40483801049048034892, 0.18773138692097435243, 1.3076355652666307172, -1.0265501463771107282, -0.0064129311024959650273, -0.00012357854486128388668, 0.0054050335391556316583, 0.010215920423662360264, 0.013794979285286890214, 0.010962842532660945286, -0.011043910496962006801, -0.057897279614306171314, -0.087970687277026074491, 0.03281813162387404309, 0.22851399974993363839, -0.01755708783615114732, -0.18443192148408357744, 0.062057810859848176743, -0.00055058219715611064586, 0.0029596755425471505882, 0.0020325770631489112686, -0.0055212477108483818539, -0.018329406838275049546, -0.024658183011501369486, -0.0034680591143291614881, 0.056636448467897407277, 0.13162665224767558514, 0.14352155514112982515, -0.13024687901437148607, -0.51787977165686871484, 0.027858904594540469246, 0.35480830946734154274, -3.3935844866350479698, 1.263223388843832895, 1.5369934524527064656, 0.92132922132440686624, 0.35802613204890065823, -0.047445516143155293254, -0.27173464995534374333, -0.33152756825133072871, -0.26975362508502243042, -0.15734323076237391459, -0.048873678054500491186, 0.02969828500308803812, 0.074803883718228894595, 0.093443155602398977888, -0.025749663812361042042, 0.032651151517135615365, -0.10576602064800970371, 0.082398139281731966332, 0.048596278200206961817, -0.013304002886184452745, -0.010062682601051062964, -0.0024660530048700507119, -0.0049736884401608508904, -0.0069006002465053107447, -0.0038049954632862728783, 5.1513389795196330365e-05, 0.0022039049088495121613, 0.0029337626741712786137, -0.10924291405511062092, 0.23207203784030805416, -0.17742192420762059601, 0.0006141537968841580394, 0.073339221536859941497, 0.00067260940317446322664, -0.016234806029693911456, -0.0029956457953847063537, -3.2342985472198348008e-05, -0.0033804522726981118576, -0.0031824408366630555986, -0.00063226199744103553802, 0.0013668222171211840549, 0.0021701950771163503948, -0.47651165473250328697, 0.16305934351665848925, 0.066215860469127127419, 0.26248899208400122696, 0.19521853630471983077, -0.0013559701529483256244, -0.1094571639850499406, -0.11494796930337469743, -0.068706074526815055781, -0.017947282698788657312, 0.015972621864802683672, 0.027616546619961868797, 0.020778497760574887732, 0.0025929772269555535248, -0.96831225992347902398, 0.65208530146313448128, 0.35609093474353309405, 0.18542834710092592765, 0.011326058181712086309, -0.12817832939746451326, -0.12592647910661680655, -0.063673713256620828593, -0.012674256097481090971, 0.010821621851933964364, 0.016579047895839919141, 0.013785443167871495873, 0.0074694813057120520647, 4.499673876452851579e-05, -0.67544051287341588807, 0.65348668524668263036, 0.23685815667439968912, -0.018268463977015439448, -0.12746066063093525189, -0.11218723990507370458, -0.028893347562903917702, 0.021408201699242391386, 0.024900912234078812629, 0.0095457409023045544649, -0.0010295696845476268438, -0.0029038209271671968498, -0.00081942665194468914964, 0.0015451783164692439558, -0.11253507780002969252, 0.40874153128445128713, -0.21143083989768032827, -0.1472601996656730805, 0.01184985068186352429, 0.024571731405627249972, 0.010957513918997885058, 0.0099400464435677963332, 0.0114822169186547373, 0.0073172251233237037976, 0.0019137386362458213337, -0.0015723922131751686369, -0.0031367953871882345469, -0.00369044955397914657, -0.44378342192654268716, 0.93449391415694371688, 0.081379426924455100512, -0.33854343638020800711, -0.28804525942798647753, -0.11843427072476098161, 0.00536873789076310233, 0.059810586081701838657, 0.075289773603805135171, 0.064131850605897988693, 0.037059310092324680708, 0.0067343803212194363467, -0.017612446077954355128, -0.033568872731157925471, -1.2221536398237231591, 1.3580258345569811151, 0.51300373537713939154, -0.20568144286964126777, -0.33136334638627662486, -0.22348878132000746244, -0.077809278112740395961, 0.024527491569737172594, 0.070689304138282352263, 0.068640149878560194341, 0.04105871189815172434, 0.0096107456491527250286, -0.013670113306270164841, -0.026767156130879737885, 13.002855565459478626, -12.316734165939713819, -3.1876182371819101746, -0.10278368800054805965, 1.0450084980539506851, 1.182043069316220274, 0.83153177521314236564, 0.38413225215147339675, 0.045199416099024558591, -0.12765509295037144599, -0.16215940803925477098, -0.11941236295472291984, -0.049900379175644438146, 0.02054125630331537361, 0.01383899285029194931, 0.22173218640728664663, 0.20686487866594968632, 0.039099213419817820114, -0.3188333756374911454, -0.31442688906076665667, -0.10852594648213456652, 0.048999279238967775874, 0.11887196887888865082, 0.12009378918844371142, 0.080235689563095291543, 0.025999303723382088255, -0.024812839087232370361, -0.06559481323125068708, -2.1940680124194642531, 1.9491704741652233768, 0.99024802530494471231, -0.14294698260949867397, -0.37170184807689254924, -0.27953376547541708419, -0.12893345211501991421, -0.0081783889582435036203, 0.046920811579103646749, 0.048155064091505564439, 0.028147276806897034152, 0.0080314696685044595359, -0.004291567658522545993, -0.0086853949815050585787, -0.5556458572935639717, -0.16092673406330407215, 0.61422133491739605571, 0.61537831958388700127, -0.10021353042748859519, -0.35874632696048330827, -0.19379796821941958074, -0.011565860594272574582, 0.056462308279356270002, 0.046684514715203309643, 0.018690789782912334976, 0.00054233498309146283868, -0.00523040669944107358, -0.0036601994183737621147, 1.239759830220376946, -0.4613156793054921101, -1.0459291105425905233, -0.2961527659231231735, 0.22696215085398291356, 0.2804513238806052855, 0.15756122964987828716, 0.042087256443155751651, -0.015095193433624446919, -0.029732348169446302932, -0.024595469300368960619, -0.014402099410581637937, -0.0046324920072267463181, 0.0036067178839243110855, 0.96796828719981531375, 0.67930941275316791828, -0.7888101951122541422, -0.94817326153534686739, -0.43112667984038577362, 0.013103644169763936644, 0.21447627977547037426, 0.24082313508240696387, 0.18667204155672242893, 0.10344364449617428348, 0.024420855786858004094, -0.02987653764619059571, -0.053533997855554696366, -0.052625450953858995962, 0.3648559273042501494, -0.11256765595817841419, -0.49089850432818299408, -0.02367562712470819633, 0.21354348359575933691, 0.12268255802444094638, 0.0041533869394650034176, -0.036985732526799465492, -0.026379725973090201868, -0.0084715182741964007579, 0.00088885806030208859216, 0.0028778318463590895487, 0.0019516685543152856655, 0.00042933772613369940597, -4.3862861898807832262, 3.2171626845916305726, 2.2998589904335671186, 0.17488862163206631006, -0.5779473293593762051, -0.59078465639537303122, -0.38011140340213461775, -0.14869658137799149356, 0.014190663539365322432, 0.084684292504136476909, 0.086607533578382522865, 0.054468641683958957111, 0.013633507389820838374, -0.024236063630546113984, 0.85298840294084288249, -0.53944016722570942157, -0.6535901454327881499, -0.062846167546453141073, 0.17169854968911726911, 0.21344368476816993896, 0.14174129128598597638, 0.025935140366054099115, -0.040913546718290741411, -0.052967076988719274944, -0.038958561550278388019, -0.017863891751488669701, 0.0033026026730428358416, 0.022029097956536208835, -4.7352094816141629963, 3.0236604950770780142, 2.434673151268463176, 0.46992968621624942038, -0.35194374854679788545, -0.50450534869568008922, -0.40588116068061219632, -0.23589504967909372124, -0.085581623486860797234, 0.009987527538588347048, 0.053922125933704358192, 0.059577784644336308217, 0.041328956426476597119, 0.010955759846349340247, 0.5461619502011644256, -0.6909220903555036708, -0.52202836670034513666, 0.18540617489312760657, 0.34807594032366928616, 0.2767762948979010651, 0.12001492327738512433, -0.044002213511253697453, -0.12235278109416172776, -0.11454333207329170374, -0.065634138103690176091, -0.013111948569045676904, 0.026370957921862631795, 0.050548488514131635629, 1.5775903155558739233, -0.45233985471790916222, -1.3391778783805978748, -0.54034813209160703629, 0.22057259721543212572, 0.39963918936311948471, 0.2587084916793223055, 0.083389155405902665175, -0.014570436103315303339, -0.045639858176387680666, -0.042280162884182070915, -0.026326162659975255176, -0.0078817771597744328005, 0.0091152966178426886312, 0.63160339977807633538, 0.06390805669899474184, -0.70390189378270440201, -0.32743547607272732725, 0.081441584719178283791, 0.16013537900687899174, 0.11448402664864723577, 0.055200396802937909402, 0.013235552401881241213, -0.0086862290103415576437, -0.01515789939893695884, -0.013195784151120984115, -0.007619575216055207538, -0.00093326037949722541204, 1.2632233893813129555, -3.0416613801774894377, 0.20303099357651954726, 0.90548917009220919461, 0.62883646815260085816, 0.31688713892833164643, 0.076368699712979354333, -0.071830361791887203937, -0.15465184270959611568, -0.16214620595012280146, -0.11740145701460469319, -0.050628205969633147121, 0.01641427330288368347, 0.074803884015019000975, 0.032651150885129834323, -0.20709296856728229286, 0.35067500119681438431, -0.23216765585561738305, -0.01104595987895941496, 0.081491394811004716492, 0.0043720931203074485383, -0.014847760394972406989, -0.0038056397688338797908, 0.0010486587825142895322, -0.0025361850130436254698, -0.0032227932956987127497, -0.00062789962307262139722, 0.0022039057610270312501, 0.23207203660431510728, -0.67603455170699888743, 0.64051528475533103801, -0.12554059955066543264, -0.15267844663641422231, 0.070126439450828526034, 0.03374171364307461396, -0.014131358768532017653, -0.0095331036337943653169, 0.0015293681457658062431, 0.00048694759751345358979, -0.0017557662444262214178, -0.00096336083616103181798, 0.001366823440662717987, 0.16305934369140745455, -0.59480943506996775128, 0.47306159252227403256, -0.13574291950037686849, -0.027731294960463705179, 0.13353046988418942886, 0.078770265756617649511, -0.0060886745676007464345, -0.04934581932344832278, -0.047134220441576182847, -0.027762397981034164784, -0.0064385823900929596564, 0.010134086614462019943, 0.02077849890905510491, 0.65208530130595376839, -1.4019793753943210479, 0.48072806290878994373, 0.18038160606379213813, 0.090675627023367830204, 0.12143949659786984485, -0.00066205856221015352281, -0.068251651194477597207, -0.059757127340798528448, -0.026722881635617038582, -0.0039798418421221606542, 0.0063510133411817139282, 0.0089485629168256312221, 0.0074694806061091261265, 0.65348668515118946143, -1.2899208960584422901, 0.41008310570604267209, 0.22271970325720991113, 0.10445790229907957192, 0.019608833616485639439, -0.071990051095399590353, -0.06042202523407468745, -0.015579156743659909948, 0.011737306494841824936, 0.012494769681746838857, 0.0042218299956753267688, -0.00099108612935978153066, -0.00081942550820962958055, 0.40874153129556117792, -0.98862685369339009522, 0.73436318047782678242, 0.041468086172424617475, -0.19832583454743871187, -0.03477380382610852777, 0.020375911304099936228, 0.011644601780100452201, 0.0021906027994863580197, 0.0051631691490661157928, 0.0056164985399890111212, 0.0025727921997354191372, -0.00090326132262792408877, -0.0031367955677774391615, 0.93449391380199831048, -1.6949417034074392152, 0.61922013563750588627, 0.4223645501013344461, -0.031712142551597169438, -0.16701094684207518082, -0.11710911614309410522, -0.026528649621482124077, 0.014656296668684905271, 0.030078721842621756327, 0.032394661687614373369, 0.023673501408651528854, 0.005396454884897630179, -0.017612446201897336007, 1.3580258341372259956, -2.4856700601766732461, 0.58089433047289595891, 0.71722940345502261827, 0.13167272593428136496, -0.107136470672548334, -0.13791422912494649688, -0.083091444425417160247, -0.028646117136125616442, 0.012651066142027354161, 0.031169893617475474129, 0.028195041026214791302, 0.010516060004856560284, -0.013670113575403156775, -12.316734165792999178, 26.233509322388744778, -10.799777330159663791, -3.520432120200243542, -1.032900716780134065, 0.078300401859025980933, 0.52456700692807456399, 0.54573985756439868933, 0.37431675820093718299, 0.17127965830538927072, 0.013615593304921069784, -0.067437424526851197371, -0.080560465145544130072, -0.049900379650904713724, 0.22173218627077984078, -0.02892851660672288161, -0.066274172790767110763, -0.23885197242577899113, 0.25815149318785424182, 0.055122473386340853363, -0.15616858934585042817, -0.1073095762683694554, -0.020608847812014163825, 0.033591694601941834986, 0.053205212784916883217, 0.043818605294611706613, 0.014185781997918353592, -0.02481283870652351356, 1.9491704744815163686, -3.6516175214914166069, 0.59321272132485092232, 1.2289086938047617537, 0.27192614287407074203, -0.09978201195663681522, -0.15591525349012377855, -0.1262395160421752871, -0.062186247995548574186, -0.0036615966521328750793, 0.021445369699712640843, 0.022230569380758756493, 0.011250691204174403243, -0.0042915681297067071256, -0.16092673354222233884, 0.54227620561481981731, -0.51664743189606776586, -0.51637208161351455526, 0.57685460736820459715, 0.37932666200287967762, -0.093874671590156208434, -0.18345825825009834387, -0.089628764207997313473, -0.0019456114468036000628, 0.028055251114797091772, 0.021901441619926756421, 0.0068598195240319528865, -0.005230407532769928286, -0.46131567959315272764, 0.47181163375460505316, 0.90056683253053926297, -0.55527125033048152503, -0.55343399685866268012, -0.089626170473557170704, 0.1226591293910461733, 0.13238890795856558813, 0.069135292330061853816, 0.017366598067319074378, -0.0086253933371808720609, -0.014691130275914442496, -0.011268353434976390026, -0.0046324925769385622357, 0.67930941254841537003, -1.1405536822064565872, 1.0143557114224577287, 0.20349314625582129445, -0.46517992005621411344, -0.40251565407448497425, -0.13841318894076917734, 0.050712069425761827079, 0.11135669201761391744, 0.10876491012203791386, 0.078556449652911516934, 0.036010718695631181718, -0.010285142073691595083, -0.053533997303214148067, -0.11256765566941953627, -0.13224934678129968146, 0.72807904162411596349, -0.33032659600790148513, -0.32795854970564025654, 0.031240684125907258045, 0.11788861855184420979, 0.058409128500156691499, -0.0023139663382517119961, -0.018051528981187746348, -0.011707259947560885033, -0.0030782773637947598647, 0.0012537377967449512585, 0.0019516676787544489421, 3.2171626851455057405, -5.8653396560669799342, 0.33088325693928405391, 2.1824581739113648737, 0.80383247579740924049, -0.043172962020031346864, -0.26716288364021256729, -0.25623131916360719629, -0.16703556150357462551, -0.062998698662726435038, 0.0093702907739863228187, 0.039889754983881836603, 0.036771469654774520397, 0.013633506674445404136, -0.53944016714298503867, 0.74042292165084222511, 0.39510087539259775014, -0.5507564041647323716, -0.18948646086444076442, -0.022280748617592743832, 0.046481373425007188604, 0.10134407174755299952, 0.061454303768708334887, 0.0046746273433145302256, -0.020853318707679230348, -0.021017681415244016119, -0.010737623542092428197, 0.0033026028698536569542, 3.0236604949306733481, -5.7297386598628436971, 0.25179077289531381911, 2.1043784917398506984, 0.88641205197545858141, 0.079311866486652227159, -0.19658054782677689132, -0.24253093218416485466, -0.18872064720277559324, -0.10307088146579759802, -0.033094835402993308393, 0.011672144146487317684, 0.034362511779704375814, 0.041328957681257882939, -0.69092209037370910796, 0.84922011217301418551, 0.30938818920505395393, -0.61041134585120282452, -0.10389082646202058013, 0.081188143223362457812, 0.097337330346003780868, 0.10836423797677480207, 0.044224581802363160943, -0.026851990682375426189, -0.055538860679616969263, -0.04454817233522676001, -0.012800090374029501894, 0.026370956510531212008, -0.45233985424220557459, 0.38426326590189363586, 1.1161017963253887508, -0.48431437957471751288, -0.72746787209275953767, -0.22264751919800718216, 0.12824546437879025262, 0.19298669476879470874, 0.11251448427464819102, 0.032734617044051381174, -0.0092814469835997177871, -0.021719290522209399458, -0.018581194017430833348, -0.0078817777060589885862, 0.063908057226919112725, -0.42863887752113000174, 0.92835867819547224222, -0.20142897799006467818, -0.43653925790921083161, -0.096608842824044763309, 0.055537854953713404471, 0.073835693495041654355, 0.050538371441314351717, 0.02458118414805402635, 0.0047407793124958838571, -0.0055476182038166361316, -0.0086885736803341655221, -0.0076195753941393799466, 1.5369934523995814057, 0.20303099388070308673, -3.2838288078383706114, -0.12174595356393473622, 0.75307272393119073506, 0.63878531535651827955, 0.42283842198405552359, 0.19098418081663862123, 0.016917840050200021185, -0.1031048152370300508, -0.14269860042391954713, -0.11620337070759577391, -0.050628205385578627096, 0.029698284803072627636, -0.10576602007460519927, 0.35067500028953629831, -0.47987187912902323772, 0.37500555270029806154, -0.17493399430318518073, -0.031338915696627306362, 0.070353311434007484437, 0.0089262501919555245489, -0.0083526323068495219454, -0.0021085933184556574296, 0.0011682001219701131436, -0.0023934767509119145301, -0.0032227926898413666867, 5.1512265045172093286e-05, -0.17742192179923269024, 0.64051528283836978162, -0.89307358909093348398, 0.54257578291509600543, -0.036367754602744960246, -0.14481737626139157271, 0.051528697964309463919, 0.031760927747186300241, -0.0081463233976279431775, -0.0072486221898255473461, 0.0015658207353497547148, 0.00039013962863569044519, -0.0017557654945410867443, -0.00063226384334268180901, 0.066215860308001153745, 0.47306159269778824683, -0.93108754483669431412, 0.32515442510285569311, -0.11574262774733094783, 0.0011428213901591294244, 0.17052365799672097868, 0.11187739615052229303, 0.014891755968898495732, -0.046367844765422050468, -0.055848172599394590998, -0.0373691864381727884, -0.0064385816439727974747, 0.027616545477925809327, 0.35609093513181555579, 0.48072806253811056765, -1.6402374134410817597, 0.38472659115252783035, 0.26950494487938375654, 0.1377279939252472607, 0.14415757441111742843, 0.013991169629028459009, -0.061259603477841317876, -0.061766345901774905025, -0.032275598424028918709, -0.0080894409350636027611, 0.006351012811620386847, 0.013785444613231275651, 0.23685815652861405556, 0.41008310567734862495, -1.3825536318777551514, 0.44795882357869482027, 0.29713284821160057536, 0.10906121897175985458, 0.012592326078396153324, -0.075510160826491354302, -0.059654268955970604638, -0.015490032212301368258, 0.011158770185121657734, 0.012296856627745642032, 0.0042218309093656324754, -0.0029038221784989208529, -0.2114308406604332724, 0.73436318073363970882, -1.0518977799420132779, 0.62875090203557171264, 0.087689631281567009813, -0.17533566548715029398, -0.039949075392397830064, 0.0096231572479878024068, 0.0077662653286906403044, 0.001024797694128685337, 0.0046302794309378281223, 0.0054103361732107425885, 0.0025727915438392267439, -0.0015723915593567759814, 0.081379427691947531232, 0.61922013525131158751, -1.5606332430728506733, 0.6631848790091954271, 0.4754237595159834795, 0.0095011608115822127063, -0.14617008219593896623, -0.13357917312617917371, -0.056122253056760398882, -0.009770771904303574007, 0.016778063403320900293, 0.027805808397977894997, 0.02367350134894662117, 0.0067343802770676768413, 0.5130037360161503468, 0.58089432984304045782, -2.3529705190310368224, 0.58813189614934424654, 0.79874425620875311171, 0.20655810444043548602, -0.072008022605424576446, -0.14762914102277646666, -0.1089002168172485685, -0.051560784430148644264, 3.0630766412999277382e-05, 0.027234082603726669086, 0.028195040904425809625, 0.0096107464001618271576, -3.1876182373385941737, -10.799777329915754009, 28.756120194991201089, -10.564431132037940486, -3.973049352121834854, -1.5016842522225803958, -0.23497319789710680027, 0.38850177166873289814, 0.53821589838599481759, 0.4230264624676445262, 0.2262141348616477643, 0.047377942923743976422, -0.067437424353151309564, -0.11941236319995308934, 0.20686487864671071479, -0.066274172896703675839, -0.3998417150604499648, 0.24978398009657787338, -0.071217322859685294745, 0.28879622273760441331, 0.084853250284849357499, -0.16537094046390676816, -0.1478811142418804403, -0.064588521845070023808, 0.0023454291237356983402, 0.038329163256629328327, 0.043818605993527937059, 0.025999303092278588934, 0.99024802444662995971, 0.59321272190381513223, -3.3559422956797519966, 0.50813423516071964947, 1.2958855197347218802, 0.37530907293436266414, -0.05783343527153880087, -0.15632250188352544984, -0.13814343768575398919, -0.075434411884935187453, -0.011330921312587942384, 0.019753337168616946079, 0.022230568472890781251, 0.0080314712468489285468, 0.6142213338243482923, -0.51664743114449585004, -0.22342876661366553592, -0.1697950847201852298, -0.30307898241503666847, 0.56625551941487051177, 0.36915648089887320715, -0.093485019550622230899, -0.18544110584464326696, -0.09497423055631508404, -0.0060963305414529073145, 0.027278923166221148017, 0.021901440348415457393, 0.00054233670559571611151, -1.045929109312647709, 0.90056683170550722028, 0.35879127417758771568, 0.76628889322361470882, -0.57495185653426761352, -0.59267271169283608501, -0.1263086449197627914, 0.1027840650929061489, 0.13178391857061166892, 0.076730843273134147209, 0.025584185730430131916, -0.0040093121660658998656, -0.014691130678557619976, -0.014402098771110781461, -0.7888101950443124899, 1.0143557116725920864, -0.75182449669181139473, 1.1218967140741704647, 0.28786576566870464289, -0.44669638314870430129, -0.44812910373051151325, -0.22041886367339139086, -0.017575694991133039569, 0.076963487765881813396, 0.10327041407701831321, 0.084724236526120239033, 0.036010718672878624935, -0.029876537368259091509, -0.49089850477161894871, 0.72807904174555149179, -0.35587350813725099341, 0.54916536075534694294, -0.3072949059786582171, -0.30767974063941017082, 0.03293547095369703942, 0.11441700648544672614, 0.05996760388515767759, 0.00036513422611197596246, -0.016813440608273687304, -0.011857814416330079293, -0.0030782784034304429624, 0.0028778329966669818907, 2.2998589890719367546, 0.33088325765516918286, -5.4693326942449136396, 0.20237098842706810475, 2.2448731602318479084, 1.0050205428855545353, 0.10240780048028419458, -0.21611932211073003884, -0.26277652446839450118, -0.1973917884118131516, -0.092825829605861609473, -0.0077359988735054008302, 0.039889754333826174526, 0.054468643301145910907, -0.65359014586949359948, 0.39510087587097891815, 0.52986812433781582676, 0.45915380305243325143, -0.61363197487628573512, -0.2877414351432425188, -0.043582589784791903043, 0.046449165044848195194, 0.10906120973479499492, 0.076723524021215683266, 0.020006475537345302929, -0.012464295222999077772, -0.021017680841832198091, -0.017863891978481660078, 2.4346731516924009497, 0.25179077307608077607, -5.4685627745502642938, 0.036902496372433563065, 2.1037636200833538247, 1.0506755323508090161, 0.22941257058736086449, -0.11073337174922537329, -0.2105713390494242665, -0.19452400367905575695, -0.1242457139229605767, -0.050412125764017962082, 0.011672145948087672696, 0.059577781596597247094, -0.52202836669670615866, 0.30938818940283452363, 0.524548147895096184, 0.31429001060151234848, -0.72310417352282230219, -0.235598677043706628, 0.062885180344397828556, 0.12722533951843284505, 0.14849286582809179125, 0.079359742400385957506, -0.0049993103239580061192, -0.047678978924330484013, -0.044548173290980916383, -0.01311194775375914244, -1.339177878670056332, 1.1161017965136315055, 0.47330676574393409917, 0.94885420798525355046, -0.56188360782243373936, -0.78778859431277437064, -0.27182637425224082106, 0.096297367626485788628, 0.18889182090635439626, 0.12508137163134266934, 0.04767130125477805902, -0.00068423291240912210986, -0.0217192906275423675, -0.026326162836885103058, -0.70390189433909799988, 0.92835867860549203812, -0.44892908659123270665, 0.79970154446780850055, -0.1716101948667274335, -0.45073627888689676846, -0.13031971097658709424, 0.033698602982120046123, 0.068232971099742245613, 0.053020026908507170149, 0.029659153647187089242, 0.00841616481069950087, -0.005547618773482664023, -0.01319578440080950224, 0.92132922123792171387, 0.90548917082177049487, -0.12174595418493036614, -3.3750136390334408709, -0.13434555938602463221, 0.80271127119126006733, 0.69721129949484139665, 0.47201616651760197607, 0.22334862832846597014, 0.034266125724258474083, -0.09699674698451647703, -0.14269859984144525611, -0.11740145703509116681, -0.048873678746553399221, 0.082398139792233177081, -0.2321676556853294604, 0.37500555245371292035, -0.44611290825442817543, 0.36516446150156139927, -0.1817798710241941007, -0.032135533840801498717, 0.074610973692729370965, 0.010661443413507774511, -0.00830425388150099128, -0.0022121128275591952356, 0.0011681992658749123376, -0.0025361847650142536892, -0.003804994550448090665, 0.00061415216396231548344, -0.12554059865235675519, 0.54257578417106666624, -0.84596185590671857391, 0.54841878781547082067, -0.043320858790238711555, -0.14887560493782237003, 0.054101951301735881628, 0.033936209035313677496, -0.0076642735043661739183, -0.0074863173850748879201, 0.00156581960954284196, 0.00048694862692368610143, -0.0031824404016293069905, 0.26248899224362148841, -0.13574291906482258141, 0.32515442492473595193, -0.92218635872432030975, 0.36960361470241176463, -0.076959387124682476777, 0.01441264101293401878, 0.1725619263905464329, 0.10741041313446898775, 0.0081671672655245913797, -0.05182138589643893456, -0.055848172167382338638, -0.027762397399800055331, 0.015972620465176300131, 0.18542834702413066306, 0.1803816061285693495, 0.38472659136283460235, -1.571128689631394959, 0.40774885706476426961, 0.27078568708394223918, 0.13242329835900557011, 0.14485381871376801222, 0.016416790653174088283, -0.060478122637464271083, -0.062733067799377095763, -0.032275596943060258504, -0.0039798420157348309012, 0.01657904692484082898, -0.018268463403635682579, 0.22271970287582235071, 0.44795882352828314543, -1.3424440982060097394, 0.41738146087575883625, 0.27128350153012037449, 0.10238492870960289038, 0.021652010624156423774, -0.067233902889113614232, -0.057304784585224574633, -0.016223483697199662934, 0.011158769659083311171, 0.012494770208129132513, -0.0010295698088615356201, -0.14726019837291426384, 0.041468085727714897037, 0.62875090129254673776, -1.061793439099393721, 0.63493585963741783207, 0.09560457798202172397, -0.17864847372946421333, -0.042753132027408186078, 0.008345621416362357034, 0.0073923703421856059095, 0.0007190471354815161989, 0.0046302793311405587157, 0.0056164990211480447443, 0.0019137382371762462051, -0.33854343778139961429, 0.42236455051595933607, 0.66318487978355600276, -1.5755953824476991443, 0.64090718213047725982, 0.48012809251746413342, 0.0057016440005478175035, -0.157376400249449655, -0.14573545577500368609, -0.063717554921575020321, -0.013089066490241746399, 0.016778063478153380783, 0.03239466202271080314, 0.037059309424580424985, -0.20568144425940573794, 0.71722940394016476695, 0.58813189659416897914, -2.3270977838409829808, 0.57567494655356976629, 0.79262907949962102538, 0.19633794697194550949, -0.080689925552607025261, -0.15347834578058511257, -0.11148189553089873916, -0.052797588233740956609, 3.0631610670614484171e-05, 0.031169892978421608748, 0.041058711890480596718, -0.10278368741057535007, -3.5204321206138717848, -10.564431132489383813, 29.078622526297522199, -10.530264504855598062, -4.0479286830499034977, -1.567696391535613909, -0.26684674343189285217, 0.3795913194928954848, 0.54383399878911320435, 0.43351131988821278718, 0.22621413427681991148, 0.013615592239005103181, -0.16215940639708020798, 0.039099213605648844294, -0.23885197247219080996, 0.24978397969116350463, -0.24846020912068578212, 0.26258938349989380256, -0.083513266492512006067, 0.26089890854610486848, 0.058962801589061752949, -0.18695955808423114775, -0.16421008076529064468, -0.074308154406665605229, 0.0023454291954755346339, 0.053205212849800564978, 0.08023568920666181159, -0.14294698264882005345, 1.2289086936704574082, 0.50813423516403044555, -3.3354122651862794235, 0.48812758840377440395, 1.274671653154920703, 0.36883464264986220993, -0.052688735004352046931, -0.15064760409256469642, -0.13464103600319521359, -0.074461806872227115361, -0.011330920597445774678, 0.021445370266676051091, 0.028147275746699697979, 0.61537832054596808273, -0.5163720820838358927, -0.1697950852270227462, 0.036940106229191436993, -0.15748525794771292552, -0.37240000368280412602, 0.51685049969743324105, 0.36730247397780746432, -0.078743111651010389007, -0.17622044630535546594, -0.093325157807445352498, -0.0060963297155003006583, 0.028055251667761824802, 0.018690788546570726025, -0.29615276740854529214, -0.55527124992296039707, 0.76628889440724656801, 0.26501555967578105388, 0.7674351087029331886, -0.54271787534912763196, -0.58399764287062017054, -0.13176277471383415762, 0.095802220380823699397, 0.12993998984910315464, 0.078055681387963926632, 0.025584186319293135414, -0.0086253934847939153774, -0.024595469327097628509, -0.94817326219569009194, 0.20349314597626028123, 1.1218967147681935259, -0.80336261943295950161, 1.0706071141241104705, 0.26923285942423519357, -0.47276916754706060919, -0.47550193172643251405, -0.23809173521796445083, -0.022205736721753105423, 0.078654741003542485989, 0.10327041394175986988, 0.078556449787389542871, 0.024420855896500753834, -0.023675626838246404543, -0.33032659572814560001, 0.54916536025075723959, -0.37687569023642297772, 0.59406467146902941501, -0.27109698709391272242, -0.3085580520179718178, 0.022002938775930139248, 0.10797568318276787314, 0.059348220878053747906, 0.0012370394698376016573, -0.016813440231352453497, -0.011707259837462621194, 0.00088885768080298652692, 0.17488862321650022746, 2.1824581734884143103, 0.20237098730125038792, -5.4132556646544873757, 0.20658204317165379527, 2.2480872247090761817, 1.0134714905290944653, 0.10907324730509526667, -0.21646176087139235023, -0.26776418441227117784, -0.20350067416062439518, -0.0928258281176750355, 0.0093702905234972766468, 0.086607532000903819647, -0.062846167063366459304, -0.55075640465979580629, 0.45915380281715223454, 0.43150517697242785076, 0.40922997155497820021, -0.59151386264301941686, -0.27141609846665426531, -0.038529309034831538128, 0.048765675752368897733, 0.11346890943855614253, 0.08112159418403552813, 0.020006475267753731639, -0.020853319903462036455, -0.03895856015358156732, 0.46992968619646008355, 2.1043784912338985293, 0.036902496807028062342, -5.4463893371539846555, 0.032939609176449169414, 2.1170517119990752697, 1.0819152201077508746, 0.25834353448279878673, -0.098783940364714828419, -0.21255290274832278063, -0.20123196425380962538, -0.12424571587854543742, -0.033094835847512493743, 0.053922127876993644624, 0.18540617553351820823, -0.61041134643647843028, 0.31429001076226342759, 0.40138441669671709455, 0.25380398931857672817, -0.69503709783244460585, -0.1996604879849482872, 0.087547036947177359067, 0.13977799868063434108, 0.15483642972918498781, 0.08286763461688706478, -0.0049993106573198442533, -0.05553886126256693534, -0.065634137352618998618, -0.54034813207417631276, -0.4843143794189486151, 0.94885420772460371541, 0.29524113568419008136, 0.92371041717851365416, -0.5206979275611112179, -0.77033373831306817259, -0.27522897170833626745, 0.089700569667473042945, 0.18815846024180679641, 0.12777208530608924475, 0.047671300801059468544, -0.0092814472233307219323, -0.042280162092886597325, -0.32743547681541151917, -0.20142897758772337902, 0.7997015448795081749, -0.49790803681957201965, 0.78771268072153433071, -0.15996581225450726138, -0.44782510586587137569, -0.13326069881050028876, 0.030271521727872587743, 0.067503834692171763021, 0.053898350661788177463, 0.029659152632740007349, 0.0047407796749930073213, -0.015157898722482623585, 0.35802613222927143211, 0.62883646736497522678, 0.75307272524120316692, -0.13434556032251621094, -3.3228625586321380858, -0.094203016392075039631, 0.82524068386724713786, 0.70696434257547724478, 0.47611616195860090306, 0.22532539603483461854, 0.034266126006587238617, -0.10310481450698384676, -0.16214620562666012815, -0.15734323196633354103, 0.048596277805624969737, -0.011045959863775116697, -0.17493399467147441317, 0.36516446235004751397, -0.45427878695819817745, 0.36806887144775751741, -0.17703407399221637863, -0.031912369724734532261, 0.07396134618309402331, 0.010529262950031467727, -0.0083042529437688545474, -0.0021085932411165995287, 0.0010486589938839717059, -0.0069006011994239063387, 0.073339222396436948292, -0.15267844721591714019, -0.036367755634279956556, 0.54841878855680814553, -0.86243302194242765779, 0.54550757203280730856, -0.038406747578603538107, -0.14745021422538248657, 0.053701722783676394957, 0.033642432736474985056, -0.0076642723896240038187, -0.0072486230839758176644, 0.001529367493814903908, -0.0033804516462274964567, 0.19521853590569407388, -0.027731295680995110553, -0.11574262713668018177, 0.36960361566450639081, -0.88484004170422525526, 0.38551432975036364903, -0.077046648346103807636, 0.0035171511300080560147, 0.16089496019785640146, 0.1007939220928556201, 0.0081671664754303929801, -0.046367846581371732761, -0.0471342203791868114, -0.017947280896660477989, 0.011326057311134885702, 0.090675627697885841094, 0.26950494459307700046, 0.40774885760046064576, -1.5795375784474676717, 0.39320717113128178521, 0.26271157860179172872, 0.12857404831625318664, 0.14412160988212810442, 0.01661625662490250413, -0.060478121949075175356, -0.061766346652592410416, -0.026722882021974678907, 0.010821622461424830128, -0.12746066086916854609, 0.10445790257470216611, 0.29713284780236332017, 0.41738146123331903992, -1.3804121287931947482, 0.39927791053309119595, 0.27349402371340097995, 0.10938927718396358746, 0.02606606083568263621, -0.066244511149064377875, -0.05730478450642781385, -0.015490032171869170738, 0.011737306471185461609, 0.0095457408336436216229, 0.011849850306318581925, -0.19832583463553665748, 0.087689631396784428752, 0.63493586011486113918, -1.069650082015058068, 0.62744503404066720442, 0.095880950312055276963, -0.17768857457924697174, -0.042633729261125630217, 0.0080574036852472039061, 0.007392370374405342337, 0.0010247980590683267773, 0.0051631693379484778009, 0.0073172246406462198784, -0.28804525800984914774, -0.031712143490607996865, 0.47542375954946725081, 0.6409071814828575242, -1.5958099665809060141, 0.61646553816290305949, 0.46365978673351948558, -0.001727246581898850275, -0.15889588099375695496, -0.14595626476784653858, -0.063717555616548904318, -0.009770772312025166606, 0.030078721629969260021, 0.064131851727457889201, -0.33136334483542495555, 0.13167272550794020813, 0.79874425576816598937, 0.57567494550615538884, -2.3609374827494047366, 0.53572585648625981047, 0.7701599569790269495, 0.19077009291474850139, -0.077674712100475093446, -0.15059596054198193249, -0.11148189525754174933, -0.051560784317015134759, 0.012651066712935265385, 0.068640149122405430537, 1.0450084971335138473, -1.0329007162166079503, -3.9730493513291484931, -10.530264505265845898, 29.251078495234381194, -10.398801298809544846, -3.9785892562782256832, -1.5482155628526252933, -0.27060089255135255648, 0.37469343964088408505, 0.54383399854989300515, 0.4230264628611920541, 0.17127965808299486228, -0.12765509276732459054, -0.31883337584767823536, 0.25815149328912412319, -0.071217322367448931963, 0.26258938304806989095, -0.26142843820285310841, 0.22733333533601407517, -0.12058289402262811763, 0.23716354158746622005, 0.048561672916204612349, -0.19110671712070925743, -0.16421008137039749819, -0.064588522089651934088, 0.03359169396677996744, 0.12009379057261841683, -0.37170184695537999975, 0.27192614266142534074, 1.2958855188912545842, 0.48812758815647050437, -3.3912362583570332397, 0.43909508469233005101, 1.2579294540470356001, 0.37150760397028104176, -0.044185046630007329682, -0.14522748821778858064, -0.13464103577782920995, -0.075434411369324300134, -0.0036615965013992966826, 0.048155063327912868421, -0.10021353115898572195, 0.57685460789017284355, -0.30307898248875581038, -0.15748525759908807431, -0.011016767601318866454, -0.20179325252790392131, -0.39118649812690348133, 0.51856335912686946088, 0.37637147244572849436, -0.073069943578001506079, -0.17622044634865413615, -0.09497423115979465591, -0.0019456120418576000608, 0.046684515911130868016, 0.2269621514098455195, -0.5534339963394865336, -0.57495185778570978563, 0.76743510862372732451, 0.28857344345347835102, 0.79310608375018964988, -0.52708695261335591908, -0.5821187860043952611, -0.1349442523338740263, 0.093776424055533949553, 0.12993999063804781424, 0.076730842969691834266, 0.017366597675772470211, -0.029732348060536276102, -0.43112667993503772657, -0.46517991990528423374, 0.2878657656965015188, 1.0706071140179624912, -0.86038565245589682107, 1.0281328182601390253, 0.25324617745873129193, -0.47110821329899910115, -0.46752347938993482623, -0.23280256284684291979, -0.022205736437881817835, 0.076963488511620506416, 0.10876491126357896999, 0.10344364253412283161, 0.21354348329465644429, -0.32795854981091260294, -0.30729490579590551524, 0.59406467178972865995, -0.34747678585402630436, 0.60229891978648275597, -0.27238333196762415378, -0.31387550283233933701, 0.018262890417330923665, 0.1069137225495352117, 0.059348220940752441366, 0.00036513403971909913492, -0.018051529377481369265, -0.0084715177188742218917, -0.57794733086949412471, 0.80383247653053047088, 2.2448731609016836508, 0.20658204359362525304, -5.4525111763089135763, 0.14133270058060731778, 2.2053602980897442087, 0.99827673418386642101, 0.108810313042373763, -0.2147748329781025145, -0.26776418567940674098, -0.19739178880359636858, -0.062998698992749596615, 0.084684294109092672631, 0.17169854938981288228, -0.18948646052039197496, -0.6136319745676941384, 0.40922997118907655789, 0.4578986953762009815, 0.44207477584466414733, -0.5736823205831002026, -0.26419356594958232076, -0.035604830378302865235, 0.049929130306648916715, 0.11346890919471475101, 0.076723525347953958065, 0.0046746285231860085219, -0.052967079184996032626, -0.35194374847883952429, 0.88641205201453165952, 2.1037636204148282282, 0.032939608789650964837, -5.4654378922221997783, -0.0019653581559379018628, 2.0960845944163786214, 1.0734996597619117509, 0.25555642105259235297, -0.10047863004155743738, -0.21255290337625637465, -0.19452400336808589598, -0.10307088108678341687, 0.009987527223140536814, 0.34807594024436511182, -0.10389082642306876614, -0.72310417348299704798, 0.25380398948995586128, 0.44237537378204766503, 0.31218970668567203042, -0.65853841391763867641, -0.1868942164700324593, 0.088003364837229036688, 0.13787668491014326566, 0.15483642927280283352, 0.07935974287112068648, -0.026851990873529352211, -0.11454333212444139933, 0.22057259657440572709, -0.72746787155486947185, -0.56188360777259149792, 0.9237104171984832357, 0.32078161809808791727, 0.95782219022887538085, -0.4962901506839449417, -0.76217862837990613123, -0.27603367451912397268, 0.087893359870094417419, 0.18815846047308132372, 0.1250813718336331315, 0.032734616171084640612, -0.045639857593390224799, 0.081441586567441437428, -0.43653925864467690854, -0.17161019607164373912, 0.78771268064470911785, -0.50303596779645642911, 0.79523950830058853789, -0.14946186163859043194, -0.44371254576963053973, -0.13315317993019684661, 0.029727393598560435117, 0.06750383491157101179, 0.05302002699417215914, 0.024581184606067733173, -0.0086862298040386778236, -0.047445518011308973616, 0.31688713985332073175, 0.6387853155530476279, 0.80271127210079640246, -0.094203016164809444399, -3.304505589709130664, -0.094079860197359671536, 0.81927488176259888331, 0.70303014764299742723, 0.47611616222045227698, 0.22334862740162181316, 0.016917838522117277511, -0.15465184409747173566, -0.26975362170383432314, -0.013304003606630318904, 0.081491395380458492292, -0.031338915519024360123, -0.18177987073649898297, 0.36806887091918372157, -0.44868356687411709016, 0.36920053108306366552, -0.17777565134728579022, -0.032363345751609109979, 0.073961345890396656966, 0.010661443361568352997, -0.0083526322874201922725, -0.0038056395301188142809, -0.0049736885523030977679, 0.00067260842307072395661, 0.070126440137971896016, -0.14481737582093967553, -0.043320858566514026944, 0.54550757144442763025, -0.85852200126687505666, 0.54753694406504860304, -0.038510275247428227274, -0.14795544452507361499, 0.053701722431141234171, 0.033936208938456127737, -0.0081463230363232521508, -0.0095331027384147337678, -3.234400503383967878e-05, -0.0013559686435274682231, 0.13353046930669482117, 0.0011428201359563461281, -0.076959386832174236925, 0.38551432994205969695, -0.88413330372057374884, 0.37547524018590272554, -0.089592481288623598834, -0.0045461652366113325385, 0.16089495907386777707, 0.10741041246309361579, 0.014891756271700848369, -0.049345818636947098856, -0.06870607459524311067, -0.12817832921117527922, 0.12143949628471155389, 0.13772799415219455499, 0.27078568731415131321, 0.39320717065991006578, -1.5908861201526169893, 0.38545776702580669681, 0.25859181961814181605, 0.12719525783763019344, 0.14412160977611601642, 0.016416790423097054469, -0.061259602743198734998, -0.059757126956019736286, -0.012674256950077764788, -0.1121872404460540068, 0.019608834060241372987, 0.10906121911221886012, 0.27128350173717119542, 0.39927791030392068006, -1.382868416716670934, 0.40141053950392696636, 0.27516015979795005464, 0.10994026932273129327, 0.026066060660768764629, -0.067233903328608290995, -0.059654268973933624598, -0.015579157015460696686, 0.024900912894632292599, 0.024571731228535245717, -0.034773803400600440672, -0.17533566572022427077, 0.095604578144991098076, 0.62744503373008475577, -1.0729398587798772979, 0.62767018495796722544, 0.09678506867120567625, -0.17751541982876720138, -0.04263372908003031786, 0.0083456211474632301106, 0.0077662652715292813249, 0.0021906026084405265213, 0.011482217272161815078, -0.11843427007590015543, -0.16701094676691791152, 0.0095011599279929692674, 0.48012809250066557087, 0.61646553833211692464, -1.6215229268798247997, 0.60204359058225709322, 0.46060971030378727864, -0.00078818779415212869566, -0.15889588061987885981, -0.14573545591025821588, -0.056122252809232442416, 0.014656296276475258566, 0.075289773627759792873, -0.22348878100817548553, -0.10713647068102423454, 0.2065581042123132427, 0.79262907927009607079, 0.53572585649575099609, -2.3950260481100174026, 0.51927360894896690269, 0.768505369439456687, 0.19360585512988767731, -0.077674711890528380653, -0.15347834603343188387, -0.10890021698040440457, -0.028646117661558858497, 0.070689304848367523482, 1.182043069255453549, 0.078300401787220919037, -1.501684251964505723, -4.047928683349283574, -10.398801298630289125, 29.407988898249801224, -10.303513744147986486, -3.9446400018029450507, -1.5449919203725843708, -0.27060089188924796311, 0.37959132026824526918, 0.53821589916568202128, 0.37431675901032823717, 0.045199413811779151628, -0.31442688788816053336, 0.055122472867993543844, 0.28879622220272194166, -0.083513266812120345861, 0.2273333352822928255, -0.29954079737901206926, 0.19985810379394308711, -0.13409568740899668882, 0.23292387590429286526, 0.048561673505714164634, -0.18695955801013819375, -0.1478811140052926909, -0.020608847718323608511, 0.11887196827093426954, -0.27953376604766561053, -0.099782011902060097075, 0.37530907357037845973, 1.2746716530872019835, 0.43909508464949847939, -3.4238867084081290315, 0.42575328324561678706, 1.2585789670698861453, 0.37583742147425519553, -0.044185046545041170574, -0.15064760407324728764, -0.13814343777498583332, -0.06218624837916879311, 0.046920811993284607111, -0.35874632745848017423, 0.3793266619969836162, 0.56625552007129031917, -0.37240000367652337232, -0.20179325269818898514, -0.023580553432370223582, -0.20143856720435551755, -0.38825606851519745, 0.52116423701916991806, 0.37637147278977484133, -0.078743111317550351647, -0.18544110479349268372, -0.08962876348286410999, 0.056462306333566268657, 0.2804513237472623377, -0.08962617055893443152, -0.59267271117113706769, -0.54271787534865900682, 0.79310608301141771737, 0.30646586647900614375, 0.80318497888927975481, -0.52317957451334418106, -0.5819504530987761326, -0.13494425253635866646, 0.095802220776258967261, 0.13178391830289115649, 0.069135292450717200641, -0.015095193463545656526, 0.013103645507445770002, -0.40251565492182728168, -0.44669638336028572034, 0.26923285880623065713, 1.028132818546025451, -0.88724252574760309464, 1.0263213117111149586, 0.26548557812695533498, -0.4611886033654626349, -0.46752347939921012898, -0.23809173566694041835, -0.017575695235881871881, 0.11135669128067099409, 0.18667204279186386251, 0.12268255886117891695, 0.031240683860011754513, -0.30767974093301581062, -0.27109698739405996282, 0.60229891954399772391, -0.34745959534579673456, 0.60160142826098272906, -0.27297034421989818664, -0.31463720635186265273, 0.018262890571185054489, 0.10797568343826810522, 0.059967603828962531831, -0.0023139655120748749834, -0.02637972698854227499, -0.59078465578061245544, -0.043172962701791439688, 1.0050205426794485142, 2.2480872245608800597, 0.14133270146215970442, -5.5195894978357173954, 0.097220598549300207836, 2.1881557271461504754, 0.99602509521515925073, 0.10881031335182267394, -0.2164617602944416408, -0.26277652411218038742, -0.16703556081814122103, 0.014190662001679524296, 0.21344368469322683168, -0.022280749138222394334, -0.28774143458114420646, -0.59151386285209583971, 0.44207477651012105913, 0.47978714246453812997, 0.45555016607242060367, -0.56634768515769229325, -0.26178629867411862264, -0.035604830503357165483, 0.048765676383037859232, 0.10906120942827957077, 0.061454303604852933418, -0.040913546686983971734, -0.50450534925764678462, 0.079311866446411749543, 1.0506755326060206457, 2.1170517122252712205, -0.0019653575683263868612, -5.5074251380126622024, -0.037357396674730480868, 2.0757962735398285403, 1.0664717298850607818, 0.25555642069597805932, -0.098783940084736454601, -0.21057133917068837636, -0.1887206478482664207, -0.085581622803620679973, 0.27677629377853468773, 0.081188143801069645455, -0.23559867682580248882, -0.69503709758543352759, 0.31218970722716221644, 0.48564343446303737117, 0.33259019901642439265, -0.65431584489042138308, -0.18872715454726651796, 0.088003364531906380641, 0.13977799896553833037, 0.14849286547042694906, 0.044224581795194811695, -0.12235278081475944867, 0.39963918980875323683, -0.22264751912366714803, -0.78778859435101245001, -0.5206979281376247215, 0.95782218995081880308, 0.3456560191398879156, 0.9749015773091233017, -0.48766216712506804676, -0.76044856757571932526, -0.27603367445059212582, 0.089700569793562182075, 0.18889182062032236487, 0.11251448434555429146, -0.014570435995163038417, 0.16013537821567083719, -0.096608842310707904177, -0.45073627861653448035, -0.15996581197559389675, 0.79523950789501751668, -0.49447722397015891893, 0.80260184147313629044, -0.14504378094412936906, -0.44247252172522533176, -0.13315318050159730134, 0.030271521549043053229, 0.068232970508579543489, 0.05053837086128505629, 0.013235553857526656182, -0.27173464922383155917, 0.076368700109101808615, 0.42283842116566872082, 0.69721129950924343177, 0.8252406829999811011, -0.094079860391485109483, -3.312895357973127286, -0.1008533314453424895, 0.81927488230180733897, 0.70696434246633632625, 0.47201616683933195207, 0.19098418163294483829, -0.071830361338750620126, -0.33152756968536251092, -0.01006268163675732609, 0.0043720924485270628787, 0.070353311473903806794, -0.032135534493885356544, -0.17703407353344560726, 0.36920053081207732104, -0.44910657576252849266, 0.36890230908390880815, -0.17777565144449644641, -0.031912369293712432305, 0.074610973386818948971, 0.0089262503285165977623, -0.014847760558223014465, -0.0024660529264940922109, -0.016234805315633680373, 0.033741713160949850236, 0.05152869787108950439, -0.14887560538951705169, -0.038406747155502435342, 0.54753694403454600259, -0.85864671384328239423, 0.5471494178319188384, -0.038510275505494122317, -0.14745021379527928351, 0.054101950852410603254, 0.031760928146334341249, -0.014131359409877905509, -0.0029956453390381264715, -0.10945716568253298362, 0.078770266608906958083, 0.17052365876858854854, 0.01441264049213930587, -0.077046647538043522285, 0.37547524089766665112, -0.89616236856371234154, 0.36763625758435858337, -0.089592482532002415407, 0.0035171511583711456694, 0.1725619270316964271, 0.11187739678437955138, -0.0060886754186113140053, -0.11494796922194595867, -0.12592647748974761246, -0.00066205971702929054118, 0.14415757426572739508, 0.13242329760454421561, 0.26271157920204618597, 0.38545776659536951092, -1.5962617265165188751, 0.38291479678463541392, 0.25859181952273824212, 0.12857404885964532348, 0.14485381829926360919, 0.013991169553421667349, -0.068251650561597126954, -0.063673713755594427144, -0.028893346502053510183, -0.071990052028798351413, 0.012592326317297165145, 0.1023849281178424997, 0.27349402364281505395, 0.40141053983320618137, -1.3830677985529036267, 0.40054674407143409098, 0.27516015955201411636, 0.1093892775341150092, 0.021652011127980402838, -0.075510160963091918496, -0.060422024575505821342, 0.021408200712935008397, 0.010957513643918679117, 0.020375911345367578309, -0.039949074971927116262, -0.17864847412803730453, 0.095880950640889320291, 0.62767018493976822757, -1.0727437557895593834, 0.62774661294606448614, 0.09678506887165666539, -0.1776885748609840221, -0.042753131641784763994, 0.0096231567546607547914, 0.011644601667281635413, 0.0099400467839138376136, 0.0053687361103321590283, -0.11710911534833637337, -0.14617008146047349992, 0.0057016443435682212515, 0.46365978692090781266, 0.60204359067042667597, -1.6275740327412289155, 0.60130306646725806274, 0.46060971039589265752, -0.0017272474297322371051, -0.15737639954205420723, -0.13357917343468017624, -0.02652865004847607816, 0.059810586684294078652, -0.077809278391393510432, -0.13791422933094893422, -0.072008022385408099186, 0.19633794761278350327, 0.77015995682766558161, 0.51927360863763927235, -2.4008637436590847258, 0.51926943084580123866, 0.76850536946036573926, 0.19077009305344316825, -0.080689925458729105934, -0.14762914103148994105, -0.083091443734985948044, 0.024527490779812706478, 0.8315317750244664019, 0.52456700697379177178, -0.23497319809369754395, -1.5676963909209300496, -3.9785892564416158734, -10.303513744283037568, 29.471524339498806455, -10.280297412431856685, -3.9446400011993971724, -1.5482155634167726888, -0.26684674378534012362, 0.38850177063802904254, 0.54573985699654459225, 0.38413225396203198869, -0.10852594801167640692, -0.15616858876856343219, 0.084853251010688746092, 0.26089890895046141672, -0.12058289373433317415, 0.19985810335092102363, -0.31400023124796305751, 0.1944723241103289979, -0.13409568691778206428, 0.23716354085170743282, 0.058962801884057114421, -0.16537094077114078416, -0.10730957631083581916, 0.048999279582899478436, -0.12893345236352335426, -0.15591525305216985542, -0.0578334354730270922, 0.36883464271829885517, 1.2579294539599810143, 0.42575328321292416067, -3.4292134445962942912, 0.4258037834177270109, 1.2585789670093501247, 0.37150760400559112995, -0.052688735096959510296, -0.15632250215339649202, -0.1262395163215954963, -0.0081783883758711414802, -0.19379796686767242209, -0.09387467172864721876, 0.36915647999686101999, 0.51685049914347303002, -0.39118649817643436073, -0.20143856718283562057, -0.020004537399830026045, -0.19991510492123204523, -0.38825606823535441237, 0.51856335905223382987, 0.36730247393633741471, -0.093485019883265740859, -0.18345825858939018849, -0.011565860003428197686, 0.15756122971044494996, 0.12265912881341965723, -0.12630864445252695805, -0.58399764280435895181, -0.52708695205764566349, 0.80318497796590704318, 0.31248931778551924943, 0.80583684593324900192, -0.52317957429931527535, -0.58211878531044880525, -0.13176277577100795058, 0.10278406523528374472, 0.13238890823797352936, 0.04208725641673579515, 0.21447628074982622026, -0.13841318910359229344, -0.4481291044803801249, -0.47276916751919267901, 0.25324617705371454957, 1.026321311875665554, -0.87772604980467694435, 1.0357383314704642352, 0.26548557805850991898, -0.47110821330552232755, -0.47550193097502824191, -0.22041886322916590113, 0.050712070069947878803, 0.24082313354037931941, 0.0041533852256036623771, 0.11788861960483495994, 0.032935471189125584135, -0.3085580514874727287, -0.27238333171510736852, 0.60160142813308681298, -0.34671571857364874569, 0.60221341422691687661, -0.27297034397464331468, -0.31387550318695028873, 0.022002938303041889856, 0.11441700647089045328, 0.058409127369279736919, -0.036985730955530968234, -0.38011140251419173186, -0.26716288384398934053, 0.10240779996997967372, 1.0134714905119424078, 2.2053602974387533919, 0.097220598880255248986, -5.5466620883097981221, 0.087169445902643305502, 2.1881557274618481657, 0.9982767337342473013, 0.10907324722744249357, -0.2161193219788190012, -0.25623131891189937503, -0.14869658153067519257, 0.1417412916402646661, 0.046481373425267757948, -0.043582590532922238502, -0.2714160976117514501, -0.57368232139174080153, 0.45555016650747592877, 0.48904337739243541883, 0.45973887492071385807, -0.56634768524740453088, -0.26419356581807251771, -0.038529309840796166498, 0.046449164769534312391, 0.10134407177189738714, 0.025935141100075775888, -0.40588116029502802062, -0.19658054788126544388, 0.22941257029972855763, 1.0819152201115900258, 2.096084593904082638, -0.03735739603940169129, -5.5365848179822680919, -0.051767020720665191114, 2.0757962730711074784, 1.0734996599605077794, 0.25834353452304026355, -0.1107333715162337473, -0.24253093171733097755, -0.23589505035012445222, 0.12001492357688647861, 0.097337330269191668641, 0.062885180094515491511, -0.19966048801159907988, -0.65853841423761627638, 0.33259019958384450355, 0.49254871116711196333, 0.33281377308246962166, -0.6543158451627684169, -0.1868942163898976716, 0.087547036871556627169, 0.12722533959923498781, 0.10836423806371039968, -0.044002213587189822896, 0.25870849203118689852, 0.12824546364962294276, -0.27182637394927938734, -0.77033373797526150106, -0.49629015083890237747, 0.97490157696926238451, 0.35620599997885349364, 0.97949502072784655482, -0.48766216689562708053, -0.76217862790882173751, -0.27522897179870137174, 0.096297367733368610665, 0.19298669561557074692, 0.083389154248472666109, 0.11448402569382806759, 0.055537855682414670466, -0.13031971061141814361, -0.44782510637714878499, -0.14946186112189288608, 0.80260184148830582274, -0.48971468518695437933, 0.80493316837316963852, -0.14504378151188629387, -0.4437125454050590001, -0.1332606983712225357, 0.033698603486286046738, 0.073835693015320308374, 0.05520039649757168404, -0.33152756885934675513, -0.071830362881124529273, 0.19098418156000876444, 0.47201616779852906269, 0.70696434275122832513, 0.81927488259077785315, -0.10085333312877778345, -3.3128953569592143324, -0.094079860469607395901, 0.82524068326634036907, 0.69721129903103906678, 0.42283842124478010494, 0.07636870048516022913, -0.2717346494746621377, -0.0024660528724438773862, -0.014847760723361382723, 0.008926250480595157577, 0.07461097330007745998, -0.031912369106863194834, -0.17777565166890213488, 0.36890230913497557008, -0.44910657570322221055, 0.3692005307826808358, -0.17703407350655539498, -0.032135534554228094295, 0.070353311545572463914, 0.0043720923872745051653, -0.010062681611574215285, -0.0029956454264127656793, -0.014131359171933557256, 0.031760927959166102008, 0.054101950743593862914, -0.14745021341337272158, -0.038510275978629389393, 0.54714941824388962921, -0.85864671405259140879, 0.547536943955666211, -0.038406746878027975267, -0.14887560573025052091, 0.051528698170877899887, 0.033741713006076402837, -0.016234805285113344114, -0.11494797010288256012, -0.0060886738069220805447, 0.11187739717531217098, 0.1725619252347616317, 0.0035171506290606613376, -0.089592481232699028304, 0.36763625896360091483, -0.89616236995470688331, 0.37547523981179908859, -0.077046647376964927378, 0.014412642364926231409, 0.17052365840862332425, 0.078770265218311041888, -0.10945716496384449989, -0.063673713914831606142, -0.068251650230981536494, 0.013991169378800610151, 0.1448538184933059747, 0.12857404823958845652, 0.25859182020166648908, 0.38291479650077675512, -1.5962617264706535636, 0.38545776668032266654, 0.2627115787326321783, 0.13242329832807506196, 0.14415757384413710462, -0.00066205965947602630345, -0.12592647747841942407, 0.021408200590681460834, -0.060422024360173798296, -0.075510160911234858183, 0.021652010827515638658, 0.10938927778207589714, 0.27516015932759102514, 0.40054674420875074903, -1.3830677984030848027, 0.40141053961065070732, 0.27349402361692837227, 0.10238492820683373075, 0.012592326447100553669, -0.071990052209550597917, -0.028893346447719410464, 0.0099400468087856661686, 0.011644601512391685588, 0.0096231571830398439371, -0.042753132262174309286, -0.17768857429583281848, 0.096785068432503965075, 0.62774661320734825853, -1.0727437557521612987, 0.62767018478358926892, 0.095880950594497249551, -0.17864847395069091696, -0.039949075010373459593, 0.020375911263214644548, 0.010957513687928464516, 0.05981058653145953502, -0.026528649693773269369, -0.13357917335179919638, -0.157376400049874271, -0.0017272473379919828229, 0.46060971035552800146, 0.60130306689958468169, -1.6275740329463568301, 0.60204359057443446179, 0.46365978694156151363, 0.0057016443929787663514, -0.14617008155119301516, -0.1171091152455160378, 0.0053687360721129560523, 0.024527490818428185348, -0.083091443853539281861, -0.14762914097090720777, -0.080689925110979748846, 0.19077009245528153691, 0.76850536981371653233, 0.51926943049786900009, -2.4008637433582373788, 0.5192736088519646076, 0.77015995668044168188, 0.19633794746161747913, -0.072008022543305669205, -0.13791422904576222663, -0.077809278468349576952, 0.38413225408159834551, 0.54573985665859570116, 0.38850177076411929189, -0.26684674360423987771, -1.5482155631553569108, -3.9446400017440170771, -10.280297412382495281, 29.4715243395497275, -10.303513743722131579, -3.9785892567301890388, -1.5676963915189894294, -0.23497319795866256076, 0.52456700746034723437, 0.83153177483112516999, 0.048999279651003327385, -0.10730957628154025318, -0.16537094111339545321, 0.058962802332716224285, 0.23716354074965129151, -0.13409568704307783804, 0.19447232387622914729, -0.31400023100921581332, 0.19985810359044764439, -0.12058289389454035501, 0.26089890877749200016, 0.084853251076140612241, -0.15616858864531274476, -0.10852594808216967093, -0.0081783881853609071561, -0.12623951680262451847, -0.15632250187176374712, -0.052688735001877887165, 0.37150760414228761741, 1.2585789666833127054, 0.42580378340603752818, -3.4292134446961077821, 0.42575328364566422668, 1.2579294538729546282, 0.36883464254364900503, -0.057833435565875002116, -0.15591525283094692655, -0.12893345244919512949, -0.011565859789707737953, -0.18345825908622090394, -0.093485019643196443084, 0.36730247393823106661, 0.51856335952595833394, -0.38825606880708796176, -0.19991510500059095357, -0.020004537347141866654, -0.20143856667251341941, -0.39118649852689246771, 0.51685049909948355129, 0.36915647991592637212, -0.093874671486196631376, -0.19379796698011572675, 0.042087256583851338776, 0.13238890813775983091, 0.10278406451112151454, -0.13176277468095803913, -0.58211878562049590435, -0.52317957424849292902, 0.80583684522713106002, 0.31248931823196923752, 0.80318497863328930375, -0.52708695260565685548, -0.58399764307718105005, -0.12630864452614884441, 0.12265912941121155655, 0.15756122942669864573, 0.24082313327563911742, 0.050712070407497006341, -0.22041886300536467358, -0.47550193124809114753, -0.47110821330738378299, 0.26548557803030525815, 1.0357383312977495038, -0.87772604947684307319, 1.0263213113239184615, 0.25324617776667507352, -0.47276916781785260913, -0.44812910432652358628, -0.13841318947893974101, 0.21447628096523610464, -0.036985730864215519975, 0.058409127184395011334, 0.11441700660274393975, 0.022002938062885314674, -0.31387550267413588667, -0.27297034450539608974, 0.60221341457808652553, -0.34671571875736595381, 0.6016014281101934591, -0.27238333146898274784, -0.30855805170270533822, 0.032935471158486120946, 0.11788861970621954123, 0.0041533852000585286604, -0.14869658177187511372, -0.25623131839069390736, -0.2161193220980497931, 0.10907324702432508179, 0.99827673338363664968, 2.1881557280577208502, 0.087169445745412188487, -5.5466620881067631998, 0.097220598191395612986, 2.2053602976829838056, 1.0134714910926783116, 0.10240779982732534092, -0.2671628843295247302, -0.38011140227297174343, 0.025935140820768316783, 0.10134407220370619807, 0.046449164851564028744, -0.038529310059074450834, -0.26419356621773792959, -0.56634768482050423799, 0.45973887521911205312, 0.48904337702102484986, 0.45555016620482402523, -0.57368232093944271366, -0.27141609766955826499, -0.043582590482164507151, 0.046481373151527172194, 0.14174129179930591849, -0.23589505047592190801, -0.2425309312895159719, -0.11073337193320598182, 0.25834353467925486036, 1.0734996595312429335, 2.0757962735592156989, -0.051767020457752832563, -5.5365848184116357444, -0.037357396196495695762, 2.0960845940947203658, 1.0819152200845789658, 0.22941257064286812994, -0.19658054823162895985, -0.40588116023393966403, -0.044002213587389329974, 0.10836423807451098233, 0.12722533962149928932, 0.087547036767161912074, -0.18689421652196935852, -0.65431584496982653221, 0.33281377343792173384, 0.4925487108702719663, 0.33259019941134537746, -0.65853841429841875055, -0.19966048778330058866, 0.062885180129323425824, 0.097337330178999426078, 0.12001492358586138276, 0.083389154584824029537, 0.19298669503766494038, 0.096297367457636298682, -0.27522897100526177994, -0.76217862766549615561, -0.4876621674994638389, 0.97949502025448798026, 0.35620600053516837669, 0.97490157742730376711, -0.49629015102473300614, -0.77033373868771304327, -0.271826373855606096, 0.12824546436900063906, 0.258708491660729345, 0.055200396668435909586, 0.073835692678219919105, 0.033698603558618811515, -0.13326069823804626013, -0.44371254527466835738, -0.14504378178484006456, 0.80493316851105767284, -0.48971468538436768725, 0.80260184186772032255, -0.14946186132416322545, -0.44782510648964940492, -0.13031971058088998028, 0.05553785581708286373, 0.11448402562872080079, -0.26975362229036237149, -0.1546518426248736644, 0.016917837716217593719, 0.22334862729479679211, 0.47611616202094475536, 0.70303014723169476863, 0.81927488304140161368, -0.094079860679165572268, -3.3045055901050348623, -0.094203015829130765102, 0.80271127189667890178, 0.63878531587463704966, 0.31688713944065960426, -0.047445517805231027242, -0.0049736886656612661062, -0.003805639283769142716, -0.0083526324039410685535, 0.010661443296361185207, 0.07396134590484590432, -0.032363345634903360748, -0.17777565149448476545, 0.36920053118774182055, -0.44868356705135470364, 0.36806887129445858831, -0.18177987115479479763, -0.031338915220141233542, 0.081491395189532425669, -0.013304003533970269052, -3.2343976706924917452e-05, -0.009533102765603184911, -0.0081463230858899362413, 0.033936209026140314038, 0.05370172232237392651, -0.14795544430314744311, -0.038510275616792273568, 0.54753694451059442549, -0.85852200164281622374, 0.54550757166495889017, -0.043320858715206286838, -0.14481737567117564103, 0.070126440033644765748, 0.00067260845371218467952, -0.068706074241648706913, -0.049345818971523666696, 0.014891755772441436778, 0.10741041300820511339, 0.16089495936703182566, -0.0045461655971711367741, -0.089592481807892643264, 0.37547524083231920144, -0.88413330345767149332, 0.38551432959778786369, -0.076959387318103977815, 0.0011428207636712123252, 0.13353046921567926497, -0.0013559687347074345781, -0.012674256938009937148, -0.05975712694753479437, -0.061259602921621095217, 0.016416790736963011998, 0.14412160968923662385, 0.12719525769361914036, 0.25859181966451144641, 0.3854577669247670757, -1.5908861198263049008, 0.39320717040978953039, 0.27078568741655273389, 0.13772799400985513696, 0.12143949641497139846, -0.12817832924771935255, 0.024900912897626494708, -0.015579156978774093278, -0.059654269125591630196, -0.067233903099745806475, 0.026066060475281337128, 0.10994026948255049281, 0.27516015957242029311, 0.40141053978367546851, -1.3828684170084106775, 0.3992779105728320177, 0.27128350147385893454, 0.10906121938057065923, 0.01960883390014587363, -0.11218724041380422063, 0.011482217201058563319, 0.0021906027401675945768, 0.0077662652232088286511, 0.0083456211410206076334, -0.042633729135420864176, -0.17751541975440013332, 0.096785068671126975315, 0.62767018493899295883, -1.0729398589549374865, 0.62744503417439856374, 0.095604577667149054365, -0.17533566543200701515, -0.034773803512028300666, 0.024571731253739553108, 0.075289773667741907714, 0.014656296251061351477, -0.056122253067635713419, -0.14573545552320738006, -0.15889588068124721465, -0.00078818795497387350579, 0.4606097104515389784, 0.60204359053751699271, -1.6215229270613376045, 0.6164655386127725345, 0.48012809207424800295, 0.0095011605050311087606, -0.16701094704588637097, -0.11843427007503433412, 0.070689304803768809871, -0.028646117586100881824, -0.10890021714099594463, -0.15347834581435604306, -0.077674711844307228747, 0.19360585481483968961, 0.76850536988569584373, 0.51927360847226500251, -2.3950260480001621666, 0.53572585659772686828, 0.79262907922586856024, 0.20655810446903008493, -0.10713647095569968359, -0.22348878095186489046, 0.045199413808828803329, 0.37431675901004063389, 0.53821589894173060831, 0.37959132059512046276, -0.2706008921259437372, -1.5449919199995738595, -3.944640001971920551, -10.30351374425905675, 29.407988898091520724, -10.3988012981895519, -4.047928683659036686, -1.5016842519755058127, 0.078300401852833045879, 1.1820430692693197905, 0.11887196834454467897, -0.020608847981466005272, -0.14788111360802402849, -0.18695955850318424218, 0.048561674001176280413, 0.23292387542962389446, -0.13409568664791926551, 0.19985810288673233615, -0.29954079670876643249, 0.22733333474249906025, -0.083513266315828893482, 0.28879622192894849331, 0.055122472912907727327, -0.31442688787897038472, 0.046920811862559065608, -0.062186248204009017537, -0.13814343774438586632, -0.15064760404519972337, -0.044185046875788683285, 0.37583742188274171792, 1.2585789668040514577, 0.42575328345279889186, -3.4238867087529341049, 0.43909508484084908142, 1.2746716533832791463, 0.37530907319705075764, -0.099782011818130011971, -0.27953376602424290231, 0.05646230623782071828, -0.089628763410281198087, -0.18544110467363322803, -0.078743111307049279413, 0.37637147251580949092, 0.52116423733855365974, -0.38825606885134367197, -0.20143856692225070582, -0.023580553514788379216, -0.20179325284577970256, -0.37240000335241107976, 0.56625551976350374517, 0.37932666212574733811, -0.35874632747318119241, -0.015095193432037860154, 0.069135292247952284672, 0.1317839186341240032, 0.095802220600986143229, -0.13494425264415407795, -0.58195045300057579585, -0.52317957415247395847, 0.80318497839067803934, 0.30646586653313834248, 0.79310608310472996418, -0.54271787519399916633, -0.5926727114156011833, -0.089626170433646207281, 0.28045132372648479135, 0.1866720427581976538, 0.11135669130010047989, -0.017575695476150982044, -0.23809173520991705475, -0.46752347954732131896, -0.46118860341754380761, 0.26548557810365275289, 1.0263213115246465623, -0.88724252522930013853, 1.0281328180760160862, 0.26923285915835937221, -0.44669638377075998603, -0.40251565468052163954, 0.013103645478201048774, -0.026379726898279064889, -0.0023139658042862238513, 0.059967604254592105084, 0.10797568306525454074, 0.018262890750000565498, -0.31463720637226466614, -0.27297034422161631229, 0.60160142829022633659, -0.34745959542876247994, 0.60229891954638292706, -0.27109698723654468422, -0.30767974113016283155, 0.031240683965521390264, 0.12268255883978955445, 0.014190662249537439044, -0.16703556147114356767, -0.2627765233835755998, -0.21646176091382751672, 0.10881031404191576772, 0.99602509458980836854, 2.1881557274274072711, 0.097220598120391188424, -5.5195894968000853709, 0.14133270031659561639, 2.2480872254143209332, 1.0050205419737632262, -0.043172962173136514297, -0.59078465596780194158, -0.040913546623125657264, 0.061454303442592006501, 0.109061209559058514, 0.048765676290380977065, -0.035604830368520690165, -0.26178629857354318577, -0.5663476855587783465, 0.45555016624602528985, 0.47978714258671117943, 0.44207477649912185758, -0.59151386284942752969, -0.28774143480295666375, -0.022280748908802022745, 0.21344368462495388883, -0.085581622768513110344, -0.18872064786243830659, -0.21057133897520402543, -0.098783940571478101589, 0.25555642109378406968, 1.06647172981763938, 2.0757962731803250023, -0.037357396222157390753, -5.5074251380477976525, -0.0019653577474997296903, 2.1170517122227390239, 1.0506755326940275808, 0.079311866481250797856, -0.5045053493131890221, -0.12235278067023613868, 0.044224581553923586474, 0.14849286559156560372, 0.13977799896686071701, 0.088003364400615957663, -0.18872715423405694857, -0.65431584543225573825, 0.33259019963504465878, 0.48564343408460386442, 0.31218970737970108686, -0.6950370977012310103, -0.23559867681017765406, 0.081188143908226761969, 0.27677629371724915552, -0.014570435979966291778, 0.11251448417848584904, 0.18889182096909223896, 0.089700569520676814994, -0.27603367446018289844, -0.76044856750092992836, -0.48766216695929814895, 0.97490157717699854611, 0.34565601907977383522, 0.95782218981077082987, -0.52069792774103484945, -0.7877885946317000343, -0.22264751911925262373, 0.39963918986629987096, 0.013235553840103279205, 0.050538370878173595468, 0.068232970556667660311, 0.030271521560064296175, -0.13315318085498017986, -0.44247252113787505223, -0.14504378142794849271, 0.80260184194834405247, -0.49447722454540010695, 0.79523950832166145908, -0.15996581219335681556, -0.45073627844943919829, -0.096608842431140956597, 0.16013537825095100486, -0.15734323111335740686, -0.1621462076305377753, -0.10310481306146274116, 0.034266125149095293112, 0.22532539644495147568, 0.47611616256970079597, 0.70696434269239172288, 0.82524068319035304153, -0.094203016776891579376, -3.3228625585378979146, -0.13434555935212830202, 0.75307272456622453838, 0.62883646741697485361, 0.35802613226569324212, -0.0069006012146815036076, 0.0010486590977086380182, -0.0021085933773993294216, -0.0083042529609813672126, 0.010529262941297473297, 0.073961346422342408302, -0.031912369889951011648, -0.17703407406837978755, 0.36806887156212730883, -0.45427878706331975556, 0.36516446247136746805, -0.17493399471351739338, -0.011045959896921050611, 0.048596277826094706753, -0.0033804517085458179668, 0.0015293677367562953043, -0.0072486234879606580364, -0.0076642720025545205825, 0.03364243240985126393, 0.053701723114620805721, -0.14745021440059399853, -0.038406747734187160481, 0.5455075723373060681, -0.86243302209493033494, 0.54841878847202119029, -0.036367755451507374764, -0.15267844732715765699, 0.073339222420165425809, -0.017947279803165086531, -0.047134222448773743785, -0.046367846663308148747, 0.0081671681878773688679, 0.10079392296708394872, 0.16089495825097471782, 0.0035171500810353140132, -0.077046646861111123261, 0.38551433090398401493, -0.88484004216232248208, 0.36960361384760465064, -0.11574262705662381745, -0.027731293510451900108, 0.19521853476008252382, 0.010821622481136081856, -0.026722882081191196013, -0.061766346549650186226, -0.060478122384000895795, 0.016616257527600590549, 0.1441216092315013575, 0.1285740483071138307, 0.26271157874063350013, 0.39320717104788005525, -1.5795375781720657482, 0.40774885730488708013, 0.2695049444910116998, 0.090675628090612983567, 0.011326057111414156164, 0.0095457409054017072075, 0.011737306183037268803, -0.015490031860168035349, -0.057304784530035436652, -0.066244511209254383632, 0.026066060822030515109, 0.10938927707959759938, 0.27349402392567367936, 0.39927791046850541479, -1.3804121290324127269, 0.41738146157507577394, 0.29713284757587615781, 0.10445790271591920417, -0.12746066092872643249, 0.0073172246231219539858, 0.0051631693365675537799, 0.0010247980450059273119, 0.0073923706439983881411, 0.0080574031948957795107, -0.042633728857228561648, -0.17768857486709010995, 0.095880950528283132028, 0.62744503405674678653, -1.069650082330725116, 0.63493586056345896473, 0.087689631005308840694, -0.19832583440757417526, 0.011849850242104092296, 0.064131851557260477481, 0.03007872178668286392, -0.0097707720884766119385, -0.063717555477502893502, -0.14595626541517092711, -0.15889588062293005799, -0.0017272470077593082838, 0.46365978708392474594, 0.61646553855886931927, -1.5958099671713261714, 0.64090718202702856576, 0.47542375876025427806, -0.031712142887551288128, -0.28804525816948683969, 0.06864014891280145747, 0.012651067028427380456, -0.051560784033457152109, -0.11148189584940283725, -0.15059596046433620931, -0.077674711923105085876, 0.19077009279444473444, 0.770159957128479844, 0.53572585636336689241, -2.3609374829451201805, 0.57567494629003357343, 0.79874425507516011447, 0.13167272551438205519, -0.3313633446942134686, -0.12765509307769010161, 0.17127965864394073536, 0.42302646341587146761, 0.54383399716089353415, 0.37469344002002413774, -0.27060089251402930088, -1.5482155621678748147, -3.9785892570573206939, -10.398801298386045389, 29.251078494621580717, -10.530264504302731865, -3.9730493516619165284, -1.0329007167365311659, 1.0450084974818230066, 0.12009379025365621296, 0.033591694693088813839, -0.064588522414332771859, -0.1642100810945326117, -0.1911067182226271477, 0.048561674008477995201, 0.23716354091136926385, -0.12058289308452563304, 0.22733333423978985266, -0.26142843751777156136, 0.26258938279320997999, -0.071217322087623435856, 0.2581514928217599758, -0.31883337559759478141, 0.048155063050192806196, -0.0036615958885524435562, -0.075434411757445668512, -0.13464103572336766357, -0.14522748851640776557, -0.044185046209238715287, 0.3715076042094063169, 1.2579294535843132952, 0.43909508445765360607, -3.3912362579800201523, 0.48812758816506157666, 1.2958855190877256458, 0.27192614222043953731, -0.37170184676442641347, 0.046684515690476074967, -0.0019456115567239159678, -0.094974231433077954678, -0.17622044639519199305, -0.073069943780300836145, 0.37637147284764543853, 0.51856335933152153395, -0.39118649855342318977, -0.20179325284606847157, -0.011016767168697406964, -0.157485257559671743, -0.30307898226928325913, 0.57685460736954330407, -0.10021353093622339947, -0.029732348436643990519, 0.017366598386957417949, 0.076730843077104496364, 0.12993998981927301628, 0.093776424242725503477, -0.13494425233737242231, -0.582118785350694945, -0.52708695289247531335, 0.79310608307205188172, 0.28857344377952676151, 0.76743510918049029357, -0.57495185782379665262, -0.55343399711803364927, 0.22696215184086390715, 0.10344364258317022276, 0.10876491108547907405, 0.076963488739740987254, -0.022205736546181159552, -0.23280256256770098799, -0.46752347970390545351, -0.47110821343536884998, 0.25324617782727593118, 1.028132817864488846, -0.86038565208242512306, 1.0706071136701023061, 0.28786576584135925572, -0.46517991970777194943, -0.43112668009099369693, -0.0084715178461692609846, -0.018051528897674806728, 0.00036513334630979693562, 0.059348221561838568339, 0.10691372197023375767, 0.01826289078633266949, -0.3138755027386821439, -0.27238333229383532297, 0.60229891997347861921, -0.34747678578271901095, 0.59406467158396447648, -0.30729490549303323244, -0.32795855015360170226, 0.21354348344562865125, 0.084684294196517212461, -0.062998698989777029977, -0.19739178951062913425, -0.26776418469288421065, -0.21477483312481637778, 0.108810312940568088, 0.99827673372008851604, 2.2053602982547362288, 0.14133270094035987885, -5.4525111759224049734, 0.20658204221168458004, 2.2448731618441621904, 0.80383247658960166326, -0.57794733105999251954, -0.052967078968292571894, 0.0046746281789347421398, 0.076723525175588669356, 0.1134689093827662143, 0.049929131123792935831, -0.035604831446190865485, -0.26419356550105221881, -0.57368232105953342881, 0.44207477663723693251, 0.45789869499963520294, 0.40922997091581791995, -0.61363197430190508097, -0.18948646041539868312, 0.17169854926540448203, 0.0099875274246974572734, -0.10307088143057584195, -0.19452400368762862159, -0.21255290259351972137, -0.10047863014982938434, 0.25555642071043249697, 1.0734996596760151277, 2.0960845946782242777, -0.0019653579478070559716, -5.4654378924195343714, 0.032939608368799722093, 2.1037636208128072113, 0.88641205210049345276, -0.35194374860320815079, -0.11454333215126161205, -0.026851990767225553114, 0.079359742693601409425, 0.15483642916334938722, 0.13787668550535681078, 0.088003364154454422064, -0.18689421602871969519, -0.65853841441408045299, 0.3121897074837112207, 0.44237537337787591341, 0.25380398918007496256, -0.72310417331908083405, -0.10389082614094344303, 0.34807594005940589721, -0.045639858065365787598, 0.032734617081281447426, 0.12508137194131033199, 0.18815845960343605126, 0.087893359611369958895, -0.27603367401698275518, -0.7621786276602005028, -0.49629015110993696114, 0.95782218948304476136, 0.32078161837059010475, 0.92371041781370677448, -0.56188360767526690509, -0.72746787248987043739, 0.22057259705838999642, -0.0086862299106342851646, 0.024581184819341767023, 0.053020026938232919789, 0.067503835011963525536, 0.029727393195071949439, -0.13315317982129731833, -0.44371254547411809188, -0.14946186154020044135, 0.79523950775865026586, -0.50303596740102818341, 0.78771268047913101107, -0.17161019582234848802, -0.43653925899157519641, 0.081441586728330683798, -0.048873679302586875428, -0.11740145592733151891, -0.14269860033387818388, -0.096996746926574922698, 0.034266125683718340333, 0.22334862796009077868, 0.47201616702304194373, 0.69721129906554368816, 0.8027112712942952033, -0.13434555956041974167, -3.3750136382206994412, -0.12174595382005301558, 0.90548916903438958492, 0.92132922214325863042, -0.0038049944278762144051, -0.0025361850459425200159, 0.0011681996116293252852, -0.002212113174669285906, -0.0083042537689875274565, 0.01066144349470056138, 0.074610973894006199503, -0.032135534322330218393, -0.18177987085245253529, 0.36516446181715306318, -0.44611290857185686987, 0.37500555250395123386, -0.23216765564466904093, 0.08239813977940305123, -0.0031824402606131941412, 0.00048694806708034026575, 0.0015658205177404558618, -0.0074863182100444510686, -0.0076642730382548247478, 0.033936208982107481102, 0.054101951189928498831, -0.14887560504266561523, -0.04332085854161549926, 0.54841878757891915352, -0.84596185545907665038, 0.54257578352281843337, -0.12554059820230678568, 0.00061415204162239007534, 0.015972620804465563482, -0.027762398334172405656, -0.055848172180174862622, -0.051821384350476021319, 0.0081671667895122368463, 0.1074104122853987292, 0.17256192608191220361, 0.014412641734709769059, -0.076959386548231561997, 0.36960361446804712582, -0.92218635965642914432, 0.32515442517379167242, -0.13574291843110875, 0.26248899191827551602, 0.01657904690884242152, -0.0039798417785093260363, -0.032275597774364075954, -0.062733066872564496874, -0.060478122956478141958, 0.016416790840924982975, 0.14485381869979402358, 0.13242329785506981588, 0.27078568723563151144, 0.40774885750031270026, -1.571128689972305148, 0.38472659149950622082, 0.18038160607886988274, 0.18542834702130947533, -0.001029569689395192552, 0.01249476992092364569, 0.011158769755679035646, -0.016223483602019743433, -0.057304784329056758507, -0.067233903498643735053, 0.021652011286115047106, 0.10238492809242383252, 0.27128350177501714402, 0.41738146121078006923, -1.3424440987380459323, 0.44795882385544172344, 0.22271970276874211714, -0.018268463387471463311, 0.0019137382893836057344, 0.0056164989977074922939, 0.0046302791650092954556, 0.0007190472480833385964, 0.007392370586697583415, 0.0083456209461587015086, -0.0427531315657010344, -0.17864847403756076161, 0.095604577941833940313, 0.63493586017404779476, -1.0617934399025408254, 0.62875090186279958271, 0.041468085540884089268, -0.14726019835232093125, 0.037059309307331207706, 0.032394662231018361842, 0.01677806318663959545, -0.01308906620595931547, -0.063717554799329403914, -0.14573545619627129022, -0.15737640013077136558, 0.0057016443118505372167, 0.48012809253555910338, 0.6409071817879394839, -1.5755953827665598599, 0.66318488027044497635, 0.4223645505880960771, -0.3385434379168766883, 0.041058711826004561096, 0.031169893252029387543, 3.0631089910464204262e-05, -0.052797587717909585936, -0.11148189581390595404, -0.15347834600573068142, -0.080689924986597827417, 0.19633794701909407232, 0.79262907908433977155, 0.57567494663349871864, -2.3270977842785574019, 0.58813189724207504128, 0.71722940388436084991, -0.20568144439153046577, -0.16215940642560841556, 0.013615592873473136848, 0.22621413358674052896, 0.433511320062328398, 0.54383399776204199849, 0.37959132109886128958, -0.26684674440073608004, -1.5676963909132015651, -4.0479286838226062883, -10.530264504041765505, 29.078622525839158186, -10.564431132307239736, -3.5204321206817454915, -0.1027836874252371635, 0.080235689253023129153, 0.053205212774120325125, 0.0023454293883267141041, -0.074308154824411110084, -0.16421008042537077465, -0.18695955859432711321, 0.05896280245930851649, 0.26089890802680271165, -0.08351326607974063343, 0.26258938278093657548, -0.24846020851183436839, 0.24978397918767286656, -0.23885197220229434367, 0.039099213613302555292, 0.028147275748947625518, 0.021445370265717092484, -0.011330920520216579417, -0.074461807282287073573, -0.13464103548290995427, -0.15064760418912748197, -0.052688735049346513417, 0.36883464258413534154, 1.2746716533019843975, 0.48812758782738885355, -3.3354122646507757821, 0.50813423533742752003, 1.2289086933237587385, -0.14294698256449836493, 0.018690788489647697235, 0.028055251749916004095, -0.0060963296383975650039, -0.093325158100836302921, -0.17622044623685251263, -0.078743111239066992946, 0.36730247394596460264, 0.51685049917259395791, -0.37240000356772151591, -0.15748525786348060995, 0.036940106396420160917, -0.16979508522524894287, -0.51637208220150587756, 0.61537832056046903872, -0.024595469447268619723, -0.0086253930870161688721, 0.025584186064783513703, 0.078055681386618586126, 0.12993998936094183771, 0.095802221119013639128, -0.13176277483525697387, -0.58399764299828815783, -0.5427178756113385516, 0.76743510878261844699, 0.26501556033164569426, 0.76628889374204856022, -0.55527124982255116059, -0.29615276733583495439, 0.024420856163663476335, 0.078556449494798341604, 0.10327041370823537847, 0.078654741018450213819, -0.022205736127242775568, -0.23809173579187709047, -0.47550193115027594715, -0.47276916789589340517, 0.26923285927145990115, 1.0706071139816550897, -0.8033626190718408111, 1.1218967146986362771, 0.20349314618502967189, -0.94817326241038191181, 0.00088885763964312194778, -0.011707259827614087747, -0.016813439952922878962, 0.0012370390255546682978, 0.059348221067449281885, 0.10797568334564736081, 0.02200293823735901752, -0.30855805124947199847, -0.27109698753155259121, 0.59406467140539953586, -0.37687569005196486227, 0.54916536012897698704, -0.33032659564468003222, -0.023675626867808618331, 0.086607532142516735485, 0.0093702900437125369315, -0.092825827932674326304, -0.20350067384114578495, -0.26776418454205191955, -0.21646176063378208543, 0.10907324693227138379, 1.0134714904717603279, 2.2480872247822030197, 0.20658204352270792548, -5.4132556649793368564, 0.20237098705322922876, 2.1824581739757631382, 0.1748886230300961131, -0.038958559985871436815, -0.020853320184114926272, 0.020006475237904886022, 0.081121594416691700236, 0.113468909141813179, 0.048765676731222340479, -0.038529310544566763852, -0.27141609742288891338, -0.59151386325965671364, 0.40922997206619571298, 0.43150517679716887676, 0.45915380265760641354, -0.55075640434072403728, -0.062846167250659876191, 0.053922127914620629485, -0.033094836444918562801, -0.12424571486161201106, -0.20123196478830096545, -0.21255290246236552276, -0.09878394060949202693, 0.25834353436474710719, 1.0819152200482018422, 2.1170517126152579301, 0.032939608646363138789, -5.4463893367122970801, 0.036902496085000596748, 2.1043784916610910329, 0.46992968617832137079, -0.065634137512091406119, -0.055538861016298013418, -0.0049993106328531378146, 0.082867634623575603392, 0.15483642948099352044, 0.13977799904304644141, 0.087547036241536480006, -0.19966048745832787326, -0.6950370978500342023, 0.25380398964849038013, 0.40138441624442766642, 0.31429001069063239315, -0.61041134622567350121, 0.18540617548228749412, -0.042280162294038553528, -0.0092814466665659730443, 0.047671300712315053949, 0.12777208487493774136, 0.18815846009313491605, 0.089700569824770398641, -0.27522897135245399935, -0.77033373824203366098, -0.52069792805138359615, 0.92371041731971359479, 0.29524113593008216583, 0.94885420768820705195, -0.48431437974338636954, -0.54034813187562258374, -0.01515789876694625081, 0.0047407796377685980024, 0.029659153184497442229, 0.053898349928783881313, 0.067503835044406615418, 0.030271521190804798196, -0.13326069815561414345, -0.44782510613295123392, -0.15996581160668146837, 0.78771267953647106097, -0.49790803603320943349, 0.79970154470964993632, -0.20142897765870926285, -0.32743547676178674788, 0.029698283900188759271, -0.050628203789992540229, -0.11620337127277270695, -0.14269860019079294577, -0.10310481444613692203, 0.01691783810344649891, 0.19098418118601664939, 0.42283842161068629917, 0.6387853166846865216, 0.75307272419660564733, -0.12174595413397372445, -3.2838288074567074659, 0.20303099203295679742, 1.5369934536685776472, 5.1512570730721123947e-05, -0.0032227933213502153731, -0.0023934761030782640903, 0.001168199203218217401, -0.002108592346151019585, -0.0083526325508130640218, 0.0089262499893033933618, 0.070353311349477004599, -0.031338915557918615995, -0.17493399400500728169, 0.37500555236574079077, -0.47987187937966568896, 0.35067500082616048784, -0.10576602030469053894, -0.00063226365885688743784, -0.0017557662191702290248, 0.00039014068083228266998, 0.0015658201269529779635, -0.0072486221654681146087, -0.0081463233538721455812, 0.031760927995151559677, 0.051528697484494408176, -0.14481737602793157138, -0.03636775450766989437, 0.54257578314644105966, -0.89307358973061734542, 0.64051528326343420972, -0.17742192188361938232, 0.027616544825089042042, -0.0064385811244743774984, -0.037369185258720377663, -0.055848173502049805461, -0.046367845836954302285, 0.014891756481588191985, 0.11187739722156457567, 0.17052365810720285233, 0.0011428195977236932368, -0.11574262705498082615, 0.32515442625630480711, -0.93108754520435610402, 0.47306159130882979857, 0.066215861242797005581, 0.013785444503429651261, 0.0063510132013316761279, -0.0080894415914133273315, -0.032275597855062544694, -0.06176634656529920997, -0.061259602459661333107, 0.013991169284947087975, 0.14415757410775348335, 0.13772799389942919079, 0.26950494455307666408, 0.38472659205268183724, -1.6402374139989155388, 0.48072806268278484021, 0.35609093509514350151, -0.0029038219304229331494, 0.0042218301976925075303, 0.012296857316033932978, 0.011158769919702687945, -0.01549003215748644309, -0.059654269015251580599, -0.075510160672644294211, 0.012592325980489328316, 0.10906121892573154841, 0.29713284827381680753, 0.4479588235539945229, -1.3825536320056357464, 0.4100831058756725378, 0.23685815645660274198, -0.0015723914583178682722, 0.002572791345770999126, 0.0054103363070471619661, 0.0046302791475217441181, 0.0010247982236819218847, 0.0077662651234129851871, 0.0096231568599040393647, -0.039949074890893637724, -0.17533566572562661601, 0.087689631485389440968, 0.6287509016272198048, -1.0518977797404096552, 0.73436318096275199352, -0.21143084084113186738, 0.0067343802443974552785, 0.023673500971872839249, 0.027805808880916205866, 0.016778063666353620897, -0.0097707725047071411484, -0.056122252821399154499, -0.13357917330701857273, -0.14617008136820702768, 0.0095011601714045795797, 0.47542375886471305302, 0.66318487999795994181, -1.5606332437267318447, 0.61922013586841129218, 0.081379427399035045299, 0.0096107467917381905059, 0.028195040435128725792, 0.027234082156285359388, 3.0631040896379746474e-05, -0.051560784061961642544, -0.10890021670176179558, -0.14762914141331440199, -0.072008021999119584122, 0.20655810446279099235, 0.79874425470788934511, 0.5881318974935129118, -2.3529705198917048037, 0.5808943309123523191, 0.51300373548747246755, -0.11941236261557844345, -0.06743742503393237997, 0.047377943315078496767, 0.22621413389194738874, 0.42302646329155046967, 0.53821589841260197851, 0.38850177123830331727, -0.23497319802192662141, -1.5016842517680066837, -3.9730493519263179181, -10.564431132454920714, 28.756120195078231916, -10.799777329340987109, -3.1876182378573920673, 0.025999303427650943821, 0.043818605402376836766, 0.03832916337121217154, 0.0023454295878123643249, -0.064588522627043953861, -0.14788111358550859453, -0.16537094130482765664, 0.084853251740831803218, 0.28879622168489971212, -0.071217322168100172242, 0.24978397894388093547, -0.39984171443808746993, -0.066274172543047960016, 0.20686487837079087226, 0.0080314713935129861816, 0.022230568466689956675, 0.019753336274649889448, -0.011330920382662028334, -0.075434411689967950654, -0.13814343789780170191, -0.15632250222693236386, -0.057833435321191221934, 0.37530907371862398669, 1.2958855188623901178, 0.50813423530613777146, -3.3559422956605313715, 0.59321272241389588764, 0.99024802409004220038, 0.00054233670520047519686, 0.021901440566381834979, 0.027278922352392782208, -0.0060963295474326835433, -0.094974231318681184844, -0.18544110495323989918, -0.093485019762289858769, 0.36915647992236799713, 0.56625552015155311558, -0.30307898226146728904, -0.16979508530005116285, -0.223428766167297288, -0.51664743107129618149, 0.61422133365201059263, -0.014402098583931759179, -0.014691130700227355427, -0.0040093128084344401701, 0.025584186397230989501, 0.076730843300352749581, 0.13178391836406455617, 0.10278406464607225956, -0.12630864449219775247, -0.59267271110706531978, -0.57495185746431787255, 0.76628889362046037625, 0.3587912740256762878, 0.90056683198594833506, -1.0459291094815046375, -0.02987653692689745677, 0.036010718195449424595, 0.08472423618755674235, 0.10327041404905117072, 0.076963488264891063984, -0.017575695025329740595, -0.22041886314939315716, -0.44812910433597008542, -0.4466963835845536579, 0.28786576553231935094, 1.1218967147652523231, -0.7518244966816685082, 1.0143557121060977622, -0.78881019560404985747, 0.002877832864288273635, -0.0030782780457273969596, -0.011857814762045233714, -0.016813440427088419127, 0.00036513424541694330938, 0.059967603738476850384, 0.11441700631556038503, 0.032935471425869930484, -0.30767974103416140252, -0.30729490553576499501, 0.5491653601976108634, -0.35587350769636627579, 0.7280790415134972271, -0.49089850470637591551, 0.054468642889974883126, 0.039889755350230260422, -0.0077360002078409040305, -0.09282582815447018576, -0.19739178998814999044, -0.26277652331364353966, -0.21611932190464971937, 0.10240780012449007819, 1.0050205422063345218, 2.2448731609523271402, 0.20237098780470583193, -5.4693326931940315916, 0.33088325660648315596, 2.299858989498705153, -0.017863892046710818512, -0.021017680401444874261, -0.01246429570769549966, 0.02000647503562247298, 0.076723524916626928416, 0.10906120967826266055, 0.046449165237719136634, -0.04358259078242865403, -0.28774143445334277036, -0.61363197515874112753, 0.4591538032813442527, 0.52986812441955799535, 0.39510087577536023851, -0.65359014591402286865, 0.059577781225901044193, 0.011672145919248824719, -0.050412124462432233996, -0.12424571478452386963, -0.19452400406698222568, -0.21057133877438005687, -0.11073337135939986808, 0.22941257009403187572, 1.050675532420257019, 2.1037636206778618231, 0.036902495951228353466, -5.4685627747627663098, 0.25179077278243949811, 2.43467315214007618, -0.013111948326513012814, -0.044548172266866511992, -0.047678979126859699633, -0.0049993104802613652637, 0.079359742372974217961, 0.1484928653164462875, 0.12722534020277342925, 0.062885179798729873291, -0.23559867656270816472, -0.7231041735581593688, 0.3142900107781229635, 0.5245481475780728875, 0.30938818901172526754, -0.52202836629667859203, -0.026326162552620727131, -0.021719290983829311892, -0.00068423281365691604461, 0.04767130128354072488, 0.12508137169174740611, 0.1888918204973614734, 0.096297367741419059861, -0.2718263739564231174, -0.78778859450214255844, -0.56188360712868057334, 0.94885420686455523853, 0.47330676634990997975, 1.1161017964661337221, -1.3391778787220283142, -0.013195784336210488602, -0.0055476190443989395473, 0.0084161653203230681747, 0.029659153221119748112, 0.053020027461455385143, 0.068232970116483757095, 0.033698603361897672137, -0.13031971081452126127, -0.45073627834614832244, -0.17161019550206840778, 0.79970154438644680539, -0.44892908634250727751, 0.92835867855571430063, -0.7039018943500890968, 0.074803884993262625969, 0.01641427104958701963, -0.050628205259802862293, -0.11740145581845518041, -0.16214620550032388113, -0.1546518444992986685, -0.071830361514812812729, 0.076368699966515740662, 0.31688714020234365343, 0.62883646756943389899, 0.90548916901979137339, 0.2030309915849524971, -3.0416613757583181865, 1.2632233875360066033, 0.002203905976895187066, -0.00062790004266668835407, -0.0032227929631399600784, -0.0025361853766954936366, 0.001048659122631242363, -0.0038056398606527708228, -0.014847760235153027036, 0.0043720927199459649892, 0.0814913950309355678, -0.011045959633689811721, -0.23216765625722171618, 0.35067500132919426914, -0.20709296848725361473, 0.032651150839617337773, 0.0013668235600269229221, -0.00096336131458402227828, -0.0017557656595773660918, 0.00048694755880036237183, 0.0015293675846304462772, -0.0095331029585805047988, -0.014131359138018141641, 0.03374171359764129019, 0.070126439713504518103, -0.15267844709039504747, -0.1255405987062919726, 0.64051528375123123471, -0.6760345510974415939, 0.2320720364594565932, 0.020778497911009795318, 0.010134087795540282911, -0.0064385816741441039002, -0.027762398555901790509, -0.04713422239704568445, -0.049345817511259965071, -0.0060886736826426396396, 0.078770265096090766321, 0.13353046807617183211, -0.027731293584286463572, -0.13574291851360209549, 0.47306159213180798417, -0.59480943684803944116, 0.16305934488821902106, 0.0074694804136058749397, 0.008948563195296077577, 0.0063510134440151848334, -0.003979842300114380324, -0.026722881367722120527, -0.059757127518382212317, -0.068251650308447278648, -0.00066205945088985279634, 0.12143949617935577423, 0.090675628058546300925, 0.18038160545272563517, 0.48072806329642275847, -1.4019793757750618202, 0.6520853014734032671, -0.00081942534719513997025, -0.00099108655722913497656, 0.0042218303043666284557, 0.01249476971635863666, 0.01173730645156878527, -0.015579156921186849277, -0.060422024841272389817, -0.07199005165459387845, 0.019608834088258163264, 0.10445790220750167387, 0.22271970313296721167, 0.41008310590196334067, -1.2899208963892589974, 0.65348668534345111247, -0.0031367956093487337453, -0.00090326101406089398006, 0.0025727915731817295637, 0.0056164989679973126649, 0.0051631691611304470976, 0.0021906028533496763391, 0.011644601518949181454, 0.020375911267017880052, -0.034773803488795815286, -0.19832583459395922154, 0.041468085710051526271, 0.73436318104968667519, -0.98862685395875915528, 0.40874153132337026628, -0.017612446498170713366, 0.0053964553261656948019, 0.023673501262896586017, 0.032394662073673217983, 0.030078721243704743121, 0.01465629685496667306, -0.026528649835453367578, -0.11710911567330506422, -0.16701094710491576345, -0.03171214299807217224, 0.42236455107098236361, 0.61922013544552445552, -1.6949417039662146856, 0.9344939140515520215, -0.013670113486683222598, 0.010516060226183756438, 0.028195040537370517214, 0.031169893510920761159, 0.012651066800745753432, -0.028646117958212698879, -0.08309144382586629296, -0.13791422924182333842, -0.10713647053293631761, 0.13167272554356909686, 0.71722940370286791811, 0.58089433085910202603, -2.4856700608728177215, 1.358025834418003841, -0.049900379706478453778, -0.080560464655057645311, -0.067437425016050378535, 0.013615593392874436596, 0.17127965797469407727, 0.37431675922386697941, 0.54573985611363728943, 0.52456700738363570835, 0.078300402339397778828, -1.0329007168525072835, -3.520432120451828073, -10.799777329646648383, 26.233509322026336008, -12.316734165838127524, -0.024812838533375996403, 0.014185781636010988405, 0.04381860539681858846, 0.053205213294794162504, 0.033591694174079533841, -0.020608847936940954959, -0.10730957639916260948, -0.15616858903242658574, 0.055122473549279349392, 0.25815149264486536573, -0.23885197188639401, -0.066274172918724033599, -0.028928516652003244403, 0.22173218622758553153, -0.0042915677954591937779, 0.011250690917608541924, 0.022230568715018501902, 0.021445370294249817278, -0.0036615961245275857605, -0.06218624851732409281, -0.12623951644708800157, -0.15591525277822382289, -0.099782012063996158857, 0.27192614282196198028, 1.2289086929385943936, 0.59321272228016108663, -3.651617521480898354, 1.9491704742542041995, -0.0052304074531186999647, 0.0068598197250265115568, 0.021901440813272061992, 0.028055251767715963518, -0.0019456117402582443279, -0.089628763570315239662, -0.18345825891737843616, -0.093874671394689843118, 0.37932666190904812398, 0.57685460778587527297, -0.51637208252227162486, -0.51664743102926391494, 0.54227620540068621757, -0.1609267336075879129, -0.0046324923742081516517, -0.01126835373762371216, -0.01469113020577507249, -0.0086253934763767473731, 0.017366598952104593534, 0.069135291301688842069, 0.13238890785563356411, 0.12265912981330820419, -0.089626170172034336225, -0.55343399715798047289, -0.55527125043274683236, 0.90056683239402091079, 0.47181163431730049895, -0.46131567992278849077, -0.053533997192377530405, -0.010285141970908750347, 0.036010718332861138546, 0.078556449256697052919, 0.10876491096437898576, 0.11135669146657403705, 0.050712070371973352878, -0.1384131894784899619, -0.4025156548523276534, -0.46517991965487731632, 0.20349314650885585998, 1.0143557118916814996, -1.1405536827236217867, 0.67930941255144916546, 0.0019516675302603009746, 0.0012537382797472369951, -0.0030782779698929419246, -0.011707259443261215301, -0.018051529530616346186, -0.0023139655322536659132, 0.058409127432289903448, 0.11788861942153214257, 0.031240683889925145722, -0.32795854980904753928, -0.33032659605952718884, 0.72807904166265480228, -0.13224934655874287515, -0.11256765583216669924, 0.013633506259671052119, 0.03677147043209395838, 0.039889754724750037251, 0.0093702905168385475232, -0.062998698995862523198, -0.16703556086884438492, -0.2562313185624888745, -0.26716288462385412439, -0.043172962173587403623, 0.8038324761382210637, 2.1824581743286630697, 0.33088325663880291394, -5.8653396565124262807, 3.2171626855210346818, 0.0033026028642230333826, -0.010737623627642613189, -0.021017680812155090098, -0.020853320022293397873, 0.004674628288331761361, 0.061454303716453662076, 0.10134407196338598989, 0.046481373147708809901, -0.022280748937869715443, -0.18948646057423318423, -0.550756404227885632, 0.39510087581913833077, 0.74042292104400075914, -0.53944016689894414007, 0.041328957646450815988, 0.034362511111121651908, 0.011672145741974365718, -0.033094836502746999285, -0.10307088133984163014, -0.1887206473918304761, -0.24253093154107302487, -0.196580548124675486, 0.079311866283683252199, 0.8864120518925658887, 2.1043784923394177611, 0.25179077242987907281, -5.7297386601447541921, 3.023660495297463946, 0.026370955978807324865, -0.012800089447403670273, -0.044548172470722957605, -0.055538861040970638472, -0.026851990872387931919, 0.044224582344889240559, 0.10836423765926117557, 0.097337330204807503975, 0.08118814353941750106, -0.1038908264189435382, -0.6104113457963662448, 0.30938818879490881386, 0.84922011216306569903, -0.6909220901552988181, -0.0078817774748985488381, -0.018581194273351187701, -0.021719290666342070534, -0.0092814466175908261897, 0.032734617081195835353, 0.11251448384542350745, 0.19298669471055687752, 0.12824546441755119153, -0.22264751840309321418, -0.72746787281244174039, -0.48431437946944749795, 1.1161017960814432204, 0.38426326651547593194, -0.45233985457153558229, -0.0076195754241245488084, -0.0086885734181095485018, -0.0055476186394230828786, 0.0047407795442499772037, 0.024581184385333576525, 0.050538371084511035658, 0.073835693019957779337, 0.055537855918056489024, -0.096608843000354618913, -0.4365392582939666144, -0.20142897801827086557, 0.92835867858589882218, -0.42863887771099346002, 0.063908057217749447698, 0.093443155423598614484, 0.074803884642666587657, 0.029698284721392076424, -0.048873679350300708313, -0.15734323206184255839, -0.26975362149313564242, -0.33152756933765575864, -0.27173464932187574306, -0.047445518390744775838, 0.35802613233679692062, 0.92132922216467527665, 1.5369934539545988539, 1.2632233873765077448, -3.3935844866063673564, 0.0029337621216802074915, 0.0022039060079911375334, 5.1512331223670755273e-05, -0.0038049940516149379291, -0.0069006015100129608139, -0.0049736883243245910355, -0.0024660530173037046447, -0.010062681776218969712, -0.01330400336786040881, 0.048596277425376907533, 0.08239814028104563981, -0.10576602066707645189, 0.032651150973698493674, -0.025749663563947578632, 0.0021701947238710688447, 0.00136682361440330696, -0.00063226395641893242891, -0.0031824398618730204494, -0.0033804519134342098097, -3.2343818399851087444e-05, -0.0029956454576227742152, -0.016234805482599045517, 0.00067260872088917094368, 0.073339222039958837374, 0.00061415255048815589785, -0.1774219223187107064, 0.23207203662152700585, -0.10924291377114739299, 0.0025929780205646447283, 0.020778497671836919369, 0.027616545033508972662, 0.015972621413905777055, -0.017947279960415181421, -0.068706075274818062537, -0.11494797044104554384, -0.10945716464877013052, -0.0013559676732163138818, 0.19521853477533890331, 0.26248899152326954365, 0.06621586083535220546, 0.16305934520395098941, -0.47651165601664746685, 4.4997067370421988297e-05, 0.0074694804390805233721, 0.013785444320706831747, 0.016579047270801325342, 0.010821622075188061865, -0.012674256709261547127, -0.063673713808021947802, -0.12592647765248835001, -0.12817832906040393937, 0.011326057097252144323, 0.18542834718331097088, 0.35609093485857584716, 0.6520853015553047527, -0.96831225996357372932, 0.001545177844425181914, -0.00081942531372431706878, -0.0029038220099672608712, -0.0010295695882551276958, 0.0095457407870040457409, 0.024900912878522061394, 0.021408200803109966176, -0.0288933466917681743, -0.11218724034971087888, -0.12746066079809015292, -0.018268463566620370825, 0.23685815639336821326, 0.65348668546458299478, -0.67544051299466822957, -0.0036904496111049840505, -0.0031367955188302525799, -0.0015723916411895379024, 0.0019137383430990184945, 0.007317224669124243637, 0.011482217204754145354, 0.0099400468124811354464, 0.010957513644086492796, 0.024571731348155684271, 0.011849850248313044052, -0.14726019843180068736, -0.21143084091049740847, 0.40874153140522334615, -0.11253507766417958358, -0.033568872198793543726, -0.01761244650454199534, 0.0067343801802470341922, 0.037059309219172331673, 0.064131851992865807066, 0.075289773312188154608, 0.059810586512991370522, 0.0053687363334848072979, -0.11843427000777541347, -0.2880452580041799604, -0.33854343816170007209, 0.081379427605095699949, 0.93449391392852154681, -0.4437834217709725193, -0.026767156406089565035, -0.013670113447910222162, 0.0096107467568891293691, 0.041058711606588998988, 0.06864014911274801678, 0.07068930495910548395, 0.024527490906548727462, -0.077809278521877911494, -0.22348878118328541209, -0.33136334466480449334, -0.20568144433585533459, 0.51300373558776113381, 1.3580258343499269635, -1.2221536395946865916, 0.020541255928375815043, -0.049900379671591055253, -0.11941236254359183233, -0.16215940683692142699, -0.1276550927472694108, 0.045199413865569165738, 0.38413225432466380216, 0.83153177485257612211, 1.1820430687735707931, 1.0450084975793956232, -0.1027836873978257709, -3.1876182377191430994, -12.316734165922001765, 13.002855565870682142, -0.065594813594227668885, -0.024812838482332830181, 0.025999303503416261452, 0.080235688882003816658, 0.12009379051293236906, 0.11887196837215847911, 0.04899927966598159601, -0.10852594796004526312, -0.31442688826007142433, -0.31883337541384842995, 0.039099213515556327514, 0.2068648785268042134, 0.2217321861548416928, 0.013838993013218808814, -0.0086853952577071059232, -0.0042915677099144523485, 0.0080314712084248382484, 0.028147275675589510796, 0.048155063213484679274, 0.046920812092612895761, -0.0081783882653817738251, -0.12893345254394564248, -0.27953376603020874125, -0.37170184703928615955, -0.14294698235354005855, 0.99024802424453883809, 1.9491704741347486429, -2.1940680120436231171, -0.0036601993410941831113, -0.0052304073395109153266, 0.00054233650084066617383, 0.01869078848696832662, 0.046684515718136421203, 0.056462306420715972777, -0.011565859768135310082, -0.1937979670423492784, -0.35874632750475088372, -0.10021353112138797758, 0.61537832074020393236, 0.61422133366479869654, -0.16092673364098297739, -0.55564585718309933426, 0.0036067177441482171744, -0.004632492219755924602, -0.01440209880568449341, -0.024595469411981738667, -0.0297323487847607551, -0.01509519278282569002, 0.042087256716271864709, 0.15756122914207496, 0.28045132349818102657, 0.22696215178918452415, -0.29615276690707564855, -1.0459291097150846817, -0.46131567990620719888, 1.2397598304735757413, -0.052625451654969325477, -0.053533997265168650759, -0.029876536924308302184, 0.02442085617995456287, 0.10344364275623238458, 0.18667204267420800523, 0.24082313336006500037, 0.2144762809465680653, 0.01310364553327931017, -0.43112668014487132151, -0.94817326259683465395, -0.78881019545533703763, 0.67930941255659438305, 0.96796828794153322306, 0.00042933804159470784993, 0.0019516676367819436096, 0.0028778327401725797841, 0.00088885755137517495975, -0.0084715176931610854105, -0.026379726896174213874, -0.036985731007229745437, 0.0041533852822479557995, 0.12268255885725613819, 0.21354348328768682497, -0.023675626639320039579, -0.49089850476480467734, -0.11256765587629924408, 0.36485592734604327392, -0.024236063020023214387, 0.013633506138483611828, 0.054468643195673543955, 0.086607531981408053534, 0.084684294292775075252, 0.01419066182465439363, -0.14869658168379951135, -0.38011140209186827787, -0.59078465602763230446, -0.57794733054907498904, 0.17488862246635059061, 2.2998589896262777721, 3.2171626855545718549, -4.3862861903956327225, 0.022029097806196690224, 0.0033026027282523932535, -0.017863891836493644982, -0.038958560030963464149, -0.052967078881841378468, -0.040913546880827343644, 0.025935140897219974288, 0.14174129193706963559, 0.21344368455567208631, 0.17169854942773624118, -0.062846167461506036167, -0.65359014586940100688, -0.53944016686279538941, 0.85298840294281630392, 0.010955760140365828487, 0.041328957594891725658, 0.059577781352731451481, 0.053922127910394135331, 0.009987527516634803898, -0.085581623213760615876, -0.23589505047444289665, -0.40588116010935920697, -0.50450534908628541508, -0.35194374844589554341, 0.46992968583331168331, 2.4346731522193070241, 3.0236604953246866145, -4.7352094823281278835, 0.050548489647570200856, 0.026370955913713665236, -0.013111948208073908803, -0.065634137519497925961, -0.11454333198690677675, -0.12235278117338327375, -0.044002213425209837916, 0.12001492362887591914, 0.27677629390866359538, 0.34807594026502164386, 0.18540617517095694944, -0.52202836610617486812, -0.69092209025587647542, 0.54616194975377096998, 0.0091152964617274880033, -0.0078817772953964793797, -0.026326162690003651612, -0.042280162540441736108, -0.045639858062760330082, -0.014570435615263659102, 0.083389154754014413462, 0.25870849151922703513, 0.39963918941026299603, 0.22057259714219623614, -0.54034813171916629138, -1.3391778786030079651, -0.45233985469624216069, 1.5775903155805364175, -0.00093326034088784301079, -0.0076195752665709068005, -0.013195784594100859119, -0.015157898748030289998, -0.0086862297549172122479, 0.013235553856464210912, 0.055200396463780346257, 0.114484025571386136, 0.16013537858003457082, 0.081441586254608802631, -0.32743547639815684391, -0.70390189443094097754, 0.063908057179264787795, 0.63160339966998757522, -3.0526377724935147562, 1.1694382568971783964, 1.4078261373524907363, 0.83670076117579261776, 0.28892633585007049257, -0.085878051192260759783, -0.26801550464371270976, -0.29834741955251964374, -0.2279794269086573566, -0.12297682828247893938, -0.029641168380397330678, 0.032827292121387824375, 0.064526368407131759475, 0.073403238895402789232, 0.001053271193030931455, -0.0047904407976332545605, -0.098962278644172757547, 0.093763282965678651504, 0.0426027033330745597, -0.017822066922302121084, -0.0089977021565921591617, -0.00079019631338804637022, -0.0041466320804431381838, -0.0064321778037111010037, -0.003515254516111966307, 0.00015176632944679819409, 0.0020831857405711725131, 0.0025363765126212695598, -0.11760043824545106383, 0.22937784254461199285, -0.15469660011731695137, 0.0002168218924262290152, 0.060945874738295595374, -0.0012348062555314156565, -0.014258274573004379657, -0.003144912673244429531, -0.00039049505338101394517, -0.0027014301072397243694, -0.0023031670431025129397, -0.00024568641932934924413, 0.0012458893081087770143, 0.00174460265809761963, -0.48628934534072565388, 0.14502533543860901544, 0.11682376069086496118, 0.29138164334998922778, 0.16863644738332986273, -0.03662903267499723059, -0.12410765323167689822, -0.11067375446615070056, -0.056648014039851382617, -0.0071302841935608381185, 0.021480224949111236671, 0.027887646828719171721, 0.017757349867735715276, -0.0017332263462958739583, -1.0654578599043249554, 0.74449369975360768503, 0.435146846161972356, 0.17406641224627850129, -0.039069266616129007252, -0.16360476694308775936, -0.13294125735091574803, -0.05328344378259173425, 0.0007948105797146822523, 0.019554725067850657616, 0.019395762291218646839, 0.01270902735919926009, 0.004892934194286510316, -0.0023678500868946112008, -0.67395669321139151098, 0.67577073777044838199, 0.23301342069968988824, -0.041258671325122696361, -0.13938065555563583509, -0.10849464416851228865, -0.020025425054169477407, 0.027321849232881362246, 0.026285026848314647224, 0.0083903812948794148641, -0.0026267277479669688851, -0.0039188617506799409973, -0.0010678552248042263821, 0.0019121980207089594398, -0.14024832554342878121, 0.3907326605108621953, -0.16603432305724483142, -0.12144960053129760791, -0.00077041540295102865256, 0.011099489497919078074, 0.0081120533093628228311, 0.010854950753542013872, 0.011747010565914916061, 0.0067682396110452435964, 0.0013912053496430343193, -0.0016190944979567005717, -0.0027337985113477228025, -0.0030365133554045814317, -0.48373677614939342151, 0.93538594155059595625, 0.099541227088457626127, -0.32355286387559012429, -0.28187363545509430685, -0.11643728812426698549, 0.0084679323452691874358, 0.061997430116782312193, 0.073215188922121698933, 0.058463822813425082858, 0.030957834799318493385, 0.0030604441287054745015, -0.017305423270349673537, -0.028820103665530658055, -0.88815827643735412167, 1.0752202786994107253, 0.34002284074519145696, -0.19069970916740891509, -0.27608070288448233187, -0.17599786583161242115, -0.047489415452106269189, 0.034695295007854412561, 0.066102949160699725484, 0.057901110121192971014, 0.030837628579380976096, 0.0035389774445315005935, -0.014276238964761781003, -0.021872986545574492345, 0.01383899255538352524, 0.22173218637665118025, 0.20686487887063156887, 0.039099213021096271281, -0.31883337503233333265, -0.31442688781596095371, -0.10852594776468710291, 0.048999277961656553804, 0.11887197034816379682, 0.12009379039219891805, 0.080235687543114664333, 0.025999303240538687576, -0.024812837997125702233, -0.065594813258319700777, 12.436896380260478168, -12.160754272981472823, -2.9610352547988880723, 0.075017979751165905777, 1.1790933283629805306, 1.2137562829852761581, 0.7764939626636592207, 0.30198452011002219209, -0.02266560596926441018, -0.16605046589200300744, -0.17228729544136078822, -0.11018127772554120458, -0.031494242409927819537, 0.040262051265378517861, -0.56892945597786981349, -0.13185938714634473645, 0.63010484591775484109, 0.60276793588185972883, -0.12051904591542261791, -0.37070924795885129566, -0.19530045303252555056, -0.0056524853038742014708, 0.06484800268802327905, 0.052540857228320941852, 0.019959371755661564629, -0.0014169028809457103087, -0.0077282962219094022624, -0.0045477873269741675433, -1.9942532618285102597, 1.9774077463556958278, 0.76362028449634467009, -0.30221330543064245377, -0.33123893922692948522, -0.18039494031970437771, -0.073371251610454699121, -0.00026069912722120652977, 0.034290371886326255491, 0.034727431232823333107, 0.020599272245420342103, 0.005751492928929890884, -0.0035295937286677417366, -0.0065329175855408302964, 1.1476589356509125928, -0.45926949462487270948, -0.96172836529827354646, -0.24668265292591992632, 0.21822763320726859537, 0.25832765568000970458, 0.14364958287384468893, 0.035593162258743454851, -0.017833329628332089678, -0.030460526405429531882, -0.023962388233274095617, -0.012907587573495298006, -0.0029813757590713063292, 0.0047363659228500344309, 0.86102169936614725287, 0.5745157144995115539, -0.73299131916371362827, -0.82628272455131335228, -0.34969747011821139537, 0.039512156949552804108, 0.20204509913964960321, 0.20838445032574151594, 0.1516977712820380686, 0.078164346074442822565, 0.013414787252509528109, -0.028369347994726512585, -0.044155659450037534897, -0.039842902273237702782, 0.40929723838602777297, -0.16981925098753966608, -0.50130802238923044545, -0.0064484153272237896073, 0.22345363547981503261, 0.12537421409159346664, 0.0020458180985068446373, -0.040835327352427640279, -0.028300763410421293786, -0.0078981293249013283242, 0.0022427554405337036224, 0.0036417287951818729311, 0.0018829126966705420423, -0.00016837045187587161652, 0.8650739005088260436, -0.5587307332713671304, -0.6617742183680577206, -0.054722300380896801708, 0.18026972913993069558, 0.21704701409639604037, 0.14197047185527600366, 0.023912995900326347587, -0.043719603571493396987, -0.054842228036602290453, -0.039348326505771670514, -0.017325993023306951457, 0.0039976901419973322194, 0.022381059680436358322, -4.3905989901059818692, 3.3751799764727858921, 2.2620335755591680638, 0.065094672853149862513, -0.62459409955177624951, -0.59790533415460511701, -0.36924113003908853869, -0.12629776777714374347, 0.036279517002628061673, 0.09760393543695390639, 0.089440284462539848542, 0.050817804263906717066, 0.0077580784475586358345, -0.029200535158670579833, 0.5098236626369693214, -0.67106876068933540846, -0.50410711552616427333, 0.18708728692176035402, 0.34811904348304079093, 0.27909182670066834397, 0.12004673512475136365, -0.047403259013372209263, -0.1267844362037432171, -0.11742467085947044803, -0.065953707871953648478, -0.011619156907060838174, 0.028028686862518845113, 0.050901221037291999316, -4.7053925512764687156, 3.2833263449195291273, 2.4128768679751213533, 0.28030214401625402898, -0.48756561838611295334, -0.55670290521149845553, -0.39518996236435754987, -0.18754656113840093012, -0.031101602113858220489, 0.048135797116102010951, 0.069244014180254942481, 0.056366552650537242453, 0.02749942790707791615, -0.0064940594404081126989, 1.5078630756890309961, -0.49970943078973883988, -1.2676678249763639883, -0.4619673467334944883, 0.23442780767107582185, 0.38115822056140385987, 0.24009180641545313506, 0.070459412816369318411, -0.023103791056595195802, -0.049750693998254255646, -0.042088834774650425719, -0.023357661325792297702, -0.0044323019255786454476, 0.011108590645101824174, 0.71093238960602955778, -0.04222089921496698095, -0.74686885662904878291, -0.29795012397173548102, 0.11987571416179829131, 0.18320188189939831669, 0.11808907886174112301, 0.045694957706091741956, 0.0012755032942221311305, -0.016338980946916669551, -0.017468220676094835947, -0.012015386471915920699, -0.0051262856576715348095, 0.0013361990110852366154, 1.1694382582315687724, -2.7496011879313719817, 0.18724560099222020693, 0.78666202587925693201, 0.58283688012600487482, 0.29478056598447016645, 0.055832599689406167687, -0.079170190623213093639, -0.1454399611222298716, -0.14365041575159306908, -0.098790124740303186801, -0.03874379287487145912, 0.017836654598081001788, 0.064526368737859593483, -0.0047904414385590964737, -0.11722086473853791933, 0.30266162988431144498, -0.25566485083173984272, 0.009641778317015637853, 0.084919626022627481277, -0.0015346418238484016081, -0.015999116585350078162, -0.0025820072231135995228, 0.0017948714813876011885, -0.002189050807977226034, -0.0030785293301696659007, -0.00067535641668608331883, 0.0020831864766934432912, 0.22937784141442496555, -0.65029089795226147075, 0.59403824118415948785, -0.11097763276230965257, -0.12974293791311272694, 0.060573646984349097111, 0.026208754018072218106, -0.012163877937987394376, -0.0080710766760398282293, 0.00093275492677893120756, 0.0002002911870160887153, -0.0013758471789994050627, -0.0006181057613772867167, 0.0012458903811021820787, 0.1450253354351623003, -0.50946686968675958251, 0.39253146077529199598, -0.16264997066384995206, 0.018173435001114519871, 0.15164315353892848792, 0.066199880279940476191, -0.020372477071023104012, -0.054740344557451955931, -0.045057112448552906481, -0.022558474269415429436, -0.0021360596308037868574, 0.01118699658503327643, 0.017757350871724982, 0.74449369956501620926, -1.5262222502801563451, 0.44066466188391956171, 0.24468601341703180574, 0.13395614848418910259, 0.11638428778074505487, -0.023260923102518905881, -0.083861758397436050871, -0.062606044305152680485, -0.02186991347643764752, 0.0020844688843075307588, 0.0099550895859692681983, 0.0092418869396208858102, 0.0048929335160801789434, 0.67577073765287876128, -1.3272552265587724829, 0.41493300885686262847, 0.24729557566535573243, 0.10245890473108804075, 0.0067094124104079009077, -0.078206483102979160171, -0.058726541552373570654, -0.011381351207611042561, 0.014409508819865506773, 0.013037500547050023814, 0.0036072013787710834921, -0.0017486079587100405042, -0.0010678542001346727486, 0.39073266037958825914, -0.92803213093954617374, 0.66137985225689333024, 0.034600404169422366296, -0.15371898410735934659, -0.02545455816606434149, 0.0093121866538484459469, 0.005114926751704356421, 0.0013375518892448632659, 0.0056289614479316666187, 0.0057469470388081126191, 0.0023985741345284566223, -0.00094813256321215439934, -0.0027337986295453277179, 0.9353859411665567114, -1.7103974730602560239, 0.60869358916053617481, 0.42746457856908931516, -0.018474959856481280862, -0.15625079266759778207, -0.11655098271783881358, -0.029429821454188549723, 0.013843985940517229694, 0.030012532540307085283, 0.03126716057224399048, 0.02185264221589660813, 0.004157671071254387743, -0.01730542336667766079, 1.0752202783023949717, -1.9799603092882496203, 0.54048197723337831899, 0.51320735225605440544, 0.089737456169653062599, -0.089321748257885677513, -0.11679773686151469214, -0.063691502890082990485, -0.016324778165968424626, 0.016011668388344758862, 0.028631008274213255727, 0.023752435464029303391, 0.0071599744738129458485, -0.014276239312323029881, 0.22173218641182165789, -0.028928516739046589157, -0.066274173078164327899, -0.2388519717135874354, 0.2581514926817354838, 0.055122473087292012117, -0.15616858909019126767, -0.10730957553994369125, -0.020608848621616882024, 0.03359169422427588092, 0.053205213600128142826, 0.043818605611140731826, 0.014185781196080299305, -0.024812838466430370277, -12.160754273128238978, 25.767588691264357692, -10.744483510406432458, -3.3564426986901367833, -0.99083308423959226552, 0.13907127305691974195, 0.57820966998206690945, 0.5476583116897449921, 0.34208244551774741327, 0.13277670067039870627, -0.01539008020042487157, -0.080580893124927194249, -0.076994939551772625941, -0.031494242097576799488, -0.13185938688312981437, 0.50758073409262671749, -0.52572462458918611805, -0.49728148210543232643, 0.58507040264017096121, 0.37352750313158672446, -0.1002817296087007426, -0.18709980347935850631, -0.090910814013594604122, -0.0002515829127025769052, 0.031612131815520423672, 0.024703009151080847217, 0.0071753633327311374876, -0.0077282966231822583514, 1.9774077468434319016, -3.8075718311786479475, 0.83276075012174788359, 1.2903367286488927057, 0.096982683374129352583, -0.18525094294069358036, -0.12631166305417135254, -0.077750948140221903815, -0.037461252688485373319, -0.00092529120316245990624, 0.015099834446727456588, 0.016018624083497256388, 0.0084468056306727726246, -0.0035295944383657903293, -0.45926949483464585278, 0.4935766765145824797, 0.81649836326181401969, -0.5487553937331923759, -0.4931266513421004305, -0.070272058096142175909, 0.11294544081414435721, 0.12119494887693851859, 0.062906197252625362104, 0.014655430505642934313, -0.0095321792750768536556, -0.014706146187047065138, -0.010515793865443883934, -0.0029813763572201887504, 0.57451571419233138371, -1.0121821650339637433, 0.94649047976313005481, 0.15253118390044129171, -0.42111638654336736032, -0.35203871832904315431, -0.11377722897177001971, 0.054668165670667040468, 0.10217081288707498266, 0.093246642912736113495, 0.06350094327974653563, 0.026846493608403240033, -0.010551918032097722458, -0.044155658931246830057, -0.16981925073669157378, -0.018929936844992258094, 0.69420533142450258079, -0.36783473708075292707, -0.32278289173743140994, 0.037124426582147540543, 0.12164222975049975017, 0.061359430876747424188, -0.0027048314484838954092, -0.020137757396610514449, -0.012768865308030591388, -0.0027607167696446889907, 0.0018782452082041641538, 0.0018829119054014258408, -0.55873073277849283347, 0.76973654504652377906, 0.39589640625870925472, -0.56678130190181152948, -0.19143698082621626866, -0.017329314541188411969, 0.048836183023681450921, 0.10235606317628392903, 0.061940456091188922372, 0.0039788909718755094413, -0.022164907875507788848, -0.021876014536204801963, -0.010742032411717568341, 0.0039976895140170169013, 3.3751799764404566417, -6.1545535914397708765, 0.4583961922238807829, 2.2875207409844136031, 0.73774026056702168841, -0.085679259132049073733, -0.27398330944014509525, -0.25813394877289208562, -0.16270875274400675714, -0.052789712178236136708, 0.019040258532716375361, 0.044773108289618862576, 0.035970252430230845175, 0.0077580785932889872319, -0.67106876086850042462, 0.80893983305087358016, 0.30609769795952623817, -0.58754654409349948985, -0.096597611350806938946, 0.077225308577536261323, 0.095617280798104581363, 0.10909718358333357624, 0.044486065936071711446, -0.027831285339774997079, -0.057234916756066303578, -0.04593708929026085741, -0.012924099835423717708, 0.02802868802009130661, 3.2833263449169929338, -6.1160483683328275006, 0.36862458578171469759, 2.2657282890844108181, 0.82774679484170088362, 0.0011939982628718093238, -0.23590834607499366893, -0.25732768095827485144, -0.18050291952885655355, -0.079863049759799992211, -0.0093115584706085219313, 0.025827310047030441198, 0.034766655086667666574, 0.027499426762276901748, -0.49970943028159758859, 0.48425923809577642309, 1.0299104784212300689, -0.52281543940003583071, -0.6696754869551334588, -0.18506236208165316448, 0.12626074723587299431, 0.18138452517039554968, 0.10446852340181961294, 0.027752720783619920508, -0.012572717512937883755, -0.02319319641165035889, -0.01770292244323571143, -0.0044323024261379607991, -0.042220898796318594159, -0.25309919821781767801, 0.90769685610325911984, -0.26943418885419911923, -0.44617328071866435568, -0.08524709346544534494, 0.069463762056671540801, 0.084065029750561251487, 0.052330526305292709988, 0.02011356799337265383, -0.00065247601641426657137, -0.0086489043476379035003, -0.0088049620794914094934, -0.0051262857569961267634, 1.4078261354729606492, 0.18724560425236011496, -3.0158415015870834175, -0.083149551336359872611, 0.67481145568384615085, 0.60386893040387301657, 0.39660170485077567148, 0.15863015311086031689, -0.0047402061065428650888, -0.1060415531789022231, -0.13181505718278843653, -0.10072323223092137512, -0.038743792255423945003, 0.032827291950530003428, -0.098962278021293592634, 0.30266162893694553704, -0.40152578837357527641, 0.3481126519154755794, -0.19991067709393797736, -0.015602957300929802542, 0.074245144127441625725, 0.0036798495902210341668, -0.0098364124762285384818, -0.0011862021989920363066, 0.0017463046970455729096, -0.0021768160119979024919, -0.0030785288924750704893, 0.0001517654142044676485, -0.15469659786413980651, 0.59403823936851041232, -0.85331260411776055363, 0.51316225886601141148, -0.032453940551492976851, -0.12436768649208401749, 0.045993091816006526074, 0.025079905617097931342, -0.0072360826795999418776, -0.0063426998836039055751, 0.00087026490413533467635, 6.4406705279693296159e-05, -0.0013758465255371589431, -0.00024568802279330195087, 0.11682376078474809855, 0.39253146048943315938, -0.88117307028840941019, 0.27985091203514389502, -0.12579915014074902202, 0.050318691918806321417, 0.18962054175822143365, 0.09721921423653470673, -0.0033141136349212479972, -0.054942067839601163581, -0.055282183204781128183, -0.032426933631374242617, -0.002136058988543278192, 0.027887645817807775461, 0.43514684658088798042, 0.44066466143161603375, -1.7448358153028036277, 0.37082254548921850246, 0.34594306101900113015, 0.18551741752224776283, 0.13847937248064601246, -0.011807488526465528822, -0.080218919496087531806, -0.066736407936253405726, -0.028117787458538490153, -0.0018644101560276711907, 0.0099550887854090702039, 0.012709028848573051054, 0.23301342066661975827, 0.41493300867257193465, -1.395015331982460749, 0.4571694981487177678, 0.31786108740241209025, 0.10388423340340660617, -0.0019805203502342369482, -0.082351422996649284447, -0.05805514988598411974, -0.011034045606109300908, 0.01419669241828541012, 0.013070496296426235144, 0.0036072022579832180127, -0.00391886289904956088, -0.1660343235044237642, 0.66137985214754169139, -1.0051562778426130773, 0.58454280352182252933, 0.080258674286908290862, -0.13680599186888492413, -0.030283360832599436063, 0.00016658805462937444075, 0.0018883817928756186331, 0.00035083603222503664462, 0.0051309533211459437685, 0.0055386176828914673517, 0.0023985736978357750405, -0.0016190940048226363894, 0.099541227928696185412, 0.60869358853402899534, -1.579798326538485842, 0.65395641128559001132, 0.48196363551734627828, 0.01787083599491091912, -0.14030887664686114702, -0.13382244062779657945, -0.057182255513067027641, -0.0078181931343056232375, 0.019429155422543066545, 0.02846968317071414678, 0.021852642372972057999, 0.0030604439738415017275, 0.34002284140715721694, 0.54048197651540330622, -1.9103079021718742947, 0.57171895609243872372, 0.58901537665844028169, 0.14195786924910369642, -0.066645674580136538001, -0.12967122511384987504, -0.087734201304296974278, -0.03522543359101159427, 0.007386820522252958432, 0.027113398986008459268, 0.0237524350586890802, 0.00353897824282475848, 0.20686487889927815398, -0.066274173164591748675, -0.39984171473763863158, 0.24978397969031485015, -0.071217322515655046011, 0.28879622156198403449, 0.084853251356913572678, -0.1653709402764209635, -0.14788111439406848024, -0.064588522785372970425, 0.0023454302169888530116, 0.038329163301046742518, 0.043818605734405491337, 0.025999302970559426407, -2.9610352549397576105, -10.744483510273873605, 28.278568919030362849, -10.646389029101721135, -3.8521357908124369551, -1.4467636525908997847, -0.1530682652889991946, 0.4671460453316484962, 0.56202827515391795554, 0.40307419225150964426, 0.19094085431489160598, 0.017249936944417787466, -0.080580892570257106611, -0.11018127823980813917, 0.63010484528169996565, -0.52572462429194222189, -0.24816905343258163774, -0.16951478579916967604, -0.27923808822613754854, 0.57875470505545278943, 0.36354339145901726216, -0.10371656485445286677, -0.19267409146181491142, -0.097415118954098017801, -0.0037676056366029442302, 0.031826472887491671915, 0.024703008300401403663, -0.0014169014551950482456, 0.76362028358050360044, 0.83276075068503696563, -3.3730904213774284273, 0.66055709179960686583, 1.3161109650056366238, 0.1984438838794370108, -0.14632992315210258116, -0.12922186454532416655, -0.088788553928195054765, -0.048264013351507199789, -0.0064526083935294593213, 0.014329276316151228626, 0.016018623063714404475, 0.0057514944124035646475, -0.96172836422683050106, 0.81649836250123575265, 0.356523423779059212, 0.70561766136320203113, -0.56765452068116106776, -0.53486554053632595629, -0.1038644364822860211, 0.096539560305230129322, 0.12237959960328591291, 0.070933004883786107686, 0.02247477174316986695, -0.005425124349374570859, -0.014706146901592402967, -0.012907586786058338618, -0.73299131893537949889, 0.94649047974899724878, -0.68604240026211438241, 1.0347055612443540884, 0.21421458977815632796, -0.41674425055847841293, -0.39463178575938634429, -0.18342485620043713634, -0.00090072776281191480052, 0.076179482453406333065, 0.091238091709131829976, 0.070116025444434890157, 0.026846493685432799847, -0.0283693478529447006, -0.50130802274796504303, 0.69420533150236962694, -0.26014045816039721437, 0.52242409821452218122, -0.34844429852561276117, -0.30386076046568716302, 0.041346972229923728559, 0.11919352671579463987, 0.062353692028740831388, -0.0005871322619010362065, -0.019093541228716801039, -0.0129595243204119661, -0.0027607176713591958725, 0.0036417298202670502633, -0.6617742196272849986, 0.39589640693575894881, 0.54921944586155835299, 0.45856094292993743178, -0.62985642364919502967, -0.29202090858147400354, -0.03888211606150981936, 0.050270726541792098985, 0.11114378641591066144, 0.077475018331653822123, 0.019203333353033535946, -0.013946160967465121017, -0.021876015171399407211, -0.017325991539767750249, 2.2620335753605473883, 0.45839619240720547166, -5.6745211775085566686, 0.30003494803878583141, 2.3671250569006012654, 0.95622775384817448874, 0.053363498685667698496, -0.2339304234657829662, -0.27167614469166273539, -0.19624771798941098577, -0.082983459565132255964, 0.0028104596743947218851, 0.044773108772437106295, 0.050817804061502713786, -0.50410711507106287232, 0.30609769801160502389, 0.49538064390280578397, 0.30858149120974182056, -0.70689856305883291832, -0.22956734630600605929, 0.060681539436857778735, 0.12788261139561429847, 0.15113467813590120148, 0.080231255051234295905, -0.0065772448144991169094, -0.050228146277732838021, -0.045937087560962014399, -0.011619159649352532246, 2.4128768680842425098, 0.36862458571096168303, -5.7266765211823287274, 0.16329349357901762341, 2.3039403354985470429, 1.0221417687407732267, 0.14530092478815981316, -0.17006774887260400764, -0.24361908735490023226, -0.19799639789666156897, -0.10568959212069667375, -0.026840693472571508504, 0.025827309467834486112, 0.056366553194968402329, -1.2676678253770976568, 1.0299104788524959808, 0.51541601283285276569, 0.88094584303634282385, -0.59604009591182771644, -0.73525747593463464113, -0.2318758061910979329, 0.100610277172623927, 0.18166801324561707154, 0.11832340712506549119, 0.041945794897749813512, -0.0051011356852180434612, -0.023193196545759742211, -0.023357661462867142232, -0.74686885691295512579, 0.90769685605082894853, -0.30150141785171685749, 0.77484558209708975784, -0.25134108520318082514, -0.46804013597277921432, -0.1175512301112413005, 0.051653341316475337441, 0.081597913711602312881, 0.056602909512346158416, 0.025720740751263473922, 0.0028061531194330498543, -0.008648904698977407865, -0.01201538680613957466, 0.83670076216292077476, 0.78666202426885689647, -0.083149550830357937814, -3.0744219411899513794, -0.084857045863973384892, 0.72081129897748785762, 0.65277068422305939421, 0.43617331771269862717, 0.18352935411132034749, 0.0076809291376158522757, -0.10242296150587128223, -0.13181505657483708061, -0.098790124936730269734, -0.02964116896626498715, 0.093763283460662219526, -0.25566485073524891636, 0.34811265177794542547, -0.36641082775594296939, 0.33536675683105582335, -0.20740892719181541759, -0.016310288550999325202, 0.078551050165760832877, 0.0053655621302947289347, -0.0099582892533369490179, -0.0014452504407327180364, 0.0017463041447530304809, -0.0021890505924081313412, -0.0035152537709627618145, 0.00021682034464899619221, -0.11097763199256364952, 0.51316226026746325495, -0.8143786936642649632, 0.51610736296688253244, -0.03714093878306828217, -0.1273759879153747443, 0.048140398972614871498, 0.026766786145022890275, -0.0068886489576595244175, -0.006556583343035131059, 0.00087026413135781952035, 0.0002002920456077508923, -0.0023031667093921528132, 0.29138164358633811268, -0.16264997026426514237, 0.27985091163162872929, -0.85042853656846051358, 0.3275602398336476262, -0.091815689617189083505, 0.058246496558606114524, 0.18883525986717600298, 0.092069575809430981295, -0.0099131931791439364954, -0.060266882124597964676, -0.055282182914780521321, -0.022558473555028987073, 0.021480223643822120294, 0.17406641211218953913, 0.24468601364496844752, 0.37082254557955385277, -1.6631050548348436102, 0.39011314636723815052, 0.33973764821164265104, 0.17639137573616972543, 0.13883241439807733619, -0.0085646395225352812963, -0.078868208991318522472, -0.067549958072159091427, -0.028117786394014930573, 0.0020844691674812617505, 0.019395761182167722825, -0.041258670866446522929, 0.24729557539490709672, 0.45716949812602825087, -1.3605839884241386351, 0.42105599803868754005, 0.29066091237500196742, 0.098390435765040418215, 0.0081857810297722190251, -0.073714333633356499131, -0.055756096690561196549, -0.011835031404269589356, 0.014196692083373655541, 0.013037500925405877406, -0.0026267278444138202585, -0.12144959937392045546, 0.034600403730842832495, 0.58454280295139515733, -1.0106047001622178172, 0.58550412864743694996, 0.084990453082567016252, -0.14008370359732752464, -0.031808842307811294181, -0.00023906624976830718327, 0.0016984712932585454941, -1.6869332032057966428e-05, 0.0051309534814273950742, 0.0057469472813662710736, 0.0013912050544977606469, -0.32355286520135684292, 0.42746457902958434616, 0.65395641199737619687, -1.5917287543583451548, 0.62710502230885312169, 0.48139756697850094591, 0.012998693236833613884, -0.14956052416556619988, -0.14323659702722096876, -0.062522443544033634577, -0.009845933282060681524, 0.019429155592049933421, 0.031267160661009485434, 0.030957834310622843466, -0.19069971064159990659, 0.51320735311980081317, 0.57171895642706838103, -1.8878033622628984656, 0.55542924106094748726, 0.58123341364102876216, 0.13071351355787436122, -0.075030440824681174372, -0.13414425641475935524, -0.088582709436555995275, -0.035114499625237105551, 0.0073868210701870593016, 0.02863100777435496555, 0.030837628598933394808, 0.0390992134283379561, -0.23885197183960912848, 0.2497839792067075293, -0.24846020858238970774, 0.26258938256833230795, -0.083513266136396424599, 0.26089890833991624675, 0.058962802520076129653, -0.18695955916335615932, -0.16421008044902785095, -0.074308154338185272714, 0.0023454295089441190569, 0.05320521262099359161, 0.080235689166549509221, 0.075017980153363403306, -3.3564426991506683962, -10.646389029246124736, 28.573479263090021618, -10.597779589686359358, -3.9087777197448678734, -1.4979070271948478066, -0.17534003419825716641, 0.4632993357634642706, 0.56955808955681053174, 0.41375441909384846895, 0.19094085431554963517, -0.015390080131903127736, -0.17228729570742393196, 0.60276793569785880322, -0.49728148192009968831, -0.16951478602094555148, 0.016239533746228926314, -0.15595314555862102579, -0.35045469542571267008, 0.52572685601149049539, 0.35946369181280679594, -0.08882411949240268112, -0.18216861173719509082, -0.094791435633541087924, -0.003767605092767425945, 0.031612132549646530799, 0.019959370663057848794, -0.30221330475966023954, 1.2903367285713045476, 0.66055709125003325699, -3.4140908851219706754, 0.63607444088973741358, 1.3106793730398846876, 0.20267891661606526887, -0.14048663855582158022, -0.12644095181620287893, -0.086911615491514992993, -0.047326622022419849445, -0.0064526075667455922574, 0.015099834861772864353, 0.02059927113320584105, -0.24668265443941950199, -0.54875539297474162925, 0.70561766223934874187, 0.27099832868351664272, 0.70464340781976764205, -0.53799003106611409386, -0.52624989022579626852, -0.10778862072860548316, 0.090755723377721780887, 0.12071536743971637839, 0.071983042854585399528, 0.022474772047424266402, -0.009532179110324124835, -0.023962388381938148907, -0.82628272517916212525, 0.15253118377330393551, 1.0347055617944478367, -0.74678205477104830923, 0.98183361280097480872, 0.1996037980036282633, -0.43586699765818337493, -0.41530672166395049549, -0.19688229698667469858, -0.0038198446564631782785, 0.07827597058038435196, 0.091238091432201989806, 0.063500943499673681569, 0.013414787383732761777, -0.0064484150982079954315, -0.36783473690838497383, 0.52242409799043321161, -0.28176701498489220565, 0.57000906169612441943, -0.30930395407375876893, -0.30468159927158672273, 0.028980648605752598601, 0.11175819864783045898, 0.061593003579865260311, 0.00045069615374157545595, -0.019093540658662977449, -0.012768865418070672602, 0.0022427551668762870166, -0.054722298862976241629, -0.56678130226208689457, 0.45856094163106980721, 0.45114410538854582411, 0.40805606208380018618, -0.60719065915958314594, -0.27454766852695533519, -0.033314941358462024645, 0.052425703820622060469, 0.1152409595616328275, 0.081711836524769654311, 0.019203334548813844052, -0.022164908052575482689, -0.03934832792840749216, 0.065094673235160188662, 2.2875207406044499869, 0.30003494785705220327, -5.6151368630020810357, 0.30744999827464875031, 2.3650376088150548703, 0.95920348733464355639, 0.056982484785950138217, -0.23500290689958069912, -0.27619137969490858975, -0.20180312328376681652, -0.082983459813056714705, 0.019040257407936883594, 0.089440285752966669497, 0.18708728701126195459, -0.58754654469650113846, 0.30858149160891662488, 0.36616706069558846437, 0.2458516732373609015, -0.67757069387120194559, -0.19115171711059730786, 0.087055999415918239315, 0.1405851351018548856, 0.15660986309259800064, 0.082927500664979669764, -0.0065772464170872901867, -0.057234917189647710467, -0.06595370618929233264, 0.28030214443312129191, 2.2657282890150995946, 0.16329349314077665212, -5.686175824369525067, 0.16337683066948471833, 2.3065875621958795705, 1.0412555861158365733, 0.16560607333506049965, -0.16258631823868308075, -0.24708581539514040681, -0.2049062870404078085, -0.10568959247165632398, -0.0093115591014898135835, 0.069244014910127216211, -0.46196734673790457171, -0.52281543936680985318, 0.8809458430034841081, 0.35119179354012053373, 0.85548180045575661357, -0.55687556804366922236, -0.71669112255011535595, -0.2328213567247749971, 0.095286260480422962926, 0.18059510729780942073, 0.12024218241942256913, 0.041945794507238856608, -0.012572717794271789579, -0.042088834038237904478, -0.29795012485001315383, -0.26943418795289292289, 0.77484558205218689864, -0.35861454123175606767, 0.76379914764926915005, -0.23439347817756883696, -0.46176331137251386272, -0.11981501876156405717, 0.047788568536261677444, 0.080418616112304691645, 0.057296728172444126581, 0.025720739838745544648, -0.00065247563925726302647, -0.017468220075498780419, 0.28892633547690843576, 0.58283688026252100656, 0.67481145666486175294, -0.084857046756997095427, -3.0266209902227796391, -0.053401008820652129394, 0.73501754960695175889, 0.65681298255752840198, 0.43723686097993408328, 0.18418278630743575852, 0.0076809295374055883143, -0.10604155254284872156, -0.14365041532689334924, -0.12297682949409949327, 0.042602702918731090442, 0.0096417783547731381044, -0.19991067749074761806, 0.33536675777682195543, -0.37388743407386737516, 0.33888514846713085893, -0.20287049469583823269, -0.01623223548428548435, 0.077826522106681908508, 0.0051286956810299034332, -0.0099582880737489540279, -0.0011862025972916292668, 0.0017948718802544508938, -0.0064321786854782125875, 0.060945875521157516497, -0.12974293828808314277, -0.032453941839348098009, 0.51610736383045230369, -0.82794643497895814566, 0.51375981568532480548, -0.033143027517408052418, -0.12633990315941753813, 0.047778449552697943581, 0.026523394239511451087, -0.0068886477020463554241, -0.0063427011398424389121, 0.00093275447102315824262, -0.0027014294871601108455, 0.16863644686357451286, 0.018173434708830038531, -0.12579914973588499083, 0.32756024060782196194, -0.81626139263287722514, 0.33796804894791748763, -0.096162254249730849676, 0.046028707874223140362, 0.17770590825365717436, 0.086126847273307743924, -0.0099131936546082433104, -0.054942069428316059332, -0.045057112558320927542, -0.0071302825389673273049, -0.039069267186541530024, 0.13395614861083043889, 0.34594306103891225845, 0.39011314692950116445, -1.6800344786361629534, 0.36904248019478008702, 0.32942051873445649113, 0.1732771166306918964, 0.13952653110118673485, -0.0075850855464076535895, -0.078868208753652108145, -0.06673640835704919283, -0.021869913922775454512, 0.019554725664503495863, -0.13938065578904651476, 0.10245890511126010125, 0.31786108678894681168, 0.42105599852408409633, -1.4006395706188619066, 0.40347807883874442503, 0.29407411169445896171, 0.10622912645009516264, 0.012845502845826778371, -0.07277976679335763921, -0.055756096541922157972, -0.011034045776930191327, 0.014409508897330385035, 0.0083903812427931438039, -0.00077041585885308152409, -0.15371898379543058599, 0.080258673944884798002, 0.58550412922301064977, -1.0199439213965246775, 0.57805741024124779326, 0.085509101688228034144, -0.1387141157656848478, -0.031391357163209433756, -0.00045239598112443079811, 0.0016984716264823508138, 0.00035083596047009459572, 0.005628961801272786003, 0.0067682391517726349373, -0.28187363418250044322, -0.018474960445245217772, 0.4819636352633663301, 0.62710502153877933917, -1.6167313491487331767, 0.60182073922157353074, 0.46673486294800919927, 0.0080794722946101182881, -0.14896136178343979961, -0.1422698769883758152, -0.062522444287784584294, -0.00781819338029719757, 0.030012532240286219493, 0.058463823881513464487, -0.27608070112681670771, 0.089737455206574029409, 0.58901537627841771272, 0.5554292402893238334, -1.9186282389004354343, 0.52117910609345308171, 0.56277200555411055483, 0.12730468860605076431, -0.070879094589252525371, -0.13064567350239320032, -0.088582709464961065238, -0.03522543347612986514, 0.016011668938487728236, 0.057901109503502003195, -0.31883337576712900141, 0.25815149318510810517, -0.071217321919831433341, 0.26258938222905514603, -0.26142843762595369128, 0.22733333474299843857, -0.1205828932872188286, 0.23716354104462422647, 0.048561673361206869615, -0.1911067177532244088, -0.16421008063678788069, -0.064588522394041558705, 0.033591693983287318481, 0.12009379053231385159, 1.1790933280855724341, -0.99083308403064807024, -3.8521357903284760837, -10.59777959014047255, 28.765020618342898473, -10.452731551452135506, -3.8326800125977849021, -1.4777707245928879409, -0.18097870339582078358, 0.4570991614173889328, 0.56955808897084980735, 0.40307419216027473574, 0.13277670020516013638, -0.16605046481974561345, -0.12051904493120250972, 0.58507040239931951131, -0.27923808870280136807, -0.15595314608423876956, -0.034212223883990908191, -0.20484570310052913067, -0.37189271859867389924, 0.52780376392234196192, 0.37022288805438074988, -0.081930250844938198873, -0.18216861157789926429, -0.097415118213305576522, -0.00025158319711138590891, 0.05254085675669049188, -0.33123893976117635418, 0.096982683556532833791, 1.316110965057344151, 0.63607444127582957183, -3.4603712256079517395, 0.5978594221800597186, 1.3000519922671649464, 0.2060031553960491868, -0.1336534489957771632, -0.12218157959674499913, -0.086911615685845225454, -0.048264013795154117603, -0.00092529187218141125543, 0.034727432392411751449, 0.21822763396021693794, -0.49312665140528327834, -0.56765452155024798131, 0.70464340785127921318, 0.29397467136175747804, 0.72988109972236758782, -0.52262478443286752849, -0.52426526288709685009, -0.11092128379475967026, 0.088626179336181290513, 0.12071536809956408265, 0.070933004704084479108, 0.014655430111661746362, -0.030460526307522651362, -0.34969747001356232818, -0.42111638677028551125, 0.21421459009737636725, 0.98183361249574518936, -0.79492895164041033063, 0.9495906605930772848, 0.18993666140480511273, -0.43218510104347823697, -0.40764443814592332682, -0.19215808546436344284, -0.0038198445235713418544, 0.076179483378049223097, 0.093246643757738820191, 0.078164344337484828462, 0.22345363512204916012, -0.32278289153665201683, -0.34844429874555610249, 0.5700090622205417068, -0.24795565912620970428, 0.57942299473913572161, -0.31183035183510154909, -0.3111716356754713475, 0.024897363207062830903, 0.11077773723750876378, 0.061593003773504302611, -0.00058713278853621264641, -0.02013775750880442833, -0.0078981288812688954515, 0.18026972781033856474, -0.19143698042293605388, -0.62985642268929842302, 0.40805606241364655862, 0.47852428644564720672, 0.44262301436539519539, -0.58867186786013725452, -0.26760961170134034415, -0.030857761698635077785, 0.053343162150650380582, 0.1152409580310920223, 0.077475018240217574661, 0.0039788909120042348366, -0.054842226799301715956, -0.62459409980642699445, 0.73774026079213750329, 2.3671250571977147104, 0.30744999808829104326, -5.6657669873197384902, 0.23311598947788580993, 2.319956361814975665, 0.9450439150633228591, 0.058433516336964697047, -0.2322427991652682433, -0.27619137998682663504, -0.19624771682865482791, -0.052789710997110605994, 0.097603933427727690431, 0.34811904332219983926, -0.096597611069001651507, -0.7068985627913927372, 0.24585167294713561326, 0.40829002914330214669, 0.30622655751061889084, -0.63965884933236216714, -0.17833533754777763924, 0.086560614636296495661, 0.13780582147166187124, 0.1566098627009687716, 0.08023125535339897807, -0.027831285193371912978, -0.11742467104542100853, -0.48756561851176977163, 0.8277467948300427647, 2.3039403359502359514, 0.1633768304414754402, -5.7249811662910738974, 0.10914828563024370212, 2.2765326496938009804, 1.0314009503217977937, 0.16448193939247990336, -0.16284257783966094024, -0.24708581591399791932, -0.19799639719733330701, -0.07986304993178919287, 0.048135797003037696096, 0.23442780716485941972, -0.66967548655841857563, -0.59604009586337747262, 0.85548180052691780162, 0.37795992738350331352, 0.89182437144831383513, -0.53153838284828025529, -0.70888892989560914337, -0.23456463607845484676, 0.092720492231205250966, 0.18059510737048115048, 0.1183234071448671515, 0.027752720244374536052, -0.049750693539728677028, 0.11987571577019351388, -0.44617328138397471804, -0.25134108607564381588, 0.76379914731760278901, -0.35612955021395609201, 0.77731494264985789844, -0.2221468770267591708, -0.45834450097392559487, -0.12088288263061183758, 0.046580809627932706629, 0.080418616655952374073, 0.056602909742991702746, 0.020113567957285329729, -0.016338981481743012486, -0.085878052308745456145, 0.29478056546331377064, 0.60386893163044530475, 0.72081129959675760421, -0.053401008815298467436, -3.0163206804264048877, -0.058419882106897684515, 0.72665330340679046284, 0.65209362720470087638, 0.43723686146120810392, 0.18352935255986624341, -0.0047402071399036960209, -0.14543996219670818348, -0.22797942403348853135, -0.017822067613964393679, 0.084919626676763101947, -0.015602957242666290472, -0.20740892690810386334, 0.33888514796558949493, -0.3681595642859901707, 0.33995681641197494915, -0.20373758051808624203, -0.016795113028498027258, 0.077826521718064173294, 0.0053655619639566814322, -0.0098364123490787460508, -0.0025820069338508269338, -0.0041466322211745699003, -0.0012348071679677660901, 0.06057364761041363066, -0.12436768607232205619, -0.037140938459980239739, 0.51375981498213108001, -0.82501273742232172914, 0.5151729541645793109, -0.03330262989956812858, -0.12676340853264675412, 0.047778449102615923316, 0.026766785998446413625, -0.007236082160345304326, -0.0080710758738055474876, -0.00039049602282796916624, -0.036629031373306558672, 0.15164315308014802985, 0.05031869088343178531, -0.091815689569017255867, 0.33796804923305368007, -0.81945911175173624663, 0.32637043018513400838, -0.10861348934716753833, 0.038506459356303313069, 0.17770590754235340847, 0.092069574988581537323, -0.0033141130815473179849, -0.054740344348697553278, -0.056648013910198836973, -0.16360476690716813097, 0.11638428792522753641, 0.18551741733716492488, 0.33973764836560671476, 0.36904247995926686698, -1.6944126783588226992, 0.36089695403827271036, 0.32607497044873168246, 0.17263954330992500874, 0.13952653082118937622, -0.0085646394970268491242, -0.080218918944328018727, -0.062606044033864216569, 0.00079480990152640284586, -0.10849464464953387233, 0.0067094128034187721504, 0.10388423360708336218, 0.29066091245192199333, 0.40347807859551709786, -1.4022359709451392984, 0.4062100656790656461, 0.29601565473795321548, 0.10683203575490418658, 0.01284550261582268621, -0.073714334016884208722, -0.058055149859806087775, -0.011381351518637689707, 0.026285027458220890789, 0.011099489349786766101, -0.025454557877354479917, -0.13680599188957504042, 0.084990453229587759054, 0.57805740975389319214, -1.0227128312629574225, 0.57847196959184621345, 0.086373554655592821305, -0.138569671473678796, -0.031391357061047418797, -0.00023906662377577844389, 0.0018883819506361975392, 0.0013375516777427674808, 0.011747010838163171251, -0.11643728755265453367, -0.15625079262358629317, 0.017870835186922444382, 0.48139756717348958226, 0.60182073917615896974, -1.6405100341268814557, 0.58988396408538246174, 0.4656685751017299868, 0.010112969998680221284, -0.1489613614007923581, -0.14323659688802176104, -0.057182255381045918319, 0.013843985494765981281, 0.073215188969184608059, -0.17599786580184251239, -0.089321747933721065693, 0.14195786899014928295, 0.5812334135865901974, 0.52117910594359084797, -1.9463481217641176446, 0.50935662449014418307, 0.56350644013671569166, 0.13096654996071227517, -0.070879094624319877171, -0.13414425640137070395, -0.087734201549861751124, -0.01632477833851221985, 0.066102949572893185404, -0.31442688787409900364, 0.055122472940250133444, 0.28879622191441789436, -0.083513266520566453721, 0.22733333514727793734, -0.29954079708470615628, 0.19985810338624387938, -0.13409568717103725888, 0.23292387562546634783, 0.048561673957543183278, -0.18695955853926737866, -0.14788111359463074201, -0.020608847873602897938, 0.11887196829500883133, 1.2137562840824711596, 0.13907127253818007961, -1.4467636530749907742, -3.9087777201025897256, -10.452731551317551606, 28.928501738794366815, -10.357445949797959628, -3.8018081579989937424, -1.4769987033556728484, -0.18097870282656591812, 0.46329933588141880563, 0.56202827532992227777, 0.3420824454304394191, -0.02266560637626419128, -0.37070924841442137732, 0.37352750314353044825, 0.57875470543986518024, -0.35045469518798810782, -0.20484570328269086081, -0.049726798888323875403, -0.20445057025410340312, -0.36739874099640146277, 0.53165873157221699152, 0.3702228882207356242, -0.088824119387918704094, -0.19267409195626472584, -0.090910813983631502677, 0.064848002949185717947, -0.18039494084690047004, -0.18525094268417108267, 0.19844388416509414763, 1.3106793732404857789, 0.59785942193240093623, -3.4882521220126183614, 0.58604683763529108464, 1.3008981660836471672, 0.21000397573864693079, -0.13365344898152564657, -0.12644095127453475191, -0.088788553193543215403, -0.037461251813263221622, 0.034290370050346394293, 0.25832765546326180717, -0.070272057964449949341, -0.53486554013108456029, -0.53799003124246358443, 0.72988109924067401391, 0.3118740130867069249, 0.739646738336909082, -0.51927697245350135091, -0.52445954933011429411, -0.11092128385105480548, 0.090755723444663607968, 0.12237959941234329009, 0.062906197538580885253, -0.017833329747210118627, 0.039512157899058916732, -0.35203871854407575714, -0.41674425118779007127, 0.19960379772214864591, 0.94959066063995689611, -0.81390539862674748939, 0.95065532214702952452, 0.20162615466065683667, -0.42337583515057763961, -0.40764443831496488535, -0.19688229706866347546, -0.00090072820314110457751, 0.10217081215557845264, 0.15169777248730514918, 0.12537421495499376234, 0.037124426062982579522, -0.30386076051741178761, -0.30930395443002450806, 0.57942299456604751118, -0.24864850951007760882, 0.57795264634065124376, -0.31264028283626510429, -0.31187353874214818195, 0.024897363067688402238, 0.11175819905924597175, 0.062353692068171359675, -0.0027048307464782473525, -0.028300764342095233173, 0.21704701452950722618, -0.017329314932257977588, -0.29202090873867159182, -0.60719065952697859334, 0.44262301524020786214, 0.50130755971017260109, 0.45599961102391461942, -0.58174972441444250393, -0.26546751140197033525, -0.030857761533751859595, 0.052425704585207344444, 0.11114378681884273803, 0.061940456289566347969, -0.043719604734894050579, -0.59790533423983061034, -0.085679259416962749718, 0.95622775415412952782, 2.3650376085120625724, 0.23311599023279716558, -5.7370601826146527813, 0.1891295993145447607, 2.3046461713593329179, 0.94423910498999408247, 0.058433516301658716685, -0.23500290626757380785, -0.27167614511949567113, -0.16270875291301545307, 0.036279517112500540843, 0.27909182623685085956, 0.077225308611829773797, -0.22956734619706281597, -0.67757069377905998486, 0.30622655816236654669, 0.45278944868152892766, 0.32646406104607450516, -0.63648892418469849019, -0.18106848384143037212, 0.086560614575658001435, 0.14058513520852233825, 0.15113467780880784908, 0.044486065632107107959, -0.1267844356833550401, -0.55670290606280736778, 0.0011939984910011580688, 1.0221417690851288818, 2.3065875622998155414, 0.10914828630496709039, -5.7792239496439590241, 0.070212518270968748979, 2.2578190734496432768, 1.0265169770119737525, 0.16448193944400379962, -0.16258631808635304328, -0.24361908766780082658, -0.18050291984061606065, -0.031101601660393440874, 0.38115822106401470171, -0.18506236215156957048, -0.73525747608478164707, -0.55687556840664798052, 0.8918243712374314125, 0.40463907103715796021, 0.9086144563159646026, -0.52439235462991895531, -0.70828634913055799061, -0.23456463619023507072, 0.095286260734006866802, 0.18166801315628211544, 0.10446852328681127675, -0.023103790958079295503, 0.18320188123442091022, -0.085247093075861632427, -0.46804013576365544047, -0.23439347776678090551, 0.7773149421838529971, -0.34484375269138445308, 0.78493024828624047284, -0.21851358446864527219, -0.45780300856033068202, -0.12088288273837931308, 0.04778856803745578502, 0.081597913025243148932, 0.052330526143420534602, 0.0012755044355162489366, -0.26801550297549359136, 0.055832598746712390481, 0.39660170372514291781, 0.65277068457038400773, 0.73501754958307297105, -0.058419882466699157852, -3.0269525744720149341, -0.065808427942960795454, 0.72665330401456251774, 0.65681298199608462163, 0.43617331836899142505, 0.15863015348087283751, -0.079170189916189750989, -0.2983474209270159383, -0.0089977012541605696366, -0.0015346423939496468998, 0.074245144118605999051, -0.016310289337557473677, -0.20287049411978810842, 0.3399568162114566805, -0.36864108265998474234, 0.33955868560985108529, -0.2037375807723426635, -0.016232234901752285589, 0.078551049768665281992, 0.0036798500356165150804, -0.015999117095496499796, -0.00079019613986776294189, -0.014258273894943879548, 0.02620875359336659674, 0.045993091772246967597, -0.12737598863958204953, -0.033143026793385932383, 0.51517295399034612746, -0.82525188325230602882, 0.51481559789165631802, -0.033302630226872340091, -0.12633990258643859605, 0.048140398437746437232, 0.025079906106181631964, -0.012163878604277932743, -0.0031449122614359151587, -0.12410765492954758948, 0.066199881179529726505, 0.18962054265517824803, 0.058246496270235315063, -0.096162254008050562781, 0.32637043089888401193, -0.83138963581109326384, 0.31896705200250929924, -0.10861349018538135613, 0.04602870781445905779, 0.18883526061604335999, 0.097219214277570173421, -0.020372477400375427203, -0.11067375451337481407, -0.13294125615499263748, -0.023260923706326191301, 0.1384793722837380503, 0.17639137491538606217, 0.32942051914548708957, 0.36089695383078390645, -1.6994023318030997416, 0.35876626587187293582, 0.32607497034725269058, 0.17327711684583813057, 0.13883241449408006973, -0.011807488627831510403, -0.083861758063175848021, -0.053283444148498372572, -0.020025424166443921148, -0.078206483986022992672, -0.0019805199250727562738, 0.098390435068041764977, 0.29407411179753090158, 0.40621006582217994429, -1.4022784880018313025, 0.40534510036485421614, 0.29601565427485826421, 0.10622912692934652346, 0.0081857813931386930878, -0.082351423006814278294, -0.05872654100517836484, 0.027321848366716458556, 0.0081120530156125209703, 0.0093121866597816842226, -0.03028336023185768458, -0.14008370428765659632, 0.085509102335038542075, 0.57847196937050171517, -1.0225704610498325131, 0.57843251929995032068, 0.086373554599834895251, -0.13871411578487050087, -0.031808842071318114975, 0.00016658770893923516634, 0.0051149264477165007001, 0.010854951192411640276, 0.0084679305746080246048, -0.11655098203388139511, -0.14030887551289827897, 0.012998693517146317489, 0.46673486291005794602, 0.58988396397082643041, -1.6445791425344589154, 0.590202983107625756, 0.46566857529544214422, 0.0080794717224099699138, -0.14956052385972520846, -0.13382244111679053589, -0.029429821424196422885, 0.061997430531440710089, -0.047489415875595381022, -0.11679773685746046297, -0.066645674233466012382, 0.13071351379834689199, 0.56277200559916429334, 0.50935662425962668909, -1.9489134001939476537, 0.51082085376492525342, 0.5635064402220711921, 0.12730468870272043658, -0.075030440871732995145, -0.12967122491609434087, -0.063691502575415448462, 0.034695294493810632797, -0.10852594804125731409, -0.15616858882209960768, 0.084853251011772323764, 0.2608989091492996959, -0.12058289390373366778, 0.1998581036237516706, -0.31400023146419009379, 0.19447232392363167364, -0.13409568650826475533, 0.23716354048385102082, 0.058962802162287664132, -0.16537094116034878333, -0.10730957596135731746, 0.048999279487534463939, 0.77649396118517410859, 0.5782096708438854149, -0.15306826506412729572, -1.4979070264444456217, -3.8326800127130238316, -10.357445949979155131, 28.988888456089721757, -10.336864046802407557, -3.8018081575306728048, -1.477770725390921136, -0.17534003385905938543, 0.46714604510188917352, 0.54765831171626633278, 0.3019845203409838863, -0.19530045321023142835, -0.10028172928814771403, 0.36354339121418766112, 0.52572685608630231879, -0.37189271847344573807, -0.20445057040121628056, -0.044709341562269244719, -0.20176416190304485854, -0.3673987409543104099, 0.52780376389308836238, 0.35946369149359813644, -0.1037165647433976462, -0.18709980387084093612, -0.0056524847802604492875, -0.073371250236722293514, -0.12631166355920422917, -0.14632992366748201474, 0.20267891586808703397, 1.3000519924712059527, 0.58604683760076270449, -3.4934031247196157821, 0.58617310190054150443, 1.3008981661134875196, 0.20600315553393683254, -0.1404866387076647849, -0.12922186473444102761, -0.07775094847539427978, -0.000260698606720895492, 0.14364958291966400949, 0.11294544044775492453, -0.10386443633584176161, -0.52624989001549760115, -0.52262478412133095418, 0.73964673778020118, 0.31729075310552473654, 0.74178411019487600075, -0.51927697235766390182, -0.52426526254567173968, -0.10778862112142614804, 0.096539560268996085779, 0.12119494896261416572, 0.035593162322358651295, 0.20204509992465630175, -0.11377722913573529162, -0.39463178629427120292, -0.43586699766565245584, 0.18993666122347102609, 0.95065532211405456842, -0.80448554608407318867, 0.95914415714431533466, 0.2016261543768849418, -0.43218510077100874556, -0.41530672133720830663, -0.18342485546454601053, 0.054668166188248559778, 0.2083844489101280284, 0.0020458165572833962934, 0.12164223084272676734, 0.041346972218880312377, -0.30468159871271677241, -0.31183035155644645009, 0.57795264614285124516, -0.24794658521455201217, 0.57873194629731739003, -0.31264028272358823646, -0.3111716355610100182, 0.028980647687836355586, 0.11919352695421210619, 0.061359429744821725183, -0.040835325913510044094, 0.14197047253417674484, 0.048836183025772639255, -0.038882116527010013307, -0.27454766864778901159, -0.58867186844659769029, 0.45599961136639333148, 0.51022616664577524759, 0.45997137498876949913, -0.58174972428473203934, -0.2676096118837106852, -0.033314941313338120032, 0.050270726321841818773, 0.10235606349425638295, 0.023912995813893375696, -0.36924112984200246901, -0.27398330927805747415, 0.053363498052532376548, 0.95920348799606447443, 2.3199563611473705826, 0.18912959961668662778, -5.7627488398011443849, 0.18032327888491894718, 2.3046461713571821939, 0.94504391497345618944, 0.056982484128817456615, -0.23393042339687453168, -0.25813394903136654968, -0.1262977670856648138, 0.12004673556483377861, 0.095617280676576488219, 0.060681539147511787125, -0.1911517170041010516, -0.63965884995599742169, 0.32646406154981555403, 0.45856853717473478049, 0.32573737199808672216, -0.63648892434601878065, -0.1783353376324298134, 0.087055999586409416047, 0.12788261168137587953, 0.10909718390015861467, -0.047403259618135018538, -0.39518996217107565938, -0.2359083460475790428, 0.14530092458494658914, 1.0412555863088648334, 2.2765326489405901533, 0.070212518903306708751, -5.8078022642789992602, 0.057442340489391785496, 2.2578190733880192376, 1.0314009503714693938, 0.16560607337605981471, -0.17006774923317802273, -0.25732768050610893562, -0.18754656131289113019, 0.24009180661595097561, 0.12626074679819987434, -0.23187580594583789839, -0.71669112245066479705, -0.53153838291101118685, 0.90861445612760494139, 0.41386724548899511689, 0.91204340383564330708, -0.52439235450783006076, -0.70888892959076421629, -0.2328213566344986829, 0.10061027714127213084, 0.18138452597353649232, 0.070459411815191691497, 0.11808907817023588105, 0.069463762420371139017, -0.11755122980641291575, -0.46176331148644228497, -0.22214687674768615611, 0.784930248289770649, -0.34049227236977558375, 0.78682242827949178121, -0.21851358481028726466, -0.45834450073586591179, -0.11981501844295543158, 0.051653341683148404351, 0.084065029498814641817, 0.045694957378754050281, -0.29834742084693016695, -0.079170190233900511467, 0.15863015345018058322, 0.43617331888011845464, 0.65681298188608849831, 0.72665330428212682179, -0.06580842871678635464, -3.0269525738230962375, -0.058419882931539074411, 0.73501754953515252566, 0.65277068447080754954, 0.39660170381772891224, 0.05583259976475599129, -0.26801550375174987284, -0.00079019621363278762932, -0.015999116945568710862, 0.0036798499448468351136, 0.078551049753291413147, -0.016232234800896160054, -0.20373758087604909583, 0.33955868563862134923, -0.36864108266767398048, 0.33995681622312723391, -0.20287049415317215395, -0.016310289228705112524, 0.074245144015840217055, -0.0015346423850494846086, -0.0089977012355796446436, -0.0031449123136482815336, -0.012163878395506380614, 0.025079905781581569857, 0.048140398640453885282, -0.12633990249415527618, -0.033302630567413227247, 0.51481559827865930323, -0.82525188349807721, 0.51517295413619090816, -0.033143026934902612701, -0.12737598865722460362, 0.045993092001471440333, 0.026208753402195673898, -0.014258273847286931715, -0.11067375495959658194, -0.020372476577530863884, 0.097219214962772831923, 0.18883525880532436858, 0.046028707722791412094, -0.10861348941011135516, 0.3189670532683711035, -0.83138963690591949351, 0.32637042973549640346, -0.096162253303241185254, 0.05824649744440690835, 0.18962054254571816814, 0.066199879695630903242, -0.12410765415730261318, -0.053283444176654953806, -0.083861758117817042368, -0.011807488351608420862, 0.13883241431100123986, 0.17327711661803657517, 0.32607497076558028537, 0.35876626558830615465, -1.6994023316911130994, 0.36089695371235713806, 0.32942051909752895211, 0.17639137533763468713, 0.13847937199483334458, -0.023260923788839572846, -0.13294125606916393223, 0.027321848276708408931, -0.058726540734423605949, -0.08235142319113088194, 0.0081857813305865004955, 0.10622912695797408211, 0.29601565432532433997, 0.40534510030458714702, -1.4022784879087180077, 0.40621006591746178227, 0.2940741113846612742, 0.098390435536153286744, -0.0019805201304347978819, -0.078206484017876387593, -0.020025424124824837968, 0.010854951120664030687, 0.0051149265804001113264, 0.00016658772869945553605, -0.031808842258068445452, -0.13871411565234401109, 0.086373554565414387141, 0.57843251935920636519, -1.0225704612115658065, 0.57847196960809266209, 0.085509102009031848213, -0.14008370379997764354, -0.030283360675795529399, 0.0093121868120835046834, 0.008112053018183247588, 0.061997430578150249103, -0.029429821418761822199, -0.13382244121211095411, -0.14956052381221543457, 0.0080794716333733584968, 0.46566857530689065303, 0.59020298334573506605, -1.6445791428855423, 0.58988396432720091589, 0.4667348626904710418, 0.012998693729408788977, -0.14030887573370465216, -0.11655098206405573036, 0.0084679306613711802926, 0.034695294582578993514, -0.063691502604814695387, -0.12967122525709554037, -0.075030440296781339415, 0.12730468822974416732, 0.56350644051316400862, 0.51082085341268834711, -1.948913399896581522, 0.50935662431986805654, 0.56277200552260742139, 0.13071351393772306926, -0.066645674613567490452, -0.11679773668287549149, -0.047489415849457455865, 0.048999279685719973609, -0.10730957643906080534, -0.16537094098528437591, 0.058962802157595195496, 0.2371635411669847926, -0.134095687396403207, 0.1944723241052339624, -0.3140002313259886435, 0.19985810380408919329, -0.12058289362347052176, 0.2608989082472719101, 0.084853251248232064441, -0.15616858855469945011, -0.10852594810886447008, 0.30198452050247787115, 0.54765831148255605942, 0.46714604507765611352, -0.17534003383325269532, -1.4777707250613012491, -3.8018081578357456607, -10.336864047011145473, 28.98888845646774115, -10.357445950103032928, -3.8326800123469886294, -1.4979070270983738666, -0.15306826473378021269, 0.5782096707407237135, 0.77649396124841085776, -0.0056524846932198974381, -0.18709980409595347584, -0.10371656470560015884, 0.35946369178417353352, 0.52780376377104365471, -0.36739874109099801558, -0.20176416186379758644, -0.044709341703682015101, -0.20445057001072602931, -0.37189271870591850133, 0.52572685612672809263, 0.36354339095688958938, -0.100281728895562805, -0.19530045337343127088, -0.00026069849158403212395, -0.077750948782730394671, -0.12922186457816289828, -0.14048663857139837008, 0.2060031555946648385, 1.3008981658842568851, 0.58617310181341797382, -3.4934031247125294506, 0.58604683819990843752, 1.3000519917798603053, 0.20267891621407826963, -0.14632992404296252253, -0.12631166310046162971, -0.073371250425575240683, 0.035593162483964754794, 0.1211949486603241255, 0.096539560100674479126, -0.10778862070870286194, -0.524265262259705489, -0.51927697268819883458, 0.74178410966758612233, 0.31729075353211699806, 0.7396467382486819897, -0.52262478439501813821, -0.52624989063492877417, -0.10386443596465212702, 0.11294544076779872155, 0.14364958269407704372, 0.20838444872873609914, 0.054668166611240950892, -0.18342485568732577739, -0.41530672121261735796, -0.4321851011789424879, 0.20162615467560671134, 0.9591441569729797223, -0.80448554589976239981, 0.95065532209357028748, 0.18993666146274038042, -0.43586699785561500153, -0.3946317864088568772, -0.11377722936064486869, 0.2020451001906600208, -0.040835325853568568699, 0.061359429615634494826, 0.11919352694902604628, 0.028980647776390283665, -0.31117163543562770256, -0.31264028298911528125, 0.5787319465729711121, -0.24794658549311524021, 0.5779526462892363714, -0.31183035154627725127, -0.30468159874408329291, 0.041346972147008123621, 0.12164223098460431871, 0.0020458164903045879431, 0.023912995690987259856, 0.10235606376780970184, 0.050270726326528847316, -0.03331494142683333326, -0.26760961232823543554, -0.58174972369089861246, 0.4599713747755962423, 0.51022616679758214797, 0.45599961113376430433, -0.58867186868603860184, -0.27454766792389817542, -0.03888211670527941255, 0.048836182594597349338, 0.14197047276252366954, -0.12629776737909045869, -0.25813394839477488141, -0.23393042369999317298, 0.056982484109001085848, 0.94504391459603798165, 2.3046461718722852652, 0.18032327893251176576, -5.762748840089731317, 0.18912959938587059483, 2.319956361654026189, 0.95920348792569087859, 0.053363497955514094251, -0.27398330942613607908, -0.36924112972010930678, -0.047403259766994894353, 0.10909718427713034627, 0.12788261152108604257, 0.087055999422477325922, -0.17833533775337440108, -0.63648892398998602182, 0.32573737205791930549, 0.45856853698665434038, 0.32646406116834292277, -0.63965884949941820281, -0.19115171723425294914, 0.060681539631002812119, 0.095617280280913014012, 0.12004673562060230729, -0.18754656124365931547, -0.25732768075856493617, -0.17006774892025711132, 0.16560607323134246549, 1.0314009503183392269, 2.2578190734029623954, 0.057442340740253783338, -5.8078022643638291811, 0.070212518561835413067, 2.2765326490462518549, 1.0412555863590076122, 0.14530092489394919042, -0.23590834629431572944, -0.39518996216133805977, 0.07045941213108240786, 0.18138452531786136701, 0.10061027714386859555, -0.23282135612547416592, -0.70888892925206326012, -0.52439235509290538229, 0.91204340342373724226, 0.4138672460552469401, 0.90861445638509363842, -0.53153838301410638589, -0.71669112319660721599, -0.23187580558843234879, 0.1262607472232785133, 0.24009180634637664054, 0.045694957510111329502, 0.084065029157863541132, 0.051653341995216359162, -0.11981501867250896054, -0.45834450039381435804, -0.2185135851933189266, 0.78682242853733030952, -0.34049227257286762693, 0.78493024855839044207, -0.22214687676504041303, -0.4617633119215152071, -0.1175512294911268546, 0.06946376247525840042, 0.11808907809768080799, -0.22797942522530786902, -0.14543995966489514649, -0.0047402081807305547489, 0.18352935242629933588, 0.43723686135593542446, 0.65209362677906712591, 0.72665330418771689747, -0.058419882727226243868, -3.0163206806060527398, -0.053401008305753991379, 0.72081129947972377714, 0.6038689313562577432, 0.29478056600865487624, -0.085878052563709950107, -0.0041466322474992631936, -0.0025820068575864163485, -0.0098364124939395326114, 0.0053655622090094320931, 0.077826521475094237457, -0.016795112867395045608, -0.20373758064761074427, 0.33995681652162429431, -0.36815956444245678814, 0.3388851482374240498, -0.20740892725413787034, -0.015602956887604612496, 0.08491962641507207199, -0.017822067523687729063, -0.00039049603373004024157, -0.0080710758108827121426, -0.007236082295297415444, 0.026766786196418184546, 0.047778448806398594872, -0.12676340807158203994, -0.033302630467938909575, 0.51517295465533252674, -0.82501273774344052736, 0.51375981518446756091, -0.037140938631936827596, -0.12436768591724939315, 0.060573647523352591726, -0.0012348071468658988747, -0.056648013653184010885, -0.054740344511735830602, -0.0033141136798091053173, 0.092069575450067747679, 0.17770590799835184193, 0.03850645884519465767, -0.10861348961626086929, 0.32637043070697047131, -0.8194591114243640062, 0.33796804861562845801, -0.091815689796806487788, 0.050318691333885293138, 0.1516431531011060152, -0.036629031481049456609, 0.00079480991820355339182, -0.062606043993618867849, -0.080218919123025894669, -0.0085646393444300562836, 0.13952653097898523593, 0.17263954299361317091, 0.32607497061882145806, 0.36089695388419851252, -1.6944126781953252614, 0.36904247993461797295, 0.33973764830989544583, 0.1855174174008299981, 0.11638428787500060269, -0.16360476688966690828, 0.02628502751037858215, -0.011381351556932274224, -0.058055150052179073372, -0.073714333649573582363, 0.012845502293276831601, 0.10683203603102220391, 0.29601565449206301839, 0.40621006583526980682, -1.4022359710864080728, 0.40347807880745156783, 0.29066091219182976779, 0.1038842338857522557, 0.006709412598405788998, -0.10849464458598680194, 0.011747010841806777501, 0.0013375517128875558191, 0.0018883817734344702892, -0.00023906633675664255498, -0.031391357300508640482, -0.13856967140774811842, 0.086373554732790513988, 0.578471969576247802, -1.0227128313686169037, 0.57805740988099885147, 0.084990453075649355363, -0.13680599169175677865, -0.025454558013950807915, 0.011099489383174458018, 0.073215189088252433103, 0.013843985418531837284, -0.057182255582603931654, -0.14323659689379908411, -0.14896136127250869663, 0.010112969952023542763, 0.46566857527264504668, 0.58988396421180355755, -1.6405100344468446227, 0.6018207393602607036, 0.4813975668515472206, 0.017870835682515899545, -0.15625079292373972306, -0.1164372875133688201, 0.066102949774448388487, -0.016324778698668544752, -0.087734201526042013275, -0.13414425623755882322, -0.070879094637416775737, 0.13096654969301207982, 0.56350644074410405349, 0.50935662396234304694, -1.9463481216294327147, 0.52117910609003803568, 0.58123341349129609146, 0.14195786914742253293, -0.089321748289151500777, -0.17599786562141075952, 0.11887196837513411274, -0.020608848065490736889, -0.14788111371553358531, -0.18695955827085852796, 0.048561673610133748724, 0.23292387613178111572, -0.13409568735646670845, 0.19985810346522248082, -0.2995407971030019656, 0.22733333506404718172, -0.083513266430954802289, 0.28879622177908048641, 0.055122472874344796701, -0.31442688776137539541, -0.022665606319789144418, 0.34208244520258590526, 0.56202827557026191485, 0.46329933552398538943, -0.18097870241297453475, -1.4769987036438969596, -3.8018081574516582322, -10.357445950572246929, 28.928501739467925802, -10.452731551776611951, -3.908777719754722213, -1.4467636533251324593, 0.13907127262126645073, 1.2137562840642288631, 0.064848002905818449837, -0.090910814043939996987, -0.19267409165400511961, -0.088824119453987368522, 0.37022288787843038449, 0.53165873188613310685, -0.36739874117982573853, -0.20445057006273847922, -0.049726799250736475333, -0.20484570293740392288, -0.35045469515667915239, 0.57875470528981831642, 0.37352750312399551902, -0.37070924836952190429, 0.034290369987533417251, -0.037461251782775248365, -0.088788552983492655124, -0.12644095152123743797, -0.13365344891310335096, 0.21000397591361985583, 1.3008981656545530736, 0.58604683801211843175, -3.4882521222698126273, 0.59785942214803211581, 1.3106793731719730278, 0.19844388418118760753, -0.18525094274022177454, -0.18039494081788923219, -0.017833329749055045615, 0.062906197429161092582, 0.12237959961574913781, 0.090755723467565399543, -0.11092128422134625576, -0.52445954901891944644, -0.5192769723083793254, 0.73964673811211611998, 0.31187401286865795713, 0.7298810995369413579, -0.53799003112537457927, -0.53486554037398148509, -0.070272057923415495728, 0.25832765549242386882, 0.15169777253619087798, 0.10217081195342000988, -0.00090072821796666202587, -0.196882296958076769, -0.40764443814716300185, -0.42337583523097510607, 0.20162615449286733216, 0.95065532218695536493, -0.81390539834863295798, 0.94959066048525275772, 0.19960379785024151555, -0.41674425140635767972, -0.35203871860494750923, 0.039512158011508036193, -0.028300764290487230579, -0.0027048309846342162566, 0.062353692501942586945, 0.11175819868871569862, 0.024897363198328095746, -0.31187353875247203483, -0.31264028284793288215, 0.5779526464025855903, -0.24864850952701247899, 0.57942299438640254561, -0.3093039540898117612, -0.30386076084941260333, 0.037124426250589689924, 0.12537421490870104268, -0.043719604530019121058, 0.061940455751648054727, 0.11114378740594715866, 0.052425704108544968207, -0.03085776092952974281, -0.2654675119651499493, -0.58174972425038840029, 0.4559996108321264785, 0.50130756029861978096, 0.44262301449871022108, -0.60719065886109502284, -0.29202090925404955435, -0.017329314608281298771, 0.21704701441976784193, 0.036279517283851286935, -0.16270875333158926757, -0.27167614477704937759, -0.23500290639216536714, 0.05843351642864158535, 0.94423910490478057955, 2.3046461713087404988, 0.18912959919255545316, -5.7370601822436544381, 0.23311598990150783628, 2.3650376088228912685, 0.95622775381236690961, -0.085679259267036761427, -0.59790533423465708207, -0.12678443559373445138, 0.044486065525877915849, 0.15113467793277363116, 0.14058513501551828107, 0.086560614882698838635, -0.18106848403981312678, -0.63648892435175130622, 0.32646406117833470795, 0.4527894488350007185, 0.30622655785240326942, -0.67757069374316569732, -0.22956734598551953308, 0.077225308661865971205, 0.27909182611724103751, -0.031101601613106016897, -0.18050291986282990253, -0.2436190876080421841, -0.16258631814371460322, 0.16448193933111765475, 1.0265169772290758665, 2.2578190731380294309, 0.070212518651995070584, -5.7792239498737751902, 0.10914828632896300675, 2.3065875621567442089, 1.022141769170214376, 0.0011939987026729492747, -0.55670290620289830663, -0.023103790974085221255, 0.1044685232059875124, 0.1816680133957798704, 0.095286260513569309705, -0.23456463617561801316, -0.70828634901373399568, -0.52439235455503130368, 0.90861445611647373433, 0.40463907102744922639, 0.89182437127655977971, -0.55687556818850802998, -0.73525747636716209943, -0.18506236205319365196, 0.38115822107041358313, 0.0012755044636709020247, 0.052330526027564147618, 0.08159791314970769871, 0.04778856810202147809, -0.12088288312150473336, -0.45780300804587237273, -0.21851358492521516519, 0.78493024886077134017, -0.34484375332296957062, 0.77731494264612954748, -0.23439347812168934149, -0.46804013549498124647, -0.085247093185524980274, 0.18320188124671979435, -0.12297682848430968594, -0.14365041760807981119, -0.10604155119922188677, 0.0076809289117563059324, 0.18418278700292323169, 0.43723686129779232301, 0.65681298214832684046, 0.73501754966277910164, -0.053401009287114400981, -3.0266209899356191215, -0.084857046084257106711, 0.67481145602831837937, 0.58283687977982512329, 0.28892633599080175255, -0.0064321786880928598015, 0.0017948718389123601187, -0.0011862023443058025349, -0.009958288541968316615, 0.0051286959605861444367, 0.077826522204609518973, -0.016232235684674949788, -0.20287049455263128639, 0.33888514836946515008, -0.37388743399554619229, 0.33536675766919565866, -0.19991067738403558995, 0.0096417783042759153816, 0.0426027029279732114, -0.0027014294943289904848, 0.00093275450447283378072, -0.0063427011998507696819, -0.0068886476794708323759, 0.026523394237253437866, 0.047778449618594037718, -0.12633990315209833732, -0.033143027672721508958, 0.51375981569854023423, -0.82794643470978857014, 0.51610736348926311567, -0.032453941572155593775, -0.12974293845225384758, 0.060945875572268749798, -0.0071302817293475949378, -0.045057114165865462529, -0.054942069632556919667, -0.0099131917872189168062, 0.086126847414002810632, 0.1777059070747959435, 0.046028706792607204656, -0.096162252942702863834, 0.33796804993319828903, -0.81626139338063385953, 0.32756023945887929871, -0.12579914960100538335, 0.018173436363009644323, 0.16863644593248242054, 0.019554725669453848491, -0.021869913910028366405, -0.066736408486935627771, -0.078868208821707752598, -0.0075850850323030162559, 0.13952653086296096241, 0.17327711627107672543, 0.32942051898987578795, 0.36904248036334341698, -1.6800344787913592537, 0.39011314683264780578, 0.34594306096981219945, 0.13395614894419230478, -0.039069267361262371985, 0.0083903813034751952016, 0.014409508601856498358, -0.01103404533610684822, -0.055756096881078631711, -0.072779766504361159685, 0.012845502649265646417, 0.10622912647475211967, 0.29407411179577869209, 0.40347807856511674895, -1.4006395703202993985, 0.42105599833514406605, 0.31786108690691389222, 0.10245890509427753312, -0.13938065581268474502, 0.0067682391885955138847, 0.005628961627721136389, 0.00035083632741742085299, 0.0016984712925822921699, -0.00045239594010004942348, -0.031391356980065419913, -0.13871411607449266046, 0.085509102087156174665, 0.57805740986606379828, -1.0199439210774332576, 0.58550412892968972667, 0.080258674167090301999, -0.15371898389944621544, -0.00077041583737339912503, 0.05846382378333141866, 0.030012532317392842662, -0.00781819322460583499, -0.062522444109877145513, -0.14226987754994760427, -0.14896136140110269319, 0.0080794719288858896089, 0.46673486310879108707, 0.60182073946314817992, -1.6167313493353405729, 0.62710502176276206043, 0.48196363466624220218, -0.018474959832401081222, -0.28187363440239054935, 0.057901109237393673379, 0.01601166924160658464, -0.035225433103790955436, -0.088582709988620444985, -0.13064567334427093503, -0.070879094760971317335, 0.12730468858830326639, 0.56277200599026855343, 0.52117910578812187694, -1.918628239165479421, 0.5554292407414159749, 0.58901537600396070005, 0.089737455420191097888, -0.27608070123918126981, 0.12009379023570614098, 0.033591694715452702358, -0.064588522315949248309, -0.16421008142258664009, -0.19110671775093579505, 0.048561673386956716314, 0.2371635418261988093, -0.1205828940245430303, 0.22733333486545892477, -0.26142843785546743263, 0.26258938295713907252, -0.071217322172147823345, 0.25815149285472971386, -0.31883337559579522091, -0.16605046511203225279, 0.13277670099167498963, 0.40307419182301529315, 0.56955808907460114821, 0.45709916040938969317, -0.18097870250826941785, -1.4777707248373932458, -3.8326800121199955385, -10.452731551995412929, 28.765020618361479166, -10.597779589843504766, -3.8521357902607391566, -0.99083308438935491225, 1.179093328236799687, 0.05254085650490404602, -0.0002515826645153422525, -0.097415118521286259146, -0.18216861163568859316, -0.081930250906305110425, 0.37022288833623978599, 0.52780376397092687579, -0.37189271883487645676, -0.20484570320025335954, -0.034212223854793472055, -0.15595314587227143699, -0.27923808853284204368, 0.58507040192180592708, -0.12051904471574670374, 0.034727432187851764445, -0.0009252914007370268562, -0.048264014147041664482, -0.086911615670975328474, -0.12218157965200911186, -0.13365344875620518983, 0.20600315562290621996, 1.3000519916891497552, 0.59785942214157938857, -3.4603712251868006256, 0.63607444122217515758, 1.316110965079098305, 0.096982683272861358592, -0.33123893959311712099, -0.030460526586772628671, 0.014655430750536733997, 0.070933004598945123553, 0.1207153676769431444, 0.088626179103621000044, -0.11092128349313666591, -0.52426526219740499091, -0.52262478512481891357, 0.72988109956825031333, 0.29397467140223992876, 0.70464340852233064982, -0.56765452192559739952, -0.49312665172036335148, 0.21822763419629159776, 0.078164344313477809867, 0.093246643727965455861, 0.076179483636809736113, -0.0038198447685568459598, -0.19215808518422289319, -0.40764443851381076955, -0.43218510099491241894, 0.18993666157586278076, 0.94959066033361727577, -0.79492895151129150388, 0.98183361250230438699, 0.21421459013990012954, -0.42111638655786765018, -0.34969747023429076371, -0.0078981289400009736801, -0.02013775727235324331, -0.00058713311247284363792, 0.061593004037604613676, 0.11077773701046057797, 0.024897363222587946818, -0.31117163542374426388, -0.31183035204974968169, 0.5794229947434854644, -0.24795565884774578569, 0.57000906168813725294, -0.34844429813491900605, -0.32278289200910903478, 0.22345363529135028879, -0.054842226760355744508, 0.0039788909470381828015, 0.077475017666986778497, 0.11524095875075190953, 0.053343162181529457655, -0.030857761938733574425, -0.26760961179151143696, -0.58867186805778781711, 0.44262301486350530411, 0.47852428661211321437, 0.40805606145115336236, -0.62985642216438664232, -0.19143698013816998205, 0.18026972757113066126, 0.097603933553537094991, -0.052789711328387389777, -0.19624771674683083589, -0.27619137994059184127, -0.23224279854297735781, 0.058433515545984970885, 0.94504391532935749787, 2.3199563616185008286, 0.23311598983410108943, -5.6657669873340781308, 0.30744999743025269723, 2.3671250578385865104, 0.73774026076834042787, -0.62459409993343351086, -0.11742467087778685197, -0.027831285527058241547, 0.080231255149363689405, 0.15660986336427185073, 0.13780582119204498603, 0.086560614628696463946, -0.17833533773859744365, -0.63965884910738202151, 0.30622655770942497355, 0.40829002898820365708, 0.24585167262887830386, -0.70689856257288430363, -0.096597611071541883421, 0.34811904333668053368, 0.048135796973308886404, -0.079863049732643687229, -0.1979963977119439944, -0.24708581562483034055, -0.16284257751616493248, 0.16448193904688324096, 1.0314009505061765282, 2.2765326492361164235, 0.10914828637739160122, -5.7249811667573693441, 0.16337683058516327961, 2.3039403355485923441, 0.82774679529657124721, -0.48756561865278741363, -0.049750693934845970079, 0.027752721134588674567, 0.11832340701289773, 0.18059510678314313892, 0.092720491983863340124, -0.23456463581351066905, -0.70888892912625267595, -0.53153838326168900696, 0.89182437095959676832, 0.37795992745472645202, 0.85548180114886762659, -0.59604009585483064271, -0.66967548732544590617, 0.23442780757179865203, -0.016338981645135173398, 0.020113568410295740163, 0.056602909345783783412, 0.080418617080187521085, 0.046580808860197539523, -0.1208828822003173431, -0.45834450079693078717, -0.22214687709376870761, 0.77731494247534649311, -0.35612955030851611848, 0.76379914779739754849, -0.25134108640026997517, -0.44617328145822393592, 0.1198757158711609011, -0.029641169607290494131, -0.098790123372701335591, -0.13181505761313813685, -0.10242296105242988669, 0.0076809288707985062805, 0.18352935333806075224, 0.43617331869289971413, 0.65277068410605110227, 0.72081129898567375403, -0.084857046149688988845, -3.0744219413868099089, -0.083149550898377083685, 0.78666202520398387232, 0.83670076163169782024, -0.0035152536667442330423, -0.0021890507596613525965, 0.0017463042765367090686, -0.0014452507085019588445, -0.0099582889259721106556, 0.0053655619387481135601, 0.078551050465516233934, -0.016310289068830759185, -0.20740892683706224631, 0.33536675687043243643, -0.36641082791577350708, 0.34811265170801369839, -0.25566485051991200139, 0.093763283364258015462, -0.0023031665946069744891, 0.00020029158015304627388, 0.0008702649087417480174, -0.0065565840626437155853, -0.0068886485322433941431, 0.026766786029954452858, 0.048140398968333303342, -0.12737598809300706981, -0.037140938446339699652, 0.5161073625961899447, -0.81437869313661659376, 0.51316225964691097872, -0.11097763160467492838, 0.00021682024852043212614, 0.021480223971407631095, -0.022558474522370391968, -0.055282182628065161534, -0.06026688114091752202, -0.0099131932225786367763, 0.092069574800471537523, 0.18883525976876236463, 0.058246497080503445432, -0.091815689306084635746, 0.32756024005860351345, -0.8504285376623690329, 0.27985091192740907395, -0.16264996975482454156, 0.29138164333234028991, 0.019395761178915137124, 0.0020844694335544295891, -0.028117787199538911425, -0.067549957245875341671, -0.078868209338806588571, -0.0085646394247520102327, 0.13883241468478707548, 0.17639137509709054696, 0.3397376484712750222, 0.39011314655387935302, -1.6631050550395278798, 0.37082254565184319395, 0.24468601369769349407, 0.17406641206451409221, -0.0026267277359078421844, 0.013037500637850635635, 0.014196692177615700264, -0.011835031193780513534, -0.055756096638419141986, -0.073714334098057623934, 0.0081857816149537958317, 0.098390435240733847833, 0.29066091262348192625, 0.42105599804838672595, -1.3605839884955375219, 0.45716949828642067244, 0.24729557518666300808, -0.041258670779236110426, 0.0013912050968022933537, 0.0057469472695504111512, 0.0051309532755579607272, -1.6869101900361714932e-05, 0.0016984713755354326203, -0.00023906659545999946886, -0.031808841848193727952, -0.14008370404124836917, 0.084990453304545007795, 0.5855041287226278035, -1.0106047004073386297, 0.58454280320714135932, 0.034600403569422795436, -0.12144959932781633494, 0.030957834160159583559, 0.031267160898932847002, 0.01942915528085365634, -0.0098459328551067648583, -0.062522443695477492032, -0.14323659723605930427, -0.1495605240418264037, 0.012998693306266850822, 0.48139756705106301338, 0.62710502203160789492, -1.5917287540943676483, 0.65395641214329280899, 0.42746457851592578603, -0.32355286493942420067, 0.030837628514425383108, 0.028631008110713077053, 0.0073868204865138106796, -0.035114499151589498305, -0.088582709716267124533, -0.13414425629688847552, -0.075030440710405266191, 0.13071351353384572103, 0.58123341356175006744, 0.55542924094961221293, -1.8878033621291694377, 0.57171895659980109805, 0.51320735280578066995, -0.19069971052043194204, 0.080235689216279520997, 0.053205212795981837726, 0.002345429408340704569, -0.0743081547338364512, -0.16421008069161979748, -0.18695955804990416205, 0.058962801578311241357, 0.2608989089252475857, -0.083513266803008079364, 0.26258938340590642913, -0.2484602089930191271, 0.24978397943695701544, -0.2388519722938152734, 0.039099213644325961359, -0.17228729559086247702, -0.015390080539057549025, 0.19094085480009947986, 0.41375441871704188301, 0.56955808986770661839, 0.46329933539597489833, -0.17534003381952167899, -1.4979070272379593209, -3.9087777198298163661, -10.59777958972921752, 28.573479263226488456, -10.646389029421484906, -3.3564426991933160593, 0.075017980261050207225, 0.019959370638680189181, 0.031612132689134611607, -0.0037676052609045898589, -0.094791435721643071965, -0.18216861155333513578, -0.088824119413937641099, 0.35946369168214553635, 0.52572685596255219753, -0.35045469513801186245, -0.15595314623771103224, 0.01623953436138511619, -0.16951478609288256338, -0.49728148201753052948, 0.60276793570183095916, 0.02059927108576229543, 0.015099835004506796665, -0.0064526076835734486742, -0.047326621980858726846, -0.086911615802109737361, -0.12644095127096011133, -0.14048663854562373765, 0.20267891612149036007, 1.3106793732280408449, 0.63607444075722718857, -3.4140908849087088228, 0.66055709140370366672, 1.290336728345145012, -0.30221330472703294978, -0.023962388475007655869, -0.0095321788444923151451, 0.022474772014640972961, 0.071983042544846082, 0.12071536742141861231, 0.090755723638674637432, -0.10778862072508610392, -0.52624989011294232011, -0.53799003147151114312, 0.70464340785557322278, 0.27099832917400384913, 0.70561766204152365223, -0.54875539330001643634, -0.24668265422370869167, 0.013414787675904985692, 0.063500943144162769194, 0.091238091011334093716, 0.078275971079158318533, -0.0038198445123457380745, -0.19688229717877284153, -0.41530672137999702365, -0.43586699798061145916, 0.19960379794027122147, 0.98183361276058200851, -0.74678205443555434329, 1.0347055615514142435, 0.15253118407262450384, -0.82628272540859015738, 0.0022427551017109468787, -0.012768865329272646378, -0.019093540453956552028, 0.00045069574007812127081, 0.061593003672733293019, 0.11175819899117375722, 0.028980647947260823605, -0.30468159853119258784, -0.30930395443184710569, 0.57000906161669451233, -0.2817670147719701923, 0.52242409779704956918, -0.36783473679873734952, -0.0064484151223954747079, -0.039348327775293512332, -0.022164908465538118998, 0.019203334564837082343, 0.081711837208141013811, 0.11524095885372120929, 0.052425704412552787126, -0.033314941810741793837, -0.27454766836006094799, -0.60719065940596950171, 0.40805606252294046588, 0.4511441049612825438, 0.45856094186769275067, -0.56678130228834144866, -0.05472229887732701914, 0.089440285882226105985, 0.019040257311456615374, -0.082983460022811927903, -0.20180312315934612055, -0.27619137977532359773, -0.23500290602565243603, 0.056982483380684123553, 0.95920348818039102845, 2.3650376085198683285, 0.30744999848831955358, -5.6151368630231814905, 0.30003494753886661428, 2.28752074105304537, 0.065094673026086069534, -0.065953706187926328108, -0.057234917614786051665, -0.0065772454485900078147, 0.082927499846837782194, 0.15660986368686663495, 0.14058513471550171392, 0.087055999412959716999, -0.19115171722299950652, -0.6775706934452933039, 0.24585167295391419096, 0.36616706080837907455, 0.30858149125160283521, -0.58754654434713293742, 0.18708728695214837945, 0.069244014758696903478, -0.0093115588849100872082, -0.1056895923184048014, -0.20490628718552339471, -0.24708581562740927762, -0.16258631793539113364, 0.16560607291426387988, 1.0412555864590362642, 2.3065875620728237827, 0.16337683098581745611, -5.6861758247505491681, 0.16329349334367562752, 2.2657282888105778618, 0.28030214456310464932, -0.0420888341812198849, -0.012572717490982560778, 0.041945794749134381052, 0.12024218174888165467, 0.18059510722689267603, 0.095286260859477139862, -0.23282135656577745975, -0.71669112256835920682, -0.55687556850683783694, 0.85548180074747948165, 0.3511917936605772339, 0.88094584302678957677, -0.52281543972638366835, -0.46196734652690935041, -0.01746822007435408497, -0.00065247589211575671847, 0.025720740651478023187, 0.057296727215955620127, 0.080418616790800076521, 0.047788567858730784943, -0.11981501823537876739, -0.4617633115532107646, -0.23439347780793032272, 0.76379914704436802086, -0.35861454094462469344, 0.77484558226929822755, -0.26943418833092130926, -0.29795012469212550643, 0.032827291131138504332, -0.03874379083007137986, -0.10072323254042950258, -0.13181505739596413584, -0.10604155254698810484, -0.0047402071361300202046, 0.15863015295736715427, 0.3966017042835870976, 0.60386893209781389302, 0.67481145557400146195, -0.083149551272583610562, -3.0158415011041133091, 0.18724560144132318396, 1.4078261372007772056, 0.0001517656913993520448, -0.0030785294465045047983, -0.0021768155404014526665, 0.0017463041668226283662, -0.0011862017198746065486, -0.0098364124762217088754, 0.0036798495045834659495, 0.074245143844986691284, -0.015602957169014277164, -0.1999106766819314096, 0.34811265161080984232, -0.40152578866883686937, 0.30266162938630458701, -0.098962278187941024599, -0.00024568789418501923603, -0.001375847034225108562, 6.4407380028818793025e-05, 0.00087026468887814586659, -0.0063427001307936083097, -0.0072360825092567856304, 0.025079905800622040446, 0.04599309136340149623, -0.12436768607657966434, -0.032453940876991570208, 0.5131622595058340508, -0.85331260503349237734, 0.59403823996598581481, -0.15469659800924831505, 0.027887645243157503649, -0.0021360584249255076816, -0.032426932845078006162, -0.055282183887238149633, -0.054942068818598005508, -0.0033141128855569967904, 0.097219214784224933235, 0.18962054198946839967, 0.050318690477291649898, -0.12579914959384705386, 0.27985091297453495995, -0.88117307062121619232, 0.3925314593484336978, 0.11682376156736033745, 0.012709028734434984001, 0.0099550889897761039166, -0.0018644103499514822278, -0.028117787366979549168, -0.066736408175850581159, -0.080218918799118535401, -0.011807488827834070055, 0.13847937232899271143, 0.1855174170496447783, 0.3459430615053035174, 0.37082254573681272536, -1.7448358155460454988, 0.44066466135953147321, 0.43514684663889163785, -0.0039188627134995540971, 0.0036072017450131483793, 0.013070496699253053921, 0.014196692456402494262, -0.01103404575208484506, -0.05805514987669051219, -0.082351423044153299058, -0.0019805202267105040392, 0.10388423348297280702, 0.31786108720322830967, 0.45716949829185726806, -1.3950153323136635919, 0.41493300906950369988, 0.23301342052460916054, -0.0016190939434438896861, 0.0023985736469413964853, 0.005538617566264243848, 0.0051309533117818665465, 0.00035083630672324726471, 0.00188838180061825546, 0.00016658760512072343617, -0.030283360455085263363, -0.13680599191606182541, 0.08025867432703258364, 0.58454280329173113895, -1.005156277933350939, 0.66137985265046328465, -0.16603432377569091183, 0.003060443924573623322, 0.021852642074448674903, 0.028469683619889104426, 0.019429155473773648821, -0.0078181933929647035031, -0.057182255361048053843, -0.13382244098958051559, -0.1403088755726255854, 0.017870835285349551125, 0.48196363470633057924, 0.65395641204346799391, -1.5797983268379467425, 0.60869358929322492369, 0.099541227463446943502, 0.0035389785947395066903, 0.023752434542023626723, 0.02711339880499314195, 0.0073868207211656364275, -0.035225433279481264959, -0.087734201429473482325, -0.1296712252109650243, -0.066645674138125568509, 0.14195786909700852663, 0.58901537576102191984, 0.57171895678249662343, -1.9103079025415299341, 0.54048197732104608093, 0.34002284094420021088, 0.025999303398644760948, 0.043818605406773875055, 0.038329163324012982539, 0.0023454297256026990226, -0.06458852252649582848, -0.14788111397039049866, -0.16537094077144565141, 0.08485325105944330204, 0.28879622241014835105, -0.071217322850298470094, 0.24978397939244256776, -0.39984171464256629269, -0.06627417239184418829, 0.20686487829731475818, -0.11018127798361670999, -0.080580892838576334558, 0.01724993662924118043, 0.19094085500864044214, 0.40307419146109024766, 0.56202827568717195206, 0.46714604494633804599, -0.15306826446206800441, -1.4467636532568390884, -3.8521357905804731736, -10.646389029561268202, 28.278568919106035651, -10.744483509749858996, -2.9610352551896244044, -0.0014169012869767313653, 0.024703008150305579005, 0.031826472411066467072, -0.0037676052998726269894, -0.097415118270518019639, -0.19267409212241731753, -0.10371656465145637216, 0.36354339115363221158, 0.57875470540569373679, -0.27923808832001784097, -0.16951478629621755112, -0.24816905288670371155, -0.52572462416248078032, 0.63010484504422192842, 0.0057514944605823008217, 0.01601862305693549024, 0.014329275892667503109, -0.0064526077434653109771, -0.048264013809107220165, -0.088788553381021345645, -0.12922186470499241739, -0.14632992373499137395, 0.19844388418673084007, 1.3161109653016804799, 0.6605570912659092242, -3.3730904209090701862, 0.83276075064281218641, 0.76362028346589616579, -0.012907586653067413054, -0.014706146922081875933, -0.0054251247989641794567, 0.022474772215925149788, 0.070933005037013940886, 0.1223795993164191187, 0.096539559806076050141, -0.10386443595519084804, -0.53486554011311659984, -0.56765452134707994336, 0.7056176615958734688, 0.35652342360563021861, 0.81649836283017085403, -0.96172836439799314068, -0.028369347392361590754, 0.026846493115533956453, 0.070116025418720959173, 0.091238091173461641215, 0.076179483383846391398, -0.00090072804277174789434, -0.18342485546097195281, -0.39463178630058493024, -0.41674425150233201931, 0.21421459013505761426, 1.0347055615666598261, -0.6860423999338187695, 0.94649047991295776061, -0.73299131934816685874, 0.0036417296627038113314, -0.0027607171637695293934, -0.012959525017321846599, -0.019093540616212573147, -0.00058713260893133653595, 0.062353692110183753239, 0.11919352663258195313, 0.041346972365948871098, -0.3038607605916366361, -0.34844429830980006013, 0.52242409795058519784, -0.26014045796559998047, 0.69420533140795781524, -0.50130802272719909851, -0.017325991868891624237, -0.021876014392423037691, -0.013946161989057542785, 0.019203334386084902619, 0.077475017298702431034, 0.1111437873091640216, 0.050270726581867708482, -0.038882116448584636004, -0.29202090895142862603, -0.62985642300189403908, 0.45856094240555511643, 0.54921944650910003993, 0.39589640624777883149, -0.66177421931759405904, 0.050817804050691445239, 0.044773108994295915397, 0.0028104593749846140671, -0.082983459943075710274, -0.19624771722847850208, -0.27167614490949182349, -0.23393042317283385678, 0.053363497904076906408, 0.95622775417265359899, 2.3671250569943391717, 0.30003494798069596516, -5.6745211771836583381, 0.4583961920979097715, 2.2620335753925759903, -0.011619159940278794493, -0.045937087671756013374, -0.050228145195615858665, -0.0065772453066183511616, 0.080231254597083301316, 0.15113467812558417647, 0.12788261188185912065, 0.060681539216266344727, -0.22956734632976685817, -0.70689856270062756405, 0.30858149114943406177, 0.49538064339973608696, 0.30609769787819596232, -0.5041071146956924709, 0.05636655273269271238, 0.025827310182773927094, -0.026840693379896030102, -0.1056895924752261906, -0.19799639775647537299, -0.24361908788658659386, -0.17006774845046868094, 0.14530092451277076826, 1.0221417693118051151, 2.3039403354440075589, 0.163293493359594033, -5.7266765212591206335, 0.36862458553577720499, 2.4128768682925305633, -0.023357661250317414414, -0.023193196750215975405, -0.0051011357782629711266, 0.041945795289808782136, 0.11832340677923645911, 0.18166801295552489903, 0.10061027731712700317, -0.23187580583088443498, -0.73525747603367697103, -0.59604009569147564651, 0.88094584254726249206, 0.51541601291596572665, 1.0299104790738629056, -1.2676678254809885527, -0.012015386698847984118, -0.0086489049649866554803, 0.002806153535087777505, 0.025720740319737332713, 0.056602910122146056493, 0.081597912677038961227, 0.051653341832007870771, -0.11755122993622516003, -0.46804013559019264079, -0.25134108580337527483, 0.77484558192642283281, -0.30150141746129616172, 0.90769685601506688855, -0.74686885696778149146, 0.064526370814191338909, 0.017836650159179120306, -0.038743791712609124134, -0.098790123137792298924, -0.14365041498914460361, -0.14543996262063765834, -0.079170190324747938271, 0.05583259940683439615, 0.29478056695636267737, 0.58283687964444008678, 0.78666202519373207291, 0.18724560094039829905, -2.7496011868834271397, 1.1694382577378621413, 0.0020831867232067387448, -0.00067535693383215732958, -0.0030785288815127413546, -0.0021890513086292992181, 0.0017948719336983907056, -0.0025820072293382964959, -0.015999116657746878523, -0.0015346420511777917293, 0.084919626181180873914, 0.0096417785584063733784, -0.2556648511726953843, 0.30266162992485023953, -0.11722086459066720365, -0.0047904415080750242101, 0.0012458905313942592562, -0.00061810634504193859122, -0.0013758463666053927579, 0.00020029078123411179352, 0.00093275471330655970043, -0.008071076167292020459, -0.012163878287298309638, 0.026208754049109762169, 0.060573647009632768101, -0.12974293792596716668, -0.11097763242505680192, 0.5940382405602793181, -0.65029089752709345884, 0.22937784131286007017, 0.017757349992252639825, 0.011186997659833067265, -0.0021360588806967825692, -0.022558474863532830545, -0.045057114218451078946, -0.054740343267239525982, -0.020372476062615767212, 0.066199879861684240989, 0.15164315185244833795, 0.018173436325133511537, -0.16264996990570898538, 0.3925314603245619316, -0.50946687114169209032, 0.14502533647973639419, 0.0048929333113227309673, 0.0092418873909466264793, 0.0099550892164256381267, 0.002084469062385897048, -0.021869913725294402168, -0.062606044299557433996, -0.083861757516385249112, -0.02326092398095888289, 0.11638428766740258058, 0.13395614881407197361, 0.24468601347650598132, 0.44066466174636947928, -1.5262222502147653191, 0.74449369954277710981, -0.0010678541056262204692, -0.0017486083179990119085, 0.0036072019045765941689, 0.013037500182383199299, 0.014409508960771342384, -0.011381351481708341239, -0.058726540921765837955, -0.078206483771180029807, 0.0067094126897734385553, 0.10245890500343378393, 0.24729557516456965316, 0.41493300913729519452, -1.3272552266682966504, 0.675770737706630098, -0.0027337987079130853137, -0.00094813232039859174013, 0.0023985738586854354545, 0.0057469471075995747292, 0.0056289616911845861622, 0.00133755166320920953, 0.0051149266156671098493, 0.0093121868215520313405, -0.025454558195174791652, -0.15371898374136158671, 0.034600403404570330423, 0.66137985277573319021, -0.92803213094897496482, 0.3907326602909746982, -0.017305423875735418543, 0.0041576717743327358881, 0.021852642315333032502, 0.031267160890294951048, 0.030012531705499615603, 0.01384398582330586755, -0.02942982135944444827, -0.11655098226157302455, -0.15625079277919429188, -0.018474960201565578805, 0.42746457905250823117, 0.60869358897212377801, -1.7103974731754048033, 0.93538594123189244822, -0.014276239447404377031, 0.0071599751652009916217, 0.023752434799752777245, 0.028631008138378859912, 0.016011668765525136071, -0.016324778540391104986, -0.063691502485988066296, -0.11679773703032203525, -0.089321747945691989945, 0.089737455449712316691, 0.5132073528231537729, 0.54048197722262059095, -1.9799603095469753278, 1.0752202784264921487, -0.02481283865158830787, 0.014185781870120137138, 0.043818605445971686718, 0.053205213242313781352, 0.033591694005589423089, -0.020608847812955577439, -0.10730957667951790402, -0.15616858868439664754, 0.055122473201199451154, 0.25815149299723627596, -0.23885197200434404863, -0.066274172883575011017, -0.028928516718571758992, 0.22173218624438625901, -0.03149424203993307636, -0.076994939712201906912, -0.080580892884337784743, -0.015390079987330163025, 0.13277670041012978874, 0.34208244529005038537, 0.54765831141420318051, 0.57820967051909422096, 0.13907127305071104173, -0.99083308429756600244, -3.3564426989785185462, -10.744483510057069253, 25.767588691224574404, -12.160754273199799513, -0.0077282963866709573561, 0.0071753632537016245191, 0.024703008140652973901, 0.031612132884154001133, -0.00025158286270897736725, -0.090910813992885031176, -0.18709980434581180431, -0.10028172877397478757, 0.37352750321694350166, 0.58507040206012883576, -0.4972814821577836164, -0.52572462424024324346, 0.50758073420982274815, -0.13185938705813740612, -0.003529594433414946341, 0.0084468059654916813339, 0.016018623131550301097, 0.015099835249757238148, -0.00092529165655490762388, -0.037461251921714594504, -0.077750948668980648937, -0.12631166319302214451, -0.18525094276832768103, 0.096982683500746166838, 1.2903367281324826799, 0.83276075065056109903, -3.8075718312170327984, 1.9774077467313826428, -0.0029813762056747510933, -0.010515793953789156004, -0.014706146454582616143, -0.0095321791186179054278, 0.014655431001301340993, 0.062906196674440847949, 0.12119494874736251588, 0.11294544099722594876, -0.07027205773039962855, -0.49312665169040426072, -0.54875539371667447774, 0.81649836300073574957, 0.49357667721864256771, -0.4592694952320758861, -0.044155659010610574799, -0.010551917629678471758, 0.026846493252538003449, 0.063500943105077672923, 0.093246643401904416582, 0.10217081228808824744, 0.054668166436361451122, -0.11377722941338747065, -0.35203871847802015171, -0.4211163868145503808, 0.15253118438151425096, 0.94649047987455858788, -1.0121821650636140255, 0.57451571404548118505, 0.0018829118062792709123, 0.0018782455549859948356, -0.0027607171646544283183, -0.012768865191952444293, -0.020137757398930110353, -0.0027048310168724457037, 0.06135942983246171073, 0.12164223083467354258, 0.037124426008723218628, -0.32278289153009398493, -0.36783473730551968783, 0.69420533166141284998, -0.018929936890357278861, -0.16981925077605317109, 0.0039976890769542203566, -0.010742031519747334556, -0.021876014950966901629, -0.022164908109114561663, 0.0039788907593820699304, 0.061940456563639501031, 0.10235606378662481797, 0.048836182176650838604, -0.017329314713104337908, -0.19143698035938638835, -0.56678130191736109111, 0.39589640617188265326, 0.76973654478315800542, -0.5587307325399130109, 0.0077580785201645224153, 0.035970252574174091387, 0.044773108595398083964, 0.01904025744576182122, -0.052789711398796040909, -0.16270875268486323373, -0.25813394864788019589, -0.2739833096706226212, -0.085679259325561391014, 0.73774026050918384279, 2.2875207415042955184, 0.45839619195182290667, -6.1545535915487503686, 3.3751799765276313536, 0.028028687946504884931, -0.012924100321322979507, -0.045937088051821520218, -0.057234917652388056508, -0.027831285358302093602, 0.044486066086487918714, 0.10909718399896291274, 0.095617280395250026093, 0.077225308620742394439, -0.096597611372101294114, -0.58754654373600900819, 0.30609769749239307268, 0.80893983295658422605, -0.67106876061932119182, 0.027499426318947885889, 0.034766655867039628069, 0.025827309834494271978, -0.0093115587106556152452, -0.079863049842679972867, -0.18050291925821393413, -0.25732768101760283841, -0.23590834619163403296, 0.001193998380885019639, 0.8277467949268509928, 2.265728289301268461, 0.3686245853148811813, -6.1160483684242654689, 3.2833263451533434285, -0.0044323022429185393936, -0.017702922629199049909, -0.023193196469891378536, -0.012572717539123328193, 0.027752721271810967124, 0.10446852279248834694, 0.18138452516129874303, 0.12626074736472411852, -0.18506236165951089889, -0.66967548721432779146, -0.52281543981926059672, 1.0299104787764441493, 0.48425923832603090524, -0.49970943051772792609, -0.005126285873035876621, -0.0088049616721433566352, -0.00864890486367738201, -0.00065247567632005509797, 0.020113568059546178823, 0.052330526160339278785, 0.084065029106620225985, 0.069463762914232826673, -0.085247093388856831631, -0.44617328129061734066, -0.26943418860330869968, 0.90769685618189133169, -0.25309919832627858494, -0.042220898770429393654, 0.073403237920959499285, 0.064526370759244403019, 0.032827291673004695138, -0.02964116986468749021, -0.12297682968460114938, -0.22797942398689335919, -0.29834742037601358344, -0.2680155037089311798, -0.085878053304693807712, 0.28892633628047093008, 0.83670076175159535659, 1.4078261374625749003, 1.169438257505282408, -3.0526377732034619683, 0.0025363759711766659978, 0.0020831868332126502429, 0.00015176531672105085611, -0.0035152532306403027305, -0.0064321789526354073915, -0.0041466320277774668807, -0.00079019625665280909376, -0.008997701420337594902, -0.017822067348405950388, 0.042602702594958456739, 0.093763283842846698035, -0.098962278553759316924, -0.0047904413805652176772, 0.0010532714467780428854, 0.0017446023275039168263, 0.0012458906084633558035, -0.00024568826906066733784, -0.0023031660352117568416, -0.002701429898599023733, -0.0003904957444887866805, -0.0031449123890880332013, -0.014258274098813243647, -0.0012348068134655316649, 0.06094587510203854086, 0.00021682086175996286926, -0.15469659846388161828, 0.22937784145533895996, -0.11760043798863209497, -0.0017332256268625528817, 0.017757349717284820478, 0.027887645392407879597, 0.021480224633571173842, -0.0071302817542767890635, -0.056648014621484735209, -0.11067375546120548913, -0.12410765399391721076, -0.036629030411379236154, 0.16863644591371546033, 0.29138164297873514474, 0.11682376106081496236, 0.14502533687374100091, -0.48628934647555027704, -0.0023678497680666653856, 0.0048929333261822525666, 0.012709028574527971023, 0.019395761467975322634, 0.019554725503728113273, 0.00079481001601276876287, -0.053283444401867031959, -0.13294125602159209687, -0.16360476666869072648, -0.039069267426382538189, 0.1740664121170790446, 0.43514684648893997565, 0.74449369961733857792, -1.0654578598477009166, 0.0019121976541579509268, -0.0010678540967356253948, -0.00391886281614876094, -0.0026267275091414951649, 0.0083903811811608906834, 0.026285027450769962026, 0.027321848327822924329, -0.020025424253288785781, -0.10849464451653710595, -0.13938065581914604874, -0.041258670799119538664, 0.23301342047748199726, 0.67577073776438500996, -0.67395669321784357209, -0.0030365133722332236441, -0.0027337986340061105894, -0.0016190941258994477267, 0.0013912052435703333494, 0.0067682391116369207407, 0.01174701091915599134, 0.010854951067243831866, 0.0081120529636849458832, 0.011099489585963608074, -0.00077041597790566984129, -0.12144959929545681709, -0.16603432380604385421, 0.39073266031378445229, -0.14024832529243067003, -0.028820103006423972158, -0.01730542389100733075, 0.0030604438510470127408, 0.030957834077009888096, 0.058463824208762983636, 0.073215188836680253348, 0.061997430382833521034, 0.0084679308333128026909, -0.11643728750320501975, -0.28187363417384364572, -0.32355286524608528609, 0.099541227627711073578, 0.93538594117033213582, -0.48373677592447711371, -0.021872986628990891134, -0.014276239398753720611, 0.0035389784794668541559, 0.030837628356593371759, 0.057901109632280317596, 0.066102949685204470676, 0.03469529449294840584, -0.047489415758245133448, -0.17599786582534351909, -0.27608070122632982768, -0.19069971060683751918, 0.34002284107682323455, 1.0752202784020434834, -0.88815827620486997862, -0.065594813460760736934, -0.024812838656212036353, 0.025999303455561076359, 0.080235688855001124908, 0.1200937906022756102, 0.11887196834888358321, 0.048999279811084463232, -0.10852594800179218848, -0.31442688827181908273, -0.31883337553476953641, 0.039099213575472843707, 0.20686487857495067244, 0.22173218611197859085, 0.013838993022205689498, 0.040262051013666429189, -0.031494242038499001279, -0.11018127790121731713, -0.17228729597518821492, -0.16605046477923768311, -0.022665606320262252082, 0.30198452049098328809, 0.77649396133975145951, 1.2137562837296484997, 1.179093328318070899, 0.075017980187974203643, -2.9610352550818115347, -12.160754273217222021, 12.436896380403220874, -0.0045477875476925234619, -0.007728296341881249823, -0.0014169013141025459807, 0.019959370520579278185, 0.052540856564805074602, 0.064848002995006245874, -0.0056524845156005233682, -0.1953004534741217546, -0.37070924846075142822, -0.12051904484803246942, 0.6027679358174369284, 0.63010484512915609923, -0.13185938711050626537, -0.56892945570176112202, -0.0065329174960208234377, -0.0035295943344625488498, 0.0057514943754350365823, 0.020599270937037610085, 0.034727432251006190578, 0.034290370176776371891, -0.00026069846799117268703, -0.073371250441058230574, -0.18039494087459859162, -0.33123893969725121122, -0.30221330461024870084, 0.76362028350455246617, 1.977407746682767975, -1.9942532617113177817, 0.0047363658267201803539, -0.002981376058913532464, -0.01290758688473305979, -0.023962388429762757763, -0.030460526853707810963, -0.017833329187872341143, 0.035593162537734118755, 0.1436495824784438935, 0.25832765530681578436, 0.21822763405803285996, -0.24668265378964948642, -0.96172836453014887148, -0.459269495278241624, 1.1476589359417186387, -0.039842902805263077926, -0.044155659056104856752, -0.028369347430476660882, 0.013414787624700795499, 0.078164344561579585102, 0.15169777240133242024, 0.20838444877420336798, 0.20204510016960186558, 0.03951215800790961602, -0.34969747019155145162, -0.82628272558267168435, -0.73299131931980798793, 0.57451571410595869693, 0.86102170009501222125, -0.00016837017257448572355, 0.0018829118506911862469, 0.0036417296382379267181, 0.0022427550659638141942, -0.0078981290087148624257, -0.028300764131843774568, -0.040835325950488818303, 0.0020458165158469647504, 0.12537421500257811546, 0.22345363505975726626, -0.0064484148573349707803, -0.50130802285248610151, -0.1698192507718853661, 0.40929723835688847089, 0.022381060243626387657, 0.0039976889529632067197, -0.017325991600396231651, -0.039348327893276247025, -0.054842226533370938557, -0.043719605053291074603, 0.023912995637767400781, 0.14197047301239498762, 0.21704701441545820595, 0.18026972793428935882, -0.054722299375246016839, -0.6617742191625884951, -0.55873073252511851194, 0.8650739001190976829, -0.029200535251030775713, 0.0077580783894433350684, 0.05081780428150724227, 0.089440285777166061987, 0.097603933775237183834, 0.03627951686497796896, -0.12629776730127975637, -0.36924112954375248874, -0.59790533428300596253, -0.62459409968708501371, 0.06509467264658064245, 2.2620335755291720581, 3.3751799765482046745, -4.3905989900286508387, 0.050901221315693051528, 0.02802868783203023989, -0.011619159676483492918, -0.065953706212673629539, -0.11742467081407551022, -0.12678443604584760385, -0.047403259704354772097, 0.12004673566412188412, 0.27909182627265899379, 0.34811904353962941361, 0.18708728665872753627, -0.50410711457467194396, -0.67106876064494114242, 0.50982366208281904019, -0.0064940584843384413505, 0.027499426253989121827, 0.056366552911553471916, 0.069244014716507054641, 0.048135797097704678682, -0.031101602015893251019, -0.18754656118134174769, -0.39518996211861889822, -0.55670290601460092894, -0.48756561841832651849, 0.28030214423114102607, 2.4128768684738055583, 3.2833263450653369375, -4.7053925516839099075, 0.011108590533447515111, -0.0044323020994146396581, -0.023357661363370023355, -0.042088834336418995408, -0.049750694047077742566, -0.023103790560105190444, 0.070459412231306348118, 0.24009180619399766998, 0.38115822074781802176, 0.23442780749650299343, -0.46196734625419899611, -1.2676678253618969272, -0.49970943067697093376, 1.5078630756985529349, 0.0013361990785015136034, -0.005126285774444796009, -0.012015386804315358535, -0.017468220136473568138, -0.016338981601202146043, 0.0012755045333355790466, 0.045694957526170858209, 0.1180890778896867277, 0.18320188141928805958, 0.11987571563626372995, -0.29795012437676215988, -0.74686885713868134129, -0.042220898758449865174, 0.71093238947475889677, 0.6926662364004434469, 0.056855795920989458947, -0.69739281736655456267, -0.35237283972917715635, 0.042696606940537340513, 0.14455098343524497295, 0.10511949841975841424, 0.054629624053406482254, 0.029109589207459696941, 0.013375577654314010564, -0.0010195335161620096027, -0.011648519322434637557, -0.014773896366685437126, -0.010857930269971371659, -0.60146356218421470619, 1.0249814841051387848, -0.12917147159111974641, -0.36185240908313598407, -0.047312267097036751762, 0.063675373856045880405, 0.043737741281516334668, 0.011321865737904692611, -0.0016796961713922021336, -0.0033846852153250226933, -0.0022993487914272289774, -0.0013164794793928840701, -0.00049736400055672375185, 0.00021975777823454319105, 0.17605209852085373923, -0.2657377446471270166, -0.061927270805137719922, 0.16898525395374108804, 0.05186743589483226724, -0.036126329287914683841, -0.037419476679364481464, -0.010116564284461886103, 0.0076968950101167447203, 0.007798000850050123417, 0.0024683491076279207889, -0.00061197154795209689031, -0.0010676608515091539926, -0.00041803739202597945063, -3.3063607094222815519, 2.5594840133252501069, 1.6205429512731064357, 0.11895803646274938359, -0.41717335041593173184, -0.47871919817593805435, -0.31076603674714936565, -0.10802494090495114165, 0.027875737282073414364, 0.080384182294334388663, 0.074692189378189027815, 0.042657734315914887235, 0.0064297338144501583329, -0.024915956669851298949, -0.089962021304843237113, -0.07543546955593986858, 0.24631821763337588549, 0.1149368408040708639, -0.16652490474014056132, -0.10487744270295080995, 0.013844439382195563443, 0.045095419909269879311, 0.027125598570640046397, 0.0021111312992846235914, -0.010006753437289702888, -0.0091043457316828252379, -0.0022986775641046070483, 0.00524302478766329981, 0.12379833011516536145, -0.18719463142880551909, -0.047493370040926595477, 0.1463280635473703517, 0.011946371374526636167, -0.045933634214511132043, -0.01445139232714590799, 0.009730491872995973221, 0.0095432917819388250746, -0.00028242545500030755276, -0.004692816430644042823, -0.0027926704078805103179, 0.00044417367171277146896, 0.0022131206998817296722, -0.3906078607511032641, 0.65941889024599098512, -0.1088406843370473831, -0.20849048881835136049, -0.020807057279500322522, 0.040652880246693681721, 0.03059774689351515975, 0.0073995621373498694784, -0.0059109773960665176751, -0.0075924384204216011432, -0.0033318662815810767656, 0.00039241448937907562116, 0.0016019245461685366252, 0.00094876284482914698213, 0.18842938016001159474, 0.20646098573648571928, -0.56448374206854956547, -0.1959522165698452667, 0.22315601450953484464, 0.21491509181700477193, 0.05564828617874616884, -0.04460565587022780748, -0.056388353478843356026, -0.030844970056064558578, -0.0067762777354719002168, 0.0048813529398471274673, 0.0075709860485262698432, 0.0059160766368202529278, 0.76867369834929955008, -0.27791692500979869873, -0.76751409063671172106, -0.19901921886245213522, 0.22818126282700640362, 0.26685220927838804172, 0.12350233795672510562, -0.0052891120354027054001, -0.051657606104817378745, -0.044867447577536169467, -0.02289558034469085096, -0.0047591859055536424664, 0.0057915441264729727519, 0.010381881735043615306, -2.1940680120246778273, 1.9491704742698887642, 0.99024802404918377263, -0.14294698229719657312, -0.37170184730395533501, -0.27953376580408811902, -0.12893345254000349032, -0.0081783879000189053932, 0.046920811560793418415, 0.048155063051097880822, 0.028147276221017234615, 0.0080314711847906082587, -0.0042915678464978280343, -0.0086853953010601874646, -0.56892945565272212693, -0.13185938720323611695, 0.63010484497770458123, 0.6027679361098960964, -0.12051904515932160278, -0.37070924797353799196, -0.19530045375755542003, -0.005652484534770326266, 0.064848002864384260424, 0.052540856723847055576, 0.019959370694025667381, -0.0014169015468489787844, -0.0077282962825917641755, -0.0045477875533456602747, 6.8666057138856890774, -9.781680870206816536, 0.76788033824422385898, 1.6947413526253400917, 0.72993004769223213124, 0.048051827403947235118, -0.0814423055027337639, -0.026764743140822690876, -0.014869210341326755814, -0.03708229257507271448, -0.040998339859265417462, -0.021657580958803449023, 0.0019550515490337309898, 0.018224528638137570818, 1.1510736095266185242, -0.26354674155904200816, -1.1576706957966373945, -0.44586374545074480835, 0.25913187767265932759, 0.4101013019249517555, 0.2424073375283249121, 0.028947894281604439476, -0.073526299913971512878, -0.073211401459785760415, -0.037194314670546936574, -0.0073819538982692230189, 0.0076480252046911998895, 0.011966426005500791288, -1.3181296444717192085, 1.770595831294483169, 0.28540736672374533711, -0.54885772612206051857, -0.3290146456714282408, -0.046806598370339111415, 0.069100335389918957074, 0.070190706704184924192, 0.040088434344748846705, 0.01363447323645698879, -0.0035833162051073699089, -0.011182178982558759264, -0.0092199322453438304487, -0.00030485384089667922312, 0.91911171366009591921, -0.43333916268897615387, -0.94705452162663561477, -0.18688725466747124759, 0.36536398253325297336, 0.37660526100119018356, 0.14652064625436889145, -0.043352091120326285978, -0.10102377665872845636, -0.075438382258640523936, -0.030695751136544408011, 0.0014738906308377447245, 0.016219257186579016938, 0.018319560875233925623, 0.24077835161893199056, -0.32430916787659780365, -0.10697725906066013479, 0.19369264012185646862, 0.069510474152348900034, -0.032911881199179170154, -0.036689213624797455127, -0.014730737162785355332, 0.0028113280433886781579, 0.0080426589481052274122, 0.0060223207629812010658, 0.0020096417520420618158, -0.0012195596578522924398, -0.0029169036627022088513, -0.67057132653506423825, 1.1623714310917754311, -0.12077755899387536409, -0.48914140598953881334, -0.073927702965844735439, 0.12843692634897871385, 0.087706172020578801463, 0.011976886740303090487, -0.021269794957703556937, -0.020795616893638388389, -0.0083637529605239211566, 0.0012073475522974316376, 0.004471904388423352146, 0.0033401591918323877452, 0.49433610631161500892, -0.65691505120121584937, -0.19319414308017729409, 0.33570264840880953416, 0.12497722180510442702, -0.047820343605928981867, -0.039665199251588061091, -0.0088807108416235746445, -0.00079167850954989628723, -0.00346811469770386821, -0.0034770748060805466324, 3.8754708280944747192e-05, 0.0028577062152678137791, 0.0028120714267750694659, -1.26610284476697732, 1.4893636549432065497, 0.38667780674633295268, -0.35590571363356537926, -0.26539170811929779781, -0.085190381330964654039, -0.0051533782809038012132, 0.019954821916897545059, 0.026831840707544123148, 0.02406401577074733858, 0.016076342665629548417, 0.005775420066775712423, -0.0036269101276367996835, -0.010431951355462737441, 0.03748190809868219564, -0.41794836888546993414, 0.20681991901774554599, 0.44857915553323812752, -0.02456434815486352799, -0.21736615107253609569, -0.11449413695350998954, -0.0040850851697243183852, 0.037044212985523967963, 0.032956350515598781259, 0.016742567297277858446, 0.0038712182157707212236, -0.0038152358827065694508, -0.0084309447028460254153, -1.2792653679117500509, 1.690504888035501363, 0.24653778068385237554, -0.47810066655762717414, -0.26473577550827048777, -0.059967181740472538798, 0.029483399201633026659, 0.051561628040403234552, 0.041625597513948432715, 0.020993549497151924033, 0.0024231397883844084731, -0.0081172030199148630797, -0.0092002785046017467735, -0.003149660858429557779, 0.32160915421431796046, -0.37459243264871455725, -0.11214945192934286378, 0.095990174517830373357, 0.052974096188395679297, 0.057051565853624065794, 0.015452782814498727121, -0.030199482552889209824, -0.028724491742717805226, -0.006420423039021740455, 0.0079505888112615056423, 0.008711893450054098223, 0.0025946050631163401073, -0.0046192777438131106885, 0.056855794956504883408, -0.35170557053803264136, 0.86877866150857141481, -0.20008885552163566124, -0.41719689257934461368, -0.13243813871704457341, 0.049709152481246569877, 0.082922014301548899184, 0.048020251417999731958, 0.021769509799146841839, 0.010562768520277236381, 0.0038650579052693033975, -0.0046935081942018862244, -0.014773896298152436671, 1.0249814844133686709, -2.2598296804284645489, 1.1202773199574058705, 0.53004554316514418399, -0.29671960055795165712, -0.16072538736124133973, -0.0020078252936655853131, 0.033171299875124501955, 0.017289357058315558968, 0.0034763449164875778047, -0.0012536067772119798783, -0.0014715878340611514199, -0.0010348370628914360225, -0.0004973641060155754726, -0.2657377442081156893, 0.48411692968766562029, -0.013497200152217071409, -0.35783977201694144865, 0.063672997309753773254, 0.1015209099811493626, 0.021529090143865715207, -0.018922774876631857227, -0.021028437483353378673, -0.0036658530094324784029, 0.004812552102849323131, 0.0037551202434049912871, 0.00062051807586329028356, -0.0010676610065568532815, 2.5594840132142748779, -4.7652241673201798378, 0.61735852980938532575, 1.5472547345069211921, 0.49836138576636795916, 0.02065406904299302987, -0.1912031580824939403, -0.21160369820564281507, -0.13402059908840840974, -0.044399295628301305161, 0.014812757722211558967, 0.036892922049565554654, 0.029880423172673566123, 0.0064297337356386680129, -0.075435469504213675807, 0.35340190803832793875, -0.41051453139010540827, -0.12117471959771580758, 0.35776971206242441648, 0.022115667338660764674, -0.11335182306194308, -0.046389174418632307673, 0.0054635107425106558057, 0.020564963934543854102, 0.012884873225814753861, 0.0012266019987236013392, -0.0042255890079790914271, -0.0022986773098881271502, -0.18719463137668479469, 0.33711736306714351619, 0.031441417745102159809, -0.34296342835200510635, 0.10233941070278004815, 0.098153554457536568156, -0.013900917488055769891, -0.027783724336623363826, -0.0085048309084326603746, 0.006838079295522728196, 0.0057162633376315911612, -0.00041758476224681799743, -0.0024801808415385682288, 0.0004441734000077643971, 0.65941889031365652496, -1.4900386055966314824, 0.80106536526875471438, 0.27342517343251421247, -0.1795267697644890581, -0.085284895382289227506, -0.0053747031994144638806, 0.01794027476228902962, 0.01364298296480223778, 0.0039877369698815496105, -0.0031793818146092322867, -0.0038000838996467456617, -0.0012121442813641917797, 0.0016019245377626976663, 0.20646098590499503689, -0.87331039888777384217, 1.0615781120593952291, -0.057195080409274810795, -0.45047572728124751773, -0.084161771194599499513, 0.12227767944978089332, 0.10609806527820178634, 0.022197567097435966926, -0.023200771503257298922, -0.026334397457575303009, -0.012492360241531955123, 0.00023202922805026489437, 0.0075709860081293687412, -0.27791692484827712484, 0.12221167324497296536, 0.7753839053058471098, -0.34866030044825946366, -0.41739558456490966076, -0.092929658403535550515, 0.11024423476029625302, 0.12936234207455021838, 0.053425263623085549047, -0.0052698025916197458948, -0.024503094208689971123, -0.019480548744144791234, -0.0070093172741191615491, 0.0057915440987636422621, 1.9491704742710693754, -3.6516175214082458034, 0.59321272216999698657, 1.2289086931450488027, 0.27192614285947142028, -0.09978201236928489759, -0.15591525288952659545, -0.12623951644994571564, -0.06218624835204308865, -0.0036615960192938439011, 0.021445370215729030183, 0.02223056886974867502, 0.011250690853688563239, -0.0042915678765707062498, -0.1318593870494906839, 0.50758073423987881689, -0.52572462407301445708, -0.49728148258672799642, 0.58507040270202570476, 0.37352750263175121814, -0.10028172874582153007, -0.1870998043069350969, -0.090910813873311763222, -0.00025158281798946802166, 0.03161213266080961648, 0.024703008448202820602, 0.0071753631968888859777, -0.0077282964727326810758, -9.7816808702366557782, 21.941568047109658579, -12.177273525280746469, -1.928203921804106713, 0.81020607780091369321, 0.90568762591642626436, 0.3066171982666893725, -0.027047108688030820078, -0.045298955484015475392, -0.00073309614881027890775, -0.00035994459471653085861, -0.016840998429793269803, -0.017437196924131537212, 0.0019550515838954595112, -0.26354674159247110099, -0.058033114899867559167, 1.1742241328749685181, -0.35051869551784159418, -0.64249780710112125437, -0.20376777715482657927, 0.10440222068845264602, 0.19966135197368053977, 0.11744631841269521921, 0.010024095891854581733, -0.037294019610810330567, -0.033782149838850895895, -0.013122468678253354502, 0.0076480253987293205301, 1.770595831427705269, -3.4989649846309123227, 1.0904262104979434156, 1.1137878779012480024, -0.095530363038634422135, -0.28425294585958454929, -0.14364443919398897642, -0.011097343091258368014, 0.031434145819163795965, 0.026941905184930760309, 0.015216698508406583182, 0.0063070891891762554879, -0.0018304297579698554539, -0.0092199320675163901756, -0.43333916250970344608, 0.25648086087962213009, 0.92441911412519339741, -0.43131737813469345211, -0.53717930002901603803, -0.10106463344621874345, 0.16740171467876199807, 0.17955068439934843139, 0.064355193219467032861, -0.022955145766089344456, -0.046733128585553888878, -0.032070393943179405227, -0.0071176508091999827338, 0.01621925687038119554, -0.32430916791438407776, 0.57374384934063804575, 0.024315392781795851923, -0.44589917308602489143, 0.061526343354789229945, 0.11990392505459848693, 0.019913474191255067791, -0.014317736977412509614, -0.016503176259875761145, -0.0058479159690693225879, 0.0014186347946273437405, 0.0034709488475335010214, 0.00211043638152865181, -0.0012195594276664127511, 1.1623714308727048916, -2.5227289380499704485, 1.1176844591372219195, 0.76393580702558994133, -0.32247313456639120632, -0.27441158238326346375, -0.0089666545101953053915, 0.06617565022370497374, 0.037690285096797865316, 0.0050143618500012327593, -0.010500180593335382145, -0.0096669947246427571919, -0.0024897086190761607595, 0.0044719045454403331605, -0.65691505121961202285, 1.1958875681089202736, -0.070254192596265127579, -0.78266523443242475011, 0.11048381628771145047, 0.2282536714717221249, 0.021862575004090042952, -0.033713203916252709735, -0.017512774938462889046, -0.00063985688654796119001, 0.0015050353395980650001, -0.0015130565605650401692, -0.0015878379443391142135, 0.0028577061720115358792, 1.489363654914997559, -2.8953201858728050411, 0.79447206028360439767, 0.94398416566569687269, -0.014383403373184956361, -0.20857945989617623916, -0.1056136520253448674, -0.027627862121491632286, -0.00029031434691346134913, 0.0098655051983747472921, 0.011263688782621726972, 0.0094022189402609406689, 0.0043410511381178212842, -0.0036269103731026183617, -0.41794836896672676874, 0.94209316698156742032, -0.33961929682006142261, -0.65530018632421549452, 0.35908205656757902213, 0.28257862323716514963, -0.054275005725224381714, -0.10688955614049205911, -0.047903500114816283117, 0.0021455568224208153294, 0.018497518611400150246, 0.014083625538079942371, 0.0042228915263472725089, -0.0038152354573717899011, 1.6905048879570905296, -3.3919645302731438186, 1.1541169801537980444, 1.004972169735251164, -0.15079906838329007468, -0.22229576248018800833, -0.10096312073909474716, -0.023105642719226302667, 0.011064709441980695795, 0.020268007364460723096, 0.016922249080718508585, 0.009322951317747638425, -3.4420899981316534569e-05, -0.0092002784320670220192, -0.37459243265235081521, 0.71180159096786110862, -0.16437431628370752223, -0.28443618932304515212, 0.084569665178410372008, -0.018209584247939204771, 0.0099904262092533491907, 0.048360952337728010941, 0.012288178728214759622, -0.015509118310345597, -0.014823918231803415169, -0.0031760074252758486134, 0.0036850783147868589572, 0.0025946050792161035568, -0.69739281614301751322, 0.86877865986815450139, -0.35637326754743253732, 0.77439575978651264965, -0.18824519498497319092, -0.41313085425326301348, -0.1483410712881284288, 0.02161773020920388505, 0.064327123379517223145, 0.042482585547826696992, 0.024630522474532106497, 0.014683338260496192965, 0.0038650577373181187538, -0.011648519552180930728, -0.12917147210019713777, 1.1202773207752731999, -2.0347952412687608792, 0.88345791340832002803, 0.53832273874998970431, -0.26081316171452983976, -0.15376528443401307755, -0.0079208301189359550665, 0.029498360763861714712, 0.016960589397012632712, 0.0044478898706328869744, -0.00070380332722557456884, -0.0014715875617498745936, -0.001316479477202546532, -0.061927271785692135364, -0.013497199069430658794, 0.18198279611942627665, 0.069357149962283454192, -0.31844849005893494098, 0.062913302319526479001, 0.097184804732808008354, 0.015834372023742333824, -0.018535193734274792166, -0.019951918132571595105, -0.0037275780586117578629, 0.0045980559496288870083, 0.0037551206259484712531, -0.0006119714741570028314, 1.6205429513218039261, 0.61735852967353999077, -4.6103613291481302383, 0.46266285551015634558, 1.6815638263955428044, 0.66256331041552041228, 0.12237234605166232015, -0.15648328706293737622, -0.22168178712263056362, -0.16174051406403253117, -0.069441810019954142152, 0.0011764845311974234977, 0.036892922276967075879, 0.042657734312535826315, 0.24631821743366144606, -0.4105145310703567918, 0.19800080010777748596, -0.091617486573462164623, -0.13164168916723537062, 0.30248620413863674949, 0.015323479418628339552, -0.11711918699385104325, -0.046983980086475377047, 0.0077316550793494935936, 0.023558401871834229457, 0.015080566584865401011, 0.0012266022811485820937, -0.0091043462406550836746, -0.047493370149111438505, 0.03144141806508948811, 0.023080825616760291402, 0.1630718521664475118, -0.29408659744198373254, 0.074939157061903599599, 0.084329582056578950011, -0.016578535080592763512, -0.023465813344457887857, -0.0054702377455170753465, 0.007244616829479849987, 0.0053619021327301783983, -0.00041758501542747932921, -0.002792670104470729945, -0.108840684352449521, 0.80106536557639584917, -1.4207371296157440543, 0.66982485205434283415, 0.28688493941786741281, -0.16804094120745760588, -0.084444726150742291382, -0.0051240200462509856438, 0.019689099488863617826, 0.013999286984844318618, 0.0038679852738824102402, -0.0034379764645136950095, -0.0038000837872815443168, 0.00039241437482233630955, -0.56448374223395936244, 1.0615781120583733799, -0.8903154661830590122, 0.77870256398760262506, -0.10121232079807115922, -0.43554393905644289697, -0.070173737374217198814, 0.12789931612334856648, 0.1122757186626993503, 0.02782465351863878944, -0.020230447535618270583, -0.026328698677448979931, -0.012492360487486513926, 0.004881353155029400509, -0.7675140907393021017, 0.77538390528297618243, 0.060568947956618920947, 0.63289046982947128051, -0.42482521470031547883, -0.45450346135943492376, -0.1001987759413305451, 0.10999135331488930911, 0.13747390017980962162, 0.063899476744353658764, 0.0020077639055012027475, -0.022178424359808480287, -0.019480548810851927577, -0.0047591859168554084131, 0.99024802421899260629, 0.59321272169195682356, -3.3559422950916308892, 0.50813423508356669878, 1.2958855194621035078, 0.37530907347771735871, -0.057833435130463885232, -0.1563225027138779033, -0.13814343753608596121, -0.075434411408713958336, -0.011330921323003367851, 0.019753336385027318867, 0.022230568746584581513, 0.0080314714889066878001, 0.63010484499989105611, -0.52572462425417554321, -0.24816905290934576644, -0.16951478594142252443, -0.27923808865978960725, 0.5787547055975490462, 0.36354339114640676911, -0.10371656438864981786, -0.19267409234844889987, -0.097415118196021582841, -0.0037676057685642957984, 0.031826472580090456321, 0.024703008189800645944, -0.0014169011750224643278, 0.76788033835387359272, -12.177273525287176881, 23.711679718920272819, -11.924537483849208641, -2.0562882863425655522, 0.60081585940520565448, 0.77455046646835223889, 0.29325094585120242696, 0.010092499208020952461, -0.015059133381981370733, 0.011776943349460368415, 0.0024885305894536504691, -0.016840998426806537414, -0.021657581243425864936, -1.1576706957039673007, 1.174224132977983448, 0.017248055317581802853, 0.97908561125847315409, -0.46580013291857197411, -0.73706384213089748414, -0.22249201362045373243, 0.11348388880899047682, 0.2110575309155986734, 0.12920229940000330093, 0.019158853254722597176, -0.034214303028186975053, -0.033782149550468300314, -0.0073819543322326472423, 0.28540736626378959695, 1.0904262109295164151, -2.9898268838936603586, 0.96132349425581364866, 1.1153916597760238361, -0.065797784600899952845, -0.27358024482784432108, -0.15580187096239267452, -0.027110589123359830205, 0.025130029039046739708, 0.030261778873308914362, 0.020325475223013454024, 0.0063070895338811962944, -0.011182179301511760428, -0.94705452174105253516, 0.924419114008217635, 0.19362333610788914906, 0.68744931166248302024, -0.56025148762132293001, -0.58914930631385131754, -0.098885136007337087749, 0.18264011017991477193, 0.19950478733624993355, 0.081205162134677877361, -0.01441502517677186769, -0.045893854364072797503, -0.032070394316151461245, 0.0014738909538320891027, -0.10697725906239743954, 0.024315392877474237276, 0.20429067989459728794, 0.12163819372532058694, -0.39929432325462604902, 0.048404243885713030271, 0.11950870292166063114, 0.020051408994617193893, -0.01456316800796896016, -0.018070585967164289748, -0.0072639869113855651608, 0.00094532388249835670128, 0.0034709494219542003958, 0.0020096413085961121504, -0.12077755843769344413, 1.1176844586605738652, -2.0729178425300549549, 0.85353525002687280487, 0.68819565627017553755, -0.2885694789865373977, -0.25976517147169675415, -0.0092582530348195024583, 0.068374579528671861062, 0.039492146791909973902, 0.0057238083741901583934, -0.010780623401839955003, -0.0096669948058785114264, 0.0012073473021672834898, -0.19319414289774597404, -0.070254192851880878368, 0.59039290148260281654, 0.10506228443391781024, -0.68830326043200484865, 0.05981155235374787521, 0.20710526786270846977, 0.025653149521468805727, -0.025155219874957425452, -0.013198253086955335361, -0.00070253701481668914819, 8.9789872757940411563e-05, -0.0015130568570685780372, 3.8754867092460334663e-05, 0.38667780675799007239, 0.79447206021927085917, -2.5209903643951818708, 0.67597227697813644376, 0.94103435172501426198, 0.044395550604591993427, -0.1798622901724202805, -0.10777036849079053848, -0.038103752643405590372, -0.010974283138354697098, 0.0034934168558163371243, 0.0090368200493241273435, 0.0094022182828244829644, 0.0057754208065467615291, 0.20681991912935590583, -0.33961929685705738446, 0.2678376777115618057, -0.043824356580456663712, -0.50461489596056730011, 0.32751846867366524174, 0.2742917179748386447, -0.05167298707615718828, -0.10911402692247884616, -0.053896034052416524895, -0.0035233845313277778466, 0.015633195740604576701, 0.014083625690821732074, 0.0038712179530230325999, 0.24653778064749251597, 1.1541169803560016316, -2.9829426918826333903, 0.96381181286443418088, 1.0466518305679075596, -0.095566323772353647947, -0.21540027172737147287, -0.11594386237318494992, -0.037916525545144394438, 0.0042626670262284963031, 0.021522815639858843689, 0.020555944065016321792, 0.0093229513060898282911, -0.0081172029306515527947, -0.11214945203867378309, -0.16437431608680305728, 0.57012313192674590745, -0.041132037757008410583, -0.37540667429661161991, 0.056325882277947728061, 0.0139590057796012499, 0.022936263477723716625, 0.047312496049211567872, 0.0085656268344793517339, -0.018188817486089418202, -0.016567865874157088057, -0.0031760072109662832937, 0.008711893474306699886, -0.35237284010517194721, -0.20008885483738247935, 0.77439575916893210739, -0.42242229725895258374, 0.77956255100950433423, -0.16291477534294074303, -0.41639785199956530493, -0.16299455900017353893, 0.011759563644563278406, 0.063514625971307608898, 0.044991839599428554153, 0.024630522282205656592, 0.010562768724224382011, -0.0010195332164386423836, -0.36185240887994613024, 0.53004554248237711889, 0.88345791361519099283, -2.1137339448568392974, 0.86174153728442448585, 0.5608456122593629134, -0.24595874484039473296, -0.15164012322830724977, -0.010023879527429512687, 0.028423444652801513333, 0.016882697356309008668, 0.0044478901026588860829, -0.0012536073926221740372, -0.0022993486758481427451, 0.16898525479460632659, -0.35783977294057472962, 0.069357149632043660348, 0.22577662384864763401, 0.094311075049259618375, -0.32518734840832469457, 0.046040635867528587055, 0.093359853154226526972, 0.01823383236374172961, -0.01634924961946271349, -0.019533213094530644388, -0.0037275778041960111464, 0.0048125512168522175879, 0.0024683494051733170967, 0.11895803669775868727, 1.5472547342378473179, 0.46266285518502603402, -4.5063129047149956818, 0.47459033545478579441, 1.6809586749418690133, 0.66183707632875077298, 0.12131285765182799052, -0.158920210053693306, -0.22568480222542813829, -0.16641031664670702384, -0.069441810062498832146, 0.014812757446782578996, 0.074692189758391022703, 0.1149368410461335116, -0.12117472009375855613, -0.091617486452484936388, 0.20424137853403637188, -0.14288027736531236367, -0.16532574940710845213, 0.29361015459024747409, 0.026401190079275958789, -0.10710801313965023773, -0.043374530293694806005, 0.0080181283973880801419, 0.023558401139318015105, 0.012884873398938593778, -0.010006753149209724613, 0.14632806368732953484, -0.3429634289548620929, 0.16307185275888327891, 0.068721431463237225845, 0.15645543333530484431, -0.31194416145974768195, 0.063492958052267386981, 0.088545282763297664119, -0.011063062359111232313, -0.02218233249979829208, -0.0063451878545584348285, 0.0072446168253472503426, 0.0057162632569550094266, -0.0046928163928510024244, -0.20849048888539645219, 0.27342517295057600091, 0.66982485270924718712, -1.4693504894546474127, 0.64793604350922517909, 0.29870021927178663779, -0.15536792079579431203, -0.079300787982646572805, -0.0062869297221999596859, 0.017729100244736974851, 0.013135321510555362026, 0.0038679852654189867019, -0.003179382046014229881, -0.0033318660988996460191, -0.19595221643725474481, -0.057195080152025068454, 0.77870256374407242905, -1.0046531793111903585, 0.79985066596743736866, -0.040428613725712041205, -0.4135121010978238254, -0.075661448529473915303, 0.11840037891177596086, 0.10905901471475405728, 0.0281976242352282927, -0.020230447734053823294, -0.026334397197928859774, -0.0067762778062684988351, -0.19901921874659128653, -0.34866030009813581136, 0.63289046941773630106, -0.073529277447482305208, 0.62871318414127952412, -0.37611228535631774372, -0.43225163203580507787, -0.1004025519483405654, 0.10349570963247944899, 0.13465818574582094636, 0.064227614539534197413, 0.0020077640903949645068, -0.024503094281121549219, -0.022895580362557378662, -0.14294698265057501074, 1.2289086937567519353, 0.50813423513596234216, -3.3354122648503561344, 0.48812758762348434916, 1.2746716536806641518, 0.36883464213420164013, -0.052688734753138011246, -0.15064760411025224229, -0.13464103555545847701, -0.074461807000918894817, -0.01133092076729167591, 0.021445370332847973199, 0.028147275669561083039, 0.60276793572362930007, -0.49728148183077292011, -0.16951478608591000197, 0.01623953399069681669, -0.15595314598698689634, -0.35045469555309538556, 0.52572685645994698955, 0.35946369114091053509, -0.088824118866532963956, -0.18216861199517453418, -0.094791435158357431612, -0.0037676054457021837529, 0.031612132639597612105, 0.019959370564309436569, 1.6947413526172305787, -1.9282039216884181432, -11.924537483991889175, 24.555162061416648811, -11.756155870115916784, -2.1986340091188503187, 0.49285786204172166158, 0.75848847708743150076, 0.31072095047141146118, 0.021199824368723561158, -0.013855652723671764426, 0.011776942789152144986, -0.0003599446200266997499, -0.040998339528532282139, -0.44586374566648617757, -0.35051869550491354666, 0.9790856114506768515, -0.19493925424684996339, 0.91624666285545985911, -0.39825839201900942177, -0.68416000926617492883, -0.21564805948276191683, 0.10078158155817081021, 0.20228502950482810108, 0.12785593802835712451, 0.019158852898388520347, -0.037294019739076957054, -0.03719431431576717545, -0.54885772556502665509, 1.1137878775700407186, 0.96132349378195047773, -3.0899346298595391502, 0.87462698438784680643, 1.1122112234787309148, -0.048114192502334209023, -0.26461978941138131205, -0.15311103478816662227, -0.024207468933526914839, 0.02795541277500315086, 0.03026177841364268567, 0.015216698510669179542, -0.0035833160739514200929, -0.18688725459881833091, -0.43131737776405520979, 0.68744931146419196999, 0.003280981711961391406, 0.68544044549768801122, -0.48323737752287970748, -0.54817854881955352386, -0.095794450298592265236, 0.17298832982553341564, 0.19364410110415811173, 0.080229648409529327746, -0.014415025081967850623, -0.046733128612023576431, -0.030695751316371325257, 0.19369264015256823508, -0.44589917316393296032, 0.12163819358542589388, 0.25285178245131045838, 0.14056433659868725061, -0.39974450450672627255, 0.03573616296656775132, 0.11389447724107901183, 0.01906352597174500868, -0.013913434422725927744, -0.017562200569734315314, -0.0072639865956790712273, 0.0014186341141386108865, 0.0060223211697197328099, -0.48914140656331411838, 0.76393580726784038326, 0.8535352505903076592, -2.2396815830144478632, 0.81765106647214624846, 0.72230870706576089102, -0.25650958157622028022, -0.24955186934583470371, -0.013275958799009279776, 0.063665889148196902481, 0.037853421290805641664, 0.0057238082876647396191, -0.010500180316844133846, -0.0083637527892220075304, 0.33570264806067406926, -0.78266523399863319188, 0.10506228456682628614, 0.69590059907762724656, 0.11196799546368231548, -0.70680345974176073121, 0.043140481394708662943, 0.20790211317410708936, 0.030633246466054454837, -0.023978192398057977808, -0.014508804137071784623, -0.00070253665050716215812, 0.0015050358357343758542, -0.0034770753652830928802, -0.35590571379680252528, 0.94398416587826239343, 0.675972276973826669, -2.5940994328614732645, 0.63931288682209697072, 0.9475995885641556038, 0.057673233289239081434, -0.17453894036475445239, -0.11038449316834680891, -0.041395126610432167458, -0.012584351851187457888, 0.0034934171421182880213, 0.011263689417470416215, 0.016076342007611096191, 0.44857915532737391295, -0.65530018644300735975, -0.043824356092973193455, 0.4472665182723827848, -0.036635481534833570783, -0.54333499443013255625, 0.29576959641248989907, 0.27022172359160862021, -0.046068577016507183952, -0.10671759257978659374, -0.054981308555282354389, -0.0035233845959138504478, 0.018497519285697108793, 0.01674256690140899298, -0.47810066640996529141, 1.0049721692023063646, 0.96381181344850519466, -3.0405790712919387175, 0.91075775190887542365, 1.0420993945439831485, -0.087336782626235467952, -0.21075362434034217163, -0.11369834289927147064, -0.034959544882050885839, 0.0066939456926796026687, 0.021522815636091052055, 0.016922249210510928452, 0.0024231396030975813592, 0.095990174776870648699, -0.28443618933352232681, -0.041132038248670808878, 0.48994603261980806952, -0.037241677339698020521, -0.33912129010026592635, 0.067486346308772801605, 0.0081493851587141896697, 0.013108968061041269618, 0.042771489947414004229, 0.0080841713589498518516, -0.018188817565942691501, -0.014823918393207211747, 0.0079505887814013184028, 0.042696607195174818283, -0.41719689317267022455, -0.18824519443212062164, 0.77956255074272728933, -0.41372367503661289234, 0.77948114732486384337, -0.16468336559278912157, -0.41796446732145370317, -0.16225135594529724314, 0.013278360765146744044, 0.06351462639876592764, 0.042482584545102168583, 0.02176951002634330487, 0.013375577892904947502, -0.047312267298973365026, -0.29671960008985326374, 0.5383227390691579517, 0.86174153654131424585, -2.1234979854510287467, 0.85951601532627264213, 0.56509878534041246834, -0.24289042236310040557, -0.15092373895334809664, -0.010326175261454606166, 0.028423443845069937264, 0.01696058982402192869, 0.0034763451897567585261, -0.0033846852736167777967, 0.051867435280407610254, 0.063672998015604359567, -0.31844848946972353909, 0.094311074315601789442, 0.22665696814112176738, 0.084996693858512176178, -0.32895251859502283587, 0.045842447953117766712, 0.094309679902408488639, 0.018832980564853486066, -0.016349250595330388353, -0.019951917186870955778, -0.0036658526487110638589, 0.0077980004113038922306, -0.41717335041279596242, 0.49836138595114964911, 1.6815638266439827397, 0.47459033478200823852, -4.5451388068004261456, 0.41871842698630262758, 1.6436284757760397657, 0.64773443102092775892, 0.12100261013432836443, -0.15712424516254897, -0.22568480207261731874, -0.16174051320531676601, -0.044399295674287388158, 0.080384181717726072014, -0.16652490477125653223, 0.35776971238029797728, -0.131641689199095413, -0.14288027772890360412, 0.18065868328292228329, -0.15282164457597657847, -0.16128124120039233347, 0.30243563399647593881, 0.032534334362559015619, -0.10538046923337859861, -0.043374530809752211147, 0.007731655899747796705, 0.020564963774464171953, 0.0021111311319322199551, 0.011946371170110386614, 0.10233941128768744844, -0.29408659800317055177, 0.15645543361192051912, 0.05727280945556770253, 0.15131417742866357101, -0.30856344531476609383, 0.066875610161050133562, 0.089506943333365507343, -0.011543835887406659127, -0.022182332758302468478, -0.005470237538927206003, 0.0068380792674152170857, -0.00028242545923442326922, -0.020807057371756917247, -0.17952676928142632939, 0.28688493892721772527, 0.64793604362497736471, -1.4781397284008670479, 0.65072834698277848275, 0.30521858121725925495, -0.15358552144903334091, -0.080252915783619718781, -0.0073471969658908242023, 0.017729100159138273113, 0.013999286905049637339, 0.0039877370479503189415, -0.0075924383779882509068, 0.22315601431990997328, -0.45047572728511769968, -0.10121232089371101637, 0.79985066637066548623, -0.96144572048453635915, 0.82491614158166925996, -0.033990279737151442296, -0.41871477537889334464, -0.081834724645641690466, 0.11568063162676685274, 0.1090590147928552911, 0.027824653771872279173, -0.023200771640138635049, -0.030844970213727066627, 0.22818126252809953214, -0.41739558461447562321, -0.4248252144672420294, 0.62871318435206713371, -0.032182244711833764284, 0.66301985383939243679, -0.3603035756084617125, -0.43238740183285861995, -0.10567087064373120731, 0.10011788667833874944, 0.13465818597992354477, 0.063899476172166222443, -0.0052698024208825394038, -0.044867447449788669389, -0.37170184694705621364, 0.27192614260647540725, 1.2958855187501105988, 0.48812758830389274545, -3.3912362580294632686, 0.43909508430781130084, 1.2579294542372496668, 0.37150760366301782689, -0.044185045876432973988, -0.14522748911045413323, -0.13464103528260107367, -0.07543441186153027056, -0.0036615960491224973372, 0.048155063228994182678, -0.12051904481931136059, 0.58507040210458360896, -0.27923808864244126227, -0.15595314571266216674, -0.034212224277371486303, -0.20484570276056851545, -0.37189271902273113302, 0.52780376428889042018, 0.37022288798076713467, -0.081930250995182848928, -0.1821686114167739845, -0.097415118554178614652, -0.00025158276109457991532, 0.052540856585891665409, 0.72993004747993794901, 0.8102060776397385089, -2.0562882858240851824, -11.756155870178787382, 24.675382200422600221, -11.723974254713766641, -2.1966748850175483732, 0.48768423221912332588, 0.75289827165504785711, 0.30751104593697231593, 0.021199823953287823264, -0.015059133170929793477, -0.000733096198531426246, -0.03708229246677376828, 0.25913187784708435757, -0.64249780727065919361, -0.46580013272943665559, 0.91624666260768083959, -0.14319611609717264322, 0.98289687898994737569, -0.36758238476561388985, -0.68466392389072416869, -0.22582052907024516286, 0.094179223751122448371, 0.20228502940534115462, 0.12920229970370056272, 0.01002409602987954472, -0.073211401778739554347, -0.32901464603547292409, -0.095530362973095445112, 1.1153916605187434996, 0.87462698401366045786, -3.1377270384385247404, 0.85557677945273746811, 1.1139864387028566295, -0.037982198451322363164, -0.2548218330739441817, -0.14825109943341416296, -0.02420746959204438456, 0.025130029863843630217, 0.026941904854035959166, 0.013634473300243053762, 0.36536398258183466758, -0.53717930009102843325, -0.56025148791503287526, 0.6854404458549411272, 0.073808890614036809508, 0.74372780363228252476, -0.45949489129157727874, -0.55109570122376827506, -0.10654776597512391478, 0.16634903491113242624, 0.1936441009894813714, 0.081205161764317232431, -0.022955145787612579039, -0.075438381850287491992, 0.069510474201489175217, 0.061526343236144427629, -0.3992943228034183667, 0.14056433607181126333, 0.26120068348558167504, 0.13742519990729543111, -0.40491681671383050745, 0.031889023867061265516, 0.11313751664183596679, 0.019649860136102415692, -0.01391343418370847608, -0.018070586368021474, -0.0058479153497594926064, 0.0080426585827521150651, -0.073927702528207558941, -0.32247313456599174808, 0.68819565543430882837, 0.81765106673983600771, -2.2419555408355926751, 0.8279024656858229303, 0.73555510136994572967, -0.2527975913870199598, -0.25200557699710196546, -0.015848288126544589349, 0.063665890284123019538, 0.039492145930239433427, 0.0050143621656591093327, -0.020795617221372412337, 0.12497722213840670169, 0.11048381589961824012, -0.68830326082821535749, 0.11196799597347309452, 0.6950358412019667087, 0.11381804378093035579, -0.7046141248811667035, 0.043182048074746501309, 0.20670271223340769273, 0.029244782110855460644, -0.023978191858524504354, -0.01319825400669314211, -0.00063985719611414464125, -0.0034681140090366026352, -0.26539170790046129333, -0.014383403735438534191, 0.94103435169379745506, 0.63931288704984079452, -2.6221070554532812658, 0.62084705458125521993, 0.94379842873054919927, 0.058526762674575995293, -0.17311668431714885563, -0.10984147330791775576, -0.041395126292015735348, -0.01097428310371736937, 0.0098655046942900583712, 0.024064016016947475246, -0.024564348082286038832, 0.35908205678260907145, -0.50461489668042247558, -0.036635481068837974505, 0.42819543725268310874, -0.057363826443256979437, -0.55564909489913194562, 0.2929876620781637353, 0.27214187431313896504, -0.044801516154416201609, -0.10671759223543392237, -0.053896034358394830088, 0.0021455566227594462941, 0.032956350699068485965, -0.26473577554162969205, -0.15079906800076986473, 1.0466518299854628005, 0.91075775218852283999, -3.0887361296046487169, 0.88543298958548000144, 1.0404676220975010992, -0.079410353152935514442, -0.20223708077205432287, -0.1091982621524430902, -0.034959545019188860426, 0.0042626671998467150632, 0.020268007298150907819, 0.02099354947529061613, 0.052974095833579940007, 0.084569664810390660081, -0.37540667304628361212, -0.037241677759699373662, 0.53637069086798982998, -0.018957177178412602803, -0.34337436694476841659, 0.057710511094046505753, 0.0024255821261667228142, 0.011519277205584573312, 0.042771489245301441695, 0.0085656276270813080237, -0.015509118650142064005, -0.0064204228897476022658, 0.144550983994423482, -0.13243813864638709354, -0.41313085484915423873, -0.16291477532342543699, 0.77948114688160219377, -0.41770357286787040962, 0.78093658280711586439, -0.15948815764941073891, -0.41444534197361188044, -0.16225135633407938851, 0.011759564135742342841, 0.064327124094347210015, 0.048020251406392065741, 0.029109588430351378274, 0.063675374356741698545, -0.16072538799198937687, -0.26081316176433544385, 0.56084561225159579312, 0.85951601568148716659, -2.1291613468195089709, 0.85776524977795531868, 0.56620082643960156776, -0.24214214379408152689, -0.15092373847261458808, -0.010023879453629574304, 0.029498360483402465138, 0.017289356801702779332, -0.0016796959282436448556, -0.036126328630272382114, 0.10152090905462465842, 0.062913302480669272554, -0.32518734860476544712, 0.084996694367231465095, 0.22641563848889023625, 0.087139410941517003728, -0.32803247450149963793, 0.045843446349206014301, 0.094309680500395454139, 0.018233832504564787685, -0.018535194307842089989, -0.021028437807760796269, 0.0076968954741564628297, -0.47871919869047330209, 0.020654069300500475792, 0.66256331084256114838, 1.6809586748633262854, 0.41871842702640438283, -4.6014169154923703076, 0.38314920983452538383, 1.6306650529876922651, 0.64648596432418203861, 0.12100261097771931873, -0.15892021007420154, -0.22168178759654533216, -0.13402059925664744355, 0.027875737487698559747, -0.10487744261979786708, 0.022115667091943458367, 0.30248620431657097241, -0.16532574946522651804, -0.15282164443508702845, 0.1859710625910406534, -0.14609915150260044392, -0.15742187857658129779, 0.30386372597359284642, 0.032534334693154420837, -0.10710801301092791327, -0.046983980573196061636, 0.0054635106683514371093, 0.027125598841442867537, -0.045933634052249067148, 0.098153554272897305943, 0.074939157137266385056, -0.31194416174668260222, 0.15131417777954780934, 0.063502470998303375826, 0.15615000554752725415, -0.30803811079956783248, 0.065915440334365263397, 0.089506943131503563515, -0.011063062039860755564, -0.023465813227021247583, -0.0085048310776607693312, 0.0095432917146990679097, 0.040652880338350509681, -0.085284895717756062372, -0.16804094087960205761, 0.29870021909629312917, 0.65072834719138084036, -1.4777217363516610416, 0.64959347293482927466, 0.30445569883857259708, -0.1540306093038699975, -0.080252915822957779368, -0.0062869296560513652714, 0.019689099596752881322, 0.013642982956270956196, -0.0059109774910767525977, 0.21491509149263118839, -0.084161771095609946847, -0.43554393868927515632, -0.040428613740113736319, 0.82491614145600877794, -0.95753082688614510332, 0.82276816983932599125, -0.036183565803729222932, -0.4204101271160640696, -0.081834724911226355837, 0.11840037888944593414, 0.11227571863922344853, 0.022197567360376310036, -0.056388353575544704344, 0.26685220915233087702, -0.092929658461994968377, -0.45450346111579409403, -0.37611228526000339834, 0.66301985368324090064, -0.015271564441783000432, 0.66887138256663414282, -0.36007746761806447289, -0.43418555669961078625, -0.10567087094169272388, 0.10349570968884788669, 0.1374739004273322629, 0.053425263719232819748, -0.051657606304788945828, -0.27953376610587232776, -0.099782011971141948425, 0.37530907355296422301, 1.2746716535603295206, 0.4390950841879994182, -3.4238867081569015483, 0.42575328309697535456, 1.2585789671128666534, 0.3758374213546683551, -0.044185046324682661356, -0.15064760421753137232, -0.13814343784890237288, -0.062186248225055369698, 0.046920811940111939453, -0.37070924848395703233, 0.37352750316103006112, 0.57875470537026474371, -0.35045469509766746796, -0.20484570308166422103, -0.049726799153544110066, -0.20445056994887189639, -0.36739874138824185623, 0.5316587319730696759, 0.37022288797317992604, -0.088824119463566719723, -0.19267409195899953822, -0.090910813862972103538, 0.064848002933404230719, 0.048051827606610537924, 0.90568762580400530293, 0.60081585925436942208, -2.1986340091517666551, -11.72397425456968989, 24.757042510033066662, -11.681778218706979189, -2.1938991257415660918, 0.48007463229198127364, 0.75289827151616151024, 0.31072095066694216392, 0.010092499160068942043, -0.045298955312994208799, -0.014869210502615337263, 0.41010130221386825733, -0.20376777734391263164, -0.73706384239421918192, -0.39825839196764045713, 0.98289687909688427947, -0.11316664729385017552, 0.98957138491346730014, -0.36952616121138859295, -0.68816335549173268671, -0.22582052904739055577, 0.10078158143391943713, 0.21105753054835854621, 0.11744631831990363924, -0.073526299465669658906, -0.046806598147745681615, -0.28425294614801743798, -0.065797784715599674787, 1.1122112234580443513, 0.85557677989293923204, -3.1531080801559512317, 0.8530786580617455428, 1.1210188549096216093, -0.031673854031898436567, -0.25482183264280072743, -0.15311103484001237218, -0.027110589351574426731, 0.031434145913660435645, 0.040088434339053263811, 0.37660526064470961627, -0.1010646332589467955, -0.58914930625574613021, -0.48323737725549464894, 0.74372780347837408321, 0.099812213774896380025, 0.74913516031262306072, -0.4631229207355954669, -0.55600689526650914907, -0.10654776616338565431, 0.17298832972292121912, 0.19950478775923458086, 0.064355193263565799167, -0.10102377690379060748, -0.032911881279037907833, 0.11990392484198281187, 0.048404243985570548359, -0.39974450403950428923, 0.13742519965624855338, 0.26021764012195758298, 0.13652028253154083193, -0.40549043759454411751, 0.031896811301219199497, 0.11313751691334589244, 0.019063525659439965743, -0.014563167889672457359, -0.01650317653293970066, 0.0028113283091302267158, 0.12843692604622269915, -0.27441158204443849566, -0.28856947885389111486, 0.72230870733438190534, 0.82790246486048335584, -2.240956009542228955, 0.82544012184411319488, 0.73392483205239888999, -0.25373580348270724505, -0.25200557738953788256, -0.013275959260256091354, 0.068374579987514158552, 0.037690284799871992538, -0.021269794641822444003, -0.047820343823649094639, 0.22825367208744687386, 0.059811552119437678965, -0.70680345992489768125, 0.1138180436561596065, 0.70524704379674185706, 0.11930924612940516094, -0.70551155443741786488, 0.040781005892612333386, 0.20670271206643581197, 0.03063324638344781109, -0.025155219898502480247, -0.01751277466012102213, -0.00079167865212640201045, -0.085190381430495509818, -0.20857945921024786951, 0.044395550090901514384, 0.9475995883690656596, 0.62084705457474953505, -2.6397863827437455519, 0.61071474345672882578, 0.9419744602988070481, 0.059681967572781643994, -0.17311668431218002473, -0.11038449319106305446, -0.038103752757487098501, -0.0002903139237313574067, 0.0268318404018012234, -0.21736615083295127149, 0.28257862316273241188, 0.32751846880678719964, -0.54333499481063407544, -0.05736382654942567827, 0.42280331325312137114, -0.058381132445894379224, -0.55662121188699509311, 0.29242438240941004413, 0.27214187391077310219, -0.046068576796354523961, -0.10911402669333333804, -0.047903500314489588785, 0.037044212957311604961, -0.059967181709880315599, -0.22229576279423152174, -0.095566323588571880365, 1.0420993946803016605, 0.8854329896200985317, -3.1062068801710203125, 0.8821015670361548322, 1.0463795506188227069, -0.073885726426662606814, -0.20223708090441366858, -0.11369834279311723246, -0.037916525479547442956, 0.011064709497135557376, 0.041625597406323926186, 0.057051566067265704896, -0.018209584230364637969, 0.056325881798832744263, -0.33912129020482645325, -0.018957176545768103515, 0.53527098835786790509, -0.026283669402014011446, -0.34794215882315182586, 0.056320833119961354052, 0.0024255825219439723292, 0.013108968234563044691, 0.047312495906748526509, 0.012288178978498697294, -0.028724492089732621414, 0.10511949724119611094, 0.049709153696210969797, -0.14834107090784043326, -0.41639785271758378293, -0.164683364892068973, 0.78093658256842435339, -0.41221018795422115222, 0.7855468620936502111, -0.15948815770003829662, -0.41796446743349663366, -0.16299455924658659556, 0.021617730013333120997, 0.082922014272559713133, 0.054629624517512458126, 0.04373774090958998384, -0.0020078249408681065821, -0.15376528452170631911, -0.24595874443483103922, 0.56509878494773180257, 0.85776524988099389635, -2.1306582714954727109, 0.85752822944062556676, 0.56620082677069338839, -0.24289042308403463455, -0.15164012282200731163, -0.0079208301641801124454, 0.033171300013867865975, 0.011321865625356685239, -0.037419477079652291307, 0.021529090733202305152, 0.097184804394050444687, 0.046040636275656424137, -0.32895251892122034931, 0.087139410931540803307, 0.22841873961817160676, 0.087614192516093730712, -0.32803247420847153615, 0.045842447289991719583, 0.093359853520257080062, 0.015834372028760094336, -0.018922774711761940897, -0.010116564458751628464, -0.31076603647834444288, -0.19120315801067658268, 0.12237234552738798787, 0.66183707681565506764, 1.6436284751473533294, 0.38314921005264596099, -4.6218208389100077227, 0.37607988453889068525, 1.6306650533054123375, 0.64773443083514403718, 0.12131285766655747205, -0.15648328715186921034, -0.21160369833237352433, -0.10802494071570671597, 0.013844438997209234168, -0.11335182275809595198, 0.015323479343738601044, 0.29361015492738085086, -0.16128124142963459753, -0.14609915137913717653, 0.1894259153015233621, -0.145453745479500407, -0.15742187829938417143, 0.30243563347740043401, 0.026401190307038135963, -0.11711918701955448263, -0.046389174503931339399, 0.045095420023028340295, -0.01445139234263545687, -0.013900917929341853954, 0.084329582643315775803, 0.063492958082316641732, -0.3085634455777646612, 0.15615000569210316095, 0.064746995016514943821, 0.15539469717013032479, -0.30803811101040445841, 0.066875610496467327581, 0.088545282622194645961, -0.016578535703670967127, -0.027783723816583540123, 0.0097304918909876404048, 0.030597747030543041846, -0.0053747032258811252084, -0.084444726351293908362, -0.15536792069443283726, 0.30521858111769267774, 0.64959347301385717, -1.4800846315788600549, 0.64854070871463886938, 0.30445569880590495071, -0.15358552136680506117, -0.079300788157456933436, -0.0051240198956037852004, 0.017940274812534646315, 0.0073995620420768976749, 0.055648286747760601634, 0.12227767929303044736, -0.070173737867782415289, -0.41351210128926790821, -0.033990279556173230713, 0.82276816991412193847, -0.95932407749908843453, 0.82225081270879762929, -0.03618356602093331098, -0.41871477496406228624, -0.075661448614061627138, 0.127899316316768763, 0.10609806501651983246, -0.044605655882002478996, 0.12350233823361028707, 0.11024423466034341534, -0.10019877621980580185, -0.43225163214737277739, -0.36030357540377466741, 0.66887138259011635899, -0.013534200625556691219, 0.66898922893400192269, -0.36007746777854493514, -0.43238740138318798589, -0.10040255238219249201, 0.10999135359710800197, 0.1293623417526231778, -0.0052891118448392315451, -0.12893345220277374619, -0.15591525305880798991, -0.057833435447178060951, 0.36883464231374374354, 1.2579294541084107273, 0.42575328311565308059, -3.4292134444031496798, 0.42580378344579639105, 1.2585789666934195097, 0.37150760436435087408, -0.052688735179111781393, -0.15632250179767256948, -0.12623951668201199849, -0.0081783883837853281473, -0.19530045309670804321, -0.10028172933259899002, 0.36354339123973311576, 0.52572685594080958982, -0.37189271848649740893, -0.20445057032419311493, -0.044709341820233283737, -0.20176416156394264001, -0.36739874122882554275, 0.52780376412749430148, 0.35946369152433166327, -0.10371656454727846974, -0.18709980422434907554, -0.0056524847093777051688, -0.081442305518662591601, 0.30661719840572598805, 0.77455046621800804907, 0.49285786218196442343, -2.1966748850923107916, -11.681778218587636431, 24.782539731358799173, -11.678338433585851774, -2.1938991255126292224, 0.48768423202368371694, 0.75848847708762467956, 0.29325094592530370763, -0.027047108564010366449, -0.026764743288307867725, 0.24240733710532813561, 0.10440222083405453768, -0.22249201342776750434, -0.6841600089352498637, -0.36758238488804217958, 0.98957138478511608159, -0.11367941263859081269, 0.98854283795352637654, -0.36952616124352088978, -0.68466392394406516697, -0.21564805946232942779, 0.1134838889446386917, 0.19966135199896853369, 0.028947894171343017955, 0.069100335362508411019, -0.14364443881356606103, -0.27358024531132291157, -0.048114192357613126683, 1.1139864384805377995, 0.85307865835159057966, -3.1516459032690193887, 0.85632681939330901333, 1.1210188553385538324, -0.037982199408028882726, -0.26461978882995124129, -0.15580187090490532631, -0.011097343070383937358, 0.070190706574313713184, 0.14652064618140847596, 0.16740171458392708037, -0.098885135804930537895, -0.54817854882096028746, -0.45949489132063747743, 0.74913516031007543194, 0.098448597467584886278, 0.74688648053057615783, -0.46312292066818477876, -0.55109570096033788644, -0.095794450434445219122, 0.18264010980971853715, 0.17955068436754642591, -0.043352090848607666151, -0.03668921346088745844, 0.019913474641105693291, 0.11950870231074404992, 0.035736162628436769773, -0.40491681632092346721, 0.13652028236376650971, 0.26097705620122130243, 0.13716996703944930869, -0.40549043750840918499, 0.031889023658144687035, 0.11389447728365381662, 0.020051409600304329756, -0.014317737259423679239, -0.014730737359886595175, 0.087706171979677824413, -0.0089666549102903544965, -0.25976517090216239003, -0.2565095821601618975, 0.73555510241873034438, 0.82544012126560006859, -2.245859113665998219, 0.82331050943685413213, 0.73392483152236076549, -0.25279759056376993609, -0.24955186975669357841, -0.0092582530631770132423, 0.06617565016753047813, 0.011976886844725045944, -0.039665199192388908567, 0.021862574556207332865, 0.20710526834963108311, 0.043140481406377446938, -0.70461412503760512127, 0.11930924625262115391, 0.70678228487637317912, 0.11804737674322196561, -0.70551155460422976251, 0.043182048332817656033, 0.2079021130393341743, 0.025653149951853547461, -0.033713204230923471405, -0.0088807108814342450298, -0.0051533781991676842066, -0.10561365290527086058, -0.17986228909135301701, 0.057673233132386432509, 0.94379842865985974498, 0.61071474347641019342, -2.6461988523157033981, 0.6091821509158152903, 0.94197446034367837697, 0.058526762684423708216, -0.17453894003867648221, -0.10777036933982633848, -0.027627861788069130872, 0.019954822098240641848, -0.11449413702219508882, -0.054275006011529500216, 0.27429171842615474031, 0.29576959641498784537, -0.55564909487396008103, -0.058381132563346038733, 0.4232845712592236409, -0.058956212869896856832, -0.55662121245432782946, 0.29298766272914111841, 0.2702217232974040706, -0.051672987210049863005, -0.10688955601222360769, -0.0040850851321862675924, 0.029483399205328848086, -0.10096312070378256343, -0.21540027165202421666, -0.087336782851451244536, 1.0404676222515434336, 0.88210156695497343726, -3.1053091030974133702, 0.88488014326851216484, 1.046379550491654431, -0.079410353051431306626, -0.21075362443326170481, -0.115943862533350317, -0.023105642720176934601, 0.051561628212248633574, 0.015452782950477864793, 0.0099904263099049282759, 0.013959005337905325272, 0.0674863469272144606, -0.34337436774643814719, -0.026283668990783881636, 0.53180650687077635208, -0.026926972801531771917, -0.34794215829931218753, 0.057710510395312011878, 0.0081493855105287740792, 0.022936263157109403749, 0.048360952536185927708, -0.030199482527632735829, 0.054629624001043680659, 0.082922015307070717771, 0.021617729539264034927, -0.16299455926893957614, -0.41796446729577790835, -0.15948815807008029366, 0.78554686216873514937, -0.41221018794106467631, 0.78093658261148146682, -0.16468336475909736705, -0.41639785223048120244, -0.14834107102707871917, 0.049709152645511088531, 0.10511949786520831396, 0.011321865618115571281, 0.033171300091373819019, -0.0079208303284387694687, -0.15164012268719301302, -0.24289042314443812232, 0.56620082680387362473, 0.85752822942617001889, -2.1306582714489508135, 0.8577652497763853523, 0.56509878498411081349, -0.24595874433849379548, -0.15376528463992239493, -0.0020078248775902574758, 0.043737740890591639276, -0.010116564458128113335, -0.018922774679189097169, 0.015834371879085783197, 0.093359853821492852499, 0.045842446917595612577, -0.32803247385363615285, 0.08761419225537685207, 0.22841873965495856913, 0.087139411143080700017, -0.32895251924216345474, 0.04604063659515787027, 0.097184804156586548918, 0.021529090832345217782, -0.03741947709460623217, -0.10802494069822331235, -0.2116036983268672067, -0.15648328721093807125, 0.12131285759777912792, 0.64773443112245032705, 1.630665053121207908, 0.37607988431480526836, -4.6218208385946155659, 0.38314921012727753968, 1.643628475091960528, 0.66183707639799449485, 0.12237234582798522942, -0.19120315791188263743, -0.31076603656820189769, 0.045095420068333245955, -0.046389174531398978674, -0.11711918707238559711, 0.0264011902215890229, 0.30243563370613463226, -0.15742187844283717091, -0.14545374539308444928, 0.18942591532303432178, -0.14609915154709143814, -0.16128124127354120931, 0.293610154682482416, 0.015323479642892138056, -0.1133518228410715506, 0.013844438964545209905, 0.0097304918996076270737, -0.027783723780395737368, -0.016578535916276882639, 0.088545283015272133187, 0.066875610093088136421, -0.30803811074993248109, 0.15539469708179071139, 0.064746994986455017052, 0.15615000582554269815, -0.30856344574921534818, 0.063492958157200449221, 0.084329582602102229338, -0.013900917837059397972, -0.014451392394787387213, 0.0073995620595503422412, 0.01794027484204995665, -0.0051240201265818916335, -0.079300787745838888232, -0.15358552177019266227, 0.30445569904291563956, 0.64854070868645496972, -1.480084631623463709, 0.64959347302873149399, 0.30521858107720029052, -0.15536792060848311769, -0.084444726479537532415, -0.0053747030966180683528, 0.030597746980908762521, -0.044605655785817134584, 0.10609806483542986999, 0.12789931629768744314, -0.075661448462449570895, -0.41871477499165188352, -0.036183566064252929706, 0.82225081281568823677, -0.95932407760383142659, 0.82276816994286183782, -0.033990279592728829539, -0.41351210129971233131, -0.070173737892110121805, 0.12227767939823067089, 0.055648286708160250857, -0.0052891118662801864594, 0.12936234188782175236, 0.10999135331188544806, -0.10040255214832745356, -0.43238740149089383014, -0.3600774677320952577, 0.66898922899867308001, -0.013534200685424749647, 0.6688713827144829871, -0.36030357576761867788, -0.43225163188095350186, -0.10019877635558895046, 0.11024423488424119066, 0.12350233811250416194, -0.0081783882609970780936, -0.12623951689320944691, -0.1563225016034426873, -0.052688735502677490341, 0.37150760456964554423, 1.2585789668035214373, 0.42580378348484437812, -3.4292134446255784219, 0.42575328321825606226, 1.2579294540083665321, 0.36883464234584284469, -0.057833435429368071357, -0.1559152529601299797, -0.12893345226436200845, -0.0056524846885935000826, -0.18709980426636044215, -0.10371656452453287267, 0.35946369149206541804, 0.52780376408974427616, -0.3673987410227841921, -0.20176416170012695384, -0.04470934176652111014, -0.20445057055433771276, -0.37189271815951174815, 0.52572685562513798985, 0.36354339132997814943, -0.10028172915032779189, -0.1953004532042202368, -0.026764743205920760172, -0.027047108832349564533, 0.2932509461887298774, 0.75848847703351407468, 0.48768423211383277183, -2.1938991257408360092, -11.678338433538964836, 24.782539731581547215, -11.681778218941483161, -2.1966748847338606332, 0.49285786194048575304, 0.7745504664562467001, 0.30661719813705368143, -0.081442305406004431023, 0.028947894142609155443, 0.19966135201684814815, 0.11348388898694378168, -0.21564805942055620402, -0.68466392410137244529, -0.36952616114789527124, 0.98854283792339048276, -0.11367941258605253341, 0.98957138471159178383, -0.36758238481522020891, -0.68416000899080919861, -0.22249201333439561079, 0.10440222071401843162, 0.24240733715392626113, 0.070190706640823288809, -0.011097343255362797712, -0.15580187062797581921, -0.2646197891720274975, -0.037982199142248061763, 1.1210188551741548935, 0.85632681962862322589, -3.1516459033394346179, 0.85307865796194348729, 1.1139864389556479729, -0.048114192568602484612, -0.2735802452433958587, -0.14364443882943372932, 0.069100335353210376454, -0.043352090692928536841, 0.17955068402739160494, 0.18264011009541861186, -0.095794450660787233698, -0.55109570069410718229, -0.46312292096085355686, 0.746886480775899253, 0.098448597503282803123, 0.74913516013428127494, -0.45949489131238302475, -0.548178548978617175, -0.0988851355658322273, 0.1674017145939502571, 0.14652064613256471359, -0.014730737396071818396, -0.014317737155779284089, 0.020051409418637585608, 0.11389447758353993256, 0.031889023308547384539, -0.40549043724833322422, 0.13716996690128493985, 0.26097705616038135989, 0.13652028261858167868, -0.40491681669478873928, 0.035736163052810666907, 0.11950870195579234301, 0.019913474794392975464, -0.036689213481888062474, 0.011976886898948940416, 0.066175650089053253411, -0.0092582532324493177489, -0.24955186918991284495, -0.25279759128764145437, 0.7339248320359973432, 0.82331050930174298763, -2.2458591138531152076, 0.82544012176827286087, 0.73555510175539851048, -0.25650958174248295052, -0.25976517105352359138, -0.0089666547940903829766, 0.087706171916410072265, -0.0088807108076432422078, -0.033713204355406076185, 0.025653150069291079383, 0.20790211271440570129, 0.043182048911982989781, -0.70551155512062435626, 0.11804737694580472174, 0.70678228499303252796, 0.11930924609075861664, -0.70461412502574183314, 0.043140481446196442261, 0.20710526823907937644, 0.021862574751770183501, -0.039665199290980800162, 0.019954822076626337751, -0.027627861875683293325, -0.10777036903162294512, -0.17453894039393982873, 0.058526763024121186696, 0.94197446009117069554, 0.60918215081935522814, -2.6461988520279882131, 0.61071474328641095486, 0.9437984288949811118, 0.057673232879014886543, -0.17986228895511685577, -0.10561365297370141347, -0.0051533781840440145289, -0.0040850851671124363204, -0.10688955595691145006, -0.051672987199197710495, 0.27022172328681537401, 0.29298766260944697493, -0.55662121223753979748, -0.058956212984749151174, 0.42328457098767718891, -0.058381132074465325044, -0.55564909509213789018, 0.29576959650927564516, 0.27429171824776110711, -0.054275005923071162339, -0.11449413703032704226, 0.05156162812182785854, -0.023105642519599019213, -0.11594386261707056995, -0.21075362444302014886, -0.079410353263042812721, 1.0463795508252948796, 0.88488014323808950046, -3.1053091034219573174, 0.88210156729683442123, 1.0404676220510766793, -0.087336782697791964791, -0.2154002718173492481, -0.10096312055628289817, 0.029483399143364744099, -0.030199482513892050156, 0.048360952325803618912, 0.02293626369314049826, 0.0081493849129142831994, 0.05771051088430236703, -0.34794215871728323419, -0.026926972620177593715, 0.5318065070159059271, -0.026283669445825035699, -0.34337436703215357126, 0.067486346342522401942, 0.01395900558014247575, 0.0099904261965409902474, 0.015452783008272413454, 0.029109589378780765384, 0.048020249358765719483, 0.064327124770087229511, 0.011759564568165407211, -0.16225135628572293078, -0.41444534187342957354, -0.15948815756635459362, 0.78093658254237363625, -0.41770357280035425029, 0.7794811467642418501, -0.1629147754983056029, -0.41313085431968749051, -0.13243813901767728658, 0.1445509840717947303, -0.0016796958613174255148, 0.017289356625271489615, 0.029498360638596506783, -0.010023879554130084996, -0.15092373833161210062, -0.2421421439450385793, 0.56620082660990300738, 0.85776524947663157494, -2.1291613464517413767, 0.85951601542712496862, 0.56084561243318764578, -0.26081316198932852357, -0.16072538780927411728, 0.063675374299429446379, 0.0076968954772448500146, -0.021028437861118996072, -0.018535194176444051706, 0.018233832371537424255, 0.0943096805783314035, 0.045843446206561819234, -0.32803247415534358966, 0.087139410474448120159, 0.22641563892958471715, 0.084996693981496895809, -0.32518734827680201294, 0.062913302244834296006, 0.10152090917658455893, -0.036126328662424128657, 0.027875737553719727446, -0.13402059957531758938, -0.22168178722103210054, -0.15892020994166444758, 0.12100261063092437874, 0.6464859641999513018, 1.6306650533079178889, 0.38314920975444938245, -4.6014169156420399176, 0.41871842729903308022, 1.6809586748035254544, 0.66256331054818606852, 0.020654069573102402935, -0.4787191987548646277, 0.027125598793284882992, 0.0054635106162931393325, -0.046983980328528712755, -0.10710801316600246369, 0.032534334712458264083, 0.3038637259147076719, -0.15742187848334540656, -0.14609915157361405558, 0.18597106252090722656, -0.15282164425012895181, -0.16532574945123690302, 0.30248620410397492053, 0.022115667212531574248, -0.10487744263038739634, 0.0095432916950020599961, -0.0085048311108328021601, -0.023465813030780720283, -0.011063062275499357945, 0.089506943233238822621, 0.065915440297376948209, -0.3080381106896460941, 0.15615000534262390985, 0.063502471227400339449, 0.15131417756271881014, -0.31194416144353159792, 0.074939156770164910837, 0.09815355449647114594, -0.04593363410203769176, -0.0059109775311766733774, 0.013642983058333136084, 0.0196890995438959715, -0.006286929739461897225, -0.080252915639663191527, -0.15403060941186344501, 0.30445569875316480513, 0.6495934729628841664, -1.4777217361179475485, 0.65072834684874680811, 0.29870021939490404295, -0.16804094110196110545, -0.085284895599649385156, 0.040652880309846463336, -0.056388353750703286427, 0.022197567626202473201, 0.11227571888446565007, 0.11840037851772988731, -0.081834724987137771879, -0.42041012692177742727, -0.036183566131012478984, 0.82276817019112180596, -0.95753082688684920676, 0.82491614128083823321, -0.040428613689655001906, -0.43554393866556184722, -0.084161771092302467556, 0.2149150914885296082, -0.051657606363381430858, 0.0534252638191260662, 0.13747390059387851191, 0.10349570941772801014, -0.10567087093369714745, -0.43418555660120944362, -0.36007746780053989655, 0.66887138284359082263, -0.015271564668272863755, 0.6630198538732375857, -0.37611228534803137213, -0.45450346125449075929, -0.092929658275725857131, 0.26685220909540519063, 0.046920811797515754593, -0.062186247775802311333, -0.13814343818799976193, -0.15064760385578096491, -0.044185046865583790798, 0.37583742168962674945, 1.2585789665578515084, 0.42575328384575805707, -3.4238867087798414701, 0.43909508466748475852, 1.274671653403737448, 0.37530907370488802988, -0.099782012230306291189, -0.27953376600863194534, 0.064848002852072858526, -0.090910813697725481597, -0.19267409189174050654, -0.088824119689659891685, 0.37022288820144166888, 0.53165873168428978701, -0.36739874122151372493, -0.20445057006255826226, -0.049726798933024479732, -0.20484570333779639206, -0.35045469489401703811, 0.57875470522339900192, 0.37352750320802713402, -0.37070924846187447432, -0.014869210543127642579, -0.045298955152710602967, 0.010092499006801539801, 0.31072095042749076077, 0.75289827196102276918, 0.48007463209962031314, -2.193899125582474241, -11.681778218857255425, 24.757042509975292433, -11.723974254512016913, -2.1986340092105955968, 0.6008158593659644886, 0.90568762580046302535, 0.048051827566640455125, -0.073526299522322674518, 0.11744631851970040171, 0.21105753035663563089, 0.10078158129356985795, -0.22582052885443737456, -0.68816335536221850955, -0.36952616122526282805, 0.9895713846778176892, -0.11316664720129551713, 0.98289687916886758767, -0.3982583921550197914, -0.7370638422014698099, -0.20376777734229020722, 0.41010130215566686918, 0.040088434277709410203, 0.031434146053172656865, -0.027110589398807435901, -0.15311103494215916321, -0.25482183236966204909, -0.031673854418043233261, 1.121018855036191475, 0.85307865814461170118, -3.1531080800903557027, 0.8555767797325339874, 1.1122112235898919952, -0.065797784724159050218, -0.28425294628982528122, -0.04680659805835041265, -0.10102377701775094865, 0.064355193550897207833, 0.19950478759456027888, 0.17298832976115927074, -0.10654776624310316469, -0.55600689509975642633, -0.46312292083364353656, 0.74913516017597503449, 0.099812213779816777448, 0.7437278034759382539, -0.48323737708769914923, -0.58914930637333784436, -0.10106463305898012728, 0.37660526049585341113, 0.0028113282549418455031, -0.016503176289878681432, -0.014563168268489039628, 0.019063525923219484803, 0.11313751679522801585, 0.031896811403614701297, -0.40549043771743242814, 0.13652028265403143292, 0.26021764006523467883, 0.13742519967696578687, -0.39974450420179025167, 0.04840424427660261486, 0.11990392464057397781, -0.032911881227834255403, -0.021269794735588792856, 0.037690285019032411518, 0.068374579640639129852, -0.013275958825174981326, -0.25200557775436344121, -0.25373580327522754496, 0.73392483208499048608, 0.82544012164942937027, -2.2409560093930216418, 0.82790246486532237391, 0.72230870721410478374, -0.28856947879419353464, -0.27441158203674897997, 0.12843692604956741254, -0.00079167877456706075082, -0.017512774339189039752, -0.025155220265330124896, 0.030633246726538269911, 0.20670271171553128431, 0.040781006199860830863, -0.70551155468699811024, 0.11930924639209672389, 0.70524704332406740193, 0.11381804435476236548, -0.70680346045552777667, 0.059811552290085515105, 0.22825367210625610581, -0.047820343855222421925, 0.0268318404695863863, -0.00029031417070066352359, -0.038103752582217646772, -0.11038449323703142246, -0.17311668414940620253, 0.059681967189465327805, 0.94197446076190849418, 0.61071474336192432819, -2.6397863828808105779, 0.62084705463027678451, 0.94759958832288060382, 0.04439555013776216863, -0.2085794593092364646, -0.085190381351845312352, 0.037044212869557405221, -0.04790350008270295723, -0.10911402697395472527, -0.046068576561173982253, 0.27214187357523555999, 0.29242438285133270925, -0.55662121222299976164, -0.058381132304404878552, 0.42280331323830949719, -0.057363826381148438904, -0.54333499505435944865, 0.3275184689103807778, 0.28257862309073461526, -0.21736615078877560836, 0.041625597370086170335, 0.011064709619542754601, -0.037916525587740126935, -0.11369834274901588489, -0.20223708097039208575, -0.073885726355989875258, 1.0463795505294741783, 0.88210156713008158746, -3.1062068801279001384, 0.88543298941082360098, 1.0420993948787911076, -0.095566323532151414843, -0.22229576302346837346, -0.059967181599602792375, -0.028724492075047642486, 0.012288179043374774027, 0.047312495855341904494, 0.013108967939836192393, 0.0024255830857695731204, 0.056320832656327389498, -0.3479421586289351831, -0.026283669401723313425, 0.53527098840592834961, -0.018957176673228565839, -0.3391212903012227331, 0.05632588209459962747, -0.018209584398119996185, 0.057051566088462318804, 0.013375579133230121667, 0.021769507587450064728, 0.042482585315976345575, 0.0635146268340449488, 0.013278360938581314568, -0.16225135602815513036, -0.4179644675961175504, -0.16468336540204231433, 0.77948114740237228748, -0.41372367506308976859, 0.77956255071606883611, -0.18824519456551011554, -0.41719689307715113191, 0.042696607183024003207, -0.0033846853098316279258, 0.0034763452864283012587, 0.016960589747359165746, 0.028423443856765220894, -0.010326175307239896656, -0.15092373883699841675, -0.24289042256116064045, 0.56509878568609683835, 0.85951601488729667544, -2.1234979850243260735, 0.86174153618041127167, 0.5383227393481192502, -0.29671960026638416474, -0.047312267242444805415, 0.0077980004549702960567, -0.003665852798540697368, -0.01995191694954663586, -0.01634925086028059249, 0.018832980762818182785, 0.094309679913924637784, 0.045842447659325040454, -0.3289525180408041094, 0.084996693109030158064, 0.22665696892057296541, 0.09431107375089617384, -0.31844848921080082249, 0.063672997944354270405, 0.051867435291205257752, 0.080384181399245205335, -0.044399295067628179179, -0.16174051337767203496, -0.22568480221295164623, -0.15712424544136921467, 0.12100261072049545152, 0.64773443089358950786, 1.643628475466771377, 0.41871842694268446294, -4.5451388066221234396, 0.47459033498369501691, 1.6815638265574781585, 0.49836138570529686653, -0.41717335026320895297, 0.0021111309631084386115, 0.020564964111448758022, 0.0077316556219433202335, -0.043374530501005537086, -0.10538046954450266957, 0.032534334493890225415, 0.30243563387029881451, -0.16128124109976335609, -0.15282164445436763334, 0.18065868314304592279, -0.14288027777829021003, -0.13164168900547018604, 0.35776971217916248014, -0.16652490469171074561, -0.00028242555469721176802, 0.0068380794649114100964, -0.0054702375649304567709, -0.022182333010381832805, -0.011543835540342750265, 0.089506943038747233121, 0.066875610281425579284, -0.30856344525410300816, 0.15131417725929346974, 0.057272809827539225558, 0.15645543306992315546, -0.29408659753512433888, 0.10233941101978207, 0.011946371251951002496, -0.0075924384372952104238, 0.0039877371454557029168, 0.013999286949681595191, 0.01772909998650934113, -0.0073471968043976937021, -0.080252915962317525334, -0.15358552125589799986, 0.30521858120177142171, 0.65072834671373780413, -1.4781397279503423192, 0.64793604308543950587, 0.2868849393806804926, -0.17952676946618101317, -0.020807057353075308903, -0.030844970336955327556, -0.023200771369225322432, 0.027824653497991513884, 0.10905901518688077001, 0.11568063136328683715, -0.081834724798835523085, -0.41871477509190957456, -0.033990279824413952015, 0.82491614139670810246, -0.96144572040006115454, 0.79985066638061919075, -0.10121232079233907775, -0.45047572728133855602, 0.22315601425825526416, -0.044867447585919005559, -0.0052698021330873610901, 0.063899475982694103893, 0.13465818619747307872, 0.10011788651781083903, -0.10567087088775523052, -0.43238740150607340995, -0.36030357576608540437, 0.6630198540063789725, -0.032182244891102045059, 0.628713184386571311, -0.42482521426329705738, -0.41739558491036587995, 0.22818126265784877127, 0.048155063148046593768, -0.0036615960522601923022, -0.075434411545322405113, -0.13464103572682933896, -0.14522748841263066755, -0.044185046755734314061, 0.37150760452209358187, 1.2579294532364897474, 0.43909508526324542155, -3.3912362587036581907, 0.48812758864737226805, 1.2958855185168287605, 0.27192614285173222211, -0.37170184704229353168, 0.052540856386253900756, -0.00025158232578830971859, -0.097415118753168133958, -0.18216861129455980128, -0.081930251339430895663, 0.37022288838588246485, 0.52780376385337213208, -0.37189271878268392868, -0.20484570281078429144, -0.03421222413477011115, -0.15595314593569067907, -0.27923808853542442243, 0.585070402168999526, -0.12051904487898074114, -0.037082292348434145457, -0.0007330963194381984703, -0.015059133381601249452, 0.021199824360439101234, 0.30751104564640258099, 0.7528982716920863405, 0.48768423221033990744, -2.1966748848650547998, -11.723974254435619358, 24.675382199595425448, -11.756155869570532602, -2.0562882859613926811, 0.81020607764849694732, 0.72993004746490497414, -0.073211401714588356104, 0.010024095937942175544, 0.12920229968747318217, 0.20228502947624477093, 0.094179223689463270608, -0.2258205290766035489, -0.68466392381953777857, -0.36758238482193400509, 0.98289687907718314985, -0.14319611615954841444, 0.91624666271357513292, -0.46580013303989964424, -0.6424978069752858012, 0.25913187775712631655, 0.013634473256398878943, 0.026941905003220828485, 0.025130029542430255135, -0.024207469131665562689, -0.14825109985580881, -0.25482183256347873401, -0.037982199135640916376, 1.113986439045442367, 0.85557677965215528371, -3.137727038670369506, 0.87462698402495075989, 1.1153916605106950488, -0.09553036289479903409, -0.32901464607515962246, -0.075438381940330631603, -0.022955145575266459185, 0.081205161682841406345, 0.19364410105509710647, 0.16634903474498216602, -0.10654776599418375183, -0.55109570111356942501, -0.4594948911659909041, 0.74372780338332722394, 0.073808890737426829931, 0.68544044568233042369, -0.56025148762069698627, -0.53717930030773652916, 0.36536398265289948828, 0.0080426586914506981857, -0.0058479156788298928757, -0.018070585930122871204, -0.01391343460457891007, 0.019649860488291235466, 0.11313751644293022769, 0.031889023910500288728, -0.40491681658254463638, 0.13742519956487153632, 0.26120068382362382486, 0.14056433597890427567, -0.39929432293278466215, 0.061526343409257562567, 0.069510474130281427185, -0.020795617373198755501, 0.0050143623703575654349, 0.039492146215702905732, 0.063665889565255789906, -0.015848287480326295196, -0.25200557748220975274, -0.25279759129364620662, 0.73555510189626582829, 0.82790246487436203182, -2.2419555402041768666, 0.81765106652109476038, 0.68819565545984240362, -0.32247313465620153217, -0.073927702466570585638, -0.0034681140871299629726, -0.00063985714732500351132, -0.013198253730830215857, -0.023978192344739586439, 0.029244782336008398604, 0.20670271232830075836, 0.043182048147191814946, -0.70461412545125812201, 0.1138180447337161505, 0.69503584022206732485, 0.11196799665006024194, -0.68830326109373274335, 0.11048381588661948527, 0.12497722218338376543, 0.024064015952897664652, 0.0098655048742883316021, -0.010974283302174116933, -0.041395126072764666658, -0.10984147390302040348, -0.1731166835087750433, 0.058526762335971739548, 0.94379842850225459383, 0.6208470548989444282, -2.6221070556633114812, 0.63931288743548297937, 0.94103435121409162445, -0.014383403480114543074, -0.26539170795834676753, 0.03295635066826777676, 0.0021455566306763894369, -0.053896034190168785882, -0.10671759277844675151, -0.044801515401438327113, 0.27214187364862441054, 0.29298766263966258272, -0.55564909519516936243, -0.057363826393482114652, 0.42819543711204571812, -0.036635480799938134611, -0.5046148969127257633, 0.35908205692488492922, -0.024564348132044774159, 0.020993549521152923137, 0.020268007182612013395, 0.0042626672986240763535, -0.034959545096964521338, -0.10919826197923104827, -0.20223708095096767923, -0.079410353201467984152, 1.0404676223621089903, 0.88543298916371626639, -3.088736128969626904, 0.91075775160280225684, 1.0466518301656200229, -0.15079906792223693457, -0.26473577558914257457, -0.0064204228918681134977, -0.015509118448012920721, 0.0085656271026868478768, 0.042771489901393289212, 0.011519276410236260061, 0.0024255829684251056441, 0.057710510711136454864, -0.34337436708097468507, -0.018957176814036410389, 0.53637069027070127802, -0.037241677026104237636, -0.37540667367483904204, 0.08456966521010680915, 0.052974095703296829318, -0.0010195335694577255334, 0.010562769092801446713, 0.024630522153357085891, 0.044991840097491103689, 0.063514625497236285034, 0.011759563760362481966, -0.1629945588252825206, -0.41639785213846708389, -0.16291477560091663368, 0.77956255094215487489, -0.42242229690706301071, 0.77439575915604619283, -0.20008885492571307219, -0.35237284010208758112, -0.002299348723657884154, -0.0012536073031431502374, 0.0044478899932073332271, 0.016882697540060744718, 0.028423444478358629584, -0.01002387946728087488, -0.15164012323593811793, -0.24595874481948834522, 0.5608456123109621938, 0.86174153705466816078, -2.1137339445890344081, 0.8834579134883238094, 0.53004554250845725694, -0.36185240888463304776, 0.0024683493376405425852, 0.0048125514892345426163, -0.0037275782819861714469, -0.019533212621554447985, -0.016349249881349246782, 0.018233832384173254137, 0.0933598532442001805, 0.046040635773141748432, -0.32518734826161649343, 0.094311074879550815808, 0.22577662386439115161, 0.06935714977311384366, -0.35783977306215319913, 0.1689852548281015332, 0.074692189768282443718, 0.014812757593034273779, -0.069441810164236020264, -0.16641031702384062596, -0.22568480178405633807, -0.15892021012850940376, 0.12131285756869129855, 0.66183707642643907487, 1.6809586749125349225, 0.47459033517491977516, -4.5063129042967462468, 0.46266285521262218161, 1.5472547339055968685, 0.11895803683679553875, -0.01000675301343754986, 0.012884873131229362259, 0.023558401435314350186, 0.0080181280733037998965, -0.043374530194341086309, -0.10710801304463567207, 0.026401190089572881825, 0.29361015459481215606, -0.16532574942991304368, -0.14288027745744411035, 0.20424137869463643291, -0.091617486722722699111, -0.12117471981952580851, 0.11493684094899425663, -0.0046928163520768163136, 0.0057162632152173414174, 0.0072446167922385852275, -0.0063451877528677390339, -0.022182332723682876563, -0.011063062043140101109, 0.088545282475655170962, 0.063492958290064058224, -0.31194416154479326364, 0.15645543316861912353, 0.068721431723275078207, 0.16307185254615014225, -0.34296342879610103305, 0.14632806362343880879, -0.0033318661186295013554, -0.0031793820347593006008, 0.0038679853933681800429, 0.013135321288043923932, 0.017729100328974810308, -0.0062869296097855338276, -0.079300788213049797082, -0.15536792063207185977, 0.29870021936379487215, 0.6479360432266478842, -1.4693504891007891366, 0.66982485241201861204, 0.27342517306033514712, -0.20849048888781479549, -0.0067762777491010795991, -0.026334397258449313006, -0.020230447463383632173, 0.028197623636652121432, 0.10905901512923293906, 0.11840037878262198168, -0.07566144858267274953, -0.41351210103707702892, -0.040428613519105732754, 0.79985066587423536699, -1.0046531793440123259, 0.77870256356180700408, -0.05719508000868225861, -0.19595221643610091777, -0.022895580267018052212, -0.024503094509657520617, 0.0020077644634419683847, 0.064227614050442582982, 0.13465818602503185075, 0.1034957097506994661, -0.10040255216027510476, -0.43225163203333821782, -0.37611228532706031391, 0.62871318428316780302, -0.07352927759609365066, 0.63289046931939585416, -0.34866029982468449422, -0.19901921888139895156, 0.028147275748025883668, 0.021445370185457540008, -0.011330920878824618514, -0.074461806527032742276, -0.13464103615462269525, -0.15064760354603554116, -0.052688735335298192419, 0.36883464289846601414, 1.2746716528741872665, 0.4881275882017094836, -3.3354122652574909047, 0.50813423548413294561, 1.2289086935357080854, -0.14294698258305707506, 0.019959370706233603432, 0.031612132392534034175, -0.0037676051945441102475, -0.094791435560631256974, -0.18216861148634566625, -0.088824119479140234423, 0.3594636918140376447, 0.52572685592869117244, -0.35045469518520938612, -0.15595314627641759797, 0.016239534294697627326, -0.16951478631193062085, -0.49728148171451391768, 0.60276793566912267863, -0.040998339605464520496, -0.00035994456752463802399, 0.011776942906214806633, -0.0138556527242397233, 0.021199824171336457967, 0.31072095065859084428, 0.75848847696245802474, 0.49285786201120274086, -2.1986340092579710337, -11.756155869445205298, 24.555162060850264538, -11.924537484000593324, -1.928203921566873591, 1.6947413526133303652, -0.037194314341261303092, -0.037294019735799557869, 0.019158852916264366983, 0.12785593811756629234, 0.20228502947835963477, 0.10078158142876417203, -0.21564805952555052282, -0.68416000908814955661, -0.39825839205671331733, 0.91624666287509248797, -0.19493925432686051713, 0.97908561150011630403, -0.35051869556555281804, -0.44586374562074732042, -0.003583315964320626934, 0.015216698257818328174, 0.030261778615695480305, 0.027955412626776067475, -0.024207468789792904035, -0.15311103496059910722, -0.26461978908810762645, -0.04811419269791812614, 1.1122112234086476423, 0.87462698448039144505, -3.0899346300532095633, 0.96132349395601701669, 1.1137878776538583381, -0.54885772566217727597, -0.030695751331275552415, -0.046733128535837553663, -0.014415025006605096392, 0.08022964818652246477, 0.19364410117246039222, 0.17298832983602019375, -0.095794450442826931114, -0.54817854893614703649, -0.48323737689091683656, 0.68544044519355362688, 0.0032809818265939396514, 0.68744931118586083407, -0.43131737779193951576, -0.18688725446273207753, 0.0060223210978762658704, 0.0014186342875953930273, -0.0072639868044796270657, -0.017562200404537792187, -0.013913434448094957538, 0.019063525890500841037, 0.11389447741836679351, 0.035736162635564124035, -0.39974450420024942865, 0.14056433647902888406, 0.25285178244976769246, 0.12163819364418758384, -0.44589917323490940726, 0.1936926401811857601, -0.0083637527127802795829, -0.010500180400214848692, 0.0057238083388399599305, 0.037853421126525163554, 0.063665889356568236557, -0.013275958988493749557, -0.24955186910998120031, -0.25650958190907108269, 0.72230870753890819014, 0.81765106597673353761, -2.2396815827137119825, 0.8535352505526264677, 0.76393580721058551575, -0.48914140654719395762, -0.0034770754260596047416, 0.0015050360375448608441, -0.00070253697601505710191, -0.01450880377776989083, -0.023978192633120017085, 0.030633246456591239903, 0.2079021132541022665, 0.043140481616882546401, -0.70680346034133412214, 0.11196799614033334236, 0.69590059861068598313, 0.10506228476555878881, -0.78266523406945454067, 0.33570264809089866986, 0.016076341966288695828, 0.011263689565492254058, 0.0034934169306286046164, -0.012584351803103813183, -0.041395126338135912469, -0.11038449359740443168, -0.17453893999910330415, 0.057673233152832736881, 0.94759958852196557455, 0.6393128868483360927, -2.5940994330524693723, 0.67597227734736065052, 0.94398416569023346767, -0.35590571379259899887, 0.016742566930242508416, 0.018497519211755790447, -0.0035233847618148551872, -0.05498130806898492845, -0.10671759302570529848, -0.046068576619217663226, 0.27022172320582027449, 0.29576959679057285291, -0.54333499488815917289, -0.036635481184727002046, 0.44726651799906380624, -0.043824355821633714791, -0.6553001865674030757, 0.44857915534172909666, 0.0024231396755375685277, 0.016922249046324035204, 0.021522815603111044441, 0.0066939458989679970213, -0.034959544940865033835, -0.1136983428924619316, -0.2107536242684217298, -0.087336782920265795815, 1.0420993949273920087, 0.91075775157253102687, -3.0405790711294526929, 0.96381181331162668613, 1.0049721694628159785, -0.47810066655296573623, 0.0079505887122727613575, -0.014823918329484101625, -0.018188817560324806871, 0.0080841713691857126217, 0.042771490061514295944, 0.013108967936617766867, 0.0081493850377070149033, 0.067486346461757829029, -0.3391212903143552837, -0.037241676876239145078, 0.48994603215287874098, -0.04113203790573590618, -0.28443618958282085307, 0.095990174867492700228, -0.011648522826126666771, 0.0038650653241822400738, 0.014683336047607263575, 0.024630519380869079354, 0.042482585405428256731, 0.064327124388755388407, 0.021617729981249118876, -0.14834107079061689682, -0.41313085444865998985, -0.1882451944315586545, 0.77439575893658485484, -0.35637326730676499587, 0.86877866042158480298, -0.69739281654044238401, -0.0013164795516118659498, -0.0014715874183838825555, -0.00070380341601501846503, 0.0044478900315407217739, 0.016960589267310182543, 0.029498360578850420233, -0.0079208298677848256591, -0.15376528460482680183, -0.26081316140367760692, 0.53832273828615262445, 0.88345791377309479397, -2.0347952413541721128, 1.1202773207012519663, -0.12917147205774653917, -0.00061197150296614984816, 0.0037551207775511795048, 0.0045980557356574197278, -0.0037275779345115169286, -0.019951918118641148031, -0.018535193925363104006, 0.015834372332025618141, 0.097184804359585805211, 0.062913302856536401642, -0.31844849068300518091, 0.069357150347539073265, 0.18198279605170447626, -0.013497199090955076459, -0.0619272717841580847, 0.042657734661590063219, 0.036892921804256684215, 0.0011764843785771954103, -0.069441809997470918536, -0.16174051364205821457, -0.22168178724873377483, -0.1564832867714328335, 0.12237234551102002533, 0.66256331074782259805, 1.681563826268202666, 0.46266285525873768192, -4.6103613289979223921, 0.61735853001434271015, 1.6205429510843598617, -0.0091043460018714623266, 0.0012266018144767919067, 0.015080566931347284126, 0.023558401667489698578, 0.0077316553875064666229, -0.046983980508710637691, -0.11711918686051965566, 0.015323479585602862302, 0.30248620410753251919, -0.13164168926686400929, -0.091617486674121423329, 0.19800080041206813419, -0.41051453129469983594, 0.24631821749004009781, -0.0027926700395612155496, -0.00041758495566290686077, 0.0053619016417802119537, 0.0072446174175931016617, -0.0054702380255214882315, -0.023465813197193288037, -0.016578535156940969048, 0.084329581923865706838, 0.074939157315701859785, -0.29408659777385198764, 0.16307185250450426106, 0.023080825491175675779, 0.031441418025217222898, -0.047493370122509051545, 0.00039241455463166898887, -0.00380008415412769196, -0.0034379765439649292297, 0.0038679859006972724077, 0.013999286450097982865, 0.01968909966161198552, -0.0051240198646264282362, -0.084444726476681553073, -0.16804094097554150844, 0.28688493911273382775, 0.66982485251459622511, -1.4207371298717286212, 0.80106536551255802525, -0.10884068427455824424, 0.0048813536978725207902, -0.012492361497238426218, -0.02632869896246401234, -0.020230446936234977584, 0.027824653870917571974, 0.11227571882078335452, 0.12789931615124236486, -0.070173738062042911867, -0.43554393908657129675, -0.10121232015416274652, 0.77870256350767530495, -0.89031546578893172583, 1.0615781116420535124, -0.56448374205972284834, -0.0047591856286282388511, -0.019480549475589033215, -0.022178424585323975454, 0.0020077647292299333724, 0.063899476606684144087, 0.13747390032455183984, 0.10999135318056942057, -0.10019877620857021994, -0.45450346138002084562, -0.42482521424320840442, 0.63289046945103821784, 0.060568948193729973961, 0.77538390502148579753, -0.76751409061440500903, 0.0080314716607494240214, 0.022230568025896416118, 0.019753336331506932827, -0.011330920766418610401, -0.075434410989104472289, -0.13814343804390327719, -0.15632250232188507622, -0.057833435333414229262, 0.37530907362896170909, 1.29588551901084581, 0.50813423535539237186, -3.355942295363425032, 0.59321272191711238442, 0.99024802421181223888, -0.0014169010069428559367, 0.02470300766194577155, 0.031826472302905139955, -0.0037676050670746655258, -0.097415118230248995412, -0.19267409184269859712, -0.10371656513865504246, 0.3635433915377210834, 0.57875470506176707186, -0.27923808807281563782, -0.1695147863446125891, -0.24816905251935089671, -0.52572462465411151644, 0.63010484515984721554, -0.02165758125899653469, -0.016840998607957492761, 0.0024885313123345501399, 0.011776942712886409939, -0.015059133248392662929, 0.010092499010173229881, 0.29325094599538270579, 0.77455046654392001315, 0.60081585918591151607, -2.0562882859245315004, -11.924537484030588885, 23.711679718753231327, -12.177273525174737046, 0.76788033835058078225, -0.0073819542356571782662, -0.033782149941702824258, -0.03421430233835116852, 0.019158852845847770818, 0.12920229961023801946, 0.21105753040081318139, 0.11348388886199406445, -0.22249201308590316195, -0.73706384229058263813, -0.46580013318303858938, 0.9790856115607251553, 0.017248055173143893304, 1.174224132900814066, -1.1576706956241720192, -0.011182179303788414987, 0.0063070893916309543267, 0.02032547554040372087, 0.030261778478097238476, 0.025130029450518412415, -0.027110589475192015119, -0.15580187060036604363, -0.27358024507521300217, -0.065797784698551464722, 1.1153916600404900628, 0.96132349406635464462, -2.9898268837053465496, 1.0904262107262556736, 0.28540736634549662698, 0.0014738910147603268182, -0.032070394938983851074, -0.045893854013547792647, -0.014415024742852165437, 0.081205161795582986306, 0.19950478756516859535, 0.18264010991623694236, -0.09888513563804729456, -0.58914930660580888677, -0.5602514872534978263, 0.68744931124996111471, 0.19362333611720766724, 0.924419113926642666, -0.94705452157061520424, 0.0020096413751493984699, 0.0034709491410976093737, 0.00094532424737964507958, -0.0072639869480741704397, -0.018070586249100784731, -0.014563167858505391902, 0.020051409142032610572, 0.11950870262077600836, 0.048404244313745140027, -0.39929432378900392431, 0.12163819425628540072, 0.20429067945697029773, 0.024315393124324454627, -0.10697725912459625508, 0.0012073476405888810929, -0.0096669954488905647516, -0.010780623119217134023, 0.0057238086683072640637, 0.039492146300976749274, 0.068374580056949046947, -0.0092582537125558000968, -0.25976517104660562518, -0.28856947882220784773, 0.68819565574633601024, 0.85353525056878931654, -2.0729178428625751884, 1.1176844588252718982, -0.12077755850319107922, 3.8755152194281609207e-05, -0.0015130575538167085822, 8.9790390931312677458e-05, -0.00070253684864211504713, -0.013198253327925311973, -0.025155220267956003577, 0.025653149851556322075, 0.20710526822887065368, 0.059811551951476062128, -0.68830326043458189833, 0.10506228455010002687, 0.59039290138276134812, -0.070254192793814243534, -0.19319414289574199373, 0.0057754210475700219068, 0.0094022182473198438052, 0.0090368197795186854665, 0.0034934169855963641271, -0.01097428301800927597, -0.038103752800607931794, -0.10777036893997340994, -0.17986228941633497502, 0.044395549913328405389, 0.94103435225324916225, 0.67597227669251880311, -2.5209903641082211934, 0.79447206015428772918, 0.38667780666833762027, 0.0038712183630270685306, 0.014083625022055774531, 0.015633196192097215116, -0.0035233846267174358247, -0.053896034204861324735, -0.10911402686530732864, -0.051672987063276165642, 0.27429171786590050841, 0.32751846895867831044, -0.50461489649956825509, -0.043824356117589169379, 0.26783767764802546285, -0.339619296793784331, 0.20681991903193752691, -0.0081172028976803818623, 0.009322951151391846436, 0.020555944091692097964, 0.02152281581266212579, 0.0042626670105719377882, -0.037916525469362784284, -0.11594386261389426351, -0.21540027179717266548, -0.095566323357613183753, 1.0466518300565825772, 0.96381181336453569664, -2.9829426922200106276, 1.1541169804825213152, 0.24653778062408202021, 0.008711893598472493852, -0.0031760077046361695138, -0.016567865294109210272, -0.018188817894812914222, 0.0085656272666232886881, 0.047312495680846548829, 0.022936263763613427835, 0.013959005628528966744, 0.056325881952608923553, -0.37540667364873187006, -0.041132038118503071178, 0.5701231318725658026, -0.16437431597743448752, -0.11214945205918203097, -0.014773893037660149052, -0.0046935153090059585079, 0.0038650599452153613922 };
	std::vector<double> l4 = { 0.010562770842879902131 , 0.021769509998438314052, 0.048020251131247852772, 0.08292201420415257207, 0.049709152430895998942, -0.13243813895931952351, -0.41719689317735159095, -0.20008885502589981997, 0.86877866071384479696, -0.35170556900630428609, 0.056855794259365502463, -0.0004973641379818123133, -0.0010348369904131723973, -0.0014715879313144028789, -0.0012536067310002971774, 0.0034763448864323595079, 0.017289357217748914375, 0.033171299595587823428, -0.0020078249000918956153, -0.16072538780493156318, -0.29671960018858539732, 0.53004554283975346962, 1.1202773202347633408, -2.2598296805808861798, 1.0249814844536302427, -0.0010676610548363185595, 0.00062051819904710911469, 0.0037551201475119369393, 0.0048125520178491044618, -0.0036658528358468039227, -0.021028437460317007252, -0.018922775206247060414, 0.021529090763510585343, 0.10152090916819557498, 0.063672998128335434043, -0.35783977273165618049, -0.013497199608734727211, 0.48411692939824146897, -0.2657377441347111291, 0.0064297335531752373636, 0.029880423592165056279, 0.036892921862841494829, 0.014812757752725336235, -0.044399295619171032234, -0.13402059914809474894, -0.21160369864569011922, -0.1912031577432153906, 0.020654069348429442143, 0.49836138558653170216, 1.5472547343958589217, 0.61735852982307559689, -4.7652241671101744913, 2.5594840130448677229, -0.0022986775265884133376, -0.0042255886457441532017, 0.0012266018295037501001, 0.012884873340642272882, 0.020564963790183909964, 0.005463510890109556789, -0.04638917470821796174, -0.11335182289433666514, 0.022115667433485745874, 0.35776971200816620744, -0.12117471951431364741, -0.41051453153530764739, 0.35340190803041104939, -0.075435469450853609441, 0.00044417321247477992152, -0.002480180633139791261, -0.00041758441529170248077, 0.0057162627162185017729, 0.0068380796011647849986, -0.0085048310044801896795, -0.027783724353497765008, -0.013900917273322975845, 0.098153554075713786431, 0.10233941112545871799, -0.34296342861608647423, 0.03144141785883328849, 0.33711736298266559153, -0.18719463133828689871, 0.0016019243730701179969, -0.0012121439761859809526, -0.0038000837870826482956, -0.0031793822991739413911, 0.0039877372235028307051, 0.013642982936205871186, 0.017940274751237477585, -0.0053747030956147103659, -0.085284895638283245156, -0.17952676940181960807, 0.2734251731209878522, 0.8010653654488091302, -1.4900386056759564735, 0.65941889033234313278, 0.0075709855625703528434, 0.00023203018241299024332, -0.01249236018852314424, -0.026334397970480972334, -0.023200771701576351075, 0.022197567057567337695, 0.10609806530230778132, 0.12227767954578794285, -0.084161770990127254821, -0.45047572741760116832, -0.057195080198992456766, 1.061578111862742535, -0.87331039893758211079, 0.20646098595497261452, 0.0057915436576816701, -0.0070093165047670161899, -0.019480548408720394643, -0.024503094883576985452, -0.0052698027431828083425, 0.053425263609348822014, 0.12936234211863703591, 0.11024423487996769244, -0.092929658379114640776, -0.41739558456104680628, -0.34866030023112332126, 0.7753839051590639686, 0.12221167300440823056, -0.27791692468437045882, -0.0042915682259915361937, 0.011250691626111992277, 0.022230569041229726601, 0.021445369663284252759, -0.0036615961372018293596, -0.062186248576181768488, -0.12623951609442035648, -0.15591525316527513056, -0.099782012085464638629, 0.27192614295784811596, 1.2289086931436132843, 0.5932127220586197458, -3.6516175215544999233, 1.9491704743855229331, -0.0077282966950273420531, 0.0071753637249852720431, 0.024703008617840985228, 0.031612132170427881261, -0.00025158274684186979209, -0.090910814376522181002, -0.1870998035719727004, -0.10028172933644539933, 0.3735275032365690806, 0.58507040213514160953, -0.49728148202100913577, -0.52572462435677225212, 0.50758073416907778519, -0.13185938698020319149, 0.0019550515502324132024, -0.017437196736537246416, -0.016840998816763784462, -0.00035994437393744660794, -0.00073309627017000673042, -0.045298955218238207143, -0.027047108623778487457, 0.30661719797273562227, 0.90568762610236996924, 0.81020607743153283398, -1.9282039214065456179, -12.17727352528811835, 21.941568046873868525, -9.7816808701113409086, 0.0076480253898937042636, -0.013122468552984854068, -0.033782150143321171276, -0.037294019520872953577, 0.010024096001883244997, 0.1174463184297298568, 0.19966135211809249439, 0.10440222052522812879, -0.20376777725179590073, -0.642497807066875537, -0.35051869545956010343, 1.1742241328567304404, -0.058033114880960537385, -0.26354674160090679758, -0.0092199321429184393634, -0.0018304296168495739778, 0.0063070890193546082919, 0.015216698984793792376, 0.026941904439224357432, 0.031434146397019475938, -0.011097343409356185198, -0.14364443909530846821, -0.28425294585551047488, -0.095530362912114932805, 1.113787877816306171, 1.0904262103930280059, -3.4989649845938330941, 1.7705958314669294484, 0.016219256645843779102, -0.0071176501008802118414, -0.032070394192139868583, -0.046733128801109337369, -0.02295514580657079251, 0.064355193272649047209, 0.17955068416907016271, 0.1674017147612814338, -0.10106463314936206532, -0.53717930028424720756, -0.43131737794768670868, 0.92441911408824406493, 0.25648086080797427577, -0.433339162498465158, -0.0012195593770473018517, 0.0021104363759073558465, 0.0034709487362253985294, 0.0014186347097240566487, -0.0058479156194035531585, -0.01650317656884247447, -0.014317736822656530749, 0.019913474280019386603, 0.11990392467795414622, 0.061526343849983441969, -0.44589917360219877285, 0.024315393246474435729, 0.57374384907209441398, -0.3243091678446819448, 0.0044719043168091254445, -0.0024897082548968139243, -0.0096669946320714038313, -0.010500181230771088794, 0.0050143626820891226692, 0.037690284308636301525, 0.066175650769643462601, -0.0089666547462977103794, -0.27441158247933239434, -0.32247313412501227292, 0.76393580660313475938, 1.1176844591975898524, -2.5227289379168338357, 1.1623714308074268864, 0.002857705922469813524, -0.0015878374569873389584, -0.0015130568425242045191, 0.0015050354596057805706, -0.00063985712875168181124, -0.017512774624822272096, -0.033713204101792862422, 0.021862574791971706167, 0.22825367200987115024, 0.11048381584563357616, -0.78266523423345213395, -0.07025419251524658848, 1.1958875678622051808, -0.65691505110235870468, -0.00362691052994681528, 0.0043410514443971036022, 0.0094022184721997241474, 0.011263689025616243841, 0.0098655050587186255556, -0.00029031387562311605777, -0.027627862131636496401, -0.10561365256338797747, -0.2085794592766544997, -0.01438340399636367159, 0.94398416592214140497, 0.79447206026941330492, -2.8953201857560260102, 1.4893636548371973483, -0.0038152355688646374075, 0.0042228916967967395496, 0.014083625253692135665, 0.018497518730064124726, 0.0021455570393309527505, -0.047903500333557683111, -0.1068895559368317344, -0.054275005625305093593, 0.28257862282039386548, 0.35908205688520244969, -0.65530018653977706311, -0.33961929667156720569, 0.94209316691995859117, -0.4179483689456297002, -0.0092002785321508885608, -3.442071333671411896e-05, 0.0093229512998947369762, 0.016922248909285682317, 0.020268007446933963139, 0.011064709466255170586, -0.023105642752422921732, -0.10096312041375940516, -0.22229576312950527495, -0.15079906780914142939, 1.0049721693345938789, 1.1541169805232245338, -3.3919645306030061782, 1.6905048880981508042, 0.0025946050198121474104, 0.0036850787265515893426, -0.0031760081163546591103, -0.014823917620513488505, -0.015509119033687988193, 0.012288179486360727502, 0.048360952087705674773, 0.0099904260777772587593, -0.018209584002996652419, 0.084569664733825863911, -0.28443618896847061572, -0.16437431637137625562, 0.71180159102700490958, -0.37459243270201797449, -0.010857931185018014283, -0.014773894567276325829, -0.011648519852052454174, -0.0010195337987449946092, 0.013375577795547882384, 0.029109588618398237608, 0.054629624322138761738, 0.10511949813799685571, 0.14455098391632376087, 0.042696607127174081298, -0.3523728397409708335, -0.69739281693670762685, 0.056855794412692506334, 0.69266623721924869272, 0.00021975790210562840096, -0.00049736425262200862924, -0.0013164792313479444473, -0.0022993490856804154644, -0.0033846849686190714525, -0.0016796962243659356064, 0.01132186583942742901, 0.04373774095545553825, 0.063675374101468937282, -0.04731226703936908945, -0.36185240915929017813, -0.12917147174242132879, 1.0249814843143680854, -0.60146356226254238386, -0.00041803729931112727043, -0.0010676611919781253679, -0.00061197109767995461899, 0.0024683489101762733087, 0.0077980007565633096406, 0.0076968951427279887548, -0.01011656421849870674, -0.037419477035311163315, -0.036126328851826025601, 0.051867435528088215446, 0.16898525443853634753, -0.061927271385935296621, -0.26573774428817420468, 0.17605209843451488649, -0.024915956664899589768, 0.00642973351644823745, 0.042657734681636957919, 0.074692189603275588317, 0.080384181834048384907, 0.027875737265727656283, -0.10802494065153506186, -0.3107660365615632081, -0.47871919867830931006, -0.41717335010596029443, 0.11895803647540513504, 1.6205429512096614086, 2.5594840130814295875, -3.3063607091956113138, 0.0052430247890629900695, -0.0022986774800685637332, -0.0091043459860401781264, -0.010006753255023311672, 0.0021111312341239234051, 0.027125598688177637169, 0.045095420084745339862, 0.013844439050658335216, -0.10487744277446023333, -0.16652490454077370496, 0.11493684075965608649, 0.2463182176193100259, -0.075435469477896005541, -0.089962021362874317054, 0.0022131208464067418477, 0.00044417335970164462718, -0.0027926703193868729867, -0.0046928162559059845987, -0.00028242551280600511864, 0.0095432916610655848383, 0.0097304920633040723615, -0.014451392546026068514, -0.045933634003124459988, 0.011946371257179198844, 0.1463280635192096002, -0.047493370019371088098, -0.18719463137615929837, 0.12379833008517378279, 0.0009487629004291785054, 0.0016019244787207108718, 0.00039241438318929979011, -0.0033318661391703997647, -0.0075924383513549017349, -0.0059109775000443110643, 0.0073995620840038196786, 0.030597746958569618098, 0.040652880306553465517, -0.020807057307484739389, -0.20849048896892763416, -0.10884068423337046883, 0.65941889032949907445, -0.39060786082123333252, 0.0059160766592087198762, 0.0075709858518331289018, 0.0048813530028410595865, -0.0067762774420265237119, -0.030844970296327160059, -0.056388353452953454692, -0.044605655943841360234, 0.055648286575317326696, 0.21491509159069438484, 0.22315601418345626383, -0.195952216352625469, -0.56448374214454721809, 0.20646098599333356782, 0.18842938001822853722, 0.010381881929353912325, 0.0057915439232698560923, -0.0047591862113405929147, -0.022895580103562875735, -0.044867447428441530055, -0.051657606210413577197, -0.0052891119272775906357, 0.12350233805609048265, 0.26685220922872948668, 0.22818126242243683133, -0.19901921866663380145, -0.76751409068162823601, -0.27791692468989215303, 0.76867369815357367013, -0.0086853950940525212476, -0.0042915679900552986328, 0.0080314710630550872894, 0.02814727611497044174, 0.048155063005288496758, 0.046920812136362510802, -0.0081783884302348355821, -0.1289334523438767921, -0.27953376592882772655, -0.37170184720142412926, -0.14294698239596723144, 0.99024802421325586188, 1.949170474317895474, -2.1940680121467988073, -0.0045477874620092999247, -0.0077282964295281655898, -0.0014169015698260469492, 0.019959370853892762637, 0.052540856452852807701, 0.064848003179334562374, -0.005652484868793049809, -0.19530045325914843768, -0.37070924835089857918, -0.1205190449595275326, 0.60276793579875709295, 0.63010484505757613505, -0.13185938696731713815, -0.56892945576924502937, 0.018224528677871526527, 0.0019550515316250471204, -0.02165758115201881287, -0.040998339676366624129, -0.037082292414691374327, -0.014869210500803285019, -0.026764743282436481014, -0.081442305329186628904, 0.048051827434713673126, 0.72993004757361568124, 1.6947413525397911904, 0.76788033840750558046, -9.781680870138099948, 6.866605713780168152, 0.011966425944097445339, 0.0076480253782775118807, -0.0073819541248623344806, -0.037194314474555803529, -0.073211401721044927493, -0.073526299472773892774, 0.028947894104255040421, 0.2424073372051198938, 0.41010130212613243872, 0.25913187783278457399, -0.44586374568238607008, -1.1576706956130007331, -0.26354674160082020018, 1.1510736094920996919, -0.00030485381141720025236, -0.0092199321773717372169, -0.011182179093648216342, -0.0035833163756206487527, 0.013634473603948349049, 0.040088434088255235632, 0.07019070674021800349, 0.069100335467037130033, -0.046806598321750575442, -0.32901464596322127498, -0.54885772578351055273, 0.28540736655603282657, 1.7705958313699041717, -1.3181296445132459905, 0.018319561102021549026, 0.016219256837515145381, 0.0014738905865936068572, -0.030695751160939626495, -0.07543838191149050687, -0.10102377692011588195, -0.043352090651099677021, 0.14652064593998553876, 0.3766052606692156246, 0.36536398262904751233, -0.18688725452371260372, -0.94705452155475255971, -0.43333916250455684072, 0.91911171344080699885, -0.0029169037497739219722, -0.001219559458129773101, 0.0020096416095233423336, 0.006022320862508193616, 0.0080426587600705510411, 0.0028113283102537437938, -0.01473073751309692328, -0.036689213338026528677, -0.032911881250103282304, 0.069510474078682951693, 0.19369264021389645536, -0.1069772591549937174, -0.32430916782298624401, 0.240778351606872415, 0.0033401591629582229673, 0.0044719044922583682522, 0.0012073471714937546347, -0.0083637523283071526364, -0.0207956175471730835, -0.021269794392797785471, 0.011976886573028398125, 0.087706171936326016159, 0.12843692629051764453, -0.073927702857715826834, -0.489141406140261914, -0.1207775587571282977, 1.1623714308798516193, -0.67057132644555461631, 0.002812071527823509047, 0.0028577061026857659282, 3.8754716354684774424e-05, -0.0034770751183724035033, -0.0034681141911474581106, -0.00079167854790866817524, -0.0088807109781918712466, -0.03966519931643813085, -0.047820343698426429069, 0.12497722205890304803, 0.33570264825352430504, -0.19319414310317950534, -0.65691505101219116458, 0.49433610618763895728, -0.01043195137033904249, -0.003626910485330840303, 0.005775420888220604565, 0.01607634223348585259, 0.024064015824236530822, 0.0268318404301683619, 0.019954822107238239653, -0.0051533782171341024902, -0.085190381408898049886, -0.26539170792329924708, -0.35590571366266632314, 0.38667780648969390933, 1.4893636549153319582, -1.2661028446151065818, -0.0084309449119995914096, -0.0038152355106011361791, 0.0038712182090812619073, 0.01674256713270043731, 0.032956350485926801108, 0.037044213054963637255, -0.0040850852826895822645, -0.11449413707127178985, -0.21736615068541709217, -0.02456434820957734827, 0.44857915546635901416, 0.20681991886240141953, -0.41794836885643255053, 0.037481908160546612008, -0.003149660837209966157, -0.0092002784534353700824, -0.0081172030046182171853, 0.0024231396968003985661, 0.020993549423555440958, 0.041625597460090445201, 0.051561628264668500632, 0.029483399044254777338, -0.059967181581232320564, -0.26473577561988986817, -0.47810066650151950007, 0.24653778061672815269, 1.6905048880903157382, -1.2792653679355563412, -0.0046192777463164328908, 0.0025946049459792865782, 0.0087118938280355776721, 0.0079505883904736531637, -0.0064204226115162927463, -0.028724492303679138167, -0.03019948236776885675, 0.015452783029535326759, 0.057051565887525385046, 0.052974096002009793405, 0.09599017452037537923, -0.11214945181295926657, -0.37459243277338283296, 0.32160915426878688983, 0.67767492951669250356, 0.0025665892141424804206, -0.65284687308809130624, -0.29771643201371844611, 0.041642013526306607962, 0.12218323886715674875, 0.093977394973496511388, 0.054632995154234399726, 0.030704114818340639409, 0.012820586394321299706, -0.0023671032297318136094, -0.012023145970161165652, -0.013926743999344551664, -0.0094028346192934070258, -0.52754921884184546865, 0.93037250498359824746, -0.15165817843887022254, -0.33056589780057660244, -0.027078485940338826998, 0.067082382458627645705, 0.038876570180744418614, 0.005713957157194856501, -0.0038514487775783354251, -0.0027526770224089200205, -0.0012011498326273805509, -0.00085423113380704270362, -0.00054093267497900253144, 7.9626740562974021673e-05, 0.15761914081320749226, -0.24323184375409323543, -0.042220427558807512025, 0.15158875872981344979, 0.033048871211621998201, -0.033385336033249168874, -0.028103875445514223119, -0.0065701717437409642986, 0.0066869504792259537237, 0.0061436153493765180075, 0.0016449601917548908273, -0.00069702228499989721255, -0.00085895329382554307975, -0.00027760839298498633841, -2.8829330997447417673, 2.3341435030928479044, 1.3527528721979142912, 0.024773253755229349832, -0.37782026252445616299, -0.39135054665941587659, -0.24321491928269545202, -0.082388545317249803346, 0.023589914921456154456, 0.065215878416321884647, 0.06083863874523380566, 0.034818374046408452627, 0.0051735134878596426844, -0.020469593244905239654, 0.012287513494102913006, -0.12757299824602805205, 0.15351200197297404371, 0.089461036492104281659, -0.12521756108355269754, -0.061759940991765273355, 0.024896904331536968968, 0.033870673817203836298, 0.014320193136529168337, -0.0026359994066779048744, -0.0083874545961641882447, -0.0060320098301071926883, -0.0008939116450472857139, 0.0039635667174717330308, 0.13277498040440008364, -0.17968941222237402866, -0.063090229139663819846, 0.13171884374674217022, 0.01644795277185441551, -0.035227478100793407911, -0.010292845835484265224, 0.0074170554545264881507, 0.0063911256191791481623, -0.001126216674523633542, -0.0038454667468715581641, -0.0019024425541652575846, 0.00058890444645844411929, 0.0016201069890321819541, -0.34680519212618876601, 0.59792125251920991147, -0.11422698975556294654, -0.18757801408401941767, -0.013114264948414411718, 0.040664123044299491094, 0.028048170361774156512, 0.0045111784504186522651, -0.0069060714303482576351, -0.0068516125332677398421, -0.0024087819249172396155, 0.0006710755352958175813, 0.0014055789063860801152, 0.00074445512617518660482, 0.037050940445650722144, 0.28733360500317511876, -0.44450710955168865057, -0.16154693999688268802, 0.18228094348527412727, 0.16458347572456369368, 0.034789412324114414587, -0.040141063252331667111, -0.045228556990773924329, -0.023649503512074329659, -0.0045501207863940596257, 0.0045583844704606621412, 0.0065726034066351827698, 0.0050400520268825202419, 0.65429739459286628822, -0.32622412295702646334, -0.61330835930099569797, -0.076198816862874885203, 0.20384681452581029926, 0.19150694747673649432, 0.079105086853112127621, -0.010969267361220728235, -0.040193010457330327623, -0.033224148326609817539, -0.016620699368198800505, -0.0030603021521710162267, 0.00488252538696147282, 0.0082266369823614446388, -0.55564585711519542954, -0.16092673366240273225, 0.61422133350372165594, 0.61537832096256428471, -0.10021353142285518989, -0.35874632734756600749, -0.19379796694930886969, -0.011565859516564502493, 0.056462305882219759878, 0.046684515625231612357, 0.018690788967678329674, 0.00054233656953988100199, -0.0052304075694336940594, -0.0036601993424298213085, -1.9942532617438395448, 1.9774077467131530028, 0.76362028335229792386, -0.30221330441007859813, -0.33123893987774105652, -0.18039494047860263337, -0.073371250762032796944, -0.00026069839021241963083, 0.034290370024538274762, 0.034727432360117736765, 0.02059927108654884334, 0.0057514942110892860322, -0.0035295942935515485679, -0.0065329175029392944421, 1.1510736095524980449, -0.26354674156404339636, -1.1576706957932623165, -0.44586374574351939071, 0.2591318781071930144, 0.41010130199638972215, 0.24240733707686437648, 0.028947894318186197932, -0.073526299421082616514, -0.073211401877238221902, -0.037194314630943886568, -0.0073819538318176738945, 0.0076480252372337883052, 0.011966425968615402448, 6.1289695309289893288, -9.2482978107958029312, 1.2414333579575957955, 1.7196781345618650239, 0.55766646761726579307, -0.10193408739664000517, -0.14160929270121569434, -0.020491584305036336655, 0.013094774845543439082, -0.014215329293325511956, -0.029601836835366445583, -0.019058791117071750137, -0.00035758558993484863787, 0.013801033433235973882, -1.1812151164552346927, 1.6220054376744439395, 0.20821567892585726711, -0.51581275892911881975, -0.27226665835193697429, -0.015579117371508193965, 0.067100119168192820518, 0.052846494383601949618, 0.026176171548788405158, 0.0092247570486640832754, -0.0015641857559390134804, -0.007487078198236761517, -0.0069339689405451257129, -0.00081206672916238754716, 0.84187980064377243572, -0.42396813737665500765, -0.84969154931862622693, -0.137098672289939022, 0.33030775155099412199, 0.3236003605636651459, 0.12322835990696268116, -0.037467451105539215095, -0.08559152885224305396, -0.064557338238720743395, -0.02749285801025833087, -0.00012559473176546135384, 0.013425422944633251185, 0.016689580245113123241, 0.19709749824623168379, -0.30514594823190249739, -0.058297514619021620352, 0.19991691557013552272, 0.043963067612672546147, -0.048447660398733702336, -0.035274587280477109896, -0.0085162199477255874636, 0.0060618281116217890767, 0.0078908474711400448248, 0.0046924967463787634753, 0.0011683035686405563933, -0.0011823442515330197765, -0.0022754039404820304224, 0.5182575117734928849, -0.67855335670239202983, -0.20072146099792542717, 0.34384741173329047559, 0.12467072996729103296, -0.054471264328746486771, -0.041006888028882672315, -0.0050025997510950090397, 0.0028386539860703243615, -0.0022946519071268514818, -0.0039738704522355424234, -0.00092588003190525428217, 0.0021567801077250702523, 0.0027402548237654257915, -0.67208592911480113941, 1.2326278601168791926, -0.18919927585289028626, -0.54136743116372787465, -0.041052567041237338075, 0.16106091979942124826, 0.088328592582869611127, 0.0014149778664894938594, -0.026835312064533933807, -0.020179638434010317011, -0.0058571130868494856656, 0.0026131654551201106279, 0.0042656500813110006382, 0.0021264946999231953972, 0.0058566070871939410886, -0.39549208264121860079, 0.23719055768691768571, 0.45616908396241867552, -0.040195401123592884585, -0.23749198286413494841, -0.12287232424461196068, 6.3314298603975066766e-05, 0.04507975442535117061, 0.038203635602786750747, 0.017958616028840253265, 0.0027155276129601515177, -0.0054163389621238733351, -0.0092132173595859730847, -1.1590087791377825521, 1.5383557547258224218, 0.21463362878002881229, -0.47923952994704122066, -0.21431493054248318408, 0.007348295616581393061, 0.035841878854495023976, 0.014653360158089160509, 0.0073069637254388926373, 0.0095715828637429421072, 0.010412534039306113251, 0.0058738434221620701783, -0.001205839836813029026, -0.007708695580385068781, -1.1295004716574637005, 1.5266825593682173956, 0.17165291926114201182, -0.43767544602242969276, -0.21465990450159119773, -0.036766173887621912686, 0.029074394634586702563, 0.040150819321617924051, 0.031078434845320884738, 0.016441742905768046146, 0.0028091225155509090897, -0.0058432290253356330284, -0.0073699624907034863922, -0.0029455238268658581399, 0.23068593931369915651, -0.3350460548545953765, -0.022250043497792117153, 0.12224637529027275462, 0.013360608731521913736, 0.016890995947740735511, 0.0057275873455112143867, -0.020224893259745198304, -0.017410730868856100911, -0.0019565883937056475009, 0.0067461144326717405531, 0.0059859019367578336981, 0.0012485706034473709105, -0.0034815098573520541873, 0.0025665881229528435085, -0.22558743391974359072, 0.78833650932966370384, -0.25089158787141263574, -0.36949799419767748176, -0.099061879795536877946, 0.04378336666738421451, 0.068782121490089290217, 0.042473955909127551755, 0.022479331796989233255, 0.011800553863180350198, 0.0037468080618883467872, -0.0050789579968903483115, -0.01392674390209068698, 0.93037250524022363507, -2.0789118735936567006, 1.0779993213102452287, 0.46063762023290399927, -0.29374783296555950773, -0.14548810614771562344, 0.0069587393471482131974, 0.036083067996621233375, 0.015053554880884013542, 0.00039688914501439555294, -0.002249683842146897074, -0.00094533005543840536219, -0.00041934178231514278356, -0.00054093277894298567865, -0.2432318433821606396, 0.45852147060554526714, -0.044021354260558802352, -0.32125431727815301208, 0.083335155500672331019, 0.082888864300899978144, 0.0095903130780454082932, -0.015466062843856135181, -0.015983459632721692956, -0.0023623367734606120008, 0.0041698296678377095095, 0.002952293528597551512, 0.00026102212113773425213, -0.00085895347570459927491, 2.3341435030797610395, -4.3877424341862436563, 0.67783342545485525399, 1.4089671414771871572, 0.38951250172220713841, -0.022113128405077446537, -0.17472116892167943725, -0.17110727827310454052, -0.1037624742172548048, -0.033995541316451674541, 0.012062850161438773769, 0.029887189230178252963, 0.024361051018511801353, 0.0051735134118589308841, -0.12757299816808870263, 0.40403999732015360991, -0.3419444881573978634, -0.14295758546783143839, 0.29251740214530902451, 0.0017215730040662634986, -0.091481193818788275607, -0.022498029467294986677, 0.012086375321176479405, 0.015176650439295898132, 0.0065325814341128498122, -0.001101300553594591761, -0.0034168808230149515064, -0.00089391139509865978965, -0.17968941217311923397, 0.31348544910536035868, 0.050765458610885071111, -0.32878348135573642885, 0.087333971395862353138, 0.086879757420077557972, -0.011290271495125694945, -0.021085886376906785161, -0.0057704779862023632478, 0.0055514530131077462552, 0.0039255187179787365964, -0.00085682122864235940075, -0.0020502471349625578899, 0.00058890415362215699918, 0.59792125257480910339, -1.3643459788769243257, 0.75708675753141629894, 0.23221260323707956785, -0.16834303331556066397, -0.077168043887074544429, -0.0030080332679837053396, 0.019407166711339685072, 0.012770236322687445255, 0.0022242557387467773618, -0.0037667929840606659969, -0.003310314282949939485, -0.00068896030148845699252, 0.0014055788942091715452, 0.28733360519155859425, -0.97671207209211097222, 0.99436999054619801264, -0.013830051263679286194, -0.38159249951770651821, -0.061109641952355547856, 0.10136126462863882802, 0.079903584751910422534, 0.012305312648964616962, -0.020389090288214707647, -0.020805710250756254132, -0.0093433507918371325879, 0.00069312551441018551129, 0.0065726033460430577024, -0.32622412276984630308, 0.29236144572213484949, 0.5927806288928875178, -0.41665234463902711681, -0.29683378464603166202, -0.022784956262144537287, 0.093034224779902280345, 0.091795227521835012774, 0.033346014248142402669, -0.0068918269237559365292, -0.018636648274540975473, -0.014196943256056726232, -0.0048853864115151694053, 0.0048825253405461021411, -0.16092673354433026378, 0.54227620542379184609, -0.51664743093548071151, -0.51637208270791479237, 0.5768546079416931871, 0.37932666170724854693, -0.093874671637461937213, -0.18345825878831770184, -0.0896287634788961729, -0.0019456115301864900652, 0.028055251712551711879, 0.021901440824615529823, 0.0068598197392295851699, -0.0052304075565745237336, 1.9774077468439674732, -3.8075718313399224968, 0.8327607508564248695, 1.2903367278884250169, 0.096982683635133626332, -0.18525094301587452517, -0.1263116629928714707, -0.077750948906419356521, -0.037461251750854213105, -0.00092529146562195501574, 0.015099834963946514102, 0.016018623386702516426, 0.0084468058990322322788, -0.003529594493749591607, -0.26354674157560681325, -0.058033115007180530176, 1.1742241329439619957, -0.35051869530522306029, -0.64249780734584838182, -0.203767777212260498, 0.10440222090528727961, 0.19966135185033007104, 0.11744631840600970596, 0.010024095924691249146, -0.037294019546682023736, -0.033782150085389227312, -0.013122468404493811689, 0.0076480252989430570765, -9.2482978108671183293, 21.080580704611357135, -12.303715153244931813, -1.5410168647928164098, 1.0252883696367129929, 0.90669338336331872785, 0.22471705859674356853, -0.098026494201007952078, -0.072365223492937974514, 0.0034666852011160371842, 0.01290225726050770079, -0.0067646587131558405873, -0.013707888979432930263, -0.00035758542948549344688, 1.6220054377907742182, -3.2508776016453460045, 1.0812401135068556712, 1.0145155875403468748, -0.12960026693790521746, -0.27338814169451441938, -0.11707282877062627491, 0.0059984947577352695433, 0.031374195569268140127, 0.018982118772446543442, 0.0089435483873858360321, 0.0044275786133970030517, -0.00066801526233130055586, -0.0069339687676063964036, -0.42396813710766401151, 0.29966087849980438396, 0.82312222561790149289, -0.445560088607611815, -0.46623252655960589763, -0.06766023864483158845, 0.15078429952291438898, 0.15350355993014105915, 0.052944723166805270076, -0.020223997224233934006, -0.039477821489391863585, -0.027167274414009883604, -0.0063849208698478366469, 0.013425422713037033287, -0.30514594831406144371, 0.56818826969571101237, -0.03239756836851796773, -0.42590308025340739562, 0.1035136428878111331, 0.12056290953488053852, 0.0033547091866799462305, -0.022314874056907386957, -0.015554895136850922094, -0.0027824914645987980193, 0.0030090383534540640978, 0.0033604787398465291691, 0.0014871973172077715495, -0.0011823440523495529642, -0.67855335687528972599, 1.2409997266458416831, -0.075009353872135350882, -0.81092114242949020486, 0.11196210128095060443, 0.23975384851104522554, 0.021271207134567253483, -0.038620107209166372464, -0.018176667709832498854, 0.0014517338346951855557, 0.003126384727028044385, -0.0011093800117971549211, -0.001941215323068132834, 0.0021567802882482430514, 1.2326278600978046729, -2.7192590778932244078, 1.2693930649758096152, 0.82223057671076815733, -0.41736006837384159729, -0.30060683009369082441, 0.01942820253497250238, 0.083530085185592642394, 0.036817625434480708013, -0.00056778965730850576923, -0.013224466524003887849, -0.0092568212086793760091, -0.0013567272746500718458, 0.0042656499904241895402, -0.39549208265220087144, 0.91953849258042408277, -0.35628336987341807651, -0.64995424348659336378, 0.37515551499824612014, 0.29180282955624220476, -0.059377731626222222949, -0.11709903357477459418, -0.052449520250985676162, 0.0040636162058751857523, 0.022255635532545173932, 0.016420633943254568654, 0.0044130456895005313114, -0.005416339164571172711, 1.5383557545877031281, -3.0970998166976277943, 1.0126971901670258625, 1.0261369010233596377, -0.16625140298096965918, -0.28253783142715888088, -0.072010411119283926351, 0.018164477934971091921, 0.017621809158186244049, 0.0063664205085208333199, 0.0018427363850143101305, 0.0030653399962011131791, 0.0029600197387885944685, -0.0012058394882820640866, 1.5266825592518844523, -3.1058982383051829679, 1.1293653056168277971, 0.88415569409388961031, -0.17264385225237607879, -0.20141981330552605201, -0.080332472581696887093, -0.011830835496212766672, 0.011854733385845103374, 0.015246967457757122297, 0.012112229910923673695, 0.007350759944418817568, 0.00048274033447347344328, -0.0073699624171729791883, -0.33504605485409222343, 0.68658250773368267872, -0.24330284359562731611, -0.26643382055544145848, 0.15109212604143593128, 0.00027548602116519288734, -0.012163433737625888725, 0.026988442888151546739, 0.0068055062874922351082, -0.01099467543671437883, -0.0093144057049237439366, -0.00091164636982818804487, 0.0031333528910591880792, 0.0012485706549734832804, -0.65284687185323508896, 0.78833650786486841877, -0.28385910543428810593, 0.72277202476661472907, -0.22502133821061923169, -0.37066789719853426854, -0.11898852342137820892, 0.016499506665748300827, 0.052281272669388645369, 0.038109085520640899325, 0.025466794994209097924, 0.015739487557463748696, 0.0037468078521671003417, -0.01202314618412141968, -0.15165817881737866935, 1.0779993218805066224, -1.8911425774239505326, 0.84847845736488403112, 0.46812753742728008266, -0.25884041049640826859, -0.13660361543977025955, 0.0014378044705094827124, 0.031865314731807571624, 0.014634218839251440097, 0.0015197111465874669697, -0.0016322209238939387656, -0.00094532992449645688888, -0.00085423107807261608113, -0.042220428321032635322, -0.044021353581512566444, 0.18856734532353502942, 0.056068575994462334366, -0.29558026647543211851, 0.076631270343063453976, 0.081013760517512384274, 0.0051551762220002281256, -0.015324073599175196125, -0.014995932024318555861, -0.0023523774985597174954, 0.0039387067910837247203, 0.0029522935701730658106, -0.00069702206956632583972, 1.3527528721392521049, 0.67783342559439119235, -4.1997526353711247538, 0.5305099077444225264, 1.5155154494493396911, 0.52478068049449522192, 0.06437112171763274826, -0.14539795019450149627, -0.1803433210804477882, -0.12723533046188736928, -0.054535180003340850974, 0.001034648679809966991, 0.029887189525418862701, 0.034818374012048881883, 0.15351200166386277934, -0.34194448760192264203, 0.27027142610763582597, -0.061977317179121083068, -0.1756534158583230032, 0.23769892169608755617, 0.0014254934532901180136, -0.091452362118988300699, -0.022479756168529409222, 0.013733128140819474833, 0.017354141692841825589, 0.0079994150061640841715, -0.0011013000166739997897, -0.0060320104058644592043, -0.063090229183747570385, 0.050765458761050874514, 0.017883505048113750024, 0.17219850730267421968, -0.28802460085114217625, 0.060077393172844544, 0.075804185027634099492, -0.012578559951011387269, -0.01708504676480742876, -0.0033751817634539884976, 0.0056844734398824618357, 0.0035029265852987885588, -0.00085682176428625257333, -0.0019024421358433928064, -0.11422698966873651161, 0.75708675755012166952, -1.3163946914162742985, 0.637750112117037804, 0.2432569353368377596, -0.15878874761793290715, -0.074390344262033780698, -0.0022106339761187370971, 0.020488399779125952144, 0.012780083487404234402, 0.0021873055416145727008, -0.003874839748242425632, -0.0033103144114557947475, 0.00067107551173048441123, -0.44450710971403817329, 0.99436999047154750464, -1.0137732960265439797, 0.73995945451762057399, -0.034447203913422326516, -0.36182677155155257287, -0.049333098528702122587, 0.10557737735707355131, 0.085354350080527371936, 0.017011661740974480528, -0.018217018084750992435, -0.021057347719693364496, -0.0093433510225370552948, 0.0045583846808365255723, -0.61330835946334361086, 0.59278062895827832168, 0.11722800214115432216, 0.51089514631455501537, -0.4560901195644962991, -0.32972978895737359295, -0.029533982419913774181, 0.093325582134294232661, 0.099008045484902759692, 0.041890020839105646111, -0.0014246208159905036049, -0.01713018043180024319, -0.01419694325970354172, -0.0030603021758192155109, 0.61422133352811347784, -0.51664743120654255115, -0.22342876603553790682, -0.16979508529466499911, -0.30307898196449556405, 0.56625552014496838282, 0.36915648009612767044, -0.09348502030215050107, -0.18544110447448244883, -0.094974231189080132731, -0.0060963304971648930863, 0.027278922495896886641, 0.021901440891955583495, 0.00054233678124837223465, 0.76362028340188792264, 0.8327607506061733833, -3.3730904210482930417, 0.6605570916296642503, 1.3161109652574785045, 0.19844388426520184754, -0.14632992395346144932, -0.1292218644033463204, -0.088788553587140522927, -0.048264013786730980415, -0.0064526079891638766367, 0.014329275950864906602, 0.016018623066960051282, 0.0057514945851495705154, -1.1576706957333857684, 1.1742241330191296456, 0.017248055349690732951, 0.97908561123309745255, -0.46580013299011263639, -0.73706384219253184842, -0.22249201342573524109, 0.11348388894404631833, 0.21105753036333657624, 0.1292022998448746629, 0.019158853134946054325, -0.034214302788519564136, -0.033782149963327290365, -0.0073819541470440085129, 1.2414333580494492093, -12.303715153145988737, 22.936152497826096663, -11.999686437844630404, -1.6334274145234324305, 0.8430391982377096749, 0.78533650516770658889, 0.21155373971466120975, -0.065166314131384017738, -0.047673431221847341333, 0.012130840131680805521, 0.014505329608643837552, -0.0067646584231022093003, -0.019058791526155978435, 0.20821567843645322204, 1.0812401140722360804, -2.7800504462864079613, 0.93291108489128371506, 1.0008784075966152827, -0.10088106829794545038, -0.25638019116041621537, -0.12481764574800692513, -0.0082611750521376196843, 0.02529636403827507235, 0.0214619500071390433, 0.01296019890521859369, 0.0044275791100218502006, -0.0074870785500755921565, -0.84969154956535353218, 0.8231222257402027731, 0.18781932826439853845, 0.62737212189219526692, -0.54647247485518757326, -0.51273902665922022059, -0.068234485372203779185, 0.16200878752176547937, 0.17043821440025203051, 0.068211649705553475953, -0.011979502992035562584, -0.038170256885529246782, -0.027167274525525728773, -0.00012559457717888709957, -0.058297514457957076706, -0.032397568640848257993, 0.20295950382112040722, 0.094421471958618746756, -0.37833794741153708996, 0.084231972532695792433, 0.11856804339596611897, 0.0038298767273983766284, -0.022170853953494142063, -0.016831241561935241652, -0.0040578435367127040079, 0.002577057669062537363, 0.0033604790310003446359, 0.001168303260904816291, -0.20072146049539607571, -0.075009354310941023236, 0.61456814516581736996, 0.10680440919310123238, -0.70574861523289633691, 0.06010323084651218728, 0.21458533106201185814, 0.023801818662709080948, -0.030101944658393688797, -0.013885711573509165739, 0.0015259307222119412872, 0.001983627834739550011, -0.0011093800324951210486, -0.00092588031461592209224, -0.18919927571571037994, 1.2693930647951539026, -2.2153503677848926934, 0.92592894689487648119, 0.73759051331503877602, -0.36496660970895333787, -0.28101761380335033014, 0.017566747606472678245, 0.084262744757048729793, 0.037994935269443079473, -1.6938884539909396243e-05, -0.013546762188332502758, -0.009256821616322120222, 0.002613165685095450181, 0.23719055762434182455, -0.35628336976557578586, 0.23426068974968519698, -0.051810081921579209663, -0.48590722540372666138, 0.34854821075181680046, 0.28131442701465586564, -0.059791248647179794096, -0.12100775718939117687, -0.059139279645089040671, -0.0015964737713628834523, 0.019770196256602586699, 0.016420633367824399862, 0.00271552827014257675, 0.21463362896277835001, 1.0126971900169796648, -2.5942019423750468654, 0.79795610461615051889, 0.97737358874231561323, -0.1049892781007473741, -0.25059525534467985697, -0.072955541471013268806, 0.0099205306184837610539, 0.0094252721730352165275, 0.0016294392761432687292, 4.9768664534457357496e-05, 0.0030653401407877999864, 0.0058738432036397553518, 0.171652919335703813, 1.1293653055649957029, -2.7528507007923863625, 0.93493014555288500578, 0.91994697765988742155, -0.12334704755494438344, -0.19276860360503617065, -0.091583768189017880634, -0.024549663105296003901, 0.005612935970627236891, 0.016140130643495748802, 0.015105049793058988872, 0.0073507598847698538483, -0.0058432289307560796379, -0.022250043646091514221, -0.24330284324019879616, 0.52524447912933580618, -0.079112603652024132694, -0.3389238027252886587, 0.12070195368561723037, 0.027627167308052775346, -0.0020685993299120481598, 0.024987929918843888866, 0.0034209236361172953161, -0.012863343993660559389, -0.010309050800030434419, -0.00091164590192974463874, 0.0059859018376406192621, -0.29771643201425918024, -0.25089158811200129584, 0.72277202484303937347, -0.32894962180160336462, 0.72652517198124566544, -0.20758471912085207922, -0.37684158889554902894, -0.13174860449876765744, 0.0089374311117457732462, 0.052125099570361391188, 0.040369329412383005151, 0.025466794799589159704, 0.011800554017032623133, -0.0023671028924807228401, -0.33056589775570804912, 0.460637619973786272, 0.84847845725962034447, -1.9665237248945148796, 0.83264729567177409031, 0.49358610156199261843, -0.24543079130060088677, -0.13659120629921389645, -0.0013357953444529694839, 0.031085735171635559687, 0.014873716335044735379, 0.0015197110731374012787, -0.0022496840577605958442, -0.0012011498654900601424, 0.15158875922900710842, -0.32125431736363924129, 0.056068575019392309644, 0.21906453584027518744, 0.072688227446352851602, -0.30105589419586981093, 0.062657016049739402974, 0.078346981282599412189, 0.007405832748749697006, -0.013567441378663117751, -0.014782940911059569422, -0.0023523778682521534478, 0.0041698295387109490101, 0.0016449602164636525567, 0.024773253890916401704, 1.4089671411224180542, 0.53050990776619011413, -4.1332126425344046794, 0.52644570967982862175, 1.5176012668231699809, 0.52926508012725692609, 0.065335877878227227589, -0.14769970232346232852, -0.18407470864279182332, -0.13115916604019725589, -0.054535179762257073033, 0.01206284967359858018, 0.060838639152013311573, 0.089461036934711260504, -0.14295758655668408532, -0.061977316450771466338, 0.25057268968677670617, -0.10795885232699191369, -0.19488272690737953874, 0.23527057336271020582, 0.010485765948125528799, -0.085344578406273186033, -0.020889562984851180183, 0.013634987649074011007, 0.017354141757483072805, 0.0065325808775093530123, -0.008387454104769770033, 0.13171884370694544875, -0.3287834814400844019, 0.17219850740380743259, 0.053935726776970997765, 0.16531743169728849163, -0.30150505631462759526, 0.05116445890873957264, 0.079403519133057293145, -0.0084642087961833041565, -0.016400974717268347602, -0.0042215409764120364819, 0.0056844728081027282723, 0.0039255192805919322113, -0.0038454669205773956385, -0.18757801441239732587, 0.23221260357688405795, 0.63775011197549691389, -1.3637768171227864222, 0.61814555536616455633, 0.25690145316578566481, -0.14677426409236035698, -0.070608700235039026372, -0.0039387313143904745943, 0.018721436823325037641, 0.012221145808603936511, 0.0021873047733861287339, -0.0037667924418547699301, -0.0024087820041834386887, -0.16154693989622673223, -0.013830050818172871449, 0.73995945403987661848, -1.0999750937069561196, 0.7596027169214638608, 0.014698157663684338278, -0.34514657553220190467, -0.053660437035877396605, 0.098524108509233973563, 0.083095185842447308322, 0.017257528313142228571, -0.018217018377558711478, -0.020805709868110713878, -0.0045501209076528282768, -0.076198816671327218586, -0.41665234443368648343, 0.51089514592443163732, 0.035685276763889557661, 0.51435064532604735099, -0.42023552930131430339, -0.31727618613845742779, -0.030925829613137126306, 0.089025417913744089859, 0.097436771302062435973, 0.042171385911963517323, -0.0014246205065467656102, -0.018636648574153043267, -0.016620699284694898795, 0.61537832057048424961, -0.51637208203444817656, -0.16979508534839515832, 0.036940106332258089428, -0.15748525795789952708, -0.37240000344972767898, 0.51685049894050250074, 0.36730247406923527409, -0.078743111129536760839, -0.17622044626459060224, -0.093325157817837040009, -0.0060963298770975882929, 0.028055251765024345756, 0.018690788438053360793, -0.30221330472959728741, 1.2903367285637141748, 0.66055709148544539122, -3.4140908854952516371, 0.63607444117605682621, 1.3106793726910690445, 0.20267891662354453608, -0.1404866387887671042, -0.12644095092669216385, -0.086911616256607829301, -0.047326621509899378937, -0.0064526077955227838928, 0.015099835040842297246, 0.020599270943199674, -0.44586374571918635512, -0.35051869534054586053, 0.97908561129221316488, -0.19493925416863755462, 0.91624666284835087904, -0.39825839216776559581, -0.68416000910677965408, -0.21564805944511838964, 0.10078158143055203355, 0.20228502960084973616, 0.12785593793437988186, 0.019158853022131713123, -0.037294019853519599916, -0.037194314271514247738, 1.7196781345127269969, -1.5410168650650262201, -11.999686437532172789, 23.82241132399909489, -11.843026230695130607, -1.7990462129398137048, 0.72770300080400107667, 0.77291642474490240833, 0.2334483527765102151, -0.052626627624508773295, -0.046846786319351838013, 0.012130839961838604454, 0.012902256868216589736, -0.029601836367459489346, -0.51581275820170391366, 1.0145155866265334055, 0.9329110850803330468, -2.8915303010158566188, 0.85720728102723298125, 1.0100934164372805668, -0.079989583537098785104, -0.24982509366697988606, -0.12497796662473326101, -0.0068562368730504877345, 0.027604560490347863216, 0.021461950118213543592, 0.008943547872952931771, -0.0015641854752469523063, -0.13709867219080942857, -0.44556008849629485935, 0.62737212196719616131, 0.032668614644018385829, 0.6289678921794047417, -0.48132259528984344676, -0.48079807362242166224, -0.066617253590804240138, 0.15447130900900921113, 0.16643970745430555769, 0.06791371789265542025, -0.011979502746746282377, -0.039477821657186190629, -0.027492858150584435772, 0.19991691532854416469, -0.42590307951564737587, 0.094421471059639827672, 0.25515409940688366097, 0.10489834213419409192, -0.38424360982103117346, 0.071528371207234459073, 0.1150843543776826905, 0.0042934612537657065173, -0.021309553024404050137, -0.01654741383431852092, -0.0040578438297727880205, 0.0030090383785071270206, 0.0046924968886771552951, 0.34384741104806570977, -0.81092114188565056487, 0.10680440951923916848, 0.73111019080209449506, 0.11375229203465817007, -0.72842094622218278577, 0.041478178595367364012, 0.21555105708175323342, 0.029373871420075955974, -0.028572866027938865346, -0.015046969000910750958, 0.0015259306073793560063, 0.003126385067388189401, -0.0039738703393511695672, -0.54136743139650977952, 0.82223057701872037661, 0.92592894689778393325, -2.3999548981952978366, 0.90494838322553605714, 0.78350704846629437572, -0.33145477632196845841, -0.27455328655895328005, 0.010819634066796642718, 0.079189317292507402368, 0.036777839185335683969, -1.693874612331697227e-05, -0.013224465978894680077, -0.0058571135979381050082, 0.45616908386974697187, -0.64995424352578723415, -0.051810081593126784683, 0.42793939217380627671, -0.041846657335692885593, -0.52989827718851512106, 0.31243321393032152944, 0.27640170923089335542, -0.053273072120783360184, -0.11768435192458068239, -0.05980947654163930316, -0.0015964733724115561832, 0.022255636016764585028, 0.017958615441691173908, -0.47923953008980030299, 1.0261369008805616421, 0.79795610501462443942, -2.7292213856134606154, 0.76135847642976617333, 1.0033573933661128486, -0.07997303301999028613, -0.24486791414551323243, -0.078807184124287901761, 0.0049562430628356333312, 0.007799075684971229841, 0.0016294393620663203889, 0.0018427366536732485425, 0.010412533863673526652, -0.43767544598431296077, 0.88415569392675741334, 0.93493014578503785827, -2.8120309269023802123, 0.88857530728298916856, 0.92240860432471394326, -0.11304075718983451482, -0.18942284419551858043, -0.091110997769708948479, -0.02259610912415337472, 0.0077011648415665854422, 0.016140130446311337642, 0.012112230180470206359, 0.0028091223099958886021, 0.12224637569877974674, -0.26643382118223118971, -0.079112603392477692799, 0.46798671313576545305, -0.081377093725274032843, -0.31511912457205543125, 0.12642977675140779081, 0.023351391418811269507, -0.0085547730903958747045, 0.022338850195620828937, 0.0033986677795372513371, -0.012863343393771063672, -0.0093144065225037857986, 0.0067461146200332045703, 0.041642013597187173268, -0.36949799414514727935, -0.22502133824537276507, 0.7265251718021499272, -0.32539309455914006142, 0.72472228880492617353, -0.20852596193363784338, -0.37718785969675289182, -0.13049951179538452073, 0.010404890779721124444, 0.052125100143867106328, 0.03810908451973140515, 0.022479332053770111066, 0.012820586525070556613, -0.027078485977597620271, -0.29374783291176737032, 0.46812753804282125625, 0.83264729491069222611, -1.9710559092903074951, 0.83104146311958393412, 0.4961758902795107562, -0.24341076602675185137, -0.13598377263491900901, -0.0014583225900812379558, 0.03108573422768844477, 0.014634219683482644139, 0.00039688892771794437525, -0.0027526768969542542674, 0.033048870952357249642, 0.083335155272063060616, -0.29558026500701917172, 0.072688226391154284189, 0.22108712694406729149, 0.065242810184814370444, -0.3040700811077680088, 0.062735012408885931778, 0.079215371912899901208, 0.0077990510702655394676, -0.013567442610847500056, -0.014995930348202072407, -0.0023623371828940861045, 0.0061436151947898923892, -0.37782026249616501579, 0.3895125019330992755, 1.5155154494505285179, 0.52644570925245326443, -4.1685627893871268768, 0.48107698979018842511, 1.4880978879561737038, 0.51841777670948774315, 0.065417373930822975137, -0.14618485150801205363, -0.18407470840883546903, -0.12723533012916135587, -0.033995541096175266316, 0.06521587788838140165, -0.12521756134453138065, 0.29251740323232139351, -0.17565341688210447946, -0.10795885220860955422, 0.24081769787235057789, -0.10997092697434657227, -0.19067942372029950127, 0.24094465896885566836, 0.014002802788814713159, -0.084622325373753373712, -0.020889562999845279895, 0.013733127930254135363, 0.015176651033662337878, -0.0026359997745315959586, 0.016447952772804304983, 0.087333971349039335452, -0.28802460069206681226, 0.16531743168251955556, 0.046261355464287307049, 0.16251625488564305777, -0.29834995757777643677, 0.053462443076162984024, 0.079733359776122059159, -0.0090481771352150571625, -0.016400975384678169455, -0.0033751807263973582565, 0.0055514523113800104404, -0.001126216472649195369, -0.013114264738342689548, -0.16834303382233781199, 0.24325693598090131209, 0.61814555503826540406, -1.3690141346467699091, 0.62145408101411481194, 0.26213232155750815222, -0.14579356468290086135, -0.071541188397014829348, -0.0047472839508882263768, 0.018721436218511904037, 0.012780084551636176557, 0.0022242548762795707681, -0.0068516122057464962092, 0.18228094323408566835, -0.3815924995830665134, -0.034447203771329794331, 0.75960271724393102488, -1.0681000911843456791, 0.77716534651644186837, 0.019542448465056663776, -0.34903321883798010816, -0.058384024667873408998, 0.096381566636102852796, 0.083095185879928978867, 0.017011662252252802291, -0.020389090745044181202, -0.023649503523118307891, 0.20384681415203068089, -0.2968337846166203553, -0.45609011926395709446, 0.51435064547608067009, 0.069174167590422386276, 0.53943289591637644254, -0.40934105467211773233, -0.31794510037855266971, -0.035007274496678279785, 0.086465543593863825045, 0.097436771504004981992, 0.041890020143823224197, -0.0068918264517738641803, -0.033224148346158881295, -0.10021353104965204095, 0.57685460753159933933, -0.30307898233115276865, -0.15748525753261394255, -0.011016767308998247354, -0.20179325280286247724, -0.39118649806294403248, 0.51856335890724059023, 0.37637147302653783942, -0.073069944312984760892, -0.17622044597644639929, -0.094974231474089884641, -0.001945611671225500397, 0.046684515803101046816, -0.33123893965914502635, 0.096982683307508643633, 1.3161109650326203724, 0.63607444157126180606, -3.4603712257763543647, 0.59785942255448909499, 1.3000519917266020187, 0.20600315562279711279, -0.13365344893223751743, -0.12218157964756973244, -0.086911615691214527679, -0.048264014137718067032, -0.000925291428614517103, 0.034727432271574647238, 0.25913187788518116106, -0.64249780734054751097, -0.46580013275389248184, 0.91624666289868705871, -0.14319611662837886112, 0.98289687934337677255, -0.36758238471771215172, -0.68466392401901488007, -0.22582052905376964747, 0.094179223692724217676, 0.20228502959942312733, 0.12920229939067548464, 0.010024096337075299665, -0.073211401902214909287, 0.55766646766642036237, 1.0252883697852994693, -1.6334274145973850523, -11.843026230898386686, 23.914123502212159877, -11.831923697223178493, -1.8037649757800509054, 0.72447684364975339921, 0.77097274602750831551, 0.23209736923060439451, -0.052626627553616148192, -0.047673431268642298131, 0.0034666857017544979933, -0.014215329751679252335, -0.27226665887599121874, -0.12960026618987330327, 1.0008784073190093444, 0.85720728119254918642, -2.924280099277808187, 0.84494733709438663016, 1.0112426048790439381, -0.072974190543932213049, -0.24258366530762243141, -0.12123997113863814457, -0.0068562371414098146824, 0.025296364075611369598, 0.018982119110264615569, 0.0092247568799992026023, 0.33030775148994667711, -0.46623252622354638408, -0.54647247556932687651, 0.62896789266215102909, 0.093197923519846237461, 0.67753285200727353388, -0.46142955012334158083, -0.48282136395738561863, -0.075079672050517465287, 0.14924793764551558262, 0.16643970733975477216, 0.068211649201345800342, -0.020223997201371680249, -0.064557337788871363427, 0.043963067952074878775, 0.10351364185356259906, -0.37833794594019032331, 0.10489834120168449461, 0.25870596206481472068, 0.10086005818661100686, -0.38841464455345248163, 0.068981492439907268754, 0.11493843364804452734, 0.0048850994988859754511, -0.0213095530782651775, -0.016831241086718660288, -0.002782491676300865354, 0.0078908473853396016856, 0.12467073057611742504, 0.11196210079437535212, -0.70574861561598267734, 0.11375229218499957606, 0.72654822062851132092, 0.11428047109813550819, -0.72601159176709606502, 0.041959326694944197589, 0.21475397630767095736, 0.02824156235246571417, -0.028572865009107391299, -0.013885712242143591344, 0.0014517339429192606566, -0.0022946521162912898555, -0.041052566777200233394, -0.41736006877886078215, 0.73759051327491242933, 0.90494838341644556756, -2.3944818098804572237, 0.91576612503926513309, 0.79487336788902107632, -0.32965274471518041244, -0.27784051633564221806, 0.0082720882815307626018, 0.079189317561418434699, 0.037994934698033909493, -0.00056779000805072966235, -0.020179637828404588251, -0.040195400966517989827, 0.37515551491671161877, -0.48590722588491169764, -0.04184665686760950104, 0.40352837551162817897, -0.067547505124260553977, -0.54380477081098466119, 0.31026652651376396275, 0.27949107709478987482, -0.051272037757133979985, -0.11768435158949211439, -0.059139279972848519951, 0.0040636160889542631558, 0.038203635725640019971, -0.21431493055186612295, -0.16625140266115420729, 0.97737358809084518008, 0.76135847677422541224, -2.74340252587771527, 0.75430394572349046545, 1.0043790687294131825, -0.07967763858511466013, -0.24560504576422778622, -0.079452729555499590486, 0.0049562433314346582974, 0.0094252717932015346591, 0.0063664206319837051903, 0.0095715828473096880441, -0.21465990445628144179, -0.17264385209268570676, 0.91994697733711849885, 0.88857530741711354683, -2.8499374014938441135, 0.86932016360446517922, 0.92107334829694409439, -0.10726895098316972799, -0.18280204004738051515, -0.08743553259093629626, -0.022596109356789347705, 0.0056129363604180181405, 0.015246967275884090037, 0.016441742889206197031, 0.013360608149205553674, 0.15109212656641990424, -0.33892380268554428424, -0.081377093586054452534, 0.50287668266353824365, -0.069827965151299173363, -0.31995288984306630153, 0.11928816459455590659, 0.02009343079282529046, -0.0091179793649821122559, 0.022338849994750338557, 0.0034209233109545526645, -0.010994674879344613905, -0.0019565885149250539753, 0.12218323912342465842, -0.099061879405615796412, -0.37066789757589135856, -0.20758471945988188345, 0.72472228852653508557, -0.32771208393964579297, 0.72752265849241237738, -0.20324759781731638331, -0.37397666363256515387, -0.13049951205621851402, 0.0089374313564565752155, 0.052281273280936783887, 0.042473956235444143559, 0.030704113946987389439, 0.067082382832050313493, -0.14548810655770674782, -0.2588404105722346138, 0.49358610141731240617, 0.83104146356841313015, -1.9774478722775248318, 0.82898581209323318753, 0.49736960026023424586, -0.24255273867873497484, -0.13598377223651689216, -0.0013357951348882087569, 0.031865314171157718692, 0.015053554973437731704, -0.0038514486811723762545, -0.033385335614450194208, 0.082888863978783353104, 0.07663126994429794725, -0.30105589422975637115, 0.065242810830871786343, 0.22122014687750121698, 0.067228398750568310605, -0.30331311403547639438, 0.062635701592960682516, 0.079215372649778542646, 0.0074058328957525582786, -0.015324074487074269835, -0.015983459539902482305, 0.0066869507674700082167, -0.39135054698549048835, -0.022113128463166865229, 0.52478068105541730848, 1.5176012667698912662, 0.48107698971374579511, -4.2158883805634230768, 0.45103759927280606679, 1.4775281599675187394, 0.51765247317961082807, 0.065417374377836678967, -0.14769970220016936935, -0.18034332120675131139, -0.10376247452749307043, 0.02358991506096207863, -0.061759940749282669914, 0.0017215722322262464686, 0.23769892244349899446, -0.19488272702262149361, -0.10997092711731165648, 0.24644600781352204999, -0.10552272521508955505, -0.18859347281284588593, 0.24158950419234878626, 0.014002802774535449623, -0.085344578241867058432, -0.022479756304870597561, 0.012086374955529666231, 0.014320193479107033718, -0.035227478053467646546, 0.086879757586346459397, 0.060077392886974743946, -0.30150505662707205445, 0.16251625555627610686, 0.052070223899884894581, 0.16629133379885507882, -0.29828185526938955086, 0.052487202912083766315, 0.079733359696217559476, -0.008464208400040097377, -0.017085047004822523364, -0.0057704778124512716084, 0.0063911254425306112184, 0.040664122986860493092, -0.07716804368617267984, -0.15878874788569202758, 0.25690145304073719323, 0.62145408151236392058, -1.3692911977681199165, 0.61990607567650457543, 0.26146320662580208039, -0.14602677214710577114, -0.071541188176709638591, -0.003938731179566755565, 0.020488399367527894424, 0.012770236752356840804, -0.0069060716406590463845, 0.16458347553385768713, -0.061109641905222229774, -0.36182677139848440362, 0.014698157727336342085, 0.77716534645039658802, -1.0661228647213523146, 0.77522554883379080071, 0.017785470951074588281, -0.35044699602235585001, -0.058384024779430665486, 0.098524108570673799012, 0.085354349843136112508, 0.012305312961730029489, -0.045228557076096825851, 0.19150694739448156922, -0.022784956265640168155, -0.3297297887969464214, -0.42023552934248298296, 0.53943289585548892529, 0.082440216052884374154, 0.54437273125074314706, -0.40923061145545930328, -0.31956545980005202079, -0.035007274787339212607, 0.08902541798616114832, 0.099008045810636033379, 0.03334601408973481701, -0.040193010522373735782, -0.35874632755549640262, 0.37932666205343340549, 0.56625552001256540535, -0.37240000335842249335, -0.20179325300627468098, -0.023580553308834791698, -0.20143856739937454625, -0.38825606822744618274, 0.52116423693049418464, 0.37637147278028626474, -0.078743111363161422123, -0.18544110490827792059, -0.089628763304781880472, 0.056462306281551619125, -0.18039494090381985592, -0.18525094270200789803, 0.19844388422867567656, 1.3106793731104906531, 0.59785942223476895574, -3.4882521222494795587, 0.58604683792762801708, 1.3008981656800577831, 0.21000397598054115766, -0.13365344892189320869, -0.12644095147210057672, -0.0887885531771052533, -0.037461251774168306816, 0.034290370073472214996, 0.41010130223196916699, -0.20376777738499179948, -0.7370638423500370795, -0.39825839213511798897, 0.98289687941157721784, -0.11316664744225973793, 0.98957138485429552155, -0.36952616124504961137, -0.68816335533937356139, -0.22582052914023853507, 0.100781581430341452, 0.21105753058932621991, 0.11744631826879665948, -0.073526299437889755528, -0.10193408718103548483, 0.90669338317258496573, 0.84303919815418859596, -1.7990462129031721261, -11.831923697243032834, 23.985075940725081978, -11.793528099294698919, -1.8009638106291365744, 0.71803039118483891201, 0.77097274611077859507, 0.23344835246985515265, -0.065166314213158132995, -0.072365223725397456733, 0.013094775292952854154, -0.015579117147201695243, -0.27338814215131068686, -0.10088106806595152631, 1.0100934163539758703, 0.84494733728047222154, -2.9389724609547109857, 0.84080043868735776513, 1.0164223009951651733, -0.067799327284411056715, -0.24258366516195453699, -0.12497796659942730924, -0.0082611750166016913177, 0.031374195297006905736, 0.026176171702350491816, 0.32360036036345823929, -0.067660238832555591459, -0.51273902630054268847, -0.48132259514568370706, 0.67753285189956446999, 0.115687309315463005, 0.68326752910267074537, -0.46361867334478951141, -0.48662687055066361497, -0.075079672328157071037, 0.15447130901640473977, 0.17043821490799290763, 0.052944723002189886252, -0.085591529007778097138, -0.048447660640566478363, 0.12056290982975921555, 0.084231972166367935495, -0.3842436093565335109, 0.10086005813337275971, 0.25859407079781776551, 0.10072531131806002069, -0.38876973555379601422, 0.068953073768702105584, 0.11493843427382847977, 0.0042934608754706081457, -0.022170854320146968219, -0.015554894885566512147, 0.0060618281991479377319, -0.05447126464622813885, 0.23975384902342350424, 0.060103230684898721614, -0.72842094594369388716, 0.11428047050216053049, 0.73725947603502628436, 0.12038390790446645839, -0.72652886121489768101, 0.039778917054236473394, 0.21475397620676273114, 0.029373870937676120618, -0.030101944410469295976, -0.01817666789747220471, 0.0028386542834535882091, 0.16106091954960194412, -0.30060682943611688467, -0.36496661000946944986, 0.78350704838518370199, 0.91576612491958664375, -2.3962527059868445889, 0.9116270573328723037, 0.79305063348052617922, -0.33039050521215590495, -0.27784051650501967456, 0.010819634012531036821, 0.084262744745060735863, 0.03681762571391607175, -0.026835312218432858661, -0.23749198276914371131, 0.29180282974141208419, 0.34854821057279772312, -0.52989827728109473171, -0.067547505303593546877, 0.39647041876736138777, -0.0682300411670459378, -0.5438981395082896908, 0.31034862757539943834, 0.2794910768735133777, -0.053273072026065723672, -0.12100775696232624057, -0.052449520310845050564, 0.045079754308408444274, 0.0073482958369327144518, -0.28253783155905526359, -0.10498927790680627481, 1.0033573931512904664, 0.754303945447371893, -2.757569936905208241, 0.74427107315920304487, 1.0019413400485481791, -0.079085289611168743673, -0.24560504625261880829, -0.078807183861709026451, 0.0099205307885423810177, 0.01762180913499534346, 0.0073069636027492763583, -0.036766173926848465248, -0.20141981348525497819, -0.12334704745309475749, 0.92240860449119155273, 0.8693201636716585412, -2.8655814856121626377, 0.86532618299702612763, 0.92568969847955084873, -0.1026083333578715201, -0.18280204011596457025, -0.091110997763397968963, -0.024549663000796636408, 0.011854733398185926183, 0.031078434777454810867, 0.016890996313513401139, 0.00027548552185153302434, 0.12070195377285693195, -0.31511912465967234498, -0.069827964967287270426, 0.50111601206265832253, -0.075331100266204945037, -0.32284495581585687107, 0.11872385419683055552, 0.02009343078755582504, -0.0085547729341936135095, 0.024987930353634849046, 0.0068055060024172133598, -0.017410731040797895047, 0.093977393918520488181, 0.043783367649609675198, -0.1189885234399802455, -0.37684158871543305125, -0.20852596151662378121, 0.72752265801549498292, -0.32199009710275516838, 0.73179125141657419551, -0.20324759785682167124, -0.3771878597809870115, -0.13174860500825233411, 0.016499506807441898654, 0.068782121112152996845, 0.054632995795606507927, 0.038876569820426268298, 0.006958739801573242921, -0.13660361584066402219, -0.24543079056367972379, 0.49617588966789794025, 0.82898581232805113039, -1.9785299481980722636, 0.82904712122406065689, 0.49736960063945495625, -0.24341076681935286397, -0.13659120580319844551, 0.0014378044264491995127, 0.036083068036602654327, 0.0057139571002335198119, -0.028103875793079512169, 0.009590313591646597649, 0.081013760109965249345, 0.062657016790412564666, -0.30407008197271689909, 0.067228399152186660293, 0.22289644858230361901, 0.067504624692223014204, -0.30331311338636046671, 0.062735011175756957846, 0.078346982237132073434, 0.0051551758398079494325, -0.015466062558812285546, -0.0065701719212198968906, -0.2432149190844784814, -0.17472116867994449962, 0.06437112116969319342, 0.52926508034227692789, 1.4880978877039299224, 0.45103759924792119484, -4.2332762594653932098, 0.44512925453953267052, 1.4775281602307008821, 0.51841777677617284503, 0.06533587765289065874, -0.14539795023010385067, -0.17110727839311778986, -0.082388545131014831457, 0.024896903945176282469, -0.091481193528459986419, 0.00142549363759408073, 0.23527057305908316565, -0.19067942327147074932, -0.10552272545585349162, 0.24837708982098546939, -0.10534701154212340601, -0.18859347302982581995, 0.24094465923906902005, 0.010485765472073577467, -0.091452361708993762379, -0.022498029719825268086, 0.033870673953418375945, -0.010292845829795676735, -0.0112902719315884012, 0.075804185430890375086, 0.051164459472490785019, -0.2983499586160704431, 0.16629133445970920024, 0.05284062947805032856, 0.16555629404036820151, -0.29828185519974709239, 0.05346244295258351098, 0.079403519472231703924, -0.012578560917094279772, -0.021085885685707815573, 0.0074170554196714508916, 0.028048170484962955951, -0.0030080332547679192352, -0.074390344667268321066, -0.14677426344031679895, 0.2621323207724663984, 0.61990607616738369057, -1.3713750953728172099, 0.61916206094794956627, 0.26146320699219427475, -0.14579356529534298614, -0.070608699732288254802, -0.0022106342298863872198, 0.019407166907319316901, 0.0045111783269599636803, 0.034789412881541775802, 0.10136126441393093767, -0.049333098878353222327, -0.34514657576371021719, 0.019542448584456917349, 0.77522554894725170804, -1.0677916477074427259, 0.77462245199101154025, 0.017785470968382077056, -0.34903321865351982778, -0.0536604370892012697, 0.10557737746243160692, 0.079903584734434610071, -0.040141063387451755839, 0.079105087130432280573, 0.09303422465573117639, -0.029533982769019341552, -0.31727618602056106845, -0.40934105460466790882, 0.54437273128326657545, 0.084181737059234498788, 0.54443198776409618134, -0.40923061178444719577, -0.31794509979804630451, -0.03092583011174591251, 0.093325582386188446482, 0.091795227324778655431, -0.010969267235359639323, -0.19379796666697940433, -0.093874671736899700791, 0.36915647989304789522, 0.5168504989140435546, -0.39118649808300964832, -0.201438567395088447, -0.020004537082456780306, -0.19991510484777139101, -0.38825606871096601802, 0.51856335940076336843, 0.36730247396858878295, -0.09348501941180992203, -0.18345825918707420188, -0.011565859914280723561, -0.073371250192878600949, -0.12631166338294319562, -0.1463299238425538884, 0.20267891588229908217, 1.3000519922817250773, 0.5860468377029537379, -3.4934031249499555294, 0.58617310228259666971, 1.3008981658281508764, 0.20600315564593232875, -0.14048663855183973759, -0.12922186451435274668, -0.077750948834935870724, -0.00026069857469762018447, 0.24240733707963799115, 0.10440222087544866192, -0.22249201338449109411, -0.68416000915373920144, -0.36758238462102865718, 0.98957138477134021226, -0.11367941288013842793, 0.98854283811715193409, -0.36952616121256820492, -0.68466392395528563597, -0.21564805960640703875, 0.11348388915438652191, 0.19966135185145272857, 0.028947894217408887824, -0.14160929306208286382, 0.22471705870184360831, 0.78533650535327925724, 0.72770300098093587682, -1.8037649756967499837, -11.79352809948842129, 24.008684742923396271, -11.790570152338547771, -1.8009638106294461046, 0.72447684365424180886, 0.77291642467360666124, 0.21155373984219044692, -0.098026494163142213756, -0.020491584401651437991, 0.067100119209211273041, -0.11707282868827861844, -0.2563801911457650462, -0.079989583955919915659, 1.0112426052019933831, 0.84080043863936182458, -2.9387959904873230954, 0.84333880318231557727, 1.0164223009766744088, -0.072974190815160558388, -0.24982509366382188443, -0.12481764541285343684, 0.0059984947352170227275, 0.052846494248618008926, 0.12322835990230218695, 0.15078429940063017844, -0.068234485102640740628, -0.48079807376019351173, -0.46142955010786546044, 0.683267529013248609, 0.115679207901786818, 0.6818870562633571053, -0.46361867336950840501, -0.48282136355595112187, -0.066617253859714703479, 0.16200878721690376016, 0.15350356010019911013, -0.037467451026396182689, -0.03527458717267160232, 0.003354709798335893476, 0.11856804238898238646, 0.071528371478445193499, -0.3884146446602693703, 0.1007253113986274623, 0.25941490101152570036, 0.10117135631221173264, -0.38876973502442374597, 0.068981491497603769636, 0.11508435499641408994, 0.003829877169965246065, -0.022314874442393918541, -0.0085162200626119118757, -0.04100688804096498663, 0.021271206667908212212, 0.21458533164615137268, 0.041478178238836510561, -0.72601159105753254597, 0.12038390743319360432, 0.73924788935107632604, 0.11936576614937496887, -0.72652886144759853959, 0.041959326920153694473, 0.21555105700825813453, 0.023801818669878883894, -0.038620107261718272795, -0.005002599704183031265, 0.088328592676222603131, 0.019428202153814026071, -0.28101761349824361247, -0.33145477629718633761, 0.79487336772556060893, 0.91162705761605256072, -2.4013970767788790361, 0.90984220240021995707, 0.79305063344100212852, -0.32965274456253679247, -0.27455328653533073169, 0.01756674772970949136, 0.083530085094507156018, 0.0014149777821293239681, -0.12287232431171700442, -0.059377732240470854075, 0.28131442809230750024, 0.31243321342367424975, -0.54380477050485775958, -0.068230041316733353041, 0.39775153176884497919, -0.068264110365566885164, -0.54389813964365163468, 0.31026652669082449565, 0.27640170938002234235, -0.059791249384256360544, -0.11709903322922786406, 6.3314453446071676697e-05, 0.035841878793263483305, -0.07201041139136694913, -0.25059525487248640596, -0.079973033318217714993, 1.0043790691020484385, 0.74427107299482286873, -2.764636802091467338, 0.74251690402853076645, 1.0019413394706619957, -0.079677637741624493528, -0.2448679146413994534, -0.072955541516352168108, 0.018164477899580463993, 0.014653360317146410519, 0.02907439464182531158, -0.080332472602779966819, -0.19276860345355795179, -0.1130407573571974994, 0.92107334821514841305, 0.86532618313331355342, -2.8653897903966103478, 0.8676437726149552887, 0.92568969836408898644, -0.10726895097107964905, -0.18942284405131076808, -0.091583768605043622268, -0.011830835317608812129, 0.040150819426317922722, 0.0057275874563087253097, -0.012163433686581778814, 0.027627167228075514421, 0.12642977659163226467, -0.3199528896797245725, -0.075331100495554997876, 0.49910764160580761706, -0.075437351446498504548, -0.32284495576556193619, 0.11928816473420011945, 0.023351391109120139128, -0.0020685993149977561362, 0.026988442923183898758, -0.020224893184269770313, 0.054632996128837207905, 0.068782120411993469267, 0.016499506915268434548, -0.13174860474645611053, -0.37718785961468132006, -0.20324759794151528447, 0.73179125127226685166, -0.3219900973448318604, 0.72752265836684959499, -0.20852596139603976622, -0.37684158852592286593, -0.11898852355588426799, 0.043783366873775333983, 0.093977394441452422047, 0.0057139571472244457193, 0.036083067890113749987, 0.0014378046515911163412, -0.13659120606870037151, -0.24341076658838259328, 0.49736960049480621882, 0.82904712128008739569, -1.9785299480746556533, 0.82898581199133569708, 0.49617589011192725357, -0.24543079100742551391, -0.13660361552187766376, 0.0069587396796109103267, 0.038876569833704986701, -0.0065701718372059006099, -0.015466062904101829031, 0.0051551764818398839937, 0.078346981485250433153, 0.062735011812960747579, -0.30331311374927522051, 0.067504624683012923181, 0.22289644896412441621, 0.06722839849242219501, -0.30407008118325090651, 0.062657016043907595715, 0.081013760691085734411, 0.0095903132582033749631, -0.028103875699621746337, -0.082388545173833704904, -0.17110727831369018648, -0.14539795042715578099, 0.065335877979505477264, 0.51841777673549949146, 1.4775281601143901433, 0.44512925431101874052, -4.2332762592613635277, 0.4510375994699500346, 1.4880978874703449932, 0.52926508022597507086, 0.06437112123948864062, -0.17472116848262952971, -0.24321491920845692003, 0.03387067387884269698, -0.02249802924961850914, -0.091452362732698633252, 0.01048576656907264544, 0.24094465840717854688, -0.18859347242926624566, -0.10534701169122906439, 0.24837708955680962153, -0.10552272499543824402, -0.19067942407329915855, 0.23527057397056791443, 0.00142549303268650813, -0.091481193199544119654, 0.024896903825681698103, 0.0074170554841802615112, -0.021085885950324449734, -0.012578560394640093956, 0.079403518795186062151, 0.053462443575343319979, -0.29828185561150871719, 0.16555629418367937, 0.052840629701293502507, 0.16629133383440500293, -0.2983499577516631196, 0.051164458596944489832, 0.07580418608288996829, -0.011290272229666362819, -0.01029284576998255353, 0.004511178379505776867, 0.019407166651844755256, -0.0022106336548037212701, -0.070608700511393202026, -0.14579356456016928645, 0.26146320645427617579, 0.61916206119288363929, -1.3713750951915362197, 0.61990607552213095044, 0.26213232173731304542, -0.14677426447628350026, -0.074390343903536937376, -0.003008033570127065584, 0.028048170537000149949, -0.040141063426151507576, 0.079903584709994909807, 0.10557737767832379994, -0.053660437238834984841, -0.34903321856751368113, 0.017785470734255799663, 0.7746224521774255356, -1.0677916476874274032, 0.77522554881302385699, 0.019542448682775898777, -0.34514657592263375818, -0.049333098700805516623, 0.10136126441600508408, 0.034789412838923408633, -0.010969267304790258563, 0.0917952274555804415, 0.093325582298913245527, -0.030925829920597565359, -0.31794510008405829637, -0.40923061169474117538, 0.54443198791152846994, 0.084181736816209454255, 0.54437273165844879408, -0.40934105513515700592, -0.3172761857351794057, -0.02953398281181135801, 0.093034224813206597982, 0.079105087013889546776, -0.011565859951400189048, -0.18345825903300577719, -0.093485019373606564885, 0.36730247367154122973, 0.51856335949150189624, -0.38825606864651285299, -0.19991510472603069015, -0.020004537321501817543, -0.20143856728999173789, -0.39118649813368011614, 0.51685049895912293927, 0.369156479871121046, -0.093874671622998839826, -0.1937979667492205349, -0.00026069859669705999772, -0.077750948744890746589, -0.12922186454385417576, -0.14048663854862786238, 0.20600315548910691499, 1.3008981659640677098, 0.58617310224419905129, -3.4934031248371746337, 0.58604683756859665689, 1.3000519923642235298, 0.20267891576808519383, -0.14632992384717691259, -0.12631166325092560432, -0.073371250243694646764, 0.02894789413504468556, 0.19966135201225862494, 0.11348388898351413623, -0.21564805932350558648, -0.6846639242708694173, -0.3695261611039131755, 0.98854283805577780608, -0.11367941267567430175, 0.98957138466986993564, -0.36758238474890564396, -0.68416000899361428811, -0.22249201343854488333, 0.10440222085881813163, 0.24240733709093825743, -0.020491584423554982941, -0.098026494098400140942, 0.21155373969823410008, 0.77291642491332313014, 0.72447684338935303128, -1.8009638104335297104, -11.790570152428260897, 24.008684742942108414, -11.793528099469801518, -1.8037649758118072807, 0.72770300121022435391, 0.78533650522218112489, 0.22471705865468186158, -0.14160929301631772215, 0.052846494207743892335, 0.0059984950642947305194, -0.12481764603564970972, -0.24982509311300182819, -0.0729741913528989955, 1.0164223015236699688, 0.84333880306330177845, -2.9387959907452256836, 0.84080043888396760465, 1.0112426048625091646, -0.079989583326879762359, -0.25638019180700283295, -0.11707282831646441401, 0.067100119116900378002, -0.037467450938841774521, 0.1535035599743262702, 0.16200878723646131019, -0.066617253762666611694, -0.48282136365234273878, -0.46361867340729923104, 0.68188705642313329758, 0.11567920776683442152, 0.68326752907868748554, -0.46142955030260462967, -0.48079807354330134084, -0.068234485147817547368, 0.15078429940147816679, 0.12322835989620337105, -0.0085162200242607684342, -0.022314874698862705721, 0.00382987776759176465, 0.11508435427670458961, 0.068981492089201262852, -0.38876973540041148691, 0.10117135636051612091, 0.25941490125494220953, 0.10072531100082431077, -0.38841464408451675805, 0.071528370800332072288, 0.11856804291845517152, 0.0033547095398020827968, -0.035274587110670517498, -0.0050025997283162587995, -0.038620107202586351214, 0.023801818589758386852, 0.21555105709191821339, 0.04195932675189972122, -0.72652886117187520654, 0.11936576604475841734, 0.73924788917342465489, 0.12038390768972935529, -0.72601159120329405994, 0.041478178219738690014, 0.21458533161866372163, 0.021271206821067908943, -0.04100688812313292575, 0.0014149778533894784294, 0.083530084887588118558, 0.017566747992109235654, -0.27455328688652952396, -0.32965274396282318037, 0.79305063274897580339, 0.90984220274491112157, -2.40139707656488266, 0.91162705719977055185, 0.7948733678960141491, -0.33145477634780867726, -0.28101761340740805117, 0.019428202171274298882, 0.088328592621104082272, 6.3314395716192639792e-05, -0.11709903315828637527, -0.059791249429743502375, 0.27640170940009822775, 0.31026652675323418418, -0.54389813968597389149, -0.068264110494280563768, 0.39775153193172807509, -0.068230041260605930886, -0.54380477065543531001, 0.3124332135968317381, 0.2813144278169237289, -0.059377732051669455948, -0.1228723243502332918, 0.014653360184187230394, 0.018164478213335246565, -0.072955541887938513135, -0.24486791421623857468, -0.079677638224196878358, 1.0019413398723058251, 0.74251690386586055581, -2.7646368023153287119, 0.74427107361452882195, 1.0043790683615598791, -0.07997303267661683146, -0.25059525531172388568, -0.072010411200274362931, 0.035841878749077279998, 0.040150819416680874618, -0.011830835263019312542, -0.091583768537209536698, -0.18942284432456280019, -0.1072689508830291516, 0.92568969850278270961, 0.86764377263079661695, -2.8653897905377276878, 0.86532618316303300254, 0.92107334827248354969, -0.11304075734388233099, -0.19276860347897636916, -0.080332472664029680454, 0.02907439468794155088, -0.020224893335265968497, 0.026988443309556926081, -0.0020685997778497381382, 0.023351391709897677956, 0.11928816402460001855, -0.32284495525785733561, -0.07543735173184756837, 0.49910764157038844946, -0.07533110002671883354, -0.31995289023963724429, 0.1264297771529977743, 0.027627166612316667732, -0.012163433323862170993, 0.0057275873886075655272, 0.030704113844387342924, 0.04247395660115643734, 0.052281273067529041021, 0.0089374310369128157072, -0.13049951191933975569, -0.37397666329557149867, -0.20324759796276742341, 0.72752265835589280396, -0.32771208382279776172, 0.72472228852644693387, -0.20758471958855381767, -0.370667897224225662, -0.0990618798900514963, 0.12218323934606128978, -0.0038514486341218495853, 0.015053554792484611161, 0.031865314492179511174, -0.0013357955135878708808, -0.13598377185637872122, -0.24255273901726948083, 0.49736960051328998889, 0.82898581189031461669, -1.9774478721041670592, 0.83104146345531204609, 0.49358610151272896971, -0.25884041067954305371, -0.14548810648858759298, 0.067082382814456623099, 0.0066869508330642838875, -0.015983459774631291611, -0.015324074090527115347, 0.0074058324234223026419, 0.079215373131676106166, 0.062635701125957912727, -0.30331311356091594789, 0.067228398283996487761, 0.22122014720949989552, 0.065242810739820120136, -0.30105589428465201474, 0.076631269973928162198, 0.082888863994662054235, -0.033385335623804211769, 0.023589915153146654575, -0.10376247474991616426, -0.18034332109842723946, -0.14769970200019730266, 0.06541737397661341602, 0.51765247342053855029, 1.4775281600704106566, 0.45103759915457980467, -4.2158883806608224987, 0.48107698978281210334, 1.5176012670140552885, 0.52478068065737493164, -0.022113128252242986171, -0.39135054701634491847, 0.014320193420738536183, 0.012086374936040363698, -0.022479756154844213883, -0.085344578140452237447, 0.014002802475280115568, 0.24158950434313361533, -0.18859347290108147144, -0.10552272529295715631, 0.24644600818527870323, -0.10997092743674047322, -0.19488272673612616614, 0.23769892197906944409, 0.0017215725736524224287, -0.061759940825166566303, 0.0063911255103597536947, -0.0057704780540754471371, -0.017085046648981017797, -0.0084642087498220195535, 0.079733360044252285515, 0.052487202594326536065, -0.29828185505267684752, 0.16629133367554030376, 0.052070223946244768853, 0.16251625555554058411, -0.30150505658349652238, 0.06007739278800620919, 0.0868797576578380637, -0.035227478071152971972, -0.0069060715458039486497, 0.012770236431497541604, 0.020488399823035775837, -0.0039387315921958175602, -0.071541187822687091202, -0.14602677238582170705, 0.26146320661585131795, 0.6199060758782838354, -1.3692911979846742465, 0.62145408167509597064, 0.25690145290716209781, -0.158788747803653707, -0.077168043691633395054, 0.040664122973847791065, -0.045228557047476275477, 0.012305312838639607953, 0.085354350040777945963, 0.098524108367397542074, -0.058384024612201983195, -0.35044699602502332736, 0.017785470693007125048, 0.77522554923311415287, -1.0661228650339564794, 0.77716534662601766126, 0.014698157559694460805, -0.36182677129146678663, -0.061109641877103069996, 0.16458347549773369445, -0.040193010560695088818, 0.033346014160088172096, 0.099008045831255164781, 0.089025417825546571127, -0.035007274706254178087, -0.31956545957842669159, -0.40923061190883702753, 0.54437273176016276466, 0.082440215618466414216, 0.53943289613798572901, -0.42023552945894510069, -0.32972978885184184294, -0.022784956138111347318, 0.19150694733870143871, 0.056462306249046995132, -0.089628763179916609571, -0.18544110516662959398, -0.07874311095319481979, 0.37637147241126123198, 0.52116423728759753153, -0.38825606885310431915, -0.20143856680312346996, -0.023580553687066251739, -0.20179325265944991763, -0.37240000358867986208, 0.56625552015239999371, 0.37932666192546521389, -0.35874632750672824644, 0.034290370108570666918, -0.037461251905791199956, -0.088788553052577351421, -0.1264409514383001143, -0.13365344888659000389, 0.210003975730040926, 1.3008981658778784318, 0.58604683791577349972, -3.4882521222384279547, 0.59785942193532914946, 1.3106793734992718825, 0.19844388415999120179, -0.18525094279131340547, -0.18039494087756252627, -0.073526299457047236685, 0.11744631838274828539, 0.21105753044751707215, 0.10078158134428620574, -0.22582052893263590554, -0.6881633553904974443, -0.36952616116643915989, 0.9895713845787962315, -0.1131666470503201738, 0.98289687905978739835, -0.39825839205183649616, -0.73706384234671284972, -0.20376777723429267963, 0.41010130212708034714, 0.013094775248488271097, -0.07236522355970374798, -0.065166314457994700682, 0.23344835263145125026, 0.77097274602584964232, 0.71803039124254497416, -1.8009638104856140473, -11.793528099584619895, 23.985075940915557169, -11.831923697309072452, -1.7990462129662692092, 0.84303919824828954521, 0.906693383190263269, -0.10193408721935791339, 0.026176171652202522705, 0.031374195364887856596, -0.0082611750676548443423, -0.12497796644391677423, -0.24258366520205712491, -0.067799327535955272728, 1.016422301152997143, 0.84080043862399655996, -2.938972460668685116, 0.84494733683449518846, 1.010093416838719671, -0.10088106838591955061, -0.27338814213638323869, -0.015579117091575256024, -0.08559152904883614299, 0.052944723120907997982, 0.17043821473441642111, 0.15447130913620910064, -0.07507967237082095735, -0.48662687037198965045, -0.46361867353338881426, 0.68326752911464982976, 0.11568730934814457079, 0.6775328516925352984, -0.48132259493195972411, -0.51273902638895174633, -0.067660238587357451889, 0.32360036018303389715, 0.0060618282085398052239, -0.015554894885793663778, -0.022170854311916270846, 0.0042934607803615017912, 0.11493843443198226151, 0.068953073611478191141, -0.38876973545051235481, 0.10072531137086580821, 0.25859407058346089991, 0.10086005846201906433, -0.38424360980681426803, 0.084231972627639220552, 0.12056290955625997841, -0.048447660571323242218, 0.0028386542164581733509, -0.018176667733604441601, -0.03010194465153906776, 0.029373871261511169672, 0.21475397590049649654, 0.03977891717021173057, -0.72652886109386294322, 0.12038390772625277847, 0.73725947603153696441, 0.1142804707555610394, -0.72842094624398745495, 0.060103230776303098737, 0.23975384907914523702, -0.0544712646761824279, -0.026835312222587389547, 0.036817625733468584137, 0.084262744628527050383, 0.010819634268753852355, -0.27784051675312199103, -0.3303905050538055721, 0.79305063324077962328, 0.9116270577261336161, -2.3962527063942609118, 0.91576612532375700759, 0.78350704796419623488, -0.36496660965659899478, -0.30060682964958879149, 0.16106091961517435274, 0.045079754294926555058, -0.052449520257956128155, -0.12100775701231547554, -0.053273072016317160604, 0.27949107682452517532, 0.31034862751620539978, -0.54389813916186080522, -0.068230041528361054515, 0.39647041887670586569, -0.067547505210252031804, -0.52989827731395910959, 0.34854821043443562356, 0.29180282983354327575, -0.23749198276918889738, 0.0073069636503833327798, 0.017621809109657667058, 0.0099205305963489982812, -0.078807183467383384867, -0.24560504663213011955, -0.07908528942577956744, 1.0019413399462562264, 0.74427107329200681285, -2.7575699369918287296, 0.75430394562196689812, 1.0033573929163273064, -0.10498927775739684554, -0.2825378317069789369, 0.0073482959222186628942, 0.031078434738564430367, 0.01185473350155979988, -0.024549663093406660225, -0.091110997696063761309, -0.18280204013496839655, -0.10260833339967073696, 0.9256896984655599292, 0.86532618298850050298, -2.8655814855081969128, 0.86932016360089270357, 0.92240860449862349668, -0.12334704739303896381, -0.2014198135862246275, -0.036766173881820254921, -0.01741073099444551131, 0.0068055060606034530404, 0.024987930085376107436, -0.0085547727902191405408, 0.020093430855730981532, 0.11872385407003169244, -0.32284495552899972193, -0.075331100590727978483, 0.50111601232657720928, -0.069827965460068225356, -0.31511912412082759838, 0.12070195358570648403, 0.00027548552762145966152, 0.016890996301715196365, 0.012820587081319978137, 0.022479330661507094385, 0.038109084803375763151, 0.052125101301552248523, 0.010404890287563865053, -0.13049951215065117016, -0.37718785927003500857, -0.20852596218781238591, 0.72472228894371104602, -0.32539309463559734636, 0.72652517166025176731, -0.22502133798262802244, -0.36949799420651924242, 0.041642013530412379241, -0.0027526769575066663102, 0.00039688920245097030924, 0.014634219138954084743, 0.031085734841664474509, -0.0014583231038394067758, -0.1359837721878486283, -0.24341076637097031066, 0.49617589039115628236, 0.83104146321002825282, -1.9710559094550141879, 0.83264729513225588242, 0.4681275378119977848, -0.29374783278908217499, -0.027078486001543295575, 0.006143615098706170681, -0.0023623368192912824227, -0.0149959310649160453, -0.013567441602591082464, 0.0077990499850280319699, 0.079215372815480689095, 0.062735011817284386249, -0.30407008083371772633, 0.065242810278793972323, 0.22108712642882835664, 0.072688227172875374826, -0.29558026572061524284, 0.083335155665326163255, 0.033048870852571264611, 0.06521587772852499576, -0.033995540844600151409, -0.12723532988109012565, -0.18407470899065619951, -0.14618485126781630146, 0.065417373837626149879, 0.5184177769085748233, 1.4880978879049788777, 0.4810769895032279142, -4.1685627889348051411, 0.52644570896702680418, 1.5155154497041656292, 0.38951250153573196977, -0.37782026228202775231, -0.0026359997958401758592, 0.015176650888230855049, 0.013733128527993907972, -0.020889563925772815928, -0.084622324441117766081, 0.014002802074601901003, 0.24094465923012736708, -0.19067942341297305919, -0.1099709277758260384, 0.24081769894780374885, -0.10795885334753035911, -0.1756534157461301926, 0.29251740240355927902, -0.1252175610792529703, -0.0011262166046293274873, 0.0055514527844287538463, -0.0033751814991920059753, -0.016400974512114335768, -0.009048177905206341437, 0.079733360235284739259, 0.053462442947703393836, -0.29834995777373335413, 0.16251625544073219909, 0.046261354793490715165, 0.16531743220728847832, -0.28802460102573351763, 0.0873339714857002819, 0.016447952755327614749, -0.0068516124236201519579, 0.0022242556466482623495, 0.012780083386770462253, 0.018721437321673074089, -0.0047472848024099449649, -0.071541187852659546276, -0.14579356478840430067, 0.2621323211174378387, 0.62145408188536532812, -1.3690141356432343756, 0.6181455559213345774, 0.24325693531243872125, -0.16834303345073156621, -0.013114264837761223717, -0.023649503606668537126, -0.02038909044316096833, 0.017011661885555414558, 0.083095186320722783102, 0.096381566139895485934, -0.058384024572439810474, -0.34903321855251828731, 0.019542448174540733019, 0.77716534685569649277, -1.0681000917899450275, 0.7596027178372812827, -0.034447204014281300466, -0.38159249954364288238, 0.18228094322765189816, -0.033224148322044601278, -0.0068918263696326814358, 0.041890020015870346737, 0.09743677149986121333, 0.086465543707862371758, -0.035007274763055401534, -0.31794509999450415449, -0.40934105510784429827, 0.53943289628041191097, 0.069174167428213820252, 0.51435064550396758509, -0.45609011916541591924, -0.29683378480917915754, 0.20384681425162456958, 0.046684515790197160523, -0.0019456116954992422677, -0.094974231119207969121, -0.17622044666978131655, -0.073069943614307478197, 0.37637147250045244196, 0.51856335943770370811, -0.39118649877433608886, -0.20179325219288912763, -0.011016767742452637791, -0.15748525718872113943, -0.30307898261007570895, 0.57685460778112229718, -0.10021353115279418855, 0.034727432237094013612, -0.00092529122607496773478, -0.048264014339353927818, -0.086911615590171270584, -0.12218157981880686802, -0.13365344874794579666, 0.20600315534188148869, 1.3000519921031341575, 0.59785942230295474609, -3.4603712254144349814, 0.63607444100913490548, 1.3161109652692357663, 0.096982683399347402453, -0.33123893970587686697, -0.073211401804258668946, 0.010024096109157346057, 0.12920229959908688389, 0.20228502936954842473, 0.094179223917748758343, -0.22582052908985852935, -0.68466392406256626479, -0.36758238462440412375, 0.98289687901689215543, -0.14319611617124103353, 0.9162466626227240285, -0.46580013277322801502, -0.64249780718451976469, 0.25913187780519997272, -0.014215329607298862746, 0.0034666853138932943353, -0.047673430885801899859, -0.05262662779436062116, 0.2320973694112535568, 0.77097274591137932021, 0.72447684361683817311, -1.8037649754969373728, -11.831923697554374897, 23.914123502284230227, -11.843026230909821095, -1.6334274143975677784, 1.0252883696053245455, 0.55766646770285954737, 0.0092247569032901828578, 0.018982118939699383492, 0.025296364493363351911, -0.0068562377739576872704, -0.1212399704534456446, -0.24258366559919419814, -0.072974190890766690853, 1.0112426054599201741, 0.84494733656176390646, -2.9242800985075336762, 0.85720728008894508942, 1.0008784082447059749, -0.12960026656735920692, -0.27226665882416972719, -0.064557337775166714899, -0.020223997202369958098, 0.068211649380979427759, 0.16643970700640822735, 0.1492479378529651679, -0.075079672272221537055, -0.48282136369625944239, -0.4614295500973397135, 0.67753285162271137398, 0.093197924025579026441, 0.62896789215057880007, -0.54647247516256935995, -0.46623252641854040501, 0.33030775154214681022, 0.007890847307667030186, -0.0027824913719310903105, -0.016831241637287400192, -0.02130955240662542674, 0.004885098748225294267, 0.11493843442599631366, 0.068981491943454459337, -0.38841464456829127849, 0.10086005865079009092, 0.25870596118933525398, 0.10489834237641436354, -0.37833794704426076949, 0.10351364251711563103, 0.043963067764477441191, -0.0022946522116566187705, 0.0014517340849267527936, -0.013885712070520504074, -0.028572865471434082119, 0.028241562676963642697, 0.21475397627172621018, 0.041959326396854845775, -0.72601159128122516595, 0.11428047086982923153, 0.72654822051091616508, 0.11375229241864548424, -0.70574861572198144266, 0.11196210070080897525, 0.12467073065594611037, -0.020179637915645965568, -0.00056778983149777424722, 0.037994934642151492743, 0.079189317471977757612, 0.0082720882495392195066, -0.27784051613508292666, -0.32965274463629978818, 0.79487336727689261728, 0.91576612578030813783, -2.3944818105600456093, 0.90494838421974699294, 0.73759051260324137722, -0.41736006852402168787, -0.041052566805497403557, 0.038203635683454702232, 0.0040636160614977955008, -0.05913927975420465305, -0.11768435184335779287, -0.051272037762838305885, 0.27949107738342499063, 0.31026652616730460155, -0.54380477054532183612, -0.067547505265846852218, 0.40352837568500088405, -0.041846657074340223859, -0.48590722578485090466, 0.37515551486807374726, -0.040195400943757959855, 0.0095715828836646123234, 0.0063664203180900363205, 0.009425272424178076508, 0.0049562425843132206643, -0.07945272877289433211, -0.24560504644005223818, -0.079677638143433468509, 1.0043790686306215409, 0.7543039454520781284, -2.743402525590801222, 0.76135847678134704886, 0.97737358800169549244, -0.16625140266374743847, -0.21431493054281078314, 0.016441742919099611164, 0.015246967196156093263, 0.0056129363593053708398, -0.022596109222690424767, -0.087435532691495274071, -0.18280203999862515429, -0.10726895110175482806, 0.92107334848879618416, 0.86932016340361217832, -2.8499374012655147581, 0.88857530718779575984, 0.91994697738580766266, -0.17264385198190690396, -0.21465990451899019664, -0.0019565883792864425215, -0.010994675296276176096, 0.0034209239885658786554, 0.022338849148156147972, -0.0091179786328221875541, 0.020093430474262544833, 0.11928816452355020528, -0.3199528894797212808, -0.069827965814938275368, 0.50287668363655890769, -0.081377094659465673554, -0.33892380202863109373, 0.15109212645719116708, 0.013360608110618062833, -0.0023671027305798520921, 0.011800553904372824976, 0.025466795626321602625, 0.040369327756369290106, 0.05212510019773094333, 0.0089374315107479170711, -0.13174860519596567898, -0.37684158816694568461, -0.20758471952075555067, 0.72652517184659282101, -0.32894962122346638189, 0.7227720247214954874, -0.25089158841897896135, -0.29771643190496938081, -0.0012011498837710592527, -0.0022496840661741040156, 0.0015197111757468458743, 0.014873716275732419903, 0.031085735119279023914, -0.0013357953425145205166, -0.13659120632011623186, -0.24543079110401783138, 0.49358610136202213381, 0.83264729570666196068, -1.9665237249030416145, 0.84847845735159255209, 0.46063761989737134206, -0.33056589774001227111, 0.0016449602058736171976, 0.0041698296269038646542, -0.0023523779925241597617, -0.014782940961982257363, -0.013567441114144741798, 0.0074058324549423244776, 0.078346981463793596845, 0.062657016001893647839, -0.30105589429972590132, 0.072688227757490037062, 0.21906453537400796816, 0.056068575440538205579, -0.32125431757950090317, 0.1515887592764523506, 0.060838638994589890052, 0.012062850019855580683, -0.05453518001581317276, -0.13115916610761221839, -0.1840747085969739183, -0.14769970199835699698, 0.065335877693632110619, 0.52926508004863781487, 1.5176012669045848558, 0.52644570934696388598, -4.1332126421599761912, 0.53050990772950235019, 1.4089671410491160231, 0.024773253899511023246, -0.0083874539671664940371, 0.0065325806870422855471, 0.017354141723097359779, 0.013634987888763626385, -0.020889563388600929095, -0.085344578050359914956, 0.010485765882148903594, 0.23527057333343270873, -0.19488272681471716097, -0.10795885263420458566, 0.25057269013005123437, -0.061977317014788028959, -0.14295758602735036957, 0.08946103673194524819, -0.0038454669147776009662, 0.0039255192328471960952, 0.0056844729921939965434, -0.004221541283999173666, -0.016400974528703062688, -0.0084642087256531113593, 0.07940351889712732314, 0.051164459187807570517, -0.30150505650877207353, 0.16531743179895505569, 0.053935726705104082568, 0.17219850736698821825, -0.32878348131834483903, 0.13171884364933619804, -0.0024087819562869450599, -0.0037667926965384932597, 0.0021873052422516979336, 0.012221145376234560098, 0.018721437068136694698, -0.0039387313643764430837, -0.07060870038836253304, -0.14677426377975574812, 0.25690145283097620688, 0.61814555560104633702, -1.363776817267805086, 0.63775011211758092511, 0.23221260344274546883, -0.18757801435900678388, -0.0045501209525616540669, -0.020805709918775887324, -0.018217018201922410836, 0.01725752800970642778, 0.083095186244093066641, 0.098524108536896026767, -0.053660437346131815983, -0.34514657547887772626, 0.014698157606917594195, 0.75960271732479967, -1.099975094129398645, 0.73995945409473229404, -0.013830050753211801939, -0.16154693988813587091, -0.016620699298873567451, -0.01863664877436901901, -0.0014246201824293094288, 0.042171385865734295484, 0.097436771204411826153, 0.089025418006517448299, -0.03092582973108648145, -0.31727618594841061617, -0.42023552951753778695, 0.51435064553382114916, 0.035685276561672953088, 0.51089514583201167675, -0.41665234415349361452, -0.076198816779720859871, 0.018690788378694117577, 0.028055251717792117211, -0.0060963300381518675808, -0.093325157139115544336, -0.17622044689038862675, -0.078743110773486280074, 0.36730247381382341576, 0.51685049935786053776, -0.37240000387894173217, -0.1574852577133288023, 0.036940105959566857385, -0.16979508511211277577, -0.51637208197713024838, 0.61537832053036567448, 0.020599270941787352351, 0.015099834816313406974, -0.0064526075072554155837, -0.04732662170646760591, -0.086911615779179912655, -0.1264409514257146816, -0.14048663829414534399, 0.20267891587923406171, 1.3106793732324928392, 0.63607444079125896597, -3.414090884916893831, 0.66055709110957516739, 1.2903367285520810359, -0.3022133046737014439, -0.037194314289778790394, -0.037294019746986914465, 0.019158852854915794928, 0.12785593815602971923, 0.20228502943672452274, 0.10078158134647378918, -0.21564805934573205715, -0.68416000918882968751, -0.39825839201162832603, 0.91624666283991362814, -0.1949392542075295276, 0.97908561135221361393, -0.3505186955453781783, -0.445863745593015115, -0.029601836397763165581, 0.012902257016792016003, 0.012130839768201057485, -0.046846786192182798825, -0.052626627579354788578, 0.23344835251135578891, 0.77291642486361011866, 0.72770300084221772874, -1.7990462128953148557, -11.843026230579159375, 23.822411323871531152, -11.999686437718759535, -1.5410168648486979315, 1.7196781344625902133, -0.0015641853860349106894, 0.0089435477435147840708, 0.021461950021310721021, 0.027604560822334787279, -0.0068562372559389447857, -0.12497796640893239772, -0.24982509350266063342, -0.07998958386026301659, 1.0100934166456108088, 0.85720728082705754947, -2.8915303008000079465, 0.93291108485666485173, 1.0145155868571824609, -0.51581275830119865944, -0.027492858291529639198, -0.039477821511638207241, -0.011979502838925189245, 0.067913718149501284271, 0.16643970725991066861, 0.15447130909169121127, -0.066617253779598331609, -0.48079807361753507111, -0.48132259489069606495, 0.62896789187858515024, 0.032668614820213034977, 0.62737212169960598906, -0.44556008848097017339, -0.13709867208614395784, 0.0046924969180440964656, 0.0030090383049267382769, -0.0040578437674064814569, -0.016547413965614935633, -0.021309552621387208016, 0.0042934606504812216493, 0.11508435494032656643, 0.071528370848007893912, -0.38424360982853533741, 0.10489834248866891642, 0.25515409890983320462, 0.094421471548197119517, -0.42590307983866571018, 0.19991691542344949828, -0.0039738702819296003316, 0.0031263850111700975873, 0.0015259305954698773947, -0.015046969088319688684, -0.028572865755002070653, 0.029373871014169723737, 0.21555105758489206957, 0.041478178103776991437, -0.72842094592394834862, 0.11375229191688139563, 0.73111019089940798565, 0.10680440943402529086, -0.81092114182968300096, 0.34384741101877969172, -0.0058571135858424255155, -0.013224465949084097255, -1.6938872721723209624e-05, 0.036777839402257254953, 0.079189317102594108944, 0.010819633941081313946, -0.27455328633730868848, -0.33145477612369705911, 0.7835070480939073656, 0.90494838355593432144, -2.3999548987113601406, 0.92592894735188624455, 0.82223057688486278494, -0.5413674313945443517, 0.017958615503450021944, 0.022255636080126050408, -0.0015964736258538346318, -0.059809476379843767857, -0.11768435189059846258, -0.053273072340766440547, 0.27640170953605619486, 0.31243321361579218243, -0.52989827698136049428, -0.041846657453312223474, 0.42793939224929089526, -0.051810081503022950566, -0.64995424359250375534, 0.45616908384229831741, 0.010412533872069109492, 0.0018427368477062656266, 0.0016294389310773068139, 0.0077990760900564625596, 0.0049562426624667581254, -0.078807183745054451562, -0.24486791442112135209, -0.079973032753825543106, 1.0033573931844381733, 0.76135847642276632818, -2.7292213857053075898, 0.79795610527741467344, 1.0261369007623191152, -0.47923953009955455595, 0.0028091223922197682168, 0.012112230007017105518, 0.016140130450016242081, 0.0077011650226164940145, -0.022596109155871475088, -0.091110997898392082073, -0.1894228440605396635, -0.11304075737281074621, 0.92240860453826067911, 0.88857530713808752232, -2.8120309268495984334, 0.93493014567797771974, 0.88415569413429639933, -0.43767544609214109608, 0.0067461144356204686204, -0.0093144060933376005701, -0.012863343927166681277, 0.0033986683802238138938, 0.022338849744156907134, -0.0085547728447472153046, 0.023351391257035128413, 0.12642977667086943705, -0.31511912431655636491, -0.081377093967043803713, 0.46798671341143610602, -0.079112603505380255098, -0.26643382133435900005, 0.12224637580978245377, -0.012023146792979085354, 0.0037468091984213997871, 0.015739486368184847642, 0.025466795904625381952, 0.038109084686142075116, 0.052281272928243256726, 0.016499507137136555646, -0.11898852358646706817, -0.370667897492160614, -0.22502133816141134393, 0.72277202465977108403, -0.28385910578746481603, 0.78833650887910144966, -0.65284687234803284728, -0.00085423115287343016986, -0.00094532984798158063971, -0.0016322208321228178982, 0.0015197109380133788474, 0.01463421917631410954, 0.031865314229032958016, 0.0014378049214817871342, -0.13660361562520614842, -0.25884041046540146036, 0.46812753732585132838, 0.84847845739871152748, -1.8911425771736707357, 1.0779993215640537585, -0.15165817869982325949, -0.00069702204050769999994, 0.0029522933955659284681, 0.0039387072848008231746, -0.0023523783036430863901, -0.014995931133260163859, -0.015324074388042872169, 0.0051551766990931696677, 0.081013760512355037502, 0.076631270069914811383, -0.29558026607274373587, 0.056068575318995984846, 0.18856734612034017906, -0.044021354072389731749, -0.042220428197698695039, 0.034818374251292250054, 0.02988718943702792924, 0.0010346483153203613926, -0.054535179930812360816, -0.12723533019440103065, -0.18034332118241000464, -0.14539795003891528569, 0.064371121158304261956, 0.52478068106056052766, 1.5155154491939630823, 0.53050990788849039514, -4.199752635631655906, 0.67783342606036012246, 1.3527528718622412462, -0.006032010459144323232, -0.0011012997462018132584, 0.0079994145199504835403, 0.017354142243382844613, 0.013733127670857275468, -0.022479755910040984113, -0.091452362246356944109, 0.0014254932957529070896, 0.23769892250118326849, -0.17565341701448480838, -0.061977316283725436552, 0.27027142563509859441, -0.34194448741973043626, 0.15351200162630368462, -0.0019024421355015724844, -0.00085682182581922920403, 0.0035029268853300687611, 0.0056844728988937522768, -0.0033751812330759739869, -0.017085047110336495924, -0.012578559844232511597, 0.075804185186056666224, 0.060077392701003706765, -0.28802460015275177518, 0.1721985065930199954, 0.017883505657375745046, 0.050765458386003700642, -0.063090229082383916803, 0.00067107562795642669313, -0.0033103148334497406111, -0.0038748390414636763902, 0.0021873047895083299642, 0.012780084103273133683, 0.020488399274624667645, -0.0022106335961090697005, -0.074390344264385968587, -0.15878874815129151443, 0.24325693611867679755, 0.63775011147111138055, -1.3163946909052699485, 0.75708675715121698246, -0.11422698952056821453, 0.0045583848999806017219, -0.009343351297226219071, -0.021057347637774531668, -0.018217018361083123223, 0.017011662122544422837, 0.085354349892192454829, 0.10557737761372008867, -0.049333098926006478668, -0.3618267715532950124, -0.034447203291074174336, 0.73995945384406192069, -1.013773295567247601, 0.99436999018170302289, -0.44450710963335954284, -0.0030603022100928020839, -0.014196943172771358047, -0.017130180602421947456, -0.0014246205220051922496, 0.041890020597105510902, 0.09900804566165824494, 0.093325582003048898616, -0.029533982630876691455, -0.32972978879191450208, -0.45609011936300236645, 0.51089514617871356617, 0.11722800235243548139, 0.59278062860883184992, -0.61330835932853067227, 0.00054233683329904145646, 0.021901440682330734766, 0.02727892244255153073, -0.0060963302553063720285, -0.094974230872024598127, -0.18544110514664219336, -0.093485019830318275003, 0.36915648001383732923, 0.56625552007556001488, -0.30307898202721478276, -0.16979508515437927163, -0.22342876611439482781, -0.51664743134424728765, 0.61422133366444076064, 0.0057514945170652753692, 0.016018623355848499168, 0.014329275598909587547, -0.0064526079001003579239, -0.048264013951836985672, -0.088788552953929220091, -0.12922186510471017629, -0.14632992360351776417, 0.19844388402231719071, 1.3161109656911043064, 0.66055709115042837709, -3.3730904206713425708, 0.83276075043131247977, 0.76362028340676202376, -0.0073819542360362647854, -0.033782150001750069213, -0.034214302391967717221, 0.019158853031309382164, 0.12920229941974631371, 0.21105753059679760453, 0.11348388887131125036, -0.22249201313287927939, -0.73706384239625322152, -0.46580013309383799802, 0.97908561155834239464, 0.017248055035226615012, 1.1742241330526386189, -1.157670695662491589, -0.019058791568329284949, -0.0067646584662764391166, 0.014505329852638656693, 0.012130840018884562689, -0.047673431301014132511, -0.065166314123311933071, 0.21155373954260414671, 0.78533650541959776348, 0.84303919844506558512, -1.6334274149688814415, -11.999686437422036889, 22.936152497312626508, -12.303715152830246637, 1.2414333580098690923, -0.0074870786341978672745, 0.0044275792936386603252, 0.012960198602577936364, 0.021461950379123313076, 0.025296363732963147303, -0.0082611748422387869945, -0.12481764569087103189, -0.25638019146990848052, -0.10088106788922023005, 1.000878407044857088, 0.93291108543457079705, -2.7800504466190103514, 1.081240114216243553, 0.20821567839960941626, -0.00012559463507277781723, -0.027167274474810369778, -0.038170256765237019214, -0.011979502920849980849, 0.068211649253446582364, 0.17043821479449330902, 0.16200878709191698923, -0.068234485133396943768, -0.51273902629952450294, -0.54647247500310980328, 0.62737212184605151144, 0.18781932811985538323, 0.82312222549850089237, -0.84969154927778678399, 0.0011683034263563027867, 0.0033604784075014285105, 0.0025770586348339234034, -0.0040578444088794456135, -0.016831240972599290084, -0.022170854318742678341, 0.0038298769086132179382, 0.11856804350150679239, 0.084231972143489028393, -0.37833794693445865764, 0.094421471506363388593, 0.20295950417922475895, -0.032397568823792910475, -0.058297514415221483863, -0.00092588018943403575106, -0.0011093803001505286518, 0.0019836281476979310094, 0.0015259305944913693215, -0.013885711761282693369, -0.030101944331581247105, 0.023801818090428146502, 0.21458533173371560632, 0.06010323051955073631, -0.70574861520863196862, 0.10680440924161686833, 0.61456814509727486495, -0.075009354176130957348, -0.20072146058623177578, 0.0026131657134533759976, -0.009256821859712929465, -0.013546761730861446929, -1.6939112312997980925e-05, 0.037994935271116678233, 0.08426274461584724873, 0.0175667476276038792, -0.28101761341948444661, -0.36496661016030235158, 0.73759051349006987675, 0.9259289469428801933, -2.2153503677561010576, 1.2693930645795208356, -0.18919927557788937533, 0.0027155282999711083218, 0.016420633629503453288, 0.019770195556302133655, -0.0015964732132748770885, -0.059139279842259945275, -0.12100775695489858486, -0.059791249048198727434, 0.2813144274926430155, 0.34854821027952898049, -0.48590722528712576578, -0.051810081813987565735, 0.23426068977328018406, -0.35628336967297408178, 0.23719055749552789281, 0.0058738433673030292573, 0.0030653398425159522275, 4.9768673416922433461e-05, 0.0016294395104215087661, 0.0094252719420933276512, 0.0099205309358217069055, -0.072955541571968790526, -0.25059525562081547134, -0.10498927771301841405, 0.9773735881973029116, 0.79795610511439862833, -2.5942019427114981767, 1.0126971903942201259, 0.2146336287652314545, -0.0058432289471851096493, 0.0073507598275401372229, 0.015105049854816660765, 0.016140130469576994404, 0.0056129363205637983897, -0.024549663062575884792, -0.091583768491602629558, -0.1927686036307875439, -0.1233470473138057466, 0.91994697743719988647, 0.93493014565052456888, -2.7528507007007596563, 1.1293653054048145012, 0.17165291940599536336, 0.0059859018583669773919, -0.00091164586842842139543, -0.010309051135921684997, -0.012863343429200847778, 0.0034209231047305668598, 0.024987930382143586083, -0.002068599448539236961, 0.027627166954272960875, 0.12070195408357042854, -0.33892380309170855313, -0.079112602970334372499, 0.52524447827254083787, -0.2433028427132390703, -0.022250043773021258714, -0.013926744161575327352, -0.0050789574419746906161, 0.0037468081714602204969, 0.011800553084851370805, 0.022479332123921259845, 0.042473956294225130581, 0.068782121067085325028, 0.043783366779798001123, -0.099061879903795946323, -0.36949799385119252992, -0.25089158858626847692, 0.78833650927001464037, -0.22558743282623958115, 0.002566587547490132605, -0.00054093281370458606936, -0.00041934164054366877437, -0.00094533038165373864566, -0.0022496833039455372931, 0.00039688854649490838038, 0.015053555399459036743, 0.036083067536025109634, 0.0069587396538182333047, -0.14548810619772534203, -0.29374783302768570081, 0.46063762031135724317, 1.0779993211582818979, -2.0789118734480425132, 0.93037250519559100415, -0.00085895358466558653469, 0.00026102258706450797744, 0.0029522926208750037771, 0.0041698307647905696535, -0.002362337755526579177, -0.015983458863243989179, -0.01546606336583946327, 0.0095903132884366561195, 0.082888864427398623036, 0.083335155043701814037, -0.32125431666473353021, -0.044021354735996248708, 0.4585214708257656091, -0.24323184343205303537, 0.0051735134643691536166, 0.024361050834103441243, 0.029887189368856915267, 0.012062850065059601765, -0.033995541002135261366, -0.1037624743953919626, -0.17110727861438407293, -0.17472116850384997178, -0.0221131284423687971, 0.38951250169178658345, 1.4089671411387114652, 0.67783342594582851603, -4.3877424344184028371, 2.3341435031007726764, -0.00089391132694510027929, -0.0034168812351594271721, -0.0011012997183892836352, 0.0065325805558177655816, 0.01517665123748215654, 0.012086374739297306088, -0.022498029539433406915, -0.091481193296559029648, 0.001721572296310996562, 0.29251740319882557628, -0.14295758654437676349, -0.34194448737865040799, 0.40403999679022695179, -0.12757299796366433631, 0.00058890409265081422423, -0.0020502468477856237473, -0.00085682182515501887233, 0.0039255194314950928383, 0.0055514523928297289379, -0.0057704775030022626564, -0.021085886576585905622, -0.011290271636371316047, 0.086879757783496172618, 0.087333970682254596918, -0.32878348038756816107, 0.050765457938249713976, 0.31348544931925653723, -0.17968941219785225516, 0.0014055788181324925176, -0.00068895995782068274096, -0.0033103149986815167891, -0.0037667920369027544335, 0.0022242548092806526359, 0.01277023697000029788, 0.019407166522117342417, -0.0030080335368499969834, -0.077168043216962162179, -0.16834303426441679652, 0.23221260421865522194, 0.75708675680532000474, -1.3643459785393465822, 0.59792125250104666279, 0.0065726033882839932782, 0.00069312554536690154303, -0.0093433510511120344832, -0.020805709978766351309, -0.020389090495633983846, 0.012305312774141322768, 0.079903584788187848043, 0.10136126454330535163, -0.061109641771771854024, -0.38159249991220428599, -0.013830050655155180089, 0.99436999017798644029, -0.97671207218401978611, 0.28733360529994722565, 0.0048825253870534417872, -0.0048853865639841538918, -0.014196942978139935956, -0.018636648631861818526, -0.0068918266670024656145, 0.033346014133419470127, 0.091795227544873528314, 0.093034224804700402234, -0.022784956345521481524, -0.29683378443822117099, -0.4166523446657312002, 0.59278062881307447274, 0.29236144557290999479, -0.32622412264373940038, -0.0052304075594656156134, 0.0068598196390293342956, 0.021901441017300581765, 0.028055251630169111138, -0.0019456115612251669023, -0.089628763606273587072, -0.18345825862995851496, -0.093874671679752941245, 0.37932666187765817778, 0.57685460771286689585, -0.51637208243245169559, -0.5166474310355479993, 0.54227620518091224078, -0.16092673338822163775, -0.0035295944527414411726, 0.0084468057343709893048, 0.016018623548948805746, 0.015099834805256047085, -0.00092529116802469914965, -0.03746125222737005378, -0.077750948302667696743, -0.12631166346748207152, -0.18525094268616096316, 0.096982683306205783036, 1.290336728240880193, 0.83276075060324694643, -3.8075718312827993017, 1.9774077468522719414, 0.0076480253823053689216, -0.013122468565377930216, -0.033782149917654102911, -0.037294019802959224585, 0.010024096195107140214, 0.11744631825141665926, 0.19966135209040034026, 0.10440222060661581704, -0.20376777722166730666, -0.64249780708707049381, -0.35051869556200382361, 1.1742241331052862829, -0.058033115028378434719, -0.26354674159808394451, -0.00035758537930123207799, -0.013707889104187574378, -0.0067646584806824120589, 0.012902256868413460034, 0.0034666856023090331096, -0.07236522375122096673, -0.098026494020034965482, 0.22471705858974311276, 0.90669338310401581538, 1.0252883698633499243, -1.5410168649947799668, -12.30371515281692929, 21.080580704211641319, -9.2482978107418656322, -0.0069339687437897775876, -0.00066801549341586523387, 0.0044275791715471506799, 0.008943547847148141175, 0.018982119058876759793, 0.031374195352342433563, 0.0059984948177386424734, -0.11707282849681252546, -0.27338814232570546325, -0.12960026596126542064, 1.0145155866583632776, 1.0812401140227803076, -3.2508776019857221762, 1.622005437939212813, 0.013425422738094739197, -0.0063849210083841215213, -0.027167274308997166332, -0.039477821547894781373, -0.020223997049533416065, 0.05294472299202862553, 0.15350356001315157428, 0.15078429961320408137, -0.067660238908372680178, -0.46623252632364448056, -0.44556008878250408234, 0.82312222573872784182, 0.29966087858867129778, -0.42396813722959647563, -0.0011823441508135013305, 0.0014871977188234879273, 0.0033604780143476785034, 0.0030090392064071307815, -0.0027824922353339638127, -0.015554894662091724861, -0.022314874145115671383, 0.0033547088977790235145, 0.12056291018643476576, 0.10351364209696371776, -0.42590307953968764565, -0.032397568961556627576, 0.56818827004569505412, -0.30514594840285869104, 0.002156780376019984749, -0.001941215562447759415, -0.0011093798396827777002, 0.003126384638884832038, 0.001451734229768825591, -0.018176668319761987713, -0.038620106778211649168, 0.02127120684313990942, 0.23975384883042447082, 0.11196210086782402982, -0.81092114203806220374, -0.075009353976825857147, 1.2409997265068550831, -0.67855335678491790485, 0.0042656498998651839333, -0.0013567270049300980597, -0.0092568215923928085648, -0.013224466099527180593, -0.00056778999308987332384, 0.036817625642722583146, 0.083530085117426516872, 0.019428202145333840767, -0.30060682943434369196, -0.41736006865732294679, 0.8222305767518636177, 1.2693930647292097635, -2.7192590775957907745, 1.2326278599951570047, -0.0054163390968974709991, 0.0044130455580813380753, 0.016420634054929064694, 0.022255635424362537472, 0.0040636163850588501861, -0.05244952045055979728, -0.11709903338629307212, -0.059377731779591352845, 0.29180282969031751117, 0.37515551479347447605, -0.64995424353909514448, -0.35628336950025846219, 0.91953849231564466216, -0.39549208258736479049, -0.0012058394256453503456, 0.0029600195976262231798, 0.0030653402207901463139, 0.001842736056302119417, 0.0063664209082646763815, 0.017621808820894063752, 0.018164478039124624942, -0.072010410939587293777, -0.28253783188385384051, -0.16625140241076982739, 1.0261369005262184206, 1.0126971905494428494, -3.0970998169414931667, 1.5383557546732180565, -0.0073699624655481467753, 0.00048274042784666987933, 0.0073507598710890803204, 0.012112230083774436912, 0.015246967151723526263, 0.011854733532840623297, -0.011830835471371363432, -0.080332472424283346557, -0.20141981363982064401, -0.17264385209211022265, 0.88415569422925333054, 1.1293653053809966647, -3.105898238175750059, 1.5266825592293240543, 0.0012485707899026060597, 0.0031333525170077008125, -0.00091164583336943681713, -0.0093144063356259410025, -0.01099467498487354325, 0.0068055060979897259374, 0.026988443035808860199, -0.012163433527878707296, 0.00027548530043921567014, 0.15109212687792281105, -0.26643382147371957869, -0.24330284275142796679, 0.68658250738250647949, -0.33504605483189509041, -0.0094028342567471778768, -0.013926744891968666912, -0.012023145748734136162, -0.0023671026539521661247, 0.012820586546652201071, 0.03070411406371427801, 0.054632995480094155571, 0.09397739471033514369, 0.12218323932058439141, 0.041642013283468996832, -0.29771643155410576487, -0.65284687290763998568, 0.0025665878839862669471, 0.6776749302634776928, 7.9626858869607757669e-05, -0.00054093293273750674231, -0.00085423082000911755921, -0.0012011502970825833642, -0.0027526765764079889551, -0.0038514489576661125811, 0.0057139573133070178934, 0.038876569907402278026, 0.067082382559926767573, -0.027078485753961104715, -0.33056589799619839898, -0.15165817844240775392, 0.93037250508591951004, -0.52754921889068517871, -0.00027760828668501349959, -0.00085895372483050776245, -0.00069702158485467947697, 0.001644959586538028741, 0.006143615674923936891, 0.0066869503402568076372, -0.0065701716104582877365, -0.028103875654647576116, -0.033385335927581485937, 0.033048871300015728036, 0.15158875873119079247, -0.042220427770917827448, -0.24323184357353258567, 0.15761914076826799569, -0.020469593334748629326, 0.0051735134406288123265, 0.034818374268087218248, 0.060838638995188348835, 0.065215877865430288418, 0.023589914946524195849, -0.082388545077410599293, -0.24321491913810888708, -0.39135054701813631883, -0.37782026227653353612, 0.024773253806348056322, 1.3527528719604122731, 2.3341435030676089823, -2.882933099615455852, 0.0039635666352892855516, -0.00089391132028112123189, -0.0060320104187332546433, -0.008387454075318575078, -0.0026359997818511883734, 0.014320193475940865002, 0.033870673940726590823, 0.024896903889401158366, -0.061759940816430228261, -0.12521756125912486546, 0.089461036859574835645, 0.1535120016346942784, -0.1275729979830743932, 0.012287513379011650097, 0.0016201070824691137569, 0.00058890411816462826747, -0.0019024421373451108272, -0.0038454670224811727218, -0.001126216436251362054, 0.0063911253040422619578, 0.0074170556971107802144, -0.01029284591331918082, -0.035227478136269627096, 0.016447953031795994189, 0.13171884329580396678, -0.063090228861014979733, -0.17968941225735846068, 0.13277498039300955623, 0.00074445514232859952199, 0.0014055788455057886387, 0.00067107569139074514653, -0.002408782250570854367, -0.006851612084769469202, -0.0069060717358579168726, 0.0045111784436889538694, 0.028048170545713704133, 0.040664122791682265345, -0.013114264529038046486, -0.1875780146591070896, -0.11422698936141087445, 0.59792125246513205816, -0.34680519216391958448, 0.0050400518661253774014, 0.0065726035373355725169, 0.0045583846229839113801, -0.0045501207429208545552, -0.023649503727247746937, -0.045228556984505223493, -0.040141063377223271114, 0.034789412728066966551, 0.16458347552519891899, 0.18228094330353461494, -0.16154693994375562971, -0.44450710955466510299, 0.28733360524330353591, 0.037050940299677558032, 0.0082266369830280173386, 0.0048825255138582124639, -0.0030603024253976623958, -0.016620699203147345363, -0.033224148360636529542, -0.04019301049873255649, -0.010969267264186600974, 0.079105086966378065916, 0.19150694746804799995, 0.20384681407968638367, -0.076198816596284635327, -0.61330835937043881589, -0.32622412267806027986, 0.65429739441851553217, -0.0036601993108219418004, -0.0052304073337220754331, 0.00054233644065022120623, 0.018690788650163953771, 0.046684515566426325328, 0.056462306467582774883, -0.011565859978087301638, -0.19379796692176035067, -0.35874632735143913154, -0.10021353120271679626, 0.61537832071780329546, 0.61422133359934427688, -0.16092673344573921179, -0.5556458573050970795, -0.0065329174769509180606, -0.0035295942939009283812, 0.0057514942173775979173, 0.020599271235528661406, 0.034727431988797026474, 0.034290370248006372955, -0.00026069862175640977527, -0.073371250299973031495, -0.18039494080676268806, -0.33123893977469409666, -0.30221330456477363269, 0.76362028343815824272, 1.9774077467769211047, -1.9942532617701824726, 0.011966425944245460619, 0.007648025403763364892, -0.0073819542408701688957, -0.037194314349951913823, -0.073211401798295591692, -0.073526299404191253206, 0.028947894101510582981, 0.24240733717584861395, 0.41010130213424211831, 0.2591318778320478855, -0.44586374563805486471, -1.1576706957142912646, -0.26354674155060453522, 1.1510736094967604082, 0.013801033388494589674, -0.00035758540924670564193, -0.019058791503211637564, -0.02960183641579570632, -0.014215329663294938578, 0.013094775276288495025, -0.020491584433433011891, -0.14160929299251950875, -0.10193408723289011653, 0.55766646768116967525, 1.7196781344909861655, 1.2414333579903424898, -9.2482978107330851003, 6.1289695308633058701, -0.00081206673558886462768, -0.0069339687543316999574, -0.0074870785438970986758, -0.0015641855615732400044, 0.0092247570252666014068, 0.026176171590034658099, 0.05284649429264345194, 0.067100119156223811401, -0.015579117098702358751, -0.27226665889607415405, -0.5158127583628553392, 0.20821567857367118126, 1.6220054378624195746, -1.1812151165276152387, 0.016689580359962583045, 0.013425422910056702963, -0.00012559490595803496794, -0.027492858060688817046, -0.064557337977012616004, -0.085591529015604433939, -0.037467450800414248135, 0.12322835963270176163, 0.32360036041151141184, 0.33030775149660690504, -0.13709867204060977031, -0.84969154930697377015, -0.42396813725316956356, 0.84187980048321309212, -0.0022754039722381927922, -0.0011823441885655406724, 0.0011683036193952615195, 0.0046924965245351184587, 0.0078908476916324556127, 0.0060618280813594868639, -0.0085162201758332538998, -0.035274586894688889083, -0.048447660810593373704, 0.043963067997140878318, 0.19991691523292648425, -0.058297514332438966911, -0.30514594841025211025, 0.1970974982946659404, 0.0027402546644102565403, 0.0021567804728788663414, -0.0009258804514958380227, -0.0039738700774633812021, -0.0022946523607691240998, 0.0028386545199645973371, -0.0050025999128579020508, -0.041006888139867386789, -0.054471264470303044913, 0.1246707304077561973, 0.34384741124222012454, -0.20072146075036820245, -0.67855335672558314553, 0.51825751176993339886, 0.002126494784565602196, 0.0042656499503232261916, 0.0026131655482321564793, -0.0058571134338927903495, -0.020179637971731660684, -0.026835312123613480617, 0.0014149777580836175699, 0.088328592652709647659, 0.16106091959820489912, -0.041052566924587530206, -0.54136743117248564694, -0.18919927572542144523, 1.2326278600296314281, -0.67208592912775610984, -0.009213217462044405448, -0.005416339100224538787, 0.0027155281695505290376, 0.017958615680586119401, 0.038203635576001815366, 0.045079754448690979862, 6.331433931984031642e-05, -0.12287232433075143123, -0.23749198271407878646, -0.040195401042698704341, 0.45616908396546435034, 0.23719055729155574808, -0.39549208246598260841, 0.0058566071510122155383, -0.0077086957987557470917, -0.0012058394521118986995, 0.0058738432718158217066, 0.010412534087240242139, 0.0095715827238705177005, 0.0073069637704943138368, 0.014653360151635189471, 0.035841878769234655766, 0.007348295941484579881, -0.21431493073562224438, -0.47923952983910206305, 0.21463362857029316677, 1.5383557547564443713, -1.1590087790789056488, -0.0029455238193185889067, -0.0073699624234437877135, -0.0058432289858595968732, 0.0028091223259167552237, 0.016441742965145639266, 0.031078434731540229702, 0.040150819514579487446, 0.029074394562914190887, -0.036766173854699511769, -0.21465990445938631925, -0.43767544617072889901, 0.17165291943682658982, 1.5266825592339461348, -1.1295004716137324596, -0.0034815099384044412116, 0.0012485707837295110384, 0.0059859018614389514906, 0.0067461144626157782539, -0.0019565883771826878511, -0.017410731102077529192, -0.020224893206083959657, 0.0057275874199555704483, 0.016890996324696771341, 0.013360608123523116594, 0.12224637571031220229, -0.022250043692727681371, -0.33504605484947735938, 0.23068593935100312198, 0.021809886803845504599, 0.04488581921937610103, -0.216860463132065584, 0.044269386036437055609, 0.098923522394568946803, 0.03140598051790401829, 0.0028875816255935909949, -0.00039719119141283900157, -0.0048309323437678020063, -0.013411696289786676287, -0.016330793824197262221, -0.010679189394987640208, 0.00031150463058675478736, 0.01275604919497941997, -0.81145104595144446069, 1.2580254514142461542, -0.027644008695816707499, -0.44026442705267160305, -0.1312673314604867747, 0.058165260741211276263, 0.079327196123260612648, 0.030381071858645448958, -0.0071609610773278277104, -0.014523441839520884883, -0.0069626462700651823093, 0.00051416113850323835049, 0.0028120587246579960283, 0.00045512653584359993907, 0.23461990998386092477, -0.312208892820896311, -0.09732176221729824106, 0.12810740607739171115, 0.085822850556588461957, 0.027224450733640499778, -0.030847131835839730618, -0.039174876849146653135, -0.011020752803509805545, 0.0073194521789394419858, 0.0083024776544401504563, 0.0031438634963200976467, -0.00050443077637328244227, -0.0012697102283473662553, -3.518687958843773167, 2.6156886265887453469, 1.7132388428349636378, 0.16126255497474481504, -0.37051188553118663549, -0.42874913207865233966, -0.30816652655321563614, -0.15338469958297523688, -0.024568473020378015642, 0.050866110189860848367, 0.071871354837681922234, 0.055542019613325263605, 0.021588725361161215099, -0.017368443492321927052, 0.60905283906434226626, -0.66202308189813308115, -0.34201089185309241358, 0.11602591937788445109, 0.2240743309996417465, 0.21070093673482015451, 0.037459094912888978024, -0.10514425760184661907, -0.10401201138225361909, -0.033710661832671941218, 0.019133054049929983792, 0.030835860441138659982, 0.015939335691713702597, -0.008941452828504101169, 0.30328436501623878829, -0.51934267669069000828, -0.18301959151665794789, 0.3876469513837394465, 0.22069948727069194083, -0.064242470616508784431, -0.12681366911382030427, -0.063449022920305145345, 0.00017076492802075099128, 0.022484420606704854512, 0.017916070060793667929, 0.0069144352527179125945, -0.00087105450547397627291, -0.0042600469807457961788, -0.47441998197326656905, 0.64605931986409481649, 0.048054956227842650363, -0.15419113492192587644, -0.098972042454288908231, -0.0298783727790228118, 0.020416523979842404052, 0.032309904765582833641, 0.015276114763060212598, -0.0035946003279177544606, -0.0077606309083389016235, -0.0028525718092253008439, 0.0012276065384548395364, 0.0011045068534606501329, 0.34340332996447436864, -0.082522493506701224963, -0.35932857238205750949, 0.0058250841189073157556, 0.053802355659528233733, -0.010708898361873461433, 0.014098181776656618786, 0.044977331197860653811, 0.036370070987917016425, 0.0072897306824938057357, -0.013506100319398514631, -0.017107642350179412538, -0.0085614399269680965809, 0.0050401112514670241838, 0.20666232429629918799, 0.39793797211819048609, -0.43286485993706330655, -0.3945051331941957784, -0.042604189730659582092, 0.13404168252665749073, 0.12771121986992059472, 0.065428655158483783882, 0.017381558250548535122, -0.0092311077817732942186, -0.019286682273445307501, -0.018111768044692996144, -0.0093271214344535242208, 0.0038759207157240149234, 1.2397598304273418357, -0.46131567999751077469, -1.0459291092757698749, -0.29615276739061263678, 0.22696215187715501616, 0.28045132353542667758, 0.1575612295819644737, 0.042087256529583512432, -0.015095193468376915183, -0.029732348247649344197, -0.024595469485217163336, -0.014402098633958101623, -0.0046324924463986066545, 0.0036067178252326310972, 1.1476589358124031914, -0.45926949511577402907, -0.96172836431750841069, -0.24668265426405855401, 0.2182276342112291212, 0.25832765536171603538, 0.14364958277304223921, 0.035593162465444055931, -0.01783332979405679794, -0.030460526342733256627, -0.02396238865429033138, -0.01290758661977456194, -0.0029813762827114466923, 0.0047363659052660558868, -1.3181296445233701142, 1.7705958315068630604, 0.28540736627352220056, -0.54885772554374545606, -0.3290146461649221532, -0.046806598120146689079, 0.069100335418842473434, 0.070190706620808937521, 0.040088434375534338916, 0.013634473132825711619, -0.0035833159892575214972, -0.011182179279194418009, -0.0092199320429453921616, -0.00030485387928619241691, -1.1812151165658357765, 1.6220054380387558535, 0.20821567830003256949, -0.51581275817505800596, -0.27226665899721169684, -0.015579117011552991168, 0.067100119101816277367, 0.05284649429364034201, 0.02617617174431125307, 0.0092247567153694344511, -0.0015641853392624281011, -0.0074870786453438246955, -0.0069339686377200959039, -0.00081206680318530414121, 7.0497397868857802905, -9.7965432661400964065, 0.62236402708075644341, 1.5648461055264888397, 0.74551776865592023213, 0.13784191694757122471, -0.046988772772479127371, -0.047828859717248914385, -0.029253374059560268566, -0.02648662562588248448, -0.025685215333385802428, -0.017675237490275195334, -0.0045677935563503682323, 0.0089739354357025822445, -2.359720287756553958, 1.9739047483257663451, 1.1927450151686584956, -0.025787926446272469577, -0.42873955788011702817, -0.38313081545641652914, -0.19139613265228377914, -0.024370435071824819717, 0.058582405139565821295, 0.070792542691594587123, 0.049939194074581111371, 0.021689876484867653267, -0.0029870992422823413291, -0.02160848002143928373, 0.0428055275713487976, 0.30424861387408991265, -0.27619971964194339753, -0.20019681057999116902, -0.0015303053747672305199, 0.064114183417930001307, 0.065325381670935794842, 0.040531488178242612574, 0.0089258853321036238371, -0.012528910583725256372, -0.016789695395533498629, -0.010556570416641099588, -0.0019429951469727641325, 0.0054076503827054740389, 0.10715660956174778706, -0.69762626660067161843, 0.062774377378478685152, 0.6681152183587077964, 0.25067821417125485395, -0.15764715197506323774, -0.20840928702267047523, -0.10733933579594603924, -0.017347937546866878944, 0.02336050061240274936, 0.031142069702654224389, 0.02307174718205654701, 0.0082107529237008831241, -0.0088453359515628313114, 0.080528393188997932084, -0.64269926434244228908, 0.11093546313075391962, 0.63069219673617238353, 0.18505286854967004095, -0.17862577846092336786, -0.18785874565879215337, -0.080391510383876138568, -0.0037635187758170698599, 0.023727515968600794427, 0.025801910838610810084, 0.018001652089254971401, 0.005949900648829871061, -0.0077856277549949630748, -1.0765786059080311166, 1.2289149385467772468, 0.23087180308763455283, -0.29245767975391012694, -0.13486643064550624116, 0.010193801850529902314, 0.026606247113002863547, -0.0019300949377543540238, -0.018594355477134713595, -0.01573854987825360971, -0.0037859503068473159514, 0.0053749705194372091391, 0.0077413287828747680622, 0.0048124340330414515018, -0.99390419451741052193, 1.1551044385648312485, 0.20977620275305053354, -0.27881830640889282424, -0.13459797359379505055, 0.0027488326993292257032, 0.023074999002224913514, 0.0017224085499535750094, -0.01100408521024505773, -0.0097863682249268516, -0.0022735256059877046592, 0.0035292290021552517182, 0.0051463290700285348883, 0.0035621187470871152302, 1.0897370159090329356, -0.12824834736442189409, -1.1211852594253663984, -0.48605846545517139567, 0.19654205493070878163, 0.34890002012476506454, 0.22542563325241299355, 0.059676485519368605015, -0.031695215875560733831, -0.049100135401459944762, -0.034754356295914300146, -0.016942512370871649213, -0.0028691157278704455466, 0.0079771146001926928476, -0.53160308515980936317, 0.24884824519203602167, 0.76780424177423389764, 0.062915021880537269627, -0.5341558248091069494, -0.28509837300408735761, 0.058177347553159453053, 0.14957077011842614955, 0.093585409452016501963, 0.017594973161592312311, -0.023948399128239632877, -0.028811758388284633425, -0.013279165512088108789, 0.0089744242717790696784, 0.044885818895134876882, -0.32794487581854192548, 0.52283278263992039037, -0.23730060899065921753, -0.10788271457191635216, 0.07285266069315905757, 0.043787428531543007293, 0.0027872700278751679978, -0.0082169646070092468987, -0.0026551880293200027551, -0.0016987320313112030858, -0.0037840891408303972861, -0.0036219881658194866443, 0.0003115047203663631227, 1.2580254514418764966, -2.6720581066683877225, 1.1362568314982270934, 0.7214759878270313731, -0.2300516485846818282, -0.2164340011260728891, -0.062924727961636200968, 0.031109752446619617222, 0.039663218198032014483, 0.013881742741532896585, -0.0043377311896139357159, -0.0076518246018588533322, -0.0036146879384211286733, 0.002812058722978194765, -0.3122088928322342416, 0.55902569504316090843, -0.061961736866710727389, -0.28072009426500382778, 0.023491193612796515411, 0.025177487102994191431, 0.0541735550586511308, 0.027700806654264754914, -0.020470709513941073854, -0.021113180598819412948, -0.0041544911291756924515, 0.0047720933910282269802, 0.0042975597749161263336, -0.00050443081457716802685, 2.6156886265618184417, -4.9280639494099496289, 0.58916745508974721623, 1.6644269874781141283, 0.54117008162126079007, -0.0082032748571559017947, -0.18871543471606316844, -0.18603571925798473297, -0.12994837643927875659, -0.063723444629558645413, -0.0075805439782996083403, 0.024931882692315131267, 0.03188255231160588149, 0.021588725293393066368, -0.66202308193467984676, 1.1044676881819814884, -0.086754668547042182447, -0.41104491056094877299, -0.040393866792162141932, -0.072626259473334620065, 0.079116391783587883269, 0.1400778823366381487, 0.032642047309856515869, -0.049790013931679449888, -0.050332619807672764056, -0.015703772609686427603, 0.0098560123356832944619, 0.01593933566363468432, -0.51934267668728173462, 0.82880838937383938259, 0.17466885501320827956, -0.73329875701531732624, -0.030646656747603417958, 0.2729774843811458851, 0.10788066583199937276, -0.037409194488923373789, -0.062619735855578284589, -0.027980489999537706824, 0.0019071604537547028405, 0.011546051367752887054, 0.007857283074060156941, -0.00087105452058721343205, 0.64605931988932085996, -1.3416812119899639288, 0.56313698149440294394, 0.27071845404532340895, -0.036259832672183460456, -0.05410161841447660297, -0.048481548707866875536, -0.023544749952531425496, 0.0066473923917511291634, 0.019083562059944238903, 0.0084675407009229854138, -0.0032685345316342275489, -0.0049940343910458272897, 0.0012276065116241226347, -0.082522493331900440006, -0.021250025631002703724, 0.55081687720675609476, -0.43133757065752698079, -0.1433897346377702442, 0.1047850670774534837, 0.023397888943138940576, -0.01962572968097293577, -0.0006416440124950574227, 0.019895821874704934518, 0.017203291072857680721, 0.0038159938726758210921, -0.0062715331979691828559, -0.0085614398753657821456, 0.39793797219028798073, -0.97135575136739804591, 0.78461344487581530238, 0.14924269754709765246, -0.2889093271528914153, -0.1889408927781598968, -0.0017281721696819573285, 0.070184467348003432008, 0.053356691079400650479, 0.025306204861941055534, 0.0053941767140348994419, -0.0045210467614772960679, -0.008398133084066949916, -0.009327121426545953789, -0.46131567982416527007, 0.47181163458970382107, 0.90056683164755502169, -0.55527124980787256892, -0.55343399736886056406, -0.089626170226813267616, 0.12265912907680136257, 0.13238890841371969698, 0.069135292167969944477, 0.017366598163217700224, -0.0086253931462152953757, -0.01469113047880654839, -0.011268353620247677921, -0.0046324924363003137928, -0.45926949500763053491, 0.49357667702111002184, 0.8164983628031096119, -0.54875539333718115209, -0.49312665183014264825, -0.070272057931995396407, 0.11294544071580242539, 0.12119494880243006363, 0.062906197719476156238, 0.014655430052963542079, -0.0095321785167952356038, -0.014706146943721482706, -0.010515793702312550839, -0.0029813763124135576887, 1.7705958314705498857, -3.4989649848226935802, 1.0904262108354838556, 1.113787877473907395, -0.095530362533996979191, -0.28425294625515540314, -0.14364443903152124271, -0.011097343121352900783, 0.031434145849129541128, 0.026941905211846545881, 0.015216698306112620351, 0.0063070893618044163859, -0.0018304297899932543605, -0.0092199320635617202507, 1.6220054379687904866, -3.25087760224405109, 1.0812401144139207609, 1.014515586512967138, -0.12960026597083257327, -0.27338814232340041821, -0.11707282848471595171, 0.0059984947842905301379, 0.031374195237609349418, 0.018982119380949097343, 0.0089435475507380936266, 0.004427579355954712817, -0.00066801568840289392726, -0.006933968631218458134, -9.796543266172456299, 21.94089908914024889, -12.105590609478429798, -1.8510323900987617485, 0.7136424291286493915, 0.77984501593283372944, 0.32616290097691535976, 0.048726902944170535914, -0.023811859226339968792, -0.018913102283318179025, -0.011279333564318486799, -0.01133595750595782492, -0.010698381805314858065, -0.0045677935640619695551, 1.9739047484802094701, -3.5729126174070509059, 0.40948464699249470522, 1.194991018671872407, 0.41365257545254524585, -0.042631082738535788779, -0.19227185079016664848, -0.16959947124032120902, -0.083389208697547650972, -0.0063514849163129431797, 0.02829133496719479085, 0.030718271453595492881, 0.016937173897321449595, -0.0029870992820003276075, 0.30424861381615464495, -0.77597708728493119956, 0.6397957204174824497, 0.033807854259069895775, -0.19332498544975659782, -0.061590066846552382618, 0.0039111538128291266012, 0.021531564064776916034, 0.025539604522858680435, 0.019240468016663129364, 0.0040589964627178141965, -0.0062595890354363989058, -0.0073172417324949964521, -0.0019429951844482490873, -0.6976262665761392423, 1.2382510228543088804, -0.1304536677268931244, -0.91495984021738629988, 0.17466499589416989435, 0.41411614702200472138, 0.096234057233312478297, -0.087569729874870744224, -0.093798581472979342122, -0.04169357538702559568, -0.0043094320830993643884, 0.010803142194914260033, 0.01275279088373188624, 0.0082107529454532781482, -0.64269926426528989349, 1.1940273903775810993, -0.20111385358137298662, -0.86403981554122777364, 0.23070430680948617042, 0.3953419371361519552, 0.058674768050603148817, -0.098936614598349445493, -0.084040079587587732113, -0.029256501380918596006, 0.0016111987187871071921, 0.010537274877398833395, 0.009969933258201904569, 0.005949900721915078218, 1.2289149385191877606, -2.537043168149392347, 0.82786871737243039604, 0.77900693886745986383, -0.070179022655472975134, -0.17947152810295954883, -0.061613752260851202547, 0.0060969482888066500736, 0.011451158910381883937, 5.7991116259506654551e-05, -0.0084649235043207735757, -0.0068591478889375228842, 0.00021271674530255654237, 0.007741328878696063405, 1.1551044384857167557, -2.3768991090480273698, 0.78484673932115112116, 0.71862870457770600119, -0.066122540334070736834, -0.16862699977212311508, -0.058794862528755625908, 0.0044452745906682260627, 0.0092238768711854716292, 0.00080662178749916265547, -0.0051691966804491840115, -0.0040729827700459152184, 0.00047523027491117101869, 0.0051463290816910114284, -0.12824834744873223058, -0.26735725894547807124, 1.2275615056029185457, -0.30523104526327421437, -0.6521865415199172844, -0.18992181760941198587, 0.094258864665212507838, 0.16735605916965037854, 0.10515361312105435398, 0.02305843334987911844, -0.018632881265196452047, -0.023365728787993929327, -0.014000273777775289197, -0.0028691157601197529378, 0.24884824510651101792, 0.13991990754450145396, -1.028086659452799978, 0.19990842593362562196, 0.81623519473449157591, -0.037207687082576969273, -0.32461630633288007797, -0.12882366482117471596, 0.028273554725750396632, 0.066064690864880884491, 0.041715097877120070613, 0.007817703400519376844, -0.011522585595818896076, -0.013279165567878222637, -0.21686046257067848053, 0.52283278253735243624, -0.58515737304462600576, 0.48064319876219779859, -0.18240719525285534197, -0.1148620196186926623, 0.053613208264813194681, 0.034647358340339176608, 0.0075380784709725024567, 0.00077824049173822764085, 0.0053621978516072395585, 0.0030702067418032449354, -0.0037840893812005563929, -0.010679189554299268333, -0.027644008667158430359, 1.1362568315213066317, -2.3377339809389061642, 0.94733664315898269415, 0.70670362722701218772, -0.2252397731605353115, -0.2084944288453971506, -0.05808700426632804753, 0.031788082745474365021, 0.039184596064970468599, 0.013125971778894679823, -0.0056474565190376708329, -0.0076518248077803710169, 0.00051416127599114244164, -0.097321762106785475455, -0.06196173703227372126, 0.37562052022326497536, -0.021206616049628267823, -0.32508138565408778442, 0.043227823980856318331, 0.049891493752896191871, 0.048853700916785133823, 0.022839535125894802614, -0.020560142749982645433, -0.02054948253812324821, -0.0039525862014799465621, 0.0047720931156278742838, 0.0031438636762225274056, 1.7132388429464586732, 0.5891674548997302141, -4.7109493671338968923, 0.40476325534435447784, 1.7225278662285943287, 0.68778279969946498618, 0.11830374697236174797, -0.12886686510231731839, -0.18148884791669805949, -0.15240330366510204674, -0.090858929609003574734, -0.02581685456408476867, 0.02493188251639455108, 0.055542019646840121161, -0.34201089178356319787, -0.086754668643880608614, 0.97951878427260186122, -0.053461884573380649699, -0.62344498792387603725, -0.084607022577647267303, 0.00566711122487798067, 0.12447589735322089899, 0.14772705796337606832, 0.026674578348643195047, -0.058919025942942980512, -0.058700075005750917967, -0.015703772928144562465, 0.030835860562508657368, -0.18301959159997527316, 0.17466885521661315583, 0.18225737745507947229, 0.15891053108978694941, -0.59978632740673853974, 0.00081375927866843495231, 0.27200819319204821989, 0.1048130378024487025, -0.037789637526758630692, -0.064433629011980395052, -0.030443570186091648694, 0.00058011568660985455507, 0.011546051406507081352, 0.0069144352491790038451, 0.048054956098345855131, 0.5631369817710417669, -1.3059876229973361728, 0.53624981359887857746, 0.30903192350985408954, -0.033724076905841976481, -0.068791600723054166511, -0.054020811323660364311, -0.01826407453027381611, 0.00933056060764443293, 0.017805796136234113003, 0.0067971468415372821373, -0.0032685343251113645018, -0.0028525718641745642992, -0.35932857265725670626, 0.55081687737201534549, -0.3810810998923045978, 0.5714865811951671537, -0.3018953980947144955, -0.15877775704380586719, 0.060867685979868239865, -0.00068867167627641786166, -0.023598433011593335923, 0.0041714087519087814787, 0.027002417896484057908, 0.022609151494453257059, 0.0038159942064986910465, -0.017107642510262494745, -0.43286486013961622543, 0.78461344504779140241, -0.7911373259022619342, 0.71551049103234576343, 0.11440248688937416921, -0.30007321166797723544, -0.20687313424483377844, -0.025148186285806315571, 0.060557170753495613569, 0.055995375653236492686, 0.033798301280747351794, 0.012448687717194319505, -0.004521046664952564284, -0.018111768059015480314, -1.0459291096216789541, 0.90056683185713148898, 0.35879127395572030235, 0.76628889402506439321, -0.57495185712637453257, -0.59267271131302379228, -0.12630864446779005439, 0.10278406424980149336, 0.13178391893933028012, 0.076730842766058973692, 0.025584186310249612539, -0.0040093130382971089998, -0.014691130199434921863, -0.014402098626238889201, -0.96172836456007959605, 0.81649836292400790327, 0.35652342310047524521, 0.70561766261572012038, -0.56765452167239227599, -0.53486553989668583764, -0.1038644364448669255, 0.096539560143989053276, 0.12237959907682463556, 0.070933005316602723567, 0.022474771656736218517, -0.0054251246676234946009, -0.014706146784476372552, -0.012907586581611155477, 0.28540736635581193115, 1.0904262106982227643, -2.9898268835144401478, 0.96132349386668958235, 1.1153916599300017776, -0.065797784515813223516, -0.27358024514120382609, -0.15580187071195927828, -0.027110589366405358352, 0.025130029543677444986, 0.030261778526156645625, 0.020325475475308728973, 0.0063070892541788915742, -0.011182179219533112599, 0.2082156784654388415, 1.0812401141312732999, -2.7800504464436777141, 0.93291108510039111312, 1.0008784073543455229, -0.10088106813714339816, -0.25638019144963458684, -0.12481764552023590087, -0.0082611750787998251139, 0.025296364067899281686, 0.021461950228670385182, 0.012960198631580025289, 0.0044275792099336156935, -0.0074870786018763988562, 0.62236402713127136899, -12.105590609548894321, 23.865210009009096837, -11.890171597471463016, -2.0764066670411116533, 0.51387661187019262332, 0.69226999587490212207, 0.3158453239343359531, 0.067894828802912685073, -0.0036769137928666416684, -0.0063297207878922128951, -0.0059450678980333794901, -0.011335957551770407814, -0.017675237423032438833, 1.1927450149247880162, 0.40948464708320936323, -3.3922446978830640418, 0.39666069608404846969, 1.299721566948748519, 0.5270348788741731072, 0.014502074087400618208, -0.18377366329159275371, -0.18355210159924742475, -0.10501542986592578288, -0.023779297842403256114, 0.01999099847567441593, 0.030718271559943156196, 0.021689876572818397177, -0.27619971958120265176, 0.63979572036807552671, -0.76849182287543360559, 0.57688292118583184287, 0.058514817281978832864, -0.20408194695592618562, -0.087542743340681461461, -0.0086567747675663376389, 0.023851500015430800172, 0.030924635959993961726, 0.023738013762285847807, 0.0067015075724640394902, -0.0062595890092862300699, -0.010556570406189874634, 0.062774377345650611559, -0.13045366770899877173, 0.34621553374952962479, -0.066207086096244913143, -0.74581638738406186295, 0.19771303154195363216, 0.42445920365566414123, 0.11504015914469722937, -0.076149933218812748681, -0.097784905926084228489, -0.05291030630295676801, -0.012800594545559209128, 0.010803142167929477568, 0.023071747119403004933, 0.11093546281968313938, -0.20111385290677857851, 0.34057221467490877975, -0.07573788456493645016, -0.69179339555409358198, 0.23837532669560829168, 0.39649193070260901894, 0.073805197800456190382, -0.088619180587408086236, -0.087564333653935191859, -0.039151135062300695955, -0.0056387928316780685373, 0.010537275361738984714, 0.018001651848433799369, 0.23087180308192839506, 0.82786871764161307219, -2.2216487853816921927, 0.67373607023284920459, 0.73786508860199973014, -0.046544444317512520537, -0.15838399444134204641, -0.05107252456358331999, 0.012614751013567442745, 0.012694918876237736363, -0.0014891099601515929603, -0.01041068436813343337, -0.0068591477055002028629, 0.0053749702933641183358, 0.20977620302216887227, 0.78484673891727374517, -2.0913978965868564863, 0.64752414506505151515, 0.68539789667478667745, -0.040431562110337541938, -0.14858217453222657967, -0.052373389265100787904, 0.0075113020972035942149, 0.0094884050610450111629, -0.000148240815854265226, -0.0062531598187760880497, -0.0040729830205647483271, 0.0035292290065835466885, -1.1211852592717566068, 1.2275615053192283632, -0.19435810382722984446, 1.0334586750750764939, -0.35948270196069931348, -0.71967628036825403992, -0.22503643305981638201, 0.083308846845983094975, 0.16830185145207082753, 0.11375745452571559002, 0.033950685899189175787, -0.012164370922442711495, -0.023365728898487512361, -0.016942512274282204437, 0.76780424207185415941, -1.0280866600307045911, 0.13514094398545700715, -0.34499893409863913174, 0.19407088995411633792, 0.69832937612699252927, -0.063331434908982547904, -0.33980368063456978245, -0.13994509663398554955, 0.028757172428581508072, 0.074305480356323672098, 0.049769638488398347531, 0.0078177032129033430241, -0.028811758248582522435, 0.044269386048261298583, -0.23730060993078300191, 0.48064319970030949802, -0.56420551262432383499, 0.48535823859917542888, -0.18813140374698086865, -0.12555420879907436271, 0.054601445248384609943, 0.041791614602767972186, 0.014460672820810475947, 0.0044152718698685375701, 0.0053621973795071461652, -0.001698731738425740892, -0.0163307935980239044, -0.44026442720745262438, 0.721475987772924654, 0.94733664366139391788, -2.4477029082366588142, 0.87720113095429053107, 0.72500349181265499876, -0.19341987797511114855, -0.19547486599163246823, -0.059605188353042445892, 0.026797130094750831047, 0.036200576125368588709, 0.013125972073367374315, -0.0043377310351459955409, -0.0069626464825300838843, 0.12810740588427671494, -0.28072009404197684201, -0.021206615935169044473, 0.34073560292656673854, 0.027592246316858630439, -0.29778052098919821633, 0.030989242148608844157, 0.037208085998867518329, 0.045087044712539069147, 0.024803016415435207714, -0.018546344120071677625, -0.02054948219931099912, -0.0041544909631683240095, 0.0083024774354146131661, 0.16126255468885747346, 1.6644269881133613165, 0.40476325499168208655, -4.6755267452776418935, 0.40076318372478242358, 1.7475409825357308691, 0.71223200623723270652, 0.12697656863660072846, -0.13279695939871785648, -0.19075266493889139752, -0.16098655945837067605, -0.090858929819826975138, -0.0075805440484689803893, 0.071871355002034620352, 0.11602591920757035493, -0.41104491023451183995, -0.053461884691444000461, 0.78064726745667956109, -0.037326700782753449714, -0.51808039487292945058, -0.037636896265094560432, -0.0047262880400330250022, 0.097735339768418486028, 0.13187174174478866195, 0.022002593221093717096, -0.058919026371412233645, -0.050332619300269630658, 0.019133053888334916603, 0.38764695155979156382, -0.7332987574726521629, 0.15891053153485376614, 0.32306330694764301414, 0.24447631390951898855, -0.59631372839827778698, -0.03202159828303573208, 0.2534585320921724505, 0.10330035705032727944, -0.033678460984508107323, -0.062316977236808512863, -0.030443569757288412719, 0.0019071600307349746532, 0.017916070213369220815, -0.15419113471658490999, 0.27071845343670564166, 0.53624981428035889408, -1.3161613839565862172, 0.50210417666775675194, 0.29094004459395489404, -0.030781865221056872728, -0.054990236283593590949, -0.047558648357454373634, -0.01997318204414294876, 0.0061730055113506029918, 0.017805796760141143209, 0.0084675400533078086507, -0.0077606306761501689731, 0.005825084172957585224, -0.43133757045601561764, 0.57148658077826097745, -0.29504703164239592939, 0.57178834440881509682, -0.33022115320337169786, -0.18388675135460122445, 0.058227212769956689098, 0.0075965825402153413712, -0.016029449468269653389, 0.0075555523820732923765, 0.027002417872485060674, 0.017203290960038645357, -0.013506100301602701197, -0.39450513305478024373, 0.14924269745439711077, 0.71551049095220575769, -0.90049349817161727927, 0.68061176405568302172, 0.12504135447190573016, -0.29367920475401865232, -0.20631638659118919921, -0.025862384530355456147, 0.063203109620043815275, 0.059669319319650949984, 0.033798301219047789667, 0.0053941766919027066621, -0.019286682299369341254, -0.29615276716639110655, -0.55527124984681008879, 0.7662888937793219668, 0.26501556024418160273, 0.76743510824530425385, -0.54271787517292813252, -0.58399764301704004676, -0.13176277474428305103, 0.095802220468975879397, 0.12993999027408159597, 0.078055681033172105709, 0.025584186402245343017, -0.0086253932450222579875, -0.024595469610874522853, -0.24668265402708677847, -0.54875539343852774987, 0.70561766243946244526, 0.27099832821770869495, 0.70464340828758098567, -0.5379900315578479697, -0.52624988969353181378, -0.10778862122763301079, 0.090755723959470402384, 0.12071536696977785974, 0.071983043226790693891, 0.022474771751921477497, -0.0095321786473312915267, -0.023962388727090361273, -0.54885772559706047513, 1.1137878775859118008, 0.96132349377995429673, -3.0899346298165148994, 0.87462698436842456484, 1.1122112234266485764, -0.048114192321464660518, -0.26461978958016457719, -0.15311103467022690872, -0.024207469150455501045, 0.027955412943149408195, 0.030261778337646871861, 0.015216698555395863704, -0.003583316076759981636, -0.51581275831691730804, 1.0145155868611097638, 0.93291108478330764253, -2.8915303007570138938, 0.85720728093323839136, 1.0100934165675743426, -0.079989583601070626817, -0.24982509385357187792, -0.12497796627207535813, -0.0068562373964979207025, 0.027604560953492524727, 0.021461949859319609524, 0.0089435479908717818026, -0.0015641854939620893853, 1.564846105516120689, -1.8510323901487515386, -11.890171597257028324, 24.60254660606861421, -11.711219956839844514, -2.1638895640745317017, 0.41863613996009202411, 0.6566733707890365368, 0.31742433887263410419, 0.077688225504071575789, 0.0016345000797619612439, -0.0063297206572472501651, -0.0112793334394790494, -0.025685215478295376351, -0.025787926396531185569, 1.194991018934598026, 0.39666069567571171639, -3.3192576208985400399, 0.38530984959884617691, 1.2756759983158416638, 0.51722789248385858674, 0.017481080186758167494, -0.18150481706826437911, -0.18537026202998876934, -0.10850689952680679651, -0.023779297785198588028, 0.028291334927440622143, 0.049939193939157168478, -0.20019681047945533936, 0.033807853944006163727, 0.57688292157496656731, -0.79811815320781043237, 0.55191055636348607116, 0.051001072972703578867, -0.20170373020335838343, -0.080177251274842778472, -0.0045790124173823755493, 0.02488363019149653424, 0.031113239492663133379, 0.0237380139241286231, 0.0040589963243757532993, -0.016789695340043327143, 0.6681152182715683896, -0.91495983983979101151, -0.066207086659092695413, 0.55502728053075744974, 0.028571803936826727088, -0.74903267493269931609, 0.16619854944126780105, 0.41114563769280992878, 0.11320846445037599515, -0.07744984793229815756, -0.10061940614707065578, -0.052910306782564510419, -0.0043094320232788044245, 0.03114206982221716033, 0.63069219714580004865, -0.86403981669744667204, -0.075737883542815034876, 0.55177665424836808139, -0.0035838411654463067646, -0.70919620038097319625, 0.20612057350689255708, 0.3876314044685687521, 0.075112687390289673006, -0.08903000593322553391, -0.090283454298572735075, -0.039151134067625929791, 0.0016111975951770471632, 0.025801911253844497796, -0.29245767966016084083, 0.77900693841268686679, 0.6737360705680737194, -2.3289072525133458491, 0.62512068115716268313, 0.75287318360355393043, -0.019060966081161572622, -0.14288498743662308987, -0.049948436106809872281, 0.0090717656628535731089, 0.010234285034088994731, -0.0014891096539315852761, -0.0084649238446201906932, -0.0037859500727298977911, -0.27881830680535968359, 0.71862870556096269947, 0.64752414405600522418, -2.1880719104122716701, 0.6051071534152885345, 0.69909256357415761052, -0.01826585537146751051, -0.13649953232102907119, -0.051411138936151963563, 0.0052227263119180951031, 0.0079690179622176476881, -0.00014824164003658223815, -0.0051691961088933223445, -0.0022735256277175044763, -0.48605846545089748112, -0.30523104510467613526, 1.0334586748214884544, -0.35268228469482898468, 0.98468526902062503137, -0.314953940971396118, -0.68757457292934098803, -0.22394321153092242027, 0.074036274536662977086, 0.16463048975987459865, 0.1154308485410941304, 0.033950685749923360801, -0.018632881010247952236, -0.034754356423178471569, 0.062915021559191874889, 0.1999084267418862193, -0.34499893469237169485, 0.16582832397646385392, -0.49729858213603506112, 0.096249621244234889739, 0.68124892302900752483, -0.03633657412354773103, -0.31523185746502890847, -0.12967640254594428595, 0.031161712513801446967, 0.074305480033910381588, 0.041715098426429460254, -0.023948399387545549893, 0.098923521922547344443, -0.10788271378692715352, -0.18240719518833725665, 0.48535823823220586304, -0.57430505822956023909, 0.47947979063247503362, -0.18359114223767983298, -0.11886332399257483472, 0.060118126502471551376, 0.044717173998153537939, 0.014460672869478004288, 0.00077824067425389558458, -0.0026551882560612461576, -0.013411696303452739359, -0.13126733136453019868, -0.23005164836914857029, 0.70670362633945638109, 0.87720113163645152543, -2.4655616781967655271, 0.88609922292529330612, 0.73766322226525360417, -0.18851103695983650876, -0.19702587741023513956, -0.062548646440315699979, 0.026797130399437544407, 0.03918459541731837692, 0.013881742989435110966, -0.014523441812774203624, 0.085822850628236121984, 0.023491193717902442845, -0.32508138615772269597, 0.027592246612519268772, 0.37573758578115173945, 0.02223793869973895998, -0.30867092849071503524, 0.025983802822015386602, 0.037089332447408342175, 0.046318739399084665753, 0.024803016977257640041, -0.0205601436476164183, -0.021113180300309661308, 0.0073194522333483863497, -0.37051188527064715661, 0.54117008102034036554, 1.7225278666261216731, 0.40076318364630530988, -4.6876426662450327854, 0.36867830771372694931, 1.7187187364051113025, 0.6957225611481346883, 0.11987436463483511218, -0.13613191570107699424, -0.19075266484401848244, -0.15240330375429625431, -0.063723444617178076621, 0.05086611021186054904, 0.22407433137682286617, -0.040393867495382491939, -0.62344498781472912352, -0.037326700421374210681, 0.89996323035441638805, 0.023559938710415843227, -0.52143511371404149823, -0.063555226979232090634, -0.024964957539113187546, 0.089292395437057706098, 0.13187174149893432507, 0.026674578542957183558, -0.049790014012510951102, -0.033710661815866807545, 0.22069948715727938371, -0.030646656345114677705, -0.59978632806213849626, 0.24447631434802769013, 0.33887761697277724027, 0.22902769639022324544, -0.61297338721146110707, -0.040743550886022793234, 0.25299921029990457555, 0.10537864066763261151, -0.033678460628875549199, -0.06443362977886545051, -0.027980489323409130087, 0.0224844203738822912, -0.098972042592453721155, -0.03625983208479128389, 0.30903192256257772108, 0.50210417722936662166, -1.3507921358328378592, 0.49403200832873916948, 0.30251048062866203736, -0.022616719757287701703, -0.05461906076871601845, -0.049854728945379234162, -0.019973181495395880142, 0.0093305597031531940788, 0.019083562736371139507, -0.0035946005431900331963, 0.053802355580215441289, -0.14338973451404346537, -0.30189539793270203649, 0.57178834406644973853, -0.32727657008472643696, 0.5547368627545614661, -0.32771119128228437534, -0.17507182441729507438, 0.06594631730073015341, 0.011279231643931941428, -0.01602944957913118515, 0.004171409128976519623, 0.019895821552393815751, 0.0072897308064978428091, -0.042604189685690706246, -0.28890932713219891204, 0.11440248667638987234, 0.68061176427212066731, -0.91021690291428336561, 0.6813241038516305581, 0.13397151089974451699, -0.28582366981913687187, -0.20104461148010646654, -0.023087276737661914516, 0.063203109285706091103, 0.055995375965488113079, 0.025306204791217704747, -0.0092311077883701474905, 0.22696215183688020489, -0.55343399735729348343, -0.5749518569440235094, 0.76743510796466418711, 0.288573444373037169, 0.79310608283413230879, -0.52708695207820588369, -0.58211878588101095833, -0.13494425262894618744, 0.093776424333447072623, 0.12993999006376669159, 0.076730843083966299156, 0.017366598079050620668, -0.029732348236368791317, 0.21822763407378339973, -0.49312665156776674946, -0.5676545218454595032, 0.70464340834543959247, 0.2939746713299297709, 0.72988109962165093059, -0.52262478479250662478, -0.52426526207523138634, -0.11092128450026056408, 0.088626180139758287124, 0.12071536688681516747, 0.070933005423551021007, 0.014655430053963766288, -0.03046052632323764317, -0.32901464618100900728, -0.095530362503089868986, 1.1153916599726876324, 0.87462698431541985311, -3.137727038621943354, 0.85557677966839251749, 1.1139864385357973742, -0.037982198402790240399, -0.25482183283914378036, -0.1482511000348351593, -0.024207469157881907756, 0.025130029863847307831, 0.026941904858988494076, 0.013634473232025471137, -0.27226665890104723156, -0.1296002662548342288, 1.0008784077616708075, 0.85720728060387263447, -2.9242800990623534219, 0.84494733705354718722, 1.0112426050141043454, -0.072974190544325467922, -0.24258366546416207377, -0.12123997094327831903, -0.0068562374980065407884, 0.025296364499884781141, 0.018982118921485605784, 0.0092247568884399697847, 0.74551776879299502809, 0.71364242902003061086, -2.0764066675140364637, -11.711219956159132138, 24.77454183773979679, -11.667798774744621682, -2.1781597768646125957, 0.4011358782922044286, 0.64994865202886642486, 0.31727372204697418256, 0.077688225616964923237, -0.0036769139796213852825, -0.018913102300306117859, -0.026486625559773400995, -0.42873955786488970876, 0.41365257537456479042, 1.2997215670737363169, 0.38530984949605284751, -3.3739417850499591189, 0.33131463184115994558, 1.2512963185148449075, 0.51338345721472888883, 0.021015636002934320187, -0.17921503612570352515, -0.18537026230899103507, -0.10501542957375660514, -0.0063514850849721556439, 0.070792542770875835356, -0.0015303055657766790124, -0.19332498483443513737, 0.058514816475512154703, 0.5519105568155241448, -0.81962363569197449475, 0.55179055780554309152, 0.063298621579144578875, -0.19365985974635679945, -0.078016152124305557236, -0.0048093276611460827191, 0.02488363023899584428, 0.030924635841383701224, 0.019240468010503435081, -0.012528910551867300863, 0.25067821429759823415, 0.17466499525792950065, -0.74581638627617952153, 0.028571803197519445028, 0.57567747054168483256, 0.01842400020107772679, -0.76636340353808229953, 0.1511394948746324074, 0.40356176511067948232, 0.11097077963243595633, -0.077449848678557586812, -0.097784905362548205088, -0.041693575178219269284, 0.02336050039770573919, 0.18505286834180401101, 0.23070430744431205672, -0.69179339638940962054, -0.0035838405958305433532, 0.5579130844936091016, -0.016574224439065352033, -0.72335009374600978127, 0.19500651528274437485, 0.38228401872279438312, 0.073484408906391129879, -0.08903000537244951107, -0.08756433491091426824, -0.029256500230955030661, 0.023727515579533575013, -0.13486643072863024884, -0.07017902261595410085, 0.73786508881780654701, 0.62512068104360796195, -2.3435766776448430093, 0.6284877002990596484, 0.76337696611589211937, -0.014766920543581696368, -0.14407361937777943273, -0.052116643076793305045, 0.0090717660425551349679, 0.012694918499658116445, 5.7991267593461353833e-05, -0.015738549918858778243, -0.13459797335068793278, -0.066122541357109321236, 0.68539789838389386656, 0.60510715227922107218, -2.2024484731650026959, 0.60455310770810288812, 0.70648328485449529701, -0.014192069399748333258, -0.13638304599904693681, -0.052513425909006913872, 0.005222725678257957721, 0.0094884060514327274449, 0.00080662133441552347213, -0.0097863682258007549802, 0.19654205491331913636, -0.65218654159762268208, -0.35948270174235313901, 0.98468526883302931463, -0.32777725055175344071, 1.0275837197590647065, -0.29111410772566553851, -0.68441877429266428923, -0.22763846394832965014, 0.071693943669239604266, 0.16463048964566795451, 0.11375745475094489745, 0.023058433213006927776, -0.049100135389259662311, -0.53415582468354061429, 0.81623519431422597581, 0.19407089050879375303, -0.49729858243905200599, 0.08140422707115908818, -0.52894803332508311122, 0.1038469174439149989, 0.7056108342007699763, -0.017044233138221821788, -0.30822264016703171396, -0.12967640243747854956, 0.028757173143026473067, 0.06606468974725092369, 0.017594973622156093868, 0.031405980304780381496, 0.072852661578156660549, -0.11486201988637417815, -0.1881314045084271358, 0.47947979099028492644, -0.56909624620095111691, 0.48802101393345703739, -0.17729844548501122459, -0.11615462113405668199, 0.060118126482233642804, 0.041791614721153350076, 0.0075380784513915865749, -0.0082169642509280372417, -0.0048309327124132364975, 0.058165260735342852461, -0.21643400119313088714, -0.2252397727451478926, 0.7250034913027099126, 0.88609922296867094094, -2.4653096761084114519, 0.88360561737611964528, 0.73567856453174740317, -0.190071702692838157, -0.19702587788464046015, -0.059605188146469152699, 0.031788082765241781891, 0.039663218166617289495, -0.0071609610481126019857, 0.027224450741593773234, 0.025177486987969382309, 0.043227824183954668968, -0.29778052077499722605, 0.022237938046638046352, 0.36802630273593339716, 0.020054887677627322717, -0.30783687025103995127, 0.026934261301129306526, 0.037089332071106421995, 0.045087044602758752365, 0.022839535565581627757, -0.020470709736376588306, -0.011020752777633672712, -0.42874913201011560782, -0.0082032750801595746143, 0.68778279973504907741, 1.7475409828701382597, 0.3686783075022731504, -4.7359740467463966596, 0.3300085540922871874, 1.6987009787158062757, 0.68917763007388832541, 0.11987436469302957531, -0.13279695971718991077, -0.18148884779716123461, -0.12994837601606029631, -0.024568473316203980061, 0.21070093647890048016, -0.072626259081232796833, -0.084607022438877382831, -0.51808039529368155041, 0.023559938903862827353, 0.90155474771461852068, 0.003290612645453215232, -0.53803429282662418309, -0.071356131408151363127, -0.024964957714474031447, 0.097735340058619502024, 0.14772705786231291047, 0.032642047194312559522, -0.10401201132032135421, -0.064242470543407634298, 0.27297748401275367947, 0.00081375982952683301955, -0.59631372860919518164, 0.2290276962612749212, 0.32971937202225365393, 0.22641015197307262641, -0.61300311527952056068, -0.040282841717356403233, 0.25299921006909797949, 0.10330035715095949034, -0.037789637256090760631, -0.062619736153712426963, 0.00017076501160839964834, -0.02987837271676218806, -0.054101618714170573998, -0.033724076600337472842, 0.2909400446772650306, 0.49403200810048397695, -1.3448151850551253172, 0.49812092499576149374, 0.30168193258069692098, -0.024645228742804965033, -0.054619060787216976161, -0.047558648616303822454, -0.018264073965553167395, 0.0066473919991928949805, 0.01527611484741646028, -0.010708898082073184735, 0.10478506653258079506, -0.15877775699624707695, -0.33022115293076764742, 0.55473686269035848984, -0.32298468232730465077, 0.56646741961173185942, -0.31879022466844730843, -0.17133273323231715257, 0.065946317393545020313, 0.0075965824588893043207, -0.023598433108133483327, -0.00064164396417218383384, 0.036370071041519021204, 0.13404168256816578753, -0.18894089301734759423, -0.30007321133336095631, 0.12504135421835393949, 0.68132410403127952048, -0.90667745982009162464, 0.68845902446828777244, 0.14199411713366910659, -0.28113991954917361138, -0.20104461167616211048, -0.025862384202031716424, 0.060557170472822226259, 0.053356691099195323325, 0.017381558338053309382, 0.28045132346774320764, -0.089626170105710889491, -0.59267271175725100285, -0.54271787477534549904, 0.79310608291316708751, 0.30646586672454129152, 0.80318497808493116707, -0.52317957374697943163, -0.58195045332421624362, -0.13494425233685131138, 0.095802220095417084389, 0.13178391902118288814, 0.069135291888736946642, -0.015095193190050936599, 0.25832765537950702628, -0.07027205811015269088, -0.53486553978412587629, -0.53799003160033476156, 0.72988109980316229208, 0.31187401282054827423, 0.73964673807383407578, -0.51927697227713354167, -0.52445954899985891551, -0.11092128422512011199, 0.09075572353885309751, 0.12237959928786668051, 0.062906197509291980619, -0.017833329616449476729, -0.046806598102160021369, -0.28425294621893537617, -0.065797784653556054546, 1.1122112235181815798, 0.85557677967268130903, -3.1531080800541788633, 0.85307865810846761345, 1.1210188549162511951, -0.031673854105587921637, -0.25482183263955210384, -0.153111034676738339, -0.027110589550730455205, 0.031434146003223334409, 0.040088434326580588141, -0.015579117103485314033, -0.27338814205994721451, -0.10088106844259897565, 1.0100934168404740454, 0.84494733684413902974, -2.9389724607399143608, 0.84080043870362286551, 1.0164223010320230234, -0.067799327406637882598, -0.24258366525069141595, -0.1249779663597084256, -0.0082611751342296212836, 0.031374195342248938079, 0.026176171671326544516, 0.13784191664574710345, 0.77984501642184267389, 0.51387661200144807427, -2.1638895646725342381, -11.667798774418599805, 24.838712735070373583, -11.637406017514202006, -2.1746637194414515371, 0.39754601755419682885, 0.64994865189037143161, 0.31742433899311933709, 0.067894828690614250633, -0.023811859162160109449, -0.029253374066399346481, -0.38313081542527860357, -0.042631083022904067703, 0.52703487907349799713, 1.2756759982107199747, 0.33131463215118200916, -3.4136717404747165538, 0.31021309202054703213, 1.2452537105291057173, 0.51395452893658410964, 0.021015635913569176429, -0.18150481699869788765, -0.18355210174177383209, -0.08338920861594144307, 0.058582405206406243381, 0.064114183525844040124, -0.061590067268983160476, -0.20408194637546589467, 0.051001072706649441812, 0.55179055778823005163, -0.81049187759351470906, 0.55948654748492965272, 0.066756194022120310394, -0.19304937043442926825, -0.078016152664845100917, -0.0045790118994634776764, 0.023851499801769700942, 0.025539604648009063548, 0.0089258852507786961422, -0.15764715179318392058, 0.41411614685847353279, 0.19771303116669039501, -0.74903267432611275733, 0.018423999940720352242, 0.56909532564990339409, 0.0097264916132721382369, -0.77552439308475218382, 0.14599247775562521068, 0.40356176504021407103, 0.11320846474323481567, -0.076149933135703090636, -0.093798581994493271385, -0.017347937302540861193, -0.1786257784037506291, 0.39534193717777715849, 0.23837532651679257945, -0.70919620045397913088, -0.016574224143090558747, 0.55382355340119482356, -0.022409844596913197634, -0.73074659420983689717, 0.19063720451392424104, 0.38228401903703290721, 0.075112687082981785536, -0.088619180284489132426, -0.08404007990102579928, -0.003763518662280989583, 0.010193801915930101459, -0.17947152789913983773, -0.046544444672042412092, 0.75287318355778332091, 0.62848770051773039658, -2.3445302770862648245, 0.62449317664942094463, 0.76021919456745201682, -0.016378137065712880527, -0.14407361898049700355, -0.049948436227546202937, 0.012614750749407963731, 0.011451159266812989054, -0.018594355662255541983, 0.0027488328312225565955, -0.16862699978416195723, -0.040431562520945499295, 0.69909256406117936677, 0.60455310750429969158, -2.2058202268685076497, 0.6008616126271485447, 0.70476893614131674859, -0.014748580711182459127, -0.13638304596548336822, -0.051411138672377211523, 0.0075113018992162905799, 0.0092238768392714585304, -0.011004085192693389278, 0.34890002000198538923, -0.18992181747397623992, -0.71967628038037934068, -0.31495394078556315787, 1.0275837195079022823, -0.3055962955064139841, 1.0369724293897844003, -0.28715459772388524096, -0.68371017651046839259, -0.22763846402192344787, 0.074036274567355842002, 0.16830185159925570293, 0.10515361288926468319, -0.03169521576656100792, -0.28509837313623415112, -0.037207686690558683384, 0.69832937563809849291, 0.09624962147917426325, -0.52894803332608275603, 0.087117622607744638064, -0.51370097372527090407, 0.11670772116780707417, 0.71244211402500501329, -0.017044232693167710169, -0.31523185780126566913, -0.13994509709898161898, 0.028273555691532812478, 0.093585409023126633898, 0.002887582473686148548, 0.043787426543752720753, 0.053613208446308999178, -0.12555420770170044875, -0.18359114241631446784, 0.48802101430541894311, -0.56230861995689929156, 0.49114986875408145339, -0.1772984456823475663, -0.11886332416737441231, 0.054601445266893755548, 0.034647357672895000291, 0.0027872705081644732865, -0.00039719101661868627109, 0.079327196123822302232, -0.06292472806885857084, -0.2084944287538987584, -0.19341987791852274792, 0.73766322236019166159, 0.88360561722586605793, -2.47091354887136605, 0.88041288217848134856, 0.73567856412724919046, -0.18851103627575416244, -0.1954748665031075805, -0.058087003862560895873, 0.031109752178975779574, 0.030381071922584473172, -0.030847131789779273087, 0.054173554899906960214, 0.049891494091024973789, 0.030989241408085635432, -0.30867092739652524092, 0.020054886954431992707, 0.3700821651196630202, 0.021774314439434745011, -0.30783687057155451061, 0.025983803245645281654, 0.037208085822476692772, 0.048853700941741018404, 0.027700806667526927524, -0.039174876864220893602, -0.30816652660582916079, -0.18871543433542833923, 0.11830374643343635754, 0.71223200623832483291, 1.7187187367407654737, 0.33000855400818557284, -4.7636211705401807492, 0.3175554158731754284, 1.6987009789205462784, 0.69572256072524552017, 0.12697656875869386961, -0.12886686482925105435, -0.18603571946735486531, -0.15338469954763220282, 0.037459094760089989284, 0.07911639224375359436, 0.0056671107950835800798, -0.037636896189197285856, -0.52143511381621876488, 0.0032906129174538039511, 0.88720753717469735822, -0.0026171640096668935364, -0.53803429286353687822, -0.063555227217289345099, -0.0047262878221886227628, 0.12447589749484500032, 0.14007788208517371609, -0.10514425750243214241, -0.12681366924491796477, 0.10788066617179914441, 0.27200819297149425591, -0.03202159844949806905, -0.61297338684054947233, 0.22641015171146680607, 0.33235428616280937586, 0.22823278500796362334, -0.61300311552632291701, -0.040743550284348659107, 0.25345853130665979558, 0.10481303836748119551, -0.037409194804819523306, -0.06344902280214413659, 0.020416523996733257351, -0.048481548859778018923, -0.068791600167639022079, -0.030781866069318413598, 0.30251048128288748895, 0.49812092468298163617, -1.34749746336525833, 0.49501974782537649933, 0.30168193251183206272, -0.022616719448270577314, -0.054990236583681871563, -0.054020811316955866244, -0.02354474987372214001, 0.03230990475512415111, 0.014098181618485423894, 0.023397889162199553975, 0.060867686085986118238, -0.18388675117632691247, -0.32771119164552531267, 0.56646741942583944684, -0.31323073368151826967, 0.570817577495635331, -0.31879022455408695391, -0.17507182416443645145, 0.058227212566380248671, -0.00068867172662575735966, -0.019625729453461599611, 0.044977331024844781349, 0.12771121961173184656, -0.0017281716606125270552, -0.20687313435782056503, -0.29367920478949638419, 0.13397151072445570574, 0.68845902449110274457, -0.89975186905001858229, 0.69322307852412434759, 0.14199411703991293709, -0.28582366955232901828, -0.20631638677812116911, -0.025148186043234967146, 0.070184467128153057991, 0.065428655185903392133, 0.1575612296217453745, 0.12265912918088801853, -0.12630864436087274605, -0.58399764292657496689, -0.52708695262128146819, 0.80318497818186007642, 0.31248931790026590605, 0.80583684553523471372, -0.52317957384408653176, -0.58211878590862897731, -0.13176277430792168732, 0.10278406420039812308, 0.13238890842678308069, 0.042087256331262708209, 0.14364958279383960305, 0.11294544082072573155, -0.10386443633017101995, -0.52624988984755316501, -0.52262478497694841995, 0.73964673808180614323, 0.31729075356377245409, 0.74178410936925875951, -0.51927697212704593266, -0.52426526224683611677, -0.10778862073643458164, 0.096539560024665571625, 0.12119494881552499155, 0.035593162302837731692, 0.069100335444228055959, -0.14364443928485950974, -0.27358024474485143074, -0.048114192526301093056, 1.1139864385519480106, 0.85307865812996175325, -3.1516459029326839847, 0.85632681910855101837, 1.1210188552200646139, -0.037982198707933297055, -0.26461978946871839158, -0.15580187096807046609, -0.011097342855405213299, 0.070190706569173394458, 0.067100119149543724362, -0.1170728286340976948, -0.25638019131362893566, -0.079989583621572615812, 1.0112426050835006119, 0.84080043858578035199, -2.9387959904648917053, 0.84333880295730345722, 1.016422301353004265, -0.072974190924521134538, -0.24982509346938572281, -0.12481764601854242192, 0.0059984950949262148476, 0.05284649424648750482, -0.0469887727041228484, 0.32616290084231058799, 0.69226999608085648674, 0.41863613980208447263, -2.1781597770699243632, -11.637406017212061471, 24.864961695588373658, -11.629508201359541886, -2.1746637196415856685, 0.4011358784457044746, 0.65667337074647502693, 0.31584532429114198226, 0.048726902490622718556, -0.047828859565801004616, -0.19139613272284894396, -0.1922718502545985586, 0.014502073572400291102, 0.51722789312562444941, 1.2512963176420868194, 0.31021309217038672834, -3.4258360145895045967, 0.30607855090175067891, 1.2452537107712444708, 0.51338345747360136784, 0.017481080084103071265, -0.18377366309107112663, -0.16959947145240464317, -0.024370435149125159457, 0.065325381716899375006, 0.0039111538201506952997, -0.087542743502653841414, -0.2017037301787986403, 0.063298621693721524029, 0.55948654741920345046, -0.80768915930067108899, 0.56008742768824260061, 0.066756193529160692823, -0.19365985887333900251, -0.080177252132700591036, -0.0086567743297694139254, 0.021531564012251962897, 0.040531488165236988219, -0.20840928738941377762, 0.096234058027179006367, 0.42445920319963365008, 0.16619854923785823408, -0.76636340314632744875, 0.0097264914102284903696, 0.5632160826255452335, 0.0058180930469909162422, -0.77552439317725263557, 0.15113949451990318917, 0.41114563836277512321, 0.1150401584821899581, -0.087569729475061558466, -0.10733933588289508443, -0.18785874565178084517, 0.058674767621466672463, 0.39649193176349628587, 0.20612057257827204326, -0.72335009366651736862, -0.022409844190170079281, 0.54929877636861068613, -0.025864742654078773376, -0.73074659418204757078, 0.19500651505873142777, 0.38763140416221447682, 0.073805198522848503018, -0.098936615060750826567, -0.080391510250961792372, 0.02660624718135536429, -0.061613752761840737127, -0.15838399398829497433, -0.019060966174994989353, 0.76337696610240834971, 0.62449317690910566281, -2.35131136323310308, 0.62092542136688955168, 0.76021919482946442859, -0.01476692101563132492, -0.14288498767988491722, -0.051072524100724332741, 0.006096948065346299156, -0.0019300948096657838914, 0.023074998724786521809, -0.058794861811331755597, -0.14858217566864800885, -0.018265854493715004414, 0.70648328468067933539, 0.60086161268519255874, -2.2109328664538732134, 0.59846793159238453885, 0.70476893615936653248, -0.014192069796822475922, -0.13649953190017127813, -0.052373389937574080233, 0.0044452751069574900078, 0.0017224084982253105424, 0.22542563322760703071, 0.094258864881834394933, -0.22503643327792208217, -0.68757457304979741064, -0.29111410747355653816, 1.0369724292134865351, -0.30060829837709895918, 1.0398162179060557175, -0.28715459782268354383, -0.68441877421842300944, -0.22394321116678878036, 0.083308846349778906126, 0.16735605932294217446, 0.059676485521313063498, 0.058177347803576077989, -0.32461630687791098726, -0.063331434695779509347, 0.68124892340637688282, 0.10384691705784312998, -0.51370097363602418294, 0.097278449361709817822, -0.50976103464269273413, 0.11670772182308598863, 0.70561083365692611746, -0.036336573807456147567, -0.33980368090685836302, -0.1288236647180194816, 0.14957077011416014534, -0.00039719119604940985972, 0.0027872705540135014196, 0.034647357899707666162, 0.054601445349944884022, -0.11886332415913158922, -0.17729844580746503824, 0.4911498687647135597, -0.56230861995129977071, 0.48802101411164922062, -0.18359114222874872713, -0.12555420851134502236, 0.053613208714819511613, 0.043787427841225814995, 0.0028875816578068677458, 0.030381071895584938619, 0.031109752320622758442, -0.058087004197701616603, -0.19547486600619748409, -0.18851103677610803189, 0.73567856445246260222, 0.88041288202497769433, -2.4709135487763984607, 0.88360561730060283026, 0.7376632220375467508, -0.19341987762364809544, -0.20849442889688232783, -0.062924727973877436771, 0.079327196082575587854, -0.03917487687127017143, 0.027700806728055055173, 0.048853700680429122671, 0.03720808640787632432, 0.02598380245091789012, -0.30783686983983638008, 0.021774314043953954534, 0.3700821650643575933, 0.020054887334301205182, -0.30867092789474476211, 0.030989241809697261332, 0.049891493894445422541, 0.054173554964701414738, -0.030847131804919103448, -0.15338469963010578634, -0.18603571935823950967, -0.12886686454951321035, 0.12697656815499883121, 0.69572256098936735569, 1.6987009789343936461, 0.31755541601889075709, -4.7636211706053028792, 0.33000855362867098641, 1.7187187373099273024, 0.71223200597419644708, 0.11830374656405744094, -0.18871543458581077624, -0.30816652647378334207, -0.10514425736255518518, 0.14007788182232713825, 0.12447589773655348688, -0.0047262882548874821445, -0.06355522672511268456, -0.53803429305174998998, -0.0026171639269019129036, 0.88720753727854273496, 0.0032906124357819270879, -0.52143511331032388334, -0.037636896708083823759, 0.0056671113589777466837, 0.079116391971996199994, 0.037459094787350523392, -0.063449022821844322517, -0.037409194570609767549, 0.10481303777345385431, 0.25345853201156032242, -0.040743550894111350835, -0.6130031150494700265, 0.22823278484056908066, 0.33235428599286370721, 0.22641015211072351709, -0.61297338752030006148, -0.03202159761897747231, 0.27200819237152001895, 0.10788066640940068008, -0.12681366928736850785, 0.032309904670203663801, -0.023544749488067225757, -0.054020812111652688481, -0.054990235615260008029, -0.022616720194369461377, 0.3016819328790985022, 0.49501974768730661047, -1.3474974634106355875, 0.49812092503528693221, 0.30251048059199975304, -0.030781865138372301732, -0.068791600928486448807, -0.048481548588620797247, 0.020416523981507721242, 0.044977331043944891731, -0.019625729497667148554, -0.00068867171971708223029, 0.058227212630025469708, -0.17507182419017675024, -0.31879022459442429849, 0.57081757740637140142, -0.31323073378846250092, 0.56646741977631664966, -0.32771119166996426353, -0.18388675100516752026, 0.060867685839602315989, 0.023397888868994978773, 0.014098181873652635906, 0.065428655257244019983, 0.070184467157773516854, -0.025148186302786351004, -0.20631638655735456966, -0.28582366977673556718, 0.14199411723754176706, 0.69322307854500142543, -0.89975186919518390649, 0.68845902471320197247, 0.13397151049052538618, -0.29367920452883150562, -0.20687313445138799661, -0.0017281718532655592543, 0.12771121973624866008, 0.042087256344945173092, 0.13238890837640290843, 0.10278406447646855348, -0.1317627748435487578, -0.58211878549905715641, -0.52317957413687388168, 0.8058368458260729561, 0.31248931753604791695, 0.80318497866005955643, -0.527086952742102266, -0.58399764298521739114, -0.12630864439546299338, 0.12265912912195500728, 0.15756122966215976855, 0.03559316226923891352, 0.12119494901903059281, 0.096539559789743933926, -0.1077886206512130296, -0.52426526241348558877, -0.51927697204797818031, 0.74178410928774951572, 0.31729075391996319766, 0.739646737947930899, -0.52262478496641828762, -0.52624988993528765135, -0.10386443634950946746, 0.11294544081270374569, 0.14364958282054215988, 0.070190706646470854557, -0.011097343044411664414, -0.15580187092851643937, -0.26461978933913060663, -0.037982198570196440413, 1.1210188549101209876, 0.85632681916132191713, -3.1516459027720316044, 0.85307865788474335211, 1.1139864387886253549, -0.048114192621774798386, -0.27358024469723407623, -0.14364443939158919039, 0.069100335510576885167, 0.052846494306103837524, 0.0059984947904723881149, -0.12481764552685188641, -0.24982509386208864854, -0.072974190578594347323, 1.0164223010650164092, 0.84333880297440821927, -2.9387959903214619928, 0.84080043839253915117, 1.0112426054159491251, -0.079989584081383333092, -0.25638019089328201172, -0.11707282887454231735, 0.067100119218658701992, -0.047828859559103757693, 0.048726902704277881562, 0.31584532377984958496, 0.65667337100076561551, 0.40113587860023186549, -2.1746637197626661475, -11.629508201366974163, 24.864961695734272951, -11.637406017463614916, -2.1781597769950167276, 0.41863613991363163302, 0.69226999586772541839, 0.32616290111956758313, -0.046988772838917190755, -0.024370435082287398437, -0.16959947167858976935, -0.18377366265443742521, 0.017481079567008186992, 0.51338345783374450537, 1.2452537103982939115, 0.30607855110229570395, -3.4258360143346950899, 0.31021309178014133678, 1.2512963181835285997, 0.5172278926166149482, 0.014502073981908048667, -0.19227185072222230389, -0.1913961325151189441, 0.040531488091873915658, 0.021531564227344333895, -0.0086567746945221575594, -0.080177251559224518851, -0.19365985956587572492, 0.066756194109573369078, 0.56008742736534844475, -0.80768915936232654751, 0.55948654781840090333, 0.063298621228626020252, -0.20170372979031866989, -0.087542743795386068761, 0.0039111539638713964384, 0.065325381689388117845, -0.10733933567933733044, -0.087569730104048004726, 0.11504015936298016276, 0.41114563770176237867, 0.15113949447892316424, -0.7755243926461976578, 0.0058180928235948559113, 0.56321608241757570301, 0.0097264916821641466543, -0.76636340323157570165, 0.16619854877308451457, 0.42445920408485182662, 0.096234057402567241279, -0.20840928722260040895, -0.080391510438482582934, -0.09893661414856734504, 0.073805196885801782614, 0.38763140539879470392, 0.19500651491607634558, -0.73074659460722590776, -0.025864742160590842301, 0.54929877611925270742, -0.022409844271428842888, -0.72335009413964734826, 0.20612057410582487749, 0.39649193014355460507, 0.058674768447664824023, -0.18785874583108311264, -0.0019300949979820004616, 0.0060969486223767434957, -0.051072524860671320024, -0.14288498712116887157, -0.014766921011958486845, 0.76021919455898900875, 0.62092542147420559662, -2.3513113630801627529, 0.624493176543903461, 0.763376966098009202, -0.019060965557137269505, -0.15838399476317877035, -0.061613752277354161269, 0.026606247067124583205, 0.0017224087304716453363, 0.0044452740854881590132, -0.052373388224957281856, -0.13649953335383124831, -0.01419206907461281926, 0.70476893603918033815, 0.59846793163055345133, -2.2109328667902734544, 0.60086161283810124445, 0.70648328512317282168, -0.018265855563517750804, -0.14858217437524381954, -0.05879486259874295434, 0.023074998923631687836, 0.0596764855871220537, 0.16735605905441566787, 0.083308846829583699245, -0.2239432115567727144, -0.68441877422985353263, -0.28715459754905192646, 1.0398162176797793865, -0.30060829834993990639, 1.0369724294924569374, -0.29111410772633933286, -0.68757457336355387589, -0.22503643271959117289, 0.094258864661792604589, 0.22542563322647493629, 0.14957077003078653177, -0.12882366475393958183, -0.33980368041674580137, -0.036336574527148432501, 0.7056108344754615791, 0.1167077209946997679, -0.50976103443043430108, 0.097278449775131517474, -0.51370097424873883973, 0.10384691787341379654, 0.68124892273037707113, -0.063331434399769029331, -0.32461630707025740383, 0.058177347905066276179, -0.0048309326259366024842, -0.0082169645835625977626, 0.0075380785969516904757, 0.041791614962782878961, 0.060118126357749761268, -0.11615462091273820455, -0.17729844581187209629, 0.48802101394862673622, -0.56909624628930777224, 0.4794797909271438785, -0.18813140386719026664, -0.11486202000744337448, 0.072852661014087438662, 0.031405980564990276793, -0.007160961112319900318, 0.039663218305097634742, 0.031788082677890912298, -0.059605188141074051666, -0.19702587785639666396, -0.19007170276012505528, 0.73567856468284431593, 0.88360561709492546623, -2.4653096757594852306, 0.88609922262257212466, 0.72500349167385436111, -0.22523977306977158164, -0.2164340010322487462, 0.058165260701227718132, -0.011020752831732460661, -0.0204707095815329157, 0.022839535407312254456, 0.045087044646432671013, 0.037089332125080663738, 0.026934261193574081383, -0.3078368700091014265, 0.020054887188450797042, 0.36802630339099046131, 0.02223793745754333237, -0.29778052035723973079, 0.043227823905592571707, 0.02517748713391369797, 0.027224450704602467382, -0.024568473263910886595, -0.12994837615721055446, -0.18148884773491205658, -0.13279695963480461773, 0.11987436473948719085, 0.68917762993272124739, 1.698700978629613445, 0.33000855418655145046, -4.7359740465158433054, 0.36867830723714356544, 1.7475409828804029377, 0.68778279981992107572, -0.0082032751209782928115, -0.4287491319984908511, -0.10401201141899324432, 0.032642047198460318047, 0.14772705817793094218, 0.097735339819318090004, -0.024964957595380889849, -0.071356131613707907135, -0.53803429277932890429, 0.0032906125561610378473, 0.9015547479677449294, 0.023559938853061186681, -0.51808039519986526233, -0.084607022662971875793, -0.072626259102207421647, 0.21070093657421229416, 0.00017076492359493599792, -0.062619735940937115704, -0.037789637500347583732, 0.10330035741959121831, 0.25299920985446500143, -0.040282841629405569117, -0.61300311535943041719, 0.22641015193095062608, 0.32971937230828413101, 0.22902769597439351368, -0.59631372836594687126, 0.00081375956107610099809, 0.27297748414760986124, -0.064242470552679745155, 0.015276114799444675749, 0.00664739215620300649, -0.018264074129989853884, -0.0475586485238336798, -0.05461906092996736406, -0.024645228576400998038, 0.30168193253698938339, 0.49812092495782883672, -1.3448151849136560365, 0.49403200775727229566, 0.29094004508253595853, -0.033724076936323073295, -0.054101618474445722917, -0.029878372803240071942, 0.036370071138006644651, -0.00064164407312994656794, -0.023598432989861455739, 0.0075965821645668472895, 0.06594631748966220397, -0.17133273308334137641, -0.31879022472647167152, 0.56646741992997284854, -0.32298468285894160745, 0.55473686303096791939, -0.33022115337066143548, -0.15877775667037816176, 0.10478506667054408863, -0.010708898228925132004, 0.017381558386605634736, 0.053356690892718473818, 0.060557170857677179721, -0.025862384548584416155, -0.20104461152155803094, -0.28113991958345058153, 0.14199411699368466588, 0.68845902474305398222, -0.90667746007802163533, 0.6813241043543930564, 0.12504135397229457216, -0.30007321137683540258, -0.18894089293087773518, 0.1340416825729938699, -0.015095193129113133185, 0.069135291940897264373, 0.13178391865265462468, 0.095802220666792947679, -0.13494425308166585475, -0.58195045248138843608, -0.52317957454737562539, 0.80318497893408313804, 0.30646586590628832614, 0.79310608355997636032, -0.54271787541991978188, -0.59267271126342568888, -0.089626170255688253352, 0.28045132348286189172, -0.017833329510029565651, 0.062906197303285754652, 0.12237959949079649236, 0.09075572324919976841, -0.1109212838060521511, -0.52445954952347562106, -0.51927697170811693006, 0.73964673759026389188, 0.3118740132315242497, 0.72988109940069889614, -0.5379900312673934204, -0.53486554003974617899, -0.070272057953074340553, 0.25832765534508583816, 0.040088434253151776498, 0.031434146199116892728, -0.027110589611621292716, -0.15311103493993366564, -0.25482183237914657337, -0.031673854202429990723, 1.1210188550886723835, 0.85307865797170501221, -3.153108080126970858, 0.8555767797164962607, 1.1122112236268875129, -0.065797784829020530828, -0.28425294604245932018, -0.046806598181672993964, 0.026176171699208702903, 0.031374195285508679754, -0.0082611750398880520047, -0.12497796652061507472, -0.24258366516217411135, -0.067799327444388129971, 1.0164223012547850544, 0.84080043848520269467, -2.9389724607538614265, 0.84494733693052515022, 1.0100934166920239043, -0.10088106822567456666, -0.27338814215764911664, -0.015579117106956197336, -0.029253374084831869351, -0.023811859174488799601, 0.06789482875215922808, 0.31742433893074278872, 0.64994865212452057612, 0.39754601719220350464, -2.1746637193214675143, -11.637406017526890523, 24.838712735170108914, -11.66779877446025715, -2.1638895645239828447, 0.51387661177974219839, 0.77984501641054204679, 0.13784191672530662909, 0.05858240521595543121, -0.083389208517058083237, -0.18355210192837839722, -0.18150481688133796032, 0.021015635787399782053, 0.51395452905589933401, 1.2452537104988348204, 0.31021309200606611567, -3.4136717402677976274, 0.33131463180439019167, 1.2756759983596108743, 0.52703487889347588879, -0.042631082688502831834, -0.38313081557250322495, 0.0089258852826593088753, 0.025539604637424155598, 0.02385149971668857824, -0.004579011845795713867, -0.078016152643650249487, -0.19304937049253492742, 0.066756194242414551576, 0.55948654716913348572, -0.81049187735069916361, 0.55179055755500616698, 0.051001072884971340726, -0.20408194641160967708, -0.061590067259644519504, 0.064114183507137392781, -0.017347937369949162356, -0.093798582014769080195, -0.076149932852563287367, 0.11320846449515691023, 0.40356176518042868873, 0.1459924775674020514, -0.77552439296881658937, 0.0097264913836863659685, 0.56909532605424772989, 0.018423999894274300376, -0.74903267468938017348, 0.19771303158070985262, 0.41411614654118816814, -0.15764715167146944874, -0.0037635188275608332138, -0.084040079410657844705, -0.088619181009211747679, 0.075112687861016408375, 0.38228401845735010545, 0.19063720494576738562, -0.73074659482303372382, -0.022409843936963644295, 0.55382355295785912208, -0.016574223967776330335, -0.70919620018676421402, 0.23837532601888514194, 0.39534193741059769911, -0.17862577841525570377, -0.018594355715698004244, 0.011451159396605568516, 0.01261475059658558398, -0.049948436202754048496, -0.14407361875976198995, -0.016378137414930001026, 0.76021919490470002412, 0.62449317647470947712, -2.3445302771952287735, 0.62848770065378911642, 0.75287318370231026687, -0.04654444492545666695, -0.17947152775168734307, 0.010193801863924257384, -0.011004085234300025817, 0.0092238768254223162113, 0.0075113020840790484939, -0.05141113879718886559, -0.13638304613898655226, -0.014748580341661260765, 0.704768935759394477, 0.600861612995514216, -2.2058202272753648643, 0.60455310811058993181, 0.69909256309045431088, -0.040431561469825030852, -0.16862700037340683235, 0.0027488329510667727437, -0.031695215732110627871, 0.10515361277118891181, 0.1683018516793771957, 0.074036274708314059323, -0.22763846424887274056, -0.6837101764440376428, -0.2871545976111838927, 1.0369724291881459166, -0.30559629541654292906, 1.0275837196715642552, -0.31495394101685580956, -0.71967628025497631938, -0.18992181753107845177, 0.34890002002369396861, 0.093585409192143101387, 0.02827355542974668498, -0.13994509697336002252, -0.3152318580636116474, -0.017044232323223745817, 0.71244211373255561526, 0.11670772156233491512, -0.51370097383881374586, 0.08711762230492420267, -0.52894803314224725366, 0.096249621251262199029, 0.69832937595741118031, -0.037207686684372062724, -0.28509837324347681031, -0.013411697248758940731, -0.0026551861827409407868, 0.00077823976323781606131, 0.01446067275204753963, 0.044717173971710710845, 0.060118125886577424011, -0.11886332333301287412, -0.18359114255538552807, 0.47947979097071113941, -0.57430505837520140489, 0.48535823845233944107, -0.1824071951781773282, -0.10788271432180240872, 0.098923522247017281472, -0.014523441712239791232, 0.013881742662363673324, 0.039184595879445817401, 0.02679712996773793432, -0.062548646062545842916, -0.19702587765213991355, -0.1885110369338347247, 0.73766322236411308033, 0.88609922261343743166, -2.4655616775507684935, 0.87720113095749985277, 0.70670362673796327524, -0.23005164852545209597, -0.13126733132545101435, 0.0073194523105376092451, -0.021113180593314042482, -0.020560143109682415402, 0.024803016310895775048, 0.046318740032848873533, 0.037089332036991815023, 0.025983802799174487291, -0.30867092793678579943, 0.022237937742049060375, 0.37573758677306223763, 0.027592245953954461918, -0.32508138586675561221, 0.023491193617389650311, 0.085822850652457080556, 0.050866110325202162912, -0.063723444645104668216, -0.1524033042171727137, -0.19075266432072768286, -0.13613191585492642743, 0.11987436462171814633, 0.69572256112486463575, 1.7187187365241973769, 0.36867830775224413875, -4.6876426663707251308, 0.40076318363806956446, 1.7225278664323779854, 0.54117008137355948616, -0.3705118854107771198, -0.033710661882469517003, -0.049790013768664556637, 0.026674577992462412263, 0.1318717423250543308, 0.089292394807958822001, -0.024964957368216880024, -0.063555227011659165548, -0.5214351137456368912, 0.023559938930233420562, 0.89996323007242506886, -0.037326700101111227881, -0.62344498814220383753, -0.040393867337880605406, 0.22407433135815446601, 0.022484420521419738737, -0.02798048980019818277, -0.064433629093024968881, -0.033678461300264826705, 0.10537864124331254578, 0.25299920982487361654, -0.040743550559617730533, -0.6129733871730871364, 0.22902769583995874481, 0.33887761791449766413, 0.24447631324225621574, -0.59978632714686852534, -0.030646656784456792477, 0.22069948724549501295, -0.0035946003532074536149, 0.019083562175874615313, 0.0093305604467312664579, -0.019973182272970190637, -0.04985472821526067827, -0.05461906126221050678, -0.022616719497950070472, 0.30251048059573737437, 0.4940320079260849262, -1.350792134823362467, 0.5021041758416466827, 0.30903192366594917084, -0.036259832528312600786, -0.098972042529871392946, 0.0072897308913139899703, 0.019895821584909999941, 0.0041714086123086531466, -0.016029449137217710586, 0.01127923153289098375, 0.065946317400394457997, -0.17507182418380851097, -0.32771119166001994039, 0.55473686308424974278, -0.32727657070546978257, 0.57178834455881488719, -0.30189539804802695278, -0.14338973438571683827, 0.05380235547386472833, -0.0092311077647824774789, 0.02530620482499838092, 0.055995375596033812593, 0.06320310972065697197, -0.023087276777352939289, -0.2010446117102721586, -0.28582366950570992081, 0.13397151067454854401, 0.68132410408580512673, -0.91021690333713412002, 0.68061176451411409261, 0.11440248670023742128, -0.28890932709695904546, -0.042604189743384542066, -0.029732348169237966112, 0.017366597844691245012, 0.076730843448675589702, 0.12993998938857884151, 0.093776425168251895115, -0.13494425329646772549, -0.5821187850187624635, -0.5270869531110244921, 0.79310608366543078329, 0.28857344360457071186, 0.76743510876742393467, -0.57495185754899458264, -0.55343399709077245596, 0.22696215178521467193, -0.030460526313214022348, 0.014655430009526062771, 0.070933005254365966152, 0.12071536743201394248, 0.088626179299970672676, -0.11092128341546028669, -0.52426526305758058832, -0.52262478407611423492, 0.72988109880982687816, 0.29397467227038243109, 0.70464340767485422745, -0.56765452152873319136, -0.49312665174154413084, 0.21822763414977688901, 0.01363447327601076571, 0.026941904778880917254, 0.025130029839805230474, -0.024207468907469303909, -0.14825110045925454783, -0.25482183237632266559, -0.037982198780358224222, 1.1139864386953681752, 0.85557677981621804708, -3.1377270389595466327, 0.87462698444298192513, 1.1153916601677380527, -0.095530362669905347084, -0.32901464615699232974, 0.0092247567963309844152, 0.018982119262741167787, 0.025296363964574162392, -0.0068562369696193959864, -0.12123997135930160429, -0.24258366510363651991, -0.072974190886624129559, 1.011242605205925571, 0.84494733712011182991, -2.9242800994705659967, 0.8572072812467991243, 1.0008784072655156905, -0.12960026605229388319, -0.27226665894457913186, -0.026486625591039307526, -0.018913102510076839485, -0.003676913432652477616, 0.077688225523914231085, 0.31727372139626708325, 0.64994865268147938853, 0.40113587812081347161, -2.1781597771188510038, -11.667798774391425098, 24.774541837566562918, -11.711219956162306488, -2.0764066673611112357, 0.71364242881536676943, 0.74551776887826337603, 0.070792542858639617709, -0.0063514851938412317245, -0.10501542973347056942, -0.18537026220240307439, -0.17921503600963775704, 0.021015636006246708745, 0.51338345725754042093, 1.2512963184393015581, 0.33131463174111769332, -3.3739417850249742159, 0.38530984962478398437, 1.2997215669988997355, 0.41365257537400379473, -0.42873955785988715483, -0.012528910506975904707, 0.019240467754093631969, 0.030924636430132892984, 0.024883629472535338106, -0.0048093269832354375651, -0.078016152482536432755, -0.1936598598752720668, 0.06329862209914606519, 0.55179055715531477588, -0.81962363508268432088, 0.55191055638911035164, 0.058514816650838942513, -0.1933249848627015266, -0.0015303055670977192947, 0.023360500299728796658, -0.041693574720188079918, -0.097784906283880757005, -0.07744984775669283561, 0.11097077933571677932, 0.40356176500739554536, 0.1511394945989693317, -0.76636340274538894413, 0.01842399930034965827, 0.5756774709973951909, 0.028571803711533342446, -0.74581638742929556862, 0.17466499606890723006, 0.250678214093724816, 0.023727516024927328142, -0.029256501864553769582, -0.087564332383231330281, -0.089030007506321881539, 0.07348440997506140715, 0.38228401832296088347, 0.19500651558458065304, -0.72335009429825170013, -0.016574223716719967137, 0.55791308453163424019, -0.0035838419925814119643, -0.69179339463687639089, 0.23070430652337295396, 0.18505286852282012977, -0.015738549731770634654, 5.799067953121358511e-05, 0.012694919324743515754, 0.0090717656232314029807, -0.052116643431460052405, -0.14407361854656397515, -0.014766921351858426636, 0.763376966316504868, 0.6284877006895884799, -2.3435766776975608394, 0.62512068025930045412, 0.7378650897993135338, -0.07017902311995150022, -0.13486643062849398844, -0.0097863683596752567423, 0.00080662211766112986572, 0.0094884045966385464899, 0.0052227268167871990917, -0.052513426022528418569, -0.13638304657428046429, -0.014192068868970337051, 0.7064832846128213939, 0.60455310813873941456, -2.20244847443785563, 0.60510715442158058419, 0.68539789621837177158, -0.066122540102268340112, -0.1345979736712417385, -0.049100135517647053307, 0.023058433552139729478, 0.11375745446225293822, 0.16463048964610901836, 0.071693943931866288555, -0.22763846426842995752, -0.68441877413268703556, -0.29111410764360773351, 1.0275837194437233979, -0.32777725041396627015, 0.98468526929744737419, -0.35948270237511043002, -0.65218654134991949078, 0.19654205490540213597, 0.01759497358773350062, 0.066064690038986739751, 0.028757172461326103302, -0.12967640154480883363, -0.30822264130857396358, -0.017044231870290027508, 0.7056108335129537279, 0.10384691727495254376, -0.52894803259680422958, 0.081404226070250926273, -0.49729858156724759954, 0.19407089000753321995, 0.81623519455473347861, -0.5341558247601435605, -0.016330791612145022923, -0.0016987357093572758721, 0.0053621992349167637779, 0.0044152713916098028715, 0.014460673240430550746, 0.041791615092811512655, 0.054601444854305278276, -0.12555420840391176562, -0.18813140369735431578, 0.48535823851173781618, -0.56420551299350518715, 0.48064319958527518262, -0.23730060994168999944, 0.044269386259227251623, -0.0069626464942764794072, -0.0043377309181603243907, 0.013125971864792940549, 0.036200576233281697713, 0.026797130003582157742, -0.059605188189433680468, -0.19547486610091058257, -0.19341987788155184935, 0.72500349179375200848, 0.87720113077352845909, -2.4477029080053118726, 0.94733664357119340416, 0.72147598774201482374, -0.44026442718003278021, 0.0083024774218790394625, -0.0041544908440042374251, -0.020549482536953163347, -0.018546343649368683681, 0.024803016012894748415, 0.045087044927095393243, 0.037208085980277111338, 0.030989242017053910544, -0.29778052072574512366, 0.027592246034016953277, 0.34073560305474670518, -0.021206615905590472243, -0.28072009410273751628, 0.12810740590560443231, 0.071871354959239394655, -0.0075805442077351032487, -0.090858929483081052325, -0.16098655939487627187, -0.19075266505762114577, -0.13279695942612929072, 0.1269765684395809946, 0.71223200631626193413, 1.7475409825388681373, 0.40076318382761694226, -4.6755267454015294604, 0.40476325513299821202, 1.6644269881410396206, 0.161262554604526126, 0.019133053948598391369, -0.050332619354192170702, -0.058919026303692764723, 0.02200259284041820862, 0.13187174203915216775, 0.097735339989188929621, -0.0047262882638284984746, -0.037636896090748640786, -0.51808039522253934717, -0.037326700538963281106, 0.78064726722944277615, -0.053461884496044748127, -0.41104491011497557018, 0.11602591906394456311, 0.017916070127008139334, 0.001907160212877305602, -0.030443569889675971968, -0.062316977151025854109, -0.033678461107131477115, 0.10330035721223064515, 0.25345853192169381929, -0.032021598176224963517, -0.59631372828353546023, 0.24447631343860232178, 0.32306330758574575146, 0.15891053103418373293, -0.73329875719007342472, 0.38764695146974470497, -0.0077606308495418033569, 0.0084675404236646441281, 0.017805796484719929595, 0.0061730057686007868203, -0.019973182368264435355, -0.047558648145818283037, -0.05499023644751253187, -0.030781865087792216473, 0.29094004466085182647, 0.50210417627841252841, -1.3161613833272347573, 0.53624981376510172382, 0.27071845362828550474, -0.15419113473499757028, -0.013506100675263747746, 0.017203291256541908272, 0.027002418458696105658, 0.0075555520995466331907, -0.016029449558485647787, 0.0075965823246313731593, 0.058227212614198484153, -0.18388675122451159094, -0.33022115323397360775, 0.57178834470656803024, -0.29504703175277896543, 0.5714865807188365121, -0.4313375704422143242, 0.0058250841736132040125, -0.019286682550908783307, 0.0053941770821612621684, 0.033798301338509487146, 0.059669318997154913164, 0.063203109715084318521, -0.025862384354642928286, -0.20631638695630247216, -0.29367920455837814853, 0.1250413541856762456, 0.68061176432193393193, -0.90049349816641932609, 0.7155104909870714236, 0.14924269746465529396, -0.3945051331179418308, -0.024595469964286683501, -0.0086253925385847904361, 0.0255841860162596621, 0.078055681606249968651, 0.12993998944281850982, 0.095802221122404648823, -0.13176277570275710116, -0.58399764205068693812, -0.54271787597795229097, 0.76743510898099975837, 0.26501555977633017847, 0.76628889413462852875, -0.55527125007534439138, -0.29615276711286020461, -0.023962388889691606242, -0.0095321784236211334895, 0.022474772070434172883, 0.071983042541954103677, 0.12071536790540912665, 0.090755722724907755827, -0.10778862013907145045, -0.52624989063266125466, -0.53799003075566931287, 0.70464340739469610853, 0.2709983290738089412, 0.70561766214792398522, -0.54875539340741885663, -0.2466826540671734902, -0.0035833160873409088294, 0.015216698654496492449, 0.030261778153800787672, 0.027955412969940543477, -0.024207469041131232657, -0.15311103481428939288, -0.26461978931390206515, -0.048114192590241411374, 1.1122112235526331325, 0.87462698447090969633, -3.0899346299617493905, 0.96132349373796965875, 1.1137878776243177459, -0.54885772557251044645, -0.0015641854781762532054, 0.0089435479814702640439, 0.02146194987762392245, 0.027604560787918164949, -0.0068562372236938232822, -0.12497796638239942746, -0.24982509366107125137, -0.079989583709045700144, 1.0100934165279233934, 0.85720728109892196933, -2.8915303010124495664, 0.9329110849211053047, 1.0145155868262676346, -0.51581275829835604441, -0.02568521554281685132, -0.011279333092102168667, -0.0063297210798984029564, 0.001634500021988365007, 0.077688225853036446678, 0.31742433866703428924, 0.65667337084954602311, 0.41863614016353717373, -2.1638895645169431425, -11.711219956517650687, 24.602546605900133869, -11.890171597153655014, -1.851032390123594551, 1.5648461054667497372, 0.049939193755784157258, 0.028291335243782359671, -0.023779297765594020059, -0.10850689956481666953, -0.18537026207606976946, -0.18150481713858340749, 0.017481080259517747105, 0.5172278922878390528, 1.2756759983951417858, 0.38530984955582170404, -3.3192576208011317362, 0.39666069561108668884, 1.1949910191529407033, -0.025787926546459574717, -0.016789695395632235619, 0.0040589965808959351315, 0.023738013446970233911, 0.031113239998115375939, 0.024883629838160521014, -0.0045790123205546370139, -0.080177251121708315029, -0.20170373048617390443, 0.051001073250100760481, 0.55191055622177775941, -0.79811815322575696552, 0.57688292169882049443, 0.033807853794598524688, -0.20019681041419762346, 0.031142069836745986389, -0.0043094320462983724471, -0.052910306668947346131, -0.10061940648931286257, -0.077449847680428451913, 0.11320846434154500926, 0.41114563828827277447, 0.16619854847822748689, -0.7490326740950140616, 0.028571803306981478215, 0.55502728060123263099, -0.06620708631644163511, -0.91495983993115115318, 0.66811521820027097718, 0.025801911006570331375, 0.0016111985033405247427, -0.039151135428085639956, -0.09028345327476035298, -0.089030006449217991005, 0.075112687768809721645, 0.3876314041521184417, 0.20612057388351173737, -0.7091962008874276302, -0.0035838408846707806715, 0.55177665435778977532, -0.075737883878196091136, -0.86403981635913729153, 0.6306921970094505614, -0.0037859500872207312375, -0.0084649237686416651388, -0.0014891098598960577498, 0.010234285021052474951, 0.0090717660484430744094, -0.049948436771622806152, -0.14288498669354296466, -0.019060966471516525844, 0.75287318387403678877, 0.62512068079237481655, -2.3289072521060938392, 0.67373607019357417691, 0.77900693852011249074, -0.2924576796325903949, -0.0022735256251749371911, -0.0051691962527553479353, -0.00014824140095285156041, 0.0079690179090456982058, 0.0052227259738986111723, -0.051411138458099817083, -0.13649953272052678388, -0.018265855005200490324, 0.69909256324617152778, 0.60510715416273019862, -2.1880719114557218852, 0.64752414475877118072, 0.7186287051668657222, -0.27881830664928924923, -0.034754356430543878465, -0.018632880930678233367, 0.033950685691547667633, 0.11543084849878675058, 0.16463048968189458732, 0.074036274624726783333, -0.22394321145802092987, -0.68757457294033008655, -0.31495394098876045019, 0.98468526898510078116, -0.35268228480469898667, 1.0334586751442065289, -0.30523104542349893675, -0.48605846533484481364, -0.023948399522524227434, 0.04171509851932404811, 0.074305480339949200874, 0.031161712131625166761, -0.12967640210096992304, -0.3152318581697128308, -0.036336573758392047562, 0.68124892310686413488, 0.096249621007451546806, -0.49729858163642598479, 0.16582832355951448999, -0.3449989345186705858, 0.19990842649450921154, 0.062915021724105554823, -0.010679191401352485305, -0.0037840857495748860678, 0.0030702049929988632464, 0.0053621984085804787895, 0.00077824014447393208803, 0.0075380781997473741626, 0.034647358292786548106, 0.053613208284470435794, -0.11486202020694906423, -0.18240719501252555812, 0.48064319919208303711, -0.58515737335623396742, 0.52283278372450703309, -0.21686046345413489966, 0.00051416123592601381802, -0.0076518247187237027271, -0.0056474566917197634164, 0.013125972145432998617, 0.039184595709292134591, 0.031788082848057570839, -0.058087004220870652738, -0.20849442895657913088, -0.22523977291793331745, 0.70670362680536080724, 0.94733664354722613155, -2.3377339811107060719, 1.136256831600642947, -0.027644008707477133596, 0.0031438636302495840714, 0.004772093226828272132, -0.0039525862947226214281, -0.020549482422488964811, -0.020560143011183116407, 0.022839535434671015546, 0.048853700838049032851, 0.049891493465905364935, 0.043227824504523261473, -0.32508138621218513054, -0.021206615654410190808, 0.37562052006203572319, -0.061961736989660981534, -0.097321762117940427417, 0.055542019724765280542, 0.024931882555067071128, -0.025816854545203382632, -0.090858930227593365236, -0.15240330310264177394, -0.18148884805998274916, -0.12886686484172082956, 0.11830374670653789893, 0.6877827998076725402, 1.7225278662821279507, 0.40476325521807865471, -4.7109493669683031314, 0.58916745458433716909, 1.7132388431318887889, 0.030835860704878526733, -0.015703773271719088933, -0.058700074540060716288, -0.058919026486836466316, 0.026674578820721710215, 0.14772705762091115078, 0.12447589761851439449, 0.0056671111154820387121, -0.084607022625245942304, -0.62344498781453150382, -0.053461884848749509358, 0.9795187847490011146, -0.086754669030796163764, -0.34201089166154496901, 0.0069144352262191267899, 0.011546051672566828503, 0.00058011508735246800333, -0.030443569565694928586, -0.064433629521701160647, -0.03778963706742759221, 0.10481303756152847428, 0.27200819304829865208, 0.0008137596436750664259, -0.59978632777726426095, 0.15891053151379949671, 0.18225737702926608774, 0.17466885543417656756, -0.18301959163809036757, -0.002852571846496530783, -0.003268534176059164232, 0.0067971462767812381237, 0.017805796839987449859, 0.0093305601212344650708, -0.018264074259172743964, -0.054020811384697213497, -0.068791600921242465616, -0.033724076436636989118, 0.30903192282610836283, 0.53624981439902863478, -1.3059876236280125728, 0.5631369819921847597, 0.048054956090453362916, -0.017107642201047667391, 0.0038159938510618834559, 0.022609151312705685205, 0.027002417817064191391, 0.0041714089990869777924, -0.023598432903669648575, -0.00068867167668758635674, 0.060867685991248005051, -0.15877775699714599678, -0.30189539799210862725, 0.57148658100116667136, -0.38108109988423177761, 0.55081687706062631982, -0.35932857237441340192, -0.018111767771906886471, -0.0045210471500383526619, 0.012448687753730845046, 0.03379830148855207661, 0.055995375481268094831, 0.060557170747585584158, -0.025148186011618563207, -0.20687313449544733612, -0.3000732113275819124, 0.1144024868559551239, 0.71551049071680450364, -0.79113732576542727948, 0.78461344486184281255, -0.43286485997848805907, -0.014402098254270488675, -0.014691130970831777977, -0.0040093126302540608649, 0.025584185866974886586, 0.076730843499681233855, 0.13178391810665673889, 0.1027840653640985924, -0.12630864525605139947, -0.59267271068117377553, -0.57495185772547330316, 0.76628889424159751798, 0.35879127373204450224, 0.90056683207387266954, -1.0459291096731824222, -0.012907586446989714746, -0.014706146875329281082, -0.0054251251087782592294, 0.022474772234663358689, 0.070933004623939699407, 0.12237959995608850561, 0.096539559459080867554, -0.10386443573127401852, -0.53486554034394639956, -0.56765452131858673557, 0.70561766207736664747, 0.3565234233783562412, 0.8164983627499079466, -0.96172836444073506179, -0.011182179299216327487, 0.0063070892084945585215, 0.020325475869773880488, 0.030261778170470633731, 0.02513002987369855265, -0.027110589766927626482, -0.15580187066993977929, -0.27358024491824717295, -0.06579778482451709154, 1.1153916602558713311, 0.96132349375267522884, -2.9898268834838312991, 1.0904262106210806937, 0.28540736639282970888, -0.0074870785105930490536, 0.0044275787069393236425, 0.012960199466222784709, 0.021461949723077293484, 0.025296364303747573971, -0.0082611753219110958119, -0.12481764550753844667, -0.25638019137879947129, -0.10088106823915110599, 1.0008784076968419985, 0.93291108462237037902, -2.7800504460006085772, 1.0812401138697445013, 0.20821567852989250613, -0.017675237319513880524, -0.011335957673972437559, -0.0059450680669098782127, -0.0063297206088572715091, -0.0036769136279980675804, 0.067894828515238009325, 0.31584532414106675224, 0.69226999591935933775, 0.51387661164502940192, -2.0764066669614251737, -11.89017159737263718, 23.865210008775342487, -12.105590609284655912, 0.6223640270265142771, 0.021689876759529994749, 0.030718271135422799828, 0.019990998741620948087, -0.023779297996563465017, -0.10501542985388848117, -0.18355210151460119028, -0.18377366341667505933, 0.014502074467388719956, 0.52703487860994224778, 1.2997215673213129428, 0.39666069558943278794, -3.3922446975198163877, 0.4094846466035089172, 1.1927450151871781259, -0.010556570416683430311, -0.0062595890308587646272, 0.0067015075426507687145, 0.023738014047743821039, 0.030924635525117063345, 0.023851500333896540063, -0.008656774848632633676, -0.087542743545762830015, -0.2040819464723394594, 0.058514816632792544804, 0.57688292179819788963, -0.76849182330010468167, 0.63979572057118727635, -0.27619971962801409537, 0.023071747370242728925, 0.010803141457259695865, -0.012800593723221346393, -0.052910306617912530869, -0.097784906305484600719, -0.076149932461050895416, 0.11504015836625937352, 0.42445920408377418864, 0.197713031441310777, -0.74581638708317266051, -0.066207086813956414484, 0.34621553464386684018, -0.1304536683720257384, 0.062774377561402833203, 0.018001651757878593674, 0.010537275930694006698, -0.0056387939992863112676, -0.039151133941777563341, -0.087564334034249527439, -0.088619180870900537972, 0.073805198168820182469, 0.3964919305750950751, 0.23837532673211414558, -0.6917933961820780242, -0.075737882955463978196, 0.3405722129322449887, -0.20111385199983372596, 0.110935462632264073, 0.0053749701185196920722, -0.0068591471098115690264, -0.010410685159665263033, -0.0014891093030526682088, 0.012694918685357889712, 0.012614751057235841503, -0.051072525057518858294, -0.15838399362654342895, -0.046544445400357910281, 0.73786508926574378986, 0.6737360705664046101, -2.2216487860439073643, 0.82786871806152295683, 0.2308718029627383761, 0.00352922923150013031, -0.0040729839565279112035, -0.0062531581586898685177, -0.00014824227998156986469, 0.0094884058011917050968, 0.0075113019673251485431, -0.052373389213511492568, -0.14858217506305052003, -0.04043156166602530216, 0.68539789683113971908, 0.64752414416454529, -2.0913978953094547464, 0.78484673824712058998, 0.20977620310886432864, -0.016942512174490029914, -0.023365729293086223706, -0.012164370436198244929, 0.033950685694397991776, 0.11375745442763829374, 0.16830185179998921607, 0.083308846510621703541, -0.2250364330485654929, -0.71967628011840434343, -0.35948270216831373913, 1.0334586748903302755, -0.19435810350243457045, 1.2275615053042550073, -1.121185259353740804, -0.028811758130755735913, 0.0078177027320457219417, 0.049769639187716908924, 0.074305479691452666624, 0.028757173174667048643, -0.13994509731343632297, -0.33980368041933023404, -0.06333143461961023013, 0.69832937549516527209, 0.19407089075992442484, -0.34499893477049781154, 0.13514094435359774726, -1.0280866601896321288, 0.76780424211513809141, 0.00031150471732814170129, -0.0036219878153334762044, -0.0037840888729637543568, -0.001698733271568499039, -0.0026551878063876348766, -0.0082169642541222477627, 0.0027872701280738415748, 0.043787428261974209454, 0.072852661346491764949, -0.10788271470922000672, -0.23730060934102956138, 0.52283278340298111608, -0.32794487736810279399, 0.044885819734941399817, 0.0028120587268211398244, -0.0036146879915965857664, -0.0076518243981211885724, -0.0043377315332793755717, 0.013881743054867881279, 0.039663218040731758329, 0.031109752429245213223, -0.062924727838465366325, -0.2164340013720542355, -0.23005164811765921318, 0.721475987372412364, 1.1362568316931129786, -2.672058106726892035, 1.2580254514665227816, -0.00050443077890597513933, 0.0042975596708546273325, 0.0047720935348418268979, -0.0041544913330982388353, -0.021113180238749647277, -0.020470709943144069914, 0.027700806808159235406, 0.054173555370055408165, 0.025177486480748318548, 0.023491194284807618425, -0.28072009474678072394, -0.06196173666564438326, 0.55902569500716547957, -0.31220889283310593321, 0.021588725158317452479, 0.031882552511722069, 0.024931882543977310074, -0.0075805436939306077709, -0.063723445124867883482, -0.12994837605578182793, -0.18603571919634775966, -0.18871543472917240436, -0.0082032751055245989313, 0.54117008141788403019, 1.6644269880098865322, 0.58916745475485821082, -4.9280639492097133569, 2.6156886264749701354, 0.015939335502891026353, 0.0098560126209116492857, -0.015703772944292856972, -0.050332619145407611616, -0.049790014529285817058, 0.032642047424728683203, 0.14007788226622139294, 0.079116391727785839727, -0.072626259113777652532, -0.040393867126103114029, -0.41104491004565302248, -0.086754669219804639302, 1.1044676884728126254, -0.66202308193814352055, -0.00087105440068529647792, 0.0078572826183998072114, 0.011546052036917815384, 0.001907159860607935244, -0.02798048943297733529, -0.062619736403436293548, -0.037409194225701609959, 0.10788066596789221141, 0.27297748395721732617, -0.030646656139956922338, -0.73329875771976515964, 0.17466885567062026552, 0.82880838895190489524, -0.51934267656008981096, 0.0012276066323801083577, -0.0049940348380703922238, -0.0032685337863560272841, 0.0084675398759972864832, 0.0190835627434304822, 0.0066473920168580750631, -0.02354474985759092906, -0.048481548542430370763, -0.054101618925017060102, -0.036259831839754036453, 0.27071845300530350542, 0.56313698235431974837, -1.341681212322815453, 0.64605931991735832121, -0.0085614398064415321576, -0.0062715333502399031942, 0.0038159937401287054477, 0.017203291442215152413, 0.019895821787560007826, -0.00064164403139391440294, -0.019625729641798737163, 0.023397888949066285996, 0.10478506685876051718, -0.14338973474904601058, -0.43133757055456006801, 0.55081687731435069466, -0.021250025300186659027, -0.082522493635300397941, -0.0093271214488516180491, -0.0083981331268898409825, -0.0045210467078258291182, 0.0053941767134307819914, 0.02530620498681438249, 0.053356691054251129125, 0.070184467163534325351, -0.0017281719586272233091, -0.18894089302353603288, -0.28890932714554867777, 0.14924269779963225679, 0.78461344459797099926, -0.97135575110614547079, 0.3979379720799537945, -0.0046324925352400833645, -0.011268353528303621386, -0.014691130718884183246, -0.0086253926425485932805, 0.017366597859189297076, 0.06913529259663216131, 0.13238890768209760074, 0.12265912954848372241, -0.0896261705168753392, -0.55343399713014251873, -0.5552712500090565273, 0.90056683185139818626, 0.47181163450182844743, -0.46131567980631110748, -0.0029813763048098505283, -0.010515793896572410773, -0.014706146638347379965, -0.009532178793959029256, 0.01465543051818714651, 0.062906197301078464745, 0.12119494907481331847, 0.11294544037665063541, -0.070272057726024572677, -0.49312665194732419049, -0.54875539326532019047, 0.81649836262377772922, 0.49357667747970712657, -0.45926949526904359278, -0.0092199319386195143639, -0.0018304299003877541087, 0.00630708918881074318, 0.01521669865111413035, 0.026941904770754324105, 0.031434146243598935777, -0.011097343254112280661, -0.14364443910093913082, -0.28425294602609496586, -0.095530362851028186033, 1.1137878776992877761, 1.090426210687200026, -3.4989649846808665856, 1.7705958314028376055, -0.0069339686908168581339, -0.00066801527244928393841, 0.0044275785961818102388, 0.0089435481550016234648, 0.018982119049810723677, 0.031374195453611274098, 0.0059984946774494091182, -0.11707282852874557888, -0.27338814208213246815, -0.12960026639213487187, 1.0145155870901558792, 1.0812401138027427638, -3.2508776018601714952, 1.6220054378645554216, -0.0045677935950762995482, -0.010698381944907655455, -0.011335957150062076285, -0.011279333603060291702, -0.018913102573277374568, -0.023811858955645828079, 0.048726902699913719752, 0.32616290081670135104, 0.7798450165184542815, 0.71364242881194961399, -1.8510323901627416809, -12.105590609247444789, 21.940899088839994846, -9.7965432660301843271, -0.0029870993613394965227, 0.016937173969289953668, 0.0307182712565629637, 0.028291335329083441841, -0.006351484988359161997, -0.08338920870139227004, -0.16959947127787100563, -0.19227185103055441373, -0.042631082408231990577, 0.413652575050336202, 1.1949910191134029969, 0.40948464666677908408, -3.572912617172143257, 1.973904748399126774, -0.0019429951392080371201, -0.0073172418550115888095, -0.0062595887578177364721, 0.0040589959491210223647, 0.019240468562853219048, 0.025539604199431287929, 0.021531564085416610849, 0.0039111541523621220703, -0.061590067469989732973, -0.19332498475886195077, 0.03380785366782734086, 0.63979572072973422969, -0.77597708729743408718, 0.30424861377077089264, 0.0082107527383266029619, 0.012752791482547026075, 0.01080314151382618694, -0.0043094318303693937566, -0.04169357491670763749, -0.093798582378485109712, -0.087569729311923735771, 0.096234057363134242746, 0.41411614658985229642, 0.17466499589318848495, -0.91495983937439961942, -0.13045366882611150738, 1.2382510234259032078, -0.69762626668098604021, 0.0059499009297377660963, 0.0099699322358621569884, 0.010537276691904398293, 0.001611197286263581609, -0.029256500998905685118, -0.084040079402490169835, -0.098936614789403756753, 0.058674767976130706715, 0.39534193726460348239, 0.23070430742466949137, -0.86403981714915467638, -0.20111385176625157234, 1.194027389222676705, -0.6426992639303940047, 0.0077413291208652951214, 0.00021271612922503762688, -0.0068591470832741075619, -0.0084649240488509497338, 5.7990931901664728026e-05, 0.011451159211761496176, 0.0060969486125215851396, -0.061613753021433792512, -0.17947152721106066076, -0.070179023083433961228, 0.77900693837196088865, 0.82786871818623108954, -2.5370431685294785318, 1.228914938546957325, 0.0051463289014875950034, 0.00047523125506829288003, -0.0040729845215742768033, -0.0051691952464103995898, 0.00080662122210143965759, 0.0092238768005758579871, 0.0044452746979128177421, -0.058794862135688699678, -0.16862700022578114289, -0.066122540688856257241, 0.71862870592820227156, 0.78484673783090552757, -2.3768991083140722509, 1.1551044383518087599, -0.0028691157624474014284, -0.014000273646585901718, -0.023365729199450128301, -0.018632880864578663105, 0.023058433470196825626, 0.1051536126940299104, 0.16735605928958627309, 0.094258864754403356989, -0.18992181767189700303, -0.65218654138096010531, -0.30523104515643884049, 1.2275615051164787683, -0.26735725859203418597, -0.12824834751859881532, -0.013279165559988408835, -0.011522585231132375511, 0.0078177024392084831089, 0.041715098891472228437, 0.066064689833016915221, 0.028273555823016386507, -0.12882366522216240345, -0.32461630670377983332, -0.037207686310890067394, 0.81623519362766439045, 0.19990842685430565417, -1.0280866600069167305, 0.13991990801613748974, 0.24884824488467655046, 0.012756049753040859043, 0.00031150326115866324665, -0.010679189092536858774, -0.016330792974219437985, -0.013411696418787194887, -0.0048309325054647642497, -0.00039719122104895314808, 0.00288758175597366969, 0.031405980222439969229, 0.098923522428543145257, 0.044269386168391017256, -0.21686046349682536194, 0.044885819848284061462, 0.02180988650993710351, 0.00045512654001833037888, 0.0028120587350227107572, 0.0005141610664221666618, -0.006962646164852900714, -0.014523441924906559108, -0.0071609610234654860403, 0.030381071872411225671, 0.079327196061624624668, 0.058165260829245030161, -0.13126733161452022736, -0.44026442689854128343, -0.027644008767933603288, 1.2580254514380821984, -0.8114510459598524017, -0.0012697102400650276018, -0.00050443074222908889336, 0.0031438634519062565523, 0.0083024777008864434485, 0.0073194520832651384368, -0.011020752652058502657, -0.039174876924564200342, -0.03084713193173787768, 0.027224450951006323035, 0.085822850313401255473, 0.12810740626423802735, -0.097321762309705669991, -0.31220889279885583045, 0.23461990998419626764, -0.017368443442930824733, 0.021588725267125959822, 0.055542019671893906596, 0.071871354828933906034, 0.05086611034347189586, -0.024568473220949363367, -0.15338469970238452489, -0.30816652645795278298, -0.42874913195548081113, -0.37051188540744206534, 0.16126255467829120338, 1.7132388429890028636, 2.615688626537546746, -3.5186879588292159227, -0.0089414527863764716592, 0.015939335634981195017, 0.030835860504756704842, 0.019133053835162647932, -0.033710661633607329202, -0.10401201137878460234, -0.10514425761666704173, 0.037459094978185517033, 0.21070093655601776494, 0.22407433115646568922, 0.11602591914692327035, -0.34201089159656500316, -0.66202308196830572662, 0.60905283904337181866, -0.0042600470599277167477, -0.00087105427483972401679, 0.0069144350045823779055, 0.017916070232444642613, 0.022484420415432977702, 0.00017076511596131422276, -0.063449022985085437587, -0.1268136691717804132, -0.064242470456356129405, 0.22069948702138761365, 0.38764695164698526053, -0.18301959175543705549, -0.51934267651246657227, 0.30328436495343519264, 0.0011045067784375528063, 0.001227606760796212497, -0.0028525720865286582038, -0.0077606306311318127428, -0.0035946005883846403675, 0.015276114901139122015, 0.032309904745874924725, 0.020416523921344645331, -0.029878372609623393696, -0.098972042734907367256, -0.15419113456141786411, 0.048054955918125953629, 0.64605931999274879285, -0.47441998198789636643, 0.0050401111337566376069, -0.0085614397027482039282, -0.017107642272964077718, -0.01350610052711238504, 0.0072897306620110912528, 0.036370070983008435939, 0.04497733119443072386, 0.014098181756551676769, -0.010708898227529600744, 0.053802355684914274225, 0.005825084106168321868, -0.35932857244268728847, -0.082522493623906942317, 0.34340333007055573455, 0.0038759206418807153928, -0.0093271212400824476951, -0.018111768102209119952, -0.019286682341109969568, -0.0092311078310125145585, 0.017381558264591909224, 0.065428655226825810365, 0.1277112197958793216, 0.13404168261128951434, -0.042604189743163732584, -0.39450513326881453402, -0.43286485980937278173, 0.39793797201303376943, 0.2066623243224687545, 0.003606717794149855421, -0.004632492288285303822, -0.01440209858515219786, -0.02459546972871740414, -0.02973234823935154772, -0.01509519351126595299, 0.042087256719807758509, 0.15756122944445652423, 0.28045132359913738096, 0.2269621518496735546, -0.29615276728665068662, -1.0459291093634943692, -0.46131568000635264637, 1.2397598304372923206, 0.004736365866142844272, -0.0029813761301424406523, -0.012907586703148156426, -0.023962388605187841945, -0.03046052651084677626, -0.017833329696086166333, 0.035593162394284991179, 0.14364958285682785655, 0.25832765530377022056, 0.21822763427052158103, -0.24668265422147384047, -0.9617283642676276445, -0.45926949536673089547, 1.1476589359520272815, -0.00030485393251563399622, -0.0092199319669219459428, -0.011182179278859623317, -0.0035833161088917962062, 0.013634473308951863477, 0.040088434270869527376, 0.070190706668658758849, 0.069100335420351724491, -0.046806598197551743668, -0.32901464606863273232, -0.54885772563576762373, 0.28540736635580826741, 1.7705958314471572645, -1.3181296444983572336, -0.00081206680001032845086, -0.0069339687308843751465, -0.0074870784371538802449, -0.0015641855611863097163, 0.0092247568607993421508, 0.026176171677043821862, 0.052846494351442070692, 0.067100119121199869321, -0.015579117138946659699, -0.27226665885719170124, -0.51581275836375028998, 0.20821567853928651948, 1.6220054378774397819, -1.1812151165203308434, 0.0089739354336379056898, -0.0045677934693598995844, -0.017675237622308232616, -0.02568521536108795203, -0.02648662551448339661, -0.029253374133411797625, -0.047828859637985249309, -0.046988772695529826273, 0.13784191670045975453, 0.74551776878572129093, 1.5648461055541527109, 0.62236402699217507895, -9.7965432660250613139, 7.0497397868299849222, -0.02160848003118739924, -0.0029870991418172467523, 0.021689876532184276109, 0.049939193869743783716, 0.070792542722030060709, 0.058582405126229829262, -0.024370435024550499842, -0.19139613258526802508, -0.38313081557109246456, -0.42873955774995925472, -0.025787926589065517796, 1.1927450152660765692, 1.9739047483093505875, -2.359720287770875391, 0.0054076503668463891603, -0.0019429950898042461661, -0.010556570548675345972, -0.016789695199500451195, -0.012528910752503816944, 0.0089258854252365991272, 0.040531488177762170499, 0.0653253815316543196, 0.064114183641622302101, -0.0015303055916765020972, -0.2001968103893032036, -0.27619971972250872971, 0.30424861382643070273, 0.042805527612202118326, -0.0088453359147898074494, 0.0082107527832498746617, 0.023071747368506038273, 0.031142069635475683881, 0.023360500434561774774, -0.017347937220079295839, -0.10733933596220078532, -0.20840928714106699005, -0.15764715174136365028, 0.25067821413951257847, 0.66811521802331264031, 0.062774377773530648517, -0.69762626675046701674, 0.10715660956951557348, -0.007785627855186325226, 0.0059499010750284362162, 0.018001651414495056791, 0.02580191128880028259, 0.0237275159001470691, -0.0037635188462958328766, -0.080391510339159727438, -0.18785874560555876345, -0.17862577849933256036, 0.18505286830460820902, 0.63069219727107661555, 0.11093546253587548145, -0.64269926392286547134, 0.080528393051619948095, 0.0048124339199794837277, 0.0077413290073094194813, 0.0053749702472841096035, -0.0037859501647014796172, -0.015738549712672713493, -0.018594355622826692909, -0.0019300950559835833818, 0.026606247363685643637, 0.010193801599090902485, -0.13486643055089012067, -0.29245767957369822598, 0.23087180281304023333, 1.2289149386475961556, -1.0765786058939788017, 0.0035621187954222206795, 0.0051463287342844990166, 0.0035292295875596320712, -0.0022735260644481487466, -0.0097863680572611005254, -0.011004085161632357059, 0.0017224085067433776534, 0.023074998886537859827, 0.0027488328568978431451, -0.13459797347837476611, -0.27881830689918818411, 0.2097762032591545811, 1.1551044383345607791, -0.99390419447525413243, 0.0079771145689773274856, -0.0028691156887172401292, -0.016942512268208882037, -0.034754356471953233576, -0.04910013544686969017, -0.03169521571199316462, 0.05967648554219812107, 0.22542563318320282817, 0.34890002009309750708, 0.19654205490850581994, -0.48605846545887571031, -1.1211852592459825573, -0.12824834754633970824, 1.0897370159627992603, 0.008974424242622108111, -0.013279165591864075871, -0.028811758037987755116, -0.023948399540165508231, 0.017594973560644953936, 0.093585409036586228448, 0.14957077027250470502, 0.058177347677363495937, -0.2850983732785922764, -0.53415582439703146456, 0.062915021544667243392, 0.76780424200232966214, 0.24884824496213941991, -0.53160308505053810446, -3.129137569949643094, 1.0974348820594306275, 1.4619154624934243003, 0.90360785697487555623, 0.30004248324306004303, -0.077273423247476422104, -0.23775913122544511946, -0.27517309058663114163, -0.23549573142258345415, -0.15399983755399795915, -0.062092025501675716936, 0.015546385155181873666, 0.067989119052919225195, 0.096545482401106913484, 0.047599506226696570999, 0.058615639751954759507, -0.2106623566272930459, 0.033836228863951121892, 0.065889632178131055107, 0.0089065289800569109879, 0.0075038849976215519361, 0.0082787539220796624306, -0.0022456668347291468095, -0.010004511210810031543, -0.0093029946763316315184, -0.0042651459750982960786, 0.0011400428201290815824, 0.0053934689517870920694, 0.10317763266248923537, 0.044689497956640865384, -0.224470940755622822, -0.027153703705596707318, 0.06775481326797015702, 0.044773230061319280182, 0.012183893436401705898, -0.00087907116814224214082, -0.0033679295809082411897, -0.005650720924184769213, -0.0058391928185258928918, -0.0034988130816494280272, -0.00016799833371477352045, 0.003129423348001069468, 0.044180054997420468987, 0.31490399505810173775, -0.1826164369795986453, -0.19025820903236120851, -0.11493371942585331458, -0.015347098664096725518, 0.065262087313388311705, 0.080329152924254437762, 0.05205957810333427388, 0.01291762969093429754, -0.014173869322524124126, -0.022393763602730917817, -0.014913508057337243259, 0.0018210780073153881586, -0.21881262224688874385, 0.34282141042647745754, -0.13245192378106182551, -0.087672111594399373957, 0.030013274720897129694, 0.063277367311864632371, 0.046649995328541707384, 0.0046824462921538408805, -0.024286395158533757854, -0.029474565577335572891, -0.019269384473199084595, -0.005252047960865099431, 0.0062669659577557724692, 0.013722208539590892429, 0.35055924608978611401, 0.14326237044016820743, -0.62336455015034075711, -0.21895717884093152117, 0.1757969373670521529, 0.17804039393456216889, 0.073047580206386408808, -4.1423434235957453308e-05, -0.023963212912622787465, -0.021620845864173371398, -0.013162253027545881001, -0.0060727151601857876084, 3.1989148452641518253e-05, 0.005931035158367807883, 0.19544430761524017592, 0.13998114317719387301, -0.33988082653922840892, -0.15581764090452124294, 0.049879394742648672234, 0.088321118755618405194, 0.044651979266567373705, 0.007687219340036462005, -0.0021846166466837689361, -0.0029270312778300589494, -0.0024631259716007018398, -0.0022091070290440216126, -0.0021070954243172512754, -0.001973878945547613388, -1.3189472230300753886, 0.74787509095380533264, 0.90010463648697014438, 0.42600666340032539159, -0.2228064340857890957, -0.43618952065740096424, -0.29031166238089728404, -0.085144584789431446148, 0.050354484905311837351, 0.10053678288583890732, 0.090765113840664907174, 0.050035927635370278099, -0.00065606680484318663418, -0.050131189309804001508, -3.3426912940986150957, 2.4291396009034946779, 1.8525738337856507698, 0.24996892678127363707, -0.50038418222905767241, -0.58009699886536925106, -0.3667637065010055375, -0.12306655097766910911, 0.038117476482156069373, 0.10268312270871071679, 0.10065465237878351568, 0.062707031170682964172, 0.010153861236838940957, -0.044491692605975158292, 0.96796828766417486456, 0.67930941263859789814, -0.78881019553091247243, -0.94817326268867541117, -0.43112667975485774452, 0.013103646082751130644, 0.21447628091488887825, 0.24082313168201385789, 0.18667204503949108574, 0.10344364211907644546, 0.024420855107245155458, -0.029876537439421015085, -0.053533996527210025995, -0.052625451396110939362, 0.8610217000096108686, 0.57451571433554082802, -0.73299131939498440858, -0.82628272547803438552, -0.34969747037021070568, 0.039512157255354229957, 0.202045101478164707, 0.20838444823422760943, 0.15169777317350871537, 0.078164343003269062815, 0.013414788448902561363, -0.028369347496848825591, -0.044155659132961405655, -0.039842902702862698083, 0.91911171319812645653, -0.43333916271769046258, -0.94705452047675353633, -0.1868872545237877103, 0.36536398144022513845, 0.37660526120032372654, 0.14652064633634054314, -0.043352091314982015013, -0.10102377740508428072, -0.075438380902955542995, -0.030695750442342423919, 0.0014738892088995746502, 0.016219257210379121187, 0.018319561167881727448, 0.84187980048366861663, -0.42396813787529469275, -0.84969154795844026307, -0.13709867309285542203, 0.33030775208279866684, 0.32360036078338405563, 0.12322835854317881754, -0.037467451392486486594, -0.085591527513970880325, -0.064557338503003186547, -0.027492857165122917829, -0.00012559625508136947384, 0.013425423583849804138, 0.016689580204319862111, -2.3597202879539231901, 1.9739047489765957355, 1.1927450143311759678, -0.025787925961830621829, -0.42873955804028346384, -0.38313081547445276831, -0.19139613257195051133, -0.024370434810960059369, 0.058582404784466454717, 0.070792542871251154524, 0.049939193937320845718, 0.021689876467226251039, -0.0029870993056438973484, -0.021608479883882782818, 13.877882909691686919, -12.87441101081978978, -3.7380343782457323698, -0.17838617179125190737, 1.1796002096147071292, 1.3155679689686898826, 0.9179557094531386463, 0.42453349220363617578, 0.048550591320977574961, -0.14598390600207869494, -0.18570550672212401966, -0.136646046952614314, -0.055378422416946550266, 0.028223490956334799928, -1.8912859683543594702, 1.7890318116039809482, 0.68903188685916261225, -0.15654453518041994142, -0.22404090823151837086, -0.19752812434127994812, -0.12419171132316984829, -0.035009935841157746594, 0.019524646666614679791, 0.03529774646945626343, 0.030320447508067675108, 0.017369887180092392909, 0.0027717230807306775936, -0.010895895479892428806, 0.56491691780335850925, -0.44641881420797602242, -0.47174980049812470506, 0.055463221451909100468, 0.18107273390203687535, 0.15030156006077630138, 0.081409609146800010482, 0.00043198583897246720098, -0.044310261928421246225, -0.049040766339898032855, -0.032062668708964660313, -0.010134626461460394947, 0.0081356198306432418732, 0.020468559062055465686, 0.47726604248785714502, -0.4228702617422572807, -0.36935065920146215035, 0.10237172633289480517, 0.14630796403839585462, 0.09820284790524166918, 0.054004747021730886702, -0.00089569350210110670885, -0.033069197343112546883, -0.036437452607553919015, -0.023462167415718847285, -0.0069192299171015230266, 0.0064302586595456241725, 0.014888788979750437053, -3.9167909494514625557, 2.5605262249807432262, 1.9005858206874279848, 0.35902159542967815309, -0.22790387777945933512, -0.37390142702446504952, -0.334628994023282611, -0.20779811238629650183, -0.0780574459453689945, 0.0059560042611286135097, 0.041775942875636261353, 0.045111625338832207766, 0.032302413610509020381, 0.013471393178415339703, -3.5316863944873113645, 2.3877769191417739236, 1.6931849728464403171, 0.27470519799108272796, -0.23104895795181257245, -0.34155084778252520206, -0.29728001890199806656, -0.18019540381059301337, -0.062385029506943079358, 0.012468632276630996214, 0.0422785155764179299, 0.041956331510499912663, 0.027101068487085648584, 0.0073268969366588870135, -0.03321211546356464539, -0.22018411497532719423, 0.22984406318875180708, 0.34586628140109526575, -0.10721517565410812467, -0.24753902709075967237, -0.10281788600660581035, 0.0284553125994780505, 0.067010154549901806975, 0.048011219734608603926, 0.017552900497274415842, -0.0030148877075810405837, -0.011468265195121521427, -0.01189257547155989081, 0.30538579470636856295, -0.23670429648866875283, -0.17590378533116537985, 0.13030534701748916482, 0.031717985333892688427, -0.055672084690605978807, -0.032694759015012647696, 0.0024385437796069402497, 0.021601930758636093272, 0.024249904076223609967, 0.015915222226931778626, 0.0038933868620515330039, -0.0062182025302012040521, -0.012760957426454262384, 1.0974348818978285625, -2.6454185987384186696, 0.085274362060912267669, 0.74576388854783115523, 0.67055100719006532906, 0.34322090350939099546, 0.041987320188750683336, -0.095280895279694466371, -0.14139131254627029799, -0.13586541288111494974, -0.10179885532907030632, -0.049784819273979472043, 0.0096408617251309818713, 0.067989119406049458583, 0.058615639267854541261, -0.31497372308270432661, 0.51195116235175996966, -0.25454555923271543483, -0.087289903531438728379, 0.078169310641241537829, 0.013371394111431653823, -0.0072688432971683191244, 0.0028805778567757189985, 0.0048688081731454863066, -0.0011609479760082488942, -0.004224887272167439714, -0.0028691844378134455373, 0.0011400438197564941121, 0.044689497007932461881, -0.26266341649968316663, 0.43699303925955296224, -0.15342599755435268327, -0.12631551871495824702, 0.010592602610907594918, 0.037715486120411957638, 0.018312513710360005942, 0.00052922992908033576146, -0.0005380445615945999259, -0.0010922326940687918616, -0.0022826145478014869816, -0.001885891912917505046, -0.0001679969690518652635, 0.31490399568197452451, -0.66596119810528919825, 0.48470683416815996125, -0.016734210900058457683, -0.058304815730351775005, -0.050776274300120063931, -0.053576004197665227924, -0.006695788861203913131, 0.025152582836030629859, 0.032558434323926535581, 0.021459983881623883456, 0.0055297657531696783884, -0.0072870617975019785773, -0.01491350724944899217, 0.34282141024077722458, -0.88960766695097714951, 0.58736763230468480579, 0.060675352956062710008, -0.092021255038960178552, -0.031052806947829546258, -0.0037295644760576979193, 0.022918088757110644665, 0.0181707815406208055, 0.00063544437242549016952, -0.011192344558957907641, -0.01159892878077053123, -0.0044731614875186155589, 0.0062669655733782626134, 0.14326237033461905468, -0.70360214488171124447, 1.0460525125748743047, -0.1643809301043493154, -0.44374662486854243859, -0.052722296031888139789, 0.096048105015748050373, 0.081383445082361352552, 0.030121177990426722254, -0.0030103077060095141324, -0.01236319731937318353, -0.0088738339592189484362, -0.003712355474949638294, 3.1990355857098207344e-05, 0.13998114324558391663, -0.47106321319683930327, 0.59938224546541052362, -0.078482797277753479914, -0.21277189536836388362, -0.06892973507736416261, 0.037253409423216921059, 0.048407075183531600437, 0.017422343094913743389, 0.0028417979189708270256, -0.0010947708138654285107, -0.0019555518788127871158, -0.0020732806936011815256, -0.0021070954485640958249, 0.74787509079405956047, -1.1455023136606685696, -0.08286721827138776586, 0.12010580337566789666, 0.48867849773070137775, 0.22479212367708542586, -0.12168236956983816133, -0.18393021138896981093, -0.11043980083656043667, -0.024727943389261194951, 0.026018001505947595253, 0.039714178519775697462, 0.02681615829621253333, -0.00065606729142220383821, 2.429139600597879145, -4.2966299562416496727, 0.18790941018236725912, 1.4806072589387824223, 0.71175641364036135172, 0.048770182591314374665, -0.23772089541007362934, -0.25081496524223934053, -0.15031837816890852855, -0.041751138119318444153, 0.022189160574901112266, 0.0427250768047413923, 0.034403886156806638696, 0.010153860883480397143, 0.67930941259601651527, -1.1405536828526123827, 1.0143557121931687792, 0.20349314641342408638, -0.46517991967311966839, -0.40251565536805167245, -0.13841318916762579416, 0.050712070831581954056, 0.1113566906113705457, 0.10876491138862988406, 0.078556449486099827473, 0.036010718669989359719, -0.01028514249419276369, -0.053533997173282242554, 0.57451571411639701381, -1.0121821652104470157, 0.94649048007124703119, 0.15253118395651119044, -0.42111638637787507378, -0.35203871852753110261, -0.1137772298069311544, 0.054668166797030258441, 0.10217081195400547211, 0.09324664424015011166, 0.063500942524690795477, 0.026846493444896361857, -0.010551917905185692961, -0.044155658916314059759, -0.43333916237464958909, 0.25648086084612536917, 0.9244191134461060555, -0.43131737773722961249, -0.53717929962599542115, -0.10106463344271969518, 0.16740171425971175911, 0.17955068461588125084, 0.064355193738278304583, -0.02295514643604725305, -0.046733128802136175706, -0.032070393664698895742, -0.0071176507339641885766, 0.016219256856966932784, -0.42396813727813437156, 0.29966087891287312983, 0.82312222485556396645, -0.44556008801444069167, -0.46623252648342772275, -0.067660238916215767468, 0.15078429975949819708, 0.15350356021673672302, 0.052944722760207341139, -0.020223997331583865034, -0.039477821754009372046, -0.027167273856433950263, -0.0063849211704682263396, 0.013425422830177361974, 1.9739047486071126247, -3.5729126181214527769, 0.40948464752024704705, 1.1949910192595145642, 0.41365257490660717332, -0.042631082731838396138, -0.19227185046840958305, -0.16959947197908242789, -0.083389208167206407829, -0.0063514849119113694562, 0.028291335094584608295, 0.030718270652528002762, 0.01693717464747960047, -0.0029870994639207225883, -12.87441101104423069, 27.158226735092611648, -10.639700122209546862, -3.968660115287104162, -1.2695764282955210867, 0.09130623214961633427, 0.59269902037072774714, 0.59953685738019835139, 0.40893281154494259821, 0.18747845360908621526, 0.014109538713889868689, -0.07579239588561462182, -0.090120486753945544534, -0.055378421995385763987, 1.7890318120182902017, -3.5240390182213707782, 0.9206025129476994584, 1.0402820012393159121, 0.047227185975074192914, -0.069871296231822047074, -0.084171516936147106058, -0.093836587136946300602, -0.056721556621713610513, -0.012162754846370357059, 0.010445387212160667928, 0.015603222139825992723, 0.011289516297315217025, 0.0027717222981705636614, -0.44641881387282716931, 0.5737030671587514119, 0.35021915020975347144, -0.51641259522017690653, -0.11962740674176947087, 0.047030505271651215637, 0.051855914598768149126, 0.062794085506234859073, 0.032806147674332236663, -0.0028057929610180026847, -0.020323129459542895597, -0.019673128089874772351, -0.0083640105444605476748, 0.0081356191034295255449, -0.42287026176249570275, 0.59286980026651381337, 0.25497276122600803783, -0.51712661676649807063, -0.051827194100506214502, 0.07274665828167439785, 0.034469533858435270901, 0.040026008506641022611, 0.022434636499465424736, -0.0020919448134876991729, -0.015079745766612646929, -0.014769737016341145247, -0.0062442930994786663423, 0.0064302587625042328251, 2.5605262249040925404, -4.9770017792266081358, 0.46039238022053513566, 1.7368031286782315004, 0.62036677903402093204, 0.060324803976603072475, -0.12054949342170442894, -0.17900424747617010857, -0.1601548262860971461, -0.095084412210981489322, -0.031750879543997953536, 0.010317718261436598093, 0.029570540672343156169, 0.032302414606808937947, 2.3877769190138962152, -4.6318836489354318076, 0.48946088157414041619, 1.6021347801755529972, 0.53683688787628924821, 0.030540699469818702316, -0.11696496815553257542, -0.15989416883064502817, -0.14063192732400556229, -0.081765695927927234399, -0.025162507242369430704, 0.011593305356478265239, 0.027036671878728602003, 0.027101067076604937189, -0.22018411462731443673, 0.61155686371523698686, -0.26322587274267467139, -0.52850148096622540095, 0.3478284789258919707, 0.24158322347631841498, -0.081575765616973841543, -0.11986464095953246189, -0.043844605614098502988, 0.01622908249599403982, 0.031289997357701992975, 0.020208668633482779198, 0.0030477998138343691606, -0.011468265545515814868, -0.23670429566510006736, 0.41529189914660430416, 0.17627151269501406539, -0.46575194393590674258, 0.014335430926898838633, 0.11265092083011261914, 0.00051780994560500569028, -0.019860937275543187608, -0.0084828802896038790066, 0.0028584548190793389322, 0.0091009401834747146592, 0.0092162519002669859164, 0.0032084140112701290716, -0.0062182027284006218676, 1.4619154633897408768, 0.085274360950720370922, -2.9532984996459608062, -0.12778468535049253285, 0.63930886680095300179, 0.65185369978875917774, 0.41500118653356116161, 0.1410130040016208397, -0.0089946359420984816485, -0.08701046594354097663, -0.11189984539042034661, -0.096774649690276257075, -0.049784819207485273007, 0.015546385219054138782, -0.21066235658881898862, 0.51195116221420688962, -0.59100268678900869102, 0.5033791217017161701, -0.18058149996741490462, -0.11184205088378171289, 0.057891717382448415474, 0.013605510679750127193, -0.0015251098720267064566, 0.0067741243664830377846, 0.0075080476226293017469, 0.00024502528264394758672, -0.0042248856318667189422, -0.0042651477013203173069, -0.22447093920712418713, 0.43699303856460242068, -0.40379145841043473375, 0.39178744917490360145, -0.13362477722695789706, -0.1233233824499189446, 0.0030864574463023127307, 0.030459142450154345905, 0.020070060752458748765, 0.0040445686897263733589, 0.0018442048924582530277, 0.00019950125657488515757, -0.0022826125847618772106, -0.0034988156057235405075, -0.18261643771557323213, 0.48470683547137410896, -0.72442979741268631599, 0.56624407964811651617, 0.040536852419315316798, -0.078660707314293365489, -0.080773582755172476655, -0.081250743986030132815, -0.020581530757798748021, 0.024913301028543055615, 0.040730693800475205324, 0.029714440470452640336, 0.005529766052594697022, -0.02239376451756954628, -0.1324519232743034125, 0.58736763157107507727, -0.93686221398699509599, 0.53926314663929653115, 0.055332695281895931394, -0.12089959433752077433, -0.037767692235637168041, 0.0017220215758577571072, 0.032520569927956974321, 0.026952397993206676091, 0.0065987102092783973817, -0.0088098923054494467089, -0.011598928043577313302, -0.0052520473228839636287, -0.6233645500685198737, 1.0460525120809522903, -0.78114357038330439753, 0.82519720396795048778, -0.14894672695371813376, -0.44470132654793331772, -0.065952838274865879797, 0.089276395704704400447, 0.082473199341964939513, 0.035304755836432985538, 0.0030618062706766059448, -0.0089014071311234749945, -0.0088738323412662929029, -0.0060727167426423396893, -0.33988082735618985186, 0.59938224582438281995, -0.51385399988105040503, 0.52823124154766598615, -0.089687157074833046, -0.20640081858775902179, -0.068795126461564382048, 0.02741129654670685159, 0.04521833110335962691, 0.01803467453321932687, 0.0038198683578609233212, -0.00060804106376828603604, -0.0019555519904146225277, -0.0022091067287853343414, 0.90010463680967733868, -0.082867217850232835485, -1.6414802199351807843, 0.069372442797198202724, 0.3074997116364116545, 0.57894753979975799307, 0.2962994894799114709, -0.092778682503419432348, -0.19793307727040840294, -0.14671495409994264603, -0.060501517459584007452, 0.0047996334289917910731, 0.039714177217178167623, 0.050035928342969139548, 1.8525738342801671976, 0.18790940978725920174, -4.2258749232607861757, 0.1765266817450961423, 1.6094173974349323952, 0.88312116885380564923, 0.16976267469486747141, -0.19176935357024196827, -0.25886747908713608757, -0.18779120374898131129, -0.082191627078343348911, -0.0027508104689595814715, 0.042725076274058193526, 0.062707032128734976895, -0.78881019537879626391, 1.0143557119634147856, -0.75182449660697858729, 1.1218967146429958959, 0.28786576568056898573, -0.44669638408308887101, -0.4481291043829163101, -0.22041886201268476309, -0.017575696505439136752, 0.076963488887657208193, 0.10327041402256849356, 0.08472423664743031213, 0.036010718083476840612, -0.029876537153316805318, -0.73299131930774130694, 0.94649047984686662804, -0.68604239996763660692, 1.034705561482744729, 0.21421459037783374679, -0.41674425118068403329, -0.39463178688706168629, -0.18342485526432675269, -0.00090072846475383605558, 0.076179484011338921556, 0.091238090738156113479, 0.070116025520701438789, 0.026846493438947027083, -0.028369347611755857397, -0.94705452141240886732, 0.92441911423308387175, 0.19362333533727782431, 0.68744931142893905829, -0.56025148707334160303, -0.58914930647834207189, -0.09888513608664657728, 0.18264011034988034132, 0.19950478781318237154, 0.081205161279235132943, -0.014415025403679313648, -0.045893853310808412072, -0.032070394410312744182, 0.0014738905796854945372, -0.84969154921477252351, 0.8231222259262979124, 0.18781932739684739642, 0.62737212226925154201, -0.54647247537261711692, -0.51273902647180280567, -0.068234484937009134975, 0.16200878774750177014, 0.17043821397940811524, 0.068211649562050419293, -0.011979503175203505203, -0.038170255990404113045, -0.02716727489527838435, -0.00012559472935478981082, 1.1927450154217955625, 0.40948464649557575479, -3.3922446976919178319, 0.39666069600343467627, 1.2997215666848613846, 0.52703487907784585254, 0.014502073830033648458, -0.18377366310108958491, -0.18355210149361503302, -0.10501542973979401119, -0.023779297966125819558, 0.019990998658895000467, 0.030718271663801227644, 0.02168987627490245873, -3.7380343781196243569, -10.639700121966868096, 29.576420709039275891, -10.419697214312316547, -4.409272392309643962, -1.7747751992626941142, -0.26226829849019284158, 0.43894729072660848557, 0.5923644491657339417, 0.46583326226041055218, 0.25155697799212345389, 0.053833295262783256718, -0.075792396240352791215, -0.1366460465022615034, 0.68903188604287957109, 0.92060251377491597591, -3.3103775395963070949, 0.69525358546117910219, 1.1275766574626981953, 0.15297658837833671974, -0.032190730322247777806, -0.075805062591185495835, -0.097339638834528716127, -0.068069603512902851494, -0.023547509334267247016, 0.0039288786968222607146, 0.01560322100535657848, 0.017369888265570392061, -0.47174980104575120476, 0.35021914971142881212, 0.28452463699695096366, 0.39104203052143610142, -0.54349526357313060476, -0.19605212308165759216, 0.027747543421084902571, 0.057693864002268460678, 0.077530601091401876723, 0.048349195076878404009, 0.0079904130288895383605, -0.015983847040139803619, -0.019673129122939198021, -0.010134624680715161385, -0.36935065911675374339, 0.2549727609331280398, 0.26232803775198793028, 0.33071500059839908925, -0.52268824316262763041, -0.12084956638844246424, 0.057172737104241311279, 0.040820339185169668417, 0.05206019963122010541, 0.033970821362687933509, 0.0054487481406245152904, -0.012301896873193239601, -0.014769736817683919161, -0.0069192299283448912423, 1.9005858212350168479, 0.46039238018802319807, -4.7698271841508459445, 0.19482731613495804512, 1.7361160231915602203, 0.77246067319947187002, 0.18732633506537546553, -0.054470866327093601933, -0.15556347503078471473, -0.16335640533389783613, -0.1082134368374237654, -0.042379134211315012959, 0.01031771938394251617, 0.045111622866800725473, 1.6931849733314841, 0.48946088096389667177, -4.4217631958366201417, 0.24771544800466266389, 1.6050340274279117914, 0.67916358052548864155, 0.14688168955742975097, -0.059499117777454466061, -0.14206028490090374938, -0.14650444036385540381, -0.095615860805148319979, -0.035693443085711866469, 0.011593304090811877571, 0.041956332497716034013, 0.22984406298950607606, -0.26322587285715093319, 0.016469420796110939564, 0.043845152725474366429, -0.35652212704582614267, 0.32837293692421498648, 0.224317484046029747, -0.092140158108256736691, -0.1305493218047852344, -0.053105571129074068126, 0.011588365594651504897, 0.030930624832387950546, 0.020208668942223215192, -0.0030148880487694878702, -0.1759037867436999969, 0.17627151452280251376, 0.020680225093069066339, 0.27804482443151301574, -0.38064136695813971523, 0.0006279957011097554373, 0.11086299953131302509, -0.001338536070056614135, -0.027825252784153233637, -0.016993073163899891087, -0.002152755220010479139, 0.0072535145085881021568, 0.0092162511913787999174, 0.0038933866671211386952, 0.90360785726267744966, 0.74576388847987684638, -0.12778468543418922132, -2.9985841408928139984, -0.16377883390175715372, 0.65376057885341176679, 0.6970844715826951532, 0.46895210254569247299, 0.1810443750857287093, 0.012416652962833998539, -0.079490041810175063319, -0.11189984493144018718, -0.10179885437931962544, -0.062092026760509469452, 0.033836230172636677649, -0.25454556003401657138, 0.50337912090902270368, -0.55319777819878057645, 0.48831938941319502634, -0.18908786971765340068, -0.11348606297440325119, 0.061921911721111846938, 0.016949017154994094198, 3.846082171966554511e-05, 0.0072297120617816268071, 0.0075080467211421685406, -0.0011609489630335090293, -0.0093029930034605416017, -0.027153704367399979225, -0.15342599783904944943, 0.39178744978764112883, -0.41503688395529053645, 0.4018866992229619961, -0.12730694853089544183, -0.12776802909842130407, 0.0022006960509455783005, 0.031364317136650526896, 0.021384295750889934645, 0.0049319522748713052682, 0.0018442036884849302963, -0.0010922334624852998025, -0.0058391912704051735991, -0.19025820895580219938, -0.016734210960194610684, 0.56624408029050798952, -0.71916432113883788091, 0.52707967881350381578, 0.014796043486015719032, -0.09379459483835769773, -0.081311228979131816574, -0.075159180355748589353, -0.013065526677934671124, 0.030085132825320059363, 0.04073069320986970332, 0.021459984137731244469, -0.014173870092495126013, -0.08767211160028341721, 0.060675353405984594479, 0.53926314567679278777, -0.98673175088916875719, 0.50025322138657402071, 0.061190834554957315716, -0.1067435476959277818, -0.026867007700356423938, 0.0064654110673078571056, 0.034147143936449290968, 0.027546902577320557526, 0.0065987120591336758876, -0.011192345688990763125, -0.019269384925199140035, -0.21895717868403283268, -0.16438092953299809906, 0.82519720313429412162, -0.84096552903948151858, 0.83038902238542677825, -0.11618964799286903011, -0.43417772065761001121, -0.070899755115881271661, 0.084245278536363057964, 0.082930801781036386777, 0.037562500228490919296, 0.0030618065297381672103, -0.012363197786493966557, -0.013162252775402664973, -0.15581763965190500998, -0.078482797646846472195, 0.52823124060138249014, -0.57314861601307698358, 0.5345733633558767961, -0.067907902455972446454, -0.20444401504565343219, -0.072510724846344465355, 0.023555957205030907814, 0.043783637877656003679, 0.018052182830822667559, 0.0038198685745292658769, -0.0010947709564802980773, -0.0024631260957715646678, 0.4260066620438658358, 0.12010580338504536768, 0.069372444564529511801, -1.4371350923676455835, 0.12484835795055242158, 0.30382634001334662965, 0.55785175683201959629, 0.28053056054726233093, -0.10737639267671388044, -0.21299969908283822306, -0.15831179186706101358, -0.060501517884062508412, 0.026018002405168230706, 0.090765113004440498012, 0.24996892542527307524, 1.4806072592719172754, 0.1765266826021442137, -4.0977734196053257065, 0.201228967357525379, 1.6059776055667205341, 0.88221614664018810359, 0.17134022296585163514, -0.19957850457982562187, -0.27414880170848221486, -0.201384891346438194, -0.082191626580073548491, 0.022189160763168815776, 0.10065465197078905801, -0.94817326232059739866, 0.2034931463254238404, 1.1218967144415303849, -0.80336261899758087957, 1.0706071138568074019, 0.26923285929672319261, -0.47276916791990136746, -0.47550193088351605519, -0.23809173616647782978, -0.022205736456170799276, 0.078654741867842381531, 0.10327041364256370204, 0.07855644934133509949, 0.024420856048534181348, -0.82628272536433300388, 0.15253118395846354538, 1.0347055615054361333, -0.74678205405369124659, 0.98183361223751042246, 0.19960379840616263025, -0.4358669982981526192, -0.41530672124504958154, -0.19688229738245693889, -0.0038198441840233132716, 0.078275970844744163779, 0.091238091048798652949, 0.063500943295403899924, 0.013414787571770669938, -0.18688725448047419087, -0.43131737811603665378, 0.68744931162339628639, 0.0032809815406317048569, 0.68544044565976436356, -0.48323737739105032674, -0.54817854827266510043, -0.09579445072598080646, 0.17298832967761676449, 0.19364410113333119234, 0.08022964785025479828, -0.014415024810780663334, -0.046733128236209862949, -0.030695751446712840615, -0.13709867210983703312, -0.44556008860066703692, 0.62737212184604007614, 0.032668614738723511515, 0.62896789201320058016, -0.4813225952560821197, -0.48079807280601966069, -0.066617253838012091549, 0.15447130840051884193, 0.16643970778585664982, 0.067913717406098389473, -0.011979502498020329349, -0.039477821561041952225, -0.027492858108320462202, -0.025787926735688675856, 1.1949910195679391833, 0.3966606954831267684, -3.3192576217429627938, 0.38530985068782763747, 1.275675998271732281, 0.51722789207505126541, 0.017481080659349373246, -0.18150481718969999068, -0.18537026278068990948, -0.10850689936537902258, -0.023779296970900126035, 0.028291334136507348407, 0.049939194264137010126, -0.17838617258906422358, -3.9686601154215619403, -10.419697213246346124, 29.895623015116321142, -10.385637623009746022, -4.4897069913817189146, -1.8472800533204103424, -0.29606179848374331032, 0.43229268162191597513, 0.60177616415504886493, 0.47939238417169999229, 0.25155697757796446901, 0.014109539154174233211, -0.18570550670857241538, -0.15654453446595509303, 1.0402820004173658486, 0.69525358578671081489, -3.2859373259819775193, 0.69443418921481459005, 1.125972531437831492, 0.15669992708486521482, -0.027737595592536017086, -0.075116497415327845855, -0.099707020598596346472, -0.071205435716455031669, -0.023547509900548532769, 0.010445387776929163787, 0.030320447118442263312, 0.055463222105965388764, -0.51641259430851094159, 0.39104202937437487764, 0.23277883330158696817, 0.35981159867413642361, -0.53076959874016527152, -0.18447142709589203768, 0.036288620535901261466, 0.062525066210142393741, 0.080348874803586350923, 0.049999966720331079362, 0.0079904146991379293441, -0.020323128886622441464, -0.032062670730285687537, 0.10237172611623376883, -0.51712661634737200345, 0.33071500049112545483, 0.23912337261809679667, 0.30447631939769637111, -0.51543873151036900815, -0.11400317235316903552, 0.06382401359710848654, 0.04477098597840477312, 0.053941171187000563902, 0.034828683528444925876, 0.005448747981702847798, -0.015079746136814967117, -0.023462166559157813378, 0.35902159476153489015, 1.7368031290957579582, 0.19482731663566640878, -4.7662593788426814712, 0.20854970986705220759, 1.7598652374674577104, 0.79545238683575603122, 0.2047378934518154503, -0.045666220173596297482, -0.1532358882275116907, -0.16504064784024213841, -0.10821343830476651271, -0.031750879801903851563, 0.041775944503350995185, 0.27470519794060527241, 1.6021347806805259495, 0.24771544793331884415, -4.4231644155835621035, 0.25942402667334801825, 1.6285274505025559844, 0.70010314831354669707, 0.16103439610307962448, -0.053678239031873548059, -0.14156172104296321779, -0.14881584849068413279, -0.095615860769896685034, -0.025162507600769566662, 0.042278516075682788522, 0.34586628151889614591, -0.52850148060222412738, 0.043845151794687030333, 0.20485649633616229148, 0.04558468729873355052, -0.40838916179942236218, 0.28842818460177277462, 0.21830394655992060859, -0.085364746882200748246, -0.12587662772549659818, -0.052402929203330374242, 0.011588365834825715278, 0.031289996947520909432, 0.017552901093316433667, 0.13030534695867726502, -0.46575194470449715611, 0.27804482520936252277, 0.092107782953478184451, 0.28983034343140007261, -0.38095646510255998241, -0.012389056439193090603, 0.10051453711139808622, -0.0066907896292442304437, -0.029625340375254660685, -0.01756522061356910816, -0.0021527568738439334384, 0.009100940387514092067, 0.015915223148605963516, 0.30004248305248693018, 0.67055100762661745684, 0.63930886625035332482, -0.16377883355286526612, -2.9781545625754834994, -0.13352321506778991322, 0.68191360092730801679, 0.71553605554695209889, 0.47874989151021818401, 0.18515923118361962696, 0.01241665350104296725, -0.087010465441463000213, -0.13586541280992581182, -0.15399983854051432774, 0.065889631093686854002, -0.087289902991048842762, -0.18058149884730090307, 0.48831938884241837817, -0.56608277103314585954, 0.49110625343739416904, -0.18247456639286310787, -0.11062976459411950902, 0.062876657848357095371, 0.017144437888896121858, 3.8460541543306869111e-05, 0.0067741257321315574008, 0.0048688086884039116428, -0.010004512611322459795, 0.067754813504376731892, -0.12631551915578739798, -0.13362477600346561735, 0.40188669804058124724, -0.41392152042147589697, 0.40027727816948044381, -0.12429088844573867034, -0.12595794838988283448, 0.0027048124503808437455, 0.031618260697553565719, 0.021384295096458727503, 0.0040445691986000708792, -0.00053804402630191182095, -0.0056507213249577959574, -0.11493372016400178748, -0.058304816049317684246, 0.040536853206959960749, 0.52707967978998471192, -0.75660958835930802735, 0.50768263082040210321, 0.015986806919559871931, -0.083490372742601764999, -0.070386197849078360744, -0.069002616290400761723, -0.013065528646460964485, 0.024913300263609071405, 0.032558434563486823854, 0.012917631106420126783, 0.030013273243944722096, -0.0920212537523194124, 0.055332696373390666267, 0.50025322014760986722, -0.99178114335497191689, 0.51153021198044545947, 0.074113579530998688338, -0.10025173001465777223, -0.025319326219434459824, 0.006086345810263064926, 0.034147144554445209297, 0.026952398053335414885, 0.00063544320362872073159, -0.029474564732257332866, 0.17579693712506239134, -0.44374662471813447473, -0.1489467268270252831, 0.83038902220186128211, -0.82519098573804461338, 0.84566511938735866583, -0.10875115202892025879, -0.43425902953925327177, -0.071950716335010697633, 0.084226669686685268212, 0.082930802019602803399, 0.035304755741633504773, -0.0030103077231120383903, -0.021620845904012843397, 0.049879394341616534181, -0.21277189558645739909, -0.089687156432677647699, 0.53457336339172123463, -0.56201117890490981388, 0.53706536374999436223, -0.065441104466064242695, -0.20467798411232016043, -0.073992916473863454274, 0.02265260971687291236, 0.043783637753374710089, 0.018034674647339734438, 0.002841798547967375882, -0.0029270318946610731671, -0.22280643208338607453, 0.48867849600048296477, 0.30749971079224674986, 0.124848358764652001, -1.4411180003279076001, 0.087423146161243345253, 0.26394031025575831473, 0.53023439854428133788, 0.26670474200439309165, -0.11326333147371754617, -0.21299969994615458591, -0.14671495523562960717, -0.024727942603249689069, 0.10053678360759309152, -0.50038418079155178031, 0.71175641366862063553, 1.609417396179721349, 0.2012289666869714333, -4.1303144810179617252, 0.14460880948389082112, 1.5653296639604328266, 0.8597880165910389838, 0.1602772939710580824, -0.20530271457300683369, -0.27414880111648437522, -0.18779120343598632026, -0.041751138543588209306, 0.10268312253629588437, -0.43112668027174994023, -0.46517991926914792344, 0.28786576541209552005, 1.0706071142991955281, -0.86038565307062975229, 1.0281328186844984618, 0.25324617735340215763, -0.47110821324214108285, -0.46752347978692554431, -0.23280256275651750597, -0.022205735810573029587, 0.076963488362665796738, 0.10876491087694914461, 0.10344364268514326388, -0.34969747033284104276, -0.42111638640073600914, 0.21421458997080033981, 0.98183361268427371549, -0.7949289517559163798, 0.94959066084103027627, 0.18993666126591207677, -0.43218510129151144206, -0.40764443828787350066, -0.19215808529933708915, -0.003819844033620983792, 0.076179482986957203217, 0.093246643606250290137, 0.078164344503622334992, 0.36536398267692593578, -0.53717930017739567994, -0.56025148766332011352, 0.68544044544031335064, 0.073808890685269995524, 0.74372780346395572781, -0.45949489088634531964, -0.55109570132597118786, -0.10654776616745370577, 0.16634903510956208783, 0.19364410103936913199, 0.081205161794551922183, -0.022955146023835575064, -0.075438381750656119396, 0.33030775159573305677, -0.46623252651254620815, -0.54647247506672769202, 0.62896789241881767563, 0.093197923268715274214, 0.6775328518079498652, -0.46142954946258590843, -0.4828213641699040104, -0.075079672481994264022, 0.1492479382301322155, 0.16643970710292418924, 0.06821164931926633479, -0.020223997260961957406, -0.064557337841270392986, -0.42873955821138182154, 0.41365257594570808974, 1.299721567378795184, 0.38530984835161236068, -3.3739417841099652584, 0.33131463148498124793, 1.2512963184368457448, 0.51338345742292157414, 0.021015635661966066555, -0.17921503601947549877, -0.18537026212908580547, -0.10501542906777652786, -0.0063514859168520183819, 0.070792543047913797372, 1.179600209825097723, -1.2695764285227839618, -4.4092723926797443568, -10.38563762253770939, 30.066663602111940889, -10.245927535023740873, -4.4106288740501620182, -1.8218135198974545474, -0.29694167030740642454, 0.42879850193307150752, 0.60177616438105541175, 0.46583326304664163242, 0.18747845458896553184, -0.14598390782592193804, -0.22404090871429607756, 0.047227186433910377239, 1.1275766569894258851, 0.69443418998221373428, -3.3243718512197419912, 0.66260084318324063979, 1.1148782529688308074, 0.15468362443012440011, -0.02732224214736363177, -0.075127374720264136765, -0.099707021789440172799, -0.068069602612785023998, -0.012162755038679381667, 0.035297746665373747055, 0.18107273268718218628, -0.11962740590506101834, -0.54349526409881299127, 0.35981159972244181056, 0.2509885100941156022, 0.38646631494848160404, -0.51321477430110040707, -0.17811758471253830161, 0.036683953189624723956, 0.061538422049962043792, 0.080348875355737037207, 0.04834919371966246715, -0.0028057943444712574177, -0.049040764056818605954, 0.14630796398983680895, -0.051827193793406230171, -0.52268824353117304682, 0.30447631929560481368, 0.25340553065269616484, 0.32524020668875142093, -0.50237002061761382166, -0.1098134654932626797, 0.063659195032007787063, 0.043732158559258632602, 0.053941172176089624557, 0.033970822118056638594, -0.0020919449918319690807, -0.036437453833959770888, -0.22790387712620877037, 0.62036677807363393988, 1.7361160233517571871, 0.20854971023007246034, -4.7735790017792858464, 0.17745899803549586293, 1.7387231444666695346, 0.78806961167941114166, 0.20548427994166618138, -0.043876978764491231244, -0.15323588809512550535, -0.16335640595384184892, -0.095084411761410470021, 0.005956004121886438428, -0.23104895774415304399, 0.53683688766363013567, 1.6050340269131655457, 0.25942402743570813772, -4.4304964550293561842, 0.22981965509637625189, 1.6080030015506980856, 0.69213860918606173733, 0.16073681508053122569, -0.052600052811793113205, -0.14156172047050880058, -0.14650444111375898748, -0.081765695977683267026, 0.012468632439470613005, -0.10721517629089730161, 0.34782847943854261485, -0.3565221265027895825, 0.045584686541674107607, 0.16699622564512706102, 0.011743857345837813155, -0.42343410132409065039, 0.28845464768658168886, 0.2243486463584180024, -0.081427955399154233751, -0.12587662691551512739, -0.053105571478901686389, 0.016229081835555003027, 0.048011220321458741855, 0.031717987246712875393, 0.014335429484778545658, -0.38064136716834540142, 0.28983034330258766742, 0.096554224042042419152, 0.28507973905399447423, -0.39055172821191375121, -0.01962539903036747066, 0.098499217389630008834, -0.0065463354944938299124, -0.029625341574754972263, -0.01699307241913856209, 0.0028584564454915140086, 0.024249902505804804553, -0.077273425462121680951, 0.34322090495217372208, 0.65185370067901815538, 0.65376057890283811869, -0.13352321523384802493, -2.952732231120455797, -0.12399146319570626684, 0.68172301838211968317, 0.71394708956938712774, 0.47874989123323230533, 0.18104437454651162165, -0.0089946373112911731651, -0.14139131398340135104, -0.23549572821030759218, 0.0089065287691197977255, 0.078169310454822935053, -0.11184205074118669376, -0.18908786966063415025, 0.49110625405688101974, -0.55896749119035926689, 0.49534672350066710766, -0.18101994764076462241, -0.11053638755391084758, 0.062876658446261904367, 0.016949017287156514094, -0.0015251102228963000357, 0.0028805775340646224379, -0.0022456665323163991881, 0.044773229602748799982, 0.010592602855678211887, -0.12332338285671859412, -0.12730694804147602017, 0.40027727874595731006, -0.41100356292899814514, 0.40354098454567977061, -0.1226431155302842857, -0.12553088997817735128, 0.0027048133009383227947, 0.03136431703338081467, 0.020070061206548497207, 0.00052922937180279328983, -0.0033679297088823438588, -0.015347096421582257994, -0.050776275860836085629, -0.07866070885564563353, 0.014796043838405957549, 0.50768263178813199943, -0.75849213624137334921, 0.51648514598172456935, 0.027241904728916126188, -0.076459892625567349445, -0.070386198873610417159, -0.075159181217424736454, -0.020581530242572764616, 0.025152583791991450779, 0.052059577780868906971, 0.063277368091359267877, -0.031052808146097538033, -0.12089959420671204937, 0.06119083495964748598, 0.51153021205311488551, -0.98173801734617149872, 0.51701959578941680817, 0.076055846773293134255, -0.10029203495855507078, -0.025319325686563132899, 0.0064654106545590289343, 0.03252057001703817507, 0.018170782497787602711, -0.024286396128976834446, 0.17804039387225670255, -0.05272229664625856238, -0.44470132562898578854, -0.11618964803210718451, 0.84566511915136000077, -0.81961886996268562555, 0.8489534476078188785, -0.10586641622704158483, -0.43311608445535687562, -0.071950716241604734114, 0.084245278319847599624, 0.082473198370184241934, 0.030121178545306284691, -0.023963212497435413401, 0.088321118671053674976, -0.068929734657380739571, -0.20640081907638280212, -0.067907902190143754062, 0.537065363566274101, -0.56134135348231450902, 0.53799964373129638417, -0.0648361049577147186, -0.2047405070849905584, -0.073992916191470095577, 0.023555956793278039901, 0.045218331281721736714, 0.017422342659473386006, -0.002184616205849847486, -0.43618952040666125747, 0.2247921242762377092, 0.5789475390621491302, 0.30382633965196825798, 0.087423146392280326467, -1.4836105646520487689, 0.051840917205316090766, 0.24218195121804592107, 0.52152573084061115694, 0.26670474157638573987, -0.10737639258892894834, -0.19793307640102592027, -0.11043980152249131821, 0.050354484871451082872, -0.58009699858458185773, 0.048770182332106108702, 0.88312116897905035273, 1.605977605705183997, 0.14460880887286947649, -4.1883576553027150879, 0.09970268634703613464, 1.5399155807759661307, 0.85048610687687919985, 0.16027729412663949127, -0.19957850484504979782, -0.25886747934586412656, -0.15031837821789678644, 0.038117476859553585389, 0.013103645475922326424, -0.40251565481967188642, -0.44669638357891550129, 0.26923285901905352757, 1.0281328182310716102, -0.8872425253304684345, 1.0263213117196505753, 0.26548557782755727308, -0.46118860307150461031, -0.46752347980037645137, -0.23809173523409166151, -0.017575695551758585555, 0.11135669136259185308, 0.18667204282361563039, 0.039512157964852252789, -0.35203871850464574234, -0.41674425152347110979, 0.19960379795073507347, 0.94959066054270147017, -0.81390539840352582601, 0.95065532208764325084, 0.20162615462236160857, -0.42337583520552923844, -0.40764443833401015072, -0.19688229700946247025, -0.00090072814613487151014, 0.1021708120570344186, 0.15169777251612012736, 0.3766052604542548532, -0.10106463279446656345, -0.58914930646391094893, -0.48323737738350064364, 0.74372780366841273469, 0.099812213705285243726, 0.74913515987455125966, -0.46312292031673313319, -0.55600689530369351576, -0.10654776622548578491, 0.17298832992173629464, 0.19950478780755820374, 0.064355192962318771843, -0.10102377679002774724, 0.32360036017048593449, -0.067660238401398065111, -0.51273902635956070117, -0.48132259543106126198, 0.6775328521099942547, 0.11568730925362993855, 0.68326752879612839831, -0.46361867314311544508, -0.48662687049428648978, -0.075079672293535154592, 0.15447130909651463071, 0.17043821492574914855, 0.052944722786306012974, -0.085591528915784267917, -0.38313081537536181065, -0.042631083134152258918, 0.52703487935825243937, 1.2756759984967105392, 0.33131463121317311638, -3.4136717403220688816, 0.3102130925672384465, 1.2452537100355050015, 0.51395452890475623597, 0.02101563643034884063, -0.18150481700678500196, -0.18355210219242407144, -0.083389208291005906259, 0.058582405082873566238, 1.3155679700364895091, 0.091306231686484010113, -1.7747751996458853707, -4.4897069917489114133, -10.245927535103877659, 30.237519305236315148, -10.138625931531837665, -4.3700027909629994838, -1.816279840494100517, -0.29694167053965292435, 0.43229268136119464172, 0.59236444856780268164, 0.40893281134605113936, 0.048550592280575621817, -0.19752812364386093669, -0.06987129622734668255, 0.15297658806100997153, 1.1259725305829386688, 0.66260084363317750533, -3.3499754047800296419, 0.64779819017639284073, 1.1094708336522467462, 0.15375628024771570423, -0.027322242421517501515, -0.075116496469322405205, -0.097339638675979486826, -0.056721557019546696432, 0.019524646344517594676, 0.15030156033440508767, 0.04703050484895858474, -0.19605212282445971805, -0.53076959851237270893, 0.38646631442565015924, 0.27277181380682152811, 0.39817984771490566587, -0.50988366492255399187, -0.17868805177493735092, 0.036683953210980890791, 0.062525065814732549985, 0.077530601908810697909, 0.032806149322806371638, -0.044310264053513952198, 0.098202847640323498446, 0.072746658126397051314, -0.1208495661782166275, -0.51543873083554792025, 0.32524020619130067367, 0.27084340132931061795, 0.33409261650600646787, -0.50044605644536066524, -0.11071478543421220397, 0.063659194911177163778, 0.044770985854214906396, 0.052060199547321286961, 0.02243463730538674053, -0.033069197896497433409, -0.37390142756716238104, 0.060324803982039473804, 0.77246067357447489865, 1.7598652376306003209, 0.17745899822220087549, -4.8126120019018792107, 0.14858977870119136067, 1.7258615505318575245, 0.78554471748199827807, 0.20548427938559618577, -0.045666220092290571442, -0.15556347534926792675, -0.16015482593435140246, -0.078057445830169630563, -0.34155084876003416472, 0.030540699648033797775, 0.67916358142223876015, 1.6285274504550579788, 0.22981965521166017319, -4.4680736748859750662, 0.20198156274116285669, 1.59550751397154289, 0.68953405315574056722, 0.16073681464737268887, -0.053678239347906470114, -0.14206028481222571225, -0.14063192686041015023, -0.062385029619799609035, -0.24753902659890036908, 0.24158322331189624466, 0.32837293700212483172, -0.40838916191314583726, 0.011743856498591618079, 0.15676054042452852721, 0.011903948039110377397, -0.42061812240414836461, 0.29088118190797751517, 0.22434864702362888877, -0.085364746968025304041, -0.13054932212232733812, -0.043844605897084977519, 0.067010154885773956512, -0.055672085524970292558, 0.11265092185991973373, 0.00062799521281781383458, -0.38095646519192299939, 0.2850797397760848062, 0.090225147029470334492, 0.28042110243802409464, -0.39223482115309488139, -0.019752765241928982309, 0.098499217019191329103, -0.0066907894567071928352, -0.027825253154269667666, -0.008482881965662318613, 0.021601932753884649924, -0.23775913006283705409, 0.04198731935281851263, 0.41500118668204255545, 0.69708447105466830873, 0.68191360067846529258, -0.12399146324331111746, -2.9550941252460432018, -0.12837396392077904395, 0.68172301814658053942, 0.71553605603077941044, 0.46895210310485790917, 0.14101300414384970727, -0.095280894952485595262, -0.27517309163073006006, 0.0075038857998297459062, 0.01337139389497674774, 0.057891716694736271298, -0.11348606243963138551, -0.18247456737663961257, 0.49534672392073186842, -0.55716405462878659804, 0.49573453105288783904, -0.18101994725208583881, -0.11062976491714607286, 0.061921911637511317661, 0.013605510486695367514, -0.0072688427720209308758, 0.0082787536706907353945, 0.012183893884839673213, 0.037715486076296191043, 0.0030864571440405710531, -0.12776802899634756572, -0.12429088902701275476, 0.40354098486536044144, -0.40904358006080349108, 0.40438361858367677382, -0.12264311494882042253, -0.12595794938238913385, 0.0022006966554105836936, 0.030459141802228990176, 0.018312514212725620327, -0.00087907119924074156748, 0.065262085815486869933, -0.053576003406925240546, -0.080773582263292661998, -0.093794595229459487573, 0.015986807741673624006, 0.51648514659171340657, -0.74819065671580764487, 0.52309250758185621422, 0.027241904143486532497, -0.083490373899321762563, -0.081311227639707189807, -0.081250743739906122753, -0.006695789083859429551, 0.080329152665262873456, 0.046649997084591880714, -0.0037295661352372642267, -0.037767692649534750915, -0.10674354654659172492, 0.074113578401903551507, 0.51701959567999944412, -0.98051013016343624429, 0.51678472271783315595, 0.076055847192900635778, -0.10025172948996076416, -0.02686700863315186899, 0.0017220211127352540864, 0.022918090479131052878, 0.0046824453105677584994, 0.073047580854802549033, 0.096048105280696044472, -0.065952839677855673584, -0.43417772010751609635, -0.10875115212389106778, 0.8489534475812292591, -0.8162522547686533736, 0.85124648097912725131, -0.10586641608010773086, -0.43425902956912165731, -0.070899755431636415914, 0.089276397140331684166, 0.081383444810505634059, -4.1424222154600110401e-05, 0.044651978940465221479, 0.037253409562443995373, -0.068795126030123332495, -0.20444401550238447407, -0.065441104169260078116, 0.5379996437588210334, -0.56074399464394486525, 0.53841615113170382401, -0.064836104785132769868, -0.20467798470460091265, -0.072510724085095165647, 0.02741129604332649683, 0.048407074996860616212, 0.0076872196592879995375, -0.29031166458682411413, -0.12168236761145485858, 0.29629948989542514992, 0.55785175607434911438, 0.26394031102094522856, 0.051840917508447603712, -1.5072958570149674884, 0.040684016867908269488, 0.24218195058690472621, 0.53023439813213502081, 0.2805305615003396702, -0.092778682230373499262, -0.18393021307194357128, -0.085144583656375699698, -0.36676370650626755054, -0.23772089621472541898, 0.16976267539257289263, 0.88221614697841643782, 1.565329664121055897, 0.099702685860603615375, -4.2211579321051360125, 0.084224033237330603319, 1.5399155809475404411, 0.85978801690247974587, 0.17134022273821752913, -0.19176935387521679144, -0.25081496397312952329, -0.12306655196894500381, 0.21447628096098994566, -0.13841318960552770334, -0.44812910400416383627, -0.47276916798126639163, 0.2532461777337488007, 1.0263213113493254713, -0.87772604976253898457, 1.035738331517600308, 0.26548557820699086873, -0.47110821300196709105, -0.47550193190733724968, -0.22041886289905901486, 0.050712070525506067198, 0.24082313327449730855, 0.20204510015357199371, -0.11377722948743246301, -0.39463178600208759228, -0.43586699797896699682, 0.18993666137729817156, 0.95065532186837908846, -0.80448554580087616728, 0.9591441571408403366, 0.20162615455141455478, -0.43218510093889195289, -0.41530672158827663942, -0.18342485540624528517, 0.054668166524079367385, 0.2083844487201263751, 0.14652064599272632273, 0.1674017146518688437, -0.098885135782012648087, -0.54817854865042892065, -0.4594948910375651896, 0.7491351599832541952, 0.098448597243742788643, 0.74688648078193353719, -0.46312292050248315967, -0.55109570128705476222, -0.095794450452554386066, 0.18264010999125041002, 0.17955068407651325613, -0.043352090613016876219, 0.12322835968678325091, 0.15078429951196775649, -0.068234485222966850815, -0.48079807357171544524, -0.46142954970345190224, 0.68326752890706565768, 0.1156792073790795794, 0.68188705653414150021, -0.46361867301468251412, -0.48282136412167464545, -0.066617253702312695385, 0.16200878718567693348, 0.15350355992591013798, -0.037467450776133011392, -0.19139613237457003825, -0.19227185139046235052, 0.01450207396130641542, 0.51722789392951729059, 1.2512963171173627774, 0.31021309225704735146, -3.4258360141332864224, 0.30607855048396875874, 1.245253710789524737, 0.51338345782030081477, 0.017481079568231097654, -0.18377366376479606358, -0.16959947041565845427, -0.024370435371640623845, 0.9179557101805232433, 0.59269902063880097742, -0.26226829932878442664, -1.8472800538063491871, -4.410628873671775807, -10.138625931656232382, 30.309637691805850324, -10.111475001203892532, -4.3700027912956418419, -1.8218135197421765348, -0.29606179813274108303, 0.4389472920695299818, 0.59953685700119896396, 0.42453349112875604643, -0.12419171273302484493, -0.084171516558029901844, -0.032190729514603860784, 0.15669992773784649498, 1.1148782524368845426, 0.64779819058137089183, -3.3603691632644960841, 0.64354103217426206207, 1.1094708333560681091, 0.15468362480896849509, -0.027737596068318756681, -0.075805063496532404388, -0.093836587164511958692, -0.035009934765984539184, 0.081409610771181459277, 0.051855912844195758971, 0.027747543704879307214, -0.18447142760673368445, -0.5132147736555233708, 0.3981798472954481416, 0.27789008443835655493, 0.39904678637143431441, -0.50988366511146310245, -0.17811758417971793045, 0.036288619792249204865, 0.057693864044675677305, 0.062794086046299213422, 0.00043198559587037839436, 0.054004747875993464046, 0.034469532721797438435, 0.057172737276407245566, -0.11400317224725023402, -0.50237002041651201978, 0.33409261630619607741, 0.27443259594526175915, 0.33438767329728080746, -0.50044605612678239037, -0.10981346555636867235, 0.063824012924203116226, 0.040820338444750019213, 0.040026009217773966808, -0.00089569312470759443467, -0.33462899403846335655, -0.12054949237645817828, 0.18732633390090913661, 0.79545238662638095839, 1.7387231446126267809, 0.14858977901363315532, -4.8326661409338109721, 0.1402442248140430836, 1.7258615501031087103, 0.78806961182476154004, 0.20473789343957382592, -0.054470865049199949925, -0.17900424847300078213, -0.20779811247257035101, -0.29728001879806009722, -0.11696496763086491888, 0.14688168881040691849, 0.70010314788341465597, 1.6080030021536013685, 0.20198156282167989417, -4.4873081236194805044, 0.19403040894183942733, 1.595507513475676209, 0.69213860946550376241, 0.16103439582061079838, -0.05949911683871629553, -0.1598941697953998875, -0.18019540349427032311, -0.10281788597965706156, -0.081575765775865227702, 0.22431748409817953149, 0.28842818483740134683, -0.42343410109544971087, 0.011903947341659238887, 0.1602953893810492858, 0.013886312526696921801, -0.42061812188285424563, 0.28845464786077096297, 0.2183039463720981288, -0.092140157768107050495, -0.1198646403339325428, 0.028455311585840494665, -0.032694760507208896205, 0.00051781177856928474641, 0.11086299967587642801, -0.012389057393543234653, -0.39055172807143795355, 0.28042110299860034672, 0.089644394815605432658, 0.28086334099501264472, -0.3922348216096444018, -0.019625399684874901551, 0.10051453883506404618, -0.0013385356872540133191, -0.019860938470108859683, 0.0024385438490766462791, -0.27517309180699350835, -0.095280894349349790073, 0.1410130035270800164, 0.46895210317720747994, 0.71553605663346142585, 0.68172301781244537544, -0.12837396447048282644, -2.9550941250410591721, -0.12399146278783285124, 0.68191360055260463735, 0.69708447056204114745, 0.41500118660919027569, 0.041987320434026580152, -0.23775913070282628237, 0.0082787536819613500444, -0.0072688427984424226064, 0.013605510515201972577, 0.061921911563864839856, -0.1106297647859729727, -0.18101994729614723223, 0.49573453099240372133, -0.55716405459911688691, 0.49534672384957062485, -0.1824745671961660598, -0.11348606255277789157, 0.057891716669669024409, 0.013371393929035718662, 0.0075038857994730963005, -0.00087907126494699689128, 0.018312514391150742227, 0.030459141595813643555, 0.0022006968192749059784, -0.12595794944187713171, -0.12264311504334174385, 0.40438361872647615725, -0.40904358012678171397, 0.4035409848781179587, -0.12429088901729715144, -0.12776802913833584796, 0.0030864574153595227338, 0.037715485885300180369, 0.012183893930876056252, 0.08032915185892916532, -0.0066957875875345487099, -0.081250743455738475007, -0.08131122936782111954, -0.083490374135656991728, 0.02724190536924121081, 0.52309250852172883839, -0.748190657639665746, 0.51648514517917487598, 0.015986808124015323973, -0.093794593461252628863, -0.080773582350299799604, -0.053576005197284524983, 0.06526208670020532121, 0.004682445229131206553, 0.022918090645647856696, 0.0017220210942492735449, -0.026867008639632858347, -0.10025172986427012545, 0.07605584780565241565, 0.51678472225697291709, -0.9805101297781781966, 0.51701959519862739079, 0.07411357866061327182, -0.10674354647421797904, -0.037767692628501561836, -0.0037295662993747230596, 0.046649997155698202933, -4.1424337551734639624e-05, 0.081383445073322346897, 0.089276396973702237814, -0.07089975538992551729, -0.43425902964466528422, -0.10586641605747129646, 0.85124648104714262242, -0.81625225467590700745, 0.84895344721229915042, -0.10875115181299394029, -0.43417772023390732716, -0.065952839415490016606, 0.096048104953415369689, 0.073047580973088027667, 0.0076872196680934524321, 0.048407074950593660667, 0.027411296213621216128, -0.072510724425824624184, -0.20467798435923398315, -0.064836105044495162542, 0.53841615138561382814, -0.56074399479637060395, 0.53799964374778064258, -0.065441104191780716248, -0.20444401533204367705, -0.068795126179252499843, 0.037253409561542681627, 0.044651978974286084256, -0.08514458362893845278, -0.18393021307025939071, -0.0927786821678699275, 0.28053056127370684658, 0.53023439813136552523, 0.24218195057708780116, 0.040684017300289629371, -1.5072958574156927192, 0.051840917814670796349, 0.26394031071112572517, 0.55785175625012406453, 0.29629948985982673682, -0.12168236772255532074, -0.29031166450059592288, -0.12306655174615185877, -0.25081496440807549142, -0.19176935372806550029, 0.17134022291424477746, 0.85978801676755423067, 1.5399155809636164705, 0.084224033067450437828, -4.221157931857391965, 0.099702685823704034762, 1.5653296642210314804, 0.88221614691257532748, 0.16976267521020915741, -0.23772089611978669543, -0.36676370648728801038, 0.24082313356209833488, 0.050712069793078365376, -0.22041886284530778917, -0.47550193128960660527, -0.47110821283648984803, 0.26548557774980380275, 1.0357383312511569962, -0.87772604941973009218, 1.0263213116485865317, 0.25324617764011103649, -0.47276916850454608898, -0.44812910387953824864, -0.13841318935090829445, 0.2144762808815906252, 0.20838444889586985043, 0.054668166138412334354, -0.18342485525728755014, -0.41530672160063458742, -0.43218510055576336848, 0.2016261541061180329, 0.9591441571217197426, -0.80448554571108454869, 0.95065532213356895852, 0.18993666137947193273, -0.43586699835145392701, -0.39463178591270808848, -0.11377722937177832396, 0.20204510011419346549, -0.043352090549384436624, 0.1795506839690474421, 0.18264010991408996532, -0.095794450116928053784, -0.55109570150334685756, -0.46312292048120157206, 0.74688648057334661168, 0.098448597325852357787, 0.74913516030491855702, -0.45949489113245495231, -0.5481785487959329739, -0.098885136066846707203, 0.16740171517795721101, 0.14652064577815065993, -0.037467450661583336091, 0.15350355972814688776, 0.16200878710511845693, -0.066617253364419024009, -0.48282136418022825186, -0.46361867309892151878, 0.68188705617153233529, 0.1156792076417213766, 0.68326752943857638201, -0.46142955034314869778, -0.48079807336939656137, -0.068234485475160658319, 0.15078429991400787991, 0.12322835951268443222, -0.024370434986526894744, -0.16959947120708857016, -0.18377366364730013393, 0.017481080130904266129, 0.51338345783590533244, 1.2452537104846534977, 0.30607855004832856816, -3.4258360136036287713, 0.31021309228735993724, 1.2512963175512841207, 0.51722789285336423504, 0.014502074277115582432, -0.19227185088763215637, -0.19139613265936691877, 0.42453349092513298357, 0.59953685727882255474, 0.43894729197865545167, -0.29606179801175613875, -1.8218135199951999148, -4.3700027911065939534, -10.11147500120036824, 30.309637691702505435, -10.138625931592599727, -4.4106288737083376716, -1.8472800535531073152, -0.26226829957563657647, 0.59269902048429912167, 0.9179557103539851548, -0.035009934584720764361, -0.093836587641062577103, -0.075805063038429959521, -0.027737596362305001751, 0.15468362517366085473, 1.1094708329102820343, 0.64354103247783445152, -3.3603691634857635329, 0.64779819099655033821, 1.1148782519172815153, 0.1566999281298374902, -0.032190729845683267729, -0.084171516298643347942, -0.12419171281811021423, 0.00043198536411188058481, 0.062794086553731434663, 0.057693863960374236477, 0.036288619608599054223, -0.17811758464669846802, -0.5098836645278044255, 0.39904678649403357849, 0.27789008416957838854, 0.39817984723529742386, -0.51321477403148607621, -0.18447142657860607362, 0.02774754309714619338, 0.051855912796269963694, 0.081409610856871622575, -0.00089569319740665474805, 0.040026009203208992582, 0.040820338869171440832, 0.063824012383630696066, -0.10981346545662132308, -0.50044605600465363882, 0.33438767350178577598, 0.27443259586120727356, 0.33409261560061093155, -0.50237001961122129678, -0.11400317249401928366, 0.057172737545824831029, 0.034469532238377759148, 0.054004748098617411101, -0.20779811259408703661, -0.17900424807414458406, -0.054470865189237646642, 0.20473789309280812643, 0.78806961194067826071, 1.7258615502287881771, 0.14024422496479349576, -4.8326661410695033183, 0.14858977873761280986, 1.7387231447988102939, 0.79545238648577731855, 0.18732633459853612057, -0.12054949308247253636, -0.3346289938416392995, -0.18019540359508923189, -0.15989416950028956244, -0.059499117138262211379, 0.16103439615793962991, 0.69213860885988853777, 1.5955075139366003967, 0.19403040909197344366, -4.4873081238839533924, 0.20198156282617524271, 1.6080030018874260644, 0.70010314831243203315, 0.14688168869541023964, -0.1169649676202528521, -0.29728001883117000093, 0.02845531198478761653, -0.11986464111227541907, -0.092140157898010513482, 0.21830394726378804515, 0.28845464801126702303, -0.42061812261401332336, 0.013886312252494559596, 0.16029538980465540821, 0.011903947900468284798, -0.42343410142072929192, 0.28842818423238014702, 0.2243174842540650582, -0.081575765189340454508, -0.10281788630159682807, 0.0024385439859227081051, -0.019860938748655885405, -0.0013385356673794842133, 0.10051453905818590051, -0.019625399719699752887, -0.39223482169101892003, 0.28086334094371090409, 0.089644394832013404861, 0.28042110315538715248, -0.39055172804356541594, -0.012389057676086211515, 0.11086299973528823681, 0.00051781194411292842672, -0.032694760584730135633, -0.23549572879562438721, -0.14139131273078670015, -0.0089946376235697615698, 0.18104437401611853109, 0.47874989109597920756, 0.7139470899304334317, 0.68172301836130833053, -0.12399146287138804179, -2.9527322320325795069, -0.13352321447014692257, 0.65376057864542069797, 0.65185370103427464628, 0.34322090447267383784, -0.077273425277122081978, -0.002245666574649559169, 0.002880577620882130048, -0.0015251102945601411297, 0.016949017345463873035, 0.06287665842871648092, -0.11053638756133438159, -0.18101994770560542181, 0.49534672357865150438, -0.55896749123131284076, 0.49110625418091957695, -0.18908786992619894263, -0.11184205038472337657, 0.078169310149844684066, 0.0089065288814828228597, -0.0033679297229067929956, 0.00052922947265169836308, 0.020070061017111846813, 0.031364317178184317192, 0.0027048132096155318854, -0.12553088982829815889, -0.12264311573407093992, 0.40354098470822002964, -0.4110035629463423823, 0.40027727863022277699, -0.1273069480390630337, -0.12332338263620928887, 0.010592602648681516062, 0.044773229660661821583, 0.052059578293182300068, 0.025152583041477105663, -0.020581530416045830389, -0.075159180659722296669, -0.070386198785718653848, -0.076459893002626788672, 0.027241904887652115819, 0.51648514579184290341, -0.75849213587206232301, 0.50768263180828754333, 0.014796043282618541459, -0.07866070867786321974, -0.050776275344677671009, -0.015347096774546195586, -0.024286396126083825697, 0.018170782467692075945, 0.032520570006648409633, 0.0064654106925610349901, -0.025319325433033917916, -0.10029203539035265458, 0.076055846988540565246, 0.51701959554020715526, -0.98173801694905937776, 0.51153021175381041541, 0.061190835264844631314, -0.12089959449971093841, -0.031052808038963038845, 0.063277368085936050446, -0.023963212362026274638, 0.03012117823491863508, 0.082473198637335903838, 0.084245278027723510816, -0.071950715793340003223, -0.43311608472543666748, -0.10586641627381157549, 0.84895344768540881297, -0.81961886995173283132, 0.84566511898723439877, -0.11618964770100172201, -0.44470132588020072628, -0.052722296573516562457, 0.17804039386425385416, -0.0021846161380275419886, 0.017422342483642092037, 0.04521833139584548189, 0.023555956811003211926, -0.073992916179718093428, -0.20474050713225575016, -0.064836104975647082527, 0.53799964380698772626, -0.5613413536321127939, 0.53706536378021374478, -0.067907902335948858163, -0.20640081904007950842, -0.068929734663116054572, 0.088321118675922280494, 0.050354484974333806047, -0.11043980163208781675, -0.19793307666419904245, -0.10737639207544567854, 0.26670474116597797654, 0.52152573105455679592, 0.24218195111952506293, 0.051840917388020990586, -1.483610564790996067, 0.087423146332852350038, 0.30382633973079886669, 0.57894753902975326643, 0.22479212432768488905, -0.43618952044012559988, 0.038117476725483413758, -0.15031837796410402186, -0.25886747952799638028, -0.19957850467921853244, 0.16027729396777148918, 0.8504861068344412578, 1.5399155809640561188, 0.099702686131033640971, -4.1883576551031556079, 0.14460880890466937254, 1.6059776055732948308, 0.88312116901250592438, 0.048770182235746681321, -0.58009699849702189844, 0.18667204280845731135, 0.11135669136298169402, -0.017575695534124607944, -0.23809173549245199464, -0.46752347939475102923, -0.46118860338442524416, 0.26548557842379111316, 1.0263213111278828293, -0.88724252506540657315, 1.0281328179772799558, 0.26923285920153339212, -0.44669638373267561704, -0.40251565471150557718, 0.013103645484924594106, 0.15169777243609836037, 0.10217081226783705783, -0.00090072833462107038116, -0.19688229696705650262, -0.4076444384647615049, -0.42337583500984748985, 0.20162615468348155101, 0.95065532192989898874, -0.81390539821938689968, 0.94959066030775818312, 0.19960379791111515502, -0.41674425130860759392, -0.35203871855312929284, 0.039512157931634483976, -0.10102377687837492404, 0.064355192994120583028, 0.1995047880023097242, 0.1729883298898931554, -0.10654776646540486951, -0.55600689513047318879, -0.46312292036985325261, 0.74913515989763490577, 0.099812213587033252193, 0.74372780376242575429, -0.48323737718095105542, -0.58914930666043785479, -0.10106463287960004704, 0.37660526054786636063, -0.08559152902565440868 };
	std::vector<double> l5 = { 0.052944722906078753311 , 0.1704382150070918589, 0.15447130907877595907, -0.075079672435713368239, -0.48662687041769148166, -0.46361867321071942305, 0.68326752891462927408, 0.1156873092042310508, 0.67753285190345691191, -0.48132259496786927722, -0.51273902671796045549, -0.067660238325043142726, 0.32360036018581561645, 0.058582404928639676689, -0.083389208029631775587, -0.18355210216045350635, -0.18150481739553334104, 0.021015636777725082984, 0.51395452880340142254, 1.2452537101485035009, 0.3102130923907877591, -3.4136717403204723809, 0.33131463133063504545, 1.2756759984261005769, 0.52703487950069105583, -0.0426310834065866881, -0.38313081522659786504, 0.04855059227832505403, 0.40893281120015162466, 0.59236444860046844063, 0.43229268166954837316, -0.29694167064377280374, -1.8162798405322597706, -4.3700027911123457969, -10.138625931472530439, 30.237519305189660912, -10.245927534911380974, -4.4897069918954457535, -1.7747751995890017618, 0.091306231652229868923, 1.3155679700488913664, 0.01952464624280460942, -0.056721556786752704016, -0.097339638841327999486, -0.075116496318314990299, -0.027322242759180159766, 0.15375628052849879546, 1.1094708335907375041, 0.64779819026983342933, -3.3499754049533780886, 0.66260084383952122522, 1.1259725303038232713, 0.15297658828362437067, -0.069871296314081537959, -0.19752812362553942571, -0.044310263931506430057, 0.032806149097972872597, 0.077530601900398621318, 0.062525066064889139872, 0.036683953156154552877, -0.17868805181218011535, -0.50988366515171057447, 0.39817984772006065342, 0.27277181428850288958, 0.38646631391286101476, -0.5307695980812244807, -0.196052123233389608, 0.047030505158023705459, 0.15030156021063736937, -0.033069197772171848859, 0.022434636938497162784, 0.05206019994568522602, 0.044770985466433738109, 0.063659195568487175754, -0.11071478607633916669, -0.50044605614324988441, 0.33409261625995934075, 0.27084340137253282155, 0.32524020660422714446, -0.51543873118892014062, -0.12084956613656568092, 0.072746658099416952603, 0.098202847685146157919, -0.078057445622261148466, -0.16015482630737379321, -0.1555634750660210841, -0.045666220291608342063, 0.2054842794863319666, 0.78554471751998167228, 1.7258615502395249219, 0.14858977915586715768, -4.8126120022586720282, 0.17745899852497498683, 1.7598652373014660455, 0.77246067378187399655, 0.06032480403680005876, -0.37390142766348283265, -0.062385029425217869425, -0.14063192712125907602, -0.14206028488752941996, -0.053678239110654779598, 0.16073681448328797083, 0.6895340533097905622, 1.595507513813724465, 0.20198156280902507254, -4.4680736746779894375, 0.2298196547369242837, 1.6285274509801386245, 0.67916358089471984272, 0.030540700080789767723, -0.34155084891553416648, 0.067010154723080556094, -0.043844605623779057957, -0.13054932206625211033, -0.085364747255108286428, 0.22434864694872341806, 0.29088118228313775138, -0.42061812262126690998, 0.011903948125867628086, 0.1567605405614374281, 0.011743855991909394124, -0.408389161460624317, 0.32837293686087121181, 0.24158322327816125763, -0.24753902655595660942, 0.02160193279868746763, -0.0084828820732611599081, -0.027825253055231334576, -0.0066907893830413346994, 0.098499216579938361771, -0.019752764723690096477, -0.39223482138065146607, 0.28042110267101266752, 0.090225146659994898135, 0.28507974009366249257, -0.38095646552540307628, 0.00062799544739941580776, 0.11265092185139692627, -0.05567208556003631903, -0.15399983807689615611, -0.13586541389155179504, -0.087010464871609433102, 0.012416653481644977092, 0.18515923103476694678, 0.47874989190958883167, 0.71553605582703561083, 0.68191360025982339366, -0.13352321493943095088, -2.9781545630231200938, -0.16377883238923562903, 0.63930886595279656781, 0.67055100669034573002, 0.30004248364115160275, -0.010004512663132113584, 0.004868808764044937204, 0.0067741258628631166594, 3.8460218811499444908e-05, 0.017144438039726615786, 0.062876657880152189706, -0.11062976451977296621, -0.18247456653286053818, 0.49110625348726621997, -0.56608277108464433169, 0.4883193889704086077, -0.18058149894205641206, -0.087289902976827982162, 0.065889631098503848028, -0.0056507212756669641246, -0.00053804413469203660452, 0.0040445692776806211471, 0.021384295008882819172, 0.031618260832132497051, 0.0027048123724527652588, -0.12595794828519565889, -0.12429088865995602253, 0.40027727821933767327, -0.41392152017818878429, 0.40188669782538094521, -0.13362477601380726155, -0.12631551907988558536, 0.067754813481495951288, 0.01291763210934158447, 0.032558432712957566246, 0.024913299910991087982, -0.013065526784718353714, -0.069002615336398370749, -0.070386199903455320959, -0.08349037390171831774, 0.015986808589314401019, 0.50768263223868304035, -0.7566095894443100045, 0.5270796782221554011, 0.040536853642991178825, -0.058304814369751926939, -0.11493372111636623867, -0.02947456472234364841, 0.00063544315739903572092, 0.026952398049034060473, 0.034147144436960909231, 0.0060863463521446193108, -0.025319326764183629208, -0.10025172975546807475, 0.07411357912882471366, 0.51153021253036423666, -0.99178114362036096097, 0.50025322018128459689, 0.0553326960944138177, -0.092021253263653263188, 0.030013273017204890325, -0.021620845845778904892, -0.0030103079476196436051, 0.035304756005928604468, 0.082930801975951387583, 0.084226669559791300657, -0.071950716167674719226, -0.43425902984111136584, -0.10875115161893608717, 0.84566511912730302214, -0.82519098563295656401, 0.83038902201625197641, -0.14894672664695746134, -0.44374662471661463492, 0.17579693708009103137, -0.0029270320236111954486, 0.0028417987710002731543, 0.018034674658863533714, 0.043783637647819874383, 0.022652609585429396566, -0.073992916258422733455, -0.20467798429658654036, -0.065441104237533784249, 0.53706536353756362256, -0.56201117879327144866, 0.53457336333753846525, -0.089687156387276964487, -0.21277189559980475009, 0.049879394335651090764, 0.10053678349102226908, -0.024727942440556399673, -0.14671495519920449424, -0.21299969994998724232, -0.11326333156862951335, 0.2667047421618170544, 0.5302343980738984941, 0.26394031077003488051, 0.087423145864403861705, -1.4411179999631305026, 0.12484835824981602159, 0.3074997110136298839, 0.48867849618607417383, -0.22280643222743609644, 0.10268312248892222383, -0.041751138446766818124, -0.18779120347356231813, -0.27414880109533373842, -0.20530271464603563891, 0.16027729390118802821, 0.85978801683041639059, 1.5653296640012961394, 0.14460880922282726146, -4.1303144811499867828, 0.20122896695716580018, 1.6094173962100042363, 0.71175641362031438764, -0.5003841808206687114, 0.10344364237737242729, 0.10876491162891610553, 0.076963488259846432604, -0.022205736202365294885, -0.23280256334276427932, -0.46752347889119316715, -0.47110821335597030668, 0.25324617728769438507, 1.0281328181845774683, -0.86038565257926302721, 1.0706071144702073994, 0.28786576544091729835, -0.46517991985296330837, -0.43112667994983505659, 0.078164344298204194161, 0.093246643981763410736, 0.076179483216996690675, -0.0038198444499800421337, -0.19215808566131048485, -0.40764443811160838704, -0.43218510079359445664, 0.18993666109105797335, 0.94959066070146169736, -0.79492895207050984219, 0.98183361332372809382, 0.21421458975331778007, -0.42111638668065298763, -0.34969747013570495753, -0.075438382001400905441, -0.022955145586709205241, 0.081205161880031462296, 0.1936441004799397958, 0.16634903524100172478, -0.1065477659603839139, -0.55109570098814408823, -0.45949489139293459017, 0.74372780334060395457, 0.073808890807275609469, 0.68544044567844331084, -0.56025148754253173422, -0.53717930057755003403, 0.36536398283398796583, -0.06455733798100654619, -0.020223997018755934807, 0.068211649371436158806, 0.16643970669558333375, 0.14924793843959738293, -0.075079672375181399913, -0.48282136379031465045, -0.46142955021283077466, 0.67753285174279531677, 0.093197923894919751375, 0.62896789220144999621, -0.54647247499377427093, -0.46623252673126741596, 0.33030775170257975493, 0.070792542683733930553, -0.0063514851626510233223, -0.1050154291607666851, -0.18537026257609998114, -0.1792150363279854941, 0.021015636009432951681, 0.51338345821114406142, 1.2512963176655722553, 0.33131463142368722252, -3.3739417846019046365, 0.38530984981610555407, 1.2997215666025900838, 0.41365257560305424001, -0.42873955790791268283, -0.14598390775602609892, 0.18747845451476477519, 0.46583326302534167063, 0.6017761644709728186, 0.42879850203919234186, -0.29694167046536268373, -1.8218135202963015029, -4.4106288734549989883, -10.245927535291844634, 30.066663602193102633, -10.385637622635471189, -4.4092723925831265319, -1.2695764284111805686, 1.1796002096955819916, 0.035297746607895523485, -0.012162754783351934981, -0.068069603015621482633, -0.099707021513753602004, -0.075127374815332964575, -0.027322242183932712867, 0.15468362469488475264, 1.1148782526751224164, 0.66260084308233102579, -3.3243718507749222546, 0.69443418959675795588, 1.127576657256327719, 0.047227186185939573626, -0.22404090860060610346, -0.049040763828165148674, -0.0028057947953446515932, 0.048349193663640613328, 0.080348875648776907354, 0.061538422511966756057, 0.03668395268653185104, -0.17811758497559973535, -0.51321477416326610754, 0.38646631539753040219, 0.2509885100949389436, 0.3598115989203924947, -0.54349526365200662159, -0.11962740563341389688, 0.18107273246994790017, -0.036437453763727735423, -0.0020919449684852625682, 0.033970821463271509699, 0.053941173182751336879, 0.043732158072838867025, 0.063659195331859930178, -0.10981346614419229546, -0.50237002034262212558, 0.32524020725296420853, 0.25340553012115274445, 0.30447631920119061499, -0.52268824354804055421, -0.051827193307271916811, 0.14630796369643170318, 0.0059560042356189297347, -0.095084412087795347479, -0.16335640593647976493, -0.15323588766315976439, -0.043876978952067741124, 0.20548427986785228194, 0.78806961163116895364, 1.7387231443138047027, 0.17745899872892798022, -4.7735790022235722319, 0.20854971008427858337, 1.7361160230667800342, 0.6203667787667195288, -0.22790387741659276499, 0.01246863244942595858, -0.081765696012335395415, -0.14650444108656204834, -0.14156172069983738449, -0.052600052364192184662, 0.16073681500571698755, 0.69213860875239607839, 1.6080030019302919975, 0.22981965489201849895, -4.4304964544705427443, 0.25942402662972058325, 1.6050340272776115746, 0.53683688770197024454, -0.23104895778983408627, 0.048011219983926704813, 0.016229082495948846804, -0.053105571335365975294, -0.12587662755458922659, -0.081427955866194334789, 0.22434864692725445257, 0.28845464838948098318, -0.423434101685884412, 0.011743856284057446401, 0.16699622637894953603, 0.045584687115680831704, -0.35652212679015249375, 0.34782847878768208272, -0.1072151758656468834, 0.024249902356339496368, 0.0028584567581464940395, -0.01699307239163382055, -0.029625341902858738641, -0.0065463354157932751587, 0.098499217273009184725, -0.019625398718000466869, -0.39055172817184596878, 0.28507973873902459072, 0.096554224041597955241, 0.28983034360575482546, -0.38064136720204999609, 0.014335429178875711345, 0.031717987423411718451, -0.062092027029106343361, -0.10179885403556478141, -0.11189984490631353642, -0.079490041711629100263, 0.012416652927876102586, 0.18104437458622663071, 0.46895210324536712454, 0.69708447112251803457, 0.65376057841026236073, -0.16377883315056052105, -2.9985841407317996854, -0.12778468555579361454, 0.74576388778111069033, 0.9036078577178392468, -0.0093029928854771303665, -0.0011609491435811713535, 0.0075080468545364988445, 0.0072297117753096270393, 3.8461069644538017215e-05, 0.016949017266925509723, 0.061921911610023791517, -0.11348606312016303044, -0.18908786970769839164, 0.48831938979911621024, -0.55319777861057706225, 0.50337912095029269111, -0.25454555986394250544, 0.033836230091450578139, -0.0058391911950046266819, -0.0010922338025323799233, 0.0018442042302784593909, 0.0049319519168793315295, 0.021384295742803240137, 0.031364317411868306018, 0.0022006957809330945745, -0.12776802910093862375, -0.12730694840198869544, 0.4018866991539655209, -0.41503688372241448468, 0.39178744930944137126, -0.1534259974547516614, -0.02715370448039189713, -0.014173869797634554291, 0.021459983282406842886, 0.040730693454500803419, 0.030085133690331188061, -0.013065526628783144525, -0.075159181294701948861, -0.08131122945119491352, -0.09379459386455436809, 0.014796043854408330587, 0.52707967865357185921, -0.71916432196095403384, 0.56624408030491757415, -0.016734210211745734892, -0.19025820926811323752, -0.0192693849257384274, -0.011192345426605302544, 0.0065987113034546747425, 0.027546903162147917676, 0.034147143888243559895, 0.0064654112350639118129, -0.026867007989298790499, -0.1067435476214880774, 0.061190834505111528496, 0.50025322154904483529, -0.98673175117352274199, 0.53926314587946322376, 0.060675353438149767782, -0.087672111660137011513, -0.013162252728493162648, -0.012363197838896687608, 0.003061806276845870263, 0.037562500665856500548, 0.082930801766084527582, 0.084245278143561350137, -0.070899754718491053707, -0.43417772097648077434, -0.11618964776350211576, 0.83038902249851698212, -0.84096552958720394155, 0.82519720367143900575, -0.16438092975002080443, -0.2189571786487272409, -0.0024631260425425085224, -0.0010947709639297835502, 0.003819868347422971603, 0.018052183057745883887, 0.043783637944890742288, 0.023555957014703532426, -0.072510724791111758059, -0.20444401501690526168, -0.067907902423318969309, 0.53457336342412997698, -0.5731486163622007135, 0.52823124099419438426, -0.078482797822950409183, -0.15581763962656972056, 0.090765112932393587908, 0.026018002365943367765, -0.060501517765755422062, -0.1583117916118952917, -0.21299969935213747263, -0.10737639299709435259, 0.28053056095767409106, 0.55785175677865439514, 0.30382634032449967743, 0.12484835718610413913, -1.4371350918286611709, 0.069372444341821132285, 0.12010580358472340567, 0.42600666194579106527, 0.10065465193982034431, 0.022189160901004402471, -0.082191626738302187016, -0.20138489135777945527, -0.27414880174856870454, -0.19957850447064784394, 0.17134022302738913246, 0.88221614677913584579, 1.6059776053948859875, 0.20122896695438652337, -4.0977734190804877557, 0.1765266823167433996, 1.480607259529234554, 0.24996892529355849155, 0.024420856115028602429, 0.078556449486508750368, 0.10327041386859185657, 0.078654741023371152475, -0.022205736500606476636, -0.23809173542898426357, -0.47550193111807281809, -0.47276916810909169753, 0.26923285937428548253, 1.0706071138941966048, -0.80336261878954162707, 1.1218967142778861756, 0.20349314638505447417, -0.94817326240009647265, 0.013414787636513016839, 0.063500943270966184184, 0.091238090947835553957, 0.078275970964202828961, -0.0038198445744344611263, -0.19688229684135771791, -0.41530672161312914836, -0.43586699781450832791, 0.19960379774748160386, 0.98183361284479575559, -0.74678205431192590247, 1.0347055613446698441, 0.15253118407609922436, -0.82628272533459323768, -0.030695751467309073668, -0.046733128083957492516, -0.014415025206482736872, 0.080229648130154901375, 0.19364410116593802624, 0.17298832977126249455, -0.095794450852117407624, -0.5481785483416041771, -0.48323737724822879436, 0.68544044528212788592, 0.0032809819239557613813, 0.68744931161927991248, -0.43131737821938526123, -0.18688725447050136275, -0.027492858189161972771, -0.039477821413752882396, -0.011979502623690983787, 0.067913717587708033463, 0.16643970730877016839, 0.15447130909185069481, -0.066617253954296420937, -0.48079807323741863279, -0.48132259521112041867, 0.62896789215337345347, 0.032668614792107440736, 0.62737212200444714139, -0.445560088865177506, -0.13709867203354270115, 0.049939194133073759596, 0.028291334599299975544, -0.023779297279954533101, -0.10850689953517973774, -0.18537026289749933272, -0.18150481674967755552, 0.017481080566972936885, 0.51722789213949360576, 1.2756759979267078275, 0.38530985063210593244, -3.3192576211873916492, 0.39666069530228198481, 1.1949910193028241423, -0.025787926590294583257, -0.18570550646101682823, 0.014109538807961034476, 0.25155697750391547984, 0.47939238413298901342, 0.60177616462235938855, 0.4322926811826475757, -0.29606179795130765919, -1.8472800537753020222, -4.4897069913800589092, -10.385637623080732794, 29.89562301541274536, -10.419697213495556554, -3.9686601150501084057, -0.17838617283807328695, 0.03032044709807469457, 0.010445387717136234498, -0.02354750956163471215, -0.071205436067458882654, -0.099707020651441824466, -0.075116497062753945224, -0.027737596087077133478, 0.15669992765616375108, 1.1259725309977961505, 0.69443418928518108046, -3.285937325710146073, 0.69525358557304672846, 1.0402820004109052388, -0.15654453443205007535, -0.032062670707415433236, -0.020323128907054652942, 0.007990414313591404083, 0.049999967299622527506, 0.08034887463706544275, 0.062525066416395075342, 0.036288620066666568698, -0.18447142672391769747, -0.53076959903104548033, 0.35981159923109434784, 0.2327788323841271656, 0.39104203002590476324, -0.51641259445375531367, 0.055463222110869757342, -0.023462166451521888899, -0.015079746207152650711, 0.0054487478180983150988, 0.034828683432384932583, 0.053941171603838211634, 0.044770986070814713931, 0.063824013106181715083, -0.11400317230387888023, -0.51543873117691219043, 0.304476319114027838, 0.23912337274230824335, 0.33071500042092200067, -0.51712661617988830987, 0.10237172599905780546, 0.041775944455810794242, -0.031750880029273745309, -0.1082134378851029588, -0.16504064783498645363, -0.15323588829560408353, -0.045666220312176736273, 0.20473789342688047954, 0.7954523867887145494, 1.7598652378575276867, 0.20854970961033764132, -4.7662593788015694685, 0.194827316496097247, 1.7368031290444141401, 0.35902159490924773211, 0.042278515875225353149, -0.02516250730364146676, -0.095615860646519668098, -0.14881584883957765797, -0.14156172059630381388, -0.053678239709039926586, 0.16103439642328673531, 0.70010314825382313764, 1.6285274508941947058, 0.25942402653024521086, -4.4231644160152683298, 0.24771544825157595926, 1.6021347806567298733, 0.27470519792922887259, 0.017552900917021115046, 0.031289997329282143168, 0.011588366064715832637, -0.052402929885955316924, -0.12587662780253533468, -0.08536474672667654362, 0.21830394713903536474, 0.28842818428057376279, -0.40838916218085452581, 0.045584687640006227305, 0.20485649631409619253, 0.043845152028388727217, -0.52850148106757977029, 0.34586628172732591935, 0.015915223101477016937, 0.0091009402763021669813, -0.002152756279034421464, -0.017565221186265925213, -0.029625340156769781191, -0.0066907901041274985693, 0.10051453774766801186, -0.012389056856408961829, -0.38095646458339399443, 0.28983034272752444815, 0.092107783509780855047, 0.27804482506264494024, -0.46575194484401322148, 0.13030534704739663021, 0.015546386286129084547, -0.049784821498218299884, -0.096774649052003891558, -0.11189984458102367193, -0.087010465017724053505, -0.0089946375107023430334, 0.14101300402764760378, 0.41500118651065959208, 0.65185370109416551632, 0.63930886627630045815, -0.12778468593833966493, -2.9532985002360470084, 0.085274362059096789346, 1.461915463107094082, -0.0042651474331159809628, -0.0042248861565507131108, 0.00024502567921054951763, 0.0075080471586360299913, 0.0067741248192004699033, -0.0015251097791386637129, 0.013605510462095212268, 0.057891717164081389624, -0.11184205076092554321, -0.18058149949870147544, 0.50337912112793459229, -0.59100268667772137832, 0.51195116240326410395, -0.21066235669047933543, -0.0034988154757745431982, -0.0022826131337950108374, 0.00019950208758543624705, 0.0018442044479175644794, 0.0040445685000444590862, 0.020070061167026274906, 0.030459142232337699241, 0.0030864573868267276835, -0.12332338233113823711, -0.13362477734298430776, 0.39178744954218963636, -0.40379145894459667065, 0.43699303886439266886, -0.22447093925745492671, -0.022393765331675925812, 0.0055297667989604466082, 0.029714441515875246458, 0.04073069347322571393, 0.024913299103868614959, -0.020581529772334523909, -0.081250743065902644724, -0.08077358276751124555, -0.078660709191508623728, 0.040536853755463495008, 0.56624408036254336718, -0.72442979780207794605, 0.48470683413375120763, -0.18261643678545144698, -0.0052520474536421586989, -0.011598927683189649146, -0.0088098928820598398221, 0.0065987108039545459837, 0.026952397275472952942, 0.032520570822623093288, 0.0017220215520812971396, -0.037767692702471038213, -0.12089959497645323683, 0.055332696167197888137, 0.53926314662548746615, -0.93686221416377790661, 0.58736763157489402243, -0.13245192327123134213, -0.0060727166053352764005, -0.0088738328844707690946, -0.0089014064974486765969, 0.0030618060906355158354, 0.035304755868673480534, 0.082473198974088973712, 0.089276396363019872027, -0.06595283890911506075, -0.44470132628518693751, -0.14894672684186374689, 0.82519720389589179543, -0.78114357048249749571, 1.0460525121052992592, -0.62336455003239454875, -0.0022091066903116526356, -0.0019555519989070233207, -0.00060804121780361653372, 0.0038198684453940173755, 0.018034674739723751541, 0.045218330957924733615, 0.027411296433223326929, -0.068795126391438976565, -0.20640081864983975124, -0.089687156712389415336, 0.52823124097948326305, -0.51385399958368982443, 0.59938224593162270359, -0.33988082747522152527, 0.050035927857992212875, 0.039714177886971700759, 0.0047996334543779174339, -0.060501517381081823532, -0.14671495456751054709, -0.19793307730890805085, -0.092778682438119708387, 0.29629948996769217562, 0.57894753967789092108, 0.30749971072637249936, 0.069372443947790740637, -1.6414802205476604069, -0.082867217637147508325, 0.9001046367465883602, 0.062707032084204195055, 0.042725076514955710738, -0.0027508108611487462961, -0.082191626962688654756, -0.18779120353697728585, -0.25886747930852915811, -0.19176935359614055709, 0.1697626750781365268, 0.88312116879323920848, 1.6094173967816383097, 0.17652668235457980028, -4.2258749235033112868, 0.18790940993679783322, 1.8525738342080038112, -0.029876536942735135161, 0.036010718262555731217, 0.084724236385591705778, 0.10327041364039629689, 0.076963488901007029441, -0.017575695733545310118, -0.22041886316309400318, -0.44812910391653204556, -0.44669638392515231828, 0.28786576609724873199, 1.1218967140240523328, -0.75182449622905089814, 1.0143557120200841215, -0.78881019561276799479, -0.028369347415984083605, 0.026846493143572788126, 0.070116025682745289349, 0.091238090870748828065, 0.076179483722220053821, -0.00090072861156834105323, -0.18342485523556009697, -0.39463178628429790296, -0.41674425142313098469, 0.21421459014374752416, 1.0347055613181137534, -0.6860423996904280175, 0.94649047995876833816, -0.73299131944190387689, 0.0014738907253537712402, -0.032070394330657857329, -0.045893854263445499342, -0.014415024750324160682, 0.081205161812039641789, 0.19950478753106029584, 0.18264010990402715362, -0.09888513581793434426, -0.58914930610431104352, -0.56025148770732957626, 0.6874493113385321541, 0.19362333586528526763, 0.92441911419493483226, -0.94705452155379266088, -0.00012559472068221429501, -0.027167274616330491788, -0.038170256899893298397, -0.01197950233677697561, 0.068211649067021054571, 0.17043821487492577527, 0.16200878715191377455, -0.068234485303624456409, -0.5127390260379637299, -0.54647247537792897898, 0.62737212181383283927, 0.18781932795417644044, 0.82312222585199679159, -0.84969154932886992171, 0.021689876531699014972, 0.030718271442965707635, 0.019990998257000937299, -0.023779297479549151711, -0.10501542963594702462, -0.18355210166223404067, -0.18377366363908448355, 0.014502074491777124446, 0.52703487907747559316, 1.2997215665683468089, 0.39666069576291257404, -3.3922446976518383366, 0.40948464688539859546, 1.1927450151734866335, -0.13664604628517418305, -0.075792396296895769847, 0.053833294990727489915, 0.25155697753346667467, 0.46583326320941581544, 0.59236444853737912908, 0.43894729173859348137, -0.26226829931071415913, -1.7747751996307816746, -4.409272392452787237, -10.419697213513403611, 29.576420708989449082, -10.639700121773245201, -3.7380343785028098402, 0.017369888156920938044, 0.015603221398262756558, 0.0039288782369784336357, -0.023547509109104351743, -0.06806960367334696782, -0.097339638413158069419, -0.075805063473855779921, -0.032190729362841646877, 0.15297658802188870975, 1.1275766574142944698, 0.69525358533052494803, -3.3103775393697163487, 0.92060251370918988467, 0.68903188603074594365, -0.010134625086843750477, -0.019673128397297207082, -0.015983847583757579414, 0.0079904136864605013191, 0.048349193921272223817, 0.077530601976707330669, 0.057693864596702781888, 0.027747542401117361588, -0.19605212317637960018, -0.54349526310949536345, 0.39104203033642048748, 0.28452463728177967983, 0.35021914927378117177, -0.4717498008182111624, -0.0069192299743202172191, -0.014769736662663637827, -0.012301896783465736396, 0.0054487471755656907701, 0.033970822453384695583, 0.052060199649344843231, 0.040820338950820878043, 0.057172736796175432916, -0.12084956656520871865, -0.52268824267145685969, 0.33071500061277492311, 0.26232803779121915966, 0.25497276065987684035, -0.36935065901467806215, 0.045111622611662516302, 0.010317719124309385204, -0.042379132826884505225, -0.10821343799237971572, -0.16335640497468853494, -0.15556347544495888191, -0.054470865551452578301, 0.18732633441966706211, 0.7724606734525084617, 1.7361160231737644555, 0.19482731637398187363, -4.7698271845255657553, 0.46039237988195058637, 1.900585821655088159, 0.041956331864149092048, 0.011593305237868195556, -0.035693443456375033296, -0.095615860806638891534, -0.14650444075094998242, -0.14206028476998774912, -0.059499117561414113786, 0.14688168908885673725, 0.67916358120441899793, 1.6050340272040706235, 0.24771544815985890864, -4.4217631961810885954, 0.48946088076083643648, 1.6931849736370818693, -0.0030148878410462885517, 0.020208668805976184174, 0.030930624700587019266, 0.011588365610685715315, -0.053105570843193533603, -0.13054932240123953702, -0.092140157894339547173, 0.22431748426257414053, 0.32837293711075404845, -0.35652212715373732266, 0.043845152250422521345, 0.016469421166188758648, -0.26322587279920717229, 0.2298440628893172466, 0.0038933866181927231226, 0.009216251352769870564, 0.0072535144658613789556, -0.0021527553660630198602, -0.016993072723066510871, -0.027825253570539023457, -0.001338536032585491576, 0.11086300011927051334, 0.00062799611948598970182, -0.38064136788020702662, 0.27804482449784212461, 0.020680225573907588116, 0.17627151419100917096, -0.17590378665668493507, 0.067989119406024270398, 0.009640861487933364421, -0.049784819479717599966, -0.10179885437311840535, -0.13586541250526948388, -0.14139131394312734469, -0.095280895038495933003, 0.041987320211038611784, 0.34322090422403050969, 0.67055100736494777092, 0.74576388753696987166, 0.085274361923688979381, -2.6454185982072666583, 1.0974348818907353476, 0.001140044059134428904, -0.0028691849633458773455, -0.0042248867118775549137, -0.0011609486614873848042, 0.0048688088508739794241, 0.0028805774841466295216, -0.0072688429253510086769, 0.01337139350314962144, 0.078169311158515622506, -0.087289903684044031773, -0.25454555914247917192, 0.51195116195639334133, -0.31497372265473144726, 0.058615639126078804277, -0.00016799683287356610337, -0.0018858924443597743607, -0.0022826137940084485767, -0.0010922331798521765171, -0.00053804437796468158016, 0.0005292297968566523703, 0.01831251401958892619, 0.037715485513903690418, 0.010592603275191635886, -0.12631551936848667284, -0.15342599655668015535, 0.43699303807146128165, -0.26266341577561980181, 0.04468949683684675378, -0.014913508170622483823, -0.0072870604961337327732, 0.0055297662239014375163, 0.021459983033915393247, 0.032558433010118348339, 0.025152584216912094589, -0.0066957879399122625236, -0.053576004848666924918, -0.050776275752591998325, -0.058304815046151620139, -0.01673420954055629406, 0.48470683405184195047, -0.66596120007408088348, 0.31490399682686898153, 0.0062669653580294188214, -0.0044731611413058028001, -0.011598928851445569968, -0.011192344733037148247, 0.00063544443890028827482, 0.018170781587758717129, 0.022918089207719410244, -0.0037295652246264994825, -0.031052806778489379369, -0.092021255032601015356, 0.060675353328725263602, 0.58736763215634046897, -0.88960766724995432586, 0.34282141043630459665, 3.1990565866587315302e-05, -0.0037123560596722476011, -0.0088738332980171572006, -0.012363197754247391355, -0.0030103075623515909015, 0.030121177997680909089, 0.081383445198081147409, 0.096048104811678011727, -0.052722295884244409214, -0.44374662507916518006, -0.16438092967647233156, 1.0460525120390842257, -0.70360214454866987221, 0.14326237025975177564, -0.0021070954139475964878, -0.0020732807221499633538, -0.0019555519235864996311, -0.001094770887597310222, 0.0028417981783428686651, 0.017422342986709048185, 0.048407075061084901824, 0.037253409415914748304, -0.068929734868738698483, -0.21277189535309010138, -0.078482797640619578194, 0.59938224566425835071, -0.47106321303851406146, 0.13998114311680548427, -0.00065606737970140983213, 0.02681615827820822695, 0.039714178558056062451, 0.026018001990440780408, -0.024727944039478581828, -0.11043980043697075588, -0.18393021210759424866, -0.12168236852137254145, 0.2247921230601482856, 0.48867849767238730196, 0.12010580350177874065, -0.08286721804396758595, -1.1455023137749571482, 0.74787509072143387723, 0.010153861070462599434, 0.03440388631225495264, 0.042725076046920169859, 0.022189160784047902431, -0.041751137714485818397, -0.1503183786902572705, -0.25081496468029251812, -0.23772089553218134217, 0.048770182750711120467, 0.71175641288526647621, 1.4806072593196466514, 0.18790941046743775411, -4.2966299562970338144, 2.4291396004642109574, -0.053533997033188965298, -0.010285142358674243673, 0.036010718452201308237, 0.078556449331047495388, 0.1087649112987205402, 0.11135669107168463809, 0.050712070205624876218, -0.13841318917401507216, -0.40251565474696987534, -0.46517991977901451683, 0.20349314617610442246, 1.0143557123019779631, -1.1405536825935482792, 0.67930941230518215868, -0.044155658734731449699, -0.010551918280144357773, 0.026846493451829246679, 0.063500943176295510106, 0.093246643676239707177, 0.10217081207660241859, 0.054668166282408114931, -0.11377722912238073805, -0.35203871859195196015, -0.42111638639730120115, 0.15253118363616136044, 0.94649048025678306217, -1.0121821651109217388, 0.57451571404145840294, 0.016219257206193170989, -0.0071176512156830174655, -0.032070394058782113189, -0.046733128181885831565, -0.022955146147662620904, 0.064355193373090577302, 0.17955068437951066107, 0.16740171457655561582, -0.10106463330377371546, -0.53717929988271717612, -0.43131737815918519408, 0.92441911379647523361, 0.25648086130959080187, -0.43333916274627565235, 0.013425422879220978364, -0.0063849210108015046553, -0.027167274472213249342, -0.039477821323579707913, -0.020223997366574084089, 0.052944723043853468558, 0.15350356004883544681, 0.15078429944535920448, -0.067660238721064566714, -0.46623252629657147006, -0.44556008860331297594, 0.82312222547638547177, 0.29966087879693215079, -0.42396813736629684888, -0.0029870992584297519946, 0.01693717427502410694, 0.030718270730201148455, 0.028291335253372608405, -0.0063514844526883496023, -0.083389209010970452574, -0.16959947197935229535, -0.19227185037217231489, -0.042631081963612593722, 0.41365257431448715364, 1.194991019346654415, 0.40948464708943610457, -3.5729126174362551005, 1.9739047483119376292, -0.055378421806870380129, -0.090120486841041000381, -0.075792396200497824665, 0.014109538607311731748, 0.18747845403823751309, 0.40893281148942589587, 0.59953685763249897622, 0.59269902029136789512, 0.091306231290325681749, -1.2695764277946472998, -3.9686601152653779856, -10.639700121720473192, 27.158226734808724956, -12.874411011149250683, 0.0027717221308330777796, 0.011289516813916631449, 0.015603221500996657806, 0.010445387616706783285, -0.01216275485206363971, -0.056721556858614102792, -0.093836587061103690632, -0.084171516897678488878, -0.069871296122358192426, 0.047227185988430168961, 1.0402820008707061028, 0.92060251331009756992, -3.5240390182862912916, 1.7890318119787207429, 0.0081356187207008087703, -0.0083640097330792643332, -0.019673128760469865695, -0.020323129025726797969, -0.0028057937791017631413, 0.032806148709002308506, 0.06279408557921008216, 0.051855913928261443413, 0.047030505048261360412, -0.11962740630364049332, -0.51641259499730662874, 0.35021915017946464443, 0.57370306650817115557, -0.44641881344050349245, 0.0064302586766426077236, -0.0062442928977963062936, -0.014769736797531712869, -0.015079746682230301541, -0.0020919442347207967137, 0.022434636570765706387, 0.040026008788323998167, 0.034469533464809257173, 0.072746658144001094537, -0.051827194013211522194, -0.51712661675742777057, 0.25497276164947069166, 0.59286979981212906132, -0.42287026164964869324, 0.032302414440726047451, 0.029570540338776134942, 0.010317719540930900268, -0.031750880436700094911, -0.095084412373180479672, -0.16015482592195420808, -0.17900424759885391612, -0.12054949305303866436, 0.060324803398547996447, 0.62036677935752215785, 1.7368031287332044155, 0.46039238005048627178, -4.9770017795053895782, 2.5605262252224578745, 0.027101066769385319388, 0.027036672380992460979, 0.011593305358167237118, -0.025162507413962885428, -0.081765696243041835478, -0.14063192684400313559, -0.15989416906759143511, -0.11696496817681378833, 0.030540699463503587219, 0.53683688810602114838, 1.6021347804237084933, 0.48946088085711275673, -4.631883648747319171, 2.3877769191424600415, -0.011468265265055356192, 0.0030477994795636825093, 0.020208668423074567511, 0.03128999790200349268, 0.016229082428250371506, -0.043844605968627971271, -0.11986464132874237454, -0.081575764995684016401, 0.24158322371419319041, 0.34782847875475370047, -0.52850148134185270798, -0.26322587280691600586, 0.61155686423986632683, -0.22018411488138001086, -0.0062182025918357461403, 0.0032084138107567449633, 0.0092162519397888005568, 0.0091009402141341930875, 0.0028584550209630756566, -0.0084828806688914161482, -0.019860937470405640171, 0.00051781061566420262011, 0.11265092071588582445, 0.014335430746660084156, -0.46575194425244748064, 0.17627151304070354243, 0.41529189925667525651, -0.23670429581340823866, 0.096545481789657813798, 0.06798912081510403993, 0.015546384807852016083, -0.062092027064309891771, -0.15399983875338393857, -0.23549572839257992829, -0.27517309144406248977, -0.23775913065782558475, -0.077273424839119406182, 0.30004248311970505014, 0.90360785793737996574, 1.4619154631569213354, 1.0974348817975911885, -3.1291375704183077566, 0.0053934684159441187781, 0.0011400439038348046152, -0.0042651470301750147793, -0.009302993270073502477, -0.010004512497493017395, -0.0022456666465671439037, 0.0082787538032185227499, 0.0075038859365113390332, 0.0089065283591098273058, 0.065889631602973203206, 0.033836229620343834246, -0.210662356427669617, 0.058615639092722451875, 0.04759950649957691432, 0.0031294230246999210285, -0.00016799704382637658043, -0.0034988149556466202617, -0.0058391916949370320511, -0.0056507209443866493123, -0.0033679300132027712236, -0.00087907111325996383985, 0.012183894168738778827, 0.044773229202897731061, 0.067754813940246294557, -0.02715370505228624845, -0.2244709388314023446, 0.044689496728341000786, 0.10317763294811961583, 0.0018210788664371764089, -0.014913508445193870283, -0.022393764966952341144, -0.014173869632616666048, 0.012917632165798565921, 0.052059577477976369531, 0.080329151906395210303, 0.065262086662743995502, -0.015347096421182556888, -0.11493372071022900371, -0.19025820972932186437, -0.18261643676733832486, 0.31490399688078657414, 0.044180053735813643034, 0.013722208890640347559, 0.0062669650636479803246, -0.0052520467712108736061, -0.019269385230273553683, -0.029474565223724869673, -0.024286395848980402157, 0.0046824459388788691044, 0.046649996637465294957, 0.063277367488868363488, 0.030013273961871433232, -0.087672111844823416571, -0.13245192345007616996, 0.34282141048716124887, -0.21881262230806644564, 0.0059310347421041905575, 3.1990410592029008168e-05, -0.0060727163671684264434, -0.013162252681481497255, -0.021620846050477409567, -0.02396321233499151479, -4.1424260582692768243e-05, 0.073047580960954872076, 0.17804039358472767418, 0.17579693731311293869, -0.21895717876862902385, -0.62336454997847945414, 0.14326237026850660583, 0.35055924611564814874, -0.0019738790158336256414, -0.0021070954222007780461, -0.0022091067529639535176, -0.0024631260517078983237, -0.0029270317422004244606, -0.0021846163308198397399, 0.0076872195691326632161, 0.044651979088535583007, 0.088321118713781468501, 0.049879394278264585494, -0.15581763976773332914, -0.33988082729603730225, 0.1399811430645460375, 0.19544430782630139998, -0.050131188725237342541, -0.0006560673725390046146, 0.050035927519992473567, 0.090765113022780216645, 0.10053678428375632803, 0.05035448457946761075, -0.085144584184994709375, -0.2903116640857888453, -0.43618951995756516382, -0.22280643284470041943, 0.4260066620819223937, 0.9001046368991928448, 0.74787509063554169497, -1.3189472227776113389, -0.044491692734356561023, 0.010153860802744600622, 0.062707032381800226206, 0.10065465201161975795, 0.10268312220939654344, 0.038117477148930156705, -0.12306655167883852331, -0.36676370680808434432, -0.58009699877592901807, -0.5003841805544532173, 0.24996892541850265768, 1.8525738339327799675, 2.4291396005983294515, -3.342691293776219652, -0.052625451733898695461, -0.053533997085840598351, -0.029876537100529808322, 0.024420856347755318799, 0.10344364237626038627, 0.1866720430788939844, 0.24082313341353592251, 0.21447628077128785828, 0.013103645441540318767, -0.43112668019502697891, -0.94817326212372443628, -0.78881019583827194364, 0.6793094124309526638, 0.9679682881244938697, -0.039842903003973521858, -0.044155658676937561968, -0.028369347633975042222, 0.013414787772092765655, 0.078164344285154077618, 0.15169777268129949754, 0.20838444874674980078, 0.20204510002058498341, 0.039512157984243650999, -0.34969747035148912584, -0.82628272506930544594, -0.73299131966287056628, 0.57451571414767221846, 0.86102170011737666488, 0.018319560879320781477, 0.016219257257237256614, 0.0014738906174855998139, -0.030695751546710330071, -0.075438381687223818317, -0.10102377699152859103, -0.043352090707604200304, 0.14652064604788531121, 0.3766052606684918147, 0.36536398247159873609, -0.18688725445944451287, -0.94705452127782563387, -0.43333916287843887716, 0.91911171359078625009, 0.01668958029445579902, 0.013425422956105896935, -0.00012559481422378722002, -0.027492858295966340204, -0.064557337757640678721, -0.085591529095897359025, -0.037467450770455247577, 0.12322835978753095276, 0.32360036016508175738, 0.33030775162111875032, -0.13709867219413493511, -0.84969154907628685613, -0.42396813746860267935, 0.84187980057978550885, -0.02160848005626009663, -0.0029870993935337229741, 0.021689876983347039507, 0.049939193712994843499, 0.070792542468670976996, 0.058582405333619337606, -0.024370434599487691973, -0.19139613295396804005, -0.38313081596232517745, -0.42873955716448486353, -0.025787926799943950568, 1.1927450151831668901, 1.9739047482988150151, -2.359720287694676788, 0.028223490343441463152, -0.055378421924254690734, -0.13664604635671220856, -0.18570550631756810311, -0.14598390756302889737, 0.048550592280912949206, 0.42453349086112623922, 0.91795571034918521658, 1.3155679702757512306, 1.1796002092258557337, -0.17838617270419865335, -3.7380343785231953113, -12.874411011063072507, 13.877882910397428162, -0.010895895176245026445, 0.0027717221622086929932, 0.017369888123035765598, 0.030320447045021934468, 0.035297746746410417484, 0.019524646249294497907, -0.035009934889987930073, -0.12419171250940745643, -0.19752812376025580265, -0.22404090846558530337, -0.15654453467965803726, 0.68903188631199918657, 1.7890318118413686133, -1.891285968381931637, 0.020468559665260913083, 0.0081356186188613760502, -0.010134624970439091995, -0.032062670384208125096, -0.049040764488962391987, -0.04431026372368985361, 0.00043198575109434055042, 0.081409610392955300551, 0.15030156029716315591, 0.18107273276064209111, 0.055463222288655231185, -0.47174980129266236295, -0.44641881325382382073, 0.56491691729671356104, 0.014888788940251321269, 0.0064302585016805550652, -0.0069192299173070704113, -0.023462165960506216333, -0.036437454357199576538, -0.033069197504406351584, -0.00089569301654524730028, 0.054004747598140265263, 0.098202847769288600932, 0.14630796377969487732, 0.10237172636979746665, -0.36935065960976765087, -0.42287026139516997736, 0.47726604250430454401, 0.01347139354183231677, 0.032302414508801363668, 0.045111622447790648838, 0.041775944694089052478, 0.0059560041934001613229, -0.07805744581796095205, -0.20779811283891302942, -0.33462899381573879554, -0.37390142720142616195, -0.22790387774653783626, 0.35902159523412346953, 1.9005858213806650081, 2.5605262252864888772, -3.9167909501232274216, 0.0073268979112767074896, 0.027101066786762710581, 0.041956331805259879097, 0.042278516021130911617, 0.012468632467446266049, -0.062385029639194400364, -0.18019540375701070811, -0.29728001860938502166, -0.34155084860587170414, -0.23104895800658037386, 0.27470519813872523729, 1.6931849735677571012, 2.3877769190750210981, -3.53168639483490443, -0.011892575644368769128, -0.011468265312653061483, -0.0030148876400892662616, 0.017552900726485784477, 0.048011219839182987801, 0.067010155025815199004, 0.028455312110535004727, -0.10281788645965715678, -0.24753902684065837159, -0.10721517580935777114, 0.34586628187237389298, 0.22984406294403764148, -0.22018411497442713642, -0.03321211544817274236, -0.012760957448458502828, -0.0062182024567254958955, 0.0038933863541858095697, 0.015915222875446305428, 0.024249903395182424715, 0.02160193210518996168, 0.0024385434240652564819, -0.03269476000079479111, -0.055672085103251819915, 0.03171798674923893141, 0.13030534684831285297, -0.1759037860260967423, -0.23670429608819631984, 0.30538579464723364376, 0.32178076843402309004, -0.16928211360804573671, -0.47160030802439339981, 0.060407975272851549553, 0.28299364045446062654, 0.12721562503781178033, -0.045902435843216821132, -0.08902813811372906061, -0.044399103713923998571, 0.0032357705185295654315, 0.020260035159570748131, 0.014664853244462655243, 0.0029076016897410716586, -0.0069136194468836674118, -0.53387460952146104098, 0.95458796708087967353, -0.20586785263121695766, -0.36098211735813728884, 0.047767706459836432875, 0.11438558116617963956, 0.012263796494493117012, -0.030349313169579110461, -0.013867055690860356429, 0.0038298350084313764405, 0.0062297494706999665542, 0.0020528625070159230853, -0.00075430606581592687075, -0.00094720803631152383079, 0.0084609296198395866212, -0.19452951049129393879, 0.1797992811358258558, 0.15454207480729267865, -0.10693515485924318109, -0.075107850575308124363, 0.0056877965866618911128, 0.016474389727014093765, 0.0087077032781850485788, 0.0024740256481326581423, -0.001217672464207483372, -0.0018055865870299754941, -0.00053434058252151898742, 0.00076062654565866086569, 0.050335737331402799488, -0.3181172929676936989, -0.010357192191110061244, 0.30820144553057188253, 0.1427818173723321149, -0.033056694179867993078, -0.093456398913365545944, -0.076970351358275762577, -0.028459790736034627778, 0.0089376265865668373667, 0.021767941430802559277, 0.017009610445761491615, 0.0056533796668648836695, -0.0060117386894270041786, -0.34266966207148874535, 0.62729268118747028016, -0.20100474574243143433, -0.17852748241061835999, 0.06095543271170715538, 0.049112507650200545872, -0.0021882460007318607804, -0.013545960739865353997, -0.0073048211031748666242, -0.0013581102284775912765, 0.0017782548711070718384, 0.0023665134713174709388, 0.0013326744685138254467, -0.00044417407302852737793, -3.247628561080527998, 3.8685583993443861672, 1.1377177376815650334, -1.1586192526336911435, -0.86222495753252814676, -0.15456571400189889065, 0.16110388502818936973, 0.16124367626795732544, 0.066390779761873255915, 0.00053146655799413751887, -0.019031049905451336707, -0.014645934920786401329, -0.0041516229616226953125, 0.0054079265671590102602, 0.099279866291095864028, -0.2623175102206601883, 0.18495082343783839085, 0.10269956199294194565, -0.1587365883944867051, -0.030788403635692036114, 0.064937312556181903123, 0.025158845806416200303, -0.012820338997700275394, -0.01344966275314008787, -0.0028316543299647988288, 0.0023876097585952297739, 0.0021786626177359084792, 6.3793748286220752093e-07, 0.85067142361767711378, -0.34255068014623368811, -0.7888078627350334493, -0.18260162404479332587, 0.22915954780367828181, 0.24813272549732495742, 0.11022910525638757173, -0.0026623669354282337822, -0.039617615149018313436, -0.033392054338940865832, -0.017177891495129535282, -0.0047099562509640728256, 0.0023053492251584322033, 0.0053318472097277670729, 0.28444581623195658304, 0.15176672737553767445, -0.49036490599097148912, -0.18713155635465450466, 0.11074316487746228843, 0.11108599259690714323, 0.04216506624346824661, 0.011147924957252957484, 0.0077665997243518275495, 0.0026667708305777885688, -0.0070970547491058175171, -0.01254242912592933544, -0.0084667683132357094561, 0.0034291096034225060363, 0.36485592713895315109, -0.1125676556055118932, -0.49089850469649359832, -0.023675626573013072718, 0.21354348308189019523, 0.12268255871766441112, 0.004153385105568670882, -0.036985730507463829286, -0.026379727160410187253, -0.0084715175402428292278, 0.00088885752576006900327, 0.0028778329398848065748, 0.0019516672328667267981, 0.00042933820188646205787, 0.40929723828012165576, -0.16981925074312204105, -0.50130802273892960397, -0.0064484148738514924376, 0.2234536350254633652, 0.12537421501140444402, 0.0020458162705494910109, -0.040835325647304833441, -0.028300764349884734855, -0.0078981286403146076791, 0.002242754806966572572, 0.0036417297466466731076, 0.0018829116421624797918, -0.00016837004788808628041, 0.24077835149755991151, -0.324309167693530076, -0.106977259077721501, 0.1936926401332685066, 0.069510474164272709197, -0.032911881457466957146, -0.036689213425495241505, -0.014730737293423437281, 0.0028113285001330531182, 0.0080426584810228338163, 0.0060223209509301742909, 0.0020096415748118668862, -0.0012195595230907916532, -0.0029169036757997856939, 0.19709749814946272983, -0.30514594816191137339, -0.058297514433128867772, 0.19991691540839157115, 0.043963067769641060722, -0.048447660914852551384, -0.035274586826420963048, -0.0085162200395881189341, 0.0060618282476601181158, 0.007890847331606260992, 0.0046924966939814113931, 0.0011683036353614902926, -0.0011823443113228311162, -0.0022754038913852222062, 0.042805527465471095161, 0.30424861373711681312, -0.27619971913681179432, -0.20019681061154698254, -0.00153030555836571611, 0.064114183204430963414, 0.065325381725678477451, 0.040531488025057064994, 0.0089258857088865221213, -0.012528910511107882453, -0.016789695337783558382, -0.010556570764542118346, -0.0019429952210615406959, 0.005407650561320375554, -1.8912859682308331699, 1.7890318118364076927, 0.68903188598552045363, -0.15654453460538059195, -0.22404090859557698745, -0.19752812326365529194, -0.12419171277806946574, -0.035009934784837340449, 0.019524646196765142042, 0.035297746484403168254, 0.030320447248276028579, 0.017369888219568564713, 0.0027717221511632502209, -0.010895895247153980501, 5.1358544100098129093, -8.1327128834066275687, 1.4976481006262176709, 1.3762605586074350139, 0.48869887764633090921, -0.014127001731573487694, -0.14491421489433808922, -0.10615790586140688712, -0.04574472520514669821, -0.004560282053987058451, 0.012371811066713421165, 0.011057887906791890459, 0.0031382207299918136838, -0.0042240662465155964095, 0.048048473217267295698, -0.3069072785687284588, 0.090367382131855147898, 0.30661337636891239367, 0.018935520986407924915, -0.11349197446878034334, -0.071200491981894639681, -0.014696283288237872028, 0.012032266846298525845, 0.014578881833351927527, 0.0087719940656770557796, 0.0031635603561773642349, -0.0005065487572403521311, -0.0027231439802752931481, 0.047530875852208068211, -0.30918022517201365629, 0.10774640638412036342, 0.30358168697213983034, 0.0015099422425438342038, -0.11922878632898352425, -0.065186366035445794664, -0.0078340843258642382102, 0.014676171642970694056, 0.01380111681047884245, 0.0069207595284750747511, 0.0018508065833710029592, -0.00077769995339817558824, -0.0020551013650689114488, -0.90052810289734774862, 1.2948833980428839929, 0.10316817286029578149, -0.40957844781831836967, -0.21025195107396818517, -0.027961330106538017509, 0.066455290777990283502, 0.076375609102915575921, 0.032745059116518926423, -0.0087096140222969707645, -0.020800756963079251871, -0.012698881628637189584, -0.00089197676926008146944, 0.0069846339530625356495, -0.78326005455827230062, 1.1672425964617023642, 0.042931896357317804458, -0.37672378501558934083, -0.16543973077495177648, -0.0096171989400780869689, 0.05912615700460085999, 0.062010066075734851521, 0.025537601093294209215, -0.0077370641519086827756, -0.017174813240845372003, -0.010504321307680275846, -0.00084395269003827806317, 0.0057575753796976422705, -0.20614357722783729909, -0.059015283357543989184, 0.46722925041172491678, 0.16333952710162125133, -0.32917564184023212492, -0.20669977493721194972, 0.042300126948455252229, 0.1039451054108288669, 0.052880059563604198125, -0.0014496281056228159465, -0.020861592445768260828, -0.015483413381942923159, -0.0034185568301074004648, 0.0066374215310591785227, 0.22401265344544460389, -0.10838509073760645152, -0.29178578612869698894, 0.066121276479729595987, 0.12622959354677229515, 0.020104182929788390333, -0.01700627676169548172, -0.0085869529001941276997, -0.00099813541196795674791, -0.0011777871791518932081, -0.0015209236745772237043, -0.00047545398729651865838, 0.00028659661924367030306, 9.1681317393997171171e-05, -0.169282113524752309, -0.060919141204245005894, 0.73709815612233897664, -0.37584418409285075269, -0.3478092605644802493, 0.05195335261314383557, 0.1598741153638882917, 0.073974547033496845794, -0.022416710353202716488, -0.043750716995143407173, -0.020776594146527500861, 0.0028639243299629140935, 0.0098488236936473145394, 0.0029076012935203600362, 0.95458796716760430101, -2.1931340497384153032, 1.1759558786298902611, 0.54994853257207565456, -0.42294156252420717079, -0.1801727943215404304, 0.078139645291502801694, 0.064404320006773249641, -0.0041231642026238347204, -0.01883539532916070694, -0.0058387454006466740428, 0.0034575978731960247889, 0.0033176081829117596254, -0.00075430621648273803777, -0.19452951036358470671, 0.53599427169495361678, -0.39029357522150631699, -0.21279192995482046347, 0.3257591511700422271, 0.031398285779164744613, -0.091347312460948015822, -0.020990433783790778122, 0.005610107676367528419, 0.0055560048551374831977, 0.0041187285128509070267, 0.0014556063749904832403, -0.00089182235498445048584, -0.00053434075578646998521, -0.31811729362067414639, 0.51478720535270161562, 0.019714109406052192364, -0.43489881781737299571, 0.076815555683536471121, 0.15021192598774968374, 0.055829196067260937475, -0.0064392408068780857683, -0.039995599456506020219, -0.03434756563994123546, -0.012426435945582497283, 0.0049043496174320530892, 0.010208167421243488637, 0.0056533801755296493596, 0.62729268123580328442, -1.5164486142622239928, 0.97282769455342443798, 0.19875651658957727008, -0.29772607033794890041, -0.03833056703359485895, 0.04927545522827828095, 0.017437374789444969742, -0.0043626253894509100725, -0.0054094138137335354838, -0.0029345850434913992553, -0.00056775720322595451661, 0.00087959154825216305778, 0.0013326745120460657592, 3.8685583993461625241, -7.3209529596783795569, 1.4933912228360410879, 2.8559034538247689561, 0.10171691237041528832, -0.67772049044707249887, -0.40256089863932686956, -0.06513009598384315868, 0.077758452943008704583, 0.07225245235164809432, 0.027084362169588732661, -0.0014003245098102427325, -0.009184659968138152511, -0.0041516230989961820305, -0.26231751024319560628, 0.72535092127725175892, -0.56401922305016105597, -0.17623661817150929054, 0.41891135680042151934, -0.049324417396996646235, -0.14437885416487014623, 0.012609860025471729406, 0.043698466966786628229, 0.0089162262914121142438, -0.0091765213211876632549, -0.0065142839815102255782, -0.00037118525444451976611, 0.0021786626390391276506, -0.342550680222245163, 0.26850394739429600532, 0.74839510212213011808, -0.4143932506687564854, -0.42675834624119052263, -0.072061997929625379311, 0.11625972520906995011, 0.12230166806296303983, 0.047781614833036507861, -0.0039370252288981220576, -0.01900634782966544381, -0.014991048293355905413, -0.0062701530529169274483, 0.0023053494423185606245, 0.15176672745244534912, -0.60227156812566562394, 0.85139390698035544158, -0.13844249991706777947, -0.35508657293788081777, -0.042066722686631183303, 0.081277608969146306062, 0.055498134359982215225, 0.010769470196138998877, -0.0010136950109562105522, 0.0020608818044148802771, 0.0029206604124781911959, -0.0016763176045238024122, -0.0084667683141393199758, -0.11256765547305584707, -0.13224934704383525808, 0.72807904169874337885, -0.3303265963143079964, -0.32795854947688951597, 0.031240683960868074337, 0.11788861999939705127, 0.058409126719725708299, -0.0023139655274026955933, -0.018051529606800564842, -0.011707259326266688387, -0.0030782782607035345454, 0.0012537388826396925721, 0.001951667257767673554, -0.16981925064920841417, -0.018929937048907503894, 0.69420533169849618638, -0.36783473748711842388, -0.32278289133281673484, 0.037124425967786263414, 0.12164223127266363123, 0.061359429286387309754, -0.0027048307896231169739, -0.020137757858800786614, -0.012768864702840311554, -0.002760717523705009252, 0.0018782460327268803157, 0.0018829115628005755907, -0.32430916761251393776, 0.57374384881099438349, 0.024315392963907205698, -0.4458991731014317339, 0.061526343345395514972, 0.11990392509882800975, 0.01991347458344706195, -0.01431773723951886533, -0.016503176826481880807, -0.0058479152908116876358, 0.0014186344305668562547, 0.0034709489693020094923, 0.0021104365821440337556, -0.0012195596018426616957, -0.30514594812927064993, 0.56818826955332524253, -0.032397568766923373618, -0.42590307963388857049, 0.10351364217780793797, 0.12056291026651702603, 0.0033547091684471932495, -0.022314874360765273725, -0.015554895313856016703, -0.0027824913064264499701, 0.0030090386035095174379, 0.003360478236824933438, 0.0014871978827502489815, -0.0011823443098609978576, 0.30424861405622100063, -0.77597708712491431093, 0.63979571955979552822, 0.033807854246143742571, -0.19332498503239634036, -0.061590066505236969185, 0.0039111537475974142083, 0.021531564380386249841, 0.025539603681921946993, 0.019240467992814903886, 0.0040589963038765684844, -0.0062595882118225793997, -0.0073172418058053386239, -0.0019429954485324313438, 1.7890318117491115224, -3.5240390181783363133, 0.92060251365386358557, 1.0402820006904680561, 0.047227186364543524699, -0.069871296972167443173, -0.084171516300593621218, -0.093836587512052455073, -0.056721556422135374631, -0.012162754890639240757, 0.010445387577878752217, 0.015603221300750702571, 0.011289516812714896557, 0.0027717222620695764444, -8.132712883392770209, 18.819858359523504276, -11.734535578335487216, -0.54694700961426090391, 0.85853380178776972365, 0.61245674051202791954, 0.2147726725538278425, -0.010698760381122985469, -0.055325942998503477299, -0.042903395048701817305, -0.02159192612498998104, -0.0018317848970025036714, 0.0066013054590520403239, 0.0031382207811068980711, -0.30690727872997702885, 0.62891606932841759292, -0.16382923518543338059, -0.46704776480219650781, 0.21867310651566740365, 0.17948123697064777282, -0.018433876106404224055, -0.054762890503585300028, -0.031468147203917662402, -0.00479729684702358012, 0.0064643590744159675626, 0.006626808405547348832, 0.0031595932881569735798, -0.0005065485655750298459, -0.30918022524124705308, 0.65438703007482712515, -0.204211502639604775, -0.46160953954453165515, 0.24272242957127024288, 0.17623718041696687098, -0.02994183529334152466, -0.057768581887370523686, -0.028535913838364936057, -0.0016759161671673463226, 0.0076515193971231198339, 0.0062300874890883258667, 0.0023250930707524649493, -0.00077769983501556208885, 1.2948833979340861333, -2.6373398811465387226, 0.9804943674081904037, 0.72934231403374816249, -0.13842670736034293211, -0.16033867260224593521, -0.092991662926542303924, -0.03173479015592198943, 0.024973228884188833354, 0.038540528053732169, 0.016276170423082637895, -0.0053030619686347257868, -0.010218058209706004646, -0.00089197677586736433264, 1.1672425964123065434, -2.4162102962348801505, 0.96694636405554923453, 0.63756128701373915302, -0.16536632890645355798, -0.14653530199474532059, -0.069483786252949819584, -0.019601214530892523136, 0.022141321633408907188, 0.031059709172119288889, 0.012607286415819540792, -0.0045611390779342914775, -0.0083333860303256620389, -0.00084395251534831938409, -0.059015283143770844165, 0.51088634534528254427, -0.74090550114718178865, -0.1383256006929769466, 0.62445379531289002539, 0.057030568218385035983, -0.23485197354703424288, -0.10265179040989512449, 0.025516823807119611706, 0.051142861062870560718, 0.024119690053307281385, -0.0023984519830911985201, -0.010446678620728197182, -0.003418557067123550227, -0.10838509079128466617, -0.0067081837149894248118, 0.48349167855681007833, -0.35717141771182953081, -0.15868030413305755921, 0.10262796675035650018, 0.059347229579818874812, -0.0012417806532252287366, -0.010315103345528016734, -0.0015043853878760127859, 0.00050369128794691339635, -0.00097140780825146887806, -0.0011001134258030152453, 0.00028659662888357099715, -0.47160030804212477173, 0.73709815618013152516, -0.34407781619632415993, 0.49867339501617657938, -0.37129306277347706722, -0.29746666889953909951, 0.085871964748910470377, 0.16541511557339841354, 0.070574097958509440787, -0.02637588900214839216, -0.04679346538668981148, -0.023022182454445667271, 0.0028639237078899979112, 0.014664853641166814349, -0.20586785274688707004, 1.1759558788902884086, -1.9111953351316210981, 0.80605085831654121886, 0.52446933405768791925, -0.33339336083349813133, -0.15424850223177474784, 0.065266873606107334016, 0.055761209677212021651, -0.004375231715166842672, -0.017133287363519836277, -0.0052745276459520958404, 0.0034575972784494680978, 0.0020528628901755898405, 0.1797992808136667231, -0.39029357474018622209, 0.33226137896159202434, -0.095611346044732328786, -0.24223204899003225599, 0.26704996813212711348, 0.04537770393758219778, -0.086063653120114172457, -0.023993094676527632059, 0.0051532810087156905032, 0.0061292693063797016803, 0.0041401723288521045246, 0.0014556056619640651666, -0.0018055861339318199279, -0.010357191590299377054, 0.019714108137874438581, 0.092111384211324295879, 0.050975857547400686698, -0.41302196127783302737, 0.088349258156023277588, 0.18256298327688874616, 0.072281539624552568202, -0.003961662891241616441, -0.044749644928451688675, -0.040642765115641944773, -0.016806002515870804848, 0.004904350472102318205, 0.017009610070463782039, -0.20100474593593478656, 0.97282769487222986449, -1.5129180815521456882, 0.73374026798781610825, 0.251694010886380104, -0.25625856520848050435, -0.039707299737428654318, 0.045446709797437465084, 0.018635867831345416523, -0.0039420222319778847053, -0.0060786562617670351211, -0.0036972622674458963631, -0.00056775713679182299165, 0.002366513394844433401, 1.1377177376436411471, 1.4933912228882240125, -5.8523911653762548113, 1.325635139975452903, 2.7682148052720521036, 0.13825129422069734608, -0.6594468186371017504, -0.40643320347560851769, -0.076497662742134472924, 0.071098753964533037419, 0.07343215277854223122, 0.030206930137378160822, -0.0014003249115546341757, -0.014645934638086296684, 0.18495082355544772978, -0.56401922320551367562, 0.61314057210517236562, -0.15763140895552502085, -0.28992652692191128727, 0.30846379741721136059, -0.019520070616607290009, -0.12068087100900520836, 0.011316824982042996903, 0.039402010321887191713, 0.0081991036783081224643, -0.009026260442840763551, -0.0065142839704757492941, 0.0023876097241388939858, -0.7888078625926498999, 0.74839510178251422712, 0.16608249924455739666, 0.60962069383433470371, -0.46777417136911664253, -0.45491774635321424736, -0.08067849599593997223, 0.1113391994451700534, 0.12596101971462772551, 0.055450542912383590788, 0.0018892386073644778371, -0.017108530289224418985, -0.014991047730491617249, -0.004709956623660052806, -0.49036490608666788438, 0.85139390716272123072, -0.68343280430982022988, 0.68520956150964063003, -0.10541724294572793474, -0.33036270000224221555, -0.053088079533812469746, 0.058693908480816500695, 0.044808591929616996274, 0.012763854165315851111, 0.0070728640391169717017, 0.0088489605130481815282, 0.0029206603300232893486, -0.01254242914168515223, -0.49089850477440716281, 0.72807904187271332841, -0.35587350829686009712, 0.54916536052178333005, -0.30729490556519034605, -0.30767974067816666883, 0.032935470782898559894, 0.11441700643948968241, 0.059967604276173591682, 0.0003651337150357356065, -0.016813440380189871581, -0.011857814484656533163, -0.0030782781908647383409, 0.0028778328524538239822, -0.50130802274727537249, 0.69420533182710286724, -0.26014045871844254654, 0.52242409850299365015, -0.34844429864146675424, -0.30386076043858972762, 0.041346972150988939299, 0.1191935269389474833, 0.062353692165945440173, -0.00058713283943097327455, -0.019093540802679137558, -0.012959524402085240147, -0.0027607176362158467042, 0.003641729771545160893, -0.10697725915748113046, 0.024315393184728865267, 0.20429067969763561874, 0.12163819334045478104, -0.39929432275508286443, 0.048404244138935778596, 0.11950870218117554122, 0.020051409428187599165, -0.014563168047680544875, -0.018070586218779535748, -0.0072639865602989010795, 0.00094532403230657223014, 0.0034709488817942640848, 0.0020096415641014898565, -0.058297514439124675789, -0.032397568641954678503, 0.20295950400227719124, 0.094421471066080023271, -0.37833794619689914418, 0.084231972301873916931, 0.11856804267069996628, 0.0038298771562200769995, -0.022170853807838022953, -0.01683124174912388471, -0.0040578436860592270036, 0.0025770583096597437199, 0.0033604783463992626488, 0.0011683035032182982876, -0.27619971957654088079, 0.63979572040359944385, -0.76849182296458562469, 0.57688292126968510054, 0.058514817186992654219, -0.20408194682796995667, -0.087542743459625690927, -0.0086567746887949072881, 0.023851500097140775764, 0.030924635713895348016, 0.023738013938673606373, 0.0067015073995392209286, -0.0062595887518548309555, -0.010556570529490749352, 0.6890318860188218153, 0.92060251358327749305, -3.3103775391193139832, 0.69525358562802253015, 1.1275766566889446896, 0.15297658835820787671, -0.032190729639823147112, -0.075805062962728139708, -0.097339639204004499429, -0.06806960245362776718, -0.023547510003081625896, 0.0039288783876650764867, 0.015603221558535078015, 0.017369888050777743116, 1.4976481006229767079, -11.734535578290062219, 21.274977342509000522, -11.73043978116607633, -0.80539766700437731561, 0.7460726688967714848, 0.59279654605878351337, 0.23991181184288445927, 0.0067679014390964273989, -0.053635968017598067981, -0.045129521709686135078, -0.023273316324356176088, -0.001831784836188561031, 0.011057887812073769959, 0.090367382351210723268, -0.16382923557215450794, 0.16304867784198334713, 0.019673338954410275053, -0.39036802879405146927, 0.19737696055363654346, 0.17947714185278970667, -0.015010320813939465046, -0.053966586878072798916, -0.03356621562885835075, -0.0074661518554339830617, 0.0050885756533188379602, 0.0066268090261114628164, 0.0031635599918399394219, 0.10774640645020901186, -0.20421150290680517569, 0.1950741014001872442, 0.0027757224175162477017, -0.38631141298470761392, 0.21495052263322006469, 0.17406335584491416935, -0.02689181847112138285, -0.056920746735433076779, -0.030236408672812666032, -0.0038868055994604336703, 0.0065623164647331437907, 0.0062300878456219548757, 0.0018508064373127593345, 0.10316817301541758578, 0.98049436699722547495, -2.3365638405710211778, 0.85771056991324012841, 0.78611479152451124719, -0.12816480555984419154, -0.19319357657292898778, -0.10748077147103989037, -0.02728500986087528668, 0.030758004147131590145, 0.041476466832700113752, 0.017413907976054226329, -0.0053030618744244343091, -0.012698881512943080777, 0.042931896391646698352, 0.9669463638924917781, -2.1573094566800214977, 0.82927857792830506067, 0.68734750605404881796, -0.15145495630593006076, -0.17300696310399243938, -0.082260613998220474818, -0.016367408307055374139, 0.026847191951723175379, 0.033686645759626702989, 0.013727038979494489424, -0.0045611386041393350843, -0.010504321540111286049, 0.46722925024267847327, -0.74090550070723770748, 0.29241279988362051112, -0.18646262840037328146, -0.1125293661213879054, 0.50122067820589688569, 0.02650678220213704625, -0.23340193746764850702, -0.098763162654554495501, 0.028136632419884694684, 0.053629645542159425087, 0.026325955299115268854, -0.0023984525053472155985, -0.015483413108208651798, -0.2917857860198755926, 0.48349167848342400333, -0.31550406993085672314, 0.41999287436921045602, -0.27334583757455116482, -0.15705877446580773782, 0.084448706555101352156, 0.054719103955757936175, 0.0030163969681288151439, -0.0078479025557453459261, -0.0015882081670202450978, -4.4812207318938360699e-06, -0.00097140783094017114051, -0.00047545396689567846153, 0.060407974953575485633, -0.37584418374805539376, 0.49867339567836316272, -0.37571244209726889496, 0.58280062309488367411, -0.30286905244471795662, -0.29700979399008164972, 0.061730827429629660241, 0.14947891070167318439, 0.068046680017493851866, -0.024705727541482556953, -0.04679346607849069889, -0.020776594274562867892, 0.020260035456490362027, -0.36098211739977648094, 0.54994853229986362386, 0.80605085888921679249, -2.0192448511085943075, 0.84430966817234287536, 0.57430934373764452872, -0.32636818210583928845, -0.16553174229961456487, 0.058928575543110192636, 0.056734861546478915217, -0.0025631327285373221338, -0.017133287681964873977, -0.0058387446530686623949, 0.0062297490680327050408, 0.154542075130555151, -0.21279193058906528679, -0.095611345718472490196, 0.32195994138061434509, -0.13869737388712802262, -0.24939920528344872164, 0.2646149637109962427, 0.049064850806555100615, -0.082401001725646488838, -0.023051928445409191476, 0.0046805804835832439184, 0.006129268889142147278, 0.0041187293681499651671, -0.0012176728951072642894, 0.30820144642889080178, -0.43489881827176846141, 0.05097585610017622143, 0.12709138831254171431, 0.10108644847641173881, -0.38367377978132899408, 0.082755923271674663999, 0.17060745464931584281, 0.064495783123462116504, -0.0064370898911590571437, -0.045331953456351846421, -0.040642764030138694986, -0.012426435398055813569, 0.021767940639371936851, -0.1785274822466361988, 0.19875651613112205185, 0.73374026844338069431, -1.5269039094738843243, 0.74718057012496252867, 0.26884105346196868203, -0.25180273723492652138, -0.041290527891587928966, 0.043211345894535187206, 0.017724525305896354777, -0.0042531967054441656428, -0.0060786564235389313471, -0.0029345850440498197527, 0.0017782549336749870969, -1.1586192525525997876, 2.8559034537463388048, 1.3256351398686716525, -6.1003411466341690428, 1.0893169810892884275, 2.7281082528568934009, 0.20019607149296422821, -0.60930100845948309551, -0.39157054764547466474, -0.078536066046496669224, 0.067887604963856165718, 0.073432152714867290899, 0.027084362393361229721, -0.01903105004182606072, 0.10269956178135980496, -0.17623661786852812261, -0.15763140899086339197, 0.52276818321127382916, -0.24982860622192326616, -0.28111273788757007441, 0.33216686784776627839, -0.009863053879585383793, -0.12415446106638872159, 0.006975000689834955081, 0.038352507333470960138, 0.0081991034367502480668, -0.0091765210666067652256, -0.0028316544088577704115, -0.18260162395295881366, -0.41439325043350666578, 0.60962069307544697327, 0.05698068502985151329, 0.61760996039197180174, -0.42108153563813499609, -0.43982037987554178926, -0.085732907252075496451, 0.10358814019096065917, 0.12320237497897827683, 0.05582182650050493411, 0.0018892388326986039784, -0.019006348014288139414, -0.017177891437228972465, -0.18713155630560782128, -0.138442500042486899, 0.68520956154804524285, -0.71341644033150786175, 0.71114059157352638163, -0.083910041592315187664, -0.34030865278214900149, -0.066729582931252667821, 0.056076031799688438872, 0.051034160546275603165, 0.018771201727423070943, 0.0070728637167727526763, 0.0020608818838066273599, -0.0070970546774602920673, -0.023675626936565438013, -0.33032659596400226754, 0.54916536066750309963, -0.37687568990573899752, 0.59406467104656290878, -0.27109698749204919066, -0.30855805134009128832, 0.02200293828220001649, 0.10797568322120465467, 0.05934822180935117597, 0.001237038628483608238, -0.016813440421821628651, -0.011707259683260950664, 0.00088885780965681734402, -0.0064484151282684226691, -0.36783473725072030058, 0.52242409856151961201, -0.28176701509249235666, 0.57000906167131770719, -0.30930395441873820284, -0.30468159863169358426, 0.028980647941815901314, 0.11175819889191111745, 0.061593004089039143922, 0.00045069572090574577805, -0.019093541089185612353, -0.012768864946465303262, 0.0022427551038969638711, 0.19369264010263642639, -0.44589917327735495434, 0.12163819374260070871, 0.25285178305919514541, 0.14056433572351414996, -0.39974450439485742503, 0.03573616318579547757, 0.11389447696661428311, 0.019063526211157054424, -0.013913434188646498293, -0.01756220077890788614, -0.0072639868287172537284, 0.001418634342387692631, 0.006022321125904809766, 0.19991691534724870261, -0.42590307988344811019, 0.094421471638417422967, 0.25515409952780837566, 0.10489834168617974408, -0.38424360997351958424, 0.071528371537460902485, 0.11508435424614762987, 0.0042934611235747755648, -0.021309552539184182379, -0.016547414141166758711, -0.004057843888338118557, 0.0030090384121808982573, 0.0046924969157038903117, -0.20019681074849202118, 0.033807853917565772184, 0.57688292223941528558, -0.79811815315969680817, 0.55191055597308469327, 0.051001072651549385351, -0.20170372992947435686, -0.080177251923101752751, -0.0045790113830345251233, 0.024883630135953908186, 0.031113239630402668434, 0.023738013137599989211, 0.0040589963894550832452, -0.016789695067198736472, -0.15654453442497245907, 1.0402820004934696385, 0.69525358535875803057, -3.2859373261329540838, 0.69443419022309471611, 1.1259725308815042855, 0.15669992715453337495, -0.027737595721978123425, -0.075116496832085696833, -0.099707021767532599821, -0.071205435102095093747, -0.02354750961990612762, 0.010445387464398846189, 0.030320447188207162031, 1.3762605585674947406, -0.54694700964954690026, -11.730439781010513656, 21.955898398289495077, -11.574158332263015936, -0.86954102007951195041, 0.68652532583508640851, 0.56320212985110018611, 0.23040183216036447256, 0.0080397223305684918648, -0.051133405077994839583, -0.045129522007111580462, -0.02159192592630034338, 0.012371811060783902136, 0.30661337636337032686, -0.46704776456972146859, 0.019673338520021572962, 0.25594426066439029954, 0.029506826388752689649, -0.40639628886257089802, 0.17884003745236159988, 0.17673594328647879159, -0.012043537602169043788, -0.052244975817117483841, -0.033658949860175702007, -0.0074661512830658189124, 0.0064643583939078593653, 0.0087719943156606625112, 0.30358168687091280358, -0.46160953892930739473, 0.0027757216086155625523, 0.28794863148180971368, 0.0052750455152029886857, -0.40619284764853186287, 0.1965601108180399792, 0.17283916951461153477, -0.02289278113696215089, -0.054957336781804234926, -0.03037139601305353237, -0.0038868048119677809857, 0.0076515190174336134357, 0.0069207594745118250837, -0.40957844778752056092, 0.7293423145183408618, 0.85771056907389631174, -2.3669884264122895701, 0.7842375930527935024, 0.75054776585855464965, -0.11886976338838384226, -0.1698076485985735018, -0.094704785146836820142, -0.02646635680680242686, 0.028160315302466119858, 0.041476467714371967332, 0.01627617021532407407, -0.020800757194348089285, -0.37672378491422381375, 0.63756128711102211248, 0.8292785775699518247, -2.1909506246739400837, 0.76954384438885525821, 0.66228733721925181488, -0.1423581809058798997, -0.15432846067713126836, -0.07246460227022609446, -0.015761946341449558501, 0.024892763471967182393, 0.033686646320954652389, 0.01260728612668757069, -0.017174813274406917907, 0.16333952674085308598, -0.13832560057862025582, -0.18646262769892454902, 0.34672735919984432051, -0.31278642930726924254, -0.19076957680918937843, 0.49532831241679192713, 0.054787844360350314299, -0.21496506417036626013, -0.09613235160586444672, 0.026001321239058478724, 0.053629644943957464898, 0.024119689950990747668, -0.020861592145990631025, 0.066121276606939546894, -0.3571714180692335816, 0.41999287466188689466, -0.26085018453168068442, 0.44179278607620459107, -0.27461936845923834438, -0.16678565739015446812, 0.081867938968020692592, 0.055370999260667443098, 0.0033094409978561248505, -0.0081593347482270747739, -0.0015882081643259237692, 0.00050369159752839602444, -0.0015209238737707649232, 0.28299364017350536749, -0.34780926002365419869, -0.37129306375530390927, 0.5828006243151916399, -0.31225468703964570638, 0.59375219528621225873, -0.31767492114917289525, -0.31361664870136485295, 0.054344711867539530203, 0.14883201391675277003, 0.068046679219580227804, -0.026375888654113645304, -0.043750717334922091528, 0.0032357707773468082427, 0.047767706407378623945, -0.42294156207324595798, 0.52446933330723810585, 0.84430966846006694215, -2.0026820102135243218, 0.83569991120653774885, 0.56815403978107459348, -0.32759641238711506572, -0.1648125432059339146, 0.059633645746535089704, 0.056734861642594489051, -0.0043752314274436935465, -0.018835396060464817958, 0.0038298353909934478717, -0.10693515513231006964, 0.32575915181390074915, -0.24223204938075801329, -0.13869737404502172073, 0.32707818306086039462, -0.13918996587115203822, -0.25150137836530356283, 0.26670008586620619351, 0.051428766841342905825, -0.082083814050951872665, -0.023051928255861230815, 0.0051532813359659070118, 0.0055560042793874209882, 0.0024740258835471641737, 0.14278181749719962013, 0.076815555762487289226, -0.41302196117388179086, 0.10108644743025196888, 0.16718045325790867328, 0.1052497697990658343, -0.39689920661209632247, 0.069385831064789516009, 0.16458930491630716086, 0.063485985907592981659, -0.0064370890512324835386, -0.044749644832527905869, -0.034347565451738927556, 0.008937626210276541569, 0.060955432868129752477, -0.29772607021500202551, 0.25169401011526065792, 0.74718057060791154278, -1.5321678039358799239, 0.74446138935891104094, 0.27084461850184032849, -0.25212457038905822237, -0.042663828858411377565, 0.042453530709943863852, 0.017724525008025970169, -0.0039420218694723651795, -0.0054094136485490816948, -0.0013581104467106060521, -0.8622249575369129726, 0.10171691230594841204, 2.7682148055094830674, 1.0893169808535403398, -6.2278495128376363255, 1.0634782500984691911, 2.7542390779932626188, 0.22964183924578854268, -0.59504103085588699873, -0.38970027611128160716, -0.0785360661373065011, 0.071098753997835995033, 0.072252452362014968346, 0.00053146655534443121814, -0.1587365882458681432, 0.41891135657347600141, -0.28992652692785514379, -0.2498286061063781649, 0.54705449920988902246, -0.22496859263624857661, -0.27772181956581543094, 0.32916271200514074513, -0.011612602175287336376, -0.12493608267187089611, 0.0069750004933943603874, 0.039402010804780750186, 0.0089162257444896017566, -0.013449662533052122657, 0.22915954768312704615, -0.42675834630787784496, -0.46777417070086346751, 0.61760995984349942312, 0.097428997585278273585, 0.64610922035587103895, -0.40992642176460430958, -0.44154207678563156536, -0.090889216613726037108, 0.10058431422016646251, 0.12320237490589447393, 0.05545054317190477311, -0.0039370253757265655584, -0.033392054343374916181, 0.11074316479916522515, -0.35508657293785350628, -0.10541724277313241354, 0.71114059160337417254, -0.70675322112077287873, 0.70382539592080095581, -0.087996907747716845982, -0.33871130671270738288, -0.061131407338465809653, 0.060861986032679588066, 0.051034160079685884914, 0.012763854495051440638, -0.0010136950052891282974, 0.0026667707753455817399, 0.21354348321767877006, -0.32795854978398214508, -0.30729490532625658439, 0.59406467117502581576, -0.34747678539339788006, 0.60229891925405132369, -0.27238333197463993063, -0.31387550249124313018, 0.018262890697981482013, 0.10691372182100372656, 0.059348221701066294909, 0.00036513356672132277203, -0.018051529360595702789, -0.0084715176372690288026, 0.22345363513139199219, -0.32278289165890594603, -0.34844429836354373631, 0.5700090617997792819, -0.24795565873378039301, 0.57942299431174704161, -0.31183035185331892114, -0.31117163544012971244, 0.024897363379164691444, 0.11077773701594248457, 0.061593003714565483508, -0.00058713259063560008233, -0.020137757730854119465, -0.0078981287750618245358, 0.069510473940957689942, 0.061526343686455584048, -0.39929432289708649551, 0.14056433583914071384, 0.26120068383560846037, 0.13742519970212407898, -0.40491681718743083307, 0.031889024577011634842, 0.11313751652935601089, 0.019649859985257759482, -0.013913434222056479672, -0.018070586267514194417, -0.005847915466655651158, 0.0080426586545041412607, 0.043963067575885667049, 0.10351364273504108648, -0.37833794672200993237, 0.1048983418193077527, 0.25870596144077157019, 0.10086005859410680785, -0.38841464495157751458, 0.068981492786563125552, 0.11493843381995715036, 0.0048850990578814526208, -0.021309552625216270083, -0.016831241588051073371, -0.0027824914045690030823, 0.0078908473564153775587, -0.0015303056394339359554, -0.19332498477823437111, 0.058514816572469596778, 0.55191055673340783105, -0.8196236354545272107, 0.55179055746254346282, 0.063298621317227646288, -0.19365985904946675378, -0.078016152799927476758, -0.0048093269462522181268, 0.024883629792716417672, 0.030924636068465630873, 0.019240467683466929566, -0.012528910371302073162, -0.22404090854594685411, 0.047227186344175414723, 1.127576656895699081, 0.69443418980603177637, -3.3243718510450248615, 0.6626008431872942861, 1.114878253217157944, 0.15468362398517515599, -0.027322241810667072248, -0.075127374870464766476, -0.099707021697284042849, -0.068069602816624275032, -0.012162754862896599239, 0.035297746632729914973, 0.48869887761067154486, 0.85853380189199157702, -0.80539766707907645138, -11.574158332310771513, 22.120871903841678119, -11.533572878979191145, -0.89535155027385582294, 0.65607734930160788167, 0.55097807194485004523, 0.22989914179983658715, 0.0080397224053902534985, -0.053635967795671112002, -0.04290339537357917149, -0.0045602819358296778698, 0.018935521314131648074, 0.2186731059216392159, -0.39036802840845641338, 0.029506826090614980684, 0.25312454712067256057, 0.019483473996519676957, -0.41347595853430774149, 0.1763865963962931438, 0.17743873600223794562, -0.011128200165438768138, -0.052244975711840190014, -0.033566216180282285042, -0.004797295597155081491, 0.014578881151683851375, 0.001509942778573563054, 0.24272242826046011577, -0.3863114118403732089, 0.0052750450813350852758, 0.28183164364784069056, -0.00489681343545260786, -0.41219080627420612295, 0.19527542773889258743, 0.17411708204935549826, -0.021909762865851965807, -0.054957336201228244865, -0.03023640963988712474, -0.0016759154296144855231, 0.013801116606847425167, -0.2102519509029162903, -0.1384267080419684659, 0.78611479240075410324, 0.78423759262464343145, -2.4372776024042708265, 0.76937821379732151961, 0.77086507601054721217, -0.10083974397203702977, -0.16426161415374879171, -0.095855515037885710705, -0.026466356110858048345, 0.030758003323468501022, 0.038540528561448156308, -0.008709614169810989126, -0.16543973081967663918, -0.16536632897265082787, 0.6873475063797236384, 0.7695438440840280947, -2.2484083327650274953, 0.75748942556426113182, 0.67923147908237235804, -0.12763756982513682625, -0.14982157003838561526, -0.073353278225493784293, -0.015761945958966499387, 0.026847191357044081217, 0.031059709648743757948, -0.0077370643211856658314, -0.32917564184020042806, 0.62445379531280797991, -0.11252936638158483884, -0.31278642877345530504, 0.2845290885992069474, -0.32194105403976086555, -0.17572047110925334557, 0.51180892882743134198, 0.062949961414275557026, -0.21404914175079770011, -0.096132352065149556664, 0.028136632534084819734, 0.051142861136123297927, -0.0014496281221360563744, 0.12622959303976763468, -0.15868030324083559623, -0.27334583789159055955, 0.44179278600587490455, -0.2642952252373245714, 0.44022530970716300391, -0.27249936608512942415, -0.16724014849547069739, 0.080370382093992109862, 0.054593003104101534007, 0.0033094414578148656503, -0.0078479025696130116224, -0.0015043862619183324264, -0.0011777866364635930941, 0.12721562598722860327, 0.05195335087079835934, -0.29746666833255536178, -0.30286905276356179684, 0.59375219611832030786, -0.32491437370191639245, 0.58267055613117435531, -0.32170333918300614284, -0.31430336184270057931, 0.054344710725240059912, 0.14947891101349011778, 0.070574097981556643822, -0.022416709231521447787, -0.044399104452957187383, 0.11438558151744682156, -0.18017279512573167177, -0.33339336035325767282, 0.574309343961104668, 0.83569991081746031192, -2.0183482116582909249, 0.83126619079388130995, 0.57041465893181664271, -0.32592319396488816929, -0.16481254309632092925, 0.058928575448729744735, 0.055761209102239819857, -0.0041231632603230755976, -0.013867056091090632433, -0.075107850261065298159, 0.031398285044914112096, 0.26704996848064282977, -0.24939920472143228825, -0.13918996660871182924, 0.32731405844716798814, -0.13791721919885127634, -0.25072657788421814917, 0.26672729247431725685, 0.051428767465441477258, -0.082401002221148350602, -0.023993095069964859301, 0.00561010828886588514, 0.0087077030809861507427, -0.033056695772064007, 0.15021192781164621755, 0.088349259295914431278, -0.38367378054894524242, 0.10524976886778245055, 0.1586599777962196256, 0.095053336761378698361, -0.40310103772717953419, 0.06722054287046107246, 0.16458930584047182233, 0.06449578382178909608, -0.0039616639083505204855, -0.039995601118184335998, -0.0284597893201763881, 0.049112507615799903815, -0.038330567192390481435, -0.25625856468114743825, 0.26884105306122535461, 0.74446138940579431598, -1.5355616248504375676, 0.74272135072829614799, 0.27025356082992912921, -0.25253563078339907877, -0.042663828908011992846, 0.043211346236729494819, 0.018635867371983295265, -0.0043626253558254682319, -0.0073048209818937330506, -0.15456571401642232866, -0.67772049043844484473, 0.13825129425190418875, 2.7281082528154478872, 1.0634782501188830839, -6.2399121261514309822, 1.0612636697189723467, 2.7596288854259558576, 0.2342625713420158351, -0.59504103110732320392, -0.39157054726455081672, -0.076497663081151676123, 0.07775845326356455034, 0.066390779609005939887, -0.03078840362218423013, -0.049324417486577745717, 0.30846379761061470948, -0.28111273805791586744, -0.2249685925609553605, 0.55247158508256632992, -0.22949726525285221723, -0.28077696354090214692, 0.32874492297795038231, -0.011612602333035217106, -0.12415446076031526335, 0.01131682453946197299, 0.043698467370186670766, -0.012820339153813285066, 0.24813272525258514567, -0.072061997603569002968, -0.45491774614358343465, -0.42108153599612163642, 0.64610922049213848073, 0.11037833696492563296, 0.65092998041141481025, -0.40950225890795072248, -0.4430267293385866112, -0.090889216791454230382, 0.10358814071512946475, 0.12596101930640590805, 0.047781614632227846495, -0.039617614935224559902, 0.11108599258422177658, -0.04206672257799023984, -0.33036270003372297799, -0.083910041944569913053, 0.70382539632307739019, -0.71257322745191342683, 0.70795434683464486536, -0.079362496952431499375, -0.33302742222963399588, -0.061131407867093319475, 0.056076032422245090936, 0.044808591721667866137, 0.010769470302894549457, 0.0077665996371495899311, 0.12268255898899689538, 0.031240683953586208255, -0.30767974125754687931, -0.27109698751361577251, 0.6022989197439098108, -0.34745959534714188077, 0.60160142857960230423, -0.27297034450073831557, -0.31463720652934257238, 0.018262890687948361845, 0.10797568309191968305, 0.059967604547247727043, -0.0023139655720090680086, -0.026379727257020570141, 0.12537421497309336749, 0.037124426155034966068, -0.30386076048812182826, -0.30930395484213130119, 0.5794229947525852964, -0.24864850927333534414, 0.57795264621947095574, -0.3126402828020690694, -0.31187353893056801724, 0.024897363157822038904, 0.11175819899868436047, 0.062353692217511733076, -0.0027048306066666215342, -0.028300764538750057542, -0.032911881136962257977, 0.11990392472394288681, 0.048404243818643048758, -0.39974450394547456078, 0.1374251996232490336, 0.26021764015199044806, 0.13652028293365930933, -0.40549043798224004842, 0.031896811414623686687, 0.11313751660180698666, 0.019063525800237998697, -0.014563167718646326601, -0.016503176533488882888, 0.0028113282346926924543, -0.048447660523760267826, 0.1205629097168255115, 0.084231972082730310891, -0.38424360942952601183, 0.10086005838880235386, 0.25859407057584365974, 0.10072531163001086085, -0.3887697356325941489, 0.068953073695636302176, 0.11493843404975559452, 0.004293460814601094723, -0.022170853844657684856, -0.015554895102413363386, 0.0060618281854440195233, 0.064114183839170113544, -0.061590067478536431045, -0.20408194677019439456, 0.051001072793782194215, 0.5517905576468200568, -0.8104918768668620821, 0.55948654748602288933, 0.066756193749841555296, -0.19304937071411867877, -0.078016152709924846409, -0.0045790120208168494723, 0.023851500273965008103, 0.025539604838313679364, 0.0089258849231020204185, -0.197528123495908009, -0.069871296641828672946, 0.1529765885016330651, 1.1259725307267136607, 0.66260084306609645655, -3.3499754045965430826, 0.64779819019580631156, 1.1094708336974516971, 0.15375628014692535594, -0.027322242283923412615, -0.075116496255545769922, -0.09733963914011989349, -0.056721556652702086188, 0.019524646191844574616, -0.014127001648475131373, 0.61245674034674635333, 0.74607266875696631736, -0.86954101953959894722, -11.533572879380262322, 22.158736264028270568, -11.53007031295686069, -0.90325498374296575577, 0.6515977016397727084, 0.55097807236069484826, 0.23040183152867341199, 0.0067679018081738739551, -0.055325943091807223972, -0.045744725190056942377, -0.1134919750044933312, 0.17948123797417905867, 0.19737696016874667349, -0.40639628894279006266, 0.019483473953959561803, 0.25089329034414342701, 0.018785569660892086075, -0.41426274346169855711, 0.1761163228204014608, 0.17743873608649840534, -0.012043537514778015346, -0.053966586199393137835, -0.031468148990496713768, 0.012032267824307161727, -0.11922878689080355918, 0.17623718179358888181, 0.21495052160467950642, -0.40619284752368678415, -0.0048968133402543766175, 0.28066065661391187414, -0.004665533381723301759, -0.41255004349020096832, 0.19510262377311399229, 0.17411708240832601824, -0.0228927814745071255, -0.056920746009175654523, -0.028535914971197610945, 0.014676172182010357656, -0.027961330573405034777, -0.16033867168094229072, -0.12816480604777258367, 0.75054776580972037969, 0.76937821392315963731, -2.4290988150648393784, 0.77893294312656624623, 0.77452845440879913852, -0.10106381801427236078, -0.1642616137669944687, -0.094704785449167050437, -0.027285009621682294878, 0.024973228273920011766, 0.032745059517359315082, -0.0096171990776331220396, -0.14653530166814623947, -0.15145495665388547524, 0.66228733744420575924, 0.75748942548100839378, -2.2424223595928758357, 0.7650001445191634275, 0.68241295663114642789, -0.12763707652772024459, -0.14982156998795576053, -0.072464602478635148364, -0.016367407701099706552, 0.02214132072843516541, 0.025537601537038115418, -0.20669977450250939799, 0.057030567674257376387, 0.50122067795138680601, -0.19076957659468615769, -0.32194105382608084742, 0.30108283803429924941, -0.31009262193032799404, -0.1720536040730163696, 0.51229813650863786911, 0.062949961091106926281, -0.21496506445252139272, -0.098763162184609509797, 0.025516824444897313884, 0.052880058968866432145, 0.020104183212542627024, 0.1026279662080383237, -0.15705877436824924875, -0.27461936801097641681, 0.44022530932164621875, -0.26013670299227892313, 0.44296659876128907651, -0.27275223586994606917, -0.16825411221690370067, 0.080370382479558633015, 0.055370998731922464742, 0.0030163968598306851315, -0.010315102443973081861, -0.0009981359443141633609, -0.045902436029571135057, 0.15987411645927376846, 0.085871963657920102886, -0.29700979379661840429, -0.31767492134661801106, 0.58267055634547415544, -0.32714061133475619281, 0.58367588863520525333, -0.32170333950761997865, -0.31361664830383845226, 0.06173082688179010874, 0.1654151160915776908, 0.073974547475342142255, -0.089028138669875633848, 0.012263796138247804249, 0.078139646110759117259, -0.15424850271733300811, -0.32636818235256342957, 0.56815404023235716036, 0.831266190645133074, -2.0181452590850561535, 0.8324943578141270617, 0.57041465888346432056, -0.32759641262750838209, -0.16553174219746741302, 0.065266874249154768761, 0.064404319184899319617, -0.030349312875168017734, 0.0056877962586305268883, -0.09134731170290459612, 0.045377703589115192528, 0.26461496311613785792, -0.25150137745922462873, -0.1379172195969972925, 0.32821065111015573779, -0.13825457963511528225, -0.25072657771063394661, 0.26670008505770492757, 0.049064851478245560812, -0.086063652823212391518, -0.020990434447390581701, 0.016474389979980852416, -0.093456399577435830817, 0.055829196887138242411, 0.18256298337295567835, 0.082755923916176546418, -0.3968992080442868553, 0.095053337144615543397, 0.15405796590255468881, 0.09364715000006743062, -0.40310103800747293379, 0.069385832445177292316, 0.17060745418811965046, 0.072281539374036390955, -0.0064392419795741588293, -0.076970350389742717145, -0.0021882465002572883128, 0.049275455971287927814, -0.039707299739379468828, -0.25180273726626212216, 0.27084461806411852525, 0.74272135097232128143, -1.5372521958626581551, 0.7420975534526981221, 0.27025356057216609607, -0.25212456990070702956, -0.041290528134952465822, 0.045446710252154104426, 0.017437373839296707589, -0.01354596024497171404, 0.16110388483625764922, -0.40256089823915619519, -0.65944681903968871417, 0.20019607194174471321, 2.7542390776978664668, 1.0612636697251009998, -6.2478076773457074111, 1.058022437871019461, 2.7596288853383010853, 0.22964183968105197908, -0.60930100886244709102, -0.40643320315630893091, -0.0651300965018991318, 0.16124367656383176151, 0.064937312489792967107, -0.14437885386279522559, -0.019520071133607402869, 0.33216686822091312514, -0.27772181973730591942, -0.22949726508085249721, 0.54930238365301509162, -0.23034918007083154956, -0.28077696375154420627, 0.32916271222357773585, -0.0098630541862854293617, -0.12068087070722587073, 0.012609859904557086524, 0.025158845821150535582, 0.11022910550356336545, 0.11625972482610946734, -0.080678496243219183226, -0.43982037923868916307, -0.40992642228723186193, 0.65092998085774766981, 0.11255554661324264354, 0.65148297781494179937, -0.40950225924668598454, -0.44154207619261798268, -0.085732907779798112058, 0.11133919947052328681, 0.12230166826264517177, -0.0026623669891207499272, 0.042165066388083968785, 0.08127760872789227542, -0.053088079690844440506, -0.34030865230077256278, -0.087996908110403948555, 0.70795434700839177111, -0.70362025089097735009, 0.71448266640956070717, -0.079362497083385205587, -0.33871130614036770723, -0.066729583592765151501, 0.05869390876239245769, 0.055498134222649292413, 0.011147925050409158987, 0.0041533855539111156491, 0.11788861897619970498, 0.032935471191814280623, -0.30855805119463625141, -0.27238333191662800203, 0.6016014286184865334, -0.34671571862716416002, 0.60221341428791552719, -0.27297034456056662455, -0.31387550302289007886, 0.02200293848246073769, 0.11441700629521654731, 0.058409127540953467705, -0.036985730997648354201, 0.0020458167553616146174, 0.1216422305027010653, 0.041346971954924954806, -0.30468159809174061303, -0.31183035194650587929, 0.57795264636146481685, -0.24794658488544563113, 0.57873194599357469237, -0.31264028294224344151, -0.3111716355690097302, 0.02898064785944391708, 0.11919352694608273402, 0.061359429694301775848, -0.040835325870421240735, -0.036689213205058537892, 0.019913474290250192389, 0.11950870231700663759, 0.035736162674288710073, -0.40491681649182675873, 0.13652028276077207702, 0.26097705630254314224, 0.1371699666975687848, -0.40549043768011500122, 0.031889023978107229607, 0.11389447715923756155, 0.020051409458876495034, -0.014317736825348158919, -0.014730737618420634838, -0.035274586815471978274, 0.0033547090679719037028, 0.1185680429661630092, 0.0715283710627451913, -0.38841464447529949755, 0.10072531141444031333, 0.259414901368130002, 0.10117135596948353959, -0.38876973548372856238, 0.068981492318204690184, 0.11508435444829456795, 0.0038298773010078269313, -0.022314874207389978622, -0.008516220243637802334, 0.065325381730446982487, 0.0039111540385843865733, -0.087542743784898194304, -0.20170373055810603136, 0.06329862223899064555, 0.55948654735094127677, -0.80768915914903860465, 0.56008742742656558722, 0.066756194030779911208, -0.1936598600479410015, -0.080177251335159016588, -0.0086567742452649046325, 0.021531563944171654146, 0.040531488086711330021, -0.12419171285150792938, -0.084171516444621438668, -0.032190729533905448889, 0.15669992752686889848, 1.1148782528183294183, 0.64779819041784136857, -3.3603691630214296282, 0.64354103182942645756, 1.1094708336187448783, 0.1546836246604241516, -0.027737596290727077608, -0.075805063185695895078, -0.093836587354524811366, -0.035009934663849273662, -0.14491421480179511594, 0.21477267251683654359, 0.59279654619708432861, 0.68652532507326613498, -0.89535154944058581439, -11.530070313269622062, 22.16834950068839305, -11.526524051825076, -0.90325498355850220022, 0.65607734847892040797, 0.56320213074204006354, 0.23991181134458969271, -0.010698759954280065085, -0.1061579060929894458, -0.071200492072792276321, -0.018433876468657819231, 0.17947714271311640544, 0.17884003757219565922, -0.41347595923254465422, 0.018785569566901319516, 0.25126104247331904062, 0.018699854194850224626, -0.41426274350426478543, 0.17638659701118772349, 0.17673594329399655578, -0.015010321693787064337, -0.054762890012692008734, -0.014696283306638753541, -0.065186366129795128566, -0.029941835571430829716, 0.17406335668464895106, 0.1965601105499688861, -0.4121908065341725047, -0.0046655334741623200046, 0.28138097532340744511, -0.0047315681468741155324, -0.4125500430576390376, 0.19527542766532363094, 0.17283916986226977031, -0.026891819412506234865, -0.057768581205768929077, -0.0078340844845571563915, 0.066455290775459877062, -0.09299166314103551878, -0.19319357605193379634, -0.11886976337957705363, 0.77086507555731698105, 0.77893294311131189289, -2.4293756310468799242, 0.77655110033983176798, 0.77452845471363029883, -0.10083974390816970168, -0.16980764842509588064, -0.10748077190293653382, -0.031734790095448231484, 0.0763756091741733395, 0.059126157082417765687, -0.069483786678954537197, -0.17300696237392643173, -0.14235818110714629214, 0.67923147891977864177, 0.76500014425924434125, -2.2426786565162997178, 0.76320166851058146573, 0.68241295658591194506, -0.12763756960945832342, -0.15432846045412629277, -0.08226061452375554861, -0.019601214296892557898, 0.062010066046882326984, 0.042300127327604429561, -0.23485197399895177406, 0.026506782030739872019, 0.49532831228195156603, -0.17572047040258759187, -0.31009262224136463493, 0.30365088645238491605, -0.31127158558430129132, -0.17205360392769383782, 0.51180892812851352947, 0.054787844732429105687, -0.23340193739927467265, -0.1026517902383970865, 0.10394510524222977388, -0.017006276513061364952, 0.059347229397566744757, 0.084448706344016966074, -0.16678565781426954451, -0.27249936516436251699, 0.44296659852355707576, -0.25934555637688033736, 0.44273656628674701308, -0.27275223548894866887, -0.16724014929128480644, 0.081867939156040639426, 0.054719104382314212587, -0.0012417808554810055334, -0.0085869529168630855803, -0.089028138009789051233, 0.073974546249097272721, 0.1654151163783026679, 0.061730827155465774203, -0.31361664825148088953, -0.32170333954616170402, 0.58367588877147524862, -0.32714061159537882739, 0.58267055628503816589, -0.31767492110274209161, -0.29700979394995113969, 0.085871964781813220347, 0.15987411468624465605, -0.045902435308099745481, -0.030349312916581633326, 0.0644043192380841234, 0.065266874279808317905, -0.16553174228659625622, -0.32759641250758969599, 0.5704146587085057174, 0.83249435802263771489, -2.0181452593483610869, 0.83126619094027109913, 0.56815404003842417957, -0.32636818230496883464, -0.15424850263462747124, 0.078139645976078406231, 0.012263796197117678502, 0.016474390023776243341, -0.020990434621317059088, -0.086063652460767747776, 0.049064851001375690831, 0.26670008540702022648, -0.25072657782942370241, -0.13825457954349631984, 0.32821065084783423993, -0.13791721922094168917, -0.25150137774132713675, 0.26461496322096561551, 0.045377703578856322386, -0.091347311705891678923, 0.0056877962575462310407, -0.076970350466633835618, -0.0064392417708887383226, 0.072281539238445033368, 0.17060745402324767328, 0.069385832776708900038, -0.40310103824699489161, 0.093647150231030984457, 0.15405796557012921277, 0.095053337385786237923, -0.39689920807774103917, 0.082755923798954009185, 0.18256298339519022589, 0.055829197089868512482, -0.093456399715901958269, -0.013545960440559891638, 0.017437374187384818008, 0.045446710004372906089, -0.041290527849598974985, -0.25212456998876303649, 0.2702535602147572158, 0.74209755376260289061, -1.5372521961056164752, 0.74272135131030858624, 0.27084461793241876304, -0.25180273706928402166, -0.039707300127367478604, 0.04927545609424306694, -0.0021882464516665158481, 0.16124367649465617958, -0.06513009641387787485, -0.40643320312312514186, -0.60930100883868276718, 0.22964183956471506676, 2.7596288852854375939, 1.0580224380001008733, -6.2478076774416742012, 1.0612636698368209665, 2.75423907779088184, 0.20019607156381510893, -0.6594468186498843032, -0.40256089844712372727, 0.16110388488734181389, 0.02515884583789027712, 0.012609859825565028837, -0.12068087055957241183, -0.0098630543090187285798, 0.32916271227935856025, -0.28077696387470479911, -0.23034917978332816735, 0.5493023833595960248, -0.22949726493311434217, -0.27772181975190146641, 0.33216686803311223919, -0.019520070782714171731, -0.14437885411492129561, 0.064937312557151308234, -0.0026623670325825036544, 0.12230166847389767337, 0.11133919920473181464, -0.085732907648413458901, -0.44154207626789077068, -0.40950225929340078324, 0.65148297817373412855, 0.11255554639897867586, 0.65092998069712748599, -0.40992642222776853878, -0.43982037945243135368, -0.080678495867242988604, 0.11625972481219382093, 0.11022910540643499067, 0.011147924860177859524, 0.055498134568615381523, 0.058693908496786316442, -0.066729583163554082037, -0.33871130650047676092, -0.079362497187626640516, 0.71448266658906633975, -0.70362025098377312116, 0.70795434715267679948, -0.08799690819496380767, -0.34030865238550711549, -0.053088079642481487974, 0.081277608810573942488, 0.042165066343673736349, -0.03698573118244039698, 0.058409127927517526246, 0.11441700617583912269, 0.022002938317403075708, -0.31387550302359434884, -0.27297034430589983689, 0.60221341411091411366, -0.34671571869430362067, 0.60160142867092891716, -0.27238333167899420051, -0.3085580515504059429, 0.032935471094410521953, 0.11788861934452976588, 0.004153385429897236622, -0.040835326025923719684, 0.061359430061367882525, 0.11919352672158184392, 0.02898064782203881723, -0.3111716355636439113, -0.31264028270901617068, 0.57873194579243281144, -0.24794658497341792724, 0.57795264640132493206, -0.31183035167408446053, -0.30468159844639175482, 0.041346971928210088376, 0.12164223082155824207, 0.0020458166119729603687, -0.014730737473373394994, -0.014317737092698266249, 0.02005140951396567478, 0.1138944772983072895, 0.03188902394675140195, -0.40549043784922866873, 0.1371699668311330278, 0.26097705647883073787, 0.13652028255890943931, -0.4049168167121924844, 0.035736163106243515331, 0.11950870212879297072, 0.019913474260432190155, -0.036689213178019139394, -0.0085162201966253146812, -0.022314874287861177771, 0.0038298773467390920602, 0.11508435441054248871, 0.068981492363075713747, -0.38876973556784732988, 0.10117135604203179866, 0.25941490144697776365, 0.1007253113571449088, -0.38841464467165048857, 0.07152837138745696266, 0.11856804270861157069, 0.00335470921631041228, -0.035274586864857231105, 0.04053148800405308172, 0.021531564018790916304, -0.0086567742060535465137, -0.080177251308034200306, -0.19365985997057641455, 0.066756193741433364486, 0.56008742752342632887, -0.80768915911080529924, 0.55948654743115255883, 0.0632986221241426672, -0.2017037302772540508, -0.087542744149348594718, 0.0039111541144468877501, 0.065325381792445527673, -0.035009934643680740152, -0.093836587379976924006, -0.075805063246314308145, -0.027737596025237634678, 0.15468362447751135558, 1.1094708333905913822, 0.64354103199964951543, -3.360369162676909216, 0.64779819007742722725, 1.1148782527523961594, 0.15669992756186559846, -0.032190729409051106746, -0.084171516538485507963, -0.1241917128059326636, -0.10615790595602264679, -0.01069876031912216259, 0.239911811753403037, 0.56320213042126432423, 0.65607734854545940451, -0.90325498322858133182, -11.526524052175661339, 22.168349500759347848, -11.530070313233517609, -0.89535154951605944174, 0.68652532519680375955, 0.59279654610841725493, 0.21477267256001927875, -0.14491421481687685735, -0.014696283449452304457, -0.054762889801495719111, -0.015010321669198415512, 0.17673594321854196876, 0.17638659686663787918, -0.4142627431171128638, 0.018699853577214677752, 0.25126104294126877514, 0.018785569432192856382, -0.41347595912601697865, 0.17884003763041481094, 0.17947714230199965368, -0.018433876077986570408, -0.071200492195962486464, -0.0078340843173392293863, -0.057768581780677563664, -0.026891818867932681947, 0.17283916996800169352, 0.19527542740294537293, -0.41255004313523152559, -0.0047315683604634700687, 0.28138097603080303166, -0.0046655339755660787948, -0.41219080651269701665, 0.19656011067402290315, 0.17406335687602095263, -0.029941836023011935558, -0.065186365909800467189, 0.076375609245666983838, -0.031734790189853673281, -0.10748077195521470173, -0.16980764837228343178, -0.10083974390557029488, 0.77452845482376908581, 0.77655110006956928892, -2.4293756305897327152, 0.77893294259925704992, 0.77086507581227325936, -0.11886976333737837852, -0.19319357611555984522, -0.092991663086214509937, 0.066455290715055251316, 0.062010065930915708732, -0.019601213937269580667, -0.082260614796466000054, -0.15432846059056296029, -0.12763756966505473978, 0.68241295721365269333, 0.76320166797232313893, -2.2426786564667655632, 0.76500014435592167406, 0.67923147896410629443, -0.14235818103367939291, -0.1730069628561932138, -0.069483786088215260257, 0.059126156835304893855, 0.10394510513552258912, -0.10265179003993082918, -0.23340193737945427732, 0.054787844480503049283, 0.5118089281996537343, -0.17205360364980321064, -0.31127158586591741063, 0.30365088635598808109, -0.3100926220238669484, -0.17572047023469503091, 0.4953283120479332613, 0.02650678180270791065, -0.23485197363215096211, 0.042300127206450065243, -0.0085869528309261370508, -0.0012417808476595218383, 0.054719104175876359664, 0.081867939095031094787, -0.16724014902631775814, -0.27275223562016359669, 0.44273656648887138587, -0.25934555644285200993, 0.44296659838728874581, -0.27249936536905827733, -0.16678565745048545832, 0.084448706064821993378, 0.059347229754248796707, -0.017006276708648425389, -0.04439910435768857494, -0.022416709674387733098, 0.070574098570852172863, 0.14947891064620852375, 0.054344710773298166018, -0.31430336139286574326, -0.32170333996858452297, 0.58267055685798929687, -0.32491437398616579424, 0.5937521958833003044, -0.30286905209459374921, -0.29746666958779227619, 0.05195335198641966995, 0.12721562566550992379, -0.013867055965864207717, -0.0041231635652351321386, 0.055761209410640867612, 0.058928575194924059932, -0.16481254294960956397, -0.3259231938871154366, 0.57041465873901797679, 0.83126619098550758036, -2.0183482118185467336, 0.8356999108512047636, 0.57430934404909972368, -0.33339336046547435322, -0.1801727950426462721, 0.11438558148756408639, 0.0087077030361009614273, 0.0056101084302042115992, -0.023993095278299455403, -0.082401002017035665514, 0.051428767364077559998, 0.26672729235847886242, -0.25072657754099625382, -0.13791721960641800671, 0.32731405866606294586, -0.13918996646640785508, -0.24939920515411279078, 0.26704996892382881768, 0.031398284803182185954, -0.075107850201679718372, -0.028459789316731411157, -0.039995601312849647646, -0.0039616634616357988483, 0.064495783618947297167, 0.16458930572680302573, 0.067220542912927908064, -0.40310103793192297861, 0.095053337285274139279, 0.1586599773907390043, 0.10524976886068330983, -0.38367378024085696531, 0.088349259034670293844, 0.15021192784550443933, -0.033056695740037757292, -0.0073048208759025805639, -0.0043626254543898809399, 0.018635867368828957363, 0.043211345943478439435, -0.042663828606231336826, -0.25253563048214178188, 0.2702535603557146282, 0.74272135117993576259, -1.5355616253174677599, 0.74446138929366012515, 0.26884105355997034437, -0.25625856508006050749, -0.038330566780539719607, 0.049112507390735443691, 0.066390779754812653812, 0.077758452888976342665, -0.076497662605611083109, -0.39157054771652305414, -0.59504103100925753722, 0.23426257171098854859, 2.7596288849770416185, 1.0612636700478614848, -6.2399121262192416282, 1.0634782498254129468, 2.7281082531989548912, 0.13825129398413207871, -0.67772049029896475059, -0.15456571405091532045, -0.012820339040637205447, 0.043698466971460209196, 0.011316825090087818256, -0.12415446112127014044, -0.011612602278195487127, 0.32874492324238296925, -0.28077696408841712472, -0.22949726472776205144, 0.55247158483410796492, -0.22496859260280688275, -0.28111273780017209667, 0.3084637973544424594, -0.049324417389470250062, -0.030788403631299084834, -0.039617615025457832145, 0.047781614497283193821, 0.12596101988561753116, 0.10358814037699208555, -0.090889216786987858665, -0.44302672917940555397, -0.40950225963079045499, 0.65092998110697419811, 0.11037833684976849913, 0.64610922041943641414, -0.42108153562905259459, -0.45491774677692670137, -0.072061997230624882937, 0.24813272517830967101, 0.0077665997500617608637, 0.010769470000631872764, 0.04480859218998147625, 0.056076031848081353548, -0.061131407630364203998, -0.33302742185133404895, -0.079362497597912751157, 0.70795434750548069669, -0.71257322796428512746, 0.70382539630878304671, -0.083910041427254597468, -0.3303627006125059995, -0.042066722230328958132, 0.11108599247840483804, -0.026379727069038322318, -0.0023139658079514395517, 0.059967604490918639815, 0.10797568328290424455, 0.0182628905712386852, -0.31463720648805104663, -0.27297034461243363523, 0.60160142874626587606, -0.34745959541465964993, 0.60229891967176885093, -0.27109698740355586688, -0.30767974114662977048, 0.031240683761581358735, 0.12268255903454622358, -0.028300764413038376893, -0.0027048308801203881091, 0.062353692415847566943, 0.11175819902614747559, 0.024897363105340450384, -0.31187353910991649775, -0.31264028266896315422, 0.57795264637075949299, -0.24864850941725011313, 0.57942299458256696365, -0.30930395447484088089, -0.303860760566775856, 0.037124425926356383054, 0.12537421509786159102, 0.0028113280543127024298, -0.016503176155750004517, -0.014563168020339556027, 0.019063525962603057595, 0.1131375165625944923, 0.031896811334437988328, -0.40549043777508408937, 0.13652028255741657792, 0.26021764038069378211, 0.13742519988071716375, -0.39974450438741393477, 0.048404244066384438661, 0.11990392462142418495, -0.032911881097545427066, 0.0060618280280472329469, -0.015554894718178061824, -0.022170854242963455166, 0.0042934610432268233149, 0.11493843397584439792, 0.068953073792580060752, -0.38876973562647393345, 0.10072531121636990603, 0.25859407101071624213, 0.10086005852772803038, -0.38424360988751143919, 0.084231972303295862825, 0.12056290972762670477, -0.048447660545203746763, 0.0089258850604712527271, 0.025539604794289162693, 0.023851499893993297757, -0.0045790116943131183025, -0.078016152721276835202, -0.19304937085501205907, 0.066756194066596052927, 0.55948654736836234136, -0.81049187699832803133, 0.55179055750525063395, 0.051001072932354708789, -0.20408194658517395115, -0.061590067588968178125, 0.064114183813010594548, 0.019524646074809890589, -0.056721556471926615206, -0.097339638968933578367, -0.075116496704507512705, -0.027322242070814010961, 0.15375628030667459756, 1.1094708334024601104, 0.64779819027562413059, -3.3499754045991441131, 0.66260084324993573368, 1.1259725306483239216, 0.15297658831443722294, -0.0698712963761402156, -0.19752812362377111799, -0.045744725322484518226, -0.055325942733505412141, 0.0067679013236754698851, 0.23040183200279742981, 0.55097807224497463707, 0.6515977011948503872, -0.90325498302884854596, -11.53007031362736079, 22.158736264315464837, -11.533572879041901871, -0.86954102027764235139, 0.7460726695340657022, 0.61245673982862780527, -0.014127001494064468298, 0.012032267704894714783, -0.031468148510301795218, -0.053966586909979685771, -0.012043537124508642971, 0.17743873611088611941, 0.17611632255317127838, -0.4142627427284602537, 0.018785568704713349192, 0.25089329084842709738, 0.019483474047527093997, -0.40639628948167710876, 0.19737696075485744451, 0.17948123774628579885, -0.11349197499692720292, 0.014676172128871125058, -0.02853591466867767823, -0.056920746519928178131, -0.022892781175190744791, 0.17411708239907525142, 0.19510262355207749763, -0.41255004286790486479, -0.0046655340361904037993, 0.28066065683468255365, -0.0048968132606450609501, -0.40619284793254678556, 0.21495052224681115982, 0.17623718144296471322, -0.11922878684988630349, 0.032745059422250893943, 0.024973228443949932465, -0.027285009836456940863, -0.094704785104413102825, -0.16426161391853094651, -0.10106381841194464477, 0.7745284550957302061, 0.77893294255191558584, -2.4290988146436749417, 0.76937821373942372194, 0.75054776560773484029, -0.12816480561014248485, -0.1603386721109154589, -0.027961330382833150116, 0.02553760141124136393, 0.022141321101287601836, -0.016367408212338993567, -0.072464602049644652459, -0.14982157011262875468, -0.12763707676880711728, 0.68241295705219551238, 0.76500014410083660987, -2.2424223593224619222, 0.75748942560636278731, 0.66228733700062691891, -0.15145495627966559793, -0.14653530187852817357, -0.0096171989943315671495, 0.052880059034817322727, 0.025516824451232416465, -0.098763162366339182774, -0.21496506435492249154, 0.062949961098653833824, 0.51229813637741350441, -0.17205360373092101756, -0.31009262215653382588, 0.30108283809670760567, -0.32194105396431504929, -0.19076957655665965907, 0.50122067817052251559, 0.057030567432635875802, -0.2066997744207089982, -0.00099813608476088445422, -0.010315102266471386228, 0.0030163967934620150232, 0.055370999051994891638, 0.080370382213667351801, -0.16825411253305391779, -0.27275223556800853864, 0.44296659863871612473, -0.26013670279222195214, 0.44022530944912457995, -0.27461936829858335507, -0.15705877419647210069, 0.10262796591324963025, 0.020104183408193844473, 0.0032357700843955515373, -0.043750715843454884846, -0.026375889592911903186, 0.06804667976077310998, 0.14883201370256074481, 0.054344710843432113279, -0.313616647214136679, -0.31767492236088623603, 0.59375219616428753788, -0.31225468702467745752, 0.5828006233731121144, -0.3712930627178495091, -0.34780926063810541082, 0.28299364037344165723, 0.0038298353161582698867, -0.018835395728522788539, -0.0043752320243787756401, 0.056734862312445170918, 0.059633645075207221409, -0.16481254257820512898, -0.3275964128358220262, 0.56815404012338299733, 0.83569991066272752978, -2.0026820093621022778, 0.84430966762967152306, 0.52446933371621518027, -0.4229415621071325182, 0.047767706375425239562, 0.0024740259327378048837, 0.0055560041596153407137, 0.0051532813345946011652, -0.023051927886442280524, -0.08208381483587738614, 0.05142876800848739105, 0.26670008437405201551, -0.25150137676127315434, -0.13918996721179868747, 0.32707818372275765384, -0.13869737392353756467, -0.24223204985565405245, 0.32575915212918021835, -0.10693515520655787199, 0.0089376261773199856819, -0.034347565263822037174, -0.044749645409748066427, -0.0064370883043041683333, 0.063485985412002815531, 0.16458930521773559863, 0.069385830932885167455, -0.39689920664299643827, 0.10524976971934932313, 0.16718045369098735731, 0.10108644672720036584, -0.41302196051266110155, 0.076815555324114714875, 0.14278181765470965292, -0.00135811031140636038, -0.0054094137068705595503, -0.003942022323605651711, 0.017724525674560162591, 0.042453530158414784501, -0.042663828529473701856, -0.25212457017499095846, 0.27084461813728133439, 0.74446138924125426772, -1.5321678033242365213, 0.74718056949125299937, 0.25169401134140567677, -0.29772607082812102552, 0.060955432962781046224, 0.00053146650632598783394, 0.072252452677448633644, 0.071098753307527517986, -0.078536065517617312248, -0.38970027616088853684, -0.59504103132624242001, 0.2296418399360018181, 2.7542390774955980426, 1.0634782500667965266, -6.2278495124507031733, 1.0893169804467055517, 2.7682148057935136443, 0.10171691218941784907, -0.86222495752117422896, -0.013449662731951248992, 0.0089162264600358403371, 0.039402009686939014343, 0.006975001428125771584, -0.1249360829974675291, -0.011612602356974918022, 0.32916271239478911426, -0.27772181999961370602, -0.22496859241622668701, 0.54705449950651829916, -0.24982860669773734874, -0.28992652656546735734, 0.4189113565231814551, -0.15873658826673930911, -0.033392054398969285567, -0.003937025157207636912, 0.055450542645020764465, 0.12320237549847939951, 0.10058431418594059037, -0.090889217319272908036, -0.44154207585150756499, -0.40992642254487093867, 0.64610922094291345363, 0.097428997382818571804, 0.61760995963722908897, -0.46777417025454598454, -0.42675834673574386535, 0.22915954783885666979, 0.0026667707996679097457, -0.0010136947078604663028, 0.012763853685061976248, 0.051034160651545604215, 0.060861986109550854362, -0.061131408003201555268, -0.33871130550429401351, -0.087996908875306922848, 0.70382539653938380741, -0.70675322125831052666, 0.711140591087070284, -0.10541724191531125221, -0.35508657339521487728, 0.11074316485496740747, -0.0084715176684178484373, -0.018051529354873457195, 0.0003651338581190481658, 0.059348221020939208314, 0.10691372240179607134, 0.01826289041716584044, -0.31387550207102771216, -0.27238333271452908058, 0.60229891991075368285, -0.34747678569355389122, 0.59406467142126173631, -0.30729490557053484867, -0.3279585497099953284, 0.21354348321324750914, -0.0078981287938970426432, -0.020137757659474828098, -0.00058713263493078218791, 0.061593003541407551071, 0.11077773725878442557, 0.024897363255067000987, -0.31117163530410746342, -0.31183035212272558478, 0.57942299466783053763, -0.24795565892039603262, 0.57000906179685006947, -0.34844429844242119643, -0.32278289150428718379, 0.22345363506446236368, 0.0080426587388573913401, -0.0058479157414292239764, -0.018070585962403084429, -0.013913434310926836926, 0.019649859906973474177, 0.11313751679384692617, 0.031889023975779057229, -0.40491681649104910301, 0.13742519941675659934, 0.26120068351977432375, 0.14056433637233475831, -0.39929432321750651935, 0.061526343794548792043, 0.069510473914968312625, 0.0078908474701347760399, -0.0027824918124526518137, -0.016831241069161641949, -0.021309552799824209302, 0.0048850988528491446489, 0.11493843415103333072, 0.068981492354624057706, -0.38841464431200006802, 0.10086005795386078498, 0.25870596155680591899, 0.10489834212947052317, -0.37833794678562943092, 0.10351364252222307327, 0.043963067684009447811, -0.012528910336248359758, 0.019240467648555185931, 0.030924636197956306, 0.024883629588243402764, -0.0048093272622046993187, -0.078016151854463602899, -0.1936598599194221948, 0.063298621797470339123, 0.55179055709431534815, -0.8196236349798921017, 0.5519105562920874064, 0.05851481666334425874, -0.19332498456203206505, -0.0015303057765941989205, 0.035297746770080927481, -0.012162755108678985003, -0.068069602996879155499, -0.099707021336249146426, -0.075127374667534621566, -0.02732224244736040572, 0.15468362480606273612, 1.1148782525374452135, 0.66260084337777924723, -3.324371851157426061, 0.69443419001088835252, 1.1275766570005936185, 0.047227186086564267031, -0.22404090846171398344, -0.0045602819113334282231, -0.042903395571015495724, -0.053635967288825248112, 0.0080397218721367270933, 0.2298991418757226346, 0.55097807252267605449, 0.65607734811849638401, -0.89535154898632163256, -11.533572879540496814, 22.120871903390330715, -11.574158331217866191, -0.80539766821387492879, 0.85853380255555300504, 0.48869887744262835438, 0.014578881522516554461, -0.0047972966267980027624, -0.033566215021521905726, -0.05224497656008500468, -0.011128199661107483071, 0.17743873585068611898, 0.17638659628424449566, -0.41347595812324133835, 0.019483473773515148919, 0.25312454638736514578, 0.029506827268076029486, -0.3903680291239505773, 0.21867310616457846795, 0.018935521263933899339, 0.013801116691671659609, -0.0016759156447006557517, -0.030236409119277260493, -0.054957337267081428356, -0.02190976198385017823, 0.17411708213970697456, 0.195275427430438131, -0.41219080651689848915, -0.0048968129298549052664, 0.28183164300754165454, 0.0052750461012306421732, -0.38631141318113138006, 0.24272242932900334189, 0.0015099424194045965673, -0.0087096141443605747806, 0.038540528271660011439, 0.030758004136712022158, -0.026466357068640825045, -0.095855514539916239491, -0.16426161399377794958, -0.10083974448550631076, 0.77086507622834843367, 0.76937821400622585788, -2.4372776027103073559, 0.78423759317949315317, 0.78611479149459240201, -0.13842670738874643388, -0.21025195105680785645, -0.0077370641572174631007, 0.031059708948116441368, 0.026847192481618174786, -0.015761946777141312009, -0.073353277760734234003, -0.14982157046403948852, -0.12763756977707763074, 0.67923147935480754356, 0.75748942561147969421, -2.2484083332822124568, 0.76954384475200265925, 0.68734750596910387799, -0.16536632895945860278, -0.16543973074560899827, -0.0014496280689774729365, 0.051142860971236665801, 0.02813663281650064657, -0.09613235232841613187, -0.21404914159961035969, 0.062949961326251593907, 0.51180892868667848816, -0.1757204707104150776, -0.32194105435782949254, 0.28452908868352388971, -0.31278642872936579522, -0.11252936643383724957, 0.62445379533047717935, -0.32917564184341052691, -0.0011777867187859850529, -0.0015043863334821487143, -0.0078479021156578482321, 0.0033094411889635147461, 0.054593002847926487553, 0.080370382739575432773, -0.16724014953710977793, -0.272499365161216478, 0.44022530931010045441, -0.26429522503883495776, 0.44179278596309767835, -0.27334583796803940681, -0.1586803033141553354, 0.12622959312710113622, 0.020260036488693209528, -0.020776596596830777114, -0.046793464625454597727, -0.024705728185960312837, 0.068046680728265251936, 0.14947891113454603307, 0.061730826164098376307, -0.29700979300611662914, -0.30286905324769258785, 0.58280062321096515188, -0.37571244094202610464, 0.4986733945942354862, -0.37584418357505133823, 0.060407975003625130617, 0.0062297491308183463102, -0.0058387448987690104829, -0.017133287219086697217, -0.0025631333188588130423, 0.056734862109638274641, 0.05892857513728676222, -0.16553174207111803762, -0.32636818230774161664, 0.57430934411969802866, 0.8443096675726200484, -2.01924485046110469, 0.80605085848102775437, 0.54994853238884477964, -0.36098211738382779412, -0.0012176729475777614206, 0.0041187295447765205625, 0.0061292686357068580302, 0.004680580603930470894, -0.023051928201763327575, -0.082401002386961463841, 0.049064851785818965058, 0.26461496258651145963, -0.24939920435233067786, -0.13869737428720768646, 0.32195994128714078286, -0.095611345470407560487, -0.21279193073303367956, 0.15454207516000850742, 0.021767940683701327703, -0.012426435728169978739, -0.040642763393256023385, -0.04533195382496512521, -0.006437089890360828473, 0.064495783049538082943, 0.17060745459906553889, 0.082755923574305154111, -0.38367377988771200759, 0.10108644820556721144, 0.12709138888616508534, 0.050975855461765272325, -0.43489881791250684095, 0.30820144634969409703, 0.0017782547862852033423, -0.0029345850845614754147, -0.0060786557646844109268, -0.0042531973361978215459, 0.017724525531353285679, 0.043211345867144403121, -0.041290528273405922854, -0.25180273680702475181, 0.26884105353546933204, 0.74718056974556990024, -1.5269039087175646507, 0.73374026752532750528, 0.19875651650210360288, -0.17852748223569925301, -0.019031049997255852085, 0.027084362056491541326, 0.073432153445923506463, 0.067887604301924975303, -0.078536065819791084364, -0.39157054746501174236, -0.6093010089537594931, 0.2001960718876459322, 2.7281082529152556049, 1.0893169807018274753, -6.1003411459946494944, 1.3256351391394933792, 2.8559034541828660636, -1.1586192526552956394, -0.0028316542974358579164, -0.0091765214965126035845, 0.0081991042587676379871, 0.03835250642985219699, 0.0069750011297349890152, -0.12415446091944853935, -0.0098630543227989971727, 0.3321668683621828988, -0.28111273812216935886, -0.24982860655393887561, 0.5227681838500407574, -0.1576314095828808548, -0.17623661748776234259, 0.1026995616632664643, -0.017177891394185653556, -0.019006348236381413741, 0.0018892394673786590178, 0.055821825768285315839, 0.12320237502317943346, 0.10358814082311280924, -0.085732907983946532449, -0.43982037907214927008, -0.42108153616572030487, 0.61760996010895008457, 0.056980685612686550467, 0.60962069247905215086, -0.41439324983639180822, -0.18260162419233008646, -0.0070970546991236957771, 0.0020608816185164477788, 0.0070728645587160255889, 0.018771200981419232379, 0.051034160651457771696, 0.056076032228340740027, -0.06672958385333820075, -0.34030865182639297162, -0.083910041946334071317, 0.71114059126929440868, -0.71341643957323130909, 0.68520956071030492129, -0.13844249963690671845, -0.18713155634556463691, 0.00088885769078648904987, -0.011707259425165122968, -0.016813440608814955723, 0.0012370389351785901488, 0.059348221394270941897, 0.10797568352521008661, 0.022002937861732456415, -0.30855805075483955235, -0.27109698798030024136, 0.59406467125927497896, -0.37687569009156279831, 0.54916536092117074208, -0.33032659605107322909, -0.023675626948544019335, 0.0022427550632702444346, -0.012768864921569721138, -0.019093540943152957634, 0.00045069568099545542772, 0.061593003941067842044, 0.11175819886970655981, 0.028980647990031797867, -0.30468159841985303915, -0.30930395457784315427, 0.5700090615386488313, -0.28176701505260726144, 0.52242409864027394839, -0.36783473720620413205, -0.0064484151742277735775, 0.0060223211229531291994, 0.0014186345971498731666, -0.0072639873772166810295, -0.017562200528224568286, -0.013913434095230159393, 0.019063526095664945492, 0.11389447738309495251, 0.035736162554388016532, -0.39974450414451156988, 0.14056433604097753887, 0.25285178241966937973, 0.12163819423913033169, -0.44589917339098533722, 0.19369264007364075941, 0.004692496851530912616, 0.003009038804008175045, -0.0040578445563691951936, -0.016547413807175632289, -0.021309552495658128091, 0.0042934610491096435875, 0.11508435458910927507, 0.071528370859438986096, -0.38424360945353625008, 0.10489834168046067997, 0.25515409912856373831, 0.094421471972368026249, -0.42590307991013492961, 0.19991691529607980016, -0.016789695231753894167, 0.0040589966452416843465, 0.023738012993989782151, 0.031113239819806667863, 0.024883630228788543998, -0.0045790119850780443278, -0.080177251408090011164, -0.20170373021890819487, 0.051001072908257262528, 0.55191055583463133338, -0.79811815309886202652, 0.57688292233017346433, 0.033807853630305063908, -0.20019681058390365047, 0.030320447186854254662, 0.010445387558708154815, -0.023547509520727358256, -0.07120543549811089723, -0.099707021698077005767, -0.075116496398863405459, -0.027737596239262921755, 0.15669992751299369194, 1.125972530769979274, 0.69443419015082974433, -3.2859373261296447311, 0.69525358523974156721, 1.040282000844821475, -0.15654453461092787681, 0.012371810945377087823, -0.02159192540259869883, -0.045129522971797640662, -0.05113340420543455167, 0.0080397220663762907628, 0.23040183169578121181, 0.56320213092303328395, 0.68652532470352312099, -0.86954101957281815238, -11.57415833192311716, 21.955898397305325886, -11.730439779969850989, -0.54694701018864166908, 1.3762605586756271325, 0.0087719939557360775834, 0.0064643593732384577524, -0.0074661524554718566166, -0.033658948847063684118, -0.052244976441058653405, -0.012043537591137937204, 0.17673594365411518514, 0.17884003693836325599, -0.40639628865955912396, 0.029506827144404314256, 0.2559442594710854002, 0.019673339330959431492, -0.46704776500732458899, 0.30661337652226383632, 0.0069207592857700945058, 0.0076515195220497201484, -0.003886805556236071213, -0.030371395092677909433, -0.054957337384025452265, -0.022892781214696778652, 0.17283916975880170286, 0.19656011074005980133, -0.40619284782617154361, 0.0052750462797055384934, 0.28794863031169654066, 0.0027757227045661497788, -0.46160953965620465933, 0.30358168710561583836, -0.020800757227406596606, 0.016276170382538132642, 0.041476467255630360131, 0.028160315873827293509, -0.02646635709565778663, -0.094704785401266075584, -0.16980764797196193361, -0.11886976380490948957, 0.750547765855873239, 0.78423759340722087874, -2.3669884270776280211, 0.85771056978242876667, 0.72934231411707084458, -0.40957844770003748458, -0.017174813424529752209, 0.012607286663389951145, 0.033686645463202977147, 0.024892764185520027675, -0.015761946709701595004, -0.072464602153904322801, -0.15432846047504519826, -0.14235818125171612047, 0.66228733722692634256, 0.76954384483263282846, -2.1909506252294752571, 0.82927857806079507963, 0.63756128676977497083, -0.37672378481088503177, -0.020861592245437418636, 0.024119690208900361472, 0.053629644614230038979, 0.026001321610611933394, -0.096132351876413349445, -0.21496506416043287269, 0.054787844516169845188, 0.49532831224272227955, -0.19076957673163999485, -0.31278642926542649105, 0.34672735904876872404, -0.18646262743356450953, -0.13832560080432379501, 0.16333952681142577257, -0.0015209237669777154762, 0.00050369170060282034574, -0.001588208713166538652, -0.0081593344676213382105, 0.0033094411681612381781, 0.05537099893323856048, 0.08186793967414457418, -0.16678565803274553181, -0.27461936831721656116, 0.44179278601543164884, -0.26085018469179255263, 0.41999287506378590873, -0.35717141815285807782, 0.066121276520633168183, 0.014664852893275348578, 0.0028639256143811507929, -0.023022183970328048697, -0.046793464658206752882, -0.026375889863227679893, 0.070574098377189434239, 0.16541511574442216426, 0.085871964582409102129, -0.2974666687136698906, -0.37129306261362532426, 0.4986733939339154098, -0.34407781471706883059, 0.73709815540071199624, -0.47160030794458407399, 0.0020528628591697010118, 0.0034575972256965185198, -0.0052745274168499757528, -0.017133287625957813877, -0.00437523126760928395, 0.055761208800963350807, 0.065266874682685382947, -0.1542485032569747283, -0.33339335988563584268, 0.52446933326849154433, 0.80605085881997384778, -1.9111953353512793896, 1.175955878954163758, -0.20586785276004440637, -0.0018055860345050443044, 0.0014556052103628772977, 0.0041401733225711941783, 0.0061292679430074220245, 0.0051532823250856236996, -0.023993095763368364054, -0.086063652177814406463, 0.045377703040557999803, 0.26704996910145667455, -0.24223205003005640612, -0.095611345181084217426, 0.33226137846192299596, -0.39029357454496022006, 0.17979928077235912687, 0.017009610092336271969, 0.0049043507924350837265, -0.016806003047006613793, -0.040642765376098066721, -0.044749644194253951346, -0.0039616632985818335533, 0.072281540296708829407, 0.18256298223399980429, 0.088349258921611467654, -0.41302196168670435306, 0.050975857526630474759, 0.092111384411360963664, 0.019714108251407073263, -0.010357191744760444255, 0.002366513291917347276, -0.00056775685820072661902, -0.0036972625768953897366, -0.0060786560785991063799, -0.0039420219752769047883, 0.018635867010605499522, 0.045446710699111304843, -0.03970730056413650233, -0.2562585644522733519, 0.25169401051383100132, 0.73374026818037485498, -1.5129180817793856928, 0.97282769492079812501, -0.20100474589295524996, -0.014645934709476129965, -0.0014003247049338720092, 0.030206929786100719659, 0.073432153231115709913, 0.071098753676124171186, -0.076497662830083842933, -0.4064332031707142967, -0.6594468188801191344, 0.1382512942813599599, 2.7682148054526236614, 1.3256351393812297879, -5.8523911644778019436, 1.4933912222385283819, 1.1377177378253766626, 0.0023876097709144119066, -0.0065142840458560814787, -0.0090262605331306155348, 0.0081991039623590614843, 0.039402010162270517668, 0.011316824713364875485, -0.12068087035158346632, -0.019520071267295185935, 0.30846379767513160131, -0.28992652674585045469, -0.15763140947671722136, 0.6131405728648617881, -0.56401922382477265305, 0.18495082375957999643, -0.0047099565145404349506, -0.014991047538176012835, -0.017108531126465934685, 0.0018892392566503088836, 0.055450542665527458031, 0.12596101999751424527, 0.11133919956714073696, -0.08067849637802400331, -0.454917746645529808, -0.46777417042148189408, 0.6096206927562862754, 0.16608250025508042302, 0.74839510109947948724, -0.78880786238226074936, -0.01254242921167328384, 0.0029206606078409323712, 0.0088489598029225114556, 0.0070728650472633271609, 0.01276385352733061182, 0.044808591986023507681, 0.058693908732586182186, -0.053088079880940788891, -0.33036270006428308799, -0.1054172421516057917, 0.6852095604766875736, -0.68343280359307334582, 0.8513939068284116507, -0.49036490599726867412, 0.0028778328857646776236, -0.003078278260585909018, -0.01185781474223508887, -0.016813439807884234672, 0.00036513330028391591253, 0.059967604504843965985, 0.11441700620909024677, 0.032935471169376173695, -0.30767974108804890809, -0.30729490514424789849, 0.54916536009594896228, -0.35587350824811109229, 0.72807904196386630247, -0.49089850474633561772, 0.0036417297190897207279, -0.0027607174887952128442, -0.012959524803196970782, -0.019093540288007207484, -0.0005871332195227958739, 0.062353692765955696931, 0.11919352625016985114, 0.041346972536835682743, -0.30386076098649078547, -0.34844429780073082892, 0.52242409792268429403, -0.26014045846656497218, 0.69420533166927800295, -0.50130802268184582182, 0.0020096416279495456052, 0.0034709485406705800972, 0.00094532468495368295568, -0.0072639870243199136657, -0.018070586113964927366, -0.014563168176242872326, 0.020051409451590521127, 0.11950870245962859351, 0.048404243951939700252, -0.39929432302580863956, 0.12163819397938117894, 0.2042906791512392739, 0.024315393370717350296, -0.1069772591670342804, 0.0011683035602765578804, 0.0033604780682186439177, 0.0025770588476043652937, -0.0040578441733774553088, -0.016831241348399767627, -0.022170854313475218261, 0.0038298773405132939714, 0.11856804295131563587, 0.08423197222563032216, -0.37833794663118230917, 0.094421471789090688498, 0.20295950330304435294, -0.0323975682590100858, -0.058297514524203848663, -0.010556570477973842292, -0.0062595890831147193989, 0.0067015079501751362648, 0.023738013477986194544, 0.030924636239276583061, 0.023851499405690777811, -0.008656774100777411482, -0.087542743991166308004, -0.20408194621864170215, 0.058514816599592643642, 0.5768829219633000438, -0.76849182367794643955, 0.63979572071301893388, -0.27619971958956118785, 0.017369887971224465956, 0.01560322167725224779, 0.0039288781717569786628, -0.023547509450812985354, -0.068069603038858061961, -0.097339638925437135808, -0.075805063211156750724, -0.03219072930782787556, 0.15297658811495618014, 1.127576656963972912, 0.6952535853516744746, -3.3103775389016116826, 0.92060251324049890975, 0.68903188623907929422, 0.011057888052599905426, -0.001831785636710536562, -0.023273315074747238929, -0.045129522942297772836, -0.053635967334427804565, 0.0067679014538644275897, 0.23991181153495200196, 0.59279654623273103642, 0.74607266900266466791, -0.80539766739761486924, -11.730439780505864888, 21.274977341791892371, -11.73453557786182877, 1.4976481005191002449, 0.0031635600428085923312, 0.0066268087686009603116, 0.00508857616058478162, -0.0074661522073300378918, -0.033566215883831442002, -0.053966585874726170913, -0.015010322352295652273, 0.17947714322027782941, 0.19737695981310485194, -0.39036802860975045038, 0.019673339270236966242, 0.16304867720989144053, -0.16382923503353305494, 0.09036738215776111971, 0.0018508066906038371166, 0.0062300870655063378054, 0.0065623172876977014026, -0.0038868056822611643519, -0.030236409168147238002, -0.056920746222382682944, -0.026891819337456348421, 0.17406335727007668512, 0.21495052154729846206, -0.38631141279936870303, 0.0027757226624245514082, 0.19507410133180666523, -0.20421150309720809135, 0.10774640657792411425, -0.012698881386166389201, -0.0053030619962727825606, 0.017413907852343150223, 0.041476466982337624467, 0.030758003982486556427, -0.027285009306781932148, -0.10748077242505367501, -0.19319357547886026327, -0.12816480645274883066, 0.78611479170590781163, 0.85771057015063922879, -2.3365638405293540636, 0.98049436691242997099, 0.10316817297098707096, -0.010504321484526065594, -0.0045611386223457228706, 0.013727039000101215818, 0.033686645549830182556, 0.026847191875837624064, -0.01636740758823730657, -0.082260614843248244799, -0.17300696248775007668, -0.15145495665361016768, 0.68734750601657235158, 0.82927857824465434078, -2.1573094572872557606, 0.96694636465064942588, 0.042931896046067348494, -0.015483413146942664324, -0.0023984525774888181861, 0.026325955509874262445, 0.053629645360817580768, 0.028136632534179247672, -0.098763162437413384853, -0.23340193791120406042, 0.02650678225679955341, 0.50122067842769568724, -0.11252936617050240953, -0.18646262823931039621, 0.2924127993498393252, -0.74090550024043766442, 0.4672292501117679131, -0.00047545384793373557169, -0.00097140810485710096887, -4.4809085664904461585e-06, -0.0015882086763823572654, -0.0078479019003876489963, 0.0030163964609986400507, 0.054719104473906376995, 0.084448706044235086199, -0.15705877407463395001, -0.27334583818711155612, 0.4199928752252625741, -0.31550407076943265405, 0.48349167910698553907, -0.29178578624369799766, 0.0029076012877525536648, 0.0098488236560891860571, 0.0028639248424952887077, -0.020776594965425883993, -0.043750716589754296004, -0.022416710490788287896, 0.073974547358983233636, 0.15987411515147226249, 0.051953352732391329394, -0.34780926097771197592, -0.37584418358131077564, 0.7370981548147554907, -0.060919139446809128513, -0.16928211420613226146, -0.00075430631323684862056, 0.0033176085787699500368, 0.0034575971490914082437, -0.0058387444942686788155, -0.018835396393866620651, -0.0041231629240985169862, 0.064404318531439005979, 0.078139646892379718834, -0.18017279579884548646, -0.42294156143796107772, 0.54994853179168012858, 1.1759558792082580592, -2.1931340500284557393, 0.95458796723050920452, -0.00053434091385429281339, -0.00089182166557430140024, 0.0014556050391429254526, 0.0041187301319953817727, 0.0055560033170228631527, 0.0056101090964156283616, -0.020990435277177255102, -0.091347310753875782252, 0.031398284013253299762, 0.32575915273094735447, -0.21279193122774991975, -0.39029357429846595062, 0.53599427122804155221, -0.1945295102516083896, 0.0056533801537800828022, 0.010208167409419358421, 0.0049043496920122752955, -0.012426435856332455898, -0.034347565961964185433, -0.039995598941863136355, -0.0064392415523402118244, 0.055829196782186001069, 0.15021192537951794099, 0.07681555618533043861, -0.43489881796509710643, 0.019714109485108110426, 0.51478720509814523965, -0.31811729348313294574, 0.0013326747645025732811, 0.0008795911349895056729, -0.00056775702453749964192, -0.0029345851258513822382, -0.0054094140259568143647, -0.0043626246400802664097, 0.017437373928391296934, 0.04927545619680198663, -0.038330567969197405775, -0.29772606989266431343, 0.19875651608125399772, 0.97282769528297186046, -1.5164486145181896859, 0.62729268118221881423, -0.004151623142090698701, -0.0091846597525138840706, -0.0014003247298424517653, 0.027084362083339853094, 0.072252452429614144602, 0.077758453231229207603, -0.065130096366571535449, -0.4025608983848736333, -0.67772049064197259494, 0.10171691241548709572, 2.8559034540929602031, 1.4933912223505305672, -7.3209529592885775884, 3.8685583992196050929, 0.0021786625507225873717, -0.0003711850602108667874, -0.0065142841389748493836, -0.0091765212151090052445, 0.0089162261440208952223, 0.043698467262949049206, 0.012609859534842199347, -0.14437885371373793131, -0.049324417584102477063, 0.41891135666706963425, -0.17623661763104783717, -0.56401922375741375681, 0.72535092169181847321, -0.26231751033560413155, 0.0023053494329813629428, -0.0062701533670095777745, -0.014991047648535774095, -0.019006348289061537893, -0.0039370246956530168675, 0.047781614136511996183, 0.12230166832301993252, 0.11625972487297366043, -0.072061997167092287087, -0.4267583467964451982, -0.4143932499900246369, 0.74839510130822850531, 0.26850394768039137494, -0.34255068021036205739, -0.0084667682247743323437, -0.0016763179719567040081, 0.0029206610875099530203, 0.0020608809821771969552, -0.0010136943756508020112, 0.010769470031996162401, 0.055498134278249504348, 0.081277608874602613742, -0.042066722408129100796, -0.35508657333419035851, -0.13844249933927035845, 0.85139390654406943426, -0.60227156800607795084, 0.15176672743476887772, 0.0019516673219109547353, 0.0012537385290344402883, -0.0030782778047590423146, -0.01170725980316273708, -0.018051528835900446796, -0.0023139664354827945564, 0.058409127589070616859, 0.11788861926525776025, 0.031240684592589225421, -0.32795855024521147758, -0.33032659557363358571, 0.72807904140626866774, -0.13224934705513172184, -0.11256765547153241291, 0.001882911669406227631, 0.0018782457920042193271, -0.00276071718894102449, -0.012768865308007696507, -0.020137757099424613561, -0.0027048315634710488886, 0.061359430154938436608, 0.12164223038039691238, 0.037124426873290095052, -0.32278289220986333685, -0.36783473674684574739, 0.69420533132628514661, -0.018929937025693448216, -0.16981925063119665542, -0.0012195597299865560629, 0.0021104368133160766353, 0.0034709488385815234074, 0.0014186346209515568809, -0.0058479155843512448804, -0.016503176639618984123, -0.01431773734008155545, 0.01991347451657924858, 0.11990392519294529317, 0.061526343655228146745, -0.44589917357169811529, 0.024315393226594952253, 0.57374384861558469506, -0.32430916750053107034, -0.0011823443097064090593, 0.0014871978272920835778, 0.0033604782808531463711, 0.0030090388019036835507, -0.002782491635343578313, -0.015554895082941545131, -0.022314874632217467781, 0.0033547095063488161643, 0.12056290994259755145, 0.10351364265288538757, -0.42590308020620143425, -0.032397568318651551178, 0.56818826924732523942, -0.30514594800409078346, -0.0019429953762934019153, -0.0073172417052549294791, -0.0062595886865245168804, 0.0040589967506657938928, 0.019240467359324105873, 0.025539604789476633845, 0.021531563435257403127, 0.0039111545253633916985, -0.061590067382682710251, -0.19332498424714666951, 0.033807853261874992734, 0.63979572051478650341, -0.77597708737026904746, 0.30424861396981661699, 0.0027717222287493002188, 0.011289516684871188715, 0.015603221696697765902, 0.010445387223047886138, -0.012162754700905643868, -0.056721556657861209316, -0.093836587283063432885, -0.084171516599291271166, -0.069871296433684171312, 0.047227186026399546603, 1.0402820008658575368, 0.92060251343522347955, -3.5240390181655025792, 1.7890318118086014909, 0.0031382206663868347611, 0.0066013058362926445299, -0.001831785603248258041, -0.021591925214901871999, -0.042903395687539924719, -0.055325942855606581594, -0.010698760410892974027, 0.21477267280822395779, 0.61245674007953765372, 0.85853380220152675228, -0.54694700993218303697, -11.734535578030328651, 18.819858359262926939, -8.1327128832961985694, -0.00050654833859004847092, 0.0031595928758546241999, 0.0066268087375189310001, 0.0064643585601176830474, -0.0047972958585836469186, -0.031468148692910431763, -0.054762888824651606567, -0.018433877347123241164, 0.179481237789174769, 0.21867310567857004, -0.46704776428049299408, -0.16382923501595822446, 0.62891606897091156814, -0.30690727861102234941, -0.00077770004011509146283, 0.0023250936482414000933, 0.0062300870044450680496, 0.0076515192769552771435, -0.0016759156467268483334, -0.028535914514979790479, -0.057768580760961515952, -0.029941836822760213577, 0.17623718174448682605, 0.24272242870922866587, -0.46160953899714946003, -0.2042115031471439246, 0.65438703054585978247, -0.30918022542696710353, -0.00089197686185623129267, -0.010218057914730226468, -0.005303062204556893873, 0.016276170394301282124, 0.038540528252019194688, 0.024973228502131003537, -0.031734789488673205426, -0.092991663935054702517, -0.16033867167786278185, -0.13842670776853072123, 0.72934231420174311378, 0.98049436711689352819, -2.6373398808936991955, 1.2948833978755316387, -0.00084395235511693252311, -0.0083333865030569519161, -0.0045611386719878604934, 0.012607286385866167677, 0.031059709410593338424, 0.022141320880828360695, -0.01960121374974160488, -0.069483786720992479768, -0.14653530182489568623, -0.16536632900809430935, 0.63756128690723268271, 0.96694636463865391018, -2.4162102968822054549, 1.167242596657626974, -0.0034185569326301160423, -0.010446678815432977258, -0.002398452039436934323, 0.024119690158957970971, 0.051142861135512723836, 0.025516823607176629396, -0.10265179008280474193, -0.23485197366231472249, 0.057030568121812522109, 0.62445379521445210091, -0.13832560071751207054, -0.74090550079323391763, 0.5108863451123014654, -0.059015283118804412466, 0.00028659644997542006208, -0.0011001132158452509632, -0.00097140793031176076496, 0.00050369190154019916929, -0.0015043864082701495933, -0.01031510237252794733, -0.0012417816737417097361, 0.059347230321365448957, 0.1026279662066173215, -0.15868030311574843338, -0.35717141875091307224, 0.4834916792352714765, -0.0067081843194007088546, -0.10838509049684644236, -0.0069136192854187868834, 0.0029076012565343060598, 0.014664853309181997976, 0.020260035542677488907, 0.0032357704648354884691, -0.044399103845150923731, -0.089028138230698966593, -0.045902435751966619293, 0.12721562497544708359, 0.28299364065673310309, 0.060407975210774449148, -0.4716003075675496703, -0.16928211455778602312, 0.32178076887556489316, -0.00094720795785696110594, -0.00075430630496576580018, 0.0020528628749013986404, 0.0062297490100792166784, 0.0038298355050442080272, -0.013867056178654104531, -0.030349312600524182282, 0.012263795809717727842, 0.11438558176577300063, 0.047767706106645796416, -0.36098211711113892264, -0.20586785288760456325, 0.95458796724631467256, -0.53387460956343968377, 0.00076062662726540078602, -0.00053434092305504278433, -0.0018055859897910346828, -0.0012176730927895690598, 0.0024740261799942289862, 0.0087077027990693431359, 0.016474390276321432963, 0.0056877958826141005594, -0.075107849834542730894, -0.106935155489835787, 0.15454207536810529722, 0.17979928065095421297, -0.19452951021817949107, 0.0084609295526155668676, -0.0060117386944921183053, 0.005653379586914694839, 0.017009610541512284326, 0.021767941519263773326, 0.0089376265188552163343, -0.028459790942951139481, -0.076970351100393782917, -0.093456399017751490277, -0.033056694055723236103, 0.14278181722913654839, 0.30820144556193412866, -0.010357192246917510184, -0.31811729288202622579, 0.0503357373103500913, -0.00044417419067838408587, 0.0013326746627325182244, 0.0023665133596357727612, 0.0017782549511303369794, -0.0013581102070340050035, -0.0073048212960738247157, -0.013545960442910098473, -0.0021882464133887898072, 0.049112507984078936085, 0.060955432606610231161, -0.17852748222355868668, -0.20100474605399551442, 0.62729268131202498004, -0.34266966205936433276, 0.0054079266264364005481, -0.0041516231648258457992, -0.014645934713363374657, -0.019031049940596338199, 0.00053146658484269908729, 0.066390779578905614877, 0.16124367647774531798, 0.16110388489055046946, -0.15456571389101961267, -0.86222495758338113525, -1.158619252684683909, 1.1377177377741647391, 3.8685583992676595422, -3.2476285610496753442, 6.3797132254002915669e-07, 0.0021786625544947639532, 0.0023876098059078215523, -0.0028316543953947129197, -0.01344966264703567789, -0.012820339117511668839, 0.025158845942791132483, 0.06493731243266584463, -0.030788403575647899785, -0.15873658832130563279, 0.10269956172516638004, 0.18495082372517199221, -0.26231751032760319831, 0.099279866294058008447, 0.005331847159962961949, 0.0023053494062213850287, -0.0047099564541412492977, -0.017177891325256627875, -0.033392054627731024286, -0.039617614861323126318, -0.0026623670800830076186, 0.11022910555016099399, 0.24813272515358619774, 0.2291595478467561009, -0.1826016241621624403, -0.7888078624975735087, -0.34255068014556194766, 0.85067142355403324583, 0.0034291095714748275926, -0.0084667681241436063694, -0.012542429440011514108, -0.0070970544775057113965, 0.0026667706299410365647, 0.007766599771126514104, 0.011147925024322068166, 0.042165066336012191339, 0.11108599246008149519, 0.11074316487275631671, -0.1871315564352096783, -0.49036490587199121904, 0.15176672738642740801, 0.28444581620650105647, 0.00042933817205187169326, 0.0019516674685685570304, 0.0028778327429416156527, 0.00088885758447241063236, -0.0084715178686263336444, -0.026379726633148685433, -0.036985730983246888071, 0.0041533853882092207355, 0.12268255850387893569, 0.21354348343118606524, -0.023675626895910531489, -0.49089850454294647886, -0.11256765562408979631, 0.36485592712419340211, -0.00016837008919267287604, 0.0018829117365831919261, 0.0036417296568884106531, 0.0022427550381408571482, -0.0078981290154467951453, -0.028300764032464936076, -0.040835325997406392295, 0.0020458167231746722699, 0.12537421460708589316, 0.22345363533503578846, -0.0064484151689305413066, -0.5013080225930610645, -0.16981925070581108117, 0.40929723825125108316, -0.0029169036147657688776, -0.0012195595732043657509, 0.0020096414655379913478, 0.0060223209731735047323, 0.0080426586441196424443, 0.0028113284251432006829, -0.014730737282214105208, -0.036689213411042406743, -0.032911881472501160195, 0.069510474001354766171, 0.19369264029223895451, -0.10697725910307258335, -0.32430916761585276697, 0.24077835142519662948, -0.0022754039121130240596, -0.0011823441962891662309, 0.0011683034592179504345, 0.0046924967139295014479, 0.0078908473907016423582, 0.0060618283116395223556, -0.0085162199704115933152, -0.035274587103062048088, -0.04844766071457265455, 0.043963067617567380685, 0.19991691555883725484, -0.058297514506512194965, -0.30514594809203643422, 0.19709749809910337426, 0.0054076505188743904592, -0.0019429951953362756788, -0.010556570647436813726, -0.016789695485086773147, -0.01252891028187391384, 0.008925885279299904107, 0.040531488392251333042, 0.065325381434588464247, 0.064114183537748989372, -0.0015303058746232720607, -0.20019681017880108786, -0.27619971954483879539, 0.30424861382498197271, 0.042805527508485145816, -0.010895895187843736285, 0.0027717221774568221414, 0.017369888017036212713, 0.030320447317104263357, 0.035297746590992627735, 0.019524646182509309417, -0.035009934771704942502, -0.12419171274603130195, -0.1975281235381092515, -0.22404090845483762284, -0.15654453462017339782, 0.68903188607236731666, 1.7890318119246668704, -1.89128596834454199, -0.0042240662491215188318, 0.0031382207188832486611, 0.011057888030734753376, 0.012371810807357549317, -0.0045602818500679216662, -0.045744725240171132952, -0.10615790587554510549, -0.14491421497507372962, -0.014127001553690673313, 0.48869887748428941787, 1.3762605586818079662, 1.4976481005796822288, -8.1327128833407318353, 5.1358544099752547751, -0.0027231441178996469016, -0.00050654848200177757765, 0.003163560057003717875, 0.0087719944469088814015, 0.014578881323407746895, 0.012032267490189051812, -0.014696283893436974666, -0.071200491608987739878, -0.11349197476368319548, 0.018935521408144348476, 0.30661337601404164976, 0.090367382169682403115, -0.30690727847381027438, 0.04804847318890247132, -0.0020551012940901013798, -0.0007777001183917228418, 0.0018508066946820202203, 0.0069207595161828493538, 0.013801116738629869499, 0.014676171912258619348, -0.0078340847955260389801, -0.065186365524791720083, -0.11922878680025474074, 0.0015099426546758653971, 0.30358168665927087737, 0.1077464066325488079, -0.30918022534244732347, 0.047530875902365057883, 0.0069846339597673147426, -0.00089197694985395975433, -0.012698881338435616742, -0.020800757078583136106, -0.0087096140645417999271, 0.032745059192896615319, 0.076375608943323874955, 0.066455291091604334897, -0.027961330385820590272, -0.21025195091534601421, -0.40957844793403686001, 0.10316817294469847466, 1.2948833979456921828, -0.90052810284037765332, 0.0057575752538660315144, -0.00084395239843057013891, -0.010504321500770904402, -0.017174813229049120528, -0.0077370642416075205428, 0.02553760132143434794, 0.062010065835620645225, 0.059126157127647578637, -0.0096171989105365372069, -0.1654397307428313868, -0.37672378500775716148, 0.042931896124980238605, 1.1672425966651629459, -0.78326005460739056563, 0.0066374214688378193813, -0.0034185566993704457825, -0.015483413378869551741, -0.020861592512576569713, -0.0014496281647246415936, 0.052880059619103997326, 0.10394510535623298353, 0.04230012701026043126, -0.20669977496152733276, -0.32917564175961411266, 0.16333952708048674052, 0.46722925033561935049, -0.059015283348891903192, -0.20614357720257803219, 9.1681376611390475654e-05, 0.00028659657387810298906, -0.00047545394814112336944, -0.001520923947205803789, -0.0011777867370207238586, -0.00099813587380807606425, -0.0085869524716165965966, -0.017006276992230843353, 0.020104183153609168216, 0.12622959306472419927, 0.066121276896613970231, -0.29178578632572771401, -0.10838509055448589113, 0.22401265334275996532, -0.57582713370730675972, 1.0409125939534651373, 0.041517091418236840017, -0.36906164779306788182, -0.20986266939862635583, -0.046509793099093268975, 0.01311908337120715369, 0.030353678951163903105, 0.043533001701577991871, 0.043644155492055358714, 0.027980562602443751624, 0.0064242675201945237395, -0.011354832575843834888, -0.02257747374515232941, 0.13615443473656391182, -0.086494130885247225149, -0.1810980528756322705, 0.075283061477185039823, 0.081580213197684395277, 0.0077800156830741184696, -0.017870449930466908439, -0.012269212926529276739, -0.0034838371655297028368, 2.0015045860314424273e-05, 0.0018549154026708709886, 0.0022466276070799142892, 0.00093869420237841069612, -0.0012574763106751260262, -0.12287864967132436367, 0.31825208509003466428, -0.21546043011310608839, -0.032215610802112139288, 0.064355718125687008668, -0.011320722668546171624, -0.012906291691947570713, 0.008814896248226539166, 0.0075871416638806322627, -0.00036380406568643070247, -0.002279009190313478761, -0.0010216902192465896447, -9.5681714490131427914e-05, -7.9773269970678052426e-05, 1.4045124124024286338, -0.38446584075938539105, -1.1566562597434077997, -0.46455779836675464844, 0.13661733314202523237, 0.28409144931287638736, 0.21961042898406044066, 0.11528806421480805589, 0.03170461155038017953, -0.020895362490203812228, -0.042367855809552643076, -0.038911263574070077687, -0.019074301830826922988, 0.0089660945633121581777, 0.90828074869815722536, -0.25821309786843177347, -1.0203464361098111279, -0.22937506196810086845, 0.3738150180261585187, 0.31182914931956290383, 0.079736809591225671356, -0.045133254864137481444, -0.057324303990688774135, -0.028472234165770646253, -0.0040540036348667282326, 0.0047265065467233524643, 0.0034879273800071833138, -0.0013536372906050170295, 0.027313516609351183584, 0.24421377932333734884, -0.23335535843582530524, -0.16457317313405503412, 0.014589478284051304521, 0.084621771968194564595, 0.05482578356653481505, 0.0084364488128065729527, -0.010184777196229321747, -0.011070264100665775861, -0.0066778969040257266793, -0.0023908862107826687159, 0.00042517535401312325349, 0.0017976540829732405666, -0.40001551101461435556, 0.5996386380919118686, -0.072901037963855858104, -0.12817232713224757656, -0.013540743058971353024, -0.004149485917864143697, -0.00055535040278833103667, 0.006634716991634139957, 0.0075891000861319918291, 0.002834674058824270676, -0.0017491840531831009172, -0.0030946887580121529987, -0.0012948051225885025012, 0.0020031981006119969034, -1.3670486409114883664, 1.5657927858065139315, 0.4287371378934536903, -0.40243041324040407236, -0.30418107853138109942, -0.053526664295719249276, 0.048702710372915611337, 0.046584469409095642656, 0.020151254454333859117, -0.0015387259481199678547, -0.010825260534691323946, -0.0092583874321832699411, -0.0014371894526072561023, 0.0085247060285134346191, 0.14423778952626986616, 0.027955634549344639261, -0.40233640768891132389, -0.024323540618043554967, 0.21389256031958528426, 0.15041035465012034256, 0.010197530760175120557, -0.06705556464600144484, -0.059620298342287249282, -0.021305610410322674275, 0.0057065276589595851156, 0.013092415775095969654, 0.0089092124196512405276, 0.00043558515348925477828, -4.3862861904628971388, 3.2171626856708339659, 2.2998589892582730343, 0.17488862315768219369, -0.57794733098091199963, -0.59078465578613692522, -0.38011140251227532039, -0.1486965816660719697, 0.014190662489754639342, 0.08468429403303677494, 0.08660753158763402626, 0.05446864303174735189, 0.013633506670111315492, -0.024236063174746639431, 0.86507390002369899396, -0.55873073230816405754, -0.66177421943744341171, -0.054722298903797608749, 0.18026972759288850678, 0.21704701434418108286, 0.14197047296114176373, 0.02391299572477965929, -0.043719604579723722604, -0.054842227036615766311, -0.039348327847305013261, -0.017325991773982002409, 0.0039976892842073013731, 0.022381060131707240096, -0.67057132662066143425, 1.1623714310370902858, -0.12077755850657814762, -0.48914140646865666984, -0.073927702627325328488, 0.12843692603535233898, 0.087706171960549764166, 0.01197688677327243649, -0.02126979469540577275, -0.020795617131708344499, -0.0083637525676510966061, 0.0012073472033720480424, 0.004471904433245462579, 0.0033401592094447936365, 0.51825751172231360187, -0.67855335682182349455, -0.20072146042005545308, 0.34384741093843601556, 0.12467073073056719745, -0.054471264638114254286, -0.041006888397850903283, -0.0050025997500472568769, 0.0028386546732442484615, -0.0022946524306718444307, -0.0039738700992724106109, -0.00092588045912829714055, 0.0021567804481590737223, 0.0027402546913098146421, 0.1071566095278904257, -0.69762626659358983883, 0.062774377350479929061, 0.6681152184844473263, 0.25067821412349267085, -0.15764715195221579158, -0.20840928708498662747, -0.107339335916729256, -0.017347937177668679154, 0.023360500253557688055, 0.031142069883656010865, 0.023071747074437650576, 0.0082107529252225530669, -0.0088453358971619499423, 0.5649169174760747536, -0.44641881382762571562, -0.47174980092630980355, 0.055463222613306287223, 0.18107273263082021564, 0.15030156006585562722, 0.081409610432018553183, 0.00043198575741963327057, -0.044310264163021827188, -0.049040763491219038772, -0.032062670958444526581, -0.01013462517083203035, 0.0081356189977498209742, 0.020468559521306592397, 0.04804847308295495667, -0.30690727847976523313, 0.090367382399736143062, 0.30661337608917349495, 0.018935521270688680101, -0.11349197493636105361, -0.071200491671316812337, -0.01469628374678148483, 0.012032267475409232083, 0.014578881628630191436, 0.0087719942947274837097, 0.0031635598799318121258, -0.00050654857278413642149, -0.002723143954788928827, 6.7883135037007606982, -9.3614052843259294434, 0.75192395182075932869, 1.3181965812189933018, 0.56127820307050535487, 0.13286372725227751923, 0.030784197359694506074, 0.025327723090681826534, -0.0023838597038387566995, -0.038769964325418723727, -0.04829122962105569794, -0.03158209435050494257, -0.006141363955344605885, 0.016709788659654493947, -0.54214329082563506024, 0.75363966621648181832, 0.025047910942182029931, -0.22801431305047667464, -0.067838372601148749785, 0.050618426439783252568, 0.027033051329983695532, -0.026554699071538789834, -0.026816070904582725576, 0.0011838292358416768746, 0.017884682049563189327, 0.015083304622837299599, 0.0027808326798358148219, -0.010697873473982549494, 0.45103589516140507598, 0.59826111658463221321, -0.68050955060226026028, -0.7435768255035842822, -0.13848707394784906533, 0.24505992023783970679, 0.26564303380707660063, 0.13380622268266870178, 0.022961354707315424772, -0.028018029326058582762, -0.037858786884224102098, -0.028799838793852818031, -0.012376396378769480636, 0.0062816551689394480723, -0.30708086690690700182, -0.35060649376595048876, 0.64046838113981430229, 0.61876005765273556225, -0.21165918353773113814, -0.4127909805832405965, -0.16840720970674116774, 0.027343463326942464564, 0.080480195139899651791, 0.05931261973433996082, 0.025944762988322548486, 0.0033781735218780407221, -0.0086821923138132826858, -0.015196447568611328241, -1.7735106281084287438, 1.6389303588295986991, 0.66390775636927878978, -0.16787300889862866171, -0.20050147229964090934, -0.12898759671703560747, -0.09388652239901053298, -0.053528145631255276415, -0.0092623594916400337623, 0.01937858216286486579, 0.028136400281561516068, 0.021724333156485114982, 0.0076887505289129475655, -0.0083730427865250627723, -1.6173518744383601486, 1.5658465287977334501, 0.63653425335137270302, -0.14875887569737475391, -0.28677793601369333532, -0.2351767089608741701, -0.081545807751693441534, 0.03775053714456196724, 0.059842151159770720636, 0.030005203113865928227, 0.003103047429016360468, -0.0058992483347023146403, -0.003847466585764370943, 0.0018542676279083363747, 1.0409125941060786147, -2.0320068955158281376, 0.80336520453862370061, 0.54597746300911742523, -0.12806113594314841309, -0.19815683451375387669, -0.072849865233053673341, 0.005299184247869062267, 0.012494149877954285194, 0.013445202879583322692, 0.018351389554686008132, 0.017024565805127657436, 0.0056603854031408206218, -0.01135483256038994325, -0.086494130931749069191, 0.022945326359974357877, 0.32060461003369311594, -0.28407349774684031996, -0.066485944295074136945, 0.069918840746928265228, 0.036718992668286966341, -0.00067741524359927205237, -0.00945383940780329321, -0.0031318484311648533969, -0.0011161572761008459621, -0.00053121178815811057397, 0.00052609757717215197871, 0.00093869422761809723048, 0.318252084979389116, -0.86607847321267084073, 0.77945063797281677331, -0.12157092248332740381, -0.1967115012488095116, 0.08286075469654256842, 0.025582802082194350257, -0.024025317653648195493, -0.0048179820281608982599, 0.0080873226499368015135, 0.003356211107974059818, -0.0010939294765260068002, -0.0014368325955352800275, -9.5681564003217671878e-05, -0.38446584105686298916, 0.36362846959828376558, 0.9679298547776048478, -0.49964852651898078451, -0.60224929564068729881, -0.14603840990676353329, 0.093963707815641561716, 0.1309079586606402279, 0.090019947485914519891, 0.045373691279776896812, 0.010673079056475794774, -0.010208939224914517774, -0.018948964484228093819, -0.019074301440538635188, -0.2582130978789121678, -0.10547956898419295901, 1.2326639302533990428, -0.43473221102662545912, -0.70628018887350618726, -0.058415542539924286691, 0.21194907778611532523, 0.15092905041980730907, 0.034307726229190024791, -0.022147643862675356341, -0.027414269811862783388, -0.013215633070646428879, -0.001028684367941674286, 0.0034879273848679092848, 0.24421377932706653247, -0.64534577838776052694, 0.53194887692657555345, 0.042947704152614847584, -0.14838389989387007484, -0.092348586783167360692, 0.01029836182985883683, 0.047589906126077843873, 0.0228938606837629198, 0.0022292068403689095125, -0.0049409708368091075081, -0.005040728289046912669, -0.0027145033285922775743, 0.0004251754563856124285, 0.599638638153935144, -1.3601447492725016186, 0.76335655940406033082, 0.17044899387793024936, -0.14599594731182552465, -0.021700264745829863505, -0.00052536432446731885088, -0.0059437707965685464967, -0.003494045159001730929, 0.0021182518571833705394, 0.0035707103025869324497, 0.0020044502221390129819, -0.00035218615710576751509, -0.0012948050981468994629, 1.5657927857741220645, -3.052294896243235911, 0.75636298340905827597, 1.0480065433727652202, 0.051416499750718339112, -0.25007282945558018117, -0.1382278025090065976, -0.01832849896822380531, 0.016429852695903263832, 0.017110925946718730239, 0.0085351661780122649315, 0.00085572763172963705464, -0.0026073547619276239533, -0.0014371893658952795275, 0.027955634576924986989, -0.43317481594756473973, 0.75731622605084847333, -0.19534630641226477521, -0.28163420206198425122, -0.013786349158561918629, 0.10094862721094625957, 0.082121610476982534466, 0.0068081276246901474558, -0.032753563662036806747, -0.027997158807179824136, -0.0084205758162516613358, 0.0048977814092117325645, 0.0089092122013205987674, 3.2171626857832271718, -5.8653396570708160596, 0.33088325734591955696, 2.182458173531398149, 0.80383247655356826478, -0.043172962548987725562, -0.26716288375690261381, -0.25623131889039529874, -0.16703556164383237537, -0.062998698801478653087, 0.0093702913972719958924, 0.0398897546868290237, 0.036771469731159411864, 0.013633506499916705351, -0.5587307323612933363, 0.76973654434104588073, 0.39589640649239654646, -0.56678130227215839376, -0.19143698013614032782, -0.017329314402124179617, 0.048836182172427217152, 0.10235606359726454051, 0.061940455924225951145, 0.0039788915225923982, -0.022164908174733155555, -0.021876014861776663101, -0.010742031866322269973, 0.0039976892263899549765, 1.1623714311829713708, -2.5227289383932305356, 1.1176844590165706528, 0.76393580707765429416, -0.32247313456824650002, -0.27441158210589217115, -0.0089666546783934829995, 0.066175650428658416846, 0.037690284785159955394, 0.0050143619417238900066, -0.01050018078550198769, -0.0096669945438203792876, -0.0024897083765975823108, 0.0044719043266136498244, -0.67855335665712956938, 1.2409997264965391128, -0.075009354219978230383, -0.81092114175227303363, 0.11196210035417632211, 0.23975384897884532975, 0.021271207533609177986, -0.038620107132806218153, -0.018176668766539536698, 0.0014517346400046756555, 0.0031263844716862324545, -0.0011093797296496473066, -0.0019412155339497737597, 0.0021567803130482848459, -0.69762626653781578678, 1.2382510229933041401, -0.13045366792646265197, -0.9149598402415547449, 0.17466499586201048033, 0.41411614708741029123, 0.096234057107011480858, -0.087569729360311654442, -0.093798582407167027664, -0.041693574527819396436, -0.0043094324388427687408, 0.01080314231770051342, 0.012752790949879938553, 0.008210752806436362114, -0.44641881367908547995, 0.57370306742134813405, 0.35021914940707477015, -0.51641259552004281641, -0.11962740586071549109, 0.047030505396011924868, 0.051855913710983185527, 0.062794085749509037653, 0.03280614946291686429, -0.0028057955150520841325, -0.020323128364737780976, -0.0196731280822104726, -0.008364010464026537528, 0.0081356189706646431237, -0.3069072784269392673, 0.6289160690400681375, -0.16382923557804712766, -0.46704776431645594892, 0.21867310602187867152, 0.17948123788088732322, -0.018433876935297605071, -0.054762889468475843546, -0.031468148294553306699, -0.004797296746441153184, 0.006464359116375970743, 0.0066268089133609316826, 0.0031595930586764435152, -0.00050654862392930546759, -9.3614052844266506526, 21.212809110047853522, -12.10538096256971663, -1.4215147759494752844, 0.7601968968817331973, 0.63689348678285373495, 0.24417984564904249289, 0.035247860414951370789, 0.0065729430552349255024, 0.017224085735410149534, 0.0018516189827407492086, -0.018165834406797920497, -0.021023236465192818995, -0.0061413639855774949938, 0.75363966619261124613, -1.5977616551680382528, 0.65509129394129539392, 0.43595415359140865519, -0.089303929675889487383, -0.17260828296360530842, -0.035585414641123178092, 0.051809252448728373985, 0.022187195615903917978, -0.014440280529304960755, -0.014958318213104962457, -0.00032674429909278213113, 0.0068153434379270017018, 0.0027808326322012284423, 0.5982611163447937308, -1.3728130617166316441, 1.0245218182605404511, 0.37494529265338899515, -0.43071030005636362636, -0.38198102526059829476, -0.053179665571077956721, 0.12876448500660980789, 0.12067377176588073295, 0.05507619229107442449, 0.0059757272525714083544, -0.014401158056338875246, -0.017428997178281391278, -0.012376396145224941026, -0.35060649395880288948, 1.0931999065932671567, -0.82733629096096483746, -0.65528212686885412008, 0.72433421127131281114, 0.37275206335317923045, -0.17111545698443605956, -0.20045446702188254862, -0.067319595878697735536, 0.019458194627284639, 0.038723752295306883286, 0.024560229282507582693, 0.0051776611897514333849, -0.0086821922739407711433, 1.6389303589670396466, -3.2267140826128608566, 0.75361611476458623127, 1.0552606678675569007, 0.074809398638719507013, -0.13528364862225830678, -0.077788940061575703733, -0.050534587088266247346, -0.041618579072417893894, -0.022150866573004003579, -0.0032376102391090530105, 0.0088988520946605836892, 0.011992958512610474645, 0.0076887504079476802488, 1.5658465288624714429, -3.004520682418422961, 0.71886299691546162727, 0.85213306928514387373, 0.12438402165477675798, -0.018667221915604496202, -0.12304522959759735257, -0.13111515662914405289, -0.045768209617872139539, 0.020894822695284663572, 0.029818292968293629019, 0.013775316697620120648, 0.00043179258004420436115, -0.0038474665670897197078, 0.041517090957516794569, 0.80336520516673393111, -1.8084504787866899633, 0.72839514096696633683, 0.55642298094168063383, -0.075483080349792425201, -0.16819964928984740271, -0.084850454438754621678, -0.018244733360109444825, -0.0049852586759151374524, 0.0048771255617784756847, 0.015497007898114166513, 0.017024565881679991353, 0.0064242675712292158369, -0.18109805291396344207, 0.32060461012965801864, -0.23687695816351236888, 0.31334451471672036815, -0.24025416380344671863, -0.065845894567540169673, 0.063798712098386625713, 0.035677115509920767633, 0.002975058426123946935, -0.0090118359431287618777, -0.0045885379000117718779, -0.0021157085020510656495, -0.00053121126977521743086, 0.0022466273056685284217, -0.21546042990225466873, 0.77945063762673361119, -1.100847090378772819, 0.65992984792301823127, 0.0054687926499708610847, -0.19282993263999478883, 0.054931864531533765306, 0.024146198722399508929, -0.018611475759072310804, -0.004445289940961875963, 0.0071787307163335943583, 0.0031039855955938928586, -0.0010939287239535055215, -0.0010216906540183496883, -1.1566562595379119571, 0.96792985435996070631, 0.35699287067989121658, 0.87849448442796040926, -0.49457328673240613659, -0.64941616493097464868, -0.21535492072515183271, 0.044732221197325415518, 0.11748125029732769953, 0.10003194386636132807, 0.063968146064065103951, 0.024780680486923251094, -0.010208938598291731129, -0.03891126391493097969, -1.0203464360395966271, 1.2326639300457209458, -0.22803440318895606964, 0.86035218637521126972, -0.45677846937511207281, -0.68490422313603327442, -0.059463389629043308937, 0.20894387416605561536, 0.15442396895055079953, 0.038206378133179993983, -0.020348292142643532865, -0.027789025893397885369, -0.01321563347778082452, 0.004726506745221702166, -0.23335535846945973382, 0.53194887698905923834, -0.62106791910846548266, 0.48482888295718923333, 0.0072609742436666267068, -0.15503707411739967026, -0.087062579002134238282, 0.0063955919203394868619, 0.046782346361805199342, 0.025171530281743490637, 0.0043291651037200753332, -0.0042251558566472838946, -0.0050407278075895681907, -0.0023908864659090657209, -0.072901038158122155597, 0.76335655977765404234, -1.4045244053294880793, 0.64949217887313992481, 0.22832781030467572125, -0.12632885589619979871, -0.025977947079216769793, -0.0064923399485615927371, -0.0078590573089060058315, -0.0028852308837238049843, 0.0038793790424836711687, 0.0048214829624360381818, 0.0020044502417836257413, -0.0030946887710140273155, 0.42873713800695295584, 0.75636298320962225272, -2.5921655471579825303, 0.69591187214003091821, 1.004802927591136763, 0.07441990527219638496, -0.23361692395749700246, -0.1450418921734507316, -0.024051591846421103982, 0.0175645350168280312, 0.021834305876764582843, 0.012395573852607113391, 0.00085572768652828184316, -0.009258387438093176186, -0.40233640762176187122, 0.75731622596071623832, -0.59562443167715384273, 0.5748666142212669472, -0.2336674632709467847, -0.27612782118657930308, 0.016514036774118209805, 0.11789126570566529906, 0.087528515627943684341, 0.0076028169831713100779, -0.034207144596241446932, -0.030683114538523369175, -0.0084205763177594081248, 0.013092416048901869483, 2.2998589892434671, 0.33088325749415620214, -5.4693326938860673536, 0.20237098754068183482, 2.2448731612546248826, 1.0050205424810760846, 0.10240779961820994926, -0.21611932119909288419, -0.26277652431956860513, -0.19739178920722083577, -0.092825828864128295814, -0.0077359991324248422728, 0.039889754704376029215, 0.054468642942890749259, -0.66177421932062130416, 0.39589640641294665535, 0.54921944655919707756, 0.45856094171974598472, -0.62985642231622906984, -0.29202090922021567421, -0.0388821162546575394, 0.050270726475397431443, 0.11114378705999716779, 0.077475017458535300197, 0.019203334288370565996, -0.013946161371455184641, -0.021876014947849853343, -0.01732599177321345052, -0.12077755867070286144, 1.117684459316836465, -2.0729178434690616051, 0.85353525073895042397, 0.68819565595642950839, -0.28856947875871724607, -0.25976517127910819793, -0.0092582536384461580625, 0.068374580099519202681, 0.039492146302443811856, 0.0057238086169386845348, -0.010780623374885646426, -0.0096669950340811557077, 0.0012073474824692074885, -0.20072146069732091367, -0.075009353769484643504, 0.61456814439798512772, 0.10680440969446237942, -0.70574861524824172854, 0.060103230805444891882, 0.21458533122200587617, 0.023801818325841796881, -0.030101944105042459376, -0.013885712323041671246, 0.0015259310824116975938, 0.0019836278904845358953, -0.001109380254757360261, -0.00092588014928739572392, 0.062774377415661178348, -0.13045366805871394522, 0.34621553431666290956, -0.066207086702836148717, -0.74581638693972285292, 0.19771303124431932607, 0.42445920430353328845, 0.1150401584588675169, -0.076149932699989208995, -0.097784906538013069421, -0.052910306091795478167, -0.01280059438058298027, 0.010803142031970634426, 0.023071747185584415979, -0.47174980093827617589, 0.35021914946747462194, 0.28452463739134936649, 0.39104203041494617343, -0.54349526364055034122, -0.19605212296816662554, 0.027747542694418994569, 0.057693864057321200822, 0.077530602108544399709, 0.048349194105698267698, 0.007990414023446662517, -0.015983848019123704509, -0.019673128269184762817, -0.01013462512322481357, 0.090367382285431188671, -0.16382923533637505642, 0.16304867758215607632, 0.01967333899319108001, -0.39036802878461568378, 0.19737696044914654347, 0.17947714239586851082, -0.015010321315189893362, -0.053966586924792572799, -0.033566215163755810935, -0.0074661527058869532614, 0.0050885764784196771751, 0.0066268087324226924648, 0.0031635599962452042033, 0.7519239519037804742, -12.105380962737145367, 23.462996126192059876, -11.837808941341124935, -1.6802286737446334719, 0.54612637357215598755, 0.52040014295520342635, 0.21963816446090561607, 0.058624054301197187078, 0.033260633625343094533, 0.033977499778230002558, 0.009590195913510984882, -0.018165835025641265121, -0.031582093969942465206, 0.025047910992949021236, 0.65509129390299902784, -1.4136707528613601248, 0.56381257164099107815, 0.35817442521888698259, -0.078024510836536720149, -0.12080333310762284005, -0.017302243232740766032, 0.045031539845662493793, 0.011918320734703530261, -0.021148563398313724448, -0.01884344517945239772, -0.00032674426551700802568, 0.015083304651731304843, -0.68050955033123217142, 1.0245218177411101745, -0.88145062912791960485, 0.96747055394113457005, 0.27154700703982703747, -0.48761087302166195023, -0.41478836961714898868, -0.082406922846289476503, 0.11608467743844645692, 0.12532872055123633892, 0.067878614548844196097, 0.01572715988657703845, -0.014401157143744576505, -0.028799839203255898912, 0.64046838142092243906, -0.8273362915373132509, 0.2756911096925112048, -0.28837102791186230988, -0.43670207192110688244, 0.67086673866366286489, 0.35913055272299659881, -0.1676679868107681104, -0.2071418371225225985, -0.079593143874895966317, 0.0094230733855907022889, 0.034349629230630090992, 0.024560229132196326196, 0.0033781736178022468003, 0.66390775619001862751, 0.75361611508641224155, -2.9216274018870098672, 0.55571027649206383714, 1.0481456392488761598, 0.16513009490045196803, -0.077248026495502525313, -0.059814001233656072243, -0.050891007556427114711, -0.051451433619395935126, -0.033316127971584195977, -0.01041659253834667162, 0.0088988519093030366697, 0.021724333219123131283, 0.63653425322573053968, 0.71886299713857815252, -2.8714015925585796296, 0.6333399837578702618, 0.99824189508856397968, 0.19145507073921805286, -0.032908051969457244312, -0.13763794752928962262, -0.133232795285505079, -0.047356886632278619165, 0.019441742624500990555, 0.030249284396393169572, 0.013775316866586931677, -0.0058992484733257364096, -0.36906164731345131313, 0.5459774620475512652, 0.72839514157472884381, -1.8729602647223344558, 0.71485618474985823667, 0.58008313166643143965, -0.070910814758626886101, -0.17840049494486226456, -0.095882441517047917179, -0.023489371672645378841, -0.0065652485888301592015, 0.0048771256977614254116, 0.018351389436764471014, 0.027980562594099367413, 0.0752830618473800256, -0.28407349835699596818, 0.31334451481881159252, -0.21271062541813806623, 0.33130236031948195885, -0.23759386144755390013, -0.07192683728894337658, 0.062334903255680772227, 0.034703570526401023022, 0.0020990014543985843765, -0.0095108455340083570412, -0.0045885378913277496266, -0.0011161580739492736949, 0.0018549159093371445951, -0.032215610768753198678, -0.12157092259553670122, 0.65992984812745869583, -1.0270706376767730905, 0.65614857283869676419, -0.008526071725407109031, -0.19553435953980694006, 0.058327482085362376829, 0.025839759005780883055, -0.018891648273384454432, -0.005106712765312698861, 0.0071787306360176182662, 0.0033562105870151423201, -0.0022790088545365755172, -0.46455779796539520277, -0.49964852677825011273, 0.87849448382032091498, 0.2705800008640772969, 0.85816180788829399262, -0.49161112646486165856, -0.65792318218521605999, -0.22264367717599656271, 0.045199115058830369174, 0.12507140449932019344, 0.10772791312600418778, 0.063968146359939706547, 0.010673079832164254419, -0.042367856434527881304, -0.22937506201892671198, -0.43473221068654693822, 0.86035218587157491488, -0.32987413444957036646, 0.91953208374336625752, -0.38190835733929368034, -0.66773022939105186602, -0.075390824590053709398, 0.19206942704709900616, 0.14851238289294688144, 0.038423988182506023381, -0.020348292640970652018, -0.027414268981765854694, -0.0040540040246561459894, -0.1645731730408779292, 0.04294770393598950492, 0.48482888303313875777, -0.7013783218103500916, 0.47036541291667977038, 0.02894018043175753882, -0.14571078040539697795, -0.086087289192643939861, 0.0041891656172677212144, 0.045479295450487032537, 0.02501232001322854781, 0.0043291655014028318504, -0.0049409712878196506564, -0.0066778967489709312871, -0.12817232681156995322, 0.17044899313584388478, 0.64949217925987035027, -1.4032919537164285639, 0.64089224301741465339, 0.23005781129626207138, -0.1270325109317900214, -0.025512045264795408378, -0.0044614563224609623746, -0.0056217434725736106782, -0.001930212377819272556, 0.0038793792480527229671, 0.0035707101206504153826, -0.0017491840153632447295, -0.40243041349571090359, 1.0480065438255521393, 0.69591187210499094729, -2.7080368640602827668, 0.61735531058650938707, 0.998463968110263278, 0.09061085654235365161, -0.21772901241060529931, -0.13659157812687705569, -0.020126467593850110027, 0.019229961228579513233, 0.021834306210625865213, 0.0085351663084655051555, -0.010825260718703121066, -0.024323541007365303418, -0.19534630568312433274, 0.57486661407477290808, -0.6789000656129038358, 0.60006683315602316053, -0.17060684364639461252, -0.25535793416484453555, 0.0078647889745796367711, 0.10557424683303959223, 0.082459393020172827571, 0.00688767296510269314, -0.034207144794834082324, -0.027997158806794729402, 0.005706527754712030151, 0.1748886229847325946, 2.1824581735978827446, 0.20237098770486650068, -5.4132556651076457754, 0.20658204352862602482, 2.2480872245519196717, 1.0134714906680515334, 0.10907324645169913513, -0.21646175966749514608, -0.26776418509348121066, -0.20350067357332235485, -0.092825828865794046685, 0.009370290800939373721, 0.08660753204336582578, -0.054722299090216544593, -0.56678130208761501319, 0.45856094186175266891, 0.45114410541354671436, 0.40805606187789367212, -0.60719065903170899023, -0.27454766893140158945, -0.033314941273229536911, 0.052425704328175282143, 0.1152409591359655483, 0.081711836855434705207, 0.019203334226981061406, -0.02216490804546203397, -0.039348327831383957121, -0.48914140658584598365, 0.76393580713965747453, 0.85353525094742377366, -2.2396815830907059741, 0.81765106619344796179, 0.72230870723360562913, -0.25650958182106647909, -0.24955186894724001867, -0.013275959159663314191, 0.063665889569886294219, 0.037853420960808813189, 0.0057238085143197087179, -0.010500180541964334105, -0.00836375269648572707, 0.34384741108327299131, -0.8109211420463230402, 0.10680440977434381422, 0.73111019066002502775, 0.11375229211935181106, -0.72842094628105247267, 0.041478178243825977733, 0.21555105759613704675, 0.029373871174613437607, -0.028572865785221848722, -0.015046969271726922529, 0.0015259308162999087391, 0.0031263848832080785682, -0.0039738702677103217992, 0.66811521819558039592, -0.9149598398029878954, -0.066207086656296085247, 0.55502728099768228187, 0.02857180324800026941, -0.74903267444386767426, 0.1661985488596426086, 0.41114563784057811535, 0.11320846473740955873, -0.077449847725880219262, -0.10061940643631681058, -0.052910306779557603196, -0.0043094321245175613855, 0.03114206991928962151, 0.055463222333725283042, -0.51641259503961189914, 0.39104203021595851375, 0.23277883287480680141, 0.35981159869128093165, -0.5307695987538129101, -0.18447142653506157228, 0.036288620058942372415, 0.06252506573096877629, 0.080348875318147397473, 0.049999967047911986207, 0.0079904140436477741338, -0.020323128420297865288, -0.032062670906941051485, 0.30661337618349571166, -0.46704776465875114111, 0.019673339298800988917, 0.25594426003200004827, 0.029506826807291639675, -0.40639628939943434061, 0.17884003755306082661, 0.17673594312265406003, -0.012043536996619527008, -0.052244976155307633336, -0.033658949190125998785, -0.0074661525172508439244, 0.0064643590051754401885, 0.0087719943036231665112, 1.3181965812731957222, -1.4215147757877395485, -11.837808941832273391, 24.150819512049356064, -11.713553042323155751, -1.8045586140176277379, 0.45527737159902642583, 0.50500775129959318832, 0.2323678142155886639, 0.069477056463377179218, 0.037193151604657415099, 0.033977499770358521314, 0.0018516199931765675579, -0.048291230251347280755, -0.22801431296830090845, 0.43595415336119514027, 0.5638125718450501811, -1.5451338752221310813, 0.53591447590578134008, 0.41315131238997054863, -0.04638054182692380456, -0.12441453382858169707, -0.031749537960562512762, 0.036735242238566723216, 0.0097952670877072267608, -0.021148562792588004866, -0.014958318929446420831, 0.01788468229542891294, -0.7435768254700084734, 0.37494529271631527045, 0.9674705539192732795, -1.0959478746144846184, 0.8651511357837470273, 0.2856141932525742555, -0.45899358321470767885, -0.40464502314593814303, -0.084086506491941437336, 0.11586584074371111819, 0.12811274828684082738, 0.067878615291203536763, 0.0059757264681910143844, -0.03785878668794936952, 0.61876005759567920261, -0.65528212642316208125, -0.28837102856758545233, 0.55200465822830757645, -0.31298744750554402527, -0.5190806504849406755, 0.62455817171733907234, 0.36054202855185946142, -0.15534583361447013017, -0.20361729248977183948, -0.082103839711148818292, 0.0094230731605882966928, 0.038723752752956447365, 0.025944762727965239857, -0.16787300899810669863, 1.0552606678471370127, 0.55571027671948947102, -2.9803007938310601688, 0.54924476768402041849, 1.071155643299589455, 0.18332100521382252434, -0.071560751078769996303, -0.063342371840648656334, -0.056060754220326726449, -0.055106713710466950773, -0.033316128456151825055, -0.0032376101282826745542, 0.028136400376123620742, -0.14875887576821447689, 0.85213306937282395914, 0.63333998381421741097, -2.7824773925945138764, 0.60121684280833520564, 0.94194578934388095703, 0.17433315925704079286, -0.018060706261897669506, -0.12106158771208885838, -0.12680010630976157326, -0.047394672291071832104, 0.019441742219149866361, 0.029818292838453989979, 0.0031030476370793062466, -0.20986266952888074644, -0.12806113561754856045, 0.55642298064925033163, 0.7148561848252257267, -1.8784277342831532387, 0.69772634047953929759, 0.56675565332163313492, -0.076730315267765913934, -0.17913872643770903381, -0.095656097882787033404, -0.023489371720619576372, -0.0049852588081244850887, 0.013445203288269723385, 0.04364415528292744928, 0.081580212813946173522, -0.066485943685520798585, -0.24025416382806261129, 0.33130236013858938016, -0.21021727133589668268, 0.33099388728646117475, -0.23659663303725039785, -0.07406106140033567875, 0.060127529681216929092, 0.033919884048102401375, 0.0020990010315114328646, -0.0090118355125431322261, -0.0031318481724239011221, 2.001479653790367132e-05, 0.064355717864921280214, -0.19671150051115204871, 0.0054687918523340919505, 0.65614857324615727396, -1.0539316823790643163, 0.65389908549794617532, -0.0024852379500013665851, -0.19433802413930545416, 0.057675722565885510773, 0.025434514458704090417, -0.018891648854312749484, -0.0044452890862257685503, 0.0080873222771233824885, -0.00036380404823170466987, 0.13661733296470698895, -0.60224929533753923661, -0.49457328649498094597, 0.85816180731645841373, 0.25977351963269346147, 0.86439428377592320007, -0.47549975689769141685, -0.64470622648209885419, -0.21371301950199611541, 0.050169807326623780397, 0.12507140492583748714, 0.10003194431690805299, 0.045373690088593751635, -0.020895361971412666768, 0.37381501795187260795, -0.70628018899112654516, -0.4567784687509294228, 0.91953208327971280855, -0.26991495307081725397, 0.95206937997785479055, -0.37630298123623290296, -0.67849894905501562103, -0.085998158967646384787, 0.18798364443616627506, 0.14851238267653699143, 0.038206378855748880652, -0.022147644648860209671, -0.028472233879486271574, 0.014589478191131934406, -0.14838389962783471443, 0.0072609741345435255266, 0.4703654127449612421, -0.69075377247338720643, 0.47711143839899000652, 0.034084428316613998855, -0.14422300189897277467, -0.086973921265674491998, 0.0031049016059576887289, 0.045479295598296436143, 0.025171529931986430517, 0.0022292071528088298026, -0.01107026419736376055, -0.013540743361704099881, -0.14599594658213130294, 0.22832780980855407749, 0.64089224307243197742, -1.4196049675552795399, 0.63222703172140592986, 0.23077765749140091711, -0.123739699267704506, -0.022551799983306943659, -0.0032245917299990982385, -0.0056217434558242518339, -0.0028852312619233001431, 0.0021182524182604520364, 0.0028346738331076434293, -0.30418107820615553072, 0.051416499246727774552, 1.004802927476049712, 0.61735531086451866933, -2.7494281254966312922, 0.59972186208196121004, 1.0049388395626921433, 0.10411811501069540009, -0.20829690762601921206, -0.13329887855411390318, -0.020126467135493167349, 0.017564534445181689826, 0.017110926052248998791, -0.0015387259121847399711, 0.21389256078575427655, -0.28163420309901676664, -0.23366746288061204795, 0.60006683339181277059, -0.62401187798935020545, 0.62649565907028104128, -0.17144593843140817913, -0.26659269034761523542, -0.00065121691948803739347, 0.10231359730303260513, 0.082459392839925288454, 0.0076028168792793109076, -0.032753563333901017374, -0.021305610534922324317, -0.57794733094593631062, 0.80383247642074817652, 2.2448731613826051756, 0.20658204361132853633, -5.4525111771020329243, 0.14133270130024111477, 2.2053602979454036692, 0.99827673383911141869, 0.10881031332339097251, -0.21477483326057233981, -0.26776418490162801067, -0.19739178963271669653, -0.062998698664631605504, 0.084684294085653394335, 0.18026972768769747302, -0.19143698032338263282, -0.62985642231951333159, 0.40805606210320999327, 0.47852428626054532845, 0.44262301431994688361, -0.58867186735744647663, -0.26760961236237212901, -0.030857760951413348849, 0.053343161055192322983, 0.11524095948153723068, 0.077475017156895642501, 0.0039788913261114000619, -0.054842226879727007383, -0.073927702509168566336, -0.32247313476533828736, 0.68819565581094799267, 0.81765106648654084509, -2.2419555405627069611, 0.82790246506021147876, 0.73555510191564266176, -0.25279759133481621847, -0.25200557751784485871, -0.015848287272991451458, 0.063665889445908174782, 0.039492146202384621756, 0.0050143622418815821107, -0.020795617251746816717, 0.12467073060044782384, 0.11196210051122111595, -0.70574861510521169627, 0.11375229195947070271, 0.72654822059928092504, 0.11428047092688933817, -0.72601159135442372428, 0.041959326367705634953, 0.21475397607160703894, 0.02824156326119558677, -0.028572865880998169624, -0.013885712105965373336, 0.0014517342323192919396, -0.0022946522550601022687, 0.25067821415045149491, 0.17466499580494002086, -0.74581638704004016294, 0.028571803475169330749, 0.57567747103400823683, 0.018423999625110855372, -0.76636340356710341837, 0.15113949534659673613, 0.40356176483138517241, 0.11097077939762227394, -0.077449848215113803462, -0.097784905708150654524, -0.041693575032586249751, 0.023360500376409538276, 0.18107273263765438776, -0.11962740608151678501, -0.54349526317636120964, 0.35981159844079446675, 0.2509885110364091787, 0.38646631456797275739, -0.51321477426418227186, -0.17811758485653450546, 0.036683952901617633557, 0.061538422746625409676, 0.080348875202610386959, 0.048349193795766895465, -0.0028057947701073254754, -0.049040763831029551834, 0.018935521133467653893, 0.21867310637065559553, -0.39036802901606254546, 0.029506826832414193018, 0.25312454676282253319, 0.019483473904863535947, -0.41347595875856862913, 0.17638659654426525059, 0.17743873631569018201, -0.011128200271380461056, -0.052244975872450201415, -0.033566215750469599599, -0.0047972962951286493224, 0.014578881498083517271, 0.56127820312884069143, 0.76019689662679135189, -1.6802286732415290249, -11.713553042674041293, 24.275831349362487543, -11.677774247743606395, -1.8006557260887223748, 0.45322407982273676952, 0.50305573852193874451, 0.23168604029271711031, 0.069477056984976442999, 0.033260633435103881261, 0.017224085038754971544, -0.038769963926958930789, -0.067838372688991871451, -0.089303929350952604427, 0.35817442478054245525, 0.5359144760865789392, -1.5102399946247984097, 0.55264070631273998213, 0.40791162334024666514, -0.057305762644935365202, -0.13126250501091130118, -0.034242200462247573967, 0.03673524245516301967, 0.011918319830427515083, -0.014440279398300281954, 0.0011838288038367936195, -0.13848707385428202232, -0.43071029997981014059, 0.27154700655767988193, 0.86515113599672321953, -1.1134439221244500917, 0.88265144753424396029, 0.30935493815571646348, -0.44440640731867919744, -0.39901027732393273828, -0.0828894518437944261, 0.11586584108945205407, 0.12532872014265794558, 0.055076192528009992877, -0.02801802940549130358, -0.21165918331687333387, 0.72433421074162340769, -0.43670207140891237341, -0.31298744783410126225, 0.49785885239814098302, -0.35269838976526202146, -0.53679218720404398013, 0.62449897936857334813, 0.36619843848144928433, -0.1527605194424501156, -0.20361729242617679358, -0.079593143486579809309, 0.019458194078765103652, 0.059312619960840196054, -0.20050147229586079423, 0.074809398665338686096, 1.0481456392494177265, 0.54924476772022390314, -2.9900491445646562028, 0.53320338962593283316, 1.0625719193871829837, 0.17827661664695507171, -0.07457429711292461183, -0.064982183788642441935, -0.056060754546890356753, -0.051451433250393710517, -0.02215086688647335128, 0.019378582313616998789, -0.28677793582134464279, 0.12438402130050343897, 0.99824189504371418913, 0.6012168430301634281, -2.8699064822502138306, 0.55313548687761293632, 0.94198822372136592485, 0.18893616317939182503, -0.0071634962522110698369, -0.11721853010576710474, -0.12680010702241284459, -0.047356885788768822432, 0.020894822646816729045, 0.030005202963268393784, -0.046509793324082560062, -0.19815683411998732955, -0.075483080961364926176, 0.58008313246753284709, 0.69772634027850222171, -1.9021239258251094739, 0.68508732555343754278, 0.56514568086987881923, -0.075316952453628821518, -0.17913872654420909258, -0.095882441604538751201, -0.018244732773866189496, 0.012494149152924344595, 0.043533001940678987818, 0.0077800157272615091952, 0.06991884060757211361, -0.065845894456888598434, -0.23759386142545557696, 0.33099388730881845744, -0.20756771072831253777, 0.33099232001739564701, -0.23812296031167215493, -0.075079475681873591997, 0.060127529998967088609, 0.034703570431619167047, 0.0029750583001001435629, -0.0094538393101768446181, -0.0034838371938978995174, -0.011320722474458476056, 0.082860753976964546252, -0.19282993163633671552, -0.0085260724287962892654, 0.65389908587234080173, -1.0495735028273533995, 0.65542461798854556587, -0.0028697234105191779183, -0.1948695361427613959, 0.057675722546646719657, 0.025839759451105295818, -0.018611476413375044325, -0.004817981563174652572, 0.0075871415284940058738, 0.28409144927232693467, -0.1460384100198756363, -0.64941616451448402891, -0.49161112677015161854, 0.86439428368215653897, 0.2766741681047287793, 0.88523521504720892317, -0.4588914038011752683, -0.63678140740350563931, -0.21371301909939177666, 0.045199115398561015566, 0.1174812497949795087, 0.090019947126577917751, 0.03170461189268221347, 0.31182914933769245724, -0.058415542460618086196, -0.68490422332677691752, -0.3819083573354569161, 0.95206938009945962786, -0.26318484175581652318, 0.94984538441708155077, -0.37964046664351980276, -0.68101442404757916993, -0.085998158706920138639, 0.19206942702244361731, 0.15442396849176015516, 0.034307726723004264291, -0.057324304126130057324, 0.084621771918130445078, -0.092348586667581361587, -0.15503707424915597057, 0.028940180722429535709, 0.47711143803340771985, -0.68844201636701252145, 0.47835687863876819037, 0.034639301528360741389, -0.1444644407232699157, -0.086973921204606438606, 0.0041891654986128835231, 0.046782346756849876868, 0.022893860078693421822, -0.010184776932620348555, -0.0041494857924109515915, -0.021700265053971816431, -0.12632885548948111221, 0.23005781103065600646, 0.63222703164454474578, -1.4237700903932750673, 0.63309124627772339622, 0.2334388237004706157, -0.12232262246202177391, -0.022551800141636192676, -0.0044614561764542886094, -0.0078590572486642567424, -0.0034940455020318304802, 0.0075891002855605401572, -0.053526664340377998197, -0.25007282932441221579, 0.074419905163630728495, 0.99846396822442029517, 0.59972186187577458671, -2.7578142862460914841, 0.60111501351987906006, 1.0108851834325245367, 0.10824806279340665194, -0.20829690770029662961, -0.13659157811023256435, -0.024051591474190850417, 0.016429852186229443323, 0.020151254648110308865, 0.15041035457214807503, -0.013786348821537194528, -0.27612782160392046471, -0.17060684349495647272, 0.62649565905432236246, -0.62498900607134877383, 0.61850964681449083216, -0.17714677244178855808, -0.26947561763004207469, -0.00065121694957021719619, 0.10557424691078684531, 0.087528515713427040845, 0.0068081274619046470986, -0.059620298306603550231, -0.59078465581588712752, -0.043172962253118452036, 1.0050205421620925783, 2.2480872245272993659, 0.14133270130212671756, -5.5195894973639747505, 0.097220598384907819245, 2.1881557272109768419, 0.99602509521922399927, 0.10881031315739375653, -0.21646176031461283884, -0.26277652383778166145, -0.16703556091182303334, 0.014190661957748464117, 0.21704701453620758866, -0.017329314565321718433, -0.29202090941715819206, -0.60719065908118230457, 0.44262301464596487399, 0.50130756057905168088, 0.45599961044749748318, -0.58174972404409475146, -0.26546751187963679719, -0.030857761088064261656, 0.052425704158215569173, 0.11114378730972623854, 0.061940456027733015176, -0.043719604713495792414, 0.12843692610865981019, -0.27441158227263190428, -0.28856947856099768313, 0.72230870702901883718, 0.82790246524109001314, -2.2409560097998939554, 0.82544012188178872425, 0.73392483216626092091, -0.25373580356713293371, -0.25200557762265740624, -0.013275958885116588559, 0.068374579783368263919, 0.037690284885434577333, -0.021269794677936035288, -0.054471264651158861347, 0.23975384895757248516, 0.060103230882474996399, -0.72842094628637565901, 0.11428047092854615174, 0.73725947575000161383, 0.12038390793032642534, -0.72652886112004833041, 0.039778917216730942663, 0.21475397572857896078, 0.029373871249657630689, -0.030101944457418275403, -0.018176667858609725503, 0.0028386542478855206206, -0.15764715164108164536, 0.41411614656575690407, 0.19771303138680293432, -0.7490326745822212251, 0.018423999979742006206, 0.56909532591079770025, 0.0097264917896231154609, -0.77552439349294022275, 0.1459924778068472373, 0.40356176513898422975, 0.11320846455997281288, -0.076149933011912113168, -0.093798581876836040205, -0.017347937402498188331, 0.15030156016903489791, 0.047030505414346598925, -0.19605212342135511228, -0.53076959825051972519, 0.38646631432079214807, 0.27277181393062854831, 0.3981798472036408576, -0.50988366431764942099, -0.17868805175545943165, 0.036683952805439012934, 0.06252506613081063036, 0.077530602076808965983, 0.032806148777864217736, -0.044310263783678256744, -0.11349197467379001925, 0.17948123744358215892, 0.19737696041273103398, -0.40639628917929965235, 0.019483473877753294812, 0.25089329094446000123, 0.018785569462074401537, -0.41426274336334562021, 0.17611632224109541323, 0.17743873639441345991, -0.012043537458175237326, -0.053966586471029226457, -0.031468148357475855814, 0.012032267445652767429, 0.13286372698257284397, 0.63689348727733130495, 0.54612637341370551258, -1.8045586140971148215, -11.6777742477346127, 24.353008418229645571, -11.63647076511705869, -1.7933731331308007118, 0.44991896730012881367, 0.50305573829987149104, 0.23236781434820202863, 0.058624054125372887514, 0.0065729434352351234772, -0.0023838598652427996083, 0.050618426245994538326, -0.17260828274043754638, -0.078024510654679912758, 0.41315131230364404713, 0.55264070604879611537, -1.5230934354131773212, 0.53706070567898167578, 0.40093603420639245538, -0.058890055796845142133, -0.13126250544786327579, -0.03174953777205191674, 0.045031540132659586551, 0.022187195280455804247, -0.026816070781230119668, 0.24505991994159959102, -0.38198102492578200895, -0.48761087282128645626, 0.28561419317131864215, 0.88265144728181033074, -1.0991845924231986587, 0.89510678414522604207, 0.31962389660215012599, -0.43947551467764689459, -0.39901027734539884495, -0.084086506642410407952, 0.11608467780467936203, 0.12067377137386121888, 0.022961354863372356039, -0.4127909810202761709, 0.372752064129291516, 0.67086673845704614116, -0.51908065051351726105, -0.35269838990642776722, 0.48771617054804616842, -0.35442122086843297346, -0.53773991836298917768, 0.62433441204788997503, 0.3661984388619278219, -0.1553458339027359858, -0.20714183720541928801, -0.067319595668237225405, 0.080480195055100928192, -0.12898759641506016282, -0.13528364912995077818, 0.16513009494511871056, 1.071155643296742177, 0.53320338991262838935, -3.01254830649031069, 0.51599693719419892357, 1.0545377720588091641, 0.17605856211483073004, -0.074574297166671923986, -0.063342371901612973262, -0.05089100762452149207, -0.041618578718891574564, -0.0092623597106614649621, -0.23517670886851221868, -0.018667222198737270372, 0.19145507118704133132, 0.94194578896550695823, 0.55313548705680570805, -2.8822077488611124174, 0.55584721726458052871, 0.94687796368599475016, 0.19189957279762756559, -0.0071634960465272526364, -0.12106158727450447299, -0.13323279593180262226, -0.04576820965925640361, 0.059842151318968375695, 0.013119083729819652762, -0.072849865637048483169, -0.16819964899176259587, -0.070910815179891117932, 0.56675565342304978778, 0.68508732560424256963, -1.9068245358827200509, 0.68497213617891761217, 0.56514568067293879938, -0.076730314905397445369, -0.17840049510314442971, -0.084850454791525253828, 0.0052991849885343873458, 0.030353678553772914933, -0.017870449845755028567, 0.036718992651774522151, 0.063798711977667926787, -0.07192683750197764736, -0.23659663268719166584, 0.33099231990318989105, -0.20833868525038767849, 0.33049461347623670937, -0.23812296000037777577, -0.074061061850879850166, 0.062334903476668507283, 0.035677115581805717015, -0.00067741537588433386324, -0.012269212882296679987, -0.012906291820851321234, 0.025582802647835273246, 0.054931863675773899558, -0.19553435898600449172, -0.0024852381899776350427, 0.65542461806049356898, -1.050166286600918264, 0.65492642247976318171, -0.0028697233230297438181, -0.19433802397980198817, 0.058327481480184253937, 0.024146199510783193654, -0.024025318214179990084, 0.0088148964159597918006, 0.21961042847573528092, 0.093963708544715376925, -0.2153549206831307794, -0.65792318198132382356, -0.47549975744108108344, 0.88523521486330769559, 0.29490850857765049664, 0.89546025605668810865, -0.4588914035877915687, -0.64470622640344221743, -0.22264367710986210369, 0.04473222084940724208, 0.13090795863998105908, 0.11528806438123027911, 0.079736809664250229979, 0.21194907771043938749, -0.059463389952260456428, -0.6677302289019594328, -0.37630298145484308847, 0.94984538449051048037, -0.26430155587523085359, 0.95004651599911427162, -0.37964046620604957605, -0.67849894964998824776, -0.075390824179802098937, 0.2089438739556009339, 0.15092905060255976113, -0.045133254956449944673, 0.054825783795153369693, 0.01029836119514766779, -0.087062578393116069053, -0.1457107809031585044, 0.034084428884371534285, 0.47835687831342110776, -0.68838644163473372561, 0.47822939827487775943, 0.034639301520718847327, -0.14422300202950510961, -0.086087289045348625027, 0.0063955915523771292297, 0.047589906737997740571, 0.0084364485358841504598, -0.00055535045174002529728, -0.00052536429138301966174, -0.025977947063452383469, -0.12703251097276513359, 0.23077765771335509437, 0.63309124605305799882, -1.4224075695058631119, 0.633923781969842004, 0.23343882338138727262, -0.12373969883411289894, -0.025512045622550959256, -0.0064923395560773818069, -0.0059437711128111437101, 0.0066347170853935499152, 0.04870271014219118566, -0.1382278022899228187, -0.23361692355417332245, 0.09061085600953819319, 1.0049388397835787856, 0.6011150134022250624, -2.7569602265323589485, 0.60223146196044330392, 1.0108851833885095228, 0.10411811520966897382, -0.21772901245718034868, -0.14504189246955725978, -0.018328498776285458149, 0.046584469410703988057, 0.010197530717296069552, 0.10094862720183109261, 0.016514036904684420426, -0.25535793429548531153, -0.17144593839476229835, 0.61850964687874054881, -0.6298336351657808363, 0.61675702195372505976, -0.17714677236725998033, -0.26659269062284296492, 0.0078647891096843419723, 0.11789126550013073746, 0.082121610885085058462, -0.067055564825613186986, -0.38011140240006408009, -0.26716288418711608976, 0.10240779989636067393, 1.0134714906196899964, 2.2053602982154267842, 0.097220598220349230267, -5.5466620883835728861, 0.087169446340796818973, 2.1881557272389740021, 0.99827673363174562748, 0.10907324693734610221, -0.21611932143938028528, -0.25623131927049441758, -0.14869658138450347895, 0.14197047274233920433, 0.048836182531101529625, -0.038882116387869758434, -0.27454766872310143455, -0.58867186781625102654, 0.45599961077323958314, 0.51022616691992961346, 0.45997137509875751782, -0.58174972436883720661, -0.2676096120526276767, -0.033314941541957132642, 0.050270726675681665085, 0.10235606338540242932, 0.023912995852093554827, 0.087706171904705976239, -0.0089666544407731984395, -0.2597651715690996177, -0.25650958171606219604, 0.73555510185316919092, 0.82544012203574990139, -2.2458591140677919284, 0.82331050916444392662, 0.73392483209875258865, -0.252797591186242121, -0.24955186922520278259, -0.0092582531632842066105, 0.066175650116856402083, 0.011976886806703838234, -0.041006888183707797257, 0.021271207269023397618, 0.214585330971366256, 0.041478178442156399264, -0.72601159128070147375, 0.12038390806516120757, 0.73924788877320934954, 0.11936576621141610499, -0.72652886116129899996, 0.041959326466451375703, 0.21555105750099343154, 0.023801818524153722789, -0.038620107317825939919, -0.0050025997081481502968, -0.20840928720116222439, 0.096234057506007400495, 0.42445920376301654597, 0.16619854921475180065, -0.76636340368038491366, 0.0097264918144233172836, 0.56321608256382726942, 0.005818092982550598577, -0.77552439312172594121, 0.15113949455239350517, 0.41114563827151123787, 0.1150401585910676705, -0.087569729495346984582, -0.10733933591881542058, 0.081409610603563747899, 0.051855913094556088661, 0.027747543206776872915, -0.18447142704065669361, -0.5132147737083696537, 0.3981798473115655268, 0.27789008428902112202, 0.39904678634044649099, -0.50988366456598444021, -0.17811758489375301151, 0.036288619879320888062, 0.057693863997173910896, 0.062794086322834369307, 0.00043198551100311587125, -0.071200491964387616073, -0.01843387650720461346, 0.17947714272350148712, 0.1788400369293590142, -0.41347595847475610009, 0.018785569243605249046, 0.25126104304058866079, 0.018699853777923422515, -0.41426274341018104463, 0.17638659648146182146, 0.17673594355454766425, -0.015010321428958736667, -0.054762889948230822346, -0.014696283484216768334, 0.03078419756935971488, 0.24417984517557950186, 0.52040014310554738586, 0.45527737199586154615, -1.8006557266779181781, -11.636470764681908108, 24.381054945383720423, -11.629475420379407069, -1.7933731331463538261, 0.45322408012974135794, 0.50500775105828954459, 0.21963816431143357022, 0.035247860615676668283, 0.025327723038859821514, 0.027033051514795302522, -0.0355854150772208222, -0.1208033327573134752, -0.046380542056789042515, 0.40791162359865951181, 0.53706070555197327199, -1.5321855021395187624, 0.53459780648269672465, 0.40093603379155617894, -0.057305761749708139519, -0.12441453432603910567, -0.017302243099506675605, 0.05180925217247758785, -0.026554698923548211709, 0.26564303398108690679, -0.053179665809659826259, -0.41478836948944403096, -0.45899358320397465327, 0.30935493810538911008, 0.89510678398666643218, -1.0910978933037314142, 0.89997812701386692424, 0.31962389644883637541, -0.44440640691580857169, -0.40464502325394591242, -0.08240692298268770255, 0.12876448508425461514, 0.13380622265499422796, -0.168407209590784257, -0.17111545716440673681, 0.35913055282624295472, 0.62455817180701089786, -0.53679218742023793176, -0.35442122080911753201, 0.48747828075499455291, -0.35580855597955207692, -0.53773991796908293672, 0.62449897897689587989, 0.36054202910489463463, -0.16766798732768523061, -0.20045446682070516831, 0.027343463286412589969, -0.093886522412418516037, -0.077788939865369916005, -0.077248027040629552165, 0.18332100576058787955, 1.0625719191127416252, 0.51599693735972240738, -3.0253074187663835382, 0.51073910290616664209, 1.0545377719834130303, 0.17827661638375141839, -0.071560751168012887158, -0.059814000821713841916, -0.050534587125590155199, -0.05352814570892590329, -0.08154580797673430248, -0.12304522895536273985, -0.032908052627831668735, 0.17433315961823142581, 0.94198822359896539069, 0.55584721724969443635, -2.8819591094778180818, 0.55547869198195032325, 0.94687796381086997144, 0.18893616278910502215, -0.018060706296577716534, -0.13763794708036936054, -0.13111515681242524867, 0.037750537134214695589, 0.030353678908775518636, 0.005299184467888423504, -0.084850454224978194562, -0.17840049592105242104, -0.076730314858859921534, 0.56514568105873574932, 0.68497213628582287459, -1.9068245360382325426, 0.6850873256017075974, 0.56675565317291465295, -0.070910814290964643836, -0.16819964938191628301, -0.072849866117442166336, 0.013119083994982556549, -0.012269212913677827667, -0.00067741524283452918265, 0.035677115354284749671, 0.062334903749309368559, -0.074061062137224270563, -0.23812295977685285497, 0.33049461326826928831, -0.20833868500599245221, 0.33099231977813753369, -0.23659663276556514022, -0.0719268373008546541, 0.063798711764458351792, 0.036718992783886315878, -0.017870449882923598872, 0.0088148963565943434795, -0.024025317980802270157, 0.024146199061522308216, 0.058327482061620070097, -0.19433802450622636449, -0.0028697229941379519789, 0.65492642227691777457, -1.0501662864449596846, 0.65542461804285934157, -0.002485238366215078952, -0.19553435871305077653, 0.054931863448289867946, 0.025582802764135537804, -0.01290629185095847585, 0.11528806450540168838, 0.13090795836502444605, 0.04473222089733333573, -0.22264367702251841585, -0.64470622617499750984, -0.45889140364650515824, 0.89546025563438191863, 0.29490850876701207861, 0.88523521506594138319, -0.47549975729399551749, -0.65792318259546833659, -0.2153549204216550772, 0.093963708805834669779, 0.2196104282957072884, -0.045133254825012458777, 0.15092905012653737296, 0.20894387458837096472, -0.075390824624070471027, -0.67849894925014031521, -0.37964046660608796735, 0.95004651602905365593, -0.26430155555207746065, 0.94984538401942109065, -0.37630298082704399798, -0.66773022952743488201, -0.059463389454155579639, 0.21194907736196455517, 0.079736809786910597952, 0.0084364484944831723279, 0.047589906877694918008, 0.0063955912847291629397, -0.08608728867464769674, -0.14422300238193958011, 0.034639301770064946651, 0.47822939813411580179, -0.68838644165626039495, 0.47835687851851066421, 0.034084428576246161846, -0.14571078054741745667, -0.087062578741084653555, 0.010298361415980356731, 0.054825783733130795117, 0.006634716984462669942, -0.0059437707760798541567, -0.0064923400897057455261, -0.025512045029132807833, -0.12373969935817211163, 0.233438823739359258, 0.63392378183494135779, -1.422407569645826042, 0.63309124651662729288, 0.23077765705683259134, -0.12703251037049329009, -0.025977947596829111671, -0.00052536388038545024214, -0.00055535059380393075011, 0.046584469375014078696, -0.018328498854398043993, -0.14504189240845105635, -0.21772901222432061719, 0.10411811508557160255, 1.0108851833350687155, 0.60223146192453302916, -2.7569602264598329633, 0.6011150135565365149, 1.0049388393039864198, 0.09061085626058952569, -0.23361692353269236655, -0.13822780222453834309, 0.048702710090982342939, -0.067055564768848913149, 0.082121610595628280471, 0.11789126587249534883, 0.007864788839995265346, -0.26659269020333470346, -0.17714677275995410755, 0.6167570220704948758, -0.62983363513457124583, 0.61850964670088059982, -0.17144593817219372367, -0.25535793444936843999, 0.016514037069451481532, 0.10094862698545127577, 0.01019753083526999185, -0.14869658159204490744, -0.25623131887866174017, -0.21611932152340196378, 0.10907324670490381457, 0.99827673375844638848, 2.1881557274046947725, 0.08716944618398336786, -5.5466620884037496353, 0.0972205980846309048, 2.2053602986381704021, 1.0134714902195511854, 0.10240779991908127711, -0.26716288407156740803, -0.38011140240505181254, 0.02391299574021585328, 0.10235606344163855619, 0.050270726890614958471, -0.03331494189338002343, -0.2676096116703439165, -0.5817497245106377779, 0.4599713752399881006, 0.51022616646957652531, 0.45599961110364017891, -0.58867186793512193965, -0.27454766867559843302, -0.038882116522868104269, 0.048836182763379198946, 0.14197047264655091037, 0.011976886829387786704, 0.066175650146504672033, -0.0092582533502579417162, -0.24955186888118563449, -0.25279759153975200903, 0.73392483210277803529, 0.82331050938604799505, -2.2458591139733390385, 0.82544012180055881256, 0.7355551017552492965, -0.25650958149869879099, -0.25976517149804539963, -0.0089666546596130543983, 0.087706171992892295597, -0.0050025996074574610034, -0.038620107494548548688, 0.023801818566764457175, 0.21555105760356171873, 0.041959326339392408189, -0.7265288611340564584, 0.11936576623296586697, 0.73924788896079873801, 0.12038390770421869558, -0.72601159110359114734, 0.041478178456871288293, 0.21458533116109862471, 0.021271206923866293331, -0.041006888036391318797, -0.10733933599039018059, -0.087569729476555779879, 0.11504015887907960425, 0.41114563786822599933, 0.15113949476549745632, -0.77552439314255916525, 0.005818092957099008089, 0.56321608258545285963, 0.0097264915203874174887, -0.76636340305456907007, 0.16619854869263714359, 0.42445920405258646957, 0.096234057288632796068, -0.20840928710371886545, 0.00043198550857146897353, 0.062794086321248110405, 0.057693863818454080294, 0.036288620315532538196, -0.17811758514558456712, -0.50988366467838497353, 0.39904678660238257493, 0.27789008418041299953, 0.39817984738047607074, -0.51321477411857974182, -0.18447142667957105999, 0.027747543084630510446, 0.0518559131960993, 0.081409610565292223061, -0.014696283563546894135, -0.054762889643751512525, -0.015010321901951440698, 0.17673594397875425077, 0.1763865961371396951, -0.41426274312202643335, 0.018699853706646143298, 0.25126104283058769973, 0.01878556952759731849, -0.41347595869701719851, 0.17884003715985863359, 0.17947714249346199966, -0.018433876380697683528, -0.071200491992155182097, 0.025327723185455581351, 0.035247860244817283415, 0.21963816459155299787, 0.50500775075348358634, 0.45322408097629501844, -1.793373134146091008, -11.629475419902203015, 24.381054945340462581, -11.636470764893216412, -1.8006557262988973633, 0.45527737169552262575, 0.52040014328918915432, 0.24417984505609791057, 0.030784197605600107511, -0.026554699025335592316, 0.051809252523127874435, -0.017302243664324090666, -0.12441453363776955321, -0.057305762348484501345, 0.40093603400976907114, 0.53459780650410060332, -1.5321855023387089823, 0.53706070612001388653, 0.40791162296657945685, -0.046380541603026792519, -0.12080333319948854953, -0.03558541472660607663, 0.027033051402264487995, 0.1338062225929368132, 0.12876448525608880091, -0.082406923080551808258, -0.40464502333090668351, -0.44440640690758570486, 0.31962389654865319688, 0.89997812693666923156, -1.0910978934338313451, 0.89510678438779534005, 0.30935493783499851128, -0.45899358320814176437, -0.41478836957408649067, -0.053179665508270169971, 0.26564303379740450417, 0.027343463398976825679, -0.20045446707201869385, -0.16766798698928903066, 0.36054202864674861351, 0.62449897947035459822, -0.53773991842535473484, -0.35580855585003856589, 0.48747828106532964387, -0.35442122123322894467, -0.53679218687487195716, 0.62455817113071709912, 0.35913055341483957061, -0.17111545748920836685, -0.16840720951881890666, -0.053528145846270885899, -0.050534586890701990725, -0.059814000656859260541, -0.071560751735560340636, 0.17827661670762354196, 1.0545377719851680709, 0.51073910300368552395, -3.0253074189537065841, 0.51599693726800588411, 1.0625719195056304578, 0.18332100543255164249, -0.077248026928294646898, -0.077788939896093062254, -0.093886522398439878345, 0.037750537103389374127, -0.13111515671754131462, -0.13763794708808230194, -0.018060706443197842297, 0.18893616275148317207, 0.94687796409040414503, 0.55547869182840947655, -2.8819591093948617733, 0.55584721705178807838, 0.94198822385790215161, 0.17433315921940209514, -0.03290805223829684073, -0.12304522907617096439, -0.081545807988587959425, 0.043533002187340623379, 0.012494148560704342391, -0.018244732957180494204, -0.095882440788248313179, -0.17913872677091061125, -0.075316952387683322456, 0.56514568073494375611, 0.68508732557613361003, -1.9021239256218847036, 0.6977263404229849808, 0.58008313141573841776, -0.075483080257933210522, -0.19815683416110624337, -0.046509793301754601136, -0.0034838372010486144334, -0.0094538393105546604522, 0.0029750583187565507112, 0.034703570364641861401, 0.060127530184505742494, -0.075079476035970121872, -0.23812295978544120723, 0.33099231945730966942, -0.20756771025650885831, 0.33099388688140068959, -0.23759386107344010464, -0.065845894628611692512, 0.069918840632807635616, 0.0077800157351827839425, 0.0075871415412895716959, -0.0048179816265473910525, -0.018611476244091998411, 0.025839759148093010416, 0.057675722927359784897, -0.19486953654263167168, -0.0028697229998942636389, 0.65542461754841363497, -1.0495735022618026822, 0.65389908518272976679, -0.0085260717933484941877, -0.19282993209739748841, 0.082860754223370886029, -0.011320722538378236166, 0.031704611843844342223, 0.090019947130838245952, 0.11748124996189202029, 0.04519911529098122327, -0.21371301902710027054, -0.6367814077972696607, -0.45889140323142901812, 0.88523521467630328363, 0.27667416815267920072, 0.86439428367673054598, -0.49161112654983563086, -0.64941616465397433711, -0.14603841006540971281, 0.28409144930087859571, -0.057324304116588092883, 0.034307726683873163154, 0.15442396869294328088, 0.19206942650737204858, -0.085998158123825163557, -0.68101442445624194022, -0.37964046623576452877, 0.94984538408248242458, -0.26318484168499128861, 0.952069380187247849, -0.38190835758077190665, -0.68490422301624565016, -0.058415542587960549159, 0.31182914933611499686, -0.010184776969497126228, 0.022893860243744811045, 0.046782346566701077384, 0.0041891655298893840098, -0.086973921214832328186, -0.14446444066161587272, 0.034639301607624531365, 0.47835687841550006416, -0.68844201617499434231, 0.47711143794460625367, 0.028940180769227399904, -0.15503707430671589962, -0.09234858662736196433, 0.084621771909140663936, 0.0075891002269112117079, -0.0034940453361982639832, -0.0078590574910951480764, -0.0044614558667801515815, -0.02255180041797726076, -0.12232262233793231299, 0.23343882371452806535, 0.63309124614204548287, -1.4237700902759142796, 0.63222703162519156006, 0.23005781111636003367, -0.12632885561579096367, -0.021700265019528503646, -0.004149485784821442709, 0.020151254509117741409, 0.016429852557104919486, -0.024051591769888178152, -0.13659157781672026655, -0.20829690810618162189, 0.1082480630022338991, 1.0108851833148619903, 0.60111501355099727917, -2.757814286330851683, 0.59972186215158274081, 0.99846396821919580766, 0.074419905016510537465, -0.25007282932518598573, -0.053526664322602980628, -0.059620298421808964884, 0.0068081278651870840574, 0.087528515396382761127, 0.10557424695257087988, -0.0006512171915677225642, -0.26947561725363455842, -0.17714677257589345194, 0.61850964680122300088, -0.62498900622100861391, 0.62649565927017525357, -0.17060684359665229115, -0.27612782142221370396, -0.013786348993406898761, 0.15041035459944199215, 0.014190662101173555365, -0.16703556118686307963, -0.26277652364158593512, -0.2164617604276571905, 0.10881031340649482964, 0.99602509479210543653, 2.188155727539248474, 0.097220598212228725998, -5.5195894971781527261, 0.14133270085809657157, 2.2480872249987782219, 1.005020542181087162, -0.043172962429510408633, -0.59078465579839600785, -0.043719604749230596419, 0.06194045623253183197, 0.11114378696663268498, 0.052425704783254367669, -0.030857761927313709371, -0.26546751130467860502, -0.58174972471582564459, 0.4559996115302771269, 0.5013075595883578206, 0.44262301519045033071, -0.6071906594410426683, -0.2920209088820666099, -0.01732931497140097199, 0.21704701461860509415, -0.021269794712422560035, 0.037690284906462798165, 0.068374579937684559772, -0.01327595924058443122, -0.2520055773232919849, -0.25373580352800395588, 0.73392483194502888821, 0.82544012183132642324, -2.2409560094422400489, 0.82790246505871545324, 0.72230870696849391877, -0.28856947866241827727, -0.27441158207468052677, 0.12843692604497869425, 0.0028386542409670864791, -0.018176667887819675934, -0.030101944287818635515, 0.02937387084743773441, 0.21475397622920702756, 0.039778916919181922518, -0.72652886101187352796, 0.12038390775984904313, 0.73725947598497332258, 0.11428047086919798647, -0.72842094636007936881, 0.06010323084594602211, 0.23975384902473403925, -0.05447126465563887493, -0.017347937291398621285, -0.093798582087353643555, -0.076149932801433908858, 0.11320846426870605217, 0.40356176544473770784, 0.14599247749121554985, -0.77552439309173304416, 0.0097264915571844117598, 0.56909532603234669335, 0.018423999648397103301, -0.74903267426816377839, 0.19771303127592815274, 0.41411614665544066449, -0.15764715170192400429, -0.044310263844719234572, 0.032806148880766321241, 0.077530602251742075648, 0.062525065767934984162, 0.036683952962475202542, -0.17868805162191353508, -0.50988366468545887056, 0.39817984745314449757, 0.27277181369505648645, 0.3864663149072503634, -0.53076959885176344667, -0.19605212307105260305, 0.047030505272355582758, 0.1503015601874371221, 0.012032267462190651339, -0.031468148435861827139, -0.053966586394316923281, -0.012043537405517997646, 0.17743873627090511746, 0.17611632237077803897, -0.41426274356507364338, 0.018785569647449221198, 0.25089329101204682626, 0.01948347352337939653, -0.40639628874494537758, 0.19737696005213561934, 0.17948123767024845576, -0.11349197474525775375, -0.0023838599284546381496, 0.006572943559417787629, 0.058624054133009889656, 0.23236781432568914818, 0.5030557378957041248, 0.44991896804779030195, -1.7933731338346312612, -11.636470764399740929, 24.353008417373995798, -11.677774246914804479, -1.8045586146579279951, 0.5461263736525867607, 0.63689348722813443615, 0.13286372698600396047, -0.026816070803406807238, 0.022187195320206604121, 0.045031540145588744295, -0.031749537865083332644, -0.13126250548769835547, -0.058890055524505990903, 0.40093603396164356756, 0.53706070588891274831, -1.523093435653769534, 0.55264070608582394062, 0.41315131247883229992, -0.078024510869696528381, -0.17260828256767957178, 0.050618426179955683386, 0.022961354777529956878, 0.12067377154020011987, 0.11608467756211983402, -0.084086506260000304369, -0.39901027778196812568, -0.43947551429001463497, 0.31962389627203752207, 0.8951067844993905176, -1.0991845927450718534, 0.88265144748149837461, 0.28561419314202440844, -0.48761087296024424598, -0.38198102485425855512, 0.24505991996282144862, 0.08048019514010962272, -0.067319595971648993338, -0.20714183691004142696, -0.15534583408887486788, 0.36619843913757088583, 0.62433441178888604561, -0.53773991817099286994, -0.35442122100450967892, 0.48771617057345773016, -0.3526983899089316532, -0.51908065041873407974, 0.67086673831845644589, 0.37275206413936406991, -0.41279098097447869398, -0.0092623596342011130234, -0.041618578828258187552, -0.050891007669975792116, -0.063342371827541057994, -0.074574297093103023015, 0.17605856205496930866, 1.0545377719591721988, 0.515996937409620271, -3.0125483066185485548, 0.53320338984057491505, 1.071155643336735519, 0.16513009505581049963, -0.13528364922318605879, -0.12898759639653104014, 0.059842151354480135894, -0.045768209848987759236, -0.13323279563320836183, -0.12106158748323858632, -0.0071634957520468871967, 0.19189957230681639611, 0.94687796408509006252, 0.55584721687616611696, -2.8822077483616146409, 0.55313548663350453616, 0.94194578939189643751, 0.1914550707832135823, -0.018667222052198555182, -0.23517670886289224197, 0.043644153924664107203, 0.013445205990722543937, -0.0049852595730682097708, -0.023489372281868703407, -0.095656097814427173964, -0.17913872670959196665, -0.076730315216828770541, 0.56675565335149513668, 0.69772634058999993822, -1.8784277345300011142, 0.71485618532367878331, 0.55642298044223459375, -0.12806113564157894924, -0.20986266954149918051, 2.0014802864612901614e-05, -0.0031318482287021648554, -0.0090118354198800987298, 0.0020990010585648869595, 0.033919883877315440412, 0.060127529970850267238, -0.074061061909501679779, -0.23659663240195247247, 0.33099388671169038822, -0.21021727081367083678, 0.33130235964060239517, -0.24025416348139044809, -0.066485943807308017783, 0.081580212825391823506, -0.00036380404543577050426, 0.0080873222940098313249, -0.0044452891722713255146, -0.018891648668938117173, 0.02543451422751801666, 0.0576757227834357733, -0.19433802440255923405, -0.0024852374907764888425, 0.65389908470997903933, -1.0539316813449923771, 0.65614857224684275394, 0.0054687925782544270781, -0.19671150086933283929, 0.064355717949661717103, -0.020895362116386193008, 0.045373690505174937848, 0.10003194406460741428, 0.12507140475444156191, 0.050169807035816374463, -0.21371301860409336326, -0.64470622688178658155, -0.47549975683836726015, 0.86439428352336478145, 0.25977351946946880545, 0.85816180786758444743, -0.49457328664129462936, -0.60224929558935158536, 0.1366173331135781821, -0.028472233902849947151, -0.022147644423358141264, 0.038206378312023808741, 0.14851238352332452042, 0.18798364338080930658, -0.085998158073533795265, -0.67849894959381729098, -0.37630298098319109323, 0.95206938021521303472, -0.26991495386283836933, 0.9195320841808403145, -0.45677846938303739988, -0.70628018869073594388, 0.37381501788093673966, -0.011070264056780692793, 0.0022292067766418290423, 0.025171530357840976588, 0.045479295216814111247, 0.0031049020094750766352, -0.086973921571459555935, -0.14422300191027048744, 0.034084428642619436778, 0.47711143794609572888, -0.69075377204477306226, 0.4703654124021388605, 0.007260974409889614338, -0.14838389982819902291, 0.0145894782634420489, 0.0028346740094015061839, 0.0021182518804279109464, -0.0028852304802272146778, -0.0056217443162226241743, -0.0032245910364582536928, -0.022551800278730847749, -0.12373969932360270874, 0.23077765781630543707, 0.63222703123174117668, -1.4196049670333430459, 0.64089224255567878163, 0.22832781029371912451, -0.14599594691470266228, -0.01354074325637667775, -0.0015387254835810444675, 0.017110925237943032384, 0.017564534859391418842, -0.020126467703311369528, -0.13329887770462628405, -0.20829690797097025201, 0.10411811521311636897, 1.0049388393945162257, 0.59972186194932264325, -2.7494281251495618079, 0.61735531048788716557, 1.0048029276611993854, 0.051416499224076608887, -0.30418107821026074689, -0.02130561020140028794, -0.032753564030726614598, 0.0076028171255142287799, 0.082459392868851094649, 0.10231359757497560103, -0.00065121711278809391032, -0.26659269039360722342, -0.17144593839486663156, 0.62649565927589689895, -0.62401187820586156985, 0.60006683336417021568, -0.23366746301915866657, -0.28163420273550426876, 0.21389256061416569987, 0.084684294142282956486, -0.062998698780130632513, -0.19739178990869965524, -0.2677641843491993523, -0.21477483373077244089, 0.10881031381294492544, 0.99827673334468047806, 2.205360298426997101, 0.14133270095542627143, -5.4525111766994864837, 0.20658204317835227037, 2.2448731613701231602, 0.80383247665492307821, -0.57794733102688577997, -0.054842226673931529501, 0.0039788908373203091262, 0.07747501751589092045, 0.1152409588002751839, 0.053343162242495356651, -0.030857761896967983439, -0.26760961136739513933, -0.58867186875678489955, 0.44262301559975159027, 0.47852428550615833114, 0.40805606264244498327, -0.62985642294890986648, -0.19143697988248975506, 0.1802697275710452296, -0.020795617189239463257, 0.0050143620220370332952, 0.039492146432853147964, 0.063665889532881214663, -0.015848287432641251782, -0.25200557766886833022, -0.25279759115542904668, 0.73555510190496098399, 0.82790246495170227714, -2.2419555403449011877, 0.81765106635798456836, 0.68819565577034647053, -0.32247313470819533032, -0.073927702523058455575, -0.0022946522355972980706, 0.001451734174911180255, -0.013885712159576270308, -0.028572865423890141112, 0.028241562740319928193, 0.21475397609558172229, 0.041959326453759174247, -0.72601159118306091145, 0.11428047085493060753, 0.72654822041731659255, 0.11375229217733301423, -0.70574861523679155439, 0.11196210057186253839, 0.12467073058275826569, 0.023360500350815469744, -0.041693574834720331657, -0.097784906261398324423, -0.077449847541248756366, 0.11097077892253236209, 0.40356176533775900417, 0.15113949469233758927, -0.76636340320738716159, 0.018423999713874834178, 0.57567747094756394066, 0.028571803338331432098, -0.7458163869371988719, 0.17466499584718819804, 0.25067821410861151898, -0.049040763739414593159, -0.0028057949968385842432, 0.048349193803706488892, 0.080348875433604499685, 0.061538422583183760572, 0.036683952773052141971, -0.17811758466976340687, -0.51321477427543893413, 0.38646631493045913164, 0.25098851037009273224, 0.35981159908973647799, -0.54349526362812539126, -0.11962740610020305976, 0.18107273276922875027, 0.014578881512723554717, -0.00479729643336014068, -0.033566215412223897319, -0.052244976315773142272, -0.011128199837953930279, 0.17743873592578179865, 0.17638659689175484324, -0.41347595891871835594, 0.019483473682089337553, 0.25312454733122913275, 0.029506826079474357538, -0.39036802832972788968, 0.2186731060159148865, 0.01893552120638973546, -0.038769963960899198052, 0.017224085195247360036, 0.033260633261464972454, 0.069477057028997646349, 0.23168604038137879941, 0.50305573814861470616, 0.45322408055386542713, -1.8006557269854461811, -11.677774246643679135, 24.275831347980041386, -11.71355304152415755, -1.6802286737099374481, 0.76019689667294298996, 0.56127820314323206841, 0.0011838289313063979358, -0.014440279759773813617, 0.011918320252080394361, 0.036735242222306285775, -0.034242200251691889967, -0.13126250547280648995, -0.057305762284086458891, 0.4079116235167783433, 0.55264070576316914885, -1.5102399940010813317, 0.53591447537833858394, 0.35817442559803985835, -0.089303929986184540923, -0.067838372474600699968, -0.028018029276078972034, 0.05507619224347040654, 0.12532872058749280897, 0.11586584050780224242, -0.082889451242358747063, -0.39901027786398185482, -0.44440640669138720131, 0.30935493748570530581, 0.88265144771679515667, -1.1134439219820728706, 0.86515113588489733765, 0.27154700677001963882, -0.43071030015855071849, -0.13848707381989966425, 0.059312619805993761768, 0.019458194570645820365, -0.07959314409797310641, -0.20361729167279557506, -0.15276052040871157534, 0.36619843929820650175, 0.62449897896953499021, -0.53679218715596577205, -0.35269838969909572679, 0.49785885211739516754, -0.31298744737106598546, -0.43670207178514264257, 0.72433421100178652097, -0.21165918342311274003, 0.019378582349875096008, -0.022150866923408066983, -0.051451433499328062815, -0.056060754025090142605, -0.064982184119061175975, -0.074574296960119954014, 0.17827661647802556244, 1.0625719192676756908, 0.5332033899772156138, -2.9900491446493959735, 0.5492447676850962246, 1.0481456389313845712, 0.074809399132818782396, -0.20050147248471503203, 0.030005202915608042424, 0.020894822827660940179, -0.047356886189879397742, -0.12680010655432594202, -0.11721853044936297872, -0.0071634958694084156783, 0.18893616273639374237, 0.94198822402406245935, 0.55313548660111511168, -2.8699064818815052114, 0.60121684280529730238, 0.9982418950109770428, 0.12438402140185256251, -0.28677793585851191205, 0.027980563028886667326, 0.018351388700368044538, 0.0048771257075205156489, -0.0065652480210733436816, -0.023489372098793349919, -0.095882441625100775506, -0.17840049466511376242, -0.070910814732390220483, 0.58008313180839354839, 0.71485618493468527923, -1.8729602655086754481, 0.72839514137920102765, 0.54597746324960305664, -0.3690616479120815141, 0.0018549158138997149346, -0.0011161578428777261651, -0.00458853817293494292, -0.0095108452307264314263, 0.0020990012619134237534, 0.034703570486168310028, 0.062334903335888681986, -0.071926837270672158708, -0.23759386140012320188, 0.33130236015598574228, -0.21271062535146639805, 0.31334451494505305424, -0.28407349851154461984, 0.075283061901078987543, -0.0022790088902400786053, 0.0033562107100379517076, 0.0071787304988945119938, -0.0051067127516765632375, -0.018891648225923641374, 0.025839759082223700409, 0.058327481837924161079, -0.19553435924413978264, -0.0085260718913075222297, 0.65614857285112582197, -1.027070637712438117, 0.65992984828706235945, -0.12157092275686667837, -0.032215610713781733798, -0.042367856551061414128, 0.010673080121352254354, 0.063968146342877257871, 0.10772791266079476125, 0.12507140487922416283, 0.045199114950357083265, -0.22264367688102237874, -0.6579231825046079063, -0.49161112656452121694, 0.85816180806379105395, 0.27058000083798039448, 0.87849448403612939718, -0.49964852711136503727, -0.46455779783293926766, -0.0040540040391999617939, -0.027414268977091170443, -0.020348292512742893745, 0.038423987914678756073, 0.14851238324450213857, 0.19206942667326099006, -0.075390824476272155774, -0.66773022921865265999, -0.38190835761625541167, 0.91953208418852783179, -0.3298741349230119857, 0.86035218620833686654, -0.43473221095118413926, -0.2293750618984565226, -0.0066778969079360128366, -0.0049409708754529239536, 0.0043291650829200521711, 0.025012320311512438453, 0.045479295131146860176, 0.0041891659291271067522, -0.08608728928191691776, -0.14571078052824987248, 0.028940180617896167636, 0.47036541267441300818, -0.70137832162747348885, 0.48482888301931215125, 0.042947703911707532654, -0.16457317304441193562, -0.0017491840985027922763, 0.0035707103767380131247, 0.0038793788910179570657, -0.0019302119734552064534, -0.0056217438664506562407, -0.0044614561822256641743, -0.025512045079795535318, -0.12703251114604965921, 0.23005781143166115221, 0.6408922427216301454, -1.4032919533316128291, 0.64949217913586798634, 0.17044899303887797726, -0.12817232675222070548, -0.010825260830832003633, 0.0085351667240381595692, 0.021834305834374317934, 0.019229961285526696313, -0.020126467744673720572, -0.13659157786378930965, -0.21772901220782489573, 0.090610855951622742976, 0.9984639684223921563, 0.61735531047334479826, -2.708036863828666263, 0.69591187211891725184, 1.0480065435172698507, -0.40243041333646301094, 0.0057065275828472993708, -0.027997158519478683447, -0.034207144892678335857, 0.0068876731560151352451, 0.082459392671535736152, 0.10557424698433340304, 0.0078647890350074523003, -0.25535793445188509354, -0.17060684338042028663, 0.60006683317930165078, -0.67890006564672733536, 0.57486661406632633131, -0.19534630579204856349, -0.024323540929575320629, 0.086607532070970716376, 0.0093702904696351951852, -0.092825828421929545975, -0.2035006738593551634, -0.26776418429759757389, -0.21646176091210889147, 0.10907324744730217958, 1.0134714899928651821, 2.2480872251513561721, 0.20658204283233139176, -5.4132556645702232245, 0.20237098743966080994, 2.182458173739195928, 0.17488862294110443796, -0.039348327876801467151, -0.022164907985826515224, 0.01920333402160617764, 0.081711837431334810589, 0.1152409584817589705, 0.052425704962250518903, -0.033314942093232380316, -0.27454766811280140182, -0.60719065974221808624, 0.40805606249323755907, 0.45114410496772194836, 0.45856094232687966139, -0.56678130253257485993, -0.054722298934634996948, -0.0083637526899597042923, -0.010500180507554978532, 0.0057238085159191619267, 0.037853420851016485182, 0.063665889552895149617, -0.013275958939870484421, -0.24955186920700808684, -0.25650958159584130724, 0.72230870702613958478, 0.81765106639618323481, -2.2396815831676484265, 0.85353525084016335089, 0.76393580727292886845, -0.4891414066304653474, -0.0039738702188977549634, 0.0031263848108016210024, 0.0015259309033957843896, -0.015046969484579487963, -0.02857286552911184746, 0.029373871100164778564, 0.21555105738424434891, 0.041478178527304822298, -0.72842094638569254705, 0.11375229222612454061, 0.73111019050418202259, 0.10680440967344931313, -0.8109211417824392365, 0.34384741097074322802, 0.031142069992664094674, -0.0043094324336166011519, -0.052910306423799499664, -0.10061940653073349339, -0.077449847632869356362, 0.11320846448047540422, 0.41114563794092057192, 0.16619854894183672123, -0.74903267457220901182, 0.028571803568817669833, 0.55502728060549844091, -0.066207086509594198498, -0.91495983978457395835, 0.66811521818477259682, -0.03206267103437622501, -0.020323128192143991022, 0.0079904140008467752776, 0.049999966946989715666, 0.08034887516720665146, 0.062525066314117611066, 0.036288619655707080036, -0.18447142665348451041, -0.53076959869360273991, 0.35981159893789743665, 0.23277883255012382646, 0.39104203041566570898, -0.5164125950804081544, 0.055463222325982906857, 0.0087719944220510815552, 0.0064643587210836449611, -0.0074661523159895298862, -0.033658949171070533246, -0.052244976224364289563, -0.01204353697737879994, 0.17673594307127632463, 0.17884003757301300541, -0.40639628926838278211, 0.029506826561293539313, 0.25594426041447781994, 0.01967333887907375331, -0.46704776445016249475, 0.30661337615473144291, -0.048291230200929985417, 0.0018516198316949616931, 0.03397750009525130066, 0.037193151253796907785, 0.069477056574901385755, 0.23236781436758574548, 0.50500775092328797911, 0.45527737209317958866, -1.8045586144721494914, -11.713553041994794413, 24.150819512009960022, -11.837808941961634801, -1.4215147757527457628, 1.3181965812897933343, 0.017884682240873587405, -0.014958318897484126908, -0.021148562616608274078, 0.0097952668377331020277, 0.036735242377661958546, -0.031749538100580676492, -0.12441453357464132834, -0.046380542249028267054, 0.41315131308712582658, 0.53591447498174860087, -1.5451338742773401691, 0.56381257107705606835, 0.43595415375048346407, -0.22801431304074615269, -0.037858786739516973874, 0.0059757268185824663487, 0.067878614720869270838, 0.12811274862734808866, 0.11586584053505905034, -0.084086506193560228706, -0.40464502355972742365, -0.45899358268414836015, 0.28561419287628986297, 0.86515113578769331504, -1.095947874448763848, 0.96747055394569381193, 0.37494529255855268968, -0.74357682540807767957, 0.025944762881944541388, 0.038723752390104247689, 0.0094230735805455601017, -0.082103840266541328052, -0.20361729186426105742, -0.1553458341380175578, 0.36054202900725107472, 0.62455817150856784981, -0.51908065050203722191, -0.31298744768119429738, 0.55200465845175528123, -0.28837102845937601092, -0.65528212663548190964, 0.61876005766558195287, 0.028136400456220313909, -0.0032376104292113532315, -0.033316128208540699007, -0.055106713770603527047, -0.056060753986198461063, -0.063342372035546337661, -0.071560751231760644231, 0.18332100538051354643, 1.0711556433324429527, 0.54924476757903950563, -2.9803007937998273746, 0.55571027680066453769, 1.0552606677622649034, -0.16787300897578483205, 0.003103047702669438964, 0.029818292682297745794, 0.019441742328539877299, -0.047394672299768403345, -0.12680010647772130272, -0.12106158732494259889, -0.018060706523306383625, 0.17433315935331000768, 0.94194578910661885818, 0.60121684307129552582, -2.7824773927752399771, 0.63333998395426527317, 0.85213306935931920627, -0.14875887580494814855, 0.0064242685382460227445, 0.017024563900224182894, 0.015497008289179790841, 0.0048771260013398820554, -0.0049852584755516234777, -0.018244732447182276935, -0.084850455699285565192, -0.16819964843697737811, -0.075483081166927104566, 0.55642298089464992117, 0.72839514144046402233, -1.808450479029780622, 0.80336520531351340857, 0.041517090912552068183, 0.0022466273544542014057, -0.00053121140759339259725, -0.0021157083313374535662, -0.0045885381167906297631, -0.0090118356257509121821, 0.0029750579792913898688, 0.035677116118535807943, 0.063798711467112911855, -0.065845894030803861496, -0.24025416437723021157, 0.31334451537778601482, -0.23687695868528724219, 0.32060461034818221737, -0.18109805294773231865, -0.0010216906716340842971, -0.0010939286773653631064, 0.0031039855981313622252, 0.007178730544409712977, -0.0044452895642910963264, -0.018611476380508598649, 0.024146199612857105476, 0.054931863496696639593, -0.19282993156270206159, 0.0054687916076909728197, 0.65992984877590288306, -1.1008470909700867058, 0.77945063795075053559, -0.21546042999388653749, -0.038911263790551750552, -0.010208938785651121284, 0.024780680610708503975, 0.063968146036541967403, 0.1000319440381876035, 0.11748124999791230427, 0.044732221122396734203, -0.21535492086342117313, -0.64941616444381122797, -0.49457328668870298438, 0.87849448417026321234, 0.35699287047077721269, 0.96792985491572691714, -1.1566562598502154735, 0.0047265068263550313876, -0.013215633889476363291, -0.027789025071488393515, -0.02034829309760391447, 0.038206378970154900032, 0.15442396830939070229, 0.208943874578242067, -0.0594633898396407512, -0.68490422310616139168, -0.45677846916040121261, 0.86035218598256646416, -0.22803440286321077779, 1.2326639299195152333, -1.0203464360259892896, -0.0023908865651912658036, -0.0050407275968020349294, -0.0042251559885176544207, 0.0043291651865303860253, 0.025171530167085190421, 0.046782346457730446254, 0.0063955917394113888375, -0.087062578881003646614, -0.15503707393170673701, 0.0072609739433704282088, 0.48482888327857437405, -0.62106791950615003461, 0.53194887727107209585, -0.23335535854522654331, -0.0030946888348684494993, 0.0020044504170800021142, 0.0048214827361609995451, 0.0038793793599843479601, -0.0028852312902004286117, -0.0078590569633004338929, -0.0064923401670278491071, -0.025977947111029030952, -0.126328855549619673, 0.22832780980324454689, 0.64949217946153914305, -1.4045244059280019844, 0.76335656014336339403, -0.072901038250046762412, -0.0092583876225644217567, 0.00085572794417283770674, 0.012395573949501768884, 0.021834305969535439812, 0.017564534530017009151, -0.024051591484485580763, -0.14504189273184064923, -0.23361692328291921994, 0.074419904990457322325, 1.0048029276212893102, 0.69591187215367789065, -2.5921655472159450539, 0.75636298329318640832, 0.42873713796839418855, 0.013092415896399469885, -0.0084205762436491086304, -0.030683114109206689746, -0.034207144986617449289, 0.0076028170856576836323, 0.087528515534517056795, 0.11789126552686868266, 0.016514037202709858088, -0.27612782173758371185, -0.23366746273420332947, 0.57486661400426530832, -0.59562443155764499547, 0.75731622575497081762, -0.4023364075222679026, 0.054468642861345853423, 0.039889755166063162184, -0.0077359996988566881448, -0.092825828316670799811, -0.19739179019798375192, -0.26277652315586452758, -0.21611932218172863474, 0.10240780043468000748, 1.0050205417278594844, 2.244873161979836329, 0.20237098706629139122, -5.4693326935119896959, 0.33088325706739885046, 2.2998589894390852884, -0.017325991952225880399, -0.021876014482544781092, -0.013946161480735214111, 0.019203334009379569025, 0.077475017508168986247, 0.1111437870127139349, 0.050270726766015072506, -0.038882116671597466429, -0.29202090884857895281, -0.62985642256668483707, 0.45856094198552571628, 0.5492194462399642152, 0.39589640667909786753, -0.66177421942417324896, 0.001207347417531557613, -0.0096669949412666895255, -0.010780623431536349652, 0.0057238086291728079713, 0.039492146404050326725, 0.068374579784515027159, -0.0092582531696632636153, -0.25976517156578998735, -0.2885694785853883948, 0.68819565563856999191, 0.85353525085287262897, -2.0729178431109924752, 1.1176844588464283081, -0.12077755848080261314, -0.00092588012617798398727, -0.0011093803892760021462, 0.001983628090050787874, 0.0015259308317516232997, -0.013885711928699452644, -0.030101944591011491753, 0.023801818784244942184, 0.21458533099515975096, 0.06010323079277192504, -0.70574861518700393592, 0.1068044096017909117, 0.61456814457014852238, -0.07500935397118850545, -0.20072146060176923599, 0.023071747143054686241, 0.010803142008058623166, -0.012800594024924122002, -0.052910306829526743133, -0.097784905756327297754, -0.076149933264020877588, 0.11504015899969252668, 0.42445920376967755105, 0.19771303164332806879, -0.74581638727074195394, -0.066207086524784727888, 0.34621553433559315582, -0.13045366815358688806, 0.062774377469602293811, -0.010134625075752681442, -0.019673128289082967429, -0.015983847812336533789, 0.0079904138852326875297, 0.048349193937812867672, 0.077530601883944672004, 0.057693864257741644952, 0.027747542967405960934, -0.19605212318522735604, -0.54349526344219167839, 0.391042030275927599, 0.2845246373638597448, 0.35021914945373688877, -0.4717498009090562161, 0.003163559909413479667, 0.0066268090760692741417, 0.0050885758820136423736, -0.0074661521395256386438, -0.033566215545621197469, -0.053966586713089549632, -0.015010321178458843269, 0.17947714199978173633, 0.19737696075981578381, -0.39036802905774004024, 0.019673339266780741352, 0.16304867745275203328, -0.16382923530832926851, 0.090367382278001603946, -0.03158209395263378988, -0.018165835109882490228, 0.0095901959711755802029, 0.033977499852313602924, 0.03326063367593354303, 0.058624054089277871649, 0.21963816443594330607, 0.52040014324336825169, 0.54612637306947686699, -1.6802286728846311803, -11.837808942332939566, 23.462996126889112958, -12.105380962976465042, 0.75192395191203864613, 0.015083304425589003778, -0.0003267435787574629602, -0.018843446149755360697, -0.021148562289083627785, 0.011918319556482676047, 0.045031540819950921506, -0.017302243956245022538, -0.12080333254501388063, -0.07802451127548826193, 0.35817442565528267995, 0.56381257125859041146, -1.4136707526538010438, 0.65509129383910202904, 0.025047910997690131518, -0.028799839365731182139, -0.014401156825229410874, 0.015727159492709946731, 0.06787861497743390149, 0.1253287202244769416, 0.11608467763566215902, -0.082406922803993531979, -0.41478836986509159956, -0.4876108727103368734, 0.27154700675333254267, 0.9674705540553347749, -0.88145062926581729812, 1.0245218180225594828, -0.68050955047577188495, 0.0033781736816794333267, 0.024560228953050732004, 0.034349629269038797763, 0.0094230733929773213831, -0.079593143686060296993, -0.20714183715462958224, -0.16766798706379748563, 0.35913055313080705089, 0.67086673808104979599, -0.43670207114497006096, -0.28837102882587672958, 0.27569111047573374718, -0.82733629194454427758, 0.64046838152034502034, 0.021724333126687794121, 0.0088988521069038786088, -0.010416592548811283436, -0.033316128261734634375, -0.051451433458505828333, -0.050891007535916535365, -0.059814001018832281542, -0.077248026704388683683, 0.16513009485253665742, 1.0481456393003740768, 0.55571027647317294829, -2.9216274018027332815, 0.75361611494206848239, 0.66390775627404008397, -0.0058992484607116530207, 0.013775316813747336131, 0.030249284595654530455, 0.019441742189780432393, -0.047356886007281155659, -0.13323279602136933741, -0.13763794678630261736, -0.032908052593190489921, 0.1914550712230879137, 0.99824189469454938095, 0.63333998399737856388, -2.8714015925480791402, 0.71886299699489586423, 0.63653425329844715019, -0.011354832701130473591, 0.0056603853801300372389, 0.017024566353245902672, 0.018351389145844672346, 0.013445203339408517973, 0.012494148681533603917, 0.0052991856257008276149, -0.072849866584246733558, -0.19815683376595868981, -0.12806113576667521592, 0.54597746300276062126, 0.8033652055338540432, -2.0320068978135013005, 1.0409125952232392009, 0.00093869436946647153135, 0.0005260973392535745613, -0.00053121161715609019598, -0.0011161575059056190462, -0.0031318484074701214565, -0.0094538388649185525758, -0.00067741590979959048768, 0.036718993046979071793, 0.069918840288515304859, -0.066485943425024396358, -0.28407349858995806224, 0.32060461028048692356, 0.02294532653646232101, -0.086494131046632921622, -9.5681442302289297771e-05, -0.0014368330448569022542, -0.0010939289560881046079, 0.003356211072178157985, 0.0080873221641358065959, -0.0048179812672371966076, -0.024025318506626862358, 0.025582802833951402699, 0.082860753958913721529, -0.19671150050121641306, -0.12157092282594980648, 0.77945063783411083769, -0.86607847302695506464, 0.31825208493405460208, -0.019074301336998406714, -0.018948965074896642613, -0.010208939100845667095, 0.010673080188191422235, 0.045373690447259348857, 0.090019947371489800192, 0.1309079583435606986, 0.093963708894223813006, -0.14603841036512335561, -0.60224929562207352163, -0.4996485269167058596, 0.96792985493277083897, 0.36362846950722305106, -0.38446584086935525759, 0.0034879272411291965122, -0.001028683779274767833, -0.013215634320699342491, -0.027414268339948807268, -0.022147645050527391825, 0.034307727156900336141, 0.15092905009781690251, 0.21194907757439465645, -0.058415542424828215129, -0.70628018907908030055, -0.43473221054133870389, 1.2326639297262853567, -0.10547956855805161458, -0.25821309804967601531, 0.00042517569673810223828, -0.0027145039847582450553, -0.00504072768958882967, -0.0049409710966595438153, 0.0022292070868160391434, 0.022893860262650632398, 0.047589906652365669693, 0.010298361507235304774, -0.092348586898504697595, -0.14838389949258171119, 0.042947703658777147684, 0.53194887729444007007, -0.64534577850616070549, 0.24421377933335725041, -0.0012948050397168214126, -0.00035218630114272744596, 0.0020044504191056426202, 0.0035707098505619043645, 0.0021182526372767714398, -0.0034940457723654058837, -0.0059437706673088489331, -0.00052536417630716033839, -0.021700265101467178241, -0.14599594649114278067, 0.17044899269682622922, 0.76335656014339681175, -1.3601447492799176864, 0.59963863803355299531, -0.0014371894394043017475, -0.0026073548536135189406, 0.00085572784628575385146, 0.0085351662644455993795, 0.017110926092234843288, 0.01642985200615446062, -0.018328498642313846567, -0.13822780227896050431, -0.25007282936190261546, 0.051416499331039360043, 1.0480065434865437624, 0.75636298329275886143, -3.052294895895346194, 1.5657927856037754388, 0.0089092122780350720829, 0.0048977817254549698908, -0.0084205765373181092059, -0.027997158642130239026, -0.032753563590226908442, 0.00680812763356040615, 0.082121610825044988324, 0.10094862696738531038, -0.01378634880765034626, -0.2816342029669643976, -0.19534630576626083043, 0.75731622586387714602, -0.43317481576802691068, 0.027955634474297670622, 0.013633506435024140072, 0.036771469993355752892, 0.039889755094004109015, 0.0093702902229862994332, -0.062998698261595792447, -0.16703556165636998521, -0.25623131856107650428, -0.26716288442302765782, -0.043172962056313851642, 0.80383247630552467911, 2.1824581738099233519, 0.33088325717825650685, -5.8653396568774738284, 3.2171626856193147326, 0.0039976893097151923115, -0.010742032040404414506, -0.021876014647157021598, -0.022164908224229867839, 0.0039788913778756040962, 0.06194045588074956199, 0.10235606356609988632, 0.048836182680377177023, -0.017329314890709546226, -0.1914369799628453106, -0.5667813024991895654, 0.39589640667335329605, 0.76973654434871807695, -0.55873073236078996118, 0.0044719043888637411907, -0.0024897083292167443425, -0.0096669949603377554326, -0.010500180480645214184, 0.0050143620099220836267, 0.03769028499431138074, 0.06617564999013154925, -0.0089666545078153601922, -0.27441158213468874738, -0.32247313464081028833, 0.7639358072439017544, 1.1176844588200745001, -2.5227289380897133242, 1.1623714310008792516, 0.0021567801066694506848, -0.0019412149038606435827, -0.0011093804861901385844, 0.0031263850140005257534, 0.001451733899723786983, -0.01817666762119905105, -0.038620107651822638273, 0.02127120713953699685, 0.23975384922182399472, 0.11196210033592032818, -0.81092114166493656224, -0.075009354049683063526, 1.2409997262847816124, -0.67855335663262017487, 0.0082107527222490175184, 0.012752791429474578597, 0.010803141643917747727, -0.0043094317515105368122, -0.041693575505076144183, -0.093798581536654462854, -0.08756972982977125064, 0.096234057538490708206, 0.41411614656050216299, 0.17466499595239795561, -0.91495983971314598371, -0.13045366832620572395, 1.2382510230510657134, -0.69762626654810311333, 0.0081356190807622928707, -0.008364010558639646542, -0.019673128423674905718, -0.020323128317771697177, -0.0028057947174158631709, 0.032806148894613536282, 0.062794086188795950054, 0.051855913222736936541, 0.047030505298460575925, -0.11962740608197733327, -0.51641259513422477223, 0.35021914945156135124, 0.57370306773124379873, -0.44641881400488608866, -0.00050654872292098859082, 0.0031595932601335267355, 0.0066268090127082042787, 0.0064643586507360534937, -0.0047972964067739534352, -0.031468148227382808368, -0.054762890059417645983, -0.018433876078055536074, 0.17948123724398240553, 0.21867310650250296189, -0.46704776478602505474, -0.16382923521505624032, 0.62891606882944661727, -0.30690727836238373971, -0.0061413641160674402319, -0.021023236115997361723, -0.018165835053591913861, 0.0018516196447909971656, 0.017224085298240876529, 0.0065729434068233921762, 0.035247860555864915511, 0.24417984501878381431, 0.63689348745931162465, 0.7601968962666977303, -1.4215147755266515084, -12.105380962931826971, 21.21280911030951799, -9.3614052844720081481, 0.0027808329322660148286, 0.0068153427244642336702, -0.00032674348623933785474, -0.014958319546080636742, -0.014440278783284007935, 0.02218719464539475883, 0.051809252829185040157, -0.035585415153535054911, -0.17260828263194058496, -0.089303929624446540259, 0.4359541533690347026, 0.65509129404185151291, -1.5977616549989299699, 0.75363966604906651447, -0.012376395965119479231, -0.017428998096977524335, -0.014401156701010976724, 0.0059757264208251115178, 0.055076192788788572163, 0.12067377124227643415, 0.128764485134909068, -0.053179665436710779614, -0.38198102516440524123, -0.43071029995737503127, 0.37494529245546437402, 1.0245218180961987997, -1.3728130617168134986, 0.59826111648947044586, -0.0086821925954931843117, 0.0051776617836572468023, 0.024560229035205341508, 0.038723752455761602242, 0.019458194256673597144, -0.067319595752064725902, -0.20045446697669705438, -0.17111545739075459394, 0.37275206438731145475, 0.72433421050933932683, -0.65528212627830872972, -0.82733629196134861328, 1.0931999072701004128, -0.35060649408286198581, 0.0076887503119230770493, 0.011992958870430952753, 0.0088988519929725923285, -0.0032376103097762039609, -0.022150866994316516234, -0.041618578855302706954, -0.050534586862231618831, -0.077788940036686071711, -0.13528364892614685999, 0.074809398756098433036, 1.0552606680279457141, 0.75361611480538093222, -3.226714082834987618, 1.6389303590440165159, -0.0038474666896199539307, 0.00043179288703107247035, 0.013775316518253246864, 0.029818293093969178847, 0.020894822424869282512, -0.045768209422280346776, -0.13111515709467164159, -0.12304522886514063229, -0.018667222143711595383, 0.12438402145303585622, 0.85213306942536282129, 0.71886299689778854205, -3.0045206824685495306, 1.5658465288990053299, -0.022577474126266781906, -0.011354831607049743064, 0.006424267014184619623, 0.02798056253797526205, 0.043644155211674133532, 0.043533002116013221061, 0.030353678508996041752, 0.013119083916226039616, -0.046509793250948734677, -0.20986266952313606393, -0.3690616478434369796, 0.041517090813336023081, 1.0409125952508437862, -0.57582713431940768523, -0.0012574764020709951976, 0.0009386943671916556705, 0.002246627485747685303, 0.0018549155584675247783, 2.0014968909001589446e-05, -0.0034838373532840158017, -0.012269212720876117306, -0.017870449980932481132, 0.007780015815283924073, 0.081580212845650923459, 0.075283061784277849338, -0.18109805286850347295, -0.086494131060187370852, 0.13615443481841749129, -7.9773333013209615349e-05, -9.5681474606025806442e-05, -0.0010216905118787468661, -0.0022790091243794627016, -0.00036380389229386314587, 0.0075871413893354227795, 0.0088148965197893877632, -0.012906291875222145049, -0.011320722514203728284, 0.064355717973985898661, -0.032215610842031269789, -0.21546042986710020606, 0.31825208489380529819, -0.12287864962082288567, 0.0089660944878357899668, -0.019074301472251087297, -0.038911263677358129409, -0.042367856433778876279, -0.020895362068367374919, 0.03170461168992075307, 0.11528806445639346034, 0.21961042837482525059, 0.2840914494443413929, 0.13661733308565277012, -0.46455779802295688041, -1.1566562597953511382, -0.3844658408373097247, 1.4045124123651244741, -0.0013536372264002811006, 0.0034879271454829369366, 0.0047265070518036477604, -0.0040540042436096809797, -0.02847223368838706567, -0.057324304333205934603, -0.045133254810097847565, 0.079736809751177112338, 0.31182914926508492615, 0.3738150180567474945, -0.22937506209049993156, -1.0203464359209106771, -0.25821309807636111389, 0.9082807487912136768, 0.001797653998286230341, 0.00042517562041837002025, -0.0023908864923810513488, -0.0066778967863974870653, -0.011070264182469495909, -0.01018477700154223585, 0.0084364485565858998023, 0.054825783731064822724, 0.084621771956250479985, 0.014589478167182047194, -0.16457317294755091641, -0.23335535852875771701, 0.2442137792902805693, 0.027313516638784757218, 0.0020031980951485143724, -0.0012948051162588100887, -0.0030946887870435114791, -0.0017491838863238698409, 0.0028346737232720071835, 0.0075891003545617017126, 0.0066347169561753131101, -0.00055535046876117649586, -0.0041494858079833846914, -0.013540743390313849867, -0.12817232662077213123, -0.072901038210863980393, 0.59963863797460537075, -0.40001551091001630311, 0.0085247061445511574523, -0.0014371895375825834412, -0.0092583875661503799159, -0.010825260577249536109, -0.0015387258979117478946, 0.020151254759515572351, 0.046584469352860286007, 0.048702710093993059615, -0.053526664312782502864, -0.30418107829479579252, -0.40243041331261936167, 0.42873713797133267134, 1.5657927856188564864, -1.367048640824480632, 0.0004355851941747800396, 0.008909212131166837495, 0.013092416076706149194, 0.0057065276871175272394, -0.02130561041770075853, -0.059620298353693812476, -0.067055564806224099428, 0.01019753081304864209, 0.15041035455293522149, 0.21389256069942244598, -0.024323540932003697573, -0.40233640760421179872, 0.027955634521809154952, 0.14423778954077401404, -0.02423606308998388012, 0.013633506365461149287, 0.054468642903756317453, 0.086607532171496859341, 0.084684293946221234073, 0.0141906622856433004, -0.14869658170027991173, -0.38011140228161660604, -0.59078465590914208683, -0.5779473309753747623, 0.17488862305698585353, 2.2998589892842540294, 3.2171626856831072594, -4.3862861904316829964, 0.022381060156372055425, 0.0039976892058877395997, -0.017325991851916161535, -0.039348327755009009088, -0.054842226925494529099, -0.043719604585518240247, 0.023912995729835413716, 0.14197047273495824737, 0.21704701456936523196, 0.18026972748064701291, -0.054722298794966033053, -0.66177421948212089564, -0.55873073235476022891, 0.86507390004039708131, 0.0033401591823122842896, 0.0044719043867359606001, 0.0012073474729394405389, -0.008363752750278116016, -0.020795617204456624122, -0.021269794739752212465, 0.011976886912246961087, 0.087706171909449487378, 0.12843692605059275902, -0.073927702518355842276, -0.48914140660171684383, -0.12077755848689267215, 1.162371430990611465, -0.67057132656709805829, 0.0027402548131120062387, 0.0021567800855672395605, -0.00092588003699833193449, -0.0039738703430929169305, -0.0022946521257105859277, 0.0028386541388067520336, -0.0050025995637754125736, -0.041006888122489239257, -0.054471264765317981005, 0.12467073071862308509, 0.34384741092706078147, -0.20072146056727857566, -0.6785533566590437049, 0.51825751168996836427, -0.0088453358374946319809, 0.0082107526654820340434, 0.023071747344198815372, 0.031142069692393522379, 0.023360500617208308088, -0.017347937524579193358, -0.10733933579766792576, -0.20840928719970860938, -0.15764715173693014116, 0.25067821413394486552, 0.66811521816437879906, 0.062774377553404869801, -0.69762626659617965608, 0.10715660952012208418, 0.020468559465814505777, 0.0081356190118610262341, -0.010134625033679095507, -0.032062670906292702055, -0.049040763909406828724, -0.04431026380957297367, 0.00043198547460228560457, 0.081409610660396397597, 0.150301560158316444, 0.18107273269503459945, 0.055463222451962577997, -0.47174980095312335493, -0.44641881401340094415, 0.56491691766281182385, -0.0027231438805823031737, -0.00050654877623462898559, 0.0031635599888717234711, 0.0087719944139611066847, 0.014578881521666392507, 0.012032267375587187355, -0.01469628342448324812, -0.071200492065814871556, -0.11349197465008503682, 0.018935521068865396321, 0.30661337626023105196, 0.090367382247680760887, -0.30690727836097225767, 0.048048473040508084264, 0.016709788733304014458, -0.006141364133967343146, -0.031582094012036654451, -0.048291229995809356601, -0.038769964103357999752, -0.0023838598370935666637, 0.025327723027476673617, 0.030784197608769697102, 0.13286372702408150159, 0.56127820317214227597, 1.3181965812500853197, 0.75192395186312510624, -9.3614052844341255621, 6.7883135037307962278, -0.010697873572743222745, 0.0027808328913630254997, 0.015083304386954375295, 0.017884682583818201085, 0.0011838284926932451935, -0.0268160705670157061, -0.026554699130774437471, 0.027033051505921931734, 0.050618426299994800921, -0.067838372661371243444, -0.22801431288660484187, 0.025047910897958883952, 0.75363966607952825871, -0.54214329072817624144, 0.0062816551132429537807, -0.012376395955461392401, -0.028799839417586174661, -0.037858786547880690643, -0.028018029543217238403, 0.022961354963043516253, 0.13380622262333474737, 0.26564303374973535776, 0.24505992012333122076, -0.13848707391824027768, -0.74357682543412750853, -0.68050955045923122722, 0.59826111649469537745, 0.45103589512525887883, -0.015196447415760495556, -0.0086821925632686453023, 0.0033781736231529631981, 0.025944762916060303687, 0.059312619858745516122, 0.080480195110346042231, 0.027343463326091103915, -0.16840720955246540891, -0.41279098103700873068, -0.21165918326821267503, 0.61876005752100937674, 0.64046838151159846131, -0.35060649404002497409, -0.30708086686494601159, -0.0083730426981859663177, 0.0076887502625468690659, 0.021724333215876776709, 0.028136400399575975384, 0.019378582368604600067, -0.0092623596431427555886, -0.053528145802900300998, -0.093886522348803291838, -0.12898759654332736524, -0.20050147235089116338, -0.16787300902033563954, 0.66390775632149989782, 1.6389303590125603449, -1.7735106281795913752, 0.0018542676993145866451, -0.0038474667652315824341, -0.0058992482676342109654, 0.0031030474815406799971, 0.03000520309582986947, 0.059842151199273926898, 0.037750537262277734929, -0.081545808017809418011, -0.23517670893704387125, -0.28677793581213689711, -0.14875887584660923979, 0.63653425334867008711, 1.5658465288828753437, -1.6173518744831614224, -0.37660767885644108599, 0.87380200493314796883, -0.078366127641851232588, -0.37032573486404318786, -0.16776826442432107855, -0.0019161506962036244095, 0.042630585965778046431, 0.038448413622416396929, 0.035472517927034361018, 0.029801800017811779647, 0.016519662235867840078, 0.00096906602486873953586, -0.010428101954724237552, -0.016193745899715818182, 0.1496670886302638448, -0.11751239148564314452, -0.16764883173710398712, 0.089337701762789031856, 0.075230609504625955974, 0.0015300530973238986917, -0.017879099806191720012, -0.0105391907033851491, -0.002236396307116173942, 0.0005299634643097410943, 0.0016700878491026349953, 0.0017399466593513589121, 0.00061062322235679847206, -0.0010808800372152762084, -0.11843490058743011417, 0.30445719465038800777, -0.20164456042207998254, -0.032389606545004979987, 0.056086596913720566937, -0.0096925758261083762996, -0.0092850481836280109732, 0.0088022918878498408041, 0.0065056042779991600511, -0.00068570146708177889382, -0.0022132062620093938959, -0.0010129405411859325904, -0.0001192840807955957913, 1.1285400971402647919e-06, 1.2744046495428174026, -0.39737234675819049778, -1.0448288699589036721, -0.3747672803962031618, 0.14425377799235572507, 0.25361276769762303163, 0.189496025066666407, 0.096157548475019352652, 0.023156962862245975693, -0.021580536524582350638, -0.038453718187906965609, -0.033679551717556939494, -0.015337906031187458974, 0.009340790655581859489, 0.71477822042268646108, -0.13422367188051356757, -0.87560778713688669495, -0.21061706374007072018, 0.31455317229598900308, 0.25881713097167757809, 0.064948092809753096843, -0.034533806764171726122, -0.043560141744458318047, -0.022425367539640601172, -0.0049241379153243244535, 0.0017700394695865163171, 0.002009221019400090312, 0.0001185447659230402101, 0.043998728127749339722, 0.23184292010978418963, -0.25448880724552180865, -0.16654112832557829704, 0.031366571792232240956, 0.097496339616897312319, 0.0544623437191665144, 0.0013088334687463573994, -0.015708498184839714273, -0.012248569446313651785, -0.0052276704903100011668, -0.00069528414747491119538, 0.0011528931124235510409, 0.0012956377998830318626, -0.37929277904780145381, 0.55942155095343881754, -0.060348648792925126128, -0.11224133327655785863, -0.01825790677022558206, -0.0088583583605057504562, -0.00084938132132694526763, 0.0066729684066916799406, 0.0074823993307333138206, 0.0032862877684002758487, -0.0010267720780637722578, -0.0026759681412057085073, -0.0013722583825938067942, 0.0015425411076538566701, -1.359954450700528783, 1.5474697281122515857, 0.42485982610010358806, -0.38860276797550996619, -0.29156025982208810809, -0.052759546882589193617, 0.041928166603980784877, 0.039247930379442456938, 0.016319230897646790773, -0.0010652437964648047049, -0.0078481966023885903899, -0.0063673795908322566514, -0.0004311103622597383311, 0.0069153436158379492232, 0.22631878373846853192, -0.1218334136123319783, -0.39446515434463302885, 0.050083593416958553657, 0.22929286499076012573, 0.13415618574230708804, -0.001728060249504154329, -0.070957616277038976227, -0.060505768123784907375, -0.021276917454205103442, 0.006661231731189628906, 0.014691186996727650244, 0.010047606274463211432, -3.0492289598066602213e-05, 0.85298840308298462531, -0.5394401672080499921, -0.65359014583086727512, -0.062846167107879560842, 0.17169854912346135811, 0.21344368480185543779, 0.1417412916724815064, 0.025935140905617465323, -0.040913546494417536714, -0.052967078949773593877, -0.038958560346515069828, -0.017863892055616142907, 0.0033026032612239152042, 0.02202909761164495786, -4.3905989899472697147, 3.3751799763942909038, 2.2620335753167104542, 0.065094673281344703208, -0.62459410011393379225, -0.59790533427515757392, -0.36924112952469556603, -0.12629776733466252514, 0.036279517223563928496, 0.097603933417052229915, 0.089440285884046621945, 0.050817804067563365766, 0.0077580787124647901387, -0.029200535389488788129, 0.49433610617710749269, -0.65691505123814852851, -0.19319414263296080803, 0.33570264811927735815, 0.12497722197026477076, -0.047820343881359757277, -0.039665199113730835456, -0.0088807109872345042084, -0.00079167872880793422816, -0.0034681137571480748043, -0.0034770754296553994106, 3.8754639349762370557e-05, 0.0028577062439375799929, 0.0028120714994557495167, -0.67208592916399578776, 1.2326278599138627001, -0.18919927533524408259, -0.5413674314544091315, -0.041052566864888360987, 0.16106091959076923592, 0.088328592613905479447, 0.0014149777004759170589, -0.026835312046995057206, -0.020179637785882522522, -0.0058571136000614863401, 0.0026131653492056203421, 0.0042656502182057010747, 0.0021264947076152828156, 0.080528393095039285665, -0.64269926406429433996, 0.11093546267573814457, 0.6306921971400795135, 0.18505286850226457251, -0.17862577857522241032, -0.18785874570374547221, -0.080391510322878168004, -0.0037635186331376078647, 0.023727515718404139367, 0.025801911113398291048, 0.018001651787976920283, 0.005949900793905114546, -0.0077856277522834332894, 0.47726604269980232864, -0.42287026199956834205, -0.3693506591765349234, 0.10237172635703795098, 0.14630796400329043605, 0.098202847451233282738, 0.054004747581415976643, -0.00089569285612323308599, -0.033069197868186905875, -0.036437453614861850526, -0.023462166486475390798, -0.0069192299952025313262, 0.0064302587643119595365, 0.014888788834112744175, 0.047530875712866578475, -0.30918022508162151807, 0.10774640663366776006, 0.30358168667756274539, 0.0015099425807452283941, -0.11922878680731945067, -0.065186365775724397498, -0.0078340845399893430101, 0.014676171936671490378, 0.01380111680749575645, 0.0069207594917129423503, 0.0018508065358688119764, -0.00077770012052050642784, -0.0020551012166838294494, -0.54214329077306822047, 0.75363966616377631169, 0.025047910895031000322, -0.22801431293518592547, -0.067838372648427042177, 0.05061842640978354152, 0.027033051240169855139, -0.026554698986827808549, -0.0268160705084496305, 0.0011838283957280280878, 0.017884682699695567815, 0.015083304311095075551, 0.0027808329202963456492, -0.010697873593119598679, 6.4465692434071275585, -9.0964316145238157674, 0.93442712499246904923, 1.312599528483175737, 0.50225150596400192082, 0.084593883386294990712, 0.012994018795896285917, 0.028213876911346014859, 0.0028579245489360749488, -0.037618301116165972497, -0.048650667549796827727, -0.030897783850584985971, -0.0044582207466017653719, 0.01829829239019897455, -0.30830874810473440917, -0.32665486381846930986, 0.63349570601640636802, 0.5943010881911693355, -0.21859162251875324801, -0.40436524225389425258, -0.1605251380112336923, 0.030188118416993606208, 0.080643839313546919478, 0.058562330778971591339, 0.024803730387313243505, 0.0022589485015148940555, -0.0091967303175409347016, -0.014612180556381030444, 0.52393703136577052248, 0.57895024503423542228, -0.83439059247740932079, -0.78088745786826863071, -0.03238496336486045013, 0.33078941580633464126, 0.2740526164553216959, 0.10371402864267202648, -0.0060482606355759180317, -0.042542739800807381767, -0.040426313290239518361, -0.025555229395724576857, -0.0076110967618100390172, 0.010463248120902608751, -1.5756324150518414839, 1.4928377584892875696, 0.56722175160825538676, -0.16735150332830692199, -0.18175946152923760923, -0.11421072413854890804, -0.081781251169772292697, -0.044638060871973625487, -0.0044315553996106035045, 0.020197126048895317624, 0.025924674028111657081, 0.018397727057325844596, 0.0053059938299971380268, -0.0083966519341664644371, -1.3968247951030736687, 1.3894628046947081756, 0.50290835701774982969, -0.14331348203638943906, -0.22824066789357075336, -0.19020593458255885544, -0.070367761503979678328, 0.02739133450505584419, 0.047604826608107331243, 0.025154596425285790512, 0.0042617004840670089094, -0.0032610289818124139384, -0.0027233600225226957045, 0.00039209692281160895546, 0.8738020046360107651, -1.7555349671510986909, 0.79934735067308193557, 0.43530767639869832886, -0.16293144729224451339, -0.19298145760291909867, -0.05831935451171723489, 0.02190912689490366258, 0.024051824217083371538, 0.016089717860281883832, 0.014388494008440746302, 0.011340226139748756096, 0.0022359143462998181187, -0.010428101971734339212, -0.11751239145148276166, 0.099532992254500254892, 0.27277671581388740485, -0.30382299512103561057, -0.043112924167123319041, 0.075320982589151458386, 0.030834208715279420865, -0.0029334607907078600122, -0.0091297921029115693126, -0.0026639569293918346354, -0.00064175216817472392067, -0.00020599860355828200891, 0.00049384591170501066643, 0.00061062320050645250372, 0.30445719463511033975, -0.82361966122083307695, 0.73497862870416585945, -0.11439685980440489343, -0.17758779763973642218, 0.073858531453643183617, 0.019962524216250471243, -0.020756790312873626969, -0.0028654870469137188986, 0.0074644305817896956096, 0.0027445251321960469071, -0.0011056320833711253803, -0.0013165163764829639417, -0.00011928398480015538056, -0.39737234692991041385, 0.4091780513817128706, 0.86958411257182355314, -0.51244384107676532736, -0.52990155081946344051, -0.10690377337414405656, 0.090328538937286109722, 0.11529838131569536408, 0.077178212181351957266, 0.037795484892245295727, 0.007548542074243484723, -0.010175142783792624421, -0.016729882006917103782, -0.015337905744978707831, -0.13422367182691091747, -0.28053812643053765985, 1.1557506189959716458, -0.34966925932582287873, -0.62266305411995925834, -0.046498310505336371001, 0.18209867589752931627, 0.12319925090065496354, 0.025835440567880209317, -0.017434091391687642908, -0.02069187661837015163, -0.010099176868706528226, -0.0013117338424912957338, 0.0020092210692592769698, 0.23184292011359408647, -0.63924099724712712955, 0.55287554014193862617, 0.040749018668029407531, -0.16856239195433123101, -0.096994648319525658686, 0.019432076857489110483, 0.055176281889621536791, 0.02328354996779873412, -0.0012093798866722616642, -0.0076224792320423980299, -0.0055178900622939314191, -0.0019310754868073283681, 0.0011528931584933168816, 0.55942155106048119251, -1.2642987719807283753, 0.70751397225833656535, 0.14754692482018358368, -0.12155818524633431321, -0.017559334785225328562, -0.0059807389202044744514, -0.0068677343138717353646, -0.0027200240795892216376, 0.0020612293405416343139, 0.0033402167965407179515, 0.0021555308762598616654, -7.6127595107713201257e-05, -0.0013722583902233086422, 1.5474697281090088463, -3.0230652614412059087, 0.75860904726810596266, 1.0319487000529738996, 0.045164740349270447728, -0.24340531241964624698, -0.13258658053085289885, -0.01746217713309659858, 0.014642219291502448636, 0.014510809382396131634, 0.0068286615868328060658, 0.0008227666359870069307, -0.0015663723084825887689, -0.00043111023465611426637, -0.121833413609937255, -0.12175618642598097507, 0.64314459271828650078, -0.31776009100046670452, -0.23662084776320202462, 0.028518726038237952186, 0.101033346827111431, 0.072589812142513282511, 0.0023570210924492331073, -0.033348364476530098499, -0.027914069043796951863, -0.0086238316278797050307, 0.0050088821341224241956, 0.010047606175322991093, -0.53944016716398479616, 0.74042292172130563799, 0.39510087558773365757, -0.55075640474229148325, -0.18948645988450940614, -0.022280749422536161308, 0.046481373364957417893, 0.10134407203431666789, 0.061454303304609325664, 0.004674627954161097998, -0.020853319111503121341, -0.021017680598399535175, -0.010737624418565711037, 0.0033026031261462196476, 3.3751799763874057447, -6.154553591293538517, 0.45839619221811694905, 2.2875207405319910592, 0.73774026125159097855, -0.085679259158058296286, -0.27398330998879943943, -0.25813394855034599296, -0.16270875315203059408, -0.052789711045551829249, 0.019040257456785031343, 0.044773108852092582954, 0.03597025217837160338, 0.0077580786732602911165, -0.65691505113022097273, 1.1958875683971985637, -0.070254193402658884904, -0.7826652341229027865, 0.11048381620246604251, 0.22825367213729577665, 0.021862574413574427379, -0.033713203818915099541, -0.017512774446392889716, -0.00063985795425419884908, 0.0015050359009035878949, -0.0015130562684146442801, -0.0015878381539773060022, 0.0028577061352937339032, 1.2326278600229749749, -2.7192590773028744167, 1.2693930640680290978, 0.82223057709183644537, -0.41736006860768937177, -0.30060682966572788155, 0.019428202371438003937, 0.083530085380456683963, 0.036817625255744626522, -0.00056779010817586744383, -0.013224466108169291925, -0.0092568208470446295549, -0.0013567277710479241394, 0.004265650124484637179, -0.64269926402466215354, 1.1940273896157229672, -0.20111385230612924269, -0.86403981668730922561, 0.23070430700198876761, 0.39534193749779017457, 0.0586747679801385702, -0.098936614700851077675, -0.084040079863731895204, -0.029256500713065625163, 0.0016111978194122981902, 0.010537275876607968966, 0.0099699327239434459152, 0.0059499007725792741469, -0.42287026182389453322, 0.59286980059608573868, 0.25497276100373794661, -0.51712661695050787802, -0.051827194098686336921, 0.072746658673162989395, 0.034469533571672844596, 0.040026008302804644279, 0.022434637162451944586, -0.0020919452706307170575, -0.015079746110720871483, -0.014769736612635097489, -0.0062442929802111711568, 0.0064302586253491434559, -0.3091802250101735039, 0.65438702986956986951, -0.20421150301194310184, -0.46160953893132056214, 0.24272242872235297306, 0.17623718162403398391, -0.029941836053938079149, -0.057768581433189920837, -0.028535914536745470016, -0.0016759157737045770123, 0.0076515193856410528006, 0.0062300872777652921419, 0.0023250935779065588435, -0.00077770013245901346843, 0.75363966608882060338, -1.5977616550945694662, 0.65509129409552424583, 0.43595415338159987373, -0.08930392968235309481, -0.17260828291643270926, -0.035585414366028272903, 0.051809252363325800084, 0.022187194580961105972, -0.014440278614366183763, -0.014958319731946834175, -0.00032674351168507365095, 0.0068153428913801175626, 0.0027808328832322455962, -9.09643161448890325, 20.738187581936426085, -12.093618335072768843, -1.2280636129421578229, 0.82312588833257915866, 0.63458064417873738527, 0.21500566634000112254, 0.0069308577638879931282, -0.0024130722420434846942, 0.02009605724718209474, 0.0049732671456469701443, -0.017604083962554910314, -0.020943719446437043114, -0.0044582207735869754453, -0.32665486396814535919, 1.0465952443307651176, -0.81594507184620279627, -0.62495977812435010712, 0.71667925318438086002, 0.35555853397456527176, -0.17290750710730862227, -0.19512217294664147005, -0.06414151252907085432, 0.020076604865090386109, 0.038572356920407023517, 0.02436114768044951584, 0.0049033675944998578178, -0.0091967301483596511358, 0.57895024487380353229, -1.4796034411563536715, 1.2393645914138080588, 0.3860807862731841511, -0.58499487141089590025, -0.40664939379486531035, 0.0088427383240661794339, 0.16896907963525753527, 0.12194882960932779126, 0.040149967586694049193, -0.0071499012107938825222, -0.019864762785646167265, -0.016385343479091367402, -0.0076110966944356586369, 1.4928377585662435667, -2.9508352696993247477, 0.73483897255562868889, 0.93930738449503681231, 0.048297050709144383196, -0.12552424031465789223, -0.068588346808793693166, -0.043392315487376216521, -0.036339936451856638677, -0.018873533162320085821, -0.0014540693741298578445, 0.0092180409949838768013, 0.010823466890210135111, 0.0053059937692613023952, 1.3894628047456207831, -2.7112554978118965643, 0.73760296517312773368, 0.73061596379178050942, 0.063551495940033966847, -0.018346305175408958443, -0.095606416462670110179, -0.10695814848513167383, -0.038809717033296241284, 0.016110141207853152789, 0.023784174097695912065, 0.011294394441202341289, 0.00085334862941077163287, -0.0027233600289392153639, -0.078366127345766284362, 0.79934735009479351842, -1.5533876591982052062, 0.73201458326843760194, 0.43496554105655405031, -0.13047207137260385945, -0.17392930996932787213, -0.071470239038407046284, 0.0020035279897382719805, 0.011407768120260822556, 0.011520069941735390159, 0.013697188383122341909, 0.01134022611511923545, 0.00096906608449701602762, -0.16764883187626578032, 0.27277671604791420279, -0.17286664074937119207, 0.2861303693930865033, -0.26025200177382246114, -0.046902453524512069394, 0.069374758626597879396, 0.030029050419887730328, 0.00021096832061045893425, -0.0087889507467884195324, -0.0038519186418987266286, -0.0013990230042142206592, -0.00020599821954493603877, 0.0017399464538199811876, -0.20164456043081774306, 0.73497862861000906598, -1.0459053398077791108, 0.63137776960695224382, 0.0012043110160808616956, -0.1770392373530119412, 0.049240163075608245347, 0.019209675319758898887, -0.016111503654675108504, -0.0026464415330858500133, 0.006763808964704007276, 0.0026067890468790232139, -0.0011056314543906647831, -0.001012940868944033632, -1.0448288699226089271, 0.86958411241354571786, 0.34547880447342932397, 0.79544488494369947063, -0.5028444344223763629, -0.57595902945412769114, -0.1689082372152230227, 0.048636881201923262696, 0.10559340687483093069, 0.087229944506834911078, 0.054397124024947840493, 0.01974007121234140974, -0.010175142357692939371, -0.033679551934934652546, -0.87560778720862053603, 1.1557506190465520746, -0.39669655404853626512, 0.82146417567687468608, -0.3525204902845104904, -0.59841169788826786835, -0.051245636814855347052, 0.17565721443137924385, 0.12541360927980785278, 0.029967004288666422573, -0.014781621915663734068, -0.020176069912143632795, -0.010099176944276241907, 0.0017700394544106464248, -0.25448880724639089124, 0.5528755401403179226, -0.60892098807840144836, 0.48804485673979941529, -0.0028071655407816316352, -0.17293485476517600485, -0.087818305986051434231, 0.017490462540166257233, 0.054549119173720371501, 0.02515310964606671737, 0.00045346050149707058108, -0.0072572281282252872753, -0.0055178896470146378272, -0.0006952843330696432695, -0.060348649001597434727, 0.70751397267265159208, -1.3209103750295601021, 0.61512110937010988643, 0.20224447721290322866, -0.1065643708239563392, -0.020339409680851294748, -0.010672043306896273132, -0.0090872746065739900406, -0.0026626835373545392274, 0.0035636729102306260487, 0.0045331978453119758932, 0.0021555307878542615506, -0.0026759681044413743088, 0.42485982615591422196, 0.7586090471885199582, -2.5764393042580717719, 0.68861306598801186318, 0.98919549592760447343, 0.070371920224727657622, -0.22396539445542390512, -0.13656537812516997277, -0.022393604486377706453, 0.014787501623373625059, 0.017742352318581310067, 0.0096214882671630114208, 0.00082276673187504749396, -0.0063673796358454989169, -0.39446515418237587669, 0.6431445924576275619, -0.37034924311110112294, 0.49979873558608972628, -0.34434981643754375424, -0.23969739580432364301, 0.056698040140399981468, 0.12051737881396362706, 0.080093996800546823911, 0.0033993098480730420119, -0.035836008829208301774, -0.031638381583098298955, -0.0086238318757318223329, 0.014691187123389333122, -0.65359014572182816405, 0.39510087554434103474, 0.52986812462255583611, 0.45915380296056484966, -0.61363197525911772345, -0.28774143426434506399, -0.043582590490883532652, 0.046449165207765652497, 0.10906120903660065924, 0.076723525620387644786, 0.020006474553609043809, -0.012464295183662155164, -0.021017680686455472716, -0.017863892062887624934, 2.2620335754997875632, 0.45839619198026682056, -5.6745211769807477609, 0.30003494758118548491, 2.3671250572489346276, 0.95622775381135793893, 0.053363498478623205834, -0.2339304234618986289, -0.27167614509288390279, -0.19624771679796793045, -0.082983460236876749327, 0.0028104596353745459059, 0.044773108749937381989, 0.050817804108142801411, -0.19319414278796118167, -0.070254193073013837534, 0.59039290150295942183, 0.1050622846780845665, -0.68830326056571700111, 0.059811552084835427345, 0.20710526832032621902, 0.025653149438598484011, -0.025155219998687798622, -0.013198253219287664484, -0.00070253679288154874394, 8.9789642960514531866e-05, -0.0015130567034174436866, 3.8754856898770692272e-05, -0.18919927555626159243, 1.2693930645065680807, -2.2153503678250632269, 0.92592894720650942286, 0.73759051316419899091, -0.36496660976359840411, -0.28101761353295245938, 0.017566747368725169065, 0.084262744841147013686, 0.037994935091539359939, -1.6938681060950887269e-05, -0.013546762491044033946, -0.0092568212660849771151, 0.0026131655618943645629, 0.11093546270994314451, -0.20111385240514115247, 0.34057221369592510785, -0.075737883631449856581, -0.69179339588838550945, 0.23837532648641124311, 0.39649193103929214477, 0.073805197758754478854, -0.088619180586863702254, -0.087564334123386780573, -0.039151134261685323346, -0.0056387936577544451114, 0.010537275888083858649, 0.01800165172062827279, -0.36935065923027415868, 0.25497276111251354624, 0.26232803763482709369, 0.33071500062630809769, -0.52268824294041793532, -0.12084956638647233185, 0.057172736527034875853, 0.040820339217291536082, 0.052060199967845610658, 0.033970821749827788405, 0.0054487478461070215641, -0.012301896932695094433, -0.014769736963626729082, -0.006919229814058880032, 0.10774640643741317803, -0.20421150261950032911, 0.1950741009368815726, 0.0027757224388191246989, -0.38631141251492878697, 0.21495052199314015895, 0.17406335631683098053, -0.026891818558046003129, -0.056920746501686006469, -0.030236409229074494298, -0.0038868057315562028892, 0.0065623175019828283716, 0.006230087005034839337, 0.0018508066503055791135, 0.025047910976184424581, 0.65509129397255227989, -1.4136707530295802293, 0.56381257171616827595, 0.35817442537098687083, -0.078024510905686961237, -0.12080333312400667878, -0.017302243481283841042, 0.045031540462967589722, 0.011918319933670901145, -0.021148562737133419498, -0.01884344557902105366, -0.00032674408268230925589, 0.015083304613363532343, 0.93442712498298197143, -12.093618335075200676, 23.044350716230390219, -11.831550304345286762, -1.4716971809763379486, 0.61779245979623564899, 0.51656255112481919145, 0.19205635617917865554, 0.033127369090650449213, 0.024838702363948639285, 0.035977131349368168811, 0.01197906888948732651, -0.017604083872671330568, -0.030897783872332947042, 0.63349570613655947771, -0.81594507216670797423, 0.25973008228700084077, -0.2793384552845604718, -0.41551729161659278766, 0.66307698978951001845, 0.34278892665995891464, -0.17056060215318305517, -0.20271415032787504229, -0.076255684683875418761, 0.010778488129757318736, 0.034822679575326892509, 0.02436114848115020215, 0.0022589482020273828761, -0.83439059229730294653, 1.2393645910238624275, -0.93947513471088162085, 1.0523824806074326688, 0.24947249200128826518, -0.62699598496328090569, -0.43156739116090669928, -0.016237708426557140307, 0.16083038053051235128, 0.13000179274223350179, 0.05433063366343321432, 0.0025164110723175647788, -0.019864762890915717286, -0.025555229335297829918, 0.5672217515035193891, 0.7348389727710649133, -2.6847961188798179499, 0.55321257601907192214, 0.93820607584355719677, 0.13134703113031626076, -0.073934631438709402795, -0.054755940337182854027, -0.045806410753422496795, -0.045967068149230475016, -0.028571083819175537022, -0.0073216948364648194031, 0.0092180409493396592291, 0.018397727068401172751, 0.5029083568439713936, 0.73760296552828219507, -2.6082024712970430791, 0.63337156931279270378, 0.86024479494025374304, 0.12491200740404138914, -0.02915272112144014216, -0.10679897644383654076, -0.1082205791765800118, -0.040757276101579462124, 0.013950263769083150536, 0.023494634781573910109, 0.011294394477295185281, -0.0032610290057408074932, -0.37032573466824303399, 0.43530767625295724299, 0.73201458332128821471, -1.6287346307056180805, 0.70976536848776039523, 0.45646698602314778048, -0.12404035853981688897, -0.18128985061821367708, -0.080230246223939066796, -0.0016609677564843050579, 0.01085778571462805292, 0.01152007002730998618, 0.014388494158835097789, 0.016519662120766942515, 0.089337702071909969037, -0.30382299568369802723, 0.2861303694717463042, -0.14595311112583783664, 0.30135042760904301451, -0.25892662768146179619, -0.053537187750788764462, 0.067815924556356738107, 0.029350017662541905833, -0.00032454331654763711457, -0.0090984418316518052283, -0.0038519187318321423778, -0.00064175278964331210034, 0.0016700882513034798899, -0.032389606387993437331, -0.11439685996231277465, 0.63137776961644309637, -0.98114418160851035733, 0.6246739744214497847, -0.010536443646289896103, -0.17867984544592016571, 0.052349228305264283723, 0.0205984549685969201, -0.016376758513231358394, -0.0031974576903519525903, 0.0067638088013127865838, 0.0027445247214260863264, -0.0022132059864344230671, -0.37476727999968717508, -0.51244384150293365021, 0.79544488471384755535, 0.2747805080287735735, 0.77915000674491996513, -0.49967348480395912258, -0.58365050190208245251, -0.17478702895427905561, 0.049563622669237944462, 0.11232905303544658826, 0.093824219141650150311, 0.054397124038895336429, 0.007548542836986943394, -0.038453718691957211595, -0.2106170636859449885, -0.34966925929402781215, 0.82146417541112182192, -0.47525549442155162749, 0.87347782587354738659, -0.29237249223277206323, -0.58734165397222004579, -0.064983280478736735186, 0.16295744415369795011, 0.1218891816607205747, 0.030810265649216232731, -0.0147816220802381975, -0.020691876364233897923, -0.0049241379952283340776, -0.16654112834180051506, 0.040749018694606460589, 0.48804485664253188792, -0.69753131303947779607, 0.47741157932095629191, 0.024977780639414779579, -0.16137535573454633853, -0.088001289642131858004, 0.013701715690273310461, 0.052581755911102805168, 0.024919212112024538563, 0.00045346089568794756969, -0.0076224797045582595376, -0.0052276703173881541864, -0.11224133310143022402, 0.14754692428984844677, 0.61512110969169186969, -1.3191914114125580237, 0.60419742644004303944, 0.20406065302923775961, -0.10685646702268909725, -0.020011694538582449482, -0.0090268087167174951291, -0.0070639573741897001072, -0.0017162274717006857465, 0.0035636728949967060205, 0.0033402167904565288094, -0.0010267720829268813052, -0.38860276812993038709, 1.0319487003237364231, 0.68861306604757077654, -2.6906941731614186786, 0.61420442230682337215, 0.98636831964382742921, 0.088202185187849252435, -0.20864000846942196143, -0.12964141457009953351, -0.019759285931311251772, 0.015787982349382376213, 0.017742352631939645874, 0.0068286616588054526583, -0.0078481967424821590801, 0.050083593116825135538, -0.31776009043521447062, 0.49979873548804665262, -0.43206726254881444582, 0.52695243496905275205, -0.28550014686593055968, -0.22111404798453718201, 0.04868177983194113545, 0.10882062071913922718, 0.074457174593072633639, 0.0018439133987038364149, -0.035836008922827206069, -0.027914069023084183557, 0.0066612317691603787143, -0.062846167276388767431, -0.55075640467457509519, 0.45915380303401509554, 0.43150517701165636009, 0.40922997180037690157, -0.59151386301971298209, -0.27141609780984871092, -0.038529310478608413959, 0.048765676856517337079, 0.11346890954050681266, 0.081121594121744633465, 0.020006474551454767052, -0.02085331956791314223, -0.038958560030467687119, 0.065094672937512615407, 2.2875207410078575165, 0.30003494753129589245, -5.6151368624220312498, 0.30744999784877768967, 2.3650376089572411331, 0.95920348758038931347, 0.056982483886456103406, -0.23500290634231646436, -0.2761913793504456871, -0.20180312354056817448, -0.082983460089675442628, 0.01904025755314228685, 0.089440285814516268248, 0.33570264809829608588, -0.78266523423651512825, 0.10506228497325001414, 0.69590059897668077316, 0.11196799535250856472, -0.70680345997801630187, 0.043140481715395577467, 0.20790211314733444969, 0.03063324627869592337, -0.023978192242214126095, -0.014508803987655430334, -0.00070253695273416946754, 0.0015050360492585134614, -0.0034770754481863986284, -0.54136743131914222271, 0.82223057670437582622, 0.92592894742204279179, -2.3999548984644789584, 0.90494838320641712848, 0.78350704828304174043, -0.33145477619598506802, -0.27455328649742699598, 0.010819634174186373507, 0.079189317162585912135, 0.036777839284598108638, -1.6938949832549614338e-05, -0.013224465754624086661, -0.0058571137000069887352, 0.63069219699097756049, -0.86403981638199334192, -0.075737883721114881985, 0.55177665412905407827, -0.003583840725603340896, -0.70919620072048872306, 0.206120573280782593, 0.38763140468056184185, 0.075112687622298446311, -0.089030006278849799228, -0.090283453831052212779, -0.039151134542943943906, 0.0016111977651402647083, 0.025801911254702984688, 0.10237172612776702929, -0.51712661663127035272, 0.3307150006096959971, 0.23912337307499648142, 0.30447631904395522362, -0.51543873140920637432, -0.11400317212960353608, 0.063824013239247578966, 0.044770985500638835197, 0.05394117222303755621, 0.034828683124520823378, 0.0054487478253113647009, -0.015079746208184394846, -0.02346216640471746398, 0.30358168686613906662, -0.46160953945406912879, 0.0027757228094761653495, 0.28794863059931197213, 0.0052750459242050171116, -0.40619284804919508769, 0.19656011100232223532, 0.17283916941682236912, -0.022892780915289668436, -0.054957337018605070811, -0.030371395248246800536, -0.0038868060552281801304, 0.0076515197347917117063, 0.0069207593647665870129, -0.22801431295765015039, 0.43595415349372057534, 0.56381257146943741798, -1.5451338746705649552, 0.53591447538957537322, 0.41315131276170302677, -0.046380542279325198685, -0.12441453351942771688, -0.031749537817121920025, 0.0367352419424182286, 0.0097952672862228196138, -0.021148562970866119359, -0.014958318726216585093, 0.017884682194842155267, 1.3125995284599798474, -1.2280636129392046296, -11.831550304301956089, 23.746354276687000606, -11.716965419439809182, -1.6085492785355735013, 0.52706393546064611577, 0.50698779628638857009, 0.20851898743240154799, 0.044323366635120098334, 0.027952582582070339967, 0.03597713175949476816, 0.0049732665532821185295, -0.048650667332584912605, 0.59430108818368521106, -0.624959777908098979, -0.27933845546361002476, 0.52500955978825547543, -0.30624728624142394384, -0.49627527040798280478, 0.61760631080505867807, 0.3440448888420561957, -0.15834297858120260916, -0.19882962039293772261, -0.078316523162360013322, 0.010778488947601805314, 0.038572356091651917309, 0.024803730559053181748, -0.78088745792814540092, 0.38608078666787926103, 1.0523824801324739298, -1.1963541254218077459, 0.97062431033311968864, 0.28737555581365326063, -0.59017460714481884132, -0.42378640932022915155, -0.021217302026110940183, 0.15984279102878645729, 0.13320221537485427143, 0.054330633575152414982, -0.0071499008077068743061, -0.040426313547520111025, -0.16735150344332833106, 0.93930738458445617312, 0.55321257610677054828, -2.7372306760867486375, 0.54846181485154643642, 0.95968108088590209359, 0.14620712360590176759, -0.07051420772883346122, -0.058468812633994830807, -0.049980375333533610949, -0.048694962992702706117, -0.028571083991907809951, -0.0014540694158520816107, 0.025924674107960597041, -0.14331348196803403439, 0.73061596360283442841, 0.63337156953937745474, -2.5355033820865959804, 0.60875627312036162397, 0.81690702130159842387, 0.112292479584490415, -0.017007483352569408019, -0.094094413895657591507, -0.10376156777057320146, -0.041216287613354135599, 0.01395026355020717243, 0.023784173998269118877, 0.0042617005835649699949, -0.16776826466213926037, -0.16293144688961500033, 0.43496554059186448571, 0.7097653689660985421, -1.6338875694764869451, 0.69871802507345126543, 0.4487582729476713439, -0.12681445130808960409, -0.18091097316645965432, -0.07966547482606183006, -0.001660967594962745042, 0.011407767882158460077, 0.016089718226600158157, 0.029801799913675736109, 0.075230609165134285088, -0.043112923551718661253, -0.26025200178517537974, 0.30135042732429950796, -0.14349239607189173684, 0.30090321292775623618, -0.25839538051171023758, -0.055539343963512746716, 0.066039389529493175424, 0.028821456073942399584, -0.00032454384633538464061, -0.0087889501548951282084, -0.002663956841162030964, 0.00052996331086412561389, 0.056086596671832772421, -0.17758779711235855858, 0.0012043105856198038379, 0.62467397464345564284, -1.004978637182290635, 0.62330532255762927107, -0.0050526845486476004363, -0.1776463688440556854, 0.051826781098341843934, 0.020272788249104081032, -0.016376759190440154379, -0.0026464404226114778343, 0.007464430076196016707, -0.00068570144781292100868, 0.14425377786232396193, -0.52990155047127285304, -0.50284443449083648936, 0.77915000648133458938, 0.26725718393088815361, 0.78636622500007824499, -0.48455580623079680391, -0.57202059621031786651, -0.16734632121757891943, 0.053588007094163947475, 0.1123290532502156519, 0.08722994506087569655, 0.037795483950953348562, -0.021580536217147179567, 0.31455317213082734096, -0.62266305393375187549, -0.35252048990034184905, 0.87347782535155749084, -0.43039365315406719059, 0.89654547128857020599, -0.28776948408541935809, -0.59475961140549726824, -0.072377558652760298763, 0.16026880034489887628, 0.12188918177685056921, 0.029967004555464776816, -0.017434091785437116906, -0.022425367412156740898, 0.03136657178407915858, -0.16856239193390998299, -0.0028071653730652691358, 0.47741157901621950055, -0.68096757014100639527, 0.48677669294133413258, 0.029517370519503843829, -0.16139886559912317021, -0.089857152052457658908, 0.012329122916326231557, 0.052581756118787348342, 0.025153109452011102043, -0.0012093797706437627894, -0.012248569484950072236, -0.018257906912855031978, -0.12155818486395579547, 0.20224447697749473041, 0.60419742644107332641, -1.3338630662835335716, 0.59648147669622841871, 0.20428526292615903359, -0.10413151590057169449, -0.017217727347383637859, -0.007757596552627856086, -0.0070639575579906676087, -0.0026626836011766482834, 0.0020612296138165873277, 0.0032862876591340361293, -0.2915602595105959427, 0.045164739730486361258, 0.98919549599523504124, 0.61420442259187890688, -2.7281545485628986292, 0.59873407257183874908, 0.99233090180347816034, 0.099846061597800850329, -0.20078458718507474767, -0.12706861818788908747, -0.019759285529993800279, 0.014787501001858049943, 0.014510809580064574012, -0.0010652438219337120699, 0.22929286510998711512, -0.23662084809002217889, -0.34434981637833955714, 0.52695243511311629092, -0.37659942389594486256, 0.55442451790644931009, -0.28613719061600539906, -0.23295426174663191876, 0.039168735741522395233, 0.1047921802774893496, 0.07445717461986829655, 0.0033993097650466645721, -0.033348364375799327552, -0.021276917485646279493, 0.17169854917871210653, -0.1894864600618838546, -0.61363197490916809862, 0.4092299715509255531, 0.45789869478141626002, 0.44207477630068992003, -0.573682320927059064, -0.26419356567904617172, -0.035604830334745818377, 0.049929130100648588808, 0.11346890945096899106, 0.076723525323137475862, 0.0046746281735036421257, -0.052967078959692388829, -0.62459409989108582995, 0.73774026091793576398, 2.3671250572913966614, 0.30744999791227789476, -5.665766987324934334, 0.23311598947917488989, 2.319956361806593037, 0.94504391502378759515, 0.058433516794098139258, -0.23224280009819775472, -0.27619137893964662567, -0.19624771726828121876, -0.052789711221340407388, 0.097603933612821922039, 0.12497722198899270662, 0.11048381618721035691, -0.68830326085992610263, 0.11196799570654397682, 0.69503584146444097325, 0.11381804384079191594, -0.70461412543107337925, 0.043182048589022740959, 0.20670271216586488805, 0.029244782231542656992, -0.023978192411322499233, -0.013198253353867237664, -0.00063985756979507632525, -0.0034681139146945493684, -0.041052566956257224851, -0.41736006844952666661, 0.73759051321913615684, 0.90494838312248093715, -2.3944818095226851895, 0.91576612507664834073, 0.79487336729593727203, -0.32965274420622941998, -0.27784051635185047502, 0.0082720882759711560073, 0.079189317160108310678, 0.037994935270958159201, -0.00056779037864419876908, -0.020179637718777371225, 0.18505286841390811148, 0.23070430707938049952, -0.69179339583368737454, -0.0035838408028094261709, 0.55791308419890039971, -0.016574224012502074277, -0.72335009399208116676, 0.19500651533293453244, 0.38228401861491634328, 0.073484409500678735117, -0.089030006529595140385, -0.087564333714536063402, -0.029256500973919446346, 0.023727515806326017478, 0.14630796412754054625, -0.051827194339549104152, -0.52268824273517078272, 0.30447631889697690522, 0.25340553040921443007, 0.32524020689863186728, -0.50237001995343155691, -0.1098134666163876616, 0.063659195085570580153, 0.043732159434201611781, 0.053941172081819255268, 0.033970821481171510625, -0.0020919446484244721973, -0.036437453873008042449, 0.0015099424714446381141, 0.2427224290361977288, -0.38631141276527308781, 0.0052750459946841327108, 0.28183164306025343393, -0.0048968132571782507778, -0.41219080638919602944, 0.19527542748720549959, 0.17411708235652217924, -0.021909762222856333375, -0.054957337229218562735, -0.030236409082033802176, -0.0016759157457840396024, 0.013801116762119737535, -0.067838372553914838559, -0.089303929811169149855, 0.35817442557339890241, 0.53591447516206924817, -1.5102399938552584224, 0.55264070576572721372, 0.40791162371221700855, -0.057305762711614138816, -0.13126250514754839172, -0.034242200472677897238, 0.036735242509304766312, 0.011918319950951522523, -0.01444027949738010963, 0.0011838288038604795338, 0.50225150596015066817, 0.82312588849351320253, -1.4716971812268631048, -11.716965419450549035, 23.855088216811836332, -11.687096359944098012, -1.6025730940997977569, 0.52787189957423164, 0.50609921891985543141, 0.20772571192669181395, 0.044323366660788288129, 0.02483870175723093543, 0.020096058126576263336, -0.037618301440351241405, -0.21859162237995274913, 0.71667925303218593669, -0.4155172918147438943, -0.30624728605591755581, 0.47159021147413682229, -0.34584827192171846422, -0.51389470020724015153, 0.61798070417456085757, 0.3502475782721510833, -0.15537505041085214463, -0.19882962006018869983, -0.076255685250283838927, 0.020076605330844904396, 0.058562330622952657455, -0.032384963154505147509, -0.58499487200338706749, 0.24947249248869982075, 0.9706243102415093027, -1.1932755511461983211, 0.99834242851067556135, 0.31298060000977484529, -0.57805676635299640953, -0.42074542947089266232, -0.020979838579538667709, 0.15984279118855360124, 0.13000179275439410764, 0.040149967268790526242, -0.042542739604892070016, -0.18175946145324153869, 0.048297050554180619153, 0.93820607593095106669, 0.54846181492964785065, -2.7463785340460220574, 0.5324428852533120704, 0.95071923860904272718, 0.14148688380406748699, -0.072594952277221264092, -0.059346334460798531862, -0.049980375438572630098, -0.045967067902519115341, -0.018873533486824946304, 0.020197126202753552804, -0.2282406677518718785, 0.063551495596154417012, 0.86024479488290361839, 0.60875627351044736546, -2.6077689852692556016, 0.56946006166334672827, 0.81713094943759345679, 0.12348884219563810438, -0.008946187240915861455, -0.091373456561690047373, -0.10376156835971758752, -0.040757275728481451305, 0.016110141457386668168, 0.025154596247243095342, -0.0019161508101103352959, -0.19298145700707358685, -0.13047207214819905863, 0.45646698668290336398, 0.69871802462164533942, -1.6506381819080218065, 0.6906698132518389599, 0.44915540806396858642, -0.12488223132058247355, -0.1809109728487610369, -0.080230246074010747037, 0.0020035281636886104889, 0.024051823331736253003, 0.035472518290720095324, 0.0015300532599772515008, 0.07532098223057942521, -0.046902453299823093635, -0.25892662774296199002, 0.30090321303873246439, -0.14099222649139223074, 0.30106992614126643071, -0.25959113800393540661, -0.056333418294768219936, 0.066039389917984539502, 0.029350017630621876713, 0.00021096807828681601177, -0.0091297919694885883157, -0.0022363963444530952571, -0.009692575630997580724, 0.073858530862022694907, -0.17703923652284359047, -0.010536444334714137328, 0.62330532291912310594, -1.0011469244818949065, 0.62456132057881430519, -0.0053334951060853919361, -0.17805865944544879698, 0.051826781245683892219, 0.020598455375692225483, -0.016111504481568953567, -0.0028654865253794179099, 0.0065056041797584831729, 0.25361276756426676116, -0.10690377335234617362, -0.57595902913369978471, -0.49967348512685128004, 0.78636622511858322859, 0.28381756255381623832, 0.80532829311241649339, -0.47024782435525402935, -0.56550723713925254987, -0.16734632092156281891, 0.04956362299659149151, 0.10559340648270404417, 0.077178211800937007725, 0.023156963196151021667, 0.25881713105931747343, -0.046498310636636758231, -0.59841169795319526514, -0.29237249208483051532, 0.89654547127970030118, -0.42546526242855797761, 0.89608003688183779367, -0.28907591367409418703, -0.59613403365948813839, -0.072377558344453113426, 0.16295744389297245336, 0.12541360911433468983, 0.025835440926303999964, -0.043560141871513996525, 0.097496339623187794343, -0.096994648305013017597, -0.17293485487334669948, 0.024977780947256826571, 0.48677669260196215761, -0.67921289473879042919, 0.48677291218518936589, 0.029333285595018888936, -0.16189648072754142527, -0.089857151756564987433, 0.013701715490450639898, 0.054549119202759815961, 0.023283549922462332821, -0.015708498160010780226, -0.0088583582635921916809, -0.017559335037219039183, -0.10656437049107607418, 0.20406065271460308019, 0.59648147680704999285, -1.3381446131437555458, 0.59686097005890759171, 0.20670728905612997806, -0.10270601763402775752, -0.017217727483509338848, -0.0090268084459054452606, -0.0090872747505775872817, -0.0027200241786502475129, 0.0074823994164432226089, -0.052759547079642338252, -0.24340531205534607362, 0.070371920006121468916, 0.98636831964623805646, 0.59873407262969757792, -2.7368308352567867203, 0.59847583371935231256, 0.99674319401129829465, 0.10320307580483084609, -0.20078458737541379775, -0.12964141475617574351, -0.022393603890866829731, 0.014642218884723109257, 0.016319231005019304431, 0.13415618584984315143, 0.028518725927489843586, -0.23969739599121792084, -0.28550014668182188693, 0.5544245179370881349, -0.37568848603111171247, 0.54669599648362698918, -0.29297526740017715285, -0.23685477814416208697, 0.039168735955094527434, 0.10882062043201647106, 0.080093997017022300344, 0.0023570209848728200519, -0.060505768103772089583, 0.21344368467052446392, -0.022280748975962827441, -0.28774143483759317963, -0.59151386274465578285, 0.44207477631279679109, 0.47978714286281987089, 0.45555016607420895092, -0.56634768543186009282, -0.26178629854192458914, -0.035604830785903152091, 0.048765676666384871751, 0.10906120932734497675, 0.061454303791320191719, -0.040913546827227462166, -0.59790533426976510967, -0.085679259169192306667, 0.9562277536649484988, 2.3650376090529889872, 0.23311598953934420386, -5.7370601819810556066, 0.18912959920849603535, 2.3046461713670809424, 0.94423910470640493031, 0.058433516559443843263, -0.23500290656859301741, -0.27167614480009316669, -0.16270875294792563892, 0.036279517042348857769, -0.047820343769353540475, 0.22825367183943762184, 0.059811552471536931497, -0.70680346018320328039, 0.11381804383310435702, 0.705247043527534756, 0.11930924652261576391, -0.7055115546424679529, 0.040781005918122385789, 0.20670271197677586583, 0.03063324650853591069, -0.025155220048356893847, -0.017512774531397587441, -0.00079167868922711687185, 0.16106091963082153073, -0.30060682972633773202, -0.36496660965373811658, 0.78350704811922633475, 0.91576612517205546737, -2.3962527063602285793, 0.91162705782342712268, 0.79305063321229685158, -0.33039050523956697836, -0.27784051650308405623, 0.010819634149583494745, 0.084262744546596463069, 0.036817625857433949887, -0.026835312258677534308, -0.17862577826471248699, 0.39534193694860236823, 0.2383753265863463866, -0.70919620058239851712, -0.016574223937893477199, 0.55382355327039423099, -0.022409844235574377658, -0.73074659459418889185, 0.19063720466643732654, 0.38228401880216911213, 0.075112687466293126537, -0.088619180769584776636, -0.084040079448472637669, -0.0037635188333485446698, 0.098202847383277336313, 0.072746658724918589201, -0.12084956658729807677, -0.51543873096662040645, 0.32524020664553188187, 0.27084340119455119211, 0.33409261587050775866, -0.5004460555174294889, -0.11071478597506187458, 0.063659195034764637366, 0.044770985678226418725, 0.052060200190362018735, 0.022434636565075341541, -0.033069197621026197309, -0.11922878668719648354, 0.17623718136968044568, 0.21495052204976836108, -0.40619284805453081955, -0.0048968131330707868887, 0.28066065702797821091, -0.004665533866605103136, -0.41255004297021780069, 0.19510262323400656692, 0.17411708235266576406, -0.022892781039783489938, -0.056920746340566485966, -0.028535914670654027869, 0.014676172022276355589, 0.050618426164365223907, -0.17260828249998572126, -0.07802451095653251123, 0.41315131260165982585, 0.55264070580103541452, -1.523093435296531073, 0.53706070561660090856, 0.40093603426371127174, -0.058890055846744449042, -0.13126250534694566818, -0.031749537909791181178, 0.045031540252000235025, 0.022187195199633719112, -0.026816070751414771561, 0.084593883253782420351, 0.63458064430518146271, 0.61779245989272291162, -1.6085492783689687712, -11.687096360333285361, 23.932263663539810494, -11.645416078901824619, -1.5956662264570455889, 0.52417892605803595441, 0.50609921850762884521, 0.20851898765548426695, 0.033127369348806223748, -0.0024130726131602886486, 0.0028579246900495985928, -0.40436524241913335276, 0.35555853423389127688, 0.66307698988470642476, -0.4962752704781601687, -0.34584827217354108253, 0.46115326616486995714, -0.34732121892724743262, -0.51418864858040491672, 0.61831501802272370583, 0.3502475781041779479, -0.15834297864337132378, -0.20271414991312208809, -0.064141512942392203689, 0.080643839442443729371, 0.33078941552410995319, -0.40664939306471004077, -0.62699598537295875289, 0.28737555576627116238, 0.99834242848650911473, -1.1772931653074900638, 1.0098480153112636692, 0.32209217145055557552, -0.57389999627172794128, -0.42074542943753889812, -0.021217302095692447406, 0.16083038056789256132, 0.1219488297151265499, -0.0060482607279902372843, -0.11421072393144883483, -0.12552424071973955488, 0.13134703128357655411, 0.95968108077685676527, 0.53244288546955098607, -2.768054373991003736, 0.51684288167571668104, 0.94427752746793691685, 0.14017137196856463421, -0.072594952259692979712, -0.058468812693273877756, -0.045806411023997899157, -0.036339935906830739265, -0.0044315556599201916255, -0.19020593452441070248, -0.018346305306477427022, 0.12491200770815755194, 0.81690702082692490116, 0.56946006198470588977, -2.6185274502251387752, 0.57057699087351643197, 0.82028120097651813225, 0.12551021480371649108, -0.0089461872342119254375, -0.094094413459790943177, -0.10822057943554645731, -0.038809717344265359795, 0.047604826796656077303, 0.042630586065151666875, -0.058319354843738853422, -0.17392930896580371947, -0.12404036020036396382, 0.44875827397296291776, 0.69066981315171782629, -1.6528158249038538141, 0.69148122321977467664, 0.44915540800075176575, -0.12681445162978371988, -0.18128985038324488266, -0.071470239185793552061, 0.021909127433253277467, 0.038448413286839268344, -0.017879099823061336827, 0.030834208800604525486, 0.069374758420285884131, -0.05353718766410254376, -0.25839538040443055245, 0.30106992602526660985, -0.14138070144416944873, 0.30079075983208969358, -0.25959113762028546146, -0.055539344524210339238, 0.067815924789626780278, 0.030029050445620049381, -0.002933460836591768664, -0.010539190678304057383, -0.0092850483461916107075, 0.019962524765830012841, 0.049240162203338183611, -0.17867984473871906248, -0.0050526848261445867633, 0.62456132060668256845, -1.0016548252643375783, 0.62416292080547230992, -0.0053334948315381762451, -0.17764636900518368523, 0.052349227944657510114, 0.01920967595336242964, -0.020756790753696408025, 0.0088022920144432449663, 0.18949602478414065421, 0.090328539390259879327, -0.1689082372113999142, -0.58365050173524490873, -0.48455580670462050641, 0.80532829304987696428, 0.29981491687946326685, 0.81401133485909982213, -0.47024782431399847482, -0.57202059603608546023, -0.17478702904976228805, 0.048636881136683879356, 0.11529838115848929647, 0.096157548642722023846, 0.064948092958448444545, 0.18209867556032735547, -0.051245636905258885874, -0.5873416534704128944, -0.28776948437532157143, 0.89608003699594784841, -0.42505720937846225294, 0.89692158398115329199, -0.28907591339458910129, -0.59475961161809487443, -0.064983280415921856887, 0.17565721418759192018, 0.12319925120475491709, -0.034533806863756413841, 0.05446234380286237714, 0.01943207665293230843, -0.087818305888904740764, -0.16137535593366181264, 0.029517371037641199766, 0.48677291185947374474, -0.67981290338377386995, 0.48648708148774377769, 0.029333285924737110412, -0.16139886604067144038, -0.088001289501029256179, 0.017490462602309672646, 0.055176281915976219117, 0.0013088334376712609103, -0.0008493813765305589748, -0.0059807387117716114683, -0.020339410032758756253, -0.10685646666782454284, 0.20428526284643780397, 0.59686096990558668018, -1.3369450397593760638, 0.59771465913256771607, 0.20670728905235680184, -0.10413151573842316377, -0.020011694882563067777, -0.010672042827485425306, -0.0068677346556948780723, 0.0066729684836460342878, 0.041928166482272745397, -0.13258658024463737535, -0.22396539438709350822, 0.088202184923316759302, 0.99233090178561711436, 0.59847583371642676386, -2.7374123956084561726, 0.59888340773779569215, 0.99674319394939858707, 0.099846061635411501034, -0.20864000822262041002, -0.13656537833911530644, -0.017462177249075273161, 0.039247930484483634084, -0.0017280602585173381915, 0.10103334675858778591, 0.05669804037530390417, -0.22111404809231033353, -0.28613719072205612282, 0.5466959965434590174, -0.38121802116147168737, 0.54409131204309113539, -0.29297526725665290481, -0.23295426212263398158, 0.04868178027730709484, 0.12051737835670167653, 0.072589812609178017144, -0.070957616475460519312, 0.14174129184561368167, 0.04648137293796866798, -0.04358259023123023379, -0.2714160977937392083, -0.57368232092769022579, 0.45555016604588005613, 0.48904337754956206386, 0.4597388751290614195, -0.56634768565118109862, -0.26419356531354298845, -0.0385293104244730511, 0.046449165021833827005, 0.10134407182132779135, 0.025935141076790124703, -0.36924112958754229386, -0.27398330997006420384, 0.053363498675752962797, 0.95920348758676543532, 2.3199563616369696106, 0.18912959917538674226, -5.7627488395289141465, 0.180323278930880182, 2.3046461710748169516, 0.9450439153341241294, 0.056982483776929493402, -0.23393042349474046926, -0.25813394868443467756, -0.12629776720260096901, -0.039665199175132830056, 0.02186257469788143587, 0.20710526802425310855, 0.043140481694735756391, -0.70461412538852896681, 0.119309246670436922, 0.70678228478776639054, 0.11804737652488145327, -0.70551155470225834776, 0.043182048651136992201, 0.20790211309996511901, 0.025653149615806638412, -0.033713203918135786763, -0.0088807110212707623953, 0.088328592695873564544, 0.019428202254616069866, -0.28101761380784451294, -0.33145477588219085519, 0.79487336728744817371, 0.91162705793788512132, -2.4013970768180543658, 0.90984220227506984457, 0.79305063321654678532, -0.32965274411257483589, -0.2745532865826668667, 0.01756674749351764403, 0.083530085294778150251, 0.0014149776944189593891, -0.18785874576076652676, 0.058674768195713572494, 0.39649193073432809076, 0.20612057350823453916, -0.72335009403199901357, -0.022409844297180590844, 0.54929877642949065386, -0.025864742502588765338, -0.7307465943872365477, 0.19500651491894990808, 0.38763140484109448458, 0.073805197868391556049, -0.09893661475697937413, -0.080391510339706873101, 0.054004747712466202869, 0.034469533162801407888, 0.057172736808717136703, -0.11400317242925736561, -0.50237001961679073059, 0.33409261599151884869, 0.27443259544229203772, 0.33438767379946665592, -0.50044605557062860068, -0.10981346671798170034, 0.063824013208835267408, 0.040820338795456871661, 0.040026009025779532569, -0.00089569307485423341197, -0.065186365903680487666, -0.029941835865308172315, 0.17406335662568761524, 0.19656011059494535309, -0.41219080640823330119, -0.0046655338598757811752, 0.28138097628637215308, -0.004731568813120427125, -0.41255004297220754239, 0.19527542732294472794, 0.17283916994225578811, -0.026891818877342425764, -0.05776858150536663683, -0.0078340844977415613481, 0.027033051481457615745, -0.035585414948998708717, -0.12080333302283272068, -0.046380541657171148184, 0.40791162308834161276, 0.53706070599622979245, -1.5321855023174335564, 0.53459780641630194609, 0.40093603412763534344, -0.057305762281755434628, -0.12441453380976114085, -0.01730224360225540714, 0.051809252568792985016, -0.026554699056948585789, 0.012994018995739542444, 0.21500566587508160299, 0.51656255147898666458, 0.52706393525504802167, -1.6025730938863556041, -11.645416079000717957, 23.959042067153514921, -11.639514046229491484, -1.5956662268997494625, 0.52787190039168296884, 0.50698779591481368811, 0.19205635612224997155, 0.0069308577456673736705, 0.028213876937971584663, -0.16052513800103046515, -0.17290750718974612354, 0.34278892680308892249, 0.6176063106941025449, -0.51389470007095794379, -0.34732121907346114043, 0.46146462633650575658, -0.3482365811419499968, -0.5141886488857829729, 0.61798070479276223654, 0.34404488853096121259, -0.17056060216939297747, -0.1951221730370631402, 0.030188118517343563718, 0.27405261654754725642, 0.0088427381562305451546, -0.43156739109314534719, -0.59017460706990731989, 0.31298059987938053883, 1.0098480153875653009, -1.1699838324058355887, 1.0144800474769395393, 0.32209217157673541987, -0.5780567664720542842, -0.42378640916715160092, -0.016237708486274593511, 0.16896907960361812195, 0.10371402864496505336, -0.081781251213959585411, -0.068588346535850774899, -0.073934631980182782818, 0.14620712403330665685, 0.95071923849407569129, 0.5168428817222002758, -2.7787636361369085769, 0.51281858328746587894, 0.94427752752729954278, 0.14148688362006828223, -0.07051420796298986382, -0.054755939869550969423, -0.043392315612368448363, -0.044638060893232849036, -0.070367761745137549778, -0.095606415823678250754, -0.029152721592094747727, 0.11229247965429610401, 0.81713094942332575865, 0.57057699086941771061, -2.6193778740128408344, 0.56980202857595096955, 0.82028120096885126511, 0.12348884193679110499, -0.017007483385184686553, -0.10679897602449760219, -0.10695814873189900263, 0.0273913345467055383, 0.038448413198561057214, 0.021909127694073278481, -0.07147023925941754785, -0.18128985058486610238, -0.12681445188288831272, 0.44915540870189724076, 0.69148122288181124162, -1.6528158250286248965, 0.6906698133416784291, 0.44875827360212494632, -0.12404035947047006849, -0.17392930939878381569, -0.058319354972762138534, 0.042630586193597497235, -0.01053919069000074614, -0.0029334607990539368801, 0.030029050395553872482, 0.067815924854266365629, -0.055539344658059965065, -0.25959113740407163684, 0.30079075959269302443, -0.14138070121492382625, 0.30106992580809771587, -0.2583953802581400172, -0.053537187668143762509, 0.069374758328779873029, 0.030834208878354904537, -0.017879099847341640289, 0.0088022920121246048947, -0.020756790717139907143, 0.019209675807944807252, 0.052349228209692774683, -0.17764636926097318592, -0.0053334946616810877812, 0.62416292061254519119, -1.0016548249795009795, 0.62456132033459377251, -0.0050526846678989852449, -0.17867984478314405394, 0.049240162185111117921, 0.019962524788633504574, -0.0092850483523066705549, 0.096157548659958139159, 0.11529838113681978307, 0.048636880958167802436, -0.17478702873118556749, -0.57202059612814193379, -0.4702478241808455417, 0.81401133445588702564, 0.29981491700582713067, 0.80532829344169987174, -0.48455580696174227473, -0.58365050199777113171, -0.16890823714354852303, 0.090328539743578489807, 0.18949602459148082167, -0.034533806790361949723, 0.12319925094628739193, 0.17565721450112586899, -0.064983280629834008768, -0.59475961131109811308, -0.28907591381178276979, 0.89692158413438616282, -0.42505720915976141772, 0.8960800366630754521, -0.28776948406432190097, -0.58734165378003855285, -0.051245636639885482067, 0.18209867539844257212, 0.064948093010605306707, 0.0013088333804004521552, 0.05517628206647758049, 0.017490462426772837579, -0.088001289280872918575, -0.16139886625102722872, 0.029333285932900035609, 0.48648708150784603088, -0.6798129032335880062, 0.48677291174943432273, 0.02951737107688705633, -0.16137535597811183963, -0.087818305975744651026, 0.019432076800763000562, 0.054462343751315277218, 0.0066729684335553006436, -0.0068677345203477886335, -0.010672043001246540178, -0.020011694704776886167, -0.10413151593080401114, 0.20670728927617387782, 0.59771465894118747109, -1.3369450396897977207, 0.59686096994653170533, 0.20428526275582245542, -0.10685646656458773485, -0.020339410202707200459, -0.0059807385023109387034, -0.00084938146952293362615, 0.039247930509122341824, -0.017462177307805727688, -0.13656537829545684071, -0.2086400083203335809, 0.099846061846943043316, 0.99674319384884868533, 0.59888340767090242345, -2.7374123955478464332, 0.59847583368063117515, 0.99233090180126692914, 0.088202184920280229941, -0.22396539440752877281, -0.13258658019298685815, 0.041928166458694626773, -0.070957616317162935937, 0.07258981225013208527, 0.12051737873616846786, 0.048681779874255370899, -0.23295426165972704657, -0.29297526767976767026, 0.54409131222042605991, -0.38121802105469093647, 0.54669599636121368746, -0.28613719048358210451, -0.22111404832837100454, 0.056698040577392326556, 0.10103334647778625111, -0.0017280600952198151184, 0.02593514094117641694, 0.10134407199822215162, 0.04644916510037500057, -0.038529310532118277166, -0.264193565509580508, -0.56634768531878920683, 0.4597388749899304905, 0.48904337745908166291, 0.45555016621416211109, -0.57368232112206274387, -0.27141609769473801217, -0.043582590233984697115, 0.046481373008528170665, 0.14174129178417110841, -0.12629776729971567417, -0.2581339485138443024, -0.23393042348838000155, 0.056982483582108778286, 0.94504391544019739069, 2.3046461711505621395, 0.18032327903893130561, -5.762748840040194942, 0.18912959962203768072, 2.3199563614451470528, 0.95920348783351927491, 0.053363498386700958243, -0.27398330976888773947, -0.36924112966701944094, -0.0088807109250311957321, -0.033713204058112060391, 0.025653149580641274735, 0.20790211320676429962, 0.043182048717360012324, -0.70551155494709583138, 0.11804737669545546652, 0.70678228490259165095, 0.11930924645465537659, -0.70461412533341860609, 0.043140481710108688485, 0.20710526816414454032, 0.021862574483069132508, -0.039665199088501662539, 0.001414977782510703358, 0.083530085096173975479, 0.017566747586034651973, -0.27455328645306875668, -0.32965274421394064053, 0.79305063304339173946, 0.90984220243146340934, -2.4013970766012020519, 0.91162705764502305339, 0.7948733673271464184, -0.33145477595497074841, -0.28101761342255671128, 0.019428201845964759159, 0.088328592835575814313, -0.080391510433936663738, -0.098936614612734202767, 0.073805197986223411433, 0.38763140456464706407, 0.19500651493190099273, -0.73074659425486465647, -0.02586474244607361167, 0.54929877629434986641, -0.022409844443912564638, -0.72335009363842506236, 0.20612057330532718158, 0.39649193072487820544, 0.058674768168152514891, -0.18785874572646207858, -0.00089569293673387798127, 0.040026008667389222084, 0.040820338981403248013, 0.063824013301818804944, -0.10981346661574264978, -0.50044605586825130494, 0.33438767398145191612, 0.27443259545689824286, 0.33409261591918532064, -0.5023700198194149813, -0.11400317215280567074, 0.057172736811256702483, 0.034469533122484685705, 0.0540047476884038119, -0.0078340845139729716612, -0.057768581392049885015, -0.026891819097947754907, 0.17283917008297455808, 0.1952754273088987691, -0.41255004291000901828, -0.0047315688936652583463, 0.28138097628546093754, -0.0046655338951851685336, -0.41219080632277621534, 0.1965601105764617218, 0.17406335660766092399, -0.029941835864548550783, -0.065186365902084306145, -0.026554698933418673795, 0.051809252254714291408, -0.017302243293973340599, -0.12441453413640463088, -0.057305761671459398698, 0.40093603338709149497, 0.53459780687968239121, -1.5321855024062038808, 0.53706070578553077866, 0.407911623520171851, -0.046380542095647903089, -0.12080333271916243731, -0.035585415146656432239, 0.027033051558775431089, 0.028213876908906383029, 0.0069308578746768850776, 0.19205635581251601085, 0.50698779639225233229, 0.52787189994371175317, -1.5956662266756174162, -11.639514046240096334, 23.959042066938184945, -11.645416078460565146, -1.6025730944260052624, 0.52706393545484375718, 0.51656255130828387934, 0.21500566613865934595, 0.012994018884920681411, 0.030188118460060215514, -0.19512217286962107998, -0.17056060237670442326, 0.34404488876032235378, 0.61798070454145781483, -0.51418864874264558207, -0.34823658113212740961, 0.46146462612030381134, -0.34732121856434616802, -0.51389470057139841064, 0.61760631093737927699, 0.3427889265734340718, -0.17290750692171019676, -0.16052513811004898159, 0.10371402867742118836, 0.16896907951246425972, -0.016237708333422640194, -0.42378640941172313061, -0.57805676614701695382, 0.32209217128642941841, 1.014480047577748012, -1.1699838322874880348, 1.009848015151348033, 0.31298060023925944328, -0.59017460748483541177, -0.43156739082935346996, 0.0088427380377962827573, 0.274052616590966025, -0.044638060954188311935, -0.043392315506766074296, -0.054755939800502619053, -0.070514208196080729874, 0.14148688375037404841, 0.94427752744609205759, 0.51281858347144704258, -2.7787636362442826865, 0.51684288152579949127, 0.95071923880207165336, 0.14620712395114274229, -0.073934631958571084276, -0.068588346663328136898, -0.08178125114463057066, 0.027391334489344047742, -0.10695814853421337087, -0.10679897626385745357, -0.017007483230158902676, 0.12348884171723321923, 0.82028120129500636803, 0.56980202842783733708, -2.6193778740905315772, 0.57057699093120517464, 0.81713094941071084953, 0.11229247966752477217, -0.02915272157982493062, -0.095606415813748790722, -0.070367761767251096883, 0.035472519078757654143, 0.024051821639847075635, 0.0020035285515902622605, -0.080230245250077658614, -0.18091097256780588887, -0.12488223249872372478, 0.44915540879827492438, 0.69066981324904452855, -1.6506381822159035266, 0.69871802504344959761, 0.45646698598162227523, -0.13047207153813888958, -0.19298145713283756941, -0.0019161508595849500221, -0.0022363963275666772119, -0.009129791994818430706, 0.00021096808851963060744, 0.029350017532104789864, 0.066039390219499352619, -0.056333418755506486919, -0.25959113746175321857, 0.30106992555561912228, -0.14099222592765608963, 0.3009032126023933329, -0.25892662746881456304, -0.046902453493785897787, 0.075320982369135439094, 0.0015300532129060141848, 0.0065056041724605032442, -0.0028654865801797471149, -0.016111504214457509493, 0.020598454896275951442, 0.051826781768529645023, -0.17805865995259551049, -0.0053334945504176639597, 0.62456131996969765918, -1.0011469238493466705, 0.62330532229098645214, -0.010536443777168369093, -0.17703923693798456762, 0.073858531088745377202, -0.0096925756929792626637, 0.023156963132067168859, 0.077178211806816804375, 0.10559340671557172764, 0.049563622846664837085, -0.16734632099793281257, -0.56550723731396601579, -0.47024782390632946516, 0.80532829287706053645, 0.28381756248375977769, 0.78636622513148213276, -0.49967348494556457217, -0.57595902917380392694, -0.10690377347219401361, 0.25361276761087664378, -0.04356014193511156124, 0.025835441077634836715, 0.12541360905786669289, 0.16295744375491516553, -0.072377558251905366182, -0.59613403362380934514, -0.28907591358224521416, 0.89608003665074320576, -0.42546526229829700894, 0.89654547130703865498, -0.29237249221668010213, -0.59841169778240510535, -0.046498310718172806977, 0.25881713106037895766, -0.015708498173064279479, 0.023283549942550222506, 0.054549119264330717116, 0.013701715311302492192, -0.089857151600953602544, -0.16189648082499291659, 0.029333285783540392899, 0.48677291189498911317, -0.67921289445025623266, 0.48677669238069026791, 0.024977781065521079129, -0.17293485494095794941, -0.096994648238635350346, 0.097496339592612613068, 0.0074823993990750618799, -0.0027200240889081227572, -0.0090872749134950862177, -0.009026808311881127575, -0.017217727473754267098, -0.10270601781034352507, 0.20670728934471074845, 0.59686096970231583558, -1.3381446128015361818, 0.59648147654565775166, 0.20406065293931141547, -0.10656437064287135885, -0.01755933502270466684, -0.0088583582397262009001, 0.016319230852007961252, 0.014642219248334326781, -0.02239360418217035964, -0.1296414144495348364, -0.20078458783856839576, 0.10320307614566408272, 0.9967431939342978886, 0.59847583356661959719, -2.7368308351545072021, 0.59873407274230894259, 0.98636831966540694516, 0.070371919865723470133, -0.24340531208703453103, -0.052759547011971032193, -0.06050576832874569877, 0.002357021547933779021, 0.080093996591706906329, 0.1088206205745894245, 0.039168735722711088554, -0.236854777797654098, -0.29297526758381320366, 0.54669599643876432005, -0.37568848596646825522, 0.55442451796538516628, -0.28550014676267865266, -0.23969739595296576939, 0.028518725982569312349, 0.13415618580419705341, -0.040913546838576411724, 0.061454303860226849876, 0.10906120938051122593, 0.048765676513899069988, -0.035604830644978546772, -0.26178629876451109393, -0.56634768534327717404, 0.45555016624794864022, 0.47978714268720656921, 0.44207477636911196583, -0.59151386286943175019, -0.28774143462211193567, -0.022280749100455826683, 0.21344368469171062785, 0.036279517017275886293, -0.16270875277687021909, -0.27167614515003646325, -0.23500290604089846269, 0.058433515998828733373, 0.94423910520696541226, 2.3046461707002601216, 0.18912959997592904138, -5.737060182553834764, 0.23311598994088800207, 2.3650376086733100323, 0.95622775403290927176, -0.085679259462591583763, -0.5979053341509168451, -0.00079167877940139810833, -0.017512774368796189817, -0.025155220038346318145, 0.030633246306835927586, 0.20670271204174467394, 0.040781006117910439945, -0.70551155481304383965, 0.11930924646520685295, 0.70524704361410406328, 0.11381804389583910497, -0.70680346018349748949, 0.059811552264907121013, 0.22825367200649826493, -0.047820343797028271038, -0.026835312292014391888, 0.036817625905865083991, 0.084262744633001249173, 0.010819633890613215496, -0.27784051633914685908, -0.33039050516643514444, 0.79305063315068502483, 0.91162705769556240387, -2.3962527061613019264, 0.91576612508261268086, 0.78350704814599836379, -0.36496660978798151076, -0.30060682957088924461, 0.16106091958260121411, -0.0037635187451655954458, -0.084040079585505050863, -0.088619180769404226616, 0.07511268747805714685, 0.38228401896244784686, 0.19063720450828777753, -0.73074659455733814717, -0.022409844302905067293, 0.55382355352632861578, -0.016574224330921479703, -0.7091962003478740062, 0.23837532656609036752, 0.39534193696674685414, -0.17862577829379211458, -0.033069197766946320516, 0.022434636916847532778, 0.052060200124269637167, 0.044770985473896685036, 0.063659194975341670819, -0.11071478572792933792, -0.50044605572318645592, 0.33409261602235884592, 0.27084340096321180003, 0.32524020718706414534, -0.51543873157765085491, -0.12084956628117977939, 0.072746658572967776157, 0.098202847468330661806, 0.014676172006283376079, -0.028535914682272567333, -0.056920746227930932304, -0.022892781183846338394, 0.17411708243403675733, 0.19510262320966756389, -0.41255004298061387358, -0.0046655338600868276333, 0.2806606571078690826, -0.0048968134020280981722, -0.40619284764132290722, 0.21495052172757639353, 0.17623718148874861766, -0.11922878670256767397, -0.026816070781272405288, 0.022187195260957470744, 0.045031540201965647441, -0.031749537844595443481, -0.13126250563429558982, -0.058890055247057593313, 0.40093603358196094888, 0.53706070625573187893, -1.5230934359402179545, 0.55264070634100248558, 0.41315131228640916694, -0.078024510779960587481, -0.17260828260620009189, 0.050618426195006643864, 0.0028579245624661270275, -0.0024130724190778718441, 0.033127369480620727504, 0.20851898732264911862, 0.50609921849957617557, 0.52417892635105001187, -1.595666226709038682, -11.645416078662879755, 23.932263663311609037, -11.68709636026548182, -1.6085492782994088579, 0.61779245977319396932, 0.63458064445052009273, 0.084593883178700132741, 0.080643839422133087314, -0.064141512944114714712, -0.20271414998672199204, -0.1583429783764888632, 0.35024757768903874711, 0.61831501848650582964, -0.51418864903447070169, -0.34732121845113095304, 0.4611532657679729974, -0.34584827202342377506, -0.49627527045418962093, 0.66307698988780805482, 0.35555853419921279457, -0.40436524241220361819, -0.0060482606826848597792, 0.12194882954236721528, 0.16083038078577804875, -0.021217302328428055969, -0.42074542922099855691, -0.57389999633205346363, 0.3220921714377281142, 1.0098480153318503127, -1.1772931654394163115, 0.99834242860934718777, 0.28737555580988999315, -0.62699598543994183864, -0.40664939304792635522, 0.33078941551364682283, -0.0044315556259796416022, -0.036339935939485458471, -0.045806411033171207103, -0.058468812751178614318, -0.072594952117625372034, 0.14017137193210024715, 0.94427752733125314144, 0.51684288181620674507, -2.7680543739239817924, 0.53244288522149862963, 0.95968108090508086327, 0.13134703133524461283, -0.12552424076635654293, -0.11421072392592900024, 0.047604826845058741125, -0.03880971749195074455, -0.10822057927532835464, -0.094094413536957383082, -0.0089461870993735592705, 0.12551021455143643313, 0.82028120116133407258, 0.57057699073640111287, -2.6185274499912196688, 0.56946006169860063917, 0.81690702112518376321, 0.12491200745796995908, -0.018346305231400700936, -0.1902059345074815222, 0.029801799307154881735, 0.016089719610943672978, 0.011407767738742875113, -0.0016609686422227529223, -0.079665475038579916522, -0.18091097235251046849, -0.12681445195307680662, 0.44875827339447660336, 0.69871802583922038021, -1.6338875702551431957, 0.70976536905193243765, 0.4349655407225770376, -0.16293144737896164687, -0.16776826436853486957, 0.00052996326860076266568, -0.0026639567342207634906, -0.008788950278085329304, -0.00032454366035801745366, 0.028821455750478439106, 0.06603938993789694456, -0.055539344443236167426, -0.25839537999357858888, 0.3009032125356210785, -0.14349239583522230279, 0.30135042709829962249, -0.26025200150229893525, -0.043112923772065721295, 0.075230609232876749504, -0.00068570145601558351624, 0.0074644301951085796631, -0.0026464407719969818694, -0.016376758677884760418, 0.020272787746530154113, 0.051826781500934379632, -0.17764636920909174234, -0.0050526841147276093519, 0.62330532204863720036, -1.0049786366294248818, 0.6246739740848663569, 0.0012043110520504113001, -0.17758779737681967825, 0.056086596742197257492, -0.021580536299122331928, 0.037795484218588706882, 0.087229944946866116995, 0.11232905296619863367, 0.053588007155217318167, -0.16734632070663527093, -0.57202059653331704947, -0.484555806318052118, 0.78636622476828277328, 0.26725718426751254864, 0.77915000656411304014, -0.50284443465188621847, -0.52990155047183984394, 0.14425377788962667758, -0.022425367340512761954, -0.017434091859387379486, 0.029967004440540814109, 0.1218891821381088425, 0.16026879980239258638, -0.072377558246563250544, -0.59475961156817369613, -0.2877694838845679115, 0.89654547118907246261, -0.43039365346115399102, 0.87347782586539046701, -0.35252049032329380029, -0.62266305368382246499, 0.31455317205136928971, -0.012248569421532884069, -0.0012093799059639987088, 0.025153109581147518881, 0.052581756015625522038, 0.012329122921536901994, -0.089857151910335206213, -0.16139886583654866614, 0.029517370874178633366, 0.4867766925430526137, -0.68096756996181795341, 0.47741157907889592016, -0.0028071654170926272839, -0.16856239198026601311, 0.031366571812972823818, 0.0032862877018894794123, 0.0020612294224750804528, -0.0026626832623789552582, -0.0070639579079750389537, -0.0077575963111953787771, -0.017217727378101895658, -0.10413151610837714789, 0.20428526332181726199, 0.59648147621687908071, -1.3338630658989190092, 0.60419742618404481682, 0.20224447719481561392, -0.1215581850170472239, -0.018257906863950387755, -0.0010652437058930831412, 0.014510809168814357839, 0.014787501462465442448, -0.019759285937894180418, -0.12706861758937715301, -0.20078458755407430081, 0.09984606157786887759, 0.99233090190160566646, 0.59873407236311715351, -2.7281545484517093492, 0.6142044224798201002, 0.98919549623047708931, 0.04516473967838053466, -0.29156025954905218089, -0.02127691733498990323, -0.033348364829655127717, 0.0033993101365956610027, 0.074457174576234005925, 0.10479218027068529229, 0.039168735886336993046, -0.23295426194149204679, -0.28613719061600234594, 0.55442451785495150407, -0.37659942387931349961, 0.52695243519826084899, -0.34434981646900508778, -0.23662084789684639796, 0.22929286498848386322, -0.052967078924460564426, 0.004674627994026014921, 0.076723525189459240714, 0.11346890993133784686, 0.049929130042792646549, -0.035604830365513318036, -0.26419356576897745637, -0.57368232110137884483, 0.44207477632781044807, 0.45789869501865188006, 0.40922997143177708423, -0.61363197510504741139, -0.18948645976819469827, 0.17169854908442080954, 0.097603933669063697165, -0.05278971158305803546, -0.19624771672164259551, -0.27619137960545381283, -0.23224279915221290871, 0.058433516066947799317, 0.94504391570743884188, 2.3199563608886837329, 0.23311599006836858372, -5.6657669877099330336, 0.30744999822268326284, 2.3671250569688999654, 0.73774026138936865671, -0.62459410011608129665, -0.0034681138240164836178, -0.00063985775065950219187, -0.013198253305690015036, -0.023978192246833653078, 0.029244782162966907746, 0.2067027118824052434, 0.043182048853322330118, -0.70461412540503065571, 0.11381804384084995285, 0.6950358413364222665, 0.11196799575644046654, -0.68830326069348080065, 0.11048381594628460534, 0.12497722208119189935, -0.020179637676269419205, -0.00056779043889222408775, 0.037994935205458602034, 0.079189317433224701293, 0.0082720880174217432251, -0.27784051642855223108, -0.32965274401603744669, 0.7948733672915788695, 0.91576612502064413945, -2.3944818095065310004, 0.90494838326676130258, 0.73759051294701505341, -0.41736006830088612229, -0.041052566977423793348, 0.023727515801162491621, -0.029256500947672160751, -0.087564333850979239604, -0.089030006367148306623, 0.073484409482363705068, 0.38228401865959643624, 0.19500651508477612195, -0.72335009365716440577, -0.016574224258279091071, 0.55791308446091669726, -0.0035838411979343277186, -0.69179339560601427195, 0.23070430712801343054, 0.18505286835504200504, -0.036437453847808637386, -0.0020919447447982977839, 0.033970821362587111381, 0.0539411724665873199, 0.043732159307314277807, 0.063659194939818197767, -0.10981346641975672784, -0.50237002014416498596, 0.32524020715397017334, 0.25340552974802565966, 0.30447631962986554388, -0.52268824305605177116, -0.051827194286583250227, 0.14630796414091401525, 0.013801116771348180198, -0.0016759157440198110822, -0.030236409215298337, -0.054957336986561404479, -0.021909762319130242358, 0.17411708226612682138, 0.1952754276128826072, -0.41219080651462058906, -0.0048968132681881347545, 0.28183164342845695716, 0.0052750453874788761494, -0.38631141228903587148, 0.24272242884278050212, 0.0015099425044711243908, 0.0011838287631621077778, -0.014440279291880131485, 0.011918319705425756139, 0.036735242613906760045, -0.034242200505889219819, -0.13126250538768852172, -0.057305762053645792342, 0.40791162295103622348, 0.552640706576400631, -1.5102399948365281546, 0.53591447599668773361, 0.35817442511889402423, -0.089303929582756208516, -0.067838372638586927343, -0.0376183012973931169, 0.020096057920400298147, 0.024838701725206663307, 0.044323366737573421936, 0.20772571217275059841, 0.50609921843549654508, 0.52787189986929405894, -1.6025730939594851066, -11.687096360406759032, 23.855088217315589816, -11.71696541976838013, -1.4716971809165770857, 0.82312588812151521456, 0.50225150612210134327, 0.058562330711639624947, 0.020076605196869837244, -0.07625568498518275895, -0.19882962064368189803, -0.15537504970456988973, 0.35024757748377011168, 0.61798070499982249437, -0.51389470076504895069, -0.34584827178929139491, 0.47159021170234804998, -0.30624728624536540211, -0.41551729179068691566, 0.71667925299428181241, -0.21859162235740092761, -0.042542739690647507489, 0.04014996761190528507, 0.13000179231708591798, 0.15984279161238160172, -0.020979839114102283304, -0.42074542914212387235, -0.57805676641339798216, 0.31298059995952670631, 0.99834242876933176891, -1.1932755513748218856, 0.97062431036033847143, 0.24947249239657581787, -0.58499487200579580737, -0.032384963136448376153, 0.020197126241344710851, -0.018873533591038330753, -0.045967068009554176711, -0.049980374994456303051, -0.05934633487332141577, -0.07259495205108229543, 0.14148688366712347531, 0.95071923861300144942, 0.53244288537542083795, -2.7463785340276158919, 0.54846181476636757246, 0.9382060758247523502, 0.048297050874256973463, -0.18175946159832773175, 0.025154596180140081224, 0.016110141572065243831, -0.04075727580909677128, -0.10376156836524864924, -0.091373456423499699319, -0.0089461872504207860801, 0.12348884195953037379, 0.81713094959862142552, 0.56946006152913608567, -2.6077689849161576063, 0.60875627314785651922, 0.86024479502459405555, 0.063551495574433694569, -0.228240667743920711, 0.016519662005470045485, 0.014388494299793688075, 0.011520069623498583461, 0.010857786448753341257, -0.0016609679795657715518, -0.08023024602060119248, -0.18128985053803578498, -0.12404035931348718347, 0.45646698575333766179, 0.70976536895491126877, -1.6287346303708942763, 0.73201458316861967024, 0.4353076763478733735, -0.37032573478261809896, 0.0016700882268168948212, -0.0006417527600026684282, -0.0038519188272666299473, -0.0090984415503751435439, -0.00032454357891389517404, 0.029350017717800241934, 0.067815924533391178097, -0.053537187607092834307, -0.2589266278858661785, 0.3013504277267290421, -0.14595311114320627666, 0.28613036949794296016, -0.30382299573535015424, 0.089337702096722176881, -0.0022132060432970284032, 0.002744524886786336125, 0.0067638086406949654572, -0.0031974576487820345096, -0.016376758519761051847, 0.020598455094218284106, 0.052349227991069086086, -0.17867984499960082312, -0.010536444066998912339, 0.62467397472350938425, -0.98114418186437024616, 0.63137776985278437625, -0.11439686009350051588, -0.032389606359385300638, -0.038453718789463714467, 0.0075485431842539293945, 0.054397123847014200448, 0.093824218835865974553, 0.11232905326763516218, 0.04956362266554027729, -0.17478702880627930427, -0.58365050206145163703, -0.49967348478855788674, 0.77915000661666733439, 0.27478050832314820928, 0.79544488480331765334, -0.51244384193719860754, -0.37476727980563268039, -0.0049241380875048104682, -0.020691876242867033447, -0.014781621940651197555, 0.030810265303610084864, 0.12188918206453823456, 0.16295744381751062213, -0.064983280507857232866, -0.587341653839900113, -0.29237249217698840775, 0.87347782601125245616, -0.47525549478428980033, 0.82146417568833252076, -0.34966925947759402993, -0.21061706360318663278, -0.0052276704381805840643, -0.0076224793599251067927, 0.00045346054393009078629, 0.024919212260220646726, 0.05258175591811065086, 0.013701715657604427237, -0.088001289658613493505, -0.16137535577392492225, 0.024977780764302279226, 0.47741157931663119607, -0.69753131321067851633, 0.48804485680633463707, 0.040749018643601704615, -0.16654112834314280245, -0.0010267721020792320175, 0.0033402168727848331395, 0.0035636727378275611343, -0.0017162272260885934501, -0.0070639576874249208546, -0.0090268085439478354393, -0.020011694485464772558, -0.10685646714445751204, 0.20406065315884377998, 0.6041974263445731852, -1.319191411379654344, 0.61512110962861643593, 0.14754692438452784997, -0.1122413331430902883, -0.0078481966661115756118, 0.0068286617572071127424, 0.017742352390961609737, 0.015787982252969769148, -0.019759285859767911075, -0.12964141467006268238, -0.2086400079909792038, 0.088202184784483189661, 0.986368319953564332, 0.61420442221260396209, -2.6906941733933678051, 0.68861306609845884807, 1.0319487004977969669, -0.38860276822247524775, 0.0066612317696293985025, -0.027914068964070462564, -0.035836009042838248839, 0.0018439135199673564736, 0.074457174546940618765, 0.10882062045477799983, 0.048681780249877917854, -0.22111404828603767325, -0.2855001465946748751, 0.52695243495391475008, -0.43206726274695506085, 0.49979873560142262789, -0.31776009055666659675, 0.050083593200961618042, -0.038958559925427869997, -0.020853319851570073507, 0.020006474752411074203, 0.081121594088018444957, 0.11346890968341549755, 0.04876567625994687738, -0.038529309796424104917, -0.27141609806402805516, -0.59151386254121873787, 0.4092299710014666303, 0.43150517749337324869, 0.45915380290630186666, -0.55075640474834641758, -0.062846167202051925194, 0.08944028585734621617, 0.019040257690687817371, -0.082983460591159352582, -0.20180312298999936171, -0.27619137984913988415, -0.23500290598192674629, 0.056982483517806215012, 0.95920348810728428557, 2.3650376087126541158, 0.30744999804390849008, -5.6151368627408189127, 0.3000349477516799368, 2.2875207408137692155, 0.065094673027210517291, -0.0034770754572496175537, 0.0015050359631934625421, -0.00070253664657991821141, -0.014508804280609306769, -0.023978192237453815894, 0.030633246514637890723, 0.20790211290240656616, 0.043140481844498028419, -0.70680346019752504638, 0.11196799578001925823, 0.69590059864853703875, 0.10506228485422849928, -0.78266523389616449169, 0.33570264795505483368, -0.0058571136920195247463, -0.013224465856577891434, -1.693869976445404324e-05, 0.036777839005802934191, 0.079189317447248386284, 0.010819633987037337655, -0.27455328666522638148, -0.33145477585927135555, 0.78350704814865779202, 0.9049483832243799819, -2.3999548985866807627, 0.92592894750627818823, 0.82223057675115429621, -0.5413674313541571026, 0.025801911295043829936, 0.0016111976269147757108, -0.039151134483175711876, -0.090283453690265941249, -0.089030006378651813104, 0.075112687602933422837, 0.38763140464384554518, 0.20612057333252795099, -0.70919620065134281184, -0.0035838407978952564054, 0.55177665411514520422, -0.075737883714917020317, -0.86403981638438942525, 0.63069219700515966043, -0.023462166443977063024, -0.015079746166019258879, 0.0054487480858216030821, 0.034828682770285057657, 0.053941172046240840054, 0.044770985840491467345, 0.063824013012185071281, -0.11400317200268569873, -0.51543873126433392695, 0.30447631938098107529, 0.23912337237268591061, 0.33071500079572752062, -0.51712661658160274936, 0.10237172613943557042, 0.0069207594598762372032, 0.0076515194386992399847, -0.0038868057039908900097, -0.030371395450384357229, -0.054957337000557215934, -0.022892780825266967337, 0.17283916933482673661, 0.19656011108479851113, -0.40619284810895073257, 0.0052750457723390008213, 0.28794863097489409309, 0.002775722432859294847, -0.46160953924996916964, 0.30358168681813407774, 0.017884682253982490258, -0.014958318998083169893, -0.02114856241664922476, 0.0097952667295534145975, 0.036735242309713811437, -0.031749537951440531813, -0.12441453385358491879, -0.046380541732159885626, 0.41315131245739766808, 0.53591447551845505792, -1.5451338747029352838, 0.56381257146047425444, 0.43595415344897758825, -0.22801431292579271704, -0.048650667428673202708, 0.0049732665957111510391, 0.035977132011823897972, 0.027952582436723272252, 0.044323366483255521864, 0.20851898745336078234, 0.50698779657033399282, 0.52706393493817282891, -1.6085492778920591483, -11.716965420109753282, 23.746354277363899143, -11.831550304783771566, -1.2280636128193906931, 1.3125995284886800007, 0.0248037304728101482, 0.038572356433933918662, 0.010778488376220751732, -0.078316522644568814515, -0.19882962072466472958, -0.15834297814062453513, 0.34404488834854618506, 0.61760631109994634791, -0.4962752706664297353, -0.30624728626036390455, 0.52500956004372845154, -0.27933845545551860834, -0.62495977801635738125, 0.59430108819387506003, -0.040426313408496505897, -0.0071499012241021675518, 0.054330634010454265104, 0.13320221503562679199, 0.15984279146478297395, -0.021217302316512809401, -0.42378640915816245816, -0.59017460730533610747, 0.28737555583034984918, 0.97062431004722471428, -1.1963541250141589423, 1.0523824800296111004, 0.38608078674884788173, -0.78088745800958558885, 0.025924674170032795117, -0.0014540695632005449026, -0.028571083998375719493, -0.048694962959750856535, -0.049980375031932269825, -0.05846881288587642278, -0.070514207831058411946, 0.14620712372690480851, 0.95968108091118708991, 0.54846181479670397252, -2.7372306760593541064, 0.55321257601465989584, 0.93930738472961727847, -0.1673515035054319311, 0.0042617007101580246103, 0.02378417378460518164, 0.013950263535250001512, -0.041216287418158026323, -0.10376156804555747293, -0.09409441356430624237, -0.017007483446539940763, 0.11229247960950822227, 0.8169070210435475099, 0.60875627332721615748, -2.5355033821046570885, 0.63337156946677675151, 0.73061596374988202474, -0.14331348205470181845, 0.00096906634959792654384, 0.011340225634147330785, 0.013697188452111049017, 0.011520069998655490737, 0.011407767984512845461, 0.0020035283318784059131, -0.071470239415395500626, -0.17392930911350909517, -0.13047207219786988164, 0.43496554139874243905, 0.73201458311825340353, -1.5533876588600632473, 0.79934734953967279125, -0.078366127107814750885, 0.0017399464297309220982, -0.00020599822810504883532, -0.0013990228523795321738, -0.0038519189384243119646, -0.0087889503199935041511, 0.00021096775225025368034, 0.030029051044789558528, 0.069374758034204606072, -0.046902452924687802749, -0.26025200240020585074, 0.28613036993770996075, -0.1728666410867761849, 0.2727767161785098482, -0.16764883190244744826, -0.001012940833664280121, -0.0011056315713872872427, 0.0026067892799323888822, 0.0067638085229565023809, -0.0026464408439860311398, -0.016111504503033478169, 0.019209676256746775064, 0.049240162091318581683, -0.17703923640762114777, 0.0012043102218529408518, 0.63137777018006113749, -1.0459053401435594033, 0.73497862874573693936, -0.20164456045830206343, -0.03367955178217127421, -0.010175142656444209621, 0.019740071412964407122, 0.054397123926630978896, 0.08722994468010880631, 0.10559340661726623889, 0.048636881405902582243, -0.16890823744512697968, -0.57595902904827611657, -0.50284443458417527051, 0.79544488465409668443, 0.34547880437748884619, 0.8695841129894505972, -1.044828870197838766, 0.0017700396009309247568, -0.010099177367423044227, -0.020176069439838087427, -0.014781622229625486498, 0.029967004578067429205, 0.12541360892047195841, 0.17565721470084991185, -0.051245636922925420775, -0.59841169783744374566, -0.35252049031722820782, 0.82146417572894536718, -0.39669655406359793925, 1.1557506189537252173, -0.87560778714319520422, -0.00069528438118191253792, -0.0055178895916615285783, -0.0072572281157677453328, 0.00045346050923580211167, 0.025153109605366839874, 0.054549119160323220057, 0.017490462613548026644, -0.087818306128530546073, -0.1729348545550162819, -0.0028071657551441281386, 0.48804485696626037639, -0.60892098834171237698, 0.55287554033436359102, -0.25448880730473655287, -0.00267596814653384878, 0.0021555308649856036625, 0.0045331978345917673182, 0.0035636729221536107837, -0.0026626836376947141984, -0.0090872745518921875518, -0.010672043225032786656, -0.020339409844187118781, -0.10656437058883963531, 0.20224447680772569824, 0.61512110985732970647, -1.3209103753073623277, 0.70751397270241589421, -0.060348648978974836488, -0.0063673796209565303866, 0.00082276680462163055696, 0.0096214880907854183345, 0.017742352595345910887, 0.014787501212069391332, -0.022393603960042897383, -0.13656537863989132164, -0.22396539413214427983, 0.070371919936105503668, 0.98919549614363555534, 0.68861306602060767812, -2.5764393043522888505, 0.7586090471799213919, 0.42485982619570672458, 0.014691187091313982066, -0.008623831915125286518, -0.031638381360231809802, -0.035836009004395479038, 0.0033993098605317323413, 0.08009399698060150341, 0.12051737839768209004, 0.056698040545737592166, -0.23969739612233409431, -0.34434981610984560785, 0.4997987353578671188, -0.37034924315843725795, 0.64314459263105949738, -0.39446515424450534493, -0.017863892127260645737, -0.021017680449965103762, -0.012464295496671468078, 0.020006474895376769041, 0.076723525170214357249, 0.109061209778696655, 0.04644916429655177037, -0.043582589669287075118, -0.28774143522199130407, -0.61363197445867156965, 0.45915380253012672007, 0.5298681248793354337, 0.39510087557123818591, -0.65359014581563767976, 0.050817803991927978924, 0.044773108895594743162, 0.002810459792395192391, -0.082983460454914281801, -0.19624771670790308598, -0.27167614515155369403, -0.23393042344729897408, 0.053363498350483373756, 0.95622775390685765817, 2.3671250572259960876, 0.3000349476477695565, -5.6745211769997441209, 0.45839619198831532687, 2.2620335754900287029, 3.8754782608499275046e-05, -0.0015130565286507451778, 8.9789421776209285553e-05, -0.00070253656312342996348, -0.013198253370661085226, -0.025155220026413016576, 0.025653149588696057271, 0.20710526812689414311, 0.05981155241663150407, -0.68830326099531169426, 0.10506228501669802322, 0.59039290137759203869, -0.070254193119008767621, -0.19319414274648716301, 0.0026131655931354693181, -0.0092568213657499079372, -0.013546762395716432231, -1.6938773315105723483e-05, 0.037994935263455105467, 0.084262744592566968249, 0.017566747591070668716, -0.28101761357110399686, -0.36496660977890388322, 0.73759051301420908153, 0.92592894736709319137, -2.2153503677746586575, 1.2693930643585160656, -0.18919927549792564525, 0.018001651702099392971, 0.010537275849826133536, -0.0056387933848471008189, -0.03915113471544859336, -0.087564333728630774911, -0.088619180885293247218, 0.073805198157283258276, 0.39649193048970909903, 0.23837532685648693542, -0.69179339596481181918, -0.075737883568483155283, 0.34057221363702439021, -0.2011138524453779941, 0.11093546274442955824, -0.0069192298484359402583, -0.014769736975032660836, -0.012301896921061636347, 0.0054487480609564792333, 0.033970821615659368975, 0.052060200019754129808, 0.040820338980529252693, 0.057172736684597269541, -0.12084956646793676427, -0.52268824303002869858, 0.33071500091312194813, 0.26232803746983696103, 0.25497276115465550284, -0.36935065923711496438, 0.0018508066294345452626, 0.0062300871286162379237, 0.0065623173503648653954, -0.0038868057579181857983, -0.03023640908319597323, -0.056920746672162764179, -0.02689181826204708331, 0.17406335592766974951, 0.21495052233250411255, -0.38631141277700797865, 0.0027757226066321156327, 0.19507410090496263289, -0.20421150266323162548, 0.10774640646166074032, 0.015083304618421968651, -0.00032674409258313941162, -0.018843445631268967988, -0.021148562661222114478, 0.011918320021424622102, 0.045031540252912172217, -0.017302243415425300199, -0.12080333296849213309, -0.078024511251590877858, 0.3581744259861849855, 0.56381257107908133719, -1.4136707527120118133, 0.65509129385749054197, 0.025047911023865068736, -0.030897783976040621434, -0.017604083497470911224, 0.011979068206068504976, 0.035977132220502305859, 0.024838701488679837404, 0.033127369837342685033, 0.19205635551247790538, 0.51656255172238396689, 0.61779245944398319956, -1.471697180705500152, -11.831550304696095921, 23.044350716280618485, -12.093618334815067428, 0.93442712484186241095, 0.0022589481411449535697, 0.024361148677800592566, 0.034822679324135255574, 0.010778488365122212977, -0.076255685049845839396, -0.20271415001919568244, -0.17056060224581973128, 0.34278892665485161118, 0.66307699009079046348, -0.41551729193813391472, -0.27933845527848344403, 0.25973008226165800183, -0.81594507203094646108, 0.63349570607062899441, -0.025555229256979609953, -0.019864762952588488343, 0.00251641093103444069, 0.054330633762607327686, 0.13000179276493167846, 0.16083038056266429905, -0.016237708591066879471, -0.4315673906975424079, -0.62699598571257431079, 0.24947249292940215648, 1.0523824796906633416, -0.93947513401858806503, 1.2393645906244215027, -0.8343905921828499439, 0.018397727034431085591, 0.0092180409373893903296, -0.0073216946307162328322, -0.028571084159084665643, -0.045967067858049590057, -0.045806410904133024187, -0.054755940174470926762, -0.073934631677200948752, 0.13134703113435966526, 0.93820607601076577708, 0.55321257594764117194, -2.6847961186594697658, 0.73483897243481410833, 0.56722175163309163093, -0.0032610290759033954723, 0.011294394617885238769, 0.023494634782070498991, 0.013950263590466940172, -0.040757275953581903472, -0.10822057928399221594, -0.10679897617882196104, -0.029152721548091412029, 0.12491200782389887125, 0.86024479458678537735, 0.63337156951838213814, -2.608202471365535402, 0.73760296560206572991, 0.50290835679501166844, -0.010428101826195015564, 0.0022359137249069248148, 0.011340226862043490536, 0.0143884938828519552, 0.016089718323731842797, 0.02405182304716113606, 0.021909127773234931408, -0.058319355133164149652, -0.19298145686777265495, -0.1629314476079247731, 0.43530767611779652704, 0.79934734972724752478, -1.7555349652541147876, 0.87380200390418238943, 0.00061062329214820257328, 0.00049384581771488226913, -0.00020599858820711603249, -0.00064175228363388828537, -0.0026639570016240171103, -0.0091297915385118149495, -0.0029334614070200756841, 0.030834209113492157955, 0.075320982081794898844, -0.043112923395324090348, -0.3038229958342484327, 0.2727767161159081466, 0.099532992267940406661, -0.11751239148700627635, -0.00011928393385240124061, -0.0013165165779025949158, -0.001105631912708536068, 0.0027445253986407881774, 0.0074644298480591349254, -0.0028654861297911318231, -0.020756791159019509496, 0.019962524865518888051, 0.073858530905101915653, -0.17758779717369466145, -0.11439685997128090644, 0.73497862855303797147, -0.82361966104147510404, 0.30445719458342634933, -0.015337905684411776153, -0.016729882436852517441, -0.010175142757550624528, 0.0075485431553021856144, 0.037795484198641482454, 0.077178212025484763115, 0.11529838089976564286, 0.090328539791246678647, -0.10690377367995372582, -0.52990155071590805402, -0.51244384131167608487, 0.86958411272563662653, 0.40917805123866352046, -0.3973723468393667857, 0.0020092209363176904112, -0.0013117333599039714652, -0.01009917767721132606, -0.020691875908119161664, -0.017434091963023059385, 0.025835441266159147461, 0.12319925060040129439, 0.18209867576193053407, -0.046498310598044989561, -0.62266305398762111789, -0.34966925935189979713, 1.1557506189987307721, -0.28053812625908447487, -0.1342236719596133776, 0.0011528933539377340424, -0.0019310759625114273103, -0.005517889714452871644, -0.007622479287347079871, -0.0012093798744172651674, 0.023283549914391001023, 0.055176281976419910835, 0.019432076835318144031, -0.096994648466281571664, -0.16856239177091772441, 0.040749018498288817891, 0.55287554039449393617, -0.63924099749351648292, 0.23184292020640898113, -0.0013722583410660208099, -7.6127641711777626655e-05, 0.0021555308072692259784, 0.0033402167314108425003, 0.0020612297022731622918, -0.002720024329803297225, -0.0068677344005173912925, -0.0059807387543007288488, -0.017559334990236877966, -0.12155818476611075785, 0.14754692416166639846, 0.70751397273316507519, -1.2642987721474194807, 0.55942155107543056758, -0.00043111038366224559322, -0.001566372321372037392, 0.00082276692688050198643, 0.0068286615208727578785, 0.014510809541307380377, 0.014642218835034109814, -0.01746217702079132239, -0.13258658045648061274, -0.24340531219146105402, 0.045164739972025719217, 1.0319487004999587931, 0.75860904710367260684, -3.0230652616544051448, 1.5474697282216067773, 0.010047606168928892301, 0.0050088823929355227524, -0.0086238320357084966017, -0.027914069044503796169, -0.033348364309794091553, 0.0023570209328165909329, 0.072589812502787579018, 0.10103334664185342762, 0.028518726121699811138, -0.23662084826517232328, -0.31776009037373398325, 0.64314459267374801676, -0.1217561867304724349, -0.12183341350069973319, 0.0033026030504680354422, -0.010737624123216125971, -0.021017680363980259517, -0.020853320051056778439, 0.0046746283625304640985, 0.061454303218453437685, 0.10134407248010851577, 0.046481372626298145256, -0.022280748717462939279, -0.18948646005459923725, -0.55075640459261254733, 0.39510087551594968991, 0.74042292137461063195, -0.53944016698004115806, 0.0077580787433407100151, 0.035970252049468902367, 0.044773108990858415113, 0.019040257423901585376, -0.052789711251952503579, -0.16270875307764115503, -0.25813394848152032557, -0.27398330978845530925, -0.085679259421824832677, 0.73774026126824976402, 2.2875207407806303905, 0.45839619210157944718, -6.1545535914699636137, 3.3751799764893948286, 0.0028577062015422514257, -0.0015878381449636888925, -0.0015130564842628950115, 0.0015050358873703057611, -0.0006398576780388357671, -0.017512774477338476087, -0.033713203902320430794, 0.021862574495491331195, 0.22825367193019349688, 0.1104838162835937998, -0.78266523421873546162, -0.070254193106837073413, 1.1958875682219622938, -0.65691505111655856819, 0.0042656499966817404412, -0.0013567273122323574366, -0.009256821430793408037, -0.013224465770528080061, -0.00056779051102892564781, 0.036817625971879971991, 0.083530085113642890682, 0.019428201976102423693, -0.30060682950317446638, -0.4173600683517491583, 0.82223057683969957843, 1.2693930642224293681, -2.7192590773576421626, 1.2326278600221931558, 0.0059499007066419438983, 0.0099699330170822351993, 0.010537275533431341226, 0.0016111981113767072198, -0.029256501226871140658, -0.084040079344359891467, -0.09893661501428552274, 0.058674768426559158196, 0.39534193697532415968, 0.23070430706466910076, -0.86403981636134952193, -0.20111385246689564288, 1.1940273895818609429, -0.64269926400865085014, 0.0064302587686314539514, -0.006244292970623718797, -0.014769736946745357781, -0.01507974630392475146, -0.0020919446168575633882, 0.02243463664286984155, 0.040026008866296723665, 0.034469533261847193351, 0.072746658532726188318, -0.051827194089452584269, -0.51712661681806237901, 0.2549727611940673655, 0.59286980044841131043, -0.42287026189324072867, -0.00077770024519902111911, 0.0023250938364286679455, 0.0062300870847956059048, 0.0076515195168926977698, -0.00167591596169515758, -0.028535914284069770208, -0.05776858191652926855, -0.029941835426405074422, 0.17623718117187320287, 0.2427224290789555261, -0.46160953932678761014, -0.20421150272838647921, 0.65438702977880292, -0.30918022500561831434, 0.0027808327733594030012, 0.0068153432077601126815, -0.00032674406651892612397, -0.014958319125498220625, -0.014440279169339292897, 0.02218719513273555155, 0.051809252345432113551, -0.035585414921089533746, -0.17260828240078490703, -0.089303930126452046867, 0.43595415368342377294, 0.65509129398595300486, -1.5977616549748150376, 0.75363966602117216098, -0.0044582205352901244436, -0.02094372000870644776, -0.01760408326458876016, 0.0049732659915397853245, 0.020096058702958094488, -0.0024130731317772574629, 0.0069308581992025136698, 0.21500566594805359255, 0.63458064435630900935, 0.82312588830954125374, -1.2280636131211488671, -12.093618334573273287, 20.738187581550111105, -9.0964316144066295067, -0.009196730038522418077, 0.0049033668886195175274, 0.024361148769911640521, 0.038572356186487528895, 0.020076605515544923564, -0.064141513081150527964, -0.19512217280162075239, -0.1729075071595402302, 0.35555853396344183626, 0.71667925336059057351, -0.62495977830539228837, -0.81594507184301989788, 1.0465952443041239839, -0.32665486387162501192, -0.007611096979429596808, -0.016385343012906249116, -0.019864762928915209728, -0.0071499010493865089444, 0.040149967252746610047, 0.12194882973688039862, 0.16896907970899058271, 0.0088427378431540637482, -0.40664939283108486601, -0.58499487229912272568, 0.38608078699953352064, 1.2393645906018608827, -1.4796034406698963526, 0.57895024481653356574, 0.0053059936468189249775, 0.010823467306756356238, 0.0092180407706145117003, -0.0014540693229484324273, -0.018873533730895779292, -0.036339936015665409708, -0.043392315357359156902, -0.068588346782279582814, -0.12552424047927501771, 0.048297050609722558256, 0.93930738471043528914, 0.73483897251332863654, -2.9508352698301250072, 1.4928377586489556261, -0.0027233600837759607514, 0.00085334873112499342197, 0.011294394469668298311, 0.023784174047166636856, 0.016110141273274158208, -0.038809717301904862741, -0.10695814856820422267, -0.09560641595072678911, -0.018346305380743235802, 0.063551495889644177284, 0.73061596380418547536, 0.7376029653350257842, -2.7112554981792067466, 1.3894628049454544882, -0.016193746107041259719, -0.010428101299185815395, 0.00096906551368203936152, 0.016519662191340631768, 0.029801799889268929672, 0.035472518400066252453, 0.038448413236375031388, 0.042630586116355229098, -0.0019161507658517909838, -0.16776826441096542863, -0.37032573469372775943, -0.078366127132651661191, 0.87380200386994366646, -0.37660767842838993591, -0.0010808800625595736157, 0.00061062323615823426923, 0.0017399466544076294585, 0.0016700879311734259138, 0.0005299634577329974958, -0.0022363965037537369453, -0.010539190507266879906, -0.017879099918042422324, 0.0015300532782820341036, 0.075230609221027533451, 0.089337702010093000826, -0.16764883182736362022, -0.11751239150393100186, 0.14966708864645653088, 1.1285143182786940941e-06, -0.00011928397972890060377, -0.0010129406373697068602, -0.0022132063374347677419, -0.00068570122053172085502, 0.006505603968127916803, 0.0088022921620565224704, -0.0092850483595818787685, -0.0096925757027799576415, 0.056086596814798370114, -0.03238960654854424609, -0.20164456030554145372, 0.30445719454339303933, -0.11843490055651970105, 0.0093407905800908727634, -0.015337905740936070079, -0.033679551737431964731, -0.038453718764169468602, -0.021580536222527570145, 0.023156963025681990415, 0.096157548680195006896, 0.18949602466617856966, 0.25361276770060953156, 0.14425377800689989649, -0.37476728018797633712, -1.0448288700246222138, -0.39737234682832345278, 1.2744046495676568664, 0.00011854480139881435675, 0.0020092208730054031407, 0.0017700397589593094057, -0.0049241381928257811629, -0.022425367317459542899, -0.04356014202227901505, -0.034533806642245840068, 0.064948092868727866356, 0.2588171310051359808, 0.31455317223381645819, -0.21061706369749302881, -0.87560778717295051354, -0.13422367193447545808, 0.71477822047358108293, 0.0012956377231895493517, 0.0011528933075319537822, -0.00069528429565485336174, -0.005227670471233560219, -0.012248569444339160034, -0.015708498158062418615, 0.0013088334137777886401, 0.054462343761968602662, 0.097496339632353462568, 0.031366571763430391162, -0.16654112829322828548, -0.25448880732898887524, 0.23184292020731933626, 0.043998728088166523531, 0.0015425410966165772897, -0.0013722583947728636024, -0.0026759680777516206031, -0.0010267720694430103844, 0.0032862876233925404287, 0.0074823994441344434531, 0.0066729684196710241026, -0.00084938135554638049575, -0.0088583583167896884358, -0.018257906919520720812, -0.11224133305189680498, -0.060348648990323668084, 0.55942155106140434295, -0.37929277907319042251, 0.0069153436728966889008, -0.00043111037161159410092, -0.0063673796939151972724, -0.0078481965850106909233, -0.001065243844673209421, 0.016319231055118843532, 0.039247930403181620351, 0.041928166552804103928, -0.052759546998781658989, -0.29156025968666665626, -0.38860276819180850083, 0.42485982621325346642, 1.5474697282338496507, -1.3599544507767165058, -3.0492267896469783806e-05, 0.010047606108528886124, 0.01469118716618790163, 0.0066612318116583552063, -0.021276917553292123281, -0.060505768071780118011, -0.070957616411578536275, -0.0017280602020486287088, 0.1341561857527674706, 0.22929286514242641637, 0.050083593175548120346, -0.39446515432382839306, -0.12183341345983987603, 0.22631878367463906243, 0.022029097681661190256, 0.0033026030301255334537, -0.017863892119334884995, -0.038958559916986483518, -0.052967079067104724799, -0.040913546549521805085, 0.025935140791605540422, 0.14174129190696996283, 0.21344368462950707999, 0.17169854906795023464, -0.062846167141689585067, -0.65359014584698027495, -0.53944016696754937268, 0.85298840296950517725, -0.029200535391540549668, 0.0077580787222393966102, 0.050817803974039607584, 0.089440285938045177572, 0.097603933527651481894, 0.036279517191465181225, -0.12629776735277609134, -0.36924112963238264751, -0.59790533414160251802, -0.62459410013293148456, 0.065094673123394272696, 2.262033575379970074, 3.3751799765254149044, -4.3905989900173549856, 0.0028120714806828721669, 0.0028577062050535906486, 3.8754773962408704119e-05, -0.0034770754333163676406, -0.003468113864667071982, -0.00079167873618139371003, -0.0088807109782415849519, -0.039665199115115006012, -0.047820343810598957279, 0.12497722197576463521, 0.33570264812585698389, -0.19319414274585200442, -0.65691505117226389743, 0.49433610617453677127, 0.0021264947765462002133, 0.004265649977173139977, 0.0026131656341164813048, -0.0058571137110753337604, -0.020179637665759111304, -0.026835312332139520802, 0.0014149777919002506574, 0.088328592784943829597, 0.16106091954002668687, -0.041052566967730318159, -0.54136743134917042486, -0.18919927543406886428, 1.232627859976445972, -0.67208592917909337761, -0.0077856277105951641332, 0.0059499006558856020604, 0.018001651887888135001, 0.025801911058063817933, 0.023727515931034601215, -0.0037635188631560214691, -0.080391510247693739877, -0.18785874581415701323, -0.17862577836225573824, 0.18505286846696877845, 0.63069219695746914223, 0.11093546276057240918, -0.6426992640154560732, 0.080528393069863202225, 0.01488878878542475824, 0.0064302587138488891821, -0.0069192298252985983895, -0.023462166394582296969, -0.036437453908852196482, -0.033069197631669294823, -0.00089569306392960588992, 0.054004747681270018345, 0.098202847487323732811, 0.14630796402801543588, 0.10237172628214877279, -0.3693506592660573129, -0.42287026190718646212, 0.47726604271841321925, -0.0020551011596641908727, -0.00077770026700715584426, 0.0018508066565396338356, 0.0069207594390242859611, 0.013801116865853859522, 0.014676171839996815169, -0.0078340843517723801614, -0.065186366009575516522, -0.11922878665417321886, 0.0015099424632010183516, 0.3035816868289924475, 0.10774640651296636673, -0.30918022503968800585, 0.047530875710292762315, -0.010697873531125499588, 0.0027808327549509563592, 0.015083304574633544376, 0.017884682433273579177, 0.0011838286047985389915, -0.026816070707117528604, -0.026554698980367830163, 0.027033051459327231414, 0.0506184262044038491, -0.067838372521963521966, -0.22801431296553401062, 0.025047910908544322728, 0.75363966609310950595, -0.5421432907354792885, 0.018298292288746617618, -0.0044582205218478944755, -0.030897784101394473733, -0.048650667110178888453, -0.03761830169653888889, 0.0028579248868601664357, 0.028213876779101223352, 0.012994018905671731595, 0.084593883315725287853, 0.50225150598703371951, 1.3125995286202787327, 0.93442712471256961226, -9.096431614355591222, 6.4465692433819823393, -0.014612180599762592675, -0.0091967300063000995136, 0.002258948041791847966, 0.024803730679729636688, 0.058562330500671284927, 0.080643839513304754751, 0.030188118424618669983, -0.16052513800989662296, -0.40436524230975989802, -0.2185916225183171524, 0.5943010882665202832, 0.63349570600480609173, -0.32665486383528452574, -0.30830874812993191991, 0.010463248234017527225, -0.0076110969415919636077, -0.025555229307439170094, -0.040426313418374375253, -0.042542739630820496088, -0.0060482607095360754745, 0.10371402862860013006, 0.274052616629515855, 0.33078941546562173892, -0.032384963073274493361, -0.78088745809270276865, -0.83439059218608224722, 0.57895024484670143394, 0.5239370313841041904, -0.008396651852744270797, 0.0053059936000227955061, 0.018397727146874865573, 0.025924674060632379308, 0.020197126282739747555, -0.0044315556094136109369, -0.044638060963675271764, -0.08178125111145041759, -0.11421072406580387582, -0.18175946149935651164, -0.16735150344853882975, 0.5672217515709642166, 1.4928377586581389469, -1.5756324151322858018, 0.00039209697098061696303, -0.0027233601279753622687, -0.0032610289607986689409, 0.004261700529915490171, 0.025154596372923963832, 0.047604826735136940297, 0.027391334507172172136, -0.070367761668883574422, -0.19020593454331474748, -0.22824066782327886993, -0.14331348209350602829, 0.50290835696587155024, 1.3894628048647359453, -1.3968247951957371011, 1.5096649810736046948, 0.47823369772043750059, -1.1206503883414831169, -1.10999525100710561, -0.41722447617782620899, 0.10422528472996744897, 0.28813942209957121854, 0.2639314511844673139, 0.17834288974670950423, 0.093447166966014780232, 0.026303904942620540131, -0.02071653576294113161, -0.049319894815771321561, -0.063577847067774181933, 0.17008001129854419253, -0.14982827615160762558, -0.19408847211977151015, 0.16339036371703907768, 0.093948564672600265446, -0.04938947382409131226, -0.048110867935202222823, -0.0063142984435533983101, 0.012942986451871085762, 0.011953693150153503824, 0.0053054670442455624177, -0.0001729179371732185032, -0.0026215083298730375641, -0.0024951845782805520332, -0.028110257119577295587, 0.13822339491430055136, -0.12992027278491610764, -0.0037105090684200884611, 0.028850307483713772544, -0.0074399716350721937022, -0.0068745464522774078417, 0.0064889153047645994551, 0.0066991720011273731342, 0.00014125579592255698997, -0.0015082257674311296859, 5.2878627191762424286e-05, 0.00041478211903239297883, -0.001266481934724563883, 1.1767655182883856035, -0.24570955902850161179, -1.0027524583604561226, -0.47187371660015353658, 0.075382856988777641183, 0.27956228197457283491, 0.24395630675379353924, 0.11791036884768013593, 0.011439162513019485515, -0.041605143206264738609, -0.049455118203272814548, -0.033338002304177341562, -0.009742707688613785541, 0.013324014941679616572, 0.083247365402812995017, 0.2417367088430356914, -0.23062190629748952064, -0.13954972530004000619, -0.034427216199847968103, 0.00464523814946356077, 0.034050173816203746502, 0.039741556660536803136, 0.028343676953315379702, 0.011355775438447372402, -0.001968806017120893273, -0.0081168373869018612538, -0.0078261346942824910888, -0.0036955649805754079047, 0.33589295088158521718, -0.12564601520133736323, -0.50002143236947682592, 0.052132663927842499973, 0.26141375699196556237, 0.091888959350831175832, -0.044519291965658680521, -0.062191446552631049483, -0.024245112285303866556, 0.0079594743814053181552, 0.015974752064119600303, 0.0086635287500490300305, -0.00069812409228003403667, -0.0064062720744851482657, 0.15489351654755056931, -0.028566669616086635963, -0.36784904515199856601, 0.029681630849249615589, 0.22837286357804442716, 0.1036612132853619167, -0.038516484246739861319, -0.066545648015686520438, -0.031039565505428991926, -0.0011771804295865819119, 0.008068808913021028803, 0.0063310203771604891751, 0.0027052259480548029695, 0.00035725024557448331174, 0.3478600895176793828, -0.34262313520073600648, -0.46572681187178044393, 0.10537284980751993047, 0.29445230529013471932, 0.19788884038525411668, 0.043528346693873891382, -0.06322456801809966942, -0.084981485460989547587, -0.05706514510447746602, -0.02182485157193637948, 0.0022517500262142086005, 0.014767181715462167693, 0.019753603717482640795, -1.4104485215996953595, 1.650464480493598618, 0.39829920448399980248, -0.45013200149556031882, -0.30515757730147669013, -0.036479503690115253067, 0.070183951743960956704, 0.06389630469093371401, 0.023753453200019321312, -0.010785078333041386162, -0.024066401099563344834, -0.018357867666985681898, -0.0023921038413815477915, 0.016660246763294431216, -4.7352094820401653408, 3.0236604951847527722, 2.4346731520709279373, 0.46992968612448515753, -0.35194374867569283571, -0.50450534941817271051, -0.40588116023232106988, -0.2358950495497401656, -0.085581624068987477427, 0.0099875274390232374189, 0.053922128388313485259, 0.059577781582321950871, 0.041328957721446964169, 0.010955759714910653924, 0.50982366217213526127, -0.67106876073368126878, -0.50410711469535030016, 0.18708728685065420527, 0.34811904352477013314, 0.27909182655833708608, 0.12004673502184609601, -0.047403259552952874134, -0.12678443608195694159, -0.11742467032725761866, -0.065953706511741622132, -0.011619159881486864835, 0.028028688315097166628, 0.050901221035620065891, -1.2661028444431297046, 1.4893636548807860365, 0.38667780613989122251, -0.35590571384185915038, -0.26539170733119993306, -0.08519038162949540971, -0.0051533782408208113346, 0.019954822346532397059, 0.026831840506664705881, 0.0240640156197919515, 0.016076341555708750786, 0.0057754216178015687841, -0.0036269104289775389374, -0.010431951549993677408, 0.0058566072514449913494, -0.39549208244326555795, 0.23719055692970694715, 0.45616908429009994652, -0.040195401301492413026, -0.23749198287996065027, -0.12287232366372038972, 6.3314636115243457665e-05, 0.045079753429498964357, 0.038203636184215204741, 0.017958614856483704619, 0.002715529047588716334, -0.0054163393795808260217, -0.0092132174473721073493, -1.0765786056829640405, 1.2289149382894695162, 0.2308718030132571597, -0.2924576799001499805, -0.13486643026433423054, 0.01019380182042903607, 0.026606247149436941524, -0.001930094991196998214, -0.018594355642577616028, -0.01573855000348995184, -0.0037859502064181765543, 0.0053749706498483779821, 0.0077413290428917863953, 0.0048124337491009614581, -3.916790949752365858, 2.5605262251951219632, 1.9005858214446211818, 0.35902159424227958473, -0.22790387724022684313, -0.37390142710886564714, -0.3346289934425228374, -0.20779811331138847708, -0.07805744522113197692, 0.0059560034745278055829, 0.041775944554150457044, 0.045111623010379610366, 0.032302414688895358941, 0.013471393230552490203, -0.90052810299176089259, 1.2948833983696887984, 0.10316817238337974705, -0.40957844739574100767, -0.21025195129792070836, -0.02796132982887199192, 0.066455290079801096481, 0.076375609653449069714, 0.032745059086420759409, -0.008709614062144483207, -0.020800757091588725767, -0.012698881114377899917, -0.00089197739057840161325, 0.0069846341724917699809, 0.45103589477056421631, 0.59826111699493511598, -0.68050955061188511674, -0.74357682511412204196, -0.13848707452863354139, 0.24505992010244798118, 0.26564303449638332344, 0.1338062223845191967, 0.022961354192637471738, -0.028018028263271207295, -0.037858787667485904738, -0.028799838696538595378, -0.012376396650510795849, 0.0062816555071046609182, -0.30830874828145982125, -0.32665486347852124016, 0.63349570592711546002, 0.59430108806331238824, -0.21859162239535137595, -0.40436524269709300805, -0.16052513770801712933, 0.030188118802698872545, 0.080643839310734682924, 0.058562330039803676962, 0.024803731066332504462, 0.0022589480030250595755, -0.0091967300109605867875, -0.014612180620436114598, 8.2144700471874259051, -10.049456675904840708, -0.10093568617431560031, 1.0830001580307382358, 0.67876938407913478546, 0.27095849516495562614, 0.10761150221487217338, 0.076536735839754815869, 0.045066005740897557841, -0.0026150395685489401243, -0.034224621294085495748, -0.039379274547862164191, -0.026768763176875209897, -0.0066375687623960880415, 0.63759473994354765036, 0.19079482257169341763, -0.67482221928907548669, -0.43116847134907643202, 0.017578910637439017911, 0.16257914789181143744, 0.096353991407237091216, 0.022747620082227061611, 0.015637855946497199178, 0.030193283691064536772, 0.026937796737586956752, 0.0065861846252461813539, -0.014810469675521673039, -0.028834263528581153407, 0.20945508108894822508, -0.69573938429661397187, -0.065282657695146864563, 0.56170500240301235451, 0.2798235992657439053, -0.059882841282108648095, -0.15168251265582743947, -0.10989692456745603355, -0.040929475348217080055, 0.0040854637083293787736, 0.019824927818939384189, 0.018245519763466114954, 0.010008896673436335975, 0.00010707718926214553859, 0.12536820734019354928, -0.34666727107751255099, -0.12328833279035698867, 0.29379998306869908875, 0.18063522792584621457, -0.00021063359867721338173, -0.060693938366035281307, -0.060913233358612955959, -0.036858995032020926141, -0.010670839478238932502, 0.0053960012640699517653, 0.010177780084069125779, 0.0082008851793668423313, 0.0033153833628626508227, 0.47823369768385159961, -0.7712394699630962247, 1.1290453524488825998, 0.073922665869895223811, -0.6168460530185970736, -0.48761142050729588959, -0.13789906138191651563, 0.098153086415143725452, 0.15490436199864612843, 0.12653321973061476546, 0.076083294390240474581, 0.027389032569691735414, -0.014769207557011320223, -0.049319894771351908969, -0.14982827593421901846, 0.13602250575121763165, 0.37709333193953725738, -0.46250093083404830718, -0.053361711012925153175, 0.16507460644292767937, 0.032898962723057638413, -0.033273997713970986101, -0.021357266989413852681, -0.00079955787583873357167, 0.0060562386592225838852, 0.0053180053152860244994, 0.0016931583467691348373, -0.002621508960311477298, 0.13822339539065836567, -0.39899101812354959007, 0.41802711039514045543, -0.12035451223261549314, -0.080430820144039447128, 0.040570640926555254224, 0.014091160006066538041, -0.012912997544642050921, -0.004903068719713705309, 0.0070459371915812070084, 0.003976079575558939469, -0.0012812797680934063377, -0.0018368035099036847304, 0.00041478134404985135108, -0.2457095590138996255, 0.14403212151696592747, 0.86088414023394943531, -0.34849253525342294147, -0.48091503185020473277, -0.18210751235087610511, 0.037588835392548383685, 0.12689287201998764032, 0.10568104192974347999, 0.049745670013533048459, 0.0026208021610465790513, -0.020060637780346314563, -0.021487035009539096758, -0.0097427084691776508207, 0.24173670897976565208, -0.58188066730198895193, 0.54978520250041817974, -0.074679106196775799087, -0.12829910155544518857, -0.024967995822157539737, -0.01402564093214495565, 0.0040061485572453333845, 0.015224026741620745595, 0.017400340828562368273, 0.012164842381616306627, 0.0042778995887930229558, -0.0030367992278934811783, -0.0078261343484584602187, -0.12564601519687704223, -0.14834563638467759827, 0.8326712445299834231, -0.4274010743424120351, -0.36544442682002481471, 0.094461818486523893568, 0.14526559127138724548, 0.046847430076628678031, -0.021254571163349913338, -0.030201865217832961352, -0.011929850999430854572, 0.0044802551273099861656, 0.0073758616701453888834, -0.00069812481029947297262, -0.028566669609594232127, -0.29175530274780103746, 0.69148147705701490384, -0.24167188298513611611, -0.30121414571093596546, 0.037852552933634313337, 0.14010845986967160326, 0.053181897369960949573, -0.026473947878352357271, -0.032916248848325488363, -0.01305007118625389588, 0.0013310744805975606789, 0.0052798948343638733849, 0.0027052259519667804823, -0.34262313490940321525, 0.2237872132775493017, 0.55891574451332870499, -0.43967767793295964296, -0.20896633494712790746, 0.054989946678632287136, 0.11270258367666222132, 0.089966809710600698091, 0.016729445836876567083, -0.033103822967154805657, -0.039716741181875167555, -0.022329925767335750963, -0.0017055281629468812404, 0.014767181953974271857, 1.650464480663792477, -3.2604644188045979192, 0.86149144888494710859, 1.1046011814142842056, 0.013643354920862479918, -0.25909490547107272285, -0.13590171854946717089, -0.018718102462525983143, 0.021090222845826608888, 0.023569307700749692058, 0.010071318008504359653, -0.0023938670742453963328, -0.0064890492539799492569, -0.0023921037790565749179, 3.023660495126423875, -5.7297386599883504132, 0.25179077255417381398, 2.1043784920619068579, 0.88641205197982941844, 0.079311866460205715512, -0.19658054824013004858, -0.2425309314555777751, -0.18872064772703531221, -0.10307088127430172331, -0.033094835835268773816, 0.01167214555148313182, 0.034362510402628561323, 0.04132895807085647244, -0.67106876071630394698, 0.80893983289255588787, 0.30609769766371247579, -0.58754654393575556082, -0.096597611310207304181, 0.077225309128801017389, 0.095617279922814713, 0.10909718433120563219, 0.044486065527345880488, -0.027831285103235514633, -0.0572349175284962014, -0.045937088001798825898, -0.012924100700007299519, 0.028028688211474479808, 1.4893636547203845666, -2.8953201854019008366, 0.79447206006322168559, 0.94398416583732147611, -0.014383403689574650319, -0.20857945962971541354, -0.10561365259307398112, -0.027627861912867826555, -0.00029031381671054332094, 0.0098655045110907103184, 0.011263689357519831788, 0.0094022189203675429758, 0.0043410507879580373033, -0.003626910232676101159, -0.39549208277826020108, 0.91953849295750911264, -0.35628337000020499081, -0.64995424340009977371, 0.37515551478415115616, 0.29180282931294193594, -0.059377731612668002703, -0.11709903326342445407, -0.052449520340210685487, 0.0040636160383682798369, 0.022255635977843597068, 0.016420634001766302307, 0.0044130451288026282738, -0.0054163389232460990713, 1.2289149382540807132, -2.5370431680772154159, 0.82786871845526277536, 0.77900693798145415059, -0.070179022965865714423, -0.17947152777449945638, -0.061613752755476224043, 0.0060969484893076929255, 0.011451159400291160279, 5.799113692346959982e-05, -0.0084649236749928039036, -0.006859147461003795368, 0.00021271575717056418831, 0.007741329364711114526, 2.5605262255359999557, -4.9770017803073498541, 0.46039238032684876023, 1.7368031292798222776, 0.6203667783471846775, 0.060324804211008997701, -0.12054949266441031264, -0.17900424815157486846, -0.16015482601961567033, -0.095084412400908147056, -0.031750880049840154995, 0.01031771893699895068, 0.029570540917135547387, 0.032302414215923738738, 1.2948833981382754654, -2.6373398815626627467, 0.98049436771148168468, 0.72934231386859527113, -0.13842670757672592696, -0.1603386719157453566, -0.092991663410709940463, -0.031734789554504802545, 0.024973227817925498784, 0.038540528809042372271, 0.016276170007440721177, -0.0053030618186039359119, -0.010218058000426973042, -0.00089197691745991321624, 0.59826111679083937567, -1.3728130618933545026, 1.0245218173295782549, 0.37494529315384322876, -0.43071029998844084785, -0.38198102500992531239, -0.053179665777605258925, 0.12876448572623647171, 0.12067377091101977316, 0.05507619237578280369, 0.0059757265364858147549, -0.014401156625268321521, -0.017428997674196155393, -0.012376396263069085191, -0.32665486338655691512, 1.0465952435197183323, -0.81594507169987107176, -0.6249597781169994315, 0.71667925317999170431, 0.35555853426646794047, -0.17290750728249798529, -0.19512217271403733965, -0.064141512881000908863, 0.020076604799044175942, 0.038572356334538322742, 0.024361149111632673936, 0.0049033669927419434864, -0.0091967302297491747287, -10.049456675828645658, 22.454085832369692355, -11.792648087176026195, -2.0117265348744939324, 0.31503040549621230682, 0.58155081000393593627, 0.34623551425493992451, 0.11710147833093130509, 0.045462895736013558823, 0.039895466101142751381, 0.021101871510994928138, -0.0057436546969116056283, -0.022884096366647375259, -0.026768763687931719508, 0.19079482257917507182, -0.5732793101961283222, 0.89933590632846782675, -0.043687468286092226433, -0.44811577917912354119, -0.22193377476456344066, 0.042797862485594057036, 0.11717216941433579447, 0.056211696203311220521, 0.007422188040141830534, 0.0024427828904342446803, 0.0095596008264311471692, 0.0042737825154711556186, -0.014810468764777449946, -0.69573938472412966139, 1.1587065871601704981, -0.033924146488214276807, -0.8369616254217990825, 0.099244663163678054074, 0.3248472668736456237, 0.10847527701899836727, -0.03214841568196356858, -0.069966280376619130466, -0.048694896664669327924, -0.016311866731803886105, 0.0043465736559708161371, 0.011307284867976217721, 0.010008897083342125842, -0.34666727152489523744, 0.47843484754920689861, 0.17104867250853045268, -0.50735829770003193495, 0.020985072814569100119, 0.17695515006420131643, 0.058267971457844680572, -0.0034024155440000800055, -0.025683747933444134315, -0.027195587751682286093, -0.016432724534998794513, -0.0032978037556322045698, 0.0050733405441301393513, 0.0082008852921479080678, -1.120650388879832926, 1.1290453527583306226, -0.31575457213433494319, 1.243538272846941739, 0.12239377505994905593, -0.62906348577523829579, -0.5400792151677871189, -0.21414884550238819094, 0.034757181466073830411, 0.11839540656333470692, 0.11321970519342557104, 0.074707530632043384067, 0.027389032749863782834, -0.020716535874571933434, -0.19408847201876944233, 0.37709333177906267931, -0.30969749869016410804, 0.39090810214207288853, -0.33749737490451947597, -0.057944947834730926861, 0.14706724591231812305, 0.029892483526567571739, -0.033118861655560154211, -0.023037017499012209937, -0.001699559851195689246, 0.0062674985945364112411, 0.0053180048232835796379, -0.00017291710265631720137, -0.12992027337394404518, 0.41802710996175601599, -0.56910188475664336138, 0.39015395962254956208, -0.062156043805283070491, -0.076676645140032334558, 0.025304204532846492903, 0.0098139162265276184971, -0.0081834357014834401001, -0.0037892751822646356108, 0.0052292336014986354198, 0.0026271780774048169096, -0.0012812802335392657095, 5.2879863009514640256e-05, -1.0027524582802169739, 0.86088414027796245076, 0.21361196632565346287, 0.83960355230433469931, -0.38545319714261994104, -0.56451992753043844786, -0.24623071062515877738, 0.0049205457137685287838, 0.12477002208420628904, 0.11940963286619485362, 0.066123874131352317063, 0.012912732314326942629, -0.020060638324930584603, -0.03333800160122539491, -0.23062190656050876592, 0.54978520287330656657, -0.68376482848500619305, 0.54978296655140168614, -0.018504049348313258289, -0.13360519316351077146, -0.040342519680050772934, -0.028754431552780970738, -0.0045741203673351294237, 0.012915756032003962439, 0.019144352585615528189, 0.014440985347180105777, 0.0042778994691345124252, -0.0081168380316770198657, -0.50002143220117223343, 0.83267124445595852578, -0.49615529119088080945, 0.60276039747348875153, -0.3580265567907847668, -0.32274322638882796399, 0.10553027347329506369, 0.14492839679748933945, 0.042972048615365340818, -0.025178304355905782586, -0.031974381682789522474, -0.01250128951216888766, 0.0044802545203964537143, 0.0086635295378118207937, -0.36784904469998502829, 0.69148147698419126783, -0.50179970407772889196, 0.46977816045414777513, -0.23191070401401578382, -0.24674863335903582384, 0.055871655947253649754, 0.13527586227713042644, 0.053086874487928981114, -0.0242927399150142595, -0.032232015516716276682, -0.013442974254128343264, 0.0013310742733998622272, 0.0063310202890427876715, -0.46572681239274987819, 0.55891574478489391975, -0.062206792321009098112, 0.45124251488804684174, -0.49656594576877555181, -0.25675048588547089157, 0.06445842743998969615, 0.13472416497695685611, 0.10911504823208251214, 0.030702198813567477265, -0.025171129562392771661, -0.037661212385176959239, -0.02232992542785955381, 0.002251749854964216549, 0.39829920427263437732, 0.8614914489784634144, -2.7592424808739153264, 0.775997690370105353, 1.0697785842922509225, 0.032749264147496320865, -0.25834272707213601361, -0.14980871979060672583, -0.022260702265329616184, 0.027605159565917750802, 0.033666155944977488312, 0.017523753674367481548, -0.0023938672177733046412, -0.018357868018901400686, 2.434673151998755003, 0.25179077274284589194, -5.4685627743314295657, 0.036902495799855911107, 2.1037636201443641326, 1.0506755330266079884, 0.22941257028871098189, -0.11073337197825322553, -0.21057133873878472463, -0.19452400321806723182, -0.1242457154995095514, -0.050412124655337620105, 0.011672146084076781558, 0.059577781334970082716, -0.50410711474510150332, 0.30609769785733864644, 0.49538064381002749892, 0.30858149094811498925, -0.7068985628377675301, -0.22956734695866271023, 0.060681540301678271732, 0.12788261123204519576, 0.15113467830023213923, 0.080231254705583510134, -0.0065772453010048970157, -0.050228145087773817812, -0.045937088079949658925, -0.011619159737650880046, 0.38667780660350575817, 0.79447205984838620285, -2.5209903637419626143, 0.67597227724690278627, 0.94103435116836442464, 0.04439555057705768526, -0.17986228926344693946, -0.10777036949663784671, -0.038103752857829437106, -0.010974282313634745045, 0.0034934172607995438165, 0.0090368185315081762943, 0.0094022189000036657247, 0.0057754209763159324725, 0.23719055755620002679, -0.35628337031275902902, 0.23426069050248718439, -0.051810081897263604578, -0.48590722554232618213, 0.34854821096455712581, 0.28131442717760679617, -0.05979124966067815572, -0.12100775648784102556, -0.059139279628317970028, -0.0015964732290138033098, 0.019770194661750409293, 0.016420634352223163754, 0.0027155282352642352156, 0.23087180271320847935, 0.82786871824018104515, -2.2216487859570857033, 0.6737360705058229593, 0.73786508920721305405, -0.046544445128096356701, -0.15838399387045262734, -0.051072524690024975746, 0.01261475083521578465, 0.012694918766682085354, -0.0014891096705791345006, -0.010410684921383141568, -0.0068591474772147376787, 0.0053749704600211979655, 1.9005858211775417121, 0.46039238059691200977, -4.769827184835053302, 0.1948273166246829724, 1.7361160232981500684, 0.77246067320586608851, 0.18732633396811670945, -0.054470865148362239416, -0.15556347548500978895, -0.16335640501518197687, -0.1082134376549852145, -0.042379133020231968265, 0.010317718803714404086, 0.045111622858478202247, 0.10316817283066319344, 0.98049436735007255983, -2.3365638406276194594, 0.85771056956667934834, 0.78611479218624968901, -0.12816480663939933127, -0.19319357526735175812, -0.10748077278588076544, -0.027285008668880776639, 0.030758003219896773439, 0.041476467563723119769, 0.017413907490262996169, -0.0053030616928413411293, -0.012698881545555117112, -0.68050955023094417129, 1.0245218177559598516, -0.8814506292924073616, 0.9674705540284355143, 0.27154700693549471691, -0.48761087263812341597, -0.41478837018199476994, -0.082406923056781850034, 0.11608467846847170257, 0.12532871952752522304, 0.067878615239285414185, 0.015727159479661995123, -0.014401156626803746086, -0.02879983955324319847, 0.63349570593912540861, -0.81594507157535600683, 0.25973008158813076784, -0.27933845480861951849, -0.41551729207326154603, 0.66307699040967904569, 0.34278892617292744394, -0.1705606021763935165, -0.20271415013333154986, -0.076255684583342142613, 0.010778488079680043077, 0.034822679378737134148, 0.024361148703906050639, 0.0022589481006197679319, -0.10093568657410165179, -11.792648087179964378, 24.61765530098942989, -11.587351506552320046, -2.2341438266709099381, 0.11903394178190168651, 0.44059433821449323077, 0.28603774610196475159, 0.11468129058489512362, 0.06005629771533524397, 0.054843926827812014557, 0.031009281853315667776, -0.0057436547368695592408, -0.039379273239280046526, -0.6748222194527426776, 0.89933590647189531886, -0.45169626588967759506, 0.76837786602218893783, -0.090849512069864551966, -0.42492490674775340587, -0.19402990844089251787, 0.029713149334917032807, 0.089188946726258794273, 0.036763647138467692166, 8.51339893837288475e-05, 0.00098792213024380404196, 0.009559601566107141235, 0.0065861840453516843491, -0.065282657437971361447, -0.033924146743535298076, 0.40196130469454016687, -0.0010553039257095008646, -0.74544809280105428506, 0.10519252615058019562, 0.34676675628075115387, 0.12909228443065443037, -0.021828078955117194426, -0.069417032457554511682, -0.053459870703702970507, -0.021049261910144008436, 0.0043465740064528111258, 0.01824551974264386478, -0.12328833221807909692, 0.17104867194914585848, 0.013479742320851612059, 0.18572121204534042249, -0.46101414796444334376, 0.0018664401549408158459, 0.18787017251384832695, 0.076778579407174016302, 0.0064155438290353482972, -0.023739915927240055893, -0.028746827823542490249, -0.018665939569030676193, -0.0032978032391298168352, 0.010177780232187223428, -1.1099952505097914202, 0.073922665174847829173, 1.2435382725294121808, -0.42075620797766055681, 1.1715483573633727055, 0.11784851793969094991, -0.63549469975199979199, -0.56086304933446573528, -0.23913894493014142828, 0.017770048545296551062, 0.11172196524860664169, 0.11321970532318842428, 0.076083294434134446371, 0.026303905155910412006, 0.16339036303751527934, -0.4625009303881160716, 0.39090810231004446962, -0.20618107798461710045, 0.4104218890754111615, -0.34935250908387577295, -0.074561529128463310823, 0.14135398881834748686, 0.030608586635664718906, -0.031062995029025385757, -0.022097798749913587918, -0.0016995590754052720961, 0.0060562384256633329527, 0.0053054664151450164553, -0.0037105090595853798888, -0.12035451080366836585, 0.39015395819473941419, -0.54089768784097103094, 0.39376531197073216806, -0.071156874627564717373, -0.081731910079980457651, 0.029229195391775965318, 0.011676794740582183141, -0.0095359865932151278295, -0.0053970842196313616179, 0.0052292345964962605107, 0.0039760792794905821385, -0.0015082263831413715918, -0.47187371728236038315, -0.34849253494225856942, 0.83960355294989574837, 0.089173739666337931142, 0.77330488838618505198, -0.3851334291079644423, -0.55375379114748157772, -0.23815408886388741982, 0.0082908896490067954255, 0.12750344746010239261, 0.12244756064227858838, 0.066123873792493376156, 0.0026208016901422253458, -0.049455117132436790528, -0.13954972524493838848, -0.074679106644441120566, 0.54978296722346198244, -0.67655851854968585002, 0.53445639502688757627, -0.023664466612996426509, -0.1403908872288616716, -0.043244561806583856067, -0.028564610239396517194, -0.0030420350276592535849, 0.014013711541768239521, 0.019144351973187732213, 0.012164842595212940063, -0.0019688055414525916509, 0.052132663549629164323, -0.42740107411172456597, 0.602760397565923145, -0.47189580429724747868, 0.67531246991157500847, -0.31270591473335829225, -0.32905655524712251747, 0.083912895881383731123, 0.13275704986365810534, 0.042440628332444461923, -0.022814687925655462164, -0.031974381902512789833, -0.01192985087005058939, 0.015974752017955725497, 0.029681630002858453571, -0.24167188266404490693, 0.46977816078885697992, -0.52755787378856811021, 0.54531463700230020653, -0.1862169261542691856, -0.25421039764488406743, 0.039672233682429447477, 0.12837588913671549307, 0.054229357213690890238, -0.022208216188807177599, -0.032232016085905985936, -0.013050070652661043685, 0.0080688088381771355673, 0.10537285039843749535, -0.43967767755928671702, 0.45124251401974485587, -0.14647326182969372388, 0.45059610812371853061, -0.44830494084350341888, -0.23309954073958091825, 0.066979531478192202987, 0.13171035134003597489, 0.10923267640748743346, 0.03248354419771778967, -0.025171129577615070777, -0.039716741376261965879, -0.021824851171610204492, -0.45013200084282811719, 1.1046011816542509187, 0.77599768977637317846, -2.875489303245884809, 0.68743359050894004003, 1.0513025804681008779, 0.046156742058502384096, -0.23818027211940145937, -0.13580967558888865088, -0.014297569338625949753, 0.031420214306963484274, 0.033666155428842409192, 0.010071318152345711677, -0.024066400923136530704, 0.46992968616992131237, 2.1043784916042516109, 0.036902495938347434912, -5.4463893369236862085, 0.032939609541470238341, 2.1170517119858027755, 1.0819152206188635823, 0.2583435334561299257, -0.098783939203819115527, -0.21255290350132510779, -0.20123196499905715573, -0.12424571492411889961, -0.033094835934817129486, 0.053922127804466028766, 0.18708728694308007778, -0.58754654419510932151, 0.30858149103592980023, 0.36616706085295591677, 0.24585167298590338003, -0.67757069362897381914, -0.19115171698556543411, 0.087055999510231019123, 0.14058513504506975345, 0.15660986274234423021, 0.082927500212611193753, -0.006577245386736596533, -0.057234917449666133926, -0.065953706322658608996, -0.3559057137891667999, 0.94398416583664412904, 0.67597227738157072974, -2.5940994333745717171, 0.63931288695269550271, 0.9475995886517774025, 0.057673232867036933191, -0.17453893996530786548, -0.11038449348753666401, -0.041395126386033785637, -0.01258435136941905147, 0.0034934166704031627271, 0.01126368953450568311, 0.016076341913302764214, 0.45616908396897448696, -0.64995424374472643425, -0.051810081215767717078, 0.42793939158974247938, -0.041846656798407357025, -0.52989827702044278723, 0.31243321301766724396, 0.27640170950444636899, -0.053273071654392850638, -0.11768435235071753042, -0.059809475973250703085, -0.0015964737134429242589, 0.022255636118345033941, 0.01795861532596391627, -0.29245767925768501039, 0.77900693811976851055, 0.67373606999451907118, -2.328907251821568547, 0.62512068065881343237, 0.75287318393547819628, -0.019060966099039518273, -0.14288498716171077474, -0.04994843693802675233, 0.0090717663356102429262, 0.010234284769902587606, -0.0014891096812568362789, -0.0084649232330330833468, -0.003785950555292555593, 0.35902159453405102507, 1.736803129448320826, 0.19482731645123571584, -4.7662593786144471508, 0.20854970997925073584, 1.7598652371299199348, 0.79545238637279147564, 0.20473789423587029912, -0.045666220765482268273, -0.15323588755915615311, -0.16504064830867093416, -0.10821343767987046114, -0.03175088069313657807, 0.041775944893084754339, -0.40957844780765084725, 0.72934231417502803918, 0.85771056958739044784, -2.3669884263122549228, 0.78423759269817061579, 0.75054776598884376249, -0.11886976360458959001, -0.16980764831869465947, -0.09470478525834351291, -0.02646635665467907822, 0.02816031548886241756, 0.041476466891387619995, 0.01627617078486787508, -0.020800757260398972798, -0.74357682558377169446, 0.37494529234053469713, 0.96747055514149471467, -1.095947875530758342, 0.86515113625414841181, 0.28561419282812994247, -0.45899358312947696747, -0.40464502364430610104, -0.084086505581073278659, 0.11586584014573522716, 0.1281127494558299551, 0.067878613791674966005, 0.0059757270001397355674, -0.037858786655020792988, 0.59430108794658198512, -0.62495977800922708489, -0.27933845491059772215, 0.52500955969990648065, -0.30624728598666361989, -0.49627527115841246763, 0.617606311537117314, 0.34404488781161945976, -0.1583429779262555126, -0.19882962010411242559, -0.07831652268211641843, 0.010778487511252793363, 0.038572356625153403353, 0.02480373070187468984, 1.0830001583377570906, -2.011726534887295248, -11.587351506858954764, 25.308265187788901329, -11.405885491641168272, -2.3364542880900964761, 0.0088273249036024403757, 0.39694207793780067473, 0.28263589461978527728, 0.1226216310857194125, 0.065914280886064241516, 0.054843928125181184985, 0.021101870967820804403, -0.034224622079705641076, -0.43116847130951896361, -0.043687468720028044267, 0.76837786624019122073, -0.58040788016661848392, 0.78351823083163496531, -0.022016708997113720869, -0.41030110744264230993, -0.21968112330282571487, 0.0032727318838908781018, 0.079148413097338332634, 0.036495065507309243369, 8.5134312732992878381e-05, 0.0024427827973864113531, 0.026937796513553241562, 0.56170500265724343514, -0.83696162579861732578, -0.0010553039760323110197, 0.52338154890326182844, 0.077200003901299163456, -0.72429760244616481213, 0.086633530898491017957, 0.33248337878517225663, 0.12574677920033688894, -0.020999028103280640428, -0.069465575705305995791, -0.053459870159198008555, -0.016311866547584909226, 0.019824927211801287447, 0.29379998331771794984, -0.50735829770424678564, 0.18572121207860634029, 0.06188600138404966311, 0.20912872924329015123, -0.44348116294593314946, 0.0014789089763957777302, 0.18370326415876139792, 0.074098161126439410218, 0.0057917062507868780075, -0.023927403335305741372, -0.02874682643298392612, -0.016432725260546059143, 0.0053960008879588503006, -0.41722447614420937745, -0.61684605282388704772, 0.1223937747504297846, 1.1715483572579425964, -0.46510669350271283129, 1.1529625010056148149, 0.11476496080105516584, -0.63711296385017490707, -0.56457110901295959327, -0.24303363144759312919, 0.01777004857920498615, 0.11839540595027214787, 0.12653322007196335375, 0.093447167366174962999, 0.093948565177749618349, -0.05336171115104679924, -0.33749737542880065, 0.41042188934085549468, -0.21627575894517503396, 0.40286198321439664616, -0.35344657723688077677, -0.076963663968350914946, 0.14200820203110445239, 0.03194573874767639432, -0.031062995645785254628, -0.023037017776431998467, -0.00079955802364690682464, 0.01195369386557154287, 0.028850307604535576494, -0.080430820752452669709, -0.06215604372612564682, 0.39376531266580316881, -0.55573392347191374885, 0.38963137233133732229, -0.065789767195955181633, -0.080136367803441341251, 0.027646235747674736266, 0.010249894346042353471, -0.0095359867675778621499, -0.0037892748061048456623, 0.0070459366270361062368, 0.00014125603065366499462, 0.075382857220419943745, -0.48091503134657193064, -0.3854531978572103279, 0.77330488842356648327, 0.07601954069274005632, 0.79465889901565278652, -0.35697878788621933976, -0.53764439650568629503, -0.23330859772915965578, 0.0085609485966306603233, 0.12750344857329176662, 0.11940963337944038303, 0.049745669289413062264, -0.041605143759445302465, -0.03442721580340156462, -0.12829910133186980814, -0.018504050219014297335, 0.53445639537015754428, -0.69269138347363357067, 0.5259704929960663522, -0.02370478704643158227, -0.13738927523366475936, -0.040056154627317194783, -0.027038202379252071483, -0.0030420346797111648407, 0.012915755292675691959, 0.01740034147371988249, 0.011355775099330691486, 0.26141375720339288247, -0.36544442706552004996, -0.35802655660409132832, 0.67531246983075310375, -0.43337441662102010875, 0.67992341247010756877, -0.3250327992483405537, -0.3420727095231678172, 0.079309406213436725586, 0.13346557231542616195, 0.042440627724339798077, -0.025178303416404784754, -0.030201865950882231943, 0.0079594746548821114496, 0.22837286399067169285, -0.3012141461274508436, -0.23191070347153708298, 0.54531463626042953052, -0.48883863714499070641, 0.54464472559913346128, -0.19473195680182869838, -0.26151954904617508157, 0.036783135946200584687, 0.12858871572563987384, 0.054229357057734198233, -0.024292739783118969671, -0.032916249250286154504, -0.0011771800421351305641, 0.29445230457411158964, -0.2089663350485337634, -0.49656594492663447271, 0.45059610816448969484, -0.095825903582730465047, 0.48886632265400892239, -0.436059674094825378, -0.23613919030602920079, 0.06216914415872350913, 0.12976230469223695718, 0.10923267709717665586, 0.030702199141762566426, -0.033103823099365395355, -0.057065145715731457887, -0.30515757808028504128, 0.013643354620007593775, 1.0697785850115952844, 0.68743359112369972763, -2.9321922279316776638, 0.66601943817428987948, 1.0631082779369041003, 0.066130765059406876327, -0.22444076306051319314, -0.13052657591284635985, -0.014297569476686561307, 0.027605159076252671485, 0.023569308079719571841, -0.010785078180134549664, -0.3519437484650758674, 0.88641205199820549687, 2.1037636204399019491, 0.032939608774647632927, -5.4654378923846458349, -0.0019653577368277108661, 2.0960845940428769474, 1.0734996600518877941, 0.25555642094207731141, -0.10047863028077330672, -0.2125529033616167518, -0.19452400319787815919, -0.10307088107423900691, 0.0099875272030652342825, 0.34811904353813216684, -0.096597611178424594436, -0.70689856290199204469, 0.24585167282661146704, 0.40829002924458890345, 0.30622655739631576832, -0.63965884893983027126, -0.17833533795195721794, 0.086560615126070494796, 0.13780582068724145373, 0.15660986326198111929, 0.080231255327837092128, -0.027831285288776957954, -0.11742467103160122988, -0.26539170787491250803, -0.014383403779608309667, 0.94103435149260084014, 0.63931288734741398283, -2.6221070554978762601, 0.62084705463208567089, 0.94379842867022822883, 0.058526762572732878021, -0.17311668400186414063, -0.10984147359497925533, -0.0413951264823210624, -0.010974282864366417117, 0.0098655047452379292933, 0.024064015967209577418, -0.04019540105525821333, 0.37515551506848632268, -0.48590722602547914422, -0.041846656932621538461, 0.40352837584487738498, -0.067547505219857847947, -0.54380477101052226985, 0.31026652658898351556, 0.27949107743032519657, -0.051272038156401633646, -0.11768435181957960511, -0.059139279611028668093, 0.0040636161254154826161, 0.038203635650065757845, -0.13486643034044396017, -0.070179023250074315521, 0.73786508873096567918, 0.62512068172736867666, -2.3435766784830205367, 0.62848770107794682538, 0.76337696621330164426, -0.014766921370083079326, -0.14407361848926369952, -0.052116643685517556606, 0.0090717660449777283438, 0.012694918430305703744, 5.7991707034143113661e-05, -0.015738550128691054797, -0.22790387738701836651, 0.62036677856548205767, 1.7361160232912247192, 0.20854970999332400616, -4.7735790017486605663, 0.17745899817300100909, 1.7387231443274600018, 0.78806961173365042139, 0.20548427967713317477, -0.043876978411380704315, -0.15323588822280442834, -0.16335640568006698503, -0.095084412226141806768, 0.0059560043384490873181, -0.21025195093697840454, -0.13842670782107077687, 0.78611479215391322217, 0.78423759263618608717, -2.4372776022150466346, 0.76937821348376467245, 0.77086507632488410202, -0.10083974435321835239, -0.16426161387679039105, -0.095855514715918133106, -0.026466356850326416827, 0.03075800394579226299, 0.038540528209017800576, -0.0087096140543015676538, -0.13848707405049109909, -0.43071029983325365142, 0.27154700700073464148, 0.86515113538042309749, -1.1134439215051461503, 0.8826514469058579504, 0.30935493819222298351, -0.4444064071018704043, -0.39901027762707502733, -0.082889451244606338065, 0.11586584090167661998, 0.12532872009992770956, 0.055076192215115366002, -0.028018029179807182882, -0.21859162244563831723, 0.71667925285620981413, -0.41551729111088608892, -0.30624728668348871752, 0.47159021156838154543, -0.34584827164541726008, -0.51389470064912790015, 0.6179807044165455121, 0.35024757776860793168, -0.15537504910817190318, -0.19882962123272115984, -0.07625568446398009459, 0.02007660454434478936, 0.058562330990284391929, 0.67876938379301765547, 0.31503040588968167501, -2.2341438269363735891, -11.40588549158277587, 25.454887924805898791, -11.373040285502353797, -2.3440858875848382681, -0.0023778390635039170498, 0.3923567718550070027, 0.2829651695273446288, 0.12262163068646020347, 0.060056298371273492132, 0.039895465685950266821, -0.0026150394599176382537, 0.017578910405563294428, -0.44811577870365637555, -0.090849512246547095562, 0.78351823066198267309, -0.53202672570847187039, 0.79652223038086022999, -0.035565010576498912176, -0.42723147323238908069, -0.22779012407753801428, 0.0019510505976116365723, 0.079148413253758043151, 0.036763646780332420727, 0.007422188667458334832, 0.030193283422251169373, 0.27982359968467923617, 0.099244662790468229918, -0.74544809266667488856, 0.077200003652083279415, 0.56572652576574233851, 0.079676956633746520708, -0.73913234834185714828, 0.073062992082854091858, 0.32686358841729751257, 0.12490900581372577083, -0.020999027937404740607, -0.069417032541078046992, -0.048694896308507262805, 0.0040854633827311220839, 0.18063522705088919107, 0.020985073153868705892, -0.46101414827593945045, 0.20912873028595058145, 0.089045633569060450574, 0.2213242299977615235, -0.4466139050485132822, -0.0062505100915650532029, 0.17933776654759125968, 0.072952127609427458443, 0.0057917063807210091703, -0.023739916964695340768, -0.027195587111637081756, -0.010670839260017637301, 0.10422528528990636365, -0.48761142017487119071, -0.62906348571602088704, 0.11784851690531818191, 1.1529625010075441605, -0.4787076034452122375, 1.1498720945650673375, 0.11765711603237576088, -0.63523303442473677549, -0.56457110828836776228, -0.23913894433934815287, 0.034757182744978028233, 0.15490436102057503209, 0.17834288869420186763, -0.049389473835407655955, 0.16507460673637822102, -0.057944947517017696137, -0.34935251005880813002, 0.402861983545810598, -0.21722521553216378454, 0.40338626600812610423, -0.35257599939638178954, -0.07629163976825908966, 0.14200820135945027567, 0.030608587455692986418, -0.033118862101800650155, -0.021357266698278708911, 0.012942986303076232588, -0.0074399716393053249089, 0.040570641577925614651, -0.076676645275204430297, -0.071156875619878490191, 0.38963137274359410656, -0.55091352040147623104, 0.39160016560903804228, -0.067210158041781917215, -0.081644945504957239235, 0.027646235231834993817, 0.011676795160742204499, -0.0081834363945576579086, -0.0049030681480417611512, 0.0066991720436763123325, 0.27956228197815380376, -0.18210751349398990317, -0.56451992683758445146, -0.3851334286548971364, 0.79465889890333007983, 0.10356052593464788736, 0.81616653030740859975, -0.34555059548234517974, -0.5343899288227584865, -0.23330859733059539085, 0.0082908895226237316889, 0.12477002226027233966, 0.10568104279047152438, 0.011439161593666476863, 0.0046452380311136744553, -0.024967995723119230933, -0.13360519310574028862, -0.023664466691649024832, 0.52597049286573682725, -0.69495068510329027767, 0.52868934251195620622, -0.019726085131075501988, -0.13522221120704752684, -0.040056154985016752257, -0.028564610384248002167, -0.0045741202538276543033, 0.01522402659025016286, 0.028343677295383753328, 0.091888959541780682039, 0.094461818513848605505, -0.32274322654344567152, -0.31270591507939632381, 0.67992341278078960443, -0.44358592212878245054, 0.67233684692201334254, -0.32657676815989988839, -0.34146272054337262203, 0.079309406216828248759, 0.13275705060823247972, 0.042972048195461923226, -0.0212545708351602608, -0.024245112741147299001, 0.10366121319734515616, 0.037852553231339368389, -0.24674863389387088897, -0.18621692572394002441, 0.54464472550062092981, -0.49645471662379858646, 0.54132925065626991135, -0.19508359667349256528, -0.2614377049617935822, 0.036783136114675499262, 0.12837588906667601885, 0.053086874702670504289, -0.026473947915618668836, -0.031039565671369728772, 0.19788884029510708795, 0.054989946334801934058, -0.2567504859132240802, -0.44830494024809575127, 0.48886632258108758675, -0.079937935533977655389, 0.49161000880976024607, -0.4375836203832755289, -0.23799777922106229999, 0.062169144509765891793, 0.1317103508293166747, 0.10911504843607136572, 0.016729445904791050148, -0.084981485428352765421, -0.036479503846127803168, -0.25909490544870661388, 0.03274926400199398846, 1.0513025808559102181, 0.66601943819772579936, -2.9360744837282393149, 0.67359321097815816337, 1.0734880609246810756, 0.072367061465199833403, -0.22444076282202779504, -0.13580967594382795105, -0.022260701562472767073, 0.021090222598056908299, 0.023753452966884650022, -0.50450534931207902112, 0.079311866639238476351, 1.0506755323072045627, 2.1170517126780161732, -0.001965357964634373289, -5.507425138060996872, -0.037357396500273365447, 2.07579627330873695, 1.0664717300270474265, 0.25555642119005261836, -0.098783940568248351788, -0.21057133906986658101, -0.18872064798276486086, -0.085581622715458327555, 0.27909182608332033793, 0.077225308821640248347, -0.22956734630251646179, -0.67757069333919062171, 0.30622655747011506833, 0.45278944915034502472, 0.32646406054242849137, -0.63648892368979459455, -0.1810684842221355062, 0.086560615162771137321, 0.14058513475876144305, 0.15113467785996481663, 0.044486065657932782891, -0.12678443567835367967, -0.08519038145060713052, -0.20857945945373329577, 0.044395550436101618996, 0.94759958845637615177, 0.6208470544697108906, -2.6397863829387095969, 0.61071474354819910157, 0.94197446044945642996, 0.059681967422045831273, -0.17311668423092524405, -0.11038449324639529325, -0.038103752781399408134, -0.00029031393592449106222, 0.026831840450145388877, -0.23749198281162225976, 0.29180282962488485188, 0.3485482107915680583, -0.52989827728713911892, -0.067547505256554812858, 0.39647041852733450051, -0.068230041057365561907, -0.54389813941797504615, 0.3103486275042030007, 0.27949107696879799123, -0.053273072039244591391, -0.12100775722482898233, -0.052449520119562643983, 0.045079754305436141315, 0.010193801518671113601, -0.17947152741591393399, -0.046544444696801481887, 0.75287318335092556776, 0.62848770114710728052, -2.3445302776620651208, 0.62449317622745426082, 0.76021919513793057011, -0.016378136814659716158, -0.14407361924030887002, -0.049948436237421213468, 0.012614750812074169289, 0.011451158782112270257, -0.018594355282629568965, -0.37390142748015858798, 0.060324803935824469758, 0.77246067338059565266, 1.7598652377536501135, 0.17745899815186130799, -4.812612001845066878, 0.14858977896737896041, 1.7258615504415963926, 0.78554471712427431651, 0.20548427964269005974, -0.045666220226747281097, -0.15556347512136953215, -0.1601548260178702332, -0.078057445875953854197, -0.027961330324927830687, -0.16033867199046339391, -0.12816480591260909216, 0.75054776548815693715, 0.76937821413126916692, -2.4290988149391909978, 0.7789329430934781584, 0.7745284544734287735, -0.10106381805884472047, -0.16426161430444757849, -0.094704784827078197784, -0.027285009796411886784, 0.024973228522027310361, 0.032745059285754760103, 0.2450599201571523611, -0.38198102508953124534, -0.48761087300848005022, 0.28561419331010179379, 0.88265144685636176547, -1.099184591690564039, 0.89510678384089192861, 0.31962389698669863014, -0.43947551534150952612, -0.39901027717491061031, -0.084086506658560600247, 0.11608467795985420723, 0.12067377149644262224, 0.022961354705644150542, -0.40436524234448251169, 0.35555853431834788525, 0.66307698948955084983, -0.49627527032274798557, -0.34584827215479208018, 0.46115326635677045264, -0.34732121897611312189, -0.51418864842519973468, 0.61831501765970964257, 0.35024757799756217613, -0.15834297843158129382, -0.20271414985107605311, -0.064141512866787653291, 0.080643839325596683443, 0.27095849573682900813, 0.58155080951643878251, 0.11903394197050240488, -2.3364542882489516273, -11.373040285828238893, 25.524715919571583811, -11.332835915319327214, -2.3334238654295802462, -0.002821976685992655387, 0.39235677181394956747, 0.28263589489220480866, 0.11468129057294439155, 0.04546289581675991287, 0.045066005505957712485, 0.16257914854014660078, -0.22193377490453825018, -0.42492490712565844646, -0.022016709133635958828, 0.79652223022364521121, -0.54982346678350912939, 0.78143665171236864619, -0.038893827376402234641, -0.42608083693669052128, -0.22779012349312377506, 0.0032727315872004836272, 0.089188946948108804857, 0.056211696038361416705, 0.015637855883500608595, -0.059882841954468503387, 0.32484726751487136287, 0.10519252592131567159, -0.72429760213447602624, 0.079676956901693574142, 0.560639657965991689, 0.072866036986385718244, -0.7449390373527884357, 0.0700027063681674111, 0.32686358862833697625, 0.12574677896815009137, -0.021828078695896046407, -0.069966280916328699169, -0.040929475072811527525, -0.00021063325958052287606, 0.17695514973533135739, 0.0018664407211433139272, -0.44348116361187178702, 0.22132422990257519246, 0.091005066968869771649, 0.21760491912561211181, -0.45065938498961338743, -0.0084407653704962719299, 0.17933776609646417954, 0.074098162035590317553, 0.0064155445232957119053, -0.025683748565627214505, -0.036858995523616940038, 0.28813942168884471018, -0.13789906200808421666, -0.5400792144290187391, -0.63549469908567635112, 0.11476496036246418286, 1.1498720951014544944, -0.47807935704528486953, 1.1519466409635925608, 0.11765711605596232414, -0.63711296413650753134, -0.56086304925677699096, -0.21414884657222768571, 0.098153086548231224806, 0.26393145200132189787, -0.04811086824385400057, 0.032898962638128886882, 0.14706724568435144973, -0.074561528125964462999, -0.35344657751506963761, 0.4033862659292316577, -0.21520114951344726517, 0.40465687111714404578, -0.35257599929979649422, -0.076963663629238232566, 0.14135398820764871752, 0.029892483923014755848, -0.033273997850329563308, -0.0063142984070363589177, -0.0068745464670234351576, 0.014091159398871384509, 0.025304204846334488882, -0.081731909419750725654, -0.065789767418959768652, 0.39160016535576841479, -0.55235297092415447384, 0.39014663916811653799, -0.067210157610663345817, -0.080136367707517627834, 0.029229195198280478324, 0.0098139162631380658097, -0.012912997456566930807, 0.0064889151791250720275, 0.24395630797495923558, 0.037588835242929032054, -0.2462307118312799803, -0.55375379162888083151, -0.35697878675783101254, 0.81616652915713705507, 0.11585303391868105594, 0.82113595429014407667, -0.34555059451048830699, -0.53764439736156965921, -0.23815408943512531015, 0.0049205465351699687049, 0.1268928723214019505, 0.11791036852024858184, 0.034050173131558017503, -0.014025640765254308515, -0.040342519443223728848, -0.1403908872743031, -0.023704786157239039757, 0.52868934203923645399, -0.69143394243406786082, 0.53081458833490891003, -0.019726085208053099895, -0.13738927523858540658, -0.043244562099277114586, -0.028754431047300341162, 0.0040061480959769522545, 0.039741556917804760074, -0.044519292476608358922, 0.14526559157724111548, 0.1055302732429150292, -0.32905655449359921816, -0.32503279952424751764, 0.67233684696294437888, -0.4434137761766223651, 0.67451765263552077023, -0.32657676788178485738, -0.34207270985918519779, 0.083912895502058174046, 0.14492839670779314454, 0.046847430113889948489, -0.062191446133502192861, -0.038516484047434795812, 0.14010845969276408685, 0.0558716559869492993, -0.25421039766537784077, -0.19473195683629615504, 0.54132925052542435473, -0.49560601414023330857, 0.54262307286001099094, -0.19508359626011223531, -0.26151954964342510879, 0.039672234295040714747, 0.13527586186262827161, 0.053181897793754241699, -0.066545648299362186728, 0.043528347655220601953, 0.11270258345842296133, 0.064458427138354257302, -0.23309954124442353041, -0.43605967418706403915, 0.49161000862836107261, -0.079530718612729117689, 0.49143049642808184219, -0.43758361990556010612, -0.23613919046507342303, 0.066979531565393851666, 0.13472416464300615568, 0.089966810455896217835, -0.063224568574440545565, 0.070183951735145105877, -0.13590171809676196424, -0.25834272707265132363, 0.046156741198413661631, 1.0631082781815603955, 0.6735932112624779533, -2.9310290812860886334, 0.67672157917383002346, 1.0734880609524888317, 0.066130764521936447053, -0.23818027159646398827, -0.14980872020896007779, -0.018718102634680912344, 0.063896305111929133091, -0.4058811603312928451, -0.19658054815540773608, 0.22941257095266648358, 1.081915219866403266, 2.0960845939196826038, -0.037357396281392563075, -5.5365848180573538073, -0.0517670203920669314, 2.0757962728114680573, 1.0734996600805537526, 0.25834353491879802345, -0.1107333721091045553, -0.24253093157604452279, -0.23589505027869567266, 0.12004673552685243831, 0.095617280488177386344, 0.0606815395169968963, -0.19115171721930912518, -0.63965884955795515587, 0.32646406111088177582, 0.4585685373642578444, 0.32573737175304962044, -0.636488924203554296, -0.17833533745196095133, 0.08705599949456310771, 0.12788261144095908195, 0.10909718405980753375, -0.047403259597957159155, -0.0051533782700409914296, -0.10561365258745329665, -0.17986228920433888767, 0.057673232898348546216, 0.94379842871190466891, 0.61071474338170550489, -2.646198852183964334, 0.60918215079949999957, 0.94197446028016995623, 0.058526763203567019889, -0.17453894034037314986, -0.10777036925026650671, -0.02762786192389472123, 0.019954822114414336176, -0.12287232427761159148, -0.059377732073839853411, 0.28131442793956212833, 0.31243321352345243547, -0.54380477094392876136, -0.068230041120993151216, 0.39775153192599982388, -0.068264110539292210955, -0.54389813974781486827, 0.31026652713789071303, 0.27640170919360396162, -0.059791249234744547891, -0.11709903344733191277, 6.3314476651889477904e-05, 0.026606247033196639418, -0.061613752438285963875, -0.15838399383650675345, -0.019060966347055247755, 0.76337696613740768647, 0.6244931762754543092, -2.3513113629908768409, 0.62092542159303643068, 0.76021919441047436372, -0.014766920380881355321, -0.14288498744499783522, -0.051072524668437868645, 0.0060969481666922472826, -0.0019300948151532587322, -0.33462899393064315934, -0.12054949277196116975, 0.18732633432911943716, 0.79545238655330985367, 1.7387231446334550089, 0.14858977886122856638, -4.8326661407652391489, 0.14024422447902995525, 1.7258615505489882658, 0.78806961168181410837, 0.20473789330013891785, -0.054470865355177790212, -0.17900424784694912717, -0.20779811272108383857, 0.066455290804919811554, -0.092991663068320282171, -0.19319357657157024133, -0.11886976274999588865, 0.77086507503957069076, 0.77893294365029286386, -2.4293756311987042551, 0.7765511004966019204, 0.77452845430279626537, -0.10083974374657168893, -0.1698076486233160709, -0.10748077158977123458, -0.031734790223572173495, 0.076375609195478671998, 0.26564303393480487303, -0.053179665548788550289, -0.414788370027153519, -0.45899358314055394015, 0.3093549383860882962, 0.89510678420827849422, -1.0910978932909880523, 0.89997812658042286582, 0.31962389711199534759, -0.44440640787781299226, -0.40464502305192956433, -0.082406922550445932396, 0.12876448499988163432, 0.13380622257883292847, -0.16052513805275703263, -0.1729075067885431638, 0.34278892607152466887, 0.61760631100671159466, -0.51389469992848169078, -0.34732121919713787594, 0.46146462651147679424, -0.34823658118062517097, -0.51418864855851809104, 0.61798070369949997804, 0.34404488935768245961, -0.17056060253655097592, -0.19512217253419539276, 0.030188118233671179524, 0.10761150204132111097, 0.34623551441848421373, 0.44059433781799223206, 0.0088273253095764725451, -2.3440858873764058856, -11.332835915529928528, 25.558591006452314076, -11.320632430080941333, -2.333423865684389753, -0.0023778389174637926118, 0.39694207767894928995, 0.28603774583136742926, 0.11710147872463755325, 0.076536735873162009547, 0.09635399138711758793, 0.042797862164333760893, -0.19402990795240088184, -0.41030110768192323834, -0.035565010392968054997, 0.78143665166879316963, -0.55331153906395702524, 0.78359486445847592595, -0.038893826968803502098, -0.42723147359313284888, -0.21968112330352618233, 0.029713149561360452378, 0.11717216925005884132, 0.022747620111876493826, -0.15168251281405414899, 0.1084752769120298499, 0.34676675704439891001, 0.086633530467298325295, -0.73913234882558254135, 0.072866037632414365488, 0.55805079470952745346, 0.071317340918716429221, -0.74493903759475232906, 0.073062992013933708568, 0.33248337910587993926, 0.1290922835707575822, -0.032148415747462973091, -0.10989692387724565859, -0.060693937824657352709, 0.05826797093081279022, 0.18787017271743225444, 0.0014789088560986062004, -0.44661390494486641334, 0.21760491858063932269, 0.08874038057184999162, 0.21635910734482888262, -0.45065938519626930203, -0.0062505089793473481152, 0.18370326283229396203, 0.076778579062607404948, -0.0034024148437103089553, -0.060913233288348155037, 0.26393145212118424014, 0.098153086044312776171, -0.21414884575158116276, -0.56086305000577185709, -0.63711296394783079045, 0.11765711578869142429, 1.1519466421480273333, -0.47807935774898341341, 1.1498720948702121358, 0.11476496044583628375, -0.63549469900220378893, -0.54007921467918196257, -0.13789906182782715738, 0.28813942173484985432, -0.0063142984015245725779, -0.033273997856242264193, 0.029892483942725495832, 0.14135398823257641632, -0.076963663867582188249, -0.35257599889985224983, 0.40465687067937006205, -0.21520114887258154357, 0.40338626506896763102, -0.35344657671542628607, -0.074561528731481557797, 0.14706724607139345928, 0.032898962486905797276, -0.048110868222425315355, 0.006488915202472747383, -0.012912997555292606011, 0.0098139164030029892721, 0.029229195149531188996, -0.080136367814133482868, -0.06721015730629154028, 0.39014663851896097535, -0.55235296988812454533, 0.39160016412369252903, -0.065789766201213462082, -0.081731910366232202181, 0.025304205279047135846, 0.014091159340296168651, -0.0068745464807078732888, 0.11791036900628910911, 0.12689287142810354303, 0.0049205460647957985354, -0.23815408774573845752, -0.5376443977452143308, -0.34555059467786219951, 0.82113595285983598249, 0.11585303489886779327, 0.81616653006649619861, -0.35697878675475425148, -0.55375379295218940534, -0.24623071200630458621, 0.0375888366709452304, 0.24395630732134948593, 0.03974155697948016791, 0.0040061480904271390413, -0.028754431442836656668, -0.043244561433898140446, -0.13738927576711523759, -0.019726084878721794613, 0.53081458814387127987, -0.69143394234346999827, 0.52868934218717278384, -0.023704786511360896595, -0.14039088708899416513, -0.04034251958057954307, -0.014025640486535369961, 0.0340501729839370465, -0.062191445971069174437, 0.046847429666942018434, 0.14492839719385752772, 0.083912895093483361175, -0.34207270940858308661, -0.32657676823464243387, 0.67451765270172225897, -0.44341377605318460509, 0.67233684677719840472, -0.32503279921781941253, -0.32905655485931539772, 0.10553027347129649571, 0.14526559148129666865, -0.044519292443119216907, -0.066545648219574246585, 0.053181897525029067642, 0.13527586223184615011, 0.039672233956599113613, -0.26151954934341309356, -0.19508359634592029042, 0.54262307258250352948, -0.49560601368624102037, 0.54132924996073783497, -0.19473195606936383184, -0.2542103985016490042, 0.055871656660820924467, 0.14010845933428286991, -0.038516483961229246169, -0.063224568578133355512, 0.089966810262159579992, 0.13472416492211999883, 0.06697953159129504952, -0.23613919031070032539, -0.43758362021763108629, 0.49143049612303296714, -0.079530718219726168061, 0.49161000861021120212, -0.43605967415327467895, -0.23309954127464607709, 0.064458427121058370357, 0.11270258347803885357, 0.043528347627444598233, 0.063896305124374205842, -0.018718102658654880988, -0.14980872010036447306, -0.23818027187352303065, 0.066130764900188265809, 1.0734880605619065985, 0.67672157956287393166, -2.931029081477775744, 0.6735932111423968971, 1.063108278038141119, 0.046156741650880253602, -0.25834272703819410877, -0.13590171845631257574, 0.070183951866301400035, -0.23589505051242898315, -0.24253093098560832286, -0.1107333721725335951, 0.25834353436989054842, 1.0734996601049675569, 2.0757962731829935343, -0.051767020304333777325, -5.5365848185875927712, -0.037357396089705452447, 2.0960845938558447799, 1.0819152204517501481, 0.22941257065445175289, -0.19658054832721089111, -0.4058811602734561097, -0.047403259727408476876, 0.10909718420784306203, 0.12788261178632626125, 0.087055998870865680672, -0.17833533715086546945, -0.63648892431850501161, 0.32573737209180375629, 0.45856853680376741256, 0.32646406140331651891, -0.63965884968872144256, -0.19115171677539066764, 0.060681539182882937578, 0.095617280468124260517, 0.12004673557113584859, 0.019954822093738000532, -0.027627861758858486552, -0.10777036958385044496, -0.17453893998980468671, 0.058526762769068409187, 0.94197446064832957013, 0.60918215080309334741, -2.6461988522125170498, 0.61071474344653231547, 0.94379842817084724782, 0.057673233399632567453, -0.17986228905154238999, -0.10561365301880058787, -0.0051533780851337127049, 6.3314358636334575579e-05, -0.11709903303404309294, -0.05979124979174774146, 0.2764017096671542717, 0.3102665266157192403, -0.54389813939071374183, -0.068264110407690092863, 0.39775153192927603651, -0.068230041582393152755, -0.54380477074756849998, 0.31243321375762450609, 0.28131442786078125762, -0.059377732258930025766, -0.12287232416381443922, -0.0019300950344489954669, 0.0060969486485393563704, -0.051072524770287980322, -0.14288498768010859941, -0.014766920834748077304, 0.76021919537920101817, 0.62092542128794292466, -2.3513113631805033776, 0.62449317617830768512, 0.76337696602435345383, -0.019060965880014081597, -0.15838399372791411857, -0.061613752972138904895, 0.026606247255467566776, -0.20779811250623669849, -0.17900424804135905976, -0.054470865588581190986, 0.20473789365432321086, 0.78806961145806631031, 1.7258615505292294046, 0.14024422470070824209, -4.8326661410118800788, 0.14858977936057546354, 1.7387231438460637456, 0.79545238717755217817, 0.18732633405238363622, -0.12054949260524913668, -0.33462899402648799141, 0.076375609107155462141, -0.031734790045599883812, -0.10748077158272190124, -0.1698076489221428953, -0.10083974341952978149, 0.77452845419159088713, 0.77655110036712726718, -2.4293756308021228207, 0.77893294287285275157, 0.77086507601768194942, -0.11886976359209579235, -0.19319357596616182349, -0.092991663389367137937, 0.066455290880459455538, 0.13380622263714012088, 0.12876448498158749034, -0.082406922902236473538, -0.40464502278764796772, -0.44440640717050927666, 0.31962389582095518392, 0.89997812742401339658, -1.0910978935771673548, 0.89510678450225911185, 0.30935493815100589821, -0.45899358325256472968, -0.41478837007643742929, -0.053179665217605170957, 0.26564303377964926245, 0.030188118554658843057, -0.19512217323668729074, -0.17056060194685948583, 0.34404488878501293669, 0.61798070435740926065, -0.51418864884405679483, -0.34823658110884037065, 0.46146462598105508768, -0.34732121819711547772, -0.51389470053142694006, 0.61760631066535554012, 0.34278892674012217956, -0.17290750708833479976, -0.16052513802526302555, 0.076536735947660389234, 0.11710147826548680805, 0.28603774646318980146, 0.39694207733769737434, -0.0023778386643709681358, -2.3334238658505110919, -11.320632429996187795, 25.558591005859497614, -11.332835914443618819, -2.3440858879939088233, 0.008827325356571658066, 0.44059433751405729929, 0.34623551480596770347, 0.10761150195669208496, 0.022747619892516425522, 0.11717216977389545429, 0.029713148899187385998, -0.21968112271789072554, -0.42723147355007784487, -0.038893827585861462381, 0.78359486473487960367, -0.55331153896149098959, 0.78143665191016453697, -0.035565010890043213809, -0.41030110723358259595, -0.19402990851152318585, 0.042797862623396631943, 0.096353991259154836424, -0.10989692429228289927, -0.032148414871247366076, 0.12909228340472550522, 0.3324833785641585493, 0.073062991954768369274, -0.7449390373674761312, 0.071317341428831088068, 0.55805079445276450922, 0.072866036952024176854, -0.73913234844373643107, 0.086633530939982078545, 0.34676675681704821619, 0.10847527660559483365, -0.15168251262987195171, -0.060913233520426032785, -0.0034024142175172371674, 0.076778578340639710986, 0.18370326365004732705, -0.0062505101081720918638, -0.4506593840295072817, 0.21635910659210297324, 0.088740380678841560003, 0.21760491886720492727, -0.44661390537665612843, 0.0014789094693313242779, 0.18787017233668742988, 0.058267970846933928031, -0.060693937709957297699, 0.17834288774015844004, 0.15490436297177748171, 0.034757181825177660417, -0.23913894400761806502, -0.56457110852605807416, -0.6352330346684078588, 0.11765711554358689039, 1.149872094901534858, -0.47870760314293370197, 1.1529625011485291619, 0.11784851673343993417, -0.62906348579102555618, -0.48761142007202390447, 0.10422528522286744757, 0.012942986415330894751, -0.021357267025050346437, -0.03311886162668953304, 0.03060858681943988413, 0.14200820227530305817, -0.076291640869918381229, -0.35257599830444502631, 0.40338626495754820178, -0.21722521459152405177, 0.40286198282719909436, -0.34935250953644136906, -0.057944947882718859933, 0.16507460693273481867, -0.04938947388947535766, 0.0066991721663262390057, -0.0049030686686718409975, -0.0081834354107039371934, 0.011676793992111784634, 0.027646236416563364779, -0.08164494675145614333, -0.067210156810062593458, 0.39160016453637236378, -0.55091351946801503381, 0.38963137181061469061, -0.071156874679114759386, -0.076676646001525405882, 0.040570641924337139117, -0.0074399717149174187519, 0.011439161400801783897, 0.10568104300275187046, 0.12477002240953691259, 0.0082908892182484705824, -0.23330859720562596715, -0.5343899291006469765, -0.34555059468647597587, 0.81616652962635216628, 0.1035605263276713206, 0.79465889820262602949, -0.38513342782388931829, -0.5645199270822965909, -0.1821075138844845398, 0.27956228226190676933, 0.028343677292805673218, 0.015224026649192194671, -0.0045741203658291908207, -0.028564610309364302232, -0.040056155018204385454, -0.13522221131763925706, -0.019726084679385570003, 0.52868934186015981158, -0.69495068458043651738, 0.52597049255032313209, -0.023664466493178874795, -0.13360519312376073486, -0.02496799589782108808, 0.0046452381447856239127, -0.02424511268207409781, -0.021254570978002242099, 0.042972048372523809212, 0.13275705040635038534, 0.079309406400188631192, -0.34146272077589367999, -0.32657676776322963619, 0.67233684653977454637, -0.44358592201273555622, 0.67992341302561476457, -0.31270591548373527502, -0.32274322634083574268, 0.094461818511766854067, 0.091888959528852107161, -0.031039565601739201733, -0.026473948040577670315, 0.053086874794470960415, 0.12837588885375650349, 0.036783136508907861295, -0.26143770540186994644, -0.19508359618319032203, 0.54132925016261290718, -0.49645471629704751582, 0.54464472547732445395, -0.18621692594372790674, -0.24674863369847963024, 0.037852553181379887393, 0.10366121319394908107, -0.084981485315314297946, 0.01672944582597681723, 0.10911504819441386616, 0.13171035090498228737, 0.062169144506041308651, -0.23799777896761950546, -0.43758362024631569653, 0.49161000859771403215, -0.079937935694027698053, 0.48886632269554752828, -0.4483049403235698227, -0.25675048574467468354, 0.054989946167398988386, 0.19788884036979589864, 0.02375345307112401369, 0.021090222587553698885, -0.022260701998875448426, -0.1358096756112627812, -0.22444076289937989199, 0.072367061798924728344, 1.0734880606601420183, 0.67359321099170565983, -2.9360744836260019852, 0.66601943814466357807, 1.0513025805861733186, 0.032749264195360283658, -0.25909490533313722649, -0.036479503931371566605, -0.085581622622345282503, -0.18872064814208500128, -0.210571339147464065, -0.098783940241214174272, 0.25555642088266383727, 1.0664717301610509015, 2.0757962730143315611, -0.037357395826535855221, -5.5074251386252539575, -0.0019653577187280779626, 2.1170517124971706124, 1.0506755324917502747, 0.079311866623915927565, -0.50450534937891988729, -0.12678443552270826644, 0.044486065468216345886, 0.1511346773339224403, 0.14058513578198539307, 0.086560614313270445308, -0.18106848335288383556, -0.63648892465913009531, 0.32646406155861495968, 0.45278944817081379082, 0.30622655827602052181, -0.67757069400641389834, -0.22956734577486195836, 0.077225308494859090658, 0.27909182618947026011, 0.02683184053137700753, -0.00029031403862168729535, -0.038103752706261415084, -0.11038449346972832588, -0.17311668397047588752, 0.05968196740638771064, 0.94197446032452558651, 0.61071474353658972145, -2.6397863827941270287, 0.62084705442636578532, 0.94759958830883184167, 0.044395550487226209668, -0.2085794592815528592, -0.085190381564479514265, 0.045079754351841028193, -0.052449520183452509603, -0.12100775714233037716, -0.053273072183753572562, 0.27949107704606646063, 0.31034862750331232428, -0.54389813929513164403, -0.068230041323330076208, 0.39647041878677724558, -0.067547505442642888362, -0.52989827720730875438, 0.34854821074730407693, 0.29180282972588644785, -0.23749198287381592665, -0.018594355220341512308, 0.011451158556714302605, 0.012614751083852061148, -0.049948436311415225775, -0.14407361916884617314, -0.016378137093672081454, 0.76021919543825322574, 0.62449317578250362892, -2.3445302770223088729, 0.6284877006419398171, 0.75287318373375600178, -0.046544445106189637207, -0.17947152709344491028, 0.010193801405029210422, -0.078057445733210259475, -0.16015482612320636191, -0.15556347524266894711, -0.045666220448215999461, 0.20548427982698516669, 0.78554471740905951194, 1.725861550324395477, 0.14858977907179099498, -4.8126120021087057665, 0.17745899842173917804, 1.7598652373262566595, 0.7724606737784568411, 0.060324803804254437112, -0.37390142748046073518, 0.03274505928711414493, 0.024973228366117601607, -0.027285009386190706843, -0.094704785288492715489, -0.16426161388502094618, -0.10106381836694476595, 0.77452845450451479614, 0.77893294315563443853, -2.4290988147885870241, 0.76937821375675530255, 0.75054776585813576251, -0.12816480612628347591, -0.16033867187739692106, -0.027961330370336018314, 0.022961354631487206068, 0.12067377172949850239, 0.11608467756239884694, -0.084086506104429359354, -0.399010278034355792, -0.4394755142064342146, 0.31962389591900486341, 0.89510678498562157301, -1.0991845930948347387, 0.88265144807828299633, 0.28561419267820797874, -0.48761087287300497461, -0.38198102513201542818, 0.24505992021184344609, 0.080643839382662563242, -0.064141513052361431857, -0.20271414956073052771, -0.15834297854304835163, 0.35024757763537961441, 0.61831501836688196327, -0.51418864925427365531, -0.34732121814951177896, 0.46115326572040232778, -0.34584827170592513657, -0.49627527074299265797, 0.66307698985631913224, 0.35555853411631327399, -0.40436524229315817802, 0.045066005507557932441, 0.045462895763737493082, 0.11468129053514289351, 0.28263589538701650028, 0.39235677098682059105, -0.0028219758103413816741, -2.3334238662097446237, -11.33283591476230967, 25.524715919045462442, -11.373040285397575389, -2.3364542883570051934, 0.1190339420383490221, 0.58155080928131486218, 0.27095849588189918622, 0.015637855792897677187, 0.056211696183925158321, 0.089188946907903410288, 0.0032727318147876549581, -0.22779012381829050193, -0.42608083691038634022, -0.038893827431977223696, 0.78143665204361512089, -0.54982346719691321812, 0.79652223033289171195, -0.022016708781271543405, -0.42492490749940575734, -0.2219337747974656494, 0.16257914854220828493, -0.040929474954444475254, -0.069966281155456236385, -0.021828078658364339498, 0.12574677919853233243, 0.32686358862383591006, 0.070002706126621516702, -0.74493903743022304997, 0.072866037243047770966, 0.56063965783052782665, 0.079676957063715886376, -0.72429760231043582852, 0.10519252600870666881, 0.32484726755127990572, -0.059882842009472220246, -0.036858995465891372656, -0.025683748716841744986, 0.0064155446661458602822, 0.074098161864684083944, 0.17933776648690208377, -0.0084407659357139397382, -0.45065938439757291523, 0.21760491842122059447, 0.091005067905981354426, 0.22132422886484154345, -0.44348116277616422831, 0.0018664401866517440001, 0.17695515006827985949, -0.00021063338618037760752, 0.093447167032542743925, 0.12653322095876293196, 0.11839540561885006364, 0.017770048264768002755, -0.2430336313659464953, -0.56457110887011741074, -0.63711296468865008524, 0.11476496164574889747, 1.1529625013147313251, -0.46510669385699687783, 1.1715483568610658427, 0.12239377478504502583, -0.61684605214751075941, -0.41722447654272898365, 0.011953693820174693396, -0.00079955791261591534907, -0.023037017980382271803, -0.0310629952456411354, 0.031945738312740792897, 0.14200820229741234479, -0.076963664239381140986, -0.35344657701728188215, 0.40286198344733770149, -0.216275759562971015, 0.41042188989283279188, -0.33749737578221022938, -0.053361710929962558048, 0.09394856510058741117, 0.0001412559435470008549, 0.0070459369896613173456, -0.0037892755077746469172, -0.0095359858834443370773, 0.01024989356829342918, 0.027646236134127179229, -0.080136367684336920481, -0.065789767743264615518, 0.38963137308226714062, -0.55573392415572708813, 0.39376531310172396827, -0.062156043874154444373, -0.080430820767587965503, 0.028850307626431127839, -0.041605144446756905929, 0.049745670509719022712, 0.11940963394043047074, 0.12750344668562155204, 0.00856094839946602848, -0.23330859637186890376, -0.53764439589317969403, -0.35697878866861343372, 0.79465889759468455456, 0.076019541666549325654, 0.77330488935915320248, -0.38545319790673299165, -0.48091503263698254234, 0.075382857875104583578, 0.011355775119963145564, 0.017400341131056026867, 0.012915756246144416072, -0.0030420356010505793243, -0.027038202075205124142, -0.040056154686583668456, -0.13738927538875525536, -0.023704786440312354545, 0.52597049216264368798, -0.69269138274659103338, 0.53445639495991070778, -0.018504049911601393175, -0.12829910170826011595, -0.034427215621788874456, 0.0079594745480899339179, -0.030201865525731209916, -0.025178304144396982528, 0.042440628538520541735, 0.13346557152545135416, 0.079309406929715761669, -0.34207271000708672215, -0.32503279918892968858, 0.67992341281467782999, -0.43337441727439834915, 0.67531247066913624444, -0.3580265572481874381, -0.36544442683635192104, 0.26141375718298831554, -0.0011771800626573154271, -0.032916249048319592974, -0.024292740309953934214, 0.054229357711869943803, 0.12858871515705772026, 0.036783136410384671688, -0.26151954936430393817, -0.19473195669179160783, 0.54464472591450407979, -0.4888386380538254361, 0.54531463741010499646, -0.23191070432303681481, -0.30121414574949428866, 0.22837286391146102704, -0.05706514549534848213, -0.033103823537905335828, 0.030702199386644261209, 0.10923267677360012917, 0.12976230509254302325, 0.06216914424490335711, -0.2361391904332021674, -0.43605967419487273729, 0.48886632258600698497, -0.095825903621135355426, 0.4505961085723712567, -0.49656594515622831665, -0.20896633522610091882, 0.29445230471679573192, -0.01078507802055143211, 0.023569307561225112668, 0.0276051595934911978, -0.01429756967650330779, -0.1305265758453047209, -0.22444076284192618354, 0.066130764448886894535, 1.0631082784798748797, 0.6660194378153297956, -2.9321922274910106054, 0.68743359083870025916, 1.0697785849075749365, 0.013643354784956105222, -0.3051575781166073198, 0.0099875274069152232548, -0.10307088155127244744, -0.19452400339915298932, -0.21255290263753673363, -0.10047863040667859291, 0.25555642099988107319, 1.0734996595516628215, 2.0960845946339032864, -0.0019653580066795184322, -5.4654378922145614439, 0.032939608230213135442, 2.1037636207513119579, 0.88641205216839080716, -0.35194374858167559728, -0.11742467091131944534, -0.027831285669179611775, 0.080231255620864749822, 0.15660986290776762742, 0.13780582155134635247, 0.086560614342522712583, -0.17833533753908048514, -0.63965884936150141371, 0.3062265579697766027, 0.40829002871053515777, 0.24585167300360577514, -0.70689856305997056385, -0.096597610720615706414, 0.348119043262740957, 0.024064016112117048857, 0.0098655044284370625035, -0.010974282734829393648, -0.041395126389071848494, -0.10984147331393144531, -0.17311668465899326952, 0.058526762864289726129, 0.94379842867875274326, 0.62084705468307177512, -2.6221070551672651661, 0.63931288675897379736, 0.94103435155045456195, -0.01438340346527020204, -0.26539170801312256254, 0.038203635863977458698, 0.0040636156611811788397, -0.059139279349824804, -0.1176843518545489381, -0.051272037761592975658, 0.27949107679751994393, 0.31026652663471071536, -0.54380477069940713619, -0.067547505185130529703, 0.40352837596436785805, -0.041846657538665621434, -0.48590722555757248635, 0.37515551493575410857, -0.040195401030096777895, -0.015738549936891468167, 5.7991286331634259513e-05 };
	std::vector<double> l6 = { 0.012694918477841744819 , 0.0090717662234274971944, -0.052116643072406813875, -0.14407361944610999616, -0.014766921329288064391, 0.76337696673084576293, 0.62848770084704097538, -2.3435766779419218153, 0.625120680587700428, 0.73786508931365812902, -0.070179023004716220568, -0.13486643055281588577, 0.0059560042842739880156, -0.095084412332657589073, -0.16335640540968732193, -0.15323588834179807661, -0.043876978450425166423, 0.2054842797211169636, 0.78806961185186685803, 1.7387231441855937053, 0.17745899817755772521, -4.7735790017323944667, 0.20854971015395451506, 1.7361160230874697064, 0.62036677845153764821, -0.22790387722581570418, -0.0087096140633702492234, 0.038540528373555267561, 0.030758003395972691174, -0.026466356016554636044, -0.095855515453084738642, -0.16426161351507945207, -0.10083974425553222931, 0.77086507589864561307, 0.76937821403457373748, -2.4372776026903752999, 0.78423759291136341254, 0.7861147919504140047, -0.13842670761797912227, -0.21025195102322991048, -0.028018029255144120959, 0.055076192220925468401, 0.1253287205983585062, 0.11586584041098826892, -0.082889451490079035789, -0.39901027736294453208, -0.44440640664347397237, 0.30935493737661207136, 0.88265144745461610043, -1.1134439218237652813, 0.86515113563771683935, 0.27154700728130881648, -0.43071030049757952352, -0.13848707375138383235, 0.05856233061818502672, 0.020076605521542695287, -0.076255685551285951052, -0.19882962010246343909, -0.15537505015836355904, 0.35024757783016052803, 0.61798070501239399377, -0.51389470054294905665, -0.34584827235847387383, 0.47159021171502829572, -0.3062472859743768927, -0.4155172918530797288, 0.71667925305350321796, -0.21859162240405496935, -0.0026150396674913846395, 0.039895466379319066696, 0.060056297875390518559, 0.12262163043110993477, 0.28296517008728339082, 0.39235677107265076735, -0.0023778382054892022168, -2.344085887856802497, -11.373040285928365023, 25.454887925075027511, -11.405885491402116827, -2.2341438269326201471, 0.31503040575884688757, 0.67876938380077622703, 0.030193283635553090638, 0.0074221881819495627575, 0.036763647332207632923, 0.079148412866806183708, 0.0019510504447395327077, -0.22779012351380012458, -0.42723147344501477551, -0.035565010796572149587, 0.79652223020326284875, -0.53202672498787739741, 0.78351822981400442991, -0.090849511235083307836, -0.44811577956911063936, 0.017578910699281496965, 0.0040854637270180691258, -0.048694896995548957175, -0.069417032563818883983, -0.020999027373195547352, 0.12490900589974365809, 0.3268635883467331249, 0.073062991554640049907, -0.7391323484988309156, 0.079676957754805766143, 0.56572652519369936996, 0.077200003252909529805, -0.74544809254490385086, 0.099244663269018359997, 0.27982359940362699913, -0.010670839052845123782, -0.027195587575746422992, -0.023739916530500507902, 0.0057917055825758931992, 0.072952128995718498738, 0.17933776543916454815, -0.0062505096323344413317, -0.44661390508598375337, 0.2213242296169134482, 0.089045634360726350076, 0.20912872928314180099, -0.4610141476304038255, 0.020985073114200083338, 0.18063522695715245048, 0.026303906162296639831, 0.07608329263043622781, 0.11321970551389921866, 0.11172196576338912222, 0.017770048465989590758, -0.23913894451858586088, -0.56086304920327323398, -0.63549469972298511244, 0.11784851746590531418, 1.1715483572313887262, -0.42075620852808598427, 1.2435382729462112206, 0.073922666353360214075, -1.1099952513672792787, 0.0053054662460192726636, 0.0060562388469812596797, -0.0016995597142472995515, -0.022097797881716788743, -0.031062995781596693112, 0.030608587050690333475, 0.14135398830221454491, -0.074561528188937867268, -0.34935251005529066592, 0.41042188964528569795, -0.20618107837353732981, 0.39090810281214011246, -0.46250093080562459935, 0.16339036317402105825, -0.0015082265425159972847, 0.0039760798496721941331, 0.0052292338501062317346, -0.0053970837883197120277, -0.0095359867958812304795, 0.011676795111927160478, 0.029229194545370960207, -0.081731908685174103901, -0.071156876038379265892, 0.39376531298250855295, -0.54089768877462529595, 0.39015395913498601699, -0.12035451134790572936, -0.0037105089361869143978, -0.049455117481997046269, 0.0026208028160565055575, 0.066123873178610820256, 0.12244755991524183814, 0.12750344755813780573, 0.0082908905278134617522, -0.23815408857761033445, -0.55375379216584619257, -0.3851334292231953782, 0.7733048885883204715, 0.089173740242735027861, 0.83960355293287691758, -0.348492535419146654, -0.47187371712800546542, -0.0019688055786921789328, 0.012164842732853920274, 0.019144351927103876732, 0.014013711543806275928, -0.0030420351865376640127, -0.028564610251653833883, -0.043244561833350202151, -0.14039088696117268951, -0.023664466466048139637, 0.53445639463325400609, -0.67655851854867288253, 0.54978296745877708318, -0.074679106812878107147, -0.13954972519077674109, 0.015974751794012694711, -0.011929850355425597463, -0.031974382063770886986, -0.022814688246133932314, 0.042440628431782277674, 0.13275705002118509679, 0.083912895775823004296, -0.3290565550763608349, -0.31270591500885874803, 0.67531247011789086176, -0.47189580441216455453, 0.60276039776386525215, -0.42740107434650215223, 0.052132663639072110429, 0.0080688087511724196355, -0.01305007058102971261, -0.032232015898099257434, -0.022208216348281371866, 0.054229357070777674255, 0.12837588926988272031, 0.039672233802708560535, -0.25421039785833549063, -0.18621692620350077663, 0.54531463730350571151, -0.52755787396806785861, 0.46977816091486951366, -0.24167188289777263854, 0.029681630128657533058, -0.021824851301079428029, -0.039716740892846574829, -0.025171129908418462401, 0.032483544116425198955, 0.10923267650985235555, 0.13171035121307816951, 0.066979531808038686447, -0.2330995415301287399, -0.44830494041986668208, 0.450596108895232772, -0.1464732626359707135, 0.45124251433073270245, -0.43967767789709510939, 0.10537285058919490521, -0.024066401099607278441, 0.010071318407608597159, 0.033666155475339708258, 0.031420214396995105322, -0.014297569449210227599, -0.13580967597929999302, -0.23818027168151278961, 0.04615674136370546965, 1.0513025811825911227, 0.68743359067871723234, -2.8754893036051174526, 0.77599768993650497517, 1.1046011812674279007, -0.45013200059293595467, 0.053922127725280502575, -0.033094836332245497479, -0.12424571443540327609, -0.20123196490523875823, -0.2125529025739995026, -0.098783941202013059169, 0.25834353485674521611, 1.081915219704189246, 2.1170517130989083832, 0.032939608542425946602, -5.4463893364129418728, 0.036902495547235236728, 2.1043784918494217173, 0.46992968616957964123, -0.065953706358989755421, -0.057234917362731146939, -0.0065772455076008040642, 0.082927500276270327273, 0.15660986301408652555, 0.14058513480457757172, 0.087055999426220664894, -0.19115171720080015305, -0.67757069336130815174, 0.24585167320592993256, 0.36616706063450876751, 0.30858149117826189078, -0.58754654443170639677, 0.18708728704546631616, 0.016076341919696771066, 0.011263689381193533676, 0.0034934170888856197479, -0.012584351535788157461, -0.041395126724369624149, -0.11038449322047819146, -0.17453894024299407284, 0.057673233292256174953, 0.94759958842129377032, 0.63931288710748856996, -2.5940994333362721314, 0.67597227710833429626, 0.94398416590690203964, -0.35590571372878909706, 0.01795861536692637192, 0.022255635959522436507, -0.0015964734582600424891, -0.059809476094960178916, -0.11768435218806583276, -0.053273072099915372291, 0.27640170935586400036, 0.31243321385007732927, -0.52989827723475035892, -0.041846657182861568547, 0.42793939197042329958, -0.051810081751124240079, -0.64995424328099327127, 0.45616908384275256516, -0.0037859503552179003989, -0.0084649237570198643948, -0.0014891094898670481465, 0.010234285148989698266, 0.009071766007707792942, -0.049948436586186334418, -0.14288498777850561217, -0.019060965703878489375, 0.75287318402821523655, 0.62512068071780957368, -2.3289072523575313767, 0.67373607031868565365, 0.77900693819188815414, -0.2924576793198840341, 0.041775944473967485671, -0.031750880088022924441, -0.1082134375402273585, -0.16504064823287983854, -0.15323588836372001309, -0.045666220268995333376, 0.20473789375020881587, 0.79545238661874972941, 1.7598652375466230513, 0.20854971009406522708, -4.7662593790865299681, 0.19482731664818325745, 1.7368031290285990131, 0.35902159485097856484, -0.020800757094391390117, 0.016276170328683524369, 0.041476467336311710621, 0.028160315166444700036, -0.026466356302789588451, -0.094704785421238862897, -0.16980764840914563951, -0.1188697634971383893, 0.75054776590109673062, 0.78423759276747240232, -2.3669884262902023409, 0.85771056943794898864, 0.72934231434434837471, -0.40957844786913538693, -0.037858786606930427965, 0.0059757268665252549766, 0.067878614521979824259, 0.12811274814010034651, 0.11586584116123249677, -0.084086506324053345018, -0.40464502316489198153, -0.45899358303601628384, 0.28561419280226657547, 0.86515113546551336565, -1.0959478741533232871, 0.96747055409731463893, 0.37494529273439697326, -0.74357682566861271756, 0.02480373068399366332, 0.038572356265818417975, 0.010778488613015474451, -0.078316523745724148098, -0.19882961962136974643, -0.15834297854955892149, 0.34404488873329874821, 0.61760631082214156873, -0.49627527055590658911, -0.30624728661802602669, 0.52500956018927458224, -0.27933845504400284243, -0.62495977823755510805, 0.59430108812211590585, -0.034224621994318318863, 0.021101871226237828294, 0.054843927471339898316, 0.065914281008307318288, 0.12262163105691334386, 0.28263589518578319248, 0.39694207764215322376, 0.0088273249654866603464, -2.3364542884277583745, -11.405885491602262505, 25.308265188016591196, -11.58735150676397474, -2.0117265349496080695, 1.0830001582570887297, 0.026937796729674050311, 0.0024427823885380811619, 8.5134434839734263178e-05, 0.036495065365645673605, 0.079148413444875886924, 0.0032727316534603700937, -0.21968112280124962865, -0.41030110786113932475, -0.022016708947401264496, 0.78351823009268617426, -0.58040787902691648448, 0.76837786562499343912, -0.043687468375882394789, -0.43116847148139048063, 0.019824927324058570777, -0.016311866864769774405, -0.05345987030232638576, -0.069465575301046078582, -0.020999027723348360347, 0.12574677890601423935, 0.33248337826472729173, 0.086633531282819226105, -0.72429760225388339911, 0.077200003702916714099, 0.5233815487653313836, -0.0010553040615029121024, -0.83696162536472928917, 0.56170500244144083712, 0.0053960008792412837014, -0.016432725127477610522, -0.028746827101413685579, -0.023927402162867580926, 0.0057917049759408853471, 0.074098162603212724964, 0.18370326297037103958, 0.0014789094346532097313, -0.44348116354088551461, 0.20912873010837601528, 0.061886001092145523361, 0.18572121181572309734, -0.50735829745247984501, 0.29379998324929529341, -0.020716535627935558916, 0.027389032149498532265, 0.074707531136607413513, 0.11321970479300007273, 0.11839540596689281104, 0.034757183010356330899, -0.21414884603222295034, -0.54007921501625055782, -0.62906348631469399812, 0.12239377485783947686, 1.2435382734440720842, -0.31575457289354630364, 1.1290453538864781979, -1.1206503894406221189, -0.00017291727618219351371, 0.0053180050889824427227, 0.0062674985222334845744, -0.0016995597662463344038, -0.023037017596132625813, -0.033118861908855448672, 0.029892483890485908177, 0.14706724587685238159, -0.057944947867383675422, -0.3374973752111005143, 0.39090810248201091071, -0.30969749861136036673, 0.37709333148968260385, -0.19408847189382408271, 5.287977504796544944e-05, -0.0012812798763335460622, 0.0026271776195331952158, 0.0052292335948128151196, -0.0037892745337667083777, -0.0081834365758885370401, 0.0098139167748891621701, 0.025304204566299181439, -0.076676645510307381293, -0.062156043274328498516, 0.39015395875509512047, -0.56910188374711356829, 0.41802710934797038256, -0.12992027322337518225, -0.033338001134435295281, -0.020060638596634892961, 0.012912731453456692543, 0.066123874509997501714, 0.11940963392010409128, 0.1247700216062564843, 0.0049205447637748825351, -0.24623071049311090475, -0.56451992601614520151, -0.38545319814056910879, 0.8396035519429765337, 0.21361196637117235708, 0.86088414120231548932, -1.0027524588705061248, -0.0081168378708985459979, 0.0042778991118942680943, 0.014440985590025878782, 0.019144352663048334884, 0.012915755954617059997, -0.0045741205441013916416, -0.02875443166186726629, -0.040342519376340771342, -0.13360519274016430247, -0.018504050280356679897, 0.54978296721344832587, -0.68376482885234446396, 0.54978520309811729216, -0.23062190663391948786, 0.0086635294309233001897, 0.0044802547389867332719, -0.01250128937562463606, -0.031974382411123221381, -0.025178303589332953066, 0.042972048382848307413, 0.14492839648745090009, 0.10553027400247443068, -0.32274322687250783792, -0.35802655648719988735, 0.60276039728568808762, -0.49615529099103500288, 0.83267124433846850806, -0.50002143218936700997, 0.0063310203456208987913, 0.0013310738718004208082, -0.013442973321947092671, -0.032232016486203267369, -0.024292739404093734862, 0.053086874257502074392, 0.13527586239971023385, 0.055871656103177805286, -0.24674863375268255017, -0.23191070378512695926, 0.46977816050941806347, -0.50179970398952644661, 0.69148147666105075348, -0.36784904453089745102, 0.0022517500363795584992, -0.022329925631890845672, -0.037661212219737297269, -0.025171129970245800789, 0.03070219924632358599, 0.10911504811932994097, 0.13472416485493077243, 0.064458427386531153003, -0.25675048587828042113, -0.49656594526107167287, 0.45124251438890150556, -0.06220679215466358436, 0.55891574467253990477, -0.46572681233370971654, -0.018357868131161907133, -0.0023938672873841590483, 0.017523754379056618452, 0.033666155377426204565, 0.027605159460574790753, -0.022260701970902244812, -0.14980872000420694667, -0.25834272710178463051, 0.032749263929425967423, 1.0697785849893055588, 0.77599768998199514236, -2.7592424807369866357, 0.861491448731650733, 0.39829920439127958209, 0.05957778117397309614, 0.011672146236825400867, -0.050412124977735284226, -0.12424571476309398399, -0.19452400370698236243, -0.21057133867712007902, -0.11073337182921977329, 0.22941257055502917162, 1.0506755322963234889, 2.1037636203940879298, 0.03690249599406267289, -5.4685627744240230541, 0.25179077248127079702, 2.4346731522482762955, -0.011619160012175640609, -0.045937087514225580231, -0.050228145203448426592, -0.0065772457540985151958, 0.080231255188968841896, 0.15113467801211621921, 0.12788261173976878116, 0.060681539167893538966, -0.22956734599553951792, -0.70689856329087152886, 0.30858149155030883959, 0.49538064336738085736, 0.30609769785138463138, -0.50410711470135094459, 0.0057754208029055682794, 0.0094022189669658341737, 0.0090368190652031308435, 0.0034934169971289827372, -0.010974282916755184053, -0.038103752540270506566, -0.10777036905886752671, -0.17986228965995978979, 0.044395550441291994903, 0.94103435150208225579, 0.67597227738500431649, -2.5209903640144002424, 0.79447205963023015407, 0.3866778068404924662, 0.0027155281749396760656, 0.016420634332668705008, 0.019770194974565138274, -0.0015964735143679216825, -0.059139279392428877047, -0.12100775713174460879, -0.059791249217177197539, 0.28131442750984930745, 0.34854821065707941585, -0.48590722576780098763, -0.0518100813295944429, 0.23426068995355767144, -0.35628337019344530345, 0.23719055763770233169, 0.0053749703430626370737, -0.006859147322520410886, -0.010410684915737202991, -0.0014891096815511403566, 0.012694918401550519746, 0.012614751339698918736, -0.051072524478819741889, -0.15838399448617668241, -0.046544444953896507711, 0.73786508918985782568, 0.67373607054071638078, -2.221648785712817542, 0.82786871791858551717, 0.23087180282775829343, 0.045111622646340540643, 0.010317718981145598822, -0.042379132809631146761, -0.10821343735275028297, -0.16335640571675852373, -0.15556347515536592119, -0.054470865805236592938, 0.18732633451889271758, 0.77246067361584747069, 1.7361160230977261687, 0.19482731653249960035, -4.7698271851790163822, 0.46039238055600284483, 1.9005858214578728038, -0.01269888142273200779, -0.0053030620816637511622, 0.017413907995378934829, 0.041476467080046590385, 0.030758003791038424773, -0.027285009487933477301, -0.10748077172776876465, -0.19319357604891881919, -0.12816480662759419107, 0.78611479266930428533, 0.85771056905842246731, -2.3365638401462542895, 0.98049436695227032423, 0.10316817297692351674, -0.028799839430498283543, -0.014401156767143180604, 0.015727159530380799279, 0.067878615006797274334, 0.12532872043644496807, 0.11608467716218701105, -0.082406922761778855779, -0.41478836960660681044, -0.48761087279360060265, 0.27154700679052840018, 0.96747055405817805607, -0.88145062942614638146, 1.02452181808715026, -0.68050955043055139093, 0.0022589482007787532042, 0.024361148376017188744, 0.03482267956470502801, 0.010778488533267654992, -0.076255685257717503722, -0.20271414985755054072, -0.17056060231476574685, 0.34278892642616531905, 0.66307699019101296045, -0.41551729153903993463, -0.27933845582140370833, 0.25973008250547596765, -0.81594507210840672151, 0.63349570612713779205, -0.039379273108637695422, -0.0057436544874868294963, 0.031009280810126993444, 0.054843927679270210174, 0.06005629745557020116, 0.1146812909036895034, 0.28603774566945727731, 0.4405943382119874574, 0.11903394221099186656, -2.2341438269276387985, -11.587351506917102029, 24.617655301617723751, -11.792648087215049202, -0.10093568679304792957, 0.0065861842512720918108, 0.0095596014646290930483, 0.00098792133350156241534, 8.5135027094035420347e-05, 0.03676364667847464851, 0.089188947011746927407, 0.029713148945747280827, -0.19402990772599790215, -0.42492490750175981873, -0.090849511969425103541, 0.76837786608242730768, -0.45169626583235905626, 0.89933590674199881398, -0.6748222196876573209, 0.018245519570364816114, 0.0043465742287789738141, -0.021049262009670891699, -0.053459870672741965913, -0.069417032528066607844, -0.021828078714535865262, 0.12909228447203574541, 0.34676675612887619682, 0.10519252583877873597, -0.74544809272182055437, -0.001055303463203836889, 0.40196130433037141616, -0.033924146646765067359, -0.065282657448085451568, 0.010177780070745757329, -0.003297802765319023284, -0.018665940287983240342, -0.028746826973369794117, -0.023739917010408756459, 0.0064155450317971568497, 0.076778578985078949071, 0.18787017200492478475, 0.0018664405376749085286, -0.46101414807272356189, 0.18572121249576434288, 0.013479741716678154267, 0.17104867222398278948, -0.12328833224692593851, -0.049319895571794655775, -0.01476920594213776243, 0.027389032120397345527, 0.076083294085059399725, 0.12653321950568507948, 0.15490436187728995332, 0.098153086709817621847, -0.13789906166727189385, -0.48761142082814989962, -0.61684605239225664963, 0.073922666666055181217, 1.1290453543009995041, -0.77123947500042955294, 0.47823370002705284376, -0.0026215089370572879253, 0.0016931583147911861194, 0.0053180053645358778269, 0.0060562385653979700084, -0.00079955783154810901835, -0.021357266956150068593, -0.033273997636028840308, 0.032898962341255381936, 0.16507460691256883867, -0.053361711248402332597, -0.46250093060846392268, 0.37709333155621260758, 0.13602250602903925847, -0.14982827600918718902, 0.00041478140684494914023, -0.0018368036983630094822, -0.0012812796184950409582, 0.003976079707903203489, 0.007045936811664018154, -0.0049030683619823182615, -0.012912997519482940015, 0.014091159396640108581, 0.040570641838802240509, -0.080430821029449428572, -0.12035451131766107535, 0.41802710950867189021, -0.39899101758359922787, 0.13822339524489019036, -0.0097427076643544600898, -0.021487036706516245838, -0.020060637512942215699, 0.002620803255519334532, 0.049745670533684414882, 0.1056810405274708764, 0.12689287167079787988, 0.037588836284081120209, -0.18210751177718917382, -0.48091503220025377807, -0.34849253649466155336, 0.86088414054548634269, 0.1440321228402135223, -0.24570955975958294615, -0.0078261343520166677667, -0.0030367992790210976417, 0.004277899728235016498, 0.01216484213919763642, 0.017400341189679442017, 0.015224026495128826705, 0.0040061485463480841357, -0.014025640873004476677, -0.02496799602610256727, -0.12829910086068804786, -0.074679107034366601359, 0.54978520276098441499, -0.58188066706738450495, 0.2417367088233824679, -0.00069812473899568880647, 0.0073758616396904699147, 0.0044802546484525550274, -0.011929849999583715461, -0.030201866035291836249, -0.021254570907207830099, 0.046847430114040231053, 0.1452655911077080364, 0.094461818907787778454, -0.36544442729415727023, -0.42740107412076833171, 0.83267124426795047221, -0.14834563597327424245, -0.1256460153901005905, 0.0027052258748438144827, 0.0052798952412250179864, 0.00133107362133822924, -0.013050070224370701258, -0.032916249483903439743, -0.026473947662064345465, 0.053181897404187356004, 0.14010845973498325945, 0.037852553302557830073, -0.3012141462746501519, -0.24167188250386273385, 0.69148147648151692124, -0.29175530211800665414, -0.028566669863079496156, 0.014767182040308031443, -0.0017055283273281901524, -0.022329925847323357779, -0.039716741029348121006, -0.033103822996466442108, 0.016729445864305934888, 0.089966809959741406155, 0.11270258360982488011, 0.05498994664107922925, -0.20896633544792342363, -0.43967767785802336356, 0.55891574458997139718, 0.22378721381526373024, -0.34262313522993803616, -0.0023921037167991573806, -0.00648904922064829498, -0.0023938674869386023306, 0.010071318402640734233, 0.023569307459774389846, 0.021090223236371510862, -0.018718102647988996146, -0.13590171845932710881, -0.25909490559136727494, 0.013643354802221644914, 1.1046011813651701594, 0.86149144878839467587, -3.260464418309811041, 1.6504644804267112335, 0.041328957925710660404, 0.034362510659990325168, 0.011672145838242220517, -0.033094836461880480782, -0.10307088140528108955, -0.18872064722369041867, -0.24253093138926890004, -0.1965805483449224178, 0.079311866109132544977, 0.88641205219615926136, 2.1043784920555239637, 0.25179077260422372264, -5.7297386600674196089, 3.0236604952026602255, 0.028028688183047739679, -0.012924100717335636201, -0.045937087948830557238, -0.057234917475627117289, -0.027831285469848116021, 0.044486066123214179635, 0.10909718394631343075, 0.095617280558676381874, 0.077225308093686356736, -0.096597610707279138453, -0.58754654427091046465, 0.30609769787281870812, 0.80893983287909210222, -0.6710687606748191314, -0.0036269103160927553854, 0.0043410510465210641953, 0.0094022187756147377474, 0.011263689081523468871, 0.0098655048053159707933, -0.0002903138420051847024, -0.027627861829248603898, -0.10561365264416036625, -0.20857945985711112691, -0.014383403586409644268, 0.94398416602786505702, 0.79447206000117454039, -2.8953201854750432176, 1.4893636547298814143, -0.0054163389965435756665, 0.0044130452252592201196, 0.016420634212608819363, 0.022255635539249179339, 0.0040636160370858638233, -0.052449519917985451545, -0.11709903322803578984, -0.059377732134950039766, 0.29180282944875340911, 0.37515551500370741822, -0.64995424344504915126, -0.35628336988552050668, 0.91953849271361376427, -0.39549208269206259647, 0.0077413290542975993214, 0.00021271628527731262443, -0.0068591472760539135695, -0.0084649243957071722971, 5.7991440106978280403e-05, 0.011451159182547467411, 0.00609694902553311463, -0.061613752892667572603, -0.17947152839186650275, -0.070179022545642538722, 0.77900693843364110514, 0.8278687180299292292, -2.5370431682859764244, 1.2289149384679989296, 0.032302414384002914316, 0.029570540714172706409, 0.010317718824120459403, -0.031750880165083039675, -0.09508441188449469339, -0.16015482631540509106, -0.17900424791278668479, -0.1205494927925642723, 0.060324804132062696815, 0.62036677827544850494, 1.7368031290156009661, 0.46039238046576025365, -4.9770017797468826259, 2.5605262251904981063, -0.00089197703237240300467, -0.010218057684886738476, -0.0053030621454504892065, 0.016276170283469847816, 0.038540528363454180927, 0.02497322841456548273, -0.031734789948119214442, -0.092991663465246648945, -0.16033867151067862644, -0.13842670824202182867, 0.7293423147862608813, 0.9804943669010860452, -2.6373398812100958821, 1.294883398085759918, -0.012376396010764026401, -0.017428998198658219448, -0.014401156641220477783, 0.005975727173344075327, 0.055076191753375011717, 0.12067377185381555904, 0.12876448451761060565, -0.05317966539236351764, -0.38198102455665949861, -0.43071029995230464271, 0.37494529230077383586, 1.0245218177556612016, -1.3728130615527891489, 0.59826111653932878554, -0.009196730250274795862, 0.0049033673352378674692, 0.024361148234643276167, 0.038572357169903474794, 0.020076604626621932792, -0.064141512928128821547, -0.1951221732030012368, -0.17290750665067339265, 0.35555853420729632841, 0.71667925294319345664, -0.62495977772722510757, -0.8159450727714437912, 1.0465952448476372183, -0.32665486394441900497, -0.026768763580329445195, -0.022884096685364410662, -0.0057436545295889637752, 0.021101871661622258192, 0.039895466350679406609, 0.045462895046599838267, 0.11710147844473101086, 0.34623551468405727727, 0.58155080977597350955, 0.31503040565334428091, -2.0117265349312951628, -11.792648087501460097, 22.454085832779014709, -10.049456676002346711, -0.014810468588235967125, 0.0042737819373979829551, 0.0095596016968879649056, 0.0024427821362948271, 0.007422188608857183123, 0.056211695775431808064, 0.11717216945331793143, 0.042797862307327322329, -0.22193377425214361187, -0.44811577899708970874, -0.043687469045150993663, 0.89933590684216091482, -0.57327931036669488307, 0.1907948225793337782, 0.010008896830598551508, 0.011307285267764331793, 0.0043465739709783570224, -0.016311867322476372721, -0.048694896979155660732, -0.069966279911221868071, -0.032148415354317130821, 0.10847527659718116111, 0.32484726669985403147, 0.099244663410152131333, -0.83696162524045181197, -0.033924146303037611139, 1.1587065863789114406, -0.69573938430726034454, 0.0082008854025232176543, 0.0050733400928196573099, -0.0032978027642551581422, -0.016432725955567401704, -0.027195586272269381412, -0.025683749318035797859, -0.0034024144955461374171, 0.058267971218482317308, 0.1769551497177244137, 0.020985073066527217683, -0.50735829785985298024, 0.17104867286356936762, 0.47843484718080864226, -0.34666727138834235733, -0.063577846766556339597, -0.049319895608708370915, -0.020716535603889162098, 0.026303905524664588506, 0.093447167816444728916, 0.17834288810931361868, 0.26393145164634829181, 0.28813942183521557006, 0.10422528575803169382, -0.41722447642214077757, -1.1099952517865563362, -1.1206503896217396843, 0.47823370025012185147, 1.5096649801958275106, -0.0024951843164302952779, -0.0026215088402011701811, -0.00017291747209272016598, 0.0053054664063219387701, 0.011953693735293136752, 0.012942986456159957104, -0.0063142985679779809263, -0.048110868122285321835, -0.049389473816727286204, 0.093948565187105800844, 0.16339036309626581156, -0.19408847193870495951, -0.14982827598433048899, 0.17008001119601501294, -0.0012664817905403767239, 0.00041478151769256467766, 5.2879546991160434688e-05, -0.0015082263641935864107, 0.00014125583609714601439, 0.0066991722521083131581, 0.0064889151112033213384, -0.0068745464688496600927, -0.0074399716683457867844, 0.02885030761988095771, -0.0037105088759333648522, -0.12992027330674746355, 0.13822339526787388864, -0.028110257192942505328, 0.013324014315498771549, -0.0097427071114737518603, -0.033338001635321967697, -0.049455118307359248464, -0.041605144621522191506, 0.011439163133040945036, 0.11791036951283619116, 0.24395630691142958235, 0.27956228060722398965, 0.075382857798978394892, -0.47187371596835753218, -1.0027524585956768544, -0.24570956024899534564, 1.1767655191185102392, -0.0036955651394073776167, -0.0078261341976976414592, -0.0081168381749099783457, -0.0019688054481644761372, 0.011355775143776810146, 0.028343677422194957227, 0.039741556779085494833, 0.034050173144957708582, 0.0046452380905615854589, -0.034427215995476496968, -0.13954972492375466997, -0.23062190651473318814, 0.24173670871366778612, 0.083247365485815238983, -0.0064062718748094105159, -0.00069812474470780892799, 0.0086635294788689354989, 0.015974751657333082638, 0.007959474717984624248, -0.024245112703573986485, -0.062191446102813838226, -0.044519292331073316571, 0.091888959356102334342, 0.26141375729583293808, 0.052132663668839278326, -0.50002143217520123031, -0.12564601544208808814, 0.33589295100687655005, 0.00035725031588077144681, 0.0027052258105215083207, 0.0063310204784216006502, 0.0080688086836268688856, -0.0011771799948039872192, -0.031039565727536758133, -0.066545648168525747823, -0.038516484080460294814, 0.10366121312782804187, 0.22837286407688126011, 0.029681630037794250881, -0.36784904445289584585, -0.028566669912680066334, 0.15489351658309125082, 0.019753603391152702029, 0.014767182022554720469, 0.0022517501219414840993, -0.021824851189411739044, -0.057065145786999157962, -0.084981485332569453583, -0.063224568340364606378, 0.043528347469263276026, 0.19788884013472646251, 0.29445230477675626846, 0.10537285059521703245, -0.46572681225789286374, -0.34262313525532223091, 0.3478600895590219233, 0.016660246828480568954, -0.0023921037566480419893, -0.018357868053696237837, -0.024066401012861931424, -0.010785077971961886528, 0.023753452682446475619, 0.063896305072618744814, 0.070183951940156585891, -0.03647950373638742666, -0.30515757812956095796, -0.45013200067690373185, 0.39829920432983278999, 1.6504644804776096301, -1.4104485216528348523, 0.010955760011178412514, 0.041328957787927222967, 0.059577781347554731317, 0.053922127898576151239, 0.0099875274629170687768, -0.085581623250153615601, -0.23589505046067402172, -0.40588116012810371291, -0.50450534896109444638, -0.35194374852482768201, 0.46992968583763095047, 2.4346731522105438117, 3.0236604953156778208, -4.7352094823131807289, 0.05090122124134335152, 0.028028688017291036177, -0.011619159807290278461, -0.06595370622417248696, -0.11742467085723326536, -0.12678443604195890893, -0.047403259580656366412, 0.12004673558321821147, 0.27909182646892505275, 0.34811904325974207808, 0.18708728680985256543, -0.5041071146376272516, -0.67106876064834741769, 0.50982366210636476112, -0.010431951429055872499, -0.0036269104402203741593, 0.0057754209269299605981, 0.01607634216343420408, 0.024064015955865721008, 0.026831840330365699238, 0.019954821958391138342, -0.0051533781284870102193, -0.085190381175113960577, -0.26539170797160233084, -0.35590571387788094704, 0.38667780658302514096, 1.4893636549216475728, -1.2661028446168256512, -0.0092132174644811940661, -0.0054163391323373815539, 0.0027155282785353389297, 0.017958615646510505542, 0.038203635706455192633, 0.045079754160733863377, 6.3314262539023368967e-05, -0.12287232414753679333, -0.23749198259857240423, -0.040195401095025090621, 0.45616908382955062562, 0.23719055746314149369, -0.39549208257656914833, 0.0058566071752116383858, 0.0048124339402107500427, 0.0077413289554508074586, 0.0053749702387649346547, -0.0037859497968417300112, -0.015738549991988769888, -0.018594355696889500879, -0.0019300953561598561091, 0.026606247443496905403, 0.010193802183943098671, -0.13486643084728133779, -0.29245767956746471228, 0.23087180276168647874, 1.228914938609217522, -1.0765786058485016241, 0.013471393542644016436, 0.032302414386062197615, 0.045111622760938309418, 0.041775944421189932509, 0.0059560041265916911085, -0.078057445660064048165, -0.20779811262527300708, -0.3346289939513411027, -0.37390142769791323563, -0.22790387706947723423, 0.35902159493282664871, 1.9005858214647266546, 2.5605262251156402087, -3.9167909500028290637, 0.0069846340734154360383, -0.00089197709544810684437, -0.012698881384484323256, -0.020800756980290727172, -0.0087096141520033587557, 0.032745059269126165769, 0.076375609094540747313, 0.066455290931954083544, -0.027961330427744984783, -0.21025195087305317831, -0.4095784480140033379, 0.10316817295400017296, 1.2948833981465359688, -0.90052810297002328088, 0.0062816551452736591257, -0.01237639593255816621, -0.028799839406536381636, -0.037858787015734078618, -0.028018028973273600979, 0.022961354710489812608, 0.13380622292017810282, 0.26564303369006231348, 0.24505991982438912746, -0.13848707394930659387, -0.74357682529588831066, -0.68050955027948856202, 0.5982611163605889848, 0.45103589511636310583, -0.014612180535361361799, -0.0091967301617110960854, 0.0022589483058939735516, 0.024803730120866347708, 0.058562331012529937613, 0.080643839402356712331, 0.030188118655896802045, -0.16052513827497075227, -0.4043652424815615265, -0.21859162228035428077, 0.59430108802089798292, 0.63349570645288066917, -0.32665486414022404382, -0.30830874807684227612, -0.0066375689890735122684, -0.026768763522299121593, -0.03937927308770131779, -0.034224622300559906507, -0.0026150396312637907981, 0.045066005898675702557, 0.076536736100863783649, 0.10761150179840833774, 0.27095849550400991079, 0.67876938391634533687, 1.0830001582193289345, -0.10093568650757445138, -10.049456676167023872, 8.2144700476082537222, -0.028834264125982441568, -0.014810468522687575757, 0.0065861841678629233499, 0.02693779670966424053, 0.030193283477474974069, 0.01563785604865737186, 0.022747620119913838921, 0.096353991262588450928, 0.16257914820099439446, 0.017578910559114584167, -0.43116847120165258112, -0.67482221973273270876, 0.19079482258084001778, 0.63759474015596051544, 0.00010707735664461883103, 0.010008896612562554265, 0.018245519677613841747, 0.019824927785116939344, 0.0040854637861331873083, -0.040929475756825366151, -0.10989692431410633122, -0.15168251234963983642, -0.059882841399037885222, 0.27982359929303035617, 0.56170500210091811599, -0.065282657575240279901, -0.695739384110829584, 0.20945508096755446892, 0.0033153833184767162248, 0.0082008852936185476501, 0.010177780076449795266, 0.0053960013881266368288, -0.010670839693155746014, -0.036858995253861139008, -0.060913233471408770314, -0.060693937808680639701, -0.00021063321557552010937, 0.18063522703135281855, 0.29379998327237261124, -0.12328833256948276698, -0.34666727119211465658, 0.12536820734806750077, 1.3097911127157226829, 0.37795345161648064014, -0.99836022568919013942, -0.92913536030555332612, -0.32710542218709148798, 0.10478782561024128916, 0.24958334106083149284, 0.22103695296212783061, 0.14544168618373251589, 0.073728266995441049669, 0.018464046601288953975, -0.019245429367067370952, -0.041118059562609513002, -0.050818138041508059688, 0.17776031129790967333, -0.17875875548828054629, -0.1780186633590629397, 0.17658502648737126051, 0.08782122505059813955, -0.052987562206978790214, -0.046164753307644745828, -0.0050412118789476016364, 0.012227291369092308307, 0.010733555641530286018, 0.0045778507522608114302, -0.00023515666091343428762, -0.0022852998727248409298, -0.0021201048365444472542, -0.045047841732990895336, 0.14256800668006977539, -0.1058869063629327284, -0.0053880346774162731285, 0.016837403139265082891, -0.0082789209248634171301, -0.0039565509379743905166, 0.006652231742980961246, 0.0059014226228892097473, -1.5097485739528455299e-05, -0.0014095047202585238386, 1.3534171847744662235e-05, 0.00037022990837620777188, -0.0010692749103307060479, 1.0403707317441592828, -0.25774474951680470536, -0.88470544840409737208, -0.37458095510652517079, 0.084673640885871592765, 0.24264122520069736355, 0.20553786831739251362, 0.095566651145767983477, 0.0056484554644730391657, -0.037483363051158090362, -0.042524020921284814978, -0.027921864237413589888, -0.0074610445582640200907, 0.012267419470278059443, 0.021785834652516485738, 0.28784127795383690485, -0.19999909516969066603, -0.14890422378405007708, -0.044353930735445044886, 0.0072853616520760474187, 0.039237176500616977037, 0.040082663453192848846, 0.024669571728233124358, 0.0073111058632311992297, -0.0042092431499387355717, -0.0083311096329166788726, -0.006654779867558554933, -0.0018104710993136112832, 0.26089896522095684661, -0.057890748085654315791, -0.44908750125872476389, 0.036245514037213093228, 0.22864039773951097545, 0.080118222152471538222, -0.03744167583910387731, -0.052716586277320724718, -0.02125845846512534193, 0.0057292471255665933932, 0.012995280916405029664, 0.0074506538729997955073, -0.00018048031104427608999, -0.005183881596568875455, 0.12018584047220591537, -0.016161696031472952417, -0.31117516208258361177, 0.038074350913786206285, 0.19173868313411213116, 0.078780298820277344651, -0.037558975862129063972, -0.055587588835146820032, -0.023091131079310458357, 0.0010158816324812038577, 0.0069675276929120621983, 0.0046763427076773915123, 0.0017707113319581089042, 0.00037452243148206716076, 0.33893917865342115814, -0.2973062533018073772, -0.44221558448749898051, 0.083244979050567102785, 0.25436585709550024736, 0.17247339416105259002, 0.043998745104205830292, -0.048272176283407658426, -0.070734645267185841666, -0.04988214587977733705, -0.020614630024074761389, 0.00055719044291394461202, 0.012314558995133548236, 0.017564352810890769552, -1.0265501462166493063, 1.3076355649601532072, 0.18773138733429586833, -0.404838011208566384, -0.20258385411868201631, 0.024401216204788477721, 0.079004952811553313952, 0.045373966641422867896, 0.0049778680635801583754, -0.0180847347881279083, -0.021976489580590848061, -0.013452896720366740863, 0.00024229316387970947777, 0.014527324181453769858, 0.5461619498887517743, -0.69092209009232807837, -0.52202836649426742621, 0.18540617544837928454, 0.34807594014230203161, 0.27677629367120243353, 0.12001492349516648672, -0.044002212795141243395, -0.12235278173585695249, -0.11454333216328903278, -0.065634136957033267779, -0.013111948051656245259, 0.026370955953712346687, 0.050548489305290780771, -4.7053925516349757174, 3.2833263450404994721, 2.4128768682375425492, 0.28030214458969937574, -0.48756561850796553692, -0.55670290577135228549, -0.39518996274199358343, -0.18754656098437103018, -0.031101601963259221939, 0.04813579731772743564, 0.06924401457790924086, 0.056366552819861664991, 0.027499426531953065356, -0.0064940586795823040944, 0.037481908281562906216, -0.41794836880555796821, 0.20681991847026967579, 0.44857915527760122698, -0.024564347496864171599, -0.21736615101856304744, -0.11449413717497630705, -0.0040850848799712175322, 0.037044213029540008775, 0.032956350223894828599, 0.016742566625001360087, 0.003871218904066891283, -0.0038152355816659568738, -0.0084309450127392372015, -1.1590087790528065259, 1.5383557549348092497, 0.21463362819611436771, -0.47923952979091366533, -0.21431493061321896243, 0.007348295652583178815, 0.035841879272760043817, 0.0146533605078508853, 0.0073069629205956762585, 0.0095715830933623124999, 0.010412533491778076453, 0.0058738440686986878245, -0.0012058398120547157879, -0.0077086957279666693671, -0.99390419431953247464, 1.1551044382096042895, 0.20977620300329002934, -0.27881830672667118343, -0.13459797338139181666, 0.0027488329005999900811, 0.023074998858451836736, 0.001722408515883343998, -0.011004085105447852094, -0.0097863685390698791633, -0.0022735256974439010079, 0.0035292294818303989769, 0.0051463290491747621866, 0.0035621185764771498387, -3.5316863945546894676, 2.3877769189873827571, 1.6931849735918480526, 0.27470519750844613061, -0.23104895780241860304, -0.34155084839224525428, -0.29728001826043154443, -0.18019540436041442399, -0.062385029084148710399, 0.012468631947619601696, 0.042278516016517636578, 0.041956332069690273345, 0.027101067067627014617, 0.0073268975975590714192, -0.78326005460730485641, 1.1672425966678974252, 0.042931896045280353025, -0.37672378482631313501, -0.16543973087956725943, -0.009617198608238751445, 0.059126156435561412483, 0.06201006643231057186, 0.025537600999853558165, -0.0077370639524687225691, -0.017174813483600379072, -0.01050432110196095406, -0.00084395289029471449657, 0.0057575754608407796187, -0.30708086716801213933, -0.35060649356733114557, 0.6404683812712765878, 0.61876005787959931226, -0.21165918384325030366, -0.41279098093105176526, -0.1684072091068951893, 0.027343463225395776839, 0.080480194567879928536, 0.059312620769705476365, 0.025944762013925905253, 0.0033781742015380287342, -0.0086821930561118559888, -0.015196447132189672483, 0.5239370313217966979, 0.57895024502126479771, -0.83439059223207689975, -0.78088745808978821117, -0.032384963248974975303, 0.33078941532240369039, 0.27405261689813187331, 0.10371402886499589413, -0.0060482607978879429633, -0.042542740158245460902, -0.040426312964720136078, -0.025555229391857330057, -0.0076110969032714336988, 0.010463248185900551543, 0.63759473987137915696, 0.19079482263548358523, -0.67482221924783980516, -0.43116847151088599777, 0.017578910680321021415, 0.16257914806884990466, 0.096353991512580672296, 0.022747619812586189481, 0.015637856095455964639, 0.030193283514071807272, 0.026937797392309545397, 0.0065861830665563132065, -0.014810468303697355702, -0.028834263893499082615, 7.5003959678562051039, -9.6177917276271731595, 0.3028159784440672686, 1.1433844676623847736, 0.59501840711406950302, 0.1906902821113289348, 0.071968765426341799185, 0.06860627679676749302, 0.038825784670470035731, -0.013678271879247427129, -0.043000864592776778805, -0.040157686954273554003, -0.01990552134906440479, 0.0047228290148046917635, 0.17362496287595471922, -0.65995314410446070941, 0.0040885706480399547358, 0.55749242677572341531, 0.21583222429494836048, -0.10201640601387239327, -0.1477748456945051081, -0.086159394344943684985, -0.022590086517019009743, 0.010454604165514049668, 0.018438700029964082139, 0.014487246939154136902, 0.0069028223318478177523, -0.0012136413932684962224, 0.14851405992301958126, -0.36957436300134482998, -0.11039729814053991785, 0.30161852098794700305, 0.16438745929982975791, -0.015659056801271973441, -0.063100452239095577989, -0.054990226824392550442, -0.029730117311713905037, -0.0061012212959446884483, 0.0068434243108349253684, 0.0095428480776625686927, 0.0066655704397556147373, 0.0016450312526030703292, 0.37795345185952655687, -0.64823810007817450618, 1.0267744653611827754, 0.010302252524096165362, -0.54086862418978953748, -0.4033895813677525477, -0.10712697220830746936, 0.089732676289456664298, 0.13252287521252437141, 0.10513172828109088042, 0.06169036149470074476, 0.021295654132293159733, -0.013136688380946590776, -0.041118059558555478117, -0.17875875525654852383, 0.20729887701994273441, 0.32631353391686068743, -0.47629652876698774788, -0.027573471016239400888, 0.16817441165936211633, 0.025953710581808026731, -0.034577182163753300559, -0.01992879704939201635, -0.00025657779060359086135, 0.0057085268807052485379, 0.0047650408733698199418, 0.0014217681509953480727, -0.0022853003790046047018, 0.14256800712697328959, -0.39014674813317773427, 0.37701616041480356945, -0.1004082071059567427, -0.058129271502114034098, 0.030694623891427668533, 0.0065307522712464392833, -0.010878157032864968837, -0.0032541953302841078802, 0.0064492446408031603444, 0.0034520520302537666645, -0.001177349410329391402, -0.0016308150859265827751, 0.00037022926390477159343, -0.25774474960204785123, 0.19097662543114413136, 0.7624783745191975326, -0.3656928907338392043, -0.41049777430457651262, -0.13864641776302041865, 0.03803487891611878613, 0.11020925743666357133, 0.089000139739072808842, 0.040007432040834345344, 0.00041041861555800307482, -0.017745443508335879967, -0.018048116617289045732, -0.0074610451730175728416, 0.28784127801159520299, -0.66853430684496750747, 0.54303709709867076239, -0.024909228321011726154, -0.11388186870952379803, -0.036146661826764896197, -0.021425107804482926982, 0.0047943040798238854491, 0.017546634505161601975, 0.017388245943437782848, 0.010436567403219104699, 0.0026843376706956294242, -0.0033912236676451289197, -0.0066547795743492335271, -0.057890748084511403637, -0.25303644484490500632, 0.80681625757330965865, -0.36804996538067208389, -0.33238585379475638071, 0.084206764585440754822, 0.12564727659447100239, 0.039334262551040180544, -0.017933950383285993807, -0.025263726497120660847, -0.010324697043339265204, 0.0032985110611620884792, 0.005990361843817472963, -0.0001804808991051160764, -0.016161695986008899756, -0.26810945662348495411, 0.61276082107627782847, -0.2312055929080292227, -0.246815523592171987, 0.042112851912092845452, 0.11788344296221693608, 0.040667326796264585687, -0.025632013439841810909, -0.027582784663955268317, -0.0093421432175642016987, 0.002123233286898348806, 0.0044282173830206108048, 0.0017707113418184358689, -0.29730625315997927238, 0.17939399075852940912, 0.5426525564101047916, -0.41878582823614440844, -0.19207933390010575869, 0.051653723199098680596, 0.095011361817950557018, 0.077344859518590977632, 0.017362711557866804257, -0.02552734501607172371, -0.033067848643742195347, -0.019604691246235310398, -0.002312952586910152486, 0.012314559185548639916, 1.3076355650529694064, -2.6667946958756334475, 0.85401356059480248994, 0.85150561707184768689, -0.068319146194688831764, -0.23425550321847621316, -0.089676097526167189433, 0.012473497065271236389, 0.027770335515153093953, 0.015933871389770287719, 0.0014380090207746047681, -0.0061601419890445008667, -0.005859141593543616168, 0.00024229324746931715263, -0.69092209017256045467, 0.84922011193915392102, 0.30938818924272287747, -0.61041134591509638163, -0.10389082668890900485, 0.081188143724826966618, 0.097337330250376774554, 0.10836423788352841924, 0.04422458164062872954, -0.026851990456319640987, -0.055538860763875624871, -0.044548172608307304299, -0.01280008984506556674, 0.026370956251213245602, 3.2833263450677203643, -6.1160483684551358863, 0.3686245856292735823, 2.2657282889538805648, 0.82774679490679037297, 0.0011939987721330536097, -0.2359083463002378811, -0.25732768093015451205, -0.18050291996707951148, -0.07986304914248612552, -0.0093115589671590448528, 0.025827309950503599123, 0.034766655702720847265, 0.027499426434574321398, -0.41794836895425790946, 0.94209316696355593912, -0.33961929659547784865, -0.65530018648475774068, 0.35908205655919589461, 0.28257862295697427912, -0.054275005494588857147, -0.1068895562049222564, -0.047903500103139984301, 0.0021455569051595044933, 0.018497518713418301117, 0.014083625453185543569, 0.0042228914769276704178, -0.0038152354514456704969, 1.5383557546793280579, -3.0970998168340053702, 1.0126971903477566261, 1.0261369011402938778, -0.16625140330650445275, -0.28253783172346902397, -0.072010410648138595269, 0.018164477833085310859, 0.017621808865918547893, 0.0063664211034959580435, 0.0018427361556129940451, 0.003065339974406497154, 0.002960019651732853814, -0.0012058394492614756769, 1.155104438209392681, -2.3768991085020925169, 0.78484673928491188732, 0.71862870428767489095, -0.066122540112841132753, -0.16862700025233406942, -0.058794862485075483072, 0.0044452748369122382227, 0.009223876730246015343, 0.00080662207754100172588, -0.0051691962179165667698, -0.004072983375274567136, 0.0004752300385832125139, 0.0051463293313697292333, 2.3877769192580209356, -4.6318836488887260572, 0.48946088077286364904, 1.6021347805220245153, 0.53683688777813187709, 0.030540699864720896595, -0.11696496795728300588, -0.15989416920247404019, -0.1406319268381017451, -0.081765696679547056336, -0.025162507206188615416, 0.011593305362925396262, 0.027036672552421030236, 0.027101066667483682954, 1.1672425965099191281, -2.4162102964031149099, 0.96694636403710920725, 0.6375612872610948445, -0.16536632928381586605, -0.14653530150280227873, -0.069483786794922813845, -0.019601213737416720712, 0.022141320894170778144, 0.031059709071302787908, 0.012607286742387580764, -0.0045611388786977648951, -0.0083333862830093242274, -0.00084395247039510353816, -0.35060649383433822868, 1.0931999070204985181, -0.82733629233916683443, -0.65528212588171286246, 0.7243342105923428198, 0.37275206421755824326, -0.17111545726918658827, -0.20045446667646793504, -0.067319596235920403648, 0.019458194277205423384, 0.038723752382106124692, 0.02456022921926370034, 0.0051776619391330491737, -0.0086821927484610848108, 0.57895024511094383968, -1.4796034411494631833, 1.2393645907812089746, 0.38608078666526890466, -0.58499487171734709801, -0.40664939300030655822, 0.0088427377176646115853, 0.16896907990398227217, 0.12194882977536436464, 0.040149966806218032045, -0.0071499009756847142771, -0.019864762607005377604, -0.016385343095824088377, -0.0076110970211228611004, 0.19079482266058681028, -0.57327931057012204707, 0.89933590689444875554, -0.043687468808775178908, -0.44811577905167182534, -0.22193377441743800116, 0.042797862192167163142, 0.11717216955579587423, 0.056211695956666161977, 0.0074221884901139734136, 0.0024427824052635355168, 0.009559601036085357828, 0.004273782528827946986, -0.01481046878492287, -9.617791727675781388, 21.672407062587591753, -11.870408480536541518, -1.6671828043011216458, 0.49431855581110539788, 0.59738619728615671711, 0.29563682167443700033, 0.066871735104384738557, 0.024187248450922266274, 0.03708527270020761657, 0.01991697720964889784, -0.009005644331578188666, -0.023470827511802686699, -0.01990552059081753325, -0.65995314445769770462, 1.1634399445827992015, -0.12442429399582817517, -0.8218938604878138765, 0.16851341583215628805, 0.3305231736977113921, 0.072755832867144903942, -0.053916642851785934187, -0.067374311062491309965, -0.037135923753157613902, -0.0080581631794983780981, 0.0065321428946314187558, 0.0095424274055794927607, 0.0069028226727616227507, -0.36957436337566546003, 0.55162045386353342202, 0.12537488906002097755, -0.52972827215094220321, 0.04269262632257658785, 0.18363386475581769397, 0.049830140223802429911, -0.0099904463344401006619, -0.026247211592801460156, -0.024344800278198484855, -0.013202439269368764982, -0.0015235005621963971893, 0.0049986494148554737327, 0.006665570497016313524, -0.99836022680741332014, 1.026774466790771223, -0.30251994323607339643, 1.1184482054705839893, 0.051163274135003863896, -0.55702210983701827551, -0.44944464286974061817, -0.17141655657780804911, 0.037427709525008019675, 0.1035151612054632797, 0.095649675981024612215, 0.061569069280663711596, 0.021295654274311581061, -0.019245429545590105741, -0.17801866334369689793, 0.3263135338986772882, -0.24590426564190012715, 0.36148513407178073376, -0.35492512142420257204, -0.03843284230503594967, 0.15048825833516890627, 0.024232176047941818287, -0.033979158966674367914, -0.021389176395602556985, -0.001088774230771521502, 0.0058515238362764074656, 0.0047650406657084093051, -0.00023515606900291171136, -0.10588690704488619432, 0.3770161603816282736, -0.5437053949146988252, 0.36451895323671251736, -0.053055357407918120327, -0.057330773716080064839, 0.018946312207387971255, 0.0030708121223465808039, -0.0069756040119042794778, -0.0023642919783642434885, 0.0048782163790079217242, 0.0022949247288536109811, -0.0011773496103547533443, 1.3535134308682419458e-05, -0.88470544820518892681, 0.76247837439695820283, 0.19472829122539836444, 0.75111789416904417571, -0.39418013051130201818, -0.48632185994347554292, -0.19258692338877811401, 0.011699288798942324152, 0.10936922476705060459, 0.10127733853440146206, 0.054235468889548860139, 0.0091968141385137293364, -0.017745444024770878899, -0.027921863629732011219, -0.19999909517695940719, 0.54303709696731250478, -0.74081997849674918744, 0.5450954031679469125, 0.021539469383692624871, -0.12297867783622047633, -0.05058024774710131316, -0.034230283958397833555, -0.0023253673600921957038, 0.016192132691168113767, 0.019685698166952404764, 0.012914041996063808926, 0.0026843374094737281095, -0.0083311101361266994003, -0.44908750118111345762, 0.80681625768556552991, -0.57031179294075018404, 0.59809781755824620308, -0.29951696642974223739, -0.29620467587346760441, 0.091832306062734958552, 0.12553235213319841046, 0.0368571515937226174, -0.020905823488107204422, -0.026819077414260008435, -0.011007347854653092012, 0.0032985107216783439779, 0.0074506544718212852976, -0.31117516179304233193, 0.61276082120600450232, -0.48175748796081141467, 0.43472016752182102151, -0.21463234544186568509, -0.19958157835743567388, 0.056627165081315225525, 0.11196889538920738738, 0.040056698316289260364, -0.023594379161899506347, -0.026796773630844445552, -0.0095291569698927949816, 0.00212323332685637918, 0.0046763425515412267511, -0.44221558471423005887, 0.54265255632458830881, -0.09850971750156595208, 0.45856945649161318102, -0.45865627711587186877, -0.23874898219412071798, 0.055587284288741989347, 0.11265935682595919209, 0.093933896352148743558, 0.029883860782726534566, -0.018157995295929763024, -0.03095716216249204053, -0.019604691022380514587, 0.00055719035950761676076, 0.18773138727532220904, 0.85401356045644616355, -2.260545749167029328, 0.75073313255686136003, 0.80161768755621543026, -0.059449298156478612432, -0.23061408965921934011, -0.098028829603049066899, 0.011742347727266268242, 0.03436494622117083203, 0.024436790378905299398, 0.0070898468412317894316, -0.0061601419393534240926, -0.01345289704625766275, -0.52202836644738270788, 0.30938818922319616389, 0.52454814761351487018, 0.31429001048854837697, -0.72310417353385836314, -0.23559867637362585935, 0.062885180048561695187, 0.12722533954550863111, 0.14849286586550347611, 0.079359742695674695412, -0.0049993110856038636136, -0.047678979189388598359, -0.044548172230118268655, -0.013111948186595924482, 2.4128768683238273063, 0.36862458551342824897, -5.7266765211798968949, 0.16329349316795827018, 2.3039403356481407137, 1.0221417688011324998, 0.14530092517099824079, -0.17006774891477183242, -0.24361908733228482271, -0.19799639824903569885, -0.10568959213575135347, -0.026840693459951048272, 0.025827309858951652544, 0.056366552947156978171, 0.2068199188784531628, -0.33961929683313635264, 0.26783767785902190539, -0.043824355901276722258, -0.50461489680035576111, 0.32751846901162628978, 0.27429171799690410527, -0.051672987321904458036, -0.10911402687797543132, -0.053896034073928414387, -0.0035233839440721691216, 0.015633195181014608588, 0.014083625318590520942, 0.0038712184029853213829, 0.2146336286683440675, 1.0126971901602788151, -2.5942019421864630502, 0.79795610470278788284, 0.97737358864916135115, -0.10498927766592980326, -0.25059525585695063832, -0.072955541938957604864, 0.009920531448639448302, 0.0094252716489018773666, 0.0016294398997349548303, 4.9768035884205710784e-05, 0.0030653401024653523406, 0.0058738434570566990314, 0.20977620274315814108, 0.78484673902006130142, -2.091397896325791983, 0.6475241451214548416, 0.68539789657110883336, -0.040431562062873141472, -0.14858217448798663463, -0.052373389420591331955, 0.0075113021754611378894, 0.0094884051747650809822, -0.00014824133300299965205, -0.006253159361060323454, -0.0040729834141149338228, 0.0035292293017104305573, 1.69318497349471353, 0.48946088076116978094, -4.4217631960706595962, 0.24771544845003529156, 1.6050340270832441636, 0.67916358099191531661, 0.14688168886757083098, -0.059499117124269501611, -0.14206028530903763718, -0.14650444001200577104, -0.095615860941509159265, -0.03569344368595966932, 0.011593305045523347174, 0.041956332072669418054, 0.042931896274694819615, 0.96694636403439515604, -2.1573094564272174978, 0.82927857740040245016, 0.68734750667349231446, -0.15145495732311584081, -0.17300696173237742337, -0.082260615423475319519, -0.01636740737500925677, 0.026847191935105221838, 0.033686645461836597915, 0.013727038993371003944, -0.0045611384861296341148, -0.01050432158953916971, 0.64046838178251130902, -0.82733629223487292759, 0.27569111035913179686, -0.28837102870574071734, -0.43670207112810155436, 0.67086673846107147678, 0.35913055249792036339, -0.16766798706820706943, -0.20714183649344289351, -0.079593144352510247774, 0.0094230737102924677639, 0.034349629163389461262, 0.024560229208382224064, 0.0033781734891513044705, -0.83439059223928002673, 1.2393645909204580313, -0.93947513465057297388, 1.0523824805569335084, 0.24947249212925168282, -0.62699598513876120265, -0.43156739108130398597, -0.016237708489980962057, 0.16083038033003579925, 0.13000179345553969368, 0.054330633164279051872, 0.002516411058415907176, -0.019864762791275254739, -0.025555229370605683781, -0.67482221952467258408, 0.89933590675431307471, -0.45169626630781789522, 0.76837786646994343798, -0.090849511961373127789, -0.42492490753887440791, -0.19402990791135360515, 0.02971314920194767506, 0.089188946887946179976, 0.036763646768808444509, 8.5134138316900198618e-05, 0.00098792275271541996062, 0.0095596008753326550123, 0.0065861842223866630397, 0.30281597836139223512, -11.870408480459921918, 23.915607679361091442, -11.656148745777747067, -1.8844574905524884567, 0.30323311120100326921, 0.45715825535627285792, 0.24274272456307216528, 0.074599675074117688389, 0.045086344713463066824, 0.054014461585388778841, 0.029965644809293701112, -0.0090056438561694068801, -0.040157687385317851481, 0.0040885708532915135405, -0.12442429421138338774, 0.40175944346969422183, -0.026497485125991843136, -0.70962430975990997695, 0.16490018508171364053, 0.34328137556256776985, 0.089200774399494781042, -0.045751684081441187923, -0.067861570782253785583, -0.04186025554111984931, -0.012202560242071744043, 0.0065321432705504056671, 0.014487246892979888449, -0.11039729764680483726, 0.12537488850660022388, 0.070692062839849273415, 0.16413741502559758834, -0.4750540010702284377, 0.022249961194023365474, 0.1929602854572172177, 0.066547274610115447757, -0.0019730177187994319549, -0.025399191189597877111, -0.026312623301208974691, -0.015458945108693517706, -0.0015235004102650937137, 0.0095428483170938976227, -0.92913535947417247307, 0.010302250954931526858, 1.1184482058322127163, -0.39584043071830821159, 1.0561306100629606064, 0.049346353546859650363, -0.56202383282610823834, -0.46680883035470571718, -0.19189870810264358925, 0.024108004969257426642, 0.098675465430857048355, 0.095649675843391637753, 0.061690361658541637002, 0.018464046863848670998, 0.17658502588348867768, -0.4762965284252506093, 0.36148513430425510684, -0.1428396884191981131, 0.37633190428324791998, -0.36786121354216766433, -0.054273635053800228789, 0.14563947308932667646, 0.025301273750077374336, -0.032016709504766954764, -0.020614283886525891498, -0.0010887733422033588765, 0.005708526317241698414, 0.0045778503838205947263, -0.0053880343598505164512, -0.10040820658108627617, 0.36451895246291426966, -0.52422528929857314939, 0.36436684200425267788, -0.060189335906897172002, -0.061204641261209925818, 0.022733408616891468551, 0.0047338314554575188983, -0.0082124808811957980548, -0.003786469197246415154, 0.0048782177610120561195, 0.0034520511651660425526, -0.0014095050063432914764, -0.37458095568121901886, -0.36569289044343816775, 0.75111789461022271386, 0.090622639255183157214, 0.69562015234959706511, -0.39159867068410514346, -0.47710779525820079527, -0.18606228701610594278, 0.014444671481718990291, 0.11181591280747552042, 0.10397611701754733149, 0.054235468363657529656, 0.00041041843665085703252, -0.042524020108955944142, -0.14890422403722042288, -0.024909227849647497255, 0.54509540294259195559, -0.74328953429912858475, 0.52274003121237400116, 0.01620199300630599526, -0.12679783038955091956, -0.05128152089950718806, -0.033317644906255043658, -0.00074923454935179809033, 0.017277779255963580246, 0.01968569697045285255, 0.010436568151658102976, -0.0042092428923890552045, 0.036245513895845293817, -0.36804996568646175525, 0.598097818071361087, -0.54713044430802670881, 0.65828671821451612001, -0.261060503699102886, -0.30108128905179920576, 0.074115216847600862149, 0.11543160875936148146, 0.036250786891683241253, -0.019149901686129781886, -0.026819077113627753234, -0.010324697410241069281, 0.012995281028854468314, 0.038074350323112647609, -0.23120559303811127894, 0.43472016821347303495, -0.49728270621989972833, 0.50051320956300815457, -0.17767530628304334384, -0.20885116555927873971, 0.042469761555308474521, 0.10693378181704840668, 0.041660230441070381091, -0.021658939280454134191, -0.026796773525336624322, -0.0093421433997102950458, 0.0069675278473301785129, 0.083244979494718790436, -0.41878582789646745521, 0.45856945598396331443, -0.16755593327483947474, 0.45371932055267466533, -0.41942889760217016981, -0.2189925877256275466, 0.058452208845078283683, 0.1107068712108776054, 0.094275451442909941058, 0.03144567576386320118, -0.018157995222466079971, -0.033067848872450164577, -0.020614629693326207843, -0.40483801075120773305, 0.85150561749692488256, 0.75073313196065682984, -2.3876330522961053404, 0.67756132226628185133, 0.80023473411906653752, -0.041311950716033704611, -0.21417132023399587659, -0.089163324314103134527, 0.016500405595931043334, 0.036946468790883348066, 0.024436790205433218948, 0.0014380090227507687922, -0.021976489477899419783, 0.1854061753714194849, -0.61041134610228475754, 0.31429001056798250291, 0.40138441594671087076, 0.25380399042508172691, -0.69503709794620671669, -0.1996604877644634346, 0.087547036006520029261, 0.13977799991524253453, 0.15483642876311742853, 0.082867634406491696986, -0.0049993104273998723031, -0.055538860890498198408, -0.065634137512266155223, 0.28030214440781847651, 2.2657282892504584382, 0.16329349305573709339, -5.6861758247463312088, 0.16337683097292643453, 2.3065875620698816917, 1.0412555863655268418, 0.1656060732107503819, -0.1625863179551840787, -0.247085816028321692, -0.20490628696859558566, -0.10568959233770430783, -0.0093115587759449447436, 0.069244014686942884085, 0.44857915531509040541, -0.65530018635586850806, -0.043824355983745906518, 0.44726651797183059056, -0.036635481280991373521, -0.54333499465423562569, 0.29576959636677807586, 0.2702217236028534586, -0.046068576970569471984, -0.10671759272436724675, -0.054981307931745292128, -0.0035233847952803216963, 0.018497519018062302115, 0.01674256695929585495, -0.47923953007350134081, 1.0261369009282295117, 0.79795610510823355987, -2.7292213861397867092, 0.76135847712720405855, 1.0033573932621318026, -0.079973033738283433314, -0.24486791393953380447, -0.07880718338282027835, 0.0049562421792683792471, 0.0077990763927873654054, 0.001629439235836077117, 0.0018427365644841198038, 0.01041253379373462129, -0.27881830629415027545, 0.71862870488187780005, 0.64752414410276348811, -2.1880719102929138131, 0.60510715310742113626, 0.69909256393444718647, -0.018265855137630369887, -0.13649953259234931524, -0.051411139033114588992, 0.0052227262854663012229, 0.0079690180106219352107, -0.00014824176340987298833, -0.0051691954418238289515, -0.0022735261106010323448, 0.27470519758166622726, 1.6021347809816857133, 0.24771544816601248629, -4.4231644154651412748, 0.25942402639924400098, 1.6285274505385747279, 0.70010314798019646076, 0.16103439650315184983, -0.053678239510206948726, -0.14156172031752822793, -0.14881584879365802876, -0.09561586098789492949, -0.025162507496263891715, 0.042278516119424736086, -0.37672378491471381068, 0.63756128685045987403, 0.82927857797579651589, -2.1909506248158878705, 0.76954384440007084223, 0.66228733732624922581, -0.14235818126833668651, -0.15432846028952557083, -0.072464602486345619514, -0.015761946219139993247, 0.024892763750629084973, 0.033686645520131085207, 0.012607286689681496475, -0.017174813370464180745, 0.61876005743686968152, -0.65528212661178719678, -0.28837102756537080772, 0.5520046574501971115, -0.31298744709931286545, -0.51908065083489351998, 0.62455817145028325044, 0.36054202865785223153, -0.15534583324000744708, -0.2036172926544894124, -0.082103839231241926289, 0.0094230725691875905081, 0.038723752731919962311, 0.025944762876969292009, -0.78088745821204352815, 0.38608078674458712332, 1.0523824805152652839, -1.1963541254243912348, 0.97062431039551522183, 0.28737555542284570764, -0.59017460686403488967, -0.42378640987548032992, -0.021217301733125415453, 0.15984279147148808264, 0.13320221539505400221, 0.054330633086732332826, -0.0071499008705700889399, -0.040426313327209029147, -0.43116847128376978304, -0.043687468589341076075, 0.76837786584355027841, -0.58040787957229666993, 0.78351823027544753586, -0.022016708730903444824, -0.41030110784272927349, -0.21968112273774983989, 0.0032727315766311049217, 0.079148413250885507608, 0.036495065243880076, 8.5134955341265028395e-05, 0.0024427821381166753278, 0.026937796717755608383, 1.1433844676920275063, -1.6671828047102994486, -11.656148745477429074, 24.620195205478875522, -11.497617585882524693, -2.0088390548312089479, 0.19484174964242728656, 0.42733846864525693032, 0.24986176672626259077, 0.085805692053134596797, 0.050671363654503309704, 0.054014461570129373991, 0.01991697750877317874, -0.043000864921560882914, 0.55749242703982426761, -0.82189386088631988247, -0.026497485134505435744, 0.54597142792276021073, 0.036599219932360732566, -0.70511614117411103297, 0.14180399712735189399, 0.3316318298634879369, 0.088659240703637412495, -0.04410827329448569617, -0.068001191245300199317, -0.041860254850337709487, -0.0080581632337297353502, 0.018438699589559678221, 0.30161852106972003584, -0.52972827166794478337, 0.16413741447674329299, 0.12954491458437925044, 0.18668430752501283254, -0.46142527096683949583, 0.01914022859691325712, 0.18804329012565193513, 0.064055362451618885755, -0.0024515948523625812061, -0.025624638226249114858, -0.026312622899460104409, -0.013202439220060759387, 0.0068434238099245288736, -0.32710542179456170153, -0.54086862466376495195, 0.051163274080306297975, 1.0561306101089589227, -0.43131471377940167056, 1.0423263668781224123, 0.047931420237117834671, -0.56268918427554315276, -0.46927733894845491402, -0.19470249821211399466, 0.024108004948739738399, 0.10351516073229448389, 0.10513172837575252005, 0.073728267455062265756, 0.087821225532736857389, -0.027573471092921911924, -0.35492512203376974655, 0.37633190454863280072, -0.15267819544145108668, 0.36989165049208855285, -0.37148593057706613685, -0.056462920783223638499, 0.14627372967204088638, 0.02649526864329647402, -0.032016709978544381776, -0.021389176869941274595, -0.00025657759440136538795, 0.010733556119298815618, 0.016837402929387981082, -0.058129271248188930932, -0.053055357931771018887, 0.36436684266956004263, -0.53672244469370333775, 0.36082693075838145891, -0.05544655927809886975, -0.059690370879544096816, 0.021436267195075350139, 0.0034997134609873179791, -0.0082124806207999126451, -0.0023642925521431324733, 0.0064492450472104986825, -1.5097621402321005591e-05, 0.084673641104212257757, -0.41049777401561649004, -0.39418013075117019905, 0.69562015198581517161, 0.082804430657018618378, 0.7155302121280946448, -0.3675608457597437928, -0.4638636397719096971, -0.18219370385884506303, 0.014622773293704482978, 0.11181591341600186085, 0.10127733944828370971, 0.040007431219189933114, -0.0374833634991808759, -0.044353930206522756974, -0.1138818691514109499, 0.021539469461414703644, 0.52274003107242228428, -0.76018846556826080718, 0.51601986498695717742, 0.017852954002023167096, -0.12294535161954089908, -0.047919324021818296422, -0.03185274513006737529, -0.00074923478057602188596, 0.016192133165922821536, 0.017388245709657669524, 0.0073111057693513855846, 0.22864039771912367249, -0.33238585338759918386, -0.29951696693827262186, 0.65828671839480878969, -0.51621071467564616508, 0.6624698410187979114, -0.27058703843089626506, -0.31184952078571120948, 0.070007069761322732182, 0.11575730947776820268, 0.036250786777886644108, -0.020905823487316357867, -0.025263726464072301658, 0.0057292471468029420104, 0.19173868330390575476, -0.24681552340956142144, -0.21463234556020310828, 0.50051320916536956585, -0.46625560602734311688, 0.49716495400169508079, -0.18556630023939327256, -0.21447791451939138274, 0.040661909938548838894, 0.10739331009824522323, 0.041660230716630664016, -0.023594379898448342159, -0.027582784267804780903, 0.0010158817168862595021, 0.25436585633791408689, -0.19207933367070748054, -0.45865627678034515124, 0.45371932086984795296, -0.12684293514588446117, 0.48671651717200709042, -0.40764253493922353311, -0.22071239076946719271, 0.054680267900496959399, 0.10905850516745647927, 0.094275452131213605123, 0.029883860840245433932, -0.025527344871081922295, -0.04988214650667633504, -0.20258385478224094634, -0.068319146554926352133, 0.80161768825684687201, 0.67756132287234704936, -2.4218031608536492172, 0.66921124453461489079, 0.81273706537662759875, -0.026426137416252118428, -0.20527518601098246331, -0.086160968607514043516, 0.016500405552389914299, 0.034364945805486386832, 0.015933871623719335353, -0.018084734615006911596, 0.34807594035985356395, -0.10389082673073660179, -0.72310417335773258252, 0.25380398989460184467, 0.44237537300705864762, 0.31218970735162132613, -0.65853841407061985969, -0.18689421641564296639, 0.088003364580915621751, 0.13787668529351815305, 0.15483642887561444024, 0.079359742564932833631, -0.026851990157250316127, -0.11454333238795297301, -0.48756561841809270552, 0.82774679476842516479, 2.3039403357109553561, 0.16337683096017519002, -5.7249811668991732461, 0.10914828591370939748, 2.276532649963215249, 1.0314009500206997583, 0.16448193948552902732, -0.16284257799277013223, -0.24708581571717258552, -0.19799639752403674908, -0.079863049546997633144, 0.048135796858033493006, -0.024564347932696642407, 0.35908205633705952975, -0.50461489634635248169, -0.036635480951304130737, 0.42819543696798367804, -0.057363826291656150325, -0.5556490948463009838, 0.29298766214731086777, 0.272141873788877664, -0.044801515273570048647, -0.10671759309713105268, -0.053896034157907848905, 0.0021455570057312359916, 0.032956350474893689184, -0.21431493033697127104, -0.16625140329065915545, 0.97737358858933376382, 0.76135847679836821111, -2.7434025259868199953, 0.75430394565903879922, 1.0043790688977802805, -0.079677638562706848413, -0.24560504620362663686, -0.07945272877596060157, 0.0049562425791161920929, 0.0094252718180116434876, 0.0063664211041885351156, 0.0095715826386154924388, -0.1345979733321929217, -0.066122540909434229794, 0.68539789698260544792, 0.6051071536436909426, -2.2024484736461364953, 0.60455310785532834217, 0.70648328468313170703, -0.014192069124730343974, -0.13638304627359795762, -0.052513425845571198858, 0.0052227257873917413283, 0.0094884056629812533329, 0.0008066217408007323765, -0.0097863683378335616231, -0.2310489579200726562, 0.53683688793760431235, 1.6050340271067926601, 0.25942402641612799519, -4.4304964536606252778, 0.22981965443825483342, 1.6080030014234494296, 0.69213860949351435625, 0.16073681469302711355, -0.052600052447202089168, -0.14156172078919024293, -0.14650444062033612935, -0.081765696545085808644, 0.012468632695369314539, -0.16543973084206159418, -0.16536632872156223351, 0.68734750584152615183, 0.76954384463213088008, -2.2484083330501922759, 0.75748942546933439868, 0.67923147936990990736, -0.12763757017370744307, -0.14982156956876035991, -0.073353278544972061459, -0.015761946093059728963, 0.026847191772147369765, 0.031059709298653981335, -0.0077370641955133052925, -0.2116591835596084159, 0.72433421114812135411, -0.43670207154610918288, -0.31298744772093839428, 0.49785885238061639013, -0.35269839013198955868, -0.53679218688255503356, 0.62449897897443362726, 0.36619843900771442691, -0.15276051981973842731, -0.20361729185223773064, -0.079593144045124825103, 0.019458194163269840193, 0.059312620027224857711, -0.032384963273685979379, -0.58499487185818710433, 0.24947249272327681746, 0.97062430979424196842, -1.1932755510394266185, 0.99834242880012658006, 0.3129805997142213192, -0.57805676654917714785, -0.42074542916258717007, -0.020979838400156436418, 0.15984279096829717703, 0.13000179302407466331, 0.04014996672954750101, -0.042542739321121250673, 0.017578910511155721591, -0.44811577895387849635, -0.090849511804537055881, 0.78351823008596321873, -0.53202672536152606497, 0.79652223037450986531, -0.035565010487447201726, -0.42723147376229475558, -0.22779012360420941574, 0.0019510506235899671701, 0.079148412824276259681, 0.036763647626528089818, 0.0074221878716534961384, 0.030193283682787393912, 0.59501840697856356321, 0.49431855620874404211, -1.8844574909197080448, -11.497617585805237184, 24.741411754830167524, -11.472909005169201535, -2.0104800890437646288, 0.19147082900280232298, 0.42653447766953389575, 0.25076949871791909263, 0.085805692009347012128, 0.045086344743747897024, 0.037085272828924820132, -0.013678271942254519727, 0.21583222453203049218, 0.16851341580218565097, -0.7096243098931203086, 0.036599219786687134381, 0.57232097290610739471, 0.032988063217102847613, -0.71924043208680099593, 0.13104064128699266512, 0.32805152469305953744, 0.088489047533530595935, -0.044108272995401598404, -0.067861571118837199501, -0.037135923299059625524, 0.010454603845995532121, 0.16438745881263244764, 0.042692625737656121621, -0.47505400026001781688, 0.18668430797233404039, 0.15376317092401309283, 0.19633277571430293729, -0.46579291690033397533, 0.011482969973428628491, 0.18413240315725165619, 0.063133470133720198669, -0.0024515954134604143327, -0.025399190963805148846, -0.024344800503052817925, -0.0061012208989567979708, 0.10478782566558507428, -0.40338958000436148099, -0.55702211110501032021, 0.04934635357396059685, 1.0423263666445297115, -0.44127776738884527674, 1.0409734990505263852, 0.051343774409608036735, -0.56060272119936982449, -0.46927733874520005664, -0.19189870709118786452, 0.037427710471702756168, 0.13252287431538703477, 0.14544168534333021259, -0.052987562198287055382, 0.16817441180858580529, -0.038432841923362839887, -0.36786121436297675169, 0.36989165067878648774, -0.15288058036386253935, 0.3706185814212045182, -0.37074343822407795912, -0.05595667403365620074, 0.146273729114031803, 0.025301274464095961647, -0.033979159178133377939, -0.019928797103181777189, 0.012227291367317934256, -0.0082789207236626041753, 0.030694623946660785224, -0.057330773533030350564, -0.060189336529487998118, 0.36082693075842753316, -0.53264636822397981586, 0.36244820950934192316, -0.056685451281685041158, -0.060975517870761532235, 0.021436266701606112445, 0.0047338315596108133942, -0.0069756039444923137113, -0.0032541955149352529489, 0.0059014229071212452246, 0.24264122509957480878, -0.13864641851196082833, -0.48632185959439028222, -0.39159867002932086111, 0.71553021191207333285, 0.10693307301957612199, 0.73392650899438938161, -0.35794305616020954952, -0.46121053826355845606, -0.18219370348311506036, 0.014444671464125258264, 0.10936922471978624549, 0.089000140572388669602, 0.0056484546684338705555, 0.0072853614353287831848, -0.036146661517668092489, -0.12297867791733139986, 0.016201992876206123412, 0.51601986510253183837, -0.76122958217422209692, 0.51909042947312367122, 0.021828137498153009355, -0.12082085968651393171, -0.047919324063686936921, -0.033317644821810363243, -0.0023253678632775189733, 0.01754663477130079946, 0.024669571924704015997, 0.080118222430306362813, 0.084206764180183815705, -0.29620467548673218294, -0.26106050414971104789, 0.66246984118718754697, -0.52427466851956250338, 0.65626483598413243303, -0.2720549487184030224, -0.31159961686810150328, 0.070007069580158731714, 0.11543160918991089703, 0.036857151841210383791, -0.017933950579014690946, -0.021258458713990995709, 0.078780298868515993838, 0.042112851843243231942, -0.19958157857653383016, -0.17767530586207092025, 0.49716495374551200603, -0.47329204805817814972, 0.49493201160578537134, -0.18533283473734943336, -0.21419852898753721782, 0.040661910007591534066, 0.10693378153123787866, 0.040056698954505899624, -0.025632013853016187566, -0.023091131083172365679, 0.17247339417810558793, 0.051653722807011234752, -0.23874898221879006788, -0.41942889704876668011, 0.48671651702924578764, -0.11185663824444851133, 0.49018320941885018538, -0.4084400082490062811, -0.22222113034288801248, 0.054680268228579104184, 0.11070687047189697871, 0.093933896744501130249, 0.017362711618238567401, -0.070734645269038401438, 0.024401216160815888262, -0.23425550321283278299, -0.059449298307285527998, 0.80023473422098156949, 0.66921124472835658015, -2.422046820087983221, 0.67500245524089586002, 0.81992299024117287765, -0.022341963005033632639, -0.20527518571693600946, -0.089163324609646738006, 0.011742348231627678518, 0.027770335341873852303, 0.0049778678702089851349, 0.27677629372404577435, 0.081188143858132499897, -0.23559867672371964664, -0.69503709760328491463, 0.31218970711475391155, 0.48564343430712675342, 0.33259019916554599661, -0.65431584498912198633, -0.18872715440600873471, 0.088003364730667832383, 0.13977799882995156633, 0.14849286544785611497, 0.044224581603900747773, -0.12235278067727603513, -0.55670290613516770772, 0.0011939986373302180489, 1.0221417691290808349, 2.3065875622837435088, 0.10914828605813142737, -5.7792239494817971845, 0.070212517932573881296, 2.2578190738263250736, 1.0265169770082822609, 0.16448193954533318895, -0.16258631820757529773, -0.24361908777906360335, -0.18050291972161347509, -0.031101601703109479913, -0.21736615089980607096, 0.28257862315654103114, 0.32751846892396435651, -0.54333499497475068463, -0.057363826250714386457, 0.4228033129834937176, -0.058381132282047110138, -0.55662121205013337377, 0.29242438272555348888, 0.27214187363213382387, -0.046068576767137138506, -0.10911402681782234048, -0.047903500154992929849, 0.037044212945249316293, 0.0073482957375892426349, -0.28253783148836647543, -0.10498927782836504907, 1.0033573929623755827, 0.75430394577777648557, -2.757569937253720127, 0.74427107340982723294, 1.0019413399711487589, -0.079085289549443160695, -0.24560504641785993463, -0.078807183763707683299, 0.0099205308613712836485, 0.017621808890741989806, 0.0073069637621180670919, 0.0027488326405728383608, -0.16862699982667814225, -0.040431561889586214997, 0.69909256338189207369, 0.6045531081558043196, -2.2058202273396978477, 0.60086161248991465378, 0.70476893614077940065, -0.014748579931594037531, -0.13638304662700431735, -0.051411138414140536423, 0.0075113016593961710876, 0.0092238767437049611064, -0.011004084997963360301, -0.34155084866854651393, 0.030540699564122593401, 0.67916358123717313067, 1.6285274508738252219, 0.22981965464544970579, -4.4680736745240983154, 0.20198156287534563269, 1.5955075137804286545, 0.68953405301211256884, 0.1607368148560383847, -0.053678239306441777323, -0.14206028488567606916, -0.14063192686865064185, -0.062385029628816299208, -0.0096171988521371390424, -0.14653530207669113183, -0.15145495622152296833, 0.66228733690080021645, 0.75748942575193645155, -2.2424223595276626675, 0.76500014453351972143, 0.68241295671302004688, -0.12763707672048013597, -0.14982157019641015316, -0.072464601982118598267, -0.01636740817507458326, 0.022141321240199601794, 0.02553760126542652803, -0.41279098084172022309, 0.37275206401096167008, 0.6708667382596480433, -0.51908065033736372484, -0.35269839031075134184, 0.48771617126430810707, -0.35442122136393094944, -0.53773991778241381212, 0.62433441120527466417, 0.36619843934338047742, -0.15534583416747599349, -0.20714183685060733042, -0.067319595771405976725, 0.080480194994789353014, 0.33078941564008002008, -0.40664939317396742169, -0.62699598561853298051, 0.28737555611692205959, 0.99834242825784136333, -1.177293165083793447, 1.0098480152367537155, 0.32209217175280974166, -0.57389999698750759904, -0.42074542894788424796, -0.021217302394189896386, 0.16083038085786768856, 0.12194882966993886297, -0.0060482607825673509461, 0.16257914845931567549, -0.22193377468387481821, -0.42492490742002797521, -0.022016708931745676558, 0.79652223020843782031, -0.54982346676089022264, 0.78143665147252372449, -0.038893826941032827449, -0.42608083721282030698, -0.22779012362954531579, 0.0032727318603533173125, 0.089188946717826261823, 0.056211696245630125846, 0.015637855795242749241, 0.19069028256240933006, 0.59738619726180486325, 0.30323311104311523323, -2.0088390552293740043, -11.472909005173926644, 24.814222936573571587, -11.42931196752516243, -1.9990187442022850917, 0.19080095463623414354, 0.42653447797561749866, 0.24986176669430937292, 0.074599674971570439919, 0.024187248633792346464, 0.038825784494116832857, -0.10201640652596689895, 0.33052317405725789579, 0.16490018507738568587, -0.70511614088614804619, 0.032988063278118311128, 0.56769312544472838589, 0.028027844986821141837, -0.72371532946645700246, 0.12863710723746124565, 0.32805152473922283285, 0.088659240564074828783, -0.045751683863193171042, -0.067374311528942687488, -0.022590086259045769751, -0.0156590566518799712, 0.18363386497264425068, 0.022249961150901914619, -0.46142527156896034901, 0.19633277595470183252, 0.15473352877489482404, 0.1926446415221768349, -0.46953430507964860618, 0.0094972027863105303036, 0.18413240295222818577, 0.064055363383244190434, -0.0019730176544536081468, -0.026247211657921022676, -0.029730117878763322875, 0.24958334041572569073, -0.10712697271169263624, -0.44944464063180272761, -0.56202383415473844419, 0.047931420349636488787, 1.040973499323002649, -0.43960348334175591845, 1.0433737809570495791, 0.051343774296528255086, -0.56268918421901603644, -0.46680883027850450517, -0.17141655761789259538, 0.089732676586763759907, 0.22103695352818314435, -0.046164753584003431763, 0.025953710427989183651, 0.15048825834547563396, -0.054273634335172314669, -0.37148593067795043776, 0.3706185811719376888, -0.15095071914548438552, 0.37170850749835593341, -0.37074343823156591382, -0.056462920284259761339, 0.14563947232301358015, 0.024232176387677557827, -0.034577182121547603288, -0.0050412119099010549786, -0.0039565509723119745081, 0.0065307516800831274043, 0.018946312854941176756, -0.061204641363614989413, -0.055446558821851496535, 0.36244820900006224917, -0.53398249323121005361, 0.36115559876928876593, -0.05668545124551783998, -0.059690370353293269179, 0.022733408147180667047, 0.0030708121640031314087, -0.01087815684242908823, 0.0066522315952269334832, 0.20553786945159902011, 0.038034878653817320859, -0.19258692450509226624, -0.47710779542731429625, -0.36756084502047869922, 0.73392650818225191323, 0.11751193912242804962, 0.73817759989650277852, -0.35794305536197140727, -0.46386364051312345413, -0.18606228731078705563, 0.011699289231538345657, 0.11020925776676228358, 0.095566650895503049701, 0.039237176020789105158, -0.021425107733427865842, -0.050580247867919446436, -0.12679782976376990922, 0.017852954099811867844, 0.51909042947274697255, -0.75790810158263943741, 0.52104855860469623252, 0.021828137877215741947, -0.1229453521603395133, -0.051281520834132537923, -0.034230283830448543503, 0.0047943039890404413189, 0.04008266356235510447, -0.037441676246867325317, 0.12564727685908835819, 0.091832305871413791309, -0.30108128855071220054, -0.27058703846331644272, 0.65626483587871675685, -0.52434877375881827621, 0.6578384680250631833, -0.27205494889505182909, -0.3118495204184920655, 0.074115216127995001139, 0.12553235211297517071, 0.039334262563370914201, -0.052716585902177558187, -0.037558975640835295939, 0.11788344264526123828, 0.05662716545720752892, -0.20885116607251708043, -0.18556629979122665763, 0.49493201121571517298, -0.47202627548845521588, 0.49622085311882901948, -0.18533283474725628648, -0.21447791456677273159, 0.042469761779532808066, 0.11196889519584148076, 0.040667327111313406052, -0.055587589093405748164, 0.043998745958559834501, 0.095011361625375961859, 0.055587283933790872237, -0.21899258833622242681, -0.40764253467176092549, 0.49018320927382852403, -0.11094990490026887531, 0.4901664682592775657, -0.40844000802803814842, -0.22071239089008479195, 0.058452209081295822268, 0.1126593566504156263, 0.077344859857128231417, -0.048272176630923376017, 0.079004952783926926663, -0.089676097080249234095, -0.23061408988235926332, -0.041311951071995055285, 0.81273706542243651096, 0.67500245536793046508, -2.418963061482199528, 0.6769042151465571866, 0.819922990422708442, -0.026426137952226681771, -0.21417131980991813145, -0.098028829902284952968, 0.012473496840398428515, 0.045373967031631538827, 0.12001492346360309593, 0.097337330352404660694, 0.062885180251031291832, -0.1996604881971820733, -0.65853841393460932352, 0.33259019919755128392, 0.49254871107676179154, 0.33281377349242013963, -0.6543158454063597862, -0.18689421643107784199, 0.087547037112369108236, 0.12722533973617411451, 0.10836423750320783865, -0.044002213297997173902, -0.39518996222638907989, -0.23590834609933852839, 0.14530092473938538422, 1.0412555860690222431, 2.2765326495331965617, 0.070212518397114287438, -5.8078022641873907617, 0.057442340524189505757, 2.2578190733646530397, 1.0314009503273657842, 0.1656060733257807005, -0.17006774881830188928, -0.25732768109338971518, -0.18754656104168071495, -0.11449413722393474158, -0.054275005365344089481, 0.27429171788663470055, 0.29576959655122625481, -0.55564909508066739896, -0.058381132183461581509, 0.42328457072393083083, -0.058956212748200689799, -0.55662121201180436714, 0.29298766234195583458, 0.27022172370545599618, -0.051672987395431267676, -0.10688955617702856904, -0.0040850850483327965507, 0.035841878641368986325, -0.072010410729265478258, -0.25059525562184009617, -0.079973032866952301068, 1.0043790686390774436, 0.74427107346393373, -2.7646368026642820226, 0.74251690421930915953, 1.0019413399307992574, -0.079677638304981521222, -0.24486791411197339485, -0.072955541668139223255, 0.018164477770668454454, 0.014653360333918659039, 0.023074998741260507179, -0.058794862066983678839, -0.14858217485370039945, -0.018265854836000780526, 0.70648328421394723797, 0.6008616127483287217, -2.2109328662554919021, 0.59846793139932030758, 0.70476893593582368158, -0.01419206893961121041, -0.13649953251193097659, -0.052373389392587822722, 0.0044452745786907428213, 0.0017224086266652499991, -0.29728001861560698904, -0.11696496811060941301, 0.14688168904546028415, 0.70010314845274590478, 1.6080030012223474056, 0.20198156314207477147, -4.4873081231289608795, 0.19403040828557446829, 1.5955075139488568148, 0.69213860922920145136, 0.16103439604215979974, -0.059499117609731491663, -0.15989416889144872358, -0.18019540381335949486, 0.05912615700938972374, -0.069483786458100271766, -0.17300696271880047772, -0.14235818096361357865, 0.67923147881557488503, 0.76500014471377764114, -2.242678656810288107, 0.76320166865382987886, 0.68241295615274710773, -0.12763756896745545588, -0.15432846127442217954, -0.082260613867413401312, -0.019601214525088447166, 0.062010066081422489381, -0.1684072094699338995, -0.17111545732113589424, 0.35913055281594019608, 0.62455817131962931832, -0.53679218655624294509, -0.35442122114294771773, 0.48747828102895396363, -0.35580855630199925699, -0.53773991770334494955, 0.62449897857774050447, 0.36054202891330250358, -0.16766798666831056464, -0.20045446714140971456, 0.027343463324538276604, 0.27405261660739860252, 0.008842738133895217123, -0.43156739120617976235, -0.59017460715295122498, 0.3129806002039000612, 1.0098480149904389691, -1.169983831992899681, 1.0144800474643307364, 0.32209217155099079122, -0.578056767003428007, -0.42378640867162137296, -0.016237708795307614906, 0.16896908000749519263, 0.10371402844206258076, 0.096353991319636650426, 0.042797862446775331025, -0.19402990846541856573, -0.41030110715456641346, -0.035565010603080926543, 0.78143665144334650829, -0.55331153870734595124, 0.78359486435975722607, -0.038893827046417639437, -0.42723147361972424463, -0.21968112311739385056, 0.029713149058553600668, 0.11717216986962215863, 0.022747619861562859422, 0.071968765424380298779, 0.29563682148469644373, 0.45715825543937160802, 0.194841749842954437, -2.0104800890625247334, -11.429311967524386162, 24.847411346847923852, -11.418633216653081064, -1.999018743873114623, 0.19147082863315501022, 0.42733846865460184405, 0.24274272497493815104, 0.066871734766534587924, 0.06860627686211151488, -0.14777484576916830039, 0.072755832654338672971, 0.34328137637533018278, 0.14180399649170233323, -0.71924043219526023751, 0.028027845358131846243, 0.56613253736151514239, 0.026811767760992241683, -0.72371532976792307679, 0.13104064134779097595, 0.33163182994675183268, 0.089200773840996408492, -0.053916642886648602495, -0.08615939383053065892, -0.063100451854348788006, 0.049830140032039130227, 0.19296028513284077532, 0.019140229146869476573, -0.46579291738594508221, 0.19264464136329353949, 0.15279305640574919467, 0.19156833589229257941, -0.4695343047720577645, 0.011482970063304027281, 0.18804328966692210323, 0.066547274167659126176, -0.0099904458261435534483, -0.054990226726506982735, 0.22103695304850712322, 0.089732677675323682664, -0.17141655806170827403, -0.4668088306069629323, -0.56268918402537371204, 0.051343774180763128434, 1.0433737810796093193, -0.43960348371773005915, 1.0409734997847608362, 0.04793142019850267499, -0.56202383348136664143, -0.4494446415498258296, -0.10712697260770785712, 0.24958334058976966441, -0.0050412118747485525419, -0.034577182097544609252, 0.02423217614622638974, 0.14563947271838978414, -0.056462920737989370779, -0.37074343773728113405, 0.37170850702327545845, -0.15095071875166585218, 0.3706185809032406242, -0.37148593054546974468, -0.054273634374994078611, 0.1504882583049666489, 0.025953710522482655726, -0.046164753633979843661, 0.0066522315244164961084, -0.010878156592930184055, 0.0030708116493693830319, 0.022733408969763450186, -0.059690371273931584761, -0.056685450498331589608, 0.36115559814478170875, -0.53398249264842700956, 0.36244820861921367428, -0.055446558787389001177, -0.061204641068403765891, 0.01894631239687254165, 0.0065307520165545180843, -0.0039565510714949644541, 0.095566651241601602296, 0.11020925699265946007, 0.011699289077951299798, -0.18606228617083814258, -0.46386364068112861947, -0.35794305559050898502, 0.7381775988488122886, 0.11751193999071155027, 0.73392650877809373355, -0.36756084503736036195, -0.47710779672841518995, -0.19258692428677490205, 0.03803487971610129692, 0.20553786890822761668, 0.04008266379378782468, 0.004794303492667371068, -0.034230283501119229683, -0.051281521030234644698, -0.12294535182210332425, 0.021828137510835378399, 0.5210485587366472382, -0.75790810120351681523, 0.51909042882014677645, 0.017852954685595567169, -0.12679783053336801668, -0.050580247110143181088, -0.021425108002006885977, 0.039237176017696676067, -0.052716585825994505266, 0.039334262414322779255, 0.12553235205922311235, 0.074115216423605329021, -0.31184952069361582261, -0.27205494870740343227, 0.65783846801008394323, -0.52434877401274959663, 0.65626483628835308171, -0.27058703889511015461, -0.30108128815928575372, 0.091832305503864883334, 0.12564727713622256267, -0.037441676338401667923, -0.055587589056845111701, 0.040667327147795383213, 0.11196889486611427689, 0.042469762308475961232, -0.21447791507366090191, -0.18533283428808927651, 0.49622085276501898266, -0.47202627540456087996, 0.49493201140206838362, -0.1855663000216268288, -0.20885116592544775149, 0.056627165318291032559, 0.11788344277262274851, -0.03755897568655518215, -0.048272176574676507743, 0.07734485961350179295, 0.11265935686270317495, 0.058452209142922470497, -0.22071239094138581094, -0.40844000810478942043, 0.49016646813842512653, -0.11094990457831764574, 0.49018320922435398845, -0.40764253493804930573, -0.21899258811599534025, 0.055587283828851558054, 0.095011361637679023207, 0.043998745986187595691, 0.045373967051205971179, 0.012473496755730317451, -0.098028829713392204304, -0.21417132022006446501, -0.026426137506571412139, 0.8199229901878744009, 0.67690421537006706387, -2.4189630615318042928, 0.67500245516369483667, 0.81273706541067813891, -0.041311950990340226553, -0.23061408971179470084, -0.089676097259963769015, 0.079004952827432334272, -0.044002213464045793201, 0.10836423796947444731, 0.12722533957838211283, 0.087547036826793434017, -0.18689421657999993887, -0.65431584499778616681, 0.33281377362109165752, 0.49254871064802918657, 0.33259019933462652396, -0.65853841412733049587, -0.19966048760613508417, 0.062885179901395527047, 0.097337330323625625983, 0.12001492348904159446, -0.18754656110946432701, -0.25732768107640013877, -0.17006774859328654959, 0.16560607296806395539, 1.0314009504914709581, 2.2578190734081817759, 0.057442340758715904059, -5.8078022646548514984, 0.070212518643451682365, 2.2765326493010502595, 1.0412555864347867729, 0.14530092449655740428, -0.23590834598711568626, -0.39518996227087688222, -0.0040850850171083039411, -0.10688955626866751547, -0.051672987182933935446, 0.27022172341593037048, 0.29298766248693686176, -0.55662121207136427969, -0.058956212640590768825, 0.42328457083516479731, -0.058381132539627358935, -0.55564909486870694977, 0.29576959637698763128, 0.27429171842654559432, -0.054275005979011026191, -0.11449413699697358338, 0.014653360327072244979, 0.018164477786700473916, -0.072955541647807903294, -0.24486791404432570185, -0.079677638539365047698, 1.0019413399177419244, 0.7425169045071864371, -2.7646368025150942493, 0.74427107281134408101, 1.0043790691251621716, -0.079973033144892041735, -0.25059525513805858132, -0.072010411235154864196, 0.03584187882240490175, 0.0017224085093094343955, 0.0044452747566351164671, -0.05237338901567257371, -0.13649953336984788077, -0.01419206873136200607, 0.70476893638184157087, 0.59846793131602449289, -2.2109328663109684143, 0.6008616121729621895, 0.70648328493102563463, -0.018265855227877082767, -0.14858217429116554742, -0.058794862702710504188, 0.023074998969612629113, -0.18019540356894472866, -0.1598941693182369117, -0.059499117484494962182, 0.16103439594528076717, 0.69213860958694251035, 1.5955075136080112408, 0.1940304086172274789, -4.4873081232858078593, 0.20198156306770742585, 1.6080030010710895105, 0.70010314857864863747, 0.14688168914380905772, -0.11696496800634964552, -0.29728001875760950989, 0.062010065890997764781, -0.019601213974166073473, -0.082260614549371974258, -0.15432846060712157588, -0.12763756962161984521, 0.68241295660477818696, 0.76320166841825298754, -2.2426786566998613281, 0.76500014479191491557, 0.67923147854209131591, -0.14235818060483232395, -0.173006962958839966, -0.0694837864166538699, 0.059126157025433195846, 0.027343463375175257324, -0.20045446717234871592, -0.1676679870479562684, 0.36054202945607910502, 0.624498978768615709, -0.53773991861503356304, -0.35580855553309764794, 0.48747828079912491894, -0.3544212208453022539, -0.53679218709402642951, 0.62455817145724301653, 0.359130552749229226, -0.17111545693464719209, -0.1684072096900891824, 0.10371402872413654217, 0.16896907938090771628, -0.016237708215113000332, -0.42378640928654992548, -0.57805676636483294129, 0.32209217135666279308, 1.0144800474998398876, -1.1699838325384757098, 1.0098480158450278132, 0.31298059980725867479, -0.59017460745509475739, -0.43156739077554562289, 0.0088427380191255511299, 0.27405261658261292901, 0.022747620054980002835, 0.11717216918861060593, 0.029713149817132522568, -0.21968112347404816775, -0.42723147329741184564, -0.038893827378174261611, 0.78359486449509119232, -0.55331153911279151725, 0.78143665213458635144, -0.03556501080832563666, -0.41030110734554597895, -0.19402990848921303213, 0.042797862516132281763, 0.096353991346071213298, 0.068606276741762395321, 0.066871735025855166823, 0.24274272477600206122, 0.42733846878153480819, 0.19147082886008537361, -1.9990187445895797325, -11.418633216266508512, 24.847411346965539991, -11.429311967520360938, -2.0104800890972414074, 0.19484174986135938124, 0.45715825520868536014, 0.29563682167326804651, 0.071968765398498321662, -0.086159394145872342241, -0.053916642100822545491, 0.089200773376145001126, 0.3316318299025002303, 0.13104064107997334676, -0.72371532944546823618, 0.026811768130531866278, 0.56613253693655163357, 0.028027845036542146928, -0.71924043201790988089, 0.14180399719780750756, 0.34328137584287005124, 0.072755832517467020071, -0.14777484562258200174, -0.054990226819603693631, -0.0099904457608058685331, 0.066547274446161361339, 0.18804328930892175165, 0.011482970050293746928, -0.46953430440887966579, 0.19156833541377787933, 0.15279305677417734377, 0.19264464088406793585, -0.46579291660985433499, 0.019140228498875110075, 0.19296028561377850941, 0.049830139566023064213, -0.063100451650728209274, 0.1454416861027962582, 0.1325228727757448155, 0.03742771091786233767, -0.19189870656134414006, -0.46927733925655878977, -0.56060272109567033105, 0.051343774845731569856, 1.0409734991890062794, -0.44127776763010406746, 1.0423263666261162186, 0.049346353191030048468, -0.55702211044528271433, -0.40338958085431275702, 0.10478782612298759291, 0.01222729140825059603, -0.019928797244170227271, -0.033979158918307786341, 0.025301274055445764632, 0.14627372966776078234, -0.055956674730813546481, -0.37074343740036946881, 0.3706185806296322105, -0.1528805796965005126, 0.36989165008540231261, -0.36786121383017511555, -0.038432842304272546441, 0.16817441196768087575, -0.052987562223583022092, 0.0059014229553691422217, -0.003254195728350597807, -0.0069756035343628303852, 0.0047338310531496010161, 0.021436267276782869362, -0.06097551860045367228, -0.056685450427568326304, 0.36244820864830873397, -0.53264636736287152008, 0.36082692987356257408, -0.060189335623038006662, -0.057330774345669174852, 0.030694624419904028428, -0.0082789208445323177188, 0.0056484544873526918579, 0.089000140783985326465, 0.10936922499100706907, 0.01444467101532449782, -0.18219370326670142335, -0.46121053853322069305, -0.35794305567793616163, 0.73392650845968687712, 0.10693307330200706173, 0.71553021175546449584, -0.39159866964436829129, -0.48632185979928566066, -0.13864641872826016789, 0.2426412252645149259, 0.024669571832468168565, 0.017546634833366769879, -0.002325367591592472985, -0.033317645307535503907, -0.047919323619765717692, -0.12082086015795491518, 0.021828138039291802852, 0.51909042894442647409, -0.76122958178312505328, 0.51601986507252262104, 0.016201992708350439198, -0.12297867782676662196, -0.036146661614857023004, 0.0072853615068387720874, -0.021258458819797518974, -0.017933950321557045465, 0.036857151584821869206, 0.11543160944506827836, 0.070007069322930520761, -0.31159961676821573695, -0.27205494869392188306, 0.65626483585962691603, -0.52427466829958868644, 0.66246984115597706833, -0.26106050430970695331, -0.29620467547149120779, 0.084206764287331953289, 0.080118222385750781367, -0.023091131161181981807, -0.025632013607653231918, 0.040056698640894644603, 0.10693378173249289131, 0.040661909959975567341, -0.21419852898537328212, -0.18533283466294461772, 0.49493201129869768273, -0.47329204764019872087, 0.49716495348724082382, -0.17767530583364207764, -0.19958157854350333515, 0.042112851879926693766, 0.078780298833908038447, -0.070734645435017381998, 0.017362712010956696251, 0.09393389636861801395, 0.11070687081926459949, 0.054680267994578410362, -0.22222113017241651445, -0.40844000835396176985, 0.4901832092262065621, -0.11185663820374160127, 0.48671651716857194936, -0.41942889709728409242, -0.23874898216722806321, 0.051653722897537723835, 0.17247339407096090591, 0.0049778677860722528314, 0.027770335589989817576, 0.011742347950758763586, -0.089163324361211507285, -0.20527518585018422748, -0.02234196292625811181, 0.81992299015311143062, 0.67500245512517753621, -2.4220468199616465022, 0.66921124470881987456, 0.80023473418995016981, -0.059449298248663837163, -0.23425550310984608049, 0.024401216053082909474, -0.12235278071882774165, 0.044224581666998746821, 0.14849286544411521849, 0.13977799906724619472, 0.088003364440820908143, -0.18872715427007991096, -0.65431584530079334172, 0.33259019952601009962, 0.48564343390529940425, 0.31218970733119455474, -0.69503709761858356586, -0.23559867651784233145, 0.08118814379842867468, 0.27677629363380740202, -0.031101601742655006488, -0.18050291957817465871, -0.2436190880452978047, -0.16258631760350572559, 0.16448193888138828989, 1.0265169774895592791, 2.257819073164821333, 0.070212518679759972073, -5.7792239502720956779, 0.10914828675793009083, 2.3065875618478317577, 1.022141769531728972, 0.0011939983766776074425, -0.55670290608729811055, 0.037044213016136917638, -0.04790350023655573497, -0.10911402681457817165, -0.046068576993012512466, 0.27214187403829975809, 0.29242438252782065877, -0.55662121201373371271, -0.058381132312715119936, 0.42280331315858288299, -0.057363826466071216048, -0.54333499500323945153, 0.32751846895460323683, 0.28257862330392402583, -0.21736615099129186213, 0.0073069637404394507246, 0.017621808988865155648, 0.0099205308399737696795, -0.078807184059387830333, -0.24560504613811479091, -0.079085289570582223417, 1.0019413400586993923, 0.74427107324397090427, -2.7575699371054360753, 0.75430394563671510078, 1.0033573929991055351, -0.10498927787444281567, -0.28253783136633858764, 0.0073482956763950444887, -0.011004084886154523806, 0.0092238763658767930487, 0.0075113021633084944145, -0.051411138823550500687, -0.13638304611309912118, -0.014748580594775547239, 0.70476893659349626464, 0.60086161216611322367, -2.2058202269391089523, 0.60455310787926030969, 0.69909256347301340639, -0.040431561943765487177, -0.16862699977911024818, 0.0027488326245682093152, -0.06238502970611405074, -0.14063192662737519778, -0.14206028498976303043, -0.053678239377547343203, 0.16073681485125540513, 0.68953405311349325046, 1.5955075136679144343, 0.20198156290218735598, -4.4680736746780409518, 0.22981965491539091406, 1.6285274506054310262, 0.67916358145303712401, 0.030540699595038571512, -0.34155084875499852615, 0.025537601294030053828, 0.022141321225613883744, -0.016367408243522722777, -0.072464601949430385019, -0.14982157017828837131, -0.12763707658707240666, 0.68241295646743660441, 0.76500014468705646031, -2.2424223595007615195, 0.75748942548650965989, 0.6622873373226998428, -0.1514549566216254195, -0.14653530180133544913, -0.0096171989489890249286, 0.08048019485584149979, -0.067319595441291929516, -0.20714183719603160294, -0.15534583372122684475, 0.36619843859472056558, 0.62433441209631057589, -0.53773991873202742209, -0.35442122026223821241, 0.48771617012607981767, -0.35269838933035191353, -0.51908065083988785826, 0.67086673829227883026, 0.3727520639872777819, -0.41279098077687409463, -0.0060482608688130767849, 0.12194882989582586652, 0.16083038061739368674, -0.021217302152707617768, -0.42074542945828019569, -0.57389999613836228942, 0.3220921708697598862, 1.009848015877538252, -1.1772931654739138274, 0.99834242861392585855, 0.28737555578635498588, -0.62699598549240143086, -0.40664939316204584685, 0.33078941562762043116, 0.015637855764344732457, 0.056211696275348541074, 0.089188946687896120125, 0.0032727319287292333527, -0.22779012394747849557, -0.42608083666281648672, -0.038893827450303675164, 0.78143665200811507354, -0.54982346723980257686, 0.79652223055977300348, -0.022016709149694779768, -0.42492490728971527014, -0.22193377493191981942, 0.16257914862243566523, 0.038825784401543494972, 0.024187248803088645666, 0.074599674856135722578, 0.24986176666382975409, 0.42653447800573823789, 0.19080095475199509991, -1.9990187443968707726, -11.429311967254875526, 24.814222936361481686, -11.472909005091901591, -2.0088390551071992896, 0.30323311070748015794, 0.59738619747613486144, 0.19069028253291975261, -0.022590086211899295526, -0.067374311614311122032, -0.045751683881269808452, 0.088659240699961949783, 0.32805152470269349774, 0.12863710719218707235, -0.72371532961335127698, 0.028027845235523519973, 0.56769312534882965249, 0.032988063290361815949, -0.70511614093071184328, 0.16490018506236470142, 0.33052317416282112061, -0.10201640658802338391, -0.029730117786117820494, -0.026247211874874498694, -0.0019730172913299832277, 0.064055362649563934996, 0.18413240400269106911, 0.0094972018483240119857, -0.46953430434694720663, 0.19264464084440713343, 0.15473352943084708877, 0.1963327752460700959, -0.46142527102543456907, 0.022249960966839364696, 0.18363386505199694176, -0.015659056709590515877, 0.07372826792738132462, 0.10513172747087792813, 0.10351516091424008137, 0.024108005134472570291, -0.19470249776829279265, -0.46927733893508594143, -0.56268918540385315374, 0.047931421026243190009, 1.0423263670311624374, -0.43131471400087600943, 1.0561306098510609974, 0.051163274105011938286, -0.54086862382725930942, -0.32710542238983147945, 0.010733556090143923575, -0.00025657758145969161365, -0.021389176824684417305, -0.032016709983606880807, 0.02649526864810315685, 0.14627372958931902924, -0.056462920773253530426, -0.37148593046557254471, 0.36989165038794968865, -0.15267819536673843395, 0.37633190448817144258, -0.35492512205314263873, -0.02757347100832747977, 0.087821225488884532839, -1.5097579342341889506e-05, 0.0064492449024578868758, -0.0023642922590653044079, -0.0082124810866550223415, 0.0034997140050185416953, 0.021436266712436889692, -0.059690370533994202562, -0.055446559404664044757, 0.36082693061667031609, -0.53672244430339899335, 0.36436684203434177576, -0.053055357190500258835, -0.05812927174568931904, 0.016837403067745598112, -0.037483364172086842958, 0.040007432480026673682, 0.10127733964409699197, 0.11181591209652329277, 0.014622772881945314549, -0.18219370255138067627, -0.46386363926715673545, -0.36756084652417569947, 0.71553021114269577119, 0.082804431150207039036, 0.69562015311822800179, -0.39418013110051808301, -0.41049777491231109749, 0.084673641608888169929, 0.0073111056408606787949, 0.017388245975675217792, 0.016192132960490764754, -0.00074923431817211158523, -0.031852746094302372692, -0.047919323155468415698, -0.12294535207256897247, 0.017852954244694563962, 0.51601986498518725988, -0.76018846601603029534, 0.52274003192418061481, 0.021539468815150085435, -0.11388186906429713374, -0.044353930135309853344, 0.0057292471072136709129, -0.025263726447960835331, -0.020905823232167358716, 0.036250786299851338323, 0.11575730984048523964, 0.070007069590860906971, -0.31184952077602035025, -0.27058703823318830395, 0.66246984066270497848, -0.51621071437575427598, 0.65828671834036867061, -0.29951696695924523439, -0.33238585347021804006, 0.22864039778050351814, 0.0010158817401481467103, -0.027582784458897533636, -0.023594379369322491385, 0.041660230022743958977, 0.10739331061298668102, 0.040661909608482725564, -0.21447791434372157071, -0.18556630006711191672, 0.49716495358763407308, -0.46625560575887792236, 0.50051320917208608208, -0.21463234566765243549, -0.24681552336703171946, 0.19173868330786428271, -0.049882146411034659905, -0.025527345118670205504, 0.029883861354787474851, 0.094275451370631091486, 0.10905850563047282675, 0.05468026785649771071, -0.22071239074094084476, -0.40764253484250473347, 0.48671651689805489616, -0.12684293497391291305, 0.45371932079336046995, -0.45865627662868369008, -0.19207933393229686958, 0.25436585648436949025, -0.018084734610075953176, 0.015933871472079391696, 0.034364946109206151448, 0.01650040546783829859, -0.08616096876804173188, -0.20527518570810410203, -0.026426137952775693996, 0.81273706575780646499, 0.66921124448085045344, -2.4218031606579057957, 0.67756132265585933272, 0.80161768815695511048, -0.068319146469304009583, -0.20258385475033033329, -0.11454333224792369317, -0.026851990481297993707, 0.079359742527649046462, 0.15483642928266772021, 0.13787668533612340571, 0.088003364332753575283, -0.18689421679826268008, -0.65853841342984531515, 0.31218970722791006267, 0.44237537317657793512, 0.25380398924497582502, -0.72310417308794461277, -0.10389082654523940086, 0.34807594027031185657, 0.048135796899245061886, -0.079863049625607696047, -0.1979963975136745935, -0.24708581596839473971, -0.16284257726340289896, 0.16448193893698959123, 1.0314009501761811638, 2.2765326496433280212, 0.1091482864597435043, -5.7249811672197701284, 0.16337683100618016763, 2.3039403354594845119, 0.82774679516537785773, -0.48756561857329089316, 0.032956350596784367291, 0.0021455569097485176905, -0.053896034575616535989, -0.1067175923653247982, -0.044801515552467738313, 0.27214187360173036634, 0.29298766219506389152, -0.55564909480850743773, -0.057363826094204289907, 0.42819543683444977189, -0.036635480960401638206, -0.5046148965346325399, 0.35908205661159176003, -0.024564348031823880075, 0.0095715827911336849587, 0.0063664208679130550317, 0.0094252716531659092025, 0.0049562430625934399131, -0.079452728893006668853, -0.24560504644070152436, -0.079677638534448147478, 1.0043790688459797167, 0.75430394608735196194, -2.7434025262625234554, 0.76135847677881951512, 0.97737358842899013656, -0.16625140302633040212, -0.2143149304297209401, -0.0097863683109866133475, 0.00080662187090490330282, 0.0094884049104635610761, 0.0052227268610431911919, -0.052513426231320228799, -0.13638304634713493968, -0.014192069538053646641, 0.7064832850862493574, 0.60455310827017672004, -2.2024484741907643937, 0.60510715374758405893, 0.68539789668761186281, -0.066122540347083161305, -0.1345979735835199087, 0.01246863259838968882, -0.08176569642354686307, -0.1465044405462831989, -0.14156172072801664896, -0.052600052928436419331, 0.16073681509583029325, 0.69213860941332816523, 1.6080030013940822542, 0.22981965469738374619, -4.430496454062721412, 0.25942402679041853819, 1.6050340268830902701, 0.53683688775825477713, -0.23104895771704234142, -0.0077370640928009631887, 0.031059709005933546638, 0.026847192116114169252, -0.015761946468870476068, -0.073353278011603545439, -0.14982157015762198604, -0.12763756965562850221, 0.67923147903191116725, 0.75748942543529074189, -2.248408332589983516, 0.76954384389493035723, 0.68734750648472031642, -0.1653663289928988811, -0.16543973080552984434, 0.059312619893779561231, 0.019458194293302186195, -0.079593143623467588199, -0.20361729251947852104, -0.15276051971507120797, 0.36619843907282179085, 0.6244989795644616537, -0.5367921877702811484, -0.35269838983536738741, 0.49785885250489908538, -0.31298744778930331911, -0.43670207113016856759, 0.72433421044864410021, -0.21165918325056320981, -0.042542739659913147987, 0.040149967493964974774, 0.13000179244935250455, 0.15984279143697974201, -0.020979838918938231718, -0.4207454293876748963, -0.57805676589079835193, 0.31298059983175380294, 0.99834242816228668804, -1.1932755508959103086, 0.97062431019319017622, 0.24947249244895960962, -0.58499487192855126327, -0.032384963186152672387, 0.030193283475313224123, 0.0074221885163681527997, 0.036763646943353628371, 0.079148413127732075711, 0.0019510505569331759723, -0.2277901238954781471, -0.42723147307367848047, -0.035565010825295395591, 0.79652223012540568359, -0.5320267254910500121, 0.78351823076965443349, -0.090849512226424122829, -0.44811577881574471416, 0.017578910436945597434, -0.013678271862765187944, 0.037085272655164425171, 0.045086344941479006287, 0.085805691887749446778, 0.25076949840358664368, 0.42653447832632784831, 0.19147082881627275341, -2.010480089381251112, -11.472909005104044766, 24.741411755126939909, -11.497617586305798554, -1.8844574900533552775, 0.49431855546743264096, 0.59501840719158383131, 0.010454604169902752611, -0.037135924028686734255, -0.067861570845276733555, -0.044108272817578106362, 0.088489047875982856928, 0.32805152432969364451, 0.13104064093964401527, -0.71924043196271092437, 0.032988063778898943368, 0.57232097277896443188, 0.036599219096441745758, -0.7096243094250138661, 0.16851341593942717467, 0.2158322243799956075, -0.0061012208786632542157, -0.024344800292532841474, -0.025399191730421884217, -0.0024515945906719502356, 0.063133469826841981387, 0.18413240324466431663, 0.011482969936696504204, -0.4657929171017329284, 0.196332776092226019, 0.15376317060985192353, 0.18668430814330916334, -0.47505400058399321894, 0.042692626209487515021, 0.16438745859975165131, 0.018464045065315348726, 0.061690365355313171725, 0.095649674427188774439, 0.09867546516336550555, 0.024108004266519239556, -0.1918987076814369086, -0.46680882959199648807, -0.56202383393034816006, 0.049346353895483252572, 1.0561306101347962549, -0.3958404313651184836, 1.118448205585455435, 0.010302253121448824019, -0.9291353607401291459, 0.0045778502556789429229, 0.0057085266251220281516, -0.0010887738472384039107, -0.020614283048652572161, -0.032016710332564674024, 0.025301274264756461879, 0.14563947254635153938, -0.054273634322765863802, -0.3678612141183534856, 0.37633190444749348158, -0.1428396885108712544, 0.36148513470945725645, -0.47629652887945489459, 0.17658502604898193566, -0.0014095051412755640798, 0.0034520516854488255548, 0.0048782169346888015066, -0.0037864684157662724037, -0.0082124814921841655935, 0.0047338320220287463322, 0.022733407877386625773, -0.061204640321445222262, -0.060189336654226904821, 0.36436684236591787567, -0.52422528966581916077, 0.36451895296263148172, -0.10040820689245620601, -0.0053880342910670019768, -0.042524020182985018679, 0.00041041904812980437023, 0.054235467822386577985, 0.10397611656121337476, 0.11181591286892172965, 0.014444672137457315042, -0.18606228674933289469, -0.4771077959946171676, -0.39159867072808429711, 0.69562015245987385281, 0.090622639686776806922, 0.75111789473543555395, -0.36569289118285336349, -0.37458095535025465184, -0.0042092427434445448334, 0.010436567839884095324, 0.019685697340069346212, 0.017277778967179517611, -0.00074923438255697014099, -0.033317645136301130004, -0.051281520898998560198, -0.1267978300061435859, 0.016201992688370324414, 0.52274003141241920378, -0.74328953465530667177, 0.54509540343832818543, -0.024909228216559966412, -0.14890422392582247646, 0.012995281019279741885, -0.010324697173775404524, -0.026819077541903733886, -0.019149901446607810185, 0.036250786693727152232, 0.11543160906167378166, 0.07411521670287331276, -0.30108128900736663702, -0.26106050372946221216, 0.65828671808239702656, -0.54713044416904887957, 0.59809781816238893892, -0.36804996585047428104, 0.03624551395169418877, 0.0069675279265941731863, -0.0093421435798915295617, -0.026796773536248919695, -0.021658938932652551518, 0.041660230055454050557, 0.10693378197718574063, 0.042469761633211457197, -0.20885116576966075708, -0.17767530611831841569, 0.50051320942686061688, -0.49728270596428486572, 0.4347201681282081287, -0.23120559325714387655, 0.03807435046496430997, -0.020614629354824373642, -0.033067849214693890991, -0.018157995571518432965, 0.031445675906697126423, 0.094275451827264628024, 0.11070687090616360138, 0.058452209226356993677, -0.21899258842663299474, -0.41942889710983022322, 0.45371932094951905601, -0.16755593366929846799, 0.45856945605378213182, -0.41878582826035370967, 0.083244979739019411835, -0.021976489233276854524, 0.0014380087026171391174, 0.024436790162914834301, 0.036946468681000842194, 0.016500405974589960056, -0.089163324576265995391, -0.21417131994038868248, -0.041311951225607795868, 0.80023473448013604514, 0.67756132259195445133, -2.3876330527163411865, 0.75073313221652049343, 0.85150561705419647307, -0.40483801050681705869, -0.06563413737735254716, -0.055538861561047536686, -0.0049993102463871674779, 0.082867634717786575749, 0.15483642949842280068, 0.13977799859579098651, 0.087547037000873073609, -0.19966048830474492082, -0.69503709722480144784, 0.25380398946013771333, 0.40138441649096839869, 0.31429001028126624107, -0.61041134602588764668, 0.18540617543354032115, 0.069244014843473783438, -0.0093115591115972562442, -0.10568959245745118691, -0.2049062866897405355, -0.247085815932215902, -0.16258631796008904402, 0.16560607317931774762, 1.0412555861061068008, 2.3065875623721305843, 0.16337683102956834791, -5.6861758249107063889, 0.16329349324502506757, 2.2657282889467227349, 0.28030214453118623696, 0.016742566987587317873, 0.018497518667845491702, -0.0035233839708880901331, -0.054981308388110106122, -0.1067175929247375088, -0.046068576838947022933, 0.27022172346822448397, 0.29576959669996683022, -0.54333499504579318984, -0.036635480767269426594, 0.44726651764565461544, -0.043824356081491641635, -0.65530018622172658826, 0.44857915531163455869, 0.010412533853382085341, 0.0018427362628961889597, 0.0016294398420588695683, 0.0077990758936523005285, 0.0049562425873853378777, -0.078807183857207710465, -0.24486791424503806525, -0.079973032694529599951, 1.0033573927184367136, 0.76135847713037285711, -2.7292213861769356598, 0.79795610505446179506, 1.0261369010575207561, -0.47923953010371272976, -0.0022735259032248236917, -0.0051691960507880430759, -0.00014824133938223446605, 0.00796901807170625337, 0.0052227261976415702283, -0.051411138829581093379, -0.13649953302928549692, -0.018265854745675338955, 0.69909256361004001867, 0.60510715369509537886, -2.1880719112114861424, 0.64752414464449026266, 0.71862870497984399076, -0.2788183064367780184, 0.04227851602753952831, -0.025162507506936993007, -0.095615860799894869526, -0.1488158489763653225, -0.14156172029521829625, -0.05367823945781943773, 0.16103439633209906678, 0.70010314815795049448, 1.6285274508051912346, 0.25942402656390817217, -4.4231644160294703028, 0.24771544838021086754, 1.6021347805327614822, 0.27470519795578601796, -0.017174813308188176947, 0.012607286409839876112, 0.033686645859670537317, 0.024892763682834630895, -0.015761946324635889988, -0.07246460233350920932, -0.15432846049621784523, -0.14235818120199478232, 0.66228733748853274577, 0.76954384410517606696, -2.1909506244301506506, 0.829278577591095023, 0.63756128708035786357, -0.37672378497508407502, 0.025944763142471936185, 0.038723752189523522049, 0.0094230733568868851879, -0.082103840232390035148, -0.20361729183334018045, -0.15534583382872485524, 0.36054202882785535422, 0.62455817167852978322, -0.51908065077945508836, -0.31298744786994819878, 0.55200465868400239167, -0.28837102843329098789, -0.6552821263859346379, 0.61876005742673290122, -0.04042631317509406963, -0.0071499014382997438943, 0.05433063401839888229, 0.13320221464709519621, 0.1598427918714214746, -0.021217302288148887079, -0.42378640917617138584, -0.59017460730834270244, 0.28737555578513618304, 0.97062430983171865684, -1.1963541248137286033, 1.0523824802305463688, 0.38608078660584038699, -0.78088745806008164063, 0.026937796825891330982, 0.0024427822380308911177, 8.5134653418778016842e-05, 0.036495065370399759619, 0.079148413196025835248, 0.0032727318258928828065, -0.21968112311535586967, -0.41030110761059029878, -0.022016709038375492646, 0.78351823037337842148, -0.58040787945097083167, 0.76837786610424996248, -0.043687468793194905725, -0.43116847132223434791, -0.043000864644511985202, 0.019916976899895258435, 0.05401446199319556718, 0.050671363605340691283, 0.085805692193032967463, 0.2498617664171943753, 0.42733846892453325461, 0.1948417496160295137, -2.0088390551308141774, -11.49761758615847107, 24.620195206271183963, -11.656148745957587209, -1.6671828044029426419, 1.1433844675310775862, 0.018438699666533130139, -0.0080581634552695461521, -0.041860255088589017058, -0.068001190727251789592, -0.044108273144920190356, 0.088659240648737924761, 0.33163182937426682884, 0.14180399726728365462, -0.70511614076707207488, 0.036599219434898827563, 0.54597142813711008191, -0.026497485422048973336, -0.82189386033337263893, 0.55749242676374166638, 0.0068434238627020941778, -0.013202439418535655627, -0.026312622949005964879, -0.025624637694231032492, -0.0024515955782041208566, 0.06405536341387124688, 0.18804328917623994166, 0.019140229232769707945, -0.46142527172817626724, 0.18668430848791195409, 0.12954491399672848906, 0.16413741455341868125, -0.52972827160702540272, 0.30161852105490832843, -0.0192454280698649699, 0.021295651391269373315, 0.061569070597683996093, 0.095649675429654051895, 0.10351516122681372889, 0.037427710607727981973, -0.17141655723178680915, -0.44944464171483300419, -0.55702211110107346936, 0.051163274172523448557, 1.1184482060427130001, -0.30251994269767878976, 1.0267744650333070489, -0.99836022593225037802, -0.00023515626143228113374, 0.0047650410949006171205, 0.0058515234214543132563, -0.0010887738248760961586, -0.021389176771885756029, -0.033979158892311775442, 0.024232176118820995814, 0.1504882584191143402, -0.038432842222450941394, -0.35492512193120340225, 0.36148513463436166004, -0.24590426587107544476, 0.32631353386076278333, -0.17801866329862933713, 1.353494671167118418e-05, -0.0011773488958566587573, 0.0022949236516733159327, 0.0048782171938695348054, -0.0023642923033629758417, -0.0069756040269001590159, 0.0030708122755356863018, 0.018946312088985031025, -0.057330773556205597008, -0.053055357604772444835, 0.36451895321051774879, -0.5437053946796283066, 0.37701616021887579544, -0.10588690701459187171, -0.027921863227984521777, -0.017745444385495572071, 0.0091968138100433881321, 0.054235468770464784205, 0.10127733978263589454, 0.10936922407860817652, 0.011699288207154004415, -0.19258692329169388446, -0.48632185904468411275, -0.39418013111566369, 0.75111789374802540031, 0.19472829142412451109, 0.76247837541285012897, -0.88470544894402203262, -0.0083311100068269080404, 0.0026843369291155304464, 0.012914042862337635903, 0.019685697194994428855, 0.016192133618003944073, -0.0023253681705647458872, -0.034230283865904334584, -0.050580247299488151347, -0.12297867779914192776, 0.021539469225372261107, 0.54509540277835644151, -0.74081997813775468664, 0.54303709698022550878, -0.19999909523820855184, 0.0074506542184557215253, 0.0032985113725550758296, -0.011007348356399319911, -0.026819077328456869536, -0.020905823665417475055, 0.036857152141664649425, 0.12553235153112937694, 0.09183230634067582332, -0.29620467576588682679, -0.29951696687034967681, 0.59809781813087581348, -0.57031179335827308918, 0.80681625793090738785, -0.44908750127885055381, 0.0046763424075654874343, 0.0021232336313171560824, -0.0095291571354998057253, -0.026796773476014026233, -0.02359437958629575019, 0.04005669874415054843, 0.11196889519904958121, 0.056627165096238191166, -0.19958157815959631787, -0.21463234600509051297, 0.4347201683033008468, -0.48175748852274319356, 0.61276082136995990712, -0.31117516178735721288, 0.00055719024119827353578, -0.019604690760929674442, -0.030957162175432383772, -0.018157995404311899623, 0.029883860795997932436, 0.093933896544389605676, 0.11265935660042884769, 0.055587284050874589958, -0.23874898212327372304, -0.45865627671771275198, 0.45856945615467320509, -0.098509717513248301612, 0.54265255657717159643, -0.44221558483542111606, -0.013452897314549605232, -0.0061601418943674785683, 0.0070898475422719355044, 0.024436790142705229478, 0.034364945915274232313, 0.011742347878266445882, -0.098028829717460019832, -0.23061408977010949828, -0.059449298321455984473, 0.80161768813314626669, 0.75073313222082560525, -2.2605457492182359225, 0.85401356069457712294, 0.18773138715545156785, -0.013111948473074494514, -0.044548171943389169858, -0.047678979212327637915, -0.0049993105480589661305, 0.079359742177016801001, 0.14849286571242337196, 0.12722533990079165633, 0.062885179982099415064, -0.23559867692335240141, -0.72310417309950625331, 0.31429001055563710043, 0.52454814748077027708, 0.3093881891460074085, -0.52202836630494453551, 0.056366552671725148049, 0.025827310306111184168, -0.026840693276043381399, -0.10568959278279188574, -0.19799639757737019785, -0.24361908782246771699, -0.17006774857862638761, 0.14530092456896281927, 1.0221417691997820576, 2.3039403355075482871, 0.16329349340353332964, -5.726676521370611006, 0.36862458571874140434, 2.4128768682012515789, 0.0038712182245270584588, 0.014083625454477357447, 0.015633195667329031348, -0.0035233844795013735673, -0.053896034278749352375, -0.10911402668072842093, -0.051672986987735793074, 0.2742917176899634657, 0.32751846891627894864, -0.50461489638977974348, -0.043824356003850713226, 0.26783767767856181408, -0.33961929698712362002, 0.20681991907033658862, 0.0058738434201981246791, 0.0030653400209449651118, 4.9768416049487344127e-05, 0.0016294395486156033545, 0.0094252718839195545608, 0.0099205307895226246817, -0.072955541309752819368, -0.25059525576767882615, -0.10498927778514052789, 0.97737358835911247734, 0.79795610521665805148, -2.5942019426414337779, 1.0126971902017116722, 0.21463362877089600667, 0.0035292292692078227473, -0.0040729835988230975174, -0.0062531587442016776934, -0.00014824215009453861266, 0.0094884057549429849654, 0.007511301907527656932, -0.052373388946625507034, -0.14858217505361639987, -0.040431562090069525506, 0.68539789714710841473, 0.64752414445052852709, -2.0913978956624439398, 0.78484673842312846492, 0.2097762029938221584, 0.041956331758352755079, 0.011593305443816388084, -0.03569344345393355572, -0.095615860632571381128, -0.14650444091845044037, -0.14206028490393876607, -0.059499117840774398691, 0.14688168945906135421, 0.67916358120045849933, 1.605034027102282046, 0.24771544820971302947, -4.421763196243938765, 0.48946088089280798128, 1.6931849735734192386, -0.010504321560985558845, -0.0045611384488331487702, 0.013727038609606235717, 0.033686646100994345476, 0.02684719152906073214, -0.016367407511577967971, -0.082260614783431246222, -0.17300696247961422358, -0.1514549567748399439, 0.68734750622974138956, 0.82927857786752512403, -2.1573094568273547544, 0.96694636420931412424, 0.042931896257051571641, 0.0033781734527773207843, 0.024560229419095816472, 0.03434962893409036766, 0.0094230736746392373959, -0.079593143826589274337, -0.20714183736663660529, -0.16766798688437523079, 0.35913055299273605181, 0.67086673857035727941, -0.43670207180349890352, -0.28837102828749072714, 0.27569110993679496469, -0.82733629151631471466, 0.64046838139015116464, -0.025555229365045242784, -0.019864762851746639583, 0.0025164109323035366295, 0.054330633900028152405, 0.13000179250633570049, 0.1608303807850590128, -0.016237708581129550733, -0.43156739111913355877, -0.62699598507973308692, 0.24947249250820280331, 1.0523824798281502524, -0.93947513415547112547, 1.2393645907124390959, -0.83439059216653310713, 0.0065861843562801305335, 0.0095596011074305167754, 0.00098792185612843474996, 8.5134373268569429882e-05, 0.036763647100648416677, 0.089188947099502230431, 0.029713148970323149944, -0.1940299081018749261, -0.42492490726336329399, -0.090849511788602121953, 0.76837786565070209654, -0.45169626546280106094, 0.89933590658299056297, -0.67482221966770727928, -0.040157687189150234608, -0.0090056439485469153006, 0.029965644150664633161, 0.054014462179414024501, 0.045086344684571622032, 0.074599675031497336697, 0.24274272460572460242, 0.45715825576534374619, 0.30323311055648405343, -1.8844574904493174294, -11.656148745669357325, 23.915607679336055469, -11.870408480268331175, 0.30281597819725764031, 0.014487246743121823256, 0.0065321435223932890202, -0.012202560484708403071, -0.041860255391736891917, -0.067861570838376350512, -0.045751683925961843247, 0.089200774391766282645, 0.34328137545324310942, 0.16490018493234726016, -0.70962430964432909875, -0.02649748489063149387, 0.40175944320077205951, -0.12442429419538612867, 0.0040885709080410821345, 0.009542848322720228621, -0.0015235004628099515014, -0.015458944942661232563, -0.026312623624783440524, -0.025399191009749248066, -0.00197301751370862824, 0.066547274697602673554, 0.1929602849856029978, 0.022249960962796362651, -0.47505400023080573968, 0.1641374145600822676, 0.07069206295233769477, 0.12537488837957111443, -0.11039729756974418773, -0.041118060166166002511, -0.013136687250017483469, 0.021295653868949378923, 0.061690361549331225455, 0.10513172792531853827, 0.1325228751721992948, 0.089732676266939592757, -0.10712697241894721112, -0.40338958143550573965, -0.54086862401429125491, 0.010302253484423764054, 1.0267744653587609349, -0.64823810150540916997, 0.37795345253392986784, -0.0022853003404393265607, 0.0014217679828605216986, 0.0047650412513877710846, 0.0057085263677643166211, -0.00025657728139398885969, -0.019928797510810843108, -0.034577181782378070551, 0.02595371028654108736, 0.16817441182063588845, -0.027573470978870442266, -0.4762965288838261757, 0.32631353395415285679, 0.20729887706118654833, -0.17875875528623663691, 0.0003702294056005941492, -0.0016308155988974863768, -0.0011773486008977063124, 0.003452051194994430576, 0.006449245375831549193, -0.0032541958643707204896, -0.010878156705186069197, 0.0065307520325277590434, 0.030694623933407067656, -0.0581292712801449174, -0.10040820731907877661, 0.37701616047418184952, -0.39014674814403810243, 0.14256800713517484014, -0.0074610446644169901853, -0.018048117694460786425, -0.01774544369437478647, 0.00041041990128922248293, 0.040007432160809437705, 0.089000138905038050918, 0.11020925696707610519, 0.038034879509348307192, -0.13864641687443962637, -0.41049777483762850405, -0.36569289165622709259, 0.76247837454803202295, 0.19097662653812594313, -0.25774475011472736119, -0.0066547796399510378978, -0.0033912233731314082169, 0.0026843370046289670959, 0.010436568239815455922, 0.017388245295795683176, 0.017546634956050084775, 0.0047943038683980841036, -0.021425108030873017684, -0.036146661379572481154, -0.1138818690712847248, -0.02490922794168816401, 0.54303709661313626089, -0.66853430647080469562, 0.2878412778917352477, -0.00018048077611085229554, 0.0059903614658129425083, 0.0032985113873364815297, -0.010324697107122169509, -0.025263726222923863635, -0.017933950983072419194, 0.03933426289099074824, 0.12564727664934777218, 0.084206764396155667218, -0.33238585334007936245, -0.36804996607078704907, 0.80681625804276613234, -0.25303644498773436577, -0.057890748062764819937, 0.001770711376160916551, 0.0044282172864107943039, 0.0021232334687777930371, -0.0093421435586468638129, -0.027582784239369273482, -0.025632013809745491512, 0.040667327055987613493, 0.11788344296218578044, 0.042112851648107164915, -0.24681552316876487208, -0.23120559347802160932, 0.61276082151409538845, -0.26810945671618013897, -0.016161696013832785312, 0.012314559336872856962, -0.002312952787089646127, -0.019604691275982327747, -0.033067848774981048465, -0.025527344820870139647, 0.017362711586641821349, 0.077344859660752066088, 0.095011361839326458067, 0.051653723077539459574, -0.19207933407891714106, -0.41878582835951178032, 0.5426525566613717988, 0.17939399086774598935, -0.29730625327407939062, 0.00024229328835821933429, -0.0058591414891815945393, -0.0061601423315321961888, 0.0014380092645736915896, 0.015933871146394745777, 0.027770335801005973514, 0.012473496923738698963, -0.089676097287405262537, -0.23425550329399208449, -0.068319146392605709095, 0.85150561706081495661, 0.85401356067991762711, -2.6667946958578534478, 1.3076355650508082462, 0.026370956181220133602, -0.012800089584636666074, -0.044548172663203253197, -0.055538861215623402945, -0.026851990429664712723, 0.044224582162427472642, 0.10836423778870546419, 0.097337330253831344518, 0.081188143490663250423, -0.10389082665036922237, -0.6104113458887490129, 0.30938818930846478938, 0.84922011183856593863, -0.69092209010580019068, 0.02749942639610123557, 0.034766655792990236451, 0.025827309831740780099, -0.0093115587331014104677, -0.079863049718662815213, -0.18050291939506690753, -0.25732768103583497643, -0.23590834605104754607, 0.0011939982440294372168, 0.82774679499323888798, 2.265728289197530998, 0.36862458560325561674, -6.11604836872167823, 3.2833263452556100681, -0.003815235536844381603, 0.004222891792311981099, 0.014083625139301560528, 0.018497518728243473457, 0.0021455569408295755568, -0.047903500012303527889, -0.10688955615922809961, -0.054275005650206875618, 0.28257862297120706074, 0.35908205637118345566, -0.65530018603899453655, -0.33961929678349561668, 0.94209316693435718459, -0.41794836895773979091, -0.0012058394898682392882, 0.002960019801080250243, 0.0030653398715552015313, 0.0018427360896154090439, 0.0063664210121326886629, 0.017621809247492178224, 0.018164477711193720288, -0.072010410934628829582, -0.28253783170611934672, -0.166251403081439042, 1.0261369011455678812, 1.0126971903726476043, -3.0970998169722867566, 1.5383557547224702144, 0.0051463290137569650193, 0.00047523080711774626965, -0.0040729838899614886905, -0.0051691958742249256506, 0.00080662126777416697648, 0.0092238772827043825908, 0.0044452750581775481165, -0.058794862664943645103, -0.16862700039466954616, -0.066122540586708231092, 0.71862870558460800741, 0.78484673840565710723, -2.3768991084981361261, 1.1551044383419621919, 0.027101066782684927331, 0.027036672420610939616, 0.011593305108391738223, -0.025162507123622529503, -0.081765696341254245905, -0.14063192693930801069, -0.15989416900161479429, -0.11696496825393207231, 0.030540699942820062618, 0.53683688764037729158, 1.6021347805017815968, 0.48946088077949689854, -4.6318836485763652533, 2.3877769190626163542, -0.00084395243719549446065, -0.0083333864665898794988, -0.0045611382878923689699, 0.012607285807850122927, 0.031059709725354611026, 0.022141320767870502351, -0.019601213753854550975, -0.069483786738067418454, -0.14653530183192392533, -0.16536632881159235575, 0.63756128691408231468, 0.96694636431209413097, -2.4162102966223768519, 1.1672425965849062557, -0.008682192557639741709, 0.0051776614432551062994, 0.024560229438221628517, 0.038723752654934696926, 0.019458193976363571953, -0.067319595691637812429, -0.20045446745542219724, -0.17111545706228833974, 0.37275206433753360624, 0.72433421119280350098, -0.65528212695541543376, -0.82733629178344147981, 1.0931999069932289981, -0.35060649386870856858, -0.0076110970107814129165, -0.016385343002451653593, -0.019864763005837227094, -0.0071499005615098948363, 0.040149966871150480352, 0.12194882957249958411, 0.16896907953475653885, 0.0088427384026213418089, -0.40664939320299497982, -0.58499487186710030784, 0.38608078674713425249, 1.2393645903997465574, -1.4796034404918676497, 0.57895024480556245283, -0.014810468771975433475, 0.0042737824046849039183, 0.0095596010716492638148, 0.0024427828237809506851, 0.0074221882705692437954, 0.056211695637285744032, 0.11717216946170221081, 0.042797862612644940739, -0.2219337744231256182, -0.44811577927443624336, -0.043687468462685674631, 0.89933590627127413253, -0.57327930999099596665, 0.19079482245533246898, -0.019905520486333289076, -0.023470827860631041789, -0.0090056438749238770791, 0.01991697698917346826, 0.037085272769062080367, 0.024187248455616872089, 0.066871734876676552917, 0.29563682152872583497, 0.59738619791167046902, 0.49431855569301430542, -1.6671828047793233463, -11.870408480136211082, 21.672407062453014959, -9.6177917276624569354, 0.0069028224932340632655, 0.0095424276385711603865, 0.006532143338774110998, -0.0080581638132536242969, -0.03713592398728831484, -0.067374310720957938936, -0.053916642464576045035, 0.072755832491805866313, 0.33052317338081271814, 0.16851341618450318971, -0.82189386036641887134, -0.12442429380227504332, 1.1634399439294145218, -0.65995314413435712808, 0.0066655704078446911809, 0.0049986496728244523766, -0.0015235007535158271059, -0.013202439264789873896, -0.024344800266074020229, -0.026247211806244400562, -0.0099904458503364927668, 0.04983013986048177163, 0.18363386512519089266, 0.042692625562611441481, -0.52972827131735211736, 0.12537488856543888005, 0.55162045406661563973, -0.36957436342823207776, -0.050818137897217965926, -0.041118059953041082022, -0.019245429477696657933, 0.018464047099613245528, 0.073728267741804742741, 0.14544168484813152498, 0.22103695348240853802, 0.24958334060878070715, 0.10478782655573251614, -0.32710542230392408714, -0.92913536117555939331, -0.99836022605151042431, 0.37795345274739033936, 1.3097911123935623845, -0.0021201046066734517519, -0.0022853002914328020279, -0.00023515636387985602547, 0.0045778503715068726357, 0.010733555957165167097, 0.012227291563074602077, -0.0050412120889385351324, -0.046164753517704687491, -0.052987562114182006068, 0.087821225454718446213, 0.17658502604387599222, -0.17801866335381910655, -0.17875875525909112884, 0.17776031119602786479, -0.0010692748102605704932, 0.0003702294945761547924, 1.3534745567983405253e-05, -0.0014095049125433438495, -1.5097812767861498512e-05, 0.0059014231119641066478, 0.0066522314933952974778, -0.0039565510640018689004, -0.0082789206756815527516, 0.016837402872147570754, -0.0053880341248481360486, -0.10588690711791805266, 0.14256800715963049986, -0.045047841846046231262, 0.012267419011307888546, -0.0074610442024536094419, -0.027921863523122425477, -0.042524021135299239216, -0.037483364158655642362, 0.0056484558814158813966, 0.095566651725815660035, 0.20553786856456060517, 0.24264122388555747767, 0.084673641670289068184, -0.3745809546177644167, -0.88470544850513832547, -0.2577447506040477232, 1.0403707324358308917, -0.0018104712116024466312, -0.0066547795611904582425, -0.0083311100465802718357, -0.0042092429777257807377, 0.0073111058761679689408, 0.024669571919641666152, 0.040082663645209307135, 0.039237176035101448135, 0.0072853613387627667169, -0.044353930195846512363, -0.14890422384686086144, -0.19999909513082447798, 0.28784127781370244525, 0.021785834705233288883, -0.0051838814445627334065, -0.00018048074853737377447, 0.0074506542430340712974, 0.012995280886224013991, 0.0057292470632879684278, -0.021258458533832201576, -0.052716585992559648888, -0.037441676203211822493, 0.080118222329943269977, 0.22864039769496177779, 0.036245514126364654361, -0.44908750135440872464, -0.057890748075157129338, 0.26089896524048822313, 0.00037452247192754203799, 0.00177071131527076122, 0.0046763425419033268979, 0.0069675278709541172556, 0.0010158816776422959546, -0.023091131098370201724, -0.055587589033991170739, -0.037558975747286546532, 0.078780298869803616624, 0.19173868325658241507, 0.038074350585314775675, -0.31117516188833982405, -0.016161695986658560636, 0.12018584041035841603, 0.017564352511427277354, 0.012314559256960278671, 0.00055719055458209806875, -0.020614629626605995155, -0.049882146540957031378, -0.070734645231027321044, -0.048272176482853285107, 0.043998745787511686633, 0.17247339396526767397, 0.25436585654544691115, 0.083244979788516609842, -0.44221558486018663903, -0.29730625326455650814, 0.33893917865285799751, 0.014527324246769239904, 0.0002422932190936859893, -0.01345289705490875995, -0.021976489512168483065, -0.018084734465561858041, 0.0049778676854172710861, 0.045373966934974509224, 0.079004952903010627896, 0.024401216171014660644, -0.20258385477410006947, -0.4048380105028516196, 0.18773138713047124448, 1.3076355650464521752, -1.0265501463055557441, 0.050548489540804084075, 0.026370955971946372087, -0.013111948098858312167, -0.065634137427592248448, -0.11454333219505957731, -0.12235278109201126739, -0.044002213500990552952, 0.12001492358117106518, 0.27677629402800274505, 0.34807594029645549938, 0.18540617521747243535, -0.52202836638930127933, -0.69092209000073956382, 0.54616194968015163713, -0.0064940585100043846795, 0.027499426286983083556, 0.056366552891885940424, 0.069244014743485793328, 0.048135796984086653971, -0.031101601920836455251, -0.18754656111347975389, -0.39518996221741342634, -0.55670290592178839351, -0.48756561850888668896, 0.28030214429545563526, 2.4128768683430581454, 3.2833263452154288764, -4.7053925517388037747, -0.0084309448867139723249, -0.0038152356685194812463, 0.0038712183822431130847, 0.016742567123769730442, 0.032956350547116652905, 0.03704421281613457656, -0.0040850851830226261527, -0.114494137000885704, -0.21736615076512907918, -0.024564347953033319194, 0.44857915518783791908, 0.20681991892132195465, -0.41794836884740893534, 0.037481908166530637783, -0.0077086957652999959015, -0.0012058395988659613776, 0.0058738435038828556989, 0.010412534054946777362, 0.0095715827026804176786, 0.0073069635708753266981, 0.014653360231162770172, 0.035841878780525159021, 0.0073482959507093969717, -0.21431493046581323592, -0.47923953018407505766, 0.21463362864969109944, 1.5383557548279396254, -1.159008779118770871, 0.0035621187497819591442, 0.0051463288859403431494, 0.0035292293465963523597, -0.0022735256923704883977, -0.0097863681362040100264, -0.011004085402078712594, 0.001722408260026501341, 0.023074999162092487553, 0.0027488330621559440257, -0.13459797358922087618, -0.27881830678131153123, 0.20977620297141461037, 1.1551044384630753115, -0.99390419447114231044, 0.0073268978890956380084, 0.027101066759938390038, 0.041956331953799531898, 0.042278515787257178737, 0.012468632596544529378, -0.062385029566447189331, -0.18019540374598774779, -0.29728001861576991427, -0.3415508489390514657, -0.23104895765040250888, 0.27470519803960913308, 1.6931849736176249888, 2.3877769189799575855, -3.5316863947872012552, 0.0057575753238461009967, -0.00084395246779504079493, -0.010504321643280859627, -0.017174812985943737481, -0.0077370643686534741612, 0.025537601358391705253, 0.062010065849188882037, 0.059126157122138242594, -0.0096171988870414337747, -0.16543973087502647501, -0.37672378494866626264, 0.042931896201611099984, 1.1672425966287403032, -0.7832600546157331145, -0.015196447398472714974, -0.0086821924830305929888, 0.0033781735206081502793, 0.025944762686386329403, 0.05931262009041876615, 0.080480195092944711854, 0.027343463571547415442, -0.1684072097279219471, -0.4127909810940595392, -0.21165918351929235985, 0.61876005781171938835, 0.64046838153663521176, -0.35060649404298821485, -0.30708086692148534036, 0.010463248235063117922, -0.0076110969257033771543, -0.025555229263085791486, -0.040426313723358836549, -0.042542739339836224688, -0.0060482606850304418411, 0.10371402871057089357, 0.27405261639558975428, 0.33078941552608776, -0.032384963142119221891, -0.78088745804214709789, -0.83439059202341514521, 0.57895024469612688112, 0.52393703140807756924, -0.028834264043271606859, -0.014810468686308319447, 0.0065861843464034754841, 0.026937796501142297761, 0.030193283577395285677, 0.015637856082375098388, 0.022747620148803594092, 0.096353991165941579866, 0.16257914824545274812, 0.017578910653765506683, -0.43116847144855846574, -0.67482221941608111493, 0.19079482233046216044, 0.63759474024424001026, 0.0047228285231981281464, -0.019905520439887403744, -0.040157687194666433217, -0.043000864808925939875, -0.013678271888507169407, 0.038825784631792349677, 0.068606276888719758267, 0.071968765322183075872, 0.1906902822980577894, 0.59501840716548071164, 1.1433844676840541066, 0.30281597811468446979, -9.6177917276177620209, 7.5003959680206060412, -0.0012136412740008585727, 0.0069028223325238525013, 0.014487246807660768624, 0.018438700030050339529, 0.010454604216881761836, -0.022590086810098119335, -0.08615939419228869911, -0.14777484541812033192, -0.10201640605703098108, 0.21583222423718598226, 0.557492426568765298, 0.0040885707143779913622, -0.65995314393348913917, 0.17362496277705166681, 0.0016450312833001233016, 0.0066655702954415982711, 0.0095428484410075998384, 0.0068434240049359352023, -0.0061012210449029758569, -0.029730117830883896757, -0.05499022681863704326, -0.063100451675738217228, -0.015659056700959156361, 0.16438745885245595324, 0.30161852084155366116, -0.11039729765430603137, -0.36957436332617232821, 0.14851406001131639556, 0.26721821519271382606, 0.28715199748904723931, -0.52629639964865349011, -0.37827702904533982986, 0.018965735354097849191, 0.20673346083958868413, 0.18715382150869874778, 0.079183024115472219751, -0.01342342745029006211, -0.055375946803326282775, -0.05320846723572916076, -0.028767446941119569764, 0.00027322484606694184654, 0.025807827433868394534, -0.24136096181131891369, 0.68222115841987240881, -0.36261323633233483177, -0.29016350586045991378, 0.11237539859110255713, 0.12480395714201247825, 0.01329167055704835558, -0.025012970148523969854, -0.010883189810243320064, 0.00079649112595473251213, 0.00078519600927528978695, -0.0016063182200673424126, -0.0013579210564940522726, 0.0011208262443898426783, 0.22733329972054280854, -0.25778694863692053207, -0.15539615073049561911, 0.15045226367008848745, 0.087529709431291236044, -0.013573466028241542095, -0.028447715263921251327, -0.010750453887744454429, 0.0015339876973784140248, 0.0022395618480241840167, -2.3394110063437940553e-06, -0.00051528180091141501714, 0.00019207525803194611764, 0.00063120106073332176311, -3.2608013257869781221, 2.4284155188690523808, 1.5248153692339891929, 0.15693704637698149829, -0.29580859348758786753, -0.37170738112987117274, -0.27726521707501222691, -0.14695275557993359161, -0.04074497497474956903, 0.028418776734735140543, 0.059498184549634888785, 0.056007971426326609576, 0.027906588430121115341, -0.013032834505443366066, -0.45686311156384334931, -0.061834285708086134115, 0.46013592865513158747, 0.39178145996436813947, -0.050182906948411827586, -0.1994409442901838303, -0.14066429441417252488, -0.046526029182399522566, 0.013853104543806768478, 0.03055517183176501006, 0.024135404856154007991, 0.01185791377385241388, 0.00097751615541755740128, -0.0069853186390235163031, 0.54273096874485293739, -0.44208149065634688712, -0.4405297604835718972, 0.046216864947921793005, 0.24151645369572002919, 0.18706392486974951495, 0.013105163245154058047, -0.087934875296940592793, -0.069887985455870424434, -0.014799138128295547673, 0.01694868816848616247, 0.01873199549986676038, 0.0068585033519266529189, -0.0065563647523959435806, -0.78027447727957277568, 1.0821549802283707553, 0.075882332990280854101, -0.33117366517671747994, -0.13795609510819356602, 0.022969776555937671675, 0.047318022033997063358, 0.010682933379486005479, -0.0069830991732663695581, -0.00049300931466626210422, 0.0059248019931459816254, 0.0036857351012360545749, -0.00091748229869390040582, -0.0029647736256106739169, -2.0612194564240140515, 1.9766154138907163773, 0.91280978741500962137, -0.22371153713645727801, -0.42460837706595649932, -0.28783083100780570973, -0.10064907856737312364, 0.025027513960694976192, 0.066123663806118754271, 0.052617894804373593898, 0.025299763643365374283, 0.0044154462351099045894, -0.0067400034540766188065, -0.010871961281512193195, 0.062057810927526484146, -0.18443192142219350105, -0.017557087926399234101, 0.22851399923171292206, 0.032818132211672589038, -0.087970687222899260349, -0.057897279686764779871, -0.01104391053220806114, 0.010962842476040113998, 0.013794979012144005168, 0.010215920616302269497, 0.0054050339326429264891, -0.00012357894555506529141, -0.006412931011660690829, 1.5775903155659394272, -0.45233985457291581156, -1.339177878546153444, -0.54034813215333843406, 0.22057259717687180411, 0.3996391896468899918, 0.25870849193882106132, 0.08338915454229699209, -0.014570436213578740609, -0.045639857944325415429, -0.042280162175032928729, -0.026326162560086845132, -0.0078817775357792087265, 0.0091152964739971992919, 1.507863075698180122, -0.49970943060608419284, -1.2676678253239601624, -0.46196734663443722591, 0.23442780758908440819, 0.38115822106040542216, 0.24009180632502510888, 0.070459412132667126327, -0.023103791143124424029, -0.049750693633780540548, -0.042088834324397431108, -0.023357661232928421391, -0.0044323022689253541509, 0.011108590563461660794, -1.2792653678317242871, 1.6905048879699191566, 0.24653778064195938646, -0.47810066665671213659, -0.26473577543176018967, -0.059967181664642189853, 0.0294833991869344475, 0.051561628245731544462, 0.041625597405608026624, 0.020993549302769996812, 0.0024231396305921205567, -0.0081172028075950277554, -0.0092002783590398318214, -0.0031496609690810535576, -1.1295004716108481002, 1.5266825593360049407, 0.1716529192568197193, -0.43767544610956177209, -0.21465990449438723808, -0.036766173882048204524, 0.029074394741764882089, 0.040150819547941363208, 0.031078434640153612678, 0.01644174273751640436, 0.0028091224109639046681, -0.0058432288015856093111, -0.0073699624539965988626, -0.0029455238751076411041, 1.0897370160182175969, -0.1282483475171906917, -1.1211852593329345584, -0.48605846559815102381, 0.19654205493272880467, 0.34890002028275063362, 0.22542563321611810978, 0.059676485592592080909, -0.03169521599825751712, -0.049100135425430714575, -0.034754356408708129977, -0.016942512144447093025, -0.0028691157006663957699, 0.0079771145029078824318, -0.033212115522349393293, -0.22018411485821789975, 0.22984406308146335118, 0.34586628159246146641, -0.10721517599282956446, -0.24753902653795548106, -0.10281788608221072157, 0.028455311636331262481, 0.067010154803936503032, 0.048011220214972782916, 0.017552900848238808806, -0.0030148878515245686877, -0.011468265425354895459, -0.011892575517121665649, -0.20614357709231279081, -0.059015283281222026546, 0.4672292499383060016, 0.163339527195109524, -0.32917564185640052488, -0.20669977452169460697, 0.042300126787616680601, 0.10394510566219895831, 0.052880059252881937704, -0.0014496284406986194217, -0.020861592344122655701, -0.015483413170013168164, -0.0034185565268450069688, 0.0066374212387800618149, -1.7735106282815089607, 1.6389303591360850909, 0.6639077563020202577, -0.16787300895191384353, -0.20050147227724224308, -0.12898759683459939951, -0.093886522377717121501, -0.053528145574673655083, -0.0092623595638890495602, 0.019378582375268276622, 0.028136400036463466573, 0.021724333531647228163, 0.007688750106145601261, -0.0083730426327191433827, -1.5756324151517935306, 1.4928377586021579493, 0.56722175179217959595, -0.16735150363669790141, -0.18175946130249989729, -0.11421072442977120431, -0.081781250945032801791, -0.044638060866830545081, -0.0044315556443690678687, 0.020197126396906059376, 0.025924673881156774791, 0.018397727020948509424, 0.0053059938814566491583, -0.0083966519618889103038, 0.2094550809058041696, -0.69573938440912808101, -0.065282656938490460363, 0.56170500199713935263, 0.27982359945492640829, -0.05988284195622978201, -0.15168251250131775665, -0.10989692410160195601, -0.040929475453598179735, 0.0040854640705372856704, 0.019824927403308253981, 0.018245519360905369799, 0.010008896842022607654, 0.00010707739936631094108, 0.173624962718840814, -0.65995314411517158604, 0.0040885711003657426721, 0.55749242662023801298, 0.21583222433017934527, -0.10201640651749914468, -0.1477748456536489563, -0.086159393887399834955, -0.022590086404173259638, 0.010454604018168332444, 0.018438699873510433491, 0.014487246929506113202, 0.0069028221080566429246, -0.0012136411218390847234, 7.0102264023249265534, -9.6845635100123068639, 0.65881391050885163008, 1.5116599625511899507, 0.65592284634110065156, 0.085685716090993818694, -0.018686591725522561092, 0.012596996477043993778, 0.0024129027052404651739, -0.030572900892078942042, -0.042116856663825245899, -0.02931597363859531763, -0.0073140989786057040642, 0.013701140144881034794, 0.4131610460118150252, -0.023843148552765761755, -0.35988107311505079799, -0.23891237644396667861, -0.012450342261135519117, 0.13124460084645814262, 0.11805360266200445873, 0.049814017751129341383, 0.00024547931431558505993, -0.021032722965180494779, -0.021624312020451993965, -0.013076591190962317396, -0.0028883716048975554766, 0.0062791899726741384063, 0.2871519970063969307, -0.87589767272136320742, 0.83565123045228340004, 0.07597692508619477425, -0.29324647055098029824, -0.16940928707171631906, -0.00024212453046635050739, 0.085406638972133547272, 0.075010641568857922001, 0.030972282251892104682, -0.0074947126108465432145, -0.023359227342332877092, -0.018154899760962690813, 0.00027322516828119154013, 0.68222115844568431697, -1.7176710795095830875, 1.195186718670389503, 0.27128296264104678759, -0.45836353128313150895, -0.1117307952271645588, 0.1039850491043780023, 0.062652755117990829858, -0.0057680884482738687541, -0.015428831144828680788, -0.0037895755253482404923, 0.0018842769742126439826, 0.00090442192821294062517, -0.0013579211711615196342, -0.25778694857660849893, 0.39783139315483456722, 0.14527803892288601584, -0.39848312329628965411, -0.0074691212381451438027, 0.11248775812058020573, 0.034500188587744680391, -0.011671491393565497016, -0.015668453735051841164, -0.0037875978794084546632, 0.0019332412108043601183, 0.0012780242300870689725, -0.00031573781249466797028, 0.00019207506295694073343, 2.4284155191377947425, -4.655968223693459862, 0.68129426874849285944, 1.5101127173871131681, 0.44929435267236755847, 0.015121471666123887864, -0.16129180461909212174, -0.17388837722380401374, -0.11544292924415300305, -0.055786848680567879821, -0.013652015419780987732, 0.013472863017348700454, 0.026420134694601429826, 0.027906587934206583679, -0.061834285701355275189, 0.27779830836898267155, -0.39701939551421100427, -0.21215561649765005137, 0.3842580821606025876, 0.17927038352763077778, -0.039905178444226294121, -0.089713308077623010739, -0.064482475989549051243, -0.018784014455271603322, 0.0080254498446306801046, 0.014307705898905259678, 0.0097747918424852114905, 0.0009775161959312335825, -0.44208149067936897136, 0.58789978470623205364, 0.29327026333793027035, -0.37561815278683668984, -0.20520564116878439731, -0.042492175095411893748, 0.1196830494611528517, 0.11876458176789415899, 0.011163702722791043337, -0.044263502435783995403, -0.034301595391395600654, -0.005894630739045189291, 0.0093459135886146439687, 0.0068585031724646043985, 1.0821549801946941383, -2.2437560517630896761, 0.85811308099067917787, 0.62879333184062513595, -0.11518826176861182708, -0.18872236813016457502, -0.069651664736532017952, 0.026562401852518711248, 0.029326299183124595465, 0.0013895108363121016527, -0.0071251234835366912893, 1.6107424800122544813e-06, 0.0037271087851033441254, -0.00091748239692370544673, 1.9766154140966518682, -3.6693244402387703573, 0.65539497784814948478, 1.2079012753648867839, 0.25034978212939906772, -0.12359304122917916247, -0.18429450513464210037, -0.13080746947961069404, -0.048081237103496474172, 0.013045316452123575934, 0.031172893727929515406, 0.023263174106562534149, 0.0078220190994042915894, -0.0067400034395751961744, -0.18443192133583954928, 0.3181607200552448167, 0.092442607360954734408, -0.43501737170019133538, 0.11471805772403946555, 0.15628613636073404591, -0.0077659435495150298118, -0.043946834304784139491, -0.023472773967592387018, -0.0014114336524377352634, 0.0061299728578400646331, 0.0051185903413092639566, 0.0020770982600803273785, -0.00012357898150990969917, -0.4523398545116643632, 0.38426326644382824416, 1.1161017960645078784, -0.48431437906952723438, -0.72746787292103354261, -0.22264751879525979117, 0.12824546372681946971, 0.19298669558449765837, 0.11251448392249768782, 0.032734617207502918201, -0.0092814467809067876386, -0.021719290667970514441, -0.018581194442205222622, -0.0078817773950394503374, -0.49970943057854727565, 0.48425923838587842107, 1.0299104787759980617, -0.52281543952394304853, -0.66967548722134562222, -0.18506236213218918429, 0.12626074705969261847, 0.18138452535563920343, 0.10446852356012079066, 0.027752720849529351199, -0.01257271744584937366, -0.023193196717209194069, -0.017702922529268819152, -0.0044323022352637190396, 1.6905048879336730394, -3.3919645302852217128, 1.1541169801845814202, 1.0049721697276872145, -0.15079906805240292367, -0.22229576303322604036, -0.10096312064767669348, -0.02310564261422497559, 0.011064709635236331248, 0.020268007364360771799, 0.016922248863526084817, 0.0093229513552951157052, -3.4420991020667532747e-05, -0.0092002783139676375807, 1.526682559221008928, -3.1058982382437338998, 1.1293653054133609981, 0.88415569448323860779, -0.17264385235961068887, -0.20141981361561006603, -0.080332472435365842589, -0.011830835402161484285, 0.011854733258289487999, 0.015246967821886649175, 0.012112229594687542664, 0.0073507599360976214053, 0.00048274029026869894532, -0.007369962323109194649, -0.12824834768677609298, -0.26735725851596292557, 1.2275615053012087774, -0.30523104504499382061, -0.65218654143534238177, -0.18992181787512621427, 0.094258864581171691865, 0.16735605915953960521, 0.10515361349580619466, 0.023058432908156487745, -0.018632880597998901367, -0.023365729244185569347, -0.01400027398431775115, -0.0028691155310347168407, -0.22018411454097747137, 0.61155686400670172809, -0.26322587307746214691, -0.52850148134255303667, 0.34782847911791908713, 0.24158322349279406915, -0.081575765310518119633, -0.11986464086505609072, -0.043844605628425549482, 0.016229081649189392583, 0.03128999786981823128, 0.020208668876419567939, 0.003047799746151903192, -0.011468265636134865326, -0.059015283411419421078, 0.51088634519799047684, -0.74090550017822198736, -0.13832560091952619818, 0.6244537953077368142, 0.057030567465796384863, -0.23485197340456570592, -0.10265179059510408377, 0.025516824311357187011, 0.051142861725162302999, 0.024119689570133481804, -0.0023984520918523370703, -0.010446679298993547316, -0.0034185564895123899364, 1.6389303592183936953, -3.2267140831085869834, 0.75361611479363876942, 1.0552606682138525596, 0.074809398373666721738, -0.1352836485760839369, -0.077788939516076557878, -0.050534587445507810066, -0.041618579349873237316, -0.022150866227407722819, -0.0032376105517245715518, 0.0088988517844241256383, 0.011992959208090457277, 0.0076887501740787560223, 1.4928377587620551559, -2.9508352698518161006, 0.73483897211098991864, 0.93930738511303424332, 0.048297050315592580938, -0.12552424001090234373, -0.068588347093589713266, -0.043392315169198523994, -0.03633993594476170319, -0.018873534489805665615, -0.0014540688901066214228, 0.0092180413674317722428, 0.0108234666631698076, 0.0053059938062319355473, -0.69573938439470084383, 1.1587065870117145838, -0.033924147410199806396, -0.83696162488554703973, 0.099244663325082735361, 0.32484726727375690114, 0.10847527682798775717, -0.032148415665029579646, -0.069966280036449474133, -0.048694897924835126024, -0.016311866323327654099, 0.0043465739908060919625, 0.011307285139646263555, 0.010008896806567737711, -0.65995314406476768276, 1.1634399440782070556, -0.12442429433008661555, -0.8218938601923648779, 0.16851341599063882071, 0.3305231738894406357, 0.072755832958060345783, -0.053916642934438471779, -0.067374311523931265655, -0.037135923337885949624, -0.0080581637483085229168, 0.0065321430958098112771, 0.0095424279362595846449, 0.0069028223501645009919, -9.6845635103133265176, 21.785095956249300286, -12.102904940975275139, -1.8163939997791656022, 0.75489509753106076584, 0.80124205374400880864, 0.27521234882327649807, 0.0023640850838580430501, -0.012280000017698026321, 0.013615661297542137886, 0.0035980591403975219732, -0.014864163575285345084, -0.018711206762992236424, -0.0073140987741517893284, -0.023843148684722443448, -0.05483525076241817231, 0.298797948894512444, -0.03781115518598660713, -0.144849529453154241, -0.14277710319499101188, -0.0010214203563748639608, 0.066465924508195939335, 0.049983185832521391523, 0.02103501342788133277, -0.001015589079839236469, -0.010089751520399137946, -0.0089692048354813001809, -0.0028883716008325536692, -0.52629639911766423754, 0.83565122996423490953, -0.71618979552829153956, 0.78017584515811222445, 0.02109248758827318368, -0.36075087757480350792, -0.2135673679181668283, -0.017159902541753105365, 0.094057231343896027798, 0.094255673781904941722, 0.04885831836103764203, 0.0025274059179955562365, -0.023359227058295843615, -0.028767447392393924455, -0.36261323609585932592, 1.1951867182505790943, -1.6042061160867360226, 0.84668777962678154481, 0.2823534705446252846, -0.3823988655873555853, -0.10179911292782599375, 0.086332641819769517788, 0.056634749659095581786, -0.0028705289939392579501, -0.012186129276402075244, -0.0024773220468729205933, 0.0018842766941745035195, -0.0016063179176129247817, -0.15539615058132891234, 0.14527803832380106752, 0.067936139077724555513, 0.18691696200009144668, -0.34262004490963732239, -0.014622500777336426211, 0.10291662412294139139, 0.03123650428450889735, -0.0092630446921215334882, -0.013457544395428342951, -0.0034159339387447428676, 0.001440438738108543084, 0.0012780237893788735579, -0.00051528134156847036837, 1.5248153690405237271, 0.68129426949604432462, -4.4871791425221365657, 0.46731798523256112832, 1.5765391716319974158, 0.57273917860097023347, 0.11907357224164732967, -0.097052843891152429379, -0.15591728817313355027, -0.13139807979527023951, -0.083346779480954924502, -0.034227002681443692567, 0.013472861953286365683, 0.056007971934619475707, 0.46013592872112485388, -0.39701939556179488555, -0.11955975654463149582, -0.2039488387344433884, -0.15421983388308183049, 0.40387462978896021548, 0.20504575898250027022, -0.031748577965949549562, -0.090778298906510179744, -0.07121019505949764683, -0.025960417260839363518, 0.0042180636117366519799, 0.014307706111847820832, 0.011857913528635601746, -0.4405297603633520076, 0.29327026314919513172, 0.43576240629749896094, 0.20082172701679867899, -0.49969095518257050648, -0.2016912520751132365, 0.013324122492942840068, 0.13970172236743708116, 0.11779056361910031359, 0.0063461519823316808048, -0.048133394792354802783, -0.037242620800737348541, -0.0058946309122854248161, 0.018731995697159147668, 0.075882333147027025699, 0.85811308083373183386, -1.9779307959734873101, 0.74120797414963657257, 0.59689565413590628395, -0.10757605882283877152, -0.16049574689468029676, -0.059110553686180865307, 0.018944156884835531424, 0.021314342136549872114, 0.00029608780095481028807, -0.0058160920707015237655, 1.6103931839638889923e-06, 0.0036857352565958512169, 0.91280978707830895402, 0.65539497812014824163, -3.3631917168670630858, 0.61146738559997104723, 1.2956481800176924413, 0.34005065658750877056, -0.094836761953116854862, -0.19199072732324989898, -0.14529249961832552129, -0.06136983290502172339, 0.0051176247015095200493, 0.028903731939903771347, 0.023263174098417771507, 0.0044154462303614694374, -0.017557088166128217294, 0.092442607631713080707, -0.12825534304426214338, 0.21595738247515028352, -0.34652140877017151288, 0.098056883485264423328, 0.15292140495446823745, -0.0041189502031540566399, -0.042450942308031210104, -0.026548173227231748605, -0.0062017234601030755622, 0.0027369956358910636426, 0.0051185901771479760947, 0.0054050340111841746471, -1.339177878797991994, 1.1161017962208037435, 0.47330676634886970078, 0.94885420756062877068, -0.56188360733777853451, -0.78778859482143559489, -0.27182637354349986625, 0.096297366998299188667, 0.18889182165673734648, 0.12508137120428727829, 0.047671300916624810562, -0.0006842333401778366575, -0.021719290161929728777, -0.026326162667853880012, -1.2676678254466946516, 1.0299104787969641794, 0.51541601277415827198, 0.88094584340277504175, -0.59604009626108267472, -0.73525747580567113459, -0.23187580608021504669, 0.10061027757719495646, 0.18166801289277081932, 0.11832340698112603361, 0.041945794706850891842, -0.0051011357441544391556, -0.023193196452880385228, -0.023357661278824940615, 0.24653778062617134892, 1.1541169802595399041, -2.9829426916599062203, 0.9638118129937360834, 1.0466518301431819715, -0.095566323439959757757, -0.21540027190770036358, -0.1159438625847492299, -0.037916525545505765094, 0.0042626674030722790446, 0.021522815932491796281, 0.020555943597941720574, 0.0093229512886194194909, -0.0081172028705021308931, 0.17165291938764226654, 1.129365305352673543, -2.7528507004096787192, 0.93493014524197071147, 0.91994697780360645822, -0.12334704744949102906, -0.19276860384676916538, -0.091583768525262482241, -0.024549662637575318802, 0.0056129359223068909027, 0.016140131013545461897, 0.015105049350124100452, 0.007350759942774966714, -0.0058432289217328833852, -1.1211852591556374925, 1.2275615047696284421, -0.19435810303567660484, 1.0334586747520773109, -0.35948270199981890993, -0.71967628054515786484, -0.22503643267624467983, 0.08330884656140090383, 0.16830185116848547788, 0.11375745550446646204, 0.033950685022210082276, -0.012164370779740595746, -0.023365728662839035934, -0.0169425123899975133, 0.22984406262840684221, -0.26322587256104368336, 0.01646942092607958652, 0.043845152744128632061, -0.35652212709636271759, 0.32837293685324125958, 0.22431748384726696299, -0.092140157405087450004, -0.13054932221367168688, -0.053105571020005286342, 0.011588365276590073538, 0.030930624845948745394, 0.020208668686277742044, -0.0030148876499878412769, 0.46722925016638616746, -0.74090550050069770016, 0.2924127995682189729, -0.18646262814596373247, -0.11252936624670611954, 0.50122067855999374864, 0.026506781662700092811, -0.23340193719655893956, -0.098763162783087041219, 0.028136632488487023362, 0.053629645840017364788, 0.026325954886426954776, -0.0023984523520991238646, -0.015483413119546235448, 0.66390775621418507413, 0.75361611523900751131, -2.9216274020967358815, 0.55571027639765702144, 1.0481456393558306051, 0.16513009542670742391, -0.077248027782718470258, -0.059814000310911330516, -0.050891007322618433961, -0.051451434454624744852, -0.03331612720520990556, -0.010416592773044169168, 0.0088988518075067586655, 0.021724333248949831421, 0.56722175152971587853, 0.73483897264566344631, -2.6847961186227657926, 0.55321257566593895483, 0.93820607607838524267, 0.13134703130619043709, -0.073934631665443673043, -0.054755940500013769445, -0.045806410928981099173, -0.045967067409042737525, -0.028571083939948137276, -0.0073216954748073180548, 0.0092180414116736827251, 0.018397726973457265848, -0.065282657012576766897, -0.033924147237429584656, 0.40196130473432895025, -0.0010553039254343390269, -0.74544809297038916451, 0.1051925266122223368, 0.34676675610994611709, 0.12909228425765445492, -0.021828079496008796767, -0.069417031634472811663, -0.053459870884518104206, -0.02104926180190432039, 0.0043465742740373871505, 0.018245519340175295891, 0.0040885709972143090968, -0.12442429408047671335, 0.40175944312072747744, -0.026497485311474368308, -0.7096243095729997119, 0.16490018532878278346, 0.34328137546946008163, 0.089200774059966056462, -0.045751683573880902456, -0.067861571621343236282, -0.041860254674869090685, -0.012202560386824721009, 0.0065321433978365473633, 0.014487246628185549216, 0.65881391088810903334, -12.102904941724803578, 23.713981554714987254, -11.823208800057496859, -1.9757781507811662536, 0.53187761433357449992, 0.66937265867439010858, 0.25468426986255066069, 0.027734172015268974965, 0.012048095775848590219, 0.028422267728829050254, 0.010933769379076238581, -0.0148641630305504438, -0.029315973837474487707, -0.35988107297961735576, 0.2987979487448619853, 0.058724507990764618359, 0.33265469522441887262, -0.11176467116215606235, -0.17846788668469862338, -0.15857106643430521342, -0.014758798979865418027, 0.066041103524108094036, 0.055939585335522429288, 0.027687158049850595781, 0.0031926214457015478047, -0.010089751573642601404, -0.013076591013858822332, -0.37827702878978258783, 0.075976924318784472279, 0.78017584532130546116, -0.84861286272634062211, 0.69963731733333212048, 0.016767282923291382257, -0.34244430639763356883, -0.19463569561826005172, -0.0061625343450246129404, 0.09996258274225314211, 0.097785737292561569767, 0.048858317842478564408, -0.0074947119151440502044, -0.053208467289298996827, -0.29016350645934291252, 0.27128296347478264572, 0.84668777952486407035, -1.675165999152232521, 0.89217242163706234059, 0.32402802408929781697, -0.38372962428464585383, -0.11334068138127309511, 0.082463444950426248448, 0.058817173718276161143, -0.00079558892857948008623, -0.012186128746445380397, -0.0037895754024857013394, 0.00078519562853409622261, 0.15045226312661746926, -0.39848312189439205611, 0.18691696113983333194, 0.11789530552398093388, 0.20727375892864433049, -0.34701979533436522196, -0.027688915504766646591, 0.099706149269768104548, 0.033261208067326748783, -0.0076437696454525715095, -0.013535915584982153789, -0.0034159335312251961139, 0.0019332415319473037364, -2.339852047447757331e-06, 0.15693704585602921697, 1.5101127170919654841, 0.46731798708382249474, -4.4486677346834282432, 0.45409027791581985101, 1.5934829076241674439, 0.60500493170248392794, 0.13845822042844960209, -0.09843617486518421722, -0.16891722876745673965, -0.14358043560382938164, -0.08334678108650679651, -0.013652015167167503781, 0.059498185094873902445, 0.39178145979137285471, -0.21215561653504272366, -0.20394883820314257949, -0.023862089530378452484, -0.17048021702539908895, -0.16849885889601845523, 0.38078240938875262955, 0.20002085472355971008, -0.029525081566475173955, -0.089130604378980599067, -0.071596373092786880221, -0.02596041690387330958, 0.008025449229686124758, 0.024135405256080013497, 0.046216864684277110531, -0.37561815220245020619, 0.20082172658521046937, 0.30001464958227419544, 0.25100690720637763853, -0.4112531896655585717, -0.1796437460130924646, -0.0055133149529584457235, 0.11816953003175886816, 0.11020226069279419334, 0.0067551111409596626203, -0.048133394619383908231, -0.034301595587089506179, 0.01694868825410530272, -0.33117366543576587379, 0.62879333216432231612, 0.74120797412030992035, -2.0829704941241913296, 0.68526216143804408176, 0.6258156749508325678, -0.078118521541283569842, -0.15794146442950882103, -0.067041242522430322115, 0.015952157175138606537, 0.022465056377444694163, 0.00029608719303313472352, -0.0071251229853647943258, 0.0059248018938093862523, -0.22371153699821000327, 1.207901275610408165, 0.61146738516922849449, -3.3338959348210548406, 0.57281647541299762771, 1.2631945900672651995, 0.33147507928865221905, -0.086833117188555603017, -0.18407869241953783712, -0.14266359355320784585, -0.062108128869954697115, 0.0051176241767360826709, 0.031172893929065463581, 0.025299763658460105303, 0.22851399961005219863, -0.43501737210099178377, 0.21595738257480473488, -0.043096747010007221967, 0.22431746734001062338, -0.35765892554825590155, 0.084387867559979429233, 0.1508203882689187203, -0.0033392581906959459097, -0.043620455526272215729, -0.028701602391977856094, -0.0062017238141556284628, 0.0061299731607096499472, 0.010215920488989539694, -0.54034813174381934875, -0.48431437956545975165, 0.94885420743673276611, 0.29524113534584739416, 0.92371041769529294019, -0.52069792744808096607, -0.77033373886775491357, -0.27522897093790454903, 0.08970056938920732692, 0.18815846019699739022, 0.12777208469005135738, 0.04767130133492786942, -0.0092814468555486029366, -0.042280162453449551818, -0.46196734634346359183, -0.52281543966798094125, 0.88094584305430101434, 0.35119179271340650317, 0.85548180159953923773, -0.55687556900407231275, -0.71669112178125327084, -0.23282135717150792553, 0.095286261263734461924, 0.18059510638374953739, 0.12024218248614003401, 0.041945794791872263729, -0.012572717548899532344, -0.042088834321496064461, -0.47810066645044740907, 1.0049721694841027286, 0.96381181294906981272, -3.0405790710913356278, 0.91075775184117879757, 1.0420993948648384908, -0.087336782554910813436, -0.21075362462726543455, -0.11369834298629451219, -0.034959545017967788572, 0.0066939458758539733055, 0.02152281589774646467, 0.016922249035618060903, 0.0024231395740261772211, -0.43767544604300395727, 0.8841556941247742385, 0.93493014554445075248, -2.8120309270225787301, 0.88857530753025160131, 0.92240860458540296918, -0.11304075740971412933, -0.18942284427575964423, -0.091110997766512671903, -0.022596109328313917358, 0.0077011648568394777681, 0.016140130909308297857, 0.012112229941419685661, 0.0028091222827722030567, -0.48605846543668651538, -0.3052310448048675684, 1.0334586742193350162, -0.35268228456796935077, 0.98468526907665177017, -0.31495394057551107103, -0.6875745733355101974, -0.22394321120680349413, 0.074036274441087332465, 0.1646304894829398735, 0.11543084839837973454, 0.033950686327565531009, -0.018632881328944857907, -0.034754356372535682995, 0.34586628179257450366, -0.52850148146323228104, 0.043845152681678593865, 0.20485649629483293488, 0.0455846872104684539, -0.40838916227251725877, 0.28842818472216874559, 0.218303946648820274, -0.085364746941235636335, -0.12587662697906706888, -0.052402929761263851438, 0.011588365396707381999, 0.031289997562011118148, 0.017552900880650607746, 0.16333952704703511016, -0.13832560061555618236, -0.18646262834125387764, 0.34672735923944569825, -0.31278642906672859958, -0.19076957685645185014, 0.49532831346549288654, 0.054787843398118575566, -0.21496506405102303039, -0.096132352265655307555, 0.026001321631763676945, 0.053629645212588596126, 0.024119690295373339556, -0.020861592556540586169, -0.16787300906485949636, 1.0552606678258855677, 0.55571027699323105598, -2.9803007941038717199, 0.54924476807509625687, 1.0711556425069084231, 0.18332100620109947897, -0.071560752031205362989, -0.063342371228392868598, -0.056060754447501373599, -0.055106713647424102054, -0.033316128209615311628, -0.0032376105009259562376, 0.028136400504874314876, -0.16735150352320937728, 0.93930738462155383139, 0.55321257609406515598, -2.7372306757513227282, 0.54846181455064313415, 0.95968108068655955112, 0.14620712421345105625, -0.070514208362709981937, -0.058468812908892421609, -0.049980374144505193557, -0.048694963804854954348, -0.02857108386044941023, -0.0014540694311155292345, 0.025924674134149301025, 0.5617050020359259932, -0.8369616249492994875, -0.0010553040509029987942, 0.5233815490952975491, 0.077200003686511614598, -0.7242976029889300893, 0.08663353148640069723, 0.33248337809382783048, 0.12574677971116401953, -0.020999027727847067309, -0.069465575937291260122, -0.053459870042792359546, -0.016311867322503253996, 0.019824927727954987089, 0.55749242654785235995, -0.82189386026110022865, -0.026497485240626277025, 0.54597142835657164195, 0.036599219395255969123, -0.70511614099996466898, 0.14180399664031156926, 0.33163183014404107318, 0.088659240728546861998, -0.044108272941757176167, -0.068001191408621181056, -0.041860254767672494536, -0.0080581638139135564752, 0.018438699977689286535, 1.5116599626122579902, -1.8163939992910822507, -11.823208801063913143, 24.509313118263516884, -11.680781058134904526, -2.1129406390927396586, 0.43004444261719410036, 0.64911638824560202732, 0.26673883432174988473, 0.03923281218285611105, 0.016516810977873403604, 0.028422268506517259223, 0.0035980589602267482735, -0.042116856837130622893, -0.23891237629044689839, -0.037811155359975569257, 0.33265469538674652394, -0.051428015266086105073, 0.30066661217551848839, -0.10484595041007255256, -0.16949845766966703531, -0.15226013369509783235, -0.014012614965915052428, 0.066110440900614214632, 0.056950598959335616001, 0.027687158651896866257, -0.001015589056047305587, -0.021624312331304681306, 0.018965735161886338594, -0.29324646962458106314, 0.021092487017485700429, 0.69963731679350582571, -0.8709641225897976069, 0.71667305590120478076, 0.046710495083352308876, -0.32233853851247984412, -0.18697816086491392262, -0.0049009355445595557954, 0.099962582660397938938, 0.094255674361951299889, 0.030972281791020035008, -0.055375946740510460786, 0.11237539904475210562, -0.45836353173939992089, 0.28235347020240680882, 0.89217242196301305146, -1.6618749543356809983, 0.88284350051489746125, 0.32041863622147348512, -0.38318793367302511621, -0.11232697610889476325, 0.08314367864143917819, 0.058817173816234462336, -0.0028705296268749891202, -0.015428831086360015115, 0.00079649138672093747378, 0.087529709879433856279, -0.0074691221316909367153, -0.34262004477242907763, 0.20727375942336026537, 0.11895010194649519031, 0.20317503150019750047, -0.34837072887690989642, -0.027932286291301358649, 0.099738395233906318449, 0.033025780230080813149, -0.0076437694167738601259, -0.013457544664965499434, -0.0037875982217346264129, 0.0022395621425700747327, -0.2958085935658203991, 0.44929435327352679197, 1.5765391701528150836, 0.45409027941300611486, -4.4673417562130639524, 0.43574397235032974462, 1.5795184304862459612, 0.59153326997393174125, 0.12663702896178333224, -0.10622344668143360313, -0.16891722970830624906, -0.13139807871320111077, -0.055786849319249293133, 0.028418776824523071106, -0.05018290694986438788, 0.38425808275924355772, -0.1542198351093983788, -0.17048021628327436616, -0.024431485378446884255, -0.19296008720308491236, -0.18507277516442444121, 0.37468754936102971476, 0.20069222664547992085, -0.027872831059677284637, -0.08913060389245412718, -0.071210195932562989074, -0.018784014029421081809, 0.030555171766623326346, 0.24151645392782361554, -0.2052056418466109422, -0.49969095445636241282, 0.25100690687721843863, 0.39279459979585357177, 0.28506011332344283193, -0.4209440098631918592, -0.1999229816084243383, -0.017356953026897967846, 0.11540525995290927541, 0.11020226088557837085, 0.0063461518614129552515, -0.044263502453135358805, -0.014799138107568292691, -0.13795609488221585726, -0.11518826211766218515, 0.59689565419271817248, 0.68526216155717789569, -2.082682047469775366, 0.69374633160664067599, 0.62557534459551611405, -0.081338167352095383511, -0.15776986965365696891, -0.065561614788437860835, 0.015952156604477393809, 0.021314342904848675164, 0.0013895102668396205458, -0.0004930091124938900532, -0.42460837718224553328, 0.25034978180087935629, 1.2956481805578752287, 0.57281647556038595148, -3.4033201503073149397, 0.52106686300086413333, 1.2501792062455718835, 0.33808153037631294424, -0.077774975847612268232, -0.17993713042101491806, -0.14266359417824170386, -0.061369831919540644705, 0.013045316002430955987, 0.052617894812125871884, 0.032818131883365589396, 0.11471805806909990411, -0.34652140898508099198, 0.22431746773397398531, -0.046994104740239739382, 0.21803364006842915201, -0.36210368209558546626, 0.080982995454268000279, 0.1487680153431198693, -0.0046533754530586578235, -0.04362045596203095732, -0.026548172731444143191, -0.0014114339326530256878, 0.013794979126123131652, 0.22057259700526254953, -0.72746787249002464737, -0.56188360747390853067, 0.92371041736993930726, 0.32078161852908687601, 0.95782218949564346122, -0.49629014998555254223, -0.76217862886935394684, -0.27603367411467999437, 0.087893359668955187147, 0.18815846007325789935, 0.12508137175770461802, 0.032734616891111442682, -0.045639857910739080327, 0.23442780743828336987, -0.66967548722398528849, -0.59604009574295946372, 0.85548180109093030499, 0.37795992702006248054, 0.8918243717755424127, -0.53153838354948268474, -0.70888892885293930402, -0.23456463690658374555, 0.092720492782126628928, 0.1805951066005785921, 0.11832340734075944, 0.027752720767096492049, -0.049750693809303789261, -0.26473577550641202993, -0.15079906812715526687, 1.0466518303245679888, 0.910757751844570973, -3.0887361296508659692, 0.88543298969211636784, 1.0404676222281379339, -0.079410353087962806673, -0.20223708103987794726, -0.10919826216775029015, -0.034959544833148566145, 0.0042626670274843320163, 0.020268007584432291834, 0.02099354929720475757, -0.21465990446103400124, -0.17264385209567068014, 0.91994697740387487706, 0.88857530750514246431, -2.8499374019482837106, 0.86932016394065192433, 0.92107334838346877071, -0.10726895104058656316, -0.18280204015722950617, -0.087435532674778937423, -0.022596108983267467613, 0.0056129358256988065232, 0.01524696777249586431, 0.016441742688327754679, 0.19654205497349333531, -0.65218654144421528418, -0.3594827023285573353, 0.98468526938640099644, -0.32777725087377890256, 1.0275837199405672973, -0.29111410765707163018, -0.68441877433345010839, -0.22763846391900691146, 0.071693943095769416263, 0.16463049051875777717, 0.11375745401217712827, 0.023058433803185972299, -0.049100135637102554353, -0.1072151761847126572, 0.34782847924879001233, -0.35652212695628165973, 0.045584687160280648599, 0.16699622613913278624, 0.011743856684722974934, -0.42343410166270428752, 0.28845464825835326339, 0.22434864655549011103, -0.081427955620992947816, -0.1258766273577831829, -0.053105571311631939724, 0.016229082007532820431, 0.048011220302180815378, -0.32917564160342710489, 0.62445379480412188666, -0.11252936585748865572, -0.31278642930031003155, 0.28452908873578747206, -0.32194105376685455688, -0.17572047124715423072, 0.51180892929743904141, 0.062949960751606262588, -0.21404914167198565988, -0.096132352223588138362, 0.028136632822735749282, 0.051142861369345725409, -0.0014496283686334546052, -0.2005014723555716416, 0.074809398602718166216, 1.0481456395978303586, 0.54924476753408890684, -2.9900491447745092266, 0.53320338998571625844, 1.0625719189153861599, 0.17827661713674769484, -0.074574297619762211498, -0.064982183207470844977, -0.056060754702906709368, -0.051451433678564521823, -0.022150866440592720252, 0.019378582166262311309, -0.18175946160169281773, 0.048297050761294722798, 0.93820607619701434921, 0.54846181423122186693, -2.7463785334923431769, 0.53244288513900095428, 0.95071923808536229483, 0.14148688444851975632, -0.072594951873898611039, -0.059346335782850687113, -0.049980374589683324027, -0.04596706763611208002, -0.018873534082960594094, 0.020197126413563769309, 0.27982359933718004008, 0.099244663224927573419, -0.74544809219057617256, 0.077200003020518659369, 0.56572652563213499111, 0.079676957138347492826, -0.73913234935097038569, 0.073062992963551204317, 0.32686358833516110378, 0.12490900547819439037, -0.020999027342410776786, -0.069417032518579196987, -0.048694897063037097251, 0.0040854637603378292712, 0.21583222440332328063, 0.16851341580858664182, -0.70962430918509877831, 0.036599219038753232069, 0.57232097300744066981, 0.032988063005423146068, -0.71924043148129745973, 0.13104064070404014553, 0.3280515246504518978, 0.088489048141425497018, -0.044108272989125035801, -0.067861571636751438774, -0.037135923132044529327, 0.010454603873935762112, 0.6559228465774362693, 0.7548950969396498456, -1.9757781499986797336, -11.680781058774551084, 24.629321161224876136, -11.645952114215406326, -2.113119240666118781, 0.42361394433762417089, 0.64617048257007769774, 0.26688221008571128001, 0.039232812804443233101, 0.012048095482980050036, 0.013615661410765348208, -0.030572901016642083094, -0.012450342280751917901, -0.14484952961169139662, -0.11176467140217496987, 0.30066661284323537018, -0.050798472579697641349, 0.31072634236899659976, -0.091170755137745806973, -0.1623400485558907369, -0.15076101342379558123, -0.014263428070026517441, 0.066110441153250210089, 0.055939584558840717288, 0.021035013883407723684, -0.0210327229702830798, 0.20673346063248157489, -0.16940928679027580928, -0.36075087668613381453, 0.016767281982176383826, 0.71667305567899008967, -0.8453307963488785326, 0.73715852147315641396, 0.05766069730691988221, -0.3191168455496952161, -0.18697816049936680227, -0.0061625336001454754029, 0.094057230886734802255, 0.075010641246409701099, -0.013423427388003835586, 0.12480395692314270872, -0.1117307947159863063, -0.38239886556265778594, 0.32402802364217175857, 0.88284350038750392109, -1.6721180104623811769, 0.88283395092203209575, 0.32379018440471146745, -0.38168001599388295952, -0.1123269767944767572, 0.082463445444209068635, 0.05663474955976777192, -0.0057680884042133782549, -0.01088318982974283318, -0.013573466223125408084, 0.11248775861008775134, -0.014622500656626064366, -0.34701979608236294128, 0.20317503165430306211, 0.12109352980209553419, 0.20591364362935343779, -0.34785343729233203502, -0.028667181921365449127, 0.099738394522415119159, 0.033261208556197401076, -0.0092630451210682657887, -0.01566845328124120626, 0.001533987558551515697, -0.37170738041491535597, 0.015121470386656074916, 0.57273917867849000185, 1.5934829077586041279, 0.43574397310593937505, -4.4992202047869698234, 0.40226664109219162357, 1.5559461055514745009, 0.58069071082906376091, 0.12663702791951397808, -0.098436174740847856568, -0.15591728793775425088, -0.11544292784268246621, -0.040744976060058855916, -0.19944094410738832934, 0.17927038297008376921, 0.4038746305013814486, -0.16849885942389089544, -0.19296008702213440777, -0.037808149572952871875, -0.20044378572798227101, -0.18821389679954977914, 0.37440649230694744709, 0.20069222594945029803, -0.029525081351169775445, -0.09077829880163003351, -0.064482475714702960912, 0.013853104327817842081, 0.18706392489793113865, -0.042492175057687389672, -0.20169125226554404651, -0.41125318939061827894, 0.28506011305739631378, 0.38648889281999254663, 0.27047964134476126796, -0.42889088666386554483, -0.20213229720005670975, -0.017356952795178320903, 0.11816952992595980421, 0.11779056361238768313, 0.011163703005371524921, -0.069887985693900409045, 0.022969776474820503137, -0.18872236822973459391, -0.10757605840388308194, 0.625815674547792411, 0.69374633192704615592, -2.0934766730666489742, 0.68355683079936180313, 0.62484735182717210389, -0.078788171799702100273, -0.15776987011406976014, -0.067041241721517833008, 0.018944156212081435936, 0.029326299723898680699, -0.0069830994337509852063, -0.28783083081830057592, -0.12359304159703547732, 0.34005065663687317201, 1.2631945898908087944, 0.52106686355539644229, -3.4328641752738544568, 0.51137318022180011656, 1.2516134908319989094, 0.34153564397812963271, -0.077774975921738404017, -0.18407869217737890533, -0.14529250018169379, -0.048081236662732444731, 0.066123663718089420449, -0.087970687129241831315, 0.15628613609498470671, 0.09805688373416443504, -0.35765892579418723241, 0.21803364032410704754, -0.047909991381152990109, 0.21634403729444656728, -0.36505101738201950479, 0.078792151446867020326, 0.14876801517198168234, -0.0033392577056252582997, -0.042450942674382687037, -0.023472773906359014479, 0.010962842490354392927, 0.39963918963579059263, -0.22264751898736878655, -0.78778859442359983056, -0.52069792776876300255, 0.95782218969173471557, 0.34565601938103518442, 0.97490157634437724088, -0.48766216598315847985, -0.76044856821431172644, -0.27603367389840194335, 0.089700569326038592788, 0.18889182093933920559, 0.11251448392751535466, -0.01457043576699631679, 0.38115822087768597148, -0.18506236189193442376, -0.73525747628745141693, -0.55687556826699258217, 0.8918243715077374123, 0.40463907063872339709, 0.90861445637713844636, -0.52439235484778545793, -0.70828634867592987767, -0.2345646361488754883, 0.095286260506297001949, 0.18166801310283289816, 0.10446852321162930499, -0.023103790827463457913, -0.059967181827590247678, -0.22229576270188258769, -0.09556632347954582829, 1.0420993947612962049, 0.8854329896026829072, -3.1062068805405154137, 0.88210156714860321614, 1.0463795504990274221, -0.073885725949862229855, -0.20223708097626016955, -0.11369834301809171595, -0.037916525388068604563, 0.011064709287338998778, 0.041625597576745311523, -0.036766173996609911867, -0.20141981345071563458, -0.12334704729874271245, 0.92240860446643324622, 0.86932016372622356037, -2.8655814860778829889, 0.86532618330628507941, 0.92568969841676762567, -0.10260833318479951892, -0.1828020401178753751, -0.091110997893324288666, -0.024549662898525427546, 0.011854733197057987226, 0.031078434907333547105, 0.34890001988930119392, -0.1899218174346815613, -0.71967628009942108402, -0.31495394098393519888, 1.0275837195901706966, -0.30559629576400604289, 1.0369724295751572285, -0.28715459796985526308, -0.6837101761385698806, -0.22763846409326357612, 0.074036274562837137148, 0.16830185156112009714, 0.10515361273266446429, -0.031695215642516184895, -0.24753902648438050327, 0.24158322320552827467, 0.32837293683037416248, -0.40838916199281311004, 0.011743856803213567272, 0.1567605405200227231, 0.011903947977695285634, -0.42061812251456481837, 0.29088118210791030771, 0.22434864666650763954, -0.085364747070332258616, -0.13054932174168809578, -0.043844605900266377108, 0.06701015478009877957, -0.20669977495871874607, 0.057030568284471470919, 0.50122067779366774598, -0.19076957610321476633, -0.32194105410840173631, 0.30108283764488286138, -0.31009262243299784423, -0.17205360354213231067, 0.51229813704537729979, 0.062949961011515689902, -0.21496506465970738331, -0.098763162136783128053, 0.025516823788406792217, 0.052880059485114622175, -0.12898759642836180039, -0.13528364894712158195, 0.16513009464317138231, 1.0711556434177609276, 0.53320338997659766367, -3.0125483067637119916, 0.51599693764705079069, 1.0545377718996435945, 0.17605856194717231511, -0.074574297038056736042, -0.063342372209218344992, -0.050891007063412097011, -0.041618579070210000304, -0.0092623596453136301798, -0.11421072391643645461, -0.12552424060641922954, 0.13134703093560620712, 0.95968108121427486701, 0.53244288510639958822, -2.7680543738442442425, 0.51684288174564041451, 0.94427752753275784325, 0.14017137156783254781, -0.072594951785728681348, -0.058468813147546386255, -0.045806410666455701919, -0.036339935984073659658, -0.0044315556934121950669, -0.059882841620160859319, 0.32484726721377449321, 0.10519252567380310559, -0.72429760195143177981, 0.079676956631996587177, 0.56063965812776750397, 0.072866038053521847706, -0.74493903834259567187, 0.070002706067365527698, 0.32686358895104505873, 0.12574677883061111117, -0.021828078668776798588, -0.069966280340535327853, -0.040929475494988612483, -0.10201640630365377271, 0.33052317385931101468, 0.16490018494804647986, -0.70511614080515250258, 0.032988062999954319288, 0.56769312592336773626, 0.02802784508305772998, -0.7237153294864574482, 0.1286371066095339244, 0.32805152499634832042, 0.088659240520552268228, -0.045751683550292167324, -0.067374311509083170679, -0.022590086427497831079, 0.085685715589299951689, 0.80124205435075401827, 0.53187761442493086683, -2.1129406391249716535, -11.645952114385366372, 24.706729597071088733, -11.60587645643693655, -2.1067689060715388827, 0.42037621098099897665, 0.64617048293098866552, 0.26673883446682611353, 0.027734171828416420702, -0.012280000499900366878, 0.0024129030697672424269, 0.13124460053433870121, -0.1427771026617080663, -0.17846788660572912577, -0.10484595079596607803, 0.31072634232233320439, -0.038548299598618711304, 0.31896594941742867357, -0.087094409921645757344, -0.16111454117801665697, -0.15076101372313879811, -0.01401261456254776494, 0.066041103858590355369, 0.049983185377142413941, 0.00024547936155526494696, 0.18715382099670418992, -0.00024212413512003736621, -0.21356736838527304312, -0.34244430541503750876, 0.046710494601546879989, 0.7371585218478862167, -0.83517159005347807454, 0.74069608761539895703, 0.057660697210919938105, -0.32233853848168059208, -0.19463569586382714616, -0.017159902781211523737, 0.085406638713513283179, 0.079183024672285620693, 0.013291670681610620597, 0.10398504848140385781, -0.10179911257020135551, -0.38372962383619468874, 0.32041863624630312346, 0.88283395047841373149, -1.6697889924778572812, 0.8845163256069215052, 0.3237901843329524243, -0.38318793320202221286, -0.11334068174589118083, 0.086332641747399282983, 0.062652755370364229814, -0.025012970262846434544, -0.028447715165762003237, 0.034500188252788763221, 0.10291662403717094509, -0.027688914795753721831, -0.3483707291162977393, 0.20591364339596929245, 0.12282993607685196991, 0.20584221949137526675, -0.3478534374182267741, -0.02793228571309880684, 0.099706148769146824007, 0.031236504511877583729, -0.011671491416911511374, -0.010750453923803722847, -0.27726521629870726215, -0.16129180532252071489, 0.11907357193503147197, 0.60500493135458155436, 1.5795184310115302218, 0.4022666413198618951, -4.5297420521026525719, 0.38607547612331527542, 1.5559461053115684059, 0.59153326985735510402, 0.13845822016933931331, -0.097052843642350963838, -0.17388837650182706329, -0.14695275613116570712, -0.14066429428930846135, -0.039905179032688198326, 0.20504575940813049573, 0.38078240959362980034, -0.18507277517904438535, -0.2004437859068973482, -0.040675847774802542656, -0.20141376529506746507, -0.18821389724024006895, 0.37468755044048807079, 0.20002085338516784052, -0.031748577333633641595, -0.089713308056979912397, -0.046526029225053576466, 0.013105163067418396788, 0.1196830499160741329, 0.013324121918005060392, -0.17964374582355174814, -0.42094400942959692147, 0.27047964102111643481, 0.3805572482533452483, 0.27012518270502017259, -0.42889088625489174555, -0.19992298203743716356, -0.0055133149606459361247, 0.13970172238904848827, 0.11876458164550625363, -0.087934875128522133059, 0.047318021942974845817, -0.069651664104406263434, -0.16049574795354590306, -0.078118520663962839889, 0.62557534388044622187, 0.68355683121107169153, -2.0973794061778319175, 0.68456828762696775303, 0.62484735166074178991, -0.081338166805540221271, -0.15794146505943620906, -0.059110553211103721016, 0.026562401554237190238, 0.010682933474654250292, -0.10064907863091118734, -0.18429450443787520619, -0.094836762675441221848, 0.33147507943871579172, 1.2501792057471203812, 0.51137318074732940332, -3.4376314697286542987, 0.5104612511906130301, 1.2516134906093563384, 0.33808153054410161609, -0.086833117150123387273, -0.19199072692722579658, -0.13080746983545959994, 0.025027513950822699584, -0.057897279666892974281, -0.0077659434803402900907, 0.15292140490576056067, 0.084387867557824292053, -0.36210368220772942704, 0.21634403736582627498, -0.049914128148426911702, 0.21439503800046258442, -0.36505101746469920121, 0.080982995877152186281, 0.15082038760403473909, -0.0041189498385594894292, -0.04394683433186725996, -0.01104391055622461551, 0.25870849167575765692, 0.12824546445524764859, -0.27182637408340593055, -0.77033373802314442091, -0.49629015128523101819, 0.97490157676564692668, 0.35620600042763639337, 0.97949502027855006592, -0.48766216645183879663, -0.76217862824442761394, -0.27522897091322473528, 0.096297367482116674742, 0.19298669495451545308, 0.083389154550902150342, 0.24009180645014782662, 0.1262607470595133452, -0.23187580558639547812, -0.71669112275199631235, -0.53153838328171099104, 0.90861445628540338237, 0.41386724537750607666, 0.91204340384799320596, -0.52439235452586840935, -0.70888892946444936882, -0.2328213561045304747, 0.10061027691535837092, 0.18138452549755665344, 0.07045941203921089746, 0.029483399107834210723, -0.100963120559743047, -0.21540027179883006192, -0.087336782788370093122, 1.040467622335544462, 0.88210156713992171618, -3.105309103633177692, 0.88488014368311285196, 1.0463795504515469581, -0.079410353022847254834, -0.21075362458634383511, -0.11594386253521178631, -0.023105642774046922933, 0.051561628322068106534, 0.029074394563125226937, -0.080332472452369782512, -0.19276860362827552553, -0.11304075732196069981, 0.92107334838000010091, 0.86532618322719767612, -2.8653897909376517816, 0.86764377312467233505, 0.92568969825705060828, -0.10726895088379607979, -0.18942284439318829437, -0.091583768193205170038, -0.011830835779166992694, 0.040150819679483001901, 0.22542563324387812629, 0.094258864320441268503, -0.22503643214298260644, -0.68757457366389385101, -0.29111410737059617571, 1.0369724290208877093, -0.300608298368804705, 1.0398162179697474361, -0.28715459809152477089, -0.6844187731762162441, -0.22394321216062409996, 0.083308846685802628862, 0.16735605905040754071, 0.059676485720213272301, -0.10281788596981437933, -0.081575765369867048182, 0.22431748397163181474, 0.28842818454776009141, -0.42343410167291906099, 0.011903947919375245865, 0.16029538955908545494, 0.013886312333013239861, -0.42061812198925463413, 0.28845464757068178807, 0.21830394713248069127, -0.092140157862479296624, -0.11986464048455799936, 0.0284553114870057626, 0.042300127189863569177, -0.23485197416967729112, 0.026506782651632032621, 0.49532831238371283211, -0.17572047094448073334, -0.31009262231143008837, 0.30365088679084989431, -0.31127158628807211072, -0.17205360341076461617, 0.51180892867397587143, 0.054787844790186522592, -0.23340193862072769471, -0.1026517895712737577, 0.10394510523849158357, -0.093886522338005776178, -0.077788939950185487238, -0.077248027115264503162, 0.18332100578780180511, 1.0625719190442766138, 0.51599693776723154492, -3.0253074190868907145, 0.5107391029486247902, 1.0545377720895712237, 0.1782766160965104929, -0.071560750983801171832, -0.059814000856677679296, -0.050534587065040062481, -0.053528145739396813807, -0.081781250997914889389, -0.068588347017120146476, -0.073934631689081209416, 0.14620712402670524299, 0.95071923835682592419, 0.51684288176434267648, -2.7787636356040388286, 0.51281858269074964696, 0.94427752737423764628, 0.14148688418991223203, -0.070514208075520695695, -0.054755940608043951467, -0.043392314717507346478, -0.044638061214342751593, -0.15168251260732393271, 0.10847527670302901937, 0.34676675661546535467, 0.086633530378773263014, -0.73913234794049853527, 0.0728660376105544072, 0.55805079431284920766, 0.071317341359997940553, -0.74493903799885174966, 0.073062991800572393841, 0.33248337913131881249, 0.12909228351802537471, -0.032148415230735541415, -0.10989692413849834474, -0.14777484572429319676, 0.072755832854444993374, 0.34328137547563375431, 0.14180399692787973298, -0.71924043160961015353, 0.028027845242287241501, 0.5661325368626171084, 0.026811768216946293975, -0.72371532936349824983, 0.13104064004447066316, 0.33163183030775561511, 0.089200774547510019996, -0.053916643377436110351, -0.086159393717699206405, -0.018686591988680341381, 0.27521234928550653009, 0.66937265841858595294, 0.43004444305262812254, -2.1131192411782349083, -11.605876456447411726, 24.735234303797366096, -11.598158310310285657, -2.1067689057700134114, 0.42361394450573608417, 0.64911638788487646679, 0.25468426992091497407, 0.0023640846805191963534, 0.012596996840828414768, 0.11805360269305099907, -0.0010214202540459497385, -0.15857106643425819548, -0.16949845807767782713, -0.091170754692406985598, 0.31896594910691605795, -0.034555330888705840242, 0.32065892098708614899, -0.087094409699538258285, -0.16234004838181562147, -0.15226013392165571592, -0.014758798841772899704, 0.06646592408456901413, 0.049814018045158343917, 0.079183025822898531199, 0.085406636029551868949, -0.01715990169457524972, -0.19463569534898703117, -0.32233853817617663662, 0.057660696707948148887, 0.74069608811671971971, -0.8351715905245342686, 0.73715852224670519455, 0.046710494326108957519, -0.3424443062154948203, -0.21356736734189629701, -0.00024212431736820981152, 0.18715382089836068991, -0.025012970214999180607, 0.062652755270411364386, 0.086332641755019451368, -0.1133406816083267904, -0.38318793344046964133, 0.32379018464604486516, 0.8845163252924116426, -1.6697889921791329026, 0.88283395015238297354, 0.32041863664417985502, -0.38372962430204504702, -0.10179911216086555414, 0.10398504825807357299, 0.013291670737436802816, -0.010750453850077602153, -0.011671491702489716555, 0.031236504959768950085, 0.099706148396534174938, -0.027932285616049167681, -0.34785343718201000573, 0.20584221903388605357, 0.12282993665785940651, 0.20591364266191755639, -0.34837072827998011304, -0.027688915500842591066, 0.10291662442735927119, 0.034500188126216045637, -0.028447715147026052945, -0.14695275629605103473, -0.1738883762052139681, -0.097052843692634505057, 0.13845822023410145341, 0.59153326943138773153, 1.5559461056714531946, 0.38607547601957120653, -4.5297420520761688678, 0.40226664105543907812, 1.5795184315624355431, 0.60500493123427656617, 0.11907357170806341651, -0.16129180542374807472, -0.27726521613900728669, -0.046526029466658368194, -0.089713307288437280085, -0.031748578491672958812, 0.20002085459685858826, 0.37468754942327864299, -0.18821389658639853248, -0.20141376555250931069, -0.040675847801838298234, -0.20044378577738275471, -0.18507277539212604878, 0.38078241010541502565, 0.20504575878522968146, -0.039905178750490384598, -0.14066429431064397804, -0.087934875163047140911, 0.11876458152306532123, 0.13970172281606663134, -0.0055133153020629158819, -0.19992298191460269874, -0.42889088649686080146, 0.27012518301633114959, 0.38055724811015190223, 0.27047964103925176138, -0.42094400929848946324, -0.17964374622682682814, 0.013324122426235814748, 0.11968304960518445779, 0.013105163145307957909, 0.010682933513118083105, 0.026562401373311547714, -0.05911055284777506913, -0.15794146538557404802, -0.081338166610955650593, 0.62484735143321223916, 0.6845682877460659288, -2.0973794060596646638, 0.68355683106826592432, 0.62557534410204218567, -0.078118521250295058778, -0.16049574712363282036, -0.069651664701310150862, 0.047318022119785280166, 0.025027513841080963053, -0.13080746962117437149, -0.19199072677572770473, -0.086833117359389297429, 0.33808153053588851922, 1.2516134902635529524, 0.51046125151112275997, -3.4376314697436622936, 0.51137318092983774509, 1.2501792055003539961, 0.33147507941428766554, -0.094836762434069271643, -0.18429450465877339482, -0.10064907855412484616, -0.011043910714399477607, -0.043946833853799695258, -0.0041189503635332934406, 0.15082038813742348893, 0.080982995389273490505, -0.36505101738529177613, 0.21439503819921837402, -0.049914128446733666, 0.21634403778476787861, -0.36210368278965843203, 0.084387868170201635332, 0.15292140454927621862, -0.0077659433591228087257, -0.057897279697629248918, 0.083389154472603879698, 0.1929866952161709559, 0.096297367667285554971, -0.27522897163748594185, -0.7621786275750642714, -0.48766216697800346402, 0.97949502062845406058, 0.3562059998851289655, 0.97490157724077974866, -0.49629015124702474671, -0.77033373794617832075, -0.27182637432934048105, 0.12824546454219473701, 0.25870849165690951166, 0.070459411896868232028, 0.18138452585801007055, 0.10061027672929352683, -0.23282135605113271626, -0.70888892955895288495, -0.52439235444664555885, 0.91204340364297009636, 0.41386724545627445737, 0.908614456280179561, -0.53153838311276402351, -0.71669112282255698076, -0.23187580568056120889, 0.12626074716047358537, 0.2400918064110599548, 0.051561628377814278101, -0.023105642813171067829, -0.11594386271130938115, -0.21075362443652703726, -0.079410353013983039916, 1.0463795505705337785, 0.88488014347646382074, -3.1053091034014630445, 0.88210156703283570945, 1.0404676221253945645, -0.087336782648531854911, -0.2154002717538454903, -0.1009631204972537416, 0.02948339903294974404, 0.040150819640623919282, -0.011830835590981449157, -0.09158376852961623571, -0.18942284421274060979, -0.10726895088661404787, 0.92568969831755887334, 0.86764377310010210032, -2.8653897908281389384, 0.86532618290821472495, 0.92107334858737732741, -0.11304075733358913414, -0.19276860372815454725, -0.080332472275134680917, 0.029074394471192931827, 0.059676485626389275119, 0.16735605943556747244, 0.083308845898677652553, -0.2239432110803527276, -0.68441877413150675746, -0.28715459743772242529, 1.0398162173477911718, -0.30060829794648774582, 1.0369724292866571158, -0.29111410821819339034, -0.68757457282854606007, -0.22503643268406861577, 0.094258864627622659937, 0.22542563314140914854, 0.028455311504825209906, -0.11986464044130371032, -0.092140157725963720003, 0.21830394673027406216, 0.28845464793501607303, -0.42061812224262168369, 0.013886312509544737614, 0.16029538948303159729, 0.011903947591027837574, -0.4234341011534628052, 0.28842818429116184875, 0.22431748422046587921, -0.081575765591488588746, -0.10281788593443751123, 0.10394510538656111243, -0.10265178995964367725, -0.23340193810370521432, 0.054787844152993044056, 0.51180892923234422298, -0.1720536034893944688, -0.31127158661848114418, 0.30365088721778316083, -0.31009262277622379189, -0.17572047046972691375, 0.49532831195539761548, 0.026506783037220545129, -0.23485197442979327542, 0.042300127268409627723, -0.053528145890684761254, -0.050534586779933589329, -0.059814000894789547125, -0.071560751380106341069, 0.17827661654491061549, 1.054537771961607806, 0.51073910298063740498, -3.0253074191490623157, 0.5159969375992045082, 1.0625719193641798288, 0.18332100576305809803, -0.077248027418874265493, -0.077788939735905129891, -0.093886522367610414364, -0.044638061128746735207, -0.043392315035991634031, -0.054755940172323033788, -0.070514208278251389039, 0.14148688412713084084, 0.94427752731142755582, 0.51281858296703075606, -2.7787636358985681184, 0.51684288229207475673, 0.95071923742742725505, 0.1462071248810691837, -0.073934631942063955279, -0.068588347164385027321, -0.081781250907344282819, -0.10989692418876345625, -0.032148415388565762529, 0.12909228385295970498, 0.33248337913757364248, 0.073062991406809441042, -0.74493903775905445652, 0.071317341493737904101, 0.55805079418049996498, 0.072866037962065879907, -0.73913234868637800457, 0.086633530993191126823, 0.34676675631810049971, 0.10847527674944822118, -0.15168251256161424667, -0.08615939395808493162, -0.053916642812655804495, 0.089200774242125274727, 0.33163182980414263046, 0.13104064088448266756, -0.72371532990061404611, 0.026811768601192595418, 0.56613253648505268423, 0.028027845413247244599, -0.71924043148516536572, 0.1418039970627759383, 0.34328137475847864124, 0.072755833452816831564, -0.14777484586456260396, 0.012596996747164402458, 0.002364084899307825928, 0.25468426984064262975, 0.6491163876564313151, 0.42361394479241909483, -2.106768905773013234, -11.598158310447084673, 24.735234303799774835, -11.605876456622191029, -2.1131192407024905755, 0.43004444274562231287, 0.66937265837077319919, 0.27521234942871064533, -0.018686592043794802986, 0.049814017959702534821, 0.066465924412243068686, -0.014758799262389408341, -0.15226013367102067542, -0.16234004854366287929, -0.087094409447293685833, 0.32065892058992456759, -0.034555330420152984006, 0.31896594882480744371, -0.091170754754736751946, -0.169498457892469645, -0.15857106659220185363, -0.0010214200860171728652, 0.11805360260847895526, -0.013423429601073963782, 0.075010645550658769376, 0.094057229996122634907, -0.0061625347656362600146, -0.18697816071085643364, -0.31911684541676377247, 0.057660696549163684144, 0.73715852241571866443, -0.84533079677403988317, 0.71667305616832288617, 0.016767282584232250792, -0.36075087792389920471, -0.169409286443926832, 0.20673346074216580881, -0.010883189685096883448, -0.0057680888298042212042, 0.056634750174868270234, 0.082463444787701220551, -0.11232697618051036559, -0.3816800165117103516, 0.32379018487164540341, 0.88283395042460388602, -1.672118009999012278, 0.88284349998291622441, 0.32402802404316655149, -0.3823988658912094718, -0.11173079454927048459, 0.12480395688399140153, 0.0015339875865333194201, -0.015668453435561825043, -0.009263044793207379779, 0.033261208151219440188, 0.099738395033066640227, -0.02866718263198840197, -0.34785343645041733041, 0.20591364275309825826, 0.12109353053510182874, 0.20317503124478483723, -0.3470197959600428983, -0.014622500604317037465, 0.11248775850515724262, -0.013573466177787193998, -0.040744975909515881995, -0.11544292818631220199, -0.15591728777982299947, -0.098436174859849540075, 0.12663702816787789085, 0.58069071078495249072, 1.5559461054606620323, 0.40226664122444671934, -4.4992202047274858501, 0.4357439726351773901, 1.5934829082667798517, 0.57273917841370203252, 0.015121470458029305159, -0.37170738041423090348, 0.013853104521840922669, -0.064482476039629504916, -0.090778298700939979993, -0.029525081493591086301, 0.20069222626927241127, 0.37440649221137156655, -0.18821389674618732579, -0.20044378597781914442, -0.037808149183365120927, -0.1929600877617655974, -0.1684988584624738106, 0.40387462971875370821, 0.17927038343534684728, -0.19944094425474781507, -0.069887985476428035603, 0.011163702471182703271, 0.11779056410282615075, 0.11816952963814610078, -0.017356952678675808405, -0.20213229701963711382, -0.42889088706672218176, 0.27047964168664434537, 0.38648889278125153623, 0.28506011284777987846, -0.41125318924047876701, -0.20169125232186779773, -0.042492174979651901878, 0.18706392485372297374, -0.006983099167040399298, 0.029326299019172913374, 0.018944156936117419882, -0.06704124225656532543, -0.15776986978484036639, -0.078788171745372975741, 0.62484735155161996722, 0.68355683108111520152, -2.0934766733111951353, 0.69374633206271141272, 0.62581567454535458328, -0.10757605854596302697, -0.18872236803259370697, 0.022969776389712228543, 0.066123664167680562542, -0.048081237685345008803, -0.14529249960008311371, -0.18407869255588824031, -0.07777497554715429029, 0.34153564431283717218, 1.2516134902866606904, 0.51137318091625882932, -3.4328641761619000938, 0.52106686406603608397, 1.2631945897169096771, 0.34005065663421008004, -0.12359304148724242767, -0.28783083087321725824, 0.010962842851224102042, -0.023472774641514346922, -0.042450942277250804102, -0.0033392581501564396959, 0.14876801569934516856, 0.078792151571127469478, -0.36505101785622340316, 0.21634403798815807218, -0.047909991934018833493, 0.21803364029423735748, -0.35765892567500734422, 0.098056883642883951868, 0.15628613630176166338, -0.087970687236764891215, -0.014570435523139474085, 0.11251448364986482709, 0.18889182034321805315, 0.089700570354937309014, -0.27603367498031161631, -0.76044856681486672212, -0.48766216717965216176, 0.9749015773754545755, 0.34565601842519383391, 0.95782219033287729903, -0.52069792826271066311, -0.78778859405639145574, -0.22264751908616239873, 0.39963918961258160234, -0.023103790507603413212, 0.10446852247524007762, 0.18166801348711691944, 0.095286260344375997566, -0.23456463582247610855, -0.70828634906760656875, -0.52439235426546637608, 0.90861445603161905549, 0.40463907079641453501, 0.89182437111288304088, -0.55687556792209191858, -0.73525747640582772569, -0.1850623618539465054, 0.38115822086356904164, 0.041625597533071267975, 0.011064709399841263227, -0.037916525555803076708, -0.11369834272579813772, -0.20223708111410471089, -0.073885726206533372529, 1.0463795508628148667, 0.88210156676916828822, -3.1062068802231430631, 0.88543298954875548912, 1.0420993946939873798, -0.095566323444442685675, -0.22229576272802806236, -0.059967181813616113328, 0.031078434822644260271, 0.011854733427861627315, -0.024549663137211623354, -0.091110997687919692556, -0.18280204023124496659, -0.10260833323491715119, 0.92568969863786387808, 0.86532618278006789847, -2.8655814855319463597, 0.86932016355940433527, 0.92240860442894423432, -0.1233470472845940441, -0.20141981344905279805, -0.036766173996319082007, -0.031695215670912692263, 0.10515361281373178381, 0.16830185147676013369, 0.074036274628186959923, -0.22763846411005647652, -0.6837101763817844402, -0.28715459740143228773, 1.0369724289673349915, -0.30559629525806303318, 1.027583719100205295, -0.31495394062956799663, -0.71967628019893692404, -0.18992181743896696666, 0.34890001988836893965, 0.067010155008648916719, -0.043844606293125566732, -0.13054932188043855135, -0.085364746881101682963, 0.22434864678495478052, 0.29088118217378655661, -0.42061812246250834812, 0.011903947757425231702, 0.15676054081299337084, 0.011743856291232040653, -0.40838916139678027584, 0.32837293625859476087, 0.24158322365874124182, -0.24753902665841390207, 0.05288005936277250657, 0.025516824045953217376, -0.09876316233662493782, -0.21496506450476113304, 0.062949960986183772782, 0.51229813667980006464, -0.17205360291290178187, -0.31009262297529610608, 0.30108283777873900977, -0.32194105367462499956, -0.19076957684480991273, 0.50122067845600981606, 0.057030567902005266379, -0.2066997748515237987, -0.0092623596504344211239, -0.041618578861665339597, -0.050891007622620915152, -0.063342371716029521878, -0.074574297125412927767, 0.1760585618173644562, 1.054537772238838933, 0.51599693720159156562, -3.0125483066727598569, 0.53320339033558161734, 1.0711556428058273127, 0.16513009526279304184, -0.13528364930821701329, -0.12898759633707046479, -0.0044315556205316774738, -0.03633993590873560442, -0.045806411328364896041, -0.058468812117299741316, -0.072594952666079320958, 0.14017137227049736303, 0.94427752698456846048, 0.51684288204992478999, -2.768054374049844224, 0.53244288552544805082, 0.95968108036629973245, 0.13134703188879828795, -0.12552424111997895206, -0.11421072381491660652, -0.040929475438660503639, -0.069966280466122937609, -0.021828078610867045206, 0.12574677916489990226, 0.32686358869873149979, 0.070002705763467912292, -0.7449390380072499207, 0.072866037838834857299, 0.56063965825977579716, 0.079676956776845136998, -0.72429760237547424762, 0.10519252621123326152, 0.32484726681034142848, -0.05988284148356541603, -0.022590086551519162478, -0.067374311117058438558, -0.045751684004143429452, 0.08865924104826612695, 0.32805152446315094528, 0.12863710704969399945, -0.72371533007151289674, 0.028027845667548137443, 0.56769312526972581789, 0.032988063913726980991, -0.70511614165366853779, 0.16490018549843155071, 0.33052317357068344839, -0.10201640621744621773, 0.0024129031015797195586, -0.01228000054961967541, 0.027734171864912057076, 0.26673883450357238623, 0.64617048293434731221, 0.42037621071002939654, -2.1067689057808927089, -11.605876456497814075, 24.706729597056444447, -11.645952114313484316, -2.1129406393582690349, 0.53187761468671312937, 0.80124205424226735417, 0.085685715596298950292, 0.00024547933541083463054, 0.049983185326275429705, 0.066041104005342410321, -0.014012614555788524204, -0.15076101368322400442, -0.16111454170111524919, -0.087094409183314056011, 0.31896594876754125192, -0.038548299000896422972, 0.31072634193845166672, -0.10484595061911233804, -0.17846788671388799652, -0.1427771026476304661, 0.13124460055446665607, -0.05537594496316027709, 0.030972278430976923208, 0.094255674434388023175, 0.09996258484197693972, -0.004900936263875952173, -0.18697816086432206273, -0.32233853780677435008, 0.046710494278732264672, 0.71667305600672115418, -0.87096412256375266292, 0.69963731637515536033, 0.021092488345281427964, -0.29324647105294932015, 0.018965735672276443774, 0.00079649122943901095022, -0.015428830641152184444, -0.0028705300958283930073, 0.058817174014632239709, 0.083143678647992477759, -0.11232697626165330518, -0.38318793333879669349, 0.32041863584353719174, 0.88284350079083506735, -1.6618749544626916226, 0.89217242201951452163, 0.28235347006871441833, -0.45836353157814202497, 0.1123753989822781213, 0.0022395621479493273316, -0.0037875981942776040801, -0.013457544770698983902, -0.0076437692939460080699, 0.033025780076130120677, 0.099738395490219775019, -0.027932286538828673267, -0.34837072882853009625, 0.20317503185120008857, 0.11895010107385135811, 0.20727376065004501915, -0.34262004591527839281, -0.0074691214795029568785, 0.087529709711139686013, 0.028418776841689214613, -0.055786849329988119628, -0.13139807886940946258, -0.168917229405303293, -0.10622344649658599369, 0.12663702823227007688, 0.5915332704392188834, 1.579518430226575898, 0.43574397280991822301, -4.4673417562411517068, 0.45409027859033473939, 1.5765391708819078698, 0.4492943532319459421, -0.29580859367426948214, 0.030555171760762684513, -0.018784014212212127998, -0.071210195363674178015, -0.089130604605827792875, -0.027872830465428370672, 0.20069222601561392816, 0.37468755004698445221, -0.18507277548900244457, -0.19296008750321297365, -0.024431484454616510943, -0.17048021772787161354, -0.15421983368653302549, 0.38425808200213562582, -0.050182906789579317941, -0.014799138330793874807, -0.044263501716356248861, 0.0063461508361698374184, 0.11020226173548770821, 0.11540525949813758289, -0.017356953077750789799, -0.19992298111210712874, -0.42094401041291451548, 0.28506011348201215627, 0.39279460004726235445, 0.25100690666755776093, -0.49969095458697748713, -0.20520564160595169967, 0.24151645383742120776, -0.00049300952565417023848, 0.0013895114828072236291, 0.021314341454900589956, 0.015952157665694171457, -0.065561615348286708671, -0.15776986965538236651, -0.081338166870882994131, 0.62557534388237323597, 0.69374633216196313068, -2.0826820476622045497, 0.6852621616826103379, 0.59689565388325294126, -0.11518826181865318703, -0.13795609498125946391, 0.052617894274984741698, 0.013045317268064016147, -0.061369833014673469407, -0.14266359315558210796, -0.17993713128133134727, -0.077774975757536349286, 0.33808153068142526854, 1.2501792054699534251, 0.5210668639553711623, -3.4033201507859276447, 0.57281647576418792678, 1.2956481804804349522, 0.2503497817471092568, -0.42460837713812665806, 0.013794978768846334963, -0.0014114331221742326683, -0.026548173347333732519, -0.043620455392694060248, -0.0046533759368408492141, 0.14876801510289100472, 0.080982996161509931965, -0.36210368289146016574, 0.21803364046541123811, -0.046994104425835431238, 0.22431746714646383367, -0.34652140848054863165, 0.11471805775629516366, 0.032818131979502285223, -0.045639858125987795012, 0.032734617221673888154, 0.12508137229627391829, 0.18815845895871918803, 0.087893360519912505646, -0.27603367504859505077, -0.76217862786014334286, -0.49629015075720073202, 0.95782219009954572275, 0.32078161794521453665, 0.92371041771297734968, -0.56188360753471178199, -0.72746787244027244501, 0.22057259697288972289, -0.049750694099522305147, 0.027752721555220315197, 0.11832340675363990645, 0.18059510715033860739, 0.092720491740682170478, -0.23456463577208233606, -0.70888893011653664189, -0.53153838232427386146, 0.89182437080416843145, 0.37795992802493388885, 0.85548180015352581762, -0.59604009515369649197, -0.66967548748764349487, 0.23442780751333583411, 0.020993549299577751732, 0.020268007499689381229, 0.0042626672593093157052, -0.034959545228306979081, -0.10919826187376258098, -0.20223708087002753575, -0.079410353536643135675, 1.0404676226558231544, 0.88543298933255598637, -3.0887361295472541833, 0.91075775209171117019, 1.046651830107616199, -0.15079906815839411172, -0.26473577544913784454, 0.016441742760845295263, 0.015246967471176283912, 0.005612936270058445018, -0.022596109373887583727, -0.087435532497905296689, -0.18280203996140698131, -0.10726895149804857588, 0.92107334891659953247, 0.86932016350065555166, -2.8499374018504246564, 0.88857530765626036029, 0.9199469774212061246, -0.1726438522944727938, -0.21465990436406401365, -0.049100135492830002115, 0.023058433299692229074, 0.11375745502154260835, 0.1646304891968338735, 0.071693944090626848809, -0.22763846439566293167, -0.68441877399392481163, -0.29111410777995694321, 1.0275837194486396875, -0.32777725003647401048, 0.98468526888242957629, -0.35948270227895184892, -0.65218654137498777157, 0.19654205495190218977, 0.048011220110737960753, 0.016229082438394732552, -0.053105571403636878114, -0.12587662741647526188, -0.081427955679877664363, 0.22434864646832311519, 0.28845464814142268706, -0.42343410128460179376, 0.011743856285178993701, 0.1669962266717737176, 0.045584686318852152276, -0.35652212612145867121, 0.3478284788548231532, -0.10721517610828416867, -0.0014496283367914539978, 0.051142861302622993902, 0.028136632846783346529, -0.09613235216238263936, -0.21404914190731946788, 0.062949961324239689375, 0.51180892848602488332, -0.17572047063674786638, -0.32194105368993031213, 0.28452908775659624396, -0.31278642771224263175, -0.11252936733587624862, 0.62445379560106484274, -0.32917564179474273489, 0.019378582294782779882, -0.022150866927958322988, -0.051451433129522883281, -0.05606075460353895229, -0.064982184004355875162, -0.074574296581287613517, 0.17827661589556592636, 1.0625719201892021015, 0.53320338936402156271, -2.9900491448998383071, 0.54924476790404896942, 1.0481456392142860512, 0.074809398908620206026, -0.20050147246707877846, 0.020197126338442981247, -0.018873534092273821366, -0.045967066971971906608, -0.049980376125730793802, -0.059346334228573133229, -0.072594952534301621094, 0.14148688445966661753, 0.95071923828650828359, 0.53244288470148248837, -2.7463785331053789385, 0.5484618144916990623, 0.938206075222355107, 0.048297051683892568141, -0.18175946191259298446, 0.0040854637112811764341, -0.048694897020237437602, -0.069417032406137252343, -0.020999027858791912554, 0.12490900588209793987, 0.32686358886290550751, 0.073062991843294344818, -0.73913234832858953904, 0.079676956394588688282, 0.56572652614031748719, 0.077200003021261148772, -0.74544809270208134411, 0.099244663664932925018, 0.27982359920907884376, 0.010454604083413639865, -0.037135923890010129333, -0.067861570746536828413, -0.044108273464333824787, 0.088489048219482224766, 0.32805152464712633575, 0.13104064087828748431, -0.71924043175293528307, 0.032988063746611305527, 0.57232097172020568454, 0.036599220052767313105, -0.70962430954429644991, 0.16851341585607890194, 0.21583222439482835364, -0.030572900996977892679, 0.013615661332504771369, 0.01204809571565417306, 0.03923281259897271056, 0.26688220989511357573, 0.64617048285040012967, 0.42361394431810095451, -2.1131192406676655438, -11.645952114060071025, 24.629321160615933906, -11.680781058059332977, -1.9757781503384082011, 0.7548950969383894094, 0.65592284662054411992, -0.021032722934810961501, 0.021035013686423419815, 0.055939585078337669688, 0.066110440410048113002, -0.014263427455160591906, -0.15076101358930413254, -0.1623400488218708082, -0.091170754851537999519, 0.31072634226354584053, -0.05079847252661205681, 0.30066661279869316692, -0.11176467133312213698, -0.14484952972940406957, -0.012450342217347769649, -0.053208467981934244595, -0.0074947104966584200333, 0.048858318468905530263, 0.097785734812175048591, 0.099962583748242225812, -0.0061625338759230359129, -0.19463569631830232587, -0.34244430575482531776, 0.016767282491962579721, 0.69963731703918541349, -0.84861286177222328142, 0.78017584422676800049, 0.075976925407384002242, -0.37827702928979428654, 0.00078519563064696035973, -0.0037895754019700348368, -0.012186128839753462144, -0.00079558862163292836008, 0.058817173240434290904, 0.082463445596473339272, -0.11334068226998984807, -0.38372962335269489298, 0.32402802342034292549, 0.89217242195274537586, -1.6751659993014063055, 0.84668777975108766576, 0.27128296323352169006, -0.29016350636881349523, -2.3399784080352628967e-06, 0.0019332419990838903105, -0.0034159342783797439803, -0.013535914863177395465, -0.0076437701225919351192, 0.033261208270588725477, 0.099706149156604931272, -0.027688915227378995548, -0.34701979587973924568, 0.2072737597983941582, 0.11789530429195153871, 0.18691696238963170784, -0.39848312262324930844, 0.15045226330782607294, 0.059498185265179040493, -0.013652015398680621486, -0.083346781078709422652, -0.14358043560584746179, -0.16891722890132704338, -0.098436174248408545018, 0.13845821974224611917, 0.60500493210431338653, 1.5934829074156289241, 0.45409027789429640132, -4.4486677343942613305, 0.46731798662661189203, 1.5101127175062949437, 0.15693704569844824093, 0.024135405266642394329, 0.0080254491829422798521, -0.02596041682987988361, -0.071596373033773197392, -0.089130604570334975856, -0.029525081137863805336, 0.2000208537265828479, 0.38078241034699289447, -0.16849885906437267469, -0.17048021736126262038, -0.023862088983598442327, -0.20394883883399139712, -0.21215561615390907124, 0.39178145970455047209, 0.016948688357136820065, -0.034301595951061192302, -0.048133394032420964703, 0.0067551105598059029561, 0.11020226098195314013, 0.11816953023129882216, -0.0055133155081386522184, -0.17964374548031947221, -0.4112531898889710269, 0.25100690697626315551, 0.300014650100838276, 0.20082172628272010328, -0.37561815223928479712, 0.046216864748079018732, 0.0059248021424697129567, -0.0071251237561778986429, 0.00029608812080657911101, 0.022465055787565652717, 0.015952157359820456212, -0.067041242330086750489, -0.15794146495679550757, -0.078118520862895349244, 0.62581567451093267351, 0.68526216133565265309, -2.08297049368051157, 0.74120797379347402867, 0.62879333223010214216, -0.33117366541957649062, 0.025299763989725290725, 0.031172893014499158948, 0.0051176251449257543352, -0.062108129621832826317, -0.14266359303625833488, -0.18407869253631800599, -0.086833117363248862874, 0.33147507971169304364, 1.2631945896336798096, 0.57281647559570125772, -3.333895934720256804, 0.61146738485837415222, 1.2079012758589020571, -0.22371153707161547941, 0.010215920653425315454, 0.0061299725520346020832, -0.0062017229201665174249, -0.028701603155191397071, -0.043620455181538099476, -0.003339257874073339738, 0.15082038740257550091, 0.0843878683249085354, -0.35765892591290499025, 0.22431746742994360111, -0.043096746571942276538, 0.21595738199659431089, -0.43501737202245127634, 0.22851399969702596016, -0.042280162464540783918, -0.0092814470172741751069, 0.047671301062988356845, 0.12777208549566487772, 0.18815845970645564411, 0.089700569786967984665, -0.2752289714757440442, -0.77033373821779582702, -0.52069792799527270244, 0.92371041796379915745, 0.29524113516153099868, 0.94885420747046367307, -0.48431437961900097866, -0.54034813165054196293, -0.04208883422249330708, -0.01257271802445326242, 0.041945795464851873269, 0.12024218150895363277, 0.18059510783467919159, 0.095286259879101109505, -0.23282135570634665545, -0.71669112335920692747, -0.5568755676480221517, 0.85548180025159747952, 0.35119179396802258086, 0.88094584226639260915, -0.52281543936921148763, -0.4619673463981465722, 0.0024231395360835481477, 0.016922249136475341763, 0.021522815865463066493, 0.006693945971835766999, -0.034959545225122297396, -0.11369834315246761525, -0.21075362408161499306, -0.08733678310036956105, 1.0420993951977084446, 0.91075775192534125235, -3.0405790714643772255, 0.96381181326491027761, 1.0049721693355586627, -0.47810066641325238423, 0.0028091222380220476525, 0.012112230112767010878, 0.016140130716512133408, 0.0077011650424502328324, -0.022596109435448551855, -0.0911109980424189142, -0.18942284385742952368, -0.11304075767043259004, 0.92240860480475461713, 0.88857530751239521827, -2.8120309272109800247, 0.93493014556910059021, 0.88415569425126605552, -0.43767544609949782242, -0.034754356560659546704, -0.018632880885708445057, 0.033950685704085173389, 0.115430849215719597, 0.16463048874844904423, 0.074036274898482315465, -0.22394321156871191403, -0.68757457299921420635, -0.31495394081126132368, 0.98468526924646204712, -0.35268228481321517442, 1.0334586746326475115, -0.30523104520128208073, -0.48605846529144403068, 0.017552900893502886015, 0.031289997312575541788, 0.011588365625654005811, -0.05240292970368681108, -0.12587662711099048551, -0.085364746794754170489, 0.21830394661572938286, 0.28842818462232278076, -0.4083891618523922129, 0.045584686578394206791, 0.20485649688995372086, 0.043845151994181319477, -0.52850148108053884854, 0.34586628177753792057, -0.020861592630242828394, 0.024119690590357970028, 0.053629644694498414259, 0.026001322155117446988, -0.096132352457675859791, -0.21496506445490143333, 0.054787844224144371941, 0.49532831276415645316, -0.19076957670337346684, -0.31278642850673488152, 0.34672735810824245606, -0.18646262721807130291, -0.13832560120411588334, 0.16333952717521144082, 0.028136400502235550669, -0.0032376103603295770805, -0.033316128360856955837, -0.055106713869069512413, -0.056060753602347482283, -0.063342372683420994361, -0.071560750384371610289, 0.18332100461320971485, 1.0711556437917642004, 0.54924476749031869716, -2.9803007940883139426, 0.55571027714798026942, 1.0552606677341116459, -0.16787300905506943871, 0.025924674063270872898, -0.0014540690970349873581, -0.028571084690476894474, -0.048694962465009941632, -0.049980375229281594207, -0.058468812805327688431, -0.07051420797113934491, 0.14620712380568678945, 0.95968108111476257971, 0.54846181447774200457, -2.7372306761952613918, 0.55321257672610302247, 0.93930738412062075859, -0.16735150333870896233, 0.019824927830946047214, -0.016311867213989042491, -0.053459870515950080883, -0.069465575571876311822, -0.020999027576656344174, 0.12574677893009386653, 0.3324833792435071822, 0.086633530272194808619, -0.72429760212459926016, 0.077200003334085109952, 0.52338154895956900958, -0.0010553035363463983409, -0.83696162523531658639, 0.5617050020276390665, 0.018438700026851821689, -0.008058163602134668646, -0.041860255465047915835, -0.068001190633114397266, -0.044108273375714733555, 0.088659240831684457684, 0.33163183014740543753, 0.14180399670142918023, -0.70511614146681544923, 0.036599220257970847825, 0.54597142753607008636, -0.026497484748170766627, -0.82189386034929101665, 0.55749242650015651268, -0.042116856908591127073, 0.0035980591384649389733, 0.028422268157726335996, 0.016516811248004099194, 0.039232812276785537786, 0.2667388342781397137, 0.64911638820181050136, 0.43004444249485451746, -2.112940639118363606, -11.680781057913575793, 24.509313118063424497, -11.823208800850556699, -1.8163939994871265426, 1.5116599626844908766, -0.021624312319690731921, -0.0010155889219638917272, 0.027687158087174985888, 0.056950599768477210005, 0.066110440144164520082, -0.014012614444977233505, -0.15226013365039123282, -0.16949845784188152797, -0.10484595075332477088, 0.30066661275797945674, -0.051428015710518693027, 0.33265469559837634872, -0.03781115536294515217, -0.23891237632229433974, -0.028767446544781066081, -0.023359228427655000931, 0.0025274059022697280448, 0.048858318718368015654, 0.094255673906506298732, 0.094057231208407768297, -0.017159901888043296858, -0.21356736849613000562, -0.36075087724261689637, 0.021092487608745456862, 0.78017584427774289146, -0.71618979462015919246, 0.83565122978659256425, -0.52629639922611082259, -0.0016063179793620939277, 0.0018842767584974792847, -0.0024773220776315532134, -0.012186129099563021427, -0.0028705291652126605287, 0.056634749512589094289, 0.086332642048050481365, -0.10179911291721265021, -0.38239886567743652801, 0.28235347057662252279, 0.84668777949343487776, -1.6042061158000202603, 1.19518671800705234, -0.36261323601926259697, -0.00051528130085626451789, 0.0012780236542963292323, 0.0014404390805227674458, -0.0034159345414012782204, -0.013457543869906477343, -0.0092630447800558651794, 0.031236503956132363052, 0.10291662467223831878, -0.014622501334144884927, -0.34262004442579674368, 0.18691696149797454307, 0.06793613957448363383, 0.14527803802255273569, -0.15539615050555236064, 0.05600797170724808316, 0.013472862678925478716, -0.03422700315384675368, -0.083346779846207336151, -0.13139807935672193073, -0.15591728829953335755, -0.097052843546800049435, 0.11907357167094832773, 0.57273917887037906205, 1.5765391713561820453, 0.46731798557652109993, -4.487179142333739712, 0.68129426918001168012, 1.5248153690863595067, 0.011857913386854788962, 0.014307706554005334559, 0.0042180629254670275186, -0.025960416414040867611, -0.071210195710816201475, -0.090778298895438869209, -0.031748577587001711175, 0.20504575849489559136, 0.40387463039178789215, -0.15421983439017131201, -0.20394883829042115275, -0.11955975699919306476, -0.39701939532005381395, 0.46013592868120617441, 0.018731995559347316277, -0.0058946305873218045671, -0.037242620996883420148, -0.048133394818406637083, 0.0063461520271557592815, 0.11779056354427076825, 0.13970172249047094204, 0.013324122441425447286, -0.20169125215970984399, -0.499690954927663189, 0.20082172647007492938, 0.43576240705017965649, 0.29327026263697136832, -0.44052976023470524769, 0.0036857351882773818456, 1.6104181364458082683e-06, -0.0058160918200547495963, 0.00029608752419357949309, 0.021314342257296947647, 0.018944156580994946787, -0.059110553129169997322, -0.1604957473561119119, -0.10757605870795752812, 0.59689565437681502136, 0.74120797356841028325, -1.9779307951160995938, 0.85811308011639497728, 0.075882333390598910583, 0.0044154460242764549324, 0.023263174708489924025, 0.028903731459216226024, 0.0051176246350485765299, -0.061369832510459784713, -0.14529250005035534987, -0.19199072678558420924, -0.094836762669725932495, 0.34005065688315838646, 1.2956481803583070889, 0.61146738495934593871, -3.3631917160884379214, 0.6553949774877447787, 0.91280978730621709083, 0.0054050337826818820389, 0.0051185905967778087128, 0.0027369951688382271326, -0.0062017226841831699785, -0.026548173729963434747, -0.042450942570635905038, -0.0041189495861137150751, 0.15292140421022101804, 0.098056883983495435331, -0.34652140874120007652, 0.21595738212896897767, -0.12825534278024169899, 0.092442607601245438409, -0.017557088185511594142, -0.02632616272823691192, -0.021719290315428407823, -0.00068423299553465807832, 0.047671300791318579892, 0.12508137182886747141, 0.18889182041269869661, 0.096297368070015473829, -0.27182637455662100878, -0.78778859391506461662, -0.56188360786650515255, 0.94885420796869379689, 0.47330676584880532198, 1.1161017965288047016, -1.3391778788291408553, -0.023357661397327252745, -0.023193196074214363078, -0.0051011362947983064642, 0.041945795374650152176, 0.1183234063491830218, 0.18166801353380995776, 0.10061027688522680701, -0.23187580559592071983, -0.73525747634642391048, -0.59604009543512093128, 0.88094584260065489367, 0.51541601320302421474, 1.0299104787303365871, -1.2676678254633655385, -0.0081172028759704008272, 0.0093229511871958346825, 0.020555943871383298982, 0.021522815825316229249, 0.004262667114717962058, -0.037916525215697396978, -0.11594386282302374114, -0.21540027158793492723, -0.095566323734428512693, 1.0466518302446137234, 0.96381181334047894005, -2.9829426923011763684, 1.1541169805769924128, 0.24653778061400907773, -0.0058432288317056685356, 0.0073507596569223855251, 0.015105049727696539044, 0.016140130673556116192, 0.0056129362308156727918, -0.024549662854994682565, -0.091583768664972267648, -0.19276860343763457806, -0.12334704757259204455, 0.91994697755773335857, 0.93493014567861210118, -2.7528507009133957872, 1.1293653056102421761, 0.17165291936396348538, -0.016942512322146271314, -0.023365728757501515178, -0.012164371116224188471, 0.033950685850308609526, 0.11375745478197810123, 0.16830185162153912803, 0.083308846265748179305, -0.22503643271107254265, -0.71967628008079054247, -0.35948270264697818055, 1.0334586753552550409, -0.1943581036298369924, 1.2275615052085193657, -1.1211852592865649836, -0.0030148879162097607862, 0.020208669154775558974, 0.030930624736492242505, 0.011588365469170630942, -0.053105571404730024521, -0.13054932205586039506, -0.092140157738123798503, 0.22431748410667315397, 0.32837293678241380457, -0.35652212660618937123, 0.043845152233480184922, 0.016469421193436247636, -0.26322587303358990773, 0.22984406295337617143, -0.015483412868500138343, -0.0023984531951352426431, 0.026325956116992916406, 0.053629644702036530224, 0.028136633166320736132, -0.098763162831236692152, -0.23340193749595666839, 0.026506781985246724576, 0.50122067828015892665, -0.11252936609277888746, -0.18646262807724617305, 0.29241279934039865473, -0.74090550032388746704, 0.46722925012178695425, 0.021724333231967686564, 0.0088988518094224242083, -0.010416592441604349134, -0.033316127935594117004, -0.051451433977418581378, -0.05089100698859731764, -0.059814001358085643056, -0.077248026559901317167, 0.16513009449696647302, 1.0481456397211443843, 0.55571027671691863858, -2.9216274027281099457, 0.7536161156717450238, 0.66390775608447427647, 0.01839772706413594916, 0.0092180408082003523901, -0.0073216944993296298461, -0.028571084155929370174, -0.045967067978339250234, -0.04580641066691948371, -0.054755940522673823834, -0.073934631452330493162, 0.13134703145510329136, 0.93820607520778198651, 0.55321257667334045038, -2.6847961190033862167, 0.73483897246706997297, 0.56722175167246180472, 0.018245519365656444333, 0.0043465744788378554014, -0.021049262336860528194, -0.053459870381396365491, -0.069417032368543685195, -0.021828078524436165392, 0.1290922834671237307, 0.34676675697043291002, 0.10519252605196822981, -0.74544809334278583357, -0.0010553033291357226398, 0.40196130457456907825, -0.033924147113513325458, -0.065282657154504611374, 0.014487246697495003225, 0.0065321432878430374033, -0.012202560142694310677, -0.041860255533445037135, -0.067861570612321689411, -0.045751683941665677313, 0.089200774139381808214, 0.34328137544196662967, 0.16490018522670765733, -0.70962430973379475585, -0.026497484757537173583, 0.40175944249354417215, -0.12442429357908521559, 0.0040885707865690127522, -0.02931597375941555686, -0.014864163266403032995, 0.010933769545108398824, 0.028422267593060553115, 0.012048096057007445098, 0.02773417204333765429, 0.25468426943519678618, 0.66937265901345066599, 0.53187761409265310597, -1.9757781502915672256, -11.823208800565732091, 23.713981554559513398, -12.102904941136099382, 0.65881391061770666617, -0.013076591007862617064, -0.010089751502197402688, 0.0031926212260672036727, 0.027687158250556530292, 0.055939585140348170589, 0.066041103909513287906, -0.014758799322214834204, -0.15857106627245259878, -0.1784678867852907691, -0.11176467133897077799, 0.33265469575362693888, 0.058724507398615059806, 0.29879794913356044761, -0.35988107309548228452, 0.00027322330229002971458, -0.018154896243425044239, -0.02335922842952803738, -0.0074947124601604372604, 0.030972281258392121012, 0.075010641938077091129, 0.085406638248176541817, -0.00024212386314698044337, -0.16940928728224757815, -0.29324647049821622735, 0.075976925777177517496, 0.83565122967786975572, -0.87589767256718242816, 0.28715199708936750556, -0.0013579211322958821464, 0.00090442185736054289159, 0.0018842770973082341272, -0.0037895757954678017596, -0.015428830776935017338, -0.0057680887707540371909, 0.062652755468951173978, 0.10398504857934669221, -0.11173079468968449834, -0.45836353170350829789, 0.27128296308052701757, 1.1951867182309705573, -1.7176710792618272716, 0.68222115839062880127, 0.00019207504591898946983, -0.00031573776137188704852, 0.001278024050036353533, 0.001933241591258787629, -0.0037875981298213356324, -0.015668453970275539322, -0.011671490780525780803, 0.034500187829972724096, 0.1124877589258495153, -0.0074691221076609408402, -0.39848312235225014089, 0.14527803807313816042, 0.39783139365609526683, -0.25778694871202939565, 0.027906588020173723336, 0.026420134338509770194, 0.013472863431363786557, -0.013652015609045189265, -0.055786848776913970649, -0.1154429291186861023, -0.17388837702024953491, -0.16129180473883492075, 0.015121471418108671861, 0.44929435291698571753, 1.5101127174166251166, 0.68129426860147856981, -4.6559682237326160958, 2.4284155192258953804, 0.00097751642414759323807, 0.009774791210833563837, 0.014307706828193843263, 0.0080254487569367989608, -0.018784013693450989785, -0.064482476169493666007, -0.089713308030499538948, -0.039905178321252891827, 0.17927038304047046569, 0.38425808260569377373, -0.21215561691959095025, -0.39701939490851978487, 0.27779830791213389052, -0.06183428557292969252, 0.0068585033131476425342, 0.0093459133509139046686, -0.0058946307645535434006, -0.034301595183216349039, -0.044263502505830491784, 0.011163702732974295148, 0.11876458171458774415, 0.11968304942796835488, -0.042492174914434993616, -0.20520564140463909419, -0.37561815240610624267, 0.29327026271039807748, 0.58789978522890373736, -0.44208149083860065476, -0.00091748235583311910958, 0.0037271088892208543118, 1.610268483966231429e-06, -0.0071251230384961278599, 0.0013895106494361483557, 0.029326299464009934759, 0.026562401482061230584, -0.069651664579370053954, -0.18872236802606845441, -0.11518826203240914441, 0.62879333234294954025, 0.85811308023985477522, -2.2437560511635914473, 1.082154980005089806, -0.0067400030893293083264, 0.0078220184769758008964, 0.023263174091387898296, 0.031172894073503713136, 0.013045316070191747757, -0.048081236508946516295, -0.13080746989216923692, -0.18429450464710264712, -0.12359304153516850744, 0.25034978189743317678, 1.2079012756729767819, 0.65539497740613072985, -3.6693244396912656491, 1.976615413863186399, -0.00012357859251171338729, 0.0020770976751709867986, 0.0051185904449578996378, 0.0061299725401357816126, -0.0014114332225674854071, -0.023472773766904757309, -0.043946834660683824214, -0.0077659430896681262008, 0.1562861359925475091, 0.11471805776863533421, -0.43501737184308103545, 0.092442607474624294284, 0.3181607201462894885, -0.18443192140255232347, -0.0078817772007274550239, -0.018581194633757354417, -0.02171929097417970092, -0.0092814464975734551483, 0.032734616811822860039, 0.11251448485206026751, 0.1929866946608638778, 0.12824546458659238923, -0.22264751949080135995, -0.72746787245170119185, -0.4843143795985135891, 1.1161017966076121066, 0.38426326616072958853, -0.45233985448064467638, -0.0044323020308297822467, -0.017702923014694116333, -0.023193196468770646557, -0.012572717566990883678, 0.027752721068948462552, 0.1044685233825753401, 0.18138452549043601603, 0.12626074709080561953, -0.18506236205338838507, -0.66967548750695715665, -0.52281543928412765876, 1.0299104787203028355, 0.4842592382912840332, -0.49970943052905875126, -0.0092002782876286492986, -3.4421031634353043121e-05, 0.0093229513409427196974, 0.016922248875184005973, 0.020268007563709743157, 0.01106470940687833554, -0.023105642602092444499, -0.10096312065657883639, -0.22229576296985617545, -0.15079906801934517269, 1.0049721694231001923, 1.1541169806343791748, -3.3919645304647807471, 1.6905048879156814312, -0.0073699623234972011163, 0.00048274028674801915811, 0.0073507598888243987426, 0.012112229806510483796, 0.015246967604155149678, 0.011854733345591950133, -0.01183083540123648289, -0.080332472558255624273, -0.20141981355318561153, -0.17264385224973191613, 0.88415569423315309994, 1.1293653057608852297, -3.105898238445793158, 1.5266825592466879424, -0.0028691154563137847081, -0.01400027423583857962, -0.023365728591641158479, -0.018632881748282655032, 0.023058434086448115208, 0.10515361267913901633, 0.16735605963282282338, 0.094258864430927513656, -0.18992181803576890786, -0.65218654091586203414, -0.30523104583511451171, 1.2275615059281919095, -0.26735725874823867976, -0.12824834765910853029, -0.011468265378957953615, 0.0030477991985641845335, 0.020208669033658647335, 0.031289997791685453254, 0.016229081966384218544, -0.043844605768472433294, -0.1198646406795893099, -0.081575765513987250066, 0.24158322351814945361, 0.34782847902652958005, -0.52850148118284800969, -0.26322587309506823017, 0.61155686410098952788, -0.22018411467975584905, -0.003418556678084352235, -0.010446678707661839758, -0.0023984529294541387899, 0.024119690363618294382, 0.051142861237968038901, 0.025516824328797389071, -0.10265179030510437896, -0.23485197376868260921, 0.057030567884787899957, 0.62445379479113072296, -0.13832560041304098331, -0.74090550052179671159, 0.51088634536128252339, -0.059015283455581678262, 0.007688750279130208802, 0.011992959005199040068, 0.0088988516990013092484, -0.0032376099842614993601, -0.02215086689476586726, -0.041618579072607159164, -0.050534587223667187805, -0.077788939866016718061, -0.1352836483288850089, 0.074809398129897675123, 1.0552606682163931939, 0.75361611523825289272, -3.2267140835964496226, 1.6389303593920192537, 0.0053059937702317370653, 0.010823467060519965585, 0.0092180407680459951203, -0.0014540690239780449181, -0.018873533819330191863, -0.036339936133034038412, -0.043392315248885718337, -0.068588347059483203982, -0.12552424060455075194, 0.048297051472248671289, 0.93930738421347870215, 0.73483897226095051813, -2.9508352695858728332, 1.4928377586203218641, 0.010008896776058165412, 0.011307285030098737991, 0.0043465744752101184017, -0.016311866819776657533, -0.048694897429971996383, -0.069966280598058594453, -0.032148415304273161852, 0.10847527648610831574, 0.32484726722507828445, 0.099244664141997784856, -0.83696162566737786204, -0.03392414720272938361, 1.1587065869224932868, -0.69573938428581760807, 0.0069028223544569727518, 0.0095424278472133108181, 0.006532143165496573714, -0.0080581631702361135317, -0.037135924217441738937, -0.067374311202638079155, -0.053916643013424331299, 0.072755833066282055155, 0.3305231739819209924, 0.16851341580255047026, -0.82189386032096356516, -0.12442429378448340016, 1.1634399434883477831, -0.65995314381802483439, -0.0073140987139730974376, -0.018711206910834249362, -0.014864163512816684631, 0.0035980593912363335006, 0.013615661073548760221, -0.012280000244858955277, 0.0023640853681360660468, 0.27521234863369714585, 0.80124205421962402252, 0.75489509687685585337, -1.8163939993558579911, -12.10290494092739344, 21.785095955934359324, -9.6845635101573268599, -0.0028883715698718405093, -0.0089692050890449043221, -0.010089751062407998142, -0.0010155894600243011057, 0.021035013856530487136, 0.049983185304140032024, 0.066465924767151907115, -0.0010214205299762953681, -0.14277710289495751383, -0.14484952939995621168, -0.037811155662066851801, 0.29879794937188991488, -0.054835251038376074673, -0.023843148606510124882, 0.025807828550008432567, 0.00027322255206700663249, -0.028767446231382526323, -0.053208466936041115547, -0.055375946280394494625, -0.013423427883080474993, 0.079183024488850520251, 0.18715382115221942949, 0.20673346098834530227, 0.018965735362310200129, -0.37827702934928025824, -0.52629639932757632526, 0.2871519972491450301, 0.26721821530596556737, 0.0011208262761333489393, -0.0013579211068901625176, -0.0016063182015150915313, 0.00078519598078109970659, 0.00079649109672201828108, -0.010883189640981366653, -0.025012970329730477986, 0.013291670707846229768, 0.12480395693987533789, 0.11237539886886882545, -0.29016350614345304271, -0.3626132361863532716, 0.68222115842382224926, -0.24136096183552005479, 0.00063120109794094101871, 0.00019207511376001539828, -0.00051528155992261329491, -2.3396410747603536666e-06, 0.0022395619090588426807, 0.0015339878775745329018, -0.010750454159834261711, -0.028447715050565359346, -0.013573466239391361945, 0.08752970970399978623, 0.15045226344282328079, -0.15539615061978040433, -0.2577869486895966733, 0.22733329974176982868, -0.013032834519223074343, 0.02790658838059019084, 0.056007971420674151908, 0.059498184883986934834, 0.028418776524853382737, -0.040744975120355478304, -0.14695275575126201528, -0.27726521671259440893, -0.37170738114195434054, -0.2958085934659676064, 0.15693704616074813374, 1.524815369301382173, 2.4284155189162941468, -3.2608013257945467345, -0.0069853187617580411034, 0.0009775164939312074551, 0.011857913266055279677, 0.02413540536491970423, 0.030555171546798869098, 0.013853104651724217761, -0.046526029204773410031, -0.14066429452502218767, -0.19944094416309410267, -0.050182906983795738665, 0.39178146004075525832, 0.46013592844686734828, -0.061834285505699596031, -0.45686311163510295863, -0.0065563647906613752026, 0.0068585033661754581796, 0.018731995598233117106, 0.016948688085143468612, -0.014799138120660618526, -0.069887985516154715881, -0.087934875171738188793, 0.013105163183845022074, 0.18706392482638589669, 0.24151645379049468954, 0.046216864794823182827, -0.44052976027456847152, -0.44208149081670433667, 0.54273096879510074331, -0.0029647736526581592621, -0.00091748228683707932531, 0.0036857352082378292031, 0.0059248018520414542573, -0.00049300920855115365048, -0.0069830993046911486039, 0.010682933469930192341, 0.047318022050133176637, 0.022969776473141089584, -0.1379560949580720397, -0.3311736654482473341, 0.075882333288054215625, 1.0821549800803444974, -0.78027447725672316459, -0.010871961513083299236, -0.0067400031145091734638, 0.0044154463292683954895, 0.02529976354236840308, 0.052617894785230739929, 0.066123663585686900945, 0.025027514024111040258, -0.10064907856164767574, -0.2878308309097095119, -0.42460837712457594195, -0.22371153704579183064, 0.9128097874277555368, 1.9766154137580296268, -2.0612194563606403008, -0.0064129312134067568626, -0.00012357858412370697912, 0.0054050338705074378945, 0.01021592062327046635, 0.01379497888520685514, 0.010962842378855809991, -0.011043910373147298315, -0.057897279783381966345, -0.087970687128867880444, 0.032818132038394683125, 0.22851399950777412373, -0.017557088057754001442, -0.18443192143879305611, 0.062057810937843800592, 0.0091152963865158073697, -0.0078817772586676944357, -0.026326162486986142558, -0.042280162511798703717, -0.045639857917653604835, -0.014570436229770786576, 0.083389154703915682743, 0.25870849166642534422, 0.39963918985466795331, 0.22057259708201026882, -0.54034813194768216427, -1.3391778787142170071, -0.45233985450502112169, 1.577590315531734344, 0.011108590466889453965, -0.0044323020248831770593, -0.023357661271105584211, -0.042088834337959901699, -0.049750693832530522298, -0.023103791046308528451, 0.070459412090157574893, 0.24009180641776839416, 0.381158220975374884, 0.234427807656163667, -0.46196734666907057765, -1.2676678253430573307, -0.4997094305458418817, 1.5078630756742190666, -0.0031496609711842314325, -0.0092002782897749428581, -0.0081172029998456633054, 0.0024231397201322038039, 0.020993549353178861394, 0.041625597439792730303, 0.051561628225087412225, 0.02948339915078752177, -0.059967181611644736206, -0.26473577558177785463, -0.47810066651046917441, 0.24653778058928454975, 1.6905048879888764368, -1.2792653678413188345, -0.0029455239250027113844, -0.0073699622912515878187, -0.0058432290010988563803, 0.0028091224438803884429, 0.016441742733749625804, 0.031078434802323563757, 0.040150819530892771503, 0.02907439459454023295, -0.036766173872208825213, -0.21465990443111515606, -0.43767544613239095508, 0.17165291927830256258, 1.5266825593265576089, -1.12950047161558631, 0.0079771144498115374027, -0.0028691155099636930087, -0.016942512440384679961, -0.034754356031231885271, -0.049100135848026651986, -0.031695215679503611905, 0.059676485417627773977, 0.22542563330241871533, 0.34890002028010047574, 0.1965420547087945724, -0.48605846514829353922, -1.1211852596546416638, -0.12824834745183658069, 1.0897370160252854987, -0.011892575548045332054, -0.01146826522667789526, -0.003014887978165693358, 0.017552900775692884677, 0.048011220265154509745, 0.067010154753095685365, 0.028455311733994816809, -0.10281788610276458573, -0.24753902663829879804, -0.10721517603309191347, 0.34586628168743827061, 0.22984406306531723896, -0.2201841147445613156, -0.033212115610387171749, 0.0066374212670948853909, -0.0034185566019625480624, -0.015483413041295834497, -0.020861592517206019315, -0.0014496283397861695844, 0.05288005934790612067, 0.10394510539300050311, 0.042300127086188597791, -0.20669977478075257649, -0.32917564158070500291, 0.16333952691212566544, 0.46722925017042199469, -0.059015283423655501283, -0.20614357704940206029, -0.0083730427404732771235, 0.007688750310790368539, 0.021724333374341046193, 0.028136400071791516087, 0.019378582442357811938, -0.0092623596026787814034, -0.053528145545059656329, -0.093886522452444609588, -0.12898759684637808864, -0.20050147206584020831, -0.16787300919451983439, 0.66390775630414256003, 1.6389303592664525855, -1.7735106283315722475, -0.0083966519514986018985, 0.0053059937980566764665, 0.018397727147574316486, 0.025924673813993714955, 0.020197126401668340223, -0.0044315556020424343511, -0.044638060922034830669, -0.08178125102015593173, -0.11421072409447703688, -0.18175946177251042779, -0.16735150334563364538, 0.5672217517771183104, 1.4928377585466183763, -1.5756324151427716362, 0.00010707737039112294236, 0.010008896757453372359, 0.018245519480445629434, 0.019824927411887460771, 0.0040854639975871371443, -0.040929475369762623183, -0.10989692415391318081, -0.15168251246867292537, -0.059882841767572358993, 0.27982359904670112138, 0.56170500234279985285, -0.065282657148286016291, -0.69573938435526661017, 0.20945508092275325063, -0.001213641229208212418, 0.0069028223094130597079, 0.014487246845267972434, 0.01843869967007554303, 0.010454604308036631946, -0.022590086505007350715, -0.086159393868528971749, -0.14777484573606899931, -0.10201640643914570983, 0.21583222443522009915, 0.55749242655503428168, 0.0040885708782745508807, -0.6599531438895880342, 0.17362496265907148563, 0.013701140071544465338, -0.0073140987594527877638, -0.029315973706489716066, -0.042116856858143106312, -0.030572900885853047243, 0.0024129028531977392107, 0.012596996513685601929, -0.018686591713249191271, 0.08568571582154305144, 0.65592284660688571218, 1.5116599624595039586, 0.65881391053218618659, -9.6845635100275959672, 7.0102264023263174408, 0.0062791899362739465179, -0.002888371459386039012, -0.013076591306495282535, -0.021624312012272689509, -0.021032723025967349495, 0.00024547936000527130658, 0.04981401774794640136, 0.11805360277374354894, 0.13124460077667099434, -0.012450342430615844724, -0.23891237620861596747, -0.35988107321092521751, -0.0238431485551682254, 0.41316104602060416129, -0.15655122602643817276, 0.1832538934848628176, -0.27610330636126495518, 0.020950540403670274425, 0.22495087878155506367, 0.12924523972216775403, -0.0015533732179366693989, -0.062619933853306344873, -0.063084530630497862025, -0.03853287114713492556, -0.012616451314674431358, 0.0046744783701308462512, 0.01289942751937150911, 0.01484293079263302069, 0.1166205099819224017, -0.2204240168069342698, 0.019948425036893382978, 0.11667617623321442644, 0.0082913343390318347548, -0.019029115167337001119, -0.026274214260180929481, -0.011727564258396187477, 0.0087266200291717752752, 0.010133735814438190118, 0.0017215658384458398713, -0.0033583048351795785971, -0.0026238443527438664107, 0.001388892639488278815, -0.83082883382830674535, 1.3476462017346111377, -0.15585865830895614947, -0.44973037555556871769, -0.031722562736353376867, 0.072977767593012138336, 0.041370768282898165435, 0.0081727129979936204152, -0.0060330193625194698803, -0.0062030017182922066177, -0.0021620871973356556653, 0.00062424092852509727777, 0.0012827078583601643144, 0.00027689924724481218164, -0.16423900866925461584, -0.21376761985297762503, 0.21413570019692035329, 0.25103398076855198973, 0.043052454820717421524, -0.0051131527985905419337, -0.048919456744397890602, -0.081262449685785290154, -0.057987807339953602315, -0.012591612866345143484, 0.018584744012142512953, 0.025497881422595462114, 0.014819875967352808682, -0.004455066607703942437, -4.374779778732241553, 4.6459224058445256489, 1.8839922564869884258, -1.0603454201756630226, -1.1329583073872826482, -0.42502576069684333548, 0.044178759228507356871, 0.17183748981938201661, 0.12982165344397864071, 0.054627543318006302475, 0.00063552263433813488067, -0.021125919262614759747, -0.016709753843204701307, 0.0028885342285279412555, -0.32383196851894135992, 0.68614320000607387229, -0.25295042981389020964, -0.25592719475201852752, 0.06074297568621357668, 0.089213420940773724421, 0.024389950565143758698, -0.0080133749182343602668, -0.01348457106299178962, -0.010392999242310602126, -0.0038354327148161075113, 0.0017020588278755417744, 0.0033067043151221672294, 0.0012275759560648029107, 0.3645394879255749454, -0.46788134129989550347, -0.24648849917550347266, 0.3269646980008857895, 0.18992805660911585286, -0.075231207210957842801, -0.10237990112918760044, -0.026868130518295799125, 0.019351979066059427081, 0.021208228019139324649, 0.008147539519780094483, -0.0012630809189772382491, -0.0038215860123660167109, -0.0019693874809553219701, -0.15656945723087939415, 0.3080219622768570753, -0.1377832699669828409, -0.038567133886130672282, 0.065782837692635293125, 0.0036130828770495485372, -0.04992535834011360002, -0.034310541517373317755, 0.0035362855449435777987, 0.022375176584503518595, 0.02035305999289764578, 0.0086510645365667146467, -0.0032279169954123874443, -0.01146910732448354342, 0.35480830959148146331, 0.02785890427454978252, -0.51787977140940677501, -0.13024687879225102516, 0.14352155456152901336, 0.13162665247217819608, 0.056636448893961299778, -0.0034680595296360996332, -0.02465818323252568639, -0.018329406369389997233, -0.0055212477237818686007, 0.0020325766328092853552, 0.0029596759498257282049, -0.00055058233772776793338, 0.63160339957469890937, 0.063908057344927882149, -0.70390189426715321641, -0.32743547687771090704, 0.081441586659174031171, 0.16013537825268264747, 0.11448402584843983543, 0.055200396342346073897, 0.013235553844581034177, -0.0086862297481588319342, -0.015157898787327017329, -0.013195784350511479549, -0.0076195755519528418978, -0.00093326024014622093513, 0.7109323895055605913, -0.042220898832960498948, -0.74686885685518999978, -0.29795012474692195159, 0.11987571582973566287, 0.18320188121968583594, 0.11808907810668071992, 0.045694957470682362466, 0.0012755044052616749169, -0.016338981530556507399, -0.017468220072275098681, -0.012015386735488742254, -0.0051262858787612108902, 0.0013361990848059389012, 0.32160915421685404292, -0.37459243266918318405, -0.11214945194435078923, 0.095990174812723136344, 0.052974095590825880098, 0.057051566131323908082, 0.015452783039797841555, -0.030199482560573080725, -0.02872449201312337469, -0.0064204229480935657431, 0.0079505888376853461885, 0.0087118933500699383826, 0.0025946052133810864997, -0.0046192778000074064329, 0.23068593929100092432, -0.33504605481429011737, -0.022250043573569897037, 0.12224637563182563782, 0.013360608080285081914, 0.016890996339442396201, 0.0057275874171536121301, -0.020224893275822448946, -0.017410730887690958196, -0.0019565885579708825956, 0.0067461145441364053083, 0.0059859018175420705687, 0.0012485707920800032111, -0.003481509934189358068, -0.53160308503286968218, 0.24884824493712098814, 0.76780424192790286408, 0.062915021861523964808, -0.53415582486587509514, -0.2850983730209991629, 0.058177347791274300337, 0.14957077001424509688, 0.093585409168498084886, 0.017594973488893803776, -0.023948399280741796552, -0.028811758397976092866, -0.013279165372507178641, 0.0089744241837689687352, 0.30538579430765949718, -0.23670429526244463059, -0.17590378651719060232, 0.13030534668063026293, 0.031717987164715424897, -0.05567208531067846522, -0.032694760462991558081, 0.002438543858189107065, 0.021601932472656272533, 0.024249902643727495183, 0.015915223374108215804, 0.0038933865580500475606, -0.0062182030244193175345, -0.012760957205325983502, 0.22401265349642168756, -0.10838509083125807964, -0.29178578611480371352, 0.066121276620179900152, 0.12622959323691546052, 0.020104183174230280673, -0.017006276771726430014, -0.0085869527055721962816, -0.00099813591294299966461, -0.001177786742252120121, -0.0015209239619278973419, -0.0004754539179169070106, 0.00028659674729141794685, 9.1681241101875625139e-05, -1.6173518744641788292, 1.5658465288196152798, 0.63653425350201042754, -0.14875887612029511931, -0.28677793553093511303, -0.23517670907360080368, -0.081545808026805124968, 0.037750537269046112454, 0.059842151330847179191, 0.030005202927328795265, 0.0031030474810164522005, -0.0058992482035710135102, -0.0038474667166562594167, 0.0018542676457546671247, -1.3968247952363932463, 1.3894628050576449674, 0.50290835663920396836, -0.14331348194301662957, -0.22824066777821278018, -0.19020593447394618103, -0.070367761759622407536, 0.027391334426021131387, 0.047604826922696602542, 0.025154596172091175932, 0.004261700720419480358, -0.0032610291689243934947, -0.0027233599808033190533, 0.00039209693606224532919, 0.12536820741395271472, -0.34666727142967040942, -0.12328833244367798372, 0.29379998344733115889, 0.18063522691963657119, -0.0002106332221187653302, -0.060693937935754227464, -0.060913233263322326405, -0.036858995506261746367, -0.010670839220493798238, 0.0053960009115553561262, 0.010177780194549479811, 0.0082008853501762766364, 0.0033153833069106313294, 0.14851406002730391243, -0.36957436348421485173, -0.11039729747324771025, 0.3016185208360984138, 0.1643874588071855547, -0.015659056675705409351, -0.063100451751693681368, -0.054990226675305460546, -0.029730117995516742307, -0.0061012207094955908313, 0.0068434235630449834209, 0.0095428485735500039738, 0.0066655703739146605208, 0.0016450312612190064188, 0.41316104603718456501, -0.023843148647182413336, -0.35988107296152421766, -0.23891237655973682208, -0.012450342142084394256, 0.13124460061012707679, 0.11805360276548612641, 0.049814017933842491548, 0.0002454791183488616883, -0.021032722773776404285, -0.021624312282228202314, -0.013076591009471932175, -0.0028883717037184717703, 0.0062791900204115800102, 5.9858804407363015443, -8.5987094888438466711, 0.80982182558534054895, 1.2080138738454686553, 0.74195638795458185299, 0.26893979315280752695, -0.039810903873396565056, -0.14538043280713891514, -0.11638011479780752122, -0.046012278422756847096, 0.0042999065016511538173, 0.020461259746213322774, 0.013536712216625002014, -0.0041356373655885101531, 0.18325389372392325571, -0.78109890217203692409, 0.80007948156062125289, -0.085262310809112423771, -0.25520652199011084615, 0.035761620881741687517, 0.10470867935237068713, 0.050578013619474400386, -0.0071250475544953860429, -0.030073095505650036807, -0.028233475462193506622, -0.014648002597295980282, 0.00044143408658313501884, 0.012899427360905539597, -0.22042401684976808962, 0.4656257186004904014, -0.17480700346101996079, -0.20137948779755276973, 0.10960928895866303701, 0.023152279621786241853, 0.01561488192156759193, 0.00023007533675999840891, -0.021259379129491971089, -0.0078919826036102178751, 0.0056778200778668597506, 0.0059023438461272394531, 0.0009455109312481002181, -0.0026238443427847304643, 1.3476462016959320778, -2.989889497045172817, 1.4854743172671311946, 0.70588113270852648196, -0.4280397856868343176, -0.16931280582529847911, 0.0088971547705819706287, 0.030346715589750140885, 0.017542571341706898747, 0.0030113319205475144215, -0.003162496529807923111, -0.0029028204441909392051, -0.0009954714604558558954, 0.0012827078948170092358, -0.21376761977456301689, 0.44777332370814848517, -0.331104577878358175, -0.14477725425735582521, 0.25593935525896699401, 0.023614545364057409893, -0.011146132229577665473, 0.018793298049151745122, -0.013894510553394030794, -0.034987982054965269252, -0.025821150610163198791, -0.0054918719937740809872, 0.0092596681465257329319, 0.014819875958536895405, 4.6459224058116399547, -8.5483991498254638941, 1.2876168981944362546, 3.3470839725657484109, 0.46966015864751381459, -0.63263828215184303438, -0.52895682325909609833, -0.18283309794263308268, 0.019699641688436908549, 0.070199464791016477627, 0.055523424775221479066, 0.02622242573365066165, 0.00019676554311398662244, -0.016709753879048262065, 0.68614319997632766679, -1.656787814352863375, 1.0426610570289536195, 0.28341598933551498396, -0.34884254906200817814, -0.089102930483734063549, 0.054619870431965242275, 0.035559649934379378156, 0.0058826021426113071083, -0.0019822351381033158357, -0.0050638906181088875688, -0.0051513548752396172425, -0.0017546125406745431263, 0.003306704356657088565, -0.46788134131532033155, 0.72620661617235338081, 0.27329512607181510875, -0.74197803715047006001, -0.06931373578018999726, 0.28165542185125119756, 0.090743545537424447067, -0.053482985893685865142, -0.052491302675492894358, -0.0097780965591262695991, 0.010851758006096548156, 0.01019997598763301358, 0.0028142778395265021472, -0.0038215860511308799921, 0.3080219623206985613, -0.78122836598936362673, 0.61688972791776941484, -0.0038466872484467420218, -0.18571190901896070247, 0.0048546405082291674282, 0.060547828617230070591, 0.0089946202849924810491, -0.024579868387209123648, -0.014944950269299454512, 0.0015955557490854625161, 0.0093301680719493386429, 0.0067045437174319795839, -0.0032279169749300246767, 0.027858904211017949948, -0.38763713302104430003, 0.81541408093503464638, -0.24606865180892925626, -0.31890153319912545316, -0.017194237124282203688, 0.062396989975228273062, 0.061657933592069003037, 0.029069393390757425344, -0.00075993870001764599574, -0.011881710085326121357, -0.0091542115763504350312, -0.002840335720992828196, 0.0029596759586206378839, 0.063908057216591082628, -0.42863887784308307705, 0.92835867839954755443, -0.20142897745341539761, -0.4365392587533425961, -0.09660884244610933036, 0.055537855335689334524, 0.073835693478503688914, 0.050538370977457620914, 0.024581184278569971624, 0.0047407795210472033001, -0.0055476188662115007849, -0.008688573207710631463, -0.0076195753913453189571, -0.042220898912698055783, -0.25309919816776366153, 0.90769685589524762293, -0.26943418828084386707, -0.44617328136277345507, -0.085247093224280645618, 0.069463762698864173206, 0.084065029262881243999, 0.052330526265815899278, 0.020113567929990267685, -0.0006524759094649508484, -0.0086489047608036447035, -0.0088049617676337395972, -0.0051262857131055767149, -0.37459243264918090643, 0.71180159092369754692, -0.16437431616620379327, -0.28443618948545634773, 0.084569665372965480143, -0.018209584328549518334, 0.0099904259991497896404, 0.048360952538523467592, 0.012288178838845729285, -0.015509118402660185265, -0.014823918470107557213, -0.0031760070764209696013, 0.0036850781336005855073, 0.0025946051125326522954, -0.33504605477858728779, 0.68658250753035454661, -0.24330284331256207819, -0.26643382099890933201, 0.15109212669648955374, 0.00027548549519508416621, -0.01216343368294544279, 0.02698844320183820325, 0.0068055058290869370127, -0.010994674913230266869, -0.0093144062647381557773, -0.00091164583988983388663, 0.0031333525561282214375, 0.0012485707460487026951, 0.24884824483351811186, 0.13991990805268464393, -1.0280866597608619983, 0.19990842603250175147, 0.81623519479710637814, -0.037207687034751940613, -0.32461630681527509079, -0.12882366459668917469, 0.028273555337030319146, 0.066064690121568514503, 0.041715098222079832757, 0.0078177034107717851413, -0.011522585930297044379, -0.013279165334528350226, -0.23670429531252090705, 0.41529189787208986573, 0.17627151387044681652, -0.46575194388264917755, 0.014335429966352702946, 0.11265092114627009434, 0.00051781152776135129685, -0.019860938326115532004, -0.0084828812429713316978, 0.0028584560900370274174, 0.0091009395404817085978, 0.0092162513942861852556, 0.0032084148673102953417, -0.0062182029488685045329, -0.10838509081097388864, -0.0067081837288476822354, 0.48349167875275422812, -0.35717141810119840128, -0.15868030359271417806, 0.10262796618207729871, 0.059347230030636674569, -0.0012417813467976641147, -0.010315102268287253129, -0.0015043863596163752246, 0.00050369189694898493324, -0.00097140797543132065446, -0.0011001135677868485729, 0.00028659672082103823925, 1.5658465289246610297, -3.0045206824121231115, 0.71886299650479767998, 0.85213307022372763999, 0.12438402066170765881, -0.018667221760836440192, -0.12304522893244317083, -0.13111515714110685837, -0.045768209560542734426, 0.020894822698161265306, 0.029818293212984257889, 0.013775316153830645394, 0.00043179303028420361592, -0.0038474666856961620837, 1.3894628050330155578, -2.7112554986374930444, 0.73760296612995324139, 0.73061596344670132197, 0.063551495743779912173, -0.018346305430793831415, -0.095606415862024840546, -0.10695814839180933786, -0.03880971773075098108, 0.016110141928677955414, 0.02378417339789093421, 0.011294394976979711676, 0.00085334847207710557218, -0.0027233600455042659104, -0.34666727136378783225, 0.47843484743871117493, 0.17104867268871298824, -0.50735829822859690452, 0.020985073344686872715, 0.1769551496416307268, 0.058267971508707987194, -0.0034024150366850215536, -0.025683748730712826353, -0.027195587017943435165, -0.016432725262629628571, -0.0032978028714158043842, 0.005073340078548250337, 0.008200885303042667121, -0.36957436339464222463, 0.55162045433775719339, 0.12537488815538572196, -0.52972827129512256583, 0.042692625712214007927, 0.18363386506145357169, 0.049830139913509867433, -0.0099904461126094357137, -0.02624721150765091035, -0.024344800714220057869, -0.01320243853637132446, -0.0015235011399074081632, 0.0049986497956858956179, 0.0066655703020079160731, -0.023843148735623233975, -0.054835250637612187319, 0.29879794869507336763, -0.037811154970337740311, -0.1448495298290616573, -0.14277710266531729033, -0.001021420508011725882, 0.066465924478565113764, 0.049983185711453090305, 0.02103501336115944087, -0.0010155889360106680561, -0.01008975158430557538, -0.0089692047434909508152, -0.0028883716496979155129, -8.5987094888518633695, 19.400749621949263002, -11.311665238199511307, -0.82625368506899765642, 0.45704215256936031597, 0.49811940851471281366, 0.33943096153055735131, 0.13828241950359226875, -0.0047520529944511747181, -0.062240350241079558935, -0.053199273655351049517, -0.021372811857606227592, 0.0028037557707775474791, 0.01353671216891007105, -0.27610330698422791418, 0.80007948229738357071, -0.93803668677284113286, 0.57373965404035254245, -0.089943309064607360193, -0.24814520832509850967, 0.042977258916304203473, 0.1188344268947589405, 0.067314980802279475447, 0.0037982677788771781677, -0.025737771818652339928, -0.028040769202255847597, -0.014648002409625078893, 0.0046744785950880570966, 0.019948425118041061832, -0.17480700358369813396, 0.31418327853593908117, -0.071479740366097610504, -0.22983233444416378832, 0.12389219147903757223, 0.034581546316999153823, 0.0027495091746610098547, -0.0047142914734448803163, -0.019134688069587044851, -0.0055089294037670563306, 0.0070500335602543392696, 0.0059023438000827916847, -0.0033583047858691005873, -0.1558586582210000071, 1.4854743172596489575, -2.6986772152096012434, 1.1159525454776522135, 0.75280438393997495261, -0.37815562948483272265, -0.16751550716672594654, 0.0061183686730838342249, 0.02903957019730060013, 0.01730617774827140562, 0.0030733345492403051948, -0.0036083678338467511069, -0.0029028204635867405764, 0.00062424090178165594044, 0.2141357000963318713, -0.33110457752887362259, 0.27142036225085131029, -0.19557754261185866795, -0.25452967310838386927, 0.2348084148673725835, 0.082579883529108097373, 0.024730485947753597792, 0.029035897617163554385, -0.017595154787420977383, -0.044446902497279373867, -0.03430260405146742797, -0.0054918724480543280875, 0.025497881510490129353, 1.883992256563322476, 1.2876168980309434797, -7.060171098362119757, 1.234108654373744951, 3.3466272299437482829, 0.55553026088719070774, -0.59163461623645541287, -0.54557588730375561337, -0.21632628856350349933, 0.0030690044705609731759, 0.073049794699669060027, 0.064117782016212154783, 0.026222425702898465721, -0.021125919157985777186, -0.25295042976858356232, 1.0426610569358654157, -1.557293069431723298, 0.7740212323125222893, 0.30718266263501775581, -0.30742293340083420849, -0.092466371600602448311, 0.049176322882891060351, 0.039954813828757484995, 0.0091787589013596175558, -0.0023370970412911369618, -0.0065398479654467387215, -0.0051513548251523157487, 0.0017020587828153910682, -0.24648849918376447565, 0.27329512607924882905, 0.0806464578730355236, 0.26443859026394939127, -0.56284693822820186515, -0.050578743036952332479, 0.25423069981490664659, 0.079184927669660454264, -0.052882509360663916553, -0.051797704279826023432, -0.0099799480968110233781, 0.011185506262263576749, 0.010199975868211797608, -0.0012630808523372623765, -0.13778326998219123101, 0.6168897280698499852, -0.89922268631613289092, 0.46607202402181713286, 0.038649048853159541361, -0.12551659792238953339, 0.022071713990103641184, 0.054564567290777007957, 0.0017726941966098557889, -0.031980600965685591908, -0.020359151610740901239, -0.00072300645775455886326, 0.0093301680399495001, 0.0086510644382890508619, -0.51787977132510965017, 0.81541408083062993928, -0.54131382675732320209, 0.68674959057343976099, -0.22590326616675218685, -0.33883314069551506709, -0.024827559768687491176, 0.065979332420951788962, 0.065890745130267491803, 0.031137706635145787765, -0.00066627558662465849026, -0.013200556538525245673, -0.0091542115714365046575, 0.0020325766652288974889, -0.70390189406908620917, 0.92835867830040674775, -0.4489290864124410585, 0.79970154485311284454, -0.1716101960428462192, -0.45073627860734843953, -0.13031971023518842956, 0.033698602572765593011, 0.06823297076220312396, 0.053020027635896957441, 0.029659153170914061282, 0.0084161650173706636924, -0.0055476186069726507802, -0.013195784649709224254, -0.74686885673028213795, 0.90769685586235548946, -0.30150141765159904628, 0.77484558241042345017, -0.25134108635495600037, -0.4680401354513828438, -0.11755123013949413113, 0.05165334185928075611, 0.081597912940050643105, 0.056602910012403376272, 0.02572074066957232702, 0.002806153174885514523, -0.0086489046165281815992, -0.012015386979768357886, -0.11214945196713171915, -0.16437431628739865297, 0.57012313215502252728, -0.041132038090758250848, -0.37540667372161845572, 0.056325881882635139841, 0.013959005814330665424, 0.022936263400893441844, 0.047312495932184540914, 0.0085656272421609520179, -0.018188817544523384201, -0.016567865995613030583, -0.0031760073308559496306, 0.0087118935808052078151, -0.022250043652635703023, -0.24330284328129950233, 0.52524447920524752753, -0.079112603720522922668, -0.33892380260955384808, 0.12070195374834781499, 0.027627167134496959794, -0.0020685996631466326175, 0.024987930473802606735, 0.003420923377756461721, -0.012863343624460749776, -0.010309051233867176992, -0.00091164580552061157473, 0.0059859018772500010089, 0.7678042421735771228, -1.0280866601927853843, 0.13514094405179849501, -0.34499893398654180032, 0.19407088981279979611, 0.69832937603020872608, -0.063331434711118045344, -0.33980368076210759609, -0.13994509682612310542, 0.028757172903402003494, 0.074305480149298400172, 0.049769638337487508151, 0.0078177035075411831916, -0.028811758419311977314, -0.1759037865091402919, 0.1762715140830650451, 0.020680225739652682848, 0.27804482453293011268, -0.38064136807896764569, 0.00062799583100484404619, 0.11086299993946270581, -0.0013385356433535033557, -0.027825253881613092738, -0.016993071901289180659, -0.0021527558038138285522, 0.0072535146605037991926, 0.0092162512467470203426, 0.003893386491299505401, -0.29178578616777139931, 0.48349167881675197966, -0.31550407032485194847, 0.41999287474889623883, -0.27334583781054144769, -0.15705877409469812256, 0.084448705778758376139, 0.054719104678062818747, 0.0030163964902678508037, -0.0078479020371412765156, -0.001588208594879332667, -4.4810571192404626516e-06, -0.00097140789435866143481, -0.00047545393321184837654, 0.63653425324263468443, 0.71886299687085142285, -2.8714015919797923893, 0.63333998308074057615, 0.99824189556350595431, 0.19145507087210514308, -0.032908052617887616209, -0.13763794673509818778, -0.13323279593065232018, -0.047356886211423768196, 0.019441742218584152219, 0.030249284817999731734, 0.013775316520526371261, -0.0058992483240201372754, 0.50290835671743383539, 0.73760296585947193737, -2.608202471796906341, 0.63337156967786967243, 0.86024479483930449408, 0.1249120076812796909, -0.029152721468645920178, -0.10679897641717402135, -0.10822057894698797198, -0.040757276398002195705, 0.013950264078775306137, 0.023494634463071470409, 0.011294394585776424897, -0.0032610289644895902074, -0.12328833240635042601, 0.17104867244426658068, 0.013479741419064602409, 0.18572121262710808431, -0.4610141480137260328, 0.0018664409063900979854, 0.18787017164580346873, 0.076778579411854300241, 0.0064155445978263453063, -0.023739917171304879273, -0.028746826941849622228, -0.01866594000517343005, -0.0032978030740861735866, 0.01017778026929696579, -0.1103972976247511173, 0.12537488833096388507, 0.070692063013721356812, 0.16413741450938035271, -0.47505400028073929697, 0.022249961207961965748, 0.19296028515471796982, 0.066547274586305660282, -0.0019730175919586805111, -0.025399191257165255886, -0.026312623709084510121, -0.015458944574608878345, -0.0015235007285958088001, 0.0095428484692588362581, -0.35988107279783815295, 0.29879794845039209239, 0.058724508070933538473, 0.33265469547811765683, -0.11176467139455563415, -0.17846788686979900285, -0.158571066148265577, -0.014758799334450116436, 0.066041103954285279576, 0.055939585207641757958, 0.027687158094746876918, 0.0031926211768178039754, -0.010089751197471821115, -0.013076591201572812517, 0.80982182570054850324, -11.311665238418708412, 21.928213359515059011, -11.485022670312662996, -1.2109320734573119083, 0.31806831897915494034, 0.49757414416880407781, 0.37641475596355900013, 0.1638010989144157159, 0.0036698896651158200224, -0.064836606184815320786, -0.059621827797830678763, -0.021372812110257027179, 0.02046125982405545482, 0.020950541348622443022, -0.085262312258510358776, 0.57373965436019336295, -0.9857142297976517753, 0.59124224165746563919, -0.054497268611176152842, -0.23278320252841722127, 0.045817087042868825941, 0.11815083681943663874, 0.067686178148419248735, 0.00453921534466917697, -0.025737771440788505956, -0.028233475879156654659, -0.012616451497002856286, 0.11667617617106153249, -0.20137948774077060743, -0.071479740174208702963, 0.2987755816987366364, -0.038395963855985905389, -0.2257951593821802827, 0.10456287226842121696, 0.027669219641939642174, 0.0055800890819421387207, -0.00029421425921537756876, -0.017102014473223306479, -0.0055089289807349747438, 0.0056778199198485321611, 0.0017215658302580608575, -0.44973037562124013045, 0.70588113257323792382, 1.1159525458405996634, -2.7627379568212186989, 1.0890984064227868799, 0.77660611958592185022, -0.35682818337109417639, -0.16350410214091043448, 0.0037473042759565027872, 0.026849076031392121039, 0.016280676863718653025, 0.003073334979155845862, -0.0031624967102649895016, -0.0021620871815098740079, 0.25103398083436673227, -0.14477725482896902642, -0.19557754152305140138, 0.19292449747631337797, -0.19017872173459327012, -0.2100157110191765153, 0.25680974688809843354, 0.083882457475752181253, 0.014936545873208634597, 0.019687716960002370203, -0.022725956109672984518, -0.044446903714800131691, -0.025821149721007730787, 0.018584743791971659183, -1.0603454202086091129, 3.3470839726999854769, 1.2341086543228660943, -7.208575228866874518, 0.99197162663085647338, 3.2734383328912310951, 0.58696145265304866534, -0.55107648088370653205, -0.52202340540170300009, -0.20585910717579944396, 0.0070443033067222682667, 0.073049794723292135745, 0.055523424935634162203, 0.00063552243885876922181, -0.25592719473800790153, 0.28341598937318435159, 0.77402123229071295718, -1.6075343081227890796, 0.78240965558344344188, 0.32784391075306740237, -0.29787261806393144914, -0.089824345053563195429, 0.047689178030282837473, 0.037478076103519809836, 0.0075115149039239222095, -0.0023370967895530651674, -0.0050638908435555714596, -0.0038354326495407490422, 0.32696469803572086876, -0.74197803726529754087, 0.2644385905502852907, 0.25542596521287425082, 0.33093706211477291745, -0.58296349759395582613, -0.087984230227842186167, 0.24131798080727084876, 0.082933889119894438346, -0.047198724720508095087, -0.049711558115936929636, -0.0099799482219351323686, 0.010851758064322378564, 0.008147539536146136685, -0.038567134116384925446, -0.003846687093871685708, 0.46607202418469390048, -0.89547865615893673663, 0.51835394039426518908, 0.06436696573472480698, -0.13379913133692250549, 0.00842265185724514788, 0.046250057016467865934, -0.00010916911885781800835, -0.031713314791308878926, -0.020359151878252955414, 0.0015955557919869911999, 0.020353060089761842999, -0.13024687886148997307, -0.24606865174004813324, 0.68674959050326545018, -0.58048124412144819217, 0.6797668860855274886, -0.19723583006933223527, -0.32426567889184682025, -0.025991993074487636239, 0.059574927566532376311, 0.061300066144296998238, 0.029258858340603419168, -0.00066627545177903575811, -0.011881710143654970052, -0.0055212477198296524422, -0.32743547699603675616, -0.20142897741564128045, 0.79970154481613897612, -0.49790803676317363369, 0.78771268046221898373, -0.15996581155020581089, -0.44782510668990971325, -0.1332606975872081545, 0.030271521038812573834, 0.067503834410668317467, 0.053898350247408174929, 0.029659153222782348724, 0.0047407794816006938732, -0.015157898555957227682, -0.29795012493915962182, -0.26943418796067897247, 0.77484558208879328323, -0.35861454110635893056, 0.76379914743830323864, -0.23439347815040034728, -0.46176331092977995763, -0.11981501873654736012, 0.04778856822101630597, 0.080418616181376703644, 0.057296727383462017769, 0.025720740672276587446, -0.00065247590633180779784, -0.017468219951321632688, 0.095990174728357774425, -0.28443618915762902688, -0.041132038498211509481, 0.48994603268184233613, -0.037241677209162965534, -0.33912129012964109531, 0.067486346336304861571, 0.0081493849788834265419, 0.013108968228434171241, 0.042771489762300995974, 0.0080841715983996070144, -0.018188817767817389015, -0.014823918206464847119, 0.0079505886837140798962, 0.12224637569620352973, -0.26643382106219215544, -0.079112603696096087114, 0.46798671331413471686, -0.081377093704425404463, -0.3151191244432242633, 0.12642977657963919147, 0.02335139139865181207, -0.0085547729403143025345, 0.022338849891872413628, 0.0033986680745238246996, -0.012863343539972460147, -0.009314406375480711528, 0.0067461145263813708864, 0.062915021634029802411, 0.19990842670405650772, -0.34499893480842086513, 0.16582832373055997333, -0.49729858172655561077, 0.096249621269606247531, 0.68124892304762107997, -0.036336574009754936398, -0.31523185781565588037, -0.12967640257105034207, 0.031161712445086971884, 0.074305480355337502618, 0.041715098327224822694, -0.02394839940691469507, 0.13030534648176089751, -0.46575194383051665703, 0.27804482452748763288, 0.092107782985477379412, 0.28983034392806056312, -0.3809564650100920602, -0.01238905726031122069, 0.10051453809658242311, -0.0066907894478000148886, -0.029625341607769983127, -0.017565220656634222135, -0.0021527558297502673128, 0.0091009395072460946841, 0.015915223583670173868, 0.066121276696527161998, -0.35717141824813214424, 0.41999287487518627282, -0.26085018484880506673, 0.44179278642057934201, -0.27461936876113113293, -0.16678565708175549509, 0.081867938861953232221, 0.055370999023787371984, 0.0033094411624280863776, -0.0081593345552722247593, -0.0015882084171814287495, 0.00050369168970026490309, -0.001520923883952663867, -0.14875887579750610157, 0.85213306956445522555, 0.63333998369864796807, -2.7824773927541270879, 0.60121684299709676758, 0.94194578911343296301, 0.17433315958324815953, -0.018060706645538673887, -0.12106158740694808462, -0.12680010642460673442, -0.047394672454659420024, 0.019441742656317713422, 0.029818292626510679855, 0.0031030475955913152525, -0.14331348204697599846, 0.73061596396681061094, 0.63337156905238134463, -2.5355033817677998798, 0.60875627294783452026, 0.81690702141543392045, 0.1122924794581618424, -0.017007483400645083405, -0.094094413721329678291, -0.10376156797726565628, -0.041216287166886474147, 0.013950263093278861271, 0.02378417426252840694, 0.0042617004774568225528, 0.29379998328180739753, -0.50735829750784233738, 0.18572121189677612274, 0.061886001617972419275, 0.20912872925146969161, -0.44348116356364414292, 0.0014789096560260412555, 0.18370326283464974426, 0.074098162513675389085, 0.005791705867594630977, -0.023927402582103886319, -0.028746827447089885049, -0.016432724823968340722, 0.0053960007450225035142, 0.30161852107335335171, -0.52972827172766923098, 0.16413741490982208804, 0.12954491392837402808, 0.18668430823072543201, -0.46142527166611113643, 0.019140228803521182332, 0.18804328968345962458, 0.064055363068431764306, -0.0024515949019933078179, -0.02562463796052385015, -0.026312623156337238556, -0.013202439284035364514, 0.006843423801275532424, -0.23891237653694991083, -0.037811154954693705632, 0.33265469533302027827, -0.05142801587151299425, 0.30066661314132303584, -0.1048459507217706782, -0.16949845819516079493, -0.15226013323148132605, -0.014012614842301237064, 0.066110440447268881137, 0.056950599240164039017, 0.02768715882270433859, -0.0010155895382098567917, -0.021624312063981920157, 1.2080138735670080674, -0.82625368453171832428, -11.48502267068427507, 22.455342777837245194, -11.316496523784628891, -1.2016035442347696716, 0.29021559625649523406, 0.46316134177653034421, 0.35349091581775876625, 0.15496784213381734063, 0.0014592362041702545278, -0.064836606596513773892, -0.053199273266403265814, 0.0042999064578233773398, 0.22495087827587700868, -0.2552065210403673956, -0.089943309285712286227, 0.59124224140051695464, -0.96024384563609677112, 0.61031710860892396386, -0.045590492657991713021, -0.2340173361705805144, 0.04223235022093038693, 0.11642769879232446606, 0.067686177700031366355, 0.0037982679766775297864, -0.030073095488414493953, -0.038532871132069268505, 0.0082913343690813427755, 0.10960928907245413855, -0.22983233491795632131, -0.038395963550312411927, 0.3104645458011436876, -0.053904337007103148016, -0.23346696194723765005, 0.10615225420826199243, 0.031326469893331182481, 0.0076371360769107773242, -0.00029421393406330259915, -0.019134688680716435, -0.0078919822558656763795, 0.010133735717580697605, -0.031722562774561417764, -0.4280397853091671978, 0.75280438320058440382, 1.0890984068828228892, -2.7816557655289000373, 1.088789830144390347, 0.78341458752708181823, -0.35350516503443035754, -0.16387002309373399633, 0.0024815623978382784706, 0.026849076343468742528, 0.017306177117244591485, 0.0030113322011079916157, -0.0062030017185233498472, 0.043052454650219285082, 0.25593935585131216692, -0.25452967435986073585, -0.19017872061793103522, 0.25063551463012040399, -0.16528356472876287731, -0.21390572705352148275, 0.24292906480205486708, 0.072203610852622182592, 0.0089943527246521322999, 0.019687716044939591181, -0.017595153563821106651, -0.03498798290487337842, -0.012591612534952036084, -1.1329583073842413032, 0.46966015861514137653, 3.346627229802168646, 0.99197162685032580498, -7.3753434485125293207, 0.91579648377778255508, 3.2731371057620761711, 0.62050617014201780997, -0.52077485973973436373, -0.50913716428027977479, -0.2058591072613070172, 0.0030690045274049926848, 0.070199464614655648909, 0.054627543466390079963, 0.060742975640017161931, -0.34884254899711231168, 0.30718266259989557287, 0.78240965556574071371, -1.6123298777326158238, 0.78332775153014744252, 0.33566412739532219822, -0.29585237362336602729, -0.092617164773917043386, 0.045148805238876474344, 0.037478076374740355858, 0.0091787583911836119477, -0.0019822347580796643995, -0.010392999354533869263, 0.18992805656750463839, -0.069313735520982425187, -0.56284693880977154645, 0.33093706253159438369, 0.24470524633761991939, 0.31075412928784978961, -0.59298482787167505759, -0.090399607137918766564, 0.2435231527415436803, 0.085410204809935899584, -0.047198724883232102967, -0.051797704141953920809, -0.0097780965719841111267, 0.021208227981795058081, 0.065782837848810712944, -0.18571190919164859423, 0.038649048775531373578, 0.51835394059139361378, -0.88165329072498499485, 0.50750229124510448653, 0.052630627892426298797, -0.14234223859420602754, 0.0054319968888086704595, 0.046565072823512243527, -0.00010916930326372456875, -0.031980600819742625862, -0.014944950243156462227, 0.022375176593954774346, 0.14352155460093640738, -0.31890153312659508256, -0.22590326628976092893, 0.67976688616370462004, -0.5625847439795569338, 0.70040474990639400854, -0.18895049714372968985, -0.32741456412672759191, -0.031852823007700842295, 0.055952718114745142453, 0.061300066252362145347, 0.03113770647199717162, -0.00075993849860543443396, -0.018329406495842776037, 0.081441586657281350714, -0.43653925873313237371, -0.17161019598156024291, 0.78771268036517771982, -0.50303596723892418652, 0.79523950740546711646, -0.14946186094812261014, -0.44371254577416791021, -0.13315318019460070653, 0.029727393649076231524, 0.067503834709539814463, 0.05302002765371344184, 0.024581184224467207688, -0.0086862298296223553612, 0.11987571604292726613, -0.44617328196546285346, -0.25134108556904877219, 0.76379914689968864128, -0.35612954987604472645, 0.77731494266804634918, -0.22214687756117867945, -0.45834450018404837568, -0.12088288321141685278, 0.046580809821757079847, 0.080418616653426033825, 0.056602909740251200477, 0.020113568070323856185, -0.016338981594849512347, 0.052974095720987497693, 0.084569665098336369025, -0.3754066732900093184, -0.037241677564649043608, 0.53637069048318652964, -0.018957176723722556771, -0.34337436720053976513, 0.057710511011583261187, 0.0024255826138199071407, 0.011519276582341008916, 0.042771489581493925269, 0.0085656275443790740737, -0.015509118593002153147, -0.0064204229233670584431, 0.01336060804383887822, 0.15109212677614983211, -0.33892380267179295084, -0.081377093778590814388, 0.50287668270397289927, -0.069827965231967464854, -0.31995288963240670066, 0.11928816454525301149, 0.020093430671343483207, -0.0091179792075104966476, 0.022338849817768954148, 0.0034209234606415242497, -0.010994674942931766809, -0.0019565885071285717654, -0.53415582461764654632, 0.81623519407770372958, 0.19407089077945474664, -0.49729858229162571259, 0.081404226561651873983, -0.52894803308587856616, 0.10384691760609365829, 0.7056108338824818027, -0.017044232954465773217, -0.30822264012616046314, -0.12967640229092694537, 0.028757172803103123182, 0.066064689908225407433, 0.017594973609036401035, 0.031717987330509858634, 0.014335429645908981042, -0.38064136771462930842, 0.2898303436364982888, 0.096554224059707607908, 0.28507973908666850438, -0.39055172863453335719, -0.019625398564893269809, 0.098499217536447358778, -0.0065463360684880995619, -0.029625341242973324968, -0.01699307238918484203, 0.0028584565079359932835, 0.024249902382649363847, 0.12622959321587814974, -0.15868030354950321548, -0.27334583785733646, 0.44179278633415036692, -0.26429522588578907571, 0.44022531020890409703, -0.27249936578183769909, -0.16724014903203290872, 0.08037038232741448851, 0.054593002909149528079, 0.0033094414133882582921, -0.0078479024300555297322, -0.0015043860672247259206, -0.0011777868157847018373, -0.28677793582163746411, 0.12438402131884905566, 0.99824189481372394006, 0.60121684340687564507, -2.8699064824573330412, 0.55313548698809278381, 0.94198822368157009155, 0.18893616298793280972, -0.0071634959003894528584, -0.11721853054139826089, -0.12680010629369647956, -0.047356886734838309072, 0.020894823226316548703, 0.030005202847832451729, -0.22824066767844228276, 0.063551495208334543574, 0.86024479567155842652, 0.60875627257348807753, -2.6077689844899292204, 0.56946006116802461161, 0.81713094971370259234, 0.12348884209375114718, -0.0089461871893323254012, -0.091373456509965533812, -0.10376156889071960765, -0.040757274902756310264, 0.016110140937765370789, 0.025154596373020553235, 0.18063522710164173213, 0.020985072501415805157, -0.46101414681048269451, 0.20912872868621479316, 0.089045634846224672176, 0.22132422938142637392, -0.44661390526354483033, -0.0062505094139614584409, 0.17933776558799180423, 0.072952128785075343043, 0.0057917055133880242235, -0.023739916503712640838, -0.027195587566765072607, -0.01067083900149074123, 0.16438745856991432426, 0.042692626272816454336, -0.47505400085954446743, 0.18668430856373569027, 0.15376317057542737698, 0.19633277575735932907, -0.46579291673421596709, 0.011482969526153484485, 0.18413240362572005915, 0.063133469813610676447, -0.002451594816377473951, -0.025399191729159380226, -0.024344800216887528543, -0.0061012208635970628656, -0.012450342254038924972, -0.14484952962765307305, -0.11176467145319370078, 0.3006666130415601712, -0.05079847283503208244, 0.31072634237910323751, -0.091170754818720861623, -0.16234004900346146738, -0.15076101342239725533, -0.014263427620835517073, 0.066110441004613468285, 0.055939584468974609677, 0.021035013860373180661, -0.021032722943985202729, 0.74195638823712628529, 0.45704215200568537592, -1.2109320729086006097, -11.316496524029243886, 22.692234419306096527, -11.231265589982609043, -1.2167050283072775585, 0.25203512506007819116, 0.43745412290772922592, 0.34418139904575523191, 0.1549678419420635056, 0.0036698901846131544956, -0.062240350706435632056, -0.046012278319502913915, 0.12924523952018598938, 0.035761621070488670837, -0.24814520782991211778, -0.054497269954654030399, 0.61031710940714500868, -0.95213476629056015277, 0.61126890438643333159, -0.047462382931871709146, -0.2359210638871252208, 0.042232349952006663363, 0.1181508366669294513, 0.067314980797919032884, -0.0071250476258108421876, -0.06308453025781536605, -0.019029115232893641096, 0.023152279642955280597, 0.1238921915856233541, -0.22579515932285576518, -0.053904337274342656317, 0.30522407589197925182, -0.051029889108673121001, -0.22987463725329901343, 0.10784543835726222161, 0.031326469666298026318, 0.0055800891980672005441, -0.0047142914150412442223, -0.021259379205982337213, 0.0087266200636342185909, 0.072977767661702788726, -0.16931280620152705518, -0.37815562889990078643, 0.7766061193052778977, 1.0887898299994276385, -2.7875967328759885255, 1.0860321555847947117, 0.78338706206910213581, -0.3536315781441076167, -0.16387002356085356158, 0.0037473047260208839498, 0.029039569888367523093, 0.017542571602417320747, -0.0060330194846971102809, -0.0051131526862685282644, 0.023614545164296788654, 0.23480841503813870697, -0.21001571131532184578, -0.16528356426013052216, 0.24941580220630349407, -0.1789758681270275098, -0.22666040274574106306, 0.2362399781287625733, 0.072203610235034007547, 0.014936546068237507856, 0.02903589760070596909, -0.013894510168774135833, -0.057987807559211963782, -0.42502576081080722936, -0.63263828201890637271, 0.55553026092448665185, 3.2734383328038179073, 0.91579648392194956674, -7.4191247344801460883, 0.90927272138896098053, 3.2887632509962956284, 0.63482114662301958319, -0.52077485985371585464, -0.52202340527505386447, -0.21632628865935221718, 0.019699641675117771089, 0.12982165346754268032, 0.089213420952348562731, -0.089102930542712177608, -0.30742293349111371459, 0.32784391106645877079, 0.78332775131322007756, -1.6109714300891220251, 0.7836948884244804292, 0.33417318837792947539, -0.29770424824435065547, -0.092617164790368300409, 0.047689177729516508963, 0.039954814304459095042, 0.005882601843393431805, -0.013484570963768670315, -0.075231207116199752538, 0.28165542147046979116, -0.050578742535021238569, -0.58296349782842704634, 0.31075412932933915711, 0.24167345679309337747, 0.31425419882755700751, -0.59030141629232002476, -0.089338724128328966811, 0.24352315242014424546, 0.08293388950526278669, -0.052882509563953829235, -0.052491302553517658891, 0.019351979005817417784, 0.0036130830768239027145, 0.0048546401748003889173, -0.1255165978012465211, 0.064366965707579118505, 0.50750229142731773102, -0.89558891494292247604, 0.49906795709663676064, 0.050231234068362579093, -0.14209518071860205435, 0.0054319968559470646619, 0.046250057161503162206, 0.0017726941496154786492, -0.024579868306308400028, 0.003536285427352825559, 0.13162665251435010094, -0.017194237345701402636, -0.33883314030689321017, -0.19723583030050637066, 0.70040474997614665664, -0.55480151916323028338, 0.70041005038981274389, -0.19174071982283472093, -0.33016409557191023216, -0.031852823094370173729, 0.05957492768199439126, 0.065890745189342903032, 0.029069393175516055478, -0.024658183122820157007, 0.16013537839477842706, -0.096608842852859658223, -0.45073627793231624272, -0.15996581213020119594, 0.79523950767168627429, -0.49447722395587723199, 0.80260184136847412262, -0.14504378089242261907, -0.44247252193978503509, -0.13315317991519731722, 0.030271521527873004875, 0.06823297008806834818, 0.05053837121578185626, 0.013235553665219060987, 0.18320188115148045016, -0.085247092909408073158, -0.46804013589137249962, -0.2343934775634442802, 0.77731494219895636011, -0.34484375326522914706, 0.78493024897947816232, -0.21851358523691616353, -0.457803008058269012, -0.12088288247425407751, 0.047788567612809342011, 0.081597913297014157408, 0.052330526061646932823, 0.001275504373168256933, 0.05705156614114630792, -0.018209584320534964036, 0.056325881749289909151, -0.33912128998264451374, -0.01895717680393228799, 0.53527098832124153649, -0.026283669147126188792, -0.34794215887823232158, 0.056320832956584278295, 0.0024255826896281921373, 0.013108968223226186156, 0.047312495813237424869, 0.012288178979570048635, -0.028724492049060849908, 0.016890996371078749461, 0.00027548545756711371102, 0.12070195365196906279, -0.31511912424623667039, -0.06982796542904613768, 0.50111601212780121362, -0.075331100088286848115, -0.32284495580679606341, 0.11872385409203135842, 0.020093430798647834296, -0.0085547728239074114637, 0.024987930206506813596, 0.0068055060172320771053, -0.017410730998505270783, -0.28509837327713688504, -0.037207686530022064664, 0.69832937579841969367, 0.096249621312445757249, -0.52894803317272909293, 0.087117622318482718091, -0.51370097380287838007, 0.11670772122758865663, 0.71244211427188863794, -0.017044232678768082845, -0.3152318578855988207, -0.13994509704538182771, 0.02827355547581223566, 0.09358540914848391401, -0.055672085086347321281, 0.11265092097715265207, 0.00062799572795735293315, -0.3809564648321001612, 0.28507973892470467314, 0.090225147255570015137, 0.28042110319315960432, -0.39223482206640436498, -0.019752765234940981853, 0.098499217710363962119, -0.0066907896815491694695, -0.027825253527573381995, -0.0084828813931651977226, 0.021601932433728564253, 0.020104183089286315617, 0.10262796651074380938, -0.15705877456934974612, -0.27461936811332188224, 0.44022530959574784903, -0.26013670311137171431, 0.44296659841474333996, -0.2727522353095998553, -0.16825411245234989099, 0.08037038245927903557, 0.055370998650612589242, 0.0030163970953064321058, -0.010315102689303471545, -0.00099813584172489012991, -0.23517670896952194703, -0.018667221945397841076, 0.19145507077526555118, 0.94194578940527862176, 0.55313548689977498629, -2.8822077489376223269, 0.55584721727003427727, 0.9468779638548591171, 0.19189957261689741252, -0.0071634957848812855913, -0.12106158781463174923, -0.1332327953433094514, -0.045768210070605601369, 0.059842151480903632566, -0.19020593459401161085, -0.018346305066401918937, 0.12491200725094930846, 0.81690702136562354241, 0.56946006164821283413, -2.6185274500217619043, 0.57057699052802068884, 0.82028120135109372502, 0.12551021467828352152, -0.0089461873322378149265, -0.094094413289266307099, -0.10822057960271121158, -0.038809717302543157713, 0.04760482682879870614, -0.00021063346419733671149, 0.17695515043218193463, 0.0018664396290502720602, -0.44348116280325189376, 0.22132422946975946476, 0.091005067415972848432, 0.21760491893632316618, -0.45065938475245975736, -0.0084407656545807622367, 0.17933776597072720316, 0.074098161948482330064, 0.0064155448177199521018, -0.025683748675704106551, -0.03685899548101675971, -0.015659056607679605611, 0.18363386481689769614, 0.022249961251255709338, -0.46142527157104729074, 0.19633277580305016907, 0.15473352931494213647, 0.19264464082263221822, -0.46953430420512604071, 0.0094972019036242173728, 0.18413240337868427843, 0.064055363051609345848, -0.001973017292908323117, -0.02624721186942595777, -0.029730117788460755368, 0.13124460065692875044, -0.14277710286254194361, -0.17846788646606237538, -0.10484595081332012678, 0.31072634206712451688, -0.038548299282029886703, 0.31896594930734489814, -0.087094409740205283588, -0.16111454165920188752, -0.15076101328794441758, -0.014012614717310878723, 0.066041103859133157283, 0.049983185509228277021, 0.00024547925278273099829, 0.26893979309597221272, 0.4981194087361780487, 0.31806831869766388765, -1.2016035442139889611, -11.23126558990681545, 22.743012812821639557, -11.228962665687920008, -1.2323793830138953798, 0.24027984292435111691, 0.43745412287900747872, 0.35349091603347704282, 0.16380109855801544549, -0.0047520525300436744151, -0.11638011498507599972, -0.001553372834793234885, 0.10470867927453737789, 0.042977257881042174592, -0.23278320060306512129, -0.045590493943319716774, 0.61126890404843881033, -0.9535449229601039578, 0.61008305651681982162, -0.047462382583963069749, -0.23401733593878221162, 0.04581708714013533329, 0.11883442743872890679, 0.050578013107629576917, -0.062619934112480513111, -0.026274214171761692271, 0.015614881750963708479, 0.034581546539491796188, 0.10456287201341107895, -0.23346696162344288128, -0.051029889395462803281, 0.30951030922228039977, -0.049067508863481684722, -0.22987463723395254456, 0.10615225431340405415, 0.027669219584216288299, 0.0027495091638441047574, 0.00023007537034998765168, -0.011727564287160477038, 0.041370768257982214189, 0.0088971549420078840237, -0.1675155073909829484, -0.35682818336880828269, 0.78341458784863915454, 1.08603215522890717, -2.7909331662343284108, 1.0846428886849803774, 0.7833870618918499229, -0.35350516450717206496, -0.16350410290522451873, 0.0061183694416931605833, 0.030346715119949053963, 0.0081727131241392421412, -0.048919456477999258281, -0.011146132980449320257, 0.082579884215226273536, 0.25680974670238632118, -0.21390572746273969762, -0.17897586752373939589, 0.23726027532507906748, -0.18557167635775068115, -0.22666040325005737399, 0.2429290653297501712, 0.083882457679016364516, 0.024730485110274546406, 0.018793298759536111037, -0.081262449926336396966, 0.044178759359430858078, -0.52895682334865989915, -0.59163461612151446722, 0.58696145253111531304, 3.273137105510441458, 0.90927272164637462648, -7.4194665367132612843, 0.91324398685049057445, 3.2887632509000512826, 0.62050617017568643341, -0.55107648086938221255, -0.54557588719672645006, -0.18283309795160057609, 0.17183748973721832409, 0.024389950591463250423, 0.05461987040447395525, -0.092466371350604689705, -0.29787261853294849656, 0.33566412771917020219, 0.78369488826805555615, -1.6137954526157873847, 0.78163698084614174899, 0.33417318828723846513, -0.29585237357726013041, -0.089824344832536470729, 0.049176322529005708772, 0.035559650184376428994, -0.0080133750246618243307, -0.10237990126277136949, 0.090743545924944019054, 0.25423069934348452303, -0.08798422995172296146, -0.59298482797687290802, 0.31425419887615507752, 0.24681478714725399448, 0.31656346911277588596, -0.59030141653964651738, -0.09039960663925070139, 0.24131798028147788404, 0.079184927956058331899, -0.053482986051457209742, -0.026868130425800166444, -0.049925358601140480685, 0.060547828869053424428, 0.022071713996860489737, -0.13379913141504934382, 0.052630627817875232088, 0.49906795740611475942, -0.89817849551215633852, 0.49945359098509711782, 0.050231233863948031304, -0.14234223843143123389, 0.0084226516092489982224, 0.054564567317257325896, 0.0089946202652084229767, -0.034310541379435108289, 0.056636448778660718195, 0.062396990147512237934, -0.024827559862103970589, -0.32426567894952584847, -0.18895049709117306413, 0.70041005053373850409, -0.55695045681126542636, 0.69868295558618076502, -0.19174071998170452713, -0.32741456389549372652, -0.025991993322953332174, 0.065979332490631731334, 0.061657933631731373647, -0.0034680595281511308013, 0.1144840255224739145, 0.055537855969979794957, -0.13031971120589583291, -0.447825105669586776, -0.14946186154282220526, 0.80260184185545324631, -0.48971468493465608596, 0.80493316783663715341, -0.14504378119384167545, -0.44371254566108742079, -0.13326069851568733693, 0.033698603567151763638, 0.073835692860971288698, 0.055200396768879778087, 0.11808907808862062483, 0.069463762309075402501, -0.11755122954410920921, -0.46176331167943485134, -0.22214687676984096187, 0.78493024891421891986, -0.34049227302542772744, 0.78682242874791830101, -0.2185135850453910622, -0.45834450103057294834, -0.11981501795882162908, 0.051653341076134492993, 0.084065029693876511252, 0.045694957550687205494, 0.015452782913971355172, 0.0099904263350590449128, 0.013959005464410707831, 0.067486346526546517333, -0.34337436718635100386, -0.026283669257457027008, 0.53180650668026063776, -0.026926972462614565118, -0.3479421585718924792, 0.057710510694062369663, 0.0081493851109978119457, 0.022936263493015709991, 0.048360952383600802618, -0.030199482494460451465, 0.0057275873778003753151, -0.012163433466748181308, 0.027627166997895236805, 0.12642977660130244599, -0.31995288957378942296, -0.0753311003264951623, 0.49910764121499617829, -0.07543735123904760087, -0.32284495564119264355, 0.11928816451643071972, 0.023351391118872896757, -0.0020685991830550372328, 0.026988442861324221655, -0.020224893184270342772, 0.058177347717846869535, -0.32461630657062462424, -0.063331435088915408183, 0.68124892328567776545, 0.10384691753366455391, -0.51370097373103218441, 0.097278449157587765228, -0.50976103418302720627, 0.11670772149899714953, 0.70561083340120278606, -0.036336573710519931568, -0.33980368064548022167, -0.12882366487591223603, 0.14957077014886399624, -0.032694760509616185851, 0.00051781128211133187374, 0.11086300041951743789, -0.012389057641593988202, -0.39055172831598133909, 0.2804211032259328884, 0.089644395567892290466, 0.28086334022600256688, -0.39223482186660552085, -0.019625399024401926629, 0.10051453841844071313, -0.0013385362466741611114, -0.019860937847545442198, 0.0024385438409884091678, -0.017006276659720348521, 0.059347229480239641064, 0.084448706630184250388, -0.16678565806719194975, -0.27249936483074216342, 0.44296659795438025897, -0.25934555619418031558, 0.44273656619308365823, -0.2727522352603507505, -0.16724014918648927308, 0.081867939191247740549, 0.054719104143144799146, -0.0012417809756892607756, -0.0085869527483731708517, -0.081545807845945603165, -0.12304522942223990178, -0.032908051669588399279, 0.17433315863852391625, 0.94198822403007453907, 0.55584721714788565272, -2.881959109560738419, 0.5554786921705386904, 0.94687796350975750315, 0.18893616324848838683, -0.018060706664650233444, -0.13763794680030860818, -0.13111515693586620057, 0.03775053710947071095, -0.070367761707092274515, -0.095606415748471645921, -0.029152721901058761822, 0.1122924801786284732, 0.8171309487906812663, 0.57057699112974713529, -2.6193778736583701594, 0.5698020279674698152, 0.82028120119604941429, 0.12348884205119547985, -0.01700748305366788643, -0.10679897685014388964, -0.10695814802062988957, 0.027391334284421114953, -0.060693937699758747362, 0.058267971060798513794, 0.18787017226032659556, 0.0014789092846534471204, -0.44661390544323387175, 0.21760491902028786249, 0.088740380263166704355, 0.21635910706977862228, -0.4506593842597426125, -0.0062505101816483898264, 0.18370326417053184365, 0.076778578225384919165, -0.0034024143766204352317, -0.060913233578321186734, -0.063100451734298124151, 0.0498301399669181036, 0.19296028535058756814, 0.019140228475739134301, -0.46579291664179289789, 0.19264464063781056025, 0.15279305690884292801, 0.19156833556369504135, -0.46953430450251326667, 0.011482970106765922821, 0.18804328914476609036, 0.066547275179353618246, -0.0099904464216658897535, -0.054990226731213689981, 0.11805360260018864238, -0.0010214200517627437109, -0.15857106663060613339, -0.1694984581658286471, -0.091170754416587285185, 0.31896594900240932136, -0.034555330835401854017, 0.32065892109322940984, -0.087094409631116267989, -0.1623400488854912227, -0.15226013384749262358, -0.014758798521490933825, 0.066465923942491761389, 0.049814018072237009616, -0.039810904013317036099, 0.33943096175205772491, 0.4975741438063412958, 0.29021559674807767548, -1.2167050284999629817, -11.228962665733241977, 22.746742905418003033, -11.230541654147055297, -1.2323793824816571263, 0.25203512477976586226, 0.46316134187172519621, 0.37641475583075473255, 0.13828241941204164056, -0.14538043268579090017, -0.062619934924687203903, 0.050578014535357450343, 0.11883442658770081157, 0.045817087856351051967, -0.23401733607375740931, -0.047462383161957888422, 0.61008305663681672293, -0.9535449229044052899, 0.61126890428984737014, -0.045590493287298743119, -0.23278320232866958617, 0.042977258264997424131, 0.10470868001130645419, -0.0015533730546679314598, -0.011727564261457016845, 0.00023007534331060671773, 0.0027495090852916056166, 0.027669219786761223984, 0.10615225407448376782, -0.22987463703073413024, -0.049067509022894938153, 0.30951030938935708559, -0.051029889492213778568, -0.23346696163740296437, 0.10456287194291229159, 0.034581546687615670466, 0.015614881683514566721, -0.026274214164962360341, 0.0081727130681814984942, 0.030346715328816529361, 0.0061183690539906594036, -0.163504102404947671, -0.35350516504408424634, 0.7833870624208127964, 1.0846428883001575372, -2.7909331660902885197, 1.0860321552208671569, 0.78341458774909167317, -0.35682818312349046108, -0.16751550763292916146, 0.0088971550364174995656, 0.041370768251286951789, -0.081262450062428770448, 0.018793299074093862411, 0.024730485119734399446, 0.083882457285279599102, 0.2429290653805011857, -0.22666040307234558426, -0.18557167617603137488, 0.23726027513990180506, -0.17897586785906241524, -0.21390572709433125498, 0.25680974679581175524, 0.082579884188443530846, -0.011146133210034044753, -0.048919456369113246963, 0.17183748987969732491, -0.18283309796974511752, -0.54557588749112262416, -0.55107648089129979141, 0.62050617030936405794, 3.2887632511071966945, 0.91324398683784580033, -7.4194665366205923007, 0.90927272113787493968, 3.2731371057213238807, 0.58696145251266895748, -0.59163461591070087842, -0.52895682342253591557, 0.044178759311496514084, -0.0080133749788879136272, 0.035559650145081875272, 0.049176322463474690661, -0.089824344742018141785, -0.295852373741463337, 0.33417318855285466084, 0.78163698078624088694, -1.6137954527919005088, 0.7836948884398321491, 0.33566412751969593176, -0.29787261833303196878, -0.092466371368740613024, 0.054619870317238070068, 0.024389950628290132245, -0.0268681303676531133, -0.053482986210223272538, 0.079184928080292038555, 0.24131798037869478524, -0.090399606921852365682, -0.59030141630636046024, 0.31656346902049459224, 0.24681478710756135642, 0.3142541990632456983, -0.59298482820056663822, -0.087984229865628593714, 0.25423069946992854495, 0.090743545744785020224, -0.10237990119849338377, -0.034310541254152240875, 0.0089946200993179185185, 0.054564567288982630811, 0.0084226517202409048601, -0.1423422385369727261, 0.050231233851880344177, 0.49945359113299681031, -0.89817849580100272977, 0.49906795763516553777, 0.052630627796059162304, -0.13379913111950411664, 0.022071713782320062647, 0.060547828653119008246, -0.049925358461450519698, -0.0034680593431883621758, 0.061657933517803568957, 0.065979332151179861943, -0.02599199301151454769, -0.32741456407217356395, -0.19174071983079934989, 0.69868295565771287858, -0.55695045689641697884, 0.70041005056689120689, -0.1889504973109471242, -0.32426567862872873738, -0.02482755994426713575, 0.062396990087953504989, 0.056636448780493391097, 0.055200396874194403163, 0.073835692714673301085, 0.033698603611227048726, -0.13326069874230986323, -0.44371254528842202225, -0.14504378144903273817, 0.80493316805160386451, -0.48971468523174532805, 0.80260184204734719149, -0.14946186156606014461, -0.44782510557171278931, -0.13031971129606884086, 0.055537856116147338303, 0.11448402537818018476, 0.045694957593999101231, 0.08406502962142325952, 0.051653341112681758207, -0.11981501802070218532, -0.45834450099920753807, -0.21851358499757306264, 0.78682242872011187718, -0.3404922731730315455, 0.784930249080127318, -0.22214687678668137982, -0.46176331170683759808, -0.11755122941113421708, 0.069463762303222376104, 0.11808907798106178799, -0.030199482420804005939, 0.048360952049504302153, 0.022936263953499826107, 0.0081493848633935415254, 0.057710510942201538598, -0.34794215893039853738, -0.026926972431583678924, 0.53180650695761144675, -0.026283669491328525564, -0.34337436691345202266, 0.067486346201094310482, 0.013959005680226396068, 0.0099904261497548649262, 0.015452783021920766734, -0.020224893167928834764, 0.02698844275228637693, -0.002068599081930592911, 0.023351391195587149763, 0.11928816451713883384, -0.32284495579129751652, -0.075437351306956432206, 0.49910764140190155835, -0.075331100286213717454, -0.31995288972558605378, 0.1264297767766846825, 0.027627166776172471657, -0.012163433400560958939, 0.0057275874168962268712, 0.14957077030917878546, -0.12882366522194046987, -0.33980368026507856705, -0.036336574277354566354, 0.70561083396444068683, 0.1167077213412355402, -0.50976103425187624385, 0.097278449422482454745, -0.5137009742091688258, 0.10384691799551004632, 0.68124892263451519714, -0.063331434405407838195, -0.32461630675451075279, 0.05817734765545384823, 0.0024385438802971668056, -0.019860937943560252883, -0.0013385361474032310934, 0.10051453834284090894, -0.019625398932295090582, -0.39223482211944205034, 0.28086334038437976712, 0.089644395661911152606, 0.28042110305921613689, -0.39055172823095779533, -0.012389057522630832903, 0.11086300050504796499, 0.00051781104154393665928, -0.032694760459716760981, -0.0085869527967253680978, -0.0012417808824745513868, 0.054719103987607681239, 0.08186793956733046107, -0.1672401496777886909, -0.27275223489210509431, 0.44273656593327914699, -0.25934555615026616548, 0.44296659828102402656, -0.27249936534524515919, -0.16678565754847246572, 0.084448706171732973447, 0.059347229740489837457, -0.017006276719066848457, 0.037750537044755991256, -0.13111515642458410058, -0.13763794777032997696, -0.018060705897074237625, 0.18893616268779594991, 0.94687796409312230406, 0.5554786920279833895, -2.8819591098237142823, 0.55584721750916532557, 0.94198822320091746274, 0.1743331599182564029, -0.032908052752802972873, -0.12304522887294760958, -0.081545807983490467552, 0.027391334552985778517, -0.10695814887664152615, -0.10679897553648519393, -0.017007484333724953718, 0.12348884277950954169, 0.82028120088946065991, 0.56980202823428893755, -2.619377873588875083, 0.57057699073116874278, 0.81713094913301109745, 0.11229247956352407523, -0.029152720950844434461, -0.095606416393485632588, -0.070367761542722978563, -0.060913233542488738614, -0.0034024147974754679127, 0.076778579208028940872, 0.18370326320938829001, -0.0062505095118037898999, -0.45065938476840600169, 0.21635910728360929767, 0.088740380477175043183, 0.21760491855216285151, -0.44661390488957009026, 0.0014789086166116186677, 0.1878701728391274639, 0.058267970681995427673, -0.060693937535927056093, -0.054990226849241596874, -0.0099904459856981174537, 0.066547274470785469647, 0.18804328981533011134, 0.011482969659954005495, -0.46953430415231367512, 0.19156833547597829082, 0.15279305653950020494, 0.19264464122073626795, -0.46579291734309419271, 0.019140229439477851603, 0.19296028432396389318, 0.049830140462738196916, -0.063100451768873425107, 0.049814018059207917921, 0.06646592409647621158, -0.014758798968870825638, -0.15226013335506596857, -0.16234004913292657268, -0.087094409429170405179, 0.32065892087455150072, -0.034555330832611128034, 0.31896594928639643296, -0.091170754867833658786, -0.1694984577416622229, -0.15857106706279933128, -0.0010214196245264138465, 0.11805360242242202151, -0.1453804327094640747, 0.13828241903643248567, 0.37641475673671115354, 0.463161341224976153, 0.2520351252650328, -1.2323793831203657678, -11.230541654142042418, 22.74674290589442549, -11.228962666097144663, -1.2167050278454338841, 0.29021559595067464299, 0.497574144320249101, 0.3394309613945002968, -0.039810903852155681004, -0.063084529715627904456, -0.0071250486760969867009, 0.067314981434004439453, 0.11815083627033876046, 0.04223235021872180639, -0.23592106362318496227, -0.047462383296799734445, 0.61126890477537609847, -0.952134766968424473, 0.61031710906807457029, -0.054497268499622643356, -0.24814520840772480437, 0.035761620957709398805, 0.12924523948159569176, 0.0087266200497900566629, -0.021259379259185411104, -0.004714291156454979842, 0.0055800887914336898266, 0.031326470073049894616, 0.1078454380206783636, -0.22987463697230842152, -0.051029889432123887172, 0.30522407618356767189, -0.053904337425260559491, -0.2257951591681150727, 0.12389219140769894278, 0.023152279715135754001, -0.019029115235418600505, -0.0060330194595526603907, 0.017542571550534378361, 0.029039569937866570315, 0.0037473045942685132177, -0.16387002323772792423, -0.35363157862224564587, 0.78338706262812396464, 1.0860321549842228972, -2.787596732303343483, 1.0887898295443523278, 0.7766061195957368879, -0.37815562905894173351, -0.16931280611791005297, 0.072977767634542028974, -0.057987807421900516847, -0.01389451046119511736, 0.029035897747801773378, 0.014936546035391146808, 0.072203610459462594373, 0.2362399778339555334, -0.22666040273077692202, -0.17897586797789913349, 0.24941580229526127965, -0.16528356442986821828, -0.21001571131003562987, 0.2348084150298833106, 0.023614545166195485132, -0.0051131526556468008884, 0.1298216532998582029, 0.019699641715635624051, -0.21632628827526237392, -0.52202340555835669011, -0.52077485951844726397, 0.63482114608561035496, 3.2887632511280764369, 0.90927272124543534471, -7.4191247341307784424, 0.91579648395123636195, 3.2734383329032090693, 0.55553026048211406884, -0.63263828188338633307, -0.42502576074339648571, -0.013484570977937771971, 0.0058826018126107073453, 0.039954814382902346936, 0.047689177757808100255, -0.092617164881540189891, -0.29770424815522261719, 0.33417318818569374805, 0.78369488853609770018, -1.6109714299644450897, 0.78332775117303421464, 0.32784391119289252314, -0.30742293368450507307, -0.089102930424446835644, 0.089213420937434284586, 0.019351978990525695035, -0.052491302564038909428, -0.052882509401291563633, 0.082933889223112039479, 0.2435231525717849177, -0.089338724068141417423, -0.59030141636470145894, 0.31425419879578042615, 0.24167345689287519916, 0.3107541292292463897, -0.58296349778533085306, -0.05057874258519392402, 0.28165542155541500868, -0.075231207156265841873, 0.003536285417068222095, -0.024579868260686147524, 0.0017726941758217206867, 0.046250057006921890812, 0.0054319968148172723815, -0.14209518041873489547, 0.050231233794077227717, 0.49906795748411481028, -0.89558891520502947881, 0.50750229135292490579, 0.064366965577210680238, -0.12551659766496181536, 0.0048546403462132545897, 0.0036130829585043697239, -0.02465818322488515929, 0.029069393244396134202, 0.065890745383050314254, 0.059574927560543590455, -0.031852823149958318927, -0.33016409555939280063, -0.19174072000689054707, 0.70041005066866712792, -0.554801519224348616, 0.70040475008460412187, -0.19723583039089889746, -0.33883314056801411285, -0.017194237117618901833, 0.13162665250114219395, 0.013235553585326787635, 0.050538371455185686354, 0.068232969755422798186, 0.03027152198671665892, -0.13315318052025773565, -0.44247252136147130486, -0.14504378140490511995, 0.80260184184059890633, -0.49447722430337603949, 0.79523950802208387412, -0.15996581240466820062, -0.45073627793344078762, -0.096608842866816757589, 0.16013537846648939755, 0.0012755043931884761394, 0.052330525960766378912, 0.081597913415538195214, 0.047788567653829071291, -0.12088288259038369732, -0.45780300812654656228, -0.21851358512948174107, 0.78493024913202058546, -0.34484375331033068068, 0.77731494204073592336, -0.23439347751982148038, -0.4680401359203206213, -0.085247092993503428837, 0.18320188127500319775, -0.028724492083788334684, 0.012288179087549953167, 0.047312495776785312707, 0.013108968043747525056, 0.0024255828353856997634, 0.056320832961887473678, -0.34794215871818162666, -0.026283669380906985596, 0.53527098835094932827, -0.018957176756742903645, -0.33912129022340725681, 0.056325882114851039129, -0.018209584359051230601, 0.057051566040690566972, -0.017410731049713929625, 0.0068055062263339015025, 0.024987929965730452941, -0.0085547728135248069653, 0.020093430810373555384, 0.11872385422814213385, -0.32284495568375259911, -0.075331100386651486667, 0.50111601210733014433, -0.069827965270850222979, -0.31511912437923927932, 0.12070195378396103569, 0.00027548552513621853621, 0.0168909962645102385, 0.09358540908359050392, 0.028273555430370814201, -0.13994509677034339123, -0.31523185806086045924, -0.017044232394154135851, 0.71244211368056375999, 0.11670772153702554941, -0.51370097387843660641, 0.0871176224397155613, -0.52894803305400517335, 0.09624962120779768926, 0.69832937574407105696, -0.037207686673986398174, -0.28509837312947428067, 0.021601932402781607451, -0.0084828813152588184582, -0.027825253637215399682, -0.0066907896336406216808, 0.098499217647940784381, -0.01975276499457294388, -0.39223482214231558673, 0.28042110308643131145, 0.090225147294127394648, 0.2850797388419094025, -0.38095646482626099871, 0.00062799566154898150971, 0.11265092126177127241, -0.055672085247289282428, -0.00099813584804204938221, -0.010315102557155272911, 0.0030163968107151167483, 0.055370998853112828042, 0.080370382418801691871, -0.16825411254819200302, -0.27275223503353140941, 0.44296659812454725058, -0.26013670294718399534, 0.44022530952919758374, -0.27461936823990390488, -0.15705877429528000699, 0.10262796634629217651, 0.02010418311546980849, 0.059842151285229627788, -0.045768209619470083538, -0.13323279590432970942, -0.12106158716664107422, -0.0071634962289402713542, 0.19189957278877112223, 0.94687796367400112185, 0.55584721719932672634, -2.8822077485971488997, 0.55313548677478707738, 0.94194578945259810343, 0.19145507059879857126, -0.018667221879479289209, -0.23517670894173081653, 0.047604826754584383952, -0.038809717223026278521, -0.10822057963706922801, -0.094094413108613134122, -0.0089461874820977831113, 0.12551021471005932573, 0.82028120114490477022, 0.57057699068559775135, -2.6185274501499518074, 0.56946006224574374777, 0.81690702046571139761, 0.12491200803166076228, -0.018346305537278430331, -0.19020593445971972768, -0.036858995389280231614, -0.02568374884738246211, 0.0064155449333362854492, 0.074098161872243967219, 0.1793377659944340452, -0.0084407656900343033812, -0.45065938460808718791, 0.21760491884724753042, 0.091005067285766086393, 0.22132422979981669631, -0.44348116334264381377, 0.0018664402992498909523, 0.17695514999146083612, -0.00021063335741093168908, -0.029730117889262022007, -0.026247211442984735158, -0.0019730179546651113731, 0.064055363620980507178, 0.18413240293715088525, 0.0094972022705551058241, -0.46953430449545918712, 0.19264464102207187746, 0.1547335291429237647, 0.19633277593059883515, -0.4614252714971935343, 0.022249961065904686613, 0.18363386497886080573, -0.015659056683243095104, 0.00024547921645498766742, 0.049983185606117322419, 0.066041103707811923962, -0.014012614452191679559, -0.1507610135289139397, -0.1611145416879650738, -0.087094409582579704332, 0.3189659492039873534, -0.038548299240254373099, 0.31072634205825505616, -0.10484595064684877619, -0.17846788667982857457, -0.14277710288204428779, 0.13124460073353408363, -0.11638011482767321447, -0.0047520526749185252463, 0.16380109860335695382, 0.3534909156233699834, 0.43745412326983401652, 0.24027984280109213056, -1.2323793825824478354, -11.228962665909630658, 22.743012812550833956, -11.231265589819686923, -1.2016035444066726079, 0.31806831910190058732, 0.49811940868767096147, 0.26893979299266584926, -0.03853287075057251071, -0.030073095857395752695, 0.0037982682624139270031, 0.067686176709957787678, 0.11642769943589179049, 0.0422323502394672326, -0.23401733572438326858, -0.045590493526426474624, 0.61031710885692824764, -0.96024384581230637359, 0.5912422424280306954, -0.089943309291543288575, -0.25520652204423799336, 0.22495087863167995135, 0.0101337357370329742, -0.0078919822606651184732, -0.01913468871985473177, -0.00029421391114987157245, 0.0076371360618196892228, 0.031326469957671160271, 0.10615225410994295896, -0.23346696185088200437, -0.053904337220914803896, 0.31046454620418661996, -0.038395963891053368211, -0.22983233481238063423, 0.10960928909198645875, 0.0082913343496109400549, -0.0062030016374555324576, 0.0030113318748137326425, 0.017306177687194050641, 0.026849075803086056013, 0.0024815626608066325941, -0.16387002303776410095, -0.35350516537467829847, 0.78341458805398822562, 1.0887898295230933332, -2.7816557648707949113, 1.089098406347683623, 0.75280438349374334539, -0.42803978540860843083, -0.031722562759996159043, -0.012591612477772906598, -0.034987983085552391838, -0.017595153666937760756, 0.019687716647780913182, 0.0089943523502241662371, 0.072203610803732956591, 0.24292906462229399422, -0.21390572672792421138, -0.16528356454047105073, 0.25063551441312942325, -0.19017872103076446932, -0.25452967415386446026, 0.25593935616223761897, 0.043052454476822231422, 0.05462754335766902003, 0.070199464744972323027, 0.0030690043848310985877, -0.20585910675744814591, -0.50913716481327697672, -0.52077485958551716916, 0.62050617003355756918, 3.2731371057194373897, 0.91579648403905888898, -7.3753434484357383027, 0.99197162655594961489, 3.3466272299121446743, 0.46966015855367565512, -1.1329583073303461926, -0.010392999354236934911, -0.0019822348030470943528, 0.0091787584493833243582, 0.037478076352730337051, 0.045148805446834174016, -0.092617165218215169453, -0.29585237332120606357, 0.33566412739382789354, 0.78332775129359066835, -1.6123298772359189179, 0.78240965497914427562, 0.30718266299032309652, -0.3488425491402292189, 0.060742975662793373404, 0.02120822794767887598, -0.0097780963816303887587, -0.051797704531517552162, -0.04719872461204025621, 0.085410204948100892919, 0.24352315239151836601, -0.090399606887334227334, -0.59298482785352935043, 0.31075412900282251005, 0.24470524674684840383, 0.33093706214728002557, -0.56284693866082879943, -0.069313735451175709223, 0.18992805651423763647, 0.022375176483870152372, -0.014944950088477260489, -0.031980601047909736068, -0.00010916894737161993677, 0.046565072737330737151, 0.0054319966465316109339, -0.14234223829819911411, 0.052630627881157049375, 0.50750229128138402146, -0.88165329110499657261, 0.51835394060687467466, 0.038649048998527446575, -0.1857119091924996912, 0.065782837828457257912, -0.018329406623774038843, -0.00075993835053392907314, 0.031137706396230958872, 0.06130006644348557493, 0.055952718080361577013, -0.031852823309432079513, -0.32741456376165251063, -0.18895049738377012183, 0.70040475015233905065, -0.56258474424257165492, 0.67976688613765001712, -0.22590326606577804114, -0.31890153333766008137, 0.14352155470858488084, -0.0086862299592455130548, 0.024581184322121883701, 0.053020027968597331636, 0.067503834136422663503, 0.029727394118554642533, -0.13315318043515095292, -0.44371254546562577392, -0.14946186144415007924, 0.7952395079139789047, -0.50303596759657231985, 0.78771268041831321582, -0.17161019596778046004, -0.43653925876369753523, 0.081441586723220188437, -0.016338981756987783905, 0.020113568398204897947, 0.056602909500621571837, 0.080418616912738521973, 0.046580809436811310142, -0.12088288289129116293, -0.45834450040886259847, -0.22214687722690440452, 0.77731494234201048421, -0.35612954968420623691, 0.76379914672655579988, -0.25134108562871676495, -0.44617328180310877883, 0.11987571602160948525, -0.0064204228945473730897, -0.015509118431159929496, 0.0085656270167864551751, 0.042771490073496461204, 0.011519276242829404322, 0.0024255830316325067628, 0.057710510736492623418, -0.34337436715413360844, -0.018957176722012261672, 0.53637069013863813982, -0.037241676802080757314, -0.3754066738770508449, 0.084569665308133548653, 0.052974095672748883201, -0.0019565884604278195591, -0.010994674995608209878, 0.003420923446490985436, 0.022338849795345554206, -0.0091179792308011507751, 0.02009343089782013661, 0.11928816433690989096, -0.31995288958443834915, -0.069827965317326573524, 0.5028766828165964764, -0.081377093878463716004, -0.33892380242180231154, 0.15109212653998704834, 0.013360608101171232362, 0.017594973432573331185, 0.066064690328736838665, 0.028757172335053104301, -0.12967640162823265704, -0.30822264103852303752, -0.017044232217620410713, 0.70561083362673049368, 0.10384691736462568767, -0.52894803271063706163, 0.081404226210888538939, -0.49729858157668449525, 0.19407088996825852534, 0.81623519438124803038, -0.5341558246149986644, 0.024249902367085154203, 0.0028584564393435668399, -0.016993072354190005141, -0.02962534116131828113, -0.006546335942080518111, 0.098499217348460760268, -0.01962539851096655949, -0.3905517285925950155, 0.28507973895395477548, 0.096554224212430164731, 0.28983034344706781749, -0.38064136765078565583, 0.014335429649982089212, 0.031717987368181370378, -0.001177786768910191054, -0.0015043862152896805319, -0.0078479021013283228186, 0.0033094408751039722238, 0.054593003369237023958, 0.080370382197800141477, -0.16724014915642926304, -0.27249936540646635397, 0.4402253096331351645, -0.26429522545003508727, 0.44179278622046347369, -0.27334583792391631274, -0.15868030347569536676, 0.12622959319214041574, 0.030005203045790126287, 0.020894822344869244668, -0.047356885336445185553, -0.12680010733807070555, -0.11721852993579179314, -0.0071634963588887365371, 0.18893616302227722592, 0.94198822399669490668, 0.55313548651957722413, -2.8699064814706853888, 0.6012168420714468775, 0.99824189573437138723, 0.12438402099740335227, -0.28677793577138616099, 0.025154596248330506753, 0.016110141443975125458, -0.040757275848728492795, -0.10376156800459436502, -0.091373456845180556218, -0.0089461870887849320733, 0.12348884198895425379, 0.81713094941282560235, 0.56946006186937170312, -2.6077689854258596647, 0.60875627409871213569, 0.86024479384255958525, 0.063551496350685646175, -0.22824066796516007116, -0.010670839060013948343, -0.027195587113481526492, -0.023739917383544108936, 0.005791706330724375601, 0.072952128153894887919, 0.17933776612422408503, -0.0062505098742285502417, -0.44661390514468646273, 0.22132422991776123933, 0.089045633846071670092, 0.20912872985843525275, -0.46101414794419492971, 0.020985073340512028217, 0.18063522678844934344, -0.0061012206347621986136, -0.024344800971167059739, -0.025399190529479795386, -0.0024515960425338535061, 0.063133470715003997542, 0.18413240299437608205, 0.011482969838293544812, -0.46579291669392297592, 0.19633277561732115357, 0.1537631709603689556, 0.18668430764741641981, -0.47505399981419338395, 0.042692625783925207406, 0.16438745861188425179, -0.021032722964311655056, 0.021035013708980483671, 0.055939585179313348895, 0.066110440092077463192, -0.014263427153531391123, -0.15076101366254218705, -0.16234004882409994175, -0.091170754716280083541, 0.31072634203135734987, -0.050798472350145749965, 0.30066661244975406708, -0.11176467077150652185, -0.14484953016118717883, -0.012450342081970077954, -0.046012278336697895464, -0.062240350329677979202, 0.0036698893438329911222, 0.15496784269943386647, 0.34418139836217348515, 0.43745412365352165374, 0.25203512477079531573, -1.2167050284476046418, -11.231265589565486707, 22.692234418389986672, -11.316496523126339468, -1.2109320733293233996, 0.45704215220730537439, 0.74195638814413744644, -0.012616451343326029813, -0.028233476468133473364, -0.025737770968398111793, 0.0045392154113567577747, 0.067686177780083137123, 0.11815083729337275131, 0.045817086717849335453, -0.23278320191848891896, -0.054497268891746622477, 0.59124224206711095686, -0.98571423134199920302, 0.5737396542045500869, -0.085262310518052317199, 0.020950540696189706713, 0.0017215658333680791767, 0.0056778199558303044103, -0.0055089290665945277181, -0.017102014401787626191, -0.0002942144108700389454, 0.005580089332904983844, 0.027669219504277087207, 0.10456287234552134879, -0.22579515947733200254, -0.038395963893213869156, 0.29877558180825586343, -0.071479740204503691703, -0.20137948777523120847, 0.1166761761961448568, -0.0021620872525358780965, -0.0031624964378127217238, 0.0030733345196991782716, 0.0162806773080036854, 0.026849075735717681246, 0.003747304423159123106, -0.16350410211308949449, -0.35682818355240958352, 0.77660611983526806057, 1.0890984061793957949, -2.7627379566509762121, 1.1159525457289980466, 0.70588113264501239907, -0.44973037564184453752, 0.018584743838149859335, -0.025821149897902483106, -0.044446903668177197011, -0.022725955878263530419, 0.019687716952907462209, 0.014936545714823211764, 0.083882457325616444077, 0.25680974701666536841, -0.21001571097759963447, -0.19017872172720956536, 0.1929244975749782598, -0.19557754168471225276, -0.14477725472473879087, 0.25103398077941119215, 0.00063552261268040111455, 0.05552342479009744064, 0.073049794554536223723, 0.0070443032594581025951, -0.20585910699520004741, -0.52202340522550927382, -0.55107648100917328016, 0.58696145277595368484, 3.2734383327024327848, 0.99197162634727331643, -7.2085752285674402629, 1.2341086543698598366, 3.3470839727940355779, -1.0603454203431530445, -0.0038354326052490268426, -0.0050638908962018554932, -0.0023370967747443375725, 0.0075115148393161811774, 0.037478076089189814313, 0.047689178213701832743, -0.089824345160299537416, -0.29787261807633558242, 0.32784391089158099186, 0.78240965517481275704, -1.6075343076720485236, 0.77402123209905671075, 0.28341598938229856097, -0.25592719472721220386, 0.0081475395313219331483, 0.010851757968741684163, -0.0099799479701155949241, -0.049711558256120551291, -0.047198724808937345121, 0.082933889278679534529, 0.24131798063393863907, -0.087984230173056815483, -0.58296349749313824962, 0.33093706195885952415, 0.2554259654386088485, 0.26443859043676021248, -0.74197803734956357946, 0.32696469809945993834, 0.020353060005462903642, 0.0015955557332053842308, -0.020359151458271454449, -0.031713314958218545569, -0.00010916910242191588988, 0.046250056862015762049, 0.008422651516946417713, -0.13379913115700872717, 0.064366965818212482087, 0.51835394073511942281, -0.89547865605801124644, 0.46607202380299128652, -0.0038466871103798499013, -0.038567134060570454346, -0.0055212476815030024288, -0.011881710147639971617, -0.0006662753517469036579, 0.029258858010225246432, 0.061300066324400701023, 0.059574927721712751805, -0.025991993314064827658, -0.32426567874330192165, -0.19723583034830974303, 0.6797668862631485176, -0.58048124379603105449, 0.68674959029631188479, -0.24606865182303033768, -0.13024687884647490055, -0.015157898567951359065, 0.0047407795820755861682, 0.029659153127273199524, 0.053898350342882539454, 0.067503834215744754066, 0.030271521232162219156, -0.13326069791134800835, -0.44782510624729776616, -0.15996581201479348477, 0.78771268075017797372, -0.49790803670641209377, 0.79970154492774969679, -0.20142897764966716245, -0.3274354769625159034, -0.017468219899991858923, -0.00065247604397000290993, 0.025720740989651665054, 0.057296726878840263664, 0.080418616699750519738, 0.047788567792152250591, -0.11981501835054246541, -0.46176331142884646797, -0.23439347778171734649, 0.76379914716415964193, -0.35861454068314690957, 0.77484558201689257562, -0.26943418812299063658, -0.29795012492977529517, 0.0079505886138328055179, -0.014823918135670507978, -0.018188817672638274425, 0.0080841714660493564637, 0.042771489902779152859, 0.013108967958440364415, 0.0081493852154773168545, 0.067486346188853726935, -0.33912129006732416592, -0.037241676982588713118, 0.48994603234064393327, -0.041132038278748110638, -0.28443618938050091449, 0.095990174862747773554, 0.0067461144908202022932, -0.0093144063107232361043, -0.012863343623169116309, 0.0033986682445078957894, 0.02233884974172690302, -0.0085547729675727237492, 0.023351391529943626296, 0.12642977638394511941, -0.31511912419147430908, -0.081377093853946966262, 0.46798671340747771685, -0.079112603820407967348, -0.26643382106429991385, 0.12224637575476531015, -0.023948399348043568347, 0.041715098391247540377, 0.074305480205055590703, 0.031161712192372702235, -0.12967640221712034565, -0.31523185791019819879, -0.036336573879165098844, 0.6812489231508137566, 0.096249620978277924954, -0.49729858181429070774, 0.1658283236869040056, -0.34499893436932294044, 0.19990842652036003302, 0.062915021588515945838, 0.01591522359974579387, 0.009100939516766484369, -0.0021527558396511898425, -0.017565220653654279453, -0.029625341538214763903, -0.0066907896037287552943, 0.10051453807365223758, -0.012389057254955111037, -0.38095646493396667642, 0.28983034386537831484, 0.092107783278979296626, 0.27804482440210920302, -0.46575194397141567171, 0.1303053465235317343, -0.0015209239959043509369, 0.00050369196271348294722, -0.001588208736466002139, -0.0081593341914506581358, 0.0033094408544311476605, 0.055370999108049580306, 0.081867938929496578138, -0.16678565724857258301, -0.27461936860925706405, 0.44179278649791758893, -0.2608501850872858574, 0.41999287509859367651, -0.35717141839149846305, 0.066121276742521536973, 0.0031030475542535868733, 0.029818293012041814649, 0.019441741931273875954, -0.047394672102183621476, -0.12680010633832811129, -0.12106158752219532693, -0.018060706294161524288, 0.17433315915218897119, 0.94194578936920192458, 0.60121684253637097672, -2.7824773922529035808, 0.63333998354551890309, 0.85213306959259660367, -0.14875887583204266962, 0.0042617005351434185043, 0.023784174143877498397, 0.013950263269053610193, -0.041216287465783367494, -0.10376156778521215818, -0.094094413708333962676, -0.017007483354877506176, 0.11229247951887515178, 0.81690702111584811984, 0.60875627333811388464, -2.5355033825046464635, 0.63337157004853872344, 0.73061596334457001856, -0.14331348190152823352, 0.0053960007659692565268, -0.016432725095139578964, -0.028746826968007690994, -0.023927402979830144497, 0.0057917062799398965484, 0.074098162025763220084, 0.18370326329764014028, 0.0014789093784529725589, -0.44348116358980360641, 0.20912872961465850397, 0.06188600105375551691, 0.18572121230485527654, -0.50735829777188912182, 0.29379998342641688858, 0.006843423675625509052, -0.013202438969844635525, -0.026312623633358421527, -0.025624637370664150871, -0.0024515953752477209066, 0.064055363290744904425, 0.1880432896575343904, 0.019140228724587687176, -0.46142527145637091124, 0.18668430797361720841, 0.12954491415692287459, 0.16413741462767347801, -0.52972827163242586224, 0.30161852113608850212, -0.021624312092913544614, -0.0010155892884602111311, 0.027687158372694253838, 0.056950599563672994752, 0.066110440141085316523, -0.01401261452597061738, -0.15226013321075262952, -0.16949845828643905765, -0.10484595073112012997, 0.30066661289169466231, -0.051428015521405599786, 0.3326546951687917586, -0.037811154898030191474, -0.23891237655287833608, 0.0042999063210808615917, -0.053199273256547940925, -0.064836606244910388597, 0.0014592361153032806353, 0.15496784210975844687, 0.35349091547637601085, 0.46316134186575275145, 0.29021559639048799539, -1.2016035443272676808, -11.316496523383447581, 22.455342777602222526, -11.485022670750867135, -0.82625368472437998868, 1.2080138737591883391, 0.0046744775945562980907, -0.014648000264912161214, -0.028040770525249170131, -0.025737771019333798017, 0.0037982674689858468217, 0.067314979912779362992, 0.11883442739471729299, 0.042977258594380805812, -0.24814520784860433222, -0.089943309406374197246, 0.57373965465477516634, -0.93803668663194195165, 0.80007948162738828923, -0.27610330680128003511, -0.003358304810179953831, 0.005902343844961701301, 0.0070500334824691174296, -0.0055089292359608143385, -0.01913468817911136155, -0.0047142915387198992133, 0.0027495092783401875111, 0.034581546152200805244, 0.12389219184272634577, -0.22983233489022264484, -0.071479740125392834948, 0.31418327854676880717, -0.17480700362678328053, 0.019948425116885975389, 0.00062424088566781564362, -0.0029028203833847757154, -0.003608368007053650451, 0.0030733348083046624008, 0.017306177476509698732, 0.029039570366618541741, 0.0061183686429466853721, -0.16751550730898820407, -0.37815562909426286842, 0.75280438333539645424, 1.1159525460536607877, -2.6986772154951332858, 1.4854743172937232565, -0.15585865820628061473, 0.025497881447222751905, -0.0054918721256388336113, -0.034302604181500787051, -0.044446903061887588215, -0.017595154357197759221, 0.029035897628727262709, 0.024730486271335030274, 0.082579883086157435113, 0.234808414839837859, -0.25452967295684969518, -0.1955775425764118558, 0.27142036233223082453, -0.33110457776613522185, 0.21413570020875108413, -0.021125919121872213657, 0.026222425744671304382, 0.064117781963057354777, 0.073049794493993333577, 0.0030690047044231792521, -0.21632628869194295285, -0.54557588708978821401, -0.59163461632450464744, 0.55553026071671696151, 3.346627230153169652, 1.2341086542109809265, -7.060171098224024, 1.2876168979314599472, 1.8839922566007867299, 0.0017020587601814065378, -0.0051513546979494546571, -0.0065398482038002128322, -0.0023370967933266126142, 0.0091787586304447488922, 0.039954814089099295604, 0.049176322862430114913, -0.092466371768242558438, -0.30742293333610115624, 0.30718266265394489345, 0.77402123229135366689, -1.5572930693709134964, 1.0426610569073468948, -0.25295042977916887272, -0.0012630807938579422488, 0.010199975725801394694, 0.011185506331163429586, -0.0099799479800159584392, -0.051797704511428475926, -0.052882509198129923345, 0.079184927704831167827, 0.25423069965697697725, -0.050578742823780269144, -0.56284693841055410957, 0.26443859028753052831, 0.080646457969819576594, 0.27329512597053362555, -0.24648849913579282123, 0.0086510645586226914833, 0.009330168020451817118, -0.00072300685438286458294, -0.020359151410131431231, -0.031980600996352394338, 0.0017726943773076818487, 0.054564567516925566171, 0.022071713676269356236, -0.12551659789182637028, 0.038649048984177189481, 0.46607202374486628171, -0.89922268619369605247, 0.61688972801109165367, -0.13778326989682335157, 0.0020325767932196374572, -0.0091542117372518348739, -0.013200556773417616852, -0.00066627523758626514168, 0.031137706477802635369, 0.065890745274174114532, 0.065979332472852508795, -0.024827559862201243474, -0.33883314066245628915, -0.22590326619640252459, 0.68674959031065974102, -0.54131382654345661098, 0.81541408095086254093, -0.51787977142359686855, -0.013195784504126766706, -0.0055476189597997715985, 0.008416165091564127565, 0.029659153214396029308, 0.053020027804786117043, 0.068232970558531794159, 0.03369860295982839149, -0.13031971059652547318, -0.45073627834366314371, -0.17161019619768647093, 0.7997015447626175666, -0.44892908653960544862, 0.92835867855010278937, -0.70390189411602810399, -0.012015386887665152615, -0.0086489047373192046164, 0.0028061529900778297425, 0.025720740887674665293, 0.056602909930935592364, 0.081597913208815167629, 0.051653341489530740782, -0.11755122976756450515, -0.46804013574910940942, -0.25134108598448540173, 0.77484558189527164007, -0.3015014174152721993, 0.90769685585545623052, -0.74686885671459457559, 0.0087118936025967017717, -0.0031760076298265613529, -0.016567865357983210112, -0.018188817980711174949, 0.0085656273696775534626, 0.04731249582571191159, 0.022936263340617903811, 0.013959006001407765762, 0.056325881829539688195, -0.37540667368868840814, -0.041132038242826240992, 0.57012313229602618048, -0.16437431632262097803, -0.112149451973008768, 0.0059859018496398819567, -0.00091164586087633743723, -0.010309051008478295822, -0.012863343790136210973, 0.0034209235155784182593, 0.024987930248414828849, -0.0020685996654225976242, 0.027627167368344286713, 0.12070195362075623879, -0.33892380269681021687, -0.079112603416482324414, 0.52524447882662017584, -0.24330284307146274325, -0.022250043693547438828, -0.028811758214769181419, 0.0078177029238342184841, 0.049769638972365130847, 0.074305479691809089848, 0.028757173283397000768, -0.13994509711685371167, -0.33980368061822191494, -0.063331434619715590295, 0.6983293756886896908, 0.1940708903482195602, -0.34499893462388458998, 0.13514094447534269472, -1.0280866602781215668, 0.76780424215535991728, 0.0038933864978084537292, 0.0092162511723399651725, 0.0072535146805819145782, -0.0021527557523259030839, -0.016993072021557906193, -0.027825253730029874538, -0.0013385357217427775942, 0.11086300009162569635, 0.00062799577557370173109, -0.38064136797198977469, 0.27804482425448151517, 0.020680225787668350679, 0.17627151403785168426, -0.17590378639571441211, -0.00047545385426194105519, -0.00097140815449294143705, -4.4807957543919975287e-06, -0.0015882085691884069347, -0.0078479022584280605512, 0.0030163966931351939536, 0.054719104498898545053, 0.084448705772742035935, -0.15705877364715414846, -0.27334583855260613516, 0.41999287546453373254, -0.31550407085005532926, 0.48349167907150641943, -0.29178578622055928449, -0.0058992482905266194204, 0.013775316678348338562, 0.030249284288664565756, 0.019441742900608304651, -0.047356886930418867765, -0.13323279518890965289, -0.13763794735392090307, -0.032908052277885804782, 0.19145507093087749118, 0.99824189495833093311, 0.63333998410582559302, -2.8714015928565235214, 0.71886299726490021644, 0.63653425316123568578, -0.0032610287557140391068, 0.011294393813005404112, 0.023494635709271946555, 0.013950263011334861074, -0.040757275856693919724, -0.10822057927432018887, -0.10679897625931115968, -0.029152721166575464723, 0.12491200714486054568, 0.86024479535780729034, 0.63337156896620516733, -2.6082024710428797221, 0.73760296548165138475, 0.50290835678360990002, 0.010177780344126636028, -0.0032978033337876821486, -0.018665939422251937724, -0.028746827550799866624, -0.023739916863308216377, 0.0064155445812376331316, 0.0767785790985301142, 0.18787017210792147925, 0.0018664402834844282322, -0.46101414726372730302, 0.18572121190478649289, 0.013479742188643019027, 0.17104867197462478301, -0.12328833233705718819, 0.0095428484021182011926, -0.0015235003719528037412, -0.015458945148139729628, -0.026312623261792386642, -0.025399191502786362673, -0.0019730172986832435894, 0.066547274298906958223, 0.19296028515840413231, 0.022249961245132642007, -0.47505400055021762551, 0.16413741540259510487, 0.070692062022824880829, 0.12537488893557557912, -0.11039729782480525722, -0.013076591165362338662, -0.010089751216278395468, 0.0031926208920053373956, 0.02768715852494701285, 0.055939585070785481646, 0.066041104115446169764, -0.014758799687024351868, -0.15857106610833782145, -0.17846788659668116894, -0.11176467165895763445, 0.33265469582634948909, 0.058724507496491308578, 0.29879794899372830175, -0.35988107300075922179, 0.020461259879813002704, -0.021372812471511751231, -0.059621827149852424377, -0.064836606801855059534, 0.003669890242304940306, 0.16380109847491780561, 0.37641475607395119551, 0.49757414424163581845, 0.31806831867468299269, -1.2109320727659207417, -11.485022671028287888, 21.928213359823860884, -11.311665238410453682, 0.80982182566371030408, 0.012899427653877585526, 0.0004414334290104305605, -0.014648002028816572637, -0.028233476082739107638, -0.030073095223537692905, -0.0071250471047668876776, 0.050578013232053915804, 0.10470867954049248305, 0.035761620849499284325, -0.25520652204648014427, -0.085262310778028677083, 0.80007948166978148929, -0.7810989026962662507, 0.18325389407644587791, -0.0026238443159749282561, 0.00094551082362580079945, 0.0059023440496206430045, 0.0056778198056795332535, -0.0078919823067220527812, -0.021259379327763009565, 0.00023007531271071192088, 0.015614882097985229373, 0.023152279352397022588, 0.10960928935048443178, -0.2013794880933052478, -0.17480700341263116782, 0.46562571861708568211, -0.22042401684267104445, 0.001282707943433755932, -0.00099547167072865853747, -0.0029028200356027628916, -0.0031624970389797744941, 0.0030113323927289901982, 0.017542571017419079926, 0.030346715687017397634, 0.0088971549945003301774, -0.16931280641819557164, -0.42803978486315552576, 0.70588113195196477978, 1.4854743177130818044, -2.9898894971879865778, 1.3476462017114321235, 0.014819875880523854933, 0.0092596681692645142758, -0.0054918717970089507563, -0.025821150660900665103, -0.03498798244285884762, -0.013894510156678805887, 0.018793298049140382683, -0.011146132168656956024, 0.023614544981961173298, 0.25593935533698153328, -0.1447772539568799588, -0.33110457795536823999, 0.44777332351954129974, -0.21376761966672830351, -0.016709754016065170301, 0.00019676555467477931018, 0.026222426026410013133, 0.05552342480615851672, 0.070199464534608124744, 0.019699641735099308187, -0.18283309807106762346, -0.52895682322230030969, -0.63263828197925076058, 0.46966015862222787458, 3.3470839726967396288, 1.2876168979109758883, -8.5483991497767792822, 4.6459224058688315395, 0.0033067043640495623509, -0.0017546126741696740914, -0.0051513545802184335445, -0.0050638909074304317218, -0.001982234820170798717, 0.0058826018197253001143, 0.035559649951666098822, 0.054619870650522414635, -0.08910293067882625595, -0.34884254884336857172, 0.28341598919683419489, 1.0426610569571201914, -1.6567878142685048548, 0.68614319996841854898, -0.0038215860807647647479, 0.0028142779843184596508, 0.010199975828588701138, 0.010851757905408047217, -0.009778096262636348443, -0.052491302848570897677, -0.053482985930365441185, 0.090743545728719787746, 0.2816554215630966973, -0.069313735566291806434, -0.74197803719123134325, 0.27329512595699845257, 0.72620661635910199738, -0.46788134140589499044, -0.0032279168936568910543, 0.0067045435714304119773, 0.0093301682315786119842, 0.0015955555359216499599, -0.014944950177288314486, -0.024579868272449418437, 0.0089946202465560586015, 0.060547828687421520466, 0.0048546403588632372136, -0.18571190934850717591, -0.0038466868913135734975, 0.61688972799885710696, -0.7812283658969100264, 0.30802196214908422744, 0.0029596759329635069502, -0.0028403357577599188701, -0.0091542114151187212873, -0.011881710183029284625, -0.00075993854528700905543, 0.02906939314540969363, 0.061657933557068417274, 0.062396990065070483122, -0.01719423702485303046, -0.31890153320252739855, -0.2460686518416420332, 0.81541408082105626409, -0.38763713293271245819, 0.02785890421067133485, -0.007619575362668004094, -0.0086885731652536889874, -0.0055476189962448910903, 0.0047407797368439995311, 0.024581184116477351048, 0.050538371068639127703, 0.073835693152385278859, 0.055537855588167268917, -0.096608842465366134844, -0.43653925878397054072, -0.20142897740605303358, 0.92835867843284591849, -0.42863887780666454219, 0.063908057141798244039, -0.0051262857226103582253, -0.008804961793099577369, -0.0086489047645892640909, -0.00065247570999919877327, 0.020113567784349761086, 0.052330526187650508452, 0.084065029421258610776, 0.069463762437183343423, -0.085247092940507751546, -0.44617328166739206319, -0.2694341879724989619, 0.90769685571075409136, -0.25309919805473063414, -0.042220898959622846225, 0.0025946051640265174569, 0.0036850783011007364094, -0.0031760076880308222083, -0.014823917963810507789, -0.015509118581897365918, 0.012288179010103790836, 0.048360952450095265331, 0.0099904259775706692054, -0.018209584274909883933, 0.084569665238777819161, -0.28443618936874859315, -0.16437431628667528716, 0.71180159110671292755, -0.37459243274345954688, 0.0012485708314736777315, 0.0031333524507074447286, -0.00091164586015259475686, -0.0093144062627207938054, -0.010994674952015354857, 0.0068055060095078730165, 0.026988443120326485736, -0.012163433701955050914, 0.00027548548140951306575, 0.15109212684852460495, -0.26643382145053962073, -0.24330284287604853066, 0.68658250746024085398, -0.33504605483510790931, -0.01327916552817072493, -0.011522585504002981829, 0.0078177029116138462006, 0.041715098848283442556, 0.066064689706657950263, 0.02827355547424365359, -0.12882366491354999005, -0.32461630683242814754, -0.037207686439724975414, 0.81623519433134061885, 0.19990842646873771549, -1.0280866603210074839, 0.13991990829629433191, 0.24884824483653153471, -0.0062182029333993447257, 0.0032084149334364563122, 0.009216251354763255188, 0.0091009394326337627462, 0.0028584561136930312225, -0.008482881193207121584, -0.019860938170285752369, 0.00051781131045605449312, 0.11265092118115131725, 0.014335429783836679818, -0.46575194372179623548, 0.17627151394696244369, 0.4152918980143164851, -0.23670429548831661837, 0.00028659673728200385934, -0.0011001135392413000525, -0.00097140802251212637983, 0.00050369171298575063887, -0.0015043859895781839468, -0.010315102563925053827, -0.0012417811491846697872, 0.059347230059365249644, 0.10262796574640647984, -0.15868030290080981093, -0.35717141880538538778, 0.48349167926966657438, -0.0067081839439270792361, -0.10838509077926367086, -0.0038474667384064895059, 0.00043179274386746010345, 0.013775317042620003477, 0.029818292363817295448, 0.020894823268907434383, -0.045768210176398198274, -0.13111515663367578899, -0.12304522909806332731, -0.018667221967094294738, 0.12438402140812408747, 0.85213306913447361968, 0.71886299738947134763, -3.0045206829074380117, 1.5658465290851018015, -0.0027233603136523412691, 0.00085334938836709433446, 0.011294393604132983017, 0.023784174633299487445, 0.016110141155337237562, -0.038809717285096245742, -0.10695814856109056579, -0.095606416235814131732, -0.018346304754354073041, 0.063551495055250414867, 0.73061596455076838197, 0.73760296484869192213, -2.7112554979771097408, 1.3894628049211648069, 0.0082008851850771738445, 0.0050733405099501758043, -0.0032978036728465477678, -0.016432724434436607464, -0.027195587631522900179, -0.025683748406226523942, -0.0034024148519002186432, 0.058267971063390391895, 0.17695515016961843591, 0.020985072532088450137, -0.50735829714359292897, 0.17104867169449344821, 0.47843484789573520466, -0.34666727142100661752, 0.0066655704103493491203, 0.0049986493308727909454, -0.0015235003209239585294, -0.013202439304731770908, -0.024344800270904673467, -0.026247211864134652182, -0.0099904458153024199019, 0.049830139964845851508, 0.18363386482334731475, 0.042692626186437994396, -0.52972827225859497879, 0.12537488923069312063, 0.55162045367397516404, -0.3695743632101692322, -0.0028883715862975042898, -0.0089692050524316207366, -0.01008975105986643396, -0.0010155893254407183292, 0.021035013724950962083, 0.049983185155715474657, 0.066465924762406411452, -0.0010214204211683086589, -0.14277710281670100145, -0.14484952941859560727, -0.037811155641030935737, 0.29879794925714775466, -0.054835250954389631872, -0.023843148636145700126, 0.013536712254321243976, 0.0028037559710767118865, -0.021372812593694001659, -0.053199273033610758321, -0.062240350697505046618, -0.0047520524800752760725, 0.13828241935150567543, 0.33943096141893064344, 0.49811940885048150429, 0.45704215176423734768, -0.82625368442934465918, -11.311665238375889331, 19.400749622109053405, -8.5987094889692912147, 0.014842931000649581263, 0.012899427115347264494, 0.0046744784318699010919, -0.012616451104921360543, -0.038532871256366472978, -0.06308453056381395363, -0.062619933814583556009, -0.0015533732255180842784, 0.12924523951539043654, 0.22495087890979875711, 0.020950540336115965057, -0.27610330640706554073, 0.18325389383904852614, -0.1565512262554416556, 0.0013888926117047581052, -0.0026238442691293765129, -0.0033583049630762332356, 0.0017215660001919173405, 0.010133735634115216551, 0.0087266201412918194269, -0.011727564244472153676, -0.026274214305596219066, -0.019029115120815006529, 0.0082913342035757127413, 0.11667617638098390365, 0.019948424990801936346, -0.22042401680844669887, 0.1166205099802532924, 0.00027689921564182334683, 0.0012827079882609055198, 0.00062424070010591928281, -0.0021620869622355066812, -0.0062030018919310937406, -0.0060330192595473832937, 0.0081727129665888909038, 0.041370768211267304471, 0.072977767805254570299, -0.031722563034206931187, -0.44973037530205761669, -0.15585865844728191698, 1.3476462017800665549, -0.83082883383460337523, -0.0044550665496640136712, 0.01481987592347258284, 0.025497881295079474762, 0.018584744130507660498, -0.01259161268908940215, -0.057987807548529272939, -0.081262449800981767511, -0.048919456709645481318, -0.0051131524891126897345, 0.043052454697510067483, 0.25103398056850656506, 0.21413570023070388482, -0.21376761965157625722, -0.16423900878309644069, 0.0028885342962455418797, -0.0167097538852233711, -0.021125919347646973656, 0.00063552263407833185088, 0.054627543406842254958, 0.12982165340317972091, 0.17183748984918942315, 0.044178759259628552769, -0.42502576072777459304, -1.1329583074298346101, -1.0603454202481135127, 1.8839922566321951614, 4.6459224058256669565, -4.3747797787611730769, 0.0012275759308309700923, 0.0033067044240816336788, 0.0017020586767536466117, -0.0038354326059151615247, -0.010392999367256851653, -0.013484570927564313095, -0.0080133749498424548807, 0.024389950510815942147, 0.089213421034035067514, 0.060742975558911241341, -0.2559271946800216746, -0.25295042979303072883, 0.6861431999793593528, -0.32383196851707607422, -0.0019693874853552438033, -0.0038215860411847271498, -0.0012630808578697211991, 0.0081475395494083596021, 0.021208227902673033177, 0.019351979119992753636, -0.026868130480040872121, -0.10237990121909121277, -0.075231207093408636322, 0.18992805651409344625, 0.32696469804740357912, -0.24648849915572343394, -0.46788134136986320177, 0.3645394879635672214, -0.011469107390790463444, -0.0032279168874166132286, 0.0086510645192157120675, 0.020353060018954059751, 0.02237517659374151785, 0.0035362854208228333003, -0.034310541517579395965, -0.049925358360251609335, 0.0036130829798829470177, 0.065782837881143724212, -0.038567134093694006858, -0.13778326999319207014, 0.30802196222154648586, -0.15656945714653527912, -0.00055058238563469798395, 0.0029596760461696007549, 0.0020325766410044603626, -0.0055212477749496079171, -0.018329406470415806479, -0.024658183102631400091, -0.0034680595208001221107, 0.056636448833995656216, 0.1316266524434483165, 0.14352155461639817813, -0.13024687878161150811, -0.51787977137428908847, 0.027858904179183970051, 0.35480830963199094796, -0.00093326032813890980825, -0.0076195754668262729095, -0.013195784283419125307, -0.015157898886016518344, -0.0086862296898555166336, 0.013235553748386977985, 0.055200396527662669299, 0.11448402575621434429, 0.16013537820012410662, 0.081441586688944800354, -0.32743547685819657245, -0.70390189426839877562, 0.063908057255144978814, 0.63160339964763156928, 0.0013361990413140376315, -0.0051262857824557854503, -0.012015386710131642153, -0.017468220237826487251, -0.016338981429226104997, 0.0012755043896784514033, 0.045694957458855697818, 0.11808907816339102304, 0.18320188112653312817, 0.11987571593526852309, -0.29795012481519828063, -0.74686885680989012481, -0.04222089888677715841, 0.71093238952698711852, -0.0046192778301532583934, 0.0025946051761618540293, 0.0087118935754462358684, 0.0079505886331143020562, -0.0064204229044403874488, -0.028724492058444815734, -0.030199482483346258566, 0.015452783007791667802, 0.057051566079478505111, 0.052974095689152948807, 0.095990174794451613161, -0.11214945194964393016, -0.37459243273761488879, 0.32160915426522440619, -0.0034815099727127413361, 0.0012485708603369415071, 0.0059859017905556505257, 0.0067461145452188059704, -0.0019565885163299098076, -0.01741073100248075739, -0.020224893206905326937, 0.0057275874301715061845, 0.016890996327868876281, 0.013360608015528949166, 0.12224637582625599852, -0.022250043738033822155, -0.33504605482445742881, 0.23068593933515879613, 0.0089744242360369833045, -0.013279165489002041009, -0.02881175822779553572, -0.023948399525762595341, 0.017594973610421373034, 0.093585409166297733874, 0.1495707701652088939, 0.058177347755938677021, -0.28509837326597875506, -0.53415582469472011606, 0.062915021730430786828, 0.76780424215310172364, 0.24884824480669612279, -0.531603085018726218, -0.01276095724756550441, -0.0062182029998522330566, 0.0038933865984447564058, 0.01591522341577054217, 0.024249902597008023142, 0.021601932461756991055, 0.0024385437332311472869, -0.032694760349284285317, -0.055672085270399122858, 0.031717987239585618275, 0.13030534660244441625, -0.17590378656702129745, -0.2367042953382553494, 0.30538579440195373627, 9.1681224945886697817e-05, 0.00028659676637208365097, -0.00047545393329486183376, -0.0015209238901272448957, -0.0011777868583985153569, -0.00099813582227404533224, -0.0085869527392002254623, -0.01700627683018168107, 0.020104183331079144431, 0.12622959300181699205, 0.066121276891701344369, -0.29178578631757512429, -0.10838509075308006002, 0.22401265348572538261, 0.0018542676724631484714, -0.0038474666179509663458, -0.0058992484978745525145, 0.0031030476989704104673, 0.030005202844101238474, 0.059842151448732193442, 0.037750537111206565466, -0.081545807952827356613, -0.23517670897834844213, -0.28677793580900229342, -0.14875887575021626286, 0.6365342531753505062, 1.5658465290457570518, -1.6173518745500716776, 0.00039209702285820534264, -0.0027233602990047412959, -0.0032610286929178329378, 0.0042617003282768217351, 0.025154596412868233946, 0.047604826731642728432, 0.027391334513142663598, -0.070367761599667913441, -0.19020593472416968961, -0.22824066754887326058, -0.14331348235018898096, 0.50290835712380943523, 1.3894628048057018344, -1.3968247951901964221, 0.0033153833616765123957, 0.0082008852255370072448, 0.010177780368121083082, 0.0053960006664756591382, -0.010670838955237516021, -0.036858995654492805738, -0.060913233339920769271, -0.060693937756346461043, -0.00021063338150388143805, 0.18063522717157884379, 0.29379998302762944107, -0.12328833206446228321, -0.34666727154104087782, 0.12536820739472490094, 0.0016450312564861629613, 0.006665570471365576781, 0.0095428482712733402049, 0.0068434238971905956422, -0.0061012208833908423283, -0.029730117866664847875, -0.054990226771675435047, -0.06310045177530788918, -0.015659056590227187628, 0.16438745861353379341, 0.3016185211723856896, -0.11039729781345862525, -0.36957436326533898985, 0.14851405996045560776, 0.0062791899689979063445, -0.002888371536131757733, -0.01307659122437935216, -0.021624312159376116171, -0.021032722932420928885, 0.00024547940343054970797, 0.049814017823744602409, 0.11805360269456093014, 0.13124460066150364645, -0.012450342326869458243, -0.23891237625812583589, -0.3598810731330707724, -0.023843148611093847172, 0.41316104603395942263, -0.0041356374245005815549, 0.013536712213205575814, 0.020461259943416777729, 0.0042999063221725647085, -0.046012278308279142003, -0.11638011490873470533, -0.1453804327816987374, -0.039810903863508460832, 0.26893979302499354533, 0.74195638825984289166, 1.2080138736719838732, 0.80982182559521498355, -8.5987094889110586848, 5.985880440794787205 };
	l1.insert(l1.end(), l2.begin(), l2.end());
	l1.insert(l1.end(), l3.begin(), l3.end());
	l1.insert(l1.end(), l4.begin(), l4.end());
	l1.insert(l1.end(), l5.begin(), l5.end());
	l1.insert(l1.end(), l6.begin(), l6.end());
	static const Matrix C_inv(308, 308, l1.data());
	constexpr double beta = 63.61571830736908594872147659771144390106201171875;
	constexpr double var = 1290.89965368065304573974572122097015380859375;
	constexpr double nugget = 0.020976231236367000321241249594095279462635517120361328125;
	constexpr double param[] = {
			8.8311817537380274956149150966666638851165771484375,
			0.1399999999999999023003738329862244427204132080078125,
			1.464528420733662184005652306950651109218597412109375,
			2.836869461034175809999169359798543155193328857421875,
			1.0802285238399818023680154510657303035259246826171875,
			0.27289868850857990079106230041361413896083831787109375};

	constexpr double y_min = 0.01;

	// compute mean and standard deviation of the Kriging model
	constexpr double scaling_factor = 2.2360679774997898051;   // sqrt(5)
	Vector k(308, 0.0);
	for (std::size_t i=0; i<308; i++)
	{
		// Matern 5/2 kernel
		double s = 0.0;
		for (std::size_t k=0; k<6; k++)
		{
			const double ecart = std::abs(xx(k) - yacht_inputs(i, k)) / (param[k] / scaling_factor);
			s += ecart - log(1.0 + ecart + ecart * ecart / 3.0);
		}
		k(i) = std::exp(-s) * var;
	}
	const double a = 1.0 - k * CF;
	const double b = var + nugget - k * (C_inv * k) + a * a / FCF;
	const double mean = beta + k * z;
	const double stddev = std::sqrt((std::max)(b, 0.0));

	// compute pdf and cdf
	const double xcr = (y_min - mean) / stddev;
	double density = std::exp(-0.5 * xcr * xcr) / std::sqrt(2.0 * M_PI);
	double probability = 0.5 * std::erfc((-xcr) * M_SQRT1_2);

	// compute objectives
	Vector f(2, 0.0);
	const double EI = (y_min - mean) * probability + stddev * density;   // expected improvement
	const double PI = probability;                                       // probability of improvement
	f(0) = (std::max)(0.0, 1.0 - 0.2 * EI);   // encoded for minimization, scaled to [0, 1]
	f(1) = 1.0 - PI;                        // encoded for minimization, scaled to [0, 1]
	return f;
}

Vector vibrating_platform_design(Vector const& x)
{
	// the order of the materials is fixed
	static constexpr double rho1 = 100;
	static constexpr double E1 = 1.6e9;
	static constexpr double c1 = 500;
	static constexpr double rho2 = 2770;
	static constexpr double E2 = 7e10;
	static constexpr double c2 = 1500;
	static constexpr double rho3 = 7780;
	static constexpr double E3 = 2e11;
	static constexpr double c3 = 800;

	// try to handle constraints through encoding
	double d1 = 0.05 + 0.45 * x[0];
	double d2min = std::max<double>(d1, 0.2);
	double d2max = std::min<double>(d1 + 0.15, 0.5);
	double d2 = d2min + (d2max - d2min) * x[1];
	double d3min = std::max<double>(d2, 0.2);
	double d3max = std::min<double>(d2 + 0.01, 0.6);   // the 0.6 part can never kick in
	double d3 = d3min + (d3max - d3min) * x[2];
	double d1q = d1*d1*d1;
	double d2q = d2*d2*d2;
	double d3q = d3*d3*d3;
	double b = 0.35 + 0.15 * x[3];
	double EI = (2.0 * b / 3.0) * (E1 * d1q + E2 * (d2q - d1q) + E3 * (d3q - d2q));
	double mu = 2.0 * b * (rho1 * d1 + rho2 * (d2 - d1) + rho3 * (d3 - d2));
	double Lmax = std::min<double>(2800.0 / mu, 6.0);
	double L = 3.0 + (Lmax - 3.0) * x[4];

	// handle constraint violation
	double violation = (std::max)(0.0, 3.0 - L);
	if (violation > 0)
	{
		return Vector(2, 0.95 + 0.05 * std::tanh(violation));
	}

	// compute objectives
	double negFreq = -(M_PI / (2.0 * L * L)) * std::sqrt(EI / mu);   // range -420 ... -20 (estimated)
	double cost = 2.0 * b * L * (c1 * d1 + c2 * (d2 - d1) + c3 * (d3 - d2));   // range 210 ... 2448 (exact)

	// shift and scale objectives
	Vector f(2, 0.0);
	if (negFreq < -400.0) f(0) = 0.05 + 0.05 * std::tanh((negFreq + 400.0) / 20.0);
	else if (negFreq > -100.0) f(0) = 0.85 + 0.05 * std::tanh((negFreq + 100.0) / 20.0);
	else f(0) = 0.05 + 0.8 * (negFreq + 400.0) / 300.0;
	f(1) = (cost - 210.0) / (2448.0 - 210.0);
	assert(f(0) >= 0.0);
	assert(f(0) <= 1.0);
	assert(f(1) >= 0.0);
	assert(f(1) <= 1.0);
	return f;
}
